/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

#define SWIG_PYTHON_2_UNICODE



#define SWIG_VERSION 0x040401
#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE
#define SWIGPYTHON_FASTPROXY

#define SWIG_name    "_openbabel"
/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if !defined(PY_SSIZE_T_CLEAN) && !defined(SWIG_NO_PY_SSIZE_T_CLEAN)
#define PY_SSIZE_T_CLEAN
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus) && __cplusplus >=201703L
#pragma GCC diagnostic ignored "-Wregister" /* For python-2.7 headers that use register */
#endif
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */

#if defined(_MSC_VER) && _MSC_VER >= 1929
/* Workaround compilation errors when redefining _DEBUG in MSVC 2019 version 16.10 and later
 * See https://github.com/swig/swig/issues/2090 */
# include <corecrt.h>
#endif

# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif


#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

#include <stdio.h>
#include <stdlib.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if PY_VERSION_HEX >= 0x03030000 && !defined(SWIG_NO_HEAPTYPES)
#if !defined(SWIG_HEAPTYPES)
#define SWIG_HEAPTYPES
#endif
#endif

/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif

#if defined(SWIG_HEAPTYPES)
#if PY_VERSION_HEX < 0x030c0000
#include <structmember.h>
#define Py_READONLY READONLY
#define Py_T_PYSSIZET T_PYSSIZET
#endif
#endif

#include <stddef.h> /* For offsetof */


/* Wrapper around PyUnicode_AsUTF8AndSize - call Py_XDECREF on the returned pbytes when finished with the returned string */
SWIGINTERN const char *
SWIG_PyUnicode_AsUTF8AndSize(PyObject *str, Py_ssize_t *psize, PyObject **pbytes)
{
#if PY_VERSION_HEX >= 0x03030000
# if !defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030A0000
  *pbytes = NULL;
  return PyUnicode_AsUTF8AndSize(str, psize);
# else
  const char *chars;
  *pbytes = PyUnicode_AsUTF8String(str);
  chars = *pbytes ? PyBytes_AsString(*pbytes) : NULL;
  if (chars && psize)
    *psize = PyBytes_Size(*pbytes);
  return chars;
# endif
#else
  char *chars = NULL;
  *pbytes = NULL;
  PyString_AsStringAndSize(str, &chars, psize);
  return chars;
#endif
}

SWIGINTERN PyObject *
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#define SWIG_RUNTIME_MODULE "swig_runtime_data" SWIG_RUNTIME_VERSION

/* SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user interface files check for it. */
# define SWIGPY_USE_CAPSULE
#ifdef SWIGPYTHON_BUILTIN
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule_builtin" SWIG_TYPE_TABLE_NAME
#else
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule" SWIG_TYPE_TABLE_NAME
#endif
#define SWIGPY_CAPSULE_NAME SWIG_RUNTIME_MODULE "." SWIGPY_CAPSULE_ATTR_NAME

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

#if defined(Py_LIMITED_API)
# define PyTuple_GET_ITEM PyTuple_GetItem
/* Note that PyTuple_SetItem() has different semantics from PyTuple_SET_ITEM as it decref's the original tuple item, so in general they cannot be used
  interchangeably. However in SWIG-generated code PyTuple_SET_ITEM is only used with newly initialized tuples without any items and for them this does work. */
# define PyTuple_SET_ITEM PyTuple_SetItem
# define PyTuple_GET_SIZE PyTuple_Size
# define PyCFunction_GET_FLAGS PyCFunction_GetFlags
# define PyCFunction_GET_FUNCTION PyCFunction_GetFunction
# define PyCFunction_GET_SELF PyCFunction_GetSelf
# define PyList_GET_ITEM PyList_GetItem
# define PyList_SET_ITEM PyList_SetItem
# define PySliceObject PyObject
#endif

/* Increment and Decrement wrappers - for portability when using the stable abi and for performance otherwise */
#ifdef Py_LIMITED_API
# define SWIG_Py_INCREF Py_IncRef
# define SWIG_Py_XINCREF Py_IncRef
# define SWIG_Py_DECREF Py_DecRef
# define SWIG_Py_XDECREF Py_DecRef
#else
# define SWIG_Py_INCREF Py_INCREF
# define SWIG_Py_XINCREF Py_XINCREF
# define SWIG_Py_DECREF Py_DECREF
# define SWIG_Py_XDECREF Py_XDECREF
#endif

#if PY_VERSION_HEX >= 0x03000000
#if (PY_VERSION_HEX >= 0x030d00a6) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a6)
# define SWIG_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
SWIGINTERN PyObject *
SWIG_PyType_GetFullyQualifiedName(PyTypeObject *type) {
  PyObject *result = NULL;
  PyObject *qualname = PyObject_GetAttrString((PyObject *)type, "__qualname__");
  if (qualname) {
    PyObject *mod = PyObject_GetAttrString((PyObject *)type, "__module__");
    if (mod) {
      if (PyUnicode_Check(mod) && PyUnicode_CompareWithASCIIString(mod, "builtins") && PyUnicode_CompareWithASCIIString(mod, "__main__")) {
        result = PyUnicode_FromFormat("%U%c%U", mod, '.', qualname);
        SWIG_Py_DECREF(qualname);
      } else {
        result = qualname;
      }
      SWIG_Py_DECREF(mod);
    } else {
      result = qualname;
    }
  }

  return result;
}
#endif
#endif

/* gh-114329 added PyList_GetItemRef() to Python 3.13.0a4 */
#if (PY_VERSION_HEX >= 0x030d00a4) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a4)
# define SWIG_PyList_GetItemRef PyList_GetItemRef
#else
SWIGINTERN PyObject *
SWIG_PyList_GetItemRef(PyObject *op, Py_ssize_t index) {
  PyObject *item = PyList_GetItem(op, index);
  Py_XINCREF(item);
  return item;
}
#endif

/* gh-106004 added PyDict_GetItemRef() and PyDict_GetItemStringRef() to Python 3.13.0a1
   functions are renamed here for compatibility with abi3audit */
#if (PY_VERSION_HEX >= 0x030d00a1) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a1)
# define SWIG_PyDict_GetItemRef PyDict_GetItemRef
# define SWIG_PyDict_GetItemStringRef PyDict_GetItemStringRef
#else
SWIGINTERN int
SWIG_PyDict_GetItemRef(PyObject *mp, PyObject *key, PyObject **result) {
#if PY_VERSION_HEX >= 0x03000000
  PyObject *item = PyDict_GetItemWithError(mp, key);
#else
  PyObject *item = _PyDict_GetItemWithError(mp, key);
#endif
  if (item != NULL) {
    *result = (PyObject *)(item);
    SWIG_Py_INCREF(*result);
    return 1;
  }
  if (!PyErr_Occurred()) {
    *result = NULL;
    return 0;
  }
  *result = NULL;
  return -1;
}

SWIGINTERN int
SWIG_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result) {
  int res;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *key_obj = PyUnicode_FromString(key);
#else
  PyObject *key_obj = PyString_FromString(key);
#endif
  if (key_obj == NULL) {
    *result = NULL;
    return -1;
  }
  res = SWIG_PyDict_GetItemRef(mp, key_obj, result);
  Py_DECREF(key_obj);
  return res;
}
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  case SWIG_NullReferenceError:
    type = PyExc_TypeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyObject *bytes = NULL;
    const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
    PyErr_Clear();
    SWIG_Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_DECREF(old_str);
    SWIG_Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    if (newvalue) {
      SWIG_Py_XDECREF(value);
      PyErr_Restore(type, newvalue, traceback);
    } else {
      PyErr_Restore(type, value, traceback);
    }
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      if PY_VERSION_HEX < 0x03070000
#        define SWIG_PYTHON_INITIALIZE_THREADS PyEval_InitThreads()
#      else
#        define SWIG_PYTHON_INITIALIZE_THREADS
#      endif
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { status = false; PyEval_RestoreThread(save); }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03050000
# error "This version of SWIG only supports Python 3 >= 3.5"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   

/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  SWIG_Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  SWIG_Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
}

#endif

/* SWIG runtime data Python module */
static PyObject *Swig_runtime_data_module_global = NULL;

/* Create/obtain the single swig_runtime_data module which is used across different SWIG generated modules */
SWIGINTERN PyObject *
SWIG_runtime_data_module(void) {
  if (!Swig_runtime_data_module_global) {
#if (PY_VERSION_HEX >= 0x030d0000) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d0000)
    /* free-threading note: the GIL is always enabled when this function is first called
       by SWIG_init, so there's no risk of race conditions */
    Swig_runtime_data_module_global = PyImport_AddModuleRef(SWIG_RUNTIME_MODULE);
#elif PY_VERSION_HEX >= 0x03000000
    Swig_runtime_data_module_global = PyImport_AddModule(SWIG_RUNTIME_MODULE);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#else
    static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
    Swig_runtime_data_module_global = Py_InitModule(SWIG_RUNTIME_MODULE, swig_empty_runtime_method_table);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#endif
  }
  assert(Swig_runtime_data_module_global);
  return Swig_runtime_data_module_global;
}

/* Append a value to the result obj */
SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj, int is_void) {
  if (!result) {
    result = obj;
  } else if (result == Py_None && is_void) {
    SWIG_Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      if (result) {
        PyList_SET_ITEM(result, 0, o2);
      } else {
        SWIG_Py_DECREF(obj);
        return o2;
      }
    }
    PyList_Append(result,obj);
    SWIG_Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
 
/* -----------------------------------------------------------------------------
 * global variable support code.
 * ----------------------------------------------------------------------------- */
 
typedef struct swig_globalvar {   
  char       *name;                  /* Name of global variable */
  PyObject *(*get_attr)(void);       /* Return the current value */
  int       (*set_attr)(PyObject *); /* Set the value */
  struct swig_globalvar *next;
} swig_globalvar;

typedef struct swig_varlinkobject {
  PyObject_HEAD
  swig_globalvar *vars;
} swig_varlinkobject;

SWIGINTERN PyObject *
SwigVarLink_repr(PyObject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_InternFromString("<Swig global variables>");
#else
  return PyString_FromString("<Swig global variables>");
#endif
}

SWIGINTERN PyObject *
SwigVarLink_str(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *str = PyUnicode_InternFromString("(");
  PyObject *tail;
  PyObject *joined;
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    tail = PyUnicode_FromString(var->name);
    joined = PyUnicode_Concat(str, tail);
    SWIG_Py_DECREF(str);
    SWIG_Py_DECREF(tail);
    str = joined;
    if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        SWIG_Py_DECREF(str);
        SWIG_Py_DECREF(tail);
        str = joined;
    }
  }
  tail = PyUnicode_InternFromString(")");
  joined = PyUnicode_Concat(str, tail);
  SWIG_Py_DECREF(str);
  SWIG_Py_DECREF(tail);
  str = joined;
#else
  PyObject *str = PyString_FromString("(");
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    PyString_ConcatAndDel(&str,PyString_FromString(var->name));
    if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
  }
  PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
  return str;
}

SWIGINTERN void
SwigVarLink_dealloc(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  swig_globalvar *var = v->vars;
  while (var) {
    swig_globalvar *n = var->next;
    free(var->name);
    free(var);
    var = n;
  }
}

SWIGINTERN PyObject *
SwigVarLink_getattr(PyObject *o, char *n) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  PyObject *res = NULL;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->get_attr)();
      break;
    }
    var = var->next;
  }
  if (res == NULL && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN int
SwigVarLink_setattr(PyObject *o, char *n, PyObject *p) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  int res = 1;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->set_attr)(p);
      break;
    }
    var = var->next;
  }
  if (res == 1 && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN PyTypeObject*
SwigVarLink_TypeOnce(void) {
  static char SwigVarLink_doc[] = "Swig variable link object";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject varlink_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                  /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigVarLink", /* tp_name */
      sizeof(swig_varlinkobject),         /* tp_basicsize */
      0,                                  /* tp_itemsize */
      (destructor) SwigVarLink_dealloc,   /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                       /* tp_print */
#else
      (Py_ssize_t)0,                      /* tp_vectorcall_offset */
#endif
      (getattrfunc) SwigVarLink_getattr,  /* tp_getattr */
      (setattrfunc) SwigVarLink_setattr,  /* tp_setattr */
      0,                                  /* tp_compare */
      (reprfunc) SwigVarLink_repr,        /* tp_repr */
      0,                                  /* tp_as_number */
      0,                                  /* tp_as_sequence */
      0,                                  /* tp_as_mapping */
      0,                                  /* tp_hash */
      0,                                  /* tp_call */
      (reprfunc) SwigVarLink_str,         /* tp_str */
      0,                                  /* tp_getattro */
      0,                                  /* tp_setattro */
      0,                                  /* tp_as_buffer */
      0,                                  /* tp_flags */
      SwigVarLink_doc,                    /* tp_doc */
      0,                                  /* tp_traverse */
      0,                                  /* tp_clear */
      0,                                  /* tp_richcompare */
      0,                                  /* tp_weaklistoffset */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
      0,                                  /* tp_del */
      0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                  /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                  /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                  /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                  /* tp_allocs */
      0,                                  /* tp_frees */
      0,                                  /* tp_maxalloc */
      0,                                  /* tp_prev */
      0                                   /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    varlink_type = tmp;
    type_init = 1;
    if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigVarLink", (PyObject *)&varlink_type) == 0)
      SWIG_Py_INCREF((PyObject *)&varlink_type);
  }
  return &varlink_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigVarLink_dealloc },
    { Py_tp_repr, (void *)SwigVarLink_repr },
    { Py_tp_getattr, (void *)SwigVarLink_getattr },
    { Py_tp_setattr, (void *)SwigVarLink_setattr },
    { Py_tp_str, (void *)SwigVarLink_str },
    { Py_tp_doc, (void *)SwigVarLink_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigVarLink",
    sizeof(swig_varlinkobject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigVarLink", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyTypeObject*
SwigVarLink_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigVarLink_TypeOnce();
  return type;
}

/* Create a variable linking object for use later */
SWIGINTERN PyObject *
SWIG_Python_newvarlink(void) {
  swig_varlinkobject *result = PyObject_New(swig_varlinkobject, SwigVarLink_Type());
  if (result) {
    result->vars = 0;
  }
  return ((PyObject*) result);
}

SWIGINTERN void 
SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
  swig_varlinkobject *v = (swig_varlinkobject *) p;
  swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
  if (gv) {
    size_t size = strlen(name)+1;
    gv->name = (char *)malloc(size);
    if (gv->name) {
      memcpy(gv->name, name, size);
      gv->get_attr = get_attr;
      gv->set_attr = set_attr;
      gv->next = v->vars;
    }
  }
  v->vars = gv;
}


static PyObject *Swig_Globals_global = NULL;
  
SWIGINTERN PyObject *
SWIG_globals(void) {
  if (Swig_Globals_global == NULL) {
    Swig_Globals_global = SWIG_newvarlink();
  }
  return Swig_Globals_global;
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  SWIG_Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    SWIG_Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      SWIG_Py_INCREF(obj);
      data->newargs = obj;
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
        data->newargs = PyTuple_New(1);
        if (data->newargs) {
          SWIG_Py_INCREF(obj);
          PyTuple_SET_ITEM(data->newargs, 0, obj);
        } else {
          SWIG_Py_DECREF(data->newraw);
          SWIG_Py_DECREF(data->klass);
          free(data);
          return 0;
        }
      } else {
        SWIG_Py_INCREF(obj);
        data->newargs = obj;
      }
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      data->delargs = !(PyCFunction_GET_FLAGS(data->destroy) & METH_O);
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data)
{
  SWIG_Py_XDECREF(data->klass);
  SWIG_Py_XDECREF(data->newraw);
  SWIG_Py_XDECREF(data->newargs);
  SWIG_Py_XDECREF(data->destroy);
  free(data);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
  PyObject *swigdict;
  PyObject *weakreflist;
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->swigdict)
    sobj->swigdict = PyDict_New();

  SWIG_Py_XINCREF(sobj->swigdict);
  return sobj->swigdict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    PyObject *val = SwigPyObject_long(v);
    if (val) {
      PyObject *ofmt;
      PyTuple_SET_ITEM(args, 0, val);
      ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
        res = PyUnicode_Format(ofmt,args);
#else
        res = PyString_Format(ofmt,args);
#endif
        SWIG_Py_DECREF(ofmt);
      }
    }
    SWIG_Py_DECREF(args);
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (repr && v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
    if (nrep) {
# if PY_VERSION_HEX >= 0x03000000
      PyObject *joined = PyUnicode_Concat(repr, nrep);
      SWIG_Py_DECREF(repr);
      SWIG_Py_DECREF(nrep);
      repr = joined;
# else
      PyString_ConcatAndDel(&repr,nrep);
# endif
    } else {
      SWIG_Py_DECREF(repr);
      repr = NULL;
    }
  }
  return repr;
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in SwigPyObject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(PyObject *v, PyObject *w)
{
  /* tp_compare is only called when both objects have the same type, so
   * the casts are guaranteed to be ok. */
  void *i = ((SwigPyObject *)v)->ptr;
  void *j = ((SwigPyObject *)w)->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

SWIGRUNTIMEINLINE int SwigPyObject_Check(PyObject *);

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(PyObject *v, PyObject *w, int op)
{
  PyObject* res = NULL;
  if (!PyErr_Occurred()) {
    /* Per https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_richcompare
     * the first argument is guaranteed to be an instance of SwigPyObject, but the
     * second is not, so we typecheck that one. */
    if ((op != Py_EQ && op != Py_NE) || !SwigPyObject_Check(w)) {
      SWIG_Py_INCREF(Py_NotImplemented);
      return Py_NotImplemented;
    }
    res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  }
  return res;
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyObject_Type();
  PyTypeObject *op_type = Py_TYPE(op);
#ifdef SWIGPYTHON_BUILTIN
  /* Only builtin types have SwigPyObject as a base type */
  return PyType_IsSubtype(op_type, target_tp);
#else
  /* Check for an exact match to SwigPyObject */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyObject");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyObject") == 0;
#endif
  }
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

static PyObject* Swig_Capsule_global = NULL;

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        if (tmp) {
          res = SWIG_Python_CallFunctor(destroy, tmp);
        } else {
          res = 0;
        }
        SWIG_Py_XDECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      SWIG_Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
    SWIG_Py_XDECREF(Swig_Capsule_global);
  }
  SWIG_Py_XDECREF(next);
  SWIG_Py_XDECREF(sobj->swigdict);
  PyObject_Free(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  ((SwigPyObject *)next)->next = sobj->next;
  sobj->next = next;
  SWIG_Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    SWIG_Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SWIG_Py_DECREF(SwigPyObject_acquire(v,args));
      } else {
        SWIG_Py_DECREF(SwigPyObject_disown(v,args));
      }
    } 
    return obj;
  }
}

static PyMethodDef
SwigPyObject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char SwigPyObject_doc[] = "Swig object holding a C/C++ pointer";
#ifndef SWIG_HEAPTYPES
  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyObject",  /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyObject_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      offsetof(SwigPyObject, weakreflist),  /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      SwigPyObject_methods,                 /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyObject", (PyObject *)&swigpyobject_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpyobject_type);
  }
  return &swigpyobject_type;
#else
  static PyMemberDef SwigPyObject_members[] = {
      { (char *)"__dictoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, swigdict), Py_READONLY, NULL },
      { (char *)"__weaklistoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, weakreflist), Py_READONLY, NULL },
      { NULL, 0, 0, 0, NULL }
  };
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyObject_dealloc },
    { Py_tp_repr, (void *)SwigPyObject_repr },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyObject_doc },
    { Py_tp_richcompare, (void *)SwigPyObject_richcompare },
    { Py_tp_methods, (void *)SwigPyObject_methods },
    { Py_nb_int, (void *)SwigPyObject_long },
    { Py_tp_members, (void *)SwigPyObject_members },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyObject",
    sizeof(SwigPyObject),
    0,
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
#if !defined(Py_LIMITED_API)
/* While this __dictoffset__ is only used with the builtin wrappers, SwigPyObject ought to be
   identical when created for use by proxy class wrappers in case it is shared across multiple modules. */
#if PY_VERSION_HEX < 0x03090000
    /* Workaround as __dictoffset__ and __weaklistoffset__ above are only supported from python-3.9 */
  if (pytype) {
    ((PyTypeObject *)pytype)->tp_dictoffset = offsetof(SwigPyObject, swigdict);
    ((PyTypeObject *)pytype)->tp_weaklistoffset = offsetof(SwigPyObject, weakreflist);
  }
#endif
#endif
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyObject", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_New(SwigPyObject, SwigPyObject_Type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
    sobj->swigdict = 0;
    sobj->weakreflist = 0;
    if (own == SWIG_POINTER_OWN) {
      /* Obtain a reference to the Python capsule wrapping the module information, so that the
       * module information is correctly destroyed after all SWIG python objects have been freed
       * by the GC (and corresponding destructors invoked) */
      SWIG_Py_XINCREF(Swig_Capsule_global);
    }
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyPacked_Type();
  PyTypeObject *op_type = Py_TYPE(op);
  /* Check for an exact match to SwigPyPacked */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyPacked");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyPacked") == 0;
#endif
  }
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_Free(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char SwigPyPacked_doc[] = "Swig object holding a C/C++ function pointer";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyPacked",  /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyPacked_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyPacked", (PyObject *)&swigpypacked_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpypacked_type);
  }
  return &swigpypacked_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyPacked_dealloc },
    { Py_tp_repr, (void *)SwigPyPacked_repr },
    { Py_tp_str, (void *)SwigPyPacked_str },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyPacked_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyPacked",
    sizeof(SwigPyPacked),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyPacked", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_New(SwigPyPacked, SwigPyPacked_Type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_Free((PyObject *)sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX >= 0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

/* Returns a borrowed reference to the 'this' object */
SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
  if (PyWeakref_CheckProxy(pyobj)) {
#if PY_VERSION_HEX >= 0x030d0000
    if (PyWeakref_GetRef(pyobj, &pyobj) > 0)
      Py_DECREF(pyobj);
    else
      pyobj = NULL;
#else
    pyobj = PyWeakref_GetObject(pyobj);
#endif
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
      if (PyWeakref_CheckProxy(pyobj)) {
        PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
      obj = PyObject_GetAttr(pyobj, SWIG_This());
      if (obj) {
	SWIG_Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj, SWIG_This());
  if (obj) {
    SWIG_Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !sobj->own) {
      res = SWIG_ERROR_RELEASE_NOT_OWNED;
    } else {
      if (own)
        *own = *own | sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
        sobj->own = 0;
      }
      if (flags & SWIG_POINTER_CLEAR) {
        sobj->ptr = 0;
      }
      res = SWIG_OK;
    }
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            SWIG_Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
#ifndef Py_LIMITED_API
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
#else
    PyObject* pystr_doc = PyObject_GetAttrString(obj, "__doc__");
    PyObject *bytes = NULL;
    const char *doc = pystr_doc ? SWIG_PyUnicode_AsUTF8AndSize(pystr_doc, NULL, &bytes) : 0;
#endif
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
#ifdef Py_LIMITED_API
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_XDECREF(pystr_doc);
#endif
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
        PyObject *dict = *dictptr;
        if (dict == NULL) {
          dict = PyDict_New();
          *dictptr = dict;
        }
        if (dict) {
          PyDict_SetItem(dict, SWIG_This(), swig_this);
        } else{
          SWIG_Py_DECREF(inst);
          inst = 0;
        }
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        SWIG_Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
#ifndef Py_LIMITED_API
        newfunc newfn = ((PyTypeObject *)data->newargs)->tp_new;
#else
        newfunc newfn = (newfunc)PyType_GetSlot((PyTypeObject *)data->newargs, Py_tp_new);
#endif
        inst = newfn((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        SWIG_Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            SWIG_Py_DECREF(inst);
            inst = 0;
          } else {
            PyType_Modified(Py_TYPE(inst));
          }
        }
      }
      SWIG_Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      SWIG_Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    if (dict) {
      return PyDict_SetItem(dict, SWIG_This(), swig_this);
    } else{
      return -1;
    }
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SWIG_Py_DECREF(SwigPyObject_append((PyObject*) sthis, obj[1]));
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
#ifndef Py_LIMITED_API
        allocfunc alloc = clientdata->pytype->tp_alloc;
#else
        allocfunc alloc = (allocfunc)PyType_GetSlot(clientdata->pytype, Py_tp_alloc);
#endif
        PyObject *next_self = alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
      if (newobj) {
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    SWIG_Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

static PyObject *Swig_TypeCache_global = NULL;

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  if (Swig_TypeCache_global == NULL) {
    Swig_TypeCache_global = PyDict_New();
  }
  return Swig_TypeCache_global;
}

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
#ifdef SWIG_LINK_RUNTIME
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
  }
#else
  void *type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
  if (PyErr_Occurred()) {
    PyErr_Clear();
    type_pointer = (void *)0;
  }
#endif
  return (swig_module_info *) type_pointer;
}


#if defined(SWIG_REFCNT_DEBUG)
#define SWIG_PYOBJ_REFCNT(OBJ) fprintf(stdout, "" #OBJ " count %ld\n", (OBJ ? Py_REFCNT(OBJ) : 0))
#else
#define SWIG_PYOBJ_REFCNT(OBJ)
#endif

static int interpreter_counter = 0; /* how many (sub-)interpreters are using swig_module's types */

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  if (--interpreter_counter != 0) /* another sub-interpreter may still be using the swig_module's types */
    return;
  for (i = 0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      ty->clientdata = 0;
      if (data) SwigPyClientData_Del(data);
    }
  }
  SWIG_PYOBJ_REFCNT(Swig_This_global);
  SWIG_Py_XDECREF(Swig_This_global);
  Swig_This_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Globals_global);
  SWIG_Py_XDECREF(Swig_Globals_global);
  Swig_Globals_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_TypeCache_global);
  SWIG_Py_XDECREF(Swig_TypeCache_global);
  Swig_TypeCache_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Capsule_global);
  Swig_Capsule_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_runtime_data_module_global);
  SWIG_Py_XDECREF(Swig_runtime_data_module_global);
  Swig_runtime_data_module_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && runtime_data_module) {
    if (PyModule_AddObject(runtime_data_module, SWIGPY_CAPSULE_ATTR_NAME, pointer) == 0) {
      ++interpreter_counter;
      Swig_Capsule_global = pointer;
    } else {
      SWIG_Py_DECREF(pointer);
    }
  } else {
    SWIG_Py_XDECREF(pointer);
  }
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  swig_type_info *descriptor;
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *obj;
  SWIG_PyDict_GetItemStringRef(cache, type, &obj);
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      if (obj) PyDict_SetItemString(cache, type, obj);
    }
  }
  SWIG_Py_XDECREF(obj);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      PyObject *bytes = NULL;
      const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      SWIG_Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Py_XDECREF(bytes);
      SWIG_Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = Py_TYPE(obj);
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(Py_TYPE(name));
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%S'", tpname);
    SWIG_Py_DECREF(tpname);
#else
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%s'", Py_TYPE(name)->tp_name);
#endif
    return -1;
  } else {
    SWIG_Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) != 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = Py_TYPE(descr)->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      SWIG_Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        goto done;
    }
#if PY_VERSION_HEX >= 0x03000000
    {
      PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(tp);
      PyErr_Format(PyExc_AttributeError, "'%S' object has no attribute '%s'", tpname, PyString_AsString(encoded_name));
      SWIG_Py_DECREF(tpname);
    }
#else
    PyErr_Format(PyExc_AttributeError, "'%s' object has no attribute '%s'", tp->tp_name, PyString_AsString(encoded_name));
#endif
    SWIG_Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  SWIG_Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



#ifdef __cplusplus
extern "C" {
#endif

/* Method creation and docstring support functions */

SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name);
SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);
SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);

#ifdef __cplusplus
}
#endif


  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_ConstRefIter swig_types[0]
#define SWIGTYPE_p_Iter swig_types[1]
#define SWIGTYPE_p_Mapping swig_types[2]
#define SWIGTYPE_p_Mappings swig_types[3]
#define SWIGTYPE_p_OpMap swig_types[4]
#define SWIGTYPE_p_OpenBabel__AliasData swig_types[5]
#define SWIGTYPE_p_OpenBabel__CharPtrLess swig_types[6]
#define SWIGTYPE_p_OpenBabel__FastSearch swig_types[7]
#define SWIGTYPE_p_OpenBabel__FastSearchIndexer swig_types[8]
#define SWIGTYPE_p_OpenBabel__FptIndex swig_types[9]
#define SWIGTYPE_p_OpenBabel__FptIndexHeader swig_types[10]
#define SWIGTYPE_p_OpenBabel__LineSearchType swig_types[11]
#define SWIGTYPE_p_OpenBabel__OBAlign swig_types[12]
#define SWIGTYPE_p_OpenBabel__OBAngle swig_types[13]
#define SWIGTYPE_p_OpenBabel__OBAngleData swig_types[14]
#define SWIGTYPE_p_OpenBabel__OBAromaticTyper swig_types[15]
#define SWIGTYPE_p_OpenBabel__OBAtom swig_types[16]
#define SWIGTYPE_p_OpenBabel__OBAtomAtomIter swig_types[17]
#define SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor swig_types[18]
#define SWIGTYPE_p_OpenBabel__OBAtomBondIter swig_types[19]
#define SWIGTYPE_p_OpenBabel__OBAtomHOF swig_types[20]
#define SWIGTYPE_p_OpenBabel__OBAtomTyper swig_types[21]
#define SWIGTYPE_p_OpenBabel__OBAtomicHeatOfFormationTable swig_types[22]
#define SWIGTYPE_p_OpenBabel__OBBase swig_types[23]
#define SWIGTYPE_p_OpenBabel__OBBitVec swig_types[24]
#define SWIGTYPE_p_OpenBabel__OBBond swig_types[25]
#define SWIGTYPE_p_OpenBabel__OBBuilder swig_types[26]
#define SWIGTYPE_p_OpenBabel__OBChainsParser swig_types[27]
#define SWIGTYPE_p_OpenBabel__OBChargeModel swig_types[28]
#define SWIGTYPE_p_OpenBabel__OBChemTsfm swig_types[29]
#define SWIGTYPE_p_OpenBabel__OBCisTransConfig swig_types[30]
#define SWIGTYPE_p_OpenBabel__OBCisTransStereo swig_types[31]
#define SWIGTYPE_p_OpenBabel__OBCommentData swig_types[32]
#define SWIGTYPE_p_OpenBabel__OBConformerData swig_types[33]
#define SWIGTYPE_p_OpenBabel__OBConformerFilter swig_types[34]
#define SWIGTYPE_p_OpenBabel__OBConformerFilters swig_types[35]
#define SWIGTYPE_p_OpenBabel__OBConformerScore swig_types[36]
#define SWIGTYPE_p_OpenBabel__OBConformerSearch swig_types[37]
#define SWIGTYPE_p_OpenBabel__OBConversion swig_types[38]
#define SWIGTYPE_p_OpenBabel__OBDOSData swig_types[39]
#define SWIGTYPE_p_OpenBabel__OBDescriptor swig_types[40]
#define SWIGTYPE_p_OpenBabel__OBElectronicTransitionData swig_types[41]
#define SWIGTYPE_p_OpenBabel__OBEnergyConformerScore swig_types[42]
#define SWIGTYPE_p_OpenBabel__OBError swig_types[43]
#define SWIGTYPE_p_OpenBabel__OBExternalBond swig_types[44]
#define SWIGTYPE_p_OpenBabel__OBExternalBondData swig_types[45]
#define SWIGTYPE_p_OpenBabel__OBFFCalculation2 swig_types[46]
#define SWIGTYPE_p_OpenBabel__OBFFCalculation3 swig_types[47]
#define SWIGTYPE_p_OpenBabel__OBFFCalculation4 swig_types[48]
#define SWIGTYPE_p_OpenBabel__OBFFConstraint swig_types[49]
#define SWIGTYPE_p_OpenBabel__OBFFConstraints swig_types[50]
#define SWIGTYPE_p_OpenBabel__OBFFParameter swig_types[51]
#define SWIGTYPE_p_OpenBabel__OBFingerprint swig_types[52]
#define SWIGTYPE_p_OpenBabel__OBForceField swig_types[53]
#define SWIGTYPE_p_OpenBabel__OBFormat swig_types[54]
#define SWIGTYPE_p_OpenBabel__OBFreeGrid swig_types[55]
#define SWIGTYPE_p_OpenBabel__OBFreeGridPoint swig_types[56]
#define SWIGTYPE_p_OpenBabel__OBGenericData swig_types[57]
#define SWIGTYPE_p_OpenBabel__OBGlobalDataBase swig_types[58]
#define SWIGTYPE_p_OpenBabel__OBGraphSym swig_types[59]
#define SWIGTYPE_p_OpenBabel__OBGridData swig_types[60]
#define SWIGTYPE_p_OpenBabel__OBInternalCoord swig_types[61]
#define SWIGTYPE_p_OpenBabel__OBIsomorphismMapper swig_types[62]
#define SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor swig_types[63]
#define SWIGTYPE_p_OpenBabel__OBMatrixData swig_types[64]
#define SWIGTYPE_p_OpenBabel__OBMessageHandler swig_types[65]
#define SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore swig_types[66]
#define SWIGTYPE_p_OpenBabel__OBMinimizingRMSDConformerScore swig_types[67]
#define SWIGTYPE_p_OpenBabel__OBMol swig_types[68]
#define SWIGTYPE_p_OpenBabel__OBMolAngleIter swig_types[69]
#define SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter swig_types[70]
#define SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter swig_types[71]
#define SWIGTYPE_p_OpenBabel__OBMolAtomIter swig_types[72]
#define SWIGTYPE_p_OpenBabel__OBMolBondBFSIter swig_types[73]
#define SWIGTYPE_p_OpenBabel__OBMolBondIter swig_types[74]
#define SWIGTYPE_p_OpenBabel__OBMolPairIter swig_types[75]
#define SWIGTYPE_p_OpenBabel__OBMolRingIter swig_types[76]
#define SWIGTYPE_p_OpenBabel__OBMolTorsionIter swig_types[77]
#define SWIGTYPE_p_OpenBabel__OBNasaThermoData swig_types[78]
#define SWIGTYPE_p_OpenBabel__OBOp swig_types[79]
#define SWIGTYPE_p_OpenBabel__OBOrbital swig_types[80]
#define SWIGTYPE_p_OpenBabel__OBOrbitalData swig_types[81]
#define SWIGTYPE_p_OpenBabel__OBPairData swig_types[82]
#define SWIGTYPE_p_OpenBabel__OBPairTemplateT_bool_t swig_types[83]
#define SWIGTYPE_p_OpenBabel__OBPairTemplateT_double_t swig_types[84]
#define SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t swig_types[85]
#define SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t swig_types[86]
#define SWIGTYPE_p_OpenBabel__OBPcharge swig_types[87]
#define SWIGTYPE_p_OpenBabel__OBPhModel swig_types[88]
#define SWIGTYPE_p_OpenBabel__OBPlugin swig_types[89]
#define SWIGTYPE_p_OpenBabel__OBQuery swig_types[90]
#define SWIGTYPE_p_OpenBabel__OBQueryAtom swig_types[91]
#define SWIGTYPE_p_OpenBabel__OBQueryBond swig_types[92]
#define SWIGTYPE_p_OpenBabel__OBRMSDConformerScore swig_types[93]
#define SWIGTYPE_p_OpenBabel__OBRTree swig_types[94]
#define SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t swig_types[95]
#define SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t swig_types[96]
#define SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t swig_types[97]
#define SWIGTYPE_p_OpenBabel__OBRateData swig_types[98]
#define SWIGTYPE_p_OpenBabel__OBReaction swig_types[99]
#define SWIGTYPE_p_OpenBabel__OBReactionFacade swig_types[100]
#define SWIGTYPE_p_OpenBabel__OBResidue swig_types[101]
#define SWIGTYPE_p_OpenBabel__OBResidueAtomIter swig_types[102]
#define SWIGTYPE_p_OpenBabel__OBResidueData swig_types[103]
#define SWIGTYPE_p_OpenBabel__OBResidueIter swig_types[104]
#define SWIGTYPE_p_OpenBabel__OBRing swig_types[105]
#define SWIGTYPE_p_OpenBabel__OBRingData swig_types[106]
#define SWIGTYPE_p_OpenBabel__OBRingSearch swig_types[107]
#define SWIGTYPE_p_OpenBabel__OBRingTyper swig_types[108]
#define SWIGTYPE_p_OpenBabel__OBRotamerList swig_types[109]
#define SWIGTYPE_p_OpenBabel__OBRotationData swig_types[110]
#define SWIGTYPE_p_OpenBabel__OBRotor swig_types[111]
#define SWIGTYPE_p_OpenBabel__OBRotorKeys swig_types[112]
#define SWIGTYPE_p_OpenBabel__OBRotorList swig_types[113]
#define SWIGTYPE_p_OpenBabel__OBRotorRule swig_types[114]
#define SWIGTYPE_p_OpenBabel__OBRotorRules swig_types[115]
#define SWIGTYPE_p_OpenBabel__OBSSMatch swig_types[116]
#define SWIGTYPE_p_OpenBabel__OBSerialNums swig_types[117]
#define SWIGTYPE_p_OpenBabel__OBSetData swig_types[118]
#define SWIGTYPE_p_OpenBabel__OBSmartsMatcher swig_types[119]
#define SWIGTYPE_p_OpenBabel__OBSmartsPattern swig_types[120]
#define SWIGTYPE_p_OpenBabel__OBSpectrophore swig_types[121]
#define SWIGTYPE_p_OpenBabel__OBSqrtTbl swig_types[122]
#define SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig swig_types[123]
#define SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo swig_types[124]
#define SWIGTYPE_p_OpenBabel__OBStereo swig_types[125]
#define SWIGTYPE_p_OpenBabel__OBStereoBase swig_types[126]
#define SWIGTYPE_p_OpenBabel__OBStereoFacade swig_types[127]
#define SWIGTYPE_p_OpenBabel__OBStereoUnit swig_types[128]
#define SWIGTYPE_p_OpenBabel__OBStericConformerFilter swig_types[129]
#define SWIGTYPE_p_OpenBabel__OBStopwatch swig_types[130]
#define SWIGTYPE_p_OpenBabel__OBSymmetryData swig_types[131]
#define SWIGTYPE_p_OpenBabel__OBTetraNonPlanarStereo swig_types[132]
#define SWIGTYPE_p_OpenBabel__OBTetraPlanarStereo swig_types[133]
#define SWIGTYPE_p_OpenBabel__OBTetrahedralConfig swig_types[134]
#define SWIGTYPE_p_OpenBabel__OBTetrahedralStereo swig_types[135]
#define SWIGTYPE_p_OpenBabel__OBTorsion swig_types[136]
#define SWIGTYPE_p_OpenBabel__OBTorsionData swig_types[137]
#define SWIGTYPE_p_OpenBabel__OBTypeTable swig_types[138]
#define SWIGTYPE_p_OpenBabel__OBUnitCell swig_types[139]
#define SWIGTYPE_p_OpenBabel__OBVectorData swig_types[140]
#define SWIGTYPE_p_OpenBabel__OBVibrationData swig_types[141]
#define SWIGTYPE_p_OpenBabel__OBVirtualBond swig_types[142]
#define SWIGTYPE_p_OpenBabel__Pattern swig_types[143]
#define SWIGTYPE_p_OpenBabel__SpaceGroup swig_types[144]
#define SWIGTYPE_p_OpenBabel__Template swig_types[145]
#define SWIGTYPE_p_OpenBabel__matrix3x3 swig_types[146]
#define SWIGTYPE_p_OpenBabel__obLogBuf swig_types[147]
#define SWIGTYPE_p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t swig_types[148]
#define SWIGTYPE_p_OpenBabel__rotor_digit swig_types[149]
#define SWIGTYPE_p_OpenBabel__transform3d swig_types[150]
#define SWIGTYPE_p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t swig_types[151]
#define SWIGTYPE_p_OpenBabel__vector3 swig_types[152]
#define SWIGTYPE_p_PluginIterator swig_types[153]
#define SWIGTYPE_p_PluginMapType swig_types[154]
#define SWIGTYPE_p_Ref swig_types[155]
#define SWIGTYPE_p_RefIter swig_types[156]
#define SWIGTYPE_p_Refs swig_types[157]
#define SWIGTYPE_p_Unit swig_types[158]
#define SWIGTYPE_p_a_3__double swig_types[159]
#define SWIGTYPE_p_a_3__doubleArray swig_types[160]
#define SWIGTYPE_p_allocator_type swig_types[161]
#define SWIGTYPE_p_bool swig_types[162]
#define SWIGTYPE_p_char swig_types[163]
#define SWIGTYPE_p_difference_type swig_types[164]
#define SWIGTYPE_p_double swig_types[165]
#define SWIGTYPE_p_doubleArray swig_types[166]
#define SWIGTYPE_p_first_type swig_types[167]
#define SWIGTYPE_p_int swig_types[168]
#define SWIGTYPE_p_iterator_category swig_types[169]
#define SWIGTYPE_p_p_OpenBabel__OBChargeModel swig_types[170]
#define SWIGTYPE_p_p_OpenBabel__OBDescriptor swig_types[171]
#define SWIGTYPE_p_p_OpenBabel__OBFingerprint swig_types[172]
#define SWIGTYPE_p_p_OpenBabel__OBForceField swig_types[173]
#define SWIGTYPE_p_p_OpenBabel__OBFormat swig_types[174]
#define SWIGTYPE_p_p_OpenBabel__OBOp swig_types[175]
#define SWIGTYPE_p_p_PyObject swig_types[176]
#define SWIGTYPE_p_p_double swig_types[177]
#define SWIGTYPE_p_p_doubleArray swig_types[178]
#define SWIGTYPE_p_p_int swig_types[179]
#define SWIGTYPE_p_second_type swig_types[180]
#define SWIGTYPE_p_size_type swig_types[181]
#define SWIGTYPE_p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t swig_types[182]
#define SWIGTYPE_p_std__dequeT_int_t swig_types[183]
#define SWIGTYPE_p_std__invalid_argument swig_types[184]
#define SWIGTYPE_p_std__istream swig_types[185]
#define SWIGTYPE_p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator swig_types[186]
#define SWIGTYPE_p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t swig_types[187]
#define SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t swig_types[188]
#define SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t swig_types[189]
#define SWIGTYPE_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t swig_types[190]
#define SWIGTYPE_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator swig_types[191]
#define SWIGTYPE_p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t swig_types[192]
#define SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t swig_types[193]
#define SWIGTYPE_p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t swig_types[194]
#define SWIGTYPE_p_std__multimapT_double_unsigned_long_t swig_types[195]
#define SWIGTYPE_p_std__ostream swig_types[196]
#define SWIGTYPE_p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t swig_types[197]
#define SWIGTYPE_p_std__pairT_int_int_t swig_types[198]
#define SWIGTYPE_p_std__pairT_std__string_std__string_t swig_types[199]
#define SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t swig_types[200]
#define SWIGTYPE_p_std__setT_OpenBabel__OBBond_p_t swig_types[201]
#define SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t swig_types[202]
#define SWIGTYPE_p_std__streampos swig_types[203]
#define SWIGTYPE_p_std__string swig_types[204]
#define SWIGTYPE_p_std__stringbuf swig_types[205]
#define SWIGTYPE_p_std__type_info swig_types[206]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t swig_types[207]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator swig_types[208]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator swig_types[209]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t swig_types[210]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t swig_types[211]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator swig_types[212]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator swig_types[213]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t swig_types[214]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t swig_types[215]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t swig_types[216]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t swig_types[217]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator swig_types[218]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t swig_types[219]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator swig_types[220]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t swig_types[221]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator swig_types[222]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t swig_types[223]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t swig_types[224]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t swig_types[225]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t swig_types[226]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator swig_types[227]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t swig_types[228]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t swig_types[229]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t__iterator swig_types[230]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t swig_types[231]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator swig_types[232]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t swig_types[233]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t swig_types[234]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t swig_types[235]
#define SWIGTYPE_p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t swig_types[236]
#define SWIGTYPE_p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t swig_types[237]
#define SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t swig_types[238]
#define SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t swig_types[239]
#define SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator swig_types[240]
#define SWIGTYPE_p_std__vectorT_double_t swig_types[241]
#define SWIGTYPE_p_std__vectorT_int_t swig_types[242]
#define SWIGTYPE_p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t swig_types[243]
#define SWIGTYPE_p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t swig_types[244]
#define SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t swig_types[245]
#define SWIGTYPE_p_std__vectorT_std__string_t swig_types[246]
#define SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t swig_types[247]
#define SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t swig_types[248]
#define SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t swig_types[249]
#define SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t swig_types[250]
#define SWIGTYPE_p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t swig_types[251]
#define SWIGTYPE_p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t swig_types[252]
#define SWIGTYPE_p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator swig_types[253]
#define SWIGTYPE_p_std__vectorT_unsigned_int_t swig_types[254]
#define SWIGTYPE_p_std__vectorT_unsigned_long_t swig_types[255]
#define SWIGTYPE_p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t swig_types[256]
#define SWIGTYPE_p_swig__SwigPyIterator swig_types[257]
#define SWIGTYPE_p_unsigned_char swig_types[258]
#define SWIGTYPE_p_unsigned_int swig_types[259]
#define SWIGTYPE_p_value_type swig_types[260]
#define SWIGTYPE_p_void swig_types[261]
#define SWIGTYPE_p_word_vector swig_types[262]
static swig_type_info *swig_types[264];
static swig_module_info swig_module = {swig_types, 263, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _openbabel.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__openbabel

#else
#  define SWIG_init    init_openbabel

#endif

#ifdef __cplusplus
#include <utility>
/* SwigValueWrapper is described in swig.swg */
template<typename T> class SwigValueWrapper {
  struct SwigSmartPointer {
    T *ptr;
    SwigSmartPointer(T *p) : ptr(p) { }
    ~SwigSmartPointer() { delete ptr; }
    SwigSmartPointer& operator=(SwigSmartPointer& rhs) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = rhs.ptr; rhs.ptr = 0; return *this; }
    void reset(T *p) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = p; }
  } pointer;
  SwigValueWrapper& operator=(const SwigValueWrapper<T>& rhs);
  SwigValueWrapper(const SwigValueWrapper<T>& rhs);
public:
  SwigValueWrapper() : pointer(0) { }
  SwigValueWrapper& operator=(const T& t) { SwigSmartPointer tmp(new T(t)); pointer = tmp; return *this; }
#if __cplusplus >=201103L
  SwigValueWrapper& operator=(T&& t) { SwigSmartPointer tmp(new T(std::move(t))); pointer = tmp; return *this; }
  operator T&&() const { return std::move(*pointer.ptr); }
#else
  operator T&() const { return *pointer.ptr; }
#endif
  T *operator&() const { return pointer.ptr; }
  static void reset(SwigValueWrapper& t, T *p) { t.pointer.reset(p); }
};

/*
 * SwigValueInit() is a generic initialisation solution as the following approach:
 * 
 *       T c_result = T();
 * 
 * doesn't compile for all types for example:
 * 
 *       unsigned int c_result = unsigned int();
 */
template <typename T> T SwigValueInit() {
  return T();
}

#if __cplusplus >=201103L
# define SWIG_STD_MOVE(OBJ) std::move(OBJ)
#else
# define SWIG_STD_MOVE(OBJ) OBJ
#endif

#endif

#if defined(__cplusplus) && (__cplusplus >=201103L)
# define SWIG_OVERRIDE override
#else
# define SWIG_OVERRIDE
#endif



#define SWIG_as_voidptr(a) const_cast< void * >(static_cast< const void * >(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),reinterpret_cast< void** >(a)) 


#include <stdexcept>


namespace swig {
  class SwigPtr_PyObject {
  protected:
    PyObject *_obj;

  public:
    SwigPtr_PyObject() :_obj(0)
    {
    }

    SwigPtr_PyObject(const SwigPtr_PyObject& item) : _obj(item._obj)
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      SWIG_Py_XINCREF(_obj);
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    SwigPtr_PyObject(PyObject *obj, bool initial_ref = true) :_obj(obj)
    {
      if (initial_ref) {
        SWIG_PYTHON_THREAD_BEGIN_BLOCK;
        SWIG_Py_XINCREF(_obj);
        SWIG_PYTHON_THREAD_END_BLOCK;
      }
    }
    
    SwigPtr_PyObject & operator=(const SwigPtr_PyObject& item) 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      SWIG_Py_XINCREF(item._obj);
      SWIG_Py_XDECREF(_obj);
      _obj = item._obj;
      SWIG_PYTHON_THREAD_END_BLOCK;
      return *this;      
    }
    
    ~SwigPtr_PyObject() 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      SWIG_Py_XDECREF(_obj);
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    operator PyObject *() const
    {
      return _obj;
    }

    PyObject *operator->() const
    {
      return _obj;
    }
  };
}


namespace swig {
  struct SwigVar_PyObject : SwigPtr_PyObject {
    SwigVar_PyObject(PyObject* obj = 0) : SwigPtr_PyObject(obj, false) { }
    
    SwigVar_PyObject & operator = (PyObject* obj)
    {
      SWIG_Py_XDECREF(_obj);
      _obj = obj;
      return *this;      
    }
  };
}


// used to set import/export for Cygwin DLLs
#ifdef WIN32
#define USING_OBDLL
#endif

#include <openbabel/obutil.h>
#include <openbabel/math/vector3.h>
#include <openbabel/math/matrix3x3.h>
#include <openbabel/math/transform3d.h>
#include <openbabel/math/spacegroup.h>

#include <openbabel/generic.h>
#include <openbabel/griddata.h>
#include <openbabel/kekulize.h>
#include <openbabel/elements.h>

#include <openbabel/base.h>
#include <openbabel/mol.h>
#include <openbabel/atom.h>
#include <openbabel/bond.h>
#include <openbabel/reaction.h>
#include <openbabel/reactionfacade.h>
#include <openbabel/residue.h>
#include <openbabel/internalcoord.h>
#include <openbabel/bondtyper.h>

#include <openbabel/ring.h>
#include <openbabel/obconversion.h>
#include <openbabel/obfunctions.h>
#include <openbabel/oberror.h>
#include <openbabel/plugin.h>
#include <openbabel/fingerprint.h>
#include <openbabel/descriptor.h>
#include <openbabel/format.h>

#include <openbabel/forcefield.h>
#include <openbabel/builder.h>
#include <openbabel/op.h>

#include <openbabel/bitvec.h>
#include <openbabel/data.h>
#include <openbabel/parsmart.h>
#include <openbabel/alias.h>

#include <openbabel/kinetics.h>
#include <openbabel/rotor.h>
#include <openbabel/rotamer.h>
#include <openbabel/spectrophore.h>

#include <openbabel/chargemodel.h>
#include <openbabel/phmodel.h>
#include <openbabel/graphsym.h>
#include <openbabel/isomorphism.h>
#include <openbabel/query.h>
#include <openbabel/canon.h>

#include <openbabel/stereo/tetrahedral.h>
#include <openbabel/stereo/cistrans.h>
#include <openbabel/stereo/squareplanar.h>
#include <openbabel/stereo/bindings.h>

#include <openbabel/chains.h>
#include <openbabel/obiter.h>


#include <openbabel/conformersearch.h>
#include <openbabel/math/align.h>


#include <iostream>

#if PY_VERSION_HEX >= 0x03020000
# define SWIGPY_SLICEOBJECT PyObject
#else
# define SWIGPY_SLICEOBJECT PySliceObject
#endif


#include <typeinfo>
#include <stdexcept>


#if defined(__GNUC__)
#  if __GNUC__ == 2 && __GNUC_MINOR <= 96
#     define SWIG_STD_NOMODERN_STL
#  endif
#endif


#include <string>


#include <stddef.h>


namespace swig {
  struct SwigPyIterator {
  private:
    SwigPtr_PyObject _seq;

  protected:
    SwigPyIterator(PyObject *seq) : _seq(seq)
    {
    }

    void raise_stop_iteration() const
    {
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
    }

  public:
    virtual ~SwigPyIterator() {}

    // Access iterator method, required by Python
    virtual PyObject *value() const = 0;

    // Forward iterator method, required by Python
    virtual SwigPyIterator *incr(size_t n = 1) = 0;
    
    // Backward iterator method, very common in C++, but not required in Python
    // This default probably ought to instead raise some sort of unsupported exception
    virtual SwigPyIterator *decr(size_t /*n*/ = 1)
    {
      raise_stop_iteration();
      return NULL;
    }

    // Random access iterator methods, but not required in Python
    virtual ptrdiff_t distance(const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }

    virtual bool equal (const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }
    
    // C++ common/needed methods
    virtual SwigPyIterator *copy() const = 0;

    PyObject *next()     
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      PyObject *obj = value();
      if (obj)
        if (!incr())
          obj = NULL;
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads
      return obj;     
    }

    /* Make an alias for Python 3.x */
    PyObject *__next__()
    {
      return next();
    }

    PyObject *previous()
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      if (!decr())
        return NULL;
      PyObject *obj = value();
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads       
      return obj;
    }

    SwigPyIterator *advance(ptrdiff_t n)
    {
      return  (n > 0) ?  incr(n) : decr(-n);
    }
      
    bool operator == (const SwigPyIterator& x)  const
    {
      return equal(x);
    }
      
    bool operator != (const SwigPyIterator& x) const
    {
      return ! operator==(x);
    }
      
    SwigPyIterator& operator += (ptrdiff_t n)
    {
      SwigPyIterator *ret = advance(n);
      return ret ? *ret : *this;
    }

    SwigPyIterator& operator -= (ptrdiff_t n)
    {
      SwigPyIterator *ret = advance(-n);
      return ret ? *ret : *this;
    }
      
    SwigPyIterator* operator + (ptrdiff_t n) const
    {
      return copy()->advance(n);
    }

    SwigPyIterator* operator - (ptrdiff_t n) const
    {
      return copy()->advance(-n);
    }
      
    ptrdiff_t operator - (const SwigPyIterator& x) const
    {
      return x.distance(*this);
    }
      
    static swig_type_info* descriptor() {
      static swig_type_info* desc = SWIG_TypeQuery("swig::SwigPyIterator *");
      return desc;
    }    
  };

#if defined(SWIGPYTHON_BUILTIN)
  inline PyObject* make_output_iterator_builtin (PyObject *pyself)
  {
    SWIG_Py_INCREF(pyself);
    return pyself;
  }
#endif
}


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <errno.h>


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than ULONG_MAX (not portably calculated easily)
      // Note that double(ULONG_MAX) is stored in a double rounded up by one (for 64-bit unsigned long)
      // 0xfffffffffffff800ULL == (uint64_t)std::nextafter(double(__uint128_t(ULONG_MAX)+1), double(0))
      const double ulong_max = sizeof(unsigned long) == 8 ? 0xfffffffffffff800ULL : ULONG_MAX;
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ulong_max)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, 0, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
  }
#endif
  return res;
}


  #define SWIG_From_long   PyInt_FromLong 


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_ptrdiff_t  (ptrdiff_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    return SWIG_From_long  (static_cast< long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(ptrdiff_t) <= sizeof(long long) */
    return SWIG_From_long_SS_long  (static_cast< long long >(value));
  }
#endif
}


SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than LONG_MAX (not portably calculated easily)
      // Note that double(LONG_MAX) is stored in a double rounded up by one (for 64-bit long)
      // 0x7ffffffffffffc00LL == (int64_t)std::nextafter(double(__uint128_t(LONG_MAX)+1), double(0))
      const double long_max = sizeof(long) == 8 ? 0x7ffffffffffffc00LL : LONG_MAX;
      // No equivalent needed for 64-bit double(LONG_MIN) is exactly LONG_MIN
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, long_max)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_long_SS_long (PyObject *obj, long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    long long v = PyLong_AsLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    long v;
    res = SWIG_AsVal_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    const double mant_min = -mant_max;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, mant_min, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, mant_min, mant_max)) {
      if (val) *val = (long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_ptrdiff_t (PyObject * obj, ptrdiff_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    long v;
    res = SWIG_AsVal_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(ptrdiff_t) <= sizeof(long long)) {
    long long v;
    res = SWIG_AsVal_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
  }
#endif
  return res;
}


#include <algorithm>


#include <list>


#include <utility>


#include <map>


#include <algorithm>


#include <vector>


namespace swig {
  template <class Type>
  struct noconst_traits {
    typedef Type noconst_type;
  };

  template <class Type>
  struct noconst_traits<const Type> {
    typedef Type noconst_type;
  };

  /*
    type categories
  */
  struct pointer_category { };
  struct value_category { };

  /*
    General traits that provides type_name and type_info
  */
  template <class Type> struct traits { };

  template <class Type>
  inline const char* type_name() {
    return traits<typename noconst_traits<Type >::noconst_type >::type_name();
  }

  template <class Type> struct traits_info {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  /*
    Partial specialization for pointers (traits_info)
  */
  template <class Type> struct traits_info<Type *> {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  template <class Type>
  inline swig_type_info *type_info() {
    return traits_info<Type>::type_info();
  }

  /*
    Partial specialization for pointers (traits)
  */
  template <class Type> struct traits <Type *> {
    typedef pointer_category category;
    static std::string make_ptr_name(const char* name) {
      std::string ptrname = name;
      ptrname += " *";
      return ptrname;
    }
    static const char* type_name() {
      static std::string name = make_ptr_name(swig::type_name<Type>());
      return name.c_str();
    }
  };

  template <class Type, class Category>
  struct traits_as { };

  template <class Type, class Category>
  struct traits_check { };

}


namespace swig {  
  /*
    Traits that provides the from method
  */
  template <class Type> struct traits_from_ptr {
    static PyObject *from(Type *val, int owner = 0) {
      return SWIG_InternalNewPointerObj(val, type_info<Type>(), owner);
    }
  };

  template <class Type> struct traits_from {
    static PyObject *from(const Type& val) {
      return traits_from_ptr<Type>::from(new Type(val), 1);
    }
  };

  template <class Type> struct traits_from<Type *> {
    static PyObject *from(Type* val) {
      return traits_from_ptr<Type>::from(val, 0);
    }
  };

  template <class Type> struct traits_from<const Type *> {
    static PyObject *from(const Type* val) {
      return traits_from_ptr<Type>::from(const_cast<Type*>(val), 0);
    }
  };


  template <class Type>
  inline PyObject *from(const Type& val) {
    return traits_from<Type>::from(val);
  }

  template <class Type>
  inline PyObject *from_ptr(Type* val, int owner) {
    return traits_from_ptr<Type>::from(val, owner);
  }

  /*
    Traits that provides the asval/as/check method
  */
  template <class Type>
  struct traits_asptr {   
    static int asptr(PyObject *obj, Type **val) {
      int res = SWIG_ERROR;
      swig_type_info *descriptor = type_info<Type>();
      if (val) {
        Type *p = 0;
        int newmem = 0;
        res = descriptor ? SWIG_ConvertPtrAndOwn(obj, (void **)&p, descriptor, 0, &newmem) : SWIG_ERROR;
        if (SWIG_IsOK(res)) {
          if (newmem & SWIG_CAST_NEW_MEMORY) {
            res |= SWIG_NEWOBJMASK;
          }
          *val = p;
        }
      } else {
        res = descriptor ? SWIG_ConvertPtr(obj, 0, descriptor, 0) : SWIG_ERROR;
      }
      return res;
    }
  }; 

  template <class Type>
  inline int asptr(PyObject *obj, Type **vptr) {
    return traits_asptr<Type>::asptr(obj, vptr);
  }

  template <class Type> 
  struct traits_asval {
    static int asval(PyObject *obj, Type *val) {
      if (val) {
	Type *p = 0;
	int res = traits_asptr<Type>::asptr(obj, &p);
	if (!SWIG_IsOK(res)) return res;	
	if (p) {
	  typedef typename noconst_traits<Type>::noconst_type noconst_type;
	  *(const_cast<noconst_type*>(val)) = *p;
	  if (SWIG_IsNewObj(res)){
	    delete p;
	    res = SWIG_DelNewMask(res);
	  }
	  return res;
	} else {
	  return SWIG_ERROR;
	}
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };

  template <class Type> struct traits_asval<Type*> {
    static int asval(PyObject *obj, Type **val) {
      if (val) {
        typedef typename noconst_traits<Type>::noconst_type noconst_type;
        noconst_type *p = 0;
        int res = traits_asptr<noconst_type>::asptr(obj,  &p);
        if (SWIG_IsOK(res)) {
          *(const_cast<noconst_type**>(val)) = p;
	}
	return res;
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };
  
  template <class Type>
  inline int asval(PyObject *obj, Type *val) {
    return traits_asval<Type>::asval(obj, val);
  }

  template <class Type> 
  struct traits_as<Type, value_category> {
    static Type as(PyObject *obj) {
      Type v;
      int res = asval(obj, &v);
      if (!obj || !SWIG_IsOK(res)) {
	if (!PyErr_Occurred()) {
	  ::SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
      return v;
    }
  };

  template <class Type> 
  struct traits_as<Type, pointer_category> {
    static Type as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res) && v) {
	if (SWIG_IsNewObj(res)) {
	  Type r(*v);
	  delete v;
	  return r;
	} else {
	  return *v;
	}
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };

  template <class Type> 
  struct traits_as<Type*, pointer_category> {
    static Type* as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res)) {
	return v;
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };
    
  template <class Type>
  inline Type as(PyObject *obj) {
    return traits_as<Type, typename traits<Type>::category>::as(obj);
  }

  template <class Type> 
  struct traits_check<Type, value_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asval(obj, (Type *)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type> 
  struct traits_check<Type, pointer_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asptr(obj, (Type **)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type>
  inline bool check(PyObject *obj) {
    return traits_check<Type, typename traits<Type>::category>::check(obj);
  }
}


#if defined(__SUNPRO_CC) && defined(_RWSTD_VER)
#  if !defined(SWIG_NO_STD_NOITERATOR_TRAITS_STL)
#    define SWIG_STD_NOITERATOR_TRAITS_STL
#  endif
#endif

#if !defined(SWIG_STD_NOITERATOR_TRAITS_STL)
#include <iterator>
#else
namespace std {
  template <class Iterator>
  struct iterator_traits {
    typedef ptrdiff_t difference_type;
    typedef typename Iterator::value_type value_type;
  };

  template <class Iterator, class Category,class T, class Reference, class Pointer, class Distance>
  struct iterator_traits<__reverse_bi_iterator<Iterator,Category,T,Reference,Pointer,Distance> > {
    typedef Distance difference_type;
    typedef T value_type;
  };

  template <class T>
  struct iterator_traits<T*> {
    typedef T value_type;
    typedef ptrdiff_t difference_type;
  };

  template<typename _InputIterator>
  inline typename iterator_traits<_InputIterator>::difference_type
  distance(_InputIterator __first, _InputIterator __last)
  {
    typename iterator_traits<_InputIterator>::difference_type __n = 0;
    while (__first != __last) {
      ++__first; ++__n;
    }
    return __n;
  }
}
#endif


namespace swig {
  template<typename OutIterator>
  class SwigPyIterator_T :  public SwigPyIterator
  {
  public:
    typedef OutIterator out_iterator;
    typedef typename std::iterator_traits<out_iterator>::value_type value_type;    
    typedef SwigPyIterator_T<out_iterator> self_type;

    SwigPyIterator_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator(seq), current(curr)
    {
    }

    const out_iterator& get_current() const
    {
      return current;
    }

    
    bool equal (const SwigPyIterator &iter) const SWIG_OVERRIDE
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return (current == iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }
    
    ptrdiff_t distance(const SwigPyIterator &iter) const SWIG_OVERRIDE
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return std::distance(current, iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }    
    
  protected:
    out_iterator current;
  };
  
  template <class ValueType>
  struct from_oper 
  {
    typedef const ValueType& argument_type;
    typedef PyObject *result_type;
    result_type operator()(argument_type v) const
    {
      return swig::from(v);
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorOpen_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq)
    {
    }
    
    PyObject *value() const SWIG_OVERRIDE {
      return from(static_cast<const value_type&>(*(base::current)));
    }
    
    SwigPyIterator *copy() const SWIG_OVERRIDE
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1) SWIG_OVERRIDE
    {
      // Open iterator does not raise StopIteration
      while (n--) {
	++base::current;
      }
      return this;
    }

  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorOpen_T :  public SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyForwardIteratorOpen_T<OutIterator>(curr, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1) SWIG_OVERRIDE
    {
      // Open iterator does not raise StopIteration
      while (n--) {
	--base::current;
      }
      return this;
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorClosed_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;    
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq), begin(first), end(last)
    {
    }
    
    PyObject *value() const SWIG_OVERRIDE {
      if (base::current == end) {
	base::raise_stop_iteration();
	return NULL;
      } else {
	return from(static_cast<const value_type&>(*(base::current)));
      }
    }
    
    SwigPyIterator *copy() const SWIG_OVERRIDE
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1) SWIG_OVERRIDE
    {
      while (n--) {
	if (base::current == end) {
	  base::raise_stop_iteration();
	  return NULL;
	} else {
	  ++base::current;
	}
      }
      return this;
    }

  protected:
    out_iterator begin;
    out_iterator end;
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorClosed_T :  public SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> base0;
    typedef SwigPyIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>(curr, first, last, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1) SWIG_OVERRIDE
    {
      while (n--) {
	if (base::current == base0::begin) {
	  base0::raise_stop_iteration();
	  return NULL;
	} else {
	  --base::current;
	}
      }
      return this;
    }
  };


  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorOpen_T<OutIter>(current, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyIteratorOpen_T<OutIter>(current, seq);
  }

}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


namespace swig {
  template <> struct traits< int > {
    typedef value_category category;
    static const char* type_name() { return"int"; }
  };
  template <>  struct traits_asval< int > {
    typedef int value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_int (obj, val);
    }
  };
  template <>  struct traits_from< int > {
    typedef int value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_int  (val);
    }
  };
}


#include <functional>

namespace std {
  template <>
  struct less <PyObject *>
  {
    bool
    operator()(PyObject * v, PyObject *w) const
    { 
      bool res;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      res = PyObject_RichCompareBool(v, w, Py_LT) ? true : false;
      /* This may fall into a case of inconsistent
               eg. ObjA > ObjX > ObjB
               but ObjA < ObjB
      */
      if( PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_TypeError) )
      {
        /* Objects can't be compared, this mostly occurred in Python 3.0 */
        /* Compare their ptr directly for a workaround */
        res = (v < w);
        PyErr_Clear();
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return res;
    }
  };

  template <>
  struct less <swig::SwigPtr_PyObject>
  {
    bool
    operator()(const swig::SwigPtr_PyObject& v, const swig::SwigPtr_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

  template <>
  struct less <swig::SwigVar_PyObject>
  {
    bool
    operator()(const swig::SwigVar_PyObject& v, const swig::SwigVar_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

}

namespace swig {
  template <> struct traits<PyObject *> {
    typedef value_category category;
    static const char* type_name() { return "PyObject *"; }
  };  

  template <>  struct traits_asval<PyObject * > {   
    typedef PyObject * value_type;
    static int asval(PyObject *obj, value_type *val) {
      if (val) *val = obj;
      return SWIG_OK;
    }
  };

  template <> 
  struct traits_check<PyObject *, value_category> {
    static bool check(PyObject *) {
      return true;
    }
  };

  template <>  struct traits_from<PyObject *> {
    typedef PyObject * value_type;
    static PyObject *from(const value_type& val) {
      SWIG_Py_XINCREF(val);
      return val;
    }
  };
  
}

namespace swig {
  template <class Difference>
  inline size_t
  check_index(Difference i, size_t size, bool insert = false) {
    if ( i < 0 ) {
      if ((size_t) (-i) <= size)
	return (size_t) (i + size);
    } else if ( (size_t) i < size ) {
      return (size_t) i;
    } else if (insert && ((size_t) i == size)) {
      return size;
    }
    throw std::out_of_range("index out of range");
  }

  template <class Difference>
  void
  slice_adjust(Difference i, Difference j, Py_ssize_t step, size_t size, Difference &ii, Difference &jj, bool insert = false) {
    if (step == 0) {
      throw std::invalid_argument("slice step cannot be zero");
    } else if (step > 0) {
      // Required range: 0 <= i < size, 0 <= j < size, i <= j
      if (i < 0) {
        ii = 0;
      } else if (i < (Difference)size) {
        ii = i;
      } else if (insert && (i >= (Difference)size)) {
        ii = (Difference)size;
      }
      if (j < 0) {
        jj = 0;
      } else {
        jj = (j < (Difference)size) ? j : (Difference)size;
      }
      if (jj < ii)
        jj = ii;
    } else {
      // Required range: -1 <= i < size-1, -1 <= j < size-1, i >= j
      if (i < -1) {
        ii = -1;
      } else if (i < (Difference) size) {
        ii = i;
      } else if (i >= (Difference)(size-1)) {
        ii = (Difference)(size-1);
      }
      if (j < -1) {
        jj = -1;
      } else {
        jj = (j < (Difference)size ) ? j : (Difference)(size-1);
      }
      if (ii < jj)
        ii = jj;
    }
  }

  template <class Sequence, class Difference>
  inline typename Sequence::iterator
  getpos(Sequence* self, Difference i)  {
    typename Sequence::iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence, class Difference>
  inline typename Sequence::const_iterator
  cgetpos(const Sequence* self, Difference i)  {
    typename Sequence::const_iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence>
  inline void
  erase(Sequence* seq, const typename Sequence::iterator& position) {
    seq->erase(position);
  }

  template <class Sequence>
  struct traits_reserve {
    static void reserve(Sequence & /*seq*/, typename Sequence::size_type /*n*/) {
      // This should be specialized for types that support reserve
    }
  };

  template <class Sequence, class Difference>
  inline Sequence*
  getslice(const Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj);

    if (step > 0) {
      typename Sequence::const_iterator sb = self->begin();
      typename Sequence::const_iterator se = self->begin();
      std::advance(sb,ii);
      std::advance(se,jj);
      if (step == 1) {
        return new Sequence(sb, se);
      } else {
        Sequence *sequence = new Sequence();
        swig::traits_reserve<Sequence>::reserve(*sequence, (jj - ii + step - 1) / step);
        typename Sequence::const_iterator it = sb;
        while (it!=se) {
          sequence->push_back(*it);
          for (Py_ssize_t c=0; c<step && it!=se; ++c)
            it++;
        }
        return sequence;
      } 
    } else {
      Sequence *sequence = new Sequence();
      swig::traits_reserve<Sequence>::reserve(*sequence, (ii - jj - step - 1) / -step);
      typename Sequence::const_reverse_iterator sb = self->rbegin();
      typename Sequence::const_reverse_iterator se = self->rbegin();
      std::advance(sb,size-ii-1);
      std::advance(se,size-jj-1);
      typename Sequence::const_reverse_iterator it = sb;
      while (it!=se) {
        sequence->push_back(*it);
        for (Py_ssize_t c=0; c<-step && it!=se; ++c)
          it++;
      }
      return sequence;
    }
  }

  template <class Sequence, class Difference, class InputSeq>
  inline void
  setslice(Sequence* self, Difference i, Difference j, Py_ssize_t step, const InputSeq& is = InputSeq()) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      if (step == 1) {
        size_t ssize = jj - ii;
        if (ssize <= is.size()) {
          // expanding/staying the same size
          swig::traits_reserve<Sequence>::reserve(*self, self->size() - ssize + is.size());
          typename Sequence::iterator sb = self->begin();
          typename InputSeq::const_iterator isit = is.begin();
          std::advance(sb,ii);
          std::advance(isit, jj - ii);
          self->insert(std::copy(is.begin(), isit, sb), isit, is.end());
        } else {
          // shrinking
          typename Sequence::iterator sb = self->begin();
          typename Sequence::iterator se = self->begin();
          std::advance(sb,ii);
          std::advance(se,jj);
          self->erase(sb,se);
          sb = self->begin();
          std::advance(sb,ii);
          self->insert(sb, is.begin(), is.end());
        }
      } else {
        size_t replacecount = (jj - ii + step - 1) / step;
        if (is.size() != replacecount) {
          char msg[1024];
          PyOS_snprintf(msg, sizeof(msg), "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
          throw std::invalid_argument(msg);
        }
        typename Sequence::const_iterator isit = is.begin();
        typename Sequence::iterator it = self->begin();
        std::advance(it,ii);
        for (size_t rc=0; rc<replacecount && it != self->end(); ++rc) {
          *it++ = *isit++;
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
        }
      }
    } else {
      size_t replacecount = (ii - jj - step - 1) / -step;
      if (is.size() != replacecount) {
        char msg[1024];
        PyOS_snprintf(msg, sizeof(msg), "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
        throw std::invalid_argument(msg);
      }
      typename Sequence::const_iterator isit = is.begin();
      typename Sequence::reverse_iterator it = self->rbegin();
      std::advance(it,size-ii-1);
      for (size_t rc=0; rc<replacecount && it != self->rend(); ++rc) {
        *it++ = *isit++;
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
      }
    }
  }

  template <class Sequence, class Difference>
  inline void
  delslice(Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      typename Sequence::iterator sb = self->begin();
      std::advance(sb,ii);
      if (step == 1) {
        typename Sequence::iterator se = self->begin();
        std::advance(se,jj);
        self->erase(sb,se);
      } else {
        typename Sequence::iterator it = sb;
        size_t delcount = (jj - ii + step - 1) / step;
        while (delcount) {
          it = self->erase(it);
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
          delcount--;
        }
      }
    } else {
      typename Sequence::reverse_iterator sb = self->rbegin();
      std::advance(sb,size-ii-1);
      typename Sequence::reverse_iterator it = sb;
      size_t delcount = (ii - jj - step - 1) / -step;
      while (delcount) {
        it = typename Sequence::reverse_iterator(self->erase((++it).base()));
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
        delcount--;
      }
    }
  }
}


namespace swig {
  template <class Seq, class T = typename Seq::value_type >
  struct IteratorProtocol {
    static void assign(PyObject *obj, Seq *seq) {
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      if (iter) {
        SwigVar_PyObject item = PyIter_Next(iter);
        while (item) {
          seq->insert(seq->end(), swig::as<T>(item));
          item = PyIter_Next(iter);
        }
      }
    }

    static bool check(PyObject *obj) {
      bool ret = false;
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      if (iter) {
        SwigVar_PyObject item = PyIter_Next(iter);
        ret = true;
        while (item) {
          ret = swig::check<T>(item);
          item = ret ? PyIter_Next(iter) : 0;
        }
      }
      return ret;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_asptr_stdseq {
    typedef Seq sequence;
    typedef T value_type;

    static bool is_iterable(PyObject *obj) {
      SwigVar_PyObject iter = PyObject_GetIter(obj);
      PyErr_Clear();
      return iter != 0;
    }

    static int asptr(PyObject *obj, sequence **seq) {
      int ret = SWIG_ERROR;
      if (obj == Py_None || SWIG_Python_GetSwigThis(obj)) {
	sequence *p;
	swig_type_info *descriptor = swig::type_info<sequence>();
	if (descriptor && SWIG_IsOK(::SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0))) {
	  if (seq) *seq = p;
	  return SWIG_OLDOBJ;
	}
      } else if (is_iterable(obj)) {
	try {
	  if (seq) {
	    *seq = new sequence();
            IteratorProtocol<Seq, T>::assign(obj, *seq);
            if (!PyErr_Occurred())
              return SWIG_NEWOBJ;
	  } else {
	    return IteratorProtocol<Seq, T>::check(obj) ? SWIG_OK : SWIG_ERROR;
	  }
	} catch (std::exception& e) {
          if (seq && !PyErr_Occurred())
            PyErr_SetString(PyExc_TypeError, e.what());
	}
        if (seq)
          delete *seq;
	return SWIG_ERROR;
      }
      return ret;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_from_stdseq {
    typedef Seq sequence;
    typedef T value_type;
    typedef typename Seq::size_type size_type;
    typedef typename sequence::const_iterator const_iterator;

    static PyObject *from(const sequence& seq) {
#ifdef SWIG_PYTHON_EXTRA_NATIVE_CONTAINERS
      swig_type_info *desc = swig::type_info<sequence>();
      if (desc && desc->clientdata) {
	return SWIG_InternalNewPointerObj(new sequence(seq), desc, SWIG_POINTER_OWN);
      }
#endif
      size_type size = seq.size();
      if (size <= (size_type)PY_SSIZE_T_MAX) {
	PyObject *obj = PyTuple_New((Py_ssize_t)size);
	Py_ssize_t i = 0;
	for (const_iterator it = seq.begin(); it != seq.end(); ++it, ++i) {
	  PyTuple_SetItem(obj,i,swig::from<value_type>(*it));
	}
	return obj;
      } else {
	PyErr_SetString(PyExc_OverflowError,"sequence size not valid in python");
	return NULL;
      }
    }
  };
}


  namespace swig {
    template <class T>
    struct traits_reserve<std::vector<T> > {
      static void reserve(std::vector<T> &seq, typename std::vector<T>::size_type n) {
        seq.reserve(n);
      }
    };

    template <class T>
    struct traits_asptr<std::vector<T> >  {
      static int asptr(PyObject *obj, std::vector<T> **vec) {
	return traits_asptr_stdseq<std::vector<T> >::asptr(obj, vec);
      }
    };
    
    template <class T>
    struct traits_from<std::vector<T> > {
      static PyObject *from(const std::vector<T>& vec) {
	return traits_from_stdseq<std::vector<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< int, std::allocator< int > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "int" "," "std::allocator< int >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_int_Sg__iterator(std::vector< int > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_int_Sg____nonzero__(std::vector< int > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_int_Sg____bool__(std::vector< int > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< int >::size_type std_vector_Sl_int_Sg____len__(std::vector< int > const *self){
      return self->size();
    }

SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong(static_cast< long >(value));
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    return SWIG_From_unsigned_SS_long  (static_cast< unsigned long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(size_t) <= sizeof(unsigned long long) */
    return SWIG_From_unsigned_SS_long_SS_long  (static_cast< unsigned long long >(value));
  }
#endif
}

SWIGINTERN std::vector< int,std::allocator< int > > *std_vector_Sl_int_Sg____getslice__(std::vector< int > *self,std::vector< int >::difference_type i,std::vector< int >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_int_Sg____setslice____SWIG_0(std::vector< int > *self,std::vector< int >::difference_type i,std::vector< int >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< int,std::allocator< int > >());
    }
SWIGINTERN void std_vector_Sl_int_Sg____setslice____SWIG_1(std::vector< int > *self,std::vector< int >::difference_type i,std::vector< int >::difference_type j,std::vector< int,std::allocator< int > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_int_Sg____delslice__(std::vector< int > *self,std::vector< int >::difference_type i,std::vector< int >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_int_Sg____delitem____SWIG_0(std::vector< int > *self,std::vector< int >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< int,std::allocator< int > > *std_vector_Sl_int_Sg____getitem____SWIG_0(std::vector< int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< int,std::allocator< int > >::difference_type id = i;
      std::vector< int,std::allocator< int > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_int_Sg____setitem____SWIG_0(std::vector< int > *self,SWIGPY_SLICEOBJECT *slice,std::vector< int,std::allocator< int > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< int,std::allocator< int > >::difference_type id = i;
      std::vector< int,std::allocator< int > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_int_Sg____setitem____SWIG_1(std::vector< int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< int,std::allocator< int > >::difference_type id = i;
      std::vector< int,std::allocator< int > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_int_Sg____delitem____SWIG_1(std::vector< int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< int,std::allocator< int > >::difference_type id = i;
      std::vector< int,std::allocator< int > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< int >::value_type const &std_vector_Sl_int_Sg____getitem____SWIG_1(std::vector< int > const *self,std::vector< int >::difference_type i){
      return *(swig::cgetpos(self, i));
    }

namespace swig {
  static PyObject* container_owner_attribute() {
    static PyObject* attr = SWIG_Python_str_FromChar("__swig_container");
    return attr;
  }

  template <typename T>
  struct container_owner {
    // By default, do not add the back-reference (for value types)
    // Specialization below will check the reference for pointer types.
    static bool back_reference(PyObject* /*child*/, PyObject* /*owner*/) {
      return false;
    }
  };

  template <>
  struct container_owner<swig::pointer_category> {  
    /*
     * Call to add a back-reference to the owning object when returning a 
     * reference from a container.  Will only set the reference if child
     * is a SWIG wrapper object that does not own the pointer.
     *
     * returns whether the reference was set or not
     */
    static bool back_reference(PyObject* child, PyObject* owner) {
      SwigPyObject* swigThis = SWIG_Python_GetSwigThis(child);
      if (swigThis && (swigThis->own & SWIG_POINTER_OWN) != SWIG_POINTER_OWN) {
        return PyObject_SetAttr(child, container_owner_attribute(), owner) != -1;
      }
      return false;
    }
  };
}

SWIGINTERN void std_vector_Sl_int_Sg____setitem____SWIG_2(std::vector< int > *self,std::vector< int >::difference_type i,std::vector< int >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}


namespace swig {
  template <> struct traits< unsigned int > {
    typedef value_category category;
    static const char* type_name() { return"unsigned int"; }
  };
  template <>  struct traits_asval< unsigned int > {
    typedef unsigned int value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_unsigned_SS_int (obj, val);
    }
  };
  template <>  struct traits_from< unsigned int > {
    typedef unsigned int value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_unsigned_SS_int  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< unsigned int, std::allocator< unsigned int > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "unsigned int" "," "std::allocator< unsigned int >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_unsigned_SS_int_Sg__iterator(std::vector< unsigned int > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_unsigned_SS_int_Sg____nonzero__(std::vector< unsigned int > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_unsigned_SS_int_Sg____bool__(std::vector< unsigned int > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< unsigned int >::size_type std_vector_Sl_unsigned_SS_int_Sg____len__(std::vector< unsigned int > const *self){
      return self->size();
    }
SWIGINTERN std::vector< unsigned int,std::allocator< unsigned int > > *std_vector_Sl_unsigned_SS_int_Sg____getslice__(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i,std::vector< unsigned int >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setslice____SWIG_0(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i,std::vector< unsigned int >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< unsigned int,std::allocator< unsigned int > >());
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setslice____SWIG_1(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i,std::vector< unsigned int >::difference_type j,std::vector< unsigned int,std::allocator< unsigned int > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____delslice__(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i,std::vector< unsigned int >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____delitem____SWIG_0(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< unsigned int,std::allocator< unsigned int > > *std_vector_Sl_unsigned_SS_int_Sg____getitem____SWIG_0(std::vector< unsigned int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type id = i;
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_0(std::vector< unsigned int > *self,SWIGPY_SLICEOBJECT *slice,std::vector< unsigned int,std::allocator< unsigned int > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type id = i;
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_1(std::vector< unsigned int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type id = i;
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____delitem____SWIG_1(std::vector< unsigned int > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type id = i;
      std::vector< unsigned int,std::allocator< unsigned int > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< unsigned int >::value_type const &std_vector_Sl_unsigned_SS_int_Sg____getitem____SWIG_1(std::vector< unsigned int > const *self,std::vector< unsigned int >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_2(std::vector< unsigned int > *self,std::vector< unsigned int >::difference_type i,std::vector< unsigned int >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

      namespace swig {
	template <>  struct traits<std::vector< std::vector< int,std::allocator< int > >, std::allocator< std::vector< int,std::allocator< int > > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::vector< int,std::allocator< int > >" "," "std::allocator< std::vector< int,std::allocator< int > > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_vector_Sl_int_Sg__Sg__iterator(std::vector< std::vector< int > > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_int_Sg__Sg____nonzero__(std::vector< std::vector< int > > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_int_Sg__Sg____bool__(std::vector< std::vector< int > > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::vector< int > >::size_type std_vector_Sl_std_vector_Sl_int_Sg__Sg____len__(std::vector< std::vector< int > > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *std_vector_Sl_std_vector_Sl_int_Sg__Sg____getslice__(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i,std::vector< std::vector< int > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____setslice____SWIG_0(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i,std::vector< std::vector< int > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >());
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____setslice____SWIG_1(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i,std::vector< std::vector< int > >::difference_type j,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____delslice__(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i,std::vector< std::vector< int > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____delitem____SWIG_0(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *std_vector_Sl_std_vector_Sl_int_Sg__Sg____getitem____SWIG_0(std::vector< std::vector< int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type id = i;
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_0(std::vector< std::vector< int > > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type id = i;
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_1(std::vector< std::vector< int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type id = i;
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____delitem____SWIG_1(std::vector< std::vector< int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type id = i;
      std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::vector< int > >::value_type const &std_vector_Sl_std_vector_Sl_int_Sg__Sg____getitem____SWIG_1(std::vector< std::vector< int > > const *self,std::vector< std::vector< int > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_2(std::vector< std::vector< int > > *self,std::vector< std::vector< int > >::difference_type i,std::vector< std::vector< int > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  #define SWIG_From_double   PyFloat_FromDouble 


namespace swig {
  template <> struct traits< double > {
    typedef value_category category;
    static const char* type_name() { return"double"; }
  };
  template <>  struct traits_asval< double > {
    typedef double value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_double (obj, val);
    }
  };
  template <>  struct traits_from< double > {
    typedef double value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_double  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< double, std::allocator< double > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "double" "," "std::allocator< double >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_double_Sg__iterator(std::vector< double > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_double_Sg____nonzero__(std::vector< double > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_double_Sg____bool__(std::vector< double > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< double >::size_type std_vector_Sl_double_Sg____len__(std::vector< double > const *self){
      return self->size();
    }
SWIGINTERN std::vector< double,std::allocator< double > > *std_vector_Sl_double_Sg____getslice__(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setslice____SWIG_0(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< double,std::allocator< double > >());
    }
SWIGINTERN void std_vector_Sl_double_Sg____setslice____SWIG_1(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j,std::vector< double,std::allocator< double > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delslice__(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delitem____SWIG_0(std::vector< double > *self,std::vector< double >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< double,std::allocator< double > > *std_vector_Sl_double_Sg____getitem____SWIG_0(std::vector< double > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_0(std::vector< double > *self,SWIGPY_SLICEOBJECT *slice,std::vector< double,std::allocator< double > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_1(std::vector< double > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delitem____SWIG_1(std::vector< double > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< double >::value_type const &std_vector_Sl_double_Sg____getitem____SWIG_1(std::vector< double > const *self,std::vector< double >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_2(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

namespace swig {
  template <> struct traits< unsigned long > {
    typedef value_category category;
    static const char* type_name() { return"unsigned long"; }
  };
  template <>  struct traits_asval< unsigned long > {
    typedef unsigned long value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_unsigned_SS_long (obj, val);
    }
  };
  template <>  struct traits_from< unsigned long > {
    typedef unsigned long value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_unsigned_SS_long  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< unsigned long, std::allocator< unsigned long > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "unsigned long" "," "std::allocator< unsigned long >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_unsigned_SS_long_Sg__iterator(std::vector< unsigned long > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_unsigned_SS_long_Sg____nonzero__(std::vector< unsigned long > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_unsigned_SS_long_Sg____bool__(std::vector< unsigned long > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< unsigned long >::size_type std_vector_Sl_unsigned_SS_long_Sg____len__(std::vector< unsigned long > const *self){
      return self->size();
    }
SWIGINTERN std::vector< unsigned long,std::allocator< unsigned long > > *std_vector_Sl_unsigned_SS_long_Sg____getslice__(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i,std::vector< unsigned long >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____setslice____SWIG_0(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i,std::vector< unsigned long >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< unsigned long,std::allocator< unsigned long > >());
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____setslice____SWIG_1(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i,std::vector< unsigned long >::difference_type j,std::vector< unsigned long,std::allocator< unsigned long > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____delslice__(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i,std::vector< unsigned long >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____delitem____SWIG_0(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< unsigned long,std::allocator< unsigned long > > *std_vector_Sl_unsigned_SS_long_Sg____getitem____SWIG_0(std::vector< unsigned long > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type id = i;
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_0(std::vector< unsigned long > *self,SWIGPY_SLICEOBJECT *slice,std::vector< unsigned long,std::allocator< unsigned long > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type id = i;
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_1(std::vector< unsigned long > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type id = i;
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____delitem____SWIG_1(std::vector< unsigned long > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type id = i;
      std::vector< unsigned long,std::allocator< unsigned long > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< unsigned long >::value_type const &std_vector_Sl_unsigned_SS_long_Sg____getitem____SWIG_1(std::vector< unsigned long > const *self,std::vector< unsigned long >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_2(std::vector< unsigned long > *self,std::vector< unsigned long >::difference_type i,std::vector< unsigned long >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


/* Return string from Python obj. NOTE: obj must remain in scope in order
   to use the returned cptr (but only when alloc is set to SWIG_OLDOBJ) */
SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char **cptr, size_t *psize, int *alloc)
{
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    PyObject *bytes = NULL;
    int ret = SWIG_OK;
    if (alloc)
      *alloc = SWIG_OLDOBJ;
#if PY_VERSION_HEX >= 0x03000000 && defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    cstr = (char *)SWIG_PyUnicode_AsUTF8AndSize(obj, &len, &bytes);
    if (!cstr)
      return SWIG_TypeError;
    /* The returned string is only duplicated if the char * returned is not owned and memory managed by obj */
    if (bytes && cptr) {
      if (alloc) {
        cstr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        *alloc = SWIG_NEWOBJ;
      } else {
        /* alloc must be set in order to clean up allocated memory */
        return SWIG_RuntimeError;
      }
    }
#endif
    if (cptr) *cptr = cstr;
    if (psize) *psize = len + 1;
    SWIG_Py_XDECREF(bytes);
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX < 0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        }
        if (psize) *psize = len + 1;

        SWIG_Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        SWIG_Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsPtr_std_string (PyObject * obj, std::string **val) 
{
  char* buf = 0 ; size_t size = 0; int alloc = SWIG_OLDOBJ;
  if (SWIG_IsOK((SWIG_AsCharPtrAndSize(obj, &buf, &size, &alloc)))) {
    if (buf) {
      if (val) *val = new std::string(buf, size - 1);
      if (alloc == SWIG_NEWOBJ) delete[] buf;
      return SWIG_NEWOBJ;
    } else {
      if (val) *val = 0;
      return SWIG_OLDOBJ;
    }
  } else {
    PyErr_Clear();
    static swig_type_info *descriptor = SWIG_TypeQuery("std::string" " *");
    if (descriptor) {
      std::string *vptr;
      int res = SWIG_ConvertPtr(obj, (void**)&vptr, descriptor, 0);
      if (SWIG_IsOK(res) && val) *val = vptr;
      return res;
    }
  }
  return SWIG_ERROR;
}


SWIGINTERN int
SWIG_AsVal_std_string (PyObject * obj, std::string *val)
{
  std::string* v = (std::string *) 0;
  int res = SWIG_AsPtr_std_string (obj, &v);
  if (!SWIG_IsOK(res)) return res;
  if (v) {
    if (val) *val = *v;
    if (SWIG_IsNewObj(res)) {
      delete v;
      res = SWIG_DelNewMask(res);
    }
    return res;
  }
  return SWIG_ERROR;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > (size_t)PY_SSIZE_T_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj(const_cast< char * >(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#else
      return PyUnicode_DecodeUTF8(carray, static_cast< Py_ssize_t >(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject *
SWIG_From_std_string  (const std::string& s)
{
  return SWIG_FromCharPtrAndSize(s.data(), s.size());
}


namespace swig {
  template <> struct traits< std::string > {
    typedef value_category category;
    static const char* type_name() { return"std::string"; }
  };
  template <>  struct traits_asval< std::string > {
    typedef std::string value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_std_string (obj, val);
    }
  };
  template <>  struct traits_from< std::string > {
    typedef std::string value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_std_string  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< std::string, std::allocator< std::string > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::string" "," "std::allocator< std::string >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_string_Sg__iterator(std::vector< std::string > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_string_Sg____nonzero__(std::vector< std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_string_Sg____bool__(std::vector< std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::string >::size_type std_vector_Sl_std_string_Sg____len__(std::vector< std::string > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::string,std::allocator< std::string > > *std_vector_Sl_std_string_Sg____getslice__(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setslice____SWIG_0(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::string,std::allocator< std::string > >());
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setslice____SWIG_1(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j,std::vector< std::string,std::allocator< std::string > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delslice__(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delitem____SWIG_0(std::vector< std::string > *self,std::vector< std::string >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::string,std::allocator< std::string > > *std_vector_Sl_std_string_Sg____getitem____SWIG_0(std::vector< std::string > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_0(std::vector< std::string > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::string,std::allocator< std::string > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_1(std::vector< std::string > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delitem____SWIG_1(std::vector< std::string > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::string >::value_type const &std_vector_Sl_std_string_Sg____getitem____SWIG_1(std::vector< std::string > const *self,std::vector< std::string >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_2(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::vector3 > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::vector3"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::vector3, std::allocator< OpenBabel::vector3 > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::vector3" "," "std::allocator< OpenBabel::vector3 >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_vector3_Sg__iterator(std::vector< OpenBabel::vector3 > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_vector3_Sg____nonzero__(std::vector< OpenBabel::vector3 > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_vector3_Sg____bool__(std::vector< OpenBabel::vector3 > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::vector3 >::size_type std_vector_Sl_OpenBabel_vector3_Sg____len__(std::vector< OpenBabel::vector3 > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *std_vector_Sl_OpenBabel_vector3_Sg____getslice__(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i,std::vector< OpenBabel::vector3 >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____setslice____SWIG_0(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i,std::vector< OpenBabel::vector3 >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____setslice____SWIG_1(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i,std::vector< OpenBabel::vector3 >::difference_type j,std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____delslice__(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i,std::vector< OpenBabel::vector3 >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____delitem____SWIG_0(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *std_vector_Sl_OpenBabel_vector3_Sg____getitem____SWIG_0(std::vector< OpenBabel::vector3 > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type id = i;
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_0(std::vector< OpenBabel::vector3 > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type id = i;
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_1(std::vector< OpenBabel::vector3 > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type id = i;
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____delitem____SWIG_1(std::vector< OpenBabel::vector3 > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type id = i;
      std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::vector3 >::value_type const &std_vector_Sl_OpenBabel_vector3_Sg____getitem____SWIG_1(std::vector< OpenBabel::vector3 > const *self,std::vector< OpenBabel::vector3 >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_2(std::vector< OpenBabel::vector3 > *self,std::vector< OpenBabel::vector3 >::difference_type i,std::vector< OpenBabel::vector3 >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

      namespace swig {
	template <>  struct traits<std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >, std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >" "," "std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg__iterator(std::vector< std::vector< OpenBabel::vector3 > > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____nonzero__(std::vector< std::vector< OpenBabel::vector3 > > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____bool__(std::vector< std::vector< OpenBabel::vector3 > > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::vector< OpenBabel::vector3 > >::size_type std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____len__(std::vector< std::vector< OpenBabel::vector3 > > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getslice__(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i,std::vector< std::vector< OpenBabel::vector3 > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setslice____SWIG_0(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i,std::vector< std::vector< OpenBabel::vector3 > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >());
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setslice____SWIG_1(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i,std::vector< std::vector< OpenBabel::vector3 > >::difference_type j,std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delslice__(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i,std::vector< std::vector< OpenBabel::vector3 > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delitem____SWIG_0(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getitem____SWIG_0(std::vector< std::vector< OpenBabel::vector3 > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type id = i;
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_0(std::vector< std::vector< OpenBabel::vector3 > > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type id = i;
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_1(std::vector< std::vector< OpenBabel::vector3 > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type id = i;
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delitem____SWIG_1(std::vector< std::vector< OpenBabel::vector3 > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type id = i;
      std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::vector< OpenBabel::vector3 > >::value_type const &std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getitem____SWIG_1(std::vector< std::vector< OpenBabel::vector3 > > const *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_2(std::vector< std::vector< OpenBabel::vector3 > > *self,std::vector< std::vector< OpenBabel::vector3 > >::difference_type i,std::vector< std::vector< OpenBabel::vector3 > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBMol > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBMol"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBMol, std::allocator< OpenBabel::OBMol > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBMol" "," "std::allocator< OpenBabel::OBMol >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBMol_Sg__iterator(std::vector< OpenBabel::OBMol > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBMol_Sg____nonzero__(std::vector< OpenBabel::OBMol > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBMol_Sg____bool__(std::vector< OpenBabel::OBMol > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBMol >::size_type std_vector_Sl_OpenBabel_OBMol_Sg____len__(std::vector< OpenBabel::OBMol > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *std_vector_Sl_OpenBabel_OBMol_Sg____getslice__(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i,std::vector< OpenBabel::OBMol >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____setslice____SWIG_0(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i,std::vector< OpenBabel::OBMol >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____setslice____SWIG_1(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i,std::vector< OpenBabel::OBMol >::difference_type j,std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____delslice__(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i,std::vector< OpenBabel::OBMol >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____delitem____SWIG_0(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *std_vector_Sl_OpenBabel_OBMol_Sg____getitem____SWIG_0(std::vector< OpenBabel::OBMol > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type id = i;
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_0(std::vector< OpenBabel::OBMol > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type id = i;
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_1(std::vector< OpenBabel::OBMol > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type id = i;
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____delitem____SWIG_1(std::vector< OpenBabel::OBMol > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type id = i;
      std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBMol >::value_type const &std_vector_Sl_OpenBabel_OBMol_Sg____getitem____SWIG_1(std::vector< OpenBabel::OBMol > const *self,std::vector< OpenBabel::OBMol >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_2(std::vector< OpenBabel::OBMol > *self,std::vector< OpenBabel::OBMol >::difference_type i,std::vector< OpenBabel::OBMol >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBBond > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBBond"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBBond, std::allocator< OpenBabel::OBBond > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBBond" "," "std::allocator< OpenBabel::OBBond >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBBond_Sg__iterator(std::vector< OpenBabel::OBBond > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBBond_Sg____nonzero__(std::vector< OpenBabel::OBBond > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBBond_Sg____bool__(std::vector< OpenBabel::OBBond > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBBond >::size_type std_vector_Sl_OpenBabel_OBBond_Sg____len__(std::vector< OpenBabel::OBBond > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *std_vector_Sl_OpenBabel_OBBond_Sg____getslice__(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i,std::vector< OpenBabel::OBBond >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____setslice____SWIG_0(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i,std::vector< OpenBabel::OBBond >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____setslice____SWIG_1(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i,std::vector< OpenBabel::OBBond >::difference_type j,std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____delslice__(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i,std::vector< OpenBabel::OBBond >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____delitem____SWIG_0(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *std_vector_Sl_OpenBabel_OBBond_Sg____getitem____SWIG_0(std::vector< OpenBabel::OBBond > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type id = i;
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_0(std::vector< OpenBabel::OBBond > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type id = i;
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_1(std::vector< OpenBabel::OBBond > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type id = i;
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____delitem____SWIG_1(std::vector< OpenBabel::OBBond > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type id = i;
      std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBBond >::value_type const &std_vector_Sl_OpenBabel_OBBond_Sg____getitem____SWIG_1(std::vector< OpenBabel::OBBond > const *self,std::vector< OpenBabel::OBBond >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_2(std::vector< OpenBabel::OBBond > *self,std::vector< OpenBabel::OBBond >::difference_type i,std::vector< OpenBabel::OBBond >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBResidue > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBResidue"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBResidue, std::allocator< OpenBabel::OBResidue > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBResidue" "," "std::allocator< OpenBabel::OBResidue >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBResidue_Sg__iterator(std::vector< OpenBabel::OBResidue > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBResidue_Sg____nonzero__(std::vector< OpenBabel::OBResidue > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBResidue_Sg____bool__(std::vector< OpenBabel::OBResidue > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBResidue >::size_type std_vector_Sl_OpenBabel_OBResidue_Sg____len__(std::vector< OpenBabel::OBResidue > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *std_vector_Sl_OpenBabel_OBResidue_Sg____getslice__(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i,std::vector< OpenBabel::OBResidue >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____setslice____SWIG_0(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i,std::vector< OpenBabel::OBResidue >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____setslice____SWIG_1(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i,std::vector< OpenBabel::OBResidue >::difference_type j,std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____delslice__(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i,std::vector< OpenBabel::OBResidue >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____delitem____SWIG_0(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *std_vector_Sl_OpenBabel_OBResidue_Sg____getitem____SWIG_0(std::vector< OpenBabel::OBResidue > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type id = i;
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_0(std::vector< OpenBabel::OBResidue > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type id = i;
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_1(std::vector< OpenBabel::OBResidue > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type id = i;
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____delitem____SWIG_1(std::vector< OpenBabel::OBResidue > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type id = i;
      std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBResidue >::value_type const &std_vector_Sl_OpenBabel_OBResidue_Sg____getitem____SWIG_1(std::vector< OpenBabel::OBResidue > const *self,std::vector< OpenBabel::OBResidue >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_2(std::vector< OpenBabel::OBResidue > *self,std::vector< OpenBabel::OBResidue >::difference_type i,std::vector< OpenBabel::OBResidue >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBRing > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBRing"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBRing, std::allocator< OpenBabel::OBRing > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBRing" "," "std::allocator< OpenBabel::OBRing >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBRing_Sg__iterator(std::vector< OpenBabel::OBRing > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBRing_Sg____nonzero__(std::vector< OpenBabel::OBRing > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBRing_Sg____bool__(std::vector< OpenBabel::OBRing > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBRing >::size_type std_vector_Sl_OpenBabel_OBRing_Sg____len__(std::vector< OpenBabel::OBRing > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *std_vector_Sl_OpenBabel_OBRing_Sg____getslice__(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i,std::vector< OpenBabel::OBRing >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____setslice____SWIG_0(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i,std::vector< OpenBabel::OBRing >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____setslice____SWIG_1(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i,std::vector< OpenBabel::OBRing >::difference_type j,std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____delslice__(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i,std::vector< OpenBabel::OBRing >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____delitem____SWIG_0(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *std_vector_Sl_OpenBabel_OBRing_Sg____getitem____SWIG_0(std::vector< OpenBabel::OBRing > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type id = i;
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_0(std::vector< OpenBabel::OBRing > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type id = i;
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_1(std::vector< OpenBabel::OBRing > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type id = i;
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____delitem____SWIG_1(std::vector< OpenBabel::OBRing > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type id = i;
      std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBRing >::value_type const &std_vector_Sl_OpenBabel_OBRing_Sg____getitem____SWIG_1(std::vector< OpenBabel::OBRing > const *self,std::vector< OpenBabel::OBRing >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_2(std::vector< OpenBabel::OBRing > *self,std::vector< OpenBabel::OBRing >::difference_type i,std::vector< OpenBabel::OBRing >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBRing*, std::allocator< OpenBabel::OBRing * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBRing" " *," "std::allocator< OpenBabel::OBRing * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBRing_Sm__Sg__iterator(std::vector< OpenBabel::OBRing * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBRing_Sm__Sg____nonzero__(std::vector< OpenBabel::OBRing * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBRing_Sm__Sg____bool__(std::vector< OpenBabel::OBRing * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBRing * >::size_type std_vector_Sl_OpenBabel_OBRing_Sm__Sg____len__(std::vector< OpenBabel::OBRing * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getslice__(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i,std::vector< OpenBabel::OBRing * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setslice____SWIG_0(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i,std::vector< OpenBabel::OBRing * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setslice____SWIG_1(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i,std::vector< OpenBabel::OBRing * >::difference_type j,std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delslice__(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i,std::vector< OpenBabel::OBRing * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delitem____SWIG_0(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getitem____SWIG_0(std::vector< OpenBabel::OBRing * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type id = i;
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_0(std::vector< OpenBabel::OBRing * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type id = i;
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_1(std::vector< OpenBabel::OBRing * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type id = i;
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delitem____SWIG_1(std::vector< OpenBabel::OBRing * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type id = i;
      std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBRing * >::value_type std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getitem____SWIG_1(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_2(std::vector< OpenBabel::OBRing * > *self,std::vector< OpenBabel::OBRing * >::difference_type i,std::vector< OpenBabel::OBRing * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBGenericData > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBGenericData"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBGenericData*, std::allocator< OpenBabel::OBGenericData * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBGenericData" " *," "std::allocator< OpenBabel::OBGenericData * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg__iterator(std::vector< OpenBabel::OBGenericData * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____nonzero__(std::vector< OpenBabel::OBGenericData * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____bool__(std::vector< OpenBabel::OBGenericData * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBGenericData * >::size_type std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____len__(std::vector< OpenBabel::OBGenericData * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getslice__(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i,std::vector< OpenBabel::OBGenericData * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setslice____SWIG_0(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i,std::vector< OpenBabel::OBGenericData * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setslice____SWIG_1(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i,std::vector< OpenBabel::OBGenericData * >::difference_type j,std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delslice__(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i,std::vector< OpenBabel::OBGenericData * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delitem____SWIG_0(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getitem____SWIG_0(std::vector< OpenBabel::OBGenericData * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type id = i;
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_0(std::vector< OpenBabel::OBGenericData * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type id = i;
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_1(std::vector< OpenBabel::OBGenericData * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type id = i;
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delitem____SWIG_1(std::vector< OpenBabel::OBGenericData * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type id = i;
      std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBGenericData * >::value_type std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getitem____SWIG_1(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_2(std::vector< OpenBabel::OBGenericData * > *self,std::vector< OpenBabel::OBGenericData * >::difference_type i,std::vector< OpenBabel::OBGenericData * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {
    template <>  struct traits< OpenBabel::OBInternalCoord > {
      typedef pointer_category category;
      static const char* type_name() { return"OpenBabel::OBInternalCoord"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< OpenBabel::OBInternalCoord*, std::allocator< OpenBabel::OBInternalCoord * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "OpenBabel::OBInternalCoord" " *," "std::allocator< OpenBabel::OBInternalCoord * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg__iterator(std::vector< OpenBabel::OBInternalCoord * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____nonzero__(std::vector< OpenBabel::OBInternalCoord * > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____bool__(std::vector< OpenBabel::OBInternalCoord * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< OpenBabel::OBInternalCoord * >::size_type std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____len__(std::vector< OpenBabel::OBInternalCoord * > const *self){
      return self->size();
    }
SWIGINTERN std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getslice__(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i,std::vector< OpenBabel::OBInternalCoord * >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setslice____SWIG_0(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i,std::vector< OpenBabel::OBInternalCoord * >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >());
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setslice____SWIG_1(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i,std::vector< OpenBabel::OBInternalCoord * >::difference_type j,std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delslice__(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i,std::vector< OpenBabel::OBInternalCoord * >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delitem____SWIG_0(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getitem____SWIG_0(std::vector< OpenBabel::OBInternalCoord * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type id = i;
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_0(std::vector< OpenBabel::OBInternalCoord * > *self,SWIGPY_SLICEOBJECT *slice,std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type id = i;
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_1(std::vector< OpenBabel::OBInternalCoord * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type id = i;
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delitem____SWIG_1(std::vector< OpenBabel::OBInternalCoord * > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type id = i;
      std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< OpenBabel::OBInternalCoord * >::value_type std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getitem____SWIG_1(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_2(std::vector< OpenBabel::OBInternalCoord * > *self,std::vector< OpenBabel::OBInternalCoord * >::difference_type i,std::vector< OpenBabel::OBInternalCoord * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }

  namespace swig {


















































    template <class T, class U >
    struct traits_asptr<std::pair<T,U> >  {
      typedef std::pair<T,U> value_type;

      static int get_pair(PyObject* first, PyObject* second,
			  std::pair<T,U> **val) 
      {
	if (val) {
	  value_type *vp = (new std::pair<T,U>());
	  T *pfirst = &(vp->first);
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) {
	    delete vp;
	    return res1;
	  }
	  U *psecond = &(vp->second);
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) {
	    delete vp;
	    return res2;
	  }
	  *val = vp;
	  return SWIG_AddNewMask(res1 > res2 ? res1 : res2);
	} else {
	  T *pfirst = 0;
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) return res1;
	  U *psecond = 0;
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) return res2;
	  return res1 > res2 ? res1 : res2;
	}
      }

      static int asptr(PyObject *obj, std::pair<T,U> **val) {
	int res = SWIG_ERROR;
	if (PyTuple_Check(obj)) {
	  if (PyTuple_GET_SIZE(obj) == 2) {
	    res = get_pair(PyTuple_GET_ITEM(obj,0),PyTuple_GET_ITEM(obj,1), val);
	  }
	} else if (PySequence_Check(obj)) {
	  if (PySequence_Size(obj) == 2) {
	    swig::SwigVar_PyObject first = PySequence_GetItem(obj,0);
	    swig::SwigVar_PyObject second = PySequence_GetItem(obj,1);
	    res = get_pair(first, second, val);
	  }
	} else {
	  value_type *p = 0;
	  swig_type_info *descriptor = swig::type_info<value_type>();
	  res = descriptor ? SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0) : SWIG_ERROR;
	  if (SWIG_IsOK(res) && val)  *val = p;
	}
	return res;
      }
    };


    template <class T, class U >
    struct traits_from<std::pair<T,U> >   {
      static PyObject *from(const std::pair<T,U>& val) {
	PyObject* obj = PyTuple_New(2);
	PyTuple_SetItem(obj,0,swig::from(val.first));
	PyTuple_SetItem(obj,1,swig::from(val.second));
	return obj;
      }
    };
  }





































      namespace swig {
	template <>  struct traits<std::pair< unsigned int, unsigned int > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "unsigned int" "," "unsigned int" " >";
	  }
	};
      }
    

      namespace swig {
	template <>  struct traits<std::vector< std::pair< unsigned int,unsigned int >, std::allocator< std::pair< unsigned int,unsigned int > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::pair< unsigned int,unsigned int >" "," "std::allocator< std::pair< unsigned int,unsigned int > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__iterator(std::vector< std::pair< unsigned int,unsigned int > > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____nonzero__(std::vector< std::pair< unsigned int,unsigned int > > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____bool__(std::vector< std::pair< unsigned int,unsigned int > > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::pair< unsigned int,unsigned int > >::size_type std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____len__(std::vector< std::pair< unsigned int,unsigned int > > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getslice__(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i,std::vector< std::pair< unsigned int,unsigned int > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setslice____SWIG_0(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i,std::vector< std::pair< unsigned int,unsigned int > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >());
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setslice____SWIG_1(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i,std::vector< std::pair< unsigned int,unsigned int > >::difference_type j,std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delslice__(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i,std::vector< std::pair< unsigned int,unsigned int > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delitem____SWIG_0(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getitem____SWIG_0(std::vector< std::pair< unsigned int,unsigned int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type id = i;
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_0(std::vector< std::pair< unsigned int,unsigned int > > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type id = i;
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_1(std::vector< std::pair< unsigned int,unsigned int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type id = i;
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delitem____SWIG_1(std::vector< std::pair< unsigned int,unsigned int > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type id = i;
      std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::pair< unsigned int,unsigned int > >::value_type const &std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getitem____SWIG_1(std::vector< std::pair< unsigned int,unsigned int > > const *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_2(std::vector< std::pair< unsigned int,unsigned int > > *self,std::vector< std::pair< unsigned int,unsigned int > >::difference_type i,std::vector< std::pair< unsigned int,unsigned int > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }

      namespace swig {
	template <>  struct traits<std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >, std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >" "," "std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__iterator(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____nonzero__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____bool__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____len__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getslice__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setslice____SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >());
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setslice____SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type j,std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delslice__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delitem____SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getitem____SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type id = i;
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,SWIGPY_SLICEOBJECT *slice,std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type id = i;
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type id = i;
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delitem____SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,SWIGPY_SLICEOBJECT *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(slice, (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type id = i;
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getitem____SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_2(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type i,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__pop(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__append(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__erase__SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__erase__SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator first,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__insert__SWIG_0(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator pos,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__insert__SWIG_1(std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *self,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator pos,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type n,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &x){ self->insert(pos, n, x); }
 /* can't use macro -- AliasData not OBAliasData */
OpenBabel::AliasData *toAliasData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::AliasData*) data;
}


OpenBabel::OBAngleData *toAngleData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBAngleData *) data;
}


OpenBabel::OBCommentData *toCommentData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBCommentData *) data;
}


OpenBabel::OBConformerData *toConformerData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBConformerData *) data;
}


OpenBabel::OBExternalBondData *toExternalBondData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBExternalBondData *) data;
}


OpenBabel::OBGridData *toGridData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBGridData *) data;
}


OpenBabel::OBMatrixData *toMatrixData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBMatrixData *) data;
}


OpenBabel::OBNasaThermoData *toNasaThermoData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBNasaThermoData *) data;
}


OpenBabel::OBPairData *toPairData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBPairData *) data;
}


OpenBabel::OBPairInteger *toPairInteger(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBPairInteger *) data;
}


OpenBabel::OBPairFloatingPoint *toPairFloatingPoint(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBPairFloatingPoint *) data;
}


OpenBabel::OBRateData *toRateData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBRateData *) data;
}


OpenBabel::OBRotamerList *toRotamerList(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBRotamerList *) data;
}


OpenBabel::OBRotationData *toRotationData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBRotationData *) data;
}


OpenBabel::OBSerialNums *toSerialNums(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBSerialNums *) data;
}


OpenBabel::OBSetData *toSetData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBSetData *) data;
}


OpenBabel::OBSymmetryData *toSymmetryData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBSymmetryData *) data;
}


OpenBabel::OBTorsionData *toTorsionData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBTorsionData *) data;
}


OpenBabel::OBUnitCell *toUnitCell(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBUnitCell *) data;
}


OpenBabel::OBVectorData *toVectorData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBVectorData *) data;
}


OpenBabel::OBVibrationData *toVibrationData(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBVibrationData *) data;
}


OpenBabel::OBVirtualBond *toVirtualBond(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBVirtualBond *) data;
}


OpenBabel::OBStereoBase *toStereoBase(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBStereoBase *) data;
}


OpenBabel::OBTetrahedralStereo *toTetrahedralStereo(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBTetrahedralStereo *) data;
}


OpenBabel::OBCisTransStereo *toCisTransStereo(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBCisTransStereo *) data;
}


OpenBabel::OBSquarePlanarStereo *toSquarePlanarStereo(OpenBabel::OBGenericData *data) {
    return (OpenBabel::OBSquarePlanarStereo *) data;
}





SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  int r;
  if (!PyBool_Check(obj))
    return SWIG_ERROR;
  r = PyObject_IsTrue(obj);
  if (r == -1)
    return SWIG_ERROR;
  if (val) *val = r ? true : false;
  return SWIG_OK;
}


/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_float (PyObject * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< float >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}


SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	delete[] cptr;
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) delete[] cptr;
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (PyObject * obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = static_cast< char >(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}


SWIGINTERN int
SWIG_AsVal_short (PyObject * obj, short *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < SHRT_MIN || v > SHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< short >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_short  (unsigned short value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_char  (unsigned char value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_short (PyObject * obj, unsigned short *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > USHRT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned short >(v);
    }
  }  
  return res;
}

SWIGINTERN void OpenBabel_OBMol_SetTorsion__SWIG_1(OpenBabel::OBMol *self,int i,int j,int k,int l,double ang){
    self->SetTorsion(self->GetAtom(i), self->GetAtom(j),
                     self->GetAtom(k), self->GetAtom(l), ang);
  }

SWIGINTERN size_t
SWIG_strnlen(const char* s, size_t maxlen)
{
  const char *p;
  for (p = s; maxlen-- && *p; p++)
    ;
  return p - s;
}

SWIGINTERN void OpenBabel_OBForceField_SetLogToStdOut(OpenBabel::OBForceField *self){
    self->SetLogFile(&std::cout);
  }
SWIGINTERN void OpenBabel_OBForceField_SetLogToStdErr(OpenBabel::OBForceField *self){
    self->SetLogFile(&std::cerr);
  }

typedef double doubleArray;

SWIGINTERN doubleArray *new_doubleArray(size_t nelements){
    return (new double[nelements]());
  }
SWIGINTERN void delete_doubleArray(doubleArray *self){
    delete[] self;
  }
SWIGINTERN double doubleArray___getitem__(doubleArray *self,size_t index){
    return self[index];
  }
SWIGINTERN void doubleArray___setitem__(doubleArray *self,size_t index,double value){
    self[index] = value;
  }
SWIGINTERN double *doubleArray_cast(doubleArray *self){
    return self;
  }
SWIGINTERN doubleArray *doubleArray_frompointer(double *t){
    return static_cast< doubleArray * >(t);
  }
SWIGINTERN void OpenBabel_OBTetrahedralStereo_SetConfig(OpenBabel::OBTetrahedralStereo *self,OpenBabel::OBTetrahedralConfig const &config){
    self->SetConfig(OpenBabel::OBTetrahedralConfig::Convert(config));
  }
SWIGINTERN OpenBabel::OBTetrahedralConfig OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_0(OpenBabel::OBTetrahedralStereo *self,OpenBabel::OBStereo::Winding winding=OpenBabel::OBStereo::Clockwise,OpenBabel::OBStereo::View view=OpenBabel::OBStereo::ViewFrom){
    OpenBabel::OBTetrahedralStereo::Config cConfig = self->GetConfig(winding, view);
    
    OpenBabel::OBTetrahedralConfig pyConfig;
    pyConfig.center = cConfig.center;
    pyConfig.from_or_towards = cConfig.from;
    pyConfig.refs = cConfig.refs;
    pyConfig.winding = cConfig.winding;
    pyConfig.view = cConfig.view;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
SWIGINTERN OpenBabel::OBTetrahedralConfig OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_3(OpenBabel::OBTetrahedralStereo *self,unsigned long from_or_towards,OpenBabel::OBStereo::Winding winding=OpenBabel::OBStereo::Clockwise,OpenBabel::OBStereo::View view=OpenBabel::OBStereo::ViewFrom){
    OpenBabel::OBTetrahedralStereo::Config cConfig = self->GetConfig(from_or_towards, winding, view);
    
    OpenBabel::OBTetrahedralConfig pyConfig;
    pyConfig.center = cConfig.center;
    pyConfig.from_or_towards = cConfig.from;
    pyConfig.refs = cConfig.refs;
    pyConfig.winding = cConfig.winding;
    pyConfig.view = cConfig.view;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
SWIGINTERN void OpenBabel_OBCisTransStereo_SetConfig(OpenBabel::OBCisTransStereo *self,OpenBabel::OBCisTransConfig const &config){
    self->SetConfig(OpenBabel::OBCisTransConfig::Convert(config));
  }
SWIGINTERN OpenBabel::OBCisTransConfig OpenBabel_OBCisTransStereo_GetConfig__SWIG_0(OpenBabel::OBCisTransStereo *self,OpenBabel::OBStereo::Shape shape=OpenBabel::OBStereo::ShapeU){
    OpenBabel::OBCisTransStereo::Config cConfig = self->GetConfig(shape);
  
    OpenBabel::OBCisTransConfig pyConfig;
    pyConfig.begin = cConfig.begin;
    pyConfig.end = cConfig.end;
    pyConfig.refs = cConfig.refs;
    pyConfig.shape = cConfig.shape;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
SWIGINTERN OpenBabel::OBCisTransConfig OpenBabel_OBCisTransStereo_GetConfig__SWIG_2(OpenBabel::OBCisTransStereo *self,unsigned long start,OpenBabel::OBStereo::Shape shape=OpenBabel::OBStereo::ShapeU){
    OpenBabel::OBCisTransStereo::Config cConfig = self->GetConfig(start, shape);
  
    OpenBabel::OBCisTransConfig pyConfig;
    pyConfig.begin = cConfig.begin;
    pyConfig.end = cConfig.end;
    pyConfig.refs = cConfig.refs;
    pyConfig.shape = cConfig.shape;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
SWIGINTERN void OpenBabel_OBSquarePlanarStereo_SetConfig(OpenBabel::OBSquarePlanarStereo *self,OpenBabel::OBSquarePlanarConfig const &config){
    self->SetConfig(OpenBabel::OBSquarePlanarConfig::Convert(config));
  }
SWIGINTERN OpenBabel::OBSquarePlanarConfig OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_0(OpenBabel::OBSquarePlanarStereo *self,OpenBabel::OBStereo::Shape shape=OpenBabel::OBStereo::ShapeU){
    OpenBabel::OBSquarePlanarStereo::Config cConfig = self->GetConfig(shape);
  
    OpenBabel::OBSquarePlanarConfig pyConfig;
    pyConfig.center = cConfig.center;
    pyConfig.refs = cConfig.refs;
    pyConfig.shape = cConfig.shape;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
SWIGINTERN OpenBabel::OBSquarePlanarConfig OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_2(OpenBabel::OBSquarePlanarStereo *self,unsigned long start,OpenBabel::OBStereo::Shape shape=OpenBabel::OBStereo::ShapeU){
    OpenBabel::OBSquarePlanarStereo::Config cConfig = self->GetConfig(start, shape);
  
    OpenBabel::OBSquarePlanarConfig pyConfig;
    pyConfig.center = cConfig.center;
    pyConfig.refs = cConfig.refs;
    pyConfig.shape = cConfig.shape;
    pyConfig.specified = cConfig.specified;

    return pyConfig;
  }
#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_delete_SwigPyIterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SwigPyIterator" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_value(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; PyObject *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_value" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (PyObject *)((swig::SwigPyIterator const *)arg1)->value();
  resultobj = result; if (!resultobj) SWIG_fail; return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_incr__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; size_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; size_t val2 ;
  int ecode2 = 0 ; swig::SwigPyIterator *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_incr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_incr" "', argument " "2"" of type '" "size_t""'"); }
    arg2 = static_cast< size_t >(val2); result = (swig::SwigPyIterator *)(arg1)->incr(SWIG_STD_MOVE(arg2));
  if (!result) SWIG_fail; resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_incr__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; swig::SwigPyIterator *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_incr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (swig::SwigPyIterator *)(arg1)->incr();
  if (!result) SWIG_fail; resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_incr(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator_incr", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_SwigPyIterator_incr__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_SwigPyIterator_incr__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SwigPyIterator_incr'.\n"
  "  Possible C/C++ prototypes are:\n" "    swig::SwigPyIterator::incr(size_t)\n" "    swig::SwigPyIterator::incr()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_decr__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; size_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; size_t val2 ;
  int ecode2 = 0 ; swig::SwigPyIterator *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_decr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_decr" "', argument " "2"" of type '" "size_t""'"); }
    arg2 = static_cast< size_t >(val2); result = (swig::SwigPyIterator *)(arg1)->decr(SWIG_STD_MOVE(arg2));
  if (!result) SWIG_fail; resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_decr__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; swig::SwigPyIterator *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_decr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (swig::SwigPyIterator *)(arg1)->decr();
  if (!result) SWIG_fail; resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_decr(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator_decr", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_SwigPyIterator_decr__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_SwigPyIterator_decr__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SwigPyIterator_decr'.\n"
  "  Possible C/C++ prototypes are:\n" "    swig::SwigPyIterator::decr(size_t)\n" "    swig::SwigPyIterator::decr()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_distance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; swig::SwigPyIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; ptrdiff_t result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_distance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_distance" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); }
   arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2); try {
    result = ((swig::SwigPyIterator const *)arg1)->distance((swig::SwigPyIterator const &)*arg2);}
   catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail; }
   resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_equal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; swig::SwigPyIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_equal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_equal" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); }
   arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2); try {
    result = (bool)((swig::SwigPyIterator const *)arg1)->equal((swig::SwigPyIterator const &)*arg2);}
   catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail; }
   resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_copy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_copy" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->copy(); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_next(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; PyObject *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_next" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (PyObject *)(arg1)->next(); resultobj = result;
  if (!resultobj) SWIG_fail; return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___next__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; PyObject *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___next__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (PyObject *)(arg1)->__next__(); resultobj = result;
  if (!resultobj) SWIG_fail; return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_previous(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; PyObject *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_previous" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); result = (PyObject *)(arg1)->previous(); resultobj = result;
  if (!resultobj) SWIG_fail; return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator_advance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; ptrdiff_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; swig::SwigPyIterator *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator_advance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_advance" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_advance" "', argument " "2"" of type '" "ptrdiff_t""'"); }
    arg2 = static_cast< ptrdiff_t >(val2); result = (swig::SwigPyIterator *)(arg1)->advance(SWIG_STD_MOVE(arg2));
  if (!result) SWIG_fail; resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; swig::SwigPyIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___eq__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); }
   arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator ==((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; swig::SwigPyIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___ne__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); }
   arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator !=((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___iadd__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; ptrdiff_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; swig::SwigPyIterator *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___iadd__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___iadd__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___iadd__" "', argument " "2"" of type '" "ptrdiff_t""'"); }
    arg2 = static_cast< ptrdiff_t >(val2); result = (swig::SwigPyIterator *) &(arg1)->operator +=(SWIG_STD_MOVE(arg2));
  if (PyErr_Occurred()) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___isub__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; ptrdiff_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; swig::SwigPyIterator *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___isub__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___isub__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___isub__" "', argument " "2"" of type '" "ptrdiff_t""'"); }
    arg2 = static_cast< ptrdiff_t >(val2); result = (swig::SwigPyIterator *) &(arg1)->operator -=(SWIG_STD_MOVE(arg2));
  if (PyErr_Occurred()) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___add__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = 0 ; ptrdiff_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; swig::SwigPyIterator *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SwigPyIterator___add__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___add__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___add__" "', argument " "2"" of type '" "ptrdiff_t""'"); }
    arg2 = static_cast< ptrdiff_t >(val2);
  result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator +(SWIG_STD_MOVE(arg2));
  if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; ptrdiff_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; swig::SwigPyIterator *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "ptrdiff_t""'"); }
    arg2 = static_cast< ptrdiff_t >(val2);
  result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator -(SWIG_STD_MOVE(arg2));
  if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; swig::SwigPyIterator *arg1 = 0 ; swig::SwigPyIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; ptrdiff_t result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'");  }
   arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); }
   arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = ((swig::SwigPyIterator const *)arg1)->operator -((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SwigPyIterator___sub__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator___sub__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_SwigPyIterator___sub____SWIG_1(self, argc, argv);} 
  check_1: if (argc == 2) { PyObject *retobj = _wrap_SwigPyIterator___sub____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail: SWIG_Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented; }
SWIGINTERN PyObject *SwigPyIterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_swig__SwigPyIterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vectorInt_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt_iterator" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_int_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___nonzero__" "', argument " "1"" of type '" "std::vector< int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1);
  result = (bool)std_vector_Sl_int_Sg____nonzero__((std::vector< int > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___bool__" "', argument " "1"" of type '" "std::vector< int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1);
  result = (bool)std_vector_Sl_int_Sg____bool__((std::vector< int > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::vector< int >::size_type result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___len__" "', argument " "1"" of type '" "std::vector< int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1);
  result = std_vector_Sl_int_Sg____len__((std::vector< int > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ; std::vector< int >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< int,std::allocator< int > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorInt___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___getslice__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___getslice__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorInt___getslice__" "', argument " "3"" of type '" "std::vector< int >::difference_type""'"); }
    arg3 = static_cast< std::vector< int >::difference_type >(val3); try {
    result = (std::vector< int,std::allocator< int > > *)std_vector_Sl_int_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ;
  std::vector< int >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ;
  int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___setslice__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___setslice__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorInt___setslice__" "', argument " "3"" of type '" "std::vector< int >::difference_type""'"); }
    arg3 = static_cast< std::vector< int >::difference_type >(val3); try {
    std_vector_Sl_int_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ;
  std::vector< int >::difference_type arg3 ; std::vector< int,std::allocator< int > > *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___setslice__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___setslice__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorInt___setslice__" "', argument " "3"" of type '" "std::vector< int >::difference_type""'"); }
    arg3 = static_cast< std::vector< int >::difference_type >(val3); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorInt___setslice__" "', argument " "4"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorInt___setslice__" "', argument " "4"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_int_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< int,std::allocator< int > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorInt___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorInt___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorInt___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorInt___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< int >::__setslice__(std::vector< int >::difference_type,std::vector< int >::difference_type)\n"
  "    std::vector< int >::__setslice__(std::vector< int >::difference_type,std::vector< int >::difference_type,std::vector< int,std::allocator< int > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorInt___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ; std::vector< int >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vectorInt___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___delslice__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___delslice__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorInt___delslice__" "', argument " "3"" of type '" "std::vector< int >::difference_type""'"); }
    arg3 = static_cast< std::vector< int >::difference_type >(val3); try {
    std_vector_Sl_int_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___delitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___delitem__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); try {
    std_vector_Sl_int_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< int,std::allocator< int > > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___getitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorInt___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< int,std::allocator< int > > *)std_vector_Sl_int_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< int,std::allocator< int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___setitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorInt___setitem__" "', argument " "3"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorInt___setitem__" "', argument " "3"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_int_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< int,std::allocator< int > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___setitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_int_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___delitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorInt___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_int_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorInt___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorInt___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorInt___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorInt___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< int >::__delitem__(std::vector< int >::difference_type)\n"
  "    std::vector< int >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorInt___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< int >::value_type *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___getitem__" "', argument " "1"" of type '" "std::vector< int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___getitem__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); try {
    result = (std::vector< int >::value_type *) &std_vector_Sl_int_Sg____getitem____SWIG_1((std::vector< int > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_From_int(static_cast< int >(*result));
  if (!swig::container_owner<swig::traits<std::vector< int >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorInt___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorInt___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorInt___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorInt___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< int >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< int >::__getitem__(std::vector< int >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorInt___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; std::vector< int >::difference_type arg2 ;
  std::vector< int >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< int >::value_type temp3 ; int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt___setitem__" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorInt___setitem__" "', argument " "2"" of type '" "std::vector< int >::difference_type""'"); }
    arg2 = static_cast< std::vector< int >::difference_type >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorInt___setitem__" "', argument " "3"" of type '" "std::vector< int >::value_type""'"); }
    temp3 = static_cast< std::vector< int >::value_type >(val3); arg3 = &temp3; try {
    std_vector_Sl_int_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(int const &)*arg3);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorInt___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorInt___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< int,std::allocator< int > >**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_vectorInt___setitem____SWIG_0(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_vectorInt___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorInt___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< int >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< int,std::allocator< int > > const &)\n"
  "    std::vector< int >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< int >::__setitem__(std::vector< int >::difference_type,std::vector< int >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< int > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< int > *)new std::vector< int >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; std::vector< int > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorInt" "', argument " "1"" of type '" "std::vector< int > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorInt" "', argument " "1"" of type '" "std::vector< int > const &""'"); }
     arg1 = ptr; }  result = (std::vector< int > *)new std::vector< int >((std::vector< int > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorInt_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorInt_clear" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ; std::vector< int > *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorInt" "', argument " "1"" of type '" "std::vector< int >::size_type""'"); }
    arg1 = static_cast< std::vector< int >::size_type >(val1);
  result = (std::vector< int > *)new std::vector< int >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorInt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< int >::size_type arg1 ; std::vector< int >::value_type *arg2 = 0 ; size_t val1 ;
  int ecode1 = 0 ; std::vector< int >::value_type temp2 ; int val2 ; int ecode2 = 0 ; std::vector< int > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorInt" "', argument " "1"" of type '" "std::vector< int >::size_type""'"); }
    arg1 = static_cast< std::vector< int >::size_type >(val1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorInt" "', argument " "2"" of type '" "std::vector< int >::value_type""'"); }
    temp2 = static_cast< std::vector< int >::value_type >(val2); arg2 = &temp2;
  result = (std::vector< int > *)new std::vector< int >(SWIG_STD_MOVE(arg1),(std::vector< int >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorInt__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { int _v = 0; { { int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } } 
    if (!_v) goto check_2; return _wrap_new_vectorInt__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorInt__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_new_vectorInt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< int >::vector()\n"
  "    std::vector< int >::vector(std::vector< int > const &)\n"
  "    std::vector< int >::vector(std::vector< int >::size_type)\n"
  "    std::vector< int >::vector(std::vector< int >::size_type,std::vector< int >::value_type const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorInt" "', argument " "1"" of type '" "std::vector< int > *""'");  }
   arg1 = reinterpret_cast< std::vector< int > * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *vectorInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_int_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorUnsignedInt_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt_iterator" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_unsigned_SS_int_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___nonzero__" "', argument " "1"" of type '" "std::vector< unsigned int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1);
  result = (bool)std_vector_Sl_unsigned_SS_int_Sg____nonzero__((std::vector< unsigned int > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___bool__" "', argument " "1"" of type '" "std::vector< unsigned int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1);
  result = (bool)std_vector_Sl_unsigned_SS_int_Sg____bool__((std::vector< unsigned int > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< unsigned int >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___len__" "', argument " "1"" of type '" "std::vector< unsigned int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1);
  result = std_vector_Sl_unsigned_SS_int_Sg____len__((std::vector< unsigned int > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  std::vector< unsigned int >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< unsigned int,std::allocator< unsigned int > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___getslice__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___getslice__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorUnsignedInt___getslice__" "', argument " "3"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned int >::difference_type >(val3); try {
    result = (std::vector< unsigned int,std::allocator< unsigned int > > *)std_vector_Sl_unsigned_SS_int_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  std::vector< unsigned int >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___setslice__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___setslice__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorUnsignedInt___setslice__" "', argument " "3"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned int >::difference_type >(val3); try {
    std_vector_Sl_unsigned_SS_int_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  std::vector< unsigned int >::difference_type arg3 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___setslice__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___setslice__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorUnsignedInt___setslice__" "', argument " "3"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned int >::difference_type >(val3); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorUnsignedInt___setslice__" "', argument " "4"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorUnsignedInt___setslice__" "', argument " "4"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_unsigned_SS_int_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorUnsignedInt___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorUnsignedInt___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorUnsignedInt___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned int >::__setslice__(std::vector< unsigned int >::difference_type,std::vector< unsigned int >::difference_type)\n"
  "    std::vector< unsigned int >::__setslice__(std::vector< unsigned int >::difference_type,std::vector< unsigned int >::difference_type,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  std::vector< unsigned int >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___delslice__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___delslice__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorUnsignedInt___delslice__" "', argument " "3"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned int >::difference_type >(val3); try {
    std_vector_Sl_unsigned_SS_int_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___delitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___delitem__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); try {
    std_vector_Sl_unsigned_SS_int_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___getitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorUnsignedInt___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< unsigned int,std::allocator< unsigned int > > *)std_vector_Sl_unsigned_SS_int_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___setitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorUnsignedInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorUnsignedInt___setitem__" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorUnsignedInt___setitem__" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___setitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorUnsignedInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___delitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorUnsignedInt___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_unsigned_SS_int_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorUnsignedInt___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorUnsignedInt___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorUnsignedInt___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned int >::__delitem__(std::vector< unsigned int >::difference_type)\n"
  "    std::vector< unsigned int >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< unsigned int >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___getitem__" "', argument " "1"" of type '" "std::vector< unsigned int > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___getitem__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2); try {
    result = (std::vector< unsigned int >::value_type *) &std_vector_Sl_unsigned_SS_int_Sg____getitem____SWIG_1((std::vector< unsigned int > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(*result));
  if (!swig::container_owner<swig::traits<std::vector< unsigned int >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorUnsignedInt___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorUnsignedInt___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorUnsignedInt___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< unsigned int >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< unsigned int >::__getitem__(std::vector< unsigned int >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; std::vector< unsigned int >::difference_type arg2 ;
  std::vector< unsigned int >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< unsigned int >::value_type temp3 ; unsigned int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt___setitem__" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorUnsignedInt___setitem__" "', argument " "2"" of type '" "std::vector< unsigned int >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorUnsignedInt___setitem__" "', argument " "3"" of type '" "std::vector< unsigned int >::value_type""'"); }
    temp3 = static_cast< std::vector< unsigned int >::value_type >(val3); arg3 = &temp3; try {
    std_vector_Sl_unsigned_SS_int_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(unsigned int const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorUnsignedInt___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorUnsignedInt___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< unsigned int,std::allocator< unsigned int > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorUnsignedInt___setitem____SWIG_0(self, argc, argv);}
   check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorUnsignedInt___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorUnsignedInt___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned int >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n"
  "    std::vector< unsigned int >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< unsigned int >::__setitem__(std::vector< unsigned int >::difference_type,std::vector< unsigned int >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorUnsignedInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< unsigned int > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< unsigned int > *)new std::vector< unsigned int >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorUnsignedInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< unsigned int > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorUnsignedInt" "', argument " "1"" of type '" "std::vector< unsigned int > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorUnsignedInt" "', argument " "1"" of type '" "std::vector< unsigned int > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< unsigned int > *)new std::vector< unsigned int >((std::vector< unsigned int > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorUnsignedInt_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorUnsignedInt_clear" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorUnsignedInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< unsigned int > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorUnsignedInt" "', argument " "1"" of type '" "std::vector< unsigned int >::size_type""'"); }
    arg1 = static_cast< std::vector< unsigned int >::size_type >(val1);
  result = (std::vector< unsigned int > *)new std::vector< unsigned int >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorUnsignedInt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int >::size_type arg1 ; std::vector< unsigned int >::value_type *arg2 = 0 ;
  size_t val1 ; int ecode1 = 0 ; std::vector< unsigned int >::value_type temp2 ; unsigned int val2 ; int ecode2 = 0 ;
  std::vector< unsigned int > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorUnsignedInt" "', argument " "1"" of type '" "std::vector< unsigned int >::size_type""'"); }
    arg1 = static_cast< std::vector< unsigned int >::size_type >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorUnsignedInt" "', argument " "2"" of type '" "std::vector< unsigned int >::value_type""'"); }
    temp2 = static_cast< std::vector< unsigned int >::value_type >(val2); arg2 = &temp2;
  result = (std::vector< unsigned int > *)new std::vector< unsigned int >(SWIG_STD_MOVE(arg1),(std::vector< unsigned int >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorUnsignedInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorUnsignedInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorUnsignedInt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorUnsignedInt__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorUnsignedInt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorUnsignedInt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorUnsignedInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< unsigned int >::vector()\n"
  "    std::vector< unsigned int >::vector(std::vector< unsigned int > const &)\n"
  "    std::vector< unsigned int >::vector(std::vector< unsigned int >::size_type)\n"
  "    std::vector< unsigned int >::vector(std::vector< unsigned int >::size_type,std::vector< unsigned int >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorUnsignedInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorUnsignedInt" "', argument " "1"" of type '" "std::vector< unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned int > * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *vectorUnsignedInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_unsigned_int_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorUnsignedInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorvInt_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt_iterator" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_vector_Sl_int_Sg__Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___nonzero__" "', argument " "1"" of type '" "std::vector< std::vector< int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_int_Sg__Sg____nonzero__((std::vector< std::vector< int > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___bool__" "', argument " "1"" of type '" "std::vector< std::vector< int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_int_Sg__Sg____bool__((std::vector< std::vector< int > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< int > >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___len__" "', argument " "1"" of type '" "std::vector< std::vector< int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1);
  result = std_vector_Sl_std_vector_Sl_int_Sg__Sg____len__((std::vector< std::vector< int > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; std::vector< std::vector< int > >::difference_type arg2 ;
  std::vector< std::vector< int > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vectorvInt___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___getslice__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___getslice__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvInt___getslice__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< int > >::difference_type >(val3); try {
    result = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)std_vector_Sl_std_vector_Sl_int_Sg__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ;
  std::vector< std::vector< int > >::difference_type arg2 ; std::vector< std::vector< int > >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvInt___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< int > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ;
  std::vector< std::vector< int > >::difference_type arg2 ; std::vector< std::vector< int > >::difference_type arg3 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvInt___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< int > >::difference_type >(val3); {
    std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *ptr = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorvInt___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvInt___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvInt___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorvInt___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorvInt___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvInt___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< int > >::__setslice__(std::vector< std::vector< int > >::difference_type,std::vector< std::vector< int > >::difference_type)\n"
  "    std::vector< std::vector< int > >::__setslice__(std::vector< std::vector< int > >::difference_type,std::vector< std::vector< int > >::difference_type,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorvInt___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; std::vector< std::vector< int > >::difference_type arg2 ;
  std::vector< std::vector< int > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorvInt___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___delslice__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___delslice__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvInt___delslice__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< int > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ;
  std::vector< std::vector< int > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___delitem__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2); try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvInt___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)std_vector_Sl_std_vector_Sl_int_Sg__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *ptr = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorvInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvInt___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_std_vector_Sl_int_Sg__Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvInt___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorvInt___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorvInt___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvInt___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< int > >::__delitem__(std::vector< std::vector< int > >::difference_type)\n"
  "    std::vector< std::vector< int > >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorvInt___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ;
  std::vector< std::vector< int > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< std::vector< int > >::value_type *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___getitem__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2); try {
    result = (std::vector< std::vector< int > >::value_type *) &std_vector_Sl_std_vector_Sl_int_Sg__Sg____getitem____SWIG_1((std::vector< std::vector< int > > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = swig::from(static_cast< std::vector< int,std::allocator< int > > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::vector< int > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvInt___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorvInt___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorvInt___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvInt___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::vector< int > >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< int > >::__getitem__(std::vector< std::vector< int > >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorvInt___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *arg1 = 0 ;
  std::vector< std::vector< int > >::difference_type arg2 ; std::vector< std::vector< int > >::value_type *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvInt___setitem__" "', argument " "2"" of type '" "std::vector< std::vector< int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< int > >::difference_type >(val2); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorvInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< int > >::value_type const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_int_Sg__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::vector< int,std::allocator< int > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvInt___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorvInt___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorvInt___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorvInt___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvInt___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< int > >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > const &)\n"
  "    std::vector< std::vector< int > >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< int > >::__setitem__(std::vector< std::vector< int > >::difference_type,std::vector< std::vector< int > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorvInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< std::vector< int > > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::vector< int > > *)new std::vector< std::vector< int > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int,std::allocator< int > > > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< std::vector< int > > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *ptr = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorvInt" "', argument " "1"" of type '" "std::vector< std::vector< int,std::allocator< int > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorvInt" "', argument " "1"" of type '" "std::vector< std::vector< int,std::allocator< int > > > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< std::vector< int > > *)new std::vector< std::vector< int > >((std::vector< std::vector< int,std::allocator< int > > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvInt_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvInt_clear" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< std::vector< int > > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorvInt" "', argument " "1"" of type '" "std::vector< std::vector< int > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< int > >::size_type >(val1);
  result = (std::vector< std::vector< int > > *)new std::vector< std::vector< int > >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvInt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< int > >::size_type arg1 ;
  std::vector< std::vector< int > >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; int res2 = SWIG_OLDOBJ ;
  std::vector< std::vector< int > > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorvInt" "', argument " "1"" of type '" "std::vector< std::vector< int > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< int > >::size_type >(val1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorvInt" "', argument " "2"" of type '" "std::vector< std::vector< int > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorvInt" "', argument " "2"" of type '" "std::vector< std::vector< int > >::value_type const &""'"); }
     arg2 = ptr; } 
  result = (std::vector< std::vector< int > > *)new std::vector< std::vector< int > >(SWIG_STD_MOVE(arg1),(std::vector< std::vector< int > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorvInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorvInt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorvInt__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorvInt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorvInt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorvInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::vector< int > >::vector()\n"
  "    std::vector< std::vector< int > >::vector(std::vector< std::vector< int,std::allocator< int > > > const &)\n"
  "    std::vector< std::vector< int > >::vector(std::vector< std::vector< int > >::size_type)\n"
  "    std::vector< std::vector< int > >::vector(std::vector< std::vector< int > >::size_type,std::vector< std::vector< int > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorvInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorvInt" "', argument " "1"" of type '" "std::vector< std::vector< int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< int > > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorvInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorvInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorDouble_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_iterator" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_double_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___nonzero__" "', argument " "1"" of type '" "std::vector< double > const *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (bool)std_vector_Sl_double_Sg____nonzero__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___bool__" "', argument " "1"" of type '" "std::vector< double > const *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (bool)std_vector_Sl_double_Sg____bool__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___len__" "', argument " "1"" of type '" "std::vector< double > const *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = std_vector_Sl_double_Sg____len__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ; std::vector< double >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorDouble___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getslice__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___getslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___getslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'"); }
    arg3 = static_cast< std::vector< double >::difference_type >(val3); try {
    result = (std::vector< double,std::allocator< double > > *)std_vector_Sl_double_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setslice__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'"); }
    arg3 = static_cast< std::vector< double >::difference_type >(val3); try {
    std_vector_Sl_double_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ; std::vector< double,std::allocator< double > > *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setslice__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'"); }
    arg3 = static_cast< std::vector< double >::difference_type >(val3); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorDouble___setslice__" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorDouble___setslice__" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_double_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< double,std::allocator< double > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorDouble___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorDouble___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< double >::__setslice__(std::vector< double >::difference_type,std::vector< double >::difference_type)\n"
  "    std::vector< double >::__setslice__(std::vector< double >::difference_type,std::vector< double >::difference_type,std::vector< double,std::allocator< double > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorDouble___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ; std::vector< double >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vectorDouble___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delslice__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___delslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___delslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'"); }
    arg3 = static_cast< std::vector< double >::difference_type >(val3); try {
    std_vector_Sl_double_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___delitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); try {
    std_vector_Sl_double_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< double,std::allocator< double > > *)std_vector_Sl_double_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< double,std::allocator< double > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_double_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< double,std::allocator< double > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_double_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_double_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorDouble___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorDouble___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< double >::__delitem__(std::vector< double >::difference_type)\n"
  "    std::vector< double >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorDouble___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< double >::value_type *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getitem__" "', argument " "1"" of type '" "std::vector< double > const *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___getitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); try {
    result = (std::vector< double >::value_type *) &std_vector_Sl_double_Sg____getitem____SWIG_1((std::vector< double > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_From_double(static_cast< double >(*result));
  if (!swig::container_owner<swig::traits<std::vector< double >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorDouble___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorDouble___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< double >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< double >::__getitem__(std::vector< double >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; std::vector< double >::difference_type arg2 ;
  std::vector< double >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< double >::value_type temp3 ; double val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'"); }
    arg2 = static_cast< std::vector< double >::difference_type >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double >::value_type""'"); }
    temp3 = static_cast< std::vector< double >::value_type >(val3); arg3 = &temp3; try {
    std_vector_Sl_double_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(double const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorDouble___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< double,std::allocator< double > >**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_vectorDouble___setitem____SWIG_0(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_vectorDouble___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< double >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< double,std::allocator< double > > const &)\n"
  "    std::vector< double >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< double >::__setitem__(std::vector< double >::difference_type,std::vector< double >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< double > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< double > *)new std::vector< double >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; std::vector< double > *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double > const &""'"); }
     arg1 = ptr; }  result = (std::vector< double > *)new std::vector< double >((std::vector< double > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorDouble_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_clear" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< double > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double >::size_type""'"); }
    arg1 = static_cast< std::vector< double >::size_type >(val1);
  result = (std::vector< double > *)new std::vector< double >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< double >::size_type arg1 ; std::vector< double >::value_type *arg2 = 0 ; size_t val1 ;
  int ecode1 = 0 ; std::vector< double >::value_type temp2 ; double val2 ; int ecode2 = 0 ; std::vector< double > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double >::size_type""'"); }
    arg1 = static_cast< std::vector< double >::size_type >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorDouble" "', argument " "2"" of type '" "std::vector< double >::value_type""'"); }
    temp2 = static_cast< std::vector< double >::value_type >(val2); arg2 = &temp2;
  result = (std::vector< double > *)new std::vector< double >(SWIG_STD_MOVE(arg1),(std::vector< double >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorDouble(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorDouble", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorDouble__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorDouble__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorDouble__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorDouble__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorDouble'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< double >::vector()\n"
  "    std::vector< double >::vector(std::vector< double > const &)\n"
  "    std::vector< double >::vector(std::vector< double >::size_type)\n"
  "    std::vector< double >::vector(std::vector< double >::size_type,std::vector< double >::value_type const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorDouble(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorDouble" "', argument " "1"" of type '" "std::vector< double > *""'");  }
   arg1 = reinterpret_cast< std::vector< double > * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *vectorDouble_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorDouble_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorULong_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong_iterator" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_unsigned_SS_long_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___nonzero__" "', argument " "1"" of type '" "std::vector< unsigned long > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1);
  result = (bool)std_vector_Sl_unsigned_SS_long_Sg____nonzero__((std::vector< unsigned long > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___bool__" "', argument " "1"" of type '" "std::vector< unsigned long > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1);
  result = (bool)std_vector_Sl_unsigned_SS_long_Sg____bool__((std::vector< unsigned long > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< unsigned long >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___len__" "', argument " "1"" of type '" "std::vector< unsigned long > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1);
  result = std_vector_Sl_unsigned_SS_long_Sg____len__((std::vector< unsigned long > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  std::vector< unsigned long >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< unsigned long,std::allocator< unsigned long > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorULong___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___getslice__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___getslice__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorULong___getslice__" "', argument " "3"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned long >::difference_type >(val3); try {
    result = (std::vector< unsigned long,std::allocator< unsigned long > > *)std_vector_Sl_unsigned_SS_long_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  std::vector< unsigned long >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___setslice__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___setslice__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorULong___setslice__" "', argument " "3"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned long >::difference_type >(val3); try {
    std_vector_Sl_unsigned_SS_long_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  std::vector< unsigned long >::difference_type arg3 ; std::vector< unsigned long,std::allocator< unsigned long > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___setslice__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___setslice__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorULong___setslice__" "', argument " "3"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned long >::difference_type >(val3); {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorULong___setslice__" "', argument " "4"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorULong___setslice__" "', argument " "4"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_unsigned_SS_long_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorULong___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorULong___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorULong___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorULong___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned long >::__setslice__(std::vector< unsigned long >::difference_type,std::vector< unsigned long >::difference_type)\n"
  "    std::vector< unsigned long >::__setslice__(std::vector< unsigned long >::difference_type,std::vector< unsigned long >::difference_type,std::vector< unsigned long,std::allocator< unsigned long > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorULong___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  std::vector< unsigned long >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorULong___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___delslice__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___delslice__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorULong___delslice__" "', argument " "3"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg3 = static_cast< std::vector< unsigned long >::difference_type >(val3); try {
    std_vector_Sl_unsigned_SS_long_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___delitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___delitem__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2); try {
    std_vector_Sl_unsigned_SS_long_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< unsigned long,std::allocator< unsigned long > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___getitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorULong___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< unsigned long,std::allocator< unsigned long > > *)std_vector_Sl_unsigned_SS_long_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< unsigned long,std::allocator< unsigned long > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___setitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorULong___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorULong___setitem__" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorULong___setitem__" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___setitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorULong___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___delitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorULong___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_unsigned_SS_long_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorULong___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorULong___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorULong___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorULong___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned long >::__delitem__(std::vector< unsigned long >::difference_type)\n"
  "    std::vector< unsigned long >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorULong___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< unsigned long >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___getitem__" "', argument " "1"" of type '" "std::vector< unsigned long > const *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___getitem__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2); try {
    result = (std::vector< unsigned long >::value_type *) &std_vector_Sl_unsigned_SS_long_Sg____getitem____SWIG_1((std::vector< unsigned long > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(*result));
  if (!swig::container_owner<swig::traits<std::vector< unsigned long >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorULong___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorULong___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorULong___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorULong___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< unsigned long >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< unsigned long >::__getitem__(std::vector< unsigned long >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorULong___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; std::vector< unsigned long >::difference_type arg2 ;
  std::vector< unsigned long >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< unsigned long >::value_type temp3 ; unsigned long val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong___setitem__" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorULong___setitem__" "', argument " "2"" of type '" "std::vector< unsigned long >::difference_type""'"); }
    arg2 = static_cast< std::vector< unsigned long >::difference_type >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorULong___setitem__" "', argument " "3"" of type '" "std::vector< unsigned long >::value_type""'"); }
    temp3 = static_cast< std::vector< unsigned long >::value_type >(val3); arg3 = &temp3; try {
    std_vector_Sl_unsigned_SS_long_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(unsigned long const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorULong___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorULong___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< unsigned long,std::allocator< unsigned long > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorULong___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorULong___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorULong___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< unsigned long >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< unsigned long,std::allocator< unsigned long > > const &)\n"
  "    std::vector< unsigned long >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< unsigned long >::__setitem__(std::vector< unsigned long >::difference_type,std::vector< unsigned long >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorULong__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< unsigned long > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< unsigned long > *)new std::vector< unsigned long >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorULong__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< unsigned long > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorULong" "', argument " "1"" of type '" "std::vector< unsigned long > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorULong" "', argument " "1"" of type '" "std::vector< unsigned long > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< unsigned long > *)new std::vector< unsigned long >((std::vector< unsigned long > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorULong_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorULong_clear" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorULong__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< unsigned long > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorULong" "', argument " "1"" of type '" "std::vector< unsigned long >::size_type""'"); }
    arg1 = static_cast< std::vector< unsigned long >::size_type >(val1);
  result = (std::vector< unsigned long > *)new std::vector< unsigned long >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorULong__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned long >::size_type arg1 ; std::vector< unsigned long >::value_type *arg2 = 0 ;
  size_t val1 ; int ecode1 = 0 ; std::vector< unsigned long >::value_type temp2 ; unsigned long val2 ; int ecode2 = 0 ;
  std::vector< unsigned long > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorULong" "', argument " "1"" of type '" "std::vector< unsigned long >::size_type""'"); }
    arg1 = static_cast< std::vector< unsigned long >::size_type >(val1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorULong" "', argument " "2"" of type '" "std::vector< unsigned long >::value_type""'"); }
    temp2 = static_cast< std::vector< unsigned long >::value_type >(val2); arg2 = &temp2;
  result = (std::vector< unsigned long > *)new std::vector< unsigned long >(SWIG_STD_MOVE(arg1),(std::vector< unsigned long >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorULong(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorULong", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorULong__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorULong__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorULong__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorULong__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorULong'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< unsigned long >::vector()\n"
  "    std::vector< unsigned long >::vector(std::vector< unsigned long > const &)\n"
  "    std::vector< unsigned long >::vector(std::vector< unsigned long >::size_type)\n"
  "    std::vector< unsigned long >::vector(std::vector< unsigned long >::size_type,std::vector< unsigned long >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorULong(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< unsigned long > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorULong" "', argument " "1"" of type '" "std::vector< unsigned long > *""'");  }
   arg1 = reinterpret_cast< std::vector< unsigned long > * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *vectorULong_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_unsigned_long_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorULong_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorString_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_iterator" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_string_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___nonzero__" "', argument " "1"" of type '" "std::vector< std::string > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (bool)std_vector_Sl_std_string_Sg____nonzero__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___bool__" "', argument " "1"" of type '" "std::vector< std::string > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (bool)std_vector_Sl_std_string_Sg____bool__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::string >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___len__" "', argument " "1"" of type '" "std::vector< std::string > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = std_vector_Sl_std_string_Sg____len__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< std::string,std::allocator< std::string > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorString___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___getslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___getslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::string >::difference_type >(val3); try {
    result = (std::vector< std::string,std::allocator< std::string > > *)std_vector_Sl_std_string_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___setslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::string >::difference_type >(val3); try {
    std_vector_Sl_std_string_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ; std::vector< std::string,std::allocator< std::string > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___setslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::string >::difference_type >(val3); {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorString___setslice__" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorString___setslice__" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_std_string_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::string,std::allocator< std::string > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorString___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorString___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::string >::__setslice__(std::vector< std::string >::difference_type,std::vector< std::string >::difference_type)\n"
  "    std::vector< std::string >::__setslice__(std::vector< std::string >::difference_type,std::vector< std::string >::difference_type,std::vector< std::string,std::allocator< std::string > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorString___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorString___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___delslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___delslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::string >::difference_type >(val3); try {
    std_vector_Sl_std_string_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___delitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); try {
    std_vector_Sl_std_string_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< std::string,std::allocator< std::string > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< std::string,std::allocator< std::string > > *)std_vector_Sl_std_string_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< std::string,std::allocator< std::string > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_string_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::string,std::allocator< std::string > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_std_string_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_std_string_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorString___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorString___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::string >::__delitem__(std::vector< std::string >::difference_type)\n"
  "    std::vector< std::string >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorString___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< std::string >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getitem__" "', argument " "1"" of type '" "std::vector< std::string > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___getitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); try {
    result = (std::vector< std::string >::value_type *) &std_vector_Sl_std_string_Sg____getitem____SWIG_1((std::vector< std::string > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::string >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorString___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorString___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::string >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::string >::__getitem__(std::vector< std::string >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::string >::difference_type >(val2); { std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); }
     arg3 = ptr; }  try { std_vector_Sl_std_string_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::string const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorString___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorString___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< std::string,std::allocator< std::string > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorString___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorString___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::string >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::string,std::allocator< std::string > > const &)\n"
  "    std::vector< std::string >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::string >::__setitem__(std::vector< std::string >::difference_type,std::vector< std::string >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< std::string > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::string > *)new std::vector< std::string >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< std::string > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< std::string > *)new std::vector< std::string >((std::vector< std::string > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorString_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_clear" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< std::string > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string >::size_type""'"); }
    arg1 = static_cast< std::vector< std::string >::size_type >(val1);
  result = (std::vector< std::string > *)new std::vector< std::string >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::string >::size_type arg1 ; std::vector< std::string >::value_type *arg2 = 0 ;
  size_t val1 ; int ecode1 = 0 ; int res2 = SWIG_OLDOBJ ; std::vector< std::string > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string >::size_type""'"); }
    arg1 = static_cast< std::vector< std::string >::size_type >(val1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorString" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorString" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); }
     arg2 = ptr; } 
  result = (std::vector< std::string > *)new std::vector< std::string >(SWIG_STD_MOVE(arg1),(std::vector< std::string >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorString(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorString", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorString__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorString__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorString__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorString__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorString'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::string >::vector()\n"
  "    std::vector< std::string >::vector(std::vector< std::string > const &)\n"
  "    std::vector< std::string >::vector(std::vector< std::string >::size_type)\n"
  "    std::vector< std::string >::vector(std::vector< std::string >::size_type,std::vector< std::string >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorString" "', argument " "1"" of type '" "std::vector< std::string > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::string > * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *vectorString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__string_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorVector3_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_vector3_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_vector3_Sg____nonzero__((std::vector< OpenBabel::vector3 > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_vector3_Sg____bool__((std::vector< OpenBabel::vector3 > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::vector3 >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1);
  result = std_vector_Sl_OpenBabel_vector3_Sg____len__((std::vector< OpenBabel::vector3 > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; std::vector< OpenBabel::vector3 >::difference_type arg2 ;
  std::vector< OpenBabel::vector3 >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorVector3___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorVector3___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)std_vector_Sl_OpenBabel_vector3_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ;
  std::vector< OpenBabel::vector3 >::difference_type arg2 ; std::vector< OpenBabel::vector3 >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorVector3___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_vector3_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ;
  std::vector< OpenBabel::vector3 >::difference_type arg2 ; std::vector< OpenBabel::vector3 >::difference_type arg3 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorVector3___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val3); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorVector3___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorVector3___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_vector3_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorVector3___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorVector3___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorVector3___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorVector3___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::vector3 >::__setslice__(std::vector< OpenBabel::vector3 >::difference_type,std::vector< OpenBabel::vector3 >::difference_type)\n"
  "    std::vector< OpenBabel::vector3 >::__setslice__(std::vector< OpenBabel::vector3 >::difference_type,std::vector< OpenBabel::vector3 >::difference_type,std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorVector3___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; std::vector< OpenBabel::vector3 >::difference_type arg2 ;
  std::vector< OpenBabel::vector3 >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorVector3___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorVector3___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_vector3_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ;
  std::vector< OpenBabel::vector3 >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_vector3_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorVector3___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)std_vector_Sl_OpenBabel_vector3_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorVector3___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorVector3___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorVector3___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorVector3___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorVector3___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_vector3_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorVector3___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorVector3___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorVector3___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorVector3___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::vector3 >::__delitem__(std::vector< OpenBabel::vector3 >::difference_type)\n"
  "    std::vector< OpenBabel::vector3 >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorVector3___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ;
  std::vector< OpenBabel::vector3 >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< OpenBabel::vector3 >::value_type *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::vector3 >::value_type *) &std_vector_Sl_OpenBabel_vector3_Sg____getitem____SWIG_1((std::vector< OpenBabel::vector3 > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!swig::container_owner<swig::traits<std::vector< OpenBabel::vector3 >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorVector3___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorVector3___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorVector3___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorVector3___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::vector3 >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::vector3 >::__getitem__(std::vector< OpenBabel::vector3 >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorVector3___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ;
  std::vector< OpenBabel::vector3 >::difference_type arg2 ; std::vector< OpenBabel::vector3 >::value_type *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorVector3___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::vector3 >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorVector3___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::value_type const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorVector3___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3 >::value_type const &""'"); }
   arg3 = reinterpret_cast< std::vector< OpenBabel::vector3 >::value_type * >(argp3); try {
    std_vector_Sl_OpenBabel_vector3_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(OpenBabel::vector3 const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorVector3___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorVector3___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorVector3___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorVector3___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorVector3___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::vector3 >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)\n"
  "    std::vector< OpenBabel::vector3 >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::vector3 >::__setitem__(std::vector< OpenBabel::vector3 >::difference_type,std::vector< OpenBabel::vector3 >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorVector3__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::vector3 > *)new std::vector< OpenBabel::vector3 >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorVector3__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::vector3 > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorVector3" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorVector3" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::vector3 > *)new std::vector< OpenBabel::vector3 >((std::vector< OpenBabel::vector3 > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorVector3_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorVector3_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorVector3__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::vector3 > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorVector3" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::vector3 >::size_type >(val1);
  result = (std::vector< OpenBabel::vector3 > *)new std::vector< OpenBabel::vector3 >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorVector3__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::vector3 >::size_type arg1 ;
  std::vector< OpenBabel::vector3 >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::vector3 > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorVector3" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::vector3 >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorVector3" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::value_type const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorVector3" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3 >::value_type const &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::vector3 >::value_type * >(argp2);
  result = (std::vector< OpenBabel::vector3 > *)new std::vector< OpenBabel::vector3 >(SWIG_STD_MOVE(arg1),(std::vector< OpenBabel::vector3 >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorVector3(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorVector3", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorVector3__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorVector3__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorVector3__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorVector3__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorVector3'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::vector3 >::vector()\n"
  "    std::vector< OpenBabel::vector3 >::vector(std::vector< OpenBabel::vector3 > const &)\n"
  "    std::vector< OpenBabel::vector3 >::vector(std::vector< OpenBabel::vector3 >::size_type)\n"
  "    std::vector< OpenBabel::vector3 >::vector(std::vector< OpenBabel::vector3 >::size_type,std::vector< OpenBabel::vector3 >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorVector3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3 > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorVector3" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3 > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::vector3 > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorVector3_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__vector3_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorVector3_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorvVector3_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3_iterator" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg__iterator(arg1,arg2);
  if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___nonzero__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____nonzero__((std::vector< std::vector< OpenBabel::vector3 > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___bool__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____bool__((std::vector< std::vector< OpenBabel::vector3 > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< OpenBabel::vector3 > >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___len__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  result = std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____len__((std::vector< std::vector< OpenBabel::vector3 > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vectorvVector3___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___getslice__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___getslice__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvVector3___getslice__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val3); try {
    result = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvVector3___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg3 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvVector3___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val3); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorvVector3___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvVector3___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvVector3___setslice__", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_vectorvVector3___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorvVector3___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvVector3___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__setslice__(std::vector< std::vector< OpenBabel::vector3 > >::difference_type,std::vector< std::vector< OpenBabel::vector3 > >::difference_type)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__setslice__(std::vector< std::vector< OpenBabel::vector3 > >::difference_type,std::vector< std::vector< OpenBabel::vector3 > >::difference_type,std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorvVector3___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorvVector3___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___delslice__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___delslice__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorvVector3___delslice__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___delitem__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2); try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvVector3___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvVector3___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorvVector3___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvVector3___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvVector3___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorvVector3___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____delitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvVector3___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorvVector3___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorvVector3___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvVector3___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__delitem__(std::vector< std::vector< OpenBabel::vector3 > >::difference_type)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorvVector3___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; std::vector< std::vector< OpenBabel::vector3 > >::value_type *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___getitem__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2); try {
    result = (std::vector< std::vector< OpenBabel::vector3 > >::value_type *) &std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____getitem____SWIG_1((std::vector< std::vector< OpenBabel::vector3 > > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = swig::from(static_cast< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::vector< OpenBabel::vector3 > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvVector3___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorvVector3___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorvVector3___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvVector3___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__getitem__(std::vector< std::vector< OpenBabel::vector3 > >::difference_type) const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > >::difference_type arg2 ;
  std::vector< std::vector< OpenBabel::vector3 > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorvVector3___setitem__" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::difference_type >(val2); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorvVector3___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorvVector3___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::value_type const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_OpenBabel_vector3_Sg__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorvVector3___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorvVector3___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorvVector3___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorvVector3___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorvVector3___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::__setitem__(std::vector< std::vector< OpenBabel::vector3 > >::difference_type,std::vector< std::vector< OpenBabel::vector3 > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorvVector3__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::vector< OpenBabel::vector3 > > *)new std::vector< std::vector< OpenBabel::vector3 > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvVector3__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ; std::vector< std::vector< OpenBabel::vector3 > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorvVector3" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorvVector3" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< std::vector< OpenBabel::vector3 > > *)new std::vector< std::vector< OpenBabel::vector3 > >((std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorvVector3_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorvVector3_clear" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); (arg1)->clear();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvVector3__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3 > > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorvVector3" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::size_type >(val1);
  result = (std::vector< std::vector< OpenBabel::vector3 > > *)new std::vector< std::vector< OpenBabel::vector3 > >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvVector3__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< OpenBabel::vector3 > >::size_type arg1 ;
  std::vector< std::vector< OpenBabel::vector3 > >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ; std::vector< std::vector< OpenBabel::vector3 > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorvVector3" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< OpenBabel::vector3 > >::size_type >(val1); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorvVector3" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorvVector3" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3 > >::value_type const &""'"); }
     arg2 = ptr; } 
  result = (std::vector< std::vector< OpenBabel::vector3 > > *)new std::vector< std::vector< OpenBabel::vector3 > >(SWIG_STD_MOVE(arg1),(std::vector< std::vector< OpenBabel::vector3 > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorvVector3(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorvVector3", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorvVector3__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorvVector3__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorvVector3__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorvVector3__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorvVector3'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::vector< OpenBabel::vector3 > >::vector()\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::vector(std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > const &)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::vector(std::vector< std::vector< OpenBabel::vector3 > >::size_type)\n"
  "    std::vector< std::vector< OpenBabel::vector3 > >::vector(std::vector< std::vector< OpenBabel::vector3 > >::size_type,std::vector< std::vector< OpenBabel::vector3 > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorvVector3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< OpenBabel::vector3 > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorvVector3" "', argument " "1"" of type '" "std::vector< std::vector< OpenBabel::vector3 > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< OpenBabel::vector3 > > * >(argp1); delete arg1;
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorvVector3_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorvVector3_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorOBMol_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBMol_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBMol_Sg____nonzero__((std::vector< OpenBabel::OBMol > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBMol_Sg____bool__((std::vector< OpenBabel::OBMol > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBMol >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBMol_Sg____len__((std::vector< OpenBabel::OBMol > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  std::vector< OpenBabel::OBMol >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBMol___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBMol___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *)std_vector_Sl_OpenBabel_OBMol_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  std::vector< OpenBabel::OBMol >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBMol___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBMol_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  std::vector< OpenBabel::OBMol >::difference_type arg3 ;
  std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBMol___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val3); {
    std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *ptr = (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorOBMol___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBMol___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBMol_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBMol___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorOBMol___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorOBMol___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBMol___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBMol >::__setslice__(std::vector< OpenBabel::OBMol >::difference_type,std::vector< OpenBabel::OBMol >::difference_type)\n"
  "    std::vector< OpenBabel::OBMol >::__setslice__(std::vector< OpenBabel::OBMol >::difference_type,std::vector< OpenBabel::OBMol >::difference_type,std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorOBMol___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  std::vector< OpenBabel::OBMol >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBMol___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBMol___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBMol_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBMol_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBMol___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *)std_vector_Sl_OpenBabel_OBMol_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBMol___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *ptr = (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBMol___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBMol___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBMol___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBMol___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBMol_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBMol___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBMol___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBMol___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBMol___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBMol >::__delitem__(std::vector< OpenBabel::OBMol >::difference_type)\n"
  "    std::vector< OpenBabel::OBMol >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBMol___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< OpenBabel::OBMol >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBMol >::value_type *) &std_vector_Sl_OpenBabel_OBMol_Sg____getitem____SWIG_1((std::vector< OpenBabel::OBMol > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!swig::container_owner<swig::traits<std::vector< OpenBabel::OBMol >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBMol___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBMol___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBMol___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBMol___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBMol >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBMol >::__getitem__(std::vector< OpenBabel::OBMol >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; std::vector< OpenBabel::OBMol >::difference_type arg2 ;
  std::vector< OpenBabel::OBMol >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBMol___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBMol >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBMol___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::value_type const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBMol___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBMol >::value_type const &""'"); }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBMol >::value_type * >(argp3); try {
    std_vector_Sl_OpenBabel_OBMol_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(OpenBabel::OBMol const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBMol___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBMol___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorOBMol___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorOBMol___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBMol___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBMol >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > const &)\n"
  "    std::vector< OpenBabel::OBMol >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBMol >::__setitem__(std::vector< OpenBabel::OBMol >::difference_type,std::vector< OpenBabel::OBMol >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorOBMol__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBMol > *)new std::vector< OpenBabel::OBMol >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBMol__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBMol > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *ptr = (std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorOBMol" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBMol" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBMol > *)new std::vector< OpenBabel::OBMol >((std::vector< OpenBabel::OBMol > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBMol_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBMol_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBMol__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBMol > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBMol" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBMol >::size_type >(val1);
  result = (std::vector< OpenBabel::OBMol > *)new std::vector< OpenBabel::OBMol >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBMol__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBMol >::size_type arg1 ;
  std::vector< OpenBabel::OBMol >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::OBMol > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBMol" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBMol >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorOBMol" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::value_type const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBMol" "', argument " "2"" of type '" "std::vector< OpenBabel::OBMol >::value_type const &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBMol >::value_type * >(argp2);
  result = (std::vector< OpenBabel::OBMol > *)new std::vector< OpenBabel::OBMol >(SWIG_STD_MOVE(arg1),(std::vector< OpenBabel::OBMol >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBMol(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorOBMol", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorOBMol__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorOBMol__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorOBMol__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorOBMol__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorOBMol'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBMol >::vector()\n"
  "    std::vector< OpenBabel::OBMol >::vector(std::vector< OpenBabel::OBMol > const &)\n"
  "    std::vector< OpenBabel::OBMol >::vector(std::vector< OpenBabel::OBMol >::size_type)\n"
  "    std::vector< OpenBabel::OBMol >::vector(std::vector< OpenBabel::OBMol >::size_type,std::vector< OpenBabel::OBMol >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorOBMol(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBMol > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorOBMol" "', argument " "1"" of type '" "std::vector< OpenBabel::OBMol > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBMol > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorOBMol_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBMol_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorOBMol_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorOBBond_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBBond_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBBond_Sg____nonzero__((std::vector< OpenBabel::OBBond > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBBond_Sg____bool__((std::vector< OpenBabel::OBBond > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBBond >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBBond_Sg____len__((std::vector< OpenBabel::OBBond > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  std::vector< OpenBabel::OBBond >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBBond___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBBond___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *)std_vector_Sl_OpenBabel_OBBond_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  std::vector< OpenBabel::OBBond >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBBond___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBBond_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  std::vector< OpenBabel::OBBond >::difference_type arg3 ;
  std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBBond___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val3); {
    std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *ptr = (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorOBBond___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBBond___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBBond_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBBond___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorOBBond___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorOBBond___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBBond___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBBond >::__setslice__(std::vector< OpenBabel::OBBond >::difference_type,std::vector< OpenBabel::OBBond >::difference_type)\n"
  "    std::vector< OpenBabel::OBBond >::__setslice__(std::vector< OpenBabel::OBBond >::difference_type,std::vector< OpenBabel::OBBond >::difference_type,std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorOBBond___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  std::vector< OpenBabel::OBBond >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBBond___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBBond___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBBond_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBBond_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBBond___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *)std_vector_Sl_OpenBabel_OBBond_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBBond___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *ptr = (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBBond___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBBond___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBBond___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBBond___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBBond_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBBond___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBBond___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBBond___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBBond___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBBond >::__delitem__(std::vector< OpenBabel::OBBond >::difference_type)\n"
  "    std::vector< OpenBabel::OBBond >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBBond___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< OpenBabel::OBBond >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBBond >::value_type *) &std_vector_Sl_OpenBabel_OBBond_Sg____getitem____SWIG_1((std::vector< OpenBabel::OBBond > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!swig::container_owner<swig::traits<std::vector< OpenBabel::OBBond >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBBond___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBBond___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBBond___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBBond___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBBond >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBBond >::__getitem__(std::vector< OpenBabel::OBBond >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; std::vector< OpenBabel::OBBond >::difference_type arg2 ;
  std::vector< OpenBabel::OBBond >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBBond___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBBond >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBond,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBBond___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::value_type const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBBond___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBBond >::value_type const &""'"); }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBBond >::value_type * >(argp3); try {
    std_vector_Sl_OpenBabel_OBBond_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(OpenBabel::OBBond const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBBond___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBBond___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorOBBond___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorOBBond___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBBond___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBBond >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > const &)\n"
  "    std::vector< OpenBabel::OBBond >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBBond >::__setitem__(std::vector< OpenBabel::OBBond >::difference_type,std::vector< OpenBabel::OBBond >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorOBBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBBond > *)new std::vector< OpenBabel::OBBond >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBBond > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *ptr = (std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorOBBond" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBBond" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBBond > *)new std::vector< OpenBabel::OBBond >((std::vector< OpenBabel::OBBond > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBBond_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBBond_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBBond > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBBond" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBBond >::size_type >(val1);
  result = (std::vector< OpenBabel::OBBond > *)new std::vector< OpenBabel::OBBond >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBBond__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBBond >::size_type arg1 ;
  std::vector< OpenBabel::OBBond >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::OBBond > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBBond" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBBond >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBond,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorOBBond" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::value_type const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBBond" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBond >::value_type const &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBBond >::value_type * >(argp2);
  result = (std::vector< OpenBabel::OBBond > *)new std::vector< OpenBabel::OBBond >(SWIG_STD_MOVE(arg1),(std::vector< OpenBabel::OBBond >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorOBBond", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorOBBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorOBBond__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorOBBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorOBBond__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorOBBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBBond >::vector()\n"
  "    std::vector< OpenBabel::OBBond >::vector(std::vector< OpenBabel::OBBond > const &)\n"
  "    std::vector< OpenBabel::OBBond >::vector(std::vector< OpenBabel::OBBond >::size_type)\n"
  "    std::vector< OpenBabel::OBBond >::vector(std::vector< OpenBabel::OBBond >::size_type,std::vector< OpenBabel::OBBond >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorOBBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBBond > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorOBBond" "', argument " "1"" of type '" "std::vector< OpenBabel::OBBond > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBBond > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorOBBond_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorOBBond_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorOBResidue_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBResidue_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBResidue_Sg____nonzero__((std::vector< OpenBabel::OBResidue > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBResidue_Sg____bool__((std::vector< OpenBabel::OBResidue > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBResidue >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBResidue_Sg____len__((std::vector< OpenBabel::OBResidue > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; std::vector< OpenBabel::OBResidue >::difference_type arg2 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBResidue___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBResidue___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *)std_vector_Sl_OpenBabel_OBResidue_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg2 ; std::vector< OpenBabel::OBResidue >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBResidue___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg2 ; std::vector< OpenBabel::OBResidue >::difference_type arg3 ;
  std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBResidue___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val3); {
    std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *ptr = (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorOBResidue___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBResidue___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBResidue___setslice__", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_vectorOBResidue___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorOBResidue___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBResidue___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBResidue >::__setslice__(std::vector< OpenBabel::OBResidue >::difference_type,std::vector< OpenBabel::OBResidue >::difference_type)\n"
  "    std::vector< OpenBabel::OBResidue >::__setslice__(std::vector< OpenBabel::OBResidue >::difference_type,std::vector< OpenBabel::OBResidue >::difference_type,std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; std::vector< OpenBabel::OBResidue >::difference_type arg2 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBResidue___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBResidue___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBResidue___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *)std_vector_Sl_OpenBabel_OBResidue_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBResidue___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *ptr = (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBResidue___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBResidue___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBResidue___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBResidue___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBResidue_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBResidue___delitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBResidue___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBResidue___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBResidue___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBResidue >::__delitem__(std::vector< OpenBabel::OBResidue >::difference_type)\n"
  "    std::vector< OpenBabel::OBResidue >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< OpenBabel::OBResidue >::value_type *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBResidue >::value_type *) &std_vector_Sl_OpenBabel_OBResidue_Sg____getitem____SWIG_1((std::vector< OpenBabel::OBResidue > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 );
  if (!swig::container_owner<swig::traits<std::vector< OpenBabel::OBResidue >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBResidue___getitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBResidue___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBResidue___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBResidue___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBResidue >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBResidue >::__getitem__(std::vector< OpenBabel::OBResidue >::difference_type) const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ;
  std::vector< OpenBabel::OBResidue >::difference_type arg2 ; std::vector< OpenBabel::OBResidue >::value_type *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBResidue___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBResidue >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBResidue,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBResidue___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::value_type const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBResidue___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBResidue >::value_type const &""'"); }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBResidue >::value_type * >(argp3); try {
    std_vector_Sl_OpenBabel_OBResidue_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(OpenBabel::OBResidue const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBResidue___setitem__", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_vectorOBResidue___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorOBResidue___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorOBResidue___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBResidue___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBResidue >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > const &)\n"
  "    std::vector< OpenBabel::OBResidue >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBResidue >::__setitem__(std::vector< OpenBabel::OBResidue >::difference_type,std::vector< OpenBabel::OBResidue >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorOBResidue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBResidue > *)new std::vector< OpenBabel::OBResidue >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBResidue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBResidue > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *ptr = (std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorOBResidue" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBResidue" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBResidue > *)new std::vector< OpenBabel::OBResidue >((std::vector< OpenBabel::OBResidue > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBResidue_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBResidue_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBResidue__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBResidue > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBResidue" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBResidue >::size_type >(val1);
  result = (std::vector< OpenBabel::OBResidue > *)new std::vector< OpenBabel::OBResidue >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBResidue__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBResidue >::size_type arg1 ;
  std::vector< OpenBabel::OBResidue >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::OBResidue > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBResidue" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBResidue >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBResidue,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorOBResidue" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::value_type const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBResidue" "', argument " "2"" of type '" "std::vector< OpenBabel::OBResidue >::value_type const &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBResidue >::value_type * >(argp2);
  result = (std::vector< OpenBabel::OBResidue > *)new std::vector< OpenBabel::OBResidue >(SWIG_STD_MOVE(arg1),(std::vector< OpenBabel::OBResidue >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBResidue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorOBResidue", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorOBResidue__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorOBResidue__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorOBResidue__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorOBResidue__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorOBResidue'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBResidue >::vector()\n"
  "    std::vector< OpenBabel::OBResidue >::vector(std::vector< OpenBabel::OBResidue > const &)\n"
  "    std::vector< OpenBabel::OBResidue >::vector(std::vector< OpenBabel::OBResidue >::size_type)\n"
  "    std::vector< OpenBabel::OBResidue >::vector(std::vector< OpenBabel::OBResidue >::size_type,std::vector< OpenBabel::OBResidue >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorOBResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBResidue > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorOBResidue" "', argument " "1"" of type '" "std::vector< OpenBabel::OBResidue > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBResidue > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorOBResidue_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBResidue_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorOBResidue_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorOBRing_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBRing_Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBRing_Sg____nonzero__((std::vector< OpenBabel::OBRing > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBRing_Sg____bool__((std::vector< OpenBabel::OBRing > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBRing_Sg____len__((std::vector< OpenBabel::OBRing > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBRing___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBRing___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *)std_vector_Sl_OpenBabel_OBRing_Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBRing___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBRing_Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing >::difference_type arg3 ;
  std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBRing___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val3); {
    std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *ptr = (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorOBRing___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBRing___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBRing_Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBRing___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorOBRing___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorOBRing___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBRing___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing >::__setslice__(std::vector< OpenBabel::OBRing >::difference_type,std::vector< OpenBabel::OBRing >::difference_type)\n"
  "    std::vector< OpenBabel::OBRing >::__setslice__(std::vector< OpenBabel::OBRing >::difference_type,std::vector< OpenBabel::OBRing >::difference_type,std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorOBRing___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorOBRing___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorOBRing___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBRing_Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBRing_Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBRing___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *)std_vector_Sl_OpenBabel_OBRing_Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBRing___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *ptr = (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBRing___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorOBRing___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBRing_Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBRing___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBRing___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBRing___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBRing___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing >::__delitem__(std::vector< OpenBabel::OBRing >::difference_type)\n"
  "    std::vector< OpenBabel::OBRing >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBRing___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< OpenBabel::OBRing >::value_type *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBRing >::value_type *) &std_vector_Sl_OpenBabel_OBRing_Sg____getitem____SWIG_1((std::vector< OpenBabel::OBRing > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!swig::container_owner<swig::traits<std::vector< OpenBabel::OBRing >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBRing___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorOBRing___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBRing___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBRing___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBRing >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBRing >::__getitem__(std::vector< OpenBabel::OBRing >::difference_type) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; std::vector< OpenBabel::OBRing >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorOBRing___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBRing,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::value_type const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing >::value_type const &""'"); }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBRing >::value_type * >(argp3); try {
    std_vector_Sl_OpenBabel_OBRing_Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(OpenBabel::OBRing const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorOBRing___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorOBRing___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorOBRing___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorOBRing___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorOBRing___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > const &)\n"
  "    std::vector< OpenBabel::OBRing >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBRing >::__setitem__(std::vector< OpenBabel::OBRing >::difference_type,std::vector< OpenBabel::OBRing >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorOBRing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBRing > *)new std::vector< OpenBabel::OBRing >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBRing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBRing > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *ptr = (std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBRing > *)new std::vector< OpenBabel::OBRing >((std::vector< OpenBabel::OBRing > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorOBRing_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorOBRing_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBRing__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBRing > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBRing >::size_type >(val1);
  result = (std::vector< OpenBabel::OBRing > *)new std::vector< OpenBabel::OBRing >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBRing__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing >::size_type arg1 ;
  std::vector< OpenBabel::OBRing >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::OBRing > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBRing >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBRing,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorOBRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::value_type const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorOBRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing >::value_type const &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBRing >::value_type * >(argp2);
  result = (std::vector< OpenBabel::OBRing > *)new std::vector< OpenBabel::OBRing >(SWIG_STD_MOVE(arg1),(std::vector< OpenBabel::OBRing >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorOBRing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorOBRing", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorOBRing__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorOBRing__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorOBRing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorOBRing__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorOBRing'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBRing >::vector()\n"
  "    std::vector< OpenBabel::OBRing >::vector(std::vector< OpenBabel::OBRing > const &)\n"
  "    std::vector< OpenBabel::OBRing >::vector(std::vector< OpenBabel::OBRing >::size_type)\n"
  "    std::vector< OpenBabel::OBRing >::vector(std::vector< OpenBabel::OBRing >::size_type,std::vector< OpenBabel::OBRing >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorOBRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorOBRing_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorOBRing_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorpOBRing_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBRing_Sm__Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBRing_Sm__Sg____nonzero__((std::vector< OpenBabel::OBRing * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBRing_Sm__Sg____bool__((std::vector< OpenBabel::OBRing * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing * >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBRing_Sm__Sg____len__((std::vector< OpenBabel::OBRing * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; std::vector< OpenBabel::OBRing * >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBRing___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBRing___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *)std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg2 ; std::vector< OpenBabel::OBRing * >::difference_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBRing___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg2 ; std::vector< OpenBabel::OBRing * >::difference_type arg3 ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBRing___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val3); {
    std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *ptr = (std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorpOBRing___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBRing___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBRing___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vectorpOBRing___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorpOBRing___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBRing___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing * >::__setslice__(std::vector< OpenBabel::OBRing * >::difference_type,std::vector< OpenBabel::OBRing * >::difference_type)\n"
  "    std::vector< OpenBabel::OBRing * >::__setslice__(std::vector< OpenBabel::OBRing * >::difference_type,std::vector< OpenBabel::OBRing * >::difference_type,std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; std::vector< OpenBabel::OBRing * >::difference_type arg2 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBRing___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBRing___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBRing___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *)std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBRing___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *ptr = (std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBRing___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBRing___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try { std_vector_Sl_OpenBabel_OBRing_Sm__Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBRing___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBRing___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBRing___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBRing___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing * >::__delitem__(std::vector< OpenBabel::OBRing * >::difference_type)\n"
  "    std::vector< OpenBabel::OBRing * >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< OpenBabel::OBRing * >::value_type result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBRing * >::value_type)std_vector_Sl_OpenBabel_OBRing_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBRing___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBRing___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBRing___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBRing___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBRing * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBRing * >::__getitem__(std::vector< OpenBabel::OBRing * >::difference_type)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ;
  std::vector< OpenBabel::OBRing * >::difference_type arg2 ; std::vector< OpenBabel::OBRing * >::value_type arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBRing___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBRing * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBRing___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBRing * >::value_type""'");  }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBRing * >::value_type >(argp3); try {
    std_vector_Sl_OpenBabel_OBRing_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBRing___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBRing___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vectorpOBRing___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vectorpOBRing___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBRing___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBRing * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > const &)\n"
  "    std::vector< OpenBabel::OBRing * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBRing * >::__setitem__(std::vector< OpenBabel::OBRing * >::difference_type,std::vector< OpenBabel::OBRing * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorpOBRing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBRing * > *)new std::vector< OpenBabel::OBRing * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBRing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBRing * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *ptr = (std::vector< OpenBabel::OBRing*,std::allocator< OpenBabel::OBRing * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorpOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorpOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBRing * > *)new std::vector< OpenBabel::OBRing * >((std::vector< OpenBabel::OBRing * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBRing_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBRing_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBRing__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBRing * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBRing * >::size_type >(val1);
  result = (std::vector< OpenBabel::OBRing * > *)new std::vector< OpenBabel::OBRing * >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBRing__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBRing * >::size_type arg1 ;
  std::vector< OpenBabel::OBRing * >::value_type arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::vector< OpenBabel::OBRing * > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBRing * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorpOBRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing * >::value_type""'");  }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBRing * >::value_type >(argp2);
  result = (std::vector< OpenBabel::OBRing * > *)new std::vector< OpenBabel::OBRing * >(SWIG_STD_MOVE(arg1),arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBRing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorpOBRing", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorpOBRing__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorpOBRing__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorpOBRing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorpOBRing__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorpOBRing'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBRing * >::vector()\n"
  "    std::vector< OpenBabel::OBRing * >::vector(std::vector< OpenBabel::OBRing * > const &)\n"
  "    std::vector< OpenBabel::OBRing * >::vector(std::vector< OpenBabel::OBRing * >::size_type)\n"
  "    std::vector< OpenBabel::OBRing * >::vector(std::vector< OpenBabel::OBRing * >::size_type,std::vector< OpenBabel::OBRing * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorpOBRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBRing * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorpOBRing" "', argument " "1"" of type '" "std::vector< OpenBabel::OBRing * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBRing * > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorpOBRing_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorpOBRing_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorpOBGenericData_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____nonzero__((std::vector< OpenBabel::OBGenericData * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____bool__((std::vector< OpenBabel::OBGenericData * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBGenericData * >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____len__((std::vector< OpenBabel::OBGenericData * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBGenericData___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBGenericData___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg3 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBGenericData___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val3); {
    std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *ptr = (std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorpOBGenericData___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBGenericData___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___setslice__", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_vectorpOBGenericData___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorpOBGenericData___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBGenericData___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBGenericData * >::__setslice__(std::vector< OpenBabel::OBGenericData * >::difference_type,std::vector< OpenBabel::OBGenericData * >::difference_type)\n"
  "    std::vector< OpenBabel::OBGenericData * >::__setslice__(std::vector< OpenBabel::OBGenericData * >::difference_type,std::vector< OpenBabel::OBGenericData * >::difference_type,std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBGenericData___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBGenericData___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBGenericData___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *ptr = (std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBGenericData___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBGenericData___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBGenericData___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBGenericData___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____delitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___delitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBGenericData___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBGenericData___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBGenericData___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBGenericData * >::__delitem__(std::vector< OpenBabel::OBGenericData * >::difference_type)\n"
  "    std::vector< OpenBabel::OBGenericData * >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; std::vector< OpenBabel::OBGenericData * >::value_type result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBGenericData * >::value_type)std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___getitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBGenericData___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBGenericData___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBGenericData___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBGenericData * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBGenericData * >::__getitem__(std::vector< OpenBabel::OBGenericData * >::difference_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData * >::difference_type arg2 ;
  std::vector< OpenBabel::OBGenericData * >::value_type arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBGenericData___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBGenericData * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBGenericData___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBGenericData * >::value_type""'");  }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * >::value_type >(argp3); try {
    std_vector_Sl_OpenBabel_OBGenericData_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBGenericData___setitem__", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_vectorpOBGenericData___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_vectorpOBGenericData___setitem____SWIG_0(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_vectorpOBGenericData___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBGenericData___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBGenericData * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > const &)\n"
  "    std::vector< OpenBabel::OBGenericData * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBGenericData * >::__setitem__(std::vector< OpenBabel::OBGenericData * >::difference_type,std::vector< OpenBabel::OBGenericData * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorpOBGenericData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBGenericData * > *)new std::vector< OpenBabel::OBGenericData * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBGenericData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBGenericData * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *ptr = (std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorpOBGenericData" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorpOBGenericData" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBGenericData * > *)new std::vector< OpenBabel::OBGenericData * >((std::vector< OpenBabel::OBGenericData * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBGenericData_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBGenericData_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBGenericData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBGenericData * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBGenericData" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBGenericData * >::size_type >(val1);
  result = (std::vector< OpenBabel::OBGenericData * > *)new std::vector< OpenBabel::OBGenericData * >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBGenericData__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBGenericData * >::size_type arg1 ;
  std::vector< OpenBabel::OBGenericData * >::value_type arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; std::vector< OpenBabel::OBGenericData * > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBGenericData" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBGenericData * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorpOBGenericData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData * >::value_type""'");  }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * >::value_type >(argp2);
  result = (std::vector< OpenBabel::OBGenericData * > *)new std::vector< OpenBabel::OBGenericData * >(SWIG_STD_MOVE(arg1),arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBGenericData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorpOBGenericData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorpOBGenericData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorpOBGenericData__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorpOBGenericData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorpOBGenericData__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorpOBGenericData'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBGenericData * >::vector()\n"
  "    std::vector< OpenBabel::OBGenericData * >::vector(std::vector< OpenBabel::OBGenericData * > const &)\n"
  "    std::vector< OpenBabel::OBGenericData * >::vector(std::vector< OpenBabel::OBGenericData * >::size_type)\n"
  "    std::vector< OpenBabel::OBGenericData * >::vector(std::vector< OpenBabel::OBGenericData * >::size_type,std::vector< OpenBabel::OBGenericData * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorpOBGenericData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBGenericData * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorpOBGenericData" "', argument " "1"" of type '" "std::vector< OpenBabel::OBGenericData * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBGenericData * > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorpOBGenericData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorpOBGenericData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord_iterator" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg__iterator(arg1,arg2); if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___nonzero__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____nonzero__((std::vector< OpenBabel::OBInternalCoord * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___bool__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  result = (bool)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____bool__((std::vector< OpenBabel::OBInternalCoord * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBInternalCoord * >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___len__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > const *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  result = std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____len__((std::vector< OpenBabel::OBInternalCoord * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___getslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___getslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBInternalCoord___getslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val3); try {
    result = (std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg3 ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val3); {
    std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *ptr = (std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBInternalCoord___setslice__" "', argument " "4"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___setslice__", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_vectorpOBInternalCoord___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vectorpOBInternalCoord___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBInternalCoord___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__setslice__(std::vector< OpenBabel::OBInternalCoord * >::difference_type,std::vector< OpenBabel::OBInternalCoord * >::difference_type)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__setslice__(std::vector< OpenBabel::OBInternalCoord * >::difference_type,std::vector< OpenBabel::OBInternalCoord * >::difference_type,std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___delslice__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___delslice__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorpOBInternalCoord___delslice__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg3 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val3); try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___delitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2); try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBInternalCoord___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *ptr = (std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___delitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); { if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorpOBInternalCoord___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____delitem____SWIG_1(arg1,arg2);}  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___delitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBInternalCoord___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBInternalCoord___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBInternalCoord___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__delitem__(std::vector< OpenBabel::OBInternalCoord * >::difference_type)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; std::vector< OpenBabel::OBInternalCoord * >::value_type result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___getitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___getitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2); try {
    result = (std::vector< OpenBabel::OBInternalCoord * >::value_type)std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____getitem____SWIG_1(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___getitem__", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vectorpOBInternalCoord___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vectorpOBInternalCoord___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBInternalCoord___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBInternalCoord * >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__getitem__(std::vector< OpenBabel::OBInternalCoord * >::difference_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * >::difference_type arg2 ;
  std::vector< OpenBabel::OBInternalCoord * >::value_type arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ;
  int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::difference_type""'"); }
    arg2 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorpOBInternalCoord___setitem__" "', argument " "3"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::value_type""'");  }
   arg3 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * >::value_type >(argp3); try {
    std_vector_Sl_OpenBabel_OBInternalCoord_Sm__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorpOBInternalCoord___setitem__", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_vectorpOBInternalCoord___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_vectorpOBInternalCoord___setitem____SWIG_0(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_vectorpOBInternalCoord___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorpOBInternalCoord___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > const &)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::__setitem__(std::vector< OpenBabel::OBInternalCoord * >::difference_type,std::vector< OpenBabel::OBInternalCoord * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vectorpOBInternalCoord__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< OpenBabel::OBInternalCoord * > *)new std::vector< OpenBabel::OBInternalCoord * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBInternalCoord__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< OpenBabel::OBInternalCoord * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *ptr = (std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorpOBInternalCoord" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vectorpOBInternalCoord" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< OpenBabel::OBInternalCoord * > *)new std::vector< OpenBabel::OBInternalCoord * >((std::vector< OpenBabel::OBInternalCoord * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vectorpOBInternalCoord_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorpOBInternalCoord_clear" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBInternalCoord__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * >::size_type arg1 ; size_t val1 ; int ecode1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord * > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBInternalCoord" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::size_type >(val1);
  result = (std::vector< OpenBabel::OBInternalCoord * > *)new std::vector< OpenBabel::OBInternalCoord * >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBInternalCoord__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< OpenBabel::OBInternalCoord * >::size_type arg1 ;
  std::vector< OpenBabel::OBInternalCoord * >::value_type arg2 = 0 ; size_t val1 ; int ecode1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; std::vector< OpenBabel::OBInternalCoord * > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorpOBInternalCoord" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::size_type""'"); }
    arg1 = static_cast< std::vector< OpenBabel::OBInternalCoord * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorpOBInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord * >::value_type""'");  }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * >::value_type >(argp2);
  result = (std::vector< OpenBabel::OBInternalCoord * > *)new std::vector< OpenBabel::OBInternalCoord * >(SWIG_STD_MOVE(arg1),arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vectorpOBInternalCoord(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorpOBInternalCoord", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vectorpOBInternalCoord__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vectorpOBInternalCoord__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vectorpOBInternalCoord__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vectorpOBInternalCoord__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorpOBInternalCoord'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< OpenBabel::OBInternalCoord * >::vector()\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::vector(std::vector< OpenBabel::OBInternalCoord * > const &)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::vector(std::vector< OpenBabel::OBInternalCoord * >::size_type)\n"
  "    std::vector< OpenBabel::OBInternalCoord * >::vector(std::vector< OpenBabel::OBInternalCoord * >::size_type,std::vector< OpenBabel::OBInternalCoord * >::value_type)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vectorpOBInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord * > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorpOBInternalCoord" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord * > *""'");  }
   arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord * > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vectorpOBInternalCoord_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorpOBInternalCoord_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_pairUIntUInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::pair< unsigned int,unsigned int > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::pair< unsigned int,unsigned int > *)new std::pair< unsigned int,unsigned int >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_pairUIntUInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int arg2 ; unsigned int val1 ; int ecode1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; std::pair< unsigned int,unsigned int > *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_pairUIntUInt" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_pairUIntUInt" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2);
  result = (std::pair< unsigned int,unsigned int > *)new std::pair< unsigned int,unsigned int >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_pairUIntUInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::pair< unsigned int,unsigned int > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::pair< unsigned int,unsigned int > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::pair< unsigned int,unsigned int > *ptr = (std::pair< unsigned int,unsigned int > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_pairUIntUInt" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_pairUIntUInt" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > const &""'"); }
     arg1 = ptr; } 
  result = (std::pair< unsigned int,unsigned int > *)new std::pair< unsigned int,unsigned int >((std::pair< unsigned int,unsigned int > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_new_pairUIntUInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_pairUIntUInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_pairUIntUInt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_pairUIntUInt__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_pairUIntUInt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_pairUIntUInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::pair< unsigned int,unsigned int >::pair()\n"
  "    std::pair< unsigned int,unsigned int >::pair(unsigned int,unsigned int)\n"
  "    std::pair< unsigned int,unsigned int >::pair(std::pair< unsigned int,unsigned int > const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_pairUIntUInt_first_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::pair< unsigned int,unsigned int > *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "pairUIntUInt_first_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairUIntUInt_first_set" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::pair< unsigned int,unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "pairUIntUInt_first_set" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); if (arg1) (arg1)->first = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_pairUIntUInt_first_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::pair< unsigned int,unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  unsigned int result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairUIntUInt_first_get" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::pair< unsigned int,unsigned int > * >(argp1); result = (unsigned int) ((arg1)->first);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_pairUIntUInt_second_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::pair< unsigned int,unsigned int > *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "pairUIntUInt_second_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairUIntUInt_second_set" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::pair< unsigned int,unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "pairUIntUInt_second_set" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); if (arg1) (arg1)->second = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_pairUIntUInt_second_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::pair< unsigned int,unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  unsigned int result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairUIntUInt_second_get" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::pair< unsigned int,unsigned int > * >(argp1); result = (unsigned int) ((arg1)->second);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_pairUIntUInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::pair< unsigned int,unsigned int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_pairUIntUInt" "', argument " "1"" of type '" "std::pair< unsigned int,unsigned int > *""'");  }
   arg1 = reinterpret_cast< std::pair< unsigned int,unsigned int > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *pairUIntUInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__pairT_unsigned_int_unsigned_int_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *pairUIntUInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vpairUIntUInt_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt_iterator" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__iterator(arg1,arg2);
  if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___nonzero__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  result = (bool)std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____nonzero__((std::vector< std::pair< unsigned int,unsigned int > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___bool__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  result = (bool)std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____bool__((std::vector< std::pair< unsigned int,unsigned int > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::pair< unsigned int,unsigned int > >::size_type result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___len__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  result = std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____len__((std::vector< std::pair< unsigned int,unsigned int > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vpairUIntUInt___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___getslice__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___getslice__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vpairUIntUInt___getslice__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val3); try {
    result = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___setslice__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___setslice__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vpairUIntUInt___setslice__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val3); try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg3 ;
  std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___setslice__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___setslice__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vpairUIntUInt___setslice__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val3); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vpairUIntUInt___setslice__" "', argument " "4"" of type '" "std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vpairUIntUInt___setslice__" "', argument " "4"" of type '" "std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vpairUIntUInt___setslice__", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vpairUIntUInt___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vpairUIntUInt___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vpairUIntUInt___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__setslice__(std::vector< std::pair< unsigned int,unsigned int > >::difference_type,std::vector< std::pair< unsigned int,unsigned int > >::difference_type)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__setslice__(std::vector< std::pair< unsigned int,unsigned int > >::difference_type,std::vector< std::pair< unsigned int,unsigned int > >::difference_type,std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vpairUIntUInt___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___delslice__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___delslice__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vpairUIntUInt___delslice__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val3); try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___delitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___delitem__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2); try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___getitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vpairUIntUInt___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vpairUIntUInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vpairUIntUInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; SWIGPY_SLICEOBJECT *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___delitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vpairUIntUInt___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vpairUIntUInt___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vpairUIntUInt___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vpairUIntUInt___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vpairUIntUInt___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__delitem__(std::vector< std::pair< unsigned int,unsigned int > >::difference_type)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; std::vector< std::pair< unsigned int,unsigned int > >::value_type *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___getitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___getitem__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2); try {
    result = (std::vector< std::pair< unsigned int,unsigned int > >::value_type *) &std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____getitem____SWIG_1((std::vector< std::pair< unsigned int,unsigned int > > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = swig::from(static_cast< std::pair< unsigned int,unsigned int > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::pair< unsigned int,unsigned int > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vpairUIntUInt___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vpairUIntUInt___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vpairUIntUInt___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vpairUIntUInt___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__getitem__(std::vector< std::pair< unsigned int,unsigned int > >::difference_type) const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ;
  std::vector< std::pair< unsigned int,unsigned int > >::difference_type arg2 ;
  std::vector< std::pair< unsigned int,unsigned int > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vpairUIntUInt___setitem__" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::difference_type >(val2); {
    std::pair< unsigned int,unsigned int > *ptr = (std::pair< unsigned int,unsigned int > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::value_type const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::pair< unsigned int,unsigned int > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vpairUIntUInt___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vpairUIntUInt___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vpairUIntUInt___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vpairUIntUInt___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vpairUIntUInt___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::__setitem__(std::vector< std::pair< unsigned int,unsigned int > >::difference_type,std::vector< std::pair< unsigned int,unsigned int > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vpairUIntUInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::pair< unsigned int,unsigned int > > *)new std::vector< std::pair< unsigned int,unsigned int > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vpairUIntUInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; int res1 = SWIG_OLDOBJ ;
  std::vector< std::pair< unsigned int,unsigned int > > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< std::pair< unsigned int,unsigned int > > *)new std::vector< std::pair< unsigned int,unsigned int > >((std::vector< std::pair< unsigned int,unsigned int > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vpairUIntUInt_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vpairUIntUInt_clear" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); (arg1)->clear();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vpairUIntUInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > >::size_type arg1 ; size_t val1 ;
  int ecode1 = 0 ; std::vector< std::pair< unsigned int,unsigned int > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::size_type >(val1);
  result = (std::vector< std::pair< unsigned int,unsigned int > > *)new std::vector< std::pair< unsigned int,unsigned int > >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vpairUIntUInt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::pair< unsigned int,unsigned int > >::size_type arg1 ;
  std::vector< std::pair< unsigned int,unsigned int > >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ; std::vector< std::pair< unsigned int,unsigned int > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::pair< unsigned int,unsigned int > >::size_type >(val1); {
    std::pair< unsigned int,unsigned int > *ptr = (std::pair< unsigned int,unsigned int > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vpairUIntUInt" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vpairUIntUInt" "', argument " "2"" of type '" "std::vector< std::pair< unsigned int,unsigned int > >::value_type const &""'"); }
     arg2 = ptr; } 
  result = (std::vector< std::pair< unsigned int,unsigned int > > *)new std::vector< std::pair< unsigned int,unsigned int > >(SWIG_STD_MOVE(arg1),(std::vector< std::pair< unsigned int,unsigned int > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vpairUIntUInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vpairUIntUInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vpairUIntUInt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vpairUIntUInt__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vpairUIntUInt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vpairUIntUInt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vpairUIntUInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::pair< unsigned int,unsigned int > >::vector()\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::vector(std::vector< std::pair< unsigned int,unsigned int > > const &)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::vector(std::vector< std::pair< unsigned int,unsigned int > >::size_type)\n"
  "    std::vector< std::pair< unsigned int,unsigned int > >::vector(std::vector< std::pair< unsigned int,unsigned int > >::size_type,std::vector< std::pair< unsigned int,unsigned int > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_delete_vpairUIntUInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::pair< unsigned int,unsigned int > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::pair< unsigned int,unsigned int > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::pair< unsigned int,unsigned int > > * >(argp1); delete arg1;
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vpairUIntUInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vpairUIntUInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vvpairUIntUInt_iterator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; PyObject **arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; PyObject *swig_obj[1] ; swig::SwigPyIterator *result = 0 ; arg2 = &swig_obj[0]; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_iterator" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__iterator(arg1,arg2);
  if (!result) SWIG_fail;
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___nonzero__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____nonzero__((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___bool__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___bool__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (bool)std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____bool__((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___len__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___len__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____len__((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___getslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___getslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___getslice__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___getslice__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vvpairUIntUInt___getslice__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val3); try {
    result = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setslice____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vvpairUIntUInt___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setslice____SWIG_0(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setslice____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg3 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___setslice__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___setslice__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vvpairUIntUInt___setslice__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val3); {
    std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *ptr = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vvpairUIntUInt___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt___setslice__" "', argument " "4"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &""'"); }
     arg4 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setslice____SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &)*arg4);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj; fail: if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setslice__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___setslice__", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_vvpairUIntUInt___setslice____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vvpairUIntUInt___setslice____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt___setslice__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__setslice__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__setslice__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type,std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___delslice__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; ptrdiff_t val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___delslice__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___delslice__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___delslice__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vvpairUIntUInt___delslice__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg3 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val3); try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delslice__(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___delitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___delitem__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2); try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delitem____SWIG_0(arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___getitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt___getitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    result = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getitem____SWIG_0(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); } 
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setitem____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  SWIGPY_SLICEOBJECT *arg2 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  {
    std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *ptr = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vvpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt___setitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___delitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  SWIGPY_SLICEOBJECT *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___delitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt___delitem__" "', argument " "2"" of type '" "SWIGPY_SLICEOBJECT *""'"); }
     arg2 = (SWIGPY_SLICEOBJECT *) swig_obj[1]; }  try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____delitem____SWIG_1(arg1,arg2);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what()); }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___delitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___delitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vvpairUIntUInt___delitem____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vvpairUIntUInt___delitem____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt___delitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__delitem__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__delitem__(SWIGPY_SLICEOBJECT *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___getitem____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___getitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___getitem__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2); try {
    result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *) &std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____getitem____SWIG_1((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1,SWIG_STD_MOVE(arg2));}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = swig::from(static_cast< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___getitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___getitem__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { { _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_1;
    return _wrap_vvpairUIntUInt___getitem____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vvpairUIntUInt___getitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt___getitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__getitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__getitem__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type) const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setitem____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  ptrdiff_t val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt___setitem__" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt___setitem__" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type >(val2); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vvpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt___setitem__" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg3 = ptr; }  try {
    std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg____setitem____SWIG_2(arg1,SWIG_STD_MOVE(arg2),(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg3);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); }  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt___setitem__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt___setitem__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vvpairUIntUInt___setitem____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        _v = PySlice_Check(argv[1]); } }  if (!_v) goto check_2; {
      int res = swig::asptr(argv[2], (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > >**)(0));
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_vvpairUIntUInt___setitem____SWIG_0(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_vvpairUIntUInt___setitem____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt___setitem__'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__setitem__(SWIGPY_SLICEOBJECT *,std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__setitem__(SWIGPY_SLICEOBJECT *)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::__setitem__(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::difference_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_pop(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_pop" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); try {
    result = std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__pop(arg1);}
   catch(std::out_of_range &_e) { SWIG_exception_fail(SWIG_IndexError, (&_e)->what()); } 
  resultobj = swig::from(static_cast< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_append(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_append" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vvpairUIntUInt_append" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_append" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg2 = ptr; } 
  std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__append(arg1,(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vvpairUIntUInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *)new std::vector< std::vector< std::pair< unsigned int,unsigned int > > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vvpairUIntUInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *ptr = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vvpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vvpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > const &""'"); }
     arg1 = ptr; } 
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *)new std::vector< std::vector< std::pair< unsigned int,unsigned int > > >((std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_empty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_empty" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (bool)((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_size" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = ((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_swap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_swap", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_swap" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vvpairUIntUInt_swap" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_swap" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > * >(argp2);
  (arg1)->swap(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_begin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_begin" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_end(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_end" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_rbegin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::reverse_iterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_rbegin" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::reverse_iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_rend(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::reverse_iterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_rend" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::reverse_iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_clear" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); (arg1)->clear();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_get_allocator(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_get_allocator" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = ((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::allocator_type(result)), SWIGTYPE_p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vvpairUIntUInt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg1 ; size_t val1 ;
  int ecode1 = 0 ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vvpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val1);
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *)new std::vector< std::vector< std::pair< unsigned int,unsigned int > > >(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_pop_back(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_pop_back" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); (arg1)->pop_back();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_resize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  size_t val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_resize" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt_resize" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val2);
  (arg1)->resize(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_erase__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ; int res2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_erase" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
   else {
    swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *>(iter2);
    if (iter_t) { arg2 = iter_t->get_current(); }  else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
     } 
  result = std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__erase__SWIG_0(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_erase__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ; int res2 ; swig::SwigPyIterator *iter3 = 0 ; int res3 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_erase" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
   else {
    swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *>(iter2);
    if (iter_t) { arg2 = iter_t->get_current(); }  else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
     }  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
   else {
    swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *>(iter3);
    if (iter_t) { arg3 = iter_t->get_current(); }  else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_erase" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
     } 
  result = std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__erase__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_erase(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_erase", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vvpairUIntUInt_erase__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_vvpairUIntUInt_erase__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt_erase'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::erase(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::erase(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_vvpairUIntUInt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg1 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg2 = 0 ; size_t val1 ; int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vvpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg1 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val1); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vvpairUIntUInt" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vvpairUIntUInt" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg2 = ptr; } 
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *)new std::vector< std::vector< std::pair< unsigned int,unsigned int > > >(SWIG_STD_MOVE(arg1),(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_vvpairUIntUInt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vvpairUIntUInt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vvpairUIntUInt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { {
        int res = SWIG_AsVal_size_t(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_vvpairUIntUInt__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_vvpairUIntUInt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vvpairUIntUInt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vvpairUIntUInt'.\n"
  "  Possible C/C++ prototypes are:\n" "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::vector()\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::vector(std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > const &)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::vector(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::vector(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_push_back(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_push_back", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_push_back" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vvpairUIntUInt_push_back" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_push_back" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg2 = ptr; } 
  (arg1)->push_back((std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_front(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_front" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *) &((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->front();
  resultobj = swig::from(static_cast< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_back(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_back" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = (std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *) &((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->back();
  resultobj = swig::from(static_cast< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > >(*result));
  if (!swig::container_owner<swig::traits<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type>::category>::back_reference(resultobj, swig_obj[0]) && PyErr_Occurred()) {
    SWIG_fail; }  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_assign(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  size_t val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_assign", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_assign" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt_assign" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val2); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vvpairUIntUInt_assign" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_assign" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg3 = ptr; } 
  (arg1)->assign(SWIG_STD_MOVE(arg2),(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_resize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  size_t val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_resize" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt_resize" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val2); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vvpairUIntUInt_resize" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_resize" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg3 = ptr; } 
  (arg1)->resize(SWIG_STD_MOVE(arg2),(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_resize(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_resize", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vvpairUIntUInt_resize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_vvpairUIntUInt_resize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt_resize'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::resize(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::resize(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_insert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ; int res2 ; int res3 = SWIG_OLDOBJ ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_insert" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_insert" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
   else {
    swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *>(iter2);
    if (iter_t) { arg2 = iter_t->get_current(); }  else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_insert" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
     }  {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vvpairUIntUInt_insert" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_insert" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg3 = ptr; } 
  result = std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__insert__SWIG_0(arg1,SWIG_STD_MOVE(arg2),(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_insert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator arg2 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg3 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ; int res2 ; size_t val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_insert" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_insert" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
   else {
    swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator > *>(iter2);
    if (iter_t) { arg2 = iter_t->get_current(); }  else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vvpairUIntUInt_insert" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator""'"); }
     }  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vvpairUIntUInt_insert" "', argument " "3"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg3 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val3); {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vvpairUIntUInt_insert" "', argument " "4"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vvpairUIntUInt_insert" "', argument " "4"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &""'"); }
     arg4 = ptr; } 
  std_vector_Sl_std_vector_Sl_std_pair_Sl_unsigned_SS_int_Sc_unsigned_SS_int_Sg__Sg__Sg__insert__SWIG_1(arg1,SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),(std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg4);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res4)) delete arg4; return resultobj; fail:
  if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_insert(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_insert", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_vvpairUIntUInt_insert__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_vvpairUIntUInt_insert__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vvpairUIntUInt_insert'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::insert(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)\n"
  "    std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::insert(std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::iterator,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type,std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::value_type const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_reserve(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  size_t val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vvpairUIntUInt_reserve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_reserve" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vvpairUIntUInt_reserve" "', argument " "2"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type""'"); }
    arg2 = static_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type >(val2);
  (arg1)->reserve(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vvpairUIntUInt_capacity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::size_type result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vvpairUIntUInt_capacity" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1);
  result = ((std::vector< std::vector< std::pair< unsigned int,unsigned int > > > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_vvpairUIntUInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vvpairUIntUInt" "', argument " "1"" of type '" "std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *""'");  }
   arg1 = reinterpret_cast< std::vector< std::vector< std::pair< unsigned int,unsigned int > > > * >(argp1); delete arg1;
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vvpairUIntUInt_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vvpairUIntUInt_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_toAliasData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::AliasData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toAliasData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::AliasData *)toAliasData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toAngleData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAngleData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toAngleData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBAngleData *)toAngleData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toCommentData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBCommentData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toCommentData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBCommentData *)toCommentData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCommentData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toConformerData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBConformerData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toConformerData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBConformerData *)toConformerData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toExternalBondData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBExternalBondData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toExternalBondData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (OpenBabel::OBExternalBondData *)toExternalBondData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBExternalBondData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toGridData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBGridData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toGridData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBGridData *)toGridData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toMatrixData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBMatrixData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toMatrixData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBMatrixData *)toMatrixData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMatrixData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toNasaThermoData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBNasaThermoData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toNasaThermoData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBNasaThermoData *)toNasaThermoData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBNasaThermoData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toPairData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBPairData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toPairData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBPairData *)toPairData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toPairInteger(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBPairInteger *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toPairInteger" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBPairInteger *)toPairInteger(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_toPairFloatingPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBPairFloatingPoint *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toPairFloatingPoint" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (OpenBabel::OBPairFloatingPoint *)toPairFloatingPoint(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairTemplateT_double_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_toRateData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBRateData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toRateData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBRateData *)toRateData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRateData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toRotamerList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBRotamerList *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toRotamerList" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBRotamerList *)toRotamerList(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toRotationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBRotationData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toRotationData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBRotationData *)toRotationData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotationData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toSerialNums(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSerialNums *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toSerialNums" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBSerialNums *)toSerialNums(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSerialNums, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toSetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSetData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toSetData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBSetData *)toSetData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toSymmetryData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSymmetryData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toSymmetryData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBSymmetryData *)toSymmetryData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toTorsionData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBTorsionData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toTorsionData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBTorsionData *)toTorsionData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toUnitCell(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBUnitCell *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toUnitCell" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBUnitCell *)toUnitCell(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toVectorData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBVectorData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toVectorData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBVectorData *)toVectorData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVectorData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toVibrationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBVibrationData *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toVibrationData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBVibrationData *)toVibrationData(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toVirtualBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBVirtualBond *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toVirtualBond" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBVirtualBond *)toVirtualBond(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVirtualBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toStereoBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereoBase *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toStereoBase" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBStereoBase *)toStereoBase(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoBase, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_toTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBTetrahedralStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (OpenBabel::OBTetrahedralStereo *)toTetrahedralStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBCisTransStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); result = (OpenBabel::OBCisTransStereo *)toCisTransStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_toSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSquarePlanarStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "toSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (OpenBabel::OBSquarePlanarStereo *)toSquarePlanarStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGlobalDataBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBGlobalDataBase", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBGlobalDataBase *)new OpenBabel::OBGlobalDataBase();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGlobalDataBase, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBGlobalDataBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBGlobalDataBase" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGlobalDataBase_Init(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGlobalDataBase_Init" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1); (arg1)->Init(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGlobalDataBase_GetSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGlobalDataBase_GetSize" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1); result = (arg1)->GetSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGlobalDataBase_SetReadDirectory(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGlobalDataBase_SetReadDirectory", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGlobalDataBase_SetReadDirectory" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGlobalDataBase_SetReadDirectory" "', argument " "2"" of type '" "char *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetReadDirectory(arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBGlobalDataBase_SetEnvironmentVariable(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGlobalDataBase_SetEnvironmentVariable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGlobalDataBase_SetEnvironmentVariable" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGlobalDataBase_SetEnvironmentVariable" "', argument " "2"" of type '" "char *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetEnvironmentVariable(arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBGlobalDataBase_ParseLine(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGlobalDataBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGlobalDataBase_ParseLine", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGlobalDataBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGlobalDataBase_ParseLine" "', argument " "1"" of type '" "OpenBabel::OBGlobalDataBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGlobalDataBase * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGlobalDataBase_ParseLine" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->ParseLine((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *OBGlobalDataBase_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBGlobalDataBase, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBGlobalDataBase_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBAtomHOF(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::string arg1 ;
  int arg2 ; std::string arg3 ; std::string arg4 ; double arg5 ; double arg6 ; int arg7 ; std::string arg8 ; int val2 ;
  int ecode2 = 0 ; double val5 ; int ecode5 = 0 ; double val6 ; int ecode6 = 0 ; int val7 ; int ecode7 = 0 ;
  PyObject *swig_obj[8] ; OpenBabel::OBAtomHOF *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBAtomHOF", 8, 8, swig_obj)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBAtomHOF" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBAtomHOF" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); { std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBAtomHOF" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[3], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBAtomHOF" "', argument " "4"" of type '" "std::string""'");  }
     arg4 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode5 = SWIG_AsVal_double(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_OBAtomHOF" "', argument " "5"" of type '" "double""'"); }  
  arg5 = static_cast< double >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_OBAtomHOF" "', argument " "6"" of type '" "double""'"); }  
  arg6 = static_cast< double >(val6); ecode7 = SWIG_AsVal_int(swig_obj[6], &val7); if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_OBAtomHOF" "', argument " "7"" of type '" "int""'"); }  
  arg7 = static_cast< int >(val7); { std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[7], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBAtomHOF" "', argument " "8"" of type '" "std::string""'");  }
     arg8 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBAtomHOF *)new OpenBabel::OBAtomHOF(SWIG_STD_MOVE(arg1),arg2,SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4),arg5,arg6,arg7,SWIG_STD_MOVE(arg8));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomHOF, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAtomHOF(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAtomHOF" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Element(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Element" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (arg1)->Element();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Charge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Charge" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (int)(arg1)->Charge();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Method(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Method" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (arg1)->Method();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Desc(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Desc" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (arg1)->Desc();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_T(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_T" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (double)(arg1)->T();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Value(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Value" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (double)(arg1)->Value();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Multiplicity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Multiplicity" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (int)(arg1)->Multiplicity();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomHOF_Unit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomHOF *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomHOF, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomHOF_Unit" "', argument " "1"" of type '" "OpenBabel::OBAtomHOF *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomHOF * >(argp1); result = (arg1)->Unit();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBAtomHOF_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomHOF, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAtomHOF_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBAtomicHeatOfFormationTable(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomicHeatOfFormationTable *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBAtomicHeatOfFormationTable", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBAtomicHeatOfFormationTable *)new OpenBabel::OBAtomicHeatOfFormationTable();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomicHeatOfFormationTable, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAtomicHeatOfFormationTable(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomicHeatOfFormationTable *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomicHeatOfFormationTable, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAtomicHeatOfFormationTable" "', argument " "1"" of type '" "OpenBabel::OBAtomicHeatOfFormationTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomicHeatOfFormationTable * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomicHeatOfFormationTable_GetHeatOfFormation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBAtomicHeatOfFormationTable *arg1 = 0 ; std::string arg2 ; int arg3 ; std::string arg4 ;
  double arg5 ; double *arg6 = 0 ; double *arg7 = 0 ; double *arg8 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int val3 ;
  int ecode3 = 0 ; double val5 ; int ecode5 = 0 ; void *argp6 = 0 ; int res6 = 0 ; void *argp7 = 0 ; int res7 = 0 ;
  void *argp8 = 0 ; int res8 = 0 ; PyObject *swig_obj[8] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtomicHeatOfFormationTable_GetHeatOfFormation", 8, 8, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomicHeatOfFormationTable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "1"" of type '" "OpenBabel::OBAtomicHeatOfFormationTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomicHeatOfFormationTable * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); { std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "4"" of type '" "std::string""'");  }
     arg4 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode5 = SWIG_AsVal_double(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "6"" of type '" "double *""'");  }
   arg6 = reinterpret_cast< double * >(argp6); res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "7"" of type '" "double *""'");  }
   arg7 = reinterpret_cast< double * >(argp7); res8 = SWIG_ConvertPtr(swig_obj[7], &argp8,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "OBAtomicHeatOfFormationTable_GetHeatOfFormation" "', argument " "8"" of type '" "double *""'");  }
   arg8 = reinterpret_cast< double * >(argp8);
  result = (int)(arg1)->GetHeatOfFormation(SWIG_STD_MOVE(arg2),arg3,SWIG_STD_MOVE(arg4),arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBAtomicHeatOfFormationTable_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomicHeatOfFormationTable, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAtomicHeatOfFormationTable_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBTypeTable(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBTypeTable", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBTypeTable *)new OpenBabel::OBTypeTable();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTypeTable, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBTypeTable(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTypeTable" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_SetFromType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTypeTable_SetFromType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_SetFromType" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTypeTable_SetFromType" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->SetFromType((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_SetToType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTypeTable_SetToType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_SetToType" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTypeTable_SetToType" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->SetToType((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_Translate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTypeTable *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_Translate" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTypeTable_Translate" "', argument " "2"" of type '" "char *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBTypeTable_Translate" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->Translate(arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_Translate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTypeTable *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_Translate" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTypeTable_Translate" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTypeTable_Translate" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); { std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBTypeTable_Translate" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTypeTable_Translate" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  result = (bool)(arg1)->Translate(*arg2,(std::string const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_Translate__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTypeTable *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; std::string result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_Translate" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTypeTable_Translate" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTypeTable_Translate" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (arg1)->Translate((std::string const &)*arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj;
  fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_Translate(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBTypeTable_Translate", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBTypeTable_Translate__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__string, SWIG_POINTER_NO_NULL); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; { int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_OBTypeTable_Translate__SWIG_1(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_OBTypeTable_Translate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBTypeTable_Translate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBTypeTable::Translate(char *,char const *)\n"
  "    OpenBabel::OBTypeTable::Translate(std::string &,std::string const &)\n"
  "    OpenBabel::OBTypeTable::Translate(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBTypeTable_GetFromType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_GetFromType" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); result = (arg1)->GetFromType();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTypeTable_GetToType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTypeTable *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTypeTable, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTypeTable_GetToType" "', argument " "1"" of type '" "OpenBabel::OBTypeTable *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTypeTable * >(argp1); result = (arg1)->GetToType();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBTypeTable_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTypeTable, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBTypeTable_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_ttab_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBTypeTable,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::ttab""' of type '""OpenBabel::OBTypeTable""'"); } 
    if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::ttab""' of type '""OpenBabel::OBTypeTable""'"); }
     else { OpenBabel::OBTypeTable * temp; temp  = reinterpret_cast< OpenBabel::OBTypeTable * >(argp); OpenBabel::ttab = *temp;
      if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_ttab_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::ttab), SWIGTYPE_p_OpenBabel__OBTypeTable,  0 ); return pyobj; }
SWIGINTERN PyObject *_wrap_new_OBResidueData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBResidueData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBResidueData *)new OpenBabel::OBResidueData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidueData_SetResName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidueData_SetResName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidueData_SetResName" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidueData_SetResName" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_SetResName" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (bool)(arg1)->SetResName((std::string const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidueData_LookupBO__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidueData_LookupBO" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidueData_LookupBO" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupBO" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (int)(arg1)->LookupBO((std::string const &)*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidueData_LookupBO__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueData *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidueData_LookupBO" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidueData_LookupBO" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupBO" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBResidueData_LookupBO" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupBO" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  result = (int)(arg1)->LookupBO((std::string const &)*arg2,(std::string const &)*arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidueData_LookupBO(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBResidueData_LookupBO", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBResidueData_LookupBO__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBResidueData_LookupBO__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBResidueData_LookupBO'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidueData::LookupBO(std::string const &)\n"
  "    OpenBabel::OBResidueData::LookupBO(std::string const &,std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBResidueData_LookupType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueData *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; int *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 = SWIG_OLDOBJ ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  PyObject *swig_obj[4] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidueData_LookupType", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidueData_LookupType" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidueData_LookupType" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupType" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBResidueData_LookupType" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupType" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBResidueData_LookupType" "', argument " "4"" of type '" "int &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_LookupType" "', argument " "4"" of type '" "int &""'"); }
   arg4 = reinterpret_cast< int * >(argp4); result = (bool)(arg1)->LookupType((std::string const &)*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidueData_AssignBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueData *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidueData_AssignBonds", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidueData_AssignBonds" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidueData_AssignBonds" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidueData_AssignBonds" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->AssignBonds(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBResidueData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBResidueData" "', argument " "1"" of type '" "OpenBabel::OBResidueData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBResidueData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBResidueData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBResidueData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_resdat_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBResidueData,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::resdat""' of type '""OpenBabel::OBResidueData""'"); } 
    if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::resdat""' of type '""OpenBabel::OBResidueData""'"); }
     else { OpenBabel::OBResidueData * temp; temp  = reinterpret_cast< OpenBabel::OBResidueData * >(argp);
      OpenBabel::resdat = *temp; if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_resdat_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::resdat), SWIGTYPE_p_OpenBabel__OBResidueData,  0 ); return pyobj; }
SWIGINTERN PyObject *_wrap_OBStopwatch_Start(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStopwatch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStopwatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStopwatch_Start" "', argument " "1"" of type '" "OpenBabel::OBStopwatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStopwatch * >(argp1); (arg1)->Start(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStopwatch_Lap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStopwatch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStopwatch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStopwatch_Lap" "', argument " "1"" of type '" "OpenBabel::OBStopwatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStopwatch * >(argp1); result = (double)(arg1)->Lap();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStopwatch_Elapsed(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStopwatch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStopwatch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStopwatch_Elapsed" "', argument " "1"" of type '" "OpenBabel::OBStopwatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStopwatch * >(argp1); result = (double)(arg1)->Elapsed();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStopwatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStopwatch *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBStopwatch", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBStopwatch *)new OpenBabel::OBStopwatch();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStopwatch, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBStopwatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStopwatch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStopwatch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStopwatch" "', argument " "1"" of type '" "OpenBabel::OBStopwatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStopwatch * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBStopwatch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStopwatch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBStopwatch_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSqrtTbl__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBSqrtTbl *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBSqrtTbl *)new OpenBabel::OBSqrtTbl();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSqrtTbl, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSqrtTbl__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double arg2 ; double val1 ; int ecode1 = 0 ; double val2 ; int ecode2 = 0 ;
  OpenBabel::OBSqrtTbl *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBSqrtTbl" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBSqrtTbl" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); result = (OpenBabel::OBSqrtTbl *)new OpenBabel::OBSqrtTbl(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSqrtTbl, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSqrtTbl(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBSqrtTbl", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBSqrtTbl__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_new_OBSqrtTbl__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBSqrtTbl'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSqrtTbl::OBSqrtTbl()\n"
  "    OpenBabel::OBSqrtTbl::OBSqrtTbl(double const,double const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBSqrtTbl(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSqrtTbl *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSqrtTbl, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSqrtTbl" "', argument " "1"" of type '" "OpenBabel::OBSqrtTbl *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSqrtTbl * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBSqrtTbl_Sqrt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSqrtTbl *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSqrtTbl_Sqrt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSqrtTbl, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSqrtTbl_Sqrt" "', argument " "1"" of type '" "OpenBabel::OBSqrtTbl const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSqrtTbl * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSqrtTbl_Sqrt" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); result = (double)((OpenBabel::OBSqrtTbl const *)arg1)->Sqrt(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSqrtTbl_Init(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSqrtTbl *arg1 = 0 ; double arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSqrtTbl_Init", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSqrtTbl, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSqrtTbl_Init" "', argument " "1"" of type '" "OpenBabel::OBSqrtTbl *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSqrtTbl * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSqrtTbl_Init" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSqrtTbl_Init" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); (arg1)->Init(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBSqrtTbl_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSqrtTbl, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSqrtTbl_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_rotate_coords(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double (*arg2)[3] = (double (*)[3])0 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "rotate_coords", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotate_coords" "', argument " "1"" of type '" "double *""'");  } 
  arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_a_3__double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "rotate_coords" "', argument " "2"" of type '" "double [3][3]""'");  }
    arg2 = reinterpret_cast< double (*)[3] >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "rotate_coords" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); OpenBabel::rotate_coords(arg1,(double (*)[3])arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_calc_rms(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ;
  int ecode3 = 0 ; PyObject *swig_obj[3] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "calc_rms", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "calc_rms" "', argument " "1"" of type '" "double *""'");  } 
  arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "calc_rms" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "calc_rms" "', argument " "3"" of type '" "unsigned int""'"); }  
  arg3 = static_cast< unsigned int >(val3); result = (double)OpenBabel::calc_rms(arg1,arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  double arg1 ; double arg2 ; double arg3 ; double val1 ; int ecode1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ;
  int ecode3 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector3" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector3" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_vector3" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); result = (OpenBabel::vector3 *)new OpenBabel::vector3(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  double arg1 ; double arg2 ; double val1 ; int ecode1 = 0 ; double val2 ; int ecode2 = 0 ; OpenBabel::vector3 *result = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector3" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector3" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); result = (OpenBabel::vector3 *)new OpenBabel::vector3(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  double arg1 ; double val1 ; int ecode1 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector3" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); result = (OpenBabel::vector3 *)new OpenBabel::vector3(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::vector3 *)new OpenBabel::vector3();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  double *arg1 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector3" "', argument " "1"" of type '" "double [3]""'");  }  
  arg1 = reinterpret_cast< double * >(argp1); result = (OpenBabel::vector3 *)new OpenBabel::vector3(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector3" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_vector3" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  result = (OpenBabel::vector3 *)new OpenBabel::vector3((OpenBabel::vector3 const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_vector3(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vector3", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_vector3__SWIG_3(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_double, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_new_vector3__SWIG_4(self, argc, argv);}  check_2:
  if (argc == 1) { int _v = 0; { int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap_new_vector3__SWIG_5(self, argc, argv);}  check_3:
  if (argc == 1) { PyObject *retobj = _wrap_new_vector3__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_vector3__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_new_vector3__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vector3'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::vector3(double const,double const,double const)\n"
  "    OpenBabel::vector3::vector3(double const,double const)\n" "    OpenBabel::vector3::vector3(double const)\n"
  "    OpenBabel::vector3::vector3()\n" "    OpenBabel::vector3::vector3(double [3])\n"
  "    OpenBabel::vector3::vector3(OpenBabel::vector3 const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_vector3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector3" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vector3_begin__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3::iterator result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_begin" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (OpenBabel::vector3::iterator)(arg1)->begin();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_end__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3::iterator result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_end" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (OpenBabel::vector3::iterator)(arg1)->end();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_begin__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::vector3::const_iterator result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_begin" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  result = (OpenBabel::vector3::const_iterator)((OpenBabel::vector3 const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_begin(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_begin", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_vector3_begin__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_vector3_begin__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_begin'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::begin()\n" "    OpenBabel::vector3::begin() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_vector3_end__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3::const_iterator result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_end" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  result = (OpenBabel::vector3::const_iterator)((OpenBabel::vector3 const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_end(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_end", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_vector3_end__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_vector3_end__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_end'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::end()\n" "    OpenBabel::vector3::end() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3_Set__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_Set" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3_Set" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector3_Set" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vector3_Set" "', argument " "4"" of type '" "double""'"); }  
  arg4 = static_cast< double >(val4); (arg1)->Set(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vector3_Set__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_Set" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3_Set" "', argument " "2"" of type '" "double const *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->Set((double const *)arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_Set(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_Set", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_vector3_Set__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 4) { PyObject *retobj = _wrap_vector3_Set__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_Set'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::Set(double const,double const,double const)\n"
  "    OpenBabel::vector3::Set(double const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3_SetX(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "vector3_SetX", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_SetX" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3_SetX" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetX(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_SetY(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "vector3_SetY", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_SetY" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3_SetY" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetY(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_SetZ(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "vector3_SetZ", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_SetZ" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3_SetZ" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetZ(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_GetX(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_GetX" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double)((OpenBabel::vector3 const *)arg1)->GetX();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_GetY(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_GetY" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double)((OpenBabel::vector3 const *)arg1)->GetY();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_GetZ(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_GetZ" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double)((OpenBabel::vector3 const *)arg1)->GetZ();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_Get(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3_Get", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_Get" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3_Get" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); (arg1)->Get(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vector3_AsArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_AsArray" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *)((OpenBabel::vector3 const *)arg1)->AsArray();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___iadd____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___iadd__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___iadd__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3___iadd__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (OpenBabel::vector3 *) &(arg1)->operator +=((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___isub____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___isub__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___isub__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3___isub__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (OpenBabel::vector3 *) &(arg1)->operator -=((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___iadd____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___iadd__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___iadd__" "', argument " "2"" of type '" "double const *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (OpenBabel::vector3 *) &(arg1)->operator +=((double const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___iadd__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3___iadd__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_vector3___iadd____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vector3___iadd____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3___iadd__'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::operator +=(OpenBabel::vector3 const &)\n"
  "    OpenBabel::vector3::operator +=(double const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3___isub____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___isub__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___isub__" "', argument " "2"" of type '" "double const *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (OpenBabel::vector3 *) &(arg1)->operator -=((double const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___isub__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3___isub__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_vector3___isub____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_vector3___isub____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3___isub__'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::operator -=(OpenBabel::vector3 const &)\n"
  "    OpenBabel::vector3::operator -=(double const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3___imul____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ;
  double val2 ; int ecode2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___imul__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3___imul__" "', argument " "2"" of type '" "double""'"); }  
  temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = (OpenBabel::vector3 *) &(arg1)->operator *=((double const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___itruediv__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ; double val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::vector3 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3___itruediv__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___itruediv__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector3___itruediv__" "', argument " "2"" of type '" "double""'"); }
    temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = (OpenBabel::vector3 *) &(arg1)->operator /=((double const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___imul____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::matrix3x3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::vector3 *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___imul__" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___imul__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3___imul__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp2);
  result = (OpenBabel::vector3 *) &(arg1)->operator *=((OpenBabel::matrix3x3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3___imul__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3___imul__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_vector3___imul____SWIG_1(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_vector3___imul____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3___imul__'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::operator *=(double const &)\n"
  "    OpenBabel::vector3::operator *=(OpenBabel::matrix3x3 const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3_randomUnitVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_randomUnitVector" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); (arg1)->randomUnitVector(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_normalize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_normalize" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (OpenBabel::vector3 *) &(arg1)->normalize();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_vector3_CanBeNormalized(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_CanBeNormalized" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (bool)((OpenBabel::vector3 const *)arg1)->CanBeNormalized();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_length_2(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_length_2" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double)((OpenBabel::vector3 const *)arg1)->length_2();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_length(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_length" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double)((OpenBabel::vector3 const *)arg1)->length();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_x__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_x" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &((OpenBabel::vector3 const *)arg1)->x();
  resultobj = SWIG_From_double(static_cast< double >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_y__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_y" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &((OpenBabel::vector3 const *)arg1)->y();
  resultobj = SWIG_From_double(static_cast< double >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_z__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_z" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &((OpenBabel::vector3 const *)arg1)->z();
  resultobj = SWIG_From_double(static_cast< double >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_x__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_x" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &(arg1)->x();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_x(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_x", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_vector3_x__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_vector3_x__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_x'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::x() const\n" "    OpenBabel::vector3::x()\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3_y__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_y" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &(arg1)->y();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_y(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_y", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_vector3_y__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_vector3_y__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_y'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::y() const\n" "    OpenBabel::vector3::y()\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3_z__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_z" "', argument " "1"" of type '" "OpenBabel::vector3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = (double *) &(arg1)->z();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_z(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "vector3_z", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_vector3_z__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_vector3_z__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vector3_z'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::vector3::z() const\n" "    OpenBabel::vector3::z()\n"); return 0; }
SWIGINTERN PyObject *_wrap_vector3___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___eq__" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___eq__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3___eq__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (bool)((OpenBabel::vector3 const *)arg1)->operator ==((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_vector3___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3___ne__" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3___ne__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3___ne__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (bool)((OpenBabel::vector3 const *)arg1)->operator !=((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_vector3_IsApprox(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double temp3 ; double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "vector3_IsApprox", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_IsApprox" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3_IsApprox" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3_IsApprox" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector3_IsApprox" "', argument " "3"" of type '" "double""'"); }  
  temp3 = static_cast< double >(val3); arg3 = &temp3;
  result = (bool)((OpenBabel::vector3 const *)arg1)->IsApprox((OpenBabel::vector3 const &)*arg2,(double const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_distSq(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3_distSq", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_distSq" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3_distSq" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3_distSq" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (double)((OpenBabel::vector3 const *)arg1)->distSq((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vector3_createOrthoVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "vector3_createOrthoVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector3_createOrthoVector" "', argument " "1"" of type '" "OpenBabel::vector3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector3_createOrthoVector" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vector3_createOrthoVector" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (bool)((OpenBabel::vector3 const *)arg1)->createOrthoVector(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *vector3_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__vector3, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vector3_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap___lshift____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::ostream *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::ostream *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__ostream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'"); }
   arg1 = reinterpret_cast< std::ostream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (std::ostream *) &OpenBabel::operator <<(*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___add__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::vector3 result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__add__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__add__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__add__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__add__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__add__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = OpenBabel::operator +((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___sub____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__sub__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__sub__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = OpenBabel::operator -((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___sub____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); result = OpenBabel::operator -((OpenBabel::vector3 const &)*arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___mul____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  double *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; double temp1 ; double val1 ; int ecode1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "__mul__" "', argument " "1"" of type '" "double""'"); }  
  temp1 = static_cast< double >(val1); arg1 = &temp1;
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = OpenBabel::operator *((double const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___mul____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ; double val2 ;
  int ecode2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "__mul__" "', argument " "2"" of type '" "double""'"); }  
  temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = OpenBabel::operator *((OpenBabel::vector3 const &)*arg1,(double const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___truediv__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ; double val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::vector3 result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__truediv__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__truediv__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__truediv__" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "__truediv__" "', argument " "2"" of type '" "double""'"); }  
  temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = OpenBabel::operator /((OpenBabel::vector3 const &)*arg1,(double const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___mul____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = OpenBabel::operator *((OpenBabel::matrix3x3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_dot(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  double result; (void)self; if (!SWIG_Python_UnpackTuple(args, "dot", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "dot" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "dot" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "dot" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "dot" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (double)OpenBabel::dot((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_cross(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::vector3 result; (void)self; if (!SWIG_Python_UnpackTuple(args, "cross", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "cross" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "cross" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "cross" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "cross" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = OpenBabel::cross((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_vectorAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  double result; (void)self; if (!SWIG_Python_UnpackTuple(args, "vectorAngle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "vectorAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (double)OpenBabel::vectorAngle((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CalcTorsionAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ; OpenBabel::vector3 *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CalcTorsionAngle", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CalcTorsionAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CalcTorsionAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CalcTorsionAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CalcTorsionAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CalcTorsionAngle" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CalcTorsionAngle" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CalcTorsionAngle" "', argument " "4"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CalcTorsionAngle" "', argument " "4"" of type '" "OpenBabel::vector3 const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4);
  result = (double)OpenBabel::CalcTorsionAngle((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2,(OpenBabel::vector3 const &)*arg3,(OpenBabel::vector3 const &)*arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_Point2PlaneSigned(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 arg1 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ; OpenBabel::vector3 arg4 ; void *argp1 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; void *argp4 ; int res4 = 0 ; PyObject *swig_obj[4] ;
  double result; (void)self; if (!SWIG_Python_UnpackTuple(args, "Point2PlaneSigned", 4, 4, swig_obj)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Point2PlaneSigned" "', argument " "1"" of type '" "OpenBabel::vector3""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneSigned" "', argument " "1"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp1); arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp; }  }  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Point2PlaneSigned" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneSigned" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Point2PlaneSigned" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneSigned" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Point2PlaneSigned" "', argument " "4"" of type '" "OpenBabel::vector3""'");  }
       if (!argp4) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneSigned" "', argument " "4"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp4); arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp; }  } 
  result = (double)OpenBabel::Point2PlaneSigned(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4));
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_Point2Plane(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 arg1 ;
  OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ; OpenBabel::vector3 arg4 ; void *argp1 ; int res1 = 0 ; void *argp2 ;
  int res2 = 0 ; void *argp3 ; int res3 = 0 ; void *argp4 ; int res4 = 0 ; PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Point2Plane", 4, 4, swig_obj)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Point2Plane" "', argument " "1"" of type '" "OpenBabel::vector3""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Plane" "', argument " "1"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp1); arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp; }  }  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Point2Plane" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Plane" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Point2Plane" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Plane" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Point2Plane" "', argument " "4"" of type '" "OpenBabel::vector3""'");  }
       if (!argp4) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Plane" "', argument " "4"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp4); arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp; }  } 
  result = (double)OpenBabel::Point2Plane(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4));
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_Point2PlaneAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 arg1 ;
  OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ; OpenBabel::vector3 arg4 ; void *argp1 ; int res1 = 0 ; void *argp2 ;
  int res2 = 0 ; void *argp3 ; int res3 = 0 ; void *argp4 ; int res4 = 0 ; PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Point2PlaneAngle", 4, 4, swig_obj)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Point2PlaneAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneAngle" "', argument " "1"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp1); arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp; }  }  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Point2PlaneAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Point2PlaneAngle" "', argument " "3"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneAngle" "', argument " "3"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Point2PlaneAngle" "', argument " "4"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp4) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2PlaneAngle" "', argument " "4"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp4); arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp; }  } 
  result = (double)OpenBabel::Point2PlaneAngle(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4));
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_Point2Line(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ;
  OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Point2Line", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Point2Line" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Line" "', argument " "1"" of type '" "OpenBabel::vector3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Point2Line" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Line" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Point2Line" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Point2Line" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  result = (double)OpenBabel::Point2Line((OpenBabel::vector3 const &)*arg1,(OpenBabel::vector3 const &)*arg2,(OpenBabel::vector3 const &)*arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN int Swig_var_VZero_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable VZero is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_VZero_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::VZero), SWIGTYPE_p_OpenBabel__vector3,  0 ); return pyobj; }
SWIGINTERN int Swig_var_VX_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable VX is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_VX_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::VX), SWIGTYPE_p_OpenBabel__vector3,  0 ); return pyobj; }
SWIGINTERN int Swig_var_VY_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable VY is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_VY_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::VY), SWIGTYPE_p_OpenBabel__vector3,  0 ); return pyobj; }
SWIGINTERN int Swig_var_VZ_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable VZ is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_VZ_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::VZ), SWIGTYPE_p_OpenBabel__vector3,  0 ); return pyobj; }
SWIGINTERN PyObject *_wrap_new_matrix3x3__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::matrix3x3 *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::matrix3x3 *)new OpenBabel::matrix3x3();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_matrix3x3__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double val1 ; int ecode1 = 0 ; OpenBabel::matrix3x3 *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_matrix3x3" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); result = (OpenBabel::matrix3x3 *)new OpenBabel::matrix3x3(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_matrix3x3__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 arg1 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ; void *argp1 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; OpenBabel::matrix3x3 *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_matrix3x3" "', argument " "1"" of type '" "OpenBabel::vector3""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_matrix3x3" "', argument " "1"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp1); arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp; }  }  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_matrix3x3" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_matrix3x3" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_matrix3x3" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_matrix3x3" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  } 
  result = (OpenBabel::matrix3x3 *)new OpenBabel::matrix3x3(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_matrix3x3__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double (*arg1)[3] = (double (*)[3])0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::matrix3x3 *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_a_3__double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_matrix3x3" "', argument " "1"" of type '" "double [3][3]""'");  }
    arg1 = reinterpret_cast< double (*)[3] >(argp1);
  result = (OpenBabel::matrix3x3 *)new OpenBabel::matrix3x3((double (*)[3])arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_matrix3x3(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_matrix3x3", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_matrix3x3__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_a_3__double, 0); _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_new_matrix3x3__SWIG_3(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_matrix3x3__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_new_matrix3x3__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_matrix3x3'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::matrix3x3::matrix3x3()\n" "    OpenBabel::matrix3x3::matrix3x3(double)\n"
  "    OpenBabel::matrix3x3::matrix3x3(OpenBabel::vector3,OpenBabel::vector3,OpenBabel::vector3)\n"
  "    OpenBabel::matrix3x3::matrix3x3(double [3][3])\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_matrix3x3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_matrix3x3" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_GetArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "matrix3x3_GetArray", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_GetArray" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "matrix3x3_GetArray" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->GetArray(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3___call____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; double *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3___call__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3___call__" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3___call__" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (double *) &((OpenBabel::matrix3x3 const *)arg1)->operator ()(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3___call____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; double *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3___call__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3___call__" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3___call__" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (double *) &(arg1)->operator ()(arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3___call__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "matrix3x3___call__", 0, 3, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_matrix3x3___call____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_matrix3x3___call____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'matrix3x3___call__'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::matrix3x3::operator ()(int,int) const\n"
  "    OpenBabel::matrix3x3::operator ()(int,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_matrix3x3_inverse(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::matrix3x3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_inverse" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); result = ((OpenBabel::matrix3x3 const *)arg1)->inverse();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_transpose(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::matrix3x3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_transpose" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); result = ((OpenBabel::matrix3x3 const *)arg1)->transpose();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_determinant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_determinant" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  result = (double)((OpenBabel::matrix3x3 const *)arg1)->determinant();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_isSymmetric(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_isSymmetric" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); result = (bool)((OpenBabel::matrix3x3 const *)arg1)->isSymmetric();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_isOrthogonal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_isOrthogonal" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  result = (bool)((OpenBabel::matrix3x3 const *)arg1)->isOrthogonal(); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_isDiagonal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_isDiagonal" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); result = (bool)((OpenBabel::matrix3x3 const *)arg1)->isDiagonal();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_isUnitMatrix(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_isUnitMatrix" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  result = (bool)((OpenBabel::matrix3x3 const *)arg1)->isUnitMatrix(); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_Get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ;
  int ecode3 = 0 ; PyObject *swig_obj[3] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_Get", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_Get" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_Get" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3_Get" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (double)((OpenBabel::matrix3x3 const *)arg1)->Get(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_Set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; int arg3 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_Set", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_Set" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_Set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3_Set" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "matrix3x3_Set" "', argument " "4"" of type '" "double""'"); }  
  arg4 = static_cast< double >(val4); (arg1)->Set(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_SetColumn(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; OpenBabel::vector3 *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_SetColumn", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_SetColumn" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_SetColumn" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "matrix3x3_SetColumn" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "matrix3x3_SetColumn" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3); (arg1)->SetColumn(arg2,(OpenBabel::vector3 const &)*arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_SetRow(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; int arg2 ; OpenBabel::vector3 *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_SetRow", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_SetRow" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_SetRow" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "matrix3x3_SetRow" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "matrix3x3_SetRow" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3); (arg1)->SetRow(arg2,(OpenBabel::vector3 const &)*arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_GetColumn(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_GetColumn", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_GetColumn" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_GetColumn" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = ((OpenBabel::matrix3x3 const *)arg1)->GetColumn(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_GetRow(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_GetRow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_GetRow" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_GetRow" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = ((OpenBabel::matrix3x3 const *)arg1)->GetRow(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3___imul__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ; double val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::matrix3x3 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3___imul__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3___imul__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3___imul__" "', argument " "2"" of type '" "double""'"); }
    temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = (OpenBabel::matrix3x3 *) &(arg1)->operator *=((double const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3___itruediv__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double temp2 ; double val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::matrix3x3 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3___itruediv__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3___itruediv__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3___itruediv__" "', argument " "2"" of type '" "double""'"); }
    temp2 = static_cast< double >(val2); arg2 = &temp2;
  result = (OpenBabel::matrix3x3 *) &(arg1)->operator /=((double const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_SetupRotMat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_SetupRotMat", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_SetupRotMat" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_SetupRotMat" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3_SetupRotMat" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "matrix3x3_SetupRotMat" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); (arg1)->SetupRotMat(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_PlaneReflection(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_PlaneReflection", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_PlaneReflection" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "matrix3x3_PlaneReflection" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "matrix3x3_PlaneReflection" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); (arg1)->PlaneReflection((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_RotAboutAxisByAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_RotAboutAxisByAngle", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_RotAboutAxisByAngle" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "matrix3x3_RotAboutAxisByAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "matrix3x3_RotAboutAxisByAngle" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3_RotAboutAxisByAngle" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->RotAboutAxisByAngle((OpenBabel::vector3 const &)*arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_FillOrth(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; double arg5 ; double arg6 ; double arg7 ;
  void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ;
  double val5 ; int ecode5 = 0 ; double val6 ; int ecode6 = 0 ; double val7 ; int ecode7 = 0 ; PyObject *swig_obj[7] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "matrix3x3_FillOrth", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_FillOrth" "', argument " "1"" of type '" "OpenBabel::matrix3x3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "matrix3x3_FillOrth" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "matrix3x3_FillOrth" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "matrix3x3_FillOrth" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "matrix3x3_FillOrth" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "matrix3x3_FillOrth" "', argument " "6"" of type '" "double""'"); }
    arg6 = static_cast< double >(val6); ecode7 = SWIG_AsVal_double(swig_obj[6], &val7); if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "matrix3x3_FillOrth" "', argument " "7"" of type '" "double""'"); }
    arg7 = static_cast< double >(val7); (arg1)->FillOrth(arg2,arg3,arg4,arg5,arg6,arg7); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_matrix3x3_findEigenvectorsIfSymmetric(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::matrix3x3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_findEigenvectorsIfSymmetric", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__matrix3x3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "matrix3x3_findEigenvectorsIfSymmetric" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "matrix3x3_findEigenvectorsIfSymmetric" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "matrix3x3_findEigenvectorsIfSymmetric" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = ((OpenBabel::matrix3x3 const *)arg1)->findEigenvectorsIfSymmetric(*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap___mul____SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::matrix3x3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::matrix3x3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__mul__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp2);
  result = operator *((OpenBabel::matrix3x3 const &)*arg1,(OpenBabel::matrix3x3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___mul__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "__mul__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; { { int res = SWIG_AsVal_double(argv[1], NULL); _v = SWIG_CheckState(res); } } 
    if (!_v) goto check_1; return _wrap___mul____SWIG_1(self, argc, argv);}  check_1: if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_2; return _wrap___mul____SWIG_2(self, argc, argv);}  check_2: if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap___mul____SWIG_3(self, argc, argv);}  check_3:
  if (argc == 2) { PyObject *retobj = _wrap___mul____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail: SWIG_Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___lshift____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::ostream *arg1 = 0 ; OpenBabel::matrix3x3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::ostream *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__ostream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'"); }
   arg1 = reinterpret_cast< std::ostream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp2);
  result = (std::ostream *) &operator <<(*arg1,(OpenBabel::matrix3x3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_matrix3x3_jacobi(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; unsigned int val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "matrix3x3_jacobi", 4, 4, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "matrix3x3_jacobi" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "matrix3x3_jacobi" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "matrix3x3_jacobi" "', argument " "3"" of type '" "double *""'");  } 
  arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "matrix3x3_jacobi" "', argument " "4"" of type '" "double *""'");  } 
  arg4 = reinterpret_cast< double * >(argp4); OpenBabel::matrix3x3::jacobi(arg1,arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *matrix3x3_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *matrix3x3_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_transform3d__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::transform3d *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::transform3d *)new OpenBabel::transform3d();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_transform3d__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::matrix3x3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::transform3d *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_transform3d" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "1"" of type '" "OpenBabel::matrix3x3 const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::matrix3x3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_transform3d" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (OpenBabel::transform3d *)new OpenBabel::transform3d((OpenBabel::matrix3x3 const &)*arg1,(OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_transform3d__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double val1 ; int ecode1 = 0 ; OpenBabel::transform3d *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_transform3d" "', argument " "1"" of type '" "double""'"); }  
  arg1 = static_cast< double >(val1); result = (OpenBabel::transform3d *)new OpenBabel::transform3d(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_transform3d__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 arg1 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  OpenBabel::vector3 arg4 ; void *argp1 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; void *argp4 ;
  int res4 = 0 ; OpenBabel::transform3d *result = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_transform3d" "', argument " "1"" of type '" "OpenBabel::vector3""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "1"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp1); arg1 = *temp;
      if (SWIG_IsNewObj(res1)) delete temp; }  }  {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_transform3d" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_transform3d" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_transform3d" "', argument " "4"" of type '" "OpenBabel::vector3""'");  }
       if (!argp4) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_transform3d" "', argument " "4"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp4); arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp; }  } 
  result = (OpenBabel::transform3d *)new OpenBabel::transform3d(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_transform3d__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double (*arg1)[3] = (double (*)[3])0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::transform3d *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_a_3__double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_transform3d" "', argument " "1"" of type '" "double [3][3]""'");  }
    arg1 = reinterpret_cast< double (*)[3] >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_transform3d" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2);
  result = (OpenBabel::transform3d *)new OpenBabel::transform3d((double (*)[3])arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_transform3d(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_transform3d", 0, 4, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_transform3d__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_transform3d__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_3; return _wrap_new_transform3d__SWIG_1(self, argc, argv);}  check_3: if (argc == 2) {
    PyObject *retobj = _wrap_new_transform3d__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_transform3d__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_transform3d'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::transform3d::transform3d()\n"
  "    OpenBabel::transform3d::transform3d(OpenBabel::matrix3x3 const &,OpenBabel::vector3 const &)\n"
  "    OpenBabel::transform3d::transform3d(double)\n"
  "    OpenBabel::transform3d::transform3d(OpenBabel::vector3,OpenBabel::vector3,OpenBabel::vector3,OpenBabel::vector3)\n"
  "    OpenBabel::transform3d::transform3d(double [3][3],double [3])\n"); return 0; }
SWIGINTERN PyObject *_wrap_transform3d___mul____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::transform3d *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "transform3d___mul__" "', argument " "1"" of type '" "OpenBabel::transform3d const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "transform3d___mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "transform3d___mul__" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = ((OpenBabel::transform3d const *)arg1)->operator *((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_transform3d___mul____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::transform3d *arg1 = 0 ; OpenBabel::transform3d *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; OpenBabel::transform3d result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "transform3d___mul__" "', argument " "1"" of type '" "OpenBabel::transform3d const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__transform3d,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "transform3d___mul__" "', argument " "2"" of type '" "OpenBabel::transform3d const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "transform3d___mul__" "', argument " "2"" of type '" "OpenBabel::transform3d const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::transform3d * >(argp2);
  result = ((OpenBabel::transform3d const *)arg1)->operator *((OpenBabel::transform3d const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::transform3d(result)), SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_transform3d___mul__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "transform3d___mul__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_transform3d___mul____SWIG_1(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_transform3d___mul____SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail: SWIG_Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_transform3d_DescribeAsString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::transform3d *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "transform3d_DescribeAsString" "', argument " "1"" of type '" "OpenBabel::transform3d const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1);
  result = ((OpenBabel::transform3d const *)arg1)->DescribeAsString();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_transform3d_DescribeAsValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::transform3d *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "transform3d_DescribeAsValues" "', argument " "1"" of type '" "OpenBabel::transform3d const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1);
  result = ((OpenBabel::transform3d const *)arg1)->DescribeAsValues();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_transform3d_Normalize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::transform3d *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "transform3d_Normalize" "', argument " "1"" of type '" "OpenBabel::transform3d *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1); (arg1)->Normalize(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_transform3d(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::transform3d *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__transform3d, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_transform3d" "', argument " "1"" of type '" "OpenBabel::transform3d *""'");  }
   arg1 = reinterpret_cast< OpenBabel::transform3d * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *transform3d_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__transform3d, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *transform3d_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_SpaceGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_SpaceGroup", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::SpaceGroup *)new OpenBabel::SpaceGroup();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_SpaceGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SpaceGroup" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_SetHMName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_SetHMName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_SetHMName" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_SetHMName" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetHMName((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_SetHallName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ;
  int alloc2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_SetHallName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_SetHallName" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_SetHallName" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetHallName((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_SetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_SetId", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_SetId" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SpaceGroup_SetId" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetId(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_AddTransform(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_AddTransform", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_AddTransform" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_AddTransform" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup_AddTransform" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->AddTransform((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetHMName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetHMName" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  result = (std::string *) &((OpenBabel::SpaceGroup const *)arg1)->GetHMName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetHallName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetHallName" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  result = (std::string *) &((OpenBabel::SpaceGroup const *)arg1)->GetHallName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetId" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  result = (unsigned int)((OpenBabel::SpaceGroup const *)arg1)->GetId();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetOriginAlternative(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetOriginAlternative" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  result = (unsigned int)((OpenBabel::SpaceGroup const *)arg1)->GetOriginAlternative();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_Transform(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ;
  SwigValueWrapper< std::list< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_Transform", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_Transform" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_Transform" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup_Transform" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = ((OpenBabel::SpaceGroup const *)arg1)->Transform((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::list< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >(result)), SWIGTYPE_p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_BeginTransform(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; OpenBabel::transform3dIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::transform3d *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_BeginTransform", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_BeginTransform" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_BeginTransform" "', argument " "2"" of type '" "OpenBabel::transform3dIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup_BeginTransform" "', argument " "2"" of type '" "OpenBabel::transform3dIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::transform3dIterator * >(argp2);
  result = (OpenBabel::transform3d *)((OpenBabel::SpaceGroup const *)arg1)->BeginTransform(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_NextTransform(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; OpenBabel::transform3dIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::transform3d *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup_NextTransform", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_NextTransform" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup_NextTransform" "', argument " "2"" of type '" "OpenBabel::transform3dIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup_NextTransform" "', argument " "2"" of type '" "OpenBabel::transform3dIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::transform3dIterator * >(argp2);
  result = (OpenBabel::transform3d *)((OpenBabel::SpaceGroup const *)arg1)->NextTransform(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__transform3d, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetSpaceGroup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::SpaceGroup *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetSpaceGroup" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::SpaceGroup *)OpenBabel::SpaceGroup::GetSpaceGroup((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetSpaceGroup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; OpenBabel::SpaceGroup *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_GetSpaceGroup" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup_GetSpaceGroup" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  result = (OpenBabel::SpaceGroup *)OpenBabel::SpaceGroup::GetSpaceGroup((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetSpaceGroup__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int val1 ; int ecode1 = 0 ; OpenBabel::SpaceGroup *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "SpaceGroup_GetSpaceGroup" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (OpenBabel::SpaceGroup *)OpenBabel::SpaceGroup::GetSpaceGroup(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_GetSpaceGroup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "SpaceGroup_GetSpaceGroup", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    int _v = 0; { { int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_1;
    return _wrap_SpaceGroup_GetSpaceGroup__SWIG_2(self, argc, argv);}  check_1: if (argc == 1) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_SpaceGroup_GetSpaceGroup__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_SpaceGroup_GetSpaceGroup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SpaceGroup_GetSpaceGroup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::SpaceGroup::GetSpaceGroup(char const *)\n"
  "    OpenBabel::SpaceGroup::GetSpaceGroup(std::string const &)\n" "    OpenBabel::SpaceGroup::GetSpaceGroup(unsigned int)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_SpaceGroup_Find(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::SpaceGroup *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_Find" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  result = (OpenBabel::SpaceGroup *)OpenBabel::SpaceGroup::Find(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; OpenBabel::SpaceGroup *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup___eq__" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__SpaceGroup,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup___eq__" "', argument " "2"" of type '" "OpenBabel::SpaceGroup const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup___eq__" "', argument " "2"" of type '" "OpenBabel::SpaceGroup const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp2);
  result = (bool)((OpenBabel::SpaceGroup const *)arg1)->operator ==((OpenBabel::SpaceGroup const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SpaceGroup___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; OpenBabel::SpaceGroup *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SpaceGroup___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup___ne__" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__SpaceGroup,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SpaceGroup___ne__" "', argument " "2"" of type '" "OpenBabel::SpaceGroup const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SpaceGroup___ne__" "', argument " "2"" of type '" "OpenBabel::SpaceGroup const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp2);
  result = (int)((OpenBabel::SpaceGroup const *)arg1)->operator !=((OpenBabel::SpaceGroup const &)*arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_SpaceGroup_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_IsValid" "', argument " "1"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); result = (bool)((OpenBabel::SpaceGroup const *)arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_SpaceGroup_HEXAGONAL_ORIGIN_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::SpaceGroup *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SpaceGroup_HEXAGONAL_ORIGIN_get" "', argument " "1"" of type '" "OpenBabel::SpaceGroup *""'");  }
   arg1 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp1); result = (int)(int) ((arg1)->HEXAGONAL_ORIGIN);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *SpaceGroup_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__SpaceGroup, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *SpaceGroup_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBBitVec__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBBitVec *)new OpenBabel::OBBitVec();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBBitVec__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int val1 ; int ecode1 = 0 ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBBitVec" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (OpenBabel::OBBitVec *)new OpenBabel::OBBitVec(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBBitVec__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBBitVec" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBBitVec" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  result = (OpenBabel::OBBitVec *)new OpenBabel::OBBitVec((OpenBabel::OBBitVec const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBBitVec(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBBitVec", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBBitVec__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_new_OBBitVec__SWIG_2(self, argc, argv);}  check_2:
  if (argc == 1) { PyObject *retobj = _wrap_new_OBBitVec__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBBitVec'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBitVec::OBBitVec()\n"
  "    OpenBabel::OBBitVec::OBBitVec(unsigned int)\n" "    OpenBabel::OBBitVec::OBBitVec(OpenBabel::OBBitVec const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBitVec_SetBitOn(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBitVec_SetBitOn", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_SetBitOn" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_SetBitOn" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetBitOn(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_SetBitOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBitVec_SetBitOff", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_SetBitOff" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_SetBitOff" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetBitOff(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_SetRangeOn(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_SetRangeOn", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_SetRangeOn" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_SetRangeOn" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBitVec_SetRangeOn" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetRangeOn(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_SetRangeOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_SetRangeOff", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_SetRangeOff" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_SetRangeOff" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBitVec_SetRangeOff" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetRangeOff(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_Fold(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBitVec_Fold", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_Fold" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_Fold" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->Fold(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_FirstBit__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_FirstBit" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_FirstBit" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (int)((OpenBabel::OBBitVec const *)arg1)->FirstBit(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_FirstBit__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_FirstBit" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); result = (int)((OpenBabel::OBBitVec const *)arg1)->FirstBit();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_FirstBit(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBitVec_FirstBit", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBitVec_FirstBit__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBitVec_FirstBit__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBitVec_FirstBit'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBitVec::FirstBit(unsigned int) const\n"
  "    OpenBabel::OBBitVec::FirstBit() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBitVec_NextBit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_NextBit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_NextBit" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_NextBit" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBBitVec const *)arg1)->NextBit(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_EndBit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_EndBit" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); result = (int)((OpenBabel::OBBitVec const *)arg1)->EndBit();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_GetSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_GetSize" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); result = ((OpenBabel::OBBitVec const *)arg1)->GetSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_CountBits(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_CountBits" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  result = (unsigned int)((OpenBabel::OBBitVec const *)arg1)->CountBits();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_IsEmpty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_IsEmpty" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); result = (bool)((OpenBabel::OBBitVec const *)arg1)->IsEmpty();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_Resize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_Resize", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_Resize" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_Resize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->Resize(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_ResizeWords(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_ResizeWords", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_ResizeWords" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_ResizeWords" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->ResizeWords(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_BitIsSet(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_BitIsSet", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_BitIsSet" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec_BitIsSet" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)((OpenBabel::OBBitVec const *)arg1)->BitIsSet(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_FromVecInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_FromVecInt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_FromVecInt" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec_FromVecInt" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec_FromVecInt" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     arg2 = ptr; }  (arg1)->FromVecInt((std::vector< int,std::allocator< int > > const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_FromString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; std::string *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_FromString", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_FromString" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec_FromString" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec_FromString" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBitVec_FromString" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->FromString((std::string const &)*arg2,arg3); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_ToVecInt(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_ToVecInt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_ToVecInt" "', argument " "1"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec_ToVecInt" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec_ToVecInt" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp2);
  ((OpenBabel::OBBitVec const *)arg1)->ToVecInt(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_Clear" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_Negate(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_Negate" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); (arg1)->Negate(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec_GetWords(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec::word_vector *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec_GetWords", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec_GetWords" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec_GetWords" "', argument " "2"" of type '" "OpenBabel::OBBitVec::word_vector &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec_GetWords" "', argument " "2"" of type '" "OpenBabel::OBBitVec::word_vector &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec::word_vector * >(argp2); (arg1)->GetWords(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___iand__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBitVec *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec___iand__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___iand__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec___iand__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec___iand__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBBitVec *) &(arg1)->operator &=((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___ior____SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::OBBitVec *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___ior__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec___ior__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec___ior__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBBitVec *) &(arg1)->operator |=((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___ior____SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBBitVec *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___ior__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBitVec___ior__" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (OpenBabel::OBBitVec *) &(arg1)->operator |=(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___ior__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBitVec___ior__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBBitVec___ior____SWIG_0(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_OBBitVec___ior____SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBitVec___ior__'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBitVec::operator |=(OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::OBBitVec::operator |=(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBitVec___ixor__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBitVec *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec___ixor__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___ixor__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec___ixor__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec___ixor__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBBitVec *) &(arg1)->operator ^=((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___isub__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBitVec *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec___isub__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___isub__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec___isub__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec___isub__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBBitVec *) &(arg1)->operator -=((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBitVec___iadd__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBitVec *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBitVec___iadd__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBitVec___iadd__" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBitVec___iadd__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBitVec___iadd__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBBitVec *) &(arg1)->operator +=((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap___or__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBBitVec result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__or__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__or__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__or__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__or__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__or__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = operator |((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___and__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBBitVec result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__and__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__and__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__and__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__and__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__and__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = operator &((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___xor__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBBitVec result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__xor__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__xor__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__xor__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__xor__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__xor__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = operator ^((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___sub____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::OBBitVec result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__sub__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__sub__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__sub__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = operator -((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___sub__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "__sub__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap___sub____SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_2; { int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap___sub____SWIG_0(self, argc, argv);}  check_2:
  if (argc == 2) { PyObject *retobj = _wrap___sub____SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail: SWIG_Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__eq__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__eq__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__eq__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__eq__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (bool)operator ==((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___lt__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "__lt__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__lt__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lt__" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__lt__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lt__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (bool)operator <((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___rshift__(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::istream *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  std::istream *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "__rshift__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__istream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__rshift__" "', argument " "1"" of type '" "std::istream &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__rshift__" "', argument " "1"" of type '" "std::istream &""'"); }
   arg1 = reinterpret_cast< std::istream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__rshift__" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__rshift__" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); result = (std::istream *) &operator >>(*arg1,*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__istream, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___lshift____SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::ostream *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::ostream *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__ostream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'"); }
   arg1 = reinterpret_cast< std::ostream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (std::ostream *) &operator <<(*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_delete_OBBitVec(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBBitVec" "', argument " "1"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBBitVec_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBBitVec_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Tanimoto(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBitVec *arg1 = 0 ;
  OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  double result; (void)self; if (!SWIG_Python_UnpackTuple(args, "Tanimoto", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tanimoto" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Tanimoto" "', argument " "1"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Tanimoto" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "Tanimoto" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (double)OpenBabel::Tanimoto((OpenBabel::OBBitVec const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReleaseVersion(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::string result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBReleaseVersion", 0, 0, 0)) SWIG_fail;
  result = OpenBabel::OBReleaseVersion(); resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGenericData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; unsigned int arg2 ; OpenBabel::DataOrigin arg3 ; unsigned int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBGenericData" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBGenericData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBGenericData" "', argument " "3"" of type '" "OpenBabel::DataOrigin""'"); }
    arg3 = static_cast< OpenBabel::DataOrigin >(val3);
  result = (OpenBabel::OBGenericData *)new OpenBabel::OBGenericData(SWIG_STD_MOVE(arg1),arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGenericData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; unsigned int arg2 ; unsigned int val2 ; int ecode2 = 0 ;
  OpenBabel::OBGenericData *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBGenericData" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBGenericData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2);
  result = (OpenBabel::OBGenericData *)new OpenBabel::OBGenericData(SWIG_STD_MOVE(arg1),arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGenericData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBGenericData" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBGenericData *)new OpenBabel::OBGenericData(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGenericData__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBGenericData *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBGenericData *)new OpenBabel::OBGenericData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGenericData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBGenericData", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBGenericData__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBGenericData__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBGenericData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBGenericData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBGenericData'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBGenericData::OBGenericData(std::string const,unsigned int const,OpenBabel::DataOrigin const)\n"
  "    OpenBabel::OBGenericData::OBGenericData(std::string const,unsigned int const)\n"
  "    OpenBabel::OBGenericData::OBGenericData(std::string const)\n" "    OpenBabel::OBGenericData::OBGenericData()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBGenericData_Clone(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGenericData_Clone", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_Clone" "', argument " "1"" of type '" "OpenBabel::OBGenericData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGenericData_Clone" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  result = (OpenBabel::OBGenericData *)((OpenBabel::OBGenericData const *)arg1)->Clone(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBGenericData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBGenericData" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_SetAttribute(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGenericData_SetAttribute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_SetAttribute" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGenericData_SetAttribute" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGenericData_SetAttribute" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetAttribute((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_SetOrigin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; OpenBabel::DataOrigin arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBGenericData_SetOrigin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_SetOrigin" "', argument " "1"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGenericData_SetOrigin" "', argument " "2"" of type '" "OpenBabel::DataOrigin""'"); }
    arg2 = static_cast< OpenBabel::DataOrigin >(val2); (arg1)->SetOrigin(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_GetAttribute(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_GetAttribute" "', argument " "1"" of type '" "OpenBabel::OBGenericData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (std::string *) &((OpenBabel::OBGenericData const *)arg1)->GetAttribute();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_GetDataType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_GetDataType" "', argument " "1"" of type '" "OpenBabel::OBGenericData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (unsigned int)((OpenBabel::OBGenericData const *)arg1)->GetDataType();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_GetValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_GetValue" "', argument " "1"" of type '" "OpenBabel::OBGenericData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (std::string *) &((OpenBabel::OBGenericData const *)arg1)->GetValue();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGenericData_GetOrigin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGenericData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::DataOrigin result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGenericData_GetOrigin" "', argument " "1"" of type '" "OpenBabel::OBGenericData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGenericData * >(argp1);
  result = (OpenBabel::DataOrigin)((OpenBabel::OBGenericData const *)arg1)->GetOrigin();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBGenericData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBGenericData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBGenericData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBBase" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_Clear" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = (bool)(arg1)->Clear();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DoTransformations(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ;
  std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > *arg2 = 0 ;
  OpenBabel::OBConversion *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; PyObject *swig_obj[3] ; OpenBabel::OBBase *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBase_DoTransformations", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DoTransformations" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_DoTransformations" "', argument " "2"" of type '" "std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > const *""'");  }
  
  arg2 = reinterpret_cast< std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBBase_DoTransformations" "', argument " "3"" of type '" "OpenBabel::OBConversion *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBConversion * >(argp3);
  result = (OpenBabel::OBBase *)(arg1)->DoTransformations((std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_ClassDescription(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBase_ClassDescription", 0, 0, 0)) SWIG_fail;
  result = (char *)OpenBabel::OBBase::ClassDescription(); resultobj = SWIG_FromCharPtr((const char *)result); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetTitle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; char *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_GetTitle" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (char *)((OpenBabel::OBBase const *)arg1)->GetTitle(arg2);
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetTitle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; char *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = (char *)((OpenBabel::OBBase const *)arg1)->GetTitle();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetTitle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBase_GetTitle", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBase_GetTitle__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBase_GetTitle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBase_GetTitle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBase::GetTitle(bool) const\n"
  "    OpenBabel::OBBase::GetTitle() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBase_SetTitle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBase_SetTitle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_SetTitle" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_SetTitle" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetTitle((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_HasData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_HasData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_HasData" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBase_HasData" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (bool)(arg1)->HasData((std::string const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_HasData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_HasData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_HasData" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->HasData((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_HasData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_HasData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_HasData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->HasData(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_HasData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBase_HasData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_1;
    return _wrap_OBBase_HasData__SWIG_2(self, argc, argv);}  check_1: if (argc == 2) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_OBBase_HasData__SWIG_0(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBBase_HasData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBase_HasData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBase::HasData(std::string const &)\n"
  "    OpenBabel::OBBase::HasData(char const *)\n" "    OpenBabel::OBBase::HasData(unsigned int const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBase_DeleteData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DeleteData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->DeleteData(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DeleteData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; OpenBabel::OBGenericData *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DeleteData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBGenericData * >(argp2); (arg1)->DeleteData(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DeleteData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DeleteData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > * >(argp2);
  (arg1)->DeleteData(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DeleteData__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DeleteData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBase_DeleteData" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (bool)(arg1)->DeleteData((std::string const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DeleteData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBase_DeleteData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0;
    { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBGenericData, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBBase_DeleteData__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 2) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_OBBase_DeleteData__SWIG_2(self, argc, argv);}  check_2:
  if (argc == 2) { int _v = 0; { { int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL); _v = SWIG_CheckState(res); } } 
    if (!_v) goto check_3; return _wrap_OBBase_DeleteData__SWIG_0(self, argc, argv);}  check_3: if (argc == 2) {
    PyObject *retobj = _wrap_OBBase_DeleteData__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBase_DeleteData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBase::DeleteData(unsigned int)\n"
  "    OpenBabel::OBBase::DeleteData(OpenBabel::OBGenericData *)\n"
  "    OpenBabel::OBBase::DeleteData(std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > &)\n"
  "    OpenBabel::OBBase::DeleteData(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBase_CloneData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; OpenBabel::OBGenericData *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBase_CloneData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_CloneData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_CloneData" "', argument " "2"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBGenericData * >(argp2); (arg1)->CloneData(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_DataSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_DataSize" "', argument " "1"" of type '" "OpenBabel::OBBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = ((OpenBabel::OBBase const *)arg1)->DataSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_GetData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (OpenBabel::OBGenericData *)(arg1)->GetData(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; OpenBabel::OBGenericData *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_GetData" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBase_GetData" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (OpenBabel::OBGenericData *)(arg1)->GetData((std::string const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBase_GetData" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (OpenBabel::OBGenericData *)(arg1)->GetData((char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetAllData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBase_GetAllData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetAllData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_GetAllData" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (arg1)->GetAllData(arg2);
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  result = (std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *) &(arg1)->GetData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ; OpenBabel::DataOrigin arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_GetData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBase_GetData" "', argument " "2"" of type '" "OpenBabel::DataOrigin""'"); }
    arg2 = static_cast< OpenBabel::DataOrigin >(val2); result = (arg1)->GetData(arg2);
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_GetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBase_GetData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBase_GetData__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_OBBase_GetData__SWIG_0(self, argc, argv);}  check_2: if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_3;
    return _wrap_OBBase_GetData__SWIG_4(self, argc, argv);}  check_3: if (argc == 2) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_4;
    return _wrap_OBBase_GetData__SWIG_1(self, argc, argv);}  check_4: if (argc == 2) {
    PyObject *retobj = _wrap_OBBase_GetData__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBase_GetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBase::GetData(unsigned int const)\n"
  "    OpenBabel::OBBase::GetData(std::string const &)\n" "    OpenBabel::OBBase::GetData(char const *)\n"
  "    OpenBabel::OBBase::GetData()\n" "    OpenBabel::OBBase::GetData(OpenBabel::DataOrigin)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBase_BeginData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBDataIterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_BeginData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = (arg1)->BeginData();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const OpenBabel::OBDataIterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBase_EndData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBDataIterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBase_EndData" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = (arg1)->EndData();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const OpenBabel::OBDataIterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBase *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBBase", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBBase *)new OpenBabel::OBBase();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBase, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBBase_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBBase, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBBase_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBCommentData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBCommentData *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBCommentData *)new OpenBabel::OBCommentData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCommentData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBCommentData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCommentData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBCommentData *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBCommentData,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBCommentData" "', argument " "1"" of type '" "OpenBabel::OBCommentData const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBCommentData" "', argument " "1"" of type '" "OpenBabel::OBCommentData const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBCommentData * >(argp1);
  result = (OpenBabel::OBCommentData *)new OpenBabel::OBCommentData((OpenBabel::OBCommentData const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCommentData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBCommentData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBCommentData", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBCommentData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBCommentData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBCommentData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBCommentData::OBCommentData()\n"
  "    OpenBabel::OBCommentData::OBCommentData(OpenBabel::OBCommentData const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBCommentData_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCommentData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCommentData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCommentData_SetData" "', argument " "1"" of type '" "OpenBabel::OBCommentData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCommentData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCommentData_SetData" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCommentData_SetData" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetData((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBCommentData_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCommentData *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCommentData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCommentData_SetData" "', argument " "1"" of type '" "OpenBabel::OBCommentData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCommentData * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCommentData_SetData" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetData((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBCommentData_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBCommentData_SetData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBCommentData_SetData__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBCommentData_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBCommentData_SetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBCommentData::SetData(std::string const &)\n"
  "    OpenBabel::OBCommentData::SetData(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBCommentData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCommentData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCommentData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCommentData_GetData" "', argument " "1"" of type '" "OpenBabel::OBCommentData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCommentData * >(argp1);
  result = (std::string *) &((OpenBabel::OBCommentData const *)arg1)->GetData();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBCommentData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCommentData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCommentData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBCommentData" "', argument " "1"" of type '" "OpenBabel::OBCommentData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCommentData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBCommentData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBCommentData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBCommentData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBExternalBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBExternalBond *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBExternalBond *)new OpenBabel::OBExternalBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBExternalBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBExternalBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::OBExternalBond *result = 0 ;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBExternalBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBExternalBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBExternalBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (OpenBabel::OBExternalBond *)new OpenBabel::OBExternalBond(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBExternalBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBExternalBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBExternalBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBExternalBond *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBExternalBond,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBExternalBond" "', argument " "1"" of type '" "OpenBabel::OBExternalBond const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBExternalBond" "', argument " "1"" of type '" "OpenBabel::OBExternalBond const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  result = (OpenBabel::OBExternalBond *)new OpenBabel::OBExternalBond((OpenBabel::OBExternalBond const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBExternalBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBExternalBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBExternalBond", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBExternalBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBExternalBond__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBExternalBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBExternalBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBExternalBond::OBExternalBond()\n"
  "    OpenBabel::OBExternalBond::OBExternalBond(OpenBabel::OBAtom *,OpenBabel::OBBond *,int)\n"
  "    OpenBabel::OBExternalBond::OBExternalBond(OpenBabel::OBExternalBond const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBExternalBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBExternalBond" "', argument " "1"" of type '" "OpenBabel::OBExternalBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_GetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_GetIdx" "', argument " "1"" of type '" "OpenBabel::OBExternalBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  result = (int)((OpenBabel::OBExternalBond const *)arg1)->GetIdx(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_GetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_GetAtom" "', argument " "1"" of type '" "OpenBabel::OBExternalBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBExternalBond const *)arg1)->GetAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_GetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBond *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_GetBond" "', argument " "1"" of type '" "OpenBabel::OBExternalBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  result = (OpenBabel::OBBond *)((OpenBabel::OBExternalBond const *)arg1)->GetBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_SetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBExternalBond_SetIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_SetIdx" "', argument " "1"" of type '" "OpenBabel::OBExternalBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBExternalBond_SetIdx" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetIdx(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_SetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBExternalBond_SetAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_SetAtom" "', argument " "1"" of type '" "OpenBabel::OBExternalBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBExternalBond_SetAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->SetAtom(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBond_SetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBond *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBExternalBond_SetBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBond_SetBond" "', argument " "1"" of type '" "OpenBabel::OBExternalBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBExternalBond_SetBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); (arg1)->SetBond(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBExternalBond_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBExternalBond, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBExternalBond_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBExternalBondData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBondData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBExternalBondData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBExternalBondData *)new OpenBabel::OBExternalBondData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBExternalBondData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBondData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBondData *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBBond *arg3 = 0 ; int arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ;
  PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBExternalBondData_SetData", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBondData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBondData_SetData" "', argument " "1"" of type '" "OpenBabel::OBExternalBondData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBondData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBExternalBondData_SetData" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBExternalBondData_SetData" "', argument " "3"" of type '" "OpenBabel::OBBond *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBond * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBExternalBondData_SetData" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->SetData(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBExternalBondData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBondData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBExternalBond,std::allocator< OpenBabel::OBExternalBond > > *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBondData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBExternalBondData_GetData" "', argument " "1"" of type '" "OpenBabel::OBExternalBondData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBondData * >(argp1);
  result = (std::vector< OpenBabel::OBExternalBond,std::allocator< OpenBabel::OBExternalBond > > *)(arg1)->GetData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBExternalBondData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBExternalBondData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBExternalBondData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBExternalBondData" "', argument " "1"" of type '" "OpenBabel::OBExternalBondData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBExternalBondData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBExternalBondData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBExternalBondData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBExternalBondData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBPairData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBPairData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBPairData *)new OpenBabel::OBPairData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPairData_SetValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPairData *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairData_SetValue" "', argument " "1"" of type '" "OpenBabel::OBPairData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairData * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPairData_SetValue" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetValue((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPairData_SetValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPairData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairData_SetValue" "', argument " "1"" of type '" "OpenBabel::OBPairData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPairData_SetValue" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPairData_SetValue" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetValue((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPairData_SetValue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBPairData_SetValue", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBPairData_SetValue__SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBPairData_SetValue__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBPairData_SetValue'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBPairData::SetValue(char const *)\n"
  "    OpenBabel::OBPairData::SetValue(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBPairData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPairData" "', argument " "1"" of type '" "OpenBabel::OBPairData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBPairData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPairData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBPairData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBSetData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBSetData *)new OpenBabel::OBSetData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSetData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_AddData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ; OpenBabel::OBGenericData *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSetData_AddData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_AddData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSetData_AddData" "', argument " "2"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBGenericData * >(argp2); (arg1)->AddData(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSetData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_SetData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBGenericData_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSetData_SetData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSetData_SetData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > * >(argp2);
  (arg1)->SetData(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_GetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSetData *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; OpenBabel::OBGenericData *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_GetData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSetData_GetData" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (OpenBabel::OBGenericData *)(arg1)->GetData((char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_GetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSetData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; OpenBabel::OBGenericData *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_GetData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSetData_GetData" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSetData_GetData" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (OpenBabel::OBGenericData *)(arg1)->GetData((std::string const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_GetData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSetData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_GetData" "', argument " "1"" of type '" "OpenBabel::OBSetData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1);
  result = (std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *) &((OpenBabel::OBSetData const *)arg1)->GetData();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBGenericData*,std::allocator< OpenBabel::OBGenericData * > > >(*result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_GetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSetData_GetData", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBSetData_GetData__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_OBSetData_GetData__SWIG_1(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBSetData_GetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSetData_GetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSetData::GetData(char const *)\n"
  "    OpenBabel::OBSetData::GetData(std::string const &)\n" "    OpenBabel::OBSetData::GetData() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSetData_GetBegin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >::iterator result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_GetBegin" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1); result = (arg1)->GetBegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_GetEnd(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >::iterator result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_GetEnd" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1); result = (arg1)->GetEnd();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSetData_DeleteData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ; OpenBabel::OBGenericData *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSetData_DeleteData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSetData_DeleteData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBGenericData, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSetData_DeleteData" "', argument " "2"" of type '" "OpenBabel::OBGenericData *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBGenericData * >(argp2); (arg1)->DeleteData(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSetData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSetData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSetData" "', argument " "1"" of type '" "OpenBabel::OBSetData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSetData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBSetData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSetData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSetData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBVirtualBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBVirtualBond *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBVirtualBond *)new OpenBabel::OBVirtualBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVirtualBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBVirtualBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int arg2 ; unsigned int arg3 ; int arg4 ; unsigned int val1 ;
  int ecode1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ;
  OpenBabel::OBVirtualBond *result = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBVirtualBond" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBVirtualBond" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBVirtualBond" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBVirtualBond" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); result = (OpenBabel::OBVirtualBond *)new OpenBabel::OBVirtualBond(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVirtualBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBVirtualBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int arg2 ; unsigned int arg3 ; unsigned int val1 ; int ecode1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; OpenBabel::OBVirtualBond *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBVirtualBond" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBVirtualBond" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBVirtualBond" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); result = (OpenBabel::OBVirtualBond *)new OpenBabel::OBVirtualBond(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVirtualBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBVirtualBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBVirtualBond", 0, 4, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBVirtualBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBVirtualBond__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBVirtualBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBVirtualBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBVirtualBond::OBVirtualBond()\n"
  "    OpenBabel::OBVirtualBond::OBVirtualBond(unsigned int,unsigned int,unsigned int,int)\n"
  "    OpenBabel::OBVirtualBond::OBVirtualBond(unsigned int,unsigned int,unsigned int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBVirtualBond_GetBgn(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVirtualBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVirtualBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVirtualBond_GetBgn" "', argument " "1"" of type '" "OpenBabel::OBVirtualBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVirtualBond * >(argp1); result = (unsigned int)(arg1)->GetBgn();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVirtualBond_GetEnd(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVirtualBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVirtualBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVirtualBond_GetEnd" "', argument " "1"" of type '" "OpenBabel::OBVirtualBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVirtualBond * >(argp1); result = (unsigned int)(arg1)->GetEnd();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVirtualBond_GetOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVirtualBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVirtualBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVirtualBond_GetOrder" "', argument " "1"" of type '" "OpenBabel::OBVirtualBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVirtualBond * >(argp1); result = (unsigned int)(arg1)->GetOrder();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVirtualBond_GetStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVirtualBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVirtualBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVirtualBond_GetStereo" "', argument " "1"" of type '" "OpenBabel::OBVirtualBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVirtualBond * >(argp1); result = (int)(arg1)->GetStereo();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBVirtualBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVirtualBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVirtualBond, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBVirtualBond" "', argument " "1"" of type '" "OpenBabel::OBVirtualBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVirtualBond * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBVirtualBond_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBVirtualBond, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBVirtualBond_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRingData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBRingData *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBRingData *)new OpenBabel::OBRingData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRingData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRingData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRingData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBRingData *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBRingData,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRingData" "', argument " "1"" of type '" "OpenBabel::OBRingData const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRingData" "', argument " "1"" of type '" "OpenBabel::OBRingData const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  result = (OpenBabel::OBRingData *)new OpenBabel::OBRingData((OpenBabel::OBRingData const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRingData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRingData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBRingData", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBRingData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBRingData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBRingData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRingData::OBRingData()\n"
  "    OpenBabel::OBRingData::OBRingData(OpenBabel::OBRingData const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBRingData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRingData" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_SetData" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingData_SetData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingData_SetData" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > * >(argp2);
  (arg1)->SetData(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_PushBack(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; OpenBabel::OBRing *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingData_PushBack", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_PushBack" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingData_PushBack" "', argument " "2"" of type '" "OpenBabel::OBRing *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBRing * >(argp2); (arg1)->PushBack(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_GetData" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  result = (std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *) &(arg1)->GetData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_BeginRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_BeginRings" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1); result = (arg1)->BeginRings();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_EndRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_EndRings" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1); result = (arg1)->EndRings();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_BeginRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBRing *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingData_BeginRing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_BeginRing" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingData_BeginRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingData_BeginRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator * >(argp2);
  result = (OpenBabel::OBRing *)(arg1)->BeginRing(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRingData_NextRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingData *arg1 = 0 ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBRing *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingData_NextRing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingData_NextRing" "', argument " "1"" of type '" "OpenBabel::OBRingData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingData_NextRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingData_NextRing" "', argument " "2"" of type '" "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator * >(argp2);
  result = (OpenBabel::OBRing *)(arg1)->NextRing(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBRingData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRingData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRingData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBUnitCell__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBUnitCell *)new OpenBabel::OBUnitCell();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBUnitCell, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBUnitCell__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBUnitCell *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBUnitCell,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBUnitCell" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBUnitCell" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (OpenBabel::OBUnitCell *)new OpenBabel::OBUnitCell((OpenBabel::OBUnitCell const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBUnitCell, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBUnitCell(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBUnitCell", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBUnitCell__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBUnitCell__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBUnitCell'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::OBUnitCell()\n"
  "    OpenBabel::OBUnitCell::OBUnitCell(OpenBabel::OBUnitCell const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBUnitCell(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBUnitCell" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; double arg5 ;
  double arg6 ; double arg7 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ;
  double val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ; double val6 ; int ecode6 = 0 ; double val7 ; int ecode7 = 0 ;
  (void)self; if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetData" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBUnitCell_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBUnitCell_SetData" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBUnitCell_SetData" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBUnitCell_SetData" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBUnitCell_SetData" "', argument " "6"" of type '" "double""'"); }
    arg6 = static_cast< double >(val6); ecode7 = SWIG_AsVal_double(swig_obj[6], &val7); if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "OBUnitCell_SetData" "', argument " "7"" of type '" "double""'"); }
    arg7 = static_cast< double >(val7); (arg1)->SetData(arg2,arg3,arg4,arg5,arg6,arg7); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  OpenBabel::vector3 arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ;
  void *argp4 ; int res4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetData" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_SetData" "', argument " "2"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_SetData" "', argument " "2"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBUnitCell_SetData" "', argument " "3"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_SetData" "', argument " "3"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  {
    res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBUnitCell_SetData" "', argument " "4"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp4) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_SetData" "', argument " "4"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp4); arg4 = *temp;
      if (SWIG_IsNewObj(res4)) delete temp; }  }  (arg1)->SetData(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),SWIG_STD_MOVE(arg4));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::matrix3x3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetData" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_SetData" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_SetData" "', argument " "2"" of type '" "OpenBabel::matrix3x3 const""'"); }
     else { OpenBabel::matrix3x3 * temp = reinterpret_cast< OpenBabel::matrix3x3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetData(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[8] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_SetData", 0, 7, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_SetData__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBUnitCell_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 7) {
    PyObject *retobj = _wrap_OBUnitCell_SetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_SetData'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBUnitCell::SetData(double const,double const,double const,double const,double const,double const)\n"
  "    OpenBabel::OBUnitCell::SetData(OpenBabel::vector3 const,OpenBabel::vector3 const,OpenBabel::vector3 const)\n"
  "    OpenBabel::OBUnitCell::SetData(OpenBabel::matrix3x3 const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetOffset(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBUnitCell_SetOffset", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetOffset" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_SetOffset" "', argument " "2"" of type '" "OpenBabel::vector3 const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_SetOffset" "', argument " "2"" of type '" "OpenBabel::vector3 const""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetOffset(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetSpaceGroup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::SpaceGroup *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "2"" of type '" "OpenBabel::SpaceGroup const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::SpaceGroup * >(argp2); (arg1)->SetSpaceGroup((OpenBabel::SpaceGroup const *)arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetSpaceGroup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "2"" of type '" "std::string const""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetSpaceGroup(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetSpaceGroup__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBUnitCell_SetSpaceGroup" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetSpaceGroup(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetSpaceGroup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_SetSpaceGroup", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__SpaceGroup, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBUnitCell_SetSpaceGroup__SWIG_0(self, argc, argv);} 
  check_1: if (argc == 2) { int _v = 0; { { int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } } 
    if (!_v) goto check_2; return _wrap_OBUnitCell_SetSpaceGroup__SWIG_2(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_SetSpaceGroup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_SetSpaceGroup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::SetSpaceGroup(OpenBabel::SpaceGroup const *)\n"
  "    OpenBabel::OBUnitCell::SetSpaceGroup(std::string const)\n" "    OpenBabel::OBUnitCell::SetSpaceGroup(int const)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_SetLatticeType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::OBUnitCell::LatticeType arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBUnitCell_SetLatticeType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_SetLatticeType" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBUnitCell_SetLatticeType" "', argument " "2"" of type '" "OpenBabel::OBUnitCell::LatticeType""'"); }
    arg2 = static_cast< OpenBabel::OBUnitCell::LatticeType >(val2); (arg1)->SetLatticeType(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_FillUnitCell(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBUnitCell_FillUnitCell", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_FillUnitCell" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_FillUnitCell" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->FillUnitCell(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetA__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetA" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetA();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetA__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetA" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetA();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetA(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetA", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetA__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetA__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetA'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetA()\n" "    OpenBabel::OBUnitCell::GetA() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetB__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetB" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetB();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetB__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetB" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetB();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetB(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetB", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetB__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetB__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetB'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetB()\n" "    OpenBabel::OBUnitCell::GetB() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetC__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetC" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetC();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetC__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetC" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetC();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetC(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetC", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetC__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetC__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetC'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetC()\n" "    OpenBabel::OBUnitCell::GetC() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetAlpha__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetAlpha" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetAlpha();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetAlpha__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetAlpha" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetAlpha();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetAlpha(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetAlpha", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetAlpha__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetAlpha__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetAlpha'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetAlpha()\n"
  "    OpenBabel::OBUnitCell::GetAlpha() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetBeta__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetBeta" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetBeta();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetBeta__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetBeta" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetBeta();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetBeta(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetBeta", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetBeta__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetBeta__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetBeta'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetBeta()\n" "    OpenBabel::OBUnitCell::GetBeta() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetGamma__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetGamma" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetGamma();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetGamma__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetGamma" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetGamma();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetGamma(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetGamma", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetGamma__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetGamma__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetGamma'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetGamma()\n"
  "    OpenBabel::OBUnitCell::GetGamma() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOffset__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOffset" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetOffset();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOffset__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOffset" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = ((OpenBabel::OBUnitCell const *)arg1)->GetOffset();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOffset(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetOffset", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetOffset__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetOffset__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetOffset'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetOffset()\n"
  "    OpenBabel::OBUnitCell::GetOffset() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::SpaceGroup *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (OpenBabel::SpaceGroup *)(arg1)->GetSpaceGroup();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::SpaceGroup *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (OpenBabel::SpaceGroup *)((OpenBabel::OBUnitCell const *)arg1)->GetSpaceGroup();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__SpaceGroup, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetSpaceGroup", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetSpaceGroup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetSpaceGroup()\n"
  "    OpenBabel::OBUnitCell::GetSpaceGroup() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupName__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; std::string result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupName" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetSpaceGroupName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupName__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; std::string result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupName" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = ((OpenBabel::OBUnitCell const *)arg1)->GetSpaceGroupName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupName(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetSpaceGroupName", 0, 1, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupName__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupName__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetSpaceGroupName'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetSpaceGroupName()\n"
  "    OpenBabel::OBUnitCell::GetSpaceGroupName() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetLatticeType__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBUnitCell::LatticeType result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetLatticeType" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBUnitCell_GetLatticeType" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (OpenBabel::OBUnitCell::LatticeType)(arg1)->GetLatticeType(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetLatticeType__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBUnitCell::LatticeType result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetLatticeType" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBUnitCell_GetLatticeType" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2);
  result = (OpenBabel::OBUnitCell::LatticeType)((OpenBabel::OBUnitCell const *)arg1)->GetLatticeType(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetLatticeType__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBUnitCell::LatticeType result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetLatticeType" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (OpenBabel::OBUnitCell::LatticeType)(arg1)->GetLatticeType(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetLatticeType__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBUnitCell::LatticeType result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetLatticeType" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (OpenBabel::OBUnitCell::LatticeType)((OpenBabel::OBUnitCell const *)arg1)->GetLatticeType();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetLatticeType(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetLatticeType", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetLatticeType__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetLatticeType__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_GetLatticeType__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_GetLatticeType__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetLatticeType'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetLatticeType(int)\n"
  "    OpenBabel::OBUnitCell::GetLatticeType(int) const\n" "    OpenBabel::OBUnitCell::GetLatticeType()\n"
  "    OpenBabel::OBUnitCell::GetLatticeType() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVectors__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellVectors" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetCellVectors();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVectors__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellVectors" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = ((OpenBabel::OBUnitCell const *)arg1)->GetCellVectors();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVectors(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetCellVectors", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellVectors__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellVectors__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetCellVectors'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetCellVectors()\n"
  "    OpenBabel::OBUnitCell::GetCellVectors() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellMatrix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetCellMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellMatrix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = ((OpenBabel::OBUnitCell const *)arg1)->GetCellMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellMatrix(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetCellMatrix", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellMatrix__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellMatrix__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetCellMatrix'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetCellMatrix()\n"
  "    OpenBabel::OBUnitCell::GetCellMatrix() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrthoMatrix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOrthoMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetOrthoMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrthoMatrix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOrthoMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = ((OpenBabel::OBUnitCell const *)arg1)->GetOrthoMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrthoMatrix(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetOrthoMatrix", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetOrthoMatrix__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetOrthoMatrix__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetOrthoMatrix'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetOrthoMatrix()\n"
  "    OpenBabel::OBUnitCell::GetOrthoMatrix() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrientationMatrix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOrientationMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetOrientationMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrientationMatrix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetOrientationMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = ((OpenBabel::OBUnitCell const *)arg1)->GetOrientationMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetOrientationMatrix(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetOrientationMatrix", 0, 1, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBUnitCell_GetOrientationMatrix__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetOrientationMatrix__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetOrientationMatrix'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetOrientationMatrix()\n"
  "    OpenBabel::OBUnitCell::GetOrientationMatrix() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetFractionalMatrix__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetFractionalMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (arg1)->GetFractionalMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetFractionalMatrix__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::matrix3x3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetFractionalMatrix" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = ((OpenBabel::OBUnitCell const *)arg1)->GetFractionalMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetFractionalMatrix(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetFractionalMatrix", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetFractionalMatrix__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetFractionalMatrix__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetFractionalMatrix'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetFractionalMatrix()\n"
  "    OpenBabel::OBUnitCell::GetFractionalMatrix() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_FractionalToCartesian__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->FractionalToCartesian(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_FractionalToCartesian__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_FractionalToCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->FractionalToCartesian(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_FractionalToCartesian(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_FractionalToCartesian", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_FractionalToCartesian__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_FractionalToCartesian__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_FractionalToCartesian'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::FractionalToCartesian(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::FractionalToCartesian(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_CartesianToFractional__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_CartesianToFractional" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_CartesianToFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_CartesianToFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->CartesianToFractional(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_CartesianToFractional__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_CartesianToFractional" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_CartesianToFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_CartesianToFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->CartesianToFractional(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_CartesianToFractional(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_CartesianToFractional", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_CartesianToFractional__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_CartesianToFractional__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_CartesianToFractional'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::CartesianToFractional(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::CartesianToFractional(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapCartesianCoordinate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->WrapCartesianCoordinate(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapCartesianCoordinate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_WrapCartesianCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->WrapCartesianCoordinate(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapCartesianCoordinate(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_WrapCartesianCoordinate", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_WrapCartesianCoordinate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_WrapCartesianCoordinate__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_WrapCartesianCoordinate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::WrapCartesianCoordinate(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::WrapCartesianCoordinate(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapFractionalCoordinate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->WrapFractionalCoordinate(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapFractionalCoordinate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_WrapFractionalCoordinate" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->WrapFractionalCoordinate(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_WrapFractionalCoordinate(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_WrapFractionalCoordinate", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_WrapFractionalCoordinate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_WrapFractionalCoordinate__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_WrapFractionalCoordinate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::WrapFractionalCoordinate(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::WrapFractionalCoordinate(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapCartesianNear__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  result = (arg1)->UnwrapCartesianNear(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapCartesianNear__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapCartesianNear" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->UnwrapCartesianNear(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapCartesianNear(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_UnwrapCartesianNear", 0, 3, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBUnitCell_UnwrapCartesianNear__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBUnitCell_UnwrapCartesianNear__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_UnwrapCartesianNear'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::UnwrapCartesianNear(OpenBabel::vector3,OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::UnwrapCartesianNear(OpenBabel::vector3,OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapFractionalNear__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  } 
  result = (arg1)->UnwrapFractionalNear(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapFractionalNear__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; OpenBabel::vector3 arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "3"" of type '" "OpenBabel::vector3""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_UnwrapFractionalNear" "', argument " "3"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->UnwrapFractionalNear(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_UnwrapFractionalNear(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_UnwrapFractionalNear", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBUnitCell_UnwrapFractionalNear__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBUnitCell_UnwrapFractionalNear__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_UnwrapFractionalNear'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBUnitCell::UnwrapFractionalNear(OpenBabel::vector3,OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::UnwrapFractionalNear(OpenBabel::vector3,OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageCartesian__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->MinimumImageCartesian(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageCartesian__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_MinimumImageCartesian" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->MinimumImageCartesian(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageCartesian(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_MinimumImageCartesian", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_MinimumImageCartesian__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_MinimumImageCartesian__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_MinimumImageCartesian'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::MinimumImageCartesian(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::MinimumImageCartesian(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageFractional__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  result = (arg1)->MinimumImageFractional(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageFractional__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBUnitCell_MinimumImageFractional" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = ((OpenBabel::OBUnitCell const *)arg1)->MinimumImageFractional(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_MinimumImageFractional(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_MinimumImageFractional", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBUnitCell_MinimumImageFractional__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_MinimumImageFractional__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_MinimumImageFractional'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::MinimumImageFractional(OpenBabel::vector3)\n"
  "    OpenBabel::OBUnitCell::MinimumImageFractional(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; int result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (int)(arg1)->GetSpaceGroupNumber(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (int)(arg1)->GetSpaceGroupNumber();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; int result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (int)((OpenBabel::OBUnitCell const *)arg1)->GetSpaceGroupNumber(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetSpaceGroupNumber" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (int)((OpenBabel::OBUnitCell const *)arg1)->GetSpaceGroupNumber();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetSpaceGroupNumber(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetSpaceGroupNumber", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBUnitCell_GetSpaceGroupNumber__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetSpaceGroupNumber'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetSpaceGroupNumber(std::string)\n"
  "    OpenBabel::OBUnitCell::GetSpaceGroupNumber()\n" "    OpenBabel::OBUnitCell::GetSpaceGroupNumber(std::string) const\n"
  "    OpenBabel::OBUnitCell::GetSpaceGroupNumber() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVolume__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellVolume" "', argument " "1"" of type '" "OpenBabel::OBUnitCell *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1); result = (double)(arg1)->GetCellVolume();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVolume__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBUnitCell *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBUnitCell, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBUnitCell_GetCellVolume" "', argument " "1"" of type '" "OpenBabel::OBUnitCell const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBUnitCell * >(argp1);
  result = (double)((OpenBabel::OBUnitCell const *)arg1)->GetCellVolume();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBUnitCell_GetCellVolume(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBUnitCell_GetCellVolume", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellVolume__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBUnitCell_GetCellVolume__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBUnitCell_GetCellVolume'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBUnitCell::GetCellVolume()\n"
  "    OpenBabel::OBUnitCell::GetCellVolume() const\n"); return 0; }
SWIGINTERN PyObject *OBUnitCell_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBUnitCell, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBUnitCell_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBConformerData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBConformerData *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBConformerData *)new OpenBabel::OBConformerData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConformerData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConformerData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBConformerData *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBConformerData,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBConformerData" "', argument " "1"" of type '" "OpenBabel::OBConformerData const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBConformerData" "', argument " "1"" of type '" "OpenBabel::OBConformerData const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1);
  result = (OpenBabel::OBConformerData *)new OpenBabel::OBConformerData((OpenBabel::OBConformerData const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConformerData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConformerData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBConformerData", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBConformerData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBConformerData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBConformerData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConformerData::OBConformerData()\n"
  "    OpenBabel::OBConformerData::OBConformerData(OpenBabel::OBConformerData const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBConformerData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConformerData" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetDimension(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ;
  SwigValueWrapper< std::vector< unsigned short,std::allocator< unsigned short > > > arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetDimension", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetDimension" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerData_SetDimension" "', argument " "2"" of type '" "std::vector< unsigned short,std::allocator< unsigned short > >""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerData_SetDimension" "', argument " "2"" of type '" "std::vector< unsigned short,std::allocator< unsigned short > >""'"); }
     else {
      std::vector< unsigned short,std::allocator< unsigned short > > * temp = reinterpret_cast< std::vector< unsigned short,std::allocator< unsigned short > > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetDimension(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetEnergies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; std::vector< double,std::allocator< double > > arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetEnergies", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetEnergies" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConformerData_SetEnergies" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetEnergies(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetForces(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetForces", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetForces" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConformerData_SetForces" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetForces(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetVelocities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetVelocities", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetVelocities" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConformerData_SetVelocities" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetVelocities(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetDisplacements(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetDisplacements", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetDisplacements" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConformerData_SetDisplacements" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetDisplacements(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; std::vector< std::string,std::allocator< std::string > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_SetData" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConformerData_SetData" "', argument " "2"" of type '" "std::vector< std::string,std::allocator< std::string > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetData(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetDimension(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< unsigned short,std::allocator< unsigned short > > > result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetDimension" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetDimension();
  resultobj = SWIG_NewPointerObj((new std::vector< unsigned short,std::allocator< unsigned short > >(result)), SWIGTYPE_p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetEnergies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetEnergies" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetEnergies();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetForces(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetForces" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetForces();
  resultobj = swig::from(static_cast< std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetVelocities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetVelocities" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetVelocities();
  resultobj = swig::from(static_cast< std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetDisplacements(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetDisplacements" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetDisplacements();
  resultobj = swig::from(static_cast< std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::string,std::allocator< std::string > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerData_GetData" "', argument " "1"" of type '" "OpenBabel::OBConformerData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerData * >(argp1); result = (arg1)->GetData();
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBConformerData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConformerData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBConformerData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSymmetryData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBSymmetryData *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBSymmetryData *)new OpenBabel::OBSymmetryData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSymmetryData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSymmetryData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSymmetryData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBSymmetryData *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBSymmetryData,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSymmetryData" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBSymmetryData" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1);
  result = (OpenBabel::OBSymmetryData *)new OpenBabel::OBSymmetryData((OpenBabel::OBSymmetryData const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSymmetryData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSymmetryData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBSymmetryData", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBSymmetryData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBSymmetryData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBSymmetryData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSymmetryData::OBSymmetryData()\n"
  "    OpenBabel::OBSymmetryData::OBSymmetryData(OpenBabel::OBSymmetryData const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBSymmetryData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSymmetryData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSymmetryData" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSymmetryData *arg1 = 0 ; std::string arg2 ; std::string arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_SetData" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBSymmetryData_SetData" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBSymmetryData_SetData" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetData(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSymmetryData *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_SetData" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBSymmetryData_SetData" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetData(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSymmetryData_SetData", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBSymmetryData_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBSymmetryData_SetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSymmetryData_SetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSymmetryData::SetData(std::string,std::string)\n"
  "    OpenBabel::OBSymmetryData::SetData(std::string)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_SetPointGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSymmetryData *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSymmetryData_SetPointGroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_SetPointGroup" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBSymmetryData_SetPointGroup" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetPointGroup(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_SetSpaceGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSymmetryData *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSymmetryData_SetSpaceGroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_SetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBSymmetryData_SetSpaceGroup" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetSpaceGroup(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_GetPointGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSymmetryData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_GetPointGroup" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); result = (arg1)->GetPointGroup();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSymmetryData_GetSpaceGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSymmetryData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSymmetryData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSymmetryData_GetSpaceGroup" "', argument " "1"" of type '" "OpenBabel::OBSymmetryData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSymmetryData * >(argp1); result = (arg1)->GetSpaceGroup();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBSymmetryData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSymmetryData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSymmetryData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBTorsion__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBTorsion *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBTorsion,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBTorsion" "', argument " "1"" of type '" "OpenBabel::OBTorsion const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBTorsion" "', argument " "1"" of type '" "OpenBabel::OBTorsion const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  result = (OpenBabel::OBTorsion *)new OpenBabel::OBTorsion((OpenBabel::OBTorsion const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTorsion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBTorsion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBTorsion", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBTorsion__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBTorsion'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBTorsion::OBTorsion(OpenBabel::OBTorsion const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBTorsion(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTorsion" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_Clear" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_Empty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_Empty" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = (bool)(arg1)->Empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_AddTorsion__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  OpenBabel::OBAtom *arg4 = 0 ; OpenBabel::OBAtom *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; bool result; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_AddTorsion" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsion_AddTorsion" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBTorsion_AddTorsion" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBTorsion_AddTorsion" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBTorsion_AddTorsion" "', argument " "5"" of type '" "OpenBabel::OBAtom *""'");  }
   arg5 = reinterpret_cast< OpenBabel::OBAtom * >(argp5); result = (bool)(arg1)->AddTorsion(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_AddTorsion__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ;
  OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_AddTorsion" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsion_AddTorsion" "', argument " "2"" of type '" "OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTorsion_AddTorsion" "', argument " "2"" of type '" "OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &""'"); }
  
  arg2 = reinterpret_cast< OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > * >(argp2);
  result = (bool)(arg1)->AddTorsion(*arg2); resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_AddTorsion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBTorsion_AddTorsion", 0, 5, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBTorsion_AddTorsion__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBTorsion_AddTorsion__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBTorsion_AddTorsion'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBTorsion::AddTorsion(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBTorsion::AddTorsion(OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBTorsion_SetAngle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; double arg2 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  double val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_SetAngle" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTorsion_SetAngle" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBTorsion_SetAngle" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); result = (bool)(arg1)->SetAngle(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_SetAngle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_SetAngle" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTorsion_SetAngle" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); result = (bool)(arg1)->SetAngle(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_SetAngle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBTorsion_SetAngle", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBTorsion_SetAngle__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBTorsion_SetAngle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBTorsion_SetAngle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBTorsion::SetAngle(double,unsigned int)\n"
  "    OpenBabel::OBTorsion::SetAngle(double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBTorsion_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTorsion_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_SetData" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsion_SetData" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->SetData(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetAngle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; double *arg2 = 0 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_double,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsion_GetAngle" "', argument " "2"" of type '" "double &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTorsion_GetAngle" "', argument " "2"" of type '" "double &""'"); }
   arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBTorsion_GetAngle" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); result = (bool)(arg1)->GetAngle(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetAngle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTorsion *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_double,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsion_GetAngle" "', argument " "2"" of type '" "double &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTorsion_GetAngle" "', argument " "2"" of type '" "double &""'"); }
   arg2 = reinterpret_cast< double * >(argp2); result = (bool)(arg1)->GetAngle(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetAngle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBTorsion_GetAngle", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBTorsion_GetAngle__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBTorsion_GetAngle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBTorsion_GetAngle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBTorsion::GetAngle(double &,unsigned int)\n"
  "    OpenBabel::OBTorsion::GetAngle(double &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetBondIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetBondIdx" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = (unsigned int)(arg1)->GetBondIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetSize" "', argument " "1"" of type '" "OpenBabel::OBTorsion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = ((OpenBabel::OBTorsion const *)arg1)->GetSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetBC(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::pair< OpenBabel::OBAtom *,OpenBabel::OBAtom * > > result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetBC" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = (arg1)->GetBC();
  resultobj = SWIG_NewPointerObj((new std::pair< OpenBabel::OBAtom *,OpenBabel::OBAtom * >(result)), SWIGTYPE_p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_GetADs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double >,std::allocator< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_GetADs" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = (arg1)->GetADs();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double >,std::allocator< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double > > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsion_IsProtonRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsion_IsProtonRotor" "', argument " "1"" of type '" "OpenBabel::OBTorsion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsion * >(argp1); result = (bool)(arg1)->IsProtonRotor();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBTorsion_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTorsion, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBTorsion_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBTorsionData_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsionData_Clear" "', argument " "1"" of type '" "OpenBabel::OBTorsionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsionData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBTorsion,std::allocator< OpenBabel::OBTorsion > > > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsionData_GetData" "', argument " "1"" of type '" "OpenBabel::OBTorsionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1); result = ((OpenBabel::OBTorsionData const *)arg1)->GetData();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTorsion,std::allocator< OpenBabel::OBTorsion > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsionData_GetSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsionData_GetSize" "', argument " "1"" of type '" "OpenBabel::OBTorsionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1); result = ((OpenBabel::OBTorsionData const *)arg1)->GetSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsionData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ; OpenBabel::OBTorsion *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTorsionData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsionData_SetData" "', argument " "1"" of type '" "OpenBabel::OBTorsionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTorsion,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsionData_SetData" "', argument " "2"" of type '" "OpenBabel::OBTorsion &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTorsionData_SetData" "', argument " "2"" of type '" "OpenBabel::OBTorsion &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTorsion * >(argp2); (arg1)->SetData(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTorsionData_FillTorsionArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ;
  std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTorsionData_FillTorsionArray", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTorsionData_FillTorsionArray" "', argument " "1"" of type '" "OpenBabel::OBTorsionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTorsionData_FillTorsionArray" "', argument " "2"" of type '" "std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTorsionData_FillTorsionArray" "', argument " "2"" of type '" "std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > * >(argp2);
  result = (bool)(arg1)->FillTorsionArray(*arg2); resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBTorsionData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTorsionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTorsionData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTorsionData" "', argument " "1"" of type '" "OpenBabel::OBTorsionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTorsionData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBTorsionData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTorsionData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBAngle__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAngle *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAngle,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAngle" "', argument " "1"" of type '" "OpenBabel::OBAngle const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAngle" "', argument " "1"" of type '" "OpenBabel::OBAngle const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1);
  result = (OpenBabel::OBAngle *)new OpenBabel::OBAngle((OpenBabel::OBAngle const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAngle, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAngle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBAngle", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBAngle__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBAngle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAngle::OBAngle(OpenBabel::OBAngle const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAngle" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; OpenBabel::OBAngle *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAngle___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle___eq__" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAngle,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngle___eq__" "', argument " "2"" of type '" "OpenBabel::OBAngle const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAngle___eq__" "', argument " "2"" of type '" "OpenBabel::OBAngle const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAngle * >(argp2);
  result = (bool)(arg1)->operator ==((OpenBabel::OBAngle const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBAngle_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle_Clear" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle_GetAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBAngle const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1); result = (double)((OpenBabel::OBAngle const *)arg1)->GetAngle();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle_SetAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngle *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAngle_SetAngle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle_SetAngle" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAngle_SetAngle" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetAngle(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle_SetAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAngle *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  OpenBabel::OBAtom *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle_SetAtoms" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngle_SetAtoms" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBAngle_SetAtoms" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBAngle_SetAtoms" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4); (arg1)->SetAtoms(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle_SetAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAngle *arg1 = 0 ;
  OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngle, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngle_SetAtoms" "', argument " "1"" of type '" "OpenBabel::OBAngle *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngle * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngle_SetAtoms" "', argument " "2"" of type '" "OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAngle_SetAtoms" "', argument " "2"" of type '" "OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &""'"); }
   arg2 = reinterpret_cast< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > * >(argp2);
  (arg1)->SetAtoms(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngle_SetAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAngle_SetAtoms", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBAngle_SetAtoms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBAngle_SetAtoms__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAngle_SetAtoms'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBAngle::SetAtoms(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBAngle::SetAtoms(OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > &)\n");
  return 0; }
SWIGINTERN PyObject *OBAngle_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAngle, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAngle_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBAngleData_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngleData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngleData_Clear" "', argument " "1"" of type '" "OpenBabel::OBAngleData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngleData_FillAngleArray__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAngleData *arg1 = 0 ; int **arg2 = 0 ; unsigned int *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; unsigned int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngleData_FillAngleArray" "', argument " "1"" of type '" "OpenBabel::OBAngleData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_p_int, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngleData_FillAngleArray" "', argument " "2"" of type '" "int **""'");  }
   arg2 = reinterpret_cast< int ** >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBAngleData_FillAngleArray" "', argument " "3"" of type '" "unsigned int &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAngleData_FillAngleArray" "', argument " "3"" of type '" "unsigned int &""'"); }
   arg3 = reinterpret_cast< unsigned int * >(argp3); result = (unsigned int)(arg1)->FillAngleArray(arg2,*arg3);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngleData_FillAngleArray__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAngleData *arg1 = 0 ;
  std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngleData_FillAngleArray" "', argument " "1"" of type '" "OpenBabel::OBAngleData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngleData_FillAngleArray" "', argument " "2"" of type '" "std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAngleData_FillAngleArray" "', argument " "2"" of type '" "std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > * >(argp2);
  result = (bool)(arg1)->FillAngleArray(*arg2); resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAngleData_FillAngleArray(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBAngleData_FillAngleArray", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBAngleData_FillAngleArray__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBAngleData_FillAngleArray__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAngleData_FillAngleArray'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAngleData::FillAngleArray(int **,unsigned int &)\n"
  "    OpenBabel::OBAngleData::FillAngleArray(std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAngleData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngleData *arg1 = 0 ; OpenBabel::OBAngle *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAngleData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngleData_SetData" "', argument " "1"" of type '" "OpenBabel::OBAngleData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAngle,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAngleData_SetData" "', argument " "2"" of type '" "OpenBabel::OBAngle &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAngleData_SetData" "', argument " "2"" of type '" "OpenBabel::OBAngle &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAngle * >(argp2); (arg1)->SetData(*arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAngleData_GetSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngleData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAngleData_GetSize" "', argument " "1"" of type '" "OpenBabel::OBAngleData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1); result = ((OpenBabel::OBAngleData const *)arg1)->GetSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAngleData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAngleData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAngleData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAngleData" "', argument " "1"" of type '" "OpenBabel::OBAngleData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAngleData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBAngleData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAngleData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBSerialNums__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBSerialNums *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBSerialNums *)new OpenBabel::OBSerialNums();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSerialNums, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSerialNums__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSerialNums *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBSerialNums *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBSerialNums,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSerialNums" "', argument " "1"" of type '" "OpenBabel::OBSerialNums const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBSerialNums" "', argument " "1"" of type '" "OpenBabel::OBSerialNums const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBSerialNums * >(argp1);
  result = (OpenBabel::OBSerialNums *)new OpenBabel::OBSerialNums((OpenBabel::OBSerialNums const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSerialNums, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSerialNums(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBSerialNums", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBSerialNums__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBSerialNums__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBSerialNums'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSerialNums::OBSerialNums()\n"
  "    OpenBabel::OBSerialNums::OBSerialNums(OpenBabel::OBSerialNums const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSerialNums_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSerialNums *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSerialNums, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSerialNums_GetData" "', argument " "1"" of type '" "OpenBabel::OBSerialNums *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSerialNums * >(argp1);
  result = (std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > *) &(arg1)->GetData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSerialNums_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSerialNums *arg1 = 0 ;
  std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSerialNums_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSerialNums, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSerialNums_SetData" "', argument " "1"" of type '" "OpenBabel::OBSerialNums *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSerialNums * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSerialNums_SetData" "', argument " "2"" of type '" "std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSerialNums_SetData" "', argument " "2"" of type '" "std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > &""'"); }
  
  arg2 = reinterpret_cast< std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > * >(argp2);
  (arg1)->SetData(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSerialNums(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSerialNums *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSerialNums, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSerialNums" "', argument " "1"" of type '" "OpenBabel::OBSerialNums *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSerialNums * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBSerialNums_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSerialNums, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSerialNums_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBVibrationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBVibrationData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBVibrationData *)new OpenBabel::OBVibrationData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVibrationData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBVibrationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBVibrationData" "', argument " "1"" of type '" "OpenBabel::OBVibrationData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBVibrationData *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *arg2 = 0 ;
  std::vector< double,std::allocator< double > > *arg3 = 0 ; std::vector< double,std::allocator< double > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; int res4 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_SetData" "', argument " "1"" of type '" "OpenBabel::OBVibrationData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBVibrationData_SetData" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     arg2 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBVibrationData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg3 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBVibrationData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg4 = ptr; } 
  (arg1)->SetData((std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)*arg2,(std::vector< double,std::allocator< double > > const &)*arg3,(std::vector< double,std::allocator< double > > const &)*arg4);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBVibrationData *arg1 = 0 ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *arg2 = 0 ;
  std::vector< double,std::allocator< double > > *arg3 = 0 ; std::vector< double,std::allocator< double > > *arg4 = 0 ;
  std::vector< double,std::allocator< double > > *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ; int res4 = SWIG_OLDOBJ ; int res5 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_SetData" "', argument " "1"" of type '" "OpenBabel::OBVibrationData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1); {
    std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *ptr = (std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBVibrationData_SetData" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "2"" of type '" "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &""'"); }
     arg2 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBVibrationData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg3 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBVibrationData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg4 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res5 = swig::asptr(swig_obj[4], &ptr); if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBVibrationData_SetData" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVibrationData_SetData" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg5 = ptr; } 
  (arg1)->SetData((std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &)*arg2,(std::vector< double,std::allocator< double > > const &)*arg3,(std::vector< double,std::allocator< double > > const &)*arg4,(std::vector< double,std::allocator< double > > const &)*arg5);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5; return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBVibrationData_SetData", 0, 5, argv))) SWIG_fail; --argc; if (argc == 4) {
    PyObject *retobj = _wrap_OBVibrationData_SetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBVibrationData_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBVibrationData_SetData'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBVibrationData::SetData(std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &,std::vector< double,std::allocator< double > > const &,std::vector< double,std::allocator< double > > const &)\n"
  "    OpenBabel::OBVibrationData::SetData(std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > const &,std::vector< double,std::allocator< double > > const &,std::vector< double,std::allocator< double > > const &,std::vector< double,std::allocator< double > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBVibrationData_GetLx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_GetLx" "', argument " "1"" of type '" "OpenBabel::OBVibrationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1); result = ((OpenBabel::OBVibrationData const *)arg1)->GetLx();
  resultobj = swig::from(static_cast< std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_GetFrequencies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_GetFrequencies" "', argument " "1"" of type '" "OpenBabel::OBVibrationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1);
  result = ((OpenBabel::OBVibrationData const *)arg1)->GetFrequencies();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_GetIntensities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_GetIntensities" "', argument " "1"" of type '" "OpenBabel::OBVibrationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1);
  result = ((OpenBabel::OBVibrationData const *)arg1)->GetIntensities();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_GetRamanActivities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_GetRamanActivities" "', argument " "1"" of type '" "OpenBabel::OBVibrationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1);
  result = ((OpenBabel::OBVibrationData const *)arg1)->GetRamanActivities();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBVibrationData_GetNumberOfFrequencies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVibrationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVibrationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVibrationData_GetNumberOfFrequencies" "', argument " "1"" of type '" "OpenBabel::OBVibrationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVibrationData * >(argp1);
  result = (unsigned int)((OpenBabel::OBVibrationData const *)arg1)->GetNumberOfFrequencies();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBVibrationData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBVibrationData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBVibrationData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBDOSData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBDOSData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBDOSData *)new OpenBabel::OBDOSData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBDOSData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBDOSData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBDOSData" "', argument " "1"" of type '" "OpenBabel::OBDOSData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBDOSData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; double arg2 ; std::vector< double,std::allocator< double > > *arg3 = 0 ;
  std::vector< double,std::allocator< double > > *arg4 = 0 ; std::vector< double,std::allocator< double > > *arg5 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ; int res3 = SWIG_OLDOBJ ; int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ; PyObject *swig_obj[5] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDOSData_SetData", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDOSData_SetData" "', argument " "1"" of type '" "OpenBabel::OBDOSData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBDOSData_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDOSData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDOSData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg3 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBDOSData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDOSData_SetData" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg4 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res5 = swig::asptr(swig_obj[4], &ptr); if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBDOSData_SetData" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDOSData_SetData" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg5 = ptr; } 
  (arg1)->SetData(arg2,(std::vector< double,std::allocator< double > > const &)*arg3,(std::vector< double,std::allocator< double > > const &)*arg4,(std::vector< double,std::allocator< double > > const &)*arg5);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return NULL; }
SWIGINTERN PyObject *_wrap_OBDOSData_GetFermiEnergy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDOSData_GetFermiEnergy" "', argument " "1"" of type '" "OpenBabel::OBDOSData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1);
  result = (double)((OpenBabel::OBDOSData const *)arg1)->GetFermiEnergy();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDOSData_GetEnergies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDOSData_GetEnergies" "', argument " "1"" of type '" "OpenBabel::OBDOSData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1); result = ((OpenBabel::OBDOSData const *)arg1)->GetEnergies();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBDOSData_GetDensities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDOSData_GetDensities" "', argument " "1"" of type '" "OpenBabel::OBDOSData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1); result = ((OpenBabel::OBDOSData const *)arg1)->GetDensities();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBDOSData_GetIntegration(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDOSData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDOSData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDOSData_GetIntegration" "', argument " "1"" of type '" "OpenBabel::OBDOSData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDOSData * >(argp1); result = ((OpenBabel::OBDOSData const *)arg1)->GetIntegration();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBDOSData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBDOSData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBDOSData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBOrbital_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBOrbital *arg1 = 0 ; double arg2 ; double arg3 ; std::string arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_SetData" "', argument " "1"" of type '" "OpenBabel::OBOrbital *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbital_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBOrbital_SetData" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[3], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbital_SetData" "', argument " "4"" of type '" "std::string""'");  }
     arg4 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetData(arg2,arg3,SWIG_STD_MOVE(arg4));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbital_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBOrbital *arg1 = 0 ; double arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_SetData" "', argument " "1"" of type '" "OpenBabel::OBOrbital *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbital_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBOrbital_SetData" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->SetData(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbital_SetData__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBOrbital *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_SetData" "', argument " "1"" of type '" "OpenBabel::OBOrbital *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbital_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetData(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbital_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBOrbital_SetData", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBOrbital_SetData__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBOrbital_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBOrbital_SetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBOrbital_SetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBOrbital::SetData(double,double,std::string)\n"
  "    OpenBabel::OBOrbital::SetData(double,double)\n" "    OpenBabel::OBOrbital::SetData(double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBOrbital_GetEnergy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbital *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_GetEnergy" "', argument " "1"" of type '" "OpenBabel::OBOrbital const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); result = (double)((OpenBabel::OBOrbital const *)arg1)->GetEnergy();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbital_GetOccupation(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbital *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_GetOccupation" "', argument " "1"" of type '" "OpenBabel::OBOrbital const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1);
  result = (double)((OpenBabel::OBOrbital const *)arg1)->GetOccupation();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbital_GetSymbol(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbital *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbital_GetSymbol" "', argument " "1"" of type '" "OpenBabel::OBOrbital const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); result = ((OpenBabel::OBOrbital const *)arg1)->GetSymbol();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBOrbital(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbital *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBOrbital", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBOrbital *)new OpenBabel::OBOrbital();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBOrbital, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBOrbital(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbital *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbital, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBOrbital" "', argument " "1"" of type '" "OpenBabel::OBOrbital *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbital * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBOrbital_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBOrbital, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBOrbital_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBOrbitalData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBOrbitalData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBOrbitalData *)new OpenBabel::OBOrbitalData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBOrbitalData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBOrbitalData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBOrbitalData" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetAlphaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_SetAlphaOrbitals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_SetAlphaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOrbitalData_SetAlphaOrbitals" "', argument " "2"" of type '" "std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBOrbitalData_SetAlphaOrbitals" "', argument " "2"" of type '" "std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >""'"); }
     else {
      std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > * temp = reinterpret_cast< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetAlphaOrbitals(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetBetaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_SetBetaOrbitals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_SetBetaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOrbitalData_SetBetaOrbitals" "', argument " "2"" of type '" "std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBOrbitalData_SetBetaOrbitals" "', argument " "2"" of type '" "std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >""'"); }
     else {
      std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > * temp = reinterpret_cast< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetBetaOrbitals(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetHOMO__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBOrbitalData *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_SetHOMO" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbitalData_SetHOMO" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBOrbitalData_SetHOMO" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); (arg1)->SetHOMO(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetHOMO__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBOrbitalData *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_SetHOMO" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbitalData_SetHOMO" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetHOMO(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetHOMO(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBOrbitalData_SetHOMO", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBOrbitalData_SetHOMO__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBOrbitalData_SetHOMO__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBOrbitalData_SetHOMO'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBOrbitalData::SetHOMO(int,int)\n"
  "    OpenBabel::OBOrbitalData::SetHOMO(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_SetOpenShell(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_SetOpenShell", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_SetOpenShell" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBOrbitalData_SetOpenShell" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetOpenShell(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_IsOpenShell(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_IsOpenShell" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); result = (bool)(arg1)->IsOpenShell();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_GetAlphaHOMO(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_GetAlphaHOMO" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); result = (unsigned int)(arg1)->GetAlphaHOMO();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_GetBetaHOMO(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_GetBetaHOMO" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); result = (unsigned int)(arg1)->GetBetaHOMO();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_GetAlphaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_GetAlphaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); result = (arg1)->GetAlphaOrbitals();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_GetBetaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_GetBetaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); result = (arg1)->GetBetaOrbitals();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_LoadClosedShellOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; std::vector< double,std::allocator< double > > arg2 ;
  std::vector< std::string,std::allocator< std::string > > arg3 ; unsigned int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_LoadClosedShellOrbitals", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_LoadClosedShellOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadClosedShellOrbitals" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    int res = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadClosedShellOrbitals" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > >""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBOrbitalData_LoadClosedShellOrbitals" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); (arg1)->LoadClosedShellOrbitals(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_LoadAlphaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; std::vector< double,std::allocator< double > > arg2 ;
  std::vector< std::string,std::allocator< std::string > > arg3 ; unsigned int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_LoadAlphaOrbitals", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_LoadAlphaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadAlphaOrbitals" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    int res = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadAlphaOrbitals" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > >""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBOrbitalData_LoadAlphaOrbitals" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); (arg1)->LoadAlphaOrbitals(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOrbitalData_LoadBetaOrbitals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOrbitalData *arg1 = 0 ; std::vector< double,std::allocator< double > > arg2 ;
  std::vector< std::string,std::allocator< std::string > > arg3 ; unsigned int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOrbitalData_LoadBetaOrbitals", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOrbitalData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOrbitalData_LoadBetaOrbitals" "', argument " "1"" of type '" "OpenBabel::OBOrbitalData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOrbitalData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadBetaOrbitals" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    int res = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBOrbitalData_LoadBetaOrbitals" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > >""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBOrbitalData_LoadBetaOrbitals" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); (arg1)->LoadBetaOrbitals(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3),arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBOrbitalData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBOrbitalData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBOrbitalData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBElectronicTransitionData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBElectronicTransitionData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBElectronicTransitionData *)new OpenBabel::OBElectronicTransitionData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBElectronicTransitionData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBElectronicTransitionData" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ;
  std::vector< double,std::allocator< double > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBElectronicTransitionData_SetData", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_SetData" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBElectronicTransitionData_SetData" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBElectronicTransitionData_SetData" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  { std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBElectronicTransitionData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBElectronicTransitionData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg3 = ptr; } 
  (arg1)->SetData((std::vector< double,std::allocator< double > > const &)*arg2,(std::vector< double,std::allocator< double > > const &)*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return resultobj;
  fail: if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_SetEDipole(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBElectronicTransitionData_SetEDipole", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_SetEDipole" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBElectronicTransitionData_SetEDipole" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBElectronicTransitionData_SetEDipole" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetEDipole((std::vector< double,std::allocator< double > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_SetRotatoryStrengthsVelocity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBElectronicTransitionData *arg1 = 0 ;
  std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBElectronicTransitionData_SetRotatoryStrengthsVelocity", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsVelocity" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsVelocity" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsVelocity" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetRotatoryStrengthsVelocity((std::vector< double,std::allocator< double > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_SetRotatoryStrengthsLength(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBElectronicTransitionData *arg1 = 0 ;
  std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBElectronicTransitionData_SetRotatoryStrengthsLength", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsLength" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsLength" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBElectronicTransitionData_SetRotatoryStrengthsLength" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetRotatoryStrengthsLength((std::vector< double,std::allocator< double > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_GetWavelengths(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_GetWavelengths" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1);
  result = ((OpenBabel::OBElectronicTransitionData const *)arg1)->GetWavelengths();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_GetForces(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_GetForces" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1);
  result = ((OpenBabel::OBElectronicTransitionData const *)arg1)->GetForces();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_GetEDipole(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_GetEDipole" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1);
  result = ((OpenBabel::OBElectronicTransitionData const *)arg1)->GetEDipole();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_GetRotatoryStrengthsVelocity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_GetRotatoryStrengthsVelocity" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1);
  result = ((OpenBabel::OBElectronicTransitionData const *)arg1)->GetRotatoryStrengthsVelocity();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBElectronicTransitionData_GetRotatoryStrengthsLength(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBElectronicTransitionData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBElectronicTransitionData_GetRotatoryStrengthsLength" "', argument " "1"" of type '" "OpenBabel::OBElectronicTransitionData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBElectronicTransitionData * >(argp1);
  result = ((OpenBabel::OBElectronicTransitionData const *)arg1)->GetRotatoryStrengthsLength();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBElectronicTransitionData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBElectronicTransitionData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBElectronicTransitionData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBRotationData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotationData *)new OpenBabel::OBRotationData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotationData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotationData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotationData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotationData" "', argument " "1"" of type '" "OpenBabel::OBRotationData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotationData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotationData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *arg1 = 0 ; OpenBabel::OBRotationData::RType arg2 ;
  std::vector< double,std::allocator< double > > arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotationData_SetData", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotationData_SetData" "', argument " "1"" of type '" "OpenBabel::OBRotationData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotationData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotationData_SetData" "', argument " "2"" of type '" "OpenBabel::OBRotationData::RType""'"); }
    arg2 = static_cast< OpenBabel::OBRotationData::RType >(val2); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBRotationData_SetData" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBRotationData_SetData" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->SetData(arg2,SWIG_STD_MOVE(arg3),arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotationData_GetRotConsts(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotationData_GetRotConsts" "', argument " "1"" of type '" "OpenBabel::OBRotationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotationData * >(argp1);
  result = ((OpenBabel::OBRotationData const *)arg1)->GetRotConsts();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotationData_GetSymmetryNumber(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotationData_GetSymmetryNumber" "', argument " "1"" of type '" "OpenBabel::OBRotationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotationData * >(argp1);
  result = (int)((OpenBabel::OBRotationData const *)arg1)->GetSymmetryNumber();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotationData_GetRotorType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotationData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBRotationData::RType result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotationData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotationData_GetRotorType" "', argument " "1"" of type '" "OpenBabel::OBRotationData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotationData * >(argp1);
  result = (OpenBabel::OBRotationData::RType)((OpenBabel::OBRotationData const *)arg1)->GetRotorType();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRotationData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotationData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotationData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBVectorData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVectorData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBVectorData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBVectorData *)new OpenBabel::OBVectorData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBVectorData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBVectorData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVectorData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVectorData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBVectorData" "', argument " "1"" of type '" "OpenBabel::OBVectorData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVectorData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBVectorData_SetData__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBVectorData *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVectorData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVectorData_SetData" "', argument " "1"" of type '" "OpenBabel::OBVectorData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVectorData * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBVectorData_SetData" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBVectorData_SetData" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBVectorData_SetData" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); (arg1)->SetData(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBVectorData_SetData__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBVectorData *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVectorData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVectorData_SetData" "', argument " "1"" of type '" "OpenBabel::OBVectorData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVectorData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBVectorData_SetData" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBVectorData_SetData" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetData(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBVectorData_SetData(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBVectorData_SetData", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBVectorData_SetData__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBVectorData_SetData__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBVectorData_SetData'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBVectorData::SetData(double,double,double)\n"
  "    OpenBabel::OBVectorData::SetData(OpenBabel::vector3)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBVectorData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBVectorData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBVectorData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBVectorData_GetData" "', argument " "1"" of type '" "OpenBabel::OBVectorData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBVectorData * >(argp1); result = ((OpenBabel::OBVectorData const *)arg1)->GetData();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBVectorData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBVectorData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBVectorData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMatrixData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMatrixData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBMatrixData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBMatrixData *)new OpenBabel::OBMatrixData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMatrixData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBMatrixData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMatrixData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMatrixData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMatrixData" "', argument " "1"" of type '" "OpenBabel::OBMatrixData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMatrixData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMatrixData_SetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMatrixData *arg1 = 0 ; OpenBabel::matrix3x3 arg2 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMatrixData_SetData", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMatrixData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMatrixData_SetData" "', argument " "1"" of type '" "OpenBabel::OBMatrixData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMatrixData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__matrix3x3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMatrixData_SetData" "', argument " "2"" of type '" "OpenBabel::matrix3x3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMatrixData_SetData" "', argument " "2"" of type '" "OpenBabel::matrix3x3""'"); }
     else { OpenBabel::matrix3x3 * temp = reinterpret_cast< OpenBabel::matrix3x3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetData(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMatrixData_GetData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMatrixData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::matrix3x3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMatrixData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMatrixData_GetData" "', argument " "1"" of type '" "OpenBabel::OBMatrixData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMatrixData * >(argp1); result = ((OpenBabel::OBMatrixData const *)arg1)->GetData();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBMatrixData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMatrixData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMatrixData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBFreeGridPoint__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBFreeGridPoint *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBFreeGridPoint *)new OpenBabel::OBFreeGridPoint();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFreeGridPoint, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFreeGridPoint__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double arg2 ; double arg3 ; double arg4 ; double val1 ; int ecode1 = 0 ; double val2 ;
  int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; OpenBabel::OBFreeGridPoint *result = 0 ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBFreeGridPoint" "', argument " "1"" of type '" "double""'"); }
    arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBFreeGridPoint" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBFreeGridPoint" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBFreeGridPoint" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4);
  result = (OpenBabel::OBFreeGridPoint *)new OpenBabel::OBFreeGridPoint(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFreeGridPoint, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFreeGridPoint(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBFreeGridPoint", 0, 4, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBFreeGridPoint__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBFreeGridPoint__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBFreeGridPoint'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBFreeGridPoint::OBFreeGridPoint()\n"
  "    OpenBabel::OBFreeGridPoint::OBFreeGridPoint(double,double,double,double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBFreeGridPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFreeGridPoint" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_GetX(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_GetX" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); result = (double)(arg1)->GetX();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_GetY(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_GetY" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); result = (double)(arg1)->GetY();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_GetZ(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_GetZ" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); result = (double)(arg1)->GetZ();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_GetV(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_GetV" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); result = (double)(arg1)->GetV();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_SetX(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFreeGridPoint_SetX", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_SetX" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFreeGridPoint_SetX" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetX(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_SetY(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFreeGridPoint_SetY", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_SetY" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFreeGridPoint_SetY" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetY(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_SetZ(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFreeGridPoint_SetZ", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_SetZ" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFreeGridPoint_SetZ" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetZ(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGridPoint_SetV(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGridPoint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFreeGridPoint_SetV", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGridPoint_SetV" "', argument " "1"" of type '" "OpenBabel::OBFreeGridPoint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGridPoint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFreeGridPoint_SetV" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetV(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFreeGridPoint_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFreeGridPoint, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFreeGridPoint_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBFreeGrid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBFreeGrid", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFreeGrid *)new OpenBabel::OBFreeGrid();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFreeGrid, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFreeGrid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFreeGrid" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_NumPoints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_NumPoints" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); result = (int)(arg1)->NumPoints();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_AddPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; double arg5 ; void *argp1 = 0 ; int res1 = 0 ;
  double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ;
  PyObject *swig_obj[5] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFreeGrid_AddPoint", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_AddPoint" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFreeGrid_AddPoint" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFreeGrid_AddPoint" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBFreeGrid_AddPoint" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBFreeGrid_AddPoint" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); (arg1)->AddPoint(arg2,arg3,arg4,arg5); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_BeginPoints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBFreeGridPoint *,std::allocator< OpenBabel::OBFreeGridPoint * > >::iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_BeginPoints" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); result = (arg1)->BeginPoints();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBFreeGridPointIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_EndPoints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBFreeGridPoint *,std::allocator< OpenBabel::OBFreeGridPoint * > >::iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_EndPoints" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); result = (arg1)->EndPoints();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBFreeGridPointIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_BeginPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; OpenBabel::OBFreeGridPointIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBFreeGridPoint *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFreeGrid_BeginPoint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_BeginPoint" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFreeGrid_BeginPoint" "', argument " "2"" of type '" "OpenBabel::OBFreeGridPointIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFreeGrid_BeginPoint" "', argument " "2"" of type '" "OpenBabel::OBFreeGridPointIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBFreeGridPointIterator * >(argp2);
  result = (OpenBabel::OBFreeGridPoint *)(arg1)->BeginPoint(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_NextPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; OpenBabel::OBFreeGridPointIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBFreeGridPoint *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFreeGrid_NextPoint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_NextPoint" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFreeGrid_NextPoint" "', argument " "2"" of type '" "OpenBabel::OBFreeGridPointIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFreeGrid_NextPoint" "', argument " "2"" of type '" "OpenBabel::OBFreeGridPointIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBFreeGridPointIterator * >(argp2);
  result = (OpenBabel::OBFreeGridPoint *)(arg1)->NextPoint(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFreeGridPoint, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFreeGrid_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFreeGrid *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFreeGrid, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFreeGrid_Clear" "', argument " "1"" of type '" "OpenBabel::OBFreeGrid *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFreeGrid * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBFreeGrid_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFreeGrid, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFreeGrid_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBPcharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPcharge *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBPcharge", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBPcharge *)new OpenBabel::OBPcharge();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPcharge, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBPcharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPcharge *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPcharge, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPcharge" "', argument " "1"" of type '" "OpenBabel::OBPcharge *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPcharge * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBPcharge_NumPartialCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPcharge *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPcharge, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPcharge_NumPartialCharges" "', argument " "1"" of type '" "OpenBabel::OBPcharge *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPcharge * >(argp1); result = (int)(arg1)->NumPartialCharges();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPcharge_AddPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPcharge *arg1 = 0 ; std::vector< double,std::allocator< double > > arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPcharge_AddPartialCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPcharge, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPcharge_AddPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBPcharge *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPcharge * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBPcharge_AddPartialCharge" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->AddPartialCharge(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPcharge_GetPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPcharge *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPcharge, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPcharge_GetPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBPcharge *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPcharge * >(argp1); result = (arg1)->GetPartialCharge();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBPcharge_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPcharge, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBPcharge_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBPairInteger(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< int > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBPairInteger", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBPairTemplate< int > *)new OpenBabel::OBPairTemplate< int >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPairInteger_SetValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< int > *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBPairInteger_SetValue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairInteger_SetValue" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< int > *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< int > * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBPairInteger_SetValue" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetValue(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPairInteger_GetGenericValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairInteger_GetGenericValue" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< int > const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< int > * >(argp1);
  result = (int *) &((OpenBabel::OBPairTemplate< int > const *)arg1)->GetGenericValue();
  resultobj = SWIG_From_int(static_cast< int >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBPairInteger(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< int > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPairInteger" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< int > *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< int > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBPairInteger_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPairTemplateT_int_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBPairInteger_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBPairFloatingPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< float > *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBPairFloatingPoint", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBPairTemplate< float > *)new OpenBabel::OBPairTemplate< float >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPairFloatingPoint_SetValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< float > *arg1 = 0 ; float arg2 ; void *argp1 = 0 ; int res1 = 0 ; float val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPairFloatingPoint_SetValue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairFloatingPoint_SetValue" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< float > *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< float > * >(argp1); ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBPairFloatingPoint_SetValue" "', argument " "2"" of type '" "float""'"); }
    arg2 = static_cast< float >(val2); (arg1)->SetValue(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPairFloatingPoint_GetGenericValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< float > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; float *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPairFloatingPoint_GetGenericValue" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< float > const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< float > * >(argp1);
  result = (float *) &((OpenBabel::OBPairTemplate< float > const *)arg1)->GetGenericValue();
  resultobj = SWIG_From_float(static_cast< float >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBPairFloatingPoint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPairTemplate< float > *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPairFloatingPoint" "', argument " "1"" of type '" "OpenBabel::OBPairTemplate< float > *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPairTemplate< float > * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBPairFloatingPoint_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPairTemplateT_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBPairFloatingPoint_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBGridData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBGridData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBGridData *)new OpenBabel::OBGridData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGridData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBGridData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBGridData" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetAxes__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; double *arg2 = (double *)0 ; double *arg3 = (double *)0 ;
  double *arg4 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetAxes" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetAxes" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBGridData_GetAxes" "', argument " "3"" of type '" "double [3]""'");  }
    arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBGridData_GetAxes" "', argument " "4"" of type '" "double [3]""'");  }
    arg4 = reinterpret_cast< double * >(argp4); ((OpenBabel::OBGridData const *)arg1)->GetAxes(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetXAxis(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetXAxis" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetXAxis();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetYAxis(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetYAxis" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetYAxis();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetZAxis(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetZAxis" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetZAxis();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetAxes__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  OpenBabel::vector3 *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetAxes" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetAxes" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetAxes" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBGridData_GetAxes" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetAxes" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBGridData_GetAxes" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetAxes" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4); ((OpenBabel::OBGridData const *)arg1)->GetAxes(*arg2,*arg3,*arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetAxes(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBGridData_GetAxes", 0, 4, argv))) SWIG_fail; --argc; if (argc == 4) { int _v = 0;
    { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_double, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_double, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_double, 0); _v = SWIG_CheckState(res);}  if (!_v) goto check_1;
    return _wrap_OBGridData_GetAxes__SWIG_0(self, argc, argv);}  check_1: if (argc == 4) {
    PyObject *retobj = _wrap_OBGridData_GetAxes__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBGridData_GetAxes'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBGridData::GetAxes(double [3],double [3],double [3]) const\n"
  "    OpenBabel::OBGridData::GetAxes(OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBGridData_GetNumberOfPoints__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; int *arg2 = 0 ; int *arg3 = 0 ; int *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetNumberOfPoints" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetNumberOfPoints" "', argument " "2"" of type '" "int &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetNumberOfPoints" "', argument " "2"" of type '" "int &""'"); }
   arg2 = reinterpret_cast< int * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBGridData_GetNumberOfPoints" "', argument " "3"" of type '" "int &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetNumberOfPoints" "', argument " "3"" of type '" "int &""'"); }
   arg3 = reinterpret_cast< int * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBGridData_GetNumberOfPoints" "', argument " "4"" of type '" "int &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetNumberOfPoints" "', argument " "4"" of type '" "int &""'"); }
   arg4 = reinterpret_cast< int * >(argp4); ((OpenBabel::OBGridData const *)arg1)->GetNumberOfPoints(*arg2,*arg3,*arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetNumberOfPoints__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetNumberOfPoints" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (int)((OpenBabel::OBGridData const *)arg1)->GetNumberOfPoints();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetNumberOfPoints(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBGridData_GetNumberOfPoints", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBGridData_GetNumberOfPoints__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBGridData_GetNumberOfPoints__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBGridData_GetNumberOfPoints'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBGridData::GetNumberOfPoints(int &,int &,int &) const\n"
  "    OpenBabel::OBGridData::GetNumberOfPoints() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBGridData_GetNumberOfSteps(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; int *arg2 = (int *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_GetNumberOfSteps", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetNumberOfSteps" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetNumberOfSteps" "', argument " "2"" of type '" "int [3]""'");  }
    arg2 = reinterpret_cast< int * >(argp2); ((OpenBabel::OBGridData const *)arg1)->GetNumberOfSteps(arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetValues" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetValues();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; double result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetValue" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_GetValue" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBGridData_GetValue" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBGridData_GetValue" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); result = (double)((OpenBabel::OBGridData const *)arg1)->GetValue(arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; OpenBabel::vector3 arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetValue" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetValue" "', argument " "2"" of type '" "OpenBabel::vector3""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_GetValue" "', argument " "2"" of type '" "OpenBabel::vector3""'"); }
     else { OpenBabel::vector3 * temp = reinterpret_cast< OpenBabel::vector3 * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = (double)((OpenBabel::OBGridData const *)arg1)->GetValue(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetValue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBGridData_GetValue", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBGridData_GetValue__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBGridData_GetValue__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBGridData_GetValue'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBGridData::GetValue(int,int,int) const\n"
  "    OpenBabel::OBGridData::GetValue(OpenBabel::vector3) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBGridData_GetUnit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBGridData::Unit result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetUnit" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (OpenBabel::OBGridData::Unit)((OpenBabel::OBGridData const *)arg1)->GetUnit();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetMinValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetMinValue" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (double)((OpenBabel::OBGridData const *)arg1)->GetMinValue();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetMaxValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetMaxValue" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (double)((OpenBabel::OBGridData const *)arg1)->GetMaxValue();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetOriginVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetOriginVector" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetOriginVector();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetOriginVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetOriginVector" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_GetOriginVector" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); ((OpenBabel::OBGridData const *)arg1)->GetOriginVector(arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetOriginVector(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBGridData_GetOriginVector", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBGridData_GetOriginVector__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBGridData_GetOriginVector__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBGridData_GetOriginVector'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBGridData::GetOriginVector() const\n"
  "    OpenBabel::OBGridData::GetOriginVector(double [3]) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBGridData_GetMaxVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::vector3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetMaxVector" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); result = ((OpenBabel::OBGridData const *)arg1)->GetMaxVector();
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetUnrestricted(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetUnrestricted" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (bool)((OpenBabel::OBGridData const *)arg1)->GetUnrestricted();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_GetNumSymmetries(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_GetNumSymmetries" "', argument " "1"" of type '" "OpenBabel::OBGridData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  result = (int)((OpenBabel::OBGridData const *)arg1)->GetNumSymmetries();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetNumberOfPoints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetNumberOfPoints", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetNumberOfPoints" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_SetNumberOfPoints" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBGridData_SetNumberOfPoints" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBGridData_SetNumberOfPoints" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->SetNumberOfPoints(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetLimits__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  OpenBabel::vector3 *arg4 = 0 ; OpenBabel::vector3 *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetLimits" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_SetLimits" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_SetLimits" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBGridData_SetLimits" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_SetLimits" "', argument " "3"" of type '" "OpenBabel::vector3 const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBGridData_SetLimits" "', argument " "4"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_SetLimits" "', argument " "4"" of type '" "OpenBabel::vector3 const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBGridData_SetLimits" "', argument " "5"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp5) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_SetLimits" "', argument " "5"" of type '" "OpenBabel::vector3 const &""'"); }
   arg5 = reinterpret_cast< OpenBabel::vector3 * >(argp5);
  (arg1)->SetLimits((OpenBabel::vector3 const &)*arg2,(OpenBabel::vector3 const &)*arg3,(OpenBabel::vector3 const &)*arg4,(OpenBabel::vector3 const &)*arg5);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetLimits__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBGridData *arg1 = 0 ; double *arg2 = (double *)0 ; double *arg3 = (double *)0 ;
  double *arg4 = (double *)0 ; double *arg5 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetLimits" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_SetLimits" "', argument " "2"" of type '" "double const [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBGridData_SetLimits" "', argument " "3"" of type '" "double const [3]""'");  }
    arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBGridData_SetLimits" "', argument " "4"" of type '" "double const [3]""'");  }
    arg4 = reinterpret_cast< double * >(argp4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBGridData_SetLimits" "', argument " "5"" of type '" "double const [3]""'");  }
    arg5 = reinterpret_cast< double * >(argp5);
  (arg1)->SetLimits((double const (*))arg2,(double const (*))arg3,(double const (*))arg4,(double const (*))arg5);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetLimits(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBGridData_SetLimits", 0, 5, argv))) SWIG_fail; --argc; if (argc == 5) {
    int _v = 0; { int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; {
      int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; { int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; {
      int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_OBGridData_SetLimits__SWIG_0(self, argc, argv);}  check_1: if (argc == 5) {
    PyObject *retobj = _wrap_OBGridData_SetLimits__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBGridData_SetLimits'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBGridData::SetLimits(OpenBabel::vector3 const &,OpenBabel::vector3 const &,OpenBabel::vector3 const &,OpenBabel::vector3 const &)\n"
  "    OpenBabel::OBGridData::SetLimits(double const [3],double const [3],double const [3],double const [3])\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBGridData_SetValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; double arg5 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ;
  PyObject *swig_obj[5] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetValue", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetValue" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_SetValue" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBGridData_SetValue" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBGridData_SetValue" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBGridData_SetValue" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); result = (bool)(arg1)->SetValue(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetValues", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetValues" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGridData_SetValues" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGridData_SetValues" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetValues((std::vector< double,std::allocator< double > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetUnit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; OpenBabel::OBGridData::Unit arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetUnit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetUnit" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_SetUnit" "', argument " "2"" of type '" "OpenBabel::OBGridData::Unit""'"); }
    arg2 = static_cast< OpenBabel::OBGridData::Unit >(val2); (arg1)->SetUnit(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetUnrestricted(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetUnrestricted", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetUnrestricted" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_SetUnrestricted" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetUnrestricted(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBGridData_SetNumSymmetries(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGridData *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGridData_SetNumSymmetries", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGridData_SetNumSymmetries" "', argument " "1"" of type '" "OpenBabel::OBGridData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGridData * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBGridData_SetNumSymmetries" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetNumSymmetries(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBGridData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBGridData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBGridData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBKekulize(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBKekulize" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)OpenBabel::OBKekulize(arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBChainsParser(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChainsParser *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBChainsParser", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBChainsParser *)new OpenBabel::OBChainsParser();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBChainsParser, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBChainsParser(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChainsParser *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChainsParser, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBChainsParser" "', argument " "1"" of type '" "OpenBabel::OBChainsParser *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChainsParser * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChainsParser_PerceiveChains__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBChainsParser *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChainsParser, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChainsParser_PerceiveChains" "', argument " "1"" of type '" "OpenBabel::OBChainsParser *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChainsParser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChainsParser_PerceiveChains" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChainsParser_PerceiveChains" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBChainsParser_PerceiveChains" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->PerceiveChains(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChainsParser_PerceiveChains__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBChainsParser *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChainsParser, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChainsParser_PerceiveChains" "', argument " "1"" of type '" "OpenBabel::OBChainsParser *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChainsParser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChainsParser_PerceiveChains" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChainsParser_PerceiveChains" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->PerceiveChains(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChainsParser_PerceiveChains(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBChainsParser_PerceiveChains", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBChainsParser_PerceiveChains__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBChainsParser_PerceiveChains__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBChainsParser_PerceiveChains'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBChainsParser::PerceiveChains(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBChainsParser::PerceiveChains(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *OBChainsParser_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBChainsParser, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBChainsParser_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_chainsparser_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBChainsParser,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::chainsparser""' of type '""OpenBabel::OBChainsParser""'"); }
     if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::chainsparser""' of type '""OpenBabel::OBChainsParser""'"); }
     else { OpenBabel::OBChainsParser * temp; temp  = reinterpret_cast< OpenBabel::OBChainsParser * >(argp);
      OpenBabel::chainsparser = *temp; if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_chainsparser_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::chainsparser), SWIGTYPE_p_OpenBabel__OBChainsParser,  0 );
  return pyobj; }
SWIGINTERN PyObject *_wrap_new_OBAtomTyper__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBAtomTyper *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBAtomTyper *)new OpenBabel::OBAtomTyper();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomTyper, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtomTyper__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomTyper *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomTyper *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAtomTyper,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAtomTyper" "', argument " "1"" of type '" "OpenBabel::OBAtomTyper const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAtomTyper" "', argument " "1"" of type '" "OpenBabel::OBAtomTyper const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAtomTyper * >(argp1);
  result = (OpenBabel::OBAtomTyper *)new OpenBabel::OBAtomTyper((OpenBabel::OBAtomTyper const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomTyper, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtomTyper(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBAtomTyper", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBAtomTyper__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBAtomTyper__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBAtomTyper'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomTyper::OBAtomTyper()\n"
  "    OpenBabel::OBAtomTyper::OBAtomTyper(OpenBabel::OBAtomTyper const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBAtomTyper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomTyper *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomTyper, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAtomTyper" "', argument " "1"" of type '" "OpenBabel::OBAtomTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomTyper * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomTyper_AssignHyb(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomTyper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtomTyper_AssignHyb", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomTyper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomTyper_AssignHyb" "', argument " "1"" of type '" "OpenBabel::OBAtomTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomTyper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtomTyper_AssignHyb" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtomTyper_AssignHyb" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->AssignHyb(*arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomTyper_AssignTypes(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomTyper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtomTyper_AssignTypes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomTyper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomTyper_AssignTypes" "', argument " "1"" of type '" "OpenBabel::OBAtomTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomTyper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtomTyper_AssignTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtomTyper_AssignTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->AssignTypes(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBAtomTyper_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomTyper, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAtomTyper_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_atomtyper_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBAtomTyper,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::atomtyper""' of type '""OpenBabel::OBAtomTyper""'"); } 
    if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::atomtyper""' of type '""OpenBabel::OBAtomTyper""'"); }
     else { OpenBabel::OBAtomTyper * temp; temp  = reinterpret_cast< OpenBabel::OBAtomTyper * >(argp);
      OpenBabel::atomtyper = *temp; if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_atomtyper_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::atomtyper), SWIGTYPE_p_OpenBabel__OBAtomTyper,  0 ); return pyobj; }
SWIGINTERN PyObject *_wrap_new_OBAromaticTyper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAromaticTyper *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBAromaticTyper", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBAromaticTyper *)new OpenBabel::OBAromaticTyper();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAromaticTyper, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAromaticTyper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAromaticTyper *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAromaticTyper, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAromaticTyper" "', argument " "1"" of type '" "OpenBabel::OBAromaticTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAromaticTyper * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAromaticTyper_AssignAromaticFlags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAromaticTyper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAromaticTyper_AssignAromaticFlags", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAromaticTyper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAromaticTyper_AssignAromaticFlags" "', argument " "1"" of type '" "OpenBabel::OBAromaticTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAromaticTyper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAromaticTyper_AssignAromaticFlags" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAromaticTyper_AssignAromaticFlags" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->AssignAromaticFlags(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBAromaticTyper_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAromaticTyper, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAromaticTyper_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_aromtyper_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBAromaticTyper,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::aromtyper""' of type '""OpenBabel::OBAromaticTyper""'"); }
     if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::aromtyper""' of type '""OpenBabel::OBAromaticTyper""'"); }
     else { OpenBabel::OBAromaticTyper * temp; temp  = reinterpret_cast< OpenBabel::OBAromaticTyper * >(argp);
      OpenBabel::aromtyper = *temp; if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_aromtyper_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::aromtyper), SWIGTYPE_p_OpenBabel__OBAromaticTyper,  0 ); return pyobj; }
SWIGINTERN PyObject *_wrap_new_OBRingTyper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingTyper *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRingTyper", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRingTyper *)new OpenBabel::OBRingTyper();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRingTyper, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRingTyper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingTyper *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingTyper, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRingTyper" "', argument " "1"" of type '" "OpenBabel::OBRingTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingTyper * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRingTyper_AssignTypes(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingTyper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingTyper_AssignTypes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingTyper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingTyper_AssignTypes" "', argument " "1"" of type '" "OpenBabel::OBRingTyper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingTyper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingTyper_AssignTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingTyper_AssignTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->AssignTypes(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRingTyper_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRingTyper, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRingTyper_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CharPtrLess___call__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::CharPtrLess *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "CharPtrLess___call__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__CharPtrLess, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharPtrLess___call__" "', argument " "1"" of type '" "OpenBabel::CharPtrLess const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::CharPtrLess * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CharPtrLess___call__" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CharPtrLess___call__" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3);
  result = (bool)((OpenBabel::CharPtrLess const *)arg1)->operator ()((char const *)arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_new_CharPtrLess(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::CharPtrLess *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_CharPtrLess", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::CharPtrLess *)new OpenBabel::CharPtrLess();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__CharPtrLess, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_CharPtrLess(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::CharPtrLess *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__CharPtrLess, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CharPtrLess" "', argument " "1"" of type '" "OpenBabel::CharPtrLess *""'");  }
   arg1 = reinterpret_cast< OpenBabel::CharPtrLess * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *CharPtrLess_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__CharPtrLess, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CharPtrLess_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBPlugin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPlugin" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Description(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_Description" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); result = (char *)(arg1)->Description();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_TypeID(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_TypeID" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); result = (char *)(arg1)->TypeID();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Display__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPlugin *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; char *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; int res4 ;
  char *buf4 = 0 ; int alloc4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_Display" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPlugin_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBPlugin_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBPlugin_Display" "', argument " "4"" of type '" "char const *""'"); }
   arg4 = reinterpret_cast< char * >(buf4); result = (bool)(arg1)->Display(*arg2,(char const *)arg3,(char const *)arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return resultobj; fail: if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Display__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPlugin *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_Display" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPlugin_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBPlugin_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->Display(*arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Display(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBPlugin_Display", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBPlugin_Display__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBPlugin_Display__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBPlugin_Display'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBPlugin::Display(std::string &,char const *,char const *)\n"
  "    OpenBabel::OBPlugin::Display(std::string &,char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBPlugin_MakeInstance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; std::vector< std::string,std::allocator< std::string > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; OpenBabel::OBPlugin *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPlugin_MakeInstance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_MakeInstance" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_MakeInstance" "', argument " "2"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPlugin_MakeInstance" "', argument " "2"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); }
     arg2 = ptr; } 
  result = (OpenBabel::OBPlugin *)(arg1)->MakeInstance((std::vector< std::string,std::allocator< std::string > > const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Init(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_Init" "', argument " "1"" of type '" "OpenBabel::OBPlugin *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); (arg1)->Init(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_GetPlugin(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  char *arg2 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBPlugin *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPlugin_GetPlugin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_GetPlugin" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_GetPlugin" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2);
  result = (OpenBabel::OBPlugin *)OpenBabel::OBPlugin::GetPlugin((char const *)arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_GetID(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_GetID" "', argument " "1"" of type '" "OpenBabel::OBPlugin const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1); result = (char *)((OpenBabel::OBPlugin const *)arg1)->GetID();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_ListAsVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  char *arg2 = 0 ; std::vector< std::string,std::allocator< std::string > > *arg3 = 0 ; int res1 ; char *buf1 = 0 ;
  int alloc1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBPlugin_ListAsVector", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_ListAsVector" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_ListAsVector" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__string_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBPlugin_ListAsVector" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPlugin_ListAsVector" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'"); }
   arg3 = reinterpret_cast< std::vector< std::string,std::allocator< std::string > > * >(argp3);
  result = (bool)OpenBabel::OBPlugin::ListAsVector((char const *)arg1,(char const *)arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_List__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; char *arg2 = 0 ; std::ostream *arg3 = 0 ; int res1 ; char *buf1 = 0 ;
  int alloc1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_List" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_List" "', argument " "2"" of type '" "char const *""'"); } 
  arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__ostream, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBPlugin_List" "', argument " "3"" of type '" "std::ostream *""'");  }
   arg3 = reinterpret_cast< std::ostream * >(argp3); OpenBabel::OBPlugin::List((char const *)arg1,(char const *)arg2,arg3);
  resultobj = SWIG_Py_Void(); if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_List__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; char *arg2 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_List" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_List" "', argument " "2"" of type '" "char const *""'"); } 
  arg2 = reinterpret_cast< char * >(buf2); OpenBabel::OBPlugin::List((char const *)arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void(); if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_List__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_List" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); OpenBabel::OBPlugin::List((char const *)arg1); resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_List(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBPlugin_List", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBPlugin_List__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_OBPlugin_List__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBPlugin_List__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBPlugin_List'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBPlugin::List(char const *,char const *,std::ostream *)\n"
  "    OpenBabel::OBPlugin::List(char const *,char const *)\n" "    OpenBabel::OBPlugin::List(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBPlugin_ListAsString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; char *arg2 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; std::string result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_ListAsString" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPlugin_ListAsString" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = OpenBabel::OBPlugin::ListAsString((char const *)arg1,(char const *)arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_ListAsString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; std::string result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_ListAsString" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); result = OpenBabel::OBPlugin::ListAsString((char const *)arg1);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_ListAsString(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBPlugin_ListAsString", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBPlugin_ListAsString__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBPlugin_ListAsString__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBPlugin_ListAsString'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBPlugin::ListAsString(char const *,char const *)\n"
  "    OpenBabel::OBPlugin::ListAsString(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBPlugin_FirstLine(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_FirstLine" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); result = OpenBabel::OBPlugin::FirstLine((char const *)arg1);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_Begin(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ;
  char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< char const *,OpenBabel::OBPlugin *,OpenBabel::CharPtrLess,std::allocator< std::pair< char const *const,OpenBabel::OBPlugin * > > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_Begin" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); result = OpenBabel::OBPlugin::Begin((char const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBPlugin::PluginIterator(result)), SWIGTYPE_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_End(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ;
  char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::map< char const *,OpenBabel::OBPlugin *,OpenBabel::CharPtrLess,std::allocator< std::pair< char const *const,OpenBabel::OBPlugin * > > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_End" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); result = OpenBabel::OBPlugin::End((char const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBPlugin::PluginIterator(result)), SWIGTYPE_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_GetMap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPlugin *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBPlugin::PluginMapType *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPlugin, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPlugin_GetMap" "', argument " "1"" of type '" "OpenBabel::OBPlugin const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPlugin * >(argp1);
  result = (OpenBabel::OBPlugin::PluginMapType *) &((OpenBabel::OBPlugin const *)arg1)->GetMap();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPlugin_LoadAllPlugins(PyObject *self, PyObject *args) { PyObject *resultobj = 0; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPlugin_LoadAllPlugins", 0, 0, 0)) SWIG_fail; OpenBabel::OBPlugin::LoadAllPlugins();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBPlugin_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPlugin, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_stringbuf(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::stringbuf *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "new_stringbuf", 0, 0, 0)) SWIG_fail;
  result = (std::stringbuf *)new std::stringbuf();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__stringbuf, SWIG_POINTER_NEW |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_stringbuf(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::stringbuf *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__stringbuf, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_stringbuf" "', argument " "1"" of type '" "std::stringbuf *""'");  }
   arg1 = reinterpret_cast< std::stringbuf * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *stringbuf_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__stringbuf, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *stringbuf_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBError__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; std::string *arg4 = 0 ; std::string *arg5 = 0 ;
  OpenBabel::obMessageLevel arg6 ; int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ; int res5 = SWIG_OLDOBJ ; int val6 ; int ecode6 = 0 ; OpenBabel::OBError *result = 0 ; (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  { std::string *ptr = (std::string *)0; res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  { std::string *ptr = (std::string *)0; res4 = SWIG_AsPtr_std_string(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     arg4 = ptr; }  { std::string *ptr = (std::string *)0; res5 = SWIG_AsPtr_std_string(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_OBError" "', argument " "5"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "5"" of type '" "std::string const &""'"); }
     arg5 = ptr; }  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_OBError" "', argument " "6"" of type '" "OpenBabel::obMessageLevel""'"); }
    arg6 = static_cast< OpenBabel::obMessageLevel >(val6);
  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1,(std::string const &)*arg2,(std::string const &)*arg3,(std::string const &)*arg4,(std::string const &)*arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; std::string *arg4 = 0 ; std::string *arg5 = 0 ;
  int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; int res4 = SWIG_OLDOBJ ; int res5 = SWIG_OLDOBJ ;
  OpenBabel::OBError *result = 0 ; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  { std::string *ptr = (std::string *)0; res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  { std::string *ptr = (std::string *)0; res4 = SWIG_AsPtr_std_string(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     arg4 = ptr; }  { std::string *ptr = (std::string *)0; res5 = SWIG_AsPtr_std_string(swig_obj[4], &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_OBError" "', argument " "5"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "5"" of type '" "std::string const &""'"); }
     arg5 = ptr; } 
  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1,(std::string const &)*arg2,(std::string const &)*arg3,(std::string const &)*arg4,(std::string const &)*arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; if (SWIG_IsNewObj(res5)) delete arg5; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; std::string *arg4 = 0 ; int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; int res4 = SWIG_OLDOBJ ; OpenBabel::OBError *result = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  { std::string *ptr = (std::string *)0; res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  { std::string *ptr = (std::string *)0; res4 = SWIG_AsPtr_std_string(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "4"" of type '" "std::string const &""'"); }
     arg4 = ptr; } 
  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1,(std::string const &)*arg2,(std::string const &)*arg3,(std::string const &)*arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ; OpenBabel::OBError *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; {
    std::string *ptr = (std::string *)0; res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  { std::string *ptr = (std::string *)0; res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; } 
  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1,(std::string const &)*arg2,(std::string const &)*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; std::string *arg2 = 0 ; int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ;
  OpenBabel::OBError *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  { std::string *ptr = (std::string *)0; res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; } 
  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1,(std::string const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; OpenBabel::OBError *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBError" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  result = (OpenBabel::OBError *)new OpenBabel::OBError((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBError *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBError *)new OpenBabel::OBError();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBError(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBError", 0, 6, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBError__SWIG_6(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_new_OBError__SWIG_5(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBError__SWIG_4(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_new_OBError__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBError__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 5) { PyObject *retobj = _wrap_new_OBError__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_new_OBError__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBError'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBError::OBError(std::string const &,std::string const &,std::string const &,std::string const &,std::string const &,OpenBabel::obMessageLevel const)\n"
  "    OpenBabel::OBError::OBError(std::string const &,std::string const &,std::string const &,std::string const &,std::string const &)\n"
  "    OpenBabel::OBError::OBError(std::string const &,std::string const &,std::string const &,std::string const &)\n"
  "    OpenBabel::OBError::OBError(std::string const &,std::string const &,std::string const &)\n"
  "    OpenBabel::OBError::OBError(std::string const &,std::string const &)\n"
  "    OpenBabel::OBError::OBError(std::string const &)\n" "    OpenBabel::OBError::OBError()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBError_message(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_message" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->message();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap___lshift____SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::ostream *arg1 = 0 ; OpenBabel::OBError *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  std::ostream *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__ostream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "1"" of type '" "std::ostream &""'"); }
   arg1 = reinterpret_cast< std::ostream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBError,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::OBError const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "__lshift__" "', argument " "2"" of type '" "OpenBabel::OBError const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBError * >(argp2);
  result = (std::ostream *) &operator <<(*arg1,(OpenBabel::OBError const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap___lshift__(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "__lshift__", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap___lshift____SWIG_0(self, argc, argv);}  check_1: if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap___lshift____SWIG_1(self, argc, argv);}  check_2:
  if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap___lshift____SWIG_2(self, argc, argv);}  check_3:
  if (argc == 2) { PyObject *retobj = _wrap___lshift____SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail: SWIG_Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBError_GetMethod(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetMethod" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->GetMethod();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError_GetError(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetError" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->GetError();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError_GetExplanation(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetExplanation" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->GetExplanation();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError_GetPossibleCause(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetPossibleCause" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->GetPossibleCause();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError_GetSuggestedRemedy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetSuggestedRemedy" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); result = ((OpenBabel::OBError const *)arg1)->GetSuggestedRemedy();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError_GetLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::obMessageLevel result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError_GetLevel" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1);
  result = (OpenBabel::obMessageLevel)((OpenBabel::OBError const *)arg1)->GetLevel();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBError___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; OpenBabel::OBError *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBError___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBError___eq__" "', argument " "1"" of type '" "OpenBabel::OBError const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBError,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBError___eq__" "', argument " "2"" of type '" "OpenBabel::OBError const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBError___eq__" "', argument " "2"" of type '" "OpenBabel::OBError const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBError * >(argp2);
  result = (bool)((OpenBabel::OBError const *)arg1)->operator ==((OpenBabel::OBError const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_delete_OBError(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBError *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBError" "', argument " "1"" of type '" "OpenBabel::OBError *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBError * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBError_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBError, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBError_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMessageHandler(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBMessageHandler", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBMessageHandler *)new OpenBabel::OBMessageHandler();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMessageHandler, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBMessageHandler(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMessageHandler" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMessageHandler *arg1 = 0 ; OpenBabel::OBError arg2 ; OpenBabel::errorQualifier arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ThrowError" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBError,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "OpenBabel::OBError""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "OpenBabel::OBError""'"); }
     else { OpenBabel::OBError * temp = reinterpret_cast< OpenBabel::OBError * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "OpenBabel::errorQualifier""'"); }
    arg3 = static_cast< OpenBabel::errorQualifier >(val3); (arg1)->ThrowError(SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMessageHandler *arg1 = 0 ; OpenBabel::OBError arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ThrowError" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBError,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "OpenBabel::OBError""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "OpenBabel::OBError""'"); }
     else { OpenBabel::OBError * temp = reinterpret_cast< OpenBabel::OBError * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->ThrowError(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMessageHandler *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ;
  OpenBabel::obMessageLevel arg4 ; OpenBabel::errorQualifier arg5 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ThrowError" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMessageHandler_ThrowError" "', argument " "4"" of type '" "OpenBabel::obMessageLevel""'"); }
    arg4 = static_cast< OpenBabel::obMessageLevel >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMessageHandler_ThrowError" "', argument " "5"" of type '" "OpenBabel::errorQualifier""'"); }
    arg5 = static_cast< OpenBabel::errorQualifier >(val5);
  (arg1)->ThrowError((std::string const &)*arg2,(std::string const &)*arg3,arg4,arg5); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMessageHandler *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ;
  OpenBabel::obMessageLevel arg4 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; int val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ThrowError" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMessageHandler_ThrowError" "', argument " "4"" of type '" "OpenBabel::obMessageLevel""'"); }
    arg4 = static_cast< OpenBabel::obMessageLevel >(val4);
  (arg1)->ThrowError((std::string const &)*arg2,(std::string const &)*arg3,arg4); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMessageHandler *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ThrowError" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  { std::string *ptr = (std::string *)0; res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMessageHandler_ThrowError" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  (arg1)->ThrowError((std::string const &)*arg2,(std::string const &)*arg3); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ThrowError(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMessageHandler_ThrowError", 0, 5, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBMessageHandler_ThrowError__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__OBError, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_2; { { int res = SWIG_AsVal_int(argv[2], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_OBMessageHandler_ThrowError__SWIG_0(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_OBMessageHandler_ThrowError__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBMessageHandler_ThrowError__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBMessageHandler_ThrowError__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMessageHandler_ThrowError'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBMessageHandler::ThrowError(OpenBabel::OBError,OpenBabel::errorQualifier)\n"
  "    OpenBabel::OBMessageHandler::ThrowError(OpenBabel::OBError)\n"
  "    OpenBabel::OBMessageHandler::ThrowError(std::string const &,std::string const &,OpenBabel::obMessageLevel,OpenBabel::errorQualifier)\n"
  "    OpenBabel::OBMessageHandler::ThrowError(std::string const &,std::string const &,OpenBabel::obMessageLevel)\n"
  "    OpenBabel::OBMessageHandler::ThrowError(std::string const &,std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetMessagesOfLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; OpenBabel::obMessageLevel arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; std::vector< std::string,std::allocator< std::string > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMessageHandler_GetMessagesOfLevel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetMessagesOfLevel" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMessageHandler_GetMessagesOfLevel" "', argument " "2"" of type '" "OpenBabel::obMessageLevel""'"); }
    arg2 = static_cast< OpenBabel::obMessageLevel >(val2); result = (arg1)->GetMessagesOfLevel(arg2);
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_StartLogging(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_StartLogging" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); (arg1)->StartLogging(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_StopLogging(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_StopLogging" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); (arg1)->StopLogging(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_SetMaxLogEntries(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMessageHandler_SetMaxLogEntries", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_SetMaxLogEntries" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMessageHandler_SetMaxLogEntries" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetMaxLogEntries(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetMaxLogEntries(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetMaxLogEntries" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetMaxLogEntries();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_ClearLog(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_ClearLog" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); (arg1)->ClearLog(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_SetOutputLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; OpenBabel::obMessageLevel arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMessageHandler_SetOutputLevel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_SetOutputLevel" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMessageHandler_SetOutputLevel" "', argument " "2"" of type '" "OpenBabel::obMessageLevel""'"); }
    arg2 = static_cast< OpenBabel::obMessageLevel >(val2); (arg1)->SetOutputLevel(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetOutputLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::obMessageLevel result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetOutputLevel" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1);
  result = (OpenBabel::obMessageLevel)(arg1)->GetOutputLevel(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_SetOutputStream(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMessageHandler_SetOutputStream", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_SetOutputStream" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMessageHandler_SetOutputStream" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); (arg1)->SetOutputStream(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetOutputStream(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::ostream *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetOutputStream" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (std::ostream *)(arg1)->GetOutputStream();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_StartErrorWrap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_StartErrorWrap" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (bool)(arg1)->StartErrorWrap();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_StopErrorWrap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_StopErrorWrap" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (bool)(arg1)->StopErrorWrap();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetErrorMessageCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetErrorMessageCount" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetErrorMessageCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetWarningMessageCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetWarningMessageCount" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetWarningMessageCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetInfoMessageCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetInfoMessageCount" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetInfoMessageCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetAuditMessageCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetAuditMessageCount" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetAuditMessageCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetDebugMessageCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetDebugMessageCount" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (unsigned int)(arg1)->GetDebugMessageCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMessageHandler_GetMessageSummary(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMessageHandler *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMessageHandler, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMessageHandler_GetMessageSummary" "', argument " "1"" of type '" "OpenBabel::OBMessageHandler *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp1); result = (arg1)->GetMessageSummary();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBMessageHandler_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMessageHandler, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMessageHandler_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_obErrorLog_set(PyObject *_val) { { void *argp = 0;
    int res = SWIG_ConvertPtr(_val, &argp, SWIGTYPE_p_OpenBabel__OBMessageHandler,  0  | 0); if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in variable '""OpenBabel::obErrorLog""' of type '""OpenBabel::OBMessageHandler""'"); }
     if (!argp) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""OpenBabel::obErrorLog""' of type '""OpenBabel::OBMessageHandler""'"); }
     else { OpenBabel::OBMessageHandler * temp; temp  = reinterpret_cast< OpenBabel::OBMessageHandler * >(argp);
      OpenBabel::obErrorLog = *temp; if (SWIG_IsNewObj(res)) delete temp; }  }  return 0; fail: return 1; }
SWIGINTERN PyObject *Swig_var_obErrorLog_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_NewPointerObj(SWIG_as_voidptr(&OpenBabel::obErrorLog), SWIGTYPE_p_OpenBabel__OBMessageHandler,  0 );
  return pyobj; }
SWIGINTERN PyObject *_wrap_delete_obLogBuf(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::obLogBuf *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__obLogBuf, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_obLogBuf" "', argument " "1"" of type '" "OpenBabel::obLogBuf *""'");  }
   arg1 = reinterpret_cast< OpenBabel::obLogBuf * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_new_obLogBuf(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::obLogBuf *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_obLogBuf", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::obLogBuf *)new OpenBabel::obLogBuf();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__obLogBuf, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *obLogBuf_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__obLogBuf, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *obLogBuf_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFormat_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat **result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFormat_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFormat **) &OpenBabel::OBFormat::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_FindType" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); result = (OpenBabel::OBFormat *)OpenBabel::OBFormat::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_ReadMolecule(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; OpenBabel::OBConversion *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFormat_ReadMolecule", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_ReadMolecule" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_ReadMolecule" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_ReadMolecule" "', argument " "3"" of type '" "OpenBabel::OBConversion *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBConversion * >(argp3); result = (bool)(arg1)->ReadMolecule(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_ReadChemObject(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; OpenBabel::OBConversion *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFormat_ReadChemObject", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_ReadChemObject" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_ReadChemObject" "', argument " "2"" of type '" "OpenBabel::OBConversion *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBConversion * >(argp2); result = (bool)(arg1)->ReadChemObject(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_WriteMolecule(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; OpenBabel::OBConversion *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFormat_WriteMolecule", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_WriteMolecule" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_WriteMolecule" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_WriteMolecule" "', argument " "3"" of type '" "OpenBabel::OBConversion *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBConversion * >(argp3); result = (bool)(arg1)->WriteMolecule(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_WriteChemObject(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; OpenBabel::OBConversion *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFormat_WriteChemObject", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_WriteChemObject" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_WriteChemObject" "', argument " "2"" of type '" "OpenBabel::OBConversion *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBConversion * >(argp2); result = (bool)(arg1)->WriteChemObject(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_TargetClassDescription(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_TargetClassDescription" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (char *)(arg1)->TargetClassDescription();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_GetType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::type_info *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_GetType" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (std::type_info *) &(arg1)->GetType();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__type_info, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_SpecificationURL(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_SpecificationURL" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (char *)(arg1)->SpecificationURL();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_GetMIMEType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_GetMIMEType" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (char *)(arg1)->GetMIMEType();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_Flags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_Flags" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (unsigned int)(arg1)->Flags();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_SkipObjects(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; int arg2 ; OpenBabel::OBConversion *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFormat_SkipObjects", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_SkipObjects" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFormat_SkipObjects" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_SkipObjects" "', argument " "3"" of type '" "OpenBabel::OBConversion *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBConversion * >(argp3); result = (int)(arg1)->SkipObjects(arg2,arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_MakeNewInstance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_MakeNewInstance" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); result = (OpenBabel::OBFormat *)(arg1)->MakeNewInstance();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_RegisterFormat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFormat *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_RegisterFormat" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_RegisterFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_RegisterFormat" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (int)(arg1)->RegisterFormat((char const *)arg2,(char const *)arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_RegisterFormat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFormat *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_RegisterFormat" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_RegisterFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (int)(arg1)->RegisterFormat((char const *)arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_RegisterFormat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBFormat_RegisterFormat", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBFormat_RegisterFormat__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBFormat_RegisterFormat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBFormat_RegisterFormat'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBFormat::RegisterFormat(char const *,char const *)\n"
  "    OpenBabel::OBFormat::RegisterFormat(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBFormat_Display__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFormat *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; char *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; int res4 ;
  char *buf4 = 0 ; int alloc4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_Display" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFormat_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBFormat_Display" "', argument " "4"" of type '" "char const *""'"); }
   arg4 = reinterpret_cast< char * >(buf4); result = (bool)(arg1)->Display(*arg2,(char const *)arg3,(char const *)arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return resultobj; fail: if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_Display__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFormat *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_Display" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFormat_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFormat_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFormat_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->Display(*arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBFormat_Display(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBFormat_Display", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBFormat_Display__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBFormat_Display__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBFormat_Display'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBFormat::Display(std::string &,char const *,char const *)\n"
  "    OpenBabel::OBFormat::Display(std::string &,char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBFormat_FormatFromMIME(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFormat_FormatFromMIME" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFormat *)OpenBabel::OBFormat::FormatFromMIME((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFormat, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFormat" "', argument " "1"" of type '" "OpenBabel::OBFormat *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFormat * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBFormat_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFormat, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::istream *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::OBConversion *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBConversion" "', argument " "1"" of type '" "std::istream *""'");  }
   arg1 = reinterpret_cast< std::istream * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBConversion" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::istream *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBConversion *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBConversion" "', argument " "1"" of type '" "std::istream *""'");  }
   arg1 = reinterpret_cast< std::istream * >(argp1); result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; std::string arg2 ; OpenBabel::OBConversion *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBConversion" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBConversion" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion(SWIG_STD_MOVE(arg1),SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBConversion *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_OBConversion" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBConversion *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBConversion,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBConversion" "', argument " "1"" of type '" "OpenBabel::OBConversion const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBConversion" "', argument " "1"" of type '" "OpenBabel::OBConversion const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (OpenBabel::OBConversion *)new OpenBabel::OBConversion((OpenBabel::OBConversion const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBConversion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBConversion", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBConversion__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__istream, 0); _v = SWIG_CheckState(res);}  if (!_v) goto check_2;
    return _wrap_new_OBConversion__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap_new_OBConversion__SWIG_5(self, argc, argv);}  check_3:
  if (argc == 1) { PyObject *retobj = _wrap_new_OBConversion__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__istream, 0); _v = SWIG_CheckState(res);}  if (!_v) goto check_5;
    { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__ostream, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_5; return _wrap_new_OBConversion__SWIG_0(self, argc, argv);}  check_5: if (argc == 2) {
    PyObject *retobj = _wrap_new_OBConversion__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBConversion'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::OBConversion(std::istream *,std::ostream *)\n"
  "    OpenBabel::OBConversion::OBConversion(std::istream *)\n" "    OpenBabel::OBConversion::OBConversion()\n"
  "    OpenBabel::OBConversion::OBConversion(std::string,std::string)\n"
  "    OpenBabel::OBConversion::OBConversion(std::string)\n"
  "    OpenBabel::OBConversion::OBConversion(OpenBabel::OBConversion const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBConversion(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConversion" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterFormat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; char *arg3 = 0 ; int res1 ; char *buf1 = 0 ;
  int alloc1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_RegisterFormat" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RegisterFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_RegisterFormat" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3);
  result = (int)OpenBabel::OBConversion::RegisterFormat((char const *)arg1,arg2,(char const *)arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterFormat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_RegisterFormat" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RegisterFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2);
  result = (int)OpenBabel::OBConversion::RegisterFormat((char const *)arg1,arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterFormat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_RegisterFormat", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBConversion_RegisterFormat__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_RegisterFormat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_RegisterFormat'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBConversion::RegisterFormat(char const *,OpenBabel::OBFormat *,char const *)\n"
  "    OpenBabel::OBConversion::RegisterFormat(char const *,OpenBabel::OBFormat *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_FindFormat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_FindFormat" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FindFormat((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FindFormat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBFormat *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_FindFormat" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FindFormat(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FindFormat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_FindFormat", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBConversion_FindFormat__SWIG_1(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_OBConversion_FindFormat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_FindFormat'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::FindFormat(char const *)\n"
  "    OpenBabel::OBConversion::FindFormat(std::string const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromExt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_FormatFromExt" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FormatFromExt((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromExt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; bool *arg2 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::OBFormat *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_FormatFromExt" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_FormatFromExt" "', argument " "2"" of type '" "bool &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConversion_FormatFromExt" "', argument " "2"" of type '" "bool &""'"); }
   arg2 = reinterpret_cast< bool * >(argp2);
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FormatFromExt((char const *)arg1,*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromExt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBFormat *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_FormatFromExt" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FormatFromExt(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromExt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; bool *arg2 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  OpenBabel::OBFormat *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_FormatFromExt" "', argument " "1"" of type '" "std::string const""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_bool,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_FormatFromExt" "', argument " "2"" of type '" "bool &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConversion_FormatFromExt" "', argument " "2"" of type '" "bool &""'"); }
   arg2 = reinterpret_cast< bool * >(argp2);
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FormatFromExt(SWIG_STD_MOVE(arg1),*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromExt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_FormatFromExt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBConversion_FormatFromExt__SWIG_2(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_OBConversion_FormatFromExt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_3;
    PyObject *retobj = _wrap_OBConversion_FormatFromExt__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_FormatFromExt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_FormatFromExt'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::FormatFromExt(char const *)\n"
  "    OpenBabel::OBConversion::FormatFromExt(char const *,bool &)\n"
  "    OpenBabel::OBConversion::FormatFromExt(std::string const)\n"
  "    OpenBabel::OBConversion::FormatFromExt(std::string const,bool &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_FormatFromMIME(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_FormatFromMIME" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::FormatFromMIME((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Description(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  char *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_Description", 0, 0, 0)) SWIG_fail;
  result = (char *)OpenBabel::OBConversion::Description(); resultobj = SWIG_FromCharPtr((const char *)result); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInStream(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::istream *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInStream" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (std::istream *)((OpenBabel::OBConversion const *)arg1)->GetInStream();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__istream, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOutStream(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::ostream *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOutStream" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (std::ostream *)((OpenBabel::OBConversion const *)arg1)->GetOutStream();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__ostream, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInStream__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; std::istream *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInStream" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInStream" "', argument " "2"" of type '" "std::istream *""'");  }
   arg2 = reinterpret_cast< std::istream * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetInStream" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); (arg1)->SetInStream(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInStream__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; std::istream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInStream" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInStream" "', argument " "2"" of type '" "std::istream *""'");  }
   arg2 = reinterpret_cast< std::istream * >(argp2); (arg1)->SetInStream(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInStream(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetInStream", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_SetInStream__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_SetInStream__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetInStream'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetInStream(std::istream *,bool)\n"
  "    OpenBabel::OBConversion::SetInStream(std::istream *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutStream__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; std::ostream *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutStream" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutStream" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetOutStream" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); (arg1)->SetOutStream(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutStream__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutStream" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutStream" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); (arg1)->SetOutStream(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutStream(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetOutStream", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_SetOutStream__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_SetOutStream__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetOutStream'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetOutStream(std::ostream *,bool)\n"
  "    OpenBabel::OBConversion::SetOutStream(std::ostream *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; bool arg4 ; bool arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ;
  bool val4 ; int ecode4 = 0 ; bool val5 ; int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "5"" of type '" "bool""'"); }
    arg5 = static_cast< bool >(val5);
  result = (bool)(arg1)->SetInAndOutFormats((char const *)arg2,(char const *)arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; bool arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool val4 ;
  int ecode4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (bool)(arg1)->SetInAndOutFormats((char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->SetInAndOutFormats((char const *)arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; OpenBabel::OBFormat *arg3 = 0 ;
  bool arg4 ; bool arg5 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  bool val4 ; int ecode4 = 0 ; bool val5 ; int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "OpenBabel::OBFormat *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBFormat * >(argp3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "5"" of type '" "bool""'"); }
    arg5 = static_cast< bool >(val5); result = (bool)(arg1)->SetInAndOutFormats(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; OpenBabel::OBFormat *arg3 = 0 ;
  bool arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool val4 ;
  int ecode4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "OpenBabel::OBFormat *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBFormat * >(argp3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (bool)(arg1)->SetInAndOutFormats(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; OpenBabel::OBFormat *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_SetInAndOutFormats" "', argument " "3"" of type '" "OpenBabel::OBFormat *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBFormat * >(argp3); result = (bool)(arg1)->SetInAndOutFormats(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInAndOutFormats(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetInAndOutFormats", 0, 5, argv))) SWIG_fail; --argc;
  if (argc == 3) { int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBConversion_SetInAndOutFormats__SWIG_5(self, argc, argv);}  check_1: if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_SetInAndOutFormats__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_3; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3;
    PyObject *retobj = _wrap_OBConversion_SetInAndOutFormats__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 4) {
    PyObject *retobj = _wrap_OBConversion_SetInAndOutFormats__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_5; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_5;
    PyObject *retobj = _wrap_OBConversion_SetInAndOutFormats__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_5: if (argc == 5) {
    PyObject *retobj = _wrap_OBConversion_SetInAndOutFormats__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetInAndOutFormats'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetInAndOutFormats(char const *,char const *,bool,bool)\n"
  "    OpenBabel::OBConversion::SetInAndOutFormats(char const *,char const *,bool)\n"
  "    OpenBabel::OBConversion::SetInAndOutFormats(char const *,char const *)\n"
  "    OpenBabel::OBConversion::SetInAndOutFormats(OpenBabel::OBFormat *,OpenBabel::OBFormat *,bool,bool)\n"
  "    OpenBabel::OBConversion::SetInAndOutFormats(OpenBabel::OBFormat *,OpenBabel::OBFormat *,bool)\n"
  "    OpenBabel::OBConversion::SetInAndOutFormats(OpenBabel::OBFormat *,OpenBabel::OBFormat *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInFormat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetInFormat" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->SetInFormat((char const *)arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInFormat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->SetInFormat((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInFormat__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetInFormat" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->SetInFormat(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInFormat__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetInFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetInFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); result = (bool)(arg1)->SetInFormat(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetInFormat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetInFormat", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBConversion_SetInFormat__SWIG_3(self, argc, argv);} 
  check_1: if (argc == 2) { PyObject *retobj = _wrap_OBConversion_SetInFormat__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_3; PyObject *retobj = _wrap_OBConversion_SetInFormat__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_SetInFormat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetInFormat'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetInFormat(char const *,bool)\n"
  "    OpenBabel::OBConversion::SetInFormat(char const *)\n"
  "    OpenBabel::OBConversion::SetInFormat(OpenBabel::OBFormat *,bool)\n"
  "    OpenBabel::OBConversion::SetInFormat(OpenBabel::OBFormat *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutFormat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetOutFormat" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->SetOutFormat((char const *)arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutFormat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutFormat" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->SetOutFormat((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutFormat__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetOutFormat" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->SetOutFormat(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutFormat__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBFormat *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOutFormat" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); result = (bool)(arg1)->SetOutFormat(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutFormat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetOutFormat", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBConversion_SetOutFormat__SWIG_3(self, argc, argv);} 
  check_1: if (argc == 2) { PyObject *retobj = _wrap_OBConversion_SetOutFormat__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBFormat, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_3; PyObject *retobj = _wrap_OBConversion_SetOutFormat__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_SetOutFormat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetOutFormat'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetOutFormat(char const *,bool)\n"
  "    OpenBabel::OBConversion::SetOutFormat(char const *)\n"
  "    OpenBabel::OBConversion::SetOutFormat(OpenBabel::OBFormat *,bool)\n"
  "    OpenBabel::OBConversion::SetOutFormat(OpenBabel::OBFormat *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (OpenBabel::OBFormat *)((OpenBabel::OBConversion const *)arg1)->GetInFormat();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOutFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFormat *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOutFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (OpenBabel::OBFormat *)((OpenBabel::OBConversion const *)arg1)->GetOutFormat();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInGzipped(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInGzipped" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (bool)((OpenBabel::OBConversion const *)arg1)->GetInGzipped();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOutGzipped(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOutGzipped" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (bool)((OpenBabel::OBConversion const *)arg1)->GetOutGzipped();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInFilename(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInFilename" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = ((OpenBabel::OBConversion const *)arg1)->GetInFilename();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOutFilename(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOutFilename" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = ((OpenBabel::OBConversion const *)arg1)->GetOutFilename();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInPos(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::streampos result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInPos" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = ((OpenBabel::OBConversion const *)arg1)->GetInPos();
  resultobj = SWIG_NewPointerObj((new std::streampos(result)), SWIGTYPE_p_std__streampos, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetInLen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetInLen" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = ((OpenBabel::OBConversion const *)arg1)->GetInLen();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetTitle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (char *)((OpenBabel::OBConversion const *)arg1)->GetTitle(); resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetAuxConv(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBConversion *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetAuxConv" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (OpenBabel::OBConversion *)((OpenBabel::OBConversion const *)arg1)->GetAuxConv();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetAuxConv(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBConversion *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_SetAuxConv", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetAuxConv" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetAuxConv" "', argument " "2"" of type '" "OpenBabel::OBConversion *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBConversion * >(argp2); (arg1)->SetAuxConv(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsOption__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; OpenBabel::OBConversion::Option_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int val3 ; int ecode3 = 0 ; char *result = 0 ;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_IsOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_IsOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_IsOption" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3);
  result = (char *)(arg1)->IsOption((char const *)arg2,arg3); resultobj = SWIG_FromCharPtr((const char *)result);
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsOption__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; char *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_IsOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_IsOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (char *)(arg1)->IsOption((char const *)arg2);
  resultobj = SWIG_FromCharPtr((const char *)result); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsOption(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_IsOption", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_IsOption__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_IsOption__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_IsOption'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBConversion::IsOption(char const *,OpenBabel::OBConversion::Option_type)\n"
  "    OpenBabel::OBConversion::IsOption(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOptions(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBConversion::Option_type arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ;
  std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_GetOptions", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOptions" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_GetOptions" "', argument " "2"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg2 = static_cast< OpenBabel::OBConversion::Option_type >(val2);
  result = (std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > *)(arg1)->GetOptions(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_AddOption__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; OpenBabel::OBConversion::Option_type arg3 ;
  char *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int val3 ; int ecode3 = 0 ;
  int res4 ; char *buf4 = 0 ; int alloc4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_AddOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_AddOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_AddOption" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBConversion_AddOption" "', argument " "4"" of type '" "char const *""'"); }
   arg4 = reinterpret_cast< char * >(buf4); (arg1)->AddOption((char const *)arg2,arg3,(char const *)arg4);
  resultobj = SWIG_Py_Void(); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_AddOption__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; OpenBabel::OBConversion::Option_type arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_AddOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_AddOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_AddOption" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3); (arg1)->AddOption((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void(); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_AddOption__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_AddOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_AddOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->AddOption((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_AddOption(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_AddOption", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_AddOption__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_AddOption__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBConversion_AddOption__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_AddOption'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBConversion::AddOption(char const *,OpenBabel::OBConversion::Option_type,char const *)\n"
  "    OpenBabel::OBConversion::AddOption(char const *,OpenBabel::OBConversion::Option_type)\n"
  "    OpenBabel::OBConversion::AddOption(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_RemoveOption(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; OpenBabel::OBConversion::Option_type arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_RemoveOption", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_RemoveOption" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RemoveOption" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_RemoveOption" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3);
  result = (bool)(arg1)->RemoveOption((char const *)arg2,arg3); resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOptions(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; char *arg2 = 0 ; OpenBabel::OBConversion::Option_type arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_SetOptions", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOptions" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_SetOptions" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_SetOptions" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3); (arg1)->SetOptions((char const *)arg2,arg3);
  resultobj = SWIG_Py_Void(); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterOptionParam__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBFormat *arg2 = 0 ; int arg3 ;
  OpenBabel::OBConversion::Option_type arg4 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_RegisterOptionParam" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RegisterOptionParam" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_RegisterOptionParam" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConversion_RegisterOptionParam" "', argument " "4"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg4 = static_cast< OpenBabel::OBConversion::Option_type >(val4);
  OpenBabel::OBConversion::RegisterOptionParam(SWIG_STD_MOVE(arg1),arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterOptionParam__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBFormat *arg2 = 0 ; int arg3 ; void *argp2 = 0 ; int res2 = 0 ;
  int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_RegisterOptionParam" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RegisterOptionParam" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_RegisterOptionParam" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); OpenBabel::OBConversion::RegisterOptionParam(SWIG_STD_MOVE(arg1),arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterOptionParam__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string arg1 ; OpenBabel::OBFormat *arg2 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_RegisterOptionParam" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_RegisterOptionParam" "', argument " "2"" of type '" "OpenBabel::OBFormat *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBFormat * >(argp2);
  OpenBabel::OBConversion::RegisterOptionParam(SWIG_STD_MOVE(arg1),arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_RegisterOptionParam(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_RegisterOptionParam", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBConversion_RegisterOptionParam__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_RegisterOptionParam__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBConversion_RegisterOptionParam__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_RegisterOptionParam'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBConversion::RegisterOptionParam(std::string,OpenBabel::OBFormat *,int,OpenBabel::OBConversion::Option_type)\n"
  "    OpenBabel::OBConversion::RegisterOptionParam(std::string,OpenBabel::OBFormat *,int)\n"
  "    OpenBabel::OBConversion::RegisterOptionParam(std::string,OpenBabel::OBFormat *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOptionParams(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::string arg1 ; OpenBabel::OBConversion::Option_type arg2 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ; int result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_GetOptionParams", 2, 2, swig_obj)) SWIG_fail; {
    std::string *ptr = (std::string *)0; int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_GetOptionParams" "', argument " "1"" of type '" "std::string""'");  }
     arg1 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_GetOptionParams" "', argument " "2"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg2 = static_cast< OpenBabel::OBConversion::Option_type >(val2);
  result = (int)OpenBabel::OBConversion::GetOptionParams(SWIG_STD_MOVE(arg1),arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_CopyOptions__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBConversion *arg2 = 0 ;
  OpenBabel::OBConversion::Option_type arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ;
  int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_CopyOptions" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_CopyOptions" "', argument " "2"" of type '" "OpenBabel::OBConversion *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBConversion * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_CopyOptions" "', argument " "3"" of type '" "OpenBabel::OBConversion::Option_type""'"); }
    arg3 = static_cast< OpenBabel::OBConversion::Option_type >(val3); (arg1)->CopyOptions(arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_CopyOptions__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBConversion *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_CopyOptions" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_CopyOptions" "', argument " "2"" of type '" "OpenBabel::OBConversion *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBConversion * >(argp2); (arg1)->CopyOptions(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_CopyOptions(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_CopyOptions", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_CopyOptions__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_CopyOptions__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_CopyOptions'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBConversion::CopyOptions(OpenBabel::OBConversion *,OpenBabel::OBConversion::Option_type)\n"
  "    OpenBabel::OBConversion::CopyOptions(OpenBabel::OBConversion *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_GetSupportedInputFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::string,std::allocator< std::string > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetSupportedInputFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (arg1)->GetSupportedInputFormat();
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetSupportedOutputFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::string,std::allocator< std::string > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetSupportedOutputFormat" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (arg1)->GetSupportedOutputFormat();
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Convert__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; std::istream *arg2 = 0 ; std::ostream *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Convert" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_Convert" "', argument " "2"" of type '" "std::istream *""'");  }
   arg2 = reinterpret_cast< std::istream * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_Convert" "', argument " "3"" of type '" "std::ostream *""'");  }
   arg3 = reinterpret_cast< std::ostream * >(argp3); result = (int)(arg1)->Convert(arg2,arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Convert__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Convert" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (int)(arg1)->Convert();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Convert(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_Convert", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBConversion_Convert__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_Convert__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_Convert'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::Convert(std::istream *,std::ostream *)\n"
  "    OpenBabel::OBConversion::Convert()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_FullConvert(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; std::vector< std::string,std::allocator< std::string > > *arg2 = 0 ;
  std::string *arg3 = 0 ; std::vector< std::string,std::allocator< std::string > > *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ;
  int result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_FullConvert", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_FullConvert" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__string_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_FullConvert" "', argument " "2"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConversion_FullConvert" "', argument " "2"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'"); }
   arg2 = reinterpret_cast< std::vector< std::string,std::allocator< std::string > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_FullConvert" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConversion_FullConvert" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_std__vectorT_std__string_t,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBConversion_FullConvert" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConversion_FullConvert" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > &""'"); }
   arg4 = reinterpret_cast< std::vector< std::string,std::allocator< std::string > > * >(argp4);
  result = (int)(arg1)->FullConvert(*arg2,*arg3,*arg4); resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_AddChemObject(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_AddChemObject", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_AddChemObject" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_AddChemObject" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (int)(arg1)->AddChemObject(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetChemObject(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBase *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetChemObject" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (OpenBabel::OBBase *)(arg1)->GetChemObject();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsLast(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_IsLast" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (bool)(arg1)->IsLast();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsFirstInput(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_IsFirstInput" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (bool)(arg1)->IsFirstInput();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetFirstInput__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetFirstInput" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_SetFirstInput" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetFirstInput(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetFirstInput__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetFirstInput" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); (arg1)->SetFirstInput(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetFirstInput(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetFirstInput", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBConversion_SetFirstInput__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_SetFirstInput__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetFirstInput'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetFirstInput(bool)\n"
  "    OpenBabel::OBConversion::SetFirstInput()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_GetOutputIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetOutputIndex" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (int)((OpenBabel::OBConversion const *)arg1)->GetOutputIndex();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOutputIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_SetOutputIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOutputIndex" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_SetOutputIndex" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetOutputIndex(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetMoreFilesToCome(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetMoreFilesToCome" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); (arg1)->SetMoreFilesToCome(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOneObjectOnly__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOneObjectOnly" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_SetOneObjectOnly" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetOneObjectOnly(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOneObjectOnly__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetOneObjectOnly" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); (arg1)->SetOneObjectOnly(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_SetOneObjectOnly(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_SetOneObjectOnly", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBConversion_SetOneObjectOnly__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_SetOneObjectOnly__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_SetOneObjectOnly'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::SetOneObjectOnly(bool)\n"
  "    OpenBabel::OBConversion::SetOneObjectOnly()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_SetLast(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBConversion_SetLast", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_SetLast" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_SetLast" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetLast(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_IsLastFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_IsLastFile" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (bool)(arg1)->IsLastFile();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_GetCount" "', argument " "1"" of type '" "OpenBabel::OBConversion const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  result = (int)((OpenBabel::OBConversion const *)arg1)->GetCount(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_GetDefaultFormat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFormat *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_GetDefaultFormat", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFormat *)OpenBabel::OBConversion::GetDefaultFormat();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Write__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::ostream *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Write" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_Write" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_Write" "', argument " "3"" of type '" "std::ostream *""'");  }
   arg3 = reinterpret_cast< std::ostream * >(argp3); result = (bool)(arg1)->Write(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Write__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Write" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_Write" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (bool)(arg1)->Write(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Write(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_Write", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_Write__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_Write__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_Write'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::Write(OpenBabel::OBBase *,std::ostream *)\n"
  "    OpenBabel::OBConversion::Write(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_WriteString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; std::string result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_WriteString" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_WriteString" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConversion_WriteString" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (arg1)->WriteString(arg2,arg3);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_WriteString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; std::string result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_WriteString" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_WriteString" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (arg1)->WriteString(arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_WriteString(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_WriteString", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_WriteString__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_WriteString__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_WriteString'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::WriteString(OpenBabel::OBBase *,bool)\n"
  "    OpenBabel::OBConversion::WriteString(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_WriteFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_WriteFile", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_WriteFile" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_WriteFile" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_WriteFile" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (bool)(arg1)->WriteFile(arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_CloseOutFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_CloseOutFile" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); (arg1)->CloseOutFile(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Read__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::istream *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Read" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_Read" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_Read" "', argument " "3"" of type '" "std::istream *""'");  }
   arg3 = reinterpret_cast< std::istream * >(argp3); result = (bool)(arg1)->Read(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Read__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_Read" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_Read" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (bool)(arg1)->Read(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_Read(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_Read", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConversion_Read__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_Read__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_Read'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::Read(OpenBabel::OBBase *,std::istream *)\n"
  "    OpenBabel::OBConversion::Read(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_ReadString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_ReadString", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_ReadString" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_ReadString" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_ReadString" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (bool)(arg1)->ReadString(arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_ReadFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_ReadFile", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_ReadFile" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConversion_ReadFile" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_ReadFile" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (bool)(arg1)->ReadFile(arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_OpenInAndOutFiles(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; std::string arg2 ; std::string arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConversion_OpenInAndOutFiles", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_OpenInAndOutFiles" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_OpenInAndOutFiles" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBConversion_OpenInAndOutFiles" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (bool)(arg1)->OpenInAndOutFiles(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_ReportNumberConverted__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; int arg2 ; OpenBabel::OBFormat *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_ReportNumberConverted" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_ReportNumberConverted" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBFormat, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConversion_ReportNumberConverted" "', argument " "3"" of type '" "OpenBabel::OBFormat *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBFormat * >(argp3); (arg1)->ReportNumberConverted(arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_ReportNumberConverted__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConversion *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_ReportNumberConverted" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConversion_ReportNumberConverted" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->ReportNumberConverted(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConversion_ReportNumberConverted(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConversion_ReportNumberConverted", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBConversion_ReportNumberConverted__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConversion_ReportNumberConverted__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConversion_ReportNumberConverted'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConversion::ReportNumberConverted(int,OpenBabel::OBFormat *)\n"
  "    OpenBabel::OBConversion::ReportNumberConverted(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConversion_NumInputObjects(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConversion *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConversion_NumInputObjects" "', argument " "1"" of type '" "OpenBabel::OBConversion *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConversion * >(argp1); result = (int)(arg1)->NumInputObjects();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBConversion_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConversion, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBConversion_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBBondGetSmallestRingSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBondGetSmallestRingSize", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBondGetSmallestRingSize" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBondGetSmallestRingSize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (unsigned int)OpenBabel::OBBondGetSmallestRingSize(arg1,arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetTypicalValence(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int arg2 ; int arg3 ; unsigned int val1 ; int ecode1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; int val3 ;
  int ecode3 = 0 ; PyObject *swig_obj[3] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetTypicalValence", 3, 3, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetTypicalValence" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GetTypicalValence" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "GetTypicalValence" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (unsigned int)OpenBabel::GetTypicalValence(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomAssignTypicalImplicitHydrogens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAssignTypicalImplicitHydrogens" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); OpenBabel::OBAtomAssignTypicalImplicitHydrogens(arg1);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetSymbol(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetSymbol" "', argument " "1"" of type '" "unsigned int""'"); }  
  arg1 = static_cast< unsigned int >(val1); result = (char *)OpenBabel::OBElements::GetSymbol(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetName(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetName" "', argument " "1"" of type '" "unsigned int""'"); }  
  arg1 = static_cast< unsigned int >(val1); result = (char *)OpenBabel::OBElements::GetName(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetMass(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetMass" "', argument " "1"" of type '" "unsigned int""'"); }  
  arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetMass(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetExactMass__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int arg2 ; unsigned int val1 ; int ecode1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetExactMass" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GetExactMass" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (double)OpenBabel::OBElements::GetExactMass(arg1,arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetExactMass__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int val1 ; int ecode1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetExactMass" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetExactMass(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetExactMass(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "GetExactMass", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_GetExactMass__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_GetExactMass__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'GetExactMass'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBElements::GetExactMass(unsigned int,unsigned int)\n"
  "    OpenBabel::OBElements::GetExactMass(unsigned int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_GetAtomicNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ;
  char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GetAtomicNum" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); result = (unsigned int)OpenBabel::OBElements::GetAtomicNum((char const *)arg1);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_GetAllredRochowElectroNeg(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  unsigned int arg1 ; unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetAllredRochowElectroNeg" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetAllredRochowElectroNeg(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetCovalentRad(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetCovalentRad" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetCovalentRad(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetVdwRad(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetVdwRad" "', argument " "1"" of type '" "unsigned int""'"); }  
  arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetVdwRad(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetElectronAffinity(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetElectronAffinity" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetElectronAffinity(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetIonization(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetIonization" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetIonization(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetMaxBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetMaxBonds" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (unsigned int)OpenBabel::OBElements::GetMaxBonds(arg1);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetElectroNeg(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  unsigned int val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetElectroNeg" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (double)OpenBabel::OBElements::GetElectroNeg(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetRGB(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned int arg1 ;
  double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; unsigned int val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "GetRGB", 4, 4, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GetRGB" "', argument " "1"" of type '" "unsigned int""'"); }  
  arg1 = static_cast< unsigned int >(val1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GetRGB" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "GetRGB" "', argument " "3"" of type '" "double *""'");  } 
  arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "GetRGB" "', argument " "4"" of type '" "double *""'");  } 
  arg4 = reinterpret_cast< double * >(argp4); OpenBabel::OBElements::GetRGB(arg1,arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN int Swig_var_Dummy_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Dummy is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Dummy_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Dummy)); return pyobj; }
SWIGINTERN int Swig_var_Hydrogen_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hydrogen is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hydrogen_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hydrogen)); return pyobj; }
SWIGINTERN int Swig_var_H_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable H is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_H_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::H)); return pyobj; }
SWIGINTERN int Swig_var_Helium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Helium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Helium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Helium)); return pyobj; }
SWIGINTERN int Swig_var_He_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable He is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_He_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::He)); return pyobj; }
SWIGINTERN int Swig_var_Lithium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lithium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lithium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lithium)); return pyobj; }
SWIGINTERN int Swig_var_Li_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Li is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Li_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Li)); return pyobj; }
SWIGINTERN int Swig_var_Beryllium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Beryllium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Beryllium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Beryllium)); return pyobj; }
SWIGINTERN int Swig_var_Be_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Be is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Be_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Be)); return pyobj; }
SWIGINTERN int Swig_var_Boron_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Boron is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Boron_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Boron)); return pyobj; }
SWIGINTERN int Swig_var_B_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable B is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_B_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::B)); return pyobj; }
SWIGINTERN int Swig_var_Carbon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Carbon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Carbon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Carbon)); return pyobj; }
SWIGINTERN int Swig_var_C_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable C is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_C_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::C)); return pyobj; }
SWIGINTERN int Swig_var_Nitrogen_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nitrogen is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nitrogen_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nitrogen)); return pyobj; }
SWIGINTERN int Swig_var_N_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable N is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_N_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::N)); return pyobj; }
SWIGINTERN int Swig_var_Oxygen_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Oxygen is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Oxygen_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Oxygen)); return pyobj; }
SWIGINTERN int Swig_var_O_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable O is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_O_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::O)); return pyobj; }
SWIGINTERN int Swig_var_Fluorine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fluorine is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fluorine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fluorine)); return pyobj; }
SWIGINTERN int Swig_var_F_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable F is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_F_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::F)); return pyobj; }
SWIGINTERN int Swig_var_Neon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Neon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Neon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Neon)); return pyobj; }
SWIGINTERN int Swig_var_Ne_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ne is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ne_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ne)); return pyobj; }
SWIGINTERN int Swig_var_Sodium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sodium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sodium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sodium)); return pyobj; }
SWIGINTERN int Swig_var_Na_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Na is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Na_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Na)); return pyobj; }
SWIGINTERN int Swig_var_Magnesium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Magnesium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Magnesium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Magnesium)); return pyobj; }
SWIGINTERN int Swig_var_Mg_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mg is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mg_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mg)); return pyobj; }
SWIGINTERN int Swig_var_Aluminium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Aluminium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Aluminium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Aluminium)); return pyobj; }
SWIGINTERN int Swig_var_Al_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Al is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Al_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Al)); return pyobj; }
SWIGINTERN int Swig_var_Silicon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Silicon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Silicon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Silicon)); return pyobj; }
SWIGINTERN int Swig_var_Si_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Si is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Si_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Si)); return pyobj; }
SWIGINTERN int Swig_var_Phosphorus_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Phosphorus is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Phosphorus_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Phosphorus)); return pyobj; }
SWIGINTERN int Swig_var_P_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable P is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_P_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::P)); return pyobj; }
SWIGINTERN int Swig_var_Sulfur_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sulfur is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sulfur_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sulfur)); return pyobj; }
SWIGINTERN int Swig_var_S_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable S is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_S_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::S)); return pyobj; }
SWIGINTERN int Swig_var_Chlorine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Chlorine is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Chlorine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Chlorine)); return pyobj; }
SWIGINTERN int Swig_var_Cl_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cl is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cl_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cl)); return pyobj; }
SWIGINTERN int Swig_var_Argon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Argon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Argon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Argon)); return pyobj; }
SWIGINTERN int Swig_var_Ar_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ar is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ar_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ar)); return pyobj; }
SWIGINTERN int Swig_var_Potassium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Potassium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Potassium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Potassium)); return pyobj; }
SWIGINTERN int Swig_var_K_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable K is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_K_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::K)); return pyobj; }
SWIGINTERN int Swig_var_Calcium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Calcium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Calcium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Calcium)); return pyobj; }
SWIGINTERN int Swig_var_Ca_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ca is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ca_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ca)); return pyobj; }
SWIGINTERN int Swig_var_Scandium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Scandium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Scandium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Scandium)); return pyobj; }
SWIGINTERN int Swig_var_Sc_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sc is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sc_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sc)); return pyobj; }
SWIGINTERN int Swig_var_Titanium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Titanium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Titanium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Titanium)); return pyobj; }
SWIGINTERN int Swig_var_Ti_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ti is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ti_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ti)); return pyobj; }
SWIGINTERN int Swig_var_Vanadium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Vanadium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Vanadium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Vanadium)); return pyobj; }
SWIGINTERN int Swig_var_V_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable V is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_V_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::V)); return pyobj; }
SWIGINTERN int Swig_var_Chromium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Chromium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Chromium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Chromium)); return pyobj; }
SWIGINTERN int Swig_var_Cr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cr)); return pyobj; }
SWIGINTERN int Swig_var_Manganese_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Manganese is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Manganese_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Manganese)); return pyobj; }
SWIGINTERN int Swig_var_Mn_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mn is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mn_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mn)); return pyobj; }
SWIGINTERN int Swig_var_Iron_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Iron is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Iron_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Iron)); return pyobj; }
SWIGINTERN int Swig_var_Fe_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fe is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fe_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fe)); return pyobj; }
SWIGINTERN int Swig_var_Cobalt_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cobalt is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cobalt_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cobalt)); return pyobj; }
SWIGINTERN int Swig_var_Co_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Co is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Co_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Co)); return pyobj; }
SWIGINTERN int Swig_var_Nickel_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nickel is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nickel_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nickel)); return pyobj; }
SWIGINTERN int Swig_var_Ni_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ni is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ni_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ni)); return pyobj; }
SWIGINTERN int Swig_var_Copper_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Copper is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Copper_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Copper)); return pyobj; }
SWIGINTERN int Swig_var_Cu_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cu is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cu_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cu)); return pyobj; }
SWIGINTERN int Swig_var_Zinc_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Zinc is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Zinc_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Zinc)); return pyobj; }
SWIGINTERN int Swig_var_Zn_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Zn is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Zn_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Zn)); return pyobj; }
SWIGINTERN int Swig_var_Gallium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Gallium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Gallium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Gallium)); return pyobj; }
SWIGINTERN int Swig_var_Ga_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ga is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ga_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ga)); return pyobj; }
SWIGINTERN int Swig_var_Germanium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Germanium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Germanium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Germanium)); return pyobj; }
SWIGINTERN int Swig_var_Ge_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ge is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ge_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ge)); return pyobj; }
SWIGINTERN int Swig_var_Arsenic_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Arsenic is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Arsenic_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Arsenic)); return pyobj; }
SWIGINTERN int Swig_var_As_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable As is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_As_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::As)); return pyobj; }
SWIGINTERN int Swig_var_Selenium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Selenium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Selenium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Selenium)); return pyobj; }
SWIGINTERN int Swig_var_Se_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Se is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Se_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Se)); return pyobj; }
SWIGINTERN int Swig_var_Bromine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bromine is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bromine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bromine)); return pyobj; }
SWIGINTERN int Swig_var_Br_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Br is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Br_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Br)); return pyobj; }
SWIGINTERN int Swig_var_Krypton_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Krypton is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Krypton_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Krypton)); return pyobj; }
SWIGINTERN int Swig_var_Kr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Kr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Kr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Kr)); return pyobj; }
SWIGINTERN int Swig_var_Rubidium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rubidium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rubidium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rubidium)); return pyobj; }
SWIGINTERN int Swig_var_Rb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rb)); return pyobj; }
SWIGINTERN int Swig_var_Strontium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Strontium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Strontium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Strontium)); return pyobj; }
SWIGINTERN int Swig_var_Sr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sr)); return pyobj; }
SWIGINTERN int Swig_var_Yttrium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Yttrium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Yttrium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Yttrium)); return pyobj; }
SWIGINTERN int Swig_var_Y_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Y is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Y_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Y)); return pyobj; }
SWIGINTERN int Swig_var_Zirconium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Zirconium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Zirconium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Zirconium)); return pyobj; }
SWIGINTERN int Swig_var_Zr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Zr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Zr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Zr)); return pyobj; }
SWIGINTERN int Swig_var_Niobium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Niobium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Niobium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Niobium)); return pyobj; }
SWIGINTERN int Swig_var_Nb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nb)); return pyobj; }
SWIGINTERN int Swig_var_Molybdenum_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Molybdenum is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Molybdenum_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Molybdenum)); return pyobj; }
SWIGINTERN int Swig_var_Mo_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mo is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mo_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mo)); return pyobj; }
SWIGINTERN int Swig_var_Technetium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Technetium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Technetium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Technetium)); return pyobj; }
SWIGINTERN int Swig_var_Tc_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tc is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tc_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tc)); return pyobj; }
SWIGINTERN int Swig_var_Ruthenium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ruthenium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Ruthenium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ruthenium)); return pyobj; }
SWIGINTERN int Swig_var_Ru_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ru is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ru_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ru)); return pyobj; }
SWIGINTERN int Swig_var_Rhodium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rhodium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rhodium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rhodium)); return pyobj; }
SWIGINTERN int Swig_var_Rh_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rh is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rh_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rh)); return pyobj; }
SWIGINTERN int Swig_var_Palladium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Palladium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Palladium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Palladium)); return pyobj; }
SWIGINTERN int Swig_var_Pd_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pd is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pd_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pd)); return pyobj; }
SWIGINTERN int Swig_var_Silver_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Silver is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Silver_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Silver)); return pyobj; }
SWIGINTERN int Swig_var_Ag_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ag is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ag_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ag)); return pyobj; }
SWIGINTERN int Swig_var_Cadmium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cadmium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cadmium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cadmium)); return pyobj; }
SWIGINTERN int Swig_var_Cd_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cd is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cd_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cd)); return pyobj; }
SWIGINTERN int Swig_var_Indium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Indium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Indium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Indium)); return pyobj; }
SWIGINTERN int Swig_var_In_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable In is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_In_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::In)); return pyobj; }
SWIGINTERN int Swig_var_Tin_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tin is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tin_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tin)); return pyobj; }
SWIGINTERN int Swig_var_Sn_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sn is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sn_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sn)); return pyobj; }
SWIGINTERN int Swig_var_Antimony_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Antimony is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Antimony_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Antimony)); return pyobj; }
SWIGINTERN int Swig_var_Sb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sb)); return pyobj; }
SWIGINTERN int Swig_var_Tellurium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tellurium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Tellurium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tellurium)); return pyobj; }
SWIGINTERN int Swig_var_Te_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Te is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Te_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Te)); return pyobj; }
SWIGINTERN int Swig_var_Iodine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Iodine is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Iodine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Iodine)); return pyobj; }
SWIGINTERN int Swig_var_I_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable I is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_I_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::I)); return pyobj; }
SWIGINTERN int Swig_var_Xenon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Xenon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Xenon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Xenon)); return pyobj; }
SWIGINTERN int Swig_var_Xe_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Xe is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Xe_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Xe)); return pyobj; }
SWIGINTERN int Swig_var_Caesium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Caesium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Caesium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Caesium)); return pyobj; }
SWIGINTERN int Swig_var_Cs_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cs is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cs_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cs)); return pyobj; }
SWIGINTERN int Swig_var_Barium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Barium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Barium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Barium)); return pyobj; }
SWIGINTERN int Swig_var_Ba_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ba is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ba_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ba)); return pyobj; }
SWIGINTERN int Swig_var_Lanthanum_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lanthanum is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Lanthanum_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lanthanum)); return pyobj; }
SWIGINTERN int Swig_var_La_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable La is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_La_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::La)); return pyobj; }
SWIGINTERN int Swig_var_Cerium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cerium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cerium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cerium)); return pyobj; }
SWIGINTERN int Swig_var_Ce_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ce is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ce_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ce)); return pyobj; }
SWIGINTERN int Swig_var_Praseodymium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Praseodymium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Praseodymium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Praseodymium)); return pyobj; }
SWIGINTERN int Swig_var_Pr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pr)); return pyobj; }
SWIGINTERN int Swig_var_Neodymium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Neodymium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Neodymium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Neodymium)); return pyobj; }
SWIGINTERN int Swig_var_Nd_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nd is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nd_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nd)); return pyobj; }
SWIGINTERN int Swig_var_Promethium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Promethium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Promethium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Promethium)); return pyobj; }
SWIGINTERN int Swig_var_Pm_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pm is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pm_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pm)); return pyobj; }
SWIGINTERN int Swig_var_Samarium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Samarium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Samarium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Samarium)); return pyobj; }
SWIGINTERN int Swig_var_Sm_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sm is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sm_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sm)); return pyobj; }
SWIGINTERN int Swig_var_Europium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Europium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Europium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Europium)); return pyobj; }
SWIGINTERN int Swig_var_Eu_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Eu is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Eu_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Eu)); return pyobj; }
SWIGINTERN int Swig_var_Gadolinium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Gadolinium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Gadolinium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Gadolinium)); return pyobj; }
SWIGINTERN int Swig_var_Gd_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Gd is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Gd_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Gd)); return pyobj; }
SWIGINTERN int Swig_var_Terbium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Terbium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Terbium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Terbium)); return pyobj; }
SWIGINTERN int Swig_var_Tb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tb)); return pyobj; }
SWIGINTERN int Swig_var_Dysprosium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Dysprosium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Dysprosium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Dysprosium)); return pyobj; }
SWIGINTERN int Swig_var_Dy_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Dy is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Dy_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Dy)); return pyobj; }
SWIGINTERN int Swig_var_Holmium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Holmium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Holmium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Holmium)); return pyobj; }
SWIGINTERN int Swig_var_Ho_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ho is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ho_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ho)); return pyobj; }
SWIGINTERN int Swig_var_Erbium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Erbium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Erbium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Erbium)); return pyobj; }
SWIGINTERN int Swig_var_Er_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Er is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Er_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Er)); return pyobj; }
SWIGINTERN int Swig_var_Thulium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Thulium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Thulium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Thulium)); return pyobj; }
SWIGINTERN int Swig_var_Tm_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tm is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tm_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tm)); return pyobj; }
SWIGINTERN int Swig_var_Ytterbium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ytterbium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Ytterbium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ytterbium)); return pyobj; }
SWIGINTERN int Swig_var_Yb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Yb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Yb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Yb)); return pyobj; }
SWIGINTERN int Swig_var_Lutetium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lutetium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lutetium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lutetium)); return pyobj; }
SWIGINTERN int Swig_var_Lu_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lu is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lu_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lu)); return pyobj; }
SWIGINTERN int Swig_var_Hafnium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hafnium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hafnium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hafnium)); return pyobj; }
SWIGINTERN int Swig_var_Hf_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hf is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hf_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hf)); return pyobj; }
SWIGINTERN int Swig_var_Tantalum_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tantalum is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tantalum_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tantalum)); return pyobj; }
SWIGINTERN int Swig_var_Ta_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ta is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ta_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ta)); return pyobj; }
SWIGINTERN int Swig_var_Tungsten_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tungsten is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tungsten_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tungsten)); return pyobj; }
SWIGINTERN int Swig_var_W_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable W is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_W_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::W)); return pyobj; }
SWIGINTERN int Swig_var_Rhenium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rhenium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rhenium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rhenium)); return pyobj; }
SWIGINTERN int Swig_var_Re_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Re is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Re_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Re)); return pyobj; }
SWIGINTERN int Swig_var_Osmium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Osmium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Osmium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Osmium)); return pyobj; }
SWIGINTERN int Swig_var_Os_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Os is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Os_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Os)); return pyobj; }
SWIGINTERN int Swig_var_Iridium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Iridium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Iridium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Iridium)); return pyobj; }
SWIGINTERN int Swig_var_Ir_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ir is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ir_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ir)); return pyobj; }
SWIGINTERN int Swig_var_Platinum_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Platinum is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Platinum_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Platinum)); return pyobj; }
SWIGINTERN int Swig_var_Pt_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pt is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pt_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pt)); return pyobj; }
SWIGINTERN int Swig_var_Gold_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Gold is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Gold_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Gold)); return pyobj; }
SWIGINTERN int Swig_var_Au_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Au is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Au_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Au)); return pyobj; }
SWIGINTERN int Swig_var_Mercury_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mercury is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mercury_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mercury)); return pyobj; }
SWIGINTERN int Swig_var_Hg_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hg is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hg_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hg)); return pyobj; }
SWIGINTERN int Swig_var_Thallium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Thallium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Thallium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Thallium)); return pyobj; }
SWIGINTERN int Swig_var_Tl_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tl is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Tl_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tl)); return pyobj; }
SWIGINTERN int Swig_var_Lead_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lead is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lead_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lead)); return pyobj; }
SWIGINTERN int Swig_var_Pb_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pb is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pb_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pb)); return pyobj; }
SWIGINTERN int Swig_var_Bismuth_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bismuth is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bismuth_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bismuth)); return pyobj; }
SWIGINTERN int Swig_var_Bi_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bi is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bi_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bi)); return pyobj; }
SWIGINTERN int Swig_var_Polonium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Polonium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Polonium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Polonium)); return pyobj; }
SWIGINTERN int Swig_var_Po_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Po is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Po_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Po)); return pyobj; }
SWIGINTERN int Swig_var_Astatine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Astatine is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Astatine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Astatine)); return pyobj; }
SWIGINTERN int Swig_var_At_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable At is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_At_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::At)); return pyobj; }
SWIGINTERN int Swig_var_Radon_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Radon is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Radon_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Radon)); return pyobj; }
SWIGINTERN int Swig_var_Rn_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rn is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rn_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rn)); return pyobj; }
SWIGINTERN int Swig_var_Francium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Francium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Francium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Francium)); return pyobj; }
SWIGINTERN int Swig_var_Fr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fr)); return pyobj; }
SWIGINTERN int Swig_var_Radium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Radium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Radium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Radium)); return pyobj; }
SWIGINTERN int Swig_var_Ra_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ra is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ra_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ra)); return pyobj; }
SWIGINTERN int Swig_var_Actinium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Actinium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Actinium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Actinium)); return pyobj; }
SWIGINTERN int Swig_var_Ac_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ac is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ac_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ac)); return pyobj; }
SWIGINTERN int Swig_var_Thorium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Thorium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Thorium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Thorium)); return pyobj; }
SWIGINTERN int Swig_var_Th_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Th is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Th_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Th)); return pyobj; }
SWIGINTERN int Swig_var_Protactinium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Protactinium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Protactinium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Protactinium)); return pyobj; }
SWIGINTERN int Swig_var_Pa_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pa is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pa_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pa)); return pyobj; }
SWIGINTERN int Swig_var_Uranium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Uranium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Uranium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Uranium)); return pyobj; }
SWIGINTERN int Swig_var_U_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable U is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_U_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::U)); return pyobj; }
SWIGINTERN int Swig_var_Neptunium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Neptunium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Neptunium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Neptunium)); return pyobj; }
SWIGINTERN int Swig_var_Np_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Np is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Np_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Np)); return pyobj; }
SWIGINTERN int Swig_var_Plutonium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Plutonium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Plutonium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Plutonium)); return pyobj; }
SWIGINTERN int Swig_var_Pu_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Pu is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Pu_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Pu)); return pyobj; }
SWIGINTERN int Swig_var_Americium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Americium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Americium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Americium)); return pyobj; }
SWIGINTERN int Swig_var_Am_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Am is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Am_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Am)); return pyobj; }
SWIGINTERN int Swig_var_Curium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Curium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Curium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Curium)); return pyobj; }
SWIGINTERN int Swig_var_Cm_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cm is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cm_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cm)); return pyobj; }
SWIGINTERN int Swig_var_Berkelium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Berkelium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Berkelium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Berkelium)); return pyobj; }
SWIGINTERN int Swig_var_Bk_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bk is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bk_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bk)); return pyobj; }
SWIGINTERN int Swig_var_Californium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Californium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Californium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Californium)); return pyobj; }
SWIGINTERN int Swig_var_Cf_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cf is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cf_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cf)); return pyobj; }
SWIGINTERN int Swig_var_Einsteinium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Einsteinium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Einsteinium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Einsteinium)); return pyobj; }
SWIGINTERN int Swig_var_Es_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Es is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Es_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Es)); return pyobj; }
SWIGINTERN int Swig_var_Fermium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fermium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fermium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fermium)); return pyobj; }
SWIGINTERN int Swig_var_Fm_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fm is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fm_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fm)); return pyobj; }
SWIGINTERN int Swig_var_Mendelevium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mendelevium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Mendelevium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mendelevium)); return pyobj; }
SWIGINTERN int Swig_var_Md_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Md is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Md_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Md)); return pyobj; }
SWIGINTERN int Swig_var_Nobelium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nobelium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nobelium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nobelium)); return pyobj; }
SWIGINTERN int Swig_var_No_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable No is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_No_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::No)); return pyobj; }
SWIGINTERN int Swig_var_Lawrencium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lawrencium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Lawrencium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lawrencium)); return pyobj; }
SWIGINTERN int Swig_var_Lr_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lr is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lr_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lr)); return pyobj; }
SWIGINTERN int Swig_var_Rutherfordium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rutherfordium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Rutherfordium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rutherfordium)); return pyobj; }
SWIGINTERN int Swig_var_Rf_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rf is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rf_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rf)); return pyobj; }
SWIGINTERN int Swig_var_Dubnium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Dubnium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Dubnium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Dubnium)); return pyobj; }
SWIGINTERN int Swig_var_Db_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Db is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Db_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Db)); return pyobj; }
SWIGINTERN int Swig_var_Seaborgium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Seaborgium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Seaborgium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Seaborgium)); return pyobj; }
SWIGINTERN int Swig_var_Sg_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Sg is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Sg_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Sg)); return pyobj; }
SWIGINTERN int Swig_var_Bohrium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bohrium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bohrium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bohrium)); return pyobj; }
SWIGINTERN int Swig_var_Bh_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Bh is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Bh_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Bh)); return pyobj; }
SWIGINTERN int Swig_var_Hassium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hassium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hassium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hassium)); return pyobj; }
SWIGINTERN int Swig_var_Hs_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Hs is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Hs_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Hs)); return pyobj; }
SWIGINTERN int Swig_var_Meitnerium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Meitnerium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Meitnerium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Meitnerium)); return pyobj; }
SWIGINTERN int Swig_var_Mt_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mt is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mt_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mt)); return pyobj; }
SWIGINTERN int Swig_var_Darmstadtium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Darmstadtium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Darmstadtium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Darmstadtium)); return pyobj; }
SWIGINTERN int Swig_var_Ds_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ds is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ds_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ds)); return pyobj; }
SWIGINTERN int Swig_var_Roentgenium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Roentgenium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Roentgenium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Roentgenium)); return pyobj; }
SWIGINTERN int Swig_var_Rg_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Rg is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Rg_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Rg)); return pyobj; }
SWIGINTERN int Swig_var_Copernicium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Copernicium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Copernicium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Copernicium)); return pyobj; }
SWIGINTERN int Swig_var_Cn_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Cn is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Cn_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Cn)); return pyobj; }
SWIGINTERN int Swig_var_Nihonium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nihonium is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nihonium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nihonium)); return pyobj; }
SWIGINTERN int Swig_var_Nh_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Nh is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Nh_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Nh)); return pyobj; }
SWIGINTERN int Swig_var_Flerovium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Flerovium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Flerovium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Flerovium)); return pyobj; }
SWIGINTERN int Swig_var_Fl_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Fl is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Fl_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Fl)); return pyobj; }
SWIGINTERN int Swig_var_Moscovium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Moscovium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Moscovium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Moscovium)); return pyobj; }
SWIGINTERN int Swig_var_Mc_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Mc is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Mc_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Mc)); return pyobj; }
SWIGINTERN int Swig_var_Livermorium_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Livermorium is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Livermorium_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Livermorium)); return pyobj; }
SWIGINTERN int Swig_var_Lv_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Lv is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Lv_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Lv)); return pyobj; }
SWIGINTERN int Swig_var_Tennessine_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Tennessine is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Tennessine_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Tennessine)); return pyobj; }
SWIGINTERN int Swig_var_Ts_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Ts is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Ts_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Ts)); return pyobj; }
SWIGINTERN int Swig_var_Oganesson_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Oganesson is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_Oganesson_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Oganesson)); return pyobj; }
SWIGINTERN int Swig_var_Og_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable Og is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_Og_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBElements::Og)); return pyobj; }
SWIGINTERN PyObject *_wrap_new_OBResidue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBResidue *)new OpenBabel::OBResidue();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBResidue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBResidue *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBResidue,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBResidue" "', argument " "1"" of type '" "OpenBabel::OBResidue const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBResidue" "', argument " "1"" of type '" "OpenBabel::OBResidue const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (OpenBabel::OBResidue *)new OpenBabel::OBResidue((OpenBabel::OBResidue const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBResidue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBResidue", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBResidue__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_new_OBResidue__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBResidue'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidue::OBResidue()\n"
  "    OpenBabel::OBResidue::OBResidue(OpenBabel::OBResidue const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBResidue" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_AddAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_AddAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_AddAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->AddAtom(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_InsertAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_InsertAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_InsertAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_InsertAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->InsertAtom(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_RemoveAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_RemoveAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_RemoveAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_RemoveAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->RemoveAtom(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetName" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_SetName" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidue_SetName" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetName((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetSegName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetSegName", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetSegName" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_SetSegName" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidue_SetSegName" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetSegName((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetNum__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetNum" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_SetNum" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetNum(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetNum__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetNum" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBResidue_SetNum" "', argument " "2"" of type '" "std::string const""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetNum(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetNum(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBResidue_SetNum", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      { int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_1;
    return _wrap_OBResidue_SetNum__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBResidue_SetNum__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBResidue_SetNum'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidue::SetNum(unsigned int const)\n"
  "    OpenBabel::OBResidue::SetNum(std::string const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBResidue_SetChain(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; char arg2 ; void *argp1 = 0 ; int res1 = 0 ; char val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetChain", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetChain" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_SetChain" "', argument " "2"" of type '" "char""'"); }  
  arg2 = static_cast< char >(val2); (arg1)->SetChain(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetChainNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetChainNum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetChainNum" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_SetChainNum" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetChainNum(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetIdx" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_SetIdx" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetIdx(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetInsertionCode(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; char arg2 ; void *argp1 = 0 ; int res1 = 0 ; char val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetInsertionCode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetInsertionCode" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_SetInsertionCode" "', argument " "2"" of type '" "char""'"); }
    arg2 = static_cast< char >(val2); (arg1)->SetInsertionCode(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetAtomID(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; std::string *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetAtomID", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetAtomID" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_SetAtomID" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); { std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBResidue_SetAtomID" "', argument " "3"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidue_SetAtomID" "', argument " "3"" of type '" "std::string const &""'"); }
     arg3 = ptr; }  (arg1)->SetAtomID(arg2,(std::string const &)*arg3); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetHetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; bool val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetHetAtom", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetHetAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_SetHetAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBResidue_SetHetAtom" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); (arg1)->SetHetAtom(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_SetSerialNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_SetSerialNum", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_SetSerialNum" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_SetSerialNum" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBResidue_SetSerialNum" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetSerialNum(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetName" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = ((OpenBabel::OBResidue const *)arg1)->GetName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetSegName(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetSegName" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = ((OpenBabel::OBResidue const *)arg1)->GetSegName();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetNum" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = (int)(arg1)->GetNum();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetNumString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetNumString" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = (arg1)->GetNumString();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetNumAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetNumAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetNumAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetNumHvyAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetNumHvyAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetNumHvyAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetChain(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetChain" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = (char)((OpenBabel::OBResidue const *)arg1)->GetChain();
  resultobj = SWIG_From_char(static_cast< char >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetChainNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetChainNum" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetChainNum();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetIdx" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetResKey(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetResKey" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetResKey();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetBonds__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_GetBonds" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = ((OpenBabel::OBResidue const *)arg1)->GetBonds(arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetBonds__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = ((OpenBabel::OBResidue const *)arg1)->GetBonds();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetBonds(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBResidue_GetBonds", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBResidue_GetBonds__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBResidue_GetBonds__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBResidue_GetBonds'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidue::GetBonds(bool) const\n"
  "    OpenBabel::OBResidue::GetBonds() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBResidue_GetAtomID(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; std::string result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_GetAtomID", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetAtomID" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_GetAtomID" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = ((OpenBabel::OBResidue const *)arg1)->GetAtomID(arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetSerialNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_GetSerialNum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetSerialNum" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_GetSerialNum" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  result = (unsigned int)((OpenBabel::OBResidue const *)arg1)->GetSerialNum(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetInsertionCode(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetInsertionCode" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (char)((OpenBabel::OBResidue const *)arg1)->GetInsertionCode();
  resultobj = SWIG_From_char(static_cast< char >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetAminoAcidProperty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_GetAminoAcidProperty", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetAminoAcidProperty" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_GetAminoAcidProperty" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)((OpenBabel::OBResidue const *)arg1)->GetAminoAcidProperty(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetAtomProperty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_GetAtomProperty", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetAtomProperty" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_GetAtomProperty" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBResidue_GetAtomProperty" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); result = (bool)((OpenBabel::OBResidue const *)arg1)->GetAtomProperty(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_GetResidueProperty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_GetResidueProperty", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_GetResidueProperty" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_GetResidueProperty" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)((OpenBabel::OBResidue const *)arg1)->GetResidueProperty(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_IsHetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_IsHetAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_IsHetAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_IsHetAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)((OpenBabel::OBResidue const *)arg1)->IsHetAtom(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_IsResidueType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_IsResidueType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_IsResidueType" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBResidue_IsResidueType" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)((OpenBabel::OBResidue const *)arg1)->IsResidueType(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_BeginAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_BeginAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = (arg1)->BeginAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_CBeginAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_CBeginAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = ((OpenBabel::OBResidue const *)arg1)->CBeginAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomConstIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_EndAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_EndAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = (arg1)->EndAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_CEndAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_CEndAtoms" "', argument " "1"" of type '" "OpenBabel::OBResidue const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1); result = ((OpenBabel::OBResidue const *)arg1)->CEndAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomConstIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_BeginAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ;
  std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_BeginAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_BeginAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_BeginAtom" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidue_BeginAtom" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator * >(argp2);
  result = (OpenBabel::OBAtom *)(arg1)->BeginAtom(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBResidue_NextAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidue *arg1 = 0 ;
  std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBResidue_NextAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBResidue_NextAtom" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBResidue_NextAtom" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBResidue_NextAtom" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator * >(argp2);
  result = (OpenBabel::OBAtom *)(arg1)->NextAtom(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBResidue_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBResidue, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBResidue_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBInternalCoord__a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__a_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBInternalCoord__a_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->_a = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__a_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->_a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__b_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBInternalCoord__b_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->_b = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__b_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->_b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__c_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBInternalCoord__c_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->_c = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__c_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->_c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__dst_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__dst_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__dst_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBInternalCoord__dst_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->_dst = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__dst_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__dst_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (double) ((arg1)->_dst);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__ang_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__ang_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__ang_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBInternalCoord__ang_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->_ang = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__ang_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__ang_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (double) ((arg1)->_ang);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__tor_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBInternalCoord__tor_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__tor_set" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBInternalCoord__tor_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->_tor = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBInternalCoord__tor_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBInternalCoord__tor_get" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); result = (double) ((arg1)->_tor);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  double arg4 ; double arg5 ; double arg6 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; double val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ; double val6 ; int ecode6 = 0 ;
  OpenBabel::OBInternalCoord *result = 0 ; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBInternalCoord" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBInternalCoord" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBInternalCoord" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_OBInternalCoord" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_OBInternalCoord" "', argument " "6"" of type '" "double""'"); }
    arg6 = static_cast< double >(val6);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  double arg4 ; double arg5 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  double val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ; OpenBabel::OBInternalCoord *result = 0 ; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBInternalCoord" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBInternalCoord" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBInternalCoord" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_OBInternalCoord" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  double arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double val4 ;
  int ecode4 = 0 ; OpenBabel::OBInternalCoord *result = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBInternalCoord" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBInternalCoord" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBInternalCoord" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  OpenBabel::OBInternalCoord *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBInternalCoord" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBInternalCoord" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::OBInternalCoord *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBInternalCoord" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBInternalCoord *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBInternalCoord *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBInternalCoord *)new OpenBabel::OBInternalCoord();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBInternalCoord(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBInternalCoord", 0, 6, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_6(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_5(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_new_OBInternalCoord__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBInternalCoord'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,double,double,double)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,double,double)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,double)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord(OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBInternalCoord::OBInternalCoord()\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBInternalCoord *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBInternalCoord *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBInternalCoord * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBInternalCoord_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBInternalCoord, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBInternalCoord_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBAtomAtomIterAdaptor__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIterAdaptor *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBAtomAtomIterAdaptor *)new OpenBabel::OBAtomAtomIterAdaptor();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtomAtomIterAdaptor__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > arg1 ;
  OpenBabel::OBAtom *arg2 = 0 ; void *argp1 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  OpenBabel::OBAtomAtomIterAdaptor *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAtomAtomIterAdaptor" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor::Iter""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAtomAtomIterAdaptor" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor::Iter""'"); }
     else { OpenBabel::OBAtomAtomIterAdaptor::Iter * temp = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor::Iter * >(argp1);
      arg1 = *temp; if (SWIG_IsNewObj(res1)) delete temp; }  } 
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBAtomAtomIterAdaptor" "', argument " "2"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  result = (OpenBabel::OBAtomAtomIterAdaptor *)new OpenBabel::OBAtomAtomIterAdaptor(SWIG_STD_MOVE(arg1),(OpenBabel::OBAtom const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtomAtomIterAdaptor__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > arg1 ;
  void *argp1 ; int res1 = 0 ; OpenBabel::OBAtomAtomIterAdaptor *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; {
    res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator,  0  | 0);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAtomAtomIterAdaptor" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor::Iter""'");  }
       if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAtomAtomIterAdaptor" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor::Iter""'"); }
     else { OpenBabel::OBAtomAtomIterAdaptor::Iter * temp = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor::Iter * >(argp1);
      arg1 = *temp; if (SWIG_IsNewObj(res1)) delete temp; }  } 
  result = (OpenBabel::OBAtomAtomIterAdaptor *)new OpenBabel::OBAtomAtomIterAdaptor(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtomAtomIterAdaptor(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBAtomAtomIterAdaptor", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBAtomAtomIterAdaptor__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBAtomAtomIterAdaptor__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBAtomAtomIterAdaptor__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBAtomAtomIterAdaptor'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomAtomIterAdaptor::OBAtomAtomIterAdaptor()\n"
  "    OpenBabel::OBAtomAtomIterAdaptor::OBAtomAtomIterAdaptor(OpenBabel::OBAtomAtomIterAdaptor::Iter,OpenBabel::OBAtom const *)\n"
  "    OpenBabel::OBAtomAtomIterAdaptor::OBAtomAtomIterAdaptor(OpenBabel::OBAtomAtomIterAdaptor::Iter)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAtomIterAdaptor___ref__" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBAtomAtomIterAdaptor const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomAtomIterAdaptor *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAtomIterAdaptor_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1);
  result = (OpenBabel::OBAtomAtomIterAdaptor *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBAtomAtomIterAdaptor result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAtomIterAdaptor_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtomAtomIterAdaptor_inc" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomAtomIterAdaptor(result)), SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtomAtomIterAdaptor_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtomAtomIterAdaptor_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtomAtomIterAdaptor_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtomAtomIterAdaptor_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomAtomIterAdaptor::operator ++()\n"
  "    OpenBabel::OBAtomAtomIterAdaptor::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; OpenBabel::OBAtomAtomIterAdaptor *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtomAtomIterAdaptor___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAtomIterAdaptor___eq__" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtomAtomIterAdaptor___eq__" "', argument " "2"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtomAtomIterAdaptor___eq__" "', argument " "2"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp2);
  result = (bool)((OpenBabel::OBAtomAtomIterAdaptor const *)arg1)->operator ==((OpenBabel::OBAtomAtomIterAdaptor const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBAtomAtomIterAdaptor___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; OpenBabel::OBAtomAtomIterAdaptor *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtomAtomIterAdaptor___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtomAtomIterAdaptor___ne__" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtomAtomIterAdaptor___ne__" "', argument " "2"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtomAtomIterAdaptor___ne__" "', argument " "2"" of type '" "OpenBabel::OBAtomAtomIterAdaptor const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp2);
  result = (bool)((OpenBabel::OBAtomAtomIterAdaptor const *)arg1)->operator !=((OpenBabel::OBAtomAtomIterAdaptor const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_delete_OBAtomAtomIterAdaptor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIterAdaptor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAtomAtomIterAdaptor" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIterAdaptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIterAdaptor * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBAtomAtomIterAdaptor_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomAtomIterAdaptor, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAtomAtomIterAdaptor_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBAtom_Visit_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_Visit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_Visit_set" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_Visit_set" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); if (arg1) (arg1)->Visit = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_Visit_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_Visit_get" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool) ((arg1)->Visit);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBAtom", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBAtom *)new OpenBabel::OBAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAtom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom___eq__" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom___eq__" "', argument " "2"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  result = (bool)((OpenBabel::OBAtom const *)arg1)->operator ==((OpenBabel::OBAtom const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBAtom_Duplicate(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_Duplicate", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_Duplicate" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_Duplicate" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->Duplicate(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetIdx" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetIdx" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetIdx(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetId", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetId" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetId" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); (arg1)->SetId(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetHyb(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetHyb", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetHyb" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetHyb" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetHyb(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetAtomicNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetAtomicNum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetAtomicNum" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetAtomicNum" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetAtomicNum(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetIsotope(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetIsotope", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetIsotope" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetIsotope" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetIsotope(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetImplicitHCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetImplicitHCount", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetImplicitHCount" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetImplicitHCount" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetImplicitHCount(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetFormalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetFormalCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetFormalCharge" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetFormalCharge" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetFormalCharge(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetSpinMultiplicity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; short arg2 ; void *argp1 = 0 ; int res1 = 0 ; short val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetSpinMultiplicity", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetSpinMultiplicity" "', argument " "2"" of type '" "short""'"); }
    arg2 = static_cast< short >(val2); (arg1)->SetSpinMultiplicity(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetType__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetType" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetType" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetType((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetType__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetType" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetType" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_SetType" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetType((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetType(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_SetType", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_1;
    return _wrap_OBAtom_SetType__SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_SetType__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_SetType'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::SetType(char const *)\n"
  "    OpenBabel::OBAtom::SetType(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_SetPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetPartialCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetPartialCharge" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetPartialCharge(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetVector" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_SetVector" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); (arg1)->SetVector((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; double arg2 ; double arg3 ; double arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetVector" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBAtom_SetVector" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBAtom_SetVector" "', argument " "4"" of type '" "double""'"); }  
  arg4 = static_cast< double >(val4); (arg1)->SetVector(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetCoordPtr(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; double **arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetCoordPtr", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetCoordPtr" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetCoordPtr" "', argument " "2"" of type '" "double **""'");  }
   arg2 = reinterpret_cast< double ** >(argp2); (arg1)->SetCoordPtr(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetVector__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->SetVector(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetVector(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_SetVector", 0, 4, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_SetVector__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_SetVector__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBAtom_SetVector__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_SetVector'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::SetVector(OpenBabel::vector3 const &)\n"
  "    OpenBabel::OBAtom::SetVector(double const,double const,double const)\n" "    OpenBabel::OBAtom::SetVector()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_SetResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetResidue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); (arg1)->SetResidue(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetParent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetParent" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_SetParent" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetParent(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetAromatic__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetAromatic" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetAromatic" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetAromatic(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetAromatic__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetAromatic" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->SetAromatic(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetAromatic(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_SetAromatic", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_SetAromatic__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_SetAromatic__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_SetAromatic'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::SetAromatic(bool)\n" "    OpenBabel::OBAtom::SetAromatic()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_SetInRing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetInRing" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetInRing" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetInRing(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetInRing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetInRing" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->SetInRing(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetInRing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_SetInRing", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_SetInRing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_SetInRing__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_SetInRing'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::SetInRing(bool)\n" "    OpenBabel::OBAtom::SetInRing()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_ClearCoordPtr(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_ClearCoordPtr" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->ClearCoordPtr(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetFormalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetFormalCharge" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (int)((OpenBabel::OBAtom const *)arg1)->GetFormalCharge();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetAtomicNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetAtomicNum" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetAtomicNum();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetIsotope(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned short result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetIsotope" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned short)((OpenBabel::OBAtom const *)arg1)->GetIsotope();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetSpinMultiplicity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (int)((OpenBabel::OBAtom const *)arg1)->GetSpinMultiplicity(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetAtomicMass(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetAtomicMass" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->GetAtomicMass();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetExactMass(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetExactMass" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->GetExactMass();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetIdx" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetIndex" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetId" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (unsigned long)((OpenBabel::OBAtom const *)arg1)->GetId();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetCoordinateIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetCoordinateIdx" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetCoordinateIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetExplicitDegree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetExplicitDegree" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetExplicitDegree();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetTotalDegree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetTotalDegree" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetTotalDegree();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetExplicitValence(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetExplicitValence" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetExplicitValence();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetTotalValence(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetTotalValence" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetTotalValence();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetHyb(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetHyb" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetHyb();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetImplicitHCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned char result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetImplicitHCount" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned char)((OpenBabel::OBAtom const *)arg1)->GetImplicitHCount();
  resultobj = SWIG_From_unsigned_SS_char(static_cast< unsigned char >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetHvyDegree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetHvyDegree" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetHvyDegree();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetHeteroDegree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetHeteroDegree" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->GetHeteroDegree();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetType" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (char *)(arg1)->GetType();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetX(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetX" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->GetX();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetY(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetY" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->GetY();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetZ(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetZ" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->GetZ();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_x(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_x" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->x();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_y(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_y" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->y();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_z(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_z" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)((OpenBabel::OBAtom const *)arg1)->z();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetCoordinate(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetCoordinate" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double *)(arg1)->GetCoordinate();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (OpenBabel::vector3 *) &(arg1)->GetVector();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetVector" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (OpenBabel::vector3 *) &((OpenBabel::OBAtom const *)arg1)->GetVector();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetVector(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_GetVector", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_GetVector__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_GetVector__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_GetVector'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::GetVector()\n" "    OpenBabel::OBAtom::GetVector() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_GetPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)(arg1)->GetPartialCharge();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBResidue *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (OpenBabel::OBResidue *)(arg1)->GetResidue();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBMol *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetParent" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (OpenBabel::OBMol *)(arg1)->GetParent();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetNewBondVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_GetNewBondVector", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetNewBondVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_GetNewBondVector" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_GetNewBondVector" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBAtom_GetNewBondVector" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); result = (bool)(arg1)->GetNewBondVector(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBBond *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_GetBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_GetBond" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (OpenBabel::OBBond *)(arg1)->GetBond(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_BeginBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_BeginBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (arg1)->BeginBonds();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBondIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_EndBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_EndBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (arg1)->EndBonds();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBondIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = ((OpenBabel::OBAtom const *)arg1)->GetBonds();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_BeginBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBondIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBond *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_BeginBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_BeginBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_BeginBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_BeginBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBondIterator * >(argp2); result = (OpenBabel::OBBond *)(arg1)->BeginBond(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_NextBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBondIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBBond *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_NextBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_NextBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_NextBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_NextBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBondIterator * >(argp2); result = (OpenBabel::OBBond *)(arg1)->NextBond(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetNbrs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,OpenBabel::OBAtomAtomIterAdaptor > > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetNbrs" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = ((OpenBabel::OBAtom const *)arg1)->GetNbrs();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_BeginNbrAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBondIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_BeginNbrAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_BeginNbrAtom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_BeginNbrAtom" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_BeginNbrAtom" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBondIterator * >(argp2); result = (OpenBabel::OBAtom *)(arg1)->BeginNbrAtom(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_NextNbrAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBondIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_NextNbrAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_NextNbrAtom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_NextNbrAtom" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_NextNbrAtom" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBondIterator * >(argp2); result = (OpenBabel::OBAtom *)(arg1)->NextNbrAtom(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetDistance__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetDistance" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_GetDistance" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (double)(arg1)->GetDistance(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetDistance__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetDistance" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_GetDistance" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (double)(arg1)->GetDistance(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetDistance__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetDistance" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_GetDistance" "', argument " "2"" of type '" "OpenBabel::vector3 *""'");  }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); result = (double)(arg1)->GetDistance(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetDistance(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_GetDistance", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0;
    { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBAtom_GetDistance__SWIG_1(self, argc, argv);}  check_1: if (argc == 2) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__vector3, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_OBAtom_GetDistance__SWIG_2(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_GetDistance__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_GetDistance'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::GetDistance(int)\n"
  "    OpenBabel::OBAtom::GetDistance(OpenBabel::OBAtom *)\n" "    OpenBabel::OBAtom::GetDistance(OpenBabel::vector3 *)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_GetAngle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; double result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_GetAngle" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBAtom_GetAngle" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (double)(arg1)->GetAngle(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetAngle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_GetAngle" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBAtom_GetAngle" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3); result = (double)(arg1)->GetAngle(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_GetAngle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_GetAngle", 0, 3, argv))) SWIG_fail; --argc; if (argc == 3) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBAtom_GetAngle__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 3) { PyObject *retobj = _wrap_OBAtom_GetAngle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_GetAngle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::GetAngle(int,int)\n"
  "    OpenBabel::OBAtom::GetAngle(OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_NewResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_NewResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->NewResidue(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_AddResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_AddResidue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_AddResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_AddResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); (arg1)->AddResidue(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_DeleteResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_DeleteResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->DeleteResidue(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_AddBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_AddBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_AddBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_AddBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); (arg1)->AddBond(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_InsertBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBondIterator *arg2 = 0 ; OpenBabel::OBBond *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_InsertBond", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_InsertBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_InsertBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAtom_InsertBond" "', argument " "2"" of type '" "OpenBabel::OBBondIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBondIterator * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBAtom_InsertBond" "', argument " "3"" of type '" "OpenBabel::OBBond *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBond * >(argp3); (arg1)->InsertBond(*arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_DeleteBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_DeleteBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_DeleteBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_DeleteBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->DeleteBond(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_ClearBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_ClearBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); (arg1)->ClearBond(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HtoMethyl(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HtoMethyl" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HtoMethyl();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SetHybAndGeom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_SetHybAndGeom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SetHybAndGeom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_SetHybAndGeom" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->SetHybAndGeom(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_CountFreeOxygens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_CountFreeOxygens" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->CountFreeOxygens();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_CountFreeSulfurs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_CountFreeSulfurs" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->CountFreeSulfurs();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_ExplicitHydrogenCount__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; unsigned int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_ExplicitHydrogenCount" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_ExplicitHydrogenCount" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->ExplicitHydrogenCount(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_ExplicitHydrogenCount__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_ExplicitHydrogenCount" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->ExplicitHydrogenCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_ExplicitHydrogenCount(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_ExplicitHydrogenCount", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBAtom_ExplicitHydrogenCount__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_ExplicitHydrogenCount__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_ExplicitHydrogenCount'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::ExplicitHydrogenCount(bool) const\n"
  "    OpenBabel::OBAtom::ExplicitHydrogenCount() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_MemberOfRingCount(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_MemberOfRingCount" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->MemberOfRingCount();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_MemberOfRingSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_MemberOfRingSize" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->MemberOfRingSize();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_CountRingBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_CountRingBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBAtom const *)arg1)->CountRingBonds();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_SmallestBondAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_SmallestBondAngle" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)(arg1)->SmallestBondAngle();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_AverageBondAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_AverageBondAngle" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (double)(arg1)->AverageBondAngle();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_LewisAcidBaseCounts(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::pair< int,int > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_LewisAcidBaseCounts" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = ((OpenBabel::OBAtom const *)arg1)->LewisAcidBaseCounts();
  resultobj = SWIG_NewPointerObj((new std::pair< int,int >(result)), SWIGTYPE_p_std__pairT_int_int_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasResidue" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasResidue();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHetAtom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHetAtom();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsElement(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_IsElement", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsElement" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_IsElement" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)((OpenBabel::OBAtom const *)arg1)->IsElement(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsAromatic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsAromatic" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)((OpenBabel::OBAtom const *)arg1)->IsAromatic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsInRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsInRing" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)((OpenBabel::OBAtom const *)arg1)->IsInRing();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsInRingSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAtom_IsInRingSize", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsInRingSize" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_IsInRingSize" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (bool)((OpenBabel::OBAtom const *)arg1)->IsInRingSize(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHeteroatom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHeteroatom" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHeteroatom();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsConnected(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_IsConnected", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsConnected" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_IsConnected" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->IsConnected(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsOneThree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_IsOneThree", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsOneThree" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_IsOneThree" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->IsOneThree(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsOneFour(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_IsOneFour", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsOneFour" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_IsOneFour" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->IsOneFour(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsCarboxylOxygen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsCarboxylOxygen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsCarboxylOxygen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsPhosphateOxygen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsPhosphateOxygen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsPhosphateOxygen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsSulfateOxygen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsSulfateOxygen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsSulfateOxygen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsNitroOxygen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsNitroOxygen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsNitroOxygen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsAmideNitrogen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsAmideNitrogen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsAmideNitrogen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsPolarHydrogen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsPolarHydrogen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsPolarHydrogen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsNonPolarHydrogen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsNonPolarHydrogen" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsNonPolarHydrogen();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsAromaticNOxide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsAromaticNOxide" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsAromaticNOxide();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsChiral(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsChiral" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsChiral();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsPeriodic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsPeriodic" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)((OpenBabel::OBAtom const *)arg1)->IsPeriodic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsAxial(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsAxial" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsAxial();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHbondAcceptor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHbondAcceptor" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHbondAcceptor();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHbondAcceptorSimple(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHbondAcceptorSimple" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHbondAcceptorSimple();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHbondDonor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHbondDonor" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHbondDonor();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsHbondDonorH(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsHbondDonorH" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsHbondDonorH();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_IsMetal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_IsMetal" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->IsMetal();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasAlphaBetaUnsat__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasAlphaBetaUnsat" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_HasAlphaBetaUnsat" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (bool)(arg1)->HasAlphaBetaUnsat(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasAlphaBetaUnsat__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasAlphaBetaUnsat" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasAlphaBetaUnsat();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasAlphaBetaUnsat(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBAtom_HasAlphaBetaUnsat", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBAtom_HasAlphaBetaUnsat__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBAtom_HasAlphaBetaUnsat__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBAtom_HasAlphaBetaUnsat'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtom::HasAlphaBetaUnsat(bool)\n"
  "    OpenBabel::OBAtom::HasAlphaBetaUnsat()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBAtom_HasBondOfOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_HasBondOfOrder", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasBondOfOrder" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_HasBondOfOrder" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->HasBondOfOrder(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_CountBondsOfOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_CountBondsOfOrder", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_CountBondsOfOrder" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAtom_CountBondsOfOrder" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (int)(arg1)->CountBondsOfOrder(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HighestBondOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HighestBondOrder" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (int)(arg1)->HighestBondOrder();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasNonSingleBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasNonSingleBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasNonSingleBond();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasSingleBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasSingleBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasSingleBond();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasDoubleBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasDoubleBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasDoubleBond();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_HasAromaticBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_HasAromaticBond" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = (bool)(arg1)->HasAromaticBond();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAtom_MatchesSMARTS(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAtom_MatchesSMARTS", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAtom_MatchesSMARTS" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAtom_MatchesSMARTS" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->MatchesSMARTS((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *OBAtom_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtom, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAtom_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBBond_Visit_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_Visit_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_Visit_set" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_Visit_set" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); if (arg1) (arg1)->Visit = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_Visit_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_Visit_get" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool) ((arg1)->Visit);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBond *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBBond", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBBond *)new OpenBabel::OBBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBBond" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetIdx" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetIdx" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetIdx(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ;
  unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetId", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetId" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetId" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); (arg1)->SetId(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetBondOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetBondOrder", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetBondOrder" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetBondOrder" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetBondOrder(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetBegin(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetBegin", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetBegin" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_SetBegin" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->SetBegin(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetEnd(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetEnd", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetEnd" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_SetEnd" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->SetEnd(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBond_SetParent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetParent" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_SetParent" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetParent(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetLength__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; double arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetLength" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_SetLength" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBond_SetLength" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); (arg1)->SetLength(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetLength__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetLength" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetLength" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetLength(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetLength(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetLength", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetLength__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBBond_SetLength__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetLength'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetLength(OpenBabel::OBAtom *,double)\n"
  "    OpenBabel::OBBond::SetLength(double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBond_Set(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ;
  int arg2 ; OpenBabel::OBAtom *arg3 = 0 ; OpenBabel::OBAtom *arg4 = 0 ; int arg5 ; int arg6 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; int val5 ; int ecode5 = 0 ;
  int val6 ; int ecode6 = 0 ; PyObject *swig_obj[6] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBond_Set", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_Set" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_Set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBBond_Set" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBBond_Set" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBBond_Set" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_int(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBBond_Set" "', argument " "6"" of type '" "int""'"); }  
  arg6 = static_cast< int >(val6); (arg1)->Set(arg2,arg3,arg4,arg5,arg6); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetAromatic__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetAromatic" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetAromatic" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetAromatic(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetAromatic__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetAromatic" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetAromatic(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetAromatic(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetAromatic", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetAromatic__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetAromatic__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetAromatic'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetAromatic(bool)\n" "    OpenBabel::OBBond::SetAromatic()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedge__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetWedge" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetWedge" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetWedge(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedge__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetWedge" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetWedge(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedge(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetWedge", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetWedge__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetWedge__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetWedge'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetWedge(bool)\n" "    OpenBabel::OBBond::SetWedge()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_SetHash__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetHash" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetHash" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetHash(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetHash__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetHash" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetHash(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetHash(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetHash", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetHash__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetHash__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetHash'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetHash(bool)\n" "    OpenBabel::OBBond::SetHash()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedgeOrHash__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetWedgeOrHash" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetWedgeOrHash" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetWedgeOrHash(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedgeOrHash__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetWedgeOrHash" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetWedgeOrHash(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetWedgeOrHash(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetWedgeOrHash", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetWedgeOrHash__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetWedgeOrHash__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetWedgeOrHash'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetWedgeOrHash(bool)\n"
  "    OpenBabel::OBBond::SetWedgeOrHash()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBond_SetInRing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetInRing" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetInRing" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetInRing(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetInRing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetInRing" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetInRing(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetInRing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetInRing", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetInRing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetInRing__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetInRing'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetInRing(bool)\n" "    OpenBabel::OBBond::SetInRing()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_SetClosure__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetClosure" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_SetClosure" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->SetClosure(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetClosure__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_SetClosure" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); (arg1)->SetClosure(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_SetClosure(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_SetClosure", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_SetClosure__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_SetClosure__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_SetClosure'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::SetClosure(bool)\n" "    OpenBabel::OBBond::SetClosure()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_GetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetIdx" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (unsigned int)((OpenBabel::OBBond const *)arg1)->GetIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetId(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetId" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (unsigned long)((OpenBabel::OBBond const *)arg1)->GetId();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetBondOrder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetBondOrder" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (unsigned int)((OpenBabel::OBBond const *)arg1)->GetBondOrder();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetFlags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetFlags" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (unsigned int)((OpenBabel::OBBond const *)arg1)->GetFlags();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetBeginAtomIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetBeginAtomIdx" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (unsigned int)((OpenBabel::OBBond const *)arg1)->GetBeginAtomIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetEndAtomIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetEndAtomIdx" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (unsigned int)((OpenBabel::OBBond const *)arg1)->GetEndAtomIdx();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetBeginAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetBeginAtom" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (OpenBabel::OBAtom *)(arg1)->GetBeginAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetBeginAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetBeginAtom" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBBond const *)arg1)->GetBeginAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetBeginAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_GetBeginAtom", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_GetBeginAtom__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_GetBeginAtom__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_GetBeginAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::GetBeginAtom()\n"
  "    OpenBabel::OBBond::GetBeginAtom() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBond_GetEndAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetEndAtom" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (OpenBabel::OBAtom *)(arg1)->GetEndAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetEndAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetEndAtom" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBBond const *)arg1)->GetEndAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetEndAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_GetEndAtom", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_GetEndAtom__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_GetEndAtom__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_GetEndAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::GetEndAtom()\n" "    OpenBabel::OBBond::GetEndAtom() const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_GetNbrAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBond_GetNbrAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetNbrAtom" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_GetNbrAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (OpenBabel::OBAtom *)(arg1)->GetNbrAtom(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBMol *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetParent" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (OpenBabel::OBMol *)(arg1)->GetParent();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetEquibLength(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetEquibLength" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (double)((OpenBabel::OBBond const *)arg1)->GetEquibLength();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetLength(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetLength" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (double)((OpenBabel::OBBond const *)arg1)->GetLength();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_GetNbrAtomIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBond_GetNbrAtomIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_GetNbrAtomIdx" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBond_GetNbrAtomIdx" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (unsigned int)(arg1)->GetNbrAtomIdx(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_FindSmallestRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBRing *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_FindSmallestRing" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1);
  result = (OpenBabel::OBRing *)((OpenBabel::OBBond const *)arg1)->FindSmallestRing();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsAromatic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsAromatic" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)((OpenBabel::OBBond const *)arg1)->IsAromatic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsInRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsInRing" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)((OpenBabel::OBBond const *)arg1)->IsInRing();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsRotor__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsRotor" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBond_IsRotor" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (bool)(arg1)->IsRotor(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsRotor__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsRotor" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsRotor();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsRotor(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBond_IsRotor", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBond_IsRotor__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBond_IsRotor__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBond_IsRotor'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBond::IsRotor(bool)\n" "    OpenBabel::OBBond::IsRotor()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBBond_IsPeriodic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsPeriodic" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)((OpenBabel::OBBond const *)arg1)->IsPeriodic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsAmide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsAmide" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsAmide();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsPrimaryAmide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsPrimaryAmide" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsPrimaryAmide();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsSecondaryAmide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsSecondaryAmide" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsSecondaryAmide();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsTertiaryAmide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsTertiaryAmide" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsTertiaryAmide();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsEster(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsEster" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsEster();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsCarbonyl(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsCarbonyl" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsCarbonyl();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsClosure(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsClosure" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsClosure();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsWedge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsWedge" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsWedge();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsHash(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsHash" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsHash();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsWedgeOrHash(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsWedgeOrHash" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)((OpenBabel::OBBond const *)arg1)->IsWedgeOrHash();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsCisOrTrans(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsCisOrTrans" "', argument " "1"" of type '" "OpenBabel::OBBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)((OpenBabel::OBBond const *)arg1)->IsCisOrTrans();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBond_IsDoubleBondGeometry(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBond_IsDoubleBondGeometry" "', argument " "1"" of type '" "OpenBabel::OBBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBond * >(argp1); result = (bool)(arg1)->IsDoubleBondGeometry();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBBond_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBBond, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBBond_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBReaction(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBReaction", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBReaction *)new OpenBabel::OBReaction();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBReaction, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_NumReactants(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_NumReactants" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1);
  result = (int)((OpenBabel::OBReaction const *)arg1)->NumReactants(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_NumProducts(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_NumProducts" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1);
  result = (int)((OpenBabel::OBReaction const *)arg1)->NumProducts(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_NumAgents(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_NumAgents" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); result = (int)((OpenBabel::OBReaction const *)arg1)->NumAgents();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_AddReactant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_AddReactant", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_AddReactant" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_AddReactant" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_AddReactant" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'"); }
     else { std::shared_ptr< OpenBabel::OBMol > * temp = reinterpret_cast< std::shared_ptr< OpenBabel::OBMol > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->AddReactant(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_AddProduct(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_AddProduct", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_AddProduct" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_AddProduct" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_AddProduct" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'"); }
     else { std::shared_ptr< OpenBabel::OBMol > * temp = reinterpret_cast< std::shared_ptr< OpenBabel::OBMol > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->AddProduct(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetTransitionState(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_SetTransitionState", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_SetTransitionState" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_SetTransitionState" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_SetTransitionState" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'"); }
     else { std::shared_ptr< OpenBabel::OBMol > * temp = reinterpret_cast< std::shared_ptr< OpenBabel::OBMol > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->SetTransitionState(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_AddAgent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_AddAgent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_AddAgent" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_AddAgent" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_AddAgent" "', argument " "2"" of type '" "std::shared_ptr< OpenBabel::OBMol > const""'"); }
     else { std::shared_ptr< OpenBabel::OBMol > * temp = reinterpret_cast< std::shared_ptr< OpenBabel::OBMol > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  (arg1)->AddAgent(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetReactant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_GetReactant", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetReactant" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReaction_GetReactant" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (arg1)->GetReactant(arg2);
  resultobj = SWIG_NewPointerObj((new std::shared_ptr< OpenBabel::OBMol >(result)), SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetProduct(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_GetProduct", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetProduct" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReaction_GetProduct" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (arg1)->GetProduct(arg2);
  resultobj = SWIG_NewPointerObj((new std::shared_ptr< OpenBabel::OBMol >(result)), SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetAgent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReaction_GetAgent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetAgent" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReaction_GetAgent" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (arg1)->GetAgent(arg2);
  resultobj = SWIG_NewPointerObj((new std::shared_ptr< OpenBabel::OBMol >(result)), SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetTransitionState(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::shared_ptr< OpenBabel::OBMol > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetTransitionState" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1);
  result = ((OpenBabel::OBReaction const *)arg1)->GetTransitionState();
  resultobj = SWIG_NewPointerObj((new std::shared_ptr< OpenBabel::OBMol >(result)), SWIGTYPE_p_std__shared_ptrT_OpenBabel__OBMol_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetTitle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); result = ((OpenBabel::OBReaction const *)arg1)->GetTitle();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_GetComment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_GetComment" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); result = ((OpenBabel::OBReaction const *)arg1)->GetComment();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetTitle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBReaction_SetTitle", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_SetTitle" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_SetTitle" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_SetTitle" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetTitle((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetComment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBReaction_SetComment", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_SetComment" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReaction_SetComment" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBReaction_SetComment" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetComment((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_IsReversible(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_IsReversible" "', argument " "1"" of type '" "OpenBabel::OBReaction const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1);
  result = (bool)((OpenBabel::OBReaction const *)arg1)->IsReversible(); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetReversible__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBReaction *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_SetReversible" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReaction_SetReversible" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetReversible(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetReversible__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReaction_SetReversible" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); (arg1)->SetReversible(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReaction_SetReversible(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBReaction_SetReversible", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBReaction_SetReversible__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBReaction_SetReversible__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBReaction_SetReversible'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBReaction::SetReversible(bool)\n"
  "    OpenBabel::OBReaction::SetReversible()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBReaction_ClassDescription(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  char *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBReaction_ClassDescription", 0, 0, 0)) SWIG_fail;
  result = (char *)OpenBabel::OBReaction::ClassDescription(); resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBReaction(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReaction *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReaction, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBReaction" "', argument " "1"" of type '" "OpenBabel::OBReaction *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReaction * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBReaction_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBReaction, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBReaction_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBReactionFacade(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBReactionFacade *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBReactionFacade" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBReactionFacade *)new OpenBabel::OBReactionFacade(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBReactionFacade, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBReactionFacade(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBReactionFacade" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_AssignComponentIds__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBReactionFacade *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_AssignComponentIds" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReactionFacade_AssignComponentIds" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->AssignComponentIds(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_AssignComponentIds__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBReactionFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_AssignComponentIds" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); (arg1)->AssignComponentIds(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_AssignComponentIds(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBReactionFacade_AssignComponentIds", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBReactionFacade_AssignComponentIds__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBReactionFacade_AssignComponentIds__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBReactionFacade_AssignComponentIds'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBReactionFacade::AssignComponentIds(bool)\n"
  "    OpenBabel::OBReactionFacade::AssignComponentIds()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_GetRole(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBReactionRole result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_GetRole", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_GetRole" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_GetRole" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (OpenBabel::OBReactionRole)(arg1)->GetRole(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_GetComponentId(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_GetComponentId", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_GetComponentId" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_GetComponentId" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (unsigned int)(arg1)->GetComponentId(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_SetRole(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBReactionRole arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_SetRole", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_SetRole" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_SetRole" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBReactionFacade_SetRole" "', argument " "3"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg3 = static_cast< OpenBabel::OBReactionRole >(val3); (arg1)->SetRole(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_SetComponentId(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_SetComponentId", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_SetComponentId" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_SetComponentId" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBReactionFacade_SetComponentId" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetComponentId(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_IsValid" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); result = (bool)(arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_AddComponent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBReactionRole arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_AddComponent", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_AddComponent" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_AddComponent" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBReactionFacade_AddComponent" "', argument " "3"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg3 = static_cast< OpenBabel::OBReactionRole >(val3); (arg1)->AddComponent(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_ClearInternalState(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_ClearInternalState" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); (arg1)->ClearInternalState(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_GetComponent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBReactionRole arg3 ; unsigned int arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; unsigned int val4 ;
  int ecode4 = 0 ; PyObject *swig_obj[4] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_GetComponent", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_GetComponent" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBReactionFacade_GetComponent" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBReactionFacade_GetComponent" "', argument " "3"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg3 = static_cast< OpenBabel::OBReactionRole >(val3); ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBReactionFacade_GetComponent" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); result = (bool)(arg1)->GetComponent(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_NumComponents(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBReactionRole arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; unsigned int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_NumComponents", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_NumComponents" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReactionFacade_NumComponents" "', argument " "2"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg2 = static_cast< OpenBabel::OBReactionRole >(val2); result = (unsigned int)(arg1)->NumComponents(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBReactionFacade_ReassignComponent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBReactionFacade *arg1 = 0 ; OpenBabel::OBReactionRole arg2 ; unsigned int arg3 ; OpenBabel::OBReactionRole arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ;
  PyObject *swig_obj[4] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBReactionFacade_ReassignComponent", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBReactionFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBReactionFacade_ReassignComponent" "', argument " "1"" of type '" "OpenBabel::OBReactionFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBReactionFacade * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBReactionFacade_ReassignComponent" "', argument " "2"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg2 = static_cast< OpenBabel::OBReactionRole >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBReactionFacade_ReassignComponent" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBReactionFacade_ReassignComponent" "', argument " "4"" of type '" "OpenBabel::OBReactionRole""'"); }
    arg4 = static_cast< OpenBabel::OBReactionRole >(val4); result = (bool)(arg1)->ReassignComponent(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBReactionFacade_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBReactionFacade, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBReactionFacade_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMol__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMol *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMol *)new OpenBabel::OBMol();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMol__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMol *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBMol" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBMol" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMol *)new OpenBabel::OBMol((OpenBabel::OBMol const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMol(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBMol", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBMol__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_new_OBMol__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBMol'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::OBMol()\n"
  "    OpenBabel::OBMol::OBMol(OpenBabel::OBMol const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBMol(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMol" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol___iadd__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBMol *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol___iadd__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol___iadd__" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol___iadd__" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol___iadd__" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  result = (OpenBabel::OBMol *) &(arg1)->operator +=((OpenBabel::OBMol const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ReserveAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_ReserveAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ReserveAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_ReserveAtoms" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->ReserveAtoms(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DestroyAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_DestroyAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DestroyAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DestroyAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); (arg1)->DestroyAtom(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DestroyBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_DestroyBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DestroyBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DestroyBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); (arg1)->DestroyBond(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DestroyResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_DestroyResidue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DestroyResidue" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DestroyResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); (arg1)->DestroyResidue(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAtom,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddAtom" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->AddAtom(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAtom,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->AddAtom(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_AddAtom", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_AddAtom__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_OBMol_AddAtom__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_AddAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::AddAtom(OpenBabel::OBAtom &,bool)\n"
  "    OpenBabel::OBMol::AddAtom(OpenBabel::OBAtom &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_InsertAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_InsertAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_InsertAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBAtom,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_InsertAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_InsertAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->InsertAtom(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; int arg6 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ;
  int val6 ; int ecode6 = 0 ; bool result; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddBond" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_AddBond" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_AddBond" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_int(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBMol_AddBond" "', argument " "6"" of type '" "int""'"); }  
  arg6 = static_cast< int >(val6); result = (bool)(arg1)->AddBond(arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ;
  bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddBond" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_AddBond" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_AddBond" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); result = (bool)(arg1)->AddBond(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddBond" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_AddBond" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); result = (bool)(arg1)->AddBond(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddBond__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBond,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddBond" "', argument " "2"" of type '" "OpenBabel::OBBond &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_AddBond" "', argument " "2"" of type '" "OpenBabel::OBBond &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->AddBond(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_AddBond", 0, 6, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_AddBond__SWIG_3(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 4) { PyObject *retobj = _wrap_OBMol_AddBond__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBMol_AddBond__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 6) { PyObject *retobj = _wrap_OBMol_AddBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_AddBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::AddBond(int,int,int,int,int)\n"
  "    OpenBabel::OBMol::AddBond(int,int,int,int)\n" "    OpenBabel::OBMol::AddBond(int,int,int)\n"
  "    OpenBabel::OBMol::AddBond(OpenBabel::OBBond &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_AddResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_AddResidue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddResidue" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBResidue,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_AddResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); result = (bool)(arg1)->AddResidue(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NewAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NewAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (OpenBabel::OBAtom *)(arg1)->NewAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NewAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBAtom *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NewAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_NewAtom" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBAtom *)(arg1)->NewAtom(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NewAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_NewAtom", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_NewAtom__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_OBMol_NewAtom__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_NewAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::NewAtom()\n" "    OpenBabel::OBMol::NewAtom(unsigned long)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_NewBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBBond *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NewBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (OpenBabel::OBBond *)(arg1)->NewBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NewBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBBond *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NewBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_NewBond" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBBond *)(arg1)->NewBond(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NewBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_NewBond", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_NewBond__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_OBMol_NewBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_NewBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::NewBond()\n" "    OpenBabel::OBMol::NewBond(unsigned long)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_NewResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBResidue *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NewResidue" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (OpenBabel::OBResidue *)(arg1)->NewResidue();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_DeleteAtom" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->DeleteAtom(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteAtom" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteAtom" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->DeleteAtom(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_DeleteAtom", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_DeleteAtom__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_DeleteAtom__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_DeleteAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::DeleteAtom(OpenBabel::OBAtom *,bool)\n"
  "    OpenBabel::OBMol::DeleteAtom(OpenBabel::OBAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_DeleteBond" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->DeleteBond(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteBond" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->DeleteBond(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_DeleteBond", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_DeleteBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_DeleteBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_DeleteBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::DeleteBond(OpenBabel::OBBond *,bool)\n"
  "    OpenBabel::OBMol::DeleteBond(OpenBabel::OBBond *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteResidue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteResidue" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_DeleteResidue" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->DeleteResidue(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteResidue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBResidue *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteResidue" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteResidue" "', argument " "2"" of type '" "OpenBabel::OBResidue *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBResidue * >(argp2); result = (bool)(arg1)->DeleteResidue(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteResidue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_DeleteResidue", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_DeleteResidue__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_DeleteResidue__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_DeleteResidue'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::DeleteResidue(OpenBabel::OBResidue *,bool)\n"
  "    OpenBabel::OBMol::DeleteResidue(OpenBabel::OBResidue *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_BeginModify(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_BeginModify" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->BeginModify(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_EndModify__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_EndModify" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_EndModify" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); (arg1)->EndModify(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_EndModify__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_EndModify" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->EndModify(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_EndModify(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_EndModify", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_EndModify__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_EndModify__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_EndModify'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::EndModify(bool)\n" "    OpenBabel::OBMol::EndModify()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetMod(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetMod" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (int)(arg1)->GetMod();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_IncrementMod(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_IncrementMod" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->IncrementMod(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DecrementMod(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DecrementMod" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->DecrementMod(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetFlags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetFlags" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (int)((OpenBabel::OBMol const *)arg1)->GetFlags();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTitle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; char *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetTitle" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (char *)((OpenBabel::OBMol const *)arg1)->GetTitle(arg2);
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTitle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; char *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTitle" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (char *)((OpenBabel::OBMol const *)arg1)->GetTitle();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTitle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetTitle", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_GetTitle__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetTitle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetTitle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetTitle(bool) const\n"
  "    OpenBabel::OBMol::GetTitle() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_NumAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)((OpenBabel::OBMol const *)arg1)->NumAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumBonds" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)((OpenBabel::OBMol const *)arg1)->NumBonds();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumHvyAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumHvyAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)((OpenBabel::OBMol const *)arg1)->NumHvyAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumResidues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumResidues" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)((OpenBabel::OBMol const *)arg1)->NumResidues();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumRotors__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; unsigned int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumRotors" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_NumRotors" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (unsigned int)(arg1)->NumRotors(arg2);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumRotors__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumRotors" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)(arg1)->NumRotors();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumRotors(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_NumRotors", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_NumRotors__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_NumRotors__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_NumRotors'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::NumRotors(bool)\n" "    OpenBabel::OBMol::NumRotors()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_GetAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetAtom" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetAtom" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (OpenBabel::OBAtom *)((OpenBabel::OBMol const *)arg1)->GetAtom(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetAtomById(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBAtom *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetAtomById", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetAtomById" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetAtomById" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBMol const *)arg1)->GetAtomById(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetFirstAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetFirstAtom" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBAtom *)((OpenBabel::OBMol const *)arg1)->GetFirstAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::OBBond *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetBond" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetBond" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (OpenBabel::OBBond *)((OpenBabel::OBMol const *)arg1)->GetBond(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBondById(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::OBBond *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetBondById", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetBondById" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetBondById" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (OpenBabel::OBBond *)((OpenBabel::OBMol const *)arg1)->GetBondById(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::OBBond *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetBond" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetBond" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_GetBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (OpenBabel::OBBond *)((OpenBabel::OBMol const *)arg1)->GetBond(arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  OpenBabel::OBBond *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetBond" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetBond" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetBond" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  result = (OpenBabel::OBBond *)((OpenBabel::OBMol const *)arg1)->GetBond(arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetBond", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetBond__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_OBMol_GetBond__SWIG_2(self, argc, argv);}  check_2:
  if (argc == 3) { PyObject *retobj = _wrap_OBMol_GetBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetBond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetBond(int) const\n"
  "    OpenBabel::OBMol::GetBond(int,int) const\n"
  "    OpenBabel::OBMol::GetBond(OpenBabel::OBAtom *,OpenBabel::OBAtom *) const\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetResidue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBResidue *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetResidue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetResidue" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetResidue" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (OpenBabel::OBResidue *)((OpenBabel::OBMol const *)arg1)->GetResidue(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (arg1)->GetInternalCoord();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTorsion__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ;
  double result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTorsion" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetTorsion" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_GetTorsion" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_GetTorsion" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_GetTorsion" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); result = (double)(arg1)->GetTorsion(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTorsion__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  OpenBabel::OBAtom *arg4 = 0 ; OpenBabel::OBAtom *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; double result; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTorsion" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetTorsion" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetTorsion" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_GetTorsion" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBMol_GetTorsion" "', argument " "5"" of type '" "OpenBabel::OBAtom *""'");  }
   arg5 = reinterpret_cast< OpenBabel::OBAtom * >(argp5); result = (double)(arg1)->GetTorsion(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTorsion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetTorsion", 0, 5, argv))) SWIG_fail; --argc; if (argc == 5) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBMol_GetTorsion__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 5) { PyObject *retobj = _wrap_OBMol_GetTorsion__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetTorsion'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetTorsion(int,int,int,int)\n"
  "    OpenBabel::OBMol::GetTorsion(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ; OpenBabel::OBAtom *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetAngle", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetAngle" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetAngle" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetAngle" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_GetAngle" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4); result = (double)(arg1)->GetAngle(arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AreInSameRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_AreInSameRing", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AreInSameRing" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AreInSameRing" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_AreInSameRing" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3); result = (int)(arg1)->AreInSameRing(arg2,arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetFormula(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (arg1)->GetFormula();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSpacedFormula__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; char *arg3 = 0 ; bool arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool val4 ; int ecode4 = 0 ; std::string result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetSpacedFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetSpacedFormula" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetSpacedFormula" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_GetSpacedFormula" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (arg1)->GetSpacedFormula(arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj; fail: if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSpacedFormula__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; char *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; std::string result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetSpacedFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetSpacedFormula" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetSpacedFormula" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (arg1)->GetSpacedFormula(arg2,(char const *)arg3);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj; fail: if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSpacedFormula__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  std::string result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetSpacedFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetSpacedFormula" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->GetSpacedFormula(arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSpacedFormula__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; std::string result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetSpacedFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (arg1)->GetSpacedFormula();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSpacedFormula(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetSpacedFormula", 0, 4, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_GetSpacedFormula__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetSpacedFormula__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_GetSpacedFormula__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBMol_GetSpacedFormula__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetSpacedFormula'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetSpacedFormula(int,char const *,bool)\n"
  "    OpenBabel::OBMol::GetSpacedFormula(int,char const *)\n" "    OpenBabel::OBMol::GetSpacedFormula(int)\n"
  "    OpenBabel::OBMol::GetSpacedFormula()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetEnergy__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetEnergy" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (double)((OpenBabel::OBMol const *)arg1)->GetEnergy();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetMolWt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetMolWt" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetMolWt" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (double)(arg1)->GetMolWt(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetMolWt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetMolWt" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (double)(arg1)->GetMolWt();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetMolWt(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetMolWt", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_GetMolWt__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetMolWt__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetMolWt'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetMolWt(bool)\n" "    OpenBabel::OBMol::GetMolWt()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetExactMass__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetExactMass" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetExactMass" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (double)(arg1)->GetExactMass(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetExactMass__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetExactMass" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (double)(arg1)->GetExactMass();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetExactMass(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetExactMass", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_GetExactMass__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetExactMass__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetExactMass'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetExactMass(bool)\n" "    OpenBabel::OBMol::GetExactMass()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetTotalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTotalCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (int)(arg1)->GetTotalCharge();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetTotalSpinMultiplicity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetTotalSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (unsigned int)(arg1)->GetTotalSpinMultiplicity();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetDimension(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned short result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetDimension" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (unsigned short)((OpenBabel::OBMol const *)arg1)->GetDimension();
  resultobj = SWIG_From_unsigned_SS_short(static_cast< unsigned short >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetCoordinates" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (double *)(arg1)->GetCoordinates();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetSSSR(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetSSSR" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *) &(arg1)->GetSSSR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetLSSR(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetLSSR" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *) &(arg1)->GetLSSR();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBRing_p_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AutomaticFormalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AutomaticFormalCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AutomaticFormalCharge();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AutomaticPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AutomaticPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AutomaticPartialCharge();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTitle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTitle" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetTitle" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetTitle((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTitle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTitle" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetTitle" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_SetTitle" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); (arg1)->SetTitle(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTitle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetTitle", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__string, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBMol_SetTitle__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_OBMol_SetTitle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetTitle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetTitle(char const *)\n"
  "    OpenBabel::OBMol::SetTitle(std::string &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetFormula(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetFormula", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetFormula" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBMol_SetFormula" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetFormula(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetEnergy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetEnergy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetEnergy" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetEnergy" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetEnergy(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetDimension(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; unsigned short arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned short val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetDimension", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetDimension" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_short(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetDimension" "', argument " "2"" of type '" "unsigned short""'"); }
    arg2 = static_cast< unsigned short >(val2); (arg1)->SetDimension(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTotalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetTotalCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTotalCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetTotalCharge" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetTotalCharge(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTotalSpinMultiplicity(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetTotalSpinMultiplicity", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTotalSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetTotalSpinMultiplicity" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetTotalSpinMultiplicity(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetInternalCoord", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *ptr = (std::vector< OpenBabel::OBInternalCoord*,std::allocator< OpenBabel::OBInternalCoord * > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBMol_SetInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetInternalCoord(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAutomaticFormalCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetAutomaticFormalCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAutomaticFormalCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetAutomaticFormalCharge" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetAutomaticFormalCharge(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAutomaticPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetAutomaticPartialCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAutomaticPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetAutomaticPartialCharge" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetAutomaticPartialCharge(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAromaticPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAromaticPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetAromaticPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetAromaticPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAromaticPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAromaticPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetAromaticPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAromaticPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetAromaticPerceived", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetAromaticPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetAromaticPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetAromaticPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetAromaticPerceived(bool)\n"
  "    OpenBabel::OBMol::SetAromaticPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetSSSRPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetSSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetSSSRPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetSSSRPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetSSSRPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetSSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetSSSRPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetSSSRPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetSSSRPerceived", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetSSSRPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetSSSRPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetSSSRPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetSSSRPerceived(bool)\n"
  "    OpenBabel::OBMol::SetSSSRPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetLSSRPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetLSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetLSSRPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetLSSRPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetLSSRPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetLSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetLSSRPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetLSSRPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetLSSRPerceived", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetLSSRPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetLSSRPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetLSSRPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetLSSRPerceived(bool)\n"
  "    OpenBabel::OBMol::SetLSSRPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingAtomsAndBondsPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetRingAtomsAndBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetRingAtomsAndBondsPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetRingAtomsAndBondsPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingAtomsAndBondsPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetRingAtomsAndBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetRingAtomsAndBondsPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingAtomsAndBondsPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetRingAtomsAndBondsPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetRingAtomsAndBondsPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetRingAtomsAndBondsPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetRingAtomsAndBondsPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetRingAtomsAndBondsPerceived(bool)\n"
  "    OpenBabel::OBMol::SetRingAtomsAndBondsPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetAtomTypesPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAtomTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetAtomTypesPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetAtomTypesPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAtomTypesPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetAtomTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetAtomTypesPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetAtomTypesPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetAtomTypesPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetAtomTypesPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetAtomTypesPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetAtomTypesPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetAtomTypesPerceived(bool)\n"
  "    OpenBabel::OBMol::SetAtomTypesPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingTypesPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetRingTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetRingTypesPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetRingTypesPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingTypesPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetRingTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetRingTypesPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetRingTypesPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetRingTypesPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetRingTypesPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetRingTypesPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetRingTypesPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetRingTypesPerceived(bool)\n"
  "    OpenBabel::OBMol::SetRingTypesPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetChainsPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetChainsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetChainsPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetChainsPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetChainsPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetChainsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetChainsPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetChainsPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetChainsPerceived", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetChainsPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetChainsPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetChainsPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetChainsPerceived(bool)\n"
  "    OpenBabel::OBMol::SetChainsPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetChiralityPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetChiralityPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetChiralityPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetChiralityPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetChiralityPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetChiralityPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetChiralityPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetChiralityPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetChiralityPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetChiralityPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetChiralityPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetChiralityPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetChiralityPerceived(bool)\n"
  "    OpenBabel::OBMol::SetChiralityPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetPartialChargesPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetPartialChargesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetPartialChargesPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetPartialChargesPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetPartialChargesPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetPartialChargesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetPartialChargesPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetPartialChargesPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetPartialChargesPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetPartialChargesPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetPartialChargesPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetPartialChargesPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetPartialChargesPerceived(bool)\n"
  "    OpenBabel::OBMol::SetPartialChargesPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetHybridizationPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetHybridizationPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetHybridizationPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetHybridizationPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetHybridizationPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetHybridizationPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetHybridizationPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetHybridizationPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetHybridizationPerceived", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetHybridizationPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetHybridizationPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetHybridizationPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetHybridizationPerceived(bool)\n"
  "    OpenBabel::OBMol::SetHybridizationPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetClosureBondsPerceived__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetClosureBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetClosureBondsPerceived" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetClosureBondsPerceived(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetClosureBondsPerceived__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetClosureBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetClosureBondsPerceived(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetClosureBondsPerceived(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetClosureBondsPerceived", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetClosureBondsPerceived__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetClosureBondsPerceived__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetClosureBondsPerceived'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetClosureBondsPerceived(bool)\n"
  "    OpenBabel::OBMol::SetClosureBondsPerceived()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetHydrogensAdded__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetHydrogensAdded" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetHydrogensAdded" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetHydrogensAdded(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetHydrogensAdded__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetHydrogensAdded" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetHydrogensAdded(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetHydrogensAdded(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetHydrogensAdded", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetHydrogensAdded__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetHydrogensAdded__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetHydrogensAdded'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetHydrogensAdded(bool)\n"
  "    OpenBabel::OBMol::SetHydrogensAdded()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetCorrectedForPH__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetCorrectedForPH" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetCorrectedForPH" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetCorrectedForPH(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetCorrectedForPH__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetCorrectedForPH" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetCorrectedForPH(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetCorrectedForPH(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetCorrectedForPH", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetCorrectedForPH__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetCorrectedForPH__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetCorrectedForPH'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetCorrectedForPH(bool)\n"
  "    OpenBabel::OBMol::SetCorrectedForPH()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetSpinMultiplicityAssigned__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetSpinMultiplicityAssigned" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetSpinMultiplicityAssigned" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetSpinMultiplicityAssigned(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetSpinMultiplicityAssigned__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetSpinMultiplicityAssigned" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetSpinMultiplicityAssigned(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetSpinMultiplicityAssigned(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetSpinMultiplicityAssigned", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetSpinMultiplicityAssigned__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetSpinMultiplicityAssigned__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetSpinMultiplicityAssigned'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetSpinMultiplicityAssigned(bool)\n"
  "    OpenBabel::OBMol::SetSpinMultiplicityAssigned()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsPatternStructure__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetIsPatternStructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetIsPatternStructure" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetIsPatternStructure(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsPatternStructure__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetIsPatternStructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetIsPatternStructure(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsPatternStructure(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetIsPatternStructure", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_SetIsPatternStructure__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetIsPatternStructure__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetIsPatternStructure'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetIsPatternStructure(bool)\n"
  "    OpenBabel::OBMol::SetIsPatternStructure()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsReaction__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetIsReaction" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetIsReaction" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetIsReaction(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsReaction__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetIsReaction" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetIsReaction(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetIsReaction(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetIsReaction", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetIsReaction__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetIsReaction__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetIsReaction'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetIsReaction(bool)\n" "    OpenBabel::OBMol::SetIsReaction()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetPeriodicMol__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetPeriodicMol" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetPeriodicMol" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetPeriodicMol(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetPeriodicMol__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetPeriodicMol" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->SetPeriodicMol(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetPeriodicMol(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetPeriodicMol", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_SetPeriodicMol__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_SetPeriodicMol__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetPeriodicMol'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::SetPeriodicMol(bool)\n"
  "    OpenBabel::OBMol::SetPeriodicMol()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_HasFlag(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_HasFlag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasFlag" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_HasFlag" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (bool)(arg1)->HasFlag(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetFlag(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetFlag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetFlag" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetFlag" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetFlag(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_UnsetFlag(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_UnsetFlag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_UnsetFlag" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_UnsetFlag" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->UnsetFlag(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetFlags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetFlags", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetFlags" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetFlags" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetFlags(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ClassDescription(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *result = 0 ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_ClassDescription", 0, 0, 0)) SWIG_fail;
  result = (char *)OpenBabel::OBMol::ClassDescription(); resultobj = SWIG_FromCharPtr((const char *)result); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_RenumberAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_RenumberAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_RenumberAtoms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_RenumberAtoms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > * >(argp2);
  (arg1)->RenumberAtoms(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_RenumberAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< int,std::allocator< int > > arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_RenumberAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBMol_RenumberAtoms" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->RenumberAtoms(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_RenumberAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_RenumberAtoms", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBMol_RenumberAtoms__SWIG_0(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_OBMol_RenumberAtoms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_RenumberAtoms'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBMol::RenumberAtoms(std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > &)\n"
  "    OpenBabel::OBMol::RenumberAtoms(std::vector< int,std::allocator< int > >)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetCoordinates", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetCoordinates" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetCoordinates" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->SetCoordinates(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ToInertialFrame__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ToInertialFrame" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_ToInertialFrame" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_ToInertialFrame" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); (arg1)->ToInertialFrame(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ToInertialFrame__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ToInertialFrame" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->ToInertialFrame(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ToInertialFrame(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_ToInertialFrame", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_ToInertialFrame__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_ToInertialFrame__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_ToInertialFrame'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::ToInertialFrame(int,double *)\n"
  "    OpenBabel::OBMol::ToInertialFrame()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_Translate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Translate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Translate" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_Translate" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); (arg1)->Translate((OpenBabel::vector3 const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Translate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Translate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Translate" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_Translate" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_Translate" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->Translate((OpenBabel::vector3 const &)*arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Translate(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_Translate", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_Translate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_Translate__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_Translate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::Translate(OpenBabel::vector3 const &)\n"
  "    OpenBabel::OBMol::Translate(OpenBabel::vector3 const &,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_Rotate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; double (*arg2)[3] = (double (*)[3])0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Rotate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_a_3__double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Rotate" "', argument " "2"" of type '" "double const [3][3]""'");  }
    arg2 = reinterpret_cast< double (*)[3] >(argp2); (arg1)->Rotate((double const (*)[3])arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Rotate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Rotate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Rotate" "', argument " "2"" of type '" "double const [9]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); (arg1)->Rotate((double const (*))arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Rotate__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; double *arg2 = (double *)0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Rotate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Rotate" "', argument " "2"" of type '" "double const [9]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_Rotate" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->Rotate((double const (*))arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Rotate(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_Rotate", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_a_3__double, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBMol_Rotate__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_Rotate__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_OBMol_Rotate__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_Rotate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::Rotate(double const [3][3])\n"
  "    OpenBabel::OBMol::Rotate(double const [9])\n" "    OpenBabel::OBMol::Rotate(double const [9],int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_Center__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Center" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->Center(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteHydrogens__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->DeleteHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteHydrogens__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteHydrogens" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->DeleteHydrogens(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteHydrogens(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_DeleteHydrogens", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_DeleteHydrogens__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_DeleteHydrogens__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_DeleteHydrogens'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::DeleteHydrogens()\n"
  "    OpenBabel::OBMol::DeleteHydrogens(OpenBabel::OBAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_DeletePolarHydrogens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeletePolarHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->DeletePolarHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteNonPolarHydrogens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteNonPolarHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->DeleteNonPolarHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteHydrogen(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_DeleteHydrogen", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteHydrogen" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_DeleteHydrogen" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->DeleteHydrogen(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; bool arg3 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  bool val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddHydrogens" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddHydrogens" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_AddHydrogens" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); result = (bool)(arg1)->AddHydrogens(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddHydrogens" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddHydrogens" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->AddHydrogens(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddHydrogens" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (bool)(arg1)->AddHydrogens(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AddHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddHydrogens" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->AddHydrogens(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddHydrogens(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_AddHydrogens", 0, 4, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_AddHydrogens__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_OBMol_AddHydrogens__SWIG_4(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_AddHydrogens__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_AddHydrogens__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBMol_AddHydrogens__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_AddHydrogens'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::AddHydrogens(bool,bool,double)\n"
  "    OpenBabel::OBMol::AddHydrogens(bool,bool)\n" "    OpenBabel::OBMol::AddHydrogens(bool)\n"
  "    OpenBabel::OBMol::AddHydrogens()\n" "    OpenBabel::OBMol::AddHydrogens(OpenBabel::OBAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_AddPolarHydrogens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddPolarHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AddPolarHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddNonPolarHydrogens(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddNonPolarHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AddNonPolarHydrogens();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddNewHydrogens__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::HydrogenType arg2 ; bool arg3 ; double arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ;
  bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddNewHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddNewHydrogens" "', argument " "2"" of type '" "OpenBabel::HydrogenType""'"); }
    arg2 = static_cast< OpenBabel::HydrogenType >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddNewHydrogens" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_AddNewHydrogens" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); result = (bool)(arg1)->AddNewHydrogens(arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddNewHydrogens__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::HydrogenType arg2 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddNewHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddNewHydrogens" "', argument " "2"" of type '" "OpenBabel::HydrogenType""'"); }
    arg2 = static_cast< OpenBabel::HydrogenType >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_AddNewHydrogens" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->AddNewHydrogens(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddNewHydrogens__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::HydrogenType arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddNewHydrogens" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AddNewHydrogens" "', argument " "2"" of type '" "OpenBabel::HydrogenType""'"); }
    arg2 = static_cast< OpenBabel::HydrogenType >(val2); result = (bool)(arg1)->AddNewHydrogens(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddNewHydrogens(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_AddNewHydrogens", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_AddNewHydrogens__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_AddNewHydrogens__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBMol_AddNewHydrogens__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_AddNewHydrogens'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::AddNewHydrogens(OpenBabel::HydrogenType,bool,double)\n"
  "    OpenBabel::OBMol::AddNewHydrogens(OpenBabel::HydrogenType,bool)\n"
  "    OpenBabel::OBMol::AddNewHydrogens(OpenBabel::HydrogenType)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_StripSalts__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_StripSalts" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_StripSalts" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->StripSalts(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_StripSalts__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_StripSalts" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->StripSalts();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_StripSalts(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_StripSalts", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_StripSalts__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_StripSalts__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_StripSalts'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::StripSalts(unsigned int)\n"
  "    OpenBabel::OBMol::StripSalts()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_Separate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Separate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_Separate" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->Separate(arg2);
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Separate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Separate" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (arg1)->Separate();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Separate(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_Separate", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_Separate__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_Separate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_Separate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::Separate(int)\n" "    OpenBabel::OBMol::Separate()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_GetNextFragment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMolAtomDFSIter *arg2 = 0 ; OpenBabel::OBMol *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_GetNextFragment", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetNextFragment" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetNextFragment" "', argument " "2"" of type '" "OpenBabel::OBMolAtomDFSIter &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_GetNextFragment" "', argument " "2"" of type '" "OpenBabel::OBMolAtomDFSIter &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_GetNextFragment" "', argument " "3"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_GetNextFragment" "', argument " "3"" of type '" "OpenBabel::OBMol &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBMol * >(argp3); result = (bool)(arg1)->GetNextFragment(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  OpenBabel::OBBitVec *arg4 = 0 ; unsigned int arg5 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg6 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg7 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; unsigned int val5 ; int ecode5 = 0 ;
  void *argp6 = 0 ; int res6 = 0 ; void *argp7 = 0 ; int res7 = 0 ; bool result; (void)self;
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopySubstructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_CopySubstructure" "', argument " "3"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_CopySubstructure" "', argument " "4"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_CopySubstructure" "', argument " "5"" of type '" "unsigned int""'"); }
    arg5 = static_cast< unsigned int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBMol_CopySubstructure" "', argument " "6"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > *""'");  }
   arg6 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp6);
  res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "OBMol_CopySubstructure" "', argument " "7"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > *""'");  }
   arg7 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp7);
  result = (bool)(arg1)->CopySubstructure(*arg2,arg3,arg4,arg5,arg6,arg7);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  OpenBabel::OBBitVec *arg4 = 0 ; unsigned int arg5 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg6 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; unsigned int val5 ; int ecode5 = 0 ; void *argp6 = 0 ; int res6 = 0 ; bool result; (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopySubstructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_CopySubstructure" "', argument " "3"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_CopySubstructure" "', argument " "4"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_CopySubstructure" "', argument " "5"" of type '" "unsigned int""'"); }
    arg5 = static_cast< unsigned int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_std__vectorT_unsigned_int_t, 0 |  0 ); if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBMol_CopySubstructure" "', argument " "6"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > *""'");  }
   arg6 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp6);
  result = (bool)(arg1)->CopySubstructure(*arg2,arg3,arg4,arg5,arg6); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  OpenBabel::OBBitVec *arg4 = 0 ; unsigned int arg5 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; unsigned int val5 ; int ecode5 = 0 ; bool result;
  (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopySubstructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_CopySubstructure" "', argument " "3"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_CopySubstructure" "', argument " "4"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_CopySubstructure" "', argument " "5"" of type '" "unsigned int""'"); }
    arg5 = static_cast< unsigned int >(val5); result = (bool)(arg1)->CopySubstructure(*arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopySubstructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_CopySubstructure" "', argument " "3"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_CopySubstructure" "', argument " "4"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); result = (bool)(arg1)->CopySubstructure(*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopySubstructure" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_CopySubstructure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_CopySubstructure" "', argument " "3"" of type '" "OpenBabel::OBBitVec *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3); result = (bool)(arg1)->CopySubstructure(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopySubstructure(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[8] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_CopySubstructure", 0, 7, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBMol_CopySubstructure__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBMol_CopySubstructure__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBMol_CopySubstructure__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_OBMol_CopySubstructure__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 7) {
    PyObject *retobj = _wrap_OBMol_CopySubstructure__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_CopySubstructure'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBMol::CopySubstructure(OpenBabel::OBMol &,OpenBabel::OBBitVec *,OpenBabel::OBBitVec *,unsigned int,std::vector< unsigned int,std::allocator< unsigned int > > *,std::vector< unsigned int,std::allocator< unsigned int > > *)\n"
  "    OpenBabel::OBMol::CopySubstructure(OpenBabel::OBMol &,OpenBabel::OBBitVec *,OpenBabel::OBBitVec *,unsigned int,std::vector< unsigned int,std::allocator< unsigned int > > *)\n"
  "    OpenBabel::OBMol::CopySubstructure(OpenBabel::OBMol &,OpenBabel::OBBitVec *,OpenBabel::OBBitVec *,unsigned int)\n"
  "    OpenBabel::OBMol::CopySubstructure(OpenBabel::OBMol &,OpenBabel::OBBitVec *,OpenBabel::OBBitVec *)\n"
  "    OpenBabel::OBMol::CopySubstructure(OpenBabel::OBMol &,OpenBabel::OBBitVec *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_ConvertDativeBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ConvertDativeBonds" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->ConvertDativeBonds();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_MakeDativeBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_MakeDativeBonds" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->MakeDativeBonds();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ConvertZeroBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ConvertZeroBonds" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->ConvertZeroBonds();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CorrectForPH__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CorrectForPH" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_CorrectForPH" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); result = (bool)(arg1)->CorrectForPH(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CorrectForPH__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CorrectForPH" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->CorrectForPH();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CorrectForPH(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_CorrectForPH", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_CorrectForPH__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_CorrectForPH__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_CorrectForPH'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::CorrectForPH(double)\n" "    OpenBabel::OBMol::CorrectForPH()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_AssignSpinMultiplicity__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AssignSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AssignSpinMultiplicity" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (bool)(arg1)->AssignSpinMultiplicity(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AssignSpinMultiplicity__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AssignSpinMultiplicity" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->AssignSpinMultiplicity();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AssignSpinMultiplicity(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_AssignSpinMultiplicity", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBMol_AssignSpinMultiplicity__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_AssignSpinMultiplicity__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_AssignSpinMultiplicity'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::AssignSpinMultiplicity(bool)\n"
  "    OpenBabel::OBMol::AssignSpinMultiplicity()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_AssignTotalChargeToAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_AssignTotalChargeToAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AssignTotalChargeToAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_AssignTotalChargeToAtoms" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->AssignTotalChargeToAtoms(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Center__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Center" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_Center" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->Center(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Center(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_Center", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_Center__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_OBMol_Center__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_Center'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::Center()\n" "    OpenBabel::OBMol::Center(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_SetTorsion__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ;
  OpenBabel::OBAtom *arg4 = 0 ; OpenBabel::OBAtom *arg5 = 0 ; double arg6 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ;
  double val6 ; int ecode6 = 0 ; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTorsion" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetTorsion" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_SetTorsion" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_SetTorsion" "', argument " "4"" of type '" "OpenBabel::OBAtom *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBAtom * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBMol_SetTorsion" "', argument " "5"" of type '" "OpenBabel::OBAtom *""'");  }
   arg5 = reinterpret_cast< OpenBabel::OBAtom * >(argp5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBMol_SetTorsion" "', argument " "6"" of type '" "double""'"); }  
  arg6 = static_cast< double >(val6); (arg1)->SetTorsion(arg2,arg3,arg4,arg5,arg6); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindSSSR(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindSSSR" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->FindSSSR(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindLSSR(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindLSSR" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->FindLSSR(); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindRingAtomsAndBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindRingAtomsAndBonds" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->FindRingAtomsAndBonds(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindChildren(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; int arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ;
  PyObject *swig_obj[4] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_FindChildren", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindChildren" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_FindChildren" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_FindChildren" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_FindChildren" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_FindChildren" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); (arg1)->FindChildren(*arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindLargestFragment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_FindLargestFragment", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindLargestFragment" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_FindLargestFragment" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_FindLargestFragment" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->FindLargestFragment(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ContigFragList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_ContigFragList", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ContigFragList" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_ContigFragList" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_ContigFragList" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp2);
  (arg1)->ContigFragList(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Align(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::OBAtom *arg3 = 0 ; OpenBabel::vector3 *arg4 = 0 ; OpenBabel::vector3 *arg5 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; PyObject *swig_obj[5] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_Align", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Align" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_Align" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBMol_Align" "', argument " "3"" of type '" "OpenBabel::OBAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBAtom * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBMol_Align" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_Align" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBMol_Align" "', argument " "5"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp5) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_Align" "', argument " "5"" of type '" "OpenBabel::vector3 &""'"); }
   arg5 = reinterpret_cast< OpenBabel::vector3 * >(argp5); (arg1)->Align(arg2,arg3,*arg4,*arg5); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_ConnectTheDots(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_ConnectTheDots" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->ConnectTheDots(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_PerceiveBondOrders(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_PerceiveBondOrders" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->PerceiveBondOrders(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindAngles(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindAngles" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->FindAngles(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_FindTorsions(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_FindTorsions" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); (arg1)->FindTorsions(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetGTDVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetGTDVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetGTDVector" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetGTDVector" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_GetGTDVector" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp2); result = (bool)(arg1)->GetGTDVector(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetGIVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetGIVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetGIVector" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetGIVector" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_GetGIVector" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp2); (arg1)->GetGIVector(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetGIDVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_GetGIDVector", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetGIDVector" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_GetGIDVector" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_GetGIDVector" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp2); (arg1)->GetGIDVector(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Has2D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ; bool result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Has2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_Has2D" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (bool)(arg1)->Has2D(arg2); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Has2D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Has2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->Has2D();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Has2D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_Has2D", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_Has2D__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_OBMol_Has2D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_Has2D'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::Has2D(bool)\n" "    OpenBabel::OBMol::Has2D()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMol_Has3D(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Has3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->Has3D();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasNonZeroCoords(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasNonZeroCoords" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasNonZeroCoords();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasAromaticPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasAromaticPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasAromaticPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasSSSRPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasSSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasSSSRPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasLSSRPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasLSSRPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasLSSRPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasRingAtomsAndBondsPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasRingAtomsAndBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasRingAtomsAndBondsPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasAtomTypesPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasAtomTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasAtomTypesPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasRingTypesPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasRingTypesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasRingTypesPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasChiralityPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasChiralityPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasChiralityPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasPartialChargesPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasPartialChargesPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasPartialChargesPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasHybridizationPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasHybridizationPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasHybridizationPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasClosureBondsPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasClosureBondsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasClosureBondsPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasChainsPerceived(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasChainsPerceived" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasChainsPerceived();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasHydrogensAdded(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasHydrogensAdded" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasHydrogensAdded();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_IsCorrectedForPH(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_IsCorrectedForPH" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->IsCorrectedForPH();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_HasSpinMultiplicityAssigned(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_HasSpinMultiplicityAssigned" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->HasSpinMultiplicityAssigned();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_IsReaction(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_IsReaction" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->IsReaction();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_IsPeriodic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_IsPeriodic" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->IsPeriodic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_Empty(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_Empty" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)(arg1)->Empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NumConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NumConformers" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (int)(arg1)->NumConformers();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< double *,std::allocator< double * > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetConformers" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetConformers" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_SetConformers" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< double *,std::allocator< double * > > * >(argp2); (arg1)->SetConformers(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_AddConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_AddConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_AddConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_AddConformer" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->AddConformer(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_SetConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetConformer" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetConformer(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CopyConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; double *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_CopyConformer", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CopyConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_CopyConformer" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_CopyConformer" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->CopyConformer(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_DeleteConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_DeleteConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_DeleteConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_DeleteConformer" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->DeleteConformer(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  double *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBMol_GetConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetConformer" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (double *)(arg1)->GetConformer(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetEnergies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_SetEnergies", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetEnergies" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_SetEnergies" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_SetEnergies" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > &""'"); }
   arg2 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp2); (arg1)->SetEnergies(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetEnergies(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetEnergies" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (arg1)->GetEnergies();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetEnergy__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetEnergy" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_GetEnergy" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (double)(arg1)->GetEnergy(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetEnergy(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_GetEnergy", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMol_GetEnergy__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMol_GetEnergy__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_GetEnergy'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMol::GetEnergy() const\n" "    OpenBabel::OBMol::GetEnergy(int)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBMol_BeginConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< double *,std::allocator< double * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; double *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_BeginConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_BeginConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_BeginConformer" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_BeginConformer" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< double *,std::allocator< double * > >::iterator * >(argp2);
  result = (double *)(arg1)->BeginConformer(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NextConformer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; std::vector< double *,std::allocator< double * > >::iterator *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; double *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_NextConformer", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NextConformer" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_NextConformer" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_NextConformer" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >::iterator &""'"); }
   arg2 = reinterpret_cast< std::vector< double *,std::allocator< double * > >::iterator * >(argp2);
  result = (double *)(arg1)->NextConformer(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double *,std::allocator< double * > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetConformers" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (std::vector< double *,std::allocator< double * > > *) &(arg1)->GetConformers();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CBeginAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CBeginAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = ((OpenBabel::OBMol const *)arg1)->CBeginAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomConstIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_CEndAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_CEndAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = ((OpenBabel::OBMol const *)arg1)->CEndAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomConstIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = ((OpenBabel::OBMol const *)arg1)->GetAtoms();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_GetBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = ((OpenBabel::OBMol const *)arg1)->GetBonds();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_BeginInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBInternalCoord *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_BeginInternalCoord", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_BeginInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_BeginInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_BeginInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator * >(argp2);
  result = (OpenBabel::OBInternalCoord *)(arg1)->BeginInternalCoord(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_NextInternalCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBInternalCoord *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBMol_NextInternalCoord", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_NextInternalCoord" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBMol_NextInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBMol_NextInternalCoord" "', argument " "2"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator * >(argp2);
  result = (OpenBabel::OBInternalCoord *)(arg1)->NextInternalCoord(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBInternalCoord, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTorsion__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; double arg6 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ;
  int val5 ; int ecode5 = 0 ; double val6 ; int ecode6 = 0 ; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMol_SetTorsion" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMol_SetTorsion" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBMol_SetTorsion" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBMol_SetTorsion" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBMol_SetTorsion" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBMol_SetTorsion" "', argument " "6"" of type '" "double""'"); }  
  arg6 = static_cast< double >(val6); OpenBabel_OBMol_SetTorsion__SWIG_1(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMol_SetTorsion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMol_SetTorsion", 0, 6, argv))) SWIG_fail; --argc; if (argc == 6) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[4], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; PyObject *retobj = _wrap_OBMol_SetTorsion__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_1: if (argc == 6) {
    PyObject *retobj = _wrap_OBMol_SetTorsion__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMol_SetTorsion'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBMol::SetTorsion(OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,double)\n"
  "    OpenBabel::OBMol::SetTorsion(int,int,int,int,double)\n"); return 0; }
SWIGINTERN PyObject *OBMol_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMol, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMol_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_ThrowError__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ThrowError" "', argument " "1"" of type '" "char *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); OpenBabel::ThrowError(arg1); resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_ThrowError__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::string *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ThrowError" "', argument " "1"" of type '" "std::string &""'");  } 
  if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "ThrowError" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1); OpenBabel::ThrowError(*arg1); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_ThrowError(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "ThrowError", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_std__string, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_ThrowError__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 1) { PyObject *retobj = _wrap_ThrowError__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'ThrowError'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::ThrowError(char *)\n" "    OpenBabel::ThrowError(std::string &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_CartesianToInternal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *arg1 = 0 ;
  OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "CartesianToInternal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CartesianToInternal" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CartesianToInternal" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > &""'"); }
  
  arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CartesianToInternal" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CartesianToInternal" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); OpenBabel::CartesianToInternal(*arg1,*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_InternalToCartesian(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *arg1 = 0 ;
  OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "InternalToCartesian", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_OpenBabel__OBInternalCoord_p_t,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "InternalToCartesian" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "InternalToCartesian" "', argument " "1"" of type '" "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > &""'"); }
  
  arg1 = reinterpret_cast< std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "InternalToCartesian" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "InternalToCartesian" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); OpenBabel::InternalToCartesian(*arg1,*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_NewExtension(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::string *arg1 = 0 ;
  char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ; char *buf2 = 0 ; int alloc2 = 0 ; PyObject *swig_obj[2] ;
  std::string result; (void)self; if (!SWIG_Python_UnpackTuple(args, "NewExtension", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "NewExtension" "', argument " "1"" of type '" "std::string &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "NewExtension" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "NewExtension" "', argument " "2"" of type '" "char *""'"); } 
  arg2 = reinterpret_cast< char * >(buf2); result = OpenBabel::NewExtension(*arg1,arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN int Swig_var_NoId_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable NoId is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_NoId_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(OpenBabel::NoId)); return pyobj; }
SWIGINTERN PyObject *_wrap_get_rmat(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; double *arg3 = 0 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "get_rmat", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "get_rmat" "', argument " "1"" of type '" "double *""'");  } 
  arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "get_rmat" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "get_rmat" "', argument " "3"" of type '" "double *""'");  } 
  arg3 = reinterpret_cast< double * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "get_rmat" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); OpenBabel::get_rmat(arg1,arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_ob_make_rmat(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double (*arg1)[3] = (double (*)[3])0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "ob_make_rmat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_a_3__double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ob_make_rmat" "', argument " "1"" of type '" "double [3][3]""'");  }
    arg1 = reinterpret_cast< double (*)[3] >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ob_make_rmat" "', argument " "2"" of type '" "double [9]""'");  }  
  arg2 = reinterpret_cast< double * >(argp2); OpenBabel::ob_make_rmat((double (*)[3])arg1,arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_qtrfit(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; int arg3 ; double (*arg4)[3] = (double (*)[3])0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int val3 ; int ecode3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "qtrfit", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "qtrfit" "', argument " "1"" of type '" "double *""'");  } 
  arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "qtrfit" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "qtrfit" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_a_3__double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "qtrfit" "', argument " "4"" of type '" "double [3][3]""'");  }  
  arg4 = reinterpret_cast< double (*)[3] >(argp4); OpenBabel::qtrfit(arg1,arg2,arg3,(double (*)[3])arg4);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_superimpose(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ;
  PyObject *swig_obj[3] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "superimpose", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "superimpose" "', argument " "1"" of type '" "double *""'");  } 
  arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "superimpose" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "superimpose" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (double)OpenBabel::superimpose(arg1,arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_ring_id_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRing_ring_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_ring_id_set" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRing_ring_id_set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); if (arg1) (arg1)->ring_id = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_ring_id_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_ring_id_get" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (int) ((arg1)->ring_id);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing__path_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRing__path_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing__path_set" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing__path_set" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing__path_set" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_path = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBRing__path_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< int,std::allocator< int > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing__path_get" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  result = (std::vector< int,std::allocator< int > > *) & ((arg1)->_path);
  resultobj = swig::from(static_cast< std::vector< int,std::allocator< int > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRing__pathset_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRing__pathset_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing__pathset_set" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing__pathset_set" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing__pathset_set" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); if (arg1) (arg1)->_pathset = *arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing__pathset_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing__pathset_get" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (OpenBabel::OBBitVec *) & ((arg1)->_pathset);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBRing *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBRing *)new OpenBabel::OBRing();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::vector< int,std::allocator< int > > *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::OBRing *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRing" "', argument " "1"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRing" "', argument " "1"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg1 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBRing" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (OpenBabel::OBRing *)new OpenBabel::OBRing(*arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRing__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::vector< int,std::allocator< int > > *arg1 = 0 ; OpenBabel::OBBitVec arg2 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 ;
  int res2 = 0 ; OpenBabel::OBRing *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRing" "', argument " "1"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRing" "', argument " "1"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg1 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBRing" "', argument " "2"" of type '" "OpenBabel::OBBitVec""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRing" "', argument " "2"" of type '" "OpenBabel::OBBitVec""'"); }
     else { OpenBabel::OBBitVec * temp = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); arg2 = *temp;
      if (SWIG_IsNewObj(res2)) delete temp; }  } 
  result = (OpenBabel::OBRing *)new OpenBabel::OBRing(*arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRing__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBRing *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBRing,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRing" "', argument " "1"" of type '" "OpenBabel::OBRing const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRing" "', argument " "1"" of type '" "OpenBabel::OBRing const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  result = (OpenBabel::OBRing *)new OpenBabel::OBRing((OpenBabel::OBRing const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBRing", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBRing__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_new_OBRing__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap_new_OBRing__SWIG_2(self, argc, argv);}  check_3:
  if (argc == 2) { PyObject *retobj = _wrap_new_OBRing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBRing'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRing::OBRing()\n"
  "    OpenBabel::OBRing::OBRing(std::vector< int,std::allocator< int > > &,int)\n"
  "    OpenBabel::OBRing::OBRing(std::vector< int,std::allocator< int > > &,OpenBabel::OBBitVec)\n"
  "    OpenBabel::OBRing::OBRing(OpenBabel::OBRing const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRing_Size(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBRing *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_Size" "', argument " "1"" of type '" "OpenBabel::OBRing const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = ((OpenBabel::OBRing const *)arg1)->Size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_PathSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_PathSize" "', argument " "1"" of type '" "OpenBabel::OBRing const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = ((OpenBabel::OBRing const *)arg1)->PathSize();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_IsAromatic(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_IsAromatic" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (bool)(arg1)->IsAromatic();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_SetType__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRing *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_SetType" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_SetType" "', argument " "2"" of type '" "char *""'"); } 
  arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetType(arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_SetType__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRing *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_SetType" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_SetType" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing_SetType" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); (arg1)->SetType(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_SetType(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRing_SetType", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__string, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBRing_SetType__SWIG_1(self, argc, argv);}  check_1:
  if (argc == 2) { PyObject *retobj = _wrap_OBRing_SetType__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRing_SetType'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRing::SetType(char *)\n"
  "    OpenBabel::OBRing::SetType(std::string &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRing_GetType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_GetType" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (char *)(arg1)->GetType();
  resultobj = SWIG_FromCharPtr((const char *)result); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_GetRootAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_GetRootAtom" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (unsigned int)(arg1)->GetRootAtom();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_IsMember__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRing *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_IsMember" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_IsMember" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (bool)(arg1)->IsMember(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_IsMember__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRing *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_IsMember" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_IsMember" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->IsMember(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_IsMember(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRing_IsMember", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0; {
      void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBRing_IsMember__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBRing_IsMember__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRing_IsMember'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRing::IsMember(OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBRing::IsMember(OpenBabel::OBBond *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRing_IsInRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRing_IsInRing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_IsInRing" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRing_IsInRing" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsInRing(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_SetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRing_SetParent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_SetParent" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_SetParent" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetParent(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_GetParent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBMol *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_GetParent" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); result = (OpenBabel::OBMol *)(arg1)->GetParent();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRing_findCenterAndNormal(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ; OpenBabel::vector3 *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; PyObject *swig_obj[4] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRing_findCenterAndNormal", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRing_findCenterAndNormal" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRing_findCenterAndNormal" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing_findCenterAndNormal" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRing_findCenterAndNormal" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing_findCenterAndNormal" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBRing_findCenterAndNormal" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRing_findCenterAndNormal" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4); result = (bool)(arg1)->findCenterAndNormal(*arg2,*arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRing" "', argument " "1"" of type '" "OpenBabel::OBRing *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBRing_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRing, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRing_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CompareRingSize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRing *arg1 = 0 ; OpenBabel::OBRing *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "CompareRingSize", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CompareRingSize" "', argument " "1"" of type '" "OpenBabel::OBRing const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRing * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CompareRingSize" "', argument " "2"" of type '" "OpenBabel::OBRing const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBRing * >(argp2);
  result = (bool)OpenBabel::CompareRingSize((OpenBabel::OBRing const *)arg1,(OpenBabel::OBRing const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBRingSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRingSearch", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRingSearch *)new OpenBabel::OBRingSearch();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRingSearch, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRingSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRingSearch" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_SortRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_SortRings" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); (arg1)->SortRings(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_RemoveRedundant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingSearch_RemoveRedundant", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_RemoveRedundant" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRingSearch_RemoveRedundant" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->RemoveRedundant(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_AddRingFromClosure(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBond *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRingSearch_AddRingFromClosure", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_AddRingFromClosure" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingSearch_AddRingFromClosure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingSearch_AddRingFromClosure" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRingSearch_AddRingFromClosure" "', argument " "3"" of type '" "OpenBabel::OBBond *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBond * >(argp3); (arg1)->AddRingFromClosure(*arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_SaveUniqueRing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; std::deque< int > *arg2 = 0 ; std::deque< int > *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRingSearch_SaveUniqueRing", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_SaveUniqueRing" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__dequeT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRingSearch_SaveUniqueRing" "', argument " "2"" of type '" "std::deque< int > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingSearch_SaveUniqueRing" "', argument " "2"" of type '" "std::deque< int > &""'"); }
   arg2 = reinterpret_cast< std::deque< int > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__dequeT_int_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRingSearch_SaveUniqueRing" "', argument " "3"" of type '" "std::deque< int > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRingSearch_SaveUniqueRing" "', argument " "3"" of type '" "std::deque< int > &""'"); }
   arg3 = reinterpret_cast< std::deque< int > * >(argp3); result = (bool)(arg1)->SaveUniqueRing(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_WriteRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_WriteRings" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); (arg1)->WriteRings(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_BeginRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_BeginRings" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); result = (arg1)->BeginRings();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRingSearch_EndRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRingSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRingSearch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRingSearch_EndRings" "', argument " "1"" of type '" "OpenBabel::OBRingSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRingSearch * >(argp1); result = (arg1)->EndRings();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRingSearch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRingSearch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRingSearch_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRTree(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ;
  OpenBabel::OBRTree *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBRTree *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRTree", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRTree" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBRTree, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBRTree" "', argument " "2"" of type '" "OpenBabel::OBRTree *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBRTree * >(argp2); result = (OpenBabel::OBRTree *)new OpenBabel::OBRTree(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRTree, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRTree(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRTree *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRTree, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRTree" "', argument " "1"" of type '" "OpenBabel::OBRTree *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRTree * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRTree_GetAtomIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRTree *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRTree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRTree_GetAtomIdx" "', argument " "1"" of type '" "OpenBabel::OBRTree *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRTree * >(argp1); result = (int)(arg1)->GetAtomIdx();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRTree_PathToRoot(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRTree *arg1 = 0 ; std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRTree_PathToRoot", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRTree, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRTree_PathToRoot" "', argument " "1"" of type '" "OpenBabel::OBRTree *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRTree * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRTree_PathToRoot" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRTree_PathToRoot" "', argument " "2"" of type '" "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > * >(argp2);
  (arg1)->PathToRoot(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRTree_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRTree, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRTree_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSmartsPattern__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBSmartsPattern *)new OpenBabel::OBSmartsPattern();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSmartsPattern, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSmartsPattern(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSmartsPattern" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSmartsPattern__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBSmartsPattern *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBSmartsPattern,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSmartsPattern" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBSmartsPattern" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (OpenBabel::OBSmartsPattern *)new OpenBabel::OBSmartsPattern((OpenBabel::OBSmartsPattern const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSmartsPattern, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSmartsPattern(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBSmartsPattern", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBSmartsPattern__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBSmartsPattern__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBSmartsPattern'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSmartsPattern::OBSmartsPattern()\n"
  "    OpenBabel::OBSmartsPattern::OBSmartsPattern(OpenBabel::OBSmartsPattern const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Init__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Init" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Init" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); result = (bool)(arg1)->Init((char const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Init__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Init" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Init" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Init" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = (bool)(arg1)->Init((std::string const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Init(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSmartsPattern_Init", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBSmartsPattern_Init__SWIG_1(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBSmartsPattern_Init__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSmartsPattern_Init'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSmartsPattern::Init(char const *)\n"
  "    OpenBabel::OBSmartsPattern::Init(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetSMARTS(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetSMARTS" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (std::string *) &((OpenBabel::OBSmartsPattern const *)arg1)->GetSMARTS();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Empty(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Empty" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (bool)((OpenBabel::OBSmartsPattern const *)arg1)->Empty(); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_IsValid" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (bool)((OpenBabel::OBSmartsPattern const *)arg1)->IsValid(); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_NumAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_NumAtoms" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (unsigned int)((OpenBabel::OBSmartsPattern const *)arg1)->NumAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_NumBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_NumBonds" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (unsigned int)((OpenBabel::OBSmartsPattern const *)arg1)->NumBonds();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; int *arg2 = 0 ; int *arg3 = 0 ; int *arg4 = 0 ; int arg5 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; int val5 ;
  int ecode5 = 0 ; PyObject *swig_obj[5] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_GetBond", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetBond" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_int,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_GetBond" "', argument " "2"" of type '" "int &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_GetBond" "', argument " "2"" of type '" "int &""'"); }
   arg2 = reinterpret_cast< int * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_GetBond" "', argument " "3"" of type '" "int &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_GetBond" "', argument " "3"" of type '" "int &""'"); }
   arg3 = reinterpret_cast< int * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_int,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBSmartsPattern_GetBond" "', argument " "4"" of type '" "int &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_GetBond" "', argument " "4"" of type '" "int &""'"); }
   arg4 = reinterpret_cast< int * >(argp4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBSmartsPattern_GetBond" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5); (arg1)->GetBond(*arg2,*arg3,*arg4,arg5); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetAtomicNum(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_GetAtomicNum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetAtomicNum" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSmartsPattern_GetAtomicNum" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)(arg1)->GetAtomicNum(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_GetCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetCharge" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSmartsPattern_GetCharge" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)(arg1)->GetCharge(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetVectorBinding(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_GetVectorBinding", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetVectorBinding" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSmartsPattern_GetVectorBinding" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBSmartsPattern const *)arg1)->GetVectorBinding(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Match__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Match" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSmartsPattern_Match" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->Match(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Match__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Match" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Match(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Match__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg3 = 0 ;
  OpenBabel::OBSmartsPattern::MatchType arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Match" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_Match" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg3 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBSmartsPattern_Match" "', argument " "4"" of type '" "OpenBabel::OBSmartsPattern::MatchType""'"); }
    arg4 = static_cast< OpenBabel::OBSmartsPattern::MatchType >(val4);
  result = (bool)((OpenBabel::OBSmartsPattern const *)arg1)->Match(*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Match__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_Match" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_Match" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_Match" "', argument " "3"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg3 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp3);
  result = (bool)((OpenBabel::OBSmartsPattern const *)arg1)->Match(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_Match(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSmartsPattern_Match", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBSmartsPattern_Match__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_OBSmartsPattern_Match__SWIG_3(self, argc, argv);} 
  check_2: if (argc == 3) { PyObject *retobj = _wrap_OBSmartsPattern_Match__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBSmartsPattern_Match__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSmartsPattern_Match'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSmartsPattern::Match(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBSmartsPattern::Match(OpenBabel::OBMol &)\n"
  "    OpenBabel::OBSmartsPattern::Match(OpenBabel::OBMol &,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &,OpenBabel::OBSmartsPattern::MatchType) const\n"
  "    OpenBabel::OBSmartsPattern::Match(OpenBabel::OBMol &,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &) const\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_HasMatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_HasMatch", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_HasMatch" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_HasMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_HasMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  result = (bool)((OpenBabel::OBSmartsPattern const *)arg1)->HasMatch(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_RestrictedMatch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > *arg3 = 0 ; bool arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool val4 ; int ecode4 = 0 ; bool result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &""'"); }
   arg3 = reinterpret_cast< std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (bool)(arg1)->RestrictedMatch(*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_RestrictedMatch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &""'"); }
   arg3 = reinterpret_cast< std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > * >(argp3);
  result = (bool)(arg1)->RestrictedMatch(*arg2,*arg3); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_RestrictedMatch__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  bool arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool val4 ;
  int ecode4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (bool)(arg1)->RestrictedMatch(*arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_RestrictedMatch__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsPattern *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_RestrictedMatch" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3); result = (bool)(arg1)->RestrictedMatch(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_RestrictedMatch(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSmartsPattern_RestrictedMatch", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1;
    return _wrap_OBSmartsPattern_RestrictedMatch__SWIG_1(self, argc, argv);}  check_1: if (argc == 3) {
    PyObject *retobj = _wrap_OBSmartsPattern_RestrictedMatch__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3;
    PyObject *retobj = _wrap_OBSmartsPattern_RestrictedMatch__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 4) {
    PyObject *retobj = _wrap_OBSmartsPattern_RestrictedMatch__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSmartsPattern_RestrictedMatch'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSmartsPattern::RestrictedMatch(OpenBabel::OBMol &,std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &,bool)\n"
  "    OpenBabel::OBSmartsPattern::RestrictedMatch(OpenBabel::OBMol &,std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > &)\n"
  "    OpenBabel::OBSmartsPattern::RestrictedMatch(OpenBabel::OBMol &,OpenBabel::OBBitVec &,bool)\n"
  "    OpenBabel::OBSmartsPattern::RestrictedMatch(OpenBabel::OBMol &,OpenBabel::OBBitVec &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_NumMatches(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_NumMatches" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (unsigned int)((OpenBabel::OBSmartsPattern const *)arg1)->NumMatches();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetMapList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetMapList" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *) &(arg1)->GetMapList();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_BeginMList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_BeginMList" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); result = (arg1)->BeginMList();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_EndMList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_EndMList" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1); result = (arg1)->EndMList();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_GetUMapList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_GetUMapList" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  result = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *) &(arg1)->GetUMapList();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsPattern_WriteMapList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsPattern *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSmartsPattern_WriteMapList", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsPattern_WriteMapList" "', argument " "1"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__ostream,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsPattern_WriteMapList" "', argument " "2"" of type '" "std::ostream &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsPattern_WriteMapList" "', argument " "2"" of type '" "std::ostream &""'"); }
   arg2 = reinterpret_cast< std::ostream * >(argp2); (arg1)->WriteMapList(*arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBSmartsPattern_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSmartsPattern, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSmartsPattern_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSmartsMatcher(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsMatcher *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBSmartsMatcher", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBSmartsMatcher *)new OpenBabel::OBSmartsMatcher();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSmartsMatcher, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSmartsMatcher(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSmartsMatcher *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsMatcher, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSmartsMatcher" "', argument " "1"" of type '" "OpenBabel::OBSmartsMatcher *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsMatcher * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsMatcher_match__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsMatcher *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::Pattern *arg3 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg4 = 0 ;
  bool arg5 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  void *argp4 = 0 ; int res4 = 0 ; bool val5 ; int ecode5 = 0 ; bool result; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsMatcher, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsMatcher_match" "', argument " "1"" of type '" "OpenBabel::OBSmartsMatcher *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsMatcher * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsMatcher_match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsMatcher_match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__Pattern, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsMatcher_match" "', argument " "3"" of type '" "OpenBabel::Pattern const *""'");  }
   arg3 = reinterpret_cast< OpenBabel::Pattern * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBSmartsMatcher_match" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsMatcher_match" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg4 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp4);
  ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBSmartsMatcher_match" "', argument " "5"" of type '" "bool""'"); }
    arg5 = static_cast< bool >(val5); result = (bool)(arg1)->match(*arg2,(OpenBabel::Pattern const *)arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsMatcher_match__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSmartsMatcher *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::Pattern *arg3 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ;
  int res4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSmartsMatcher, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSmartsMatcher_match" "', argument " "1"" of type '" "OpenBabel::OBSmartsMatcher *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSmartsMatcher * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSmartsMatcher_match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsMatcher_match" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__Pattern, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBSmartsMatcher_match" "', argument " "3"" of type '" "OpenBabel::Pattern const *""'");  }
   arg3 = reinterpret_cast< OpenBabel::Pattern * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBSmartsMatcher_match" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSmartsMatcher_match" "', argument " "4"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg4 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp4);
  result = (bool)(arg1)->match(*arg2,(OpenBabel::Pattern const *)arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSmartsMatcher_match(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSmartsMatcher_match", 0, 5, argv))) SWIG_fail; --argc; if (argc == 4) {
    PyObject *retobj = _wrap_OBSmartsMatcher_match__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBSmartsMatcher_match__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSmartsMatcher_match'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSmartsMatcher::match(OpenBabel::OBMol &,OpenBabel::Pattern const *,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &,bool)\n"
  "    OpenBabel::OBSmartsMatcher::match(OpenBabel::OBMol &,OpenBabel::Pattern const *,std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &)\n");
  return 0; }
SWIGINTERN PyObject *OBSmartsMatcher_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSmartsMatcher, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSmartsMatcher_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSSMatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  OpenBabel::Pattern *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  OpenBabel::OBSSMatch *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBSSMatch", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSSMatch" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBSSMatch" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__Pattern, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBSSMatch" "', argument " "2"" of type '" "OpenBabel::Pattern const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::Pattern * >(argp2);
  result = (OpenBabel::OBSSMatch *)new OpenBabel::OBSSMatch(*arg1,(OpenBabel::Pattern const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSSMatch, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSSMatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSSMatch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSSMatch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSSMatch" "', argument " "1"" of type '" "OpenBabel::OBSSMatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSSMatch * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBSSMatch_Match__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSSMatch *arg1 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg2 = 0 ;
  int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSSMatch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSSMatch_Match" "', argument " "1"" of type '" "OpenBabel::OBSSMatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSSMatch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSSMatch_Match" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSSMatch_Match" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSSMatch_Match" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->Match(*arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSSMatch_Match__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSSMatch *arg1 = 0 ;
  std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSSMatch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSSMatch_Match" "', argument " "1"" of type '" "OpenBabel::OBSSMatch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSSMatch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_int_t_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSSMatch_Match" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSSMatch_Match" "', argument " "2"" of type '" "std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > * >(argp2);
  (arg1)->Match(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSSMatch_Match(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSSMatch_Match", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBSSMatch_Match__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBSSMatch_Match__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSSMatch_Match'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSSMatch::Match(std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &,int)\n"
  "    OpenBabel::OBSSMatch::Match(std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > &)\n");
  return 0; }
SWIGINTERN PyObject *OBSSMatch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSSMatch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSSMatch_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_SmartsLexReplace(PyObject *self, PyObject *args) { PyObject *resultobj = 0; std::string *arg1 = 0 ;
  std::vector< std::pair< std::string,std::string >,std::allocator< std::pair< std::string,std::string > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SmartsLexReplace", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SmartsLexReplace" "', argument " "1"" of type '" "std::string &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SmartsLexReplace" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SmartsLexReplace" "', argument " "2"" of type '" "std::vector< std::pair< std::string,std::string >,std::allocator< std::pair< std::string,std::string > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "SmartsLexReplace" "', argument " "2"" of type '" "std::vector< std::pair< std::string,std::string >,std::allocator< std::pair< std::string,std::string > > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< std::pair< std::string,std::string >,std::allocator< std::pair< std::string,std::string > > > * >(argp2);
  OpenBabel::SmartsLexReplace(*arg1,*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN int Swig_var_AliasDataType_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable AliasDataType is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_AliasDataType_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::AliasDataType)); return pyobj; }
SWIGINTERN PyObject *_wrap_new_AliasData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_AliasData", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::AliasData *)new OpenBabel::AliasData();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__AliasData, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_SetAlias__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::AliasData *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_SetAlias" "', argument " "1"" of type '" "OpenBabel::AliasData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AliasData_SetAlias" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "AliasData_SetAlias" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetAlias((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_SetAlias__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::AliasData *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_SetAlias" "', argument " "1"" of type '" "OpenBabel::AliasData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AliasData_SetAlias" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->SetAlias((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_SetAlias(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "AliasData_SetAlias", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) { int _v = 0;
    { int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);}  if (!_v) goto check_1;
    return _wrap_AliasData_SetAlias__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_AliasData_SetAlias__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'AliasData_SetAlias'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::AliasData::SetAlias(std::string const &)\n"
  "    OpenBabel::AliasData::SetAlias(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_AliasData_GetAlias__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::AliasData *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; std::string result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_GetAlias" "', argument " "1"" of type '" "OpenBabel::AliasData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "AliasData_GetAlias" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = ((OpenBabel::AliasData const *)arg1)->GetAlias(arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_GetAlias__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::AliasData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; std::string result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_GetAlias" "', argument " "1"" of type '" "OpenBabel::AliasData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); result = ((OpenBabel::AliasData const *)arg1)->GetAlias();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_GetAlias(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "AliasData_GetAlias", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_AliasData_GetAlias__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_AliasData_GetAlias__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'AliasData_GetAlias'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::AliasData::GetAlias(bool) const\n"
  "    OpenBabel::AliasData::GetAlias() const\n"); return 0; }
SWIGINTERN PyObject *_wrap_AliasData_GetColor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_GetColor" "', argument " "1"" of type '" "OpenBabel::AliasData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); result = ((OpenBabel::AliasData const *)arg1)->GetColor();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_SetColor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "AliasData_SetColor", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_SetColor" "', argument " "1"" of type '" "OpenBabel::AliasData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "AliasData_SetColor" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->SetColor(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_IsExpanded(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_IsExpanded" "', argument " "1"" of type '" "OpenBabel::AliasData const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); result = (bool)((OpenBabel::AliasData const *)arg1)->IsExpanded();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_RevertToAliasForm(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_RevertToAliasForm" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "AliasData_RevertToAliasForm" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); OpenBabel::AliasData::RevertToAliasForm(*arg1);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_Expand(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "AliasData_Expand", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_Expand" "', argument " "1"" of type '" "OpenBabel::AliasData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AliasData_Expand" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "AliasData_Expand" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "AliasData_Expand" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); result = (bool)(arg1)->Expand(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AliasData_AddAliases(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AliasData_AddAliases" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)OpenBabel::AliasData::AddAliases(arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_AliasData(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::AliasData *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__AliasData, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_AliasData" "', argument " "1"" of type '" "OpenBabel::AliasData *""'");  }
   arg1 = reinterpret_cast< OpenBabel::AliasData * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *AliasData_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__AliasData, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *AliasData_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFingerprint_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint **result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFingerprint **) &OpenBabel::OBFingerprint::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBFingerprint, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFingerprint *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_FindType" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFingerprint *)OpenBabel::OBFingerprint::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFingerprint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFingerprint" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_SetBit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ;
  PyObject *swig_obj[3] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_SetBit", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_SetBit" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_SetBit" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_SetBit" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFingerprint_SetBit" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetBit(*arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_GetBit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; unsigned int val3 ; int ecode3 = 0 ;
  PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_GetBit", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_GetBit" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_GetBit" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_GetBit" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFingerprint_GetBit" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3);
  result = (bool)(arg1)->GetBit((std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Fold(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ;
  PyObject *swig_obj[3] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_Fold", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_Fold" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_Fold" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_Fold" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFingerprint_Fold" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->Fold(*arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_GetFingerprint__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFingerprint *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_GetFingerprint" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_GetFingerprint" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFingerprint_GetFingerprint" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_GetFingerprint" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBFingerprint_GetFingerprint" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); result = (bool)(arg1)->GetFingerprint(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_GetFingerprint__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFingerprint *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_GetFingerprint" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_GetFingerprint" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBFingerprint_GetFingerprint" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_GetFingerprint" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  result = (bool)(arg1)->GetFingerprint(arg2,*arg3); resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_GetFingerprint(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBFingerprint_GetFingerprint", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBFingerprint_GetFingerprint__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBFingerprint_GetFingerprint__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBFingerprint_GetFingerprint'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBFingerprint::GetFingerprint(OpenBabel::OBBase *,std::vector< unsigned int,std::allocator< unsigned int > > &,int)\n"
  "    OpenBabel::OBFingerprint::GetFingerprint(OpenBabel::OBBase *,std::vector< unsigned int,std::allocator< unsigned int > > &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Flags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_Flags" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); result = (unsigned int)(arg1)->Flags();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_SetFlags(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFingerprint *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_SetFlags", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_SetFlags" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFingerprint_SetFlags" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetFlags(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_DescribeBits__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFingerprint *arg1 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > arg2 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; bool val3 ;
  int ecode3 = 0 ; std::string result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_DescribeBits" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBFingerprint_DescribeBits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFingerprint_DescribeBits" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (arg1)->DescribeBits(SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_DescribeBits__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBFingerprint *arg1 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > arg2 ; void *argp1 = 0 ; int res1 = 0 ; std::string result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_DescribeBits" "', argument " "1"" of type '" "OpenBabel::OBFingerprint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFingerprint * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBFingerprint_DescribeBits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (arg1)->DescribeBits(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_DescribeBits(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBFingerprint_DescribeBits", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBFingerprint_DescribeBits__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBFingerprint_DescribeBits__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBFingerprint_DescribeBits'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBFingerprint::DescribeBits(std::vector< unsigned int,std::allocator< unsigned int > > const,bool)\n"
  "    OpenBabel::OBFingerprint::DescribeBits(std::vector< unsigned int,std::allocator< unsigned int > > const)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Tanimoto__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int,std::allocator< unsigned int > > *arg1 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ; int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ;
  double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_Tanimoto" "', argument " "1"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_Tanimoto" "', argument " "1"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg1 = ptr; }  {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_Tanimoto" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_Tanimoto" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; } 
  result = (double)OpenBabel::OBFingerprint::Tanimoto((std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Tanimoto__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::vector< unsigned int,std::allocator< unsigned int > > *arg1 = 0 ; unsigned int *arg2 = 0 ;
  int res1 = SWIG_OLDOBJ ; void *argp2 = 0 ; int res2 = 0 ; double result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_Tanimoto" "', argument " "1"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFingerprint_Tanimoto" "', argument " "1"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg1 = ptr; }  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_int, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFingerprint_Tanimoto" "', argument " "2"" of type '" "unsigned int const *""'");  }
   arg2 = reinterpret_cast< unsigned int * >(argp2);
  result = (double)OpenBabel::OBFingerprint::Tanimoto((std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg1,(unsigned int const *)arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Tanimoto(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBFingerprint_Tanimoto", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_int, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; return _wrap_OBFingerprint_Tanimoto__SWIG_1(self, argc, argv);} 
  check_1: if (argc == 2) { PyObject *retobj = _wrap_OBFingerprint_Tanimoto__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBFingerprint_Tanimoto'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBFingerprint::Tanimoto(std::vector< unsigned int,std::allocator< unsigned int > > const &,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n"
  "    OpenBabel::OBFingerprint::Tanimoto(std::vector< unsigned int,std::allocator< unsigned int > > const &,unsigned int const *)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBFingerprint_Getbitsperint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  unsigned int result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFingerprint_Getbitsperint", 0, 0, 0)) SWIG_fail;
  result = (unsigned int)OpenBabel::OBFingerprint::Getbitsperint();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFingerprint_FindFingerprint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBFingerprint *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFingerprint_FindFingerprint" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBFingerprint *)OpenBabel::OBFingerprint::FindFingerprint((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *OBFingerprint_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFingerprint, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_FptIndexHeader_headerlength_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_headerlength_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_headerlength_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FptIndexHeader_headerlength_set" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); if (arg1) (arg1)->headerlength = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_headerlength_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_headerlength_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (unsigned int) ((arg1)->headerlength);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_nEntries_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_nEntries_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_nEntries_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FptIndexHeader_nEntries_set" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); if (arg1) (arg1)->nEntries = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_nEntries_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_nEntries_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (unsigned int) ((arg1)->nEntries);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_words_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_words_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_words_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FptIndexHeader_words_set" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); if (arg1) (arg1)->words = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_words_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_words_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (unsigned int) ((arg1)->words);
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_fpid_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; char *arg2 = (char *)0 ; void *argp1 = 0 ; int res1 = 0 ; char temp2[15] ; int res2 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_fpid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_fpid_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); res2 = SWIG_AsCharArray(swig_obj[1], temp2, 15);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FptIndexHeader_fpid_set" "', argument " "2"" of type '" "char [15]""'"); }
   arg2 = reinterpret_cast< char * >(temp2); if (arg2) memcpy(arg1->fpid,arg2,15*sizeof(char));
  else memset(arg1->fpid,0,15*sizeof(char)); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_fpid_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_fpid_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (char *)(char *) ((arg1)->fpid); {
    size_t size = SWIG_strnlen(result, 15); resultobj = SWIG_FromCharPtrAndSize(result, size); }  return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_seek64_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; char arg2 ; void *argp1 = 0 ; int res1 = 0 ; char val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_seek64_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_seek64_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); ecode2 = SWIG_AsVal_char(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FptIndexHeader_seek64_set" "', argument " "2"" of type '" "char""'"); }
    arg2 = static_cast< char >(val2); if (arg1) (arg1)->seek64 = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_seek64_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_seek64_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (char) ((arg1)->seek64);
  resultobj = SWIG_From_char(static_cast< char >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_datafilename_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; char *arg2 = (char *)0 ; void *argp1 = 0 ; int res1 = 0 ; char temp2[256] ; int res2 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FptIndexHeader_datafilename_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_datafilename_set" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); res2 = SWIG_AsCharArray(swig_obj[1], temp2, 256);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FptIndexHeader_datafilename_set" "', argument " "2"" of type '" "char [256]""'"); }
   arg2 = reinterpret_cast< char * >(temp2); if (arg2) memcpy(arg1->datafilename,arg2,256*sizeof(char));
  else memset(arg1->datafilename,0,256*sizeof(char)); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FptIndexHeader_datafilename_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; char *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FptIndexHeader_datafilename_get" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); result = (char *)(char *) ((arg1)->datafilename); {
    size_t size = SWIG_strnlen(result, 256); resultobj = SWIG_FromCharPtrAndSize(result, size); }  return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_new_FptIndexHeader(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_FptIndexHeader", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::FptIndexHeader *)new OpenBabel::FptIndexHeader();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FptIndexHeader, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_FptIndexHeader(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FptIndexHeader *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndexHeader, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FptIndexHeader" "', argument " "1"" of type '" "OpenBabel::FptIndexHeader *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndexHeader * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *FptIndexHeader_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__FptIndexHeader, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *FptIndexHeader_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_FastSearch_ReadIndexFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ;
  std::string result; (void)self; if (!SWIG_Python_UnpackTuple(args, "FastSearch_ReadIndexFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_ReadIndexFile" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "FastSearch_ReadIndexFile" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (arg1)->ReadIndexFile(SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_ReadIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; std::istream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; std::string result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FastSearch_ReadIndex", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_ReadIndex" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__istream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_ReadIndex" "', argument " "2"" of type '" "std::istream *""'");  }
   arg2 = reinterpret_cast< std::istream * >(argp2); result = (arg1)->ReadIndex(arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_FastSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FastSearch" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_Find(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::vector< unsigned long,std::allocator< unsigned long > > *arg3 = 0 ; unsigned int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; unsigned int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "FastSearch_Find", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_Find" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_Find" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_Find" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_Find" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned long,std::allocator< unsigned long > > * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastSearch_Find" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); result = (bool)(arg1)->Find(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindMatch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::vector< unsigned long,std::allocator< unsigned long > > *arg3 = 0 ; unsigned int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; unsigned int val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "FastSearch_FindMatch", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_FindMatch" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_FindMatch" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_FindMatch" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_FindMatch" "', argument " "3"" of type '" "std::vector< unsigned long,std::allocator< unsigned long > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned long,std::allocator< unsigned long > > * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastSearch_FindMatch" "', argument " "4"" of type '" "unsigned int""'"); }
    arg4 = static_cast< unsigned int >(val4); result = (bool)(arg1)->FindMatch(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindSimilar__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::multimap< double,unsigned long > *arg3 = 0 ; double arg4 ; double arg5 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double val4 ; int ecode4 = 0 ; double val5 ;
  int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_FindSimilar" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_FindSimilar" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__multimapT_double_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'"); }
   arg3 = reinterpret_cast< std::multimap< double,unsigned long > * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastSearch_FindSimilar" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FastSearch_FindSimilar" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); result = (bool)(arg1)->FindSimilar(arg2,*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindSimilar__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::multimap< double,unsigned long > *arg3 = 0 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_FindSimilar" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_FindSimilar" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__multimapT_double_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'"); }
   arg3 = reinterpret_cast< std::multimap< double,unsigned long > * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastSearch_FindSimilar" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); result = (bool)(arg1)->FindSimilar(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindSimilar__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::multimap< double,unsigned long > *arg3 = 0 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_FindSimilar" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_FindSimilar" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__multimapT_double_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'"); }
   arg3 = reinterpret_cast< std::multimap< double,unsigned long > * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastSearch_FindSimilar" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); result = (bool)(arg1)->FindSimilar(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindSimilar__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FastSearch *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ;
  std::multimap< double,unsigned long > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_FindSimilar" "', argument " "1"" of type '" "OpenBabel::FastSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearch_FindSimilar" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__multimapT_double_unsigned_long_t,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearch_FindSimilar" "', argument " "3"" of type '" "std::multimap< double,unsigned long > &""'"); }
   arg3 = reinterpret_cast< std::multimap< double,unsigned long > * >(argp3); result = (bool)(arg1)->FindSimilar(arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_FindSimilar(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastSearch_FindSimilar", 0, 5, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_FastSearch_FindSimilar__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[3], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_FastSearch_FindSimilar__SWIG_2(self, argc, argv);}  check_2: if (argc == 4) {
    PyObject *retobj = _wrap_FastSearch_FindSimilar__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_FastSearch_FindSimilar__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastSearch_FindSimilar'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::FastSearch::FindSimilar(OpenBabel::OBBase *,std::multimap< double,unsigned long > &,double,double)\n"
  "    OpenBabel::FastSearch::FindSimilar(OpenBabel::OBBase *,std::multimap< double,unsigned long > &,double)\n"
  "    OpenBabel::FastSearch::FindSimilar(OpenBabel::OBBase *,std::multimap< double,unsigned long > &,int)\n"
  "    OpenBabel::FastSearch::FindSimilar(OpenBabel::OBBase *,std::multimap< double,unsigned long > &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_FastSearch_GetFingerprint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBFingerprint *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_GetFingerprint" "', argument " "1"" of type '" "OpenBabel::FastSearch const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  result = (OpenBabel::OBFingerprint *)((OpenBabel::FastSearch const *)arg1)->GetFingerprint();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFingerprint, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_FastSearch_GetIndexHeader(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::FptIndexHeader *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearch_GetIndexHeader" "', argument " "1"" of type '" "OpenBabel::FastSearch const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearch * >(argp1);
  result = (OpenBabel::FptIndexHeader *) &((OpenBabel::FastSearch const *)arg1)->GetIndexHeader();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FptIndexHeader, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearch *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_FastSearch", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::FastSearch *)new OpenBabel::FastSearch();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearch, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *FastSearch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__FastSearch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *FastSearch_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_FastSearchIndexer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; std::ostream *arg2 = 0 ; std::string *arg3 = 0 ; int arg4 ; int arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ;
  int val5 ; int ecode5 = 0 ; OpenBabel::FastSearchIndexer *result = 0 ; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_FastSearchIndexer" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_FastSearchIndexer" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_FastSearchIndexer" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5);
  result = (OpenBabel::FastSearchIndexer *)new OpenBabel::FastSearchIndexer(*arg1,arg2,*arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearchIndexer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; std::ostream *arg2 = 0 ; std::string *arg3 = 0 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int val4 ; int ecode4 = 0 ;
  OpenBabel::FastSearchIndexer *result = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_FastSearchIndexer" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_FastSearchIndexer" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4);
  result = (OpenBabel::FastSearchIndexer *)new OpenBabel::FastSearchIndexer(*arg1,arg2,*arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearchIndexer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; std::ostream *arg2 = 0 ; std::string *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; OpenBabel::FastSearchIndexer *result = 0 ;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "std::string &""'"); }
   arg1 = reinterpret_cast< std::string * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_FastSearchIndexer" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3);
  result = (OpenBabel::FastSearchIndexer *)new OpenBabel::FastSearchIndexer(*arg1,arg2,*arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearchIndexer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FptIndex *arg1 = 0 ; std::ostream *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::FastSearchIndexer *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndex, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "OpenBabel::FptIndex *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndex * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_FastSearchIndexer" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_FastSearchIndexer" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); result = (OpenBabel::FastSearchIndexer *)new OpenBabel::FastSearchIndexer(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearchIndexer__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::FptIndex *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::FastSearchIndexer *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FptIndex, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FastSearchIndexer" "', argument " "1"" of type '" "OpenBabel::FptIndex *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FptIndex * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_FastSearchIndexer" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2);
  result = (OpenBabel::FastSearchIndexer *)new OpenBabel::FastSearchIndexer(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_FastSearchIndexer(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_FastSearchIndexer", 0, 5, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_new_FastSearchIndexer__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__FptIndex, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; { { int res = SWIG_AsVal_int(argv[2], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_new_FastSearchIndexer__SWIG_3(self, argc, argv);}  check_2: if (argc == 3) {
    PyObject *retobj = _wrap_new_FastSearchIndexer__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_FastSearchIndexer__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_new_FastSearchIndexer__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_FastSearchIndexer'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::FastSearchIndexer::FastSearchIndexer(std::string &,std::ostream *,std::string &,int,int)\n"
  "    OpenBabel::FastSearchIndexer::FastSearchIndexer(std::string &,std::ostream *,std::string &,int)\n"
  "    OpenBabel::FastSearchIndexer::FastSearchIndexer(std::string &,std::ostream *,std::string &)\n"
  "    OpenBabel::FastSearchIndexer::FastSearchIndexer(OpenBabel::FptIndex *,std::ostream *,int)\n"
  "    OpenBabel::FastSearchIndexer::FastSearchIndexer(OpenBabel::FptIndex *,std::ostream *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_FastSearchIndexer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearchIndexer *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FastSearchIndexer" "', argument " "1"" of type '" "OpenBabel::FastSearchIndexer *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearchIndexer * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_FastSearchIndexer_Add(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::FastSearchIndexer *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::streampos arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "FastSearchIndexer_Add", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__FastSearchIndexer, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastSearchIndexer_Add" "', argument " "1"" of type '" "OpenBabel::FastSearchIndexer *""'");  }
   arg1 = reinterpret_cast< OpenBabel::FastSearchIndexer * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastSearchIndexer_Add" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); {
    res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__streampos,  0  | 0); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastSearchIndexer_Add" "', argument " "3"" of type '" "std::streampos""'");  }
       if (!argp3) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FastSearchIndexer_Add" "', argument " "3"" of type '" "std::streampos""'"); }
     else { std::streampos * temp = reinterpret_cast< std::streampos * >(argp3); arg3 = *temp;
      if (SWIG_IsNewObj(res3)) delete temp; }  }  result = (bool)(arg1)->Add(arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *FastSearchIndexer_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__FastSearchIndexer, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *FastSearchIndexer_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBDescriptor_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDescriptor **result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDescriptor_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBDescriptor **) &OpenBabel::OBDescriptor::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBDescriptor, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBDescriptor__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; bool arg2 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; bool val2 ;
  int ecode2 = 0 ; OpenBabel::OBDescriptor *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBDescriptor" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBDescriptor" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (OpenBabel::OBDescriptor *)new OpenBabel::OBDescriptor((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBDescriptor, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBDescriptor__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::OBDescriptor *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBDescriptor" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); result = (OpenBabel::OBDescriptor *)new OpenBabel::OBDescriptor((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBDescriptor, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBDescriptor(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBDescriptor", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBDescriptor__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBDescriptor__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBDescriptor'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBDescriptor::OBDescriptor(char const *,bool)\n"
  "    OpenBabel::OBDescriptor::OBDescriptor(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBDescriptor *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_FindType" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBDescriptor *)OpenBabel::OBDescriptor::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Predict__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Predict" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Predict" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__string, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_Predict" "', argument " "3"" of type '" "std::string *""'");  }
   arg3 = reinterpret_cast< std::string * >(argp3); result = (double)(arg1)->Predict(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Predict__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Predict" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Predict" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (double)(arg1)->Predict(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Predict(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_Predict", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBDescriptor_Predict__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBDescriptor_Predict__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_Predict'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBDescriptor::Predict(OpenBabel::OBBase *,std::string *)\n"
  "    OpenBabel::OBDescriptor::Predict(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_PredictAndSave__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_PredictAndSave" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_PredictAndSave" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__string, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_PredictAndSave" "', argument " "3"" of type '" "std::string *""'");  }
   arg3 = reinterpret_cast< std::string * >(argp3); result = (double)(arg1)->PredictAndSave(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_PredictAndSave__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_PredictAndSave" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_PredictAndSave" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (double)(arg1)->PredictAndSave(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_PredictAndSave(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_PredictAndSave", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBDescriptor_PredictAndSave__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBDescriptor_PredictAndSave__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_PredictAndSave'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBDescriptor::PredictAndSave(OpenBabel::OBBase *,std::string *)\n"
  "    OpenBabel::OBDescriptor::PredictAndSave(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_GetStringValue__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string *arg3 = 0 ;
  std::string *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  void *argp4 = 0 ; int res4 = 0 ; double result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_GetStringValue" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_GetStringValue" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_GetStringValue" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_GetStringValue" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_std__string, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBDescriptor_GetStringValue" "', argument " "4"" of type '" "std::string *""'");  }
   arg4 = reinterpret_cast< std::string * >(argp4); result = (double)(arg1)->GetStringValue(arg2,*arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_GetStringValue__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::string *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_GetStringValue" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_GetStringValue" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_GetStringValue" "', argument " "3"" of type '" "std::string &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_GetStringValue" "', argument " "3"" of type '" "std::string &""'"); }
   arg3 = reinterpret_cast< std::string * >(argp3); result = (double)(arg1)->GetStringValue(arg2,*arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_GetStringValue(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_GetStringValue", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBDescriptor_GetStringValue__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBDescriptor_GetStringValue__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_GetStringValue'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBDescriptor::GetStringValue(OpenBabel::OBBase *,std::string &,std::string *)\n"
  "    OpenBabel::OBDescriptor::GetStringValue(OpenBabel::OBBase *,std::string &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Compare__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::istream *arg3 = 0 ;
  bool arg4 ; std::string *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; bool val4 ; int ecode4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; bool result; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Compare" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Compare" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__istream,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_Compare" "', argument " "3"" of type '" "std::istream &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_Compare" "', argument " "3"" of type '" "std::istream &""'"); }
   arg3 = reinterpret_cast< std::istream * >(argp3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBDescriptor_Compare" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_std__string, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBDescriptor_Compare" "', argument " "5"" of type '" "std::string *""'");  }
   arg5 = reinterpret_cast< std::string * >(argp5); result = (bool)(arg1)->Compare(arg2,*arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Compare__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; std::istream *arg3 = 0 ;
  bool arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool val4 ;
  int ecode4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Compare" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Compare" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__istream,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_Compare" "', argument " "3"" of type '" "std::istream &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_Compare" "', argument " "3"" of type '" "std::istream &""'"); }
   arg3 = reinterpret_cast< std::istream * >(argp3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBDescriptor_Compare" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); result = (bool)(arg1)->Compare(arg2,*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Compare(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_Compare", 0, 5, argv))) SWIG_fail; --argc; if (argc == 4) {
    PyObject *retobj = _wrap_OBDescriptor_Compare__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBDescriptor_Compare__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_Compare'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBDescriptor::Compare(OpenBabel::OBBase *,std::istream &,bool,std::string *)\n"
  "    OpenBabel::OBDescriptor::Compare(OpenBabel::OBBase *,std::istream &,bool)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Display__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; char *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; int res4 ;
  char *buf4 = 0 ; int alloc4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Display" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBDescriptor_Display" "', argument " "4"" of type '" "char const *""'"); }
   arg4 = reinterpret_cast< char * >(buf4); result = (bool)(arg1)->Display(*arg2,(char const *)arg3,(char const *)arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return resultobj; fail: if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Display__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; std::string *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Display" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_Display" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_Display" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBDescriptor_Display" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->Display(*arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Display(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_Display", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBDescriptor_Display__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBDescriptor_Display__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_Display'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBDescriptor::Display(std::string &,char const *,char const *)\n"
  "    OpenBabel::OBDescriptor::Display(std::string &,char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Order__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; double arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Order" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBDescriptor_Order" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBDescriptor_Order" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); result = (bool)(arg1)->Order(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Order__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBDescriptor *arg1 = 0 ; std::string arg2 ; std::string arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_Order" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBDescriptor_Order" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBDescriptor_Order" "', argument " "3"" of type '" "std::string""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (bool)(arg1)->Order(SWIG_STD_MOVE(arg2),SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_Order(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBDescriptor_Order", 0, 3, argv))) SWIG_fail; --argc; if (argc == 3) { int _v = 0;
    { { int res = SWIG_AsVal_double(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_1; { {
        int res = SWIG_AsVal_double(argv[2], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_1;
    return _wrap_OBDescriptor_Order__SWIG_0(self, argc, argv);}  check_1: if (argc == 3) {
    PyObject *retobj = _wrap_OBDescriptor_Order__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBDescriptor_Order'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBDescriptor::Order(double,double)\n"
  "    OpenBabel::OBDescriptor::Order(std::string,std::string)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBDescriptor_FilterCompare(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; std::istream *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; bool val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDescriptor_FilterCompare", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_FilterCompare" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__istream,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_FilterCompare" "', argument " "2"" of type '" "std::istream &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_FilterCompare" "', argument " "2"" of type '" "std::istream &""'"); }
   arg2 = reinterpret_cast< std::istream * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBDescriptor_FilterCompare" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)OpenBabel::OBDescriptor::FilterCompare(arg1,*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_AddProperties(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDescriptor_AddProperties", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_AddProperties" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_AddProperties" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_AddProperties" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  OpenBabel::OBDescriptor::AddProperties(arg1,(std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_DeleteProperties(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDescriptor_DeleteProperties", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_DeleteProperties" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_DeleteProperties" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_DeleteProperties" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  OpenBabel::OBDescriptor::DeleteProperties(arg1,(std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_GetValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; std::string result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBDescriptor_GetValues", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_GetValues" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBDescriptor_GetValues" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_GetValues" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  result = OpenBabel::OBDescriptor::GetValues(arg1,(std::string const &)*arg2);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj;
  fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBDescriptor_GetIdentifier(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::istream *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::pair< std::string,std::string > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__istream,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBDescriptor_GetIdentifier" "', argument " "1"" of type '" "std::istream &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBDescriptor_GetIdentifier" "', argument " "1"" of type '" "std::istream &""'"); }
   arg1 = reinterpret_cast< std::istream * >(argp1); result = OpenBabel::OBDescriptor::GetIdentifier(*arg1);
  resultobj = SWIG_NewPointerObj((new std::pair< std::string,std::string >(result)), SWIGTYPE_p_std__pairT_std__string_std__string_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBDescriptor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBDescriptor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBDescriptor, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBDescriptor" "', argument " "1"" of type '" "OpenBabel::OBDescriptor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBDescriptor * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBDescriptor_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBDescriptor, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBDescriptor_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_KCAL_TO_KJ_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable KCAL_TO_KJ is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_KCAL_TO_KJ_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_double(static_cast< double >(OpenBabel::KCAL_TO_KJ)); return pyobj; }
SWIGINTERN int Swig_var_GAS_CONSTANT_set(PyObject *) { SWIG_Error(SWIG_AttributeError,"Variable GAS_CONSTANT is read-only.");
  return 1; }
SWIGINTERN PyObject *Swig_var_GAS_CONSTANT_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_double(static_cast< double >(OpenBabel::GAS_CONSTANT)); return pyobj; }
SWIGINTERN PyObject *_wrap_new_LineSearchType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::LineSearchType *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_LineSearchType", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::LineSearchType *)new OpenBabel::LineSearchType();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__LineSearchType, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_LineSearchType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::LineSearchType *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__LineSearchType, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_LineSearchType" "', argument " "1"" of type '" "OpenBabel::LineSearchType *""'");  }
   arg1 = reinterpret_cast< OpenBabel::LineSearchType * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *LineSearchType_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__LineSearchType, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *LineSearchType_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFFParameter_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFParameter_a_set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); if (arg1) (arg1)->a = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (int) ((arg1)->a);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFParameter_b_set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); if (arg1) (arg1)->b = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (int) ((arg1)->b);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFParameter_c_set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); if (arg1) (arg1)->c = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (int) ((arg1)->c);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFParameter_d_set" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); if (arg1) (arg1)->d = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (int) ((arg1)->d);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__a_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__a_set" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__a_set" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_a = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__a_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (std::string *) & ((arg1)->_a);
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__b_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__b_set" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__b_set" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_b = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__b_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (std::string *) & ((arg1)->_b);
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__c_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__c_set" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__c_set" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_c = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__c_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (std::string *) & ((arg1)->_c);
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__d_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__d_set" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__d_set" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_d = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__d_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); result = (std::string *) & ((arg1)->_d);
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__ipar_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__ipar_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__ipar_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__ipar_set" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__ipar_set" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_ipar = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__ipar_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< int,std::allocator< int > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__ipar_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1);
  result = (std::vector< int,std::allocator< int > > *) & ((arg1)->_ipar);
  resultobj = swig::from(static_cast< std::vector< int,std::allocator< int > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__dpar_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFParameter__dpar_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__dpar_set" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFParameter__dpar_set" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFParameter__dpar_set" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->_dpar = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter__dpar_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter__dpar_get" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) & ((arg1)->_dpar);
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFParameter_clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFParameter_clear" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); (arg1)->clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFFParameter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFParameter", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFParameter *)new OpenBabel::OBFFParameter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFParameter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFParameter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFParameter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFParameter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFParameter" "', argument " "1"" of type '" "OpenBabel::OBFFParameter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFParameter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBFFParameter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFParameter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFParameter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFFCalculation2_energy_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_energy_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_energy_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFCalculation2_energy_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->energy = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_energy_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_energy_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (double) ((arg1)->energy);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_a_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->a = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_b_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->b = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_idx_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_idx_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_idx_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFCalculation2_idx_a_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->idx_a = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_idx_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_idx_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (int) ((arg1)->idx_a);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_idx_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_idx_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_idx_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFCalculation2_idx_b_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->idx_b = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_idx_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_idx_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (int) ((arg1)->idx_b);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_pos_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_pos_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_pos_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_pos_a_set" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); if (arg1) (arg1)->pos_a = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_pos_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_pos_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (double *) ((arg1)->pos_a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_pos_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_pos_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_pos_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_pos_b_set" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); if (arg1) (arg1)->pos_b = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_pos_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_pos_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (double *) ((arg1)->pos_b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_force_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_force_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_force_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_force_a_set" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); { if (arg2) { size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(double *)&arg1->force_a[ii] = *((double *)arg2 + ii); }  else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""force_a""' of type '""double [3]""'"); }
     }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_force_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_force_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (double *)(double *) ((arg1)->force_a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_force_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation2_force_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_force_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation2_force_b_set" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); { if (arg2) { size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(double *)&arg1->force_b[ii] = *((double *)arg2 + ii); }  else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""force_b""' of type '""double [3]""'"); }
     }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_force_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_force_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); result = (double *)(double *) ((arg1)->force_b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFCalculation2(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFCalculation2" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation2_SetupPointers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation2, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation2_SetupPointers" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation2 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation2 * >(argp1); (arg1)->SetupPointers(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFFCalculation2(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation2 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFCalculation2", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFCalculation2 *)new OpenBabel::OBFFCalculation2();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFCalculation2, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFFCalculation2_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFCalculation2, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFCalculation2_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFFCalculation3_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation3_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation3_c_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->c = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_idx_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation3_idx_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_idx_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFCalculation3_idx_c_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->idx_c = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_idx_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_idx_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); result = (int) ((arg1)->idx_c);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_pos_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation3_pos_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_pos_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation3_pos_c_set" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); if (arg1) (arg1)->pos_c = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_pos_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_pos_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); result = (double *) ((arg1)->pos_c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_force_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation3_force_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_force_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation3_force_c_set" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); { if (arg2) { size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(double *)&arg1->force_c[ii] = *((double *)arg2 + ii); }  else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""force_c""' of type '""double [3]""'"); }
     }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation3_force_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation3_force_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); result = (double *)(double *) ((arg1)->force_c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFCalculation3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFCalculation3" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation3 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation3 * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFFCalculation3(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation3 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFCalculation3", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFCalculation3 *)new OpenBabel::OBFFCalculation3();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFCalculation3, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFFCalculation3_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFCalculation3, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFCalculation3_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFFCalculation4_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation4_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation4_d_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->d = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->d);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_idx_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation4_idx_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_idx_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFCalculation4_idx_d_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->idx_d = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_idx_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_idx_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); result = (int) ((arg1)->idx_d);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_pos_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation4_pos_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_pos_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation4_pos_d_set" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); if (arg1) (arg1)->pos_d = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_pos_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_pos_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); result = (double *) ((arg1)->pos_d);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_force_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; double *arg2 = (double *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFCalculation4_force_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_force_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFCalculation4_force_d_set" "', argument " "2"" of type '" "double [3]""'");  }
    arg2 = reinterpret_cast< double * >(argp2); { if (arg2) { size_t ii = 0;
      for (; ii < (size_t)3; ++ii) *(double *)&arg1->force_d[ii] = *((double *)arg2 + ii); }  else {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in variable '""force_d""' of type '""double [3]""'"); }
     }  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFCalculation4_force_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFCalculation4_force_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); result = (double *)(double *) ((arg1)->force_d);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFCalculation4(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFCalculation4, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFCalculation4" "', argument " "1"" of type '" "OpenBabel::OBFFCalculation4 *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFCalculation4 * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFFCalculation4(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFCalculation4 *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFCalculation4", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFCalculation4 *)new OpenBabel::OBFFCalculation4();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFCalculation4, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFFCalculation4_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFCalculation4, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFCalculation4_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBFFConstraint_factor_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_factor_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_factor_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_factor_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->factor = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_factor_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_factor_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (double) ((arg1)->factor);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_constraint_value_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_constraint_value_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_constraint_value_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_constraint_value_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->constraint_value = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_constraint_value_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_constraint_value_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (double) ((arg1)->constraint_value);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_rab0_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_rab0_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_rab0_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_rab0_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->rab0 = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_rab0_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_rab0_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (double) ((arg1)->rab0);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_rbc0_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_rbc0_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_rbc0_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_rbc0_set" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); if (arg1) (arg1)->rbc0 = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_rbc0_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_rbc0_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (double) ((arg1)->rbc0);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_type_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_type_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_type_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->type = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_type_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_type_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (int) ((arg1)->type);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ia_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_ia_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ia_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_ia_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->ia = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ia_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ia_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (int) ((arg1)->ia);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ib_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_ib_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ib_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_ib_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->ib = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ib_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ib_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (int) ((arg1)->ib);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ic_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_ic_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ic_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_ic_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->ic = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_ic_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_ic_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (int) ((arg1)->ic);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_id_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_id_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_id_set" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); if (arg1) (arg1)->id = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_id_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_id_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (int) ((arg1)->id);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_a_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_a_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_a_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_a_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->a = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_a_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_a_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->a);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_b_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_b_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_b_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_b_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->b = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_b_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_b_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->b);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_c_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_c_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_c_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_c_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->c = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_c_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_c_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->c);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_d_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_d_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_d_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_d_set" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); if (arg1) (arg1)->d = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_d_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_d_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::OBAtom *) ((arg1)->d);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_grada_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_grada_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_grada_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_grada_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFConstraint_grada_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); if (arg1) (arg1)->grada = *arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_grada_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::vector3 *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_grada_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::vector3 *) & ((arg1)->grada);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradb_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_gradb_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradb_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_gradb_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFConstraint_gradb_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); if (arg1) (arg1)->gradb = *arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradb_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::vector3 *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradb_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::vector3 *) & ((arg1)->gradb);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradc_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_gradc_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradc_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_gradc_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFConstraint_gradc_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); if (arg1) (arg1)->gradc = *arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradc_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::vector3 *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradc_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::vector3 *) & ((arg1)->gradc);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradd_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_gradd_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradd_set" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraint_gradd_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFConstraint_gradd_set" "', argument " "2"" of type '" "OpenBabel::vector3 const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2); if (arg1) (arg1)->gradd = *arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_gradd_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::vector3 *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_gradd_get" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); result = (OpenBabel::vector3 *) & ((arg1)->gradd);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__vector3, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_new_OBFFConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFConstraint", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFConstraint *)new OpenBabel::OBFFConstraint();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFConstraint, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraint_GetGradient(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraint *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraint_GetGradient", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraint, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraint_GetGradient" "', argument " "1"" of type '" "OpenBabel::OBFFConstraint *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraint * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraint_GetGradient" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->GetGradient(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFFConstraint_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFConstraint, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFConstraint_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBFFConstraints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBFFConstraints", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBFFConstraints *)new OpenBabel::OBFFConstraints();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFConstraints, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBFFConstraints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBFFConstraints" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_Clear" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintEnergy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintEnergy" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); result = (double)(arg1)->GetConstraintEnergy();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetGradient(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetGradient", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetGradient" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetGradient" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->GetGradient(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_Setup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_Setup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_Setup" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBFFConstraints_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBFFConstraints_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->Setup(*arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_SetFactor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_SetFactor", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_SetFactor" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_SetFactor" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetFactor(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddIgnore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddIgnore", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddIgnore" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddIgnore" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->AddIgnore(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddAtomConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddAtomConstraint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddAtomConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddAtomConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->AddAtomConstraint(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddAtomXConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddAtomXConstraint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddAtomXConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddAtomXConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->AddAtomXConstraint(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddAtomYConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddAtomYConstraint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddAtomYConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddAtomYConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->AddAtomYConstraint(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddAtomZConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddAtomZConstraint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddAtomZConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddAtomZConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->AddAtomZConstraint(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddDistanceConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; int arg3 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddDistanceConstraint", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddDistanceConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddDistanceConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFFConstraints_AddDistanceConstraint" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBFFConstraints_AddDistanceConstraint" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); (arg1)->AddDistanceConstraint(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddAngleConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; double arg5 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; double val5 ; int ecode5 = 0 ;
  PyObject *swig_obj[5] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddAngleConstraint", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddAngleConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddAngleConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFFConstraints_AddAngleConstraint" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBFFConstraints_AddAngleConstraint" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBFFConstraints_AddAngleConstraint" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); (arg1)->AddAngleConstraint(arg2,arg3,arg4,arg5); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_AddTorsionConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; double arg6 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ;
  double val6 ; int ecode6 = 0 ; PyObject *swig_obj[6] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_AddTorsionConstraint", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBFFConstraints_AddTorsionConstraint" "', argument " "6"" of type '" "double""'"); }
    arg6 = static_cast< double >(val6); (arg1)->AddTorsionConstraint(arg2,arg3,arg4,arg5,arg6); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_DeleteConstraint(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_DeleteConstraint", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_DeleteConstraint" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_DeleteConstraint" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->DeleteConstraint(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetFactor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetFactor" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); result = (double)(arg1)->GetFactor();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_Size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_Size" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1);
  result = (int)((OpenBabel::OBFFConstraints const *)arg1)->Size(); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintType" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintType" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintType(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintValue(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintValue", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintValue" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintValue" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (double)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintValue(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintAtomA(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintAtomA", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintAtomA" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintAtomA" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintAtomA(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintAtomB(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintAtomB", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintAtomB" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintAtomB" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintAtomB(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintAtomC(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintAtomC", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintAtomC" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintAtomC" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintAtomC(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetConstraintAtomD(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_GetConstraintAtomD", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetConstraintAtomD" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_GetConstraintAtomD" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (int)((OpenBabel::OBFFConstraints const *)arg1)->GetConstraintAtomD(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_IsIgnored(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_IsIgnored", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_IsIgnored" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_IsIgnored" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsIgnored(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_IsFixed(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_IsFixed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_IsFixed" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_IsFixed" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsFixed(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_IsXFixed(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_IsXFixed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_IsXFixed" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_IsXFixed" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsXFixed(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_IsYFixed(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_IsYFixed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_IsYFixed" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_IsYFixed" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsYFixed(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_IsZFixed(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBFFConstraints_IsZFixed", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_IsZFixed" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBFFConstraints_IsZFixed" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->IsZFixed(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetIgnoredBitVec(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetIgnoredBitVec" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); result = (arg1)->GetIgnoredBitVec();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBFFConstraints_GetFixedBitVec(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBFFConstraints *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBFFConstraints_GetFixedBitVec" "', argument " "1"" of type '" "OpenBabel::OBFFConstraints *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp1); result = (arg1)->GetFixedBitVec();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBFFConstraints_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBFFConstraints, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBFFConstraints_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBForceField_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField **result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBForceField **) &OpenBabel::OBForceField::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBForceField, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBForceField *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_FindType" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBForceField *)OpenBabel::OBForceField::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_MakeNewInstance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBForceField *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_MakeNewInstance" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (OpenBabel::OBForceField *)(arg1)->MakeNewInstance();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBForceField(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBForceField" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FindForceField__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; OpenBabel::OBForceField *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_FindForceField" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_FindForceField" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  result = (OpenBabel::OBForceField *)OpenBabel::OBForceField::FindForceField((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FindForceField__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::OBForceField *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_FindForceField" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBForceField *)OpenBabel::OBForceField::FindForceField((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FindForceField(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_FindForceField", 0, 1, argv))) SWIG_fail; --argc;
  if (argc == 1) { int _v = 0; { int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBForceField_FindForceField__SWIG_0(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_FindForceField__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_FindForceField'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::FindForceField(std::string const &)\n"
  "    OpenBabel::OBForceField::FindForceField(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SetParameterFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetParameterFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetParameterFile" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_SetParameterFile" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_SetParameterFile" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; }  (arg1)->SetParameterFile((std::string const &)*arg2); resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetUnit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetUnit" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (arg1)->GetUnit();
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_HasAnalyticalGradients(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_HasAnalyticalGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->HasAnalyticalGradients();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Setup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Setup" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Setup(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Setup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBFFConstraints *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Setup" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBFFConstraints,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_Setup" "', argument " "3"" of type '" "OpenBabel::OBFFConstraints &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_Setup" "', argument " "3"" of type '" "OpenBabel::OBFFConstraints &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp3); result = (bool)(arg1)->Setup(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Setup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_Setup", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_Setup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_Setup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_Setup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::Setup(OpenBabel::OBMol &)\n"
  "    OpenBabel::OBForceField::Setup(OpenBabel::OBMol &,OpenBabel::OBFFConstraints &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_ParseParamFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ParseParamFile" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->ParseParamFile();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetTypes(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetTypes" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SetTypes();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetFormalCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetFormalCharges" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SetFormalCharges();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetPartialCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetPartialCharges" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SetPartialCharges();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetupCalculations(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetupCalculations" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SetupCalculations();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetupPointers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetupPointers" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SetupPointers();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IsSetupNeeded(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_IsSetupNeeded", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_IsSetupNeeded" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_IsSetupNeeded" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_IsSetupNeeded" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->IsSetupNeeded(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetAtomTypes(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_GetAtomTypes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetAtomTypes" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetAtomTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_GetAtomTypes" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->GetAtomTypes(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetPartialCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_GetPartialCharges", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetPartialCharges" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetPartialCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_GetPartialCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->GetPartialCharges(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_GetCoordinates", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetCoordinates" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_GetCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->GetCoordinates(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_UpdateCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_UpdateCoordinates", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_UpdateCoordinates" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_UpdateCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_UpdateCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->UpdateCoordinates(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_GetConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetConformers" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_GetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->GetConformers(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_UpdateConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_UpdateConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_UpdateConformers" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_UpdateConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_UpdateConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->UpdateConformers(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetCoordinates", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetCoordinates" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_SetCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_SetCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->SetCoordinates(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetConformers" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_SetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_SetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->SetConformers(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetGrid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; double arg3 ; char *arg4 = 0 ; double arg5 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int res4 ; char *buf4 = 0 ; int alloc4 = 0 ;
  double val5 ; int ecode5 = 0 ; PyObject *swig_obj[5] ; OpenBabel::OBGridData *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_GetGrid", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetGrid" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_GetGrid" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_GetGrid" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_GetGrid" "', argument " "4"" of type '" "char const *""'"); }
   arg4 = reinterpret_cast< char * >(buf4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBForceField_GetGrid" "', argument " "5"" of type '" "double""'"); }
    arg5 = static_cast< double >(val5); result = (OpenBabel::OBGridData *)(arg1)->GetGrid(arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGridData, 0 |  0 );
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return resultobj; fail: if (alloc4 == SWIG_NEWOBJ) delete[] buf4; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_AddIntraGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_AddIntraGroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_AddIntraGroup" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_AddIntraGroup" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_AddIntraGroup" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->AddIntraGroup(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_AddInterGroup(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_AddInterGroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_AddInterGroup" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_AddInterGroup" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_AddInterGroup" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->AddInterGroup(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_AddInterGroups(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; OpenBabel::OBBitVec *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_AddInterGroups", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_AddInterGroups" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_AddInterGroups" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_AddInterGroups" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_AddInterGroups" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_AddInterGroups" "', argument " "3"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3); (arg1)->AddInterGroups(*arg2,*arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ClearGroups(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ClearGroups" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->ClearGroups(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_HasGroups(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_HasGroups" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->HasGroups();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_EnableCutOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_EnableCutOff", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_EnableCutOff" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_EnableCutOff" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->EnableCutOff(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IsCutOffEnabled(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_IsCutOffEnabled" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->IsCutOffEnabled();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetVDWCutOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetVDWCutOff", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetVDWCutOff" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetVDWCutOff" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetVDWCutOff(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetVDWCutOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetVDWCutOff" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->GetVDWCutOff();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetElectrostaticCutOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetElectrostaticCutOff", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetElectrostaticCutOff" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetElectrostaticCutOff" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetElectrostaticCutOff(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetElectrostaticCutOff(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetElectrostaticCutOff" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->GetElectrostaticCutOff();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetDielectricConstant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetDielectricConstant", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetDielectricConstant" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetDielectricConstant" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetDielectricConstant(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetDielectricConstant(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetDielectricConstant" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->GetDielectricConstant();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetUpdateFrequency(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetUpdateFrequency", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetUpdateFrequency" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetUpdateFrequency" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetUpdateFrequency(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetUpdateFrequency(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetUpdateFrequency" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (int)(arg1)->GetUpdateFrequency();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_UpdatePairsSimple(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_UpdatePairsSimple" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->UpdatePairsSimple(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetNumPairs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetNumPairs" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (unsigned int)(arg1)->GetNumPairs();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetNumElectrostaticPairs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetNumElectrostaticPairs" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (unsigned int)(arg1)->GetNumElectrostaticPairs();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetNumVDWPairs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetNumVDWPairs" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (unsigned int)(arg1)->GetNumVDWPairs();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_EnableAllPairs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_EnableAllPairs" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->EnableAllPairs(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetGradient__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::vector3 result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetGradient" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetGradient" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_GetGradient" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); result = (arg1)->GetGradient(arg2,arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetGradient__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetGradient" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_GetGradient" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); result = (arg1)->GetGradient(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetGradient(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_GetGradient", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_GetGradient__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_GetGradient__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_GetGradient'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::GetGradient(OpenBabel::OBAtom *,int)\n"
  "    OpenBabel::OBForceField::GetGradient(OpenBabel::OBAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_GetGradientPtr(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetGradientPtr" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double *)(arg1)->GetGradientPtr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Energy__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Energy" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_Energy" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->Energy(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Energy__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Energy" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->Energy();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Energy(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_Energy", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_Energy__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_Energy__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_Energy'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::Energy(bool)\n" "    OpenBabel::OBForceField::Energy()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Bond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Bond" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_Bond" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_Bond(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Bond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Bond" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_Bond();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Bond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_Bond", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_Bond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_Bond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_Bond'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_Bond(bool)\n" "    OpenBabel::OBForceField::E_Bond()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Angle__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Angle" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_Angle" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_Angle(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Angle__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Angle" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_Angle();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Angle(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_Angle", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_Angle__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_Angle__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_Angle'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_Angle(bool)\n"
  "    OpenBabel::OBForceField::E_Angle()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_StrBnd__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_StrBnd" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_StrBnd" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_StrBnd(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_StrBnd__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_StrBnd" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_StrBnd();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_StrBnd(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_StrBnd", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_StrBnd__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_StrBnd__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_StrBnd'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_StrBnd(bool)\n"
  "    OpenBabel::OBForceField::E_StrBnd()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Torsion__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Torsion" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_Torsion" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_Torsion(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Torsion__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Torsion" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_Torsion();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Torsion(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_Torsion", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_Torsion__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_Torsion__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_Torsion'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_Torsion(bool)\n"
  "    OpenBabel::OBForceField::E_Torsion()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_OOP__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_OOP" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_OOP" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_OOP(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_OOP__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_OOP" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_OOP();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_OOP(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_OOP", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_OOP__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_OOP__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_OOP'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_OOP(bool)\n" "    OpenBabel::OBForceField::E_OOP()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_VDW__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_VDW" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_VDW" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_VDW(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_VDW__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_VDW" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_VDW();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_VDW(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_VDW", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBForceField_E_VDW__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_VDW__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_VDW'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_VDW(bool)\n" "    OpenBabel::OBForceField::E_VDW()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Electrostatic__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; double result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Electrostatic" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_E_Electrostatic" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (double)(arg1)->E_Electrostatic(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Electrostatic__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_E_Electrostatic" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (double)(arg1)->E_Electrostatic();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_E_Electrostatic(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_E_Electrostatic", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_E_Electrostatic__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_E_Electrostatic__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_E_Electrostatic'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::E_Electrostatic(bool)\n"
  "    OpenBabel::OBForceField::E_Electrostatic()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_PrintTypes(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_PrintTypes" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->PrintTypes(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_PrintFormalCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_PrintFormalCharges" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->PrintFormalCharges(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_PrintPartialCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_PrintPartialCharges" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->PrintPartialCharges(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_PrintVelocities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_PrintVelocities" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->PrintVelocities(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetLogFile(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetLogFile", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetLogFile" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_SetLogFile" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); result = (bool)(arg1)->SetLogFile(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetLogLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetLogLevel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetLogLevel" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetLogLevel" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->SetLogLevel(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetLogLevel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetLogLevel" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (int)(arg1)->GetLogLevel();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_OBFFLog__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_OBFFLog" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBForceField_OBFFLog" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->OBFFLog(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_OBFFLog__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 ;
  char *buf2 = 0 ; int alloc2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_OBFFLog" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_OBFFLog" "', argument " "2"" of type '" "char const *""'"); }
   arg2 = reinterpret_cast< char * >(buf2); (arg1)->OBFFLog((char const *)arg2); resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return resultobj; fail: if (alloc2 == SWIG_NEWOBJ) delete[] buf2; return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_OBFFLog(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_OBFFLog", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0)); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBForceField_OBFFLog__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_OBFFLog__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_OBFFLog'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::OBFFLog(std::string)\n"
  "    OpenBabel::OBForceField::OBFFLog(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_DistanceGeometry(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DistanceGeometry" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->DistanceGeometry(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); (arg1)->SystematicRotorSearch(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SystematicRotorSearch(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearch__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->SystematicRotorSearch(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearch(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_SystematicRotorSearch", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_SystematicRotorSearch__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_SystematicRotorSearch__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_SystematicRotorSearch__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_SystematicRotorSearch'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::SystematicRotorSearch(unsigned int,bool)\n"
  "    OpenBabel::OBForceField::SystematicRotorSearch(unsigned int)\n"
  "    OpenBabel::OBForceField::SystematicRotorSearch()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (int)(arg1)->SystematicRotorSearchInitialize(arg2,arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (int)(arg1)->SystematicRotorSearchInitialize(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (int)(arg1)->SystematicRotorSearchInitialize();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchInitialize(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_SystematicRotorSearchInitialize", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_SystematicRotorSearchInitialize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_SystematicRotorSearchInitialize'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::SystematicRotorSearchInitialize(unsigned int,bool)\n"
  "    OpenBabel::OBForceField::SystematicRotorSearchInitialize(unsigned int)\n"
  "    OpenBabel::OBForceField::SystematicRotorSearchInitialize()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchNextConformer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearchNextConformer" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SystematicRotorSearchNextConformer" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->SystematicRotorSearchNextConformer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchNextConformer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SystematicRotorSearchNextConformer" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->SystematicRotorSearchNextConformer();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SystematicRotorSearchNextConformer(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_SystematicRotorSearchNextConformer", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_SystematicRotorSearchNextConformer__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_SystematicRotorSearchNextConformer__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_SystematicRotorSearchNextConformer'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::SystematicRotorSearchNextConformer(unsigned int)\n"
  "    OpenBabel::OBForceField::SystematicRotorSearchNextConformer()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; bool arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; bool val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_RandomRotorSearch" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_RandomRotorSearch" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); (arg1)->RandomRotorSearch(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_RandomRotorSearch" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->RandomRotorSearch(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearch__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->RandomRotorSearch(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearch(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_RandomRotorSearch", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearch__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearch__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearch__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_RandomRotorSearch'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::RandomRotorSearch(unsigned int,unsigned int,bool)\n"
  "    OpenBabel::OBForceField::RandomRotorSearch(unsigned int,unsigned int)\n"
  "    OpenBabel::OBForceField::RandomRotorSearch(unsigned int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchInitialize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; bool arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; bool val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); (arg1)->RandomRotorSearchInitialize(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchInitialize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->RandomRotorSearchInitialize(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchInitialize__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearchInitialize" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->RandomRotorSearchInitialize(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchInitialize(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_RandomRotorSearchInitialize", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBForceField_RandomRotorSearchInitialize__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearchInitialize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearchInitialize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_RandomRotorSearchInitialize'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBForceField::RandomRotorSearchInitialize(unsigned int,unsigned int,bool)\n"
  "    OpenBabel::OBForceField::RandomRotorSearchInitialize(unsigned int,unsigned int)\n"
  "    OpenBabel::OBForceField::RandomRotorSearchInitialize(unsigned int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchNextConformer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearchNextConformer" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_RandomRotorSearchNextConformer" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); result = (bool)(arg1)->RandomRotorSearchNextConformer(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchNextConformer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_RandomRotorSearchNextConformer" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->RandomRotorSearchNextConformer();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_RandomRotorSearchNextConformer(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_RandomRotorSearchNextConformer", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_RandomRotorSearchNextConformer__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_RandomRotorSearchNextConformer__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_RandomRotorSearchNextConformer'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::RandomRotorSearchNextConformer(unsigned int)\n"
  "    OpenBabel::OBForceField::RandomRotorSearchNextConformer()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_WeightedRotorSearch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; bool arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; bool val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "4"" of type '" "bool""'"); }
    arg4 = static_cast< bool >(val4); (arg1)->WeightedRotorSearch(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_WeightedRotorSearch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_WeightedRotorSearch" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->WeightedRotorSearch(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_WeightedRotorSearch(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_WeightedRotorSearch", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBForceField_WeightedRotorSearch__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_WeightedRotorSearch__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_WeightedRotorSearch'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::WeightedRotorSearch(unsigned int,unsigned int,bool)\n"
  "    OpenBabel::OBForceField::WeightedRotorSearch(unsigned int,unsigned int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_FastRotorSearch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_FastRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_FastRotorSearch" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (int)(arg1)->FastRotorSearch(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FastRotorSearch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_FastRotorSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (int)(arg1)->FastRotorSearch();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_FastRotorSearch(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_FastRotorSearch", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_FastRotorSearch__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_FastRotorSearch__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_FastRotorSearch'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::FastRotorSearch(bool)\n"
  "    OpenBabel::OBForceField::FastRotorSearch()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_DiverseConfGen__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; unsigned int arg3 ; double arg4 ; bool arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; double val4 ;
  int ecode4 = 0 ; bool val5 ; int ecode5 = 0 ; int result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DiverseConfGen" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_DiverseConfGen" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_DiverseConfGen" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_DiverseConfGen" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBForceField_DiverseConfGen" "', argument " "5"" of type '" "bool""'"); }
    arg5 = static_cast< bool >(val5); result = (int)(arg1)->DiverseConfGen(arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_DiverseConfGen__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; unsigned int arg3 ; double arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; int result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DiverseConfGen" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_DiverseConfGen" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_DiverseConfGen" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_DiverseConfGen" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); result = (int)(arg1)->DiverseConfGen(arg2,arg3,arg4);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_DiverseConfGen__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; double val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; int result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DiverseConfGen" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_DiverseConfGen" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_DiverseConfGen" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); result = (int)(arg1)->DiverseConfGen(arg2,arg3);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_DiverseConfGen__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; int result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DiverseConfGen" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_DiverseConfGen" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); result = (int)(arg1)->DiverseConfGen(arg2);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_DiverseConfGen(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_DiverseConfGen", 0, 5, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBForceField_DiverseConfGen__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_DiverseConfGen__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_DiverseConfGen__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBForceField_DiverseConfGen__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_DiverseConfGen'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::DiverseConfGen(double,unsigned int,double,bool)\n"
  "    OpenBabel::OBForceField::DiverseConfGen(double,unsigned int,double)\n"
  "    OpenBabel::OBForceField::DiverseConfGen(double,unsigned int)\n" "    OpenBabel::OBForceField::DiverseConfGen(double)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SetLineSearchType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetLineSearchType", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetLineSearchType" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetLineSearchType" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetLineSearchType(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GetLineSearchType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetLineSearchType" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (int)(arg1)->GetLineSearchType();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LineSearch__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; OpenBabel::vector3 result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LineSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_LineSearch" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_LineSearch" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_LineSearch" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3); result = (arg1)->LineSearch(arg2,*arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LineSearch__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LineSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_LineSearch" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_LineSearch" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); result = (double)(arg1)->LineSearch(arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LineSearch(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_LineSearch", 0, 3, argv))) SWIG_fail; --argc; if (argc == 3) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_1; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_OBForceField_LineSearch__SWIG_0(self, argc, argv);}  check_1: if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_LineSearch__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_LineSearch'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::LineSearch(OpenBabel::OBAtom *,OpenBabel::vector3 &)\n"
  "    OpenBabel::OBForceField::LineSearch(double *,double *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_Newton2NumLineSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_Newton2NumLineSearch", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Newton2NumLineSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_Newton2NumLineSearch" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (double)(arg1)->Newton2NumLineSearch(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LineSearchTakeStep(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double val4 ; int ecode4 = 0 ; PyObject *swig_obj[4] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_LineSearchTakeStep", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LineSearchTakeStep" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_LineSearchTakeStep" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_LineSearchTakeStep" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_LineSearchTakeStep" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); (arg1)->LineSearchTakeStep(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescent__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescent" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescent" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SteepestDescent" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_SteepestDescent" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->SteepestDescent(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescent__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescent" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescent" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SteepestDescent" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->SteepestDescent(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescent__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescent" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescent" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SteepestDescent(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescent(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_SteepestDescent", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBForceField_SteepestDescent__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_SteepestDescent__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_SteepestDescent__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_SteepestDescent'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::SteepestDescent(int,double,int)\n"
  "    OpenBabel::OBForceField::SteepestDescent(int,double)\n" "    OpenBabel::OBForceField::SteepestDescent(int)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentInitialize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->SteepestDescentInitialize(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentInitialize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->SteepestDescentInitialize(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentInitialize__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SteepestDescentInitialize(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentInitialize__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescentInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->SteepestDescentInitialize(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentInitialize(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_SteepestDescentInitialize", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_SteepestDescentInitialize__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_SteepestDescentInitialize__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_SteepestDescentInitialize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_SteepestDescentInitialize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_SteepestDescentInitialize'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::SteepestDescentInitialize(int,double,int)\n"
  "    OpenBabel::OBForceField::SteepestDescentInitialize(int,double)\n"
  "    OpenBabel::OBForceField::SteepestDescentInitialize(int)\n" "    OpenBabel::OBForceField::SteepestDescentInitialize()\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_SteepestDescentTakeNSteps(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SteepestDescentTakeNSteps", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SteepestDescentTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SteepestDescentTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->SteepestDescentTakeNSteps(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradients__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradients" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_ConjugateGradients" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_ConjugateGradients" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->ConjugateGradients(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradients__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradients" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_ConjugateGradients" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->ConjugateGradients(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradients__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradients" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->ConjugateGradients(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradients(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_ConjugateGradients", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBForceField_ConjugateGradients__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_ConjugateGradients__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_ConjugateGradients__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_ConjugateGradients'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::ConjugateGradients(int,double,int)\n"
  "    OpenBabel::OBForceField::ConjugateGradients(int,double)\n" "    OpenBabel::OBForceField::ConjugateGradients(int)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsInitialize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->ConjugateGradientsInitialize(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsInitialize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->ConjugateGradientsInitialize(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsInitialize__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->ConjugateGradientsInitialize(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsInitialize__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradientsInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->ConjugateGradientsInitialize();
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsInitialize(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_ConjugateGradientsInitialize", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_ConjugateGradientsInitialize__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_ConjugateGradientsInitialize__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_ConjugateGradientsInitialize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_ConjugateGradientsInitialize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_ConjugateGradientsInitialize'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::ConjugateGradientsInitialize(int,double,int)\n"
  "    OpenBabel::OBForceField::ConjugateGradientsInitialize(int,double)\n"
  "    OpenBabel::OBForceField::ConjugateGradientsInitialize(int)\n"
  "    OpenBabel::OBForceField::ConjugateGradientsInitialize()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_ConjugateGradientsTakeNSteps(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_ConjugateGradientsTakeNSteps", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ConjugateGradientsTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ConjugateGradientsTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->ConjugateGradientsTakeNSteps(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGS__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGS" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGS" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_LBFGS" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_LBFGS" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); (arg1)->LBFGS(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGS__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGS" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGS" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_LBFGS" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->LBFGS(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGS__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGS" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGS" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->LBFGS(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGS(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_LBFGS", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_LBFGS__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_LBFGS__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_LBFGS__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_LBFGS'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::LBFGS(int,double,int)\n"
  "    OpenBabel::OBForceField::LBFGS(int,double)\n" "    OpenBabel::OBForceField::LBFGS(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSInitialize__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGSInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGSInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_LBFGSInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_LBFGSInitialize" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); (arg1)->LBFGSInitialize(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSInitialize__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGSInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGSInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_LBFGSInitialize" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->LBFGSInitialize(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSInitialize__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGSInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGSInitialize" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->LBFGSInitialize(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSInitialize__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGSInitialize" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->LBFGSInitialize(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSInitialize(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_LBFGSInitialize", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBForceField_LBFGSInitialize__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_LBFGSInitialize__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_LBFGSInitialize__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_LBFGSInitialize__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_LBFGSInitialize'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::LBFGSInitialize(int,double,int)\n"
  "    OpenBabel::OBForceField::LBFGSInitialize(int,double)\n" "    OpenBabel::OBForceField::LBFGSInitialize(int)\n"
  "    OpenBabel::OBForceField::LBFGSInitialize()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_LBFGSTakeNSteps(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_LBFGSTakeNSteps", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_LBFGSTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_LBFGSTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)(arg1)->LBFGSTakeNSteps(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_GenerateVelocities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GenerateVelocities" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->GenerateVelocities(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_CorrectVelocities(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_CorrectVelocities" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->CorrectVelocities(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; double arg4 ; int arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ;
  int val5 ; int ecode5 = 0 ; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5); (arg1)->MolecularDynamicsTakeNSteps(arg2,arg3,arg4,arg5); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; double arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "4"" of type '" "double""'"); }
    arg4 = static_cast< double >(val4); (arg1)->MolecularDynamicsTakeNSteps(arg2,arg3,arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_MolecularDynamicsTakeNSteps" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->MolecularDynamicsTakeNSteps(arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_MolecularDynamicsTakeNSteps(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_MolecularDynamicsTakeNSteps", 0, 5, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBForceField_MolecularDynamicsTakeNSteps__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_MolecularDynamicsTakeNSteps'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::MolecularDynamicsTakeNSteps(int,double,double,int)\n"
  "    OpenBabel::OBForceField::MolecularDynamicsTakeNSteps(int,double,double)\n"
  "    OpenBabel::OBForceField::MolecularDynamicsTakeNSteps(int,double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_GetConstraints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBFFConstraints *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_GetConstraints" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  result = (OpenBabel::OBFFConstraints *) &(arg1)->GetConstraints();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBFFConstraints, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetConstraints(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBFFConstraints *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetConstraints", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetConstraints" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBFFConstraints,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_SetConstraints" "', argument " "2"" of type '" "OpenBabel::OBFFConstraints &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_SetConstraints" "', argument " "2"" of type '" "OpenBabel::OBFFConstraints &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBFFConstraints * >(argp2); (arg1)->SetConstraints(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetFixAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetFixAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetFixAtom" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetFixAtom" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetFixAtom(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_UnsetFixAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_UnsetFixAtom" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->UnsetFixAtom(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetIgnoreAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_SetIgnoreAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetIgnoreAtom" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_SetIgnoreAtom" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetIgnoreAtom(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_UnsetIgnoreAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_UnsetIgnoreAtom" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); (arg1)->UnsetIgnoreAtom(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IgnoreCalculation__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; int arg2 ; int val1 ; int ecode1 = 0 ; int val2 ; int ecode2 = 0 ; bool result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBForceField_IgnoreCalculation" "', argument " "1"" of type '" "int""'"); }
    arg1 = static_cast< int >(val1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_IgnoreCalculation" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (bool)OpenBabel::OBForceField::IgnoreCalculation(arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IgnoreCalculation__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; int arg2 ; int arg3 ; int val1 ; int ecode1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ;
  int ecode3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBForceField_IgnoreCalculation" "', argument " "1"" of type '" "int""'"); }
    arg1 = static_cast< int >(val1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_IgnoreCalculation" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_IgnoreCalculation" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); result = (bool)OpenBabel::OBForceField::IgnoreCalculation(arg1,arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IgnoreCalculation__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; int arg2 ; int arg3 ; int arg4 ; int val1 ; int ecode1 = 0 ; int val2 ; int ecode2 = 0 ;
  int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBForceField_IgnoreCalculation" "', argument " "1"" of type '" "int""'"); }
    arg1 = static_cast< int >(val1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_IgnoreCalculation" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBForceField_IgnoreCalculation" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBForceField_IgnoreCalculation" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); result = (bool)OpenBabel::OBForceField::IgnoreCalculation(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_IgnoreCalculation(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_IgnoreCalculation", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBForceField_IgnoreCalculation__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBForceField_IgnoreCalculation__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBForceField_IgnoreCalculation__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_IgnoreCalculation'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBForceField::IgnoreCalculation(int,int)\n"
  "    OpenBabel::OBForceField::IgnoreCalculation(int,int,int)\n"
  "    OpenBabel::OBForceField::IgnoreCalculation(int,int,int,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_DetectExplosion(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_DetectExplosion" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->DetectExplosion();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ValidateLineSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ;
  OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_ValidateLineSearch", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ValidateLineSearch" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_ValidateLineSearch" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_ValidateLineSearch" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_ValidateLineSearch" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3); result = (arg1)->ValidateLineSearch(arg2,*arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ValidateSteepestDescent(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_ValidateSteepestDescent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ValidateSteepestDescent" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ValidateSteepestDescent" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->ValidateSteepestDescent(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ValidateConjugateGradients(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_ValidateConjugateGradients", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ValidateConjugateGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_ValidateConjugateGradients" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->ValidateConjugateGradients(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_Validate(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_Validate" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->Validate();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ValidateGradients(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ValidateGradients" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); result = (bool)(arg1)->ValidateGradients();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_ValidateGradientError(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ;
  OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_ValidateGradientError", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_ValidateGradientError" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_ValidateGradientError" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_ValidateGradientError" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_ValidateGradientError" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_ValidateGradientError" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3); result = (arg1)->ValidateGradientError(*arg2,*arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorBondDerivative(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; double result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorBondDerivative", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorBondDerivative" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorBondDerivative" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorBondDerivative" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorBondDerivative" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4);
  result = (double)OpenBabel::OBForceField::VectorBondDerivative(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorDistanceDerivative(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = (double *)0 ; double *arg2 = (double *)0 ; double *arg3 = 0 ; double *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorDistanceDerivative", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorDistanceDerivative" "', argument " "1"" of type '" "double const *const""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorDistanceDerivative" "', argument " "2"" of type '" "double const *const""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorDistanceDerivative" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorDistanceDerivative" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4);
  result = (double)OpenBabel::OBForceField::VectorDistanceDerivative((double const *)arg1,(double const *)arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorLengthDerivative(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorLengthDerivative", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorLengthDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorLengthDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorLengthDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorLengthDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  result = (double)OpenBabel::OBForceField::VectorLengthDerivative(*arg1,*arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorAngleDerivative__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; double *arg5 = 0 ;
  double *arg6 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; void *argp6 = 0 ; int res6 = 0 ; double result; (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "5"" of type '" "double *""'");  }
   arg5 = reinterpret_cast< double * >(argp5); res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "6"" of type '" "double *""'");  }
   arg6 = reinterpret_cast< double * >(argp6);
  result = (double)OpenBabel::OBForceField::VectorAngleDerivative(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorAngleDerivative__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorAngleDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorAngleDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorAngleDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorAngleDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  result = (double)OpenBabel::OBForceField::VectorAngleDerivative(*arg1,*arg2,*arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorAngleDerivative(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_VectorAngleDerivative", 0, 6, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBForceField_VectorAngleDerivative__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_OBForceField_VectorAngleDerivative__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_VectorAngleDerivative'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBForceField::VectorAngleDerivative(double *,double *,double *,double *,double *,double *)\n"
  "    OpenBabel::OBForceField::VectorAngleDerivative(OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorOOPDerivative__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; double *arg5 = 0 ;
  double *arg6 = 0 ; double *arg7 = 0 ; double *arg8 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; void *argp6 = 0 ;
  int res6 = 0 ; void *argp7 = 0 ; int res7 = 0 ; void *argp8 = 0 ; int res8 = 0 ; double result; (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "5"" of type '" "double *""'");  }
   arg5 = reinterpret_cast< double * >(argp5); res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "6"" of type '" "double *""'");  }
   arg6 = reinterpret_cast< double * >(argp6); res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "7"" of type '" "double *""'");  }
   arg7 = reinterpret_cast< double * >(argp7); res8 = SWIG_ConvertPtr(swig_obj[7], &argp8,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "8"" of type '" "double *""'");  }
   arg8 = reinterpret_cast< double * >(argp8);
  result = (double)OpenBabel::OBForceField::VectorOOPDerivative(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorOOPDerivative__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  OpenBabel::vector3 *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; double result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorOOPDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorOOPDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorOOPDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorOOPDerivative" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorOOPDerivative" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4);
  result = (double)OpenBabel::OBForceField::VectorOOPDerivative(*arg1,*arg2,*arg3,*arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorOOPDerivative(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[9] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_VectorOOPDerivative", 0, 8, argv))) SWIG_fail; --argc;
  if (argc == 4) { PyObject *retobj = _wrap_OBForceField_VectorOOPDerivative__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 8) {
    PyObject *retobj = _wrap_OBForceField_VectorOOPDerivative__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_VectorOOPDerivative'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBForceField::VectorOOPDerivative(double *,double *,double *,double *,double *,double *,double *,double *)\n"
  "    OpenBabel::OBForceField::VectorOOPDerivative(OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorTorsionDerivative__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; double *arg5 = 0 ;
  double *arg6 = 0 ; double *arg7 = 0 ; double *arg8 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; void *argp6 = 0 ;
  int res6 = 0 ; void *argp7 = 0 ; int res7 = 0 ; void *argp8 = 0 ; int res8 = 0 ; double result; (void)self;
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "5"" of type '" "double *""'");  }
   arg5 = reinterpret_cast< double * >(argp5); res6 = SWIG_ConvertPtr(swig_obj[5], &argp6,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "6"" of type '" "double *""'");  }
   arg6 = reinterpret_cast< double * >(argp6); res7 = SWIG_ConvertPtr(swig_obj[6], &argp7,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "7"" of type '" "double *""'");  }
   arg7 = reinterpret_cast< double * >(argp7); res8 = SWIG_ConvertPtr(swig_obj[7], &argp8,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "8"" of type '" "double *""'");  }
   arg8 = reinterpret_cast< double * >(argp8);
  result = (double)OpenBabel::OBForceField::VectorTorsionDerivative(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorTorsionDerivative__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::vector3 *arg1 = 0 ; OpenBabel::vector3 *arg2 = 0 ; OpenBabel::vector3 *arg3 = 0 ;
  OpenBabel::vector3 *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; double result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "1"" of type '" "OpenBabel::vector3 &""'"); }
   arg1 = reinterpret_cast< OpenBabel::vector3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "2"" of type '" "OpenBabel::vector3 &""'"); }
   arg2 = reinterpret_cast< OpenBabel::vector3 * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "3"" of type '" "OpenBabel::vector3 &""'"); }
   arg3 = reinterpret_cast< OpenBabel::vector3 * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBForceField_VectorTorsionDerivative" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4);
  result = (double)OpenBabel::OBForceField::VectorTorsionDerivative(*arg1,*arg2,*arg3,*arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorTorsionDerivative(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[9] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBForceField_VectorTorsionDerivative", 0, 8, argv))) SWIG_fail; --argc;
  if (argc == 4) { PyObject *retobj = _wrap_OBForceField_VectorTorsionDerivative__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 8) {
    PyObject *retobj = _wrap_OBForceField_VectorTorsionDerivative__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBForceField_VectorTorsionDerivative'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBForceField::VectorTorsionDerivative(double *,double *,double *,double *,double *,double *,double *,double *)\n"
  "    OpenBabel::OBForceField::VectorTorsionDerivative(OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &,OpenBabel::vector3 &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorSubtract(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorSubtract", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorSubtract" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorSubtract" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorSubtract" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); OpenBabel::OBForceField::VectorSubtract(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorAdd(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorAdd", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorAdd" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorAdd" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorAdd" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); OpenBabel::OBForceField::VectorAdd(arg1,arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorDivide(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double arg2 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorDivide", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorDivide" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_VectorDivide" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorDivide" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); OpenBabel::OBForceField::VectorDivide(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorMultiply(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double arg2 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorMultiply", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorMultiply" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_VectorMultiply" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorMultiply" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); OpenBabel::OBForceField::VectorMultiply(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorSelfMultiply(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorSelfMultiply", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorSelfMultiply" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBForceField_VectorSelfMultiply" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); OpenBabel::OBForceField::VectorSelfMultiply(arg1,arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorNormalize(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorNormalize" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); OpenBabel::OBForceField::VectorNormalize(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorCopy(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorCopy", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorCopy" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorCopy" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); OpenBabel::OBForceField::VectorCopy(arg1,arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorLength(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorLength" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); result = (double)OpenBabel::OBForceField::VectorLength(arg1);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorDistance(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorDistance", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorDistance" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorDistance" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (double)OpenBabel::OBForceField::VectorDistance(arg1,arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorAngle", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorAngle" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorAngle" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorAngle" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); result = (double)OpenBabel::OBForceField::VectorAngle(arg1,arg2,arg3);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorTorsion(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; double result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorTorsion", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorTorsion" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorTorsion" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorTorsion" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorTorsion" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4); result = (double)OpenBabel::OBForceField::VectorTorsion(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorOOP(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; double *arg3 = 0 ; double *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; PyObject *swig_obj[4] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorOOP", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorOOP" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorOOP" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorOOP" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBForceField_VectorOOP" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4); result = (double)OpenBabel::OBForceField::VectorOOP(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorClear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorClear" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); OpenBabel::OBForceField::VectorClear(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorDot(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; double result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorDot", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorDot" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorDot" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (double)OpenBabel::OBForceField::VectorDot(arg1,arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_VectorCross(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; double *arg2 = 0 ; double *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBForceField_VectorCross", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_VectorCross" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBForceField_VectorCross" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_double, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBForceField_VectorCross" "', argument " "3"" of type '" "double *""'");  }
   arg3 = reinterpret_cast< double * >(argp3); OpenBabel::OBForceField::VectorCross(arg1,arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_PrintVector(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  double *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_PrintVector" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); OpenBabel::OBForceField::PrintVector(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetLogToStdOut(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetLogToStdOut" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); OpenBabel_OBForceField_SetLogToStdOut(arg1);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBForceField_SetLogToStdErr(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBForceField *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBForceField, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBForceField_SetLogToStdErr" "', argument " "1"" of type '" "OpenBabel::OBForceField *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBForceField * >(argp1); OpenBabel_OBForceField_SetLogToStdErr(arg1);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBForceField_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBForceField, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBBuilder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBBuilder", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBBuilder *)new OpenBabel::OBBuilder();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBuilder, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Build__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBuilder *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Build" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBuilder_Build" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Build" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Build" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->Build(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Build__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBBuilder *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Build" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBuilder_Build" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Build" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Build(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Build(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBuilder_Build", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBBuilder_Build__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBBuilder_Build__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBuilder_Build'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBuilder::Build(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBBuilder::Build(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBuilder_SetKeepRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_SetKeepRings" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1); (arg1)->SetKeepRings(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_UnsetKeepRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_UnsetKeepRings" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1); (arg1)->UnsetKeepRings(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_AddRingFragment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; OpenBabel::OBSmartsPattern *arg2 = 0 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBuilder_AddRingFragment", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_AddRingFragment" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBuilder_AddRingFragment" "', argument " "2"" of type '" "OpenBabel::OBSmartsPattern *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBSmartsPattern * >(argp2); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBBuilder_AddRingFragment" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_AddRingFragment" "', argument " "3"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg3 = ptr; } 
  (arg1)->AddRingFragment(arg2,(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_LoadFragments(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_LoadFragments" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1); (arg1)->LoadFragments(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_GetFragmentCoord(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; std::string arg2 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[2] ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBuilder_GetFragmentCoord", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_GetFragmentCoord" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1); { std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBBuilder_GetFragmentCoord" "', argument " "2"" of type '" "std::string""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  result = (arg1)->GetFragmentCoord(SWIG_STD_MOVE(arg2));
  resultobj = swig::from(static_cast< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_GetNewBondVector__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::vector3 result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_GetNewBondVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = OpenBabel::OBBuilder::GetNewBondVector(arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_GetNewBondVector__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; OpenBabel::vector3 result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_GetNewBondVector" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_GetNewBondVector" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); result = OpenBabel::OBBuilder::GetNewBondVector(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_GetNewBondVector(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBBuilder_GetNewBondVector", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBBuilder_GetNewBondVector__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBuilder_GetNewBondVector__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBuilder_GetNewBondVector'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBuilder::GetNewBondVector(OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBBuilder::GetNewBondVector(OpenBabel::OBAtom *,double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBuilder_Connect__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; OpenBabel::vector3 *arg4 = 0 ; int arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  int val5 ; int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_Connect" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Connect" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBBuilder_Connect" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBBuilder_Connect" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); result = (bool)OpenBabel::OBBuilder::Connect(*arg1,arg2,arg3,*arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Connect__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; OpenBabel::vector3 *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; bool result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_Connect" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Connect" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__vector3,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBBuilder_Connect" "', argument " "4"" of type '" "OpenBabel::vector3 &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "4"" of type '" "OpenBabel::vector3 &""'"); }
   arg4 = reinterpret_cast< OpenBabel::vector3 * >(argp4); result = (bool)OpenBabel::OBBuilder::Connect(*arg1,arg2,arg3,*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Connect__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_Connect" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Connect" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBBuilder_Connect" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); result = (bool)OpenBabel::OBBuilder::Connect(*arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Connect__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Connect" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_Connect" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Connect" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (bool)OpenBabel::OBBuilder::Connect(*arg1,arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_Connect(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBBuilder_Connect", 0, 5, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBBuilder_Connect__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_NO_NULL); _v = SWIG_CheckState(res);}
     if (!_v) goto check_2; return _wrap_OBBuilder_Connect__SWIG_1(self, argc, argv);}  check_2: if (argc == 4) {
    PyObject *retobj = _wrap_OBBuilder_Connect__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBBuilder_Connect__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBuilder_Connect'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBBuilder::Connect(OpenBabel::OBMol &,int,int,OpenBabel::vector3 &,int)\n"
  "    OpenBabel::OBBuilder::Connect(OpenBabel::OBMol &,int,int,OpenBabel::vector3 &)\n"
  "    OpenBabel::OBBuilder::Connect(OpenBabel::OBMol &,int,int,int)\n"
  "    OpenBabel::OBBuilder::Connect(OpenBabel::OBMol &,int,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBuilder_Swap(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; int arg2 ; int arg3 ; int arg4 ; int arg5 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; int val5 ; int ecode5 = 0 ; PyObject *swig_obj[5] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBuilder_Swap", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_Swap" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_Swap" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_Swap" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBBuilder_Swap" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBBuilder_Swap" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBBuilder_Swap" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); result = (bool)OpenBabel::OBBuilder::Swap(*arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_CorrectStereoBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_CorrectStereoBonds" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_CorrectStereoBonds" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)OpenBabel::OBBuilder::CorrectStereoBonds(*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_CorrectStereoAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_CorrectStereoAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_CorrectStereoAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBBuilder_CorrectStereoAtoms" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); result = (bool)OpenBabel::OBBuilder::CorrectStereoAtoms(*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_CorrectStereoAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; bool result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_CorrectStereoAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_CorrectStereoAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (bool)OpenBabel::OBBuilder::CorrectStereoAtoms(*arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_CorrectStereoAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBBuilder_CorrectStereoAtoms", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBBuilder_CorrectStereoAtoms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBBuilder_CorrectStereoAtoms__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBBuilder_CorrectStereoAtoms'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBBuilder::CorrectStereoAtoms(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBBuilder::CorrectStereoAtoms(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBBuilder_IsSpiroAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0; unsigned long arg1 ;
  OpenBabel::OBMol *arg2 = 0 ; unsigned long val1 ; int ecode1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBBuilder_IsSpiroAtom", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBBuilder_IsSpiroAtom" "', argument " "1"" of type '" "unsigned long""'"); }
    arg1 = static_cast< unsigned long >(val1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBuilder_IsSpiroAtom" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_IsSpiroAtom" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)OpenBabel::OBBuilder::IsSpiroAtom(arg1,*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_GetFragment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_GetFragment" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1); result = OpenBabel::OBBuilder::GetFragment(arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBBitVec(result)), SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBBuilder_AddNbrs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBitVec *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBBuilder_AddNbrs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBBuilder_AddNbrs" "', argument " "1"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBBuilder_AddNbrs" "', argument " "1"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBBitVec * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBBuilder_AddNbrs" "', argument " "2"" of type '" "OpenBabel::OBAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2); OpenBabel::OBBuilder::AddNbrs(*arg1,arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBBuilder(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBuilder *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBuilder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBBuilder" "', argument " "1"" of type '" "OpenBabel::OBBuilder *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBuilder * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBBuilder_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBBuilder, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBBuilder_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBOp_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOp **result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBOp_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBOp **) &OpenBabel::OBOp::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBOp, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ;
  char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBOp *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_FindType" "', argument " "1"" of type '" "char const *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); result = (OpenBabel::OBOp *)OpenBabel::OBOp::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_Do__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBOp *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; char *arg3 = 0 ; OpenBabel::OBOp::OpMap *arg4 = 0 ;
  OpenBabel::OBConversion *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ;
  char *buf3 = 0 ; int alloc3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ; int res5 = 0 ; bool result; (void)self;
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_Do" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  } 
  arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_Do" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBOp_Do" "', argument " "3"" of type '" "char const *""'"); } 
  arg3 = reinterpret_cast< char * >(buf3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBOp_Do" "', argument " "4"" of type '" "OpenBabel::OBOp::OpMap *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBOp::OpMap * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBOp_Do" "', argument " "5"" of type '" "OpenBabel::OBConversion *""'");  }
   arg5 = reinterpret_cast< OpenBabel::OBConversion * >(argp5); result = (bool)(arg1)->Do(arg2,(char const *)arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_Do__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBOp *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; char *arg3 = 0 ; OpenBabel::OBOp::OpMap *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ;
  void *argp4 = 0 ; int res4 = 0 ; bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_Do" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  } 
  arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_Do" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBOp_Do" "', argument " "3"" of type '" "char const *""'"); } 
  arg3 = reinterpret_cast< char * >(buf3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBOp_Do" "', argument " "4"" of type '" "OpenBabel::OBOp::OpMap *""'");  }
   arg4 = reinterpret_cast< OpenBabel::OBOp::OpMap * >(argp4); result = (bool)(arg1)->Do(arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_Do__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBOp *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_Do" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  } 
  arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_Do" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBOp_Do" "', argument " "3"" of type '" "char const *""'"); } 
  arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->Do(arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_Do__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBOp *arg1 = 0 ; OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_Do" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  } 
  arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_Do" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (bool)(arg1)->Do(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_Do(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBOp_Do", 0, 5, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBOp_Do__SWIG_3(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_OBOp_Do__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBOp_Do__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 5) { PyObject *retobj = _wrap_OBOp_Do__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBOp_Do'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBOp::Do(OpenBabel::OBBase *,char const *,OpenBabel::OBOp::OpMap *,OpenBabel::OBConversion *)\n"
  "    OpenBabel::OBOp::Do(OpenBabel::OBBase *,char const *,OpenBabel::OBOp::OpMap *)\n"
  "    OpenBabel::OBOp::Do(OpenBabel::OBBase *,char const *)\n" "    OpenBabel::OBOp::Do(OpenBabel::OBBase *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBOp_WorksWith(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBOp *arg1 = 0 ;
  OpenBabel::OBBase *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBOp_WorksWith", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_WorksWith" "', argument " "1"" of type '" "OpenBabel::OBOp const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_WorksWith" "', argument " "2"" of type '" "OpenBabel::OBBase *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBase * >(argp2); result = (bool)((OpenBabel::OBOp const *)arg1)->WorksWith(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_ProcessVec(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBOp *arg1 = 0 ; std::vector< OpenBabel::OBBase *,std::allocator< OpenBabel::OBBase * > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOp_ProcessVec", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_ProcessVec" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_ProcessVec" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBase *,std::allocator< OpenBabel::OBBase * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBOp_ProcessVec" "', argument " "2"" of type '" "std::vector< OpenBabel::OBBase *,std::allocator< OpenBabel::OBBase * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< OpenBabel::OBBase *,std::allocator< OpenBabel::OBBase * > > * >(argp2);
  result = (bool)(arg1)->ProcessVec(*arg2); resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_OpOptions(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_OpOptions" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1); result = OpenBabel::OBOp::OpOptions(arg1);
  resultobj = SWIG_From_std_string(static_cast< std::string >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBOp_DoOps(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBBase *arg1 = 0 ;
  OpenBabel::OBOp::OpMap *arg2 = 0 ; OpenBabel::OBConversion *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBOp_DoOps", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBOp_DoOps" "', argument " "1"" of type '" "OpenBabel::OBBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBOp_DoOps" "', argument " "2"" of type '" "OpenBabel::OBOp::OpMap *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBOp::OpMap * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBConversion, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBOp_DoOps" "', argument " "3"" of type '" "OpenBabel::OBConversion *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBConversion * >(argp3);
  result = (bool)OpenBabel::OBOp::DoOps(arg1,(std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > const *)arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBOp(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBOp *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBOp, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBOp" "', argument " "1"" of type '" "OpenBabel::OBOp *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBOp * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBOp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBOp, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_OBChargeModel_Default(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChargeModel **result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBChargeModel_Default", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBChargeModel **) &OpenBabel::OBChargeModel::Default();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_p_OpenBabel__OBChargeModel, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBChargeModel__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; bool arg2 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; bool val2 ;
  int ecode2 = 0 ; OpenBabel::OBChargeModel *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBChargeModel" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBChargeModel" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (OpenBabel::OBChargeModel *)new OpenBabel::OBChargeModel((char const *)arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBChargeModel, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBChargeModel__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; char *arg1 = 0 ; int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; OpenBabel::OBChargeModel *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBChargeModel" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBChargeModel *)new OpenBabel::OBChargeModel((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBChargeModel, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBChargeModel(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBChargeModel", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBChargeModel__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBChargeModel__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBChargeModel'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBChargeModel::OBChargeModel(char const *,bool)\n"
  "    OpenBabel::OBChargeModel::OBChargeModel(char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBChargeModel_FindType(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int res1 ; char *buf1 = 0 ; int alloc1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBChargeModel *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_FindType" "', argument " "1"" of type '" "char const *""'"); }
   arg1 = reinterpret_cast< char * >(buf1);
  result = (OpenBabel::OBChargeModel *)OpenBabel::OBChargeModel::FindType((char const *)arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_OBChargeModel_ComputeCharges__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBChargeModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_ComputeCharges" "', argument " "1"" of type '" "OpenBabel::OBChargeModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChargeModel_ComputeCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChargeModel_ComputeCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->ComputeCharges(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChargeModel_ComputeCharges__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBChargeModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; char *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 ; char *buf3 = 0 ; int alloc3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_ComputeCharges" "', argument " "1"" of type '" "OpenBabel::OBChargeModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChargeModel_ComputeCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChargeModel_ComputeCharges" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBChargeModel_ComputeCharges" "', argument " "3"" of type '" "char const *""'"); }
   arg3 = reinterpret_cast< char * >(buf3); result = (bool)(arg1)->ComputeCharges(*arg2,(char const *)arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return resultobj; fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3; return NULL; }
SWIGINTERN PyObject *_wrap_OBChargeModel_ComputeCharges(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBChargeModel_ComputeCharges", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBChargeModel_ComputeCharges__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBChargeModel_ComputeCharges__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBChargeModel_ComputeCharges'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBChargeModel::ComputeCharges(OpenBabel::OBMol &)\n"
  "    OpenBabel::OBChargeModel::ComputeCharges(OpenBabel::OBMol &,char const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBChargeModel_GetFormalCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChargeModel *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_GetFormalCharges" "', argument " "1"" of type '" "OpenBabel::OBChargeModel const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) &((OpenBabel::OBChargeModel const *)arg1)->GetFormalCharges();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBChargeModel_GetPartialCharges(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChargeModel *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_GetPartialCharges" "', argument " "1"" of type '" "OpenBabel::OBChargeModel const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) &((OpenBabel::OBChargeModel const *)arg1)->GetPartialCharges();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBChargeModel_GetDipoleMoment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChargeModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::vector3 result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBChargeModel_GetDipoleMoment", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChargeModel_GetDipoleMoment" "', argument " "1"" of type '" "OpenBabel::OBChargeModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChargeModel_GetDipoleMoment" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChargeModel_GetDipoleMoment" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (arg1)->GetDipoleMoment(*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::vector3(result)), SWIGTYPE_p_OpenBabel__vector3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBChargeModel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChargeModel *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChargeModel, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBChargeModel" "', argument " "1"" of type '" "OpenBabel::OBChargeModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChargeModel * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBChargeModel_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBChargeModel, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBChargeModel_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBChemTsfm(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBChemTsfm", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBChemTsfm *)new OpenBabel::OBChemTsfm();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBChemTsfm, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBChemTsfm(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChemTsfm, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBChemTsfm" "', argument " "1"" of type '" "OpenBabel::OBChemTsfm *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChemTsfm * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBChemTsfm_Init(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *arg1 = 0 ; std::string *arg2 = 0 ; std::string *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBChemTsfm_Init", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChemTsfm, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChemTsfm_Init" "', argument " "1"" of type '" "OpenBabel::OBChemTsfm *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChemTsfm * >(argp1); res2 = SWIG_AsPtr_std_string(swig_obj[1], &arg2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChemTsfm_Init" "', argument " "2"" of type '" "std::string &""'"); }
   if (!arg2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChemTsfm_Init" "', argument " "2"" of type '" "std::string &""'"); }
   res2 = SWIG_AddTmpMask(res2); res3 = SWIG_AsPtr_std_string(swig_obj[2], &arg3); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBChemTsfm_Init" "', argument " "3"" of type '" "std::string &""'"); }
   if (!arg3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChemTsfm_Init" "', argument " "3"" of type '" "std::string &""'"); }
   res3 = SWIG_AddTmpMask(res3); result = (bool)(arg1)->Init(*arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_OBChemTsfm_Apply(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBChemTsfm_Apply", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChemTsfm, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChemTsfm_Apply" "', argument " "1"" of type '" "OpenBabel::OBChemTsfm *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChemTsfm * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBChemTsfm_Apply" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBChemTsfm_Apply" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Apply(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChemTsfm_IsAcid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChemTsfm, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChemTsfm_IsAcid" "', argument " "1"" of type '" "OpenBabel::OBChemTsfm *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChemTsfm * >(argp1); result = (bool)(arg1)->IsAcid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBChemTsfm_IsBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBChemTsfm *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBChemTsfm, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBChemTsfm_IsBase" "', argument " "1"" of type '" "OpenBabel::OBChemTsfm *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBChemTsfm * >(argp1); result = (bool)(arg1)->IsBase();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBChemTsfm_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBChemTsfm, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBChemTsfm_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBPhModel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPhModel *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBPhModel", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBPhModel *)new OpenBabel::OBPhModel();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBPhModel, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBPhModel(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPhModel *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPhModel, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBPhModel" "', argument " "1"" of type '" "OpenBabel::OBPhModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPhModel * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBPhModel_AssignSeedPartialCharge(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBPhModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBPhModel_AssignSeedPartialCharge", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPhModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPhModel_AssignSeedPartialCharge" "', argument " "1"" of type '" "OpenBabel::OBPhModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPhModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPhModel_AssignSeedPartialCharge" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPhModel_AssignSeedPartialCharge" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->AssignSeedPartialCharge(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPhModel_CorrectForPH__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPhModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; double arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; double val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPhModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPhModel_CorrectForPH" "', argument " "1"" of type '" "OpenBabel::OBPhModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPhModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPhModel_CorrectForPH" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPhModel_CorrectForPH" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBPhModel_CorrectForPH" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->CorrectForPH(*arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBPhModel_CorrectForPH__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBPhModel *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBPhModel, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBPhModel_CorrectForPH" "', argument " "1"" of type '" "OpenBabel::OBPhModel *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBPhModel * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBPhModel_CorrectForPH" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBPhModel_CorrectForPH" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->CorrectForPH(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBPhModel_CorrectForPH(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBPhModel_CorrectForPH", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBPhModel_CorrectForPH__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBPhModel_CorrectForPH__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBPhModel_CorrectForPH'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBPhModel::CorrectForPH(OpenBabel::OBMol &,double)\n"
  "    OpenBabel::OBPhModel::CorrectForPH(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *OBPhModel_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBPhModel, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBPhModel_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBGraphSym__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::OBGraphSym *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBGraphSym" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBGraphSym" "', argument " "2"" of type '" "OpenBabel::OBBitVec const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBGraphSym *)new OpenBabel::OBGraphSym(arg1,(OpenBabel::OBBitVec const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGraphSym, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGraphSym__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBGraphSym *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBGraphSym" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (OpenBabel::OBGraphSym *)new OpenBabel::OBGraphSym(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBGraphSym, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBGraphSym(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBGraphSym", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBGraphSym__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBGraphSym__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBGraphSym'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBGraphSym::OBGraphSym(OpenBabel::OBMol *,OpenBabel::OBBitVec const *)\n"
  "    OpenBabel::OBGraphSym::OBGraphSym(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBGraphSym(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGraphSym *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGraphSym, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBGraphSym" "', argument " "1"" of type '" "OpenBabel::OBGraphSym *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGraphSym * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN int Swig_var_OBGraphSym_NoSymmetryClass_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable OBGraphSym_NoSymmetryClass is read-only."); return 1; }
SWIGINTERN PyObject *Swig_var_OBGraphSym_NoSymmetryClass_get(void) { PyObject *pyobj = 0;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(OpenBabel::OBGraphSym::NoSymmetryClass)); return pyobj; }
SWIGINTERN PyObject *_wrap_OBGraphSym_GetSymmetry(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGraphSym *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; int result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBGraphSym_GetSymmetry", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGraphSym, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGraphSym_GetSymmetry" "', argument " "1"" of type '" "OpenBabel::OBGraphSym *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGraphSym * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBGraphSym_GetSymmetry" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBGraphSym_GetSymmetry" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp2);
  result = (int)(arg1)->GetSymmetry(*arg2); resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBGraphSym_ClearSymmetry(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBGraphSym *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBGraphSym, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBGraphSym_ClearSymmetry" "', argument " "1"" of type '" "OpenBabel::OBGraphSym *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBGraphSym * >(argp1); (arg1)->ClearSymmetry(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBGraphSym_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBGraphSym, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBGraphSym_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBIsomorphismMapper(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBIsomorphismMapper *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBIsomorphismMapper" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_GetInstance__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBQuery *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; OpenBabel::OBIsomorphismMapper *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_GetInstance" "', argument " "1"" of type '" "OpenBabel::OBQuery *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1); { std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_GetInstance" "', argument " "2"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_GetInstance" "', argument " "2"" of type '" "std::string const &""'"); }
     arg2 = ptr; } 
  result = (OpenBabel::OBIsomorphismMapper *)OpenBabel::OBIsomorphismMapper::GetInstance(arg1,(std::string const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_GetInstance__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBIsomorphismMapper *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_GetInstance" "', argument " "1"" of type '" "OpenBabel::OBQuery *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  result = (OpenBabel::OBIsomorphismMapper *)OpenBabel::OBIsomorphismMapper::GetInstance(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_GetInstance(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_GetInstance", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBIsomorphismMapper_GetInstance__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_GetInstance__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBIsomorphismMapper_GetInstance'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBIsomorphismMapper::GetInstance(OpenBabel::OBQuery *,std::string const &)\n"
  "    OpenBabel::OBIsomorphismMapper::GetInstance(OpenBabel::OBQuery *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapFirst__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mapping *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mapping * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  (arg1)->MapFirst((OpenBabel::OBMol const *)arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapFirst__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mapping *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapFirst" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mapping * >(argp3);
  (arg1)->MapFirst((OpenBabel::OBMol const *)arg2,*arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapFirst(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_MapFirst", 0, 4, argv))) SWIG_fail; --argc;
  if (argc == 3) { PyObject *retobj = _wrap_OBIsomorphismMapper_MapFirst__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapFirst__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBIsomorphismMapper_MapFirst'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBIsomorphismMapper::MapFirst(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mapping &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::OBIsomorphismMapper::MapFirst(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mapping &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapUnique__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mappings *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mappings * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  (arg1)->MapUnique((OpenBabel::OBMol const *)arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapUnique__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mappings *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapUnique" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mappings * >(argp3);
  (arg1)->MapUnique((OpenBabel::OBMol const *)arg2,*arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapUnique(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_MapUnique", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapUnique__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapUnique__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBIsomorphismMapper_MapUnique'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBIsomorphismMapper::MapUnique(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mappings &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::OBIsomorphismMapper::MapUnique(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mappings &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapAll__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mappings *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; std::size_t arg5 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  size_t val5 ; int ecode5 = 0 ; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mappings * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapAll" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_size_t(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "5"" of type '" "std::size_t""'"); }
    arg5 = static_cast< std::size_t >(val5);
  (arg1)->MapAll((OpenBabel::OBMol const *)arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4,SWIG_STD_MOVE(arg5));
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapAll__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mappings *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mappings * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapAll" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  (arg1)->MapAll((OpenBabel::OBMol const *)arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapAll__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  OpenBabel::OBIsomorphismMapper::Mappings *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "2"" of type '" "OpenBabel::OBMol const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapAll" "', argument " "3"" of type '" "OpenBabel::OBIsomorphismMapper::Mappings &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Mappings * >(argp3);
  (arg1)->MapAll((OpenBabel::OBMol const *)arg2,*arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapAll(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_MapAll", 0, 5, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapAll__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapAll__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapAll__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBIsomorphismMapper_MapAll'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBIsomorphismMapper::MapAll(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mappings &,OpenBabel::OBBitVec const &,std::size_t)\n"
  "    OpenBabel::OBIsomorphismMapper::MapAll(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mappings &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::OBIsomorphismMapper::MapAll(OpenBabel::OBMol const *,OpenBabel::OBIsomorphismMapper::Mappings &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapGeneric__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBIsomorphismMapper::Functor *arg2 = 0 ;
  OpenBabel::OBMol *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "2"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "2"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Functor * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "3"" of type '" "OpenBabel::OBMol const *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBMol * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  (arg1)->MapGeneric(*arg2,(OpenBabel::OBMol const *)arg3,(OpenBabel::OBBitVec const &)*arg4); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapGeneric__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper *arg1 = 0 ; OpenBabel::OBIsomorphismMapper::Functor *arg2 = 0 ;
  OpenBabel::OBMol *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "2"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "2"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Functor * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBIsomorphismMapper_MapGeneric" "', argument " "3"" of type '" "OpenBabel::OBMol const *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBMol * >(argp3); (arg1)->MapGeneric(*arg2,(OpenBabel::OBMol const *)arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_MapGeneric(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_MapGeneric", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapGeneric__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBIsomorphismMapper_MapGeneric__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBIsomorphismMapper_MapGeneric'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBIsomorphismMapper::MapGeneric(OpenBabel::OBIsomorphismMapper::Functor &,OpenBabel::OBMol const *,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::OBIsomorphismMapper::MapGeneric(OpenBabel::OBIsomorphismMapper::Functor &,OpenBabel::OBMol const *)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBIsomorphismMapper_SetTimeout(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBIsomorphismMapper *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBIsomorphismMapper_SetTimeout", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBIsomorphismMapper_SetTimeout" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBIsomorphismMapper_SetTimeout" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->SetTimeout(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBIsomorphismMapper_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBIsomorphismMapper, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_MapsTo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBIsomorphismMapper::Mapping *arg1 = 0 ; unsigned int arg2 ; unsigned int *arg3 = 0 ; int res1 = SWIG_OLDOBJ ;
  unsigned int val2 ; int ecode2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "MapsTo", 3, 3, swig_obj)) SWIG_fail; {
    std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *ptr = (std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MapsTo" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "MapsTo" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Mapping const &""'"); }
     arg1 = ptr; }  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MapsTo" "', argument " "2"" of type '" "unsigned int""'"); }  
  arg2 = static_cast< unsigned int >(val2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "MapsTo" "', argument " "3"" of type '" "unsigned int &""'");  } 
  if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "MapsTo" "', argument " "3"" of type '" "unsigned int &""'"); }
   arg3 = reinterpret_cast< unsigned int * >(argp3);
  result = (bool)OpenBabel::MapsTo((std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > const &)*arg1,arg2,*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; std::size_t arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; void *argp4 = 0 ; int res4 = 0 ;
  size_t val5 ; int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; }  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_size_t(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FindAutomorphisms" "', argument " "5"" of type '" "std::size_t""'"); }
    arg5 = static_cast< std::size_t >(val5);
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3,(OpenBabel::OBBitVec const &)*arg4,SWIG_STD_MOVE(arg5));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; void *argp4 = 0 ; int res4 = 0 ; bool result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; }  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3,(OpenBabel::OBBitVec const &)*arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int res3 = SWIG_OLDOBJ ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; } 
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  OpenBabel::OBBitVec *arg3 = 0 ; std::size_t arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  void *argp3 = 0 ; int res3 = 0 ; size_t val4 ; int ecode4 = 0 ; bool result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3); ecode4 = SWIG_AsVal_size_t(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "std::size_t""'"); }
    arg4 = static_cast< std::size_t >(val4);
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2,(OpenBabel::OBBitVec const &)*arg3,SWIG_STD_MOVE(arg4));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  OpenBabel::OBBitVec *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ;
  int res3 = 0 ; bool result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBBitVec * >(argp3);
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2,(OpenBabel::OBBitVec const &)*arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &""'"); }
  
  arg2 = reinterpret_cast< std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > * >(argp2);
  result = (bool)OpenBabel::FindAutomorphisms(arg1,*arg2); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper::Functor *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Functor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; }  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  OpenBabel::FindAutomorphisms(*arg1,arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3,(OpenBabel::OBBitVec const &)*arg4);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms__SWIG_7(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBIsomorphismMapper::Functor *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int res3 = SWIG_OLDOBJ ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "1"" of type '" "OpenBabel::OBIsomorphismMapper::Functor &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBIsomorphismMapper::Functor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindAutomorphisms" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindAutomorphisms" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg3 = ptr; } 
  OpenBabel::FindAutomorphisms(*arg1,arg2,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindAutomorphisms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[6] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "FindAutomorphisms", 0, 5, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_FindAutomorphisms__SWIG_5(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; {
      int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_2; return _wrap_FindAutomorphisms__SWIG_4(self, argc, argv);}  check_2:
  if (argc == 3) { int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; PyObject *retobj = _wrap_FindAutomorphisms__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_3: if (argc == 3) {
    PyObject *retobj = _wrap_FindAutomorphisms__SWIG_7(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_5; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_5; {
      int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_OpenBabel__OBBitVec, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_5; { { int res = SWIG_AsVal_size_t(argv[3], NULL);
        _v = SWIG_CheckState(res); } }  if (!_v) goto check_5; return _wrap_FindAutomorphisms__SWIG_3(self, argc, argv);} 
  check_5: if (argc == 4) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBIsomorphismMapper__Functor, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_6; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_6; PyObject *retobj = _wrap_FindAutomorphisms__SWIG_6(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_6: if (argc == 4) {
    PyObject *retobj = _wrap_FindAutomorphisms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_FindAutomorphisms__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FindAutomorphisms'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &,std::vector< unsigned int,std::allocator< unsigned int > > const &,OpenBabel::OBBitVec const &,std::size_t)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &,std::vector< unsigned int,std::allocator< unsigned int > > const &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &,OpenBabel::OBBitVec const &,std::size_t)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBMol *,std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > &)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBIsomorphismMapper::Functor &,OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::FindAutomorphisms(OpenBabel::OBIsomorphismMapper::Functor &,OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_OBQueryAtom__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; bool arg2 ; bool arg3 ; int val1 ; int ecode1 = 0 ; bool val2 ; int ecode2 = 0 ;
  bool val3 ; int ecode3 = 0 ; OpenBabel::OBQueryAtom *result = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBQueryAtom" "', argument " "1"" of type '" "int""'"); }  
  arg1 = static_cast< int >(val1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBQueryAtom" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBQueryAtom" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (OpenBabel::OBQueryAtom *)new OpenBabel::OBQueryAtom(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryAtom__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; bool arg2 ; int val1 ; int ecode1 = 0 ; bool val2 ; int ecode2 = 0 ;
  OpenBabel::OBQueryAtom *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBQueryAtom" "', argument " "1"" of type '" "int""'"); }  
  arg1 = static_cast< int >(val1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBQueryAtom" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (OpenBabel::OBQueryAtom *)new OpenBabel::OBQueryAtom(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryAtom__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; int arg1 ; int val1 ; int ecode1 = 0 ; OpenBabel::OBQueryAtom *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBQueryAtom" "', argument " "1"" of type '" "int""'"); }  
  arg1 = static_cast< int >(val1); result = (OpenBabel::OBQueryAtom *)new OpenBabel::OBQueryAtom(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryAtom__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBQueryAtom *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBQueryAtom *)new OpenBabel::OBQueryAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryAtom(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBQueryAtom", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBQueryAtom__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBQueryAtom__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBQueryAtom__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBQueryAtom__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBQueryAtom'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBQueryAtom::OBQueryAtom(int,bool,bool)\n"
  "    OpenBabel::OBQueryAtom::OBQueryAtom(int,bool)\n" "    OpenBabel::OBQueryAtom::OBQueryAtom(int)\n"
  "    OpenBabel::OBQueryAtom::OBQueryAtom()\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBQueryAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBQueryAtom" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryAtom_GetIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryAtom_GetIndex" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  result = (unsigned int)((OpenBabel::OBQueryAtom const *)arg1)->GetIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryAtom_GetBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > > *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryAtom_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  result = (std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > > *) &((OpenBabel::OBQueryAtom const *)arg1)->GetBonds();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryAtom_GetNbrs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > > *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryAtom_GetNbrs" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  result = (std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > > *) &((OpenBabel::OBQueryAtom const *)arg1)->GetNbrs();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryAtom_Matches(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryAtom *arg1 = 0 ; OpenBabel::OBAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBQueryAtom_Matches", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryAtom_Matches" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBQueryAtom_Matches" "', argument " "2"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBAtom * >(argp2);
  result = (bool)((OpenBabel::OBQueryAtom const *)arg1)->Matches((OpenBabel::OBAtom const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBQueryAtom_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBQueryAtom, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBQueryAtom_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBQueryBond__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBQueryAtom *arg1 = 0 ; OpenBabel::OBQueryAtom *arg2 = 0 ; int arg3 ; bool arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; bool val4 ; int ecode4 = 0 ;
  OpenBabel::OBQueryBond *result = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBQueryBond" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBQueryBond" "', argument " "2"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBQueryBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBQueryBond" "', argument " "4"" of type '" "bool""'"); }  
  arg4 = static_cast< bool >(val4); result = (OpenBabel::OBQueryBond *)new OpenBabel::OBQueryBond(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryBond__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBQueryAtom *arg1 = 0 ; OpenBabel::OBQueryAtom *arg2 = 0 ; int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; OpenBabel::OBQueryBond *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBQueryBond" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBQueryBond" "', argument " "2"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBQueryBond" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); result = (OpenBabel::OBQueryBond *)new OpenBabel::OBQueryBond(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryBond__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBQueryAtom *arg1 = 0 ; OpenBabel::OBQueryAtom *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; OpenBabel::OBQueryBond *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBQueryBond" "', argument " "1"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBQueryBond" "', argument " "2"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp2);
  result = (OpenBabel::OBQueryBond *)new OpenBabel::OBQueryBond(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryBond, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQueryBond(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBQueryBond", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_new_OBQueryBond__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBQueryBond__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBQueryBond__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBQueryBond'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBQueryBond::OBQueryBond(OpenBabel::OBQueryAtom *,OpenBabel::OBQueryAtom *,int,bool)\n"
  "    OpenBabel::OBQueryBond::OBQueryBond(OpenBabel::OBQueryAtom *,OpenBabel::OBQueryAtom *,int)\n"
  "    OpenBabel::OBQueryBond::OBQueryBond(OpenBabel::OBQueryAtom *,OpenBabel::OBQueryAtom *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBQueryBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryBond, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBQueryBond" "', argument " "1"" of type '" "OpenBabel::OBQueryBond *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryBond_GetIndex(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryBond_GetIndex" "', argument " "1"" of type '" "OpenBabel::OBQueryBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp1);
  result = (unsigned int)((OpenBabel::OBQueryBond const *)arg1)->GetIndex();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryBond_GetBeginAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBQueryAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryBond_GetBeginAtom" "', argument " "1"" of type '" "OpenBabel::OBQueryBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp1);
  result = (OpenBabel::OBQueryAtom *)((OpenBabel::OBQueryBond const *)arg1)->GetBeginAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryBond_GetEndAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryBond *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBQueryAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryBond_GetEndAtom" "', argument " "1"" of type '" "OpenBabel::OBQueryBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp1);
  result = (OpenBabel::OBQueryAtom *)((OpenBabel::OBQueryBond const *)arg1)->GetEndAtom();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQueryBond_Matches(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQueryBond *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBQueryBond_Matches", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQueryBond_Matches" "', argument " "1"" of type '" "OpenBabel::OBQueryBond const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBQueryBond_Matches" "', argument " "2"" of type '" "OpenBabel::OBBond const *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2);
  result = (bool)((OpenBabel::OBQueryBond const *)arg1)->Matches((OpenBabel::OBBond const *)arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBQueryBond_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBQueryBond, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBQueryBond_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBQuery(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBQuery" "', argument " "1"" of type '" "OpenBabel::OBQuery *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_NumAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_NumAtoms" "', argument " "1"" of type '" "OpenBabel::OBQuery const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  result = (unsigned int)((OpenBabel::OBQuery const *)arg1)->NumAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_NumBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_NumBonds" "', argument " "1"" of type '" "OpenBabel::OBQuery const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  result = (unsigned int)((OpenBabel::OBQuery const *)arg1)->NumBonds();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_GetAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > > *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_GetAtoms" "', argument " "1"" of type '" "OpenBabel::OBQuery const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  result = (std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > > *) &((OpenBabel::OBQuery const *)arg1)->GetAtoms();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_GetBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > > *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_GetBonds" "', argument " "1"" of type '" "OpenBabel::OBQuery const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  result = (std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > > *) &((OpenBabel::OBQuery const *)arg1)->GetBonds();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_GetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; OpenBabel::OBQueryAtom *arg2 = 0 ; OpenBabel::OBQueryAtom *arg3 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ;
  OpenBabel::OBQueryBond *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBQuery_GetBond", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_GetBond" "', argument " "1"" of type '" "OpenBabel::OBQuery const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBQuery_GetBond" "', argument " "2"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBQuery_GetBond" "', argument " "3"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp3);
  result = (OpenBabel::OBQueryBond *)((OpenBabel::OBQuery const *)arg1)->GetBond(arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_AddAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; OpenBabel::OBQueryAtom *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBQuery_AddAtom", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_AddAtom" "', argument " "1"" of type '" "OpenBabel::OBQuery *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryAtom, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBQuery_AddAtom" "', argument " "2"" of type '" "OpenBabel::OBQueryAtom *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryAtom * >(argp2); (arg1)->AddAtom(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBQuery_AddBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *arg1 = 0 ; OpenBabel::OBQueryBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBQuery_AddBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBQuery_AddBond" "', argument " "1"" of type '" "OpenBabel::OBQuery *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBQuery * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBQueryBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBQuery_AddBond" "', argument " "2"" of type '" "OpenBabel::OBQueryBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBQueryBond * >(argp2); (arg1)->AddBond(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBQuery(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBQuery *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBQuery", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBQuery *)new OpenBabel::OBQuery();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQuery, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBQuery_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBQuery, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBQuery_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_CompileMoleculeQuery__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; OpenBabel::OBQuery *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CompileMoleculeQuery" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CompileMoleculeQuery" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CompileMoleculeQuery" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBQuery *)OpenBabel::CompileMoleculeQuery(arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_CompileMoleculeQuery__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBQuery *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CompileMoleculeQuery" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = (OpenBabel::OBQuery *)OpenBabel::CompileMoleculeQuery(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_CompileMoleculeQuery(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CompileMoleculeQuery", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_CompileMoleculeQuery__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_CompileMoleculeQuery__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CompileMoleculeQuery'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::CompileMoleculeQuery(OpenBabel::OBMol *,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::CompileMoleculeQuery(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_CompileSmilesQuery__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; int res1 = SWIG_OLDOBJ ; void *argp2 = 0 ;
  int res2 = 0 ; OpenBabel::OBQuery *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::string *ptr = (std::string *)0; res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CompileSmilesQuery" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CompileSmilesQuery" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CompileSmilesQuery" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CompileSmilesQuery" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2);
  result = (OpenBabel::OBQuery *)OpenBabel::CompileSmilesQuery((std::string const &)*arg1,(OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_CompileSmilesQuery__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; std::string *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; OpenBabel::OBQuery *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; { std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CompileSmilesQuery" "', argument " "1"" of type '" "std::string const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CompileSmilesQuery" "', argument " "1"" of type '" "std::string const &""'"); }
     arg1 = ptr; }  result = (OpenBabel::OBQuery *)OpenBabel::CompileSmilesQuery((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBQuery, 0 |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_CompileSmilesQuery(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CompileSmilesQuery", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_CompileSmilesQuery__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_CompileSmilesQuery__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CompileSmilesQuery'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::CompileSmilesQuery(std::string const &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::CompileSmilesQuery(std::string const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_CanonicalLabels__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; int arg5 ; bool arg6 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  int val5 ; int ecode5 = 0 ; bool val6 ; int ecode6 = 0 ; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CanonicalLabels" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CanonicalLabels" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_bool(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "CanonicalLabels" "', argument " "6"" of type '" "bool""'"); }  
  arg6 = static_cast< bool >(val6);
  OpenBabel::CanonicalLabels(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4,arg5,arg6);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_CanonicalLabels__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; int arg5 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ;
  int val5 ; int ecode5 = 0 ; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CanonicalLabels" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "CanonicalLabels" "', argument " "5"" of type '" "int""'"); }  
  arg5 = static_cast< int >(val5);
  OpenBabel::CanonicalLabels(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4,arg5);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_CanonicalLabels__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; OpenBabel::OBBitVec *arg4 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; int res2 = SWIG_OLDOBJ ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CanonicalLabels" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp4) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "4"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg4 = reinterpret_cast< OpenBabel::OBBitVec * >(argp4);
  OpenBabel::CanonicalLabels(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,*arg3,(OpenBabel::OBBitVec const &)*arg4);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_CanonicalLabels__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  std::vector< unsigned int,std::allocator< unsigned int > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; void *argp3 = 0 ; int res3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CanonicalLabels" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_unsigned_int_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CanonicalLabels" "', argument " "3"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > &""'"); }
   arg3 = reinterpret_cast< std::vector< unsigned int,std::allocator< unsigned int > > * >(argp3);
  OpenBabel::CanonicalLabels(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,*arg3);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_CanonicalLabels(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CanonicalLabels", 0, 6, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_CanonicalLabels__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_CanonicalLabels__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_CanonicalLabels__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_CanonicalLabels__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CanonicalLabels'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::CanonicalLabels(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,std::vector< unsigned int,std::allocator< unsigned int > > &,OpenBabel::OBBitVec const &,int,bool)\n"
  "    OpenBabel::CanonicalLabels(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,std::vector< unsigned int,std::allocator< unsigned int > > &,OpenBabel::OBBitVec const &,int)\n"
  "    OpenBabel::CanonicalLabels(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,std::vector< unsigned int,std::allocator< unsigned int > > &,OpenBabel::OBBitVec const &)\n"
  "    OpenBabel::CanonicalLabels(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,std::vector< unsigned int,std::allocator< unsigned int > > &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBStereo_MakeRefs__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBStereo::Ref arg1 ; OpenBabel::OBStereo::Ref arg2 ; OpenBabel::OBStereo::Ref arg3 ;
  OpenBabel::OBStereo::Ref arg4 ; unsigned long val1 ; int ecode1 = 0 ; unsigned long val2 ; int ecode2 = 0 ;
  unsigned long val3 ; int ecode3 = 0 ; unsigned long val4 ; int ecode4 = 0 ; OpenBabel::OBStereo::Refs result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBStereo_MakeRefs" "', argument " "1"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg1 = static_cast< OpenBabel::OBStereo::Ref >(val1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereo_MakeRefs" "', argument " "2"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Ref >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBStereo_MakeRefs" "', argument " "3"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Ref >(val3); ecode4 = SWIG_AsVal_unsigned_SS_long(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBStereo_MakeRefs" "', argument " "4"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg4 = static_cast< OpenBabel::OBStereo::Ref >(val4); result = OpenBabel::OBStereo::MakeRefs(arg1,arg2,arg3,arg4);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_MakeRefs__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBStereo::Ref arg1 ; OpenBabel::OBStereo::Ref arg2 ; OpenBabel::OBStereo::Ref arg3 ;
  unsigned long val1 ; int ecode1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ;
  OpenBabel::OBStereo::Refs result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "OBStereo_MakeRefs" "', argument " "1"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg1 = static_cast< OpenBabel::OBStereo::Ref >(val1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereo_MakeRefs" "', argument " "2"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Ref >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBStereo_MakeRefs" "', argument " "3"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Ref >(val3); result = OpenBabel::OBStereo::MakeRefs(arg1,arg2,arg3);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_MakeRefs(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBStereo_MakeRefs", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBStereo_MakeRefs__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBStereo_MakeRefs__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBStereo_MakeRefs'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBStereo::MakeRefs(OpenBabel::OBStereo::Ref,OpenBabel::OBStereo::Ref,OpenBabel::OBStereo::Ref,OpenBabel::OBStereo::Ref)\n"
  "    OpenBabel::OBStereo::MakeRefs(OpenBabel::OBStereo::Ref,OpenBabel::OBStereo::Ref,OpenBabel::OBStereo::Ref)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBStereo_ContainsSameRefs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo::Refs *arg1 = 0 ; OpenBabel::OBStereo::Refs *arg2 = 0 ; int res1 = SWIG_OLDOBJ ; int res2 = SWIG_OLDOBJ ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereo_ContainsSameRefs", 2, 2, swig_obj)) SWIG_fail; {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereo_ContainsSameRefs" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_ContainsSameRefs" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg1 = ptr; }  {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBStereo_ContainsSameRefs" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_ContainsSameRefs" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg2 = ptr; } 
  result = (bool)OpenBabel::OBStereo::ContainsSameRefs((std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg1,(std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_ContainsRef(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo::Refs *arg1 = 0 ; unsigned long arg2 ; int res1 = SWIG_OLDOBJ ; unsigned long val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereo_ContainsRef", 2, 2, swig_obj)) SWIG_fail; {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereo_ContainsRef" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_ContainsRef" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg1 = ptr; }  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereo_ContainsRef" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (bool)OpenBabel::OBStereo::ContainsRef((std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg1,arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_NumInversions(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo::Refs *arg1 = 0 ; int res1 = SWIG_OLDOBJ ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereo_NumInversions" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_NumInversions" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg1 = ptr; } 
  result = (int)OpenBabel::OBStereo::NumInversions((std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg1);
  resultobj = SWIG_From_int(static_cast< int >(result)); if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_Permutate(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo::Refs *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned int val2 ; int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereo_Permutate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_unsigned_long_t,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereo_Permutate" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_Permutate" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBStereo::Refs * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereo_Permutate" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBStereo_Permutate" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); OpenBabel::OBStereo::Permutate(*arg1,arg2,arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereo_Permutated(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo::Refs *arg1 = 0 ; unsigned int arg2 ; unsigned int arg3 ; int res1 = SWIG_OLDOBJ ; unsigned int val2 ;
  int ecode2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; OpenBabel::OBStereo::Refs result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereo_Permutated", 3, 3, swig_obj)) SWIG_fail; {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereo_Permutated" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBStereo_Permutated" "', argument " "1"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg1 = ptr; }  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereo_Permutated" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBStereo_Permutated" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3);
  result = OpenBabel::OBStereo::Permutated((std::vector< unsigned long,std::allocator< unsigned long > > const &)*arg1,arg2,arg3);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(result));
  if (SWIG_IsNewObj(res1)) delete arg1; return resultobj; fail: if (SWIG_IsNewObj(res1)) delete arg1; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBStereo", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBStereo *)new OpenBabel::OBStereo();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereo, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStereo" "', argument " "1"" of type '" "OpenBabel::OBStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBStereo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBStereoUnit__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBStereoUnit *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBStereoUnit *)new OpenBabel::OBStereoUnit();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoUnit, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereoUnit__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBStereo::Type arg1 ; unsigned long arg2 ; bool arg3 ; int val1 ; int ecode1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; bool val3 ; int ecode3 = 0 ; OpenBabel::OBStereoUnit *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBStereoUnit" "', argument " "1"" of type '" "OpenBabel::OBStereo::Type""'"); }
    arg1 = static_cast< OpenBabel::OBStereo::Type >(val1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBStereoUnit" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBStereoUnit" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (OpenBabel::OBStereoUnit *)new OpenBabel::OBStereoUnit(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoUnit, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereoUnit__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBStereo::Type arg1 ; unsigned long arg2 ; int val1 ; int ecode1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBStereoUnit *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_int(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBStereoUnit" "', argument " "1"" of type '" "OpenBabel::OBStereo::Type""'"); }
    arg1 = static_cast< OpenBabel::OBStereo::Type >(val1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBStereoUnit" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBStereoUnit *)new OpenBabel::OBStereoUnit(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoUnit, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereoUnit(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBStereoUnit", 0, 3, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBStereoUnit__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBStereoUnit__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBStereoUnit__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBStereoUnit'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBStereoUnit::OBStereoUnit()\n"
  "    OpenBabel::OBStereoUnit::OBStereoUnit(OpenBabel::OBStereo::Type,unsigned long,bool)\n"
  "    OpenBabel::OBStereoUnit::OBStereoUnit(OpenBabel::OBStereo::Type,unsigned long)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_type_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; OpenBabel::OBStereo::Type arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoUnit_type_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_type_set" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoUnit_type_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Type""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Type >(val2); if (arg1) (arg1)->type = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_type_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBStereo::Type result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_type_get" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); result = (OpenBabel::OBStereo::Type) ((arg1)->type);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_id_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoUnit_id_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_id_set" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoUnit_id_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->id = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_id_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_id_get" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); result = (unsigned long) ((arg1)->id);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_para_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBStereoUnit_para_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_para_set" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoUnit_para_set" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); if (arg1) (arg1)->para = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoUnit_para_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoUnit_para_get" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); result = (bool) ((arg1)->para);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBStereoUnit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoUnit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoUnit, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStereoUnit" "', argument " "1"" of type '" "OpenBabel::OBStereoUnit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoUnit * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBStereoUnit_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStereoUnit, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBStereoUnit_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBStereoBase(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoBase, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStereoBase" "', argument " "1"" of type '" "OpenBabel::OBStereoBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoBase * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoBase_GetMolecule(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBMol *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoBase_GetMolecule" "', argument " "1"" of type '" "OpenBabel::OBStereoBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoBase * >(argp1);
  result = (OpenBabel::OBMol *)((OpenBabel::OBStereoBase const *)arg1)->GetMolecule();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoBase_GetType(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBStereo::Type result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoBase_GetType" "', argument " "1"" of type '" "OpenBabel::OBStereoBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoBase * >(argp1);
  result = (OpenBabel::OBStereo::Type)((OpenBabel::OBStereoBase const *)arg1)->GetType();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoBase_SetSpecified(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoBase *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoBase_SetSpecified", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoBase_SetSpecified" "', argument " "1"" of type '" "OpenBabel::OBStereoBase *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoBase * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoBase_SetSpecified" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetSpecified(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoBase_IsSpecified(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoBase *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoBase, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoBase_IsSpecified" "', argument " "1"" of type '" "OpenBabel::OBStereoBase const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoBase * >(argp1);
  result = (bool)((OpenBabel::OBStereoBase const *)arg1)->IsSpecified();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBStereoBase_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStereoBase, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBStereoFacade__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; OpenBabel::OBStereoFacade *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBStereoFacade" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBStereoFacade" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (OpenBabel::OBStereoFacade *)new OpenBabel::OBStereoFacade(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoFacade, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereoFacade__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBStereoFacade *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBStereoFacade" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBStereoFacade *)new OpenBabel::OBStereoFacade(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStereoFacade, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStereoFacade(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBStereoFacade", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_new_OBStereoFacade__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBStereoFacade__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBStereoFacade'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBStereoFacade::OBStereoFacade(OpenBabel::OBMol *,bool)\n"
  "    OpenBabel::OBStereoFacade::OBStereoFacade(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_NumTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_NumTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (unsigned int)(arg1)->NumTetrahedralStereo();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetAllTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetAllTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (arg1)->GetAllTetrahedralStereo();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_HasTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_HasTetrahedralStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_HasTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_HasTetrahedralStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (bool)(arg1)->HasTetrahedralStereo(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBTetrahedralStereo *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_GetTetrahedralStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_GetTetrahedralStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBTetrahedralStereo *)(arg1)->GetTetrahedralStereo(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_NumCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_NumCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (unsigned int)(arg1)->NumCisTransStereo();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetAllCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetAllCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (arg1)->GetAllCisTransStereo();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_HasCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_HasCisTransStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_HasCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_HasCisTransStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (bool)(arg1)->HasCisTransStereo(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBCisTransStereo *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_GetCisTransStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_GetCisTransStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBCisTransStereo *)(arg1)->GetCisTransStereo(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_NumSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_NumSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (unsigned int)(arg1)->NumSquarePlanarStereo();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetAllSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBSquarePlanarStereo *,std::allocator< OpenBabel::OBSquarePlanarStereo * > > > result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetAllSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); result = (arg1)->GetAllSquarePlanarStereo();
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBSquarePlanarStereo *,std::allocator< OpenBabel::OBSquarePlanarStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_HasSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_HasSquarePlanarStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_HasSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_HasSquarePlanarStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (bool)(arg1)->HasSquarePlanarStereo(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBStereoFacade_GetSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBSquarePlanarStereo *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBStereoFacade_GetSquarePlanarStereo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBStereoFacade_GetSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBStereoFacade_GetSquarePlanarStereo" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = (OpenBabel::OBSquarePlanarStereo *)(arg1)->GetSquarePlanarStereo(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBStereoFacade(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStereoFacade *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStereoFacade, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStereoFacade" "', argument " "1"" of type '" "OpenBabel::OBStereoFacade *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStereoFacade * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBStereoFacade_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStereoFacade, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBStereoFacade_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_PerceiveStereo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PerceiveStereo" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PerceiveStereo" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); OpenBabel::PerceiveStereo(arg1,arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_PerceiveStereo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PerceiveStereo" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); OpenBabel::PerceiveStereo(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_PerceiveStereo(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "PerceiveStereo", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_PerceiveStereo__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_PerceiveStereo__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PerceiveStereo'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::PerceiveStereo(OpenBabel::OBMol *,bool)\n"
  "    OpenBabel::PerceiveStereo(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_StereoFrom2D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *arg2 = 0 ;
  bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StereoFrom2D" "', argument " "2"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *""'");  }
  
  arg2 = reinterpret_cast< std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "StereoFrom2D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); OpenBabel::StereoFrom2D(arg1,arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_StereoFrom2D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StereoFrom2D" "', argument " "2"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *""'");  }
  
  arg2 = reinterpret_cast< std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > * >(argp2);
  OpenBabel::StereoFrom2D(arg1,arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_StereoFrom2D__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); OpenBabel::StereoFrom2D(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_StereoFrom2D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "StereoFrom2D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_StereoFrom2D__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_StereoFrom2D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_StereoFrom2D__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'StereoFrom2D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::StereoFrom2D(OpenBabel::OBMol *,std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *,bool)\n"
  "    OpenBabel::StereoFrom2D(OpenBabel::OBMol *,std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *)\n"
  "    OpenBabel::StereoFrom2D(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_StereoFrom3D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StereoFrom3D" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); OpenBabel::StereoFrom3D(arg1,arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_StereoFrom3D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); OpenBabel::StereoFrom3D(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_StereoFrom3D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "StereoFrom3D", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_StereoFrom3D__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 2) { PyObject *retobj = _wrap_StereoFrom3D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'StereoFrom3D'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::StereoFrom3D(OpenBabel::OBMol *,bool)\n"
  "    OpenBabel::StereoFrom3D(OpenBabel::OBMol *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_StereoFrom0D(PyObject *self, PyObject *args) { PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoFrom0D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); OpenBabel::StereoFrom0D(arg1); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom3D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "TetrahedralFrom3D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = OpenBabel::TetrahedralFrom3D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom3D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::TetrahedralFrom3D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom3D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "TetrahedralFrom3D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_TetrahedralFrom3D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_TetrahedralFrom3D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'TetrahedralFrom3D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::TetrahedralFrom3D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,bool)\n"
  "    OpenBabel::TetrahedralFrom3D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom2D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "TetrahedralFrom2D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = OpenBabel::TetrahedralFrom2D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom2D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::TetrahedralFrom2D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom2D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "TetrahedralFrom2D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_TetrahedralFrom2D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_TetrahedralFrom2D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'TetrahedralFrom2D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::TetrahedralFrom2D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,bool)\n"
  "    OpenBabel::TetrahedralFrom2D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom0D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom0D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "TetrahedralFrom0D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = OpenBabel::TetrahedralFrom0D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom0D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetrahedralFrom0D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetrahedralFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetrahedralFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::TetrahedralFrom0D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_TetrahedralFrom0D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "TetrahedralFrom0D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_TetrahedralFrom0D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_TetrahedralFrom0D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'TetrahedralFrom0D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::TetrahedralFrom0D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,bool)\n"
  "    OpenBabel::TetrahedralFrom0D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_CisTransFrom3D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CisTransFrom3D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = OpenBabel::CisTransFrom3D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom3D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom3D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom3D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::CisTransFrom3D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom3D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CisTransFrom3D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_CisTransFrom3D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_CisTransFrom3D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CisTransFrom3D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::CisTransFrom3D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,bool)\n"
  "    OpenBabel::CisTransFrom3D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_CisTransFrom2D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ;
  std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *arg3 = 0 ;
  bool arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; bool val4 ;
  int ecode4 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CisTransFrom2D" "', argument " "3"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *""'");  }
  
  arg3 = reinterpret_cast< std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CisTransFrom2D" "', argument " "4"" of type '" "bool""'"); }  
  arg4 = static_cast< bool >(val4);
  result = OpenBabel::CisTransFrom2D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,(std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom2D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ;
  std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "CisTransFrom2D" "', argument " "3"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *""'");  }
  
  arg3 = reinterpret_cast< std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > * >(argp3);
  result = OpenBabel::CisTransFrom2D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,(std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *)arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom2D__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom2D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom2D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::CisTransFrom2D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom2D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CisTransFrom2D", 0, 4, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_CisTransFrom2D__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_CisTransFrom2D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_CisTransFrom2D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CisTransFrom2D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::CisTransFrom2D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *,bool)\n"
  "    OpenBabel::CisTransFrom2D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > const *)\n"
  "    OpenBabel::CisTransFrom2D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_TetStereoToWedgeHash(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ;
  std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *arg2 = 0 ;
  std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > > *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ;
  bool result; (void)self; if (!SWIG_Python_UnpackTuple(args, "TetStereoToWedgeHash", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TetStereoToWedgeHash" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetStereoToWedgeHash" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "TetStereoToWedgeHash" "', argument " "2"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetStereoToWedgeHash" "', argument " "2"" of type '" "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > &""'"); }
  
  arg2 = reinterpret_cast< std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "TetStereoToWedgeHash" "', argument " "3"" of type '" "std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "TetStereoToWedgeHash" "', argument " "3"" of type '" "std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > > &""'"); }
  
  arg3 = reinterpret_cast< std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > > * >(argp3);
  result = (bool)OpenBabel::TetStereoToWedgeHash(*arg1,*arg2,*arg3); resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_GetUnspecifiedCisTrans(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::set< OpenBabel::OBBond * > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GetUnspecifiedCisTrans" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "GetUnspecifiedCisTrans" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = OpenBabel::GetUnspecifiedCisTrans(*arg1);
  resultobj = SWIG_NewPointerObj((new std::set< OpenBabel::OBBond * >(result)), SWIGTYPE_p_std__setT_OpenBabel__OBBond_p_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_StereoRefToImplicit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereo::Ref arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "StereoRefToImplicit", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StereoRefToImplicit" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "StereoRefToImplicit" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StereoRefToImplicit" "', argument " "2"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Ref >(val2); OpenBabel::StereoRefToImplicit(*arg1,arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_ImplicitRefToStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereo::Ref arg2 ; OpenBabel::OBStereo::Ref arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "ImplicitRefToStereo", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ImplicitRefToStereo" "', argument " "1"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "ImplicitRefToStereo" "', argument " "1"" of type '" "OpenBabel::OBMol &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ImplicitRefToStereo" "', argument " "2"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Ref >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ImplicitRefToStereo" "', argument " "3"" of type '" "OpenBabel::OBStereo::Ref""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Ref >(val3); OpenBabel::ImplicitRefToStereo(*arg1,arg2,arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom0D__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom0D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CisTransFrom0D" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = OpenBabel::CisTransFrom0D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom0D__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBStereoUnitSet *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ;
  SwigValueWrapper< std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > > result;
  (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CisTransFrom0D" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CisTransFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "CisTransFrom0D" "', argument " "2"" of type '" "OpenBabel::OBStereoUnitSet const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBStereoUnitSet * >(argp2);
  result = OpenBabel::CisTransFrom0D(arg1,(std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > >(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_CisTransFrom0D(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "CisTransFrom0D", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_CisTransFrom0D__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_CisTransFrom0D__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'CisTransFrom0D'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::CisTransFrom0D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &,bool)\n"
  "    OpenBabel::CisTransFrom0D(OpenBabel::OBMol *,OpenBabel::OBStereoUnitSet const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_FindStereogenicUnits__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ;
  SwigValueWrapper< std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > > result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindStereogenicUnits" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindStereogenicUnits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindStereogenicUnits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; } 
  result = OpenBabel::FindStereogenicUnits(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBStereoUnitSet(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_FindStereogenicUnits__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; std::vector< unsigned int,std::allocator< unsigned int > > *arg2 = 0 ;
  OpenBabel::Automorphisms *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; int res3 = SWIG_OLDOBJ ;
  SwigValueWrapper< std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > > result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FindStereogenicUnits" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); {
    std::vector< unsigned int,std::allocator< unsigned int > > *ptr = (std::vector< unsigned int,std::allocator< unsigned int > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FindStereogenicUnits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindStereogenicUnits" "', argument " "2"" of type '" "std::vector< unsigned int,std::allocator< unsigned int > > const &""'"); }
     arg2 = ptr; }  {
    std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *ptr = (std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FindStereogenicUnits" "', argument " "3"" of type '" "OpenBabel::Automorphisms const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "FindStereogenicUnits" "', argument " "3"" of type '" "OpenBabel::Automorphisms const &""'"); }
     arg3 = ptr; } 
  result = OpenBabel::FindStereogenicUnits(arg1,(std::vector< unsigned int,std::allocator< unsigned int > > const &)*arg2,(std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > const &)*arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBStereoUnitSet(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t, SWIG_POINTER_OWN |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_FindStereogenicUnits(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "FindStereogenicUnits", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_FindStereogenicUnits__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_FindStereogenicUnits__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FindStereogenicUnits'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::FindStereogenicUnits(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &)\n"
  "    OpenBabel::FindStereogenicUnits(OpenBabel::OBMol *,std::vector< unsigned int,std::allocator< unsigned int > > const &,OpenBabel::Automorphisms const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_new_OBRotorRule(PyObject *self, PyObject *args) { PyObject *resultobj = 0; char *arg1 = 0 ;
  int *arg2 = (int *)0 ; std::vector< double,std::allocator< double > > *arg3 = 0 ; double arg4 ; int res1 ; char *buf1 = 0 ;
  int alloc1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; double val4 ; int ecode4 = 0 ;
  PyObject *swig_obj[4] ; OpenBabel::OBRotorRule *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBRotorRule", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBRotorRule" "', argument " "1"" of type '" "char *""'"); } 
  arg1 = reinterpret_cast< char * >(buf1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBRotorRule" "', argument " "2"" of type '" "int [4]""'");  }  
  arg2 = reinterpret_cast< int * >(argp2); res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_double_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_OBRotorRule" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBRotorRule" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > &""'"); }
   arg3 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp3);
  ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBRotorRule" "', argument " "4"" of type '" "double""'"); }  
  arg4 = static_cast< double >(val4); result = (OpenBabel::OBRotorRule *)new OpenBabel::OBRotorRule(arg1,arg2,*arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotorRule, SWIG_POINTER_NEW |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return resultobj; fail: if (alloc1 == SWIG_NEWOBJ) delete[] buf1; return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotorRule(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotorRule" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_IsValid" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1); result = (bool)(arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_GetReferenceAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; int *arg2 = (int *)0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorRule_GetReferenceAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_GetReferenceAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorRule_GetReferenceAtoms" "', argument " "2"" of type '" "int [4]""'");  }
    arg2 = reinterpret_cast< int * >(argp2); (arg1)->GetReferenceAtoms(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_SetDelta(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotorRule_SetDelta", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_SetDelta" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotorRule_SetDelta" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetDelta(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_GetDelta(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_GetDelta" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1); result = (double)(arg1)->GetDelta();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_GetTorsionVals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_GetTorsionVals" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) &(arg1)->GetTorsionVals();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_GetSmartsString(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; std::string *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_GetSmartsString" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1); result = (std::string *) &(arg1)->GetSmartsString();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__string, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRule_GetSmartsPattern(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRule *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSmartsPattern *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRule, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRule_GetSmartsPattern" "', argument " "1"" of type '" "OpenBabel::OBRotorRule *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRule * >(argp1);
  result = (OpenBabel::OBSmartsPattern *)(arg1)->GetSmartsPattern();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSmartsPattern, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBRotorRule_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotorRule, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotorRule_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotorRules(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRules *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRotorRules", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotorRules *)new OpenBabel::OBRotorRules();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotorRules, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotorRules(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRules *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRules, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotorRules" "', argument " "1"" of type '" "OpenBabel::OBRotorRules *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRules * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRules_SetFilename(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRules *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotorRules_SetFilename", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRules, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRules_SetFilename" "', argument " "1"" of type '" "OpenBabel::OBRotorRules *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorRules_SetFilename" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorRules_SetFilename" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); (arg1)->SetFilename(*arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRules_GetRotorIncrements(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRules *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBBond *arg3 = 0 ; int *arg4 = (int *)0 ;
  std::vector< double,std::allocator< double > > *arg5 = 0 ; double *arg6 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; void *argp4 = 0 ; int res4 = 0 ; void *argp5 = 0 ;
  int res5 = 0 ; void *argp6 = 0 ; int res6 = 0 ; PyObject *swig_obj[6] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorRules_GetRotorIncrements", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRules, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "1"" of type '" "OpenBabel::OBRotorRules *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRules * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "3"" of type '" "OpenBabel::OBBond *""'");  }
   arg3 = reinterpret_cast< OpenBabel::OBBond * >(argp3); res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "4"" of type '" "int [4]""'");  }
    arg4 = reinterpret_cast< int * >(argp4); res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_std__vectorT_double_t,  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > &""'");  }
   if (!argp5) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "5"" of type '" "std::vector< double,std::allocator< double > > &""'"); }
   arg5 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6, SWIGTYPE_p_double,  0 ); if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "6"" of type '" "double &""'");  }
   if (!argp6) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorRules_GetRotorIncrements" "', argument " "6"" of type '" "double &""'"); }
   arg6 = reinterpret_cast< double * >(argp6); (arg1)->GetRotorIncrements(*arg2,arg3,arg4,*arg5,*arg6);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorRules_Quiet(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorRules *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorRules, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorRules_Quiet" "', argument " "1"" of type '" "OpenBabel::OBRotorRules *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorRules * >(argp1); (arg1)->Quiet(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBRotorRules_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotorRules, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotorRules_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRotor", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotor *)new OpenBabel::OBRotor();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotor, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotor" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetBond" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); (arg1)->SetBond(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetRings(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetRings" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); (arg1)->SetRings(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetIdx", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetIdx" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotor_SetIdx" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); (arg1)->SetIdx(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetDihedralAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetDihedralAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetDihedralAtoms" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetDihedralAtoms" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp2); (arg1)->SetDihedralAtoms(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetDihedralAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; int *arg2 = (int *)0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetDihedralAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetDihedralAtoms" "', argument " "2"" of type '" "int [4]""'");  }
    arg2 = reinterpret_cast< int * >(argp2); (arg1)->SetDihedralAtoms(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetDihedralAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotor_SetDihedralAtoms", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__vectorT_int_t, SWIG_POINTER_NO_NULL); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_OBRotor_SetDihedralAtoms__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    PyObject *retobj = _wrap_OBRotor_SetDihedralAtoms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotor_SetDihedralAtoms'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotor::SetDihedralAtoms(std::vector< int,std::allocator< int > > &)\n"
  "    OpenBabel::OBRotor::SetDihedralAtoms(int [4])\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotor_SetRotAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; std::vector< int,std::allocator< int > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetRotAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetRotAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_int_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetRotAtoms" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetRotAtoms" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > > &""'"); }
   arg2 = reinterpret_cast< std::vector< int,std::allocator< int > > * >(argp2); (arg1)->SetRotAtoms(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetTorsionValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; std::vector< double,std::allocator< double > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetTorsionValues", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetTorsionValues" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_t,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetTorsionValues" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetTorsionValues" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > &""'"); }
   arg2 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp2); (arg1)->SetTorsionValues(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetFixedBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetFixedBonds", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetFixedBonds" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetFixedBonds(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetToAngle(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetToAngle", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetToAngle" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetToAngle" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_SetToAngle" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); (arg1)->SetToAngle(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetRotor__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; int arg3 ; int arg4 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetRotor" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetRotor" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_SetRotor" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBRotor_SetRotor" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); (arg1)->SetRotor(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetRotor__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetRotor" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetRotor" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_SetRotor" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->SetRotor(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetRotor(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotor_SetRotor", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBRotor_SetRotor__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBRotor_SetRotor__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotor_SetRotor'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotor::SetRotor(double *,int,int)\n"
  "    OpenBabel::OBRotor::SetRotor(double *,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotor_Set__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; double arg3 ; double arg4 ; double arg5 ; double arg6 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; double val3 ; int ecode3 = 0 ; double val4 ; int ecode4 = 0 ; double val5 ;
  int ecode5 = 0 ; double val6 ; int ecode6 = 0 ; (void)self; if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Set" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_Set" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_Set" "', argument " "3"" of type '" "double""'"); }  
  arg3 = static_cast< double >(val3); ecode4 = SWIG_AsVal_double(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBRotor_Set" "', argument " "4"" of type '" "double""'"); }  
  arg4 = static_cast< double >(val4); ecode5 = SWIG_AsVal_double(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBRotor_Set" "', argument " "5"" of type '" "double""'"); }  
  arg5 = static_cast< double >(val5); ecode6 = SWIG_AsVal_double(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBRotor_Set" "', argument " "6"" of type '" "double""'"); }  
  arg6 = static_cast< double >(val6); (arg1)->Set(arg2,arg3,arg4,arg5,arg6); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Precompute(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_Precompute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Precompute" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_Precompute" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->Precompute(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Set__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  int val3 ; int ecode3 = 0 ; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Set" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_Set" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_Set" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); (arg1)->Set(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Precalc(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; std::vector< double *,std::allocator< double * > > *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_Precalc", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Precalc" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_Precalc" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > > &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_Precalc" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > > &""'"); }
   arg2 = reinterpret_cast< std::vector< double *,std::allocator< double * > > * >(argp2); (arg1)->Precalc(*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Set__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; int arg3 ; int arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Set" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_Set" "', argument " "2"" of type '" "double *""'");  } 
  arg2 = reinterpret_cast< double * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotor_Set" "', argument " "3"" of type '" "int""'"); }  
  arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBRotor_Set" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); (arg1)->Set(arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Set(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotor_Set", 0, 6, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBRotor_Set__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 4) { PyObject *retobj = _wrap_OBRotor_Set__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_OBRotor_Set__SWIG_0(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotor_Set'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotor::Set(double *,double,double,double,double)\n"
  "    OpenBabel::OBRotor::Set(double *,int)\n" "    OpenBabel::OBRotor::Set(double *,int,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotor_GetBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBond *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetBond" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (OpenBabel::OBBond *)(arg1)->GetBond();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_Size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_Size" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (arg1)->Size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetIdx(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetIdx" "', argument " "1"" of type '" "OpenBabel::OBRotor const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (int)((OpenBabel::OBRotor const *)arg1)->GetIdx();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetDihedralAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; int *arg2 = (int *)0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetDihedralAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_GetDihedralAtoms" "', argument " "2"" of type '" "int [4]""'");  }
    arg2 = reinterpret_cast< int * >(argp2); (arg1)->GetDihedralAtoms(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetDihedralAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  std::vector< int,std::allocator< int > > *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetDihedralAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  result = (std::vector< int,std::allocator< int > > *) &(arg1)->GetDihedralAtoms();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_int_t, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetDihedralAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotor_GetDihedralAtoms", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBRotor_GetDihedralAtoms__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBRotor_GetDihedralAtoms__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotor_GetDihedralAtoms'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotor::GetDihedralAtoms(int [4])\n"
  "    OpenBabel::OBRotor::GetDihedralAtoms()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotor_GetTorsionValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetTorsionValues" "', argument " "1"" of type '" "OpenBabel::OBRotor const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) &((OpenBabel::OBRotor const *)arg1)->GetTorsionValues();
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(*result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetFixedBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetFixedBonds" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (OpenBabel::OBBitVec *) &(arg1)->GetFixedBonds();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_CalcTorsion(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_CalcTorsion", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_CalcTorsion" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_CalcTorsion" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (double)(arg1)->CalcTorsion(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_CalcBondLength(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_CalcBondLength", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_CalcBondLength" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_CalcBondLength" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); result = (double)(arg1)->CalcBondLength(arg2);
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_BeginTorIncrement(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > >::iterator result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_BeginTorIncrement" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (arg1)->BeginTorIncrement();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double,std::allocator< double > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_EndTorIncrement(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > >::iterator result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_EndTorIncrement" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (arg1)->EndTorIncrement();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double,std::allocator< double > >::iterator & >(result)),
  swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_RemoveSymTorsionValues(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_RemoveSymTorsionValues", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_RemoveSymTorsionValues" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotor_RemoveSymTorsionValues" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->RemoveSymTorsionValues(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetDelta(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetDelta", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetDelta" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotor_SetDelta" "', argument " "2"" of type '" "double""'"); }  
  arg2 = static_cast< double >(val2); (arg1)->SetDelta(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetDelta(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetDelta" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (double)(arg1)->GetDelta();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetFixedAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetFixedAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (OpenBabel::OBBitVec *) &(arg1)->GetFixedAtoms();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetFixedAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetFixedAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetFixedAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetFixedAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetFixedAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetFixedAtoms(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetEvalAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBitVec *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetEvalAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (OpenBabel::OBBitVec *) &(arg1)->GetEvalAtoms();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBitVec, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetEvalAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetEvalAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetEvalAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotor_SetEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotor_SetEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetEvalAtoms(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetRotAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; void *result = 0 ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetRotAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); result = (void *)(arg1)->GetRotAtoms();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_void, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_GetResolution(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_GetResolution" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1);
  result = (std::vector< double,std::allocator< double > > *) &(arg1)->GetResolution();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_t, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotor_SetNumCoords(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotor *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotor_SetNumCoords", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotor_SetNumCoords" "', argument " "1"" of type '" "OpenBabel::OBRotor *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotor * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotor_SetNumCoords" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetNumCoords(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBRotor_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotor, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotor_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotorList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRotorList", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotorList *)new OpenBabel::OBRotorList();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotorList, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotorList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotorList" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_Clear" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; size_t result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_Size" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (arg1)->Size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_IsFixedBond(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBBond *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_IsFixedBond", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_IsFixedBond" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_IsFixedBond" "', argument " "2"" of type '" "OpenBabel::OBBond *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBBond * >(argp2); result = (bool)(arg1)->IsFixedBond(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_HasFixedBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_HasFixedBonds" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (bool)(arg1)->HasFixedBonds();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_RemoveSymVals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_RemoveSymVals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_RemoveSymVals" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_RemoveSymVals" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_RemoveSymVals" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->RemoveSymVals(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_HasRingRotors(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_HasRingRotors" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (bool)(arg1)->HasRingRotors();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Setup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_Setup" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotorList_Setup" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); result = (bool)(arg1)->Setup(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Setup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_Setup" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Setup(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Setup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotorList_Setup", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBRotorList_Setup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBRotorList_Setup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotorList_Setup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotorList::Setup(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBRotorList::Setup(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetFixedBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_SetFixedBonds", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetFixedBonds" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetFixedBonds(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_Init(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; std::string *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotorList_Init", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_Init" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__string,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_Init" "', argument " "2"" of type '" "std::string &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_Init" "', argument " "2"" of type '" "std::string &""'"); }
   arg2 = reinterpret_cast< std::string * >(argp2); (arg1)->Init(*arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetQuiet(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetQuiet" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); (arg1)->SetQuiet(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetRotAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_SetRotAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetRotAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_SetRotAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_SetRotAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->SetRotAtoms(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_FindRotors__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_FindRotors" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_FindRotors" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_FindRotors" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotorList_FindRotors" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3); result = (bool)(arg1)->FindRotors(*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_FindRotors__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_FindRotors" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_FindRotors" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_FindRotors" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->FindRotors(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_FindRotors(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotorList_FindRotors", 0, 3, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBRotorList_FindRotors__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBRotorList_FindRotors__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotorList_FindRotors'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotorList::FindRotors(OpenBabel::OBMol &,bool)\n"
  "    OpenBabel::OBRotorList::FindRotors(OpenBabel::OBMol &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetEvalAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_SetEvalAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetEvalAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_SetEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_SetEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->SetEvalAtoms(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_AssignTorVals(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_AssignTorVals", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_AssignTorVals" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_AssignTorVals" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_AssignTorVals" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->AssignTorVals(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_BeginRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBRotorIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBRotor *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_BeginRotor", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_BeginRotor" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_BeginRotor" "', argument " "2"" of type '" "OpenBabel::OBRotorIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_BeginRotor" "', argument " "2"" of type '" "OpenBabel::OBRotorIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBRotorIterator * >(argp2); result = (OpenBabel::OBRotor *)(arg1)->BeginRotor(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_NextRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBRotorIterator *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; OpenBabel::OBRotor *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_NextRotor", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_NextRotor" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_NextRotor" "', argument " "2"" of type '" "OpenBabel::OBRotorIterator &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_NextRotor" "', argument " "2"" of type '" "OpenBabel::OBRotorIterator &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBRotorIterator * >(argp2); result = (OpenBabel::OBRotor *)(arg1)->NextRotor(*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotor, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_BeginRotors(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBRotor *,std::allocator< OpenBabel::OBRotor * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_BeginRotors" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (arg1)->BeginRotors();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBRotorIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_EndRotors(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< OpenBabel::OBRotor *,std::allocator< OpenBabel::OBRotor * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_EndRotors" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (arg1)->EndRotors();
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBRotorIterator(result)), SWIGTYPE_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_IdentifyEvalAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_IdentifyEvalAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_IdentifyEvalAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_IdentifyEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_IdentifyEvalAtoms" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->IdentifyEvalAtoms(*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetFixAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_SetFixAtoms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetFixAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_SetFixAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_SetFixAtoms" "', argument " "2"" of type '" "OpenBabel::OBBitVec &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetFixAtoms(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_HasFixedAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_HasFixedAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); result = (bool)(arg1)->HasFixedAtoms();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_IgnoreSymmetryRemoval(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_IgnoreSymmetryRemoval" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1); (arg1)->IgnoreSymmetryRemoval(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorList_SetRotAtomsByFix(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotorList_SetRotAtomsByFix", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorList_SetRotAtomsByFix" "', argument " "1"" of type '" "OpenBabel::OBRotorList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotorList_SetRotAtomsByFix" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotorList_SetRotAtomsByFix" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetRotAtomsByFix(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRotorList_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotorList, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotorList_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_rotor_digit__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; unsigned int arg1 ; unsigned int val1 ; int ecode1 = 0 ; OpenBabel::rotor_digit *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_rotor_digit" "', argument " "1"" of type '" "unsigned int""'"); }
    arg1 = static_cast< unsigned int >(val1); result = (OpenBabel::rotor_digit *)new OpenBabel::rotor_digit(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__rotor_digit, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_rotor_digit__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::rotor_digit *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::rotor_digit *)new OpenBabel::rotor_digit();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__rotor_digit, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_rotor_digit(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_rotor_digit", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_rotor_digit__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_rotor_digit__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_rotor_digit'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::rotor_digit::rotor_digit(unsigned int)\n"
  "    OpenBabel::rotor_digit::rotor_digit()\n"); return 0; }
SWIGINTERN PyObject *_wrap_rotor_digit_set_size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "rotor_digit_set_size", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotor_digit_set_size" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rotor_digit_set_size" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->set_size(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_rotor_digit_set_state(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "rotor_digit_set_state", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotor_digit_set_state" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "rotor_digit_set_state" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->set_state(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_rotor_digit_get_state(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotor_digit_get_state" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); result = (int)(arg1)->get_state();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_rotor_digit_size(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotor_digit_size" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); result = (unsigned int)(arg1)->size();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_rotor_digit_next(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "rotor_digit_next" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); result = (bool)(arg1)->next();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_rotor_digit(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::rotor_digit *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__rotor_digit, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_rotor_digit" "', argument " "1"" of type '" "OpenBabel::rotor_digit *""'");  }
   arg1 = reinterpret_cast< OpenBabel::rotor_digit * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *rotor_digit_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__rotor_digit, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *rotor_digit_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotorKeys(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *result = 0 ; (void)self; if (!SWIG_Python_UnpackTuple(args, "new_OBRotorKeys", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotorKeys *)new OpenBabel::OBRotorKeys();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotorKeys, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorKeys_Clear(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorKeys_Clear" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); (arg1)->Clear(); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorKeys_NumKeys(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorKeys_NumKeys" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); result = (unsigned int)(arg1)->NumKeys();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorKeys_AddRotor(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotorKeys_AddRotor", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorKeys_AddRotor" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotorKeys_AddRotor" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2); (arg1)->AddRotor(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorKeys_Next(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorKeys_Next" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); result = (bool)(arg1)->Next();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotorKeys_GetKey(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< int,std::allocator< int > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotorKeys_GetKey" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); result = (arg1)->GetKey();
  resultobj = swig::from(static_cast< std::vector< int,std::allocator< int > > >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotorKeys(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotorKeys *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotorKeys, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotorKeys" "', argument " "1"" of type '" "OpenBabel::OBRotorKeys *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotorKeys * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBRotorKeys_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotorKeys, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotorKeys_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBRotamerList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBRotamerList", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRotamerList *)new OpenBabel::OBRotamerList();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRotamerList, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRotamerList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRotamerList" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_Setup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::OBRotorList *arg3 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_Setup" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_OpenBabel__OBRotorList,  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRotamerList_Setup" "', argument " "3"" of type '" "OpenBabel::OBRotorList &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_Setup" "', argument " "3"" of type '" "OpenBabel::OBRotorList &""'"); }
   arg3 = reinterpret_cast< OpenBabel::OBRotorList * >(argp3); (arg1)->Setup(*arg2,*arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_Setup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; unsigned char *arg3 = 0 ;
  int arg4 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; int val4 ;
  int ecode4 = 0 ; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_Setup" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_unsigned_char, 0 |  0 ); if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRotamerList_Setup" "', argument " "3"" of type '" "unsigned char *""'");  }
   arg3 = reinterpret_cast< unsigned char * >(argp3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBRotamerList_Setup" "', argument " "4"" of type '" "int""'"); }  
  arg4 = static_cast< int >(val4); (arg1)->Setup(*arg2,arg3,arg4); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_Setup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotamerList_Setup", 0, 4, argv))) SWIG_fail; --argc; if (argc == 3) {
    PyObject *retobj = _wrap_OBRotamerList_Setup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBRotamerList_Setup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotamerList_Setup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotamerList::Setup(OpenBabel::OBMol &,OpenBabel::OBRotorList &)\n"
  "    OpenBabel::OBRotamerList::Setup(OpenBabel::OBMol &,unsigned char *,int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotamerList_NumRotors(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_NumRotors" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  result = (unsigned int)((OpenBabel::OBRotamerList const *)arg1)->NumRotors();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_NumRotamers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_NumRotamers" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  result = (unsigned int)((OpenBabel::OBRotamerList const *)arg1)->NumRotamers();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamer__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; double *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_AddRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_AddRotamer" "', argument " "2"" of type '" "double *""'");  }
   arg2 = reinterpret_cast< double * >(argp2); (arg1)->AddRotamer(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamer__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; int *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_AddRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_int, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_AddRotamer" "', argument " "2"" of type '" "int *""'");  }
   arg2 = reinterpret_cast< int * >(argp2); (arg1)->AddRotamer(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamer__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; std::vector< int,std::allocator< int > > arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_AddRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBRotamerList_AddRotamer" "', argument " "2"" of type '" "std::vector< int,std::allocator< int > >""'");  }
     arg2 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; }  (arg1)->AddRotamer(SWIG_STD_MOVE(arg2)); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamer__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; unsigned char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_AddRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_AddRotamer" "', argument " "2"" of type '" "unsigned char *""'");  }
   arg2 = reinterpret_cast< unsigned char * >(argp2); (arg1)->AddRotamer(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamer(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotamerList_AddRotamer", 0, 2, argv))) SWIG_fail; --argc; if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_double, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_OBRotamerList_AddRotamer__SWIG_0(self, argc, argv);}  check_1: if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_int, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_OBRotamerList_AddRotamer__SWIG_1(self, argc, argv);}  check_2: if (argc == 2) {
    int _v = 0; { void *vptr = 0; int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_unsigned_char, 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap_OBRotamerList_AddRotamer__SWIG_3(self, argc, argv);} 
  check_3: if (argc == 2) { PyObject *retobj = _wrap_OBRotamerList_AddRotamer__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotamerList_AddRotamer'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotamerList::AddRotamer(double *)\n"
  "    OpenBabel::OBRotamerList::AddRotamer(int *)\n"
  "    OpenBabel::OBRotamerList::AddRotamer(std::vector< int,std::allocator< int > >)\n"
  "    OpenBabel::OBRotamerList::AddRotamer(unsigned char *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBRotamerList_AddRotamers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; unsigned char *arg2 = 0 ; int arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; int val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_AddRotamers", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_AddRotamers" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_AddRotamers" "', argument " "2"" of type '" "unsigned char *""'");  }
   arg2 = reinterpret_cast< unsigned char * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotamerList_AddRotamers" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); (arg1)->AddRotamers(arg2,arg3); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_GetReferenceArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; unsigned char *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_GetReferenceArray", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_GetReferenceArray" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_unsigned_char, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_GetReferenceArray" "', argument " "2"" of type '" "unsigned char *""'");  }
   arg2 = reinterpret_cast< unsigned char * >(argp2); ((OpenBabel::OBRotamerList const *)arg1)->GetReferenceArray(arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_BeginRotamer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< unsigned char *,std::allocator< unsigned char * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_BeginRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); result = (arg1)->BeginRotamer();
  resultobj = SWIG_NewPointerObj((new std::vector< unsigned char *,std::allocator< unsigned char * > >::iterator(result)), SWIGTYPE_p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_EndRotamer(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  SwigValueWrapper< std::vector< unsigned char *,std::allocator< unsigned char * > >::iterator > result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_EndRotamer" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); result = (arg1)->EndRotamer();
  resultobj = SWIG_NewPointerObj((new std::vector< unsigned char *,std::allocator< unsigned char * > >::iterator(result)), SWIGTYPE_p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_CreateConformerList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; SwigValueWrapper< std::vector< double *,std::allocator< double * > > > result;
  (void)self; if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_CreateConformerList", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_CreateConformerList" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_CreateConformerList" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_CreateConformerList" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (arg1)->CreateConformerList(*arg2);
  resultobj = SWIG_NewPointerObj((new std::vector< double *,std::allocator< double * > >(result)), SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_ExpandConformerList(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ;
  std::vector< double *,std::allocator< double * > > *arg3 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; void *argp3 = 0 ; int res3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_ExpandConformerList", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_ExpandConformerList" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_ExpandConformerList" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_ExpandConformerList" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBRotamerList_ExpandConformerList" "', argument " "3"" of type '" "std::vector< double *,std::allocator< double * > > &""'");  }
   if (!argp3) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_ExpandConformerList" "', argument " "3"" of type '" "std::vector< double *,std::allocator< double * > > &""'"); }
   arg3 = reinterpret_cast< std::vector< double *,std::allocator< double * > > * >(argp3);
  (arg1)->ExpandConformerList(*arg2,*arg3); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_SetCurrentCoordinates(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; std::vector< int,std::allocator< int > > arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_SetCurrentCoordinates", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_SetCurrentCoordinates" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_SetCurrentCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_SetCurrentCoordinates" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    int res = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "OBRotamerList_SetCurrentCoordinates" "', argument " "3"" of type '" "std::vector< int,std::allocator< int > >""'");  }
     arg3 = *ptr; if (SWIG_IsNewObj(res)) delete ptr; } 
  result = (bool)(arg1)->SetCurrentCoordinates(*arg2,SWIG_STD_MOVE(arg3));
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_SetBaseCoordinateSets__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetBaseCoordinateSets(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_SetBaseCoordinateSets__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBRotamerList *arg1 = 0 ;
  SwigValueWrapper< std::vector< double *,std::allocator< double * > > > arg2 ; unsigned int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 ; int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "1"" of type '" "OpenBabel::OBRotamerList *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); {
    res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t,  0  | 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >""'");  }
       if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "2"" of type '" "std::vector< double *,std::allocator< double * > >""'"); }
     else {
      std::vector< double *,std::allocator< double * > > * temp = reinterpret_cast< std::vector< double *,std::allocator< double * > > * >(argp2);
      arg2 = *temp; if (SWIG_IsNewObj(res2)) delete temp; }  }  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBRotamerList_SetBaseCoordinateSets" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); (arg1)->SetBaseCoordinateSets(SWIG_STD_MOVE(arg2),arg3);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_SetBaseCoordinateSets(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBRotamerList_SetBaseCoordinateSets", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 2) { PyObject *retobj = _wrap_OBRotamerList_SetBaseCoordinateSets__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBRotamerList_SetBaseCoordinateSets__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBRotamerList_SetBaseCoordinateSets'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBRotamerList::SetBaseCoordinateSets(OpenBabel::OBMol &)\n"
  "    OpenBabel::OBRotamerList::SetBaseCoordinateSets(std::vector< double *,std::allocator< double * > >,unsigned int)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBRotamerList_NumBaseCoordinateSets(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_NumBaseCoordinateSets" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  result = (unsigned int)((OpenBabel::OBRotamerList const *)arg1)->NumBaseCoordinateSets();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_GetBaseCoordinateSet(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; unsigned int arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; double *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBRotamerList_GetBaseCoordinateSet", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_GetBaseCoordinateSet" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBRotamerList_GetBaseCoordinateSet" "', argument " "2"" of type '" "unsigned int""'"); }
    arg2 = static_cast< unsigned int >(val2);
  result = (double *)((OpenBabel::OBRotamerList const *)arg1)->GetBaseCoordinateSet(arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBRotamerList_NumAtoms(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRotamerList *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRotamerList, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBRotamerList_NumAtoms" "', argument " "1"" of type '" "OpenBabel::OBRotamerList const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRotamerList * >(argp1);
  result = (unsigned int)((OpenBabel::OBRotamerList const *)arg1)->NumAtoms();
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRotamerList_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRotamerList, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRotamerList_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSpectrophore__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBSpectrophore *)new OpenBabel::OBSpectrophore();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSpectrophore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSpectrophore__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBSpectrophore *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBSpectrophore,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSpectrophore" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBSpectrophore" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  result = (OpenBabel::OBSpectrophore *)new OpenBabel::OBSpectrophore((OpenBabel::OBSpectrophore const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSpectrophore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSpectrophore(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBSpectrophore", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBSpectrophore__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBSpectrophore__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBSpectrophore'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSpectrophore::OBSpectrophore()\n"
  "    OpenBabel::OBSpectrophore::OBSpectrophore(OpenBabel::OBSpectrophore const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBSpectrophore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSpectrophore" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetResolution__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetResolution" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSpectrophore_SetResolution" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetResolution(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetResolution__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetResolution" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); (arg1)->SetResolution(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetResolution(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBSpectrophore_SetResolution", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBSpectrophore_SetResolution__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBSpectrophore_SetResolution__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSpectrophore_SetResolution'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSpectrophore::SetResolution(double const)\n"
  "    OpenBabel::OBSpectrophore::SetResolution()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetAccuracy__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; OpenBabel::OBSpectrophore::AccuracyOption arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetAccuracy" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSpectrophore_SetAccuracy" "', argument " "2"" of type '" "OpenBabel::OBSpectrophore::AccuracyOption""'"); }
    arg2 = static_cast< OpenBabel::OBSpectrophore::AccuracyOption >(val2); (arg1)->SetAccuracy(arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetAccuracy__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetAccuracy" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); (arg1)->SetAccuracy(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetAccuracy(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBSpectrophore_SetAccuracy", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBSpectrophore_SetAccuracy__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBSpectrophore_SetAccuracy__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSpectrophore_SetAccuracy'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSpectrophore::SetAccuracy(OpenBabel::OBSpectrophore::AccuracyOption const)\n"
  "    OpenBabel::OBSpectrophore::SetAccuracy()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetStereo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; OpenBabel::OBSpectrophore::StereoOption arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetStereo" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSpectrophore_SetStereo" "', argument " "2"" of type '" "OpenBabel::OBSpectrophore::StereoOption""'"); }
    arg2 = static_cast< OpenBabel::OBSpectrophore::StereoOption >(val2); (arg1)->SetStereo(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetStereo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetStereo" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); (arg1)->SetStereo(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetStereo(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBSpectrophore_SetStereo", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBSpectrophore_SetStereo__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBSpectrophore_SetStereo__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSpectrophore_SetStereo'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSpectrophore::SetStereo(OpenBabel::OBSpectrophore::StereoOption const)\n"
  "    OpenBabel::OBSpectrophore::SetStereo()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetNormalization__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; OpenBabel::OBSpectrophore::NormalizationOption arg2 ;
  void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetNormalization" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSpectrophore_SetNormalization" "', argument " "2"" of type '" "OpenBabel::OBSpectrophore::NormalizationOption""'"); }
    arg2 = static_cast< OpenBabel::OBSpectrophore::NormalizationOption >(val2); (arg1)->SetNormalization(arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetNormalization__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_SetNormalization" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1); (arg1)->SetNormalization(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_SetNormalization(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSpectrophore_SetNormalization", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBSpectrophore_SetNormalization__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBSpectrophore_SetNormalization__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSpectrophore_SetNormalization'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBSpectrophore::SetNormalization(OpenBabel::OBSpectrophore::NormalizationOption const)\n"
  "    OpenBabel::OBSpectrophore::SetNormalization()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_GetAccuracy(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSpectrophore::AccuracyOption result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_GetAccuracy" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  result = (OpenBabel::OBSpectrophore::AccuracyOption)((OpenBabel::OBSpectrophore const *)arg1)->GetAccuracy();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_GetResolution(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_GetResolution" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  result = (double)((OpenBabel::OBSpectrophore const *)arg1)->GetResolution();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_GetStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSpectrophore::StereoOption result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_GetStereo" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  result = (OpenBabel::OBSpectrophore::StereoOption)((OpenBabel::OBSpectrophore const *)arg1)->GetStereo();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_GetNormalization(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSpectrophore::NormalizationOption result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_GetNormalization" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  result = (OpenBabel::OBSpectrophore::NormalizationOption)((OpenBabel::OBSpectrophore const *)arg1)->GetNormalization();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSpectrophore_GetSpectrophore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSpectrophore *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; std::vector< double,std::allocator< double > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSpectrophore_GetSpectrophore", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSpectrophore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSpectrophore_GetSpectrophore" "', argument " "1"" of type '" "OpenBabel::OBSpectrophore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSpectrophore * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSpectrophore_GetSpectrophore" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (arg1)->GetSpectrophore(arg2);
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result)); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBSpectrophore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSpectrophore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSpectrophore_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBConformerFilter_IsGood(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerFilter *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; OpenBabel::RotorKey *arg3 = 0 ; double *arg4 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int res3 = SWIG_OLDOBJ ; void *argp4 = 0 ; int res4 = 0 ;
  PyObject *swig_obj[4] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerFilter_IsGood", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerFilter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerFilter_IsGood" "', argument " "1"" of type '" "OpenBabel::OBConformerFilter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerFilter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerFilter_IsGood" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerFilter_IsGood" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); {
    std::vector< int,std::allocator< int > > *ptr = (std::vector< int,std::allocator< int > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr); if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "OBConformerFilter_IsGood" "', argument " "3"" of type '" "OpenBabel::RotorKey const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerFilter_IsGood" "', argument " "3"" of type '" "OpenBabel::RotorKey const &""'"); }
     arg3 = ptr; }  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBConformerFilter_IsGood" "', argument " "4"" of type '" "double *""'");  }
   arg4 = reinterpret_cast< double * >(argp4);
  result = (bool)(arg1)->IsGood((OpenBabel::OBMol const &)*arg2,(OpenBabel::RotorKey const &)*arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); if (SWIG_IsNewObj(res3)) delete arg3; return resultobj; fail:
  if (SWIG_IsNewObj(res3)) delete arg3; return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBConformerFilter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerFilter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerFilter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConformerFilter" "', argument " "1"" of type '" "OpenBabel::OBConformerFilter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerFilter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBConformerFilter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConformerFilter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBConformerFilters(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > *arg1 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBConformerFilters *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBConformerFilters" "', argument " "1"" of type '" "std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBConformerFilters" "', argument " "1"" of type '" "std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > const &""'"); }
  
  arg1 = reinterpret_cast< std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > * >(argp1);
  result = (OpenBabel::OBConformerFilters *)new OpenBabel::OBConformerFilters((std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConformerFilters, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBConformerFilters(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerFilters *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerFilters, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConformerFilters" "', argument " "1"" of type '" "OpenBabel::OBConformerFilters *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerFilters * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBConformerFilters_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConformerFilters, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBConformerFilters_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBStericConformerFilter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBStericConformerFilter *result = 0 ; (void)self;
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBStericConformerFilter *)new OpenBabel::OBStericConformerFilter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStericConformerFilter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double arg2 ; bool arg3 ; double val1 ; int ecode1 = 0 ; double val2 ; int ecode2 = 0 ;
  bool val3 ; int ecode3 = 0 ; OpenBabel::OBStericConformerFilter *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBStericConformerFilter" "', argument " "1"" of type '" "double""'"); }
    arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBStericConformerFilter" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBStericConformerFilter" "', argument " "3"" of type '" "bool""'"); }
    arg3 = static_cast< bool >(val3);
  result = (OpenBabel::OBStericConformerFilter *)new OpenBabel::OBStericConformerFilter(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStericConformerFilter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double arg2 ; double val1 ; int ecode1 = 0 ; double val2 ; int ecode2 = 0 ;
  OpenBabel::OBStericConformerFilter *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_double(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBStericConformerFilter" "', argument " "1"" of type '" "double""'"); }
    arg1 = static_cast< double >(val1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBStericConformerFilter" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2);
  result = (OpenBabel::OBStericConformerFilter *)new OpenBabel::OBStericConformerFilter(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStericConformerFilter__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; double arg1 ; double val1 ; int ecode1 = 0 ; OpenBabel::OBStericConformerFilter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_double(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBStericConformerFilter" "', argument " "1"" of type '" "double""'"); }
    arg1 = static_cast< double >(val1);
  result = (OpenBabel::OBStericConformerFilter *)new OpenBabel::OBStericConformerFilter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBStericConformerFilter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBStericConformerFilter", 0, 3, argv))) SWIG_fail; --argc;
  if (argc == 0) { PyObject *retobj = _wrap_new_OBStericConformerFilter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) {
    PyObject *retobj = _wrap_new_OBStericConformerFilter__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBStericConformerFilter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_new_OBStericConformerFilter__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBStericConformerFilter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBStericConformerFilter::OBStericConformerFilter()\n"
  "    OpenBabel::OBStericConformerFilter::OBStericConformerFilter(double,double,bool)\n"
  "    OpenBabel::OBStericConformerFilter::OBStericConformerFilter(double,double)\n"
  "    OpenBabel::OBStericConformerFilter::OBStericConformerFilter(double)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBStericConformerFilter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBStericConformerFilter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBStericConformerFilter" "', argument " "1"" of type '" "OpenBabel::OBStericConformerFilter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBStericConformerFilter * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBStericConformerFilter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBStericConformerFilter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBStericConformerFilter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBConformerScore_GetPreferred(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBConformerScore::Preferred result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerScore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerScore_GetPreferred" "', argument " "1"" of type '" "OpenBabel::OBConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerScore * >(argp1);
  result = (OpenBabel::OBConformerScore::Preferred)(arg1)->GetPreferred();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerScore_GetConvergence(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBConformerScore::Convergence result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerScore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerScore_GetConvergence" "', argument " "1"" of type '" "OpenBabel::OBConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerScore * >(argp1);
  result = (OpenBabel::OBConformerScore::Convergence)(arg1)->GetConvergence();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerScore_Score(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerScore *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; unsigned int arg3 ; OpenBabel::RotorKeys *arg4 = 0 ;
  std::vector< double *,std::allocator< double * > > *arg5 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; unsigned int val3 ; int ecode3 = 0 ; int res4 = SWIG_OLDOBJ ; void *argp5 = 0 ; int res5 = 0 ;
  PyObject *swig_obj[5] ; double result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerScore_Score", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerScore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerScore_Score" "', argument " "1"" of type '" "OpenBabel::OBConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerScore * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerScore_Score" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerScore_Score" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConformerScore_Score" "', argument " "3"" of type '" "unsigned int""'"); }
    arg3 = static_cast< unsigned int >(val3); {
    std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *ptr = (std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr); if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "OBConformerScore_Score" "', argument " "4"" of type '" "OpenBabel::RotorKeys const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerScore_Score" "', argument " "4"" of type '" "OpenBabel::RotorKeys const &""'"); }
     arg4 = ptr; } 
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_std__vectorT_double_p_std__allocatorT_double_p_t_t,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "OBConformerScore_Score" "', argument " "5"" of type '" "std::vector< double *,std::allocator< double * > > const &""'");  }
   if (!argp5) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerScore_Score" "', argument " "5"" of type '" "std::vector< double *,std::allocator< double * > > const &""'"); }
   arg5 = reinterpret_cast< std::vector< double *,std::allocator< double * > > * >(argp5);
  result = (double)(arg1)->Score(*arg2,arg3,(OpenBabel::RotorKeys const &)*arg4,(std::vector< double *,std::allocator< double * > > const &)*arg5);
  resultobj = SWIG_From_double(static_cast< double >(result)); if (SWIG_IsNewObj(res4)) delete arg4; return resultobj; fail:
  if (SWIG_IsNewObj(res4)) delete arg4; return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConformerScore" "', argument " "1"" of type '" "OpenBabel::OBConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerScore * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBConformerScore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConformerScore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBRMSDConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRMSDConformerScore *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBRMSDConformerScore", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBRMSDConformerScore *)new OpenBabel::OBRMSDConformerScore();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRMSDConformerScore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBRMSDConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBRMSDConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBRMSDConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBRMSDConformerScore" "', argument " "1"" of type '" "OpenBabel::OBRMSDConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBRMSDConformerScore * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBRMSDConformerScore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBRMSDConformerScore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBRMSDConformerScore_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBEnergyConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBEnergyConformerScore *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBEnergyConformerScore", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBEnergyConformerScore *)new OpenBabel::OBEnergyConformerScore();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBEnergyConformerScore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBEnergyConformerScore_GetNbEnergyCompute(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBEnergyConformerScore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBEnergyConformerScore_GetNbEnergyCompute" "', argument " "1"" of type '" "OpenBabel::OBEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBEnergyConformerScore * >(argp1); result = (unsigned long)(arg1)->GetNbEnergyCompute();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBEnergyConformerScore_GetNbEnergyRequest(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBEnergyConformerScore, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBEnergyConformerScore_GetNbEnergyRequest" "', argument " "1"" of type '" "OpenBabel::OBEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBEnergyConformerScore * >(argp1); result = (unsigned long)(arg1)->GetNbEnergyRequest();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBEnergyConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBEnergyConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBEnergyConformerScore" "', argument " "1"" of type '" "OpenBabel::OBEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBEnergyConformerScore * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBEnergyConformerScore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBEnergyConformerScore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBEnergyConformerScore_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMinimizingEnergyConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMinimizingEnergyConformerScore *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBMinimizingEnergyConformerScore", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBMinimizingEnergyConformerScore *)new OpenBabel::OBMinimizingEnergyConformerScore();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMinimizingEnergyConformerScore_GetNbEnergyCompute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBMinimizingEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; unsigned long result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMinimizingEnergyConformerScore_GetNbEnergyCompute" "', argument " "1"" of type '" "OpenBabel::OBMinimizingEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMinimizingEnergyConformerScore * >(argp1);
  result = (unsigned long)(arg1)->GetNbEnergyCompute();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMinimizingEnergyConformerScore_GetNbEnergyRequest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0; OpenBabel::OBMinimizingEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  PyObject *swig_obj[1] ; unsigned long result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMinimizingEnergyConformerScore_GetNbEnergyRequest" "', argument " "1"" of type '" "OpenBabel::OBMinimizingEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMinimizingEnergyConformerScore * >(argp1);
  result = (unsigned long)(arg1)->GetNbEnergyRequest();
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBMinimizingEnergyConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMinimizingEnergyConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMinimizingEnergyConformerScore" "', argument " "1"" of type '" "OpenBabel::OBMinimizingEnergyConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMinimizingEnergyConformerScore * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBMinimizingEnergyConformerScore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMinimizingEnergyConformerScore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMinimizingEnergyConformerScore_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMinimizingRMSDConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMinimizingRMSDConformerScore *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBMinimizingRMSDConformerScore", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBMinimizingRMSDConformerScore *)new OpenBabel::OBMinimizingRMSDConformerScore();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMinimizingRMSDConformerScore, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBMinimizingRMSDConformerScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMinimizingRMSDConformerScore *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMinimizingRMSDConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMinimizingRMSDConformerScore" "', argument " "1"" of type '" "OpenBabel::OBMinimizingRMSDConformerScore *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMinimizingRMSDConformerScore * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBMinimizingRMSDConformerScore_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMinimizingRMSDConformerScore, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMinimizingRMSDConformerScore_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBConformerSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBConformerSearch", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBConformerSearch *)new OpenBabel::OBConformerSearch();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBConformerSearch, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBConformerSearch(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBConformerSearch" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; int arg3 ; int arg4 ;
  int arg5 ; int arg6 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ;
  int ecode4 = 0 ; int val5 ; int ecode5 = 0 ; int val6 ; int ecode6 = 0 ; bool result; (void)self;
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Setup" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConformerSearch_Setup" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConformerSearch_Setup" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBConformerSearch_Setup" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5); ecode6 = SWIG_AsVal_int(swig_obj[5], &val6); if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "OBConformerSearch_Setup" "', argument " "6"" of type '" "int""'"); }
    arg6 = static_cast< int >(val6); result = (bool)(arg1)->Setup((OpenBabel::OBMol const &)*arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; int arg3 ; int arg4 ;
  int arg5 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ;
  int ecode4 = 0 ; int val5 ; int ecode5 = 0 ; bool result; (void)self; if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Setup" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConformerSearch_Setup" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConformerSearch_Setup" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); ecode5 = SWIG_AsVal_int(swig_obj[4], &val5); if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "OBConformerSearch_Setup" "', argument " "5"" of type '" "int""'"); }
    arg5 = static_cast< int >(val5); result = (bool)(arg1)->Setup((OpenBabel::OBMol const &)*arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; int arg3 ; int arg4 ;
  void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; int val4 ; int ecode4 = 0 ;
  bool result; (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Setup" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConformerSearch_Setup" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBConformerSearch_Setup" "', argument " "4"" of type '" "int""'"); }
    arg4 = static_cast< int >(val4); result = (bool)(arg1)->Setup((OpenBabel::OBMol const &)*arg2,arg3,arg4);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; int arg3 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; int val3 ; int ecode3 = 0 ; bool result; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Setup" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBConformerSearch_Setup" "', argument " "3"" of type '" "int""'"); }
    arg3 = static_cast< int >(val3); result = (bool)(arg1)->Setup((OpenBabel::OBMol const &)*arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ;
  int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ; bool result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Setup" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_Setup" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->Setup((OpenBabel::OBMol const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Setup(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[7] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConformerSearch_Setup", 0, 6, argv))) SWIG_fail; --argc; if (argc == 2) {
    PyObject *retobj = _wrap_OBConformerSearch_Setup__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBConformerSearch_Setup__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBConformerSearch_Setup__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 5) {
    PyObject *retobj = _wrap_OBConformerSearch_Setup__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 6) {
    PyObject *retobj = _wrap_OBConformerSearch_Setup__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConformerSearch_Setup'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConformerSearch::Setup(OpenBabel::OBMol const &,int,int,int,int)\n"
  "    OpenBabel::OBConformerSearch::Setup(OpenBabel::OBMol const &,int,int,int)\n"
  "    OpenBabel::OBConformerSearch::Setup(OpenBabel::OBMol const &,int,int)\n"
  "    OpenBabel::OBConformerSearch::Setup(OpenBabel::OBMol const &,int)\n"
  "    OpenBabel::OBConformerSearch::Setup(OpenBabel::OBMol const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetNumConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetNumConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetNumConformers" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetNumConformers" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetNumConformers(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetNumChildren(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetNumChildren", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetNumChildren" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetNumChildren" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetNumChildren(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetMutability(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetMutability", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetMutability" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetMutability" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetMutability(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetConvergence(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetConvergence", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetConvergence" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetConvergence" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetConvergence(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetFixedBonds(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBBitVec *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetFixedBonds", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetFixedBonds" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBBitVec,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_SetFixedBonds" "', argument " "2"" of type '" "OpenBabel::OBBitVec const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBBitVec * >(argp2); (arg1)->SetFixedBonds((OpenBabel::OBBitVec const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetFilter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBConformerFilter *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetFilter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetFilter" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&arg2), SWIGTYPE_p_OpenBabel__OBConformerFilter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_SetFilter" "', argument " "2"" of type '" "OpenBabel::OBConformerFilter *""'"); }
   (arg1)->SetFilter(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetScore(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBConformerScore *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetScore", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetScore" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&arg2), SWIGTYPE_p_OpenBabel__OBConformerScore, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_SetScore" "', argument " "2"" of type '" "OpenBabel::OBConformerScore *""'"); }
   (arg1)->SetScore(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_PrintRotors(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_PrintRotors", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_PrintRotors" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_PrintRotors" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->PrintRotors(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_Search(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_Search" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); (arg1)->Search(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetRotorKeys(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::RotorKeys *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetRotorKeys" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  result = (OpenBabel::RotorKeys *) &((OpenBabel::OBConformerSearch const *)arg1)->GetRotorKeys();
  resultobj = swig::from(static_cast< std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > >(*result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetConformers(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_GetConformers", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetConformers" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_GetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBConformerSearch_GetConformers" "', argument " "2"" of type '" "OpenBabel::OBMol &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->GetConformers(*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetLogStream(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; std::ostream *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetLogStream", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetLogStream" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__ostream, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBConformerSearch_SetLogStream" "', argument " "2"" of type '" "std::ostream *""'");  }
   arg2 = reinterpret_cast< std::ostream * >(argp2); (arg1)->SetLogStream(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetSharing__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetSharing" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); (arg1)->SetSharing(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetSharing__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); (arg1)->SetSharing(); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetSharing(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = {
    0} ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetSharing", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBConformerSearch_SetSharing__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBConformerSearch_SetSharing__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConformerSearch_SetSharing'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConformerSearch::SetSharing(bool)\n"
  "    OpenBabel::OBConformerSearch::SetSharing()\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetNbNiches(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetNbNiches" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (int)(arg1)->GetNbNiches();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetNbNiches(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetNbNiches", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetNbNiches" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetNbNiches" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetNbNiches(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetNicheRadius(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetNicheRadius" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (double)(arg1)->GetNicheRadius();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetNicheRadius(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetNicheRadius", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetNicheRadius" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetNicheRadius" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetNicheRadius(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetAlphaSharing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetAlphaSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (double)(arg1)->GetAlphaSharing();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetAlphaSharing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetAlphaSharing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetAlphaSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetAlphaSharing" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetAlphaSharing(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetSigmaSharing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetSigmaSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (double)(arg1)->GetSigmaSharing();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetSigmaSharing(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetSigmaSharing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetSigmaSharing" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetSigmaSharing" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetSigmaSharing(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetCrossoverProbability(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetCrossoverProbability" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (double)(arg1)->GetCrossoverProbability();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetCrossoverProbability(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetCrossoverProbability", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetCrossoverProbability" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetCrossoverProbability" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetCrossoverProbability(arg2); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_GetNicheMating(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_GetNicheMating" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (double)(arg1)->GetNicheMating();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetNicheMating(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBConformerSearch *arg1 = 0 ; double arg2 ; void *argp1 = 0 ; int res1 = 0 ; double val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetNicheMating", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetNicheMating" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetNicheMating" "', argument " "2"" of type '" "double""'"); }
    arg2 = static_cast< double >(val2); (arg1)->SetNicheMating(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetLocalOptRate__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetLocalOptRate" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBConformerSearch_SetLocalOptRate" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); (arg1)->SetLocalOptRate(arg2); resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetLocalOptRate__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBConformerSearch *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; int result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBConformerSearch, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBConformerSearch_SetLocalOptRate" "', argument " "1"" of type '" "OpenBabel::OBConformerSearch *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBConformerSearch * >(argp1); result = (int)(arg1)->SetLocalOptRate();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBConformerSearch_SetLocalOptRate(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBConformerSearch_SetLocalOptRate", 0, 2, argv))) SWIG_fail; --argc;
  if (argc == 1) { PyObject *retobj = _wrap_OBConformerSearch_SetLocalOptRate__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBConformerSearch_SetLocalOptRate__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBConformerSearch_SetLocalOptRate'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBConformerSearch::SetLocalOptRate(int)\n"
  "    OpenBabel::OBConformerSearch::SetLocalOptRate()\n"); return 0; }
SWIGINTERN PyObject *OBConformerSearch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBConformerSearch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBConformerSearch_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  bool arg1 ; bool arg2 ; bool val1 ; int ecode1 = 0 ; bool val2 ; int ecode2 = 0 ; OpenBabel::OBAlign *result = 0 ; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "bool""'"); }  
  arg1 = static_cast< bool >(val1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBAlign" "', argument " "2"" of type '" "bool""'"); }  
  arg2 = static_cast< bool >(val2); result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  bool arg1 ; bool val1 ; int ecode1 = 0 ; OpenBabel::OBAlign *result = 0 ; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "bool""'"); }  
  arg1 = static_cast< bool >(val1); result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBAlign *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; bool arg4 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; bool val3 ; int ecode3 = 0 ; bool val4 ; int ecode4 = 0 ; OpenBabel::OBAlign *result = 0 ;
  (void)self; if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBAlign" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3); ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4); if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_OBAlign" "', argument " "4"" of type '" "bool""'"); }  
  arg4 = static_cast< bool >(val4);
  result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign((OpenBabel::OBMol const &)*arg1,(OpenBabel::OBMol const &)*arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; bool arg3 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ;
  int res2 = 0 ; bool val3 ; int ecode3 = 0 ; OpenBabel::OBAlign *result = 0 ; (void)self;
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_OBAlign" "', argument " "3"" of type '" "bool""'"); }  
  arg3 = static_cast< bool >(val3);
  result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign((OpenBabel::OBMol const &)*arg1,(OpenBabel::OBMol const &)*arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  OpenBabel::OBAlign *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2);
  result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign((OpenBabel::OBMol const &)*arg1,(OpenBabel::OBMol const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign__SWIG_6(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg1 = 0 ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg2 = 0 ; int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ; OpenBabel::OBAlign *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail; {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr); if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBAlign" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "1"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg1 = ptr; }  {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_OBAlign" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBAlign" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg2 = ptr; } 
  result = (OpenBabel::OBAlign *)new OpenBabel::OBAlign((std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg1,(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res1)) delete arg1; if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_new_OBAlign(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBAlign", 0, 4, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBAlign__SWIG_2(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 1) { PyObject *retobj = _wrap_new_OBAlign__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_3; { int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);}  if (!_v) goto check_3; return _wrap_new_OBAlign__SWIG_5(self, argc, argv);}  check_3:
  if (argc == 2) { int _v = 0; { { int res = SWIG_AsVal_bool(argv[0], NULL); _v = SWIG_CheckState(res); } } 
    if (!_v) goto check_4; { { int res = SWIG_AsVal_bool(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_4;
    return _wrap_new_OBAlign__SWIG_0(self, argc, argv);}  check_4: if (argc == 2) {
    PyObject *retobj = _wrap_new_OBAlign__SWIG_6(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  if (argc == 3) { PyObject *retobj = _wrap_new_OBAlign__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_new_OBAlign__SWIG_3(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBAlign'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAlign::OBAlign(bool,bool)\n" "    OpenBabel::OBAlign::OBAlign(bool)\n"
  "    OpenBabel::OBAlign::OBAlign()\n"
  "    OpenBabel::OBAlign::OBAlign(OpenBabel::OBMol const &,OpenBabel::OBMol const &,bool,bool)\n"
  "    OpenBabel::OBAlign::OBAlign(OpenBabel::OBMol const &,OpenBabel::OBMol const &,bool)\n"
  "    OpenBabel::OBAlign::OBAlign(OpenBabel::OBMol const &,OpenBabel::OBMol const &)\n"
  "    OpenBabel::OBAlign::OBAlign(std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &,std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)\n");
  return 0; }
SWIGINTERN PyObject *_wrap_OBAlign_SetRef(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAlign_SetRef", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_SetRef" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAlign_SetRef" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAlign_SetRef" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetRef((std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_SetTarget(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *arg2 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAlign_SetTarget", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_SetTarget" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); {
    std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *ptr = (std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAlign_SetTarget" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAlign_SetTarget" "', argument " "2"" of type '" "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &""'"); }
     arg2 = ptr; }  (arg1)->SetTarget((std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > const &)*arg2);
  resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2; return resultobj; fail:
  if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_SetRefMol(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAlign_SetRefMol", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_SetRefMol" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAlign_SetRefMol" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAlign_SetRefMol" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetRefMol((OpenBabel::OBMol const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_SetTargetMol(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; (void)self; if (!SWIG_Python_UnpackTuple(args, "OBAlign_SetTargetMol", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_SetTargetMol" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAlign_SetTargetMol" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBAlign_SetTargetMol" "', argument " "2"" of type '" "OpenBabel::OBMol const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); (arg1)->SetTargetMol((OpenBabel::OBMol const &)*arg2);
  resultobj = SWIG_Py_Void(); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_Align(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_Align" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); result = (bool)(arg1)->Align();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_SetMethod(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; enum OpenBabel::OBAlign::AlignMethod arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAlign_SetMethod", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_SetMethod" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBAlign_SetMethod" "', argument " "2"" of type '" "enum OpenBabel::OBAlign::AlignMethod""'"); }
    arg2 = static_cast< enum OpenBabel::OBAlign::AlignMethod >(val2); (arg1)->SetMethod(arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_GetRMSD(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_GetRMSD" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); result = (double)(arg1)->GetRMSD();
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_GetRotMatrix(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::matrix3x3 result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_GetRotMatrix" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); result = (arg1)->GetRotMatrix();
  resultobj = SWIG_NewPointerObj((new OpenBabel::matrix3x3(result)), SWIGTYPE_p_OpenBabel__matrix3x3, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_GetAlignment(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > result; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_GetAlignment" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); result = (arg1)->GetAlignment();
  resultobj = swig::from(static_cast< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBAlign_UpdateCoords(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; OpenBabel::OBMol *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ; void *argp2 = 0 ; int res2 = 0 ;
  PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBAlign_UpdateCoords", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBAlign_UpdateCoords" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBAlign_UpdateCoords" "', argument " "2"" of type '" "OpenBabel::OBMol *""'");  }
   arg2 = reinterpret_cast< OpenBabel::OBMol * >(argp2); result = (bool)(arg1)->UpdateCoords(arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBAlign(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAlign *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAlign, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBAlign" "', argument " "1"" of type '" "OpenBabel::OBAlign *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAlign * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *OBAlign_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAlign, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBAlign_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolAtomIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolAtomIter *)new OpenBabel::OBMolAtomIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMolAtomIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolAtomIter *)new OpenBabel::OBMolAtomIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolAtomIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolAtomIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1);
  result = (OpenBabel::OBMolAtomIter *)new OpenBabel::OBMolAtomIter((OpenBabel::OBMolAtomIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolAtomIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolAtomIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolAtomIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolAtomIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolAtomIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomIter::OBMolAtomIter()\n"
  "    OpenBabel::OBMolAtomIter::OBMolAtomIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolAtomIter::OBMolAtomIter(OpenBabel::OBMolAtomIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolAtomIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolAtomIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1);
  result = (bool)((OpenBabel::OBMolAtomIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1); result = (OpenBabel::OBMolAtomIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomIter, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolAtomIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBMolAtomIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomIter(result)), SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBMolAtomIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBMolAtomIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBMolAtomIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBMolAtomIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomIter::operator ++()\n"
  "    OpenBabel::OBMolAtomIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBMolAtomIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomIter * >(argp1);
  result = (OpenBabel::OBAtom *) &((OpenBabel::OBMolAtomIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolAtomIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolAtomIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolAtomIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolAtomDFSIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomDFSIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolAtomDFSIter *)new OpenBabel::OBMolAtomDFSIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomDFSIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::OBMolAtomDFSIter *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomDFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new__OBMolAtomDFSIter" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (OpenBabel::OBMolAtomDFSIter *)new OpenBabel::OBMolAtomDFSIter(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomDFSIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomDFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomDFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolAtomDFSIter *)new OpenBabel::OBMolAtomDFSIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomDFSIter__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomDFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomDFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolAtomDFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1);
  result = (OpenBabel::OBMolAtomDFSIter *)new OpenBabel::OBMolAtomDFSIter((OpenBabel::OBMolAtomDFSIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomDFSIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolAtomDFSIter", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolAtomDFSIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolAtomDFSIter__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolAtomDFSIter__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new__OBMolAtomDFSIter__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolAtomDFSIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomDFSIter::OBMolAtomDFSIter()\n"
  "    OpenBabel::OBMolAtomDFSIter::OBMolAtomDFSIter(OpenBabel::OBMol *,int)\n"
  "    OpenBabel::OBMolAtomDFSIter::OBMolAtomDFSIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolAtomDFSIter::OBMolAtomDFSIter(OpenBabel::OBMolAtomDFSIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolAtomDFSIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolAtomDFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomDFSIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1);
  result = (bool)((OpenBabel::OBMolAtomDFSIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomDFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomDFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1);
  result = (OpenBabel::OBMolAtomDFSIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolAtomDFSIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomDFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBMolAtomDFSIter_inc" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomDFSIter(result)), SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBMolAtomDFSIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBMolAtomDFSIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBMolAtomDFSIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBMolAtomDFSIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomDFSIter::operator ++()\n"
  "    OpenBabel::OBMolAtomDFSIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomDFSIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1);
  result = (OpenBabel::OBAtom *) &((OpenBabel::OBMolAtomDFSIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap__OBMolAtomDFSIter_next(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomDFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomDFSIter_next" "', argument " "1"" of type '" "OpenBabel::OBMolAtomDFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomDFSIter * >(argp1); result = (OpenBabel::OBAtom *)(arg1)->next();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_OBMolAtomDFSIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolAtomDFSIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolAtomDFSIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolAtomBFSIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomBFSIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolAtomBFSIter *)new OpenBabel::OBMolAtomBFSIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomBFSIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::OBMolAtomBFSIter *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new__OBMolAtomBFSIter" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (OpenBabel::OBMolAtomBFSIter *)new OpenBabel::OBMolAtomBFSIter(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomBFSIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolAtomBFSIter *)new OpenBabel::OBMolAtomBFSIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomBFSIter__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAtomBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolAtomBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1);
  result = (OpenBabel::OBMolAtomBFSIter *)new OpenBabel::OBMolAtomBFSIter((OpenBabel::OBMolAtomBFSIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAtomBFSIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolAtomBFSIter", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolAtomBFSIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolAtomBFSIter__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolAtomBFSIter__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new__OBMolAtomBFSIter__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolAtomBFSIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomBFSIter::OBMolAtomBFSIter()\n"
  "    OpenBabel::OBMolAtomBFSIter::OBMolAtomBFSIter(OpenBabel::OBMol *,int)\n"
  "    OpenBabel::OBMolAtomBFSIter::OBMolAtomBFSIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolAtomBFSIter::OBMolAtomBFSIter(OpenBabel::OBMolAtomBFSIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolAtomBFSIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolAtomBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomBFSIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1);
  result = (bool)((OpenBabel::OBMolAtomBFSIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAtomBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomBFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1);
  result = (OpenBabel::OBMolAtomBFSIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolAtomBFSIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomBFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBMolAtomBFSIter_inc" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomBFSIter(result)), SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBMolAtomBFSIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBMolAtomBFSIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBMolAtomBFSIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBMolAtomBFSIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAtomBFSIter::operator ++()\n"
  "    OpenBabel::OBMolAtomBFSIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomBFSIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1);
  result = (OpenBabel::OBAtom *) &((OpenBabel::OBMolAtomBFSIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap__OBMolAtomBFSIter_CurrentDepth(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAtomBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAtomBFSIter_CurrentDepth" "', argument " "1"" of type '" "OpenBabel::OBMolAtomBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAtomBFSIter * >(argp1);
  result = (int)((OpenBabel::OBMolAtomBFSIter const *)arg1)->CurrentDepth();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_OBMolAtomBFSIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolAtomBFSIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolAtomBFSIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBMolBondBFSIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondBFSIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolBondBFSIter *)new OpenBabel::OBMolBondBFSIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMolBondBFSIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ;
  OpenBabel::OBMolBondBFSIter *result = 0 ; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBMolBondBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_OBMolBondBFSIter" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (OpenBabel::OBMolBondBFSIter *)new OpenBabel::OBMolBondBFSIter(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMolBondBFSIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolBondBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBMolBondBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolBondBFSIter *)new OpenBabel::OBMolBondBFSIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMolBondBFSIter__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolBondBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolBondBFSIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBMolBondBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new_OBMolBondBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1);
  result = (OpenBabel::OBMolBondBFSIter *)new OpenBabel::OBMolBondBFSIter((OpenBabel::OBMolBondBFSIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBMolBondBFSIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_OBMolBondBFSIter", 0, 2, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new_OBMolBondBFSIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new_OBMolBondBFSIter__SWIG_2(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new_OBMolBondBFSIter__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_new_OBMolBondBFSIter__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_OBMolBondBFSIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolBondBFSIter::OBMolBondBFSIter()\n"
  "    OpenBabel::OBMolBondBFSIter::OBMolBondBFSIter(OpenBabel::OBMol *,int)\n"
  "    OpenBabel::OBMolBondBFSIter::OBMolBondBFSIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolBondBFSIter::OBMolBondBFSIter(OpenBabel::OBMolBondBFSIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete_OBMolBondBFSIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBMolBondBFSIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMolBondBFSIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1);
  result = (bool)((OpenBabel::OBMolBondBFSIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolBondBFSIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMolBondBFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1);
  result = (OpenBabel::OBMolBondBFSIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondBFSIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolBondBFSIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMolBondBFSIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBMolBondBFSIter_inc" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolBondBFSIter(result)), SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBMolBondBFSIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBMolBondBFSIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap_OBMolBondBFSIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBMolBondBFSIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolBondBFSIter::operator ++()\n"
  "    OpenBabel::OBMolBondBFSIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBBond *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMolBondBFSIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1);
  result = (OpenBabel::OBBond *) &((OpenBabel::OBMolBondBFSIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBMolBondBFSIter_CurrentDepth(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondBFSIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; int result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBMolBondBFSIter_CurrentDepth" "', argument " "1"" of type '" "OpenBabel::OBMolBondBFSIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondBFSIter * >(argp1);
  result = (int)((OpenBabel::OBMolBondBFSIter const *)arg1)->CurrentDepth();
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBMolBondBFSIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolBondBFSIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBMolBondBFSIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolBondIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolBondIter *)new OpenBabel::OBMolBondIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolBondIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMolBondIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolBondIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolBondIter *)new OpenBabel::OBMolBondIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolBondIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolBondIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolBondIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolBondIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolBondIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1);
  result = (OpenBabel::OBMolBondIter *)new OpenBabel::OBMolBondIter((OpenBabel::OBMolBondIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolBondIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolBondIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolBondIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolBondIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolBondIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolBondIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolBondIter::OBMolBondIter()\n"
  "    OpenBabel::OBMolBondIter::OBMolBondIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolBondIter::OBMolBondIter(OpenBabel::OBMolBondIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolBondIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolBondIter" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolBondIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolBondIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1);
  result = (bool)((OpenBabel::OBMolBondIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolBondIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolBondIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolBondIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1); result = (OpenBabel::OBMolBondIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolBondIter, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap__OBMolBondIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolBondIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolBondIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolBondIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBMolBondIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolBondIter(result)), SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolBondIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBMolBondIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBMolBondIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBMolBondIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBMolBondIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolBondIter::operator ++()\n"
  "    OpenBabel::OBMolBondIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBMolBondIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBond *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolBondIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolBondIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolBondIter * >(argp1);
  result = (OpenBabel::OBBond *) &((OpenBabel::OBMolBondIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolBondIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolBondIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolBondIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBAtomAtomIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBAtomAtomIter *)new OpenBabel::OBAtomAtomIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomAtomIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBAtomAtomIter" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (OpenBabel::OBAtomAtomIter *)new OpenBabel::OBAtomAtomIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomAtomIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAtomAtomIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBAtomAtomIter" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBAtomAtomIter" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1);
  result = (OpenBabel::OBAtomAtomIter *)new OpenBabel::OBAtomAtomIter((OpenBabel::OBAtomAtomIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomAtomIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBAtomAtomIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBAtomAtomIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBAtomAtomIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBAtomAtomIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBAtomAtomIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomAtomIter::OBAtomAtomIter()\n"
  "    OpenBabel::OBAtomAtomIter::OBAtomAtomIter(OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBAtomAtomIter::OBAtomAtomIter(OpenBabel::OBAtomAtomIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBAtomAtomIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBAtomAtomIter" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomAtomIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomAtomIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1);
  result = (bool)((OpenBabel::OBAtomAtomIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomAtomIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1); result = (OpenBabel::OBAtomAtomIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomAtomIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomAtomIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomAtomIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBAtomAtomIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBAtomAtomIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomAtomIter(result)), SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomAtomIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBAtomAtomIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBAtomAtomIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBAtomAtomIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBAtomAtomIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomAtomIter::operator ++()\n"
  "    OpenBabel::OBAtomAtomIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBAtomAtomIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBAtom *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomAtomIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBAtomAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomAtomIter * >(argp1);
  result = (OpenBabel::OBAtom *) &((OpenBabel::OBAtomAtomIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBAtomAtomIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomAtomIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBAtomAtomIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBAtomBondIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBAtomBondIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBAtomBondIter *)new OpenBabel::OBAtomBondIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomBondIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomBondIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBAtomBondIter" "', argument " "1"" of type '" "OpenBabel::OBAtom *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = (OpenBabel::OBAtomBondIter *)new OpenBabel::OBAtomBondIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomBondIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomBondIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAtomBondIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBAtomBondIter" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBAtomBondIter" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1);
  result = (OpenBabel::OBAtomBondIter *)new OpenBabel::OBAtomBondIter((OpenBabel::OBAtomBondIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBAtomBondIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBAtomBondIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBAtomBondIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBAtom, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBAtomBondIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBAtomBondIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBAtomBondIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomBondIter::OBAtomBondIter()\n"
  "    OpenBabel::OBAtomBondIter::OBAtomBondIter(OpenBabel::OBAtom *)\n"
  "    OpenBabel::OBAtomBondIter::OBAtomBondIter(OpenBabel::OBAtomBondIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBAtomBondIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBAtomBondIter" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomBondIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomBondIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1);
  result = (bool)((OpenBabel::OBAtomBondIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomBondIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBAtomBondIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomBondIter_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1); result = (OpenBabel::OBAtomBondIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtomBondIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomBondIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtomBondIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBAtomBondIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomBondIter_inc" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBAtomBondIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomBondIter(result)), SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBAtomBondIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBAtomBondIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBAtomBondIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBAtomBondIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBAtomBondIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBAtomBondIter::operator ++()\n"
  "    OpenBabel::OBAtomBondIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBAtomBondIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBAtomBondIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBBond *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtomBondIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBAtomBondIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBAtomBondIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtomBondIter * >(argp1);
  result = (OpenBabel::OBBond *) &((OpenBabel::OBAtomBondIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBBond, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBAtomBondIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBAtomBondIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBAtomBondIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBResidueIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBResidueIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBResidueIter *)new OpenBabel::OBResidueIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBResidueIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBResidueIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBResidueIter *)new OpenBabel::OBResidueIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBResidueIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBResidueIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBResidueIter" "', argument " "1"" of type '" "OpenBabel::OBResidueIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBResidueIter" "', argument " "1"" of type '" "OpenBabel::OBResidueIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1);
  result = (OpenBabel::OBResidueIter *)new OpenBabel::OBResidueIter((OpenBabel::OBResidueIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBResidueIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBResidueIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBResidueIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBResidueIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBResidueIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidueIter::OBResidueIter()\n"
  "    OpenBabel::OBResidueIter::OBResidueIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBResidueIter::OBResidueIter(OpenBabel::OBResidueIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBResidueIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBResidueIter" "', argument " "1"" of type '" "OpenBabel::OBResidueIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBResidueIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1);
  result = (bool)((OpenBabel::OBResidueIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBResidueIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueIter_inc" "', argument " "1"" of type '" "OpenBabel::OBResidueIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1); result = (OpenBabel::OBResidueIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueIter, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBResidueIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueIter_inc" "', argument " "1"" of type '" "OpenBabel::OBResidueIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBResidueIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBResidueIter(result)), SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBResidueIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBResidueIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBResidueIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBResidueIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidueIter::operator ++()\n"
  "    OpenBabel::OBResidueIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBResidueIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBResidue *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBResidueIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueIter * >(argp1);
  result = (OpenBabel::OBResidue *) &((OpenBabel::OBResidueIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBResidueIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBResidueIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBResidueIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBResidueAtomIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBResidueAtomIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBResidueAtomIter *)new OpenBabel::OBResidueAtomIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueAtomIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidue *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBResidueAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidue, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBResidueAtomIter" "', argument " "1"" of type '" "OpenBabel::OBResidue *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidue * >(argp1);
  result = (OpenBabel::OBResidueAtomIter *)new OpenBabel::OBResidueAtomIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueAtomIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBResidueAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBResidueAtomIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBResidueAtomIter" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBResidueAtomIter" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1);
  result = (OpenBabel::OBResidueAtomIter *)new OpenBabel::OBResidueAtomIter((OpenBabel::OBResidueAtomIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBResidueAtomIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBResidueAtomIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBResidueAtomIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBResidue, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBResidueAtomIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBResidueAtomIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBResidueAtomIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidueAtomIter::OBResidueAtomIter()\n"
  "    OpenBabel::OBResidueAtomIter::OBResidueAtomIter(OpenBabel::OBResidue *)\n"
  "    OpenBabel::OBResidueAtomIter::OBResidueAtomIter(OpenBabel::OBResidueAtomIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBResidueAtomIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBResidueAtomIter" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueAtomIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueAtomIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1);
  result = (bool)((OpenBabel::OBResidueAtomIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueAtomIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBResidueAtomIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1);
  result = (OpenBabel::OBResidueAtomIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBResidueAtomIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueAtomIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBResidueAtomIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBResidueAtomIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueAtomIter_inc" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBResidueAtomIter_inc" "', argument " "2"" of type '" "int""'"); }
    arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBResidueAtomIter(result)), SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBResidueAtomIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBResidueAtomIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBResidueAtomIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBResidueAtomIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBResidueAtomIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBResidueAtomIter::operator ++()\n"
  "    OpenBabel::OBResidueAtomIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBResidueAtomIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBResidueAtomIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBAtom *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBResidueAtomIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBResidueAtomIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBResidueAtomIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBResidueAtomIter * >(argp1);
  result = (OpenBabel::OBAtom *) &((OpenBabel::OBResidueAtomIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBResidueAtomIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBResidueAtomIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBResidueAtomIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolAngleIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolAngleIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolAngleIter *)new OpenBabel::OBMolAngleIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAngleIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAngleIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMolAngleIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAngleIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolAngleIter *)new OpenBabel::OBMolAngleIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAngleIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAngleIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolAngleIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolAngleIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolAngleIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolAngleIter" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolAngleIter" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolAngleIter * >(argp1);
  result = (OpenBabel::OBMolAngleIter *)new OpenBabel::OBMolAngleIter((OpenBabel::OBMolAngleIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAngleIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolAngleIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolAngleIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolAngleIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolAngleIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolAngleIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolAngleIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolAngleIter::OBMolAngleIter()\n"
  "    OpenBabel::OBMolAngleIter::OBMolAngleIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolAngleIter::OBMolAngleIter(OpenBabel::OBMolAngleIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolAngleIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAngleIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAngleIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolAngleIter" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAngleIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAngleIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAngleIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAngleIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAngleIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAngleIter * >(argp1);
  result = (bool)((OpenBabel::OBMolAngleIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAngleIter_inc(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAngleIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBMolAngleIter *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAngleIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAngleIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAngleIter * >(argp1); result = (OpenBabel::OBMolAngleIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolAngleIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolAngleIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolAngleIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< unsigned int,std::allocator< unsigned int > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolAngleIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolAngleIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolAngleIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolAngleIter * >(argp1);
  result = ((OpenBabel::OBMolAngleIter const *)arg1)->operator *();
  resultobj = swig::from(static_cast< std::vector< unsigned int,std::allocator< unsigned int > > >(result)); return resultobj;
  fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolAngleIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolAngleIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolAngleIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolTorsionIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolTorsionIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolTorsionIter *)new OpenBabel::OBMolTorsionIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolTorsionIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolTorsionIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolTorsionIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolTorsionIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolTorsionIter *)new OpenBabel::OBMolTorsionIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolTorsionIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolTorsionIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolTorsionIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolTorsionIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolTorsionIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolTorsionIter" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolTorsionIter" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolTorsionIter * >(argp1);
  result = (OpenBabel::OBMolTorsionIter *)new OpenBabel::OBMolTorsionIter((OpenBabel::OBMolTorsionIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolTorsionIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolTorsionIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolTorsionIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolTorsionIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolTorsionIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolTorsionIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolTorsionIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolTorsionIter::OBMolTorsionIter()\n"
  "    OpenBabel::OBMolTorsionIter::OBMolTorsionIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolTorsionIter::OBMolTorsionIter(OpenBabel::OBMolTorsionIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolTorsionIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolTorsionIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolTorsionIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolTorsionIter" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolTorsionIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolTorsionIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolTorsionIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolTorsionIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolTorsionIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolTorsionIter * >(argp1);
  result = (bool)((OpenBabel::OBMolTorsionIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolTorsionIter_inc(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolTorsionIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBMolTorsionIter *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolTorsionIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolTorsionIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolTorsionIter * >(argp1);
  result = (OpenBabel::OBMolTorsionIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolTorsionIter, 0 |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolTorsionIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolTorsionIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< unsigned int,std::allocator< unsigned int > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolTorsionIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolTorsionIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolTorsionIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolTorsionIter * >(argp1);
  result = ((OpenBabel::OBMolTorsionIter const *)arg1)->operator *();
  resultobj = swig::from(static_cast< std::vector< unsigned int,std::allocator< unsigned int > > >(result)); return resultobj;
  fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolTorsionIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolTorsionIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolTorsionIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolPairIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolPairIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolPairIter *)new OpenBabel::OBMolPairIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolPairIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolPairIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMolPairIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolPairIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolPairIter *)new OpenBabel::OBMolPairIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolPairIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolPairIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolPairIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolPairIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolPairIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolPairIter" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolPairIter" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolPairIter * >(argp1);
  result = (OpenBabel::OBMolPairIter *)new OpenBabel::OBMolPairIter((OpenBabel::OBMolPairIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolPairIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolPairIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolPairIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolPairIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolPairIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolPairIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolPairIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolPairIter::OBMolPairIter()\n"
  "    OpenBabel::OBMolPairIter::OBMolPairIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolPairIter::OBMolPairIter(OpenBabel::OBMolPairIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolPairIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolPairIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolPairIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolPairIter" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolPairIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolPairIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolPairIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolPairIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolPairIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolPairIter * >(argp1);
  result = (bool)((OpenBabel::OBMolPairIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolPairIter_inc(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolPairIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBMolPairIter *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolPairIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolPairIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolPairIter * >(argp1); result = (OpenBabel::OBMolPairIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolPairIter, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap__OBMolPairIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolPairIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  std::vector< unsigned int,std::allocator< unsigned int > > result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolPairIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolPairIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolPairIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolPairIter * >(argp1);
  result = ((OpenBabel::OBMolPairIter const *)arg1)->operator *();
  resultobj = swig::from(static_cast< std::vector< unsigned int,std::allocator< unsigned int > > >(result)); return resultobj;
  fail: if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolPairIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolPairIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolPairIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new__OBMolRingIter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0; OpenBabel::OBMolRingIter *result = 0 ; (void)self; if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (OpenBabel::OBMolRingIter *)new OpenBabel::OBMolRingIter();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolRingIter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; OpenBabel::OBMolRingIter *result = 0 ;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolRingIter" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBMolRingIter *)new OpenBabel::OBMolRingIter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolRingIter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolRingIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolRingIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMolRingIter,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new__OBMolRingIter" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "new__OBMolRingIter" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1);
  result = (OpenBabel::OBMolRingIter *)new OpenBabel::OBMolRingIter((OpenBabel::OBMolRingIter const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new__OBMolRingIter(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new__OBMolRingIter", 0, 1, argv))) SWIG_fail; --argc; if (argc == 0) {
    PyObject *retobj = _wrap_new__OBMolRingIter__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 1) { int _v = 0; { void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_OpenBabel__OBMol, 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_2; return _wrap_new__OBMolRingIter__SWIG_1(self, argc, argv);}  check_2: if (argc == 1) {
    PyObject *retobj = _wrap_new__OBMolRingIter__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new__OBMolRingIter'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolRingIter::OBMolRingIter()\n"
  "    OpenBabel::OBMolRingIter::OBMolRingIter(OpenBabel::OBMol *)\n"
  "    OpenBabel::OBMolRingIter::OBMolRingIter(OpenBabel::OBMolRingIter const &)\n"); return 0; }
SWIGINTERN PyObject *_wrap_delete__OBMolRingIter(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolRingIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete__OBMolRingIter" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolRingIter___nonzero__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolRingIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolRingIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolRingIter___nonzero__" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1);
  result = (bool)((OpenBabel::OBMolRingIter const *)arg1)->operator bool();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolRingIter_inc__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolRingIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBMolRingIter *result = 0 ; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolRingIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolRingIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1); result = (OpenBabel::OBMolRingIter *) &(arg1)->operator ++();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBMolRingIter, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap__OBMolRingIter_inc__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBMolRingIter *arg1 = 0 ; int arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; OpenBabel::OBMolRingIter result; (void)self; if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolRingIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolRingIter_inc" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "_OBMolRingIter_inc" "', argument " "2"" of type '" "int""'"); }  
  arg2 = static_cast< int >(val2); result = (arg1)->operator ++(arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolRingIter(result)), SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap__OBMolRingIter_inc(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[3] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "_OBMolRingIter_inc", 0, 2, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap__OBMolRingIter_inc__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) {
    PyObject *retobj = _wrap__OBMolRingIter_inc__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function '_OBMolRingIter_inc'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBMolRingIter::operator ++()\n"
  "    OpenBabel::OBMolRingIter::operator ++(int)\n"); return 0; }
SWIGINTERN PyObject *_wrap__OBMolRingIter___ref__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMolRingIter *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; OpenBabel::OBRing *result = 0 ;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMolRingIter, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "_OBMolRingIter___ref__" "', argument " "1"" of type '" "OpenBabel::OBMolRingIter const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMolRingIter * >(argp1);
  result = (OpenBabel::OBRing *) &((OpenBabel::OBMolRingIter const *)arg1)->operator *();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBRing, 0 |  0 ); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_OBMolRingIter_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBMolRingIter, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_OBMolRingIter_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_MolGetAtoms__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MolGetAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "MolGetAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = OpenBabel::impl::MolGetAtoms((OpenBabel::OBMol const &)*arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_MolGetAtoms__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MolGetAtoms" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = OpenBabel::impl::MolGetAtoms((OpenBabel::OBMol const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_MolGetAtoms(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "MolGetAtoms", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_MolGetAtoms__SWIG_0(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_MolGetAtoms__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MolGetAtoms'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::impl::MolGetAtoms(OpenBabel::OBMol const &)\n"
  "    OpenBabel::impl::MolGetAtoms(OpenBabel::OBMol const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_MolGetBonds__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBMol,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MolGetBonds" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "MolGetBonds" "', argument " "1"" of type '" "OpenBabel::OBMol const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = OpenBabel::impl::MolGetBonds((OpenBabel::OBMol const &)*arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_MolGetBonds__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MolGetBonds" "', argument " "1"" of type '" "OpenBabel::OBMol const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1); result = OpenBabel::impl::MolGetBonds((OpenBabel::OBMol const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBMolBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_MolGetBonds(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "MolGetBonds", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBMol, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);} 
    if (!_v) goto check_1; return _wrap_MolGetBonds__SWIG_0(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_MolGetBonds__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MolGetBonds'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::impl::MolGetBonds(OpenBabel::OBMol const &)\n"
  "    OpenBabel::impl::MolGetBonds(OpenBabel::OBMol const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_AtomGetBonds__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAtom,  0  | 0); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AtomGetBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "AtomGetBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = OpenBabel::impl::AtomGetBonds((OpenBabel::OBAtom const &)*arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AtomGetBonds__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > > result;
  (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AtomGetBonds" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = OpenBabel::impl::AtomGetBonds((OpenBabel::OBAtom const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomBondRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AtomGetBonds(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "AtomGetBonds", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_AtomGetBonds__SWIG_0(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_AtomGetBonds__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'AtomGetBonds'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::impl::AtomGetBonds(OpenBabel::OBAtom const &)\n"
  "    OpenBabel::impl::AtomGetBonds(OpenBabel::OBAtom const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_AtomGetNbrs__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,OpenBabel::OBAtomAtomIterAdaptor > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_OpenBabel__OBAtom,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AtomGetNbrs" "', argument " "1"" of type '" "OpenBabel::OBAtom const &""'");  }
   if (!argp1) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "AtomGetNbrs" "', argument " "1"" of type '" "OpenBabel::OBAtom const &""'"); }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = OpenBabel::impl::AtomGetNbrs((OpenBabel::OBAtom const &)*arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AtomGetNbrs__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) { PyObject *resultobj = 0;
  OpenBabel::OBAtom *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  SwigValueWrapper< OpenBabel::OBRange< OpenBabel::OBAtom *,OpenBabel::OBAtomAtomIterAdaptor > > result; (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBAtom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AtomGetNbrs" "', argument " "1"" of type '" "OpenBabel::OBAtom const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBAtom * >(argp1);
  result = OpenBabel::impl::AtomGetNbrs((OpenBabel::OBAtom const *)arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBAtomAtomRange(result)), SWIGTYPE_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_AtomGetNbrs(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[2] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "AtomGetNbrs", 0, 1, argv))) SWIG_fail; --argc; if (argc == 1) { int _v = 0; {
      int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_OpenBabel__OBAtom, SWIG_POINTER_NO_NULL | 0); _v = SWIG_CheckState(res);}
     if (!_v) goto check_1; return _wrap_AtomGetNbrs__SWIG_0(self, argc, argv);}  check_1: if (argc == 1) {
    PyObject *retobj = _wrap_AtomGetNbrs__SWIG_1(self, argc, argv); if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj;
    SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'AtomGetNbrs'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::impl::AtomGetNbrs(OpenBabel::OBAtom const &)\n"
  "    OpenBabel::impl::AtomGetNbrs(OpenBabel::OBAtom const *)\n"); return 0; }
SWIGINTERN PyObject *_wrap_new_doubleArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0; size_t arg1 ;
  size_t val1 ; int ecode1 = 0 ; PyObject *swig_obj[1] ; doubleArray *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1); if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_doubleArray" "', argument " "1"" of type '" "size_t""'"); }  
  arg1 = static_cast< size_t >(val1); result = (doubleArray *)new_doubleArray(SWIG_STD_MOVE(arg1));
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_doubleArray, SWIG_POINTER_NEW |  0 ); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_doubleArray(PyObject *self, PyObject *args) { PyObject *resultobj = 0; doubleArray *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_doubleArray, SWIG_POINTER_DISOWN |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_doubleArray" "', argument " "1"" of type '" "doubleArray *""'");  }
   arg1 = reinterpret_cast< doubleArray * >(argp1); delete_doubleArray(arg1); resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_doubleArray___getitem__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  doubleArray *arg1 = 0 ; size_t arg2 ; void *argp1 = 0 ; int res1 = 0 ; size_t val2 ; int ecode2 = 0 ; PyObject *swig_obj[2] ;
  double result; (void)self; if (!SWIG_Python_UnpackTuple(args, "doubleArray___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_doubleArray, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "doubleArray___getitem__" "', argument " "1"" of type '" "doubleArray *""'");  }
   arg1 = reinterpret_cast< doubleArray * >(argp1); ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "doubleArray___getitem__" "', argument " "2"" of type '" "size_t""'"); }
    arg2 = static_cast< size_t >(val2); result = (double)doubleArray___getitem__(arg1,SWIG_STD_MOVE(arg2));
  resultobj = SWIG_From_double(static_cast< double >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_doubleArray___setitem__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  doubleArray *arg1 = 0 ; size_t arg2 ; double arg3 ; void *argp1 = 0 ; int res1 = 0 ; size_t val2 ; int ecode2 = 0 ;
  double val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "doubleArray___setitem__", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_doubleArray, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "doubleArray___setitem__" "', argument " "1"" of type '" "doubleArray *""'");  }
   arg1 = reinterpret_cast< doubleArray * >(argp1); ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "doubleArray___setitem__" "', argument " "2"" of type '" "size_t""'"); }
    arg2 = static_cast< size_t >(val2); ecode3 = SWIG_AsVal_double(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "doubleArray___setitem__" "', argument " "3"" of type '" "double""'"); }
    arg3 = static_cast< double >(val3); doubleArray___setitem__(arg1,SWIG_STD_MOVE(arg2),arg3); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_doubleArray_cast(PyObject *self, PyObject *args) { PyObject *resultobj = 0; doubleArray *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; double *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_doubleArray, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "doubleArray_cast" "', argument " "1"" of type '" "doubleArray *""'");  }
   arg1 = reinterpret_cast< doubleArray * >(argp1); result = (double *)doubleArray_cast(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_double, 0 |  0 ); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_doubleArray_frompointer(PyObject *self, PyObject *args) { PyObject *resultobj = 0; double *arg1 = 0 ;
  void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; doubleArray *result = 0 ; (void)self; if (!args) SWIG_fail;
  swig_obj[0] = args; res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_double, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "doubleArray_frompointer" "', argument " "1"" of type '" "double *""'");  }
   arg1 = reinterpret_cast< double * >(argp1); result = (doubleArray *)doubleArray_frompointer(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_doubleArray, 0 |  0 ); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *doubleArray_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_doubleArray, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *doubleArray_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_OBTetraNonPlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetraNonPlanarStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetraNonPlanarStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTetraNonPlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBTetraNonPlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetraNonPlanarStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBTetraNonPlanarStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTetraNonPlanarStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_OBTetraPlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetraPlanarStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetraPlanarStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTetraPlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBTetraPlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetraPlanarStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBTetraPlanarStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTetraPlanarStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_OBTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBTetrahedralStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBTetrahedralStereo *)new OpenBabel::OBTetrahedralStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBTetrahedralStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTetrahedralStereo" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_IsValid" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1);
  result = (bool)((OpenBabel::OBTetrahedralStereo const *)arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralStereo *arg1 = 0 ; OpenBabel::OBTetrahedralStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralStereo___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo___eq__" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTetrahedralStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp2);
  result = (bool)((OpenBabel::OBTetrahedralStereo const *)arg1)->operator ==((OpenBabel::OBTetrahedralStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralStereo *arg1 = 0 ; OpenBabel::OBTetrahedralStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralStereo___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo___ne__" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTetrahedralStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp2);
  result = (bool)((OpenBabel::OBTetrahedralStereo const *)arg1)->operator !=((OpenBabel::OBTetrahedralStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_SetConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralStereo *arg1 = 0 ; OpenBabel::OBTetrahedralConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralStereo_SetConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_SetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTetrahedralConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp2);
  OpenBabel_OBTetrahedralStereo_SetConfig(arg1,(OpenBabel::OBTetrahedralConfig const &)*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; OpenBabel::OBStereo::Winding arg2 ;
  OpenBabel::OBStereo::View arg3 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ;
  OpenBabel::OBTetrahedralConfig result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "2"" of type '" "OpenBabel::OBStereo::Winding""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Winding >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "3"" of type '" "OpenBabel::OBStereo::View""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::View >(val3);
  result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_0(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; OpenBabel::OBStereo::Winding arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; OpenBabel::OBTetrahedralConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "2"" of type '" "OpenBabel::OBStereo::Winding""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Winding >(val2);
  result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBTetrahedralConfig result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1);
  result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_0(arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; unsigned long arg2 ; OpenBabel::OBStereo::Winding arg3 ;
  OpenBabel::OBStereo::View arg4 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; int val3 ;
  int ecode3 = 0 ; int val4 ; int ecode4 = 0 ; OpenBabel::OBTetrahedralConfig result; (void)self;
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "3"" of type '" "OpenBabel::OBStereo::Winding""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Winding >(val3); ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "4"" of type '" "OpenBabel::OBStereo::View""'"); }
    arg4 = static_cast< OpenBabel::OBStereo::View >(val4);
  result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_3(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_4(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; unsigned long arg2 ; OpenBabel::OBStereo::Winding arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ;
  OpenBabel::OBTetrahedralConfig result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "3"" of type '" "OpenBabel::OBStereo::Winding""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Winding >(val3);
  result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_3(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig__SWIG_5(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBTetrahedralStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBTetrahedralConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = OpenBabel_OBTetrahedralStereo_GetConfig__SWIG_3(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBTetrahedralConfig(result)), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralStereo_GetConfig(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[5] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBTetrahedralStereo_GetConfig", 0, 4, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBTetrahedralStereo_GetConfig__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_OBTetrahedralStereo_GetConfig__SWIG_1(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBTetrahedralStereo_GetConfig__SWIG_5(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_4;
    PyObject *retobj = _wrap_OBTetrahedralStereo_GetConfig__SWIG_0(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  check_4: if (argc == 3) {
    PyObject *retobj = _wrap_OBTetrahedralStereo_GetConfig__SWIG_4(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 4) {
    PyObject *retobj = _wrap_OBTetrahedralStereo_GetConfig__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBTetrahedralStereo_GetConfig'.\n"
  "  Possible C/C++ prototypes are:\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig(OpenBabel::OBStereo::Winding,OpenBabel::OBStereo::View)\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig(OpenBabel::OBStereo::Winding)\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig()\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig(unsigned long,OpenBabel::OBStereo::Winding,OpenBabel::OBStereo::View)\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig(unsigned long,OpenBabel::OBStereo::Winding)\n"
  "    OpenBabel::OBTetrahedralStereo::GetConfig(unsigned long)\n"); return 0; }
SWIGINTERN PyObject *OBTetrahedralStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTetrahedralStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBTetrahedralStereo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBCisTransStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBCisTransStereo *)new OpenBabel::OBCisTransStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCisTransStereo, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBCisTransStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBCisTransStereo" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_IsValid" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1);
  result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; OpenBabel::OBCisTransStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo___eq__" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBCisTransStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBCisTransStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBCisTransStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp2);
  result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->operator ==((OpenBabel::OBCisTransStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; OpenBabel::OBCisTransStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo___ne__" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBCisTransStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBCisTransStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBCisTransStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp2);
  result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->operator !=((OpenBabel::OBCisTransStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_IsOnSameAtom(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; unsigned long arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_IsOnSameAtom", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_IsOnSameAtom" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_IsOnSameAtom" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBCisTransStereo_IsOnSameAtom" "', argument " "3"" of type '" "unsigned long""'"); }
    arg3 = static_cast< unsigned long >(val3);
  result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->IsOnSameAtom(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_IsTrans(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; unsigned long arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_IsTrans", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_IsTrans" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_IsTrans" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBCisTransStereo_IsTrans" "', argument " "3"" of type '" "unsigned long""'"); }
    arg3 = static_cast< unsigned long >(val3); result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->IsTrans(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_IsCis(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; unsigned long arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_IsCis", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_IsCis" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_IsCis" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBCisTransStereo_IsCis" "', argument " "3"" of type '" "unsigned long""'"); }
    arg3 = static_cast< unsigned long >(val3); result = (bool)((OpenBabel::OBCisTransStereo const *)arg1)->IsCis(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetTransRef(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; unsigned long result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_GetTransRef", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetTransRef" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_GetTransRef" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (unsigned long)((OpenBabel::OBCisTransStereo const *)arg1)->GetTransRef(arg2);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetCisRef(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; unsigned long result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_GetCisRef", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetCisRef" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_GetCisRef" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (unsigned long)((OpenBabel::OBCisTransStereo const *)arg1)->GetCisRef(arg2);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_SetConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransStereo *arg1 = 0 ; OpenBabel::OBCisTransConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransStereo_SetConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_SetConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBCisTransConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp2);
  OpenBabel_OBCisTransStereo_SetConfig(arg1,(OpenBabel::OBCisTransConfig const &)*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetConfig__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCisTransStereo *arg1 = 0 ; OpenBabel::OBStereo::Shape arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; OpenBabel::OBCisTransConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_GetConfig" "', argument " "2"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Shape >(val2); result = OpenBabel_OBCisTransStereo_GetConfig__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBCisTransConfig(result)), SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetConfig__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCisTransStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBCisTransConfig result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); result = OpenBabel_OBCisTransStereo_GetConfig__SWIG_0(arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBCisTransConfig(result)), SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetConfig__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; OpenBabel::OBStereo::Shape arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ;
  OpenBabel::OBCisTransConfig result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBCisTransStereo_GetConfig" "', argument " "3"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Shape >(val3);
  result = OpenBabel_OBCisTransStereo_GetConfig__SWIG_2(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBCisTransConfig(result)), SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetConfig__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBCisTransStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBCisTransConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransStereo * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = OpenBabel_OBCisTransStereo_GetConfig__SWIG_2(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBCisTransConfig(result)), SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransStereo_GetConfig(PyObject *self, PyObject *args) { Py_ssize_t argc; PyObject *argv[4] = { 0}
  ; if (!(argc = SWIG_Python_UnpackTuple(args, "OBCisTransStereo_GetConfig", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBCisTransStereo_GetConfig__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_OBCisTransStereo_GetConfig__SWIG_0(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBCisTransStereo_GetConfig__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBCisTransStereo_GetConfig__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBCisTransStereo_GetConfig'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBCisTransStereo::GetConfig(OpenBabel::OBStereo::Shape)\n"
  "    OpenBabel::OBCisTransStereo::GetConfig()\n"
  "    OpenBabel::OBCisTransStereo::GetConfig(unsigned long,OpenBabel::OBStereo::Shape)\n"
  "    OpenBabel::OBCisTransStereo::GetConfig(unsigned long)\n"); return 0; }
SWIGINTERN PyObject *OBCisTransStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBCisTransStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBCisTransStereo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_OBSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBMol *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBSquarePlanarStereo *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBMol, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_OBSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBMol *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBMol * >(argp1);
  result = (OpenBabel::OBSquarePlanarStereo *)new OpenBabel::OBSquarePlanarStereo(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSquarePlanarStereo(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSquarePlanarStereo" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_IsValid(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_IsValid" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  result = (bool)((OpenBabel::OBSquarePlanarStereo const *)arg1)->IsValid();
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; OpenBabel::OBSquarePlanarStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo___eq__" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarStereo___eq__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp2);
  result = (bool)((OpenBabel::OBSquarePlanarStereo const *)arg1)->operator ==((OpenBabel::OBSquarePlanarStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; OpenBabel::OBSquarePlanarStereo *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo___ne__" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarStereo const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarStereo___ne__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarStereo const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp2);
  result = (bool)((OpenBabel::OBSquarePlanarStereo const *)arg1)->operator !=((OpenBabel::OBSquarePlanarStereo const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_IsTrans(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; unsigned long arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_IsTrans", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_IsTrans" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_IsTrans" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSquarePlanarStereo_IsTrans" "', argument " "3"" of type '" "unsigned long""'"); }
    arg3 = static_cast< unsigned long >(val3);
  result = (bool)((OpenBabel::OBSquarePlanarStereo const *)arg1)->IsTrans(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_IsCis(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; unsigned long arg3 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; unsigned long val3 ; int ecode3 = 0 ; PyObject *swig_obj[3] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_IsCis", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_IsCis" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_IsCis" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSquarePlanarStereo_IsCis" "', argument " "3"" of type '" "unsigned long""'"); }
    arg3 = static_cast< unsigned long >(val3); result = (bool)((OpenBabel::OBSquarePlanarStereo const *)arg1)->IsCis(arg2,arg3);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetTransRef(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; unsigned long result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_GetTransRef", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetTransRef" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_GetTransRef" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2);
  result = (unsigned long)((OpenBabel::OBSquarePlanarStereo const *)arg1)->GetTransRef(arg2);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetCisRefs(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; std::vector< unsigned long,std::allocator< unsigned long > > result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_GetCisRefs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetCisRefs" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_GetCisRefs" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = ((OpenBabel::OBSquarePlanarStereo const *)arg1)->GetCisRefs(arg2);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(result)); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_SetConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; OpenBabel::OBSquarePlanarConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_SetConfig", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_SetConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarStereo_SetConfig" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp2);
  OpenBabel_OBSquarePlanarStereo_SetConfig(arg1,(OpenBabel::OBSquarePlanarConfig const &)*arg2); resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetConfig__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; OpenBabel::OBStereo::Shape arg2 ; void *argp1 = 0 ;
  int res1 = 0 ; int val2 ; int ecode2 = 0 ; OpenBabel::OBSquarePlanarConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "2"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Shape >(val2);
  result = OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBSquarePlanarConfig(result)), SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetConfig__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  OpenBabel::OBSquarePlanarConfig result; (void)self; if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  result = OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_0(arg1);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBSquarePlanarConfig(result)), SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetConfig__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; OpenBabel::OBStereo::Shape arg3 ;
  void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ; int ecode2 = 0 ; int val3 ; int ecode3 = 0 ;
  OpenBabel::OBSquarePlanarConfig result; (void)self; if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); ecode3 = SWIG_AsVal_int(swig_obj[2], &val3); if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "3"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg3 = static_cast< OpenBabel::OBStereo::Shape >(val3);
  result = OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_2(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBSquarePlanarConfig(result)), SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetConfig__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0; OpenBabel::OBSquarePlanarStereo *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ;
  unsigned long val2 ; int ecode2 = 0 ; OpenBabel::OBSquarePlanarConfig result; (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarStereo *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarStereo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarStereo_GetConfig" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); result = OpenBabel_OBSquarePlanarStereo_GetConfig__SWIG_2(arg1,arg2);
  resultobj = SWIG_NewPointerObj((new OpenBabel::OBSquarePlanarConfig(result)), SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_OWN |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarStereo_GetConfig(PyObject *self, PyObject *args) { Py_ssize_t argc;
  PyObject *argv[4] = { 0} ;
  if (!(argc = SWIG_Python_UnpackTuple(args, "OBSquarePlanarStereo_GetConfig", 0, 3, argv))) SWIG_fail; --argc; if (argc == 1) {
    PyObject *retobj = _wrap_OBSquarePlanarStereo_GetConfig__SWIG_1(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 2) { int _v = 0; { {
        int res = SWIG_AsVal_int(argv[1], NULL); _v = SWIG_CheckState(res); } }  if (!_v) goto check_2;
    return _wrap_OBSquarePlanarStereo_GetConfig__SWIG_0(self, argc, argv);}  check_2: if (argc == 2) {
    PyObject *retobj = _wrap_OBSquarePlanarStereo_GetConfig__SWIG_3(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  if (argc == 3) {
    PyObject *retobj = _wrap_OBSquarePlanarStereo_GetConfig__SWIG_2(self, argc, argv);
    if (!SWIG_Python_TypeErrorOccurred(retobj)) return retobj; SWIG_fail;}  fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'OBSquarePlanarStereo_GetConfig'.\n"
  "  Possible C/C++ prototypes are:\n" "    OpenBabel::OBSquarePlanarStereo::GetConfig(OpenBabel::OBStereo::Shape)\n"
  "    OpenBabel::OBSquarePlanarStereo::GetConfig()\n"
  "    OpenBabel::OBSquarePlanarStereo::GetConfig(unsigned long,OpenBabel::OBStereo::Shape)\n"
  "    OpenBabel::OBSquarePlanarStereo::GetConfig(unsigned long)\n"); return 0; }
SWIGINTERN PyObject *OBSquarePlanarStereo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSquarePlanarStereo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSquarePlanarStereo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBTetrahedralConfig___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; OpenBabel::OBTetrahedralConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig___eq__" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTetrahedralConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp2);
  result = (bool)((OpenBabel::OBTetrahedralConfig const *)arg1)->operator ==((OpenBabel::OBTetrahedralConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; OpenBabel::OBTetrahedralConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig___ne__" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBTetrahedralConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBTetrahedralConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp2);
  result = (bool)((OpenBabel::OBTetrahedralConfig const *)arg1)->operator !=((OpenBabel::OBTetrahedralConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_center_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_center_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_center_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralConfig_center_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->center = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_center_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_center_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); result = (unsigned long) ((arg1)->center);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_from_or_towards_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_from_or_towards_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_from_or_towards_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralConfig_from_or_towards_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->from_or_towards = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_from_or_towards_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_from_or_towards_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); result = (unsigned long) ((arg1)->from_or_towards);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_refs_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; OpenBabel::OBStereo::Refs *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_refs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_refs_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBTetrahedralConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBTetrahedralConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->refs = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_refs_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Refs *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_refs_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); result = (OpenBabel::OBStereo::Refs *) & ((arg1)->refs);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(*result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_winding_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; OpenBabel::OBStereo::Winding arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_winding_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_winding_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralConfig_winding_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Winding""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Winding >(val2); if (arg1) (arg1)->winding = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_winding_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Winding result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_winding_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1);
  result = (OpenBabel::OBStereo::Winding) ((arg1)->winding); resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_view_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; OpenBabel::OBStereo::View arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_view_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_view_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralConfig_view_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::View""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::View >(val2); if (arg1) (arg1)->view = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_view_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::View result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_view_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); result = (OpenBabel::OBStereo::View) ((arg1)->view);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_specified_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBTetrahedralConfig_specified_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_specified_set" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBTetrahedralConfig_specified_set" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); if (arg1) (arg1)->specified = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBTetrahedralConfig_specified_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBTetrahedralConfig_specified_get" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); result = (bool) ((arg1)->specified);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBTetrahedralConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBTetrahedralConfig", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBTetrahedralConfig *)new OpenBabel::OBTetrahedralConfig();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBTetrahedralConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBTetrahedralConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBTetrahedralConfig" "', argument " "1"" of type '" "OpenBabel::OBTetrahedralConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBTetrahedralConfig * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBTetrahedralConfig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBTetrahedralConfig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBTetrahedralConfig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBCisTransConfig___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; OpenBabel::OBCisTransConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig___eq__" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBCisTransConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp2);
  result = (bool)((OpenBabel::OBCisTransConfig const *)arg1)->operator ==((OpenBabel::OBCisTransConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; OpenBabel::OBCisTransConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig___ne__" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBCisTransConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBCisTransConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp2);
  result = (bool)((OpenBabel::OBCisTransConfig const *)arg1)->operator !=((OpenBabel::OBCisTransConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_begin_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig_begin_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_begin_set" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransConfig_begin_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->begin = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_begin_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_begin_get" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); result = (unsigned long) ((arg1)->begin);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_end_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig_end_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_end_set" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransConfig_end_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->end = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_end_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_end_get" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); result = (unsigned long) ((arg1)->end);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_refs_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; OpenBabel::OBStereo::Refs *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig_refs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_refs_set" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBCisTransConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBCisTransConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->refs = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_refs_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Refs *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_refs_get" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); result = (OpenBabel::OBStereo::Refs *) & ((arg1)->refs);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(*result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_shape_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; OpenBabel::OBStereo::Shape arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig_shape_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_shape_set" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransConfig_shape_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Shape >(val2); if (arg1) (arg1)->shape = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_shape_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Shape result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_shape_get" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); result = (OpenBabel::OBStereo::Shape) ((arg1)->shape);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_specified_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBCisTransConfig_specified_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_specified_set" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBCisTransConfig_specified_set" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); if (arg1) (arg1)->specified = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBCisTransConfig_specified_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBCisTransConfig_specified_get" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); result = (bool) ((arg1)->specified);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBCisTransConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBCisTransConfig", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBCisTransConfig *)new OpenBabel::OBCisTransConfig();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBCisTransConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBCisTransConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBCisTransConfig" "', argument " "1"" of type '" "OpenBabel::OBCisTransConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBCisTransConfig * >(argp1); delete arg1; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *OBCisTransConfig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBCisTransConfig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBCisTransConfig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig___eq__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; OpenBabel::OBSquarePlanarConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig___eq__" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarConfig___eq__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp2);
  result = (bool)((OpenBabel::OBSquarePlanarConfig const *)arg1)->operator ==((OpenBabel::OBSquarePlanarConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig___ne__(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; OpenBabel::OBSquarePlanarConfig *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  void *argp2 = 0 ; int res2 = 0 ; PyObject *swig_obj[2] ; bool result; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig___ne__" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig const *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig,  0  | 0); if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'");  }
   if (!argp2) {
    SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarConfig___ne__" "', argument " "2"" of type '" "OpenBabel::OBSquarePlanarConfig const &""'"); }
   arg2 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp2);
  result = (bool)((OpenBabel::OBSquarePlanarConfig const *)arg1)->operator !=((OpenBabel::OBSquarePlanarConfig const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail:
  if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_TypeError)) { return NULL; }  PyErr_Clear();
  SWIG_Py_INCREF(Py_NotImplemented); return Py_NotImplemented; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_center_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; unsigned long arg2 ; void *argp1 = 0 ; int res1 = 0 ; unsigned long val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig_center_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_center_set" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val2); if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarConfig_center_set" "', argument " "2"" of type '" "unsigned long""'"); }
    arg2 = static_cast< unsigned long >(val2); if (arg1) (arg1)->center = arg2; resultobj = SWIG_Py_Void(); return resultobj;
  fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_center_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; unsigned long result;
  (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_center_get" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); result = (unsigned long) ((arg1)->center);
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_refs_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; OpenBabel::OBStereo::Refs *arg2 = 0 ; void *argp1 = 0 ; int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig_refs_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_refs_set" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); {
    std::vector< unsigned long,std::allocator< unsigned long > > *ptr = (std::vector< unsigned long,std::allocator< unsigned long > > *)0;
    res2 = swig::asptr(swig_obj[1], &ptr); if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "OBSquarePlanarConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     if (!ptr) {
      SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "OBSquarePlanarConfig_refs_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Refs const &""'"); }
     arg2 = ptr; }  if (arg1) (arg1)->refs = *arg2; resultobj = SWIG_Py_Void(); if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj; fail: if (SWIG_IsNewObj(res2)) delete arg2; return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_refs_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Refs *result = 0 ; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_refs_get" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); result = (OpenBabel::OBStereo::Refs *) & ((arg1)->refs);
  resultobj = swig::from(static_cast< std::vector< unsigned long,std::allocator< unsigned long > > >(*result));
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_shape_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; OpenBabel::OBStereo::Shape arg2 ; void *argp1 = 0 ; int res1 = 0 ; int val2 ;
  int ecode2 = 0 ; PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig_shape_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_shape_set" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarConfig_shape_set" "', argument " "2"" of type '" "OpenBabel::OBStereo::Shape""'"); }
    arg2 = static_cast< OpenBabel::OBStereo::Shape >(val2); if (arg1) (arg1)->shape = arg2; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_shape_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ;
  OpenBabel::OBStereo::Shape result; (void)self; if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_shape_get" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); result = (OpenBabel::OBStereo::Shape) ((arg1)->shape);
  resultobj = SWIG_From_int(static_cast< int >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_specified_set(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; bool arg2 ; void *argp1 = 0 ; int res1 = 0 ; bool val2 ; int ecode2 = 0 ;
  PyObject *swig_obj[2] ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "OBSquarePlanarConfig_specified_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_specified_set" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "OBSquarePlanarConfig_specified_set" "', argument " "2"" of type '" "bool""'"); }
    arg2 = static_cast< bool >(val2); if (arg1) (arg1)->specified = arg2; resultobj = SWIG_Py_Void(); return resultobj; fail:
  return NULL; }
SWIGINTERN PyObject *_wrap_OBSquarePlanarConfig_specified_get(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; bool result; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, 0 |  0 ); if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "OBSquarePlanarConfig_specified_get" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); result = (bool) ((arg1)->specified);
  resultobj = SWIG_From_bool(static_cast< bool >(result)); return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_new_OBSquarePlanarConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *result = 0 ; (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_OBSquarePlanarConfig", 0, 0, 0)) SWIG_fail;
  result = (OpenBabel::OBSquarePlanarConfig *)new OpenBabel::OBSquarePlanarConfig();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_NEW |  0 );
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *_wrap_delete_OBSquarePlanarConfig(PyObject *self, PyObject *args) { PyObject *resultobj = 0;
  OpenBabel::OBSquarePlanarConfig *arg1 = 0 ; void *argp1 = 0 ; int res1 = 0 ; PyObject *swig_obj[1] ; (void)self;
  if (!args) SWIG_fail; swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_OBSquarePlanarConfig" "', argument " "1"" of type '" "OpenBabel::OBSquarePlanarConfig *""'");  }
   arg1 = reinterpret_cast< OpenBabel::OBSquarePlanarConfig * >(argp1); delete arg1; resultobj = SWIG_Py_Void();
  return resultobj; fail: return NULL; }
SWIGINTERN PyObject *OBSquarePlanarConfig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_OpenBabel__OBSquarePlanarConfig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *OBSquarePlanarConfig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

static PyMethodDef SwigMethods[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "SWIG_PyStaticMethod_New", SWIG_PyStaticMethod_New, METH_O, NULL},
	 { "delete_SwigPyIterator", _wrap_delete_SwigPyIterator, METH_O, NULL},
	 { "SwigPyIterator_value", _wrap_SwigPyIterator_value, METH_O, NULL},
	 { "SwigPyIterator_incr", _wrap_SwigPyIterator_incr, METH_VARARGS, NULL},
	 { "SwigPyIterator_decr", _wrap_SwigPyIterator_decr, METH_VARARGS, NULL},
	 { "SwigPyIterator_distance", _wrap_SwigPyIterator_distance, METH_VARARGS, NULL},
	 { "SwigPyIterator_equal", _wrap_SwigPyIterator_equal, METH_VARARGS, NULL},
	 { "SwigPyIterator_copy", _wrap_SwigPyIterator_copy, METH_O, NULL},
	 { "SwigPyIterator_next", _wrap_SwigPyIterator_next, METH_O, NULL},
	 { "SwigPyIterator___next__", _wrap_SwigPyIterator___next__, METH_O, NULL},
	 { "SwigPyIterator_previous", _wrap_SwigPyIterator_previous, METH_O, NULL},
	 { "SwigPyIterator_advance", _wrap_SwigPyIterator_advance, METH_VARARGS, NULL},
	 { "SwigPyIterator___eq__", _wrap_SwigPyIterator___eq__, METH_VARARGS, NULL},
	 { "SwigPyIterator___ne__", _wrap_SwigPyIterator___ne__, METH_VARARGS, NULL},
	 { "SwigPyIterator___iadd__", _wrap_SwigPyIterator___iadd__, METH_VARARGS, NULL},
	 { "SwigPyIterator___isub__", _wrap_SwigPyIterator___isub__, METH_VARARGS, NULL},
	 { "SwigPyIterator___add__", _wrap_SwigPyIterator___add__, METH_VARARGS, NULL},
	 { "SwigPyIterator___sub__", _wrap_SwigPyIterator___sub__, METH_VARARGS, NULL},
	 { "SwigPyIterator_swigregister", SwigPyIterator_swigregister, METH_O, NULL},
	 { "vectorInt_iterator", _wrap_vectorInt_iterator, METH_O, NULL},
	 { "vectorInt___nonzero__", _wrap_vectorInt___nonzero__, METH_O, NULL},
	 { "vectorInt___bool__", _wrap_vectorInt___bool__, METH_O, NULL},
	 { "vectorInt___len__", _wrap_vectorInt___len__, METH_O, NULL},
	 { "vectorInt___getslice__", _wrap_vectorInt___getslice__, METH_VARARGS, NULL},
	 { "vectorInt___setslice__", _wrap_vectorInt___setslice__, METH_VARARGS, NULL},
	 { "vectorInt___delslice__", _wrap_vectorInt___delslice__, METH_VARARGS, NULL},
	 { "vectorInt___delitem__", _wrap_vectorInt___delitem__, METH_VARARGS, NULL},
	 { "vectorInt___getitem__", _wrap_vectorInt___getitem__, METH_VARARGS, NULL},
	 { "vectorInt___setitem__", _wrap_vectorInt___setitem__, METH_VARARGS, NULL},
	 { "vectorInt_clear", _wrap_vectorInt_clear, METH_O, NULL},
	 { "new_vectorInt", _wrap_new_vectorInt, METH_VARARGS, NULL},
	 { "delete_vectorInt", _wrap_delete_vectorInt, METH_O, NULL},
	 { "vectorInt_swigregister", vectorInt_swigregister, METH_O, NULL},
	 { "vectorInt_swiginit", vectorInt_swiginit, METH_VARARGS, NULL},
	 { "vectorUnsignedInt_iterator", _wrap_vectorUnsignedInt_iterator, METH_O, NULL},
	 { "vectorUnsignedInt___nonzero__", _wrap_vectorUnsignedInt___nonzero__, METH_O, NULL},
	 { "vectorUnsignedInt___bool__", _wrap_vectorUnsignedInt___bool__, METH_O, NULL},
	 { "vectorUnsignedInt___len__", _wrap_vectorUnsignedInt___len__, METH_O, NULL},
	 { "vectorUnsignedInt___getslice__", _wrap_vectorUnsignedInt___getslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___setslice__", _wrap_vectorUnsignedInt___setslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___delslice__", _wrap_vectorUnsignedInt___delslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___delitem__", _wrap_vectorUnsignedInt___delitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___getitem__", _wrap_vectorUnsignedInt___getitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___setitem__", _wrap_vectorUnsignedInt___setitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt_clear", _wrap_vectorUnsignedInt_clear, METH_O, NULL},
	 { "new_vectorUnsignedInt", _wrap_new_vectorUnsignedInt, METH_VARARGS, NULL},
	 { "delete_vectorUnsignedInt", _wrap_delete_vectorUnsignedInt, METH_O, NULL},
	 { "vectorUnsignedInt_swigregister", vectorUnsignedInt_swigregister, METH_O, NULL},
	 { "vectorUnsignedInt_swiginit", vectorUnsignedInt_swiginit, METH_VARARGS, NULL},
	 { "vectorvInt_iterator", _wrap_vectorvInt_iterator, METH_O, NULL},
	 { "vectorvInt___nonzero__", _wrap_vectorvInt___nonzero__, METH_O, NULL},
	 { "vectorvInt___bool__", _wrap_vectorvInt___bool__, METH_O, NULL},
	 { "vectorvInt___len__", _wrap_vectorvInt___len__, METH_O, NULL},
	 { "vectorvInt___getslice__", _wrap_vectorvInt___getslice__, METH_VARARGS, NULL},
	 { "vectorvInt___setslice__", _wrap_vectorvInt___setslice__, METH_VARARGS, NULL},
	 { "vectorvInt___delslice__", _wrap_vectorvInt___delslice__, METH_VARARGS, NULL},
	 { "vectorvInt___delitem__", _wrap_vectorvInt___delitem__, METH_VARARGS, NULL},
	 { "vectorvInt___getitem__", _wrap_vectorvInt___getitem__, METH_VARARGS, NULL},
	 { "vectorvInt___setitem__", _wrap_vectorvInt___setitem__, METH_VARARGS, NULL},
	 { "vectorvInt_clear", _wrap_vectorvInt_clear, METH_O, NULL},
	 { "new_vectorvInt", _wrap_new_vectorvInt, METH_VARARGS, NULL},
	 { "delete_vectorvInt", _wrap_delete_vectorvInt, METH_O, NULL},
	 { "vectorvInt_swigregister", vectorvInt_swigregister, METH_O, NULL},
	 { "vectorvInt_swiginit", vectorvInt_swiginit, METH_VARARGS, NULL},
	 { "vectorDouble_iterator", _wrap_vectorDouble_iterator, METH_O, NULL},
	 { "vectorDouble___nonzero__", _wrap_vectorDouble___nonzero__, METH_O, NULL},
	 { "vectorDouble___bool__", _wrap_vectorDouble___bool__, METH_O, NULL},
	 { "vectorDouble___len__", _wrap_vectorDouble___len__, METH_O, NULL},
	 { "vectorDouble___getslice__", _wrap_vectorDouble___getslice__, METH_VARARGS, NULL},
	 { "vectorDouble___setslice__", _wrap_vectorDouble___setslice__, METH_VARARGS, NULL},
	 { "vectorDouble___delslice__", _wrap_vectorDouble___delslice__, METH_VARARGS, NULL},
	 { "vectorDouble___delitem__", _wrap_vectorDouble___delitem__, METH_VARARGS, NULL},
	 { "vectorDouble___getitem__", _wrap_vectorDouble___getitem__, METH_VARARGS, NULL},
	 { "vectorDouble___setitem__", _wrap_vectorDouble___setitem__, METH_VARARGS, NULL},
	 { "vectorDouble_clear", _wrap_vectorDouble_clear, METH_O, NULL},
	 { "new_vectorDouble", _wrap_new_vectorDouble, METH_VARARGS, NULL},
	 { "delete_vectorDouble", _wrap_delete_vectorDouble, METH_O, NULL},
	 { "vectorDouble_swigregister", vectorDouble_swigregister, METH_O, NULL},
	 { "vectorDouble_swiginit", vectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorULong_iterator", _wrap_vectorULong_iterator, METH_O, NULL},
	 { "vectorULong___nonzero__", _wrap_vectorULong___nonzero__, METH_O, NULL},
	 { "vectorULong___bool__", _wrap_vectorULong___bool__, METH_O, NULL},
	 { "vectorULong___len__", _wrap_vectorULong___len__, METH_O, NULL},
	 { "vectorULong___getslice__", _wrap_vectorULong___getslice__, METH_VARARGS, NULL},
	 { "vectorULong___setslice__", _wrap_vectorULong___setslice__, METH_VARARGS, NULL},
	 { "vectorULong___delslice__", _wrap_vectorULong___delslice__, METH_VARARGS, NULL},
	 { "vectorULong___delitem__", _wrap_vectorULong___delitem__, METH_VARARGS, NULL},
	 { "vectorULong___getitem__", _wrap_vectorULong___getitem__, METH_VARARGS, NULL},
	 { "vectorULong___setitem__", _wrap_vectorULong___setitem__, METH_VARARGS, NULL},
	 { "vectorULong_clear", _wrap_vectorULong_clear, METH_O, NULL},
	 { "new_vectorULong", _wrap_new_vectorULong, METH_VARARGS, NULL},
	 { "delete_vectorULong", _wrap_delete_vectorULong, METH_O, NULL},
	 { "vectorULong_swigregister", vectorULong_swigregister, METH_O, NULL},
	 { "vectorULong_swiginit", vectorULong_swiginit, METH_VARARGS, NULL},
	 { "vectorString_iterator", _wrap_vectorString_iterator, METH_O, NULL},
	 { "vectorString___nonzero__", _wrap_vectorString___nonzero__, METH_O, NULL},
	 { "vectorString___bool__", _wrap_vectorString___bool__, METH_O, NULL},
	 { "vectorString___len__", _wrap_vectorString___len__, METH_O, NULL},
	 { "vectorString___getslice__", _wrap_vectorString___getslice__, METH_VARARGS, NULL},
	 { "vectorString___setslice__", _wrap_vectorString___setslice__, METH_VARARGS, NULL},
	 { "vectorString___delslice__", _wrap_vectorString___delslice__, METH_VARARGS, NULL},
	 { "vectorString___delitem__", _wrap_vectorString___delitem__, METH_VARARGS, NULL},
	 { "vectorString___getitem__", _wrap_vectorString___getitem__, METH_VARARGS, NULL},
	 { "vectorString___setitem__", _wrap_vectorString___setitem__, METH_VARARGS, NULL},
	 { "vectorString_clear", _wrap_vectorString_clear, METH_O, NULL},
	 { "new_vectorString", _wrap_new_vectorString, METH_VARARGS, NULL},
	 { "delete_vectorString", _wrap_delete_vectorString, METH_O, NULL},
	 { "vectorString_swigregister", vectorString_swigregister, METH_O, NULL},
	 { "vectorString_swiginit", vectorString_swiginit, METH_VARARGS, NULL},
	 { "vectorVector3_iterator", _wrap_vectorVector3_iterator, METH_O, NULL},
	 { "vectorVector3___nonzero__", _wrap_vectorVector3___nonzero__, METH_O, NULL},
	 { "vectorVector3___bool__", _wrap_vectorVector3___bool__, METH_O, NULL},
	 { "vectorVector3___len__", _wrap_vectorVector3___len__, METH_O, NULL},
	 { "vectorVector3___getslice__", _wrap_vectorVector3___getslice__, METH_VARARGS, NULL},
	 { "vectorVector3___setslice__", _wrap_vectorVector3___setslice__, METH_VARARGS, NULL},
	 { "vectorVector3___delslice__", _wrap_vectorVector3___delslice__, METH_VARARGS, NULL},
	 { "vectorVector3___delitem__", _wrap_vectorVector3___delitem__, METH_VARARGS, NULL},
	 { "vectorVector3___getitem__", _wrap_vectorVector3___getitem__, METH_VARARGS, NULL},
	 { "vectorVector3___setitem__", _wrap_vectorVector3___setitem__, METH_VARARGS, NULL},
	 { "vectorVector3_clear", _wrap_vectorVector3_clear, METH_O, NULL},
	 { "new_vectorVector3", _wrap_new_vectorVector3, METH_VARARGS, NULL},
	 { "delete_vectorVector3", _wrap_delete_vectorVector3, METH_O, NULL},
	 { "vectorVector3_swigregister", vectorVector3_swigregister, METH_O, NULL},
	 { "vectorVector3_swiginit", vectorVector3_swiginit, METH_VARARGS, NULL},
	 { "vectorvVector3_iterator", _wrap_vectorvVector3_iterator, METH_O, NULL},
	 { "vectorvVector3___nonzero__", _wrap_vectorvVector3___nonzero__, METH_O, NULL},
	 { "vectorvVector3___bool__", _wrap_vectorvVector3___bool__, METH_O, NULL},
	 { "vectorvVector3___len__", _wrap_vectorvVector3___len__, METH_O, NULL},
	 { "vectorvVector3___getslice__", _wrap_vectorvVector3___getslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___setslice__", _wrap_vectorvVector3___setslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___delslice__", _wrap_vectorvVector3___delslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___delitem__", _wrap_vectorvVector3___delitem__, METH_VARARGS, NULL},
	 { "vectorvVector3___getitem__", _wrap_vectorvVector3___getitem__, METH_VARARGS, NULL},
	 { "vectorvVector3___setitem__", _wrap_vectorvVector3___setitem__, METH_VARARGS, NULL},
	 { "vectorvVector3_clear", _wrap_vectorvVector3_clear, METH_O, NULL},
	 { "new_vectorvVector3", _wrap_new_vectorvVector3, METH_VARARGS, NULL},
	 { "delete_vectorvVector3", _wrap_delete_vectorvVector3, METH_O, NULL},
	 { "vectorvVector3_swigregister", vectorvVector3_swigregister, METH_O, NULL},
	 { "vectorvVector3_swiginit", vectorvVector3_swiginit, METH_VARARGS, NULL},
	 { "vectorOBMol_iterator", _wrap_vectorOBMol_iterator, METH_O, NULL},
	 { "vectorOBMol___nonzero__", _wrap_vectorOBMol___nonzero__, METH_O, NULL},
	 { "vectorOBMol___bool__", _wrap_vectorOBMol___bool__, METH_O, NULL},
	 { "vectorOBMol___len__", _wrap_vectorOBMol___len__, METH_O, NULL},
	 { "vectorOBMol___getslice__", _wrap_vectorOBMol___getslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___setslice__", _wrap_vectorOBMol___setslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___delslice__", _wrap_vectorOBMol___delslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___delitem__", _wrap_vectorOBMol___delitem__, METH_VARARGS, NULL},
	 { "vectorOBMol___getitem__", _wrap_vectorOBMol___getitem__, METH_VARARGS, NULL},
	 { "vectorOBMol___setitem__", _wrap_vectorOBMol___setitem__, METH_VARARGS, NULL},
	 { "vectorOBMol_clear", _wrap_vectorOBMol_clear, METH_O, NULL},
	 { "new_vectorOBMol", _wrap_new_vectorOBMol, METH_VARARGS, NULL},
	 { "delete_vectorOBMol", _wrap_delete_vectorOBMol, METH_O, NULL},
	 { "vectorOBMol_swigregister", vectorOBMol_swigregister, METH_O, NULL},
	 { "vectorOBMol_swiginit", vectorOBMol_swiginit, METH_VARARGS, NULL},
	 { "vectorOBBond_iterator", _wrap_vectorOBBond_iterator, METH_O, NULL},
	 { "vectorOBBond___nonzero__", _wrap_vectorOBBond___nonzero__, METH_O, NULL},
	 { "vectorOBBond___bool__", _wrap_vectorOBBond___bool__, METH_O, NULL},
	 { "vectorOBBond___len__", _wrap_vectorOBBond___len__, METH_O, NULL},
	 { "vectorOBBond___getslice__", _wrap_vectorOBBond___getslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___setslice__", _wrap_vectorOBBond___setslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___delslice__", _wrap_vectorOBBond___delslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___delitem__", _wrap_vectorOBBond___delitem__, METH_VARARGS, NULL},
	 { "vectorOBBond___getitem__", _wrap_vectorOBBond___getitem__, METH_VARARGS, NULL},
	 { "vectorOBBond___setitem__", _wrap_vectorOBBond___setitem__, METH_VARARGS, NULL},
	 { "vectorOBBond_clear", _wrap_vectorOBBond_clear, METH_O, NULL},
	 { "new_vectorOBBond", _wrap_new_vectorOBBond, METH_VARARGS, NULL},
	 { "delete_vectorOBBond", _wrap_delete_vectorOBBond, METH_O, NULL},
	 { "vectorOBBond_swigregister", vectorOBBond_swigregister, METH_O, NULL},
	 { "vectorOBBond_swiginit", vectorOBBond_swiginit, METH_VARARGS, NULL},
	 { "vectorOBResidue_iterator", _wrap_vectorOBResidue_iterator, METH_O, NULL},
	 { "vectorOBResidue___nonzero__", _wrap_vectorOBResidue___nonzero__, METH_O, NULL},
	 { "vectorOBResidue___bool__", _wrap_vectorOBResidue___bool__, METH_O, NULL},
	 { "vectorOBResidue___len__", _wrap_vectorOBResidue___len__, METH_O, NULL},
	 { "vectorOBResidue___getslice__", _wrap_vectorOBResidue___getslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___setslice__", _wrap_vectorOBResidue___setslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___delslice__", _wrap_vectorOBResidue___delslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___delitem__", _wrap_vectorOBResidue___delitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue___getitem__", _wrap_vectorOBResidue___getitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue___setitem__", _wrap_vectorOBResidue___setitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue_clear", _wrap_vectorOBResidue_clear, METH_O, NULL},
	 { "new_vectorOBResidue", _wrap_new_vectorOBResidue, METH_VARARGS, NULL},
	 { "delete_vectorOBResidue", _wrap_delete_vectorOBResidue, METH_O, NULL},
	 { "vectorOBResidue_swigregister", vectorOBResidue_swigregister, METH_O, NULL},
	 { "vectorOBResidue_swiginit", vectorOBResidue_swiginit, METH_VARARGS, NULL},
	 { "vectorOBRing_iterator", _wrap_vectorOBRing_iterator, METH_O, NULL},
	 { "vectorOBRing___nonzero__", _wrap_vectorOBRing___nonzero__, METH_O, NULL},
	 { "vectorOBRing___bool__", _wrap_vectorOBRing___bool__, METH_O, NULL},
	 { "vectorOBRing___len__", _wrap_vectorOBRing___len__, METH_O, NULL},
	 { "vectorOBRing___getslice__", _wrap_vectorOBRing___getslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___setslice__", _wrap_vectorOBRing___setslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___delslice__", _wrap_vectorOBRing___delslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___delitem__", _wrap_vectorOBRing___delitem__, METH_VARARGS, NULL},
	 { "vectorOBRing___getitem__", _wrap_vectorOBRing___getitem__, METH_VARARGS, NULL},
	 { "vectorOBRing___setitem__", _wrap_vectorOBRing___setitem__, METH_VARARGS, NULL},
	 { "vectorOBRing_clear", _wrap_vectorOBRing_clear, METH_O, NULL},
	 { "new_vectorOBRing", _wrap_new_vectorOBRing, METH_VARARGS, NULL},
	 { "delete_vectorOBRing", _wrap_delete_vectorOBRing, METH_O, NULL},
	 { "vectorOBRing_swigregister", vectorOBRing_swigregister, METH_O, NULL},
	 { "vectorOBRing_swiginit", vectorOBRing_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBRing_iterator", _wrap_vectorpOBRing_iterator, METH_O, NULL},
	 { "vectorpOBRing___nonzero__", _wrap_vectorpOBRing___nonzero__, METH_O, NULL},
	 { "vectorpOBRing___bool__", _wrap_vectorpOBRing___bool__, METH_O, NULL},
	 { "vectorpOBRing___len__", _wrap_vectorpOBRing___len__, METH_O, NULL},
	 { "vectorpOBRing___getslice__", _wrap_vectorpOBRing___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___setslice__", _wrap_vectorpOBRing___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___delslice__", _wrap_vectorpOBRing___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___delitem__", _wrap_vectorpOBRing___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing___getitem__", _wrap_vectorpOBRing___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing___setitem__", _wrap_vectorpOBRing___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing_clear", _wrap_vectorpOBRing_clear, METH_O, NULL},
	 { "new_vectorpOBRing", _wrap_new_vectorpOBRing, METH_VARARGS, NULL},
	 { "delete_vectorpOBRing", _wrap_delete_vectorpOBRing, METH_O, NULL},
	 { "vectorpOBRing_swigregister", vectorpOBRing_swigregister, METH_O, NULL},
	 { "vectorpOBRing_swiginit", vectorpOBRing_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBGenericData_iterator", _wrap_vectorpOBGenericData_iterator, METH_O, NULL},
	 { "vectorpOBGenericData___nonzero__", _wrap_vectorpOBGenericData___nonzero__, METH_O, NULL},
	 { "vectorpOBGenericData___bool__", _wrap_vectorpOBGenericData___bool__, METH_O, NULL},
	 { "vectorpOBGenericData___len__", _wrap_vectorpOBGenericData___len__, METH_O, NULL},
	 { "vectorpOBGenericData___getslice__", _wrap_vectorpOBGenericData___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___setslice__", _wrap_vectorpOBGenericData___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___delslice__", _wrap_vectorpOBGenericData___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___delitem__", _wrap_vectorpOBGenericData___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___getitem__", _wrap_vectorpOBGenericData___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___setitem__", _wrap_vectorpOBGenericData___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData_clear", _wrap_vectorpOBGenericData_clear, METH_O, NULL},
	 { "new_vectorpOBGenericData", _wrap_new_vectorpOBGenericData, METH_VARARGS, NULL},
	 { "delete_vectorpOBGenericData", _wrap_delete_vectorpOBGenericData, METH_O, NULL},
	 { "vectorpOBGenericData_swigregister", vectorpOBGenericData_swigregister, METH_O, NULL},
	 { "vectorpOBGenericData_swiginit", vectorpOBGenericData_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord_iterator", _wrap_vectorpOBInternalCoord_iterator, METH_O, NULL},
	 { "vectorpOBInternalCoord___nonzero__", _wrap_vectorpOBInternalCoord___nonzero__, METH_O, NULL},
	 { "vectorpOBInternalCoord___bool__", _wrap_vectorpOBInternalCoord___bool__, METH_O, NULL},
	 { "vectorpOBInternalCoord___len__", _wrap_vectorpOBInternalCoord___len__, METH_O, NULL},
	 { "vectorpOBInternalCoord___getslice__", _wrap_vectorpOBInternalCoord___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___setslice__", _wrap_vectorpOBInternalCoord___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___delslice__", _wrap_vectorpOBInternalCoord___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___delitem__", _wrap_vectorpOBInternalCoord___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___getitem__", _wrap_vectorpOBInternalCoord___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___setitem__", _wrap_vectorpOBInternalCoord___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord_clear", _wrap_vectorpOBInternalCoord_clear, METH_O, NULL},
	 { "new_vectorpOBInternalCoord", _wrap_new_vectorpOBInternalCoord, METH_VARARGS, NULL},
	 { "delete_vectorpOBInternalCoord", _wrap_delete_vectorpOBInternalCoord, METH_O, NULL},
	 { "vectorpOBInternalCoord_swigregister", vectorpOBInternalCoord_swigregister, METH_O, NULL},
	 { "vectorpOBInternalCoord_swiginit", vectorpOBInternalCoord_swiginit, METH_VARARGS, NULL},
	 { "new_pairUIntUInt", _wrap_new_pairUIntUInt, METH_VARARGS, NULL},
	 { "pairUIntUInt_first_set", _wrap_pairUIntUInt_first_set, METH_VARARGS, NULL},
	 { "pairUIntUInt_first_get", _wrap_pairUIntUInt_first_get, METH_O, NULL},
	 { "pairUIntUInt_second_set", _wrap_pairUIntUInt_second_set, METH_VARARGS, NULL},
	 { "pairUIntUInt_second_get", _wrap_pairUIntUInt_second_get, METH_O, NULL},
	 { "delete_pairUIntUInt", _wrap_delete_pairUIntUInt, METH_O, NULL},
	 { "pairUIntUInt_swigregister", pairUIntUInt_swigregister, METH_O, NULL},
	 { "pairUIntUInt_swiginit", pairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "vpairUIntUInt_iterator", _wrap_vpairUIntUInt_iterator, METH_O, NULL},
	 { "vpairUIntUInt___nonzero__", _wrap_vpairUIntUInt___nonzero__, METH_O, NULL},
	 { "vpairUIntUInt___bool__", _wrap_vpairUIntUInt___bool__, METH_O, NULL},
	 { "vpairUIntUInt___len__", _wrap_vpairUIntUInt___len__, METH_O, NULL},
	 { "vpairUIntUInt___getslice__", _wrap_vpairUIntUInt___getslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___setslice__", _wrap_vpairUIntUInt___setslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___delslice__", _wrap_vpairUIntUInt___delslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___delitem__", _wrap_vpairUIntUInt___delitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___getitem__", _wrap_vpairUIntUInt___getitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___setitem__", _wrap_vpairUIntUInt___setitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt_clear", _wrap_vpairUIntUInt_clear, METH_O, NULL},
	 { "new_vpairUIntUInt", _wrap_new_vpairUIntUInt, METH_VARARGS, NULL},
	 { "delete_vpairUIntUInt", _wrap_delete_vpairUIntUInt, METH_O, NULL},
	 { "vpairUIntUInt_swigregister", vpairUIntUInt_swigregister, METH_O, NULL},
	 { "vpairUIntUInt_swiginit", vpairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_iterator", _wrap_vvpairUIntUInt_iterator, METH_O, NULL},
	 { "vvpairUIntUInt___nonzero__", _wrap_vvpairUIntUInt___nonzero__, METH_O, NULL},
	 { "vvpairUIntUInt___bool__", _wrap_vvpairUIntUInt___bool__, METH_O, NULL},
	 { "vvpairUIntUInt___len__", _wrap_vvpairUIntUInt___len__, METH_O, NULL},
	 { "vvpairUIntUInt___getslice__", _wrap_vvpairUIntUInt___getslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___setslice__", _wrap_vvpairUIntUInt___setslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___delslice__", _wrap_vvpairUIntUInt___delslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___delitem__", _wrap_vvpairUIntUInt___delitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___getitem__", _wrap_vvpairUIntUInt___getitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___setitem__", _wrap_vvpairUIntUInt___setitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_pop", _wrap_vvpairUIntUInt_pop, METH_O, NULL},
	 { "vvpairUIntUInt_append", _wrap_vvpairUIntUInt_append, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_empty", _wrap_vvpairUIntUInt_empty, METH_O, NULL},
	 { "vvpairUIntUInt_size", _wrap_vvpairUIntUInt_size, METH_O, NULL},
	 { "vvpairUIntUInt_swap", _wrap_vvpairUIntUInt_swap, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_begin", _wrap_vvpairUIntUInt_begin, METH_O, NULL},
	 { "vvpairUIntUInt_end", _wrap_vvpairUIntUInt_end, METH_O, NULL},
	 { "vvpairUIntUInt_rbegin", _wrap_vvpairUIntUInt_rbegin, METH_O, NULL},
	 { "vvpairUIntUInt_rend", _wrap_vvpairUIntUInt_rend, METH_O, NULL},
	 { "vvpairUIntUInt_clear", _wrap_vvpairUIntUInt_clear, METH_O, NULL},
	 { "vvpairUIntUInt_get_allocator", _wrap_vvpairUIntUInt_get_allocator, METH_O, NULL},
	 { "vvpairUIntUInt_pop_back", _wrap_vvpairUIntUInt_pop_back, METH_O, NULL},
	 { "vvpairUIntUInt_erase", _wrap_vvpairUIntUInt_erase, METH_VARARGS, NULL},
	 { "new_vvpairUIntUInt", _wrap_new_vvpairUIntUInt, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_push_back", _wrap_vvpairUIntUInt_push_back, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_front", _wrap_vvpairUIntUInt_front, METH_O, NULL},
	 { "vvpairUIntUInt_back", _wrap_vvpairUIntUInt_back, METH_O, NULL},
	 { "vvpairUIntUInt_assign", _wrap_vvpairUIntUInt_assign, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_resize", _wrap_vvpairUIntUInt_resize, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_insert", _wrap_vvpairUIntUInt_insert, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_reserve", _wrap_vvpairUIntUInt_reserve, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_capacity", _wrap_vvpairUIntUInt_capacity, METH_O, NULL},
	 { "delete_vvpairUIntUInt", _wrap_delete_vvpairUIntUInt, METH_O, NULL},
	 { "vvpairUIntUInt_swigregister", vvpairUIntUInt_swigregister, METH_O, NULL},
	 { "vvpairUIntUInt_swiginit", vvpairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "toAliasData", _wrap_toAliasData, METH_O, NULL},
	 { "toAngleData", _wrap_toAngleData, METH_O, NULL},
	 { "toCommentData", _wrap_toCommentData, METH_O, NULL},
	 { "toConformerData", _wrap_toConformerData, METH_O, NULL},
	 { "toExternalBondData", _wrap_toExternalBondData, METH_O, NULL},
	 { "toGridData", _wrap_toGridData, METH_O, NULL},
	 { "toMatrixData", _wrap_toMatrixData, METH_O, NULL},
	 { "toNasaThermoData", _wrap_toNasaThermoData, METH_O, NULL},
	 { "toPairData", _wrap_toPairData, METH_O, NULL},
	 { "toPairInteger", _wrap_toPairInteger, METH_O, NULL},
	 { "toPairFloatingPoint", _wrap_toPairFloatingPoint, METH_O, NULL},
	 { "toRateData", _wrap_toRateData, METH_O, NULL},
	 { "toRotamerList", _wrap_toRotamerList, METH_O, NULL},
	 { "toRotationData", _wrap_toRotationData, METH_O, NULL},
	 { "toSerialNums", _wrap_toSerialNums, METH_O, NULL},
	 { "toSetData", _wrap_toSetData, METH_O, NULL},
	 { "toSymmetryData", _wrap_toSymmetryData, METH_O, NULL},
	 { "toTorsionData", _wrap_toTorsionData, METH_O, NULL},
	 { "toUnitCell", _wrap_toUnitCell, METH_O, NULL},
	 { "toVectorData", _wrap_toVectorData, METH_O, NULL},
	 { "toVibrationData", _wrap_toVibrationData, METH_O, NULL},
	 { "toVirtualBond", _wrap_toVirtualBond, METH_O, NULL},
	 { "toStereoBase", _wrap_toStereoBase, METH_O, NULL},
	 { "toTetrahedralStereo", _wrap_toTetrahedralStereo, METH_O, NULL},
	 { "toCisTransStereo", _wrap_toCisTransStereo, METH_O, NULL},
	 { "toSquarePlanarStereo", _wrap_toSquarePlanarStereo, METH_O, NULL},
	 { "new_OBGlobalDataBase", _wrap_new_OBGlobalDataBase, METH_NOARGS, "Constructor"},
	 { "delete_OBGlobalDataBase", _wrap_delete_OBGlobalDataBase, METH_O, "Destructor"},
	 { "OBGlobalDataBase_Init", _wrap_OBGlobalDataBase_Init, METH_O, "Read in the data file, falling back as needed"},
	 { "OBGlobalDataBase_GetSize", _wrap_OBGlobalDataBase_GetSize, METH_O, "\n"
		":rtype: int\n"
		":return: the size of the database (for error checking)\n"
		""},
	 { "OBGlobalDataBase_SetReadDirectory", _wrap_OBGlobalDataBase_SetReadDirectory, METH_VARARGS, "Set the directory before calling Init()"},
	 { "OBGlobalDataBase_SetEnvironmentVariable", _wrap_OBGlobalDataBase_SetEnvironmentVariable, METH_VARARGS, "Set the environment variable to use before calling Init()"},
	 { "OBGlobalDataBase_ParseLine", _wrap_OBGlobalDataBase_ParseLine, METH_VARARGS, "Specified by particular table classes (parses an individual data line)"},
	 { "OBGlobalDataBase_swigregister", OBGlobalDataBase_swigregister, METH_O, NULL},
	 { "OBGlobalDataBase_swiginit", OBGlobalDataBase_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomHOF", _wrap_new_OBAtomHOF, METH_VARARGS, "\n"
		"Initialize Heat of Formation for atom\n"
		"\n"
		"    :type element: string\n"
		"    :param element: The element string\n"
		"    :type charge: int\n"
		"    :param charge:  The formal charge of the particle (if an ion)\n"
		"    :type method: string\n"
		"    :param method:  The method used for determining the value\n"
		"    :type desc: string\n"
		"    :param desc:    Description of the value\n"
		"    :type T: float\n"
		"    :param T:       Temperature\n"
		"    :type value: float\n"
		"    :param value:   The value of the property (energy)\n"
		"    :type multiplicity: int\n"
		"    :param multiplicity: The multiplicity of the atomic system\n"
		"    :type unit: string\n"
		"    :param unit:    The (energy) unit\n"
		""},
	 { "delete_OBAtomHOF", _wrap_delete_OBAtomHOF, METH_O, "Destructor"},
	 { "OBAtomHOF_Element", _wrap_OBAtomHOF_Element, METH_O, "Return the chemical element"},
	 { "OBAtomHOF_Charge", _wrap_OBAtomHOF_Charge, METH_O, "Return the formal charge"},
	 { "OBAtomHOF_Method", _wrap_OBAtomHOF_Method, METH_O, "Return the method used for the measurement/calculation"},
	 { "OBAtomHOF_Desc", _wrap_OBAtomHOF_Desc, METH_O, "Return specification of the measurement/calculation type"},
	 { "OBAtomHOF_T", _wrap_OBAtomHOF_T, METH_O, "Return the temperature"},
	 { "OBAtomHOF_Value", _wrap_OBAtomHOF_Value, METH_O, "Return the (energy) value"},
	 { "OBAtomHOF_Multiplicity", _wrap_OBAtomHOF_Multiplicity, METH_O, "Return the multiplicity"},
	 { "OBAtomHOF_Unit", _wrap_OBAtomHOF_Unit, METH_O, "Return the (energy) unit"},
	 { "OBAtomHOF_swigregister", OBAtomHOF_swigregister, METH_O, NULL},
	 { "OBAtomHOF_swiginit", OBAtomHOF_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomicHeatOfFormationTable", _wrap_new_OBAtomicHeatOfFormationTable, METH_NOARGS, "Constructor"},
	 { "delete_OBAtomicHeatOfFormationTable", _wrap_delete_OBAtomicHeatOfFormationTable, METH_O, "Destructor"},
	 { "OBAtomicHeatOfFormationTable_GetHeatOfFormation", _wrap_OBAtomicHeatOfFormationTable_GetHeatOfFormation, METH_VARARGS, "\n"
		"Extract heat of formation and entropy for an atom\n"
		"      :type elem: string\n"
		"      :param elem:         The chemical element we're looking for\n"
		"      :type charge: int\n"
		"      :param charge:       At this formal charge\n"
		"      :type method: string\n"
		"      :param method:       The method used for computing/measuring\n"
		"      :type T: float\n"
		"      :param T:            The temperature\n"
		"      :type dhof0: float\n"
		"      :param dhof0:        The output energy at 0K\n"
		"      :param dhof1:        The output energy at T\n"
		"      :type S0T: float\n"
		"      :param S0T:          The entropy at T (it is 0 at 0K)\n"
		"      :rtype: int\n"
		"      :return: 1 if the contribution to the Heat of Formation for this atom\n"
		"          is known at temperature T. If 1 the values\n"
		"          including all corrections are returned in the dhof variable.\n"
		""},
	 { "OBAtomicHeatOfFormationTable_swigregister", OBAtomicHeatOfFormationTable_swigregister, METH_O, NULL},
	 { "OBAtomicHeatOfFormationTable_swiginit", OBAtomicHeatOfFormationTable_swiginit, METH_VARARGS, NULL},
	 { "new_OBTypeTable", _wrap_new_OBTypeTable, METH_NOARGS, NULL},
	 { "delete_OBTypeTable", _wrap_delete_OBTypeTable, METH_O, NULL},
	 { "OBTypeTable_SetFromType", _wrap_OBTypeTable_SetFromType, METH_VARARGS, "Set the initial atom type to be translated"},
	 { "OBTypeTable_SetToType", _wrap_OBTypeTable_SetToType, METH_VARARGS, "Set the destination atom type for translation"},
	 { "OBTypeTable_Translate", _wrap_OBTypeTable_Translate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Translate atom types\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Translate atom types :rtype: boolean\n"
		":return: whether the translation was successful\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Translate atom types :rtype: string\n"
		":return: the translated atom type, or an empty string if not possible\n"
		""},
	 { "OBTypeTable_GetFromType", _wrap_OBTypeTable_GetFromType, METH_O, "\n"
		":rtype: string\n"
		":return: the initial atom type to be translated\n"
		""},
	 { "OBTypeTable_GetToType", _wrap_OBTypeTable_GetToType, METH_O, "\n"
		":rtype: string\n"
		":return: the destination atom type for translation\n"
		""},
	 { "OBTypeTable_swigregister", OBTypeTable_swigregister, METH_O, NULL},
	 { "OBTypeTable_swiginit", OBTypeTable_swiginit, METH_VARARGS, NULL},
	 { "new_OBResidueData", _wrap_new_OBResidueData, METH_NOARGS, NULL},
	 { "OBResidueData_SetResName", _wrap_OBResidueData_SetResName, METH_VARARGS, "\n"
		" Sets the table to access the residue information for a specified  residue name :rtype: boolean\n"
		":return: whether this residue name is in the table\n"
		""},
	 { "OBResidueData_LookupBO", _wrap_OBResidueData_LookupBO, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: int\n"
		":return: the bond order for the bond specified in the current residue Deprecated: Easier to use the two-argument form\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: the bond order for the bond specified between the two specified atom labels\n"
		""},
	 { "OBResidueData_LookupType", _wrap_OBResidueData_LookupType, METH_VARARGS, "\n"
		" Look up the atom type and hybridization for the atom label specified in the first argument for the current residue :rtype: boolean\n"
		":return: whether the atom label specified is found in the current residue\n"
		""},
	 { "OBResidueData_AssignBonds", _wrap_OBResidueData_AssignBonds, METH_VARARGS, "Assign bond orders, atom types and residues for the supplied OBMol based on the residue information assigned to atoms"},
	 { "delete_OBResidueData", _wrap_delete_OBResidueData, METH_O, NULL},
	 { "OBResidueData_swigregister", OBResidueData_swigregister, METH_O, NULL},
	 { "OBResidueData_swiginit", OBResidueData_swiginit, METH_VARARGS, NULL},
	 { "OBStopwatch_Start", _wrap_OBStopwatch_Start, METH_O, "Mark the start of \"stopwatch\" timing"},
	 { "OBStopwatch_Lap", _wrap_OBStopwatch_Lap, METH_O, "\n"
		":rtype: float\n"
		":return: The time since calling OBStopwatch::Start() in seconds.\n"
		""},
	 { "OBStopwatch_Elapsed", _wrap_OBStopwatch_Elapsed, METH_O, "\n"
		":rtype: float\n"
		":return: The time since calling OBStopwatch::Start() in seconds.\n"
		""},
	 { "new_OBStopwatch", _wrap_new_OBStopwatch, METH_NOARGS, NULL},
	 { "delete_OBStopwatch", _wrap_delete_OBStopwatch, METH_O, NULL},
	 { "OBStopwatch_swigregister", OBStopwatch_swigregister, METH_O, NULL},
	 { "OBStopwatch_swiginit", OBStopwatch_swiginit, METH_VARARGS, NULL},
	 { "new_OBSqrtTbl", _wrap_new_OBSqrtTbl, METH_VARARGS, "\n"
		" Create a square root table to handle up to the square root of ``max`` (e.g., if you want the square root of 144, supply 12 for max) :type max: float\n"
		":param max: The maximum square root stored in the lookup table :type incr: float\n"
		":param incr: The floating point resolution of the lookup table\n"
		""},
	 { "delete_OBSqrtTbl", _wrap_delete_OBSqrtTbl, METH_O, NULL},
	 { "OBSqrtTbl_Sqrt", _wrap_OBSqrtTbl_Sqrt, METH_VARARGS, "\n"
		" Fast square root calculation using a lookup table :rtype: float\n"
		":return: Square root of ``d2``\n"
		""},
	 { "OBSqrtTbl_Init", _wrap_OBSqrtTbl_Init, METH_VARARGS, "\n"
		" Initialize the square root lookup table :type max: float\n"
		":param max: The maximum square root stored in the lookup table (e.g., if you want the square root of 144, supply 12 for max) :type incr: float\n"
		":param incr: The floating point resolution of the lookup table\n"
		""},
	 { "OBSqrtTbl_swigregister", OBSqrtTbl_swigregister, METH_O, NULL},
	 { "OBSqrtTbl_swiginit", OBSqrtTbl_swiginit, METH_VARARGS, NULL},
	 { "rotate_coords", _wrap_rotate_coords, METH_VARARGS, NULL},
	 { "calc_rms", _wrap_calc_rms, METH_VARARGS, NULL},
	 { "new_vector3", _wrap_new_vector3, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy Constructor\n"
		""},
	 { "delete_vector3", _wrap_delete_vector3, METH_O, "Destructor"},
	 { "vector3_begin", _wrap_vector3_begin, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" :rtype: float\n"
		" :return: iterator to beginning\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"return const_iterator to beginning\n"
		""},
	 { "vector3_end", _wrap_vector3_end, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" :rtype: float\n"
		" :return: iterator to end\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"return const_iterator to end\n"
		""},
	 { "vector3_Set", _wrap_vector3_Set, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set x,y and z-component of a vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set x,y and z-component of a vector from c[0]..c[2]\n"
		""},
	 { "vector3_SetX", _wrap_vector3_SetX, METH_VARARGS, "Access function to set the x-coordinate of the vector"},
	 { "vector3_SetY", _wrap_vector3_SetY, METH_VARARGS, "Access function to set the y-coordinate of the vector"},
	 { "vector3_SetZ", _wrap_vector3_SetZ, METH_VARARGS, "Access function to set the z-coordinate of the vector"},
	 { "vector3_GetX", _wrap_vector3_GetX, METH_O, "Access function to get the x-coordinate of the vector"},
	 { "vector3_GetY", _wrap_vector3_GetY, METH_O, "Access function to get the y-coordinate of the vector"},
	 { "vector3_GetZ", _wrap_vector3_GetZ, METH_O, "Access function to get the z-coordinate of the vector"},
	 { "vector3_Get", _wrap_vector3_Get, METH_VARARGS, "Set c[0]..c[2] to the components of the vector Warning: No error checking is performed"},
	 { "vector3_AsArray", _wrap_vector3_AsArray, METH_O, "\n"
		":rtype: float\n"
		":return: the vector as a const double *\n"
		""},
	 { "vector3___iadd__", _wrap_vector3___iadd__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Vector addition (add ``v`` to *this) :rtype: :py:class:`vector3`\n"
		":return: *this + v\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Scalar addition (add ``f`` to *this) :rtype: :py:class:`vector3`\n"
		":return: *this + f\n"
		""},
	 { "vector3___isub__", _wrap_vector3___isub__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Vector subtraction (subtract ``v`` from *this) :rtype: :py:class:`vector3`\n"
		":return: *this - v\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Scalar subtraction (subtract ``f`` from *this) :rtype: :py:class:`vector3`\n"
		":return: *this - f\n"
		""},
	 { "vector3___itruediv__", _wrap_vector3___itruediv__, METH_VARARGS, "\n"
		" Scalar division (divide *this by ``c)`` :rtype: :py:class:`vector3`\n"
		":return: *this divided by c\n"
		""},
	 { "vector3___imul__", _wrap_vector3___imul__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Scalar multiplication (multiply *this by ``c)`` :rtype: :py:class:`vector3`\n"
		":return: *this * c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Multiplication of matrix and vector :rtype: :py:class:`vector3`\n"
		":return: the result (i.e., the updated vector) TODO: Currently unimplemented\n"
		""},
	 { "vector3_randomUnitVector", _wrap_vector3_randomUnitVector, METH_O, "Create a random unit vector"},
	 { "vector3_normalize", _wrap_vector3_normalize, METH_O, "\n"
		" Scales a vector to give it length one. :rtype: :py:class:`vector3`\n"
		":return: the result (i.e., the normalized vector)\n"
		""},
	 { "vector3_CanBeNormalized", _wrap_vector3_CanBeNormalized, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether a vector can be normalized\n"
		""},
	 { "vector3_length_2", _wrap_vector3_length_2, METH_O, "\n"
		":rtype: float\n"
		":return: The length of the vector squared\n"
		""},
	 { "vector3_length", _wrap_vector3_length, METH_O, "\n"
		":rtype: float\n"
		":return: The vector length\n"
		""},
	 { "vector3_x", _wrap_vector3_x, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the x-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the x-coordinate of the vector\n"
		""},
	 { "vector3_y", _wrap_vector3_y, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the y-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the y-coordinate of the vector\n"
		""},
	 { "vector3_z", _wrap_vector3_z, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the z-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the z-coordinate of the vector\n"
		""},
	 { "vector3___eq__", _wrap_vector3___eq__, METH_VARARGS, "\n"
		" Comparison Methods Equivalence of vectors Deprecated: This method uses unreliable floating point == comparisons    Use vector3::IsApprox() instead. :rtype: boolean\n"
		":return: true if every component is equal\n"
		""},
	 { "vector3___ne__", _wrap_vector3___ne__, METH_VARARGS, "\n"
		" Deprecated: This method uses unreliable floating point == comparisons    Use vector3::IsApprox() instead. :rtype: boolean\n"
		":return: true if at least one component of the two vectors is !=\n"
		""},
	 { "vector3_IsApprox", _wrap_vector3_IsApprox, METH_VARARGS, "\n"
		" Safe comparison for floating-point vector3 :rtype: boolean\n"
		":return: true if the vector *this is approximately equal to the vector         ``other``, to the precision ``precision``. More specifically,         this method works exactly like the OpenBabel::IsApprox()         function, replacing the absolute value for doubles by the norm         for vectors. :type other: :py:class:`vector3`\n"
		":param other: The vector for comparison :type precision: float\n"
		":param precision: This parameter plays the same role as in        OpenBabel::IsApprox().\n"
		""},
	 { "vector3_distSq", _wrap_vector3_distSq, METH_VARARGS, "\n"
		" }@ :rtype: float\n"
		":return: square of the distance between *this and vv equivalent to length_2(*this-vv)\n"
		""},
	 { "vector3_createOrthoVector", _wrap_vector3_createOrthoVector, METH_VARARGS, "\n"
		" Creates a vector of length one, orthogonal to *this. :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "vector3_swigregister", vector3_swigregister, METH_O, NULL},
	 { "vector3_swiginit", vector3_swiginit, METH_VARARGS, NULL},
	 { "__add__", _wrap___add__, METH_VARARGS, "Vector addition"},
	 { "__truediv__", _wrap___truediv__, METH_VARARGS, "Division by a scalar"},
	 { "dot", _wrap_dot, METH_VARARGS, "Dot product of two vectors"},
	 { "cross", _wrap_cross, METH_VARARGS, "Cross product of two vectors"},
	 { "vectorAngle", _wrap_vectorAngle, METH_VARARGS, "Calculate the angle between vectors (in degrees)"},
	 { "CalcTorsionAngle", _wrap_CalcTorsionAngle, METH_VARARGS, "Calculate the torsion angle between vectors (in degrees)"},
	 { "Point2PlaneSigned", _wrap_Point2PlaneSigned, METH_VARARGS, "Calculate the signed distance of point a to the plane determined by b,c,d"},
	 { "Point2Plane", _wrap_Point2Plane, METH_VARARGS, "Calculate the distance of point a to the plane determined by b,c,d"},
	 { "Point2PlaneAngle", _wrap_Point2PlaneAngle, METH_VARARGS, "Calculate the angle between point a and the plane determined by b,c,d"},
	 { "Point2Line", _wrap_Point2Line, METH_VARARGS, "Calculate the distance of a point a to a line determined by b and c"},
	 { "new_matrix3x3", _wrap_new_matrix3x3, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructs the zero-matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructs s times the unit matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Constructs a matrix from row vectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Constructs a matrix from a 3x3-array of doubles The first index represents the row, the second index the column\n"
		""},
	 { "delete_matrix3x3", _wrap_delete_matrix3x3, METH_O, "Destructor"},
	 { "matrix3x3_GetArray", _wrap_matrix3x3_GetArray, METH_VARARGS, "\n"
		"Access function Writes the matrix into the 1-dimensional array m, row by\n"
		"       row. The array must be able to hold 9 doubles, otherwise your\n"
		"       program will segfault.\n"
		""},
	 { "matrix3x3___call__", _wrap_matrix3x3___call__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: a constant reference to an element of the matrix.\n"
		"             row and column must be between 0 and 2. No check is done.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: a non-constant reference to an element of the matrix.\n"
		"             row and column must be between 0 and 2. No check is done.\n"
		""},
	 { "matrix3x3_inverse", _wrap_matrix3x3_inverse, METH_O, "Calculates the inverse of a matrix."},
	 { "matrix3x3_transpose", _wrap_matrix3x3_transpose, METH_O, "Calculates the transpose of a matrix."},
	 { "matrix3x3_determinant", _wrap_matrix3x3_determinant, METH_O, "\n"
		":rtype: float\n"
		":return: The determinant of the matrix\n"
		""},
	 { "matrix3x3_isSymmetric", _wrap_matrix3x3_isSymmetric, METH_O, "Checks if a matrix is symmetric"},
	 { "matrix3x3_isOrthogonal", _wrap_matrix3x3_isOrthogonal, METH_O, "\n"
		"Checks if a matrix is orthogonal This method checks if a matrix is orthogonal, i.e.\n"
		"       if all column vectors are normalized and\n"
		"       are mutually orthogonal. A matrix is orthogonal if,\n"
		"       and only if the transformation it describes is orthonormal.\n"
		"       An orthonormal transformation is a\n"
		"       transformation that preserves length and angle.\n"
		"\n"
		"       The check is performed using the method isUnitMatrix() to\n"
		"       check if\n"
		"\n"
		"       .. code-block:: c++\n"
		"\n"
		"   this * transpose()\n"
		"       is a unit matrix. The criterion is therefore numerically quite\n"
		"       tight.\n"
		""},
	 { "matrix3x3_isDiagonal", _wrap_matrix3x3_isDiagonal, METH_O, "\n"
		":rtype: boolean\n"
		":return: if a matrix is diagonal\n"
		""},
	 { "matrix3x3_isUnitMatrix", _wrap_matrix3x3_isUnitMatrix, METH_O, "\n"
		":rtype: boolean\n"
		":return: if a matrix is the unit matrix\n"
		""},
	 { "matrix3x3_Get", _wrap_matrix3x3_Get, METH_VARARGS, "\n"
		" Access function Warning: row or column are not in the range 0..2, zero is returned\n"
		"Deprecated: use the constant operator() instead\n"
		""},
	 { "matrix3x3_Set", _wrap_matrix3x3_Set, METH_VARARGS, "\n"
		" Access function Warning: if row or column are not in the range 0..2, nothing will happen\n"
		"Deprecated: use the non-constant operator() instead\n"
		""},
	 { "matrix3x3_SetColumn", _wrap_matrix3x3_SetColumn, METH_VARARGS, "\n"
		"Access function Warning: If column is not in the range 0..2, the vector\n"
		"       remains unchanged and an exception is thrown.\n"
		""},
	 { "matrix3x3_SetRow", _wrap_matrix3x3_SetRow, METH_VARARGS, "\n"
		"Access function Warning: If column is not in the range 0..2, the vector\n"
		"       remains unchanged and an exception is thrown.\n"
		""},
	 { "matrix3x3_GetColumn", _wrap_matrix3x3_GetColumn, METH_VARARGS, "\n"
		"Access function Warning: If col is not in the range 0..2, an exception is\n"
		"       thrown.\n"
		""},
	 { "matrix3x3_GetRow", _wrap_matrix3x3_GetRow, METH_VARARGS, "\n"
		"Access function Warning: If row is not in the range 0..2, an exception is\n"
		"       thrown.\n"
		""},
	 { "matrix3x3___imul__", _wrap_matrix3x3___imul__, METH_VARARGS, "Multiplies all entries of the matrix by a scalar c"},
	 { "matrix3x3___itruediv__", _wrap_matrix3x3___itruediv__, METH_VARARGS, "Divides all entries of the matrix by a scalar c"},
	 { "matrix3x3_SetupRotMat", _wrap_matrix3x3_SetupRotMat, METH_VARARGS, "Calculate a rotation matrix for rotation about the x, y, and z axes by the angles specified (in degrees)"},
	 { "matrix3x3_PlaneReflection", _wrap_matrix3x3_PlaneReflection, METH_VARARGS, "Calculates a matrix that represents reflection on a plane"},
	 { "matrix3x3_RotAboutAxisByAngle", _wrap_matrix3x3_RotAboutAxisByAngle, METH_VARARGS, "Calculates a rotation matrix, rotating around the specified axis by the specified angle (in degrees)"},
	 { "matrix3x3_FillOrth", _wrap_matrix3x3_FillOrth, METH_VARARGS, "Calculate an orthogonalisation matrix for a unit cell specified by the parameters alpha, beta, gamma, a, b, c where alpha, beta, and gamma are the cell angles (in degrees) and a, b, and c are the cell vector lengths Used by OBUnitCell"},
	 { "matrix3x3_findEigenvectorsIfSymmetric", _wrap_matrix3x3_findEigenvectorsIfSymmetric, METH_VARARGS, "Find the eigenvalues and -vectors of a symmetric matrix"},
	 { "__mul__", _wrap___mul__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Multiplication with a scalar\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Multiplication with a scalar\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Multiplication of matrix and vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Matrix-matrix multiplication\n"
		""},
	 { "matrix3x3_jacobi", _wrap_matrix3x3_jacobi, METH_VARARGS, "Eigenvalue calculation"},
	 { "matrix3x3_swigregister", matrix3x3_swigregister, METH_O, NULL},
	 { "matrix3x3_swiginit", matrix3x3_swiginit, METH_VARARGS, NULL},
	 { "new_transform3d", _wrap_new_transform3d, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructs a matrix from row vectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructs a matrix from a 3x3-array of doubles The first index represents the row, the second index the column\n"
		""},
	 { "transform3d___mul__", _wrap_transform3d___mul__, METH_VARARGS, NULL},
	 { "transform3d_DescribeAsString", _wrap_transform3d_DescribeAsString, METH_O, NULL},
	 { "transform3d_DescribeAsValues", _wrap_transform3d_DescribeAsValues, METH_O, NULL},
	 { "transform3d_Normalize", _wrap_transform3d_Normalize, METH_O, NULL},
	 { "delete_transform3d", _wrap_delete_transform3d, METH_O, NULL},
	 { "transform3d_swigregister", transform3d_swigregister, METH_O, NULL},
	 { "transform3d_swiginit", transform3d_swiginit, METH_VARARGS, NULL},
	 { "new_SpaceGroup", _wrap_new_SpaceGroup, METH_NOARGS, NULL},
	 { "delete_SpaceGroup", _wrap_delete_SpaceGroup, METH_O, NULL},
	 { "SpaceGroup_SetHMName", _wrap_SpaceGroup_SetHMName, METH_VARARGS, NULL},
	 { "SpaceGroup_SetHallName", _wrap_SpaceGroup_SetHallName, METH_VARARGS, NULL},
	 { "SpaceGroup_SetId", _wrap_SpaceGroup_SetId, METH_VARARGS, NULL},
	 { "SpaceGroup_AddTransform", _wrap_SpaceGroup_AddTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_GetHMName", _wrap_SpaceGroup_GetHMName, METH_O, NULL},
	 { "SpaceGroup_GetHallName", _wrap_SpaceGroup_GetHallName, METH_O, NULL},
	 { "SpaceGroup_GetId", _wrap_SpaceGroup_GetId, METH_O, NULL},
	 { "SpaceGroup_GetOriginAlternative", _wrap_SpaceGroup_GetOriginAlternative, METH_O, NULL},
	 { "SpaceGroup_Transform", _wrap_SpaceGroup_Transform, METH_VARARGS, NULL},
	 { "SpaceGroup_BeginTransform", _wrap_SpaceGroup_BeginTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_NextTransform", _wrap_SpaceGroup_NextTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_GetSpaceGroup", _wrap_SpaceGroup_GetSpaceGroup, METH_VARARGS, NULL},
	 { "SpaceGroup_Find", _wrap_SpaceGroup_Find, METH_O, NULL},
	 { "SpaceGroup___eq__", _wrap_SpaceGroup___eq__, METH_VARARGS, NULL},
	 { "SpaceGroup___ne__", _wrap_SpaceGroup___ne__, METH_VARARGS, NULL},
	 { "SpaceGroup_IsValid", _wrap_SpaceGroup_IsValid, METH_O, NULL},
	 { "SpaceGroup_HEXAGONAL_ORIGIN_get", _wrap_SpaceGroup_HEXAGONAL_ORIGIN_get, METH_O, NULL},
	 { "SpaceGroup_swigregister", SpaceGroup_swigregister, METH_O, NULL},
	 { "SpaceGroup_swiginit", SpaceGroup_swiginit, METH_VARARGS, NULL},
	 { "new_OBBitVec", _wrap_new_OBBitVec, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Construct a bit vector of the default size Construct a bit vector of STARTWORDS size,\n"
		"	      cleared to all zero bits.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Construct a bit vector of maxbits bits Construct a bit vector with a size in bits\n"
		"	      of ``size_in_bits`` rounded up to the nearest word\n"
		"		  and cleared to all zero bits.\n"
		"		  :type size_in_bits: int, in\n"
		":param size_in_bits: The number of bits for which to reserve space\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Copy constructor (result has same number of bits) Construct a bit vector which is an exact\n"
		"	      duplicate of ``bv``.\n"
		"		  :type bv: :py:class:`OBBitVec`, in\n"
		":param bv: The other bit vector to copy to this\n"
		""},
	 { "OBBitVec_SetBitOn", _wrap_OBBitVec_SetBitOn, METH_VARARGS, "Set the ``bit_offset`` 'th bit to 1"},
	 { "OBBitVec_SetBitOff", _wrap_OBBitVec_SetBitOff, METH_VARARGS, "Set the ``bit_offset`` 'th bit to 0"},
	 { "OBBitVec_SetRangeOn", _wrap_OBBitVec_SetRangeOn, METH_VARARGS, "Set the range of bits from ``lo_bit_offset`` to ``hi_bit_offset`` to 1"},
	 { "OBBitVec_SetRangeOff", _wrap_OBBitVec_SetRangeOff, METH_VARARGS, "Set the range of bits from ``lo_bit_offset`` to ``hi_bit_offset`` to 0"},
	 { "OBBitVec_Fold", _wrap_OBBitVec_Fold, METH_VARARGS, "Reduce the size of the vector by or-ing the excess bits over the start"},
	 { "OBBitVec_FirstBit", _wrap_OBBitVec_FirstBit, METH_VARARGS, "\n"
		" Find the first true bit at or after ``bit_offset`` Searches the vector for the first true value, starting at the ``bit_offset`` 'th bit\n"
		"          :type bit_offset: int, in, optional\n"
		"          :param bit_offset: the first bit to consider\n"
		"    		  :rtype: int\n"
		":return: the bit offset of the first true bit at or after ``bit_offset``, or -1 if there is none\n"
		""},
	 { "OBBitVec_NextBit", _wrap_OBBitVec_NextBit, METH_VARARGS, "Find the next true bit after ``last_bit_offset``"},
	 { "OBBitVec_EndBit", _wrap_OBBitVec_EndBit, METH_O, "Return the bit offset of the last bit (for iterating) i.e. -1"},
	 { "OBBitVec_GetSize", _wrap_OBBitVec_GetSize, METH_O, "Return the number of words ( NOT the number of bits )."},
	 { "OBBitVec_CountBits", _wrap_OBBitVec_CountBits, METH_O, "Return the number of bits which are set to 1 in the vector"},
	 { "OBBitVec_IsEmpty", _wrap_OBBitVec_IsEmpty, METH_O, "Are there no bits set to 1 in this vector?"},
	 { "OBBitVec_Resize", _wrap_OBBitVec_Resize, METH_VARARGS, "\n"
		" Reserve space for ``size_in_bits`` bits Reserve space for ``size_in_bits`` bits rounded up\n"
		"	      :type size_in_bits: int, in\n"
		":param size_in_bits: the number of bits\n"
		"    	  :rtype: boolean\n"
		":return: true if enlargement was necessary, false otherwise\n"
		""},
	 { "OBBitVec_ResizeWords", _wrap_OBBitVec_ResizeWords, METH_VARARGS, "\n"
		" Reserve space for ``size_in_words`` words Reserve space for ``size_in_words`` words\n"
		"	      :type size_in_words: int, in\n"
		":param size_in_words: the number of words\n"
		"    	  :rtype: boolean\n"
		":return: true if enlargement was necessary, false otherwise\n"
		""},
	 { "OBBitVec_BitIsSet", _wrap_OBBitVec_BitIsSet, METH_VARARGS, "\n"
		" Asks if the ``bit_offset`` 'th bit is set Is the ``bit_offset`` 'th bit set ?\n"
		"          :type bit_offset: int, in\n"
		"          :param bit_offset: a zero based offset into the bit vector\n"
		"    		  :rtype: boolean\n"
		":return: true if it is set, false otherwise\n"
		""},
	 { "OBBitVec_FromVecInt", _wrap_OBBitVec_FromVecInt, METH_VARARGS, "Sets the bits listed as bit offsets"},
	 { "OBBitVec_FromString", _wrap_OBBitVec_FromString, METH_VARARGS, "Sets the bits listed as a string of integers"},
	 { "OBBitVec_ToVecInt", _wrap_OBBitVec_ToVecInt, METH_VARARGS, "List the offsets of the bits which are set"},
	 { "OBBitVec_Clear", _wrap_OBBitVec_Clear, METH_O, "Set all bits to zero"},
	 { "OBBitVec_Negate", _wrap_OBBitVec_Negate, METH_O, "\n"
		"Inverts every bit in the vector Inverts the entire vector.\n"
		"      Note that this may give unexpected results, as the vector\n"
		"	  can be considered to end in an arbitrary number of zero bits.\n"
		""},
	 { "OBBitVec_GetWords", _wrap_OBBitVec_GetWords, METH_VARARGS, "\n"
		" Return a copy of the internal vector of words, at the end of ``vec`` Copy the internal word vector.\n"
		"	      The copy is appended to ``vec``.\n"
		"		  :type vec: OpenBabel::OBBitVec::word_vector, out\n"
		":param vec: a vector of words to which to append the data\n"
		""},
	 { "OBBitVec___iand__", _wrap_OBBitVec___iand__, METH_VARARGS, "And-equals operator"},
	 { "OBBitVec___ior__", _wrap_OBBitVec___ior__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Or-equals operator\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Or-equals operator for integer Or the bit at offset ``bit_offset`` with 1\n"
		""},
	 { "OBBitVec___ixor__", _wrap_OBBitVec___ixor__, METH_VARARGS, "Exclusive-or-equals operator"},
	 { "OBBitVec___isub__", _wrap_OBBitVec___isub__, METH_VARARGS, "Minus-equals operator"},
	 { "OBBitVec___iadd__", _wrap_OBBitVec___iadd__, METH_VARARGS, "Plus-equals operator"},
	 { "__or__", _wrap___or__, METH_VARARGS, "Or operator"},
	 { "__and__", _wrap___and__, METH_VARARGS, "And operator"},
	 { "__xor__", _wrap___xor__, METH_VARARGS, "Exclusive-or operator"},
	 { "__sub__", _wrap___sub__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Vector subtraction\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Unary minus\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Minus operator\n"
		""},
	 { "__eq__", _wrap___eq__, METH_VARARGS, "Equivalency operator"},
	 { "__lt__", _wrap___lt__, METH_VARARGS, "Smaller-than operator"},
	 { "__rshift__", _wrap___rshift__, METH_VARARGS, "Input from a stream"},
	 { "delete_OBBitVec", _wrap_delete_OBBitVec, METH_O, NULL},
	 { "OBBitVec_swigregister", OBBitVec_swigregister, METH_O, NULL},
	 { "OBBitVec_swiginit", OBBitVec_swiginit, METH_VARARGS, NULL},
	 { "Tanimoto", _wrap_Tanimoto, METH_VARARGS, "The Tanimoto coefficient, which may be regarded as the proportion of the \"on-bits\" which are shared."},
	 { "OBReleaseVersion", _wrap_OBReleaseVersion, METH_NOARGS, "\n"
		":rtype: string\n"
		":return: the version of the Open Babel library for feature-detection (e.g. \"2.3.1\")\n"
		""},
	 { "new_OBGenericData", _wrap_new_OBGenericData, METH_VARARGS, NULL},
	 { "OBGenericData_Clone", _wrap_OBGenericData_Clone, METH_VARARGS, NULL},
	 { "delete_OBGenericData", _wrap_delete_OBGenericData, METH_O, NULL},
	 { "OBGenericData_SetAttribute", _wrap_OBGenericData_SetAttribute, METH_VARARGS, "Set the attribute (key), which can be used to retrieve this data"},
	 { "OBGenericData_SetOrigin", _wrap_OBGenericData_SetOrigin, METH_VARARGS, "Set the origin of this data, which can be used to filter the data"},
	 { "OBGenericData_GetAttribute", _wrap_OBGenericData_GetAttribute, METH_O, "\n"
		":rtype: string\n"
		":return: The attribute (key), which can be used to retrieve this data\n"
		""},
	 { "OBGenericData_GetDataType", _wrap_OBGenericData_GetDataType, METH_O, "\n"
		":rtype: int\n"
		":return: the data type for this object as defined in OBGenericDataType\n"
		""},
	 { "OBGenericData_GetValue", _wrap_OBGenericData_GetValue, METH_O, "Base class returns a default value (the attribute type) but should never be called"},
	 { "OBGenericData_GetOrigin", _wrap_OBGenericData_GetOrigin, METH_O, NULL},
	 { "OBGenericData_swigregister", OBGenericData_swigregister, METH_O, NULL},
	 { "OBGenericData_swiginit", OBGenericData_swiginit, METH_VARARGS, NULL},
	 { "delete_OBBase", _wrap_delete_OBBase, METH_O, NULL},
	 { "OBBase_Clear", _wrap_OBBase_Clear, METH_O, "Clear any and all data associated with this object"},
	 { "OBBase_DoTransformations", _wrap_OBBase_DoTransformations, METH_VARARGS, "Perform a set of transformations specified by the user Typically these are program options to filter or modify an object For example, see OBMol::DoTransformations() and OBMol::ClassDescription() Base type does nothing"},
	 { "OBBase_ClassDescription", _wrap_OBBase_ClassDescription, METH_NOARGS, "\n"
		":rtype: string\n"
		":return: A list of descriptions of command-line options for DoTransformations()\n"
		""},
	 { "OBBase_GetTitle", _wrap_OBBase_GetTitle, METH_VARARGS, "Base type does nothing Made virtual around r3535 to simplify code which passes around OBBase*."},
	 { "OBBase_SetTitle", _wrap_OBBase_SetTitle, METH_VARARGS, NULL},
	 { "OBBase_HasData", _wrap_OBBase_HasData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists, for a given OBGenericDataType\n"
		""},
	 { "OBBase_DeleteData", _wrap_OBBase_DeleteData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Delete any data matching the given OBGenericDataType\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Delete the given generic data from this object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Delete all of the given generic data from this object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Deletes the generic data with the specified attribute, returning false if not found\n"
		""},
	 { "OBBase_CloneData", _wrap_OBBase_CloneData, METH_VARARGS, "Adds a copy of a data object; does nothing if d == NULL Since: version 2.2"},
	 { "OBBase_DataSize", _wrap_OBBase_DataSize, METH_O, "\n"
		":rtype: int\n"
		":return: the number of OBGenericData items attached to this molecule.\n"
		""},
	 { "OBBase_GetAllData", _wrap_OBBase_GetAllData, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: the all matching data for a given type from OBGenericDataType    or an empty vector if nothing matches Since: version 2.2\n"
		""},
	 { "OBBase_GetData", _wrap_OBBase_GetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the first matching data for a given type from OBGenericDataType    or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: any data matching the given attribute name or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: any data matching the given attribute name or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: all data, suitable for iterating\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: all data with a specific origin, suitable for iterating\n"
		""},
	 { "OBBase_BeginData", _wrap_OBBase_BeginData, METH_O, "\n"
		":rtype: OpenBabel::OBDataIterator\n"
		":return: An iterator pointing to the beginning of the data\n"
		""},
	 { "OBBase_EndData", _wrap_OBBase_EndData, METH_O, "\n"
		":rtype: OpenBabel::OBDataIterator\n"
		":return: An iterator pointing to the end of the data\n"
		""},
	 { "new_OBBase", _wrap_new_OBBase, METH_NOARGS, NULL},
	 { "OBBase_swigregister", OBBase_swigregister, METH_O, NULL},
	 { "OBBase_swiginit", OBBase_swiginit, METH_VARARGS, NULL},
	 { "new_OBCommentData", _wrap_new_OBCommentData, METH_VARARGS, NULL},
	 { "OBCommentData_SetData", _wrap_OBCommentData_SetData, METH_VARARGS, NULL},
	 { "OBCommentData_GetData", _wrap_OBCommentData_GetData, METH_O, NULL},
	 { "delete_OBCommentData", _wrap_delete_OBCommentData, METH_O, NULL},
	 { "OBCommentData_swigregister", OBCommentData_swigregister, METH_O, NULL},
	 { "OBCommentData_swiginit", OBCommentData_swiginit, METH_VARARGS, NULL},
	 { "new_OBExternalBond", _wrap_new_OBExternalBond, METH_VARARGS, NULL},
	 { "delete_OBExternalBond", _wrap_delete_OBExternalBond, METH_O, NULL},
	 { "OBExternalBond_GetIdx", _wrap_OBExternalBond_GetIdx, METH_O, NULL},
	 { "OBExternalBond_GetAtom", _wrap_OBExternalBond_GetAtom, METH_O, NULL},
	 { "OBExternalBond_GetBond", _wrap_OBExternalBond_GetBond, METH_O, NULL},
	 { "OBExternalBond_SetIdx", _wrap_OBExternalBond_SetIdx, METH_VARARGS, NULL},
	 { "OBExternalBond_SetAtom", _wrap_OBExternalBond_SetAtom, METH_VARARGS, NULL},
	 { "OBExternalBond_SetBond", _wrap_OBExternalBond_SetBond, METH_VARARGS, NULL},
	 { "OBExternalBond_swigregister", OBExternalBond_swigregister, METH_O, NULL},
	 { "OBExternalBond_swiginit", OBExternalBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBExternalBondData", _wrap_new_OBExternalBondData, METH_NOARGS, NULL},
	 { "OBExternalBondData_SetData", _wrap_OBExternalBondData_SetData, METH_VARARGS, NULL},
	 { "OBExternalBondData_GetData", _wrap_OBExternalBondData_GetData, METH_O, NULL},
	 { "delete_OBExternalBondData", _wrap_delete_OBExternalBondData, METH_O, NULL},
	 { "OBExternalBondData_swigregister", OBExternalBondData_swigregister, METH_O, NULL},
	 { "OBExternalBondData_swiginit", OBExternalBondData_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairData", _wrap_new_OBPairData, METH_NOARGS, NULL},
	 { "OBPairData_SetValue", _wrap_OBPairData_SetValue, METH_VARARGS, NULL},
	 { "delete_OBPairData", _wrap_delete_OBPairData, METH_O, NULL},
	 { "OBPairData_swigregister", OBPairData_swigregister, METH_O, NULL},
	 { "OBPairData_swiginit", OBPairData_swiginit, METH_VARARGS, NULL},
	 { "new_OBSetData", _wrap_new_OBSetData, METH_NOARGS, NULL},
	 { "OBSetData_AddData", _wrap_OBSetData_AddData, METH_VARARGS, "Add an OBGenericData element to the set."},
	 { "OBSetData_SetData", _wrap_OBSetData_SetData, METH_VARARGS, "Set the array of data to a new vector"},
	 { "OBSetData_GetData", _wrap_OBSetData_GetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the OBGenericData associate with the attribute name parameter.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the OBGenericData associate with the attribute name parameter.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Gets the entire set.\n"
		""},
	 { "OBSetData_GetBegin", _wrap_OBSetData_GetBegin, METH_O, "Get the begin iterator."},
	 { "OBSetData_GetEnd", _wrap_OBSetData_GetEnd, METH_O, "Get the end iterator."},
	 { "OBSetData_DeleteData", _wrap_OBSetData_DeleteData, METH_VARARGS, "Delete the matching OBGenericData element."},
	 { "delete_OBSetData", _wrap_delete_OBSetData, METH_O, NULL},
	 { "OBSetData_swigregister", OBSetData_swigregister, METH_O, NULL},
	 { "OBSetData_swiginit", OBSetData_swiginit, METH_VARARGS, NULL},
	 { "new_OBVirtualBond", _wrap_new_OBVirtualBond, METH_VARARGS, NULL},
	 { "OBVirtualBond_GetBgn", _wrap_OBVirtualBond_GetBgn, METH_O, NULL},
	 { "OBVirtualBond_GetEnd", _wrap_OBVirtualBond_GetEnd, METH_O, NULL},
	 { "OBVirtualBond_GetOrder", _wrap_OBVirtualBond_GetOrder, METH_O, NULL},
	 { "OBVirtualBond_GetStereo", _wrap_OBVirtualBond_GetStereo, METH_O, NULL},
	 { "delete_OBVirtualBond", _wrap_delete_OBVirtualBond, METH_O, NULL},
	 { "OBVirtualBond_swigregister", OBVirtualBond_swigregister, METH_O, NULL},
	 { "OBVirtualBond_swiginit", OBVirtualBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBRingData", _wrap_new_OBRingData, METH_VARARGS, NULL},
	 { "delete_OBRingData", _wrap_delete_OBRingData, METH_O, NULL},
	 { "OBRingData_SetData", _wrap_OBRingData_SetData, METH_VARARGS, NULL},
	 { "OBRingData_PushBack", _wrap_OBRingData_PushBack, METH_VARARGS, NULL},
	 { "OBRingData_GetData", _wrap_OBRingData_GetData, METH_O, NULL},
	 { "OBRingData_BeginRings", _wrap_OBRingData_BeginRings, METH_O, NULL},
	 { "OBRingData_EndRings", _wrap_OBRingData_EndRings, METH_O, NULL},
	 { "OBRingData_BeginRing", _wrap_OBRingData_BeginRing, METH_VARARGS, NULL},
	 { "OBRingData_NextRing", _wrap_OBRingData_NextRing, METH_VARARGS, NULL},
	 { "OBRingData_swigregister", OBRingData_swigregister, METH_O, NULL},
	 { "OBRingData_swiginit", OBRingData_swiginit, METH_VARARGS, NULL},
	 { "new_OBUnitCell", _wrap_new_OBUnitCell, METH_VARARGS, "public constructor"},
	 { "delete_OBUnitCell", _wrap_delete_OBUnitCell, METH_O, NULL},
	 { "OBUnitCell_SetData", _wrap_OBUnitCell_SetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Constructs the cell matrix in lower triangular form from the values supplied\n"
		":type a: float\n"
		":param a: The length a\n"
		":type b: float\n"
		":param b: The length b\n"
		":type c: float\n"
		":param c: The length c\n"
		":type alpha: float\n"
		":param alpha: The angle alpha\n"
		":type beta: float\n"
		":param beta: The angle beta\n"
		":type gamma: float\n"
		":param gamma: The angle gamma\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Constructs the cell matrix using the supplied row vectors\n"
		":type v1: :py:class:`vector3`\n"
		":param v1: The x-vector\n"
		":type v2: :py:class:`vector3`\n"
		":param v2: The y-vector\n"
		":type v3: :py:class:`vector3`\n"
		":param v3: The z-vector\n"
		"See also: OBUnitCell::GetCellVectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Sets the unit cell matrix\n"
		":type m: :py:class:`matrix3x3`\n"
		":param m: The unit cell matrix (row vectors)\n"
		"See also: OBUnitCell::GetCellMatrix\n"
		""},
	 { "OBUnitCell_SetOffset", _wrap_OBUnitCell_SetOffset, METH_VARARGS, "Set the offset to the origin to ``v1``"},
	 { "OBUnitCell_SetSpaceGroup", _wrap_OBUnitCell_SetSpaceGroup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the space group for this unit cell. Does not create an OBSymmetryData entry\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the space group symbol for this unit cell. Does not create an OBSymmetryData entry or attempt to convert  between different symbol notations\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Set the space group for this unit cell. Each spacegroup-symbol has a numeric equivalent which is easier to use to convert between notations. Does not create an OBSymmetryData entry or attempt to convert  between different symbol notations\n"
		""},
	 { "OBUnitCell_SetLatticeType", _wrap_OBUnitCell_SetLatticeType, METH_VARARGS, "Set the Bravais lattice type for this unit cell"},
	 { "OBUnitCell_FillUnitCell", _wrap_OBUnitCell_FillUnitCell, METH_VARARGS, "Duplicate symmetry-unique atoms to fill out the unit cell of the molecule, based on the known space group"},
	 { "OBUnitCell_GetA", _wrap_OBUnitCell_GetA, METH_VARARGS, "\n"
		" TODO: Remove nonconst overloads in OBUnitCell on next version bump :rtype: float\n"
		":return: vector a\n"
		""},
	 { "OBUnitCell_GetB", _wrap_OBUnitCell_GetB, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: vector b\n"
		""},
	 { "OBUnitCell_GetC", _wrap_OBUnitCell_GetC, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: vector c\n"
		""},
	 { "OBUnitCell_GetAlpha", _wrap_OBUnitCell_GetAlpha, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle alpha\n"
		""},
	 { "OBUnitCell_GetBeta", _wrap_OBUnitCell_GetBeta, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle beta\n"
		""},
	 { "OBUnitCell_GetGamma", _wrap_OBUnitCell_GetGamma, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle gamma\n"
		""},
	 { "OBUnitCell_GetOffset", _wrap_OBUnitCell_GetOffset, METH_VARARGS, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: any offset in the origin of the periodic boundaries\n"
		""},
	 { "OBUnitCell_GetSpaceGroup", _wrap_OBUnitCell_GetSpaceGroup, METH_VARARGS, "\n"
		":rtype: :py:class:`SpaceGroup`\n"
		":return: the text representation of the space group for this unit cell\n"
		""},
	 { "OBUnitCell_GetSpaceGroupName", _wrap_OBUnitCell_GetSpaceGroupName, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the text representation of the space group for this unit cell\n"
		""},
	 { "OBUnitCell_GetLatticeType", _wrap_OBUnitCell_GetLatticeType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: int\n"
		":return: lattice type (based on the ``spacegroup)``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: lattice type (based on angles and distances)\n"
		""},
	 { "OBUnitCell_GetCellVectors", _wrap_OBUnitCell_GetCellVectors, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >\n"
		":return: v1, v2, v3 cell vectors\n"
		""},
	 { "OBUnitCell_GetCellMatrix", _wrap_OBUnitCell_GetCellMatrix, METH_VARARGS, "\n"
		" Access to the cell matrix as row vectors, useful for writing input files. Equivalent to the transpose of GetOrientationMatrix() * GetOrthoMatrix() :rtype: :py:class:`matrix3x3`\n"
		":return: The cell matrix with row vectors. See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetOrthoMatrix", _wrap_OBUnitCell_GetOrthoMatrix, METH_VARARGS, "\n"
		":rtype: :py:class:`matrix3x3`\n"
		":return: The orthogonalization matrix, used for converting from fractional to Cartesian coords. See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetOrientationMatrix", _wrap_OBUnitCell_GetOrientationMatrix, METH_VARARGS, "\n"
		" Used to convert fractional and cartesian coordinates if the cell is not oriented in standard form (a parallel to x axis, b in xy plane) :rtype: :py:class:`matrix3x3`\n"
		":return: The orientation matrix See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetFractionalMatrix", _wrap_OBUnitCell_GetFractionalMatrix, METH_VARARGS, "\n"
		":rtype: :py:class:`matrix3x3`\n"
		":return: The fractionalization matrix, used for converting from Cartesian to fractional coords. See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_FractionalToCartesian", _wrap_OBUnitCell_FractionalToCartesian, METH_VARARGS, "\n"
		" Convenience function to convert fractional coordinates to cartesian coordinates. Returns GetOrientationMatrix() * GetOrthoMatrix() * frac + GetOffset() :type frac: :py:class:`vector3`\n"
		":param frac: Vector of fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian coordinates\n"
		""},
	 { "OBUnitCell_CartesianToFractional", _wrap_OBUnitCell_CartesianToFractional, METH_VARARGS, "\n"
		" Convenience function to convert cartesian coordinates to fractional coordinates. Returns GetFractionalMatrix() * GetOrientationMatrix().inverse() * (cart - GetOffset()) :type cart: :py:class:`vector3`\n"
		":param cart: Vector of cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional coordinates\n"
		""},
	 { "OBUnitCell_WrapCartesianCoordinate", _wrap_OBUnitCell_WrapCartesianCoordinate, METH_VARARGS, "\n"
		" Wraps cartesian coordinate to fall within the unit cell. :type cart: :py:class:`vector3`\n"
		":param cart: Vector of cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian coordinates within cell boundaries.\n"
		""},
	 { "OBUnitCell_WrapFractionalCoordinate", _wrap_OBUnitCell_WrapFractionalCoordinate, METH_VARARGS, "\n"
		" Wraps fractional coordinate to fall within the unit cell. :type frac: :py:class:`vector3`\n"
		":param frac: Vector of fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional coordinates within cell boundaries (between 0 and 1). TODO: Make OBUnitCell::WrapFractionalCoordinate static in the next ABI break\n"
		""},
	 { "OBUnitCell_UnwrapCartesianNear", _wrap_OBUnitCell_UnwrapCartesianNear, METH_VARARGS, "\n"
		" Unwraps Cartesian coordinates near a reference location :type new_loc: :py:class:`vector3`\n"
		":param new_loc: Cartesian coordinates of target :type ref_loc: :py:class:`vector3`\n"
		":param ref_loc: Cartesian coordinates of the reference location :rtype: :py:class:`vector3`\n"
		":return: Unwrapped coordinates of new_loc near ref_loc\n"
		""},
	 { "OBUnitCell_UnwrapFractionalNear", _wrap_OBUnitCell_UnwrapFractionalNear, METH_VARARGS, "\n"
		" Unwraps fractional coordinates near a reference location. :type new_loc: :py:class:`vector3`\n"
		":param new_loc: Fractional coordinates of target :type ref_loc: :py:class:`vector3`\n"
		":param ref_loc: Fractional coordinates of the reference location :rtype: :py:class:`vector3`\n"
		":return: Unwrapped coordinates of new_loc near ref_loc TODO: Add a simple test case/example, like unwrapNear(<0.9, 0.2, 0.2>, <0.3, 0.9, 0.2>) -> <-0.1, 1.2, 0.2>\n"
		""},
	 { "OBUnitCell_MinimumImageCartesian", _wrap_OBUnitCell_MinimumImageCartesian, METH_VARARGS, "\n"
		" Applies the minimum image convention to a Cartesian displacement vector :type cart: :py:class:`vector3`\n"
		":param cart: Displacement vector between two atoms in Cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian difference, wrapped within half the unit cell\n"
		""},
	 { "OBUnitCell_MinimumImageFractional", _wrap_OBUnitCell_MinimumImageFractional, METH_VARARGS, "\n"
		" Applies the minimum image convention to a fractional displacement vector :param cart: Displacement vector between two atoms in fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional difference, wrapped within half the unit cell (-0.5 to 0.5)\n"
		""},
	 { "OBUnitCell_GetSpaceGroupNumber", _wrap_OBUnitCell_GetSpaceGroupNumber, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The numeric value of the given spacegroup\n"
		""},
	 { "OBUnitCell_GetCellVolume", _wrap_OBUnitCell_GetCellVolume, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: The cell volume (in Angstroms^3)\n"
		""},
	 { "OBUnitCell_swigregister", OBUnitCell_swigregister, METH_O, NULL},
	 { "OBUnitCell_swiginit", OBUnitCell_swiginit, METH_VARARGS, NULL},
	 { "new_OBConformerData", _wrap_new_OBConformerData, METH_VARARGS, NULL},
	 { "delete_OBConformerData", _wrap_delete_OBConformerData, METH_O, NULL},
	 { "OBConformerData_SetDimension", _wrap_OBConformerData_SetDimension, METH_VARARGS, NULL},
	 { "OBConformerData_SetEnergies", _wrap_OBConformerData_SetEnergies, METH_VARARGS, NULL},
	 { "OBConformerData_SetForces", _wrap_OBConformerData_SetForces, METH_VARARGS, NULL},
	 { "OBConformerData_SetVelocities", _wrap_OBConformerData_SetVelocities, METH_VARARGS, NULL},
	 { "OBConformerData_SetDisplacements", _wrap_OBConformerData_SetDisplacements, METH_VARARGS, NULL},
	 { "OBConformerData_SetData", _wrap_OBConformerData_SetData, METH_VARARGS, NULL},
	 { "OBConformerData_GetDimension", _wrap_OBConformerData_GetDimension, METH_O, NULL},
	 { "OBConformerData_GetEnergies", _wrap_OBConformerData_GetEnergies, METH_O, NULL},
	 { "OBConformerData_GetForces", _wrap_OBConformerData_GetForces, METH_O, NULL},
	 { "OBConformerData_GetVelocities", _wrap_OBConformerData_GetVelocities, METH_O, NULL},
	 { "OBConformerData_GetDisplacements", _wrap_OBConformerData_GetDisplacements, METH_O, NULL},
	 { "OBConformerData_GetData", _wrap_OBConformerData_GetData, METH_O, NULL},
	 { "OBConformerData_swigregister", OBConformerData_swigregister, METH_O, NULL},
	 { "OBConformerData_swiginit", OBConformerData_swiginit, METH_VARARGS, NULL},
	 { "new_OBSymmetryData", _wrap_new_OBSymmetryData, METH_VARARGS, NULL},
	 { "delete_OBSymmetryData", _wrap_delete_OBSymmetryData, METH_O, NULL},
	 { "OBSymmetryData_SetData", _wrap_OBSymmetryData_SetData, METH_VARARGS, NULL},
	 { "OBSymmetryData_SetPointGroup", _wrap_OBSymmetryData_SetPointGroup, METH_VARARGS, NULL},
	 { "OBSymmetryData_SetSpaceGroup", _wrap_OBSymmetryData_SetSpaceGroup, METH_VARARGS, NULL},
	 { "OBSymmetryData_GetPointGroup", _wrap_OBSymmetryData_GetPointGroup, METH_O, NULL},
	 { "OBSymmetryData_GetSpaceGroup", _wrap_OBSymmetryData_GetSpaceGroup, METH_O, NULL},
	 { "OBSymmetryData_swigregister", OBSymmetryData_swigregister, METH_O, NULL},
	 { "OBSymmetryData_swiginit", OBSymmetryData_swiginit, METH_VARARGS, NULL},
	 { "new_OBTorsion", _wrap_new_OBTorsion, METH_VARARGS, NULL},
	 { "delete_OBTorsion", _wrap_delete_OBTorsion, METH_O, NULL},
	 { "OBTorsion_Clear", _wrap_OBTorsion_Clear, METH_O, NULL},
	 { "OBTorsion_Empty", _wrap_OBTorsion_Empty, METH_O, NULL},
	 { "OBTorsion_AddTorsion", _wrap_OBTorsion_AddTorsion, METH_VARARGS, NULL},
	 { "OBTorsion_SetAngle", _wrap_OBTorsion_SetAngle, METH_VARARGS, NULL},
	 { "OBTorsion_SetData", _wrap_OBTorsion_SetData, METH_VARARGS, NULL},
	 { "OBTorsion_GetAngle", _wrap_OBTorsion_GetAngle, METH_VARARGS, NULL},
	 { "OBTorsion_GetBondIdx", _wrap_OBTorsion_GetBondIdx, METH_O, "\n"
		" Gets the bond index of the central bond :rtype: int\n"
		":return: int bond index\n"
		""},
	 { "OBTorsion_GetSize", _wrap_OBTorsion_GetSize, METH_O, NULL},
	 { "OBTorsion_GetBC", _wrap_OBTorsion_GetBC, METH_O, "\n"
		" Gets the two central atoms of ABCD torsion   :rtype: std::pair< OpenBabel::OBAtom *,OpenBabel::OBAtom * >\n"
		":return: pair<OBAtom*,OBAtom*>\n"
		""},
	 { "OBTorsion_GetADs", _wrap_OBTorsion_GetADs, METH_O, "\n"
		" Gets the vector of distal atoms of ABCD torsion :rtype: std::vector< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double >,std::allocator< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double > > >\n"
		":return: vector of A,D atom pointers and a double\n"
		""},
	 { "OBTorsion_IsProtonRotor", _wrap_OBTorsion_IsProtonRotor, METH_O, NULL},
	 { "OBTorsion_swigregister", OBTorsion_swigregister, METH_O, NULL},
	 { "OBTorsion_swiginit", OBTorsion_swiginit, METH_VARARGS, NULL},
	 { "OBTorsionData_Clear", _wrap_OBTorsionData_Clear, METH_O, NULL},
	 { "OBTorsionData_GetData", _wrap_OBTorsionData_GetData, METH_O, "\n"
		" Gets a vector of the OBTorsion objects :rtype: std::vector< OpenBabel::OBTorsion,std::allocator< OpenBabel::OBTorsion > >\n"
		":return: the vector of torsions\n"
		""},
	 { "OBTorsionData_GetSize", _wrap_OBTorsionData_GetSize, METH_O, "\n"
		" Gets the number of torsion structs :rtype: int\n"
		":return: integer count of the number of torsions\n"
		""},
	 { "OBTorsionData_SetData", _wrap_OBTorsionData_SetData, METH_VARARGS, NULL},
	 { "OBTorsionData_FillTorsionArray", _wrap_OBTorsionData_FillTorsionArray, METH_VARARGS, NULL},
	 { "delete_OBTorsionData", _wrap_delete_OBTorsionData, METH_O, NULL},
	 { "OBTorsionData_swigregister", OBTorsionData_swigregister, METH_O, NULL},
	 { "new_OBAngle", _wrap_new_OBAngle, METH_VARARGS, NULL},
	 { "delete_OBAngle", _wrap_delete_OBAngle, METH_O, NULL},
	 { "OBAngle___eq__", _wrap_OBAngle___eq__, METH_VARARGS, NULL},
	 { "OBAngle_Clear", _wrap_OBAngle_Clear, METH_O, NULL},
	 { "OBAngle_GetAngle", _wrap_OBAngle_GetAngle, METH_O, "\n"
		" Gets the OBAngle angle value :rtype: float\n"
		":return: angle in radians\n"
		""},
	 { "OBAngle_SetAngle", _wrap_OBAngle_SetAngle, METH_VARARGS, "\n"
		" Sets the OBAngle to ``radians`` :type angle: float\n"
		":param angle: in radians\n"
		""},
	 { "OBAngle_SetAtoms", _wrap_OBAngle_SetAtoms, METH_VARARGS, NULL},
	 { "OBAngle_swigregister", OBAngle_swigregister, METH_O, NULL},
	 { "OBAngle_swiginit", OBAngle_swiginit, METH_VARARGS, NULL},
	 { "OBAngleData_Clear", _wrap_OBAngleData_Clear, METH_O, NULL},
	 { "OBAngleData_FillAngleArray", _wrap_OBAngleData_FillAngleArray, METH_VARARGS, NULL},
	 { "OBAngleData_SetData", _wrap_OBAngleData_SetData, METH_VARARGS, NULL},
	 { "OBAngleData_GetSize", _wrap_OBAngleData_GetSize, METH_O, "\n"
		" Gets the number of angles stored :rtype: int\n"
		":return: integer count of the number of angles\n"
		""},
	 { "delete_OBAngleData", _wrap_delete_OBAngleData, METH_O, NULL},
	 { "OBAngleData_swigregister", OBAngleData_swigregister, METH_O, NULL},
	 { "new_OBSerialNums", _wrap_new_OBSerialNums, METH_VARARGS, NULL},
	 { "OBSerialNums_GetData", _wrap_OBSerialNums_GetData, METH_O, NULL},
	 { "OBSerialNums_SetData", _wrap_OBSerialNums_SetData, METH_VARARGS, NULL},
	 { "delete_OBSerialNums", _wrap_delete_OBSerialNums, METH_O, NULL},
	 { "OBSerialNums_swigregister", OBSerialNums_swigregister, METH_O, NULL},
	 { "OBSerialNums_swiginit", OBSerialNums_swiginit, METH_VARARGS, NULL},
	 { "new_OBVibrationData", _wrap_new_OBVibrationData, METH_NOARGS, NULL},
	 { "delete_OBVibrationData", _wrap_delete_OBVibrationData, METH_O, NULL},
	 { "OBVibrationData_SetData", _wrap_OBVibrationData_SetData, METH_VARARGS, NULL},
	 { "OBVibrationData_GetLx", _wrap_OBVibrationData_GetLx, METH_O, NULL},
	 { "OBVibrationData_GetFrequencies", _wrap_OBVibrationData_GetFrequencies, METH_O, NULL},
	 { "OBVibrationData_GetIntensities", _wrap_OBVibrationData_GetIntensities, METH_O, NULL},
	 { "OBVibrationData_GetRamanActivities", _wrap_OBVibrationData_GetRamanActivities, METH_O, NULL},
	 { "OBVibrationData_GetNumberOfFrequencies", _wrap_OBVibrationData_GetNumberOfFrequencies, METH_O, NULL},
	 { "OBVibrationData_swigregister", OBVibrationData_swigregister, METH_O, NULL},
	 { "OBVibrationData_swiginit", OBVibrationData_swiginit, METH_VARARGS, NULL},
	 { "new_OBDOSData", _wrap_new_OBDOSData, METH_NOARGS, NULL},
	 { "delete_OBDOSData", _wrap_delete_OBDOSData, METH_O, NULL},
	 { "OBDOSData_SetData", _wrap_OBDOSData_SetData, METH_VARARGS, NULL},
	 { "OBDOSData_GetFermiEnergy", _wrap_OBDOSData_GetFermiEnergy, METH_O, NULL},
	 { "OBDOSData_GetEnergies", _wrap_OBDOSData_GetEnergies, METH_O, NULL},
	 { "OBDOSData_GetDensities", _wrap_OBDOSData_GetDensities, METH_O, NULL},
	 { "OBDOSData_GetIntegration", _wrap_OBDOSData_GetIntegration, METH_O, NULL},
	 { "OBDOSData_swigregister", OBDOSData_swigregister, METH_O, NULL},
	 { "OBDOSData_swiginit", OBDOSData_swiginit, METH_VARARGS, NULL},
	 { "OBOrbital_SetData", _wrap_OBOrbital_SetData, METH_VARARGS, NULL},
	 { "OBOrbital_GetEnergy", _wrap_OBOrbital_GetEnergy, METH_O, NULL},
	 { "OBOrbital_GetOccupation", _wrap_OBOrbital_GetOccupation, METH_O, NULL},
	 { "OBOrbital_GetSymbol", _wrap_OBOrbital_GetSymbol, METH_O, NULL},
	 { "new_OBOrbital", _wrap_new_OBOrbital, METH_NOARGS, NULL},
	 { "delete_OBOrbital", _wrap_delete_OBOrbital, METH_O, NULL},
	 { "OBOrbital_swigregister", OBOrbital_swigregister, METH_O, NULL},
	 { "OBOrbital_swiginit", OBOrbital_swiginit, METH_VARARGS, NULL},
	 { "new_OBOrbitalData", _wrap_new_OBOrbitalData, METH_NOARGS, NULL},
	 { "delete_OBOrbitalData", _wrap_delete_OBOrbitalData, METH_O, NULL},
	 { "OBOrbitalData_SetAlphaOrbitals", _wrap_OBOrbitalData_SetAlphaOrbitals, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetBetaOrbitals", _wrap_OBOrbitalData_SetBetaOrbitals, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetHOMO", _wrap_OBOrbitalData_SetHOMO, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetOpenShell", _wrap_OBOrbitalData_SetOpenShell, METH_VARARGS, NULL},
	 { "OBOrbitalData_IsOpenShell", _wrap_OBOrbitalData_IsOpenShell, METH_O, NULL},
	 { "OBOrbitalData_GetAlphaHOMO", _wrap_OBOrbitalData_GetAlphaHOMO, METH_O, NULL},
	 { "OBOrbitalData_GetBetaHOMO", _wrap_OBOrbitalData_GetBetaHOMO, METH_O, NULL},
	 { "OBOrbitalData_GetAlphaOrbitals", _wrap_OBOrbitalData_GetAlphaOrbitals, METH_O, NULL},
	 { "OBOrbitalData_GetBetaOrbitals", _wrap_OBOrbitalData_GetBetaOrbitals, METH_O, NULL},
	 { "OBOrbitalData_LoadClosedShellOrbitals", _wrap_OBOrbitalData_LoadClosedShellOrbitals, METH_VARARGS, "Convenience function for common cases of closed-shell calculations -- pass the energies and symmetries This method will fill the OBOrbital objects for you"},
	 { "OBOrbitalData_LoadAlphaOrbitals", _wrap_OBOrbitalData_LoadAlphaOrbitals, METH_VARARGS, "Convenience function to load alpha orbitals in an open-shell calculation"},
	 { "OBOrbitalData_LoadBetaOrbitals", _wrap_OBOrbitalData_LoadBetaOrbitals, METH_VARARGS, "Convenience function to load beta orbitals in an open-shell calculation"},
	 { "OBOrbitalData_swigregister", OBOrbitalData_swigregister, METH_O, NULL},
	 { "OBOrbitalData_swiginit", OBOrbitalData_swiginit, METH_VARARGS, NULL},
	 { "new_OBElectronicTransitionData", _wrap_new_OBElectronicTransitionData, METH_NOARGS, NULL},
	 { "delete_OBElectronicTransitionData", _wrap_delete_OBElectronicTransitionData, METH_O, NULL},
	 { "OBElectronicTransitionData_SetData", _wrap_OBElectronicTransitionData_SetData, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetEDipole", _wrap_OBElectronicTransitionData_SetEDipole, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetRotatoryStrengthsVelocity", _wrap_OBElectronicTransitionData_SetRotatoryStrengthsVelocity, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetRotatoryStrengthsLength", _wrap_OBElectronicTransitionData_SetRotatoryStrengthsLength, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_GetWavelengths", _wrap_OBElectronicTransitionData_GetWavelengths, METH_O, NULL},
	 { "OBElectronicTransitionData_GetForces", _wrap_OBElectronicTransitionData_GetForces, METH_O, NULL},
	 { "OBElectronicTransitionData_GetEDipole", _wrap_OBElectronicTransitionData_GetEDipole, METH_O, NULL},
	 { "OBElectronicTransitionData_GetRotatoryStrengthsVelocity", _wrap_OBElectronicTransitionData_GetRotatoryStrengthsVelocity, METH_O, NULL},
	 { "OBElectronicTransitionData_GetRotatoryStrengthsLength", _wrap_OBElectronicTransitionData_GetRotatoryStrengthsLength, METH_O, NULL},
	 { "OBElectronicTransitionData_swigregister", OBElectronicTransitionData_swigregister, METH_O, NULL},
	 { "OBElectronicTransitionData_swiginit", OBElectronicTransitionData_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotationData", _wrap_new_OBRotationData, METH_NOARGS, NULL},
	 { "delete_OBRotationData", _wrap_delete_OBRotationData, METH_O, NULL},
	 { "OBRotationData_SetData", _wrap_OBRotationData_SetData, METH_VARARGS, NULL},
	 { "OBRotationData_GetRotConsts", _wrap_OBRotationData_GetRotConsts, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: Rotational constants in GHz\n"
		""},
	 { "OBRotationData_GetSymmetryNumber", _wrap_OBRotationData_GetSymmetryNumber, METH_O, NULL},
	 { "OBRotationData_GetRotorType", _wrap_OBRotationData_GetRotorType, METH_O, NULL},
	 { "OBRotationData_swigregister", OBRotationData_swigregister, METH_O, NULL},
	 { "OBRotationData_swiginit", OBRotationData_swiginit, METH_VARARGS, NULL},
	 { "new_OBVectorData", _wrap_new_OBVectorData, METH_NOARGS, NULL},
	 { "delete_OBVectorData", _wrap_delete_OBVectorData, METH_O, NULL},
	 { "OBVectorData_SetData", _wrap_OBVectorData_SetData, METH_VARARGS, NULL},
	 { "OBVectorData_GetData", _wrap_OBVectorData_GetData, METH_O, NULL},
	 { "OBVectorData_swigregister", OBVectorData_swigregister, METH_O, NULL},
	 { "OBVectorData_swiginit", OBVectorData_swiginit, METH_VARARGS, NULL},
	 { "new_OBMatrixData", _wrap_new_OBMatrixData, METH_NOARGS, NULL},
	 { "delete_OBMatrixData", _wrap_delete_OBMatrixData, METH_O, NULL},
	 { "OBMatrixData_SetData", _wrap_OBMatrixData_SetData, METH_VARARGS, NULL},
	 { "OBMatrixData_GetData", _wrap_OBMatrixData_GetData, METH_O, NULL},
	 { "OBMatrixData_swigregister", OBMatrixData_swigregister, METH_O, NULL},
	 { "OBMatrixData_swiginit", OBMatrixData_swiginit, METH_VARARGS, NULL},
	 { "new_OBFreeGridPoint", _wrap_new_OBFreeGridPoint, METH_VARARGS, NULL},
	 { "delete_OBFreeGridPoint", _wrap_delete_OBFreeGridPoint, METH_O, NULL},
	 { "OBFreeGridPoint_GetX", _wrap_OBFreeGridPoint_GetX, METH_O, NULL},
	 { "OBFreeGridPoint_GetY", _wrap_OBFreeGridPoint_GetY, METH_O, NULL},
	 { "OBFreeGridPoint_GetZ", _wrap_OBFreeGridPoint_GetZ, METH_O, NULL},
	 { "OBFreeGridPoint_GetV", _wrap_OBFreeGridPoint_GetV, METH_O, NULL},
	 { "OBFreeGridPoint_SetX", _wrap_OBFreeGridPoint_SetX, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetY", _wrap_OBFreeGridPoint_SetY, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetZ", _wrap_OBFreeGridPoint_SetZ, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetV", _wrap_OBFreeGridPoint_SetV, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_swigregister", OBFreeGridPoint_swigregister, METH_O, NULL},
	 { "OBFreeGridPoint_swiginit", OBFreeGridPoint_swiginit, METH_VARARGS, NULL},
	 { "new_OBFreeGrid", _wrap_new_OBFreeGrid, METH_NOARGS, NULL},
	 { "delete_OBFreeGrid", _wrap_delete_OBFreeGrid, METH_O, NULL},
	 { "OBFreeGrid_NumPoints", _wrap_OBFreeGrid_NumPoints, METH_O, NULL},
	 { "OBFreeGrid_AddPoint", _wrap_OBFreeGrid_AddPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_BeginPoints", _wrap_OBFreeGrid_BeginPoints, METH_O, NULL},
	 { "OBFreeGrid_EndPoints", _wrap_OBFreeGrid_EndPoints, METH_O, NULL},
	 { "OBFreeGrid_BeginPoint", _wrap_OBFreeGrid_BeginPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_NextPoint", _wrap_OBFreeGrid_NextPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_Clear", _wrap_OBFreeGrid_Clear, METH_O, NULL},
	 { "OBFreeGrid_swigregister", OBFreeGrid_swigregister, METH_O, NULL},
	 { "OBFreeGrid_swiginit", OBFreeGrid_swiginit, METH_VARARGS, NULL},
	 { "new_OBPcharge", _wrap_new_OBPcharge, METH_NOARGS, NULL},
	 { "delete_OBPcharge", _wrap_delete_OBPcharge, METH_O, NULL},
	 { "OBPcharge_NumPartialCharges", _wrap_OBPcharge_NumPartialCharges, METH_O, NULL},
	 { "OBPcharge_AddPartialCharge", _wrap_OBPcharge_AddPartialCharge, METH_VARARGS, NULL},
	 { "OBPcharge_GetPartialCharge", _wrap_OBPcharge_GetPartialCharge, METH_O, NULL},
	 { "OBPcharge_swigregister", OBPcharge_swigregister, METH_O, NULL},
	 { "OBPcharge_swiginit", OBPcharge_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairInteger", _wrap_new_OBPairInteger, METH_NOARGS, NULL},
	 { "OBPairInteger_SetValue", _wrap_OBPairInteger_SetValue, METH_VARARGS, NULL},
	 { "OBPairInteger_GetGenericValue", _wrap_OBPairInteger_GetGenericValue, METH_O, NULL},
	 { "delete_OBPairInteger", _wrap_delete_OBPairInteger, METH_O, NULL},
	 { "OBPairInteger_swigregister", OBPairInteger_swigregister, METH_O, NULL},
	 { "OBPairInteger_swiginit", OBPairInteger_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairFloatingPoint", _wrap_new_OBPairFloatingPoint, METH_NOARGS, NULL},
	 { "OBPairFloatingPoint_SetValue", _wrap_OBPairFloatingPoint_SetValue, METH_VARARGS, NULL},
	 { "OBPairFloatingPoint_GetGenericValue", _wrap_OBPairFloatingPoint_GetGenericValue, METH_O, NULL},
	 { "delete_OBPairFloatingPoint", _wrap_delete_OBPairFloatingPoint, METH_O, NULL},
	 { "OBPairFloatingPoint_swigregister", OBPairFloatingPoint_swigregister, METH_O, NULL},
	 { "OBPairFloatingPoint_swiginit", OBPairFloatingPoint_swiginit, METH_VARARGS, NULL},
	 { "new_OBGridData", _wrap_new_OBGridData, METH_NOARGS, "Constructor."},
	 { "delete_OBGridData", _wrap_delete_OBGridData, METH_O, "Destructor"},
	 { "OBGridData_GetXAxis", _wrap_OBGridData_GetXAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: X axis direction.\n"
		""},
	 { "OBGridData_GetYAxis", _wrap_OBGridData_GetYAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: Y axis direction.\n"
		""},
	 { "OBGridData_GetZAxis", _wrap_OBGridData_GetZAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: Z axis direction.\n"
		""},
	 { "OBGridData_GetAxes", _wrap_OBGridData_GetAxes, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: void\n"
		":return: the three axes parallel to the grid edges. The\n"
		"    length of the returned vector is the step along that\n"
		"    direction.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: void\n"
		":return: the three axes parallel to the grid edges\n"
		""},
	 { "OBGridData_GetNumberOfPoints", _wrap_OBGridData_GetNumberOfPoints, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: void\n"
		":return: number of points along the three axes parallel to the grid edges.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: total number of points in the grid.\n"
		""},
	 { "OBGridData_GetNumberOfSteps", _wrap_OBGridData_GetNumberOfSteps, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: number of points along the three axes parallel to the grid edges.\n"
		""},
	 { "OBGridData_GetValues", _wrap_OBGridData_GetValues, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: grid values as a vector of doubles.\n"
		""},
	 { "OBGridData_GetValue", _wrap_OBGridData_GetValue, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the value at position i, j, k in the grid.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the value at a position in the grid (by interpolation)\n"
		""},
	 { "OBGridData_GetUnit", _wrap_OBGridData_GetUnit, METH_O, "\n"
		":rtype: int\n"
		":return: the unit of measure for grid spacings.\n"
		""},
	 { "OBGridData_GetMinValue", _wrap_OBGridData_GetMinValue, METH_O, "\n"
		":rtype: float\n"
		":return: the minimum value of all points in the grid.\n"
		""},
	 { "OBGridData_GetMaxValue", _wrap_OBGridData_GetMaxValue, METH_O, "\n"
		":rtype: float\n"
		":return: the maximum value of all points in the grid.\n"
		""},
	 { "OBGridData_GetOriginVector", _wrap_OBGridData_GetOriginVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: vector3 of the origin (i.e., the minimum x, y, and z coords of the grid).\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":type o: float\n"
		":param o: set to the origin (i.e., the minimum x, y, and z coords of the grid).\n"
		"Deprecated: Will be removed.\n"
		"See also: GetOriginVector()\n"
		""},
	 { "OBGridData_GetMaxVector", _wrap_OBGridData_GetMaxVector, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: The maximum point in the grid.\n"
		""},
	 { "OBGridData_GetUnrestricted", _wrap_OBGridData_GetUnrestricted, METH_O, "\n"
		":rtype: boolean\n"
		":return: the unrestricted flag.\n"
		""},
	 { "OBGridData_GetNumSymmetries", _wrap_OBGridData_GetNumSymmetries, METH_O, "\n"
		":rtype: int\n"
		":return: the number of symmetries.\n"
		""},
	 { "OBGridData_SetNumberOfPoints", _wrap_OBGridData_SetNumberOfPoints, METH_VARARGS, "Set number of points along the three axes."},
	 { "OBGridData_SetLimits", _wrap_OBGridData_SetLimits, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the limits (i.e., the origin point and the axes)\n"
		"NOTE: You must set the number of points first,\n"
		"      with SetNumberOfPoints\n"
		"      so the grid spacing can be calculated\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Deprecated: Will be removed.\n"
		"See also: SetLimits(const vector3 &origin, const vector3 &x, const vector3 &y, const vector3 &z)\n"
		""},
	 { "OBGridData_SetValue", _wrap_OBGridData_SetValue, METH_VARARGS, "Set an individual value, grid must have been initialised"},
	 { "OBGridData_SetValues", _wrap_OBGridData_SetValues, METH_VARARGS, "Set the values, this vector must match the dimensions of the grid"},
	 { "OBGridData_SetUnit", _wrap_OBGridData_SetUnit, METH_VARARGS, "Set the unit of measure"},
	 { "OBGridData_SetUnrestricted", _wrap_OBGridData_SetUnrestricted, METH_VARARGS, "Set the unrestricted flag"},
	 { "OBGridData_SetNumSymmetries", _wrap_OBGridData_SetNumSymmetries, METH_VARARGS, "Set the number of symmetries"},
	 { "OBGridData_swigregister", OBGridData_swigregister, METH_O, NULL},
	 { "OBGridData_swiginit", OBGridData_swiginit, METH_VARARGS, NULL},
	 { "OBKekulize", _wrap_OBKekulize, METH_O, "\n"
		"Kekulize a molecule by assigning bond orders of 1 or 2 to aromatic bonds\n"
		"\n"
		"Some file formats describe bond orders as aromatic. Such bonds require kekulization\n"
		"before the molecule is returned by the reader. Normally, a user should never need\n"
		"to call this function themselves.\n"
		"\n"
		"This function takes an OBMol which has atoms and bonds marked as aromatic,\n"
		"aromatic bonds whose bond orders have been set to single, and aromaticity set as perceived.\n"
		"The function assumes that atoms joined by aromatic bonds have been marked as aromatic\n"
		"(if they are so intended).\n"
		"\n"
		"The purpose of the function is to set the bond orders of the aromatic bonds to either 1 or 2\n"
		"in such a way that the valencies of all of the aromatic atoms are satisfied. Failure to do\n"
		"this will result in one or more atoms having unsatisfied valences, indicated by a radical.\n"
		"Such a failure can only occur if an atom is incorrectly marked as aromatic, or is correctly\n"
		"marked as aromatic but has incorrect valence (e.g. 'n' instead of '[nH]' in SMILES).\n"
		"\n"
		":rtype: boolean\n"
		":return: Whether kekulization was successful\n"
		""},
	 { "new_OBChainsParser", _wrap_new_OBChainsParser, METH_NOARGS, NULL},
	 { "delete_OBChainsParser", _wrap_delete_OBChainsParser, METH_O, NULL},
	 { "OBChainsParser_PerceiveChains", _wrap_OBChainsParser_PerceiveChains, METH_VARARGS, "\n"
		"Perceive macromolecular (peptide and nucleotide) residues and chains\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to parse and update\n"
		":type nukeSingleResidue: boolean, optional\n"
		":param nukeSingleResidue: If only one residue is found, clear information\n"
		"    default = false  -- single residue files should still be recognized.\n"
		""},
	 { "OBChainsParser_swigregister", OBChainsParser_swigregister, METH_O, NULL},
	 { "OBChainsParser_swiginit", OBChainsParser_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomTyper", _wrap_new_OBAtomTyper, METH_VARARGS, NULL},
	 { "delete_OBAtomTyper", _wrap_delete_OBAtomTyper, METH_O, NULL},
	 { "OBAtomTyper_AssignHyb", _wrap_OBAtomTyper_AssignHyb, METH_VARARGS, "Assign atomic hybridization (1 = sp, 2 = sp2, 3 = sp3...)"},
	 { "OBAtomTyper_AssignTypes", _wrap_OBAtomTyper_AssignTypes, METH_VARARGS, "Assign external atomic types (i.e., EXTTYP lines in atomtyp.txt)"},
	 { "OBAtomTyper_swigregister", OBAtomTyper_swigregister, METH_O, NULL},
	 { "OBAtomTyper_swiginit", OBAtomTyper_swiginit, METH_VARARGS, NULL},
	 { "new_OBAromaticTyper", _wrap_new_OBAromaticTyper, METH_NOARGS, NULL},
	 { "delete_OBAromaticTyper", _wrap_delete_OBAromaticTyper, METH_O, NULL},
	 { "OBAromaticTyper_AssignAromaticFlags", _wrap_OBAromaticTyper_AssignAromaticFlags, METH_VARARGS, "Assign aromaticity flag to atoms and bonds"},
	 { "OBAromaticTyper_swigregister", OBAromaticTyper_swigregister, METH_O, NULL},
	 { "OBAromaticTyper_swiginit", OBAromaticTyper_swiginit, METH_VARARGS, NULL},
	 { "new_OBRingTyper", _wrap_new_OBRingTyper, METH_NOARGS, NULL},
	 { "delete_OBRingTyper", _wrap_delete_OBRingTyper, METH_O, NULL},
	 { "OBRingTyper_AssignTypes", _wrap_OBRingTyper_AssignTypes, METH_VARARGS, "Assign external atomic types (ringtyp.txt)"},
	 { "OBRingTyper_swigregister", OBRingTyper_swigregister, METH_O, NULL},
	 { "OBRingTyper_swiginit", OBRingTyper_swiginit, METH_VARARGS, NULL},
	 { "CharPtrLess___call__", _wrap_CharPtrLess___call__, METH_VARARGS, NULL},
	 { "new_CharPtrLess", _wrap_new_CharPtrLess, METH_NOARGS, NULL},
	 { "delete_CharPtrLess", _wrap_delete_CharPtrLess, METH_O, NULL},
	 { "CharPtrLess_swigregister", CharPtrLess_swigregister, METH_O, NULL},
	 { "CharPtrLess_swiginit", CharPtrLess_swiginit, METH_VARARGS, NULL},
	 { "delete_OBPlugin", _wrap_delete_OBPlugin, METH_O, "Virtual destructor necessary for classes with virtual functions"},
	 { "OBPlugin_Description", _wrap_OBPlugin_Description, METH_O, "Required description of a sub-type"},
	 { "OBPlugin_TypeID", _wrap_OBPlugin_TypeID, METH_O, "Redefined by each plugin type: \"formats\", \"fingerprints\", etc."},
	 { "OBPlugin_Display", _wrap_OBPlugin_Display, METH_VARARGS, "\n"
		"Write information on a plugin class to the string txt.\n"
		"Return false if not written.\n"
		"The default implementation outputs:\n"
		" the ID, a tab character, and the first line of the Description.\n"
		"The param string can be used in derived types to provide different outputs.\n"
		""},
	 { "OBPlugin_MakeInstance", _wrap_OBPlugin_MakeInstance, METH_VARARGS, "\n"
		"Make a new instance of the class.\n"
		"See OpTransform, OBGroupContrib, SmartsDescriptor classes for derived versions.\n"
		"Usually, the first parameter is the classname, the next three are\n"
		"parameters(ID, filename, description) for a constructor, and the rest data.\n"
		""},
	 { "OBPlugin_Init", _wrap_OBPlugin_Init, METH_O, "\n"
		"Initialize the plugin.\n"
		"The default version does nothing.\n"
		""},
	 { "OBPlugin_GetPlugin", _wrap_OBPlugin_GetPlugin, METH_VARARGS, "\n"
		"Get a pointer to a plugin from its type and ID. Return NULL if not found.\n"
		"If Type is NULL, search all types. Not cast to Type*\n"
		""},
	 { "OBPlugin_GetID", _wrap_OBPlugin_GetID, METH_O, "Return the ID of the sub-type instance."},
	 { "OBPlugin_ListAsVector", _wrap_OBPlugin_ListAsVector, METH_VARARGS, "\n"
		"Output a list of sub-type classes of the the type PluginID,\n"
		"or, if PluginID is \"plugins\" or empty, a list of the base types.\n"
		"If PluginID is not recognized or is NULL, the base types are output and the return is false.\n"
		""},
	 { "OBPlugin_List", _wrap_OBPlugin_List, METH_VARARGS, "As ListAsVector but sent to an ostream with a default of cout if not specified"},
	 { "OBPlugin_ListAsString", _wrap_OBPlugin_ListAsString, METH_VARARGS, "As ListAsVector but returns a string containing the list"},
	 { "OBPlugin_FirstLine", _wrap_OBPlugin_FirstLine, METH_O, "Utility function to return only the first line of a string"},
	 { "OBPlugin_Begin", _wrap_OBPlugin_Begin, METH_O, "\n"
		"Return an iterator at the start of the map of the plugin types PluginID or,\n"
		" if there is no such map, the end of the top level plugin map.\n"
		""},
	 { "OBPlugin_End", _wrap_OBPlugin_End, METH_O, NULL},
	 { "OBPlugin_GetMap", _wrap_OBPlugin_GetMap, METH_O, "Returns the map of the subtypes"},
	 { "OBPlugin_LoadAllPlugins", _wrap_OBPlugin_LoadAllPlugins, METH_NOARGS, "Load all plugins (formats, fingerprints, forcefields etc.)"},
	 { "OBPlugin_swigregister", OBPlugin_swigregister, METH_O, NULL},
	 { "new_stringbuf", _wrap_new_stringbuf, METH_NOARGS, NULL},
	 { "delete_stringbuf", _wrap_delete_stringbuf, METH_O, NULL},
	 { "stringbuf_swigregister", stringbuf_swigregister, METH_O, NULL},
	 { "stringbuf_swiginit", stringbuf_swiginit, METH_VARARGS, NULL},
	 { "new_OBError", _wrap_new_OBError, METH_VARARGS, "Constructor for an error message e.g. OBError(__FUNCTION__, \" message \")"},
	 { "OBError_message", _wrap_OBError_message, METH_O, "\n"
		":rtype: string\n"
		":return: A formatted message string, including optional explanations, etc.\n"
		""},
	 { "__lshift__", _wrap___lshift__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Prints a representation of the vector as a row vector of the form \"<0.1,1,2>\"\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Output a text representation of a matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Output to a stream\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Output a formatted message string\n"
		""},
	 { "OBError_GetMethod", _wrap_OBError_GetMethod, METH_O, "\n"
		":rtype: string\n"
		":return: The method which caused this error\n"
		"             (typically supplied via the __FUNCTION__ compiler macro *\n"
		""},
	 { "OBError_GetError", _wrap_OBError_GetError, METH_O, "\n"
		":rtype: string\n"
		":return: The main error message\n"
		""},
	 { "OBError_GetExplanation", _wrap_OBError_GetExplanation, METH_O, "\n"
		":rtype: string\n"
		":return: A more detailed explanation of the error (optional)\n"
		""},
	 { "OBError_GetPossibleCause", _wrap_OBError_GetPossibleCause, METH_O, "\n"
		":rtype: string\n"
		":return: A possible cause for the error (optional)\n"
		""},
	 { "OBError_GetSuggestedRemedy", _wrap_OBError_GetSuggestedRemedy, METH_O, "\n"
		":rtype: string\n"
		":return: The suggested workaround or remedy for the error (optional)\n"
		""},
	 { "OBError_GetLevel", _wrap_OBError_GetLevel, METH_O, "\n"
		":rtype: int\n"
		":return: The severity level of this error\n"
		""},
	 { "OBError___eq__", _wrap_OBError___eq__, METH_VARARGS, NULL},
	 { "delete_OBError", _wrap_delete_OBError, METH_O, NULL},
	 { "OBError_swigregister", OBError_swigregister, METH_O, NULL},
	 { "OBError_swiginit", OBError_swiginit, METH_VARARGS, NULL},
	 { "new_OBMessageHandler", _wrap_new_OBMessageHandler, METH_NOARGS, NULL},
	 { "delete_OBMessageHandler", _wrap_delete_OBMessageHandler, METH_O, NULL},
	 { "OBMessageHandler_ThrowError", _wrap_OBMessageHandler_ThrowError, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Throw an error with an already-formatted OBError object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		""},
	 { "OBMessageHandler_GetMessagesOfLevel", _wrap_OBMessageHandler_GetMessagesOfLevel, METH_VARARGS, "\n"
		":rtype: std::vector< std::string,std::allocator< std::string > >\n"
		":return: all messages matching a specified level\n"
		""},
	 { "OBMessageHandler_StartLogging", _wrap_OBMessageHandler_StartLogging, METH_O, "Start logging messages (default)"},
	 { "OBMessageHandler_StopLogging", _wrap_OBMessageHandler_StopLogging, METH_O, "Stop logging messages completely"},
	 { "OBMessageHandler_SetMaxLogEntries", _wrap_OBMessageHandler_SetMaxLogEntries, METH_VARARGS, "Set the maximum number of entries (or 0 for no limit)"},
	 { "OBMessageHandler_GetMaxLogEntries", _wrap_OBMessageHandler_GetMaxLogEntries, METH_O, "\n"
		":rtype: int\n"
		":return: the current maximum number of entries (default = 0 for no limit)\n"
		""},
	 { "OBMessageHandler_ClearLog", _wrap_OBMessageHandler_ClearLog, METH_O, "Clear the current message log entirely"},
	 { "OBMessageHandler_SetOutputLevel", _wrap_OBMessageHandler_SetOutputLevel, METH_VARARGS, "Set the level of messages to output (i.e., messages with at least this priority will be output)"},
	 { "OBMessageHandler_GetOutputLevel", _wrap_OBMessageHandler_GetOutputLevel, METH_O, "\n"
		":rtype: int\n"
		":return: the current output level\n"
		""},
	 { "OBMessageHandler_SetOutputStream", _wrap_OBMessageHandler_SetOutputStream, METH_VARARGS, NULL},
	 { "OBMessageHandler_GetOutputStream", _wrap_OBMessageHandler_GetOutputStream, METH_O, NULL},
	 { "OBMessageHandler_StartErrorWrap", _wrap_OBMessageHandler_StartErrorWrap, METH_O, "Start \"wrapping\" messages to cerr into ThrowError calls"},
	 { "OBMessageHandler_StopErrorWrap", _wrap_OBMessageHandler_StopErrorWrap, METH_O, "Turn off \"wrapping\" messages, restoring normal cerr use (default)"},
	 { "OBMessageHandler_GetErrorMessageCount", _wrap_OBMessageHandler_GetErrorMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obError level\n"
		""},
	 { "OBMessageHandler_GetWarningMessageCount", _wrap_OBMessageHandler_GetWarningMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obWarning level\n"
		""},
	 { "OBMessageHandler_GetInfoMessageCount", _wrap_OBMessageHandler_GetInfoMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obInfo level\n"
		""},
	 { "OBMessageHandler_GetAuditMessageCount", _wrap_OBMessageHandler_GetAuditMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obAuditMsg level\n"
		""},
	 { "OBMessageHandler_GetDebugMessageCount", _wrap_OBMessageHandler_GetDebugMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obDebug level\n"
		""},
	 { "OBMessageHandler_GetMessageSummary", _wrap_OBMessageHandler_GetMessageSummary, METH_O, "\n"
		":rtype: string\n"
		":return: Summary of messages received at all levels\n"
		""},
	 { "OBMessageHandler_swigregister", OBMessageHandler_swigregister, METH_O, NULL},
	 { "OBMessageHandler_swiginit", OBMessageHandler_swiginit, METH_VARARGS, NULL},
	 { "delete_obLogBuf", _wrap_delete_obLogBuf, METH_O, "Close the output buffer, flush, and call OBMessageHandler::ThrowError()"},
	 { "new_obLogBuf", _wrap_new_obLogBuf, METH_NOARGS, NULL},
	 { "obLogBuf_swigregister", obLogBuf_swigregister, METH_O, NULL},
	 { "obLogBuf_swiginit", obLogBuf_swiginit, METH_VARARGS, NULL},
	 { "OBFormat_Default", _wrap_OBFormat_Default, METH_NOARGS, NULL},
	 { "OBFormat_FindType", _wrap_OBFormat_FindType, METH_O, NULL},
	 { "OBFormat_ReadMolecule", _wrap_OBFormat_ReadMolecule, METH_VARARGS, "\n"
		"The \"API\" interface Read function.\n"
		"Reads a single object.\n"
		"Does not make a new object on the heap;\n"
		"can be used with a pointer to an chem object on the heap or the stack.\n"
		""},
	 { "OBFormat_ReadChemObject", _wrap_OBFormat_ReadChemObject, METH_VARARGS, "\n"
		"The \"Convert\" interface Read function.\n"
		"Possibly reads multiple new objects on the heap and subjects them\n"
		"to its DoTransformations() function, which may delete them again.\n"
		"Sends result to pConv->AddChemObject()\n"
		""},
	 { "OBFormat_WriteMolecule", _wrap_OBFormat_WriteMolecule, METH_VARARGS, "\n"
		"The \"API\" interface Write function.\n"
		"Writes a single object\n"
		"Does not delete the object;\n"
		"can be used with a pointer to an chem object on the heap or the stack.\n"
		":rtype: boolean\n"
		":return: false on error.\n"
		""},
	 { "OBFormat_WriteChemObject", _wrap_OBFormat_WriteChemObject, METH_VARARGS, "\n"
		"The \"Convert\" interface Write function.\n"
		"Writes a single object\n"
		"Deletes the object after writing\n"
		":rtype: boolean\n"
		":return: false on error\n"
		"\n"
		"Defined out-of-line in format.cpp: it needs OBConversion::GetChemObject(),\n"
		"which is only forward-declared here to avoid a header cycle with\n"
		"obconversion.h. Formats that don't override this must still consume\n"
		"and delete the pending object themselves, or it leaks (see\n"
		"OBConversion::Convert()'s final flush and AddChemObject()).\n"
		""},
	 { "OBFormat_TargetClassDescription", _wrap_OBFormat_TargetClassDescription, METH_O, "\n"
		"A decription of the chemical object converted by this format.\n"
		"If not provided, the object type used by the default format is used (usually OBMol).\n"
		""},
	 { "OBFormat_GetType", _wrap_OBFormat_GetType, METH_O, "\n"
		":rtype: std::type_info\n"
		":return: the type of chemical object used by the format.\n"
		"    Defaults to that used by the default format. Useful for checking\n"
		"    that a format can handle a particular object.\n"
		""},
	 { "OBFormat_SpecificationURL", _wrap_OBFormat_SpecificationURL, METH_O, "Web address where the format is defined."},
	 { "OBFormat_GetMIMEType", _wrap_OBFormat_GetMIMEType, METH_O, "Chemical MIME type associated with this file type (if any)"},
	 { "OBFormat_Flags", _wrap_OBFormat_Flags, METH_O, "\n"
		"Decribes the capabilities of the format (Read only etc.)\n"
		"Currently, can be a bitwise OR of any of the following\n"
		"NOTREADABLE READONEONLY NOTWRITABLE WRITEONEONLY DEFAULTFORMAT\n"
		"READBINARY WRITEBINARY READXML\n"
		""},
	 { "OBFormat_SkipObjects", _wrap_OBFormat_SkipObjects, METH_VARARGS, "\n"
		"Skip past first n objects in input stream (or current one with n=0)\n"
		":rtype: int\n"
		":return: 1 on success, -1 on error and 0 if not implemented\n"
		""},
	 { "OBFormat_MakeNewInstance", _wrap_OBFormat_MakeNewInstance, METH_O, "\n"
		":rtype: :py:class:`OBFormat`\n"
		":return: a pointer to a new instance of the format, or NULL if fails.\n"
		"    Normally a single global instance is used but this may cause problems\n"
		"    if there are member variables and the format is used in more than one place\n"
		"    in the program.\n"
		""},
	 { "OBFormat_RegisterFormat", _wrap_OBFormat_RegisterFormat, METH_VARARGS, NULL},
	 { "OBFormat_Display", _wrap_OBFormat_Display, METH_VARARGS, "\n"
		"Provides a description in txt of the format specified by itr.\n"
		"If param starts with \"in\", \"read\", \"out\" or \"write\" only the\n"
		"appropriate formats are output. The others return false.\n"
		"If param contains \"verbose\", the whole description is output.\n"
		""},
	 { "OBFormat_FormatFromMIME", _wrap_OBFormat_FormatFromMIME, METH_O, NULL},
	 { "delete_OBFormat", _wrap_delete_OBFormat, METH_O, NULL},
	 { "OBFormat_swigregister", OBFormat_swigregister, METH_O, NULL},
	 { "new_OBConversion", _wrap_new_OBConversion, METH_VARARGS, "Copy constructor.  Stream *ownership* is not copied. Source remains responsible for the memory."},
	 { "delete_OBConversion", _wrap_delete_OBConversion, METH_O, NULL},
	 { "OBConversion_RegisterFormat", _wrap_OBConversion_RegisterFormat, METH_VARARGS, "Called once by each format class"},
	 { "OBConversion_FindFormat", _wrap_OBConversion_FindFormat, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Searches registered formats\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Searches registered formats\n"
		"Since: version 2.3\n"
		""},
	 { "OBConversion_FormatFromExt", _wrap_OBConversion_FormatFromExt, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Searches registered formats for an ID the same as the file extension\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Searches registered formats for an ID the same as the file extension\n"
		"Since: version 2.3\n"
		""},
	 { "OBConversion_FormatFromMIME", _wrap_OBConversion_FormatFromMIME, METH_O, "Searches registered formats for a MIME the same as the chemical MIME type passed"},
	 { "OBConversion_Description", _wrap_OBConversion_Description, METH_NOARGS, "Deprecated!.Repeatedly called to recover available Formats"},
	 { "OBConversion_GetInStream", _wrap_OBConversion_GetInStream, METH_O, "These return a filtered stream for reading/writing (possible filters include compression, decompression, and newline transformation)"},
	 { "OBConversion_GetOutStream", _wrap_OBConversion_GetOutStream, METH_O, NULL},
	 { "OBConversion_SetInStream", _wrap_OBConversion_SetInStream, METH_VARARGS, "\n"
		"Set input stream.  If takeOwnership is true, will deallocate when done.\n"
		"If isGzipped is true, will treat as a gzipped stream regardless of option settings,\n"
		""},
	 { "OBConversion_SetOutStream", _wrap_OBConversion_SetOutStream, METH_VARARGS, NULL},
	 { "OBConversion_SetInAndOutFormats", _wrap_OBConversion_SetInAndOutFormats, METH_VARARGS, "Sets the formats from their ids, e g CML"},
	 { "OBConversion_SetInFormat", _wrap_OBConversion_SetInFormat, METH_VARARGS, "Sets the input format from an id e.g. CML"},
	 { "OBConversion_SetOutFormat", _wrap_OBConversion_SetOutFormat, METH_VARARGS, "Sets the output format from an id e.g. CML"},
	 { "OBConversion_GetInFormat", _wrap_OBConversion_GetInFormat, METH_O, NULL},
	 { "OBConversion_GetOutFormat", _wrap_OBConversion_GetOutFormat, METH_O, NULL},
	 { "OBConversion_GetInGzipped", _wrap_OBConversion_GetInGzipped, METH_O, NULL},
	 { "OBConversion_GetOutGzipped", _wrap_OBConversion_GetOutGzipped, METH_O, NULL},
	 { "OBConversion_GetInFilename", _wrap_OBConversion_GetInFilename, METH_O, NULL},
	 { "OBConversion_GetOutFilename", _wrap_OBConversion_GetOutFilename, METH_O, NULL},
	 { "OBConversion_GetInPos", _wrap_OBConversion_GetInPos, METH_O, "Get the position in the input stream of the object being read"},
	 { "OBConversion_GetInLen", _wrap_OBConversion_GetInLen, METH_O, "Get the length in the input stream of the object being read"},
	 { "OBConversion_GetTitle", _wrap_OBConversion_GetTitle, METH_O, "\n"
		":rtype: string\n"
		":return: a default title which is the filename\n"
		""},
	 { "OBConversion_GetAuxConv", _wrap_OBConversion_GetAuxConv, METH_O, "Extension method: deleted in ~OBConversion()"},
	 { "OBConversion_SetAuxConv", _wrap_OBConversion_SetAuxConv, METH_VARARGS, NULL},
	 { "OBConversion_IsOption", _wrap_OBConversion_IsOption, METH_VARARGS, "\n"
		"Determine whether an option is set. :rtype: char\n"
		":return: NULL if option not and a pointer to the associated text if it is\n"
		""},
	 { "OBConversion_GetOptions", _wrap_OBConversion_GetOptions, METH_VARARGS, "Access the map with option name as key and any associated text as value"},
	 { "OBConversion_AddOption", _wrap_OBConversion_AddOption, METH_VARARGS, "Set an option of specified type, with optional text"},
	 { "OBConversion_RemoveOption", _wrap_OBConversion_RemoveOption, METH_VARARGS, NULL},
	 { "OBConversion_SetOptions", _wrap_OBConversion_SetOptions, METH_VARARGS, "Set several single character options of specified type from string like ab\"btext\"c\"ctext\""},
	 { "OBConversion_RegisterOptionParam", _wrap_OBConversion_RegisterOptionParam, METH_VARARGS, "For example -h takes 0 parameters; -f takes 1. Call in a format constructor."},
	 { "OBConversion_GetOptionParams", _wrap_OBConversion_GetOptionParams, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the number of parameters registered for the option, or 0 if not found\n"
		""},
	 { "OBConversion_CopyOptions", _wrap_OBConversion_CopyOptions, METH_VARARGS, "Copies the options (by default of all types) from one OBConversion Object to another."},
	 { "OBConversion_GetSupportedInputFormat", _wrap_OBConversion_GetSupportedInputFormat, METH_O, NULL},
	 { "OBConversion_GetSupportedOutputFormat", _wrap_OBConversion_GetSupportedOutputFormat, METH_O, NULL},
	 { "OBConversion_Convert", _wrap_OBConversion_Convert, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Conversion for single input and output stream\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Conversion with existing streams\n"
		""},
	 { "OBConversion_FullConvert", _wrap_OBConversion_FullConvert, METH_VARARGS, "\n"
		"Conversion with multiple input/output files:\n"
		"makes input and output streams, and carries out normal, batch, aggregation, and splitting conversion.\n"
		""},
	 { "OBConversion_AddChemObject", _wrap_OBConversion_AddChemObject, METH_VARARGS, "Adds to internal array during input"},
	 { "OBConversion_GetChemObject", _wrap_OBConversion_GetChemObject, METH_O, "Retrieve from internal array during output"},
	 { "OBConversion_IsLast", _wrap_OBConversion_IsLast, METH_O, "True if no more objects to be output"},
	 { "OBConversion_IsFirstInput", _wrap_OBConversion_IsFirstInput, METH_O, "True if the first input object is being processed"},
	 { "OBConversion_SetFirstInput", _wrap_OBConversion_SetFirstInput, METH_VARARGS, "Setwhether or not is the first input"},
	 { "OBConversion_GetOutputIndex", _wrap_OBConversion_GetOutputIndex, METH_O, "Retrieves number of ChemObjects that have been actually output"},
	 { "OBConversion_SetOutputIndex", _wrap_OBConversion_SetOutputIndex, METH_VARARGS, "Sets output index (maybe to control whether seen as first object)"},
	 { "OBConversion_SetMoreFilesToCome", _wrap_OBConversion_SetMoreFilesToCome, METH_O, "Used with multiple input files. Off by default."},
	 { "OBConversion_SetOneObjectOnly", _wrap_OBConversion_SetOneObjectOnly, METH_VARARGS, "Used with multiple input files. Off by default."},
	 { "OBConversion_SetLast", _wrap_OBConversion_SetLast, METH_VARARGS, "Synonym for SetOneObjectOnly()"},
	 { "OBConversion_IsLastFile", _wrap_OBConversion_IsLastFile, METH_O, "True if no more files to be read"},
	 { "OBConversion_GetCount", _wrap_OBConversion_GetCount, METH_O, "\n"
		"Number of objects read and processed\n"
		"Incremented after options are processed, so 0 for first object.  Returns -1 if Convert interface not used.\n"
		""},
	 { "OBConversion_GetDefaultFormat", _wrap_OBConversion_GetDefaultFormat, METH_NOARGS, "The default format is set in a single OBFormat class (generally it is OBMol)"},
	 { "OBConversion_Write", _wrap_OBConversion_Write, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase.\n"
		"Part of \"API\" interface.\n"
		"The output stream can be specified and the change is retained in the OBConversion instance\n"
		""},
	 { "OBConversion_WriteString", _wrap_OBConversion_WriteString, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase as a string\n"
		"Part of \"API\" interface.\n"
		"The output stream is temporarily changed to the string and then restored\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		"The optional \"trimWhitespace\" parameter allows trailing whitespace to be removed\n"
		"(e.g., in a SMILES string or InChI, etc.)\n"
		""},
	 { "OBConversion_WriteFile", _wrap_OBConversion_WriteFile, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase as a file (with the supplied path)\n"
		"Part of \"API\" interface.\n"
		"The output stream is changed to the supplied file and the change is retained in the\n"
		"OBConversion instance.\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		""},
	 { "OBConversion_CloseOutFile", _wrap_OBConversion_CloseOutFile, METH_O, "\n"
		"Manually closes and deletes the output stream\n"
		"The file is closed anyway when in the OBConversion destructor or when WriteFile\n"
		"is called again.\n"
		"Since: version 2.1\n"
		""},
	 { "OBConversion_Read", _wrap_OBConversion_Read, METH_VARARGS, "\n"
		"Reads an object of a class derived from OBBase into pOb.\n"
		"Part of \"API\" interface.\n"
		"The input stream can be specified and the change is retained in the OBConversion instance\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		""},
	 { "OBConversion_ReadString", _wrap_OBConversion_ReadString, METH_VARARGS, "\n"
		"Part of \"API\" interface.\n"
		"The input stream can be specified and the change is retained in the OBConversion instance\n"
		":rtype: boolean\n"
		":return: NULL on error\n"
		"Reads an object of a class derived from OBBase into pOb from the supplied string\n"
		"Part of \"API\" interface.\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		"    This method is primarily intended for scripting languages without \"stream\" classes\n"
		"    Any existing input stream will be replaced by stringstream.\n"
		""},
	 { "OBConversion_ReadFile", _wrap_OBConversion_ReadFile, METH_VARARGS, "\n"
		"Reads an object of a class derived from OBBase into pOb from the file specified\n"
		"Part of \"API\" interface.\n"
		"The output stream is changed to the supplied file and the change is retained in the\n"
		"OBConversion instance. For multi-molecule files, the remaining molecules\n"
		"can be read by repeatedly calling the Read() method.\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		"    This method is primarily intended for scripting languages without \"stream\" classes\n"
		""},
	 { "OBConversion_OpenInAndOutFiles", _wrap_OBConversion_OpenInAndOutFiles, METH_VARARGS, "\n"
		"Part of the \"Convert\" interface.\n"
		"Open the files and update the streams in the OBConversion object.\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		"and will usually followed by a call to Convert().\n"
		"Will set format from file extension if format has not already been set.\n"
		"Files will be opened even if format cannot be determined, but not if file path is empty.\n"
		":rtype: boolean\n"
		":return: false if unsuccessful.\n"
		""},
	 { "OBConversion_ReportNumberConverted", _wrap_OBConversion_ReportNumberConverted, METH_VARARGS, "\n"
		"Sends a message like \"2 molecules converted\" to clog\n"
		"The type of object is taken from the TargetClassDescription\n"
		"of the specified class (or the output format if not specified)and\n"
		"is appropriately singular or plural.\n"
		""},
	 { "OBConversion_NumInputObjects", _wrap_OBConversion_NumInputObjects, METH_O, "\n"
		":rtype: int\n"
		":return: the number of objects in the inputstream,\n"
		"    or -1 if error or if SkipObjects for the input format is not implemented\n"
		"    Adjusts for the value of -f and -l options (first and last objects).\n"
		""},
	 { "OBConversion_swigregister", OBConversion_swigregister, METH_O, NULL},
	 { "OBConversion_swiginit", OBConversion_swiginit, METH_VARARGS, NULL},
	 { "OBBondGetSmallestRingSize", _wrap_OBBondGetSmallestRingSize, METH_VARARGS, "\n"
		"Return the size of the smallest ring in which a bond appears\n"
		"\n"
		"This function returns the size of the smallest ring in which a bond appears. The\n"
		"search is bounded by the specified bound. A value of 0 is returned if the bond\n"
		"is not in a ring or if no ring is found of size less than or equal to the bound.\n"
		"\n"
		"Note that alternative algorithms may be more appropriate if you wish to calculate\n"
		"this value for all atoms in a molecule.\n"
		"\n"
		":rtype: int\n"
		":return: The size of the smallest ring, or 0\n"
		""},
	 { "GetTypicalValence", _wrap_GetTypicalValence, METH_VARARGS, "\n"
		"Return the typical valence of an atom of a particular element\n"
		"\n"
		"This function returns the typical valence of an atom given its element, current\n"
		"valence (that is, the current sum of the bond orders of its bonds) and formal\n"
		"charge.\n"
		"\n"
		"This is typically used on atoms that are missing hydrogens, to decide how many\n"
		"implicit hydrogens should be assigned (should one have to guess). For example,\n"
		"the value 3 is returned for a positively charged carbon with no attached atoms.\n"
		"\n"
		":rtype: int\n"
		":return: A value for the typical valence\n"
		""},
	 { "OBAtomAssignTypicalImplicitHydrogens", _wrap_OBAtomAssignTypicalImplicitHydrogens, METH_O, "\n"
		"Assign implicit hydrogens to an OBAtom based on typical valences\n"
		"\n"
		"This function uses the return value of GetTypicalValence to determine how many\n"
		"implicit hydrogens to assign to an OBAtom. Note that most file formats describe\n"
		"exactly how many hydrogens are present, and do not require this function.\n"
		""},
	 { "GetSymbol", _wrap_GetSymbol, METH_O, "\n"
		":rtype: string\n"
		":return: the element symbol matching the atomic number passed\n"
		""},
	 { "GetName", _wrap_GetName, METH_O, "\n"
		":rtype: string\n"
		":return: the name of this element\n"
		""},
	 { "GetMass", _wrap_GetMass, METH_O, "\n"
		":rtype: float\n"
		":return: the average atomic mass for this element. For exact isotope masses, use GetExactMass()\n"
		""},
	 { "GetExactMass", _wrap_GetExactMass, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the exact mass of the specified isotope         or by default (i.e. \"isotope 0\") the mass of the most abundant isotope\n"
		""},
	 { "GetAtomicNum", _wrap_GetAtomicNum, METH_O, NULL},
	 { "GetAllredRochowElectroNeg", _wrap_GetAllredRochowElectroNeg, METH_O, "\n"
		":rtype: float\n"
		":return: the Allred-Rochow electronegativity for this element\n"
		""},
	 { "GetCovalentRad", _wrap_GetCovalentRad, METH_O, "\n"
		":rtype: float\n"
		":return: the covalent radius (in Angstrom) for this atomic number\n"
		""},
	 { "GetVdwRad", _wrap_GetVdwRad, METH_O, "\n"
		":rtype: float\n"
		":return: the van der Waals radius (in Angstrom) for this atomic number\n"
		""},
	 { "GetElectronAffinity", _wrap_GetElectronAffinity, METH_O, "\n"
		":rtype: float\n"
		":return: the electron affinity (in eV) for this element\n"
		""},
	 { "GetIonization", _wrap_GetIonization, METH_O, "\n"
		":rtype: float\n"
		":return: the ionization potential (in eV) for this element\n"
		""},
	 { "GetMaxBonds", _wrap_GetMaxBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the maximum expected number of bonds to this element\n"
		""},
	 { "GetElectroNeg", _wrap_GetElectroNeg, METH_O, "\n"
		":rtype: float\n"
		":return: the Pauling electronegativity for this element\n"
		""},
	 { "GetRGB", _wrap_GetRGB, METH_VARARGS, "Sets the red, green, and blue color values for this element"},
	 { "new_OBResidue", _wrap_new_OBResidue, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor\n"
		""},
	 { "delete_OBResidue", _wrap_delete_OBResidue, METH_O, "Destructor"},
	 { "OBResidue_AddAtom", _wrap_OBResidue_AddAtom, METH_VARARGS, "Add ``atom`` to this residue. Updates the atom via OBAtom::SetResidue()"},
	 { "OBResidue_InsertAtom", _wrap_OBResidue_InsertAtom, METH_VARARGS, "Add ``atom`` to this residue. Updates the atom via OBAtom::SetResidue()"},
	 { "OBResidue_RemoveAtom", _wrap_OBResidue_RemoveAtom, METH_VARARGS, "Remove ``atom`` from this residue and update the atom."},
	 { "OBResidue_SetName", _wrap_OBResidue_SetName, METH_VARARGS, "Set the name of this residue (e.g., \"ALA\"). Use 3-char PDB standard names. http://www.rcsb.org/pdb/file_formats/pdb/pdbguide2.2/part_79.html MODRES records for modified residues: http://www.rcsb.org/pdb/file_formats/pdb/pdbguide2.2/part_36.html"},
	 { "OBResidue_SetSegName", _wrap_OBResidue_SetSegName, METH_VARARGS, "Set the segment name of this residue (max four characters)"},
	 { "OBResidue_SetNum", _wrap_OBResidue_SetNum, METH_VARARGS, "Set the residue number (in the sequence)"},
	 { "OBResidue_SetChain", _wrap_OBResidue_SetChain, METH_VARARGS, "Set the chain ID for this residue"},
	 { "OBResidue_SetChainNum", _wrap_OBResidue_SetChainNum, METH_VARARGS, "Set the chain number for this residue"},
	 { "OBResidue_SetIdx", _wrap_OBResidue_SetIdx, METH_VARARGS, "Set the internal index of this residue in the parent OBMol. Intended mostly for internal use"},
	 { "OBResidue_SetInsertionCode", _wrap_OBResidue_SetInsertionCode, METH_VARARGS, "Set  PDB insertion code information for this residue. This allows consecutive residues to have the same number. Some communities that work in a well-conserved structural world use this, e.g. for immunoglobulins."},
	 { "OBResidue_SetAtomID", _wrap_OBResidue_SetAtomID, METH_VARARGS, "Set the character code ID for an ATOM record for the supplied atom This does nothing if the supplied atom is not found in the residue"},
	 { "OBResidue_SetHetAtom", _wrap_OBResidue_SetHetAtom, METH_VARARGS, NULL},
	 { "OBResidue_SetSerialNum", _wrap_OBResidue_SetSerialNum, METH_VARARGS, "Set the atomic serial number for a given atom (see OBSerialNums)"},
	 { "OBResidue_GetName", _wrap_OBResidue_GetName, METH_O, "\n"
		":rtype: string\n"
		":return: The residue name\n"
		""},
	 { "OBResidue_GetSegName", _wrap_OBResidue_GetSegName, METH_O, "\n"
		":rtype: string\n"
		":return: The residue segment name\n"
		""},
	 { "OBResidue_GetNum", _wrap_OBResidue_GetNum, METH_O, "\n"
		":rtype: int\n"
		":return: The residue number (in the sequence)\n"
		""},
	 { "OBResidue_GetNumString", _wrap_OBResidue_GetNumString, METH_O, NULL},
	 { "OBResidue_GetNumAtoms", _wrap_OBResidue_GetNumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in this residue\n"
		""},
	 { "OBResidue_GetNumHvyAtoms", _wrap_OBResidue_GetNumHvyAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of heavy atoms in this residue\n"
		""},
	 { "OBResidue_GetChain", _wrap_OBResidue_GetChain, METH_O, "\n"
		":rtype: char\n"
		":return: The ID of the chain which includes this residue\n"
		""},
	 { "OBResidue_GetChainNum", _wrap_OBResidue_GetChainNum, METH_O, "\n"
		":rtype: int\n"
		":return: The number of the chain which includes this residue\n"
		""},
	 { "OBResidue_GetIdx", _wrap_OBResidue_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The internal index of this residue in the parent OBMol\n"
		""},
	 { "OBResidue_GetResKey", _wrap_OBResidue_GetResKey, METH_O, "\n"
		":rtype: int\n"
		":return: The residue key (i.e., an entry in the OBResidueIndex namespace)\n"
		""},
	 { "OBResidue_GetBonds", _wrap_OBResidue_GetBonds, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >\n"
		":return: all bonds in this residue. ``exterior`` includes bonds to atoms  outside this residue (default is true)\n"
		""},
	 { "OBResidue_GetAtomID", _wrap_OBResidue_GetAtomID, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the atom ID (character code) for the supplied atom or \"\"  if the atom is not found in this residue\n"
		""},
	 { "OBResidue_GetSerialNum", _wrap_OBResidue_GetSerialNum, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the serial number of the supplied atom (uses OBSerialNums)\n"
		""},
	 { "OBResidue_GetInsertionCode", _wrap_OBResidue_GetInsertionCode, METH_O, "\n"
		":rtype: char\n"
		":return: The Insertion Code (i.e., an extra position motivated by a multiple sequence alignment against a template with defined numbers)\n"
		""},
	 { "OBResidue_GetAminoAcidProperty", _wrap_OBResidue_GetAminoAcidProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this residue has the supplied amino acid property  defined from the OBAminoAcidProperty namespace\n"
		""},
	 { "OBResidue_GetAtomProperty", _wrap_OBResidue_GetAtomProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether atom ``a`` has the supplied residue atom property  defined from the OBResidueAtomProperty namespace\n"
		""},
	 { "OBResidue_GetResidueProperty", _wrap_OBResidue_GetResidueProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this residue has the supplied property  defined from the OBResidueProperty namespace\n"
		""},
	 { "OBResidue_IsHetAtom", _wrap_OBResidue_IsHetAtom, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: If the given atom is a HETATM record\n"
		""},
	 { "OBResidue_IsResidueType", _wrap_OBResidue_IsResidueType, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: If this residue matches the supplied ``restype`` Set by SetResidueKeys()\n"
		""},
	 { "OBResidue_BeginAtoms", _wrap_OBResidue_BeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomIterator\n"
		":return: An iterator to the beginning of the atom list in this residue\n"
		""},
	 { "OBResidue_CBeginAtoms", _wrap_OBResidue_CBeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant iterator to the beginning of the atom list in this residue\n"
		""},
	 { "OBResidue_EndAtoms", _wrap_OBResidue_EndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomIterator\n"
		":return: An iterator to the end of the atom list in this residue\n"
		""},
	 { "OBResidue_CEndAtoms", _wrap_OBResidue_CEndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant iterator to the end of the atom list in this residue\n"
		""},
	 { "OBResidue_BeginAtom", _wrap_OBResidue_BeginAtom, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the atom list in this residue :rtype: :py:class:`OBAtom`\n"
		":return: The first atom (or NULL if none exist)\n"
		""},
	 { "OBResidue_NextAtom", _wrap_OBResidue_NextAtom, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBAtom`\n"
		":return: The next atom (or NULL if none exist)\n"
		""},
	 { "OBResidue_swigregister", OBResidue_swigregister, METH_O, NULL},
	 { "OBResidue_swiginit", OBResidue_swiginit, METH_VARARGS, NULL},
	 { "OBInternalCoord__a_set", _wrap_OBInternalCoord__a_set, METH_VARARGS, "First connection for this atom (i.e., distance)"},
	 { "OBInternalCoord__a_get", _wrap_OBInternalCoord__a_get, METH_O, "First connection for this atom (i.e., distance)"},
	 { "OBInternalCoord__b_set", _wrap_OBInternalCoord__b_set, METH_VARARGS, "Second reference atom (i.e., angle)"},
	 { "OBInternalCoord__b_get", _wrap_OBInternalCoord__b_get, METH_O, "Second reference atom (i.e., angle)"},
	 { "OBInternalCoord__c_set", _wrap_OBInternalCoord__c_set, METH_VARARGS, "Third reference atom (i.e., dihedral / torsion angle)"},
	 { "OBInternalCoord__c_get", _wrap_OBInternalCoord__c_get, METH_O, "Third reference atom (i.e., dihedral / torsion angle)"},
	 { "OBInternalCoord__dst_set", _wrap_OBInternalCoord__dst_set, METH_VARARGS, "Distance between this atom and _a"},
	 { "OBInternalCoord__dst_get", _wrap_OBInternalCoord__dst_get, METH_O, "Distance between this atom and _a"},
	 { "OBInternalCoord__ang_set", _wrap_OBInternalCoord__ang_set, METH_VARARGS, "Angle between this, _a, and _b (i.e., _a is the vertex)"},
	 { "OBInternalCoord__ang_get", _wrap_OBInternalCoord__ang_get, METH_O, "Angle between this, _a, and _b (i.e., _a is the vertex)"},
	 { "OBInternalCoord__tor_set", _wrap_OBInternalCoord__tor_set, METH_VARARGS, "Torsional/dihedral angle between this, _a, _b, and _c"},
	 { "OBInternalCoord__tor_get", _wrap_OBInternalCoord__tor_get, METH_O, "Torsional/dihedral angle between this, _a, _b, and _c"},
	 { "new_OBInternalCoord", _wrap_new_OBInternalCoord, METH_VARARGS, "Constructor"},
	 { "delete_OBInternalCoord", _wrap_delete_OBInternalCoord, METH_O, NULL},
	 { "OBInternalCoord_swigregister", OBInternalCoord_swigregister, METH_O, NULL},
	 { "OBInternalCoord_swiginit", OBInternalCoord_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomAtomIterAdaptor", _wrap_new_OBAtomAtomIterAdaptor, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___ref__", _wrap_OBAtomAtomIterAdaptor___ref__, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_inc", _wrap_OBAtomAtomIterAdaptor_inc, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___eq__", _wrap_OBAtomAtomIterAdaptor___eq__, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___ne__", _wrap_OBAtomAtomIterAdaptor___ne__, METH_VARARGS, NULL},
	 { "delete_OBAtomAtomIterAdaptor", _wrap_delete_OBAtomAtomIterAdaptor, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_swigregister", OBAtomAtomIterAdaptor_swigregister, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_swiginit", OBAtomAtomIterAdaptor_swiginit, METH_VARARGS, NULL},
	 { "OBAtom_Visit_set", _wrap_OBAtom_Visit_set, METH_VARARGS, "Used internally by graph traversal algorithms"},
	 { "OBAtom_Visit_get", _wrap_OBAtom_Visit_get, METH_O, "Used internally by graph traversal algorithms"},
	 { "new_OBAtom", _wrap_new_OBAtom, METH_NOARGS, "Constructor"},
	 { "delete_OBAtom", _wrap_delete_OBAtom, METH_O, "Destructor"},
	 { "OBAtom___eq__", _wrap_OBAtom___eq__, METH_VARARGS, "Equivalence"},
	 { "OBAtom_Duplicate", _wrap_OBAtom_Duplicate, METH_VARARGS, "Duplicate another atom. Copies all information with the exception of index Since: version 2.2"},
	 { "OBAtom_SetIdx", _wrap_OBAtom_SetIdx, METH_VARARGS, "Set atom index (i.e., in an OBMol)"},
	 { "OBAtom_SetId", _wrap_OBAtom_SetId, METH_VARARGS, NULL},
	 { "OBAtom_SetHyb", _wrap_OBAtom_SetHyb, METH_VARARGS, "Set atom hybridization (i.e., 1 = sp, 2 = sp2, 3 = sp3 ...)"},
	 { "OBAtom_SetAtomicNum", _wrap_OBAtom_SetAtomicNum, METH_VARARGS, "Set atomic number"},
	 { "OBAtom_SetIsotope", _wrap_OBAtom_SetIsotope, METH_VARARGS, "Set isotope number (actual atomic weight is tabulated automatically, 0 = most abundant)"},
	 { "OBAtom_SetImplicitHCount", _wrap_OBAtom_SetImplicitHCount, METH_VARARGS, "Set the implicit hydrogen count to ``val``"},
	 { "OBAtom_SetFormalCharge", _wrap_OBAtom_SetFormalCharge, METH_VARARGS, "Set the formal charge of the atom to ``fcharge``"},
	 { "OBAtom_SetSpinMultiplicity", _wrap_OBAtom_SetSpinMultiplicity, METH_VARARGS, "Set the atomic spin to ``spin``. See _spinmultiplicity"},
	 { "OBAtom_SetType", _wrap_OBAtom_SetType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the atomic type symbol (see OBTypeTable and OBAtomTyper for more)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the atomic type symbol (see OBTypeTable and OBAtomTyper for more)\n"
		""},
	 { "OBAtom_SetPartialCharge", _wrap_OBAtom_SetPartialCharge, METH_VARARGS, "Set the partial charge to ``pcharge``"},
	 { "OBAtom_SetCoordPtr", _wrap_OBAtom_SetCoordPtr, METH_VARARGS, "Set the position of this atom from a pointer-driven array of coordinates"},
	 { "OBAtom_SetVector", _wrap_OBAtom_SetVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the coordinate vector for this atom to ``v`` as a vector3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the coordinate vector for this atom based on ``x`` ``y`` & ``z``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Set the position of this atom based on the internal pointer array (i.e. from SetCoordPtr() )\n"
		""},
	 { "OBAtom_SetResidue", _wrap_OBAtom_SetResidue, METH_VARARGS, "Attach an OBResidue ``res`` as containing this atom"},
	 { "OBAtom_SetParent", _wrap_OBAtom_SetParent, METH_VARARGS, "Attach an OBMol ``ptr`` as the parent container for this atom"},
	 { "OBAtom_SetAromatic", _wrap_OBAtom_SetAromatic, METH_VARARGS, "Mark atom as being aromatic"},
	 { "OBAtom_SetInRing", _wrap_OBAtom_SetInRing, METH_VARARGS, "Mark an atom as belonging to at least one ring"},
	 { "OBAtom_ClearCoordPtr", _wrap_OBAtom_ClearCoordPtr, METH_O, "Clear the internal coordinate pointer"},
	 { "OBAtom_GetFormalCharge", _wrap_OBAtom_GetFormalCharge, METH_O, "\n"
		":rtype: int\n"
		":return: the formal charge for this atom\n"
		""},
	 { "OBAtom_GetAtomicNum", _wrap_OBAtom_GetAtomicNum, METH_O, "\n"
		":rtype: int\n"
		":return: the atomic number for this atom\n"
		""},
	 { "OBAtom_GetIsotope", _wrap_OBAtom_GetIsotope, METH_O, "\n"
		":rtype: int\n"
		":return: the isotope for this atom, if specified, or 0 for unspecified\n"
		""},
	 { "OBAtom_GetSpinMultiplicity", _wrap_OBAtom_GetSpinMultiplicity, METH_O, "\n"
		":rtype: int\n"
		":return: the atomic spin, e.g., 0 (default) for normal atoms - note that this value is a convention,   2 for radical  1 or 3 for carbene\n"
		""},
	 { "OBAtom_GetAtomicMass", _wrap_OBAtom_GetAtomicMass, METH_O, "\n"
		":rtype: float\n"
		":return: the atomic mass of this atom given by standard IUPAC  average molar mass\n"
		""},
	 { "OBAtom_GetExactMass", _wrap_OBAtom_GetExactMass, METH_O, "\n"
		":rtype: float\n"
		":return: the atomic mass of this atom given by the isotope (default of 0 gives the most abundant isotope)\n"
		""},
	 { "OBAtom_GetIdx", _wrap_OBAtom_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the internal atom index (e.g., inside an OBMol)\n"
		""},
	 { "OBAtom_GetIndex", _wrap_OBAtom_GetIndex, METH_O, NULL},
	 { "OBAtom_GetId", _wrap_OBAtom_GetId, METH_O, NULL},
	 { "OBAtom_GetCoordinateIdx", _wrap_OBAtom_GetCoordinateIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the index into a pointer-driven array as used by   GetCoordPtr() or SetCoordPtr()\n"
		""},
	 { "OBAtom_GetExplicitDegree", _wrap_OBAtom_GetExplicitDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of explicit bonds to this atom\n"
		""},
	 { "OBAtom_GetTotalDegree", _wrap_OBAtom_GetTotalDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The total number of bonds to this atom including bonds to implicit hydrogens\n"
		""},
	 { "OBAtom_GetExplicitValence", _wrap_OBAtom_GetExplicitValence, METH_O, "\n"
		":rtype: int\n"
		":return: The sum of the bond orders of the explicit bonds to this atom\n"
		""},
	 { "OBAtom_GetTotalValence", _wrap_OBAtom_GetTotalValence, METH_O, "\n"
		":rtype: int\n"
		":return: The sum of the bond orders of all bonds to this atom including bonds to implicit hydrogens\n"
		""},
	 { "OBAtom_GetHyb", _wrap_OBAtom_GetHyb, METH_O, "\n"
		":rtype: int\n"
		":return: The hybridization of this atom: 1 for sp, 2 for sp2, 3 for sp3, 4 for sq. planar, 5 for trig. bipy, 6 for octahedral\n"
		""},
	 { "OBAtom_GetImplicitHCount", _wrap_OBAtom_GetImplicitHCount, METH_O, "\n"
		":rtype: unsigned char\n"
		":return: The number of implicit hydrogens attached to this atom\n"
		""},
	 { "OBAtom_GetHvyDegree", _wrap_OBAtom_GetHvyDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of non-hydrogens connected to this atom\n"
		""},
	 { "OBAtom_GetHeteroDegree", _wrap_OBAtom_GetHeteroDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of heteroatoms connected to an atom\n"
		""},
	 { "OBAtom_GetType", _wrap_OBAtom_GetType, METH_O, "\n"
		":rtype: string\n"
		":return: the atomic type (e.g., for molecular mechanics)\n"
		""},
	 { "OBAtom_GetX", _wrap_OBAtom_GetX, METH_O, "\n"
		":rtype: float\n"
		":return: the x coordinate\n"
		""},
	 { "OBAtom_GetY", _wrap_OBAtom_GetY, METH_O, "\n"
		":rtype: float\n"
		":return: the y coordinate\n"
		""},
	 { "OBAtom_GetZ", _wrap_OBAtom_GetZ, METH_O, "\n"
		":rtype: float\n"
		":return: the z coordinate\n"
		""},
	 { "OBAtom_x", _wrap_OBAtom_x, METH_O, "\n"
		":rtype: float\n"
		":return: the x coordinate\n"
		""},
	 { "OBAtom_y", _wrap_OBAtom_y, METH_O, "\n"
		":rtype: float\n"
		":return: the y coordinate\n"
		""},
	 { "OBAtom_z", _wrap_OBAtom_z, METH_O, "\n"
		":rtype: float\n"
		":return: the z coordinate\n"
		""},
	 { "OBAtom_GetCoordinate", _wrap_OBAtom_GetCoordinate, METH_O, "\n"
		":rtype: float\n"
		":return: the coordinates as a double* or NULL if none. See SetCoordPtr() for more. If no coordinate pointer is used (e.g., only vector3), NULL will be returned.\n"
		""},
	 { "OBAtom_GetVector", _wrap_OBAtom_GetVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the coordinates as a vector3 object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the coordinates as a vector3 object\n"
		""},
	 { "OBAtom_GetPartialCharge", _wrap_OBAtom_GetPartialCharge, METH_O, "\n"
		":rtype: float\n"
		":return: the partial charge of this atom, calculating a Gasteiger charge if needed\n"
		""},
	 { "OBAtom_GetResidue", _wrap_OBAtom_GetResidue, METH_O, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: the residue which contains this atom, or NULL if none exists\n"
		""},
	 { "OBAtom_GetParent", _wrap_OBAtom_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: the molecule which contains this atom, or NULL if none exists\n"
		""},
	 { "OBAtom_GetNewBondVector", _wrap_OBAtom_GetNewBondVector, METH_VARARGS, "\n"
		" Create a vector for a new bond from this atom, with length given by the supplied parameter :rtype: boolean\n"
		":return: success or failure\n"
		""},
	 { "OBAtom_GetBond", _wrap_OBAtom_GetBond, METH_VARARGS, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the OBBond object between this atom and that supplied, or NULL if the two atoms are not bonded\n"
		""},
	 { "OBAtom_BeginBonds", _wrap_OBAtom_BeginBonds, METH_O, "\n"
		":rtype: OpenBabel::OBBondIterator\n"
		":return: An iterator to the beginning of the bonds to this atom\n"
		""},
	 { "OBAtom_EndBonds", _wrap_OBAtom_EndBonds, METH_O, "\n"
		":rtype: OpenBabel::OBBondIterator\n"
		":return: An iterator to the end of the bonds to this atom\n"
		""},
	 { "OBAtom_GetBonds", _wrap_OBAtom_GetBonds, METH_O, "\n"
		":rtype: OpenBabel::OBAtomBondRange\n"
		":return: A range over the bonds to this atom. This range can be used in a range-based for loop.\n"
		""},
	 { "OBAtom_BeginBond", _wrap_OBAtom_BeginBond, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the bonds :rtype: :py:class:`OBBond`\n"
		":return: The first bond to this atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_NextBond", _wrap_OBAtom_NextBond, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBBond`\n"
		":return: The next bond to this atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_GetNbrs", _wrap_OBAtom_GetNbrs, METH_O, "\n"
		":rtype: OpenBabel::OBAtomAtomRange\n"
		":return: A range over the neighbors of this atom. This range can be used in a range-based for loop.\n"
		""},
	 { "OBAtom_BeginNbrAtom", _wrap_OBAtom_BeginNbrAtom, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the bonds :rtype: :py:class:`OBAtom`\n"
		":return: The first neighboring atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_NextNbrAtom", _wrap_OBAtom_NextNbrAtom, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBAtom`\n"
		":return: The next neighboring atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_GetDistance", _wrap_OBAtom_GetDistance, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the distance to the atom defined by OBMol::GetAtom()\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the distance to the supplied OBAtom\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: float\n"
		":return: the distance to the coordinates of the supplied vector3 Since: version 2.4\n"
		""},
	 { "OBAtom_GetAngle", _wrap_OBAtom_GetAngle, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the angle defined by this atom -> b (vertex) -> c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the angle defined by this atom -> b (vertex) -> c\n"
		""},
	 { "OBAtom_NewResidue", _wrap_OBAtom_NewResidue, METH_O, "If no residue has been set for this atom, create a new one"},
	 { "OBAtom_AddResidue", _wrap_OBAtom_AddResidue, METH_VARARGS, "Add (set) the residue for this atom"},
	 { "OBAtom_DeleteResidue", _wrap_OBAtom_DeleteResidue, METH_O, "Delete any residue associated with this atom"},
	 { "OBAtom_AddBond", _wrap_OBAtom_AddBond, METH_VARARGS, "Add a bond to the internal list. Does not update the bond."},
	 { "OBAtom_InsertBond", _wrap_OBAtom_InsertBond, METH_VARARGS, "Insert ``bond`` into the internal list at the position from ``i`` Does not modify the bond"},
	 { "OBAtom_DeleteBond", _wrap_OBAtom_DeleteBond, METH_VARARGS, "Find ``bond`` and remove it from the internal list. Does not update the bond."},
	 { "OBAtom_ClearBond", _wrap_OBAtom_ClearBond, METH_O, "Clear all bonding information in this atom (does not delete them)"},
	 { "OBAtom_HtoMethyl", _wrap_OBAtom_HtoMethyl, METH_O, "\n"
		" If this is a hydrogen atom, transform into a methyl group :rtype: boolean\n"
		":return: success or failure\n"
		""},
	 { "OBAtom_SetHybAndGeom", _wrap_OBAtom_SetHybAndGeom, METH_VARARGS, "\n"
		" Change the hybridization of this atom and modify the geometry accordingly :rtype: boolean\n"
		":return: success or failure Deprecated: This will be removed in future versions of Open Babel\n"
		""},
	 { "OBAtom_CountFreeOxygens", _wrap_OBAtom_CountFreeOxygens, METH_O, "\n"
		":rtype: int\n"
		":return: The number of oxygen atoms connected that only have one heavy valence\n"
		""},
	 { "OBAtom_CountFreeSulfurs", _wrap_OBAtom_CountFreeSulfurs, METH_O, "\n"
		":rtype: int\n"
		":return: The number of sulfur atoms connected that only have one heavy valence Since: version 2.4\n"
		""},
	 { "OBAtom_ExplicitHydrogenCount", _wrap_OBAtom_ExplicitHydrogenCount, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The number of hydrogens explicitly bound to this atom, optionally excluding D,T and isotope explicitly set to 1\n"
		""},
	 { "OBAtom_MemberOfRingCount", _wrap_OBAtom_MemberOfRingCount, METH_O, "\n"
		":rtype: int\n"
		":return: The number of rings that contain this atom\n"
		""},
	 { "OBAtom_MemberOfRingSize", _wrap_OBAtom_MemberOfRingSize, METH_O, "\n"
		":rtype: int\n"
		":return: The size of the smallest ring that contains this atom (0 if not in a ring)\n"
		""},
	 { "OBAtom_CountRingBonds", _wrap_OBAtom_CountRingBonds, METH_O, "\n"
		":rtype: int\n"
		":return: The number of explicit ring connections to this atom\n"
		""},
	 { "OBAtom_SmallestBondAngle", _wrap_OBAtom_SmallestBondAngle, METH_O, "\n"
		":rtype: float\n"
		":return: The smallest angle of bonds to this atom\n"
		""},
	 { "OBAtom_AverageBondAngle", _wrap_OBAtom_AverageBondAngle, METH_O, "\n"
		":rtype: float\n"
		":return: The average angle of bonds to this atom\n"
		""},
	 { "OBAtom_LewisAcidBaseCounts", _wrap_OBAtom_LewisAcidBaseCounts, METH_O, "\n"
		"Lewis acid/base vacancies for this atom\n"
		":rtype: std::pair< int,int >\n"
		":return: A pair of integers, where first is acid count and second is base count\n"
		"Since: version 2.3\n"
		""},
	 { "OBAtom_HasResidue", _wrap_OBAtom_HasResidue, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is there any residue information?\n"
		""},
	 { "OBAtom_IsHetAtom", _wrap_OBAtom_IsHetAtom, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this a HETATM in a residue (returns false if not in a residue) Since: version 2.4\n"
		""},
	 { "OBAtom_IsElement", _wrap_OBAtom_IsElement, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the specified element, as specified by atom number (see OBElement namespace)?\n"
		""},
	 { "OBAtom_IsAromatic", _wrap_OBAtom_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom aromatic?\n"
		""},
	 { "OBAtom_IsInRing", _wrap_OBAtom_IsInRing, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom in a ring?\n"
		""},
	 { "OBAtom_IsInRingSize", _wrap_OBAtom_IsInRingSize, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the atom in a ring of a given size?\n"
		""},
	 { "OBAtom_IsHeteroatom", _wrap_OBAtom_IsHeteroatom, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an element in the 15th or 16th main groups  (i.e., N, O, P, S ...) ?\n"
		""},
	 { "OBAtom_IsConnected", _wrap_OBAtom_IsConnected, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom directly connected to the supplied OBAtom?\n"
		""},
	 { "OBAtom_IsOneThree", _wrap_OBAtom_IsOneThree, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom related to the supplied OBAtom in  a 1,3 bonding pattern?\n"
		""},
	 { "OBAtom_IsOneFour", _wrap_OBAtom_IsOneFour, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom related to the supplied OBAtom in  a 1,4 bonding pattern?\n"
		""},
	 { "OBAtom_IsCarboxylOxygen", _wrap_OBAtom_IsCarboxylOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a carboxyl (-CO2 or CO2H) group?\n"
		""},
	 { "OBAtom_IsPhosphateOxygen", _wrap_OBAtom_IsPhosphateOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a phosphate (R-PO3) group?\n"
		""},
	 { "OBAtom_IsSulfateOxygen", _wrap_OBAtom_IsSulfateOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a sulfate (-SO3) group?\n"
		""},
	 { "OBAtom_IsNitroOxygen", _wrap_OBAtom_IsNitroOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a nitro (-NO2) group?\n"
		""},
	 { "OBAtom_IsAmideNitrogen", _wrap_OBAtom_IsAmideNitrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a nitrogen in an amide (-C(=O)NR2) group?\n"
		""},
	 { "OBAtom_IsPolarHydrogen", _wrap_OBAtom_IsPolarHydrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen connected to a polar atom  (i.e., N, O, P, S)\n"
		""},
	 { "OBAtom_IsNonPolarHydrogen", _wrap_OBAtom_IsNonPolarHydrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen connected to a non-polar atom  (i.e., C)\n"
		""},
	 { "OBAtom_IsAromaticNOxide", _wrap_OBAtom_IsAromaticNOxide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an aromatic nitrogen with at least one  double bond to an oxygen atom\n"
		""},
	 { "OBAtom_IsChiral", _wrap_OBAtom_IsChiral, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom chiral?\n"
		""},
	 { "OBAtom_IsPeriodic", _wrap_OBAtom_IsPeriodic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom part of a periodic unit cell?\n"
		""},
	 { "OBAtom_IsAxial", _wrap_OBAtom_IsAxial, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an axial atom in a ring\n"
		""},
	 { "OBAtom_IsHbondAcceptor", _wrap_OBAtom_IsHbondAcceptor, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond acceptor  (considering also atom surrounding)\n"
		""},
	 { "OBAtom_IsHbondAcceptorSimple", _wrap_OBAtom_IsHbondAcceptorSimple, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond acceptor (old function)?\n"
		""},
	 { "OBAtom_IsHbondDonor", _wrap_OBAtom_IsHbondDonor, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond donor?\n"
		""},
	 { "OBAtom_IsHbondDonorH", _wrap_OBAtom_IsHbondDonorH, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this a hydrogen atom attached to a hydrogen-bond donor?\n"
		""},
	 { "OBAtom_IsMetal", _wrap_OBAtom_IsMetal, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a metal? Since: version 2.4\n"
		""},
	 { "OBAtom_HasAlphaBetaUnsat", _wrap_OBAtom_HasAlphaBetaUnsat, METH_VARARGS, "\n"
		" :rtype: boolean\n"
		" :return: Whether a neighboring atom (alpha) has an unsaturated bond   to a third atom (beta). :type includePandS: boolean, optional\n"
		":param includePandS: Whether to include phosphorus and sulfur neighbors in this determination (or to exclude them)\n"
		""},
	 { "OBAtom_HasBondOfOrder", _wrap_OBAtom_HasBondOfOrder, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this atom is connected to any bond with order == ``bo``\n"
		""},
	 { "OBAtom_CountBondsOfOrder", _wrap_OBAtom_CountBondsOfOrder, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The count of bonds connected to this atom with order == ``bo``\n"
		""},
	 { "OBAtom_HighestBondOrder", _wrap_OBAtom_HighestBondOrder, METH_O, "\n"
		":rtype: int\n"
		":return: The maximum bond order for this atom\n"
		""},
	 { "OBAtom_HasNonSingleBond", _wrap_OBAtom_HasNonSingleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this atom is connected to any bond with order >1\n"
		""},
	 { "OBAtom_HasSingleBond", _wrap_OBAtom_HasSingleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have a single bond\n"
		""},
	 { "OBAtom_HasDoubleBond", _wrap_OBAtom_HasDoubleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have a double bond\n"
		""},
	 { "OBAtom_HasAromaticBond", _wrap_OBAtom_HasAromaticBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have an aromatic bond\n"
		""},
	 { "OBAtom_MatchesSMARTS", _wrap_OBAtom_MatchesSMARTS, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this atom matches the first atom in a given SMARTS pattern\n"
		""},
	 { "OBAtom_swigregister", OBAtom_swigregister, METH_O, NULL},
	 { "OBAtom_swiginit", OBAtom_swiginit, METH_VARARGS, NULL},
	 { "OBBond_Visit_set", _wrap_OBBond_Visit_set, METH_VARARGS, "Whether this bond has been visited by a graph algorithm Deprecated: Use OBBitVec objects instead to be fully thread-safe. *"},
	 { "OBBond_Visit_get", _wrap_OBBond_Visit_get, METH_O, "Whether this bond has been visited by a graph algorithm Deprecated: Use OBBitVec objects instead to be fully thread-safe. *"},
	 { "new_OBBond", _wrap_new_OBBond, METH_NOARGS, "Constructor"},
	 { "delete_OBBond", _wrap_delete_OBBond, METH_O, "Destructor"},
	 { "OBBond_SetIdx", _wrap_OBBond_SetIdx, METH_VARARGS, "\n"
		"Set the internal bond index Warning: This will not update the index in the parent OBMol.\n"
		"         Intended mainly for internal use. Use with care. *\n"
		""},
	 { "OBBond_SetId", _wrap_OBBond_SetId, METH_VARARGS, NULL},
	 { "OBBond_SetBondOrder", _wrap_OBBond_SetBondOrder, METH_VARARGS, "Set the bond order to ``order`` (i.e., 1 = single, 2 = double, 5 = aromatic)"},
	 { "OBBond_SetBegin", _wrap_OBBond_SetBegin, METH_VARARGS, "Set the beginning atom of this bond to ``begin``. Does not update ``begin``."},
	 { "OBBond_SetEnd", _wrap_OBBond_SetEnd, METH_VARARGS, "Set the ending atom of this bond to ``end``. Does not update ``end``."},
	 { "OBBond_SetParent", _wrap_OBBond_SetParent, METH_VARARGS, "Set the parent molecule to ``ptr``. Does not update parent."},
	 { "OBBond_SetLength", _wrap_OBBond_SetLength, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Change the bond length to ``length``, while keeping ``fixed`` stationary\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Change the bond length to ``length``, moving both atoms halfway Since: version 2.2\n"
		""},
	 { "OBBond_Set", _wrap_OBBond_Set, METH_VARARGS, "Set the main bond information (i.e., when creating a bond)"},
	 { "OBBond_SetAromatic", _wrap_OBBond_SetAromatic, METH_VARARGS, "Mark that this bond is aromatic. Does not update atoms or validate."},
	 { "OBBond_SetWedge", _wrap_OBBond_SetWedge, METH_VARARGS, "\n"
		"Mark that this bond has 2D \"wedge\" notation (i.e., goes in a positive\n"
		"Z direction from the beginning to end atoms)\n"
		""},
	 { "OBBond_SetHash", _wrap_OBBond_SetHash, METH_VARARGS, "\n"
		"Mark that this bond has 2D \"hash\" notation (i.e., goes in a negative\n"
		"Z direction from the beginning to end atoms)\n"
		""},
	 { "OBBond_SetWedgeOrHash", _wrap_OBBond_SetWedgeOrHash, METH_VARARGS, "Set the WedgeOrHash flag on a bond (??)"},
	 { "OBBond_SetInRing", _wrap_OBBond_SetInRing, METH_VARARGS, "Mark that this bond is in a ring. Primarily for internal use."},
	 { "OBBond_SetClosure", _wrap_OBBond_SetClosure, METH_VARARGS, "\n"
		"Mark that this bond indicates a ring closure when walking the molecule Warning: This is for internal use only. All closure bonds are marked\n"
		"         automatically by lazy evaluation when requesting\n"
		"         OBBond::IsClosure() *\n"
		""},
	 { "OBBond_GetIdx", _wrap_OBBond_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The unique bond index in a molecule.\n"
		""},
	 { "OBBond_GetId", _wrap_OBBond_GetId, METH_O, NULL},
	 { "OBBond_GetBondOrder", _wrap_OBBond_GetBondOrder, METH_O, "\n"
		":rtype: int\n"
		":return: The bond order for the bond\n"
		""},
	 { "OBBond_GetFlags", _wrap_OBBond_GetFlags, METH_O, "\n"
		":rtype: int\n"
		":return: The set of property flags defined for this bond.\n"
		""},
	 { "OBBond_GetBeginAtomIdx", _wrap_OBBond_GetBeginAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The atom index for the end atom in this bond (from OBAtom::GetIdx()\n"
		""},
	 { "OBBond_GetEndAtomIdx", _wrap_OBBond_GetEndAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The atom index for the end atom in this bond (from OBAtom::GetIdx()\n"
		""},
	 { "OBBond_GetBeginAtom", _wrap_OBBond_GetBeginAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The \"beginning\" atom for this bond\n"
		""},
	 { "OBBond_GetEndAtom", _wrap_OBBond_GetEndAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The \"end\" atom for this bond\n"
		""},
	 { "OBBond_GetNbrAtom", _wrap_OBBond_GetNbrAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The neighboring atom to ``ptr`` (i.e., the end if ``ptr`` is the start) Warning: If ``ptr`` is not part of the bond, the beginning atom\n"
		"         will always be returned *\n"
		""},
	 { "OBBond_GetParent", _wrap_OBBond_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: The enclosing OBMol for this bond, or NULL if none is defined.\n"
		""},
	 { "OBBond_GetEquibLength", _wrap_OBBond_GetEquibLength, METH_O, "\n"
		":rtype: float\n"
		":return: The expected \"equilibrium\" length based on the covalent radii and bond order Length is given in Angstroms *\n"
		""},
	 { "OBBond_GetLength", _wrap_OBBond_GetLength, METH_O, "\n"
		":rtype: float\n"
		":return: The current length of this bond in Angstroms\n"
		""},
	 { "OBBond_GetNbrAtomIdx", _wrap_OBBond_GetNbrAtomIdx, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The index to the neighboring atom of ``ptr`` (i.e., the end if ``ptr`` is the start) Warning: If ``ptr`` is not part of the bond, the beginning atom\n"
		"         index will always be returned *\n"
		""},
	 { "OBBond_FindSmallestRing", _wrap_OBBond_FindSmallestRing, METH_O, "Find the smallest ring containing this bond (returns a NULL pointer if none exists)"},
	 { "OBBond_IsAromatic", _wrap_OBBond_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond aromatic?  (Note that the two atoms of the bond may be aromatic,   but not the bond)\n"
		""},
	 { "OBBond_IsInRing", _wrap_OBBond_IsInRing, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond part of a ring?\n"
		""},
	 { "OBBond_IsRotor", _wrap_OBBond_IsRotor, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the bond a rotatable bond?  Currently, this function classifies any bond with at least one heavy\n"
		"              atom, no sp-hybrid atoms (e.g., a triple bond somewhere) not in a ring\n"
		"              as a potential rotor if includeRingsBonds is false.  If true, rotors in\n"
		"              rings with more than 3 atoms may be included. No other bond typing is attempted.\n"
		"              For more detailed rotor detection, check the OBRotorList and\n"
		"              OBRotorRules classes *\n"
		""},
	 { "OBBond_IsPeriodic", _wrap_OBBond_IsPeriodic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond an amide link (i.e., between a carbonyl C and a N)?\n"
		"              No distinction is made between primary, secondary, and tertiary amides. *\n"
		""},
	 { "OBBond_IsAmide", _wrap_OBBond_IsAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond within a periodic unit cell?\n"
		""},
	 { "OBBond_IsPrimaryAmide", _wrap_OBBond_IsPrimaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a primary amide (i.e., between carbonyl C and a NH2)?\n"
		"              In versions prior to 2.3, this function incorrectly identified secondary amides. *\n"
		""},
	 { "OBBond_IsSecondaryAmide", _wrap_OBBond_IsSecondaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a secondary amide (i.e., between a carbonyl C and a NH1)?\n"
		"              In versions prior to 2.3, this function incorrectly identified tertiary amides. *\n"
		""},
	 { "OBBond_IsTertiaryAmide", _wrap_OBBond_IsTertiaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a teriary amide (i.e., between a carbonyl C and a NH0)?  Since: version 2.3.\n"
		""},
	 { "OBBond_IsEster", _wrap_OBBond_IsEster, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond an ester link (i.e., between a carbonyl C and an O)?\n"
		""},
	 { "OBBond_IsCarbonyl", _wrap_OBBond_IsCarbonyl, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a carbonyl C=O?\n"
		""},
	 { "OBBond_IsClosure", _wrap_OBBond_IsClosure, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this bond \"close\" a ring when walking the molecular graph?\n"
		""},
	 { "OBBond_IsWedge", _wrap_OBBond_IsWedge, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this bond is a \"wedge\" in 2D representations\n"
		"             (i.e., goes in a positive Z direction from the beginning to end atoms) *\n"
		""},
	 { "OBBond_IsHash", _wrap_OBBond_IsHash, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this bond is a \"hash\" in 2D representations\n"
		"             (i.e., goes in a negative Z direction from the beginning to end atoms) *\n"
		""},
	 { "OBBond_IsWedgeOrHash", _wrap_OBBond_IsWedgeOrHash, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this bond is either a wedge or hash.\n"
		"Notes: : This is a separate bond type\n"
		"Since: version 2.3\n"
		""},
	 { "OBBond_IsCisOrTrans", _wrap_OBBond_IsCisOrTrans, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this bond is either a cis or trans.\n"
		"Since: version 2.3\n"
		""},
	 { "OBBond_IsDoubleBondGeometry", _wrap_OBBond_IsDoubleBondGeometry, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether the geometry around this bond \"looks\" unsaturated\n"
		""},
	 { "OBBond_swigregister", OBBond_swigregister, METH_O, NULL},
	 { "OBBond_swiginit", OBBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBReaction", _wrap_new_OBReaction, METH_NOARGS, NULL},
	 { "OBReaction_NumReactants", _wrap_OBReaction_NumReactants, METH_O, NULL},
	 { "OBReaction_NumProducts", _wrap_OBReaction_NumProducts, METH_O, NULL},
	 { "OBReaction_NumAgents", _wrap_OBReaction_NumAgents, METH_O, NULL},
	 { "OBReaction_AddReactant", _wrap_OBReaction_AddReactant, METH_VARARGS, NULL},
	 { "OBReaction_AddProduct", _wrap_OBReaction_AddProduct, METH_VARARGS, NULL},
	 { "OBReaction_SetTransitionState", _wrap_OBReaction_SetTransitionState, METH_VARARGS, NULL},
	 { "OBReaction_AddAgent", _wrap_OBReaction_AddAgent, METH_VARARGS, NULL},
	 { "OBReaction_GetReactant", _wrap_OBReaction_GetReactant, METH_VARARGS, NULL},
	 { "OBReaction_GetProduct", _wrap_OBReaction_GetProduct, METH_VARARGS, NULL},
	 { "OBReaction_GetAgent", _wrap_OBReaction_GetAgent, METH_VARARGS, NULL},
	 { "OBReaction_GetTransitionState", _wrap_OBReaction_GetTransitionState, METH_O, NULL},
	 { "OBReaction_GetTitle", _wrap_OBReaction_GetTitle, METH_O, NULL},
	 { "OBReaction_GetComment", _wrap_OBReaction_GetComment, METH_O, NULL},
	 { "OBReaction_SetTitle", _wrap_OBReaction_SetTitle, METH_VARARGS, NULL},
	 { "OBReaction_SetComment", _wrap_OBReaction_SetComment, METH_VARARGS, NULL},
	 { "OBReaction_IsReversible", _wrap_OBReaction_IsReversible, METH_O, NULL},
	 { "OBReaction_SetReversible", _wrap_OBReaction_SetReversible, METH_VARARGS, NULL},
	 { "OBReaction_ClassDescription", _wrap_OBReaction_ClassDescription, METH_NOARGS, NULL},
	 { "delete_OBReaction", _wrap_delete_OBReaction, METH_O, NULL},
	 { "OBReaction_swigregister", OBReaction_swigregister, METH_O, NULL},
	 { "OBReaction_swiginit", OBReaction_swiginit, METH_VARARGS, NULL},
	 { "new_OBReactionFacade", _wrap_new_OBReactionFacade, METH_O, "\n"
		"The constructor requires an OBMol. Member functions retrieve or manipulate reaction\n"
		"data stored in this molecule.\n"
		""},
	 { "delete_OBReactionFacade", _wrap_delete_OBReactionFacade, METH_O, NULL},
	 { "OBReactionFacade_AssignComponentIds", _wrap_OBReactionFacade_AssignComponentIds, METH_VARARGS, "\n"
		"These are convenience functions that set/get reaction-related data on individual atoms.\n"
		"The actual data is stored in the underlying OBMol as OBPairInteger data with the\n"
		"keys \"rxnrole\" or \"rxncomp\" (reaction role, reaction component).\n"
		"The other methods of this class are implemented using these methods. Care should be\n"
		"taken using the setter functions, as (unlike the other\n"
		"methods) their use can result in an inconsistent state. IsValid() may be used to check\n"
		"whether this is the case.\n"
		"\n"
		"\n"
		"Assigns a component Id to every atom in the molecule based on connected components.\n"
		"If ``wipe`` is ``false``, then any atoms that already have a component Id are skipped.\n"
		""},
	 { "OBReactionFacade_GetRole", _wrap_OBReactionFacade_GetRole, METH_VARARGS, "Return the reaction role of an atom."},
	 { "OBReactionFacade_GetComponentId", _wrap_OBReactionFacade_GetComponentId, METH_VARARGS, "Return the component Id of an atom."},
	 { "OBReactionFacade_SetRole", _wrap_OBReactionFacade_SetRole, METH_VARARGS, "Set the reaction role of an atom."},
	 { "OBReactionFacade_SetComponentId", _wrap_OBReactionFacade_SetComponentId, METH_VARARGS, "Set the component Id of an atom."},
	 { "OBReactionFacade_IsValid", _wrap_OBReactionFacade_IsValid, METH_O, "\n"
		"Check whether this OBMol is in a valid state to represent a reaction.\n"
		"\n"
		"Note, first of all, that this does not check whether a reaction is\n"
		"chemically valid. It simply checks whether the OBMol has the following\n"
		"properties:\n"
		"- (a) it is marked as a reaction\n"
		"- (b) every atom is marked as having a reaction role\n"
		"- (c) every atom is marked as belonging to a particular reaction component\n"
		"- (d) every atom in a connected component is marked as belong to the same\n"
		"      reaction component and having the same reaction role\n"
		"\n"
		"If (a) is not true, while it does not affect any of the methods of the\n"
		"OBReactionFacade, the reaction may not be written appropriately by an\n"
		"output format.\n"
		"\n"
		"If (b), (c) or (d) is not true, then the behavior of the toolkit is undefined\n"
		"when handling this molecule.\n"
		"\n"
		"While not checked by this function, it is also a good idea to avoid reusing\n"
		"the same component Id for different reaction roles, e.g. having a reactant\n"
		"with component Id 1 as well as a product with component Id 1. If you reassign\n"
		"roles, then the duplication of component Ids may cause two components to\n"
		"be merged into one.\n"
		""},
	 { "OBReactionFacade_AddComponent", _wrap_OBReactionFacade_AddComponent, METH_VARARGS, "\n"
		"These are a set of methods that manipulate reactions at the component level.\n"
		"If you use these high-level methods, and subsequently modify the underlying\n"
		"OBMol or its reaction data, you may need to clear cached information on the\n"
		"reaction components via ClearInternalState().\n"
		"\n"
		"\n"
		"Add a new component to the reaction\n"
		""},
	 { "OBReactionFacade_ClearInternalState", _wrap_OBReactionFacade_ClearInternalState, METH_O, "\n"
		"Clear the list of found components.\n"
		"\n"
		"On first\n"
		"use any of the high-level methods,\n"
		"the number of components for each reaction role is\n"
		"determined and each is assigned a number from 0 to the number of components.\n"
		"This information is cached for future accesses. If you modify the underlying\n"
		"OBMol in a way that invalidates this information (e.g. using the low-level\n"
		"methods), you should call ClearInternalState() or create a new OBReactionFacade.\n"
		""},
	 { "OBReactionFacade_GetComponent", _wrap_OBReactionFacade_GetComponent, METH_VARARGS, "\n"
		"Copy a component from the reaction into the provided OBMol.\n"
		"\n"
		"Note that all\n"
		"reaction data is copied, and so this may be useful for building up a new\n"
		"reaction composed of just some of the components of the original reaction.\n"
		"If copying components from multiple reactions into a single one, to avoid\n"
		"duplicate component Ids it may be easier to copy into an empty OBMol, and\n"
		"then use AddComponent() to add this OBMol into the new reaction.\n"
		"\n"
		":rtype: boolean\n"
		":return: A boolean indicating whether the reaction has a component with that number.\n"
		""},
	 { "OBReactionFacade_NumComponents", _wrap_OBReactionFacade_NumComponents, METH_VARARGS, "Return the number of reaction components with a particular reaction role."},
	 { "OBReactionFacade_ReassignComponent", _wrap_OBReactionFacade_ReassignComponent, METH_VARARGS, "\n"
		"Reassign the reaction role of a given component. If assigned to #NO_REACTIONROLE,\n"
		"this may be used to hide the reaction component when writing to a reaction\n"
		"output format.\n"
		"\n"
		"It should be noted that reassigning a component changes the numbers assigned.\n"
		"For example,\n"
		"in a reaction with 2 reactants and 2 products (each numbered from 0 to 1), if you\n"
		"reassign reactant number\n"
		"0 to be a product, it will now be product number 2 while former reactant number 1\n"
		"will now be reactant number 0.\n"
		"\n"
		":rtype: boolean\n"
		":return: A boolean indicating whether the reaction has a component with that number.\n"
		""},
	 { "OBReactionFacade_swigregister", OBReactionFacade_swigregister, METH_O, NULL},
	 { "OBReactionFacade_swiginit", OBReactionFacade_swiginit, METH_VARARGS, NULL},
	 { "new_OBMol", _wrap_new_OBMol, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor, copies atoms,bonds and OBGenericData\n"
		""},
	 { "delete_OBMol", _wrap_delete_OBMol, METH_O, "Destructor"},
	 { "OBMol___iadd__", _wrap_OBMol___iadd__, METH_VARARGS, "Copies atoms and bonds but not OBGenericData"},
	 { "OBMol_ReserveAtoms", _wrap_OBMol_ReserveAtoms, METH_VARARGS, "Reserve a minimum number of atoms for internal storage This improves performance since the internal atom vector does not grow. Reservation is purely a hint, so an unreasonably large value (e.g. a count read from a malformed file) is silently clamped to a sane upper bound to avoid OOM on untrusted input."},
	 { "OBMol_DestroyAtom", _wrap_OBMol_DestroyAtom, METH_VARARGS, "Free an OBAtom pointer if defined. Does no bookkeeping See also: DeleteAtom which ensures internal connections"},
	 { "OBMol_DestroyBond", _wrap_OBMol_DestroyBond, METH_VARARGS, "Free an OBBond pointer if defined. Does no bookkeeping See also: DeleteBond which ensures internal connections"},
	 { "OBMol_DestroyResidue", _wrap_OBMol_DestroyResidue, METH_VARARGS, "Free an OBResidue pointer if defined. Does no bookkeeping See also: DeleteResidue which ensures internal connections"},
	 { "OBMol_AddAtom", _wrap_OBMol_AddAtom, METH_VARARGS, "\n"
		" Add the specified atom to this molecule :type atom: :py:class:`OBAtom`\n"
		":param atom:        the atom to add :type forceNewId: boolean, optional\n"
		":param forceNewId:  whether to make a new atom Id even if the atom already has one (default is false) :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_InsertAtom", _wrap_OBMol_InsertAtom, METH_VARARGS, "Add a new atom to this molecule (like AddAtom) Calls BeginModify() before insertion and EndModify() after insertion"},
	 { "OBMol_AddBond", _wrap_OBMol_AddBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Add a new bond to the molecule with the specified parameters :type beginIdx: int\n"
		":param beginIdx:  the atom index of the \"start\" atom :type endIdx: int\n"
		":param endIdx:    the atom index of the \"end\" atom :type order: int\n"
		":param order:     the bond order (see OBBond::GetBondOrder()) :type flags: int, optional\n"
		":param flags:     any bond flags such as stereochemistry (default = none) :type insertpos: int, optional\n"
		":param insertpos: the position index to insert the bond (default = none) :rtype: boolean\n"
		":return: Whether the new bond creation was successful\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Add the specified residue to this molecule and update connections :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_AddResidue", _wrap_OBMol_AddResidue, METH_VARARGS, "\n"
		" Add the specified residue to this molecule and update connections :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_NewAtom", _wrap_OBMol_NewAtom, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Create a new OBAtom in this molecule and ensure connections (e.g. OBAtom::GetParent(). A new unique id will be assigned to this atom.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Create a new OBAtom in this molecule and ensure connections. (e.g. OBAtom::GetParent(). The ``id`` will be assigned to this atom.\n"
		""},
	 { "OBMol_NewBond", _wrap_OBMol_NewBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Create a new OBBond in this molecule and ensure connections (e.g. OBBond::GetParent(). A new unique id will be assigned to this bond.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Create a new OBBond in this molecule and ensure connections (e.g. OBBond::GetParent(). The ``id`` will be assigned to this bond.\n"
		""},
	 { "OBMol_NewResidue", _wrap_OBMol_NewResidue, METH_O, "Create a new OBResidue in this molecule and ensure connections."},
	 { "OBMol_DeleteAtom", _wrap_OBMol_DeleteAtom, METH_VARARGS, "\n"
		" Deletes an atom from this molecule and all appropriate bonds. Updates the molecule and atom and bond indexes accordingly. Warning: Does not update any residues which may contain this atom :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_DeleteBond", _wrap_OBMol_DeleteBond, METH_VARARGS, "\n"
		" Deletes an bond from this molecule and updates accordingly :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_DeleteResidue", _wrap_OBMol_DeleteResidue, METH_VARARGS, "\n"
		" Deletes a residue from this molecule and updates accordingly. :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_BeginModify", _wrap_OBMol_BeginModify, METH_O, "Call when making many modifications -- clears conformer/rotomer data. The method \"turns off\" perception routines, improving performance. Changes in molecular structure will be re-considered after modifications."},
	 { "OBMol_EndModify", _wrap_OBMol_EndModify, METH_VARARGS, "Call when done with modificaions -- re-perceive data as needed. This method \"turns on\" perception routines and re-evaluates molecular structure."},
	 { "OBMol_GetMod", _wrap_OBMol_GetMod, METH_O, "\n"
		":rtype: int\n"
		":return: The number of nested BeginModify() calls. Used internally.\n"
		""},
	 { "OBMol_IncrementMod", _wrap_OBMol_IncrementMod, METH_O, "Increase the number of nested BeginModify calls. Dangerous! Instead, properly use BeginModify as needed."},
	 { "OBMol_DecrementMod", _wrap_OBMol_DecrementMod, METH_O, "Decrease the number of nested BeginModify calls. Dangerous! Instead, properly use EndModify as needed."},
	 { "OBMol_GetFlags", _wrap_OBMol_GetFlags, METH_O, "\n"
		":rtype: int\n"
		":return: the entire set of flags. (Internal use, mainly.)\n"
		""},
	 { "OBMol_GetTitle", _wrap_OBMol_GetTitle, METH_VARARGS, "\n"
		" :rtype: char\n"
		" :return: the title of this molecule (often the filename) :type replaceNewlines: boolean, optional\n"
		":param replaceNewlines: whether to replace any newline characters with spaces\n"
		""},
	 { "OBMol_NumAtoms", _wrap_OBMol_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of atoms (i.e. OBAtom children)\n"
		""},
	 { "OBMol_NumBonds", _wrap_OBMol_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the number of bonds (i.e. OBBond children)\n"
		""},
	 { "OBMol_NumHvyAtoms", _wrap_OBMol_NumHvyAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of non-hydrogen atoms\n"
		""},
	 { "OBMol_NumResidues", _wrap_OBMol_NumResidues, METH_O, "\n"
		":rtype: int\n"
		":return: the number of residues (i.e. OBResidue substituents)\n"
		""},
	 { "OBMol_NumRotors", _wrap_OBMol_NumRotors, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the number of rotatable bonds. If sampleRingBonds is true, will include rotors within rings (see OBBond::IsRotor() for details)\n"
		""},
	 { "OBMol_GetAtom", _wrap_OBMol_GetAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the atom at index ``idx`` or NULL if it does not exist. Warning: Atom indexing will change. Use iterator methods instead.\n"
		""},
	 { "OBMol_GetAtomById", _wrap_OBMol_GetAtomById, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the atom with ``id`` or NULL if it does not exist.\n"
		""},
	 { "OBMol_GetFirstAtom", _wrap_OBMol_GetFirstAtom, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the first atom in this molecule, or NULL if none exist. Deprecated: Will be removed in favor of more standard iterator methods\n"
		""},
	 { "OBMol_GetBondById", _wrap_OBMol_GetBondById, METH_VARARGS, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond with ``id`` or NULL if it does not exist.\n"
		""},
	 { "OBMol_GetBond", _wrap_OBMol_GetBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond at index ``idx`` or NULL if it does not exist. Warning: Bond indexing may change. Use iterator methods instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond connecting the atom indexed by ``a`` and ``b`` or NULL if none exists. Warning: Atom indexing will change. Use atom objects and iterators instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond between the atoms ``bgn`` and ``end`` or NULL if none exists\n"
		""},
	 { "OBMol_GetResidue", _wrap_OBMol_GetResidue, METH_VARARGS, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: the residue indexed by ``idx``, or NULL if none exists Warning: Residue indexing may change. Use iterator methods instead.\n"
		""},
	 { "OBMol_GetInternalCoord", _wrap_OBMol_GetInternalCoord, METH_O, NULL},
	 { "OBMol_GetTorsion", _wrap_OBMol_GetTorsion, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the dihedral angle (in degrees) between the four atoms supplied a1-a2-a3-a4)\n"
		"    WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"    in degrees\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the dihedral angle (in degrees) between the four atoms ``a``, ``b``, ``c``, and ``d)``\n"
		"    WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"    in degrees\n"
		""},
	 { "OBMol_GetAngle", _wrap_OBMol_GetAngle, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the angle (in degrees) between the three atoms ``a``, ``b`` and ``c``  (where  a-> b (vertex) -> c )\n"
		""},
	 { "OBMol_AreInSameRing", _wrap_OBMol_AreInSameRing, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the size of the smallest ring if a and b are in the same ring, 0 otherwise Since: version 2.4\n"
		""},
	 { "OBMol_GetFormula", _wrap_OBMol_GetFormula, METH_O, "\n"
		":rtype: string\n"
		":return: the stochoimetric formula (e.g., C4H6O)\n"
		""},
	 { "OBMol_GetSpacedFormula", _wrap_OBMol_GetSpacedFormula, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the stochoimetric formula in spaced format e.g. C 4 H 6 O 1\n"
		""},
	 { "OBMol_GetMolWt", _wrap_OBMol_GetMolWt, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the standard molar mass given by IUPAC atomic masses (amu)\n"
		""},
	 { "OBMol_GetExactMass", _wrap_OBMol_GetExactMass, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the mass given by isotopes (or most abundant isotope, if not specified)\n"
		""},
	 { "OBMol_GetTotalCharge", _wrap_OBMol_GetTotalCharge, METH_O, "\n"
		":rtype: int\n"
		":return: the total charge on this molecule (i.e., 0 = neutral, +1, -1...)\n"
		""},
	 { "OBMol_GetTotalSpinMultiplicity", _wrap_OBMol_GetTotalSpinMultiplicity, METH_O, "\n"
		":rtype: int\n"
		":return: the total spin on this molecule (i.e., 1 = singlet, 2 = doublet...)\n"
		""},
	 { "OBMol_GetDimension", _wrap_OBMol_GetDimension, METH_O, "\n"
		":rtype: int\n"
		":return: the dimensionality of coordinates (i.e., 0 = unknown or no coord, 2=2D, 3=3D)\n"
		""},
	 { "OBMol_GetCoordinates", _wrap_OBMol_GetCoordinates, METH_O, "\n"
		":rtype: float\n"
		":return: the set of all atomic coordinates. See OBAtom::GetCoordPtr for more\n"
		""},
	 { "OBMol_GetSSSR", _wrap_OBMol_GetSSSR, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >\n"
		":return: the Smallest Set of Smallest Rings has been run (see OBRing class)\n"
		""},
	 { "OBMol_GetLSSR", _wrap_OBMol_GetLSSR, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >\n"
		":return: the Largest Set of Smallest Rings has been run (see OBRing class)\n"
		""},
	 { "OBMol_AutomaticFormalCharge", _wrap_OBMol_AutomaticFormalCharge, METH_O, "Get the current flag for whether formal charges are set with pH correction"},
	 { "OBMol_AutomaticPartialCharge", _wrap_OBMol_AutomaticPartialCharge, METH_O, "Get the current flag for whether partial charges are auto-determined"},
	 { "OBMol_SetTitle", _wrap_OBMol_SetTitle, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the title of this molecule to ``title``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the title of this molecule to ``title``\n"
		""},
	 { "OBMol_SetFormula", _wrap_OBMol_SetFormula, METH_VARARGS, "Set the stochiometric formula for this molecule"},
	 { "OBMol_SetEnergy", _wrap_OBMol_SetEnergy, METH_VARARGS, "Set the heat of formation for this molecule (in kcal/mol)"},
	 { "OBMol_SetDimension", _wrap_OBMol_SetDimension, METH_VARARGS, "Set the dimension of this molecule (i.e., 0, 1 , 2, 3)"},
	 { "OBMol_SetTotalCharge", _wrap_OBMol_SetTotalCharge, METH_VARARGS, "Set the total charge of this molecule to ``charge``"},
	 { "OBMol_SetTotalSpinMultiplicity", _wrap_OBMol_SetTotalSpinMultiplicity, METH_VARARGS, "Set the total spin multiplicity of this molecule to ``spinMultiplicity`` Overrides the calculation from spin multiplicity of OBAtoms"},
	 { "OBMol_SetInternalCoord", _wrap_OBMol_SetInternalCoord, METH_VARARGS, "Set the internal coordinates to ``int_coord`` (Does not call InternalToCartesian to update the 3D cartesian coordinates). The size of the ``int_coord`` has to be the same as the number of atoms in molecule (+ NULL at the beginning)."},
	 { "OBMol_SetAutomaticFormalCharge", _wrap_OBMol_SetAutomaticFormalCharge, METH_VARARGS, "Set the flag for determining automatic formal charges with pH (default=true)"},
	 { "OBMol_SetAutomaticPartialCharge", _wrap_OBMol_SetAutomaticPartialCharge, METH_VARARGS, "Set the flag for determining partial charges automatically (default=true)"},
	 { "OBMol_SetAromaticPerceived", _wrap_OBMol_SetAromaticPerceived, METH_VARARGS, "Mark that aromaticity has been perceived for this molecule (see OBAromaticTyper)"},
	 { "OBMol_SetSSSRPerceived", _wrap_OBMol_SetSSSRPerceived, METH_VARARGS, "Mark that Smallest Set of Smallest Rings has been run (see OBRing class)"},
	 { "OBMol_SetLSSRPerceived", _wrap_OBMol_SetLSSRPerceived, METH_VARARGS, "Mark that Largest Set of Smallest Rings has been run (see OBRing class)"},
	 { "OBMol_SetRingAtomsAndBondsPerceived", _wrap_OBMol_SetRingAtomsAndBondsPerceived, METH_VARARGS, "Mark that rings have been perceived (see OBRing class for details)"},
	 { "OBMol_SetAtomTypesPerceived", _wrap_OBMol_SetAtomTypesPerceived, METH_VARARGS, "Mark that atom types have been perceived (see OBAtomTyper for details)"},
	 { "OBMol_SetRingTypesPerceived", _wrap_OBMol_SetRingTypesPerceived, METH_VARARGS, "Mark that ring types have been perceived (see OBRingTyper for details)"},
	 { "OBMol_SetChainsPerceived", _wrap_OBMol_SetChainsPerceived, METH_VARARGS, "Mark that chains and residues have been perceived (see OBChainsParser)"},
	 { "OBMol_SetChiralityPerceived", _wrap_OBMol_SetChiralityPerceived, METH_VARARGS, "Mark that chirality has been perceived"},
	 { "OBMol_SetPartialChargesPerceived", _wrap_OBMol_SetPartialChargesPerceived, METH_VARARGS, "Mark that partial charges have been assigned"},
	 { "OBMol_SetHybridizationPerceived", _wrap_OBMol_SetHybridizationPerceived, METH_VARARGS, "Mark that hybridization of all atoms has been assigned"},
	 { "OBMol_SetClosureBondsPerceived", _wrap_OBMol_SetClosureBondsPerceived, METH_VARARGS, "Mark that ring closure bonds have been assigned by graph traversal"},
	 { "OBMol_SetHydrogensAdded", _wrap_OBMol_SetHydrogensAdded, METH_VARARGS, "Mark that explicit hydrogen atoms have been added"},
	 { "OBMol_SetCorrectedForPH", _wrap_OBMol_SetCorrectedForPH, METH_VARARGS, NULL},
	 { "OBMol_SetSpinMultiplicityAssigned", _wrap_OBMol_SetSpinMultiplicityAssigned, METH_VARARGS, NULL},
	 { "OBMol_SetIsPatternStructure", _wrap_OBMol_SetIsPatternStructure, METH_VARARGS, "The OBMol is a pattern, not a complete molecule. Left unchanged by Clear()."},
	 { "OBMol_SetIsReaction", _wrap_OBMol_SetIsReaction, METH_VARARGS, NULL},
	 { "OBMol_SetPeriodicMol", _wrap_OBMol_SetPeriodicMol, METH_VARARGS, "Mark that distance calculations, etc., should apply periodic boundary conditions through the minimimum image convention. Does not automatically recalculate bonding."},
	 { "OBMol_HasFlag", _wrap_OBMol_HasFlag, METH_VARARGS, NULL},
	 { "OBMol_SetFlag", _wrap_OBMol_SetFlag, METH_VARARGS, NULL},
	 { "OBMol_UnsetFlag", _wrap_OBMol_UnsetFlag, METH_VARARGS, NULL},
	 { "OBMol_SetFlags", _wrap_OBMol_SetFlags, METH_VARARGS, NULL},
	 { "OBMol_ClassDescription", _wrap_OBMol_ClassDescription, METH_NOARGS, NULL},
	 { "OBMol_RenumberAtoms", _wrap_OBMol_RenumberAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Renumber the atoms of this molecule according to the order in the supplied vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Renumber the atoms of this molecule using the initial indexes in the supplied vector\n"
		""},
	 { "OBMol_SetCoordinates", _wrap_OBMol_SetCoordinates, METH_VARARGS, "Set the coordinates for all atoms in this conformer. See also: OBMol::GetCoordinates()"},
	 { "OBMol_ToInertialFrame", _wrap_OBMol_ToInertialFrame, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translate one conformer and rotate by a rotation matrix (which is returned) to the inertial frame-of-reference\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Translate all conformers to the inertial frame-of-reference\n"
		""},
	 { "OBMol_Translate", _wrap_OBMol_Translate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translates all conformers in the molecule by the supplied vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Translates one conformer in the molecule by the supplied vector\n"
		""},
	 { "OBMol_Rotate", _wrap_OBMol_Rotate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Rotate all conformers using the supplied matrix ``u`` (a 3x3 array of double)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Rotate all conformers using the supplied matrix ``m`` (a linear 3x3 row-major array of double)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Rotate a specific conformer ``nconf`` using the supplied rotation matrix ``m``\n"
		""},
	 { "OBMol_DeleteHydrogens", _wrap_OBMol_DeleteHydrogens, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Suppress hydrogens by converting explicit hydrogen atoms to implicit :rtype: boolean\n"
		":return: Success\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Suppress explicit hydrogen atoms on the supplied atom :rtype: boolean\n"
		":return: Success\n"
		""},
	 { "OBMol_DeletePolarHydrogens", _wrap_OBMol_DeletePolarHydrogens, METH_O, "Suppress explicit hydrogen atoms connected to a polar atom See also: OBAtom::IsPolarHydrogen Since: version 2.4"},
	 { "OBMol_DeleteNonPolarHydrogens", _wrap_OBMol_DeleteNonPolarHydrogens, METH_O, "Suppress explicit hydrogen atoms connected to a non-polar atom See also: OBAtom::IsNonPolarHydrogen"},
	 { "OBMol_DeleteHydrogen", _wrap_OBMol_DeleteHydrogen, METH_VARARGS, "Suppress the supplied atom if it is a hydrogen (Helper function for DeleteHydrogens)"},
	 { "OBMol_AddHydrogens", _wrap_OBMol_AddHydrogens, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Convert implicit hydrogens to explicit atoms in the molecular graph :type polaronly: boolean, optional\n"
		":param polaronly:    Whether to add hydrogens only to polar atoms (i.e., not to C atoms) :type correctForPH: boolean, optional\n"
		":param correctForPH: Whether to call CorrectForPH() first :type pH: float, optional\n"
		":param pH: The pH to use for CorrectForPH() modification :rtype: boolean\n"
		":return: Whether any hydrogens were added\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" For a particular atom, convert implicit hydrogens to explicit atoms in the molecular graph\n"
		""},
	 { "OBMol_AddPolarHydrogens", _wrap_OBMol_AddPolarHydrogens, METH_O, "For polar atoms only, convert implicit hydrogens to explicit atoms in the molecular graph"},
	 { "OBMol_AddNonPolarHydrogens", _wrap_OBMol_AddNonPolarHydrogens, METH_O, "For non-polar atoms only, convert implicit hydrogens to explicit atoms in the molecular graph Since: version 2.4"},
	 { "OBMol_AddNewHydrogens", _wrap_OBMol_AddNewHydrogens, METH_VARARGS, "For polar and/or non-polar atoms, convert implicit hydrogens to explicit atoms in the molecular graph Since: verison 2.4"},
	 { "OBMol_StripSalts", _wrap_OBMol_StripSalts, METH_VARARGS, "If ``threshold`` is not specified or is zero, remove all but the largest contiguous fragment. If ``threshold`` is non-zero, remove any fragments with fewer than ``threshold`` atoms."},
	 { "OBMol_Separate", _wrap_OBMol_Separate, METH_VARARGS, "Copies each disconnected fragment as a separate OBMol"},
	 { "OBMol_GetNextFragment", _wrap_OBMol_GetNextFragment, METH_VARARGS, "Iterative component of Separate to copy one fragment at a time"},
	 { "OBMol_CopySubstructure", _wrap_OBMol_CopySubstructure, METH_VARARGS, NULL},
	 { "OBMol_ConvertDativeBonds", _wrap_OBMol_ConvertDativeBonds, METH_O, "Converts the charged form of coordinate bonds, e.g.[N+]([O-])=O to N(=O)=O"},
	 { "OBMol_MakeDativeBonds", _wrap_OBMol_MakeDativeBonds, METH_O, "\n"
		" Converts 5-valent N and P only. Return true if conversion occurred. :rtype: boolean\n"
		":return: has charged form of dative bonds(e.g.[N+]([O-])=O from N(=O)=O). Since: version 2.4\n"
		""},
	 { "OBMol_ConvertZeroBonds", _wrap_OBMol_ConvertZeroBonds, METH_O, "\n"
		"Convert zero-order bonds to single or double bonds and adjust adjacent atom\n"
		"charges in an attempt to achieve the correct valence state.\n"
		":rtype: boolean\n"
		":return: Whether any modifications were made\n"
		"Since: version 2.4\n"
		""},
	 { "OBMol_CorrectForPH", _wrap_OBMol_CorrectForPH, METH_VARARGS, "Correct for pH by applying the OBPhModel transformations"},
	 { "OBMol_AssignSpinMultiplicity", _wrap_OBMol_AssignSpinMultiplicity, METH_VARARGS, NULL},
	 { "OBMol_AssignTotalChargeToAtoms", _wrap_OBMol_AssignTotalChargeToAtoms, METH_VARARGS, "Put the specified molecular charge on appropriate atoms. Assumes all the hydrogen is explicitly included in the molecule. Since: version 2.4"},
	 { "OBMol_Center", _wrap_OBMol_Center, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translate to the center of all coordinates (for this conformer)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the center of the supplied conformer ``nconf`` See also: Center() to actually center all conformers at the origin\n"
		""},
	 { "OBMol_FindSSSR", _wrap_OBMol_FindSSSR, METH_O, "Find Smallest Set of Smallest Rings (see OBRing class for more details)"},
	 { "OBMol_FindLSSR", _wrap_OBMol_FindLSSR, METH_O, "Find Largest Set of Smallest Rings"},
	 { "OBMol_FindRingAtomsAndBonds", _wrap_OBMol_FindRingAtomsAndBonds, METH_O, "Find all ring atoms and bonds. Does not need to call FindSSSR()."},
	 { "OBMol_FindChildren", _wrap_OBMol_FindChildren, METH_VARARGS, NULL},
	 { "OBMol_FindLargestFragment", _wrap_OBMol_FindLargestFragment, METH_VARARGS, "\n"
		" Find the largest fragment in OBMol (which may include multiple non-connected fragments) :type frag: :py:class:`OBBitVec`\n"
		":param frag:   Return (by reference) a bit vector indicating the atoms in the largest fragment\n"
		""},
	 { "OBMol_ContigFragList", _wrap_OBMol_ContigFragList, METH_VARARGS, "Sort a list of contig fragments by size from largest to smallest Each vector<int> contains the atom numbers of a contig fragment"},
	 { "OBMol_Align", _wrap_OBMol_Align, METH_VARARGS, "Aligns atom a on p1 and atom b along p1->p2 vector"},
	 { "OBMol_ConnectTheDots", _wrap_OBMol_ConnectTheDots, METH_O, "Adds single bonds based on atom proximity"},
	 { "OBMol_PerceiveBondOrders", _wrap_OBMol_PerceiveBondOrders, METH_O, "Attempts to perceive multiple bonds based on geometries"},
	 { "OBMol_FindAngles", _wrap_OBMol_FindAngles, METH_O, "Fills out an OBAngleData with angles from the molecule"},
	 { "OBMol_FindTorsions", _wrap_OBMol_FindTorsions, METH_O, "Fills out an OBTorsionData with angles from the molecule"},
	 { "OBMol_GetGTDVector", _wrap_OBMol_GetGTDVector, METH_VARARGS, NULL},
	 { "OBMol_GetGIVector", _wrap_OBMol_GetGIVector, METH_VARARGS, NULL},
	 { "OBMol_GetGIDVector", _wrap_OBMol_GetGIDVector, METH_VARARGS, NULL},
	 { "OBMol_Has2D", _wrap_OBMol_Has2D, METH_VARARGS, "Are there non-zero coordinates in two dimensions (i.e. X and Y)- and, if Not3D is true, no Z coordinates?"},
	 { "OBMol_Has3D", _wrap_OBMol_Has3D, METH_O, "Are there non-zero coordinates in all three dimensions (i.e. X, Y, Z)?"},
	 { "OBMol_HasNonZeroCoords", _wrap_OBMol_HasNonZeroCoords, METH_O, "Are there any non-zero coordinates?"},
	 { "OBMol_HasAromaticPerceived", _wrap_OBMol_HasAromaticPerceived, METH_O, "Has aromatic perception been performed?"},
	 { "OBMol_HasSSSRPerceived", _wrap_OBMol_HasSSSRPerceived, METH_O, "Has the smallest set of smallest rings (FindSSSR) been performed?"},
	 { "OBMol_HasLSSRPerceived", _wrap_OBMol_HasLSSRPerceived, METH_O, "Has the largest set of smallest rings (FindLSSR) been performed?"},
	 { "OBMol_HasRingAtomsAndBondsPerceived", _wrap_OBMol_HasRingAtomsAndBondsPerceived, METH_O, "Have ring atoms and bonds been assigned?"},
	 { "OBMol_HasAtomTypesPerceived", _wrap_OBMol_HasAtomTypesPerceived, METH_O, "Have atom types been assigned by OBAtomTyper?"},
	 { "OBMol_HasRingTypesPerceived", _wrap_OBMol_HasRingTypesPerceived, METH_O, "Have ring types been assigned by OBRingTyper?"},
	 { "OBMol_HasChiralityPerceived", _wrap_OBMol_HasChiralityPerceived, METH_O, "Has atom chirality been assigned?"},
	 { "OBMol_HasPartialChargesPerceived", _wrap_OBMol_HasPartialChargesPerceived, METH_O, "Have atomic Gasteiger partial charges been assigned by OBGastChrg?"},
	 { "OBMol_HasHybridizationPerceived", _wrap_OBMol_HasHybridizationPerceived, METH_O, "Has atomic hybridization been assigned by OBAtomTyper?"},
	 { "OBMol_HasClosureBondsPerceived", _wrap_OBMol_HasClosureBondsPerceived, METH_O, "Have ring \"closure\" bonds been assigned? (e.g., OBBond::IsClosure())"},
	 { "OBMol_HasChainsPerceived", _wrap_OBMol_HasChainsPerceived, METH_O, "Have biomolecule chains and residues been assigned by OBChainsParser?"},
	 { "OBMol_HasHydrogensAdded", _wrap_OBMol_HasHydrogensAdded, METH_O, "Have hydrogens been added to the molecule?"},
	 { "OBMol_IsCorrectedForPH", _wrap_OBMol_IsCorrectedForPH, METH_O, "Has the molecule been corrected for pH by CorrectForPH?"},
	 { "OBMol_HasSpinMultiplicityAssigned", _wrap_OBMol_HasSpinMultiplicityAssigned, METH_O, "Has total spin multiplicity been assigned?"},
	 { "OBMol_IsReaction", _wrap_OBMol_IsReaction, METH_O, "Does this OBMol represent a reaction?"},
	 { "OBMol_IsPeriodic", _wrap_OBMol_IsPeriodic, METH_O, "Is this molecule periodic? Should periodic boundary conditions be applied?"},
	 { "OBMol_Empty", _wrap_OBMol_Empty, METH_O, "Are there any atoms in this molecule?"},
	 { "OBMol_NumConformers", _wrap_OBMol_NumConformers, METH_O, "\n"
		":rtype: int\n"
		":return: the number of conformers in this molecule\n"
		""},
	 { "OBMol_SetConformers", _wrap_OBMol_SetConformers, METH_VARARGS, "Set the entire set of conformers for this molecule to ``v``"},
	 { "OBMol_AddConformer", _wrap_OBMol_AddConformer, METH_VARARGS, "Add a new set of coordinates ``f`` as a new conformer"},
	 { "OBMol_SetConformer", _wrap_OBMol_SetConformer, METH_VARARGS, "Set the molecule's current conformer to ``i`` Does nothing if ``i`` is larger than NumConformers()"},
	 { "OBMol_CopyConformer", _wrap_OBMol_CopyConformer, METH_VARARGS, "Copy the conformer ``nconf`` into the array ``c`` Warning: Does no checking to see if ``c`` is large enough"},
	 { "OBMol_DeleteConformer", _wrap_OBMol_DeleteConformer, METH_VARARGS, "Delete the conformer ``nconf``"},
	 { "OBMol_GetConformer", _wrap_OBMol_GetConformer, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the coordinates to conformer ``i``\n"
		""},
	 { "OBMol_SetEnergies", _wrap_OBMol_SetEnergies, METH_VARARGS, "Set the entire set of conformer energies"},
	 { "OBMol_GetEnergies", _wrap_OBMol_GetEnergies, METH_O, "Set the entire set of conformer energies"},
	 { "OBMol_GetEnergy", _wrap_OBMol_GetEnergy, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the heat of formation for this molecule (in kcal/mol)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Get the energy for conformer ci Title: ci conformer index\n"
		""},
	 { "OBMol_BeginConformer", _wrap_OBMol_BeginConformer, METH_VARARGS, "\n"
		" Set the iterator to the beginning of the conformer list :rtype: float\n"
		":return: the array of coordinates for the first conformer\n"
		""},
	 { "OBMol_NextConformer", _wrap_OBMol_NextConformer, METH_VARARGS, "\n"
		" Advance the iterator to the next confomer, if possible :rtype: float\n"
		":return: The array of coordinates for the next conformer, or NULL if none exist\n"
		""},
	 { "OBMol_GetConformers", _wrap_OBMol_GetConformers, METH_O, "\n"
		":rtype: std::vector< double *,std::allocator< double * > >\n"
		":return: the entire set of conformers for this molecule as a vector of floating point arrays\n"
		""},
	 { "OBMol_CBeginAtoms", _wrap_OBMol_CBeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant atom iterator pointing to the beginning of the atom list\n"
		""},
	 { "OBMol_CEndAtoms", _wrap_OBMol_CEndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant atom iterator pointing to the end of the atom list\n"
		""},
	 { "OBMol_GetAtoms", _wrap_OBMol_GetAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBMolAtomRange\n"
		":return: A range over the atoms. This range can be used in a range-based for loop.\n"
		""},
	 { "OBMol_GetBonds", _wrap_OBMol_GetBonds, METH_O, "\n"
		":rtype: OpenBabel::OBMolBondRange\n"
		":return: A range over the bonds. This range can be used in a range-based for loop.\n"
		""},
	 { "OBMol_BeginInternalCoord", _wrap_OBMol_BeginInternalCoord, METH_VARARGS, "\n"
		" Set the iterator to the beginning of the internal coordinate list :rtype: :py:class:`OBInternalCoord`\n"
		":return: the first internal coordinate record, or NULL if none exist See also: SetInternalCoord\n"
		""},
	 { "OBMol_NextInternalCoord", _wrap_OBMol_NextInternalCoord, METH_VARARGS, "\n"
		" Advance the iterator to the next internal coordinate record :rtype: :py:class:`OBInternalCoord`\n"
		":return: the next first internal coordinate record, or NULL if none exist See also: SetInternalCoord\n"
		""},
	 { "OBMol_SetTorsion", _wrap_OBMol_SetTorsion, METH_VARARGS, "\n"
		"Set the torsion defined by these atoms, rotating bonded neighbors\n"
		"Title: ang The torsion angle in radians\n"
		"WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"in degrees\n"
		""},
	 { "OBMol_swigregister", OBMol_swigregister, METH_O, NULL},
	 { "OBMol_swiginit", OBMol_swiginit, METH_VARARGS, NULL},
	 { "ThrowError", _wrap_ThrowError, METH_VARARGS, NULL},
	 { "CartesianToInternal", _wrap_CartesianToInternal, METH_VARARGS, "Convert Cartesian XYZ to a set of OBInternalCoord coordinates"},
	 { "InternalToCartesian", _wrap_InternalToCartesian, METH_VARARGS, "Convert set of OBInternalCoord coordinates into Cartesian XYZ"},
	 { "NewExtension", _wrap_NewExtension, METH_VARARGS, NULL},
	 { "get_rmat", _wrap_get_rmat, METH_VARARGS, NULL},
	 { "ob_make_rmat", _wrap_ob_make_rmat, METH_VARARGS, NULL},
	 { "qtrfit", _wrap_qtrfit, METH_VARARGS, NULL},
	 { "superimpose", _wrap_superimpose, METH_VARARGS, NULL},
	 { "OBRing_ring_id_set", _wrap_OBRing_ring_id_set, METH_VARARGS, "a unique id to ensure a stable sort in SSSR determination"},
	 { "OBRing_ring_id_get", _wrap_OBRing_ring_id_get, METH_O, "a unique id to ensure a stable sort in SSSR determination"},
	 { "OBRing__path_set", _wrap_OBRing__path_set, METH_VARARGS, "the path of this ring (i.e., the atom indexes)"},
	 { "OBRing__path_get", _wrap_OBRing__path_get, METH_O, "the path of this ring (i.e., the atom indexes)"},
	 { "OBRing__pathset_set", _wrap_OBRing__pathset_set, METH_VARARGS, "the path of this ring as a redundant bit vector"},
	 { "OBRing__pathset_get", _wrap_OBRing__pathset_get, METH_O, "the path of this ring as a redundant bit vector"},
	 { "new_OBRing", _wrap_new_OBRing, METH_VARARGS, "Initialize a ring from a set of atom indexes ``path`` and with ``size``"},
	 { "OBRing_Size", _wrap_OBRing_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the size of this ring (i.e., how many atoms in the cycle)\n"
		""},
	 { "OBRing_PathSize", _wrap_OBRing_PathSize, METH_O, "\n"
		":rtype: int\n"
		":return: the size of this ring (i.e., how many atoms in the cycle) Deprecated: Use Size() instead\n"
		""},
	 { "OBRing_IsAromatic", _wrap_OBRing_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this ring is aromatic If all atoms in this ring are aromatic, the ring will be considered aromatic TODO: This method uses implicit bonding -- bond info is not stored in OBRing\n"
		""},
	 { "OBRing_SetType", _wrap_OBRing_SetType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the ring type (see OBRingTyper for more)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the ring type (see OBRingTyper for more)\n"
		""},
	 { "OBRing_GetType", _wrap_OBRing_GetType, METH_O, "\n"
		":rtype: string\n"
		":return: the ring type\n"
		""},
	 { "OBRing_GetRootAtom", _wrap_OBRing_GetRootAtom, METH_O, "\n"
		":rtype: int\n"
		":return: the index for the root atom. O for furan, S for thiazole, N for pyrrole. For 6 membered aromatic rings, the first non carbon atom is used for root. For 5 members rings the O, S or N (BOSum=3, valence=3) will be used for root\n"
		""},
	 { "OBRing_IsMember", _wrap_OBRing_IsMember, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: Whether atom ``a`` is a member of this ring\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: Whether both atoms in bond ``b`` are in this ring TODO: This method uses implicit bonding -- bond info is not stored in OBRing\n"
		""},
	 { "OBRing_IsInRing", _wrap_OBRing_IsInRing, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether ``i`` as an atom index is in this ring\n"
		""},
	 { "OBRing_SetParent", _wrap_OBRing_SetParent, METH_VARARGS, "Set the parent of this ring to ``m``"},
	 { "OBRing_GetParent", _wrap_OBRing_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: the parent of this ring, or NULL if none has been defined\n"
		""},
	 { "OBRing_findCenterAndNormal", _wrap_OBRing_findCenterAndNormal, METH_VARARGS, "\n"
		" Set the supplied vectors to the ``center`` of this ring, along with the ``normal`` (in both directions). :type center: :py:class:`vector3`\n"
		":param center: The center of the ring :type norm1: :py:class:`vector3`\n"
		":param norm1: The normal of the best-fit plane for this ring :type norm2: :py:class:`vector3`\n"
		":param norm2: -1 * norm1 (i.e., the opposite direction of norm1) :rtype: boolean\n"
		":return: True (success)\n"
		""},
	 { "delete_OBRing", _wrap_delete_OBRing, METH_O, NULL},
	 { "OBRing_swigregister", OBRing_swigregister, METH_O, NULL},
	 { "OBRing_swiginit", OBRing_swiginit, METH_VARARGS, NULL},
	 { "CompareRingSize", _wrap_CompareRingSize, METH_VARARGS, "\n"
		" Comparison function for rings, used by OBRingSearch::SortRings() :rtype: boolean\n"
		":return: true if a.size() > b.size()\n"
		""},
	 { "new_OBRingSearch", _wrap_new_OBRingSearch, METH_NOARGS, NULL},
	 { "delete_OBRingSearch", _wrap_delete_OBRingSearch, METH_O, NULL},
	 { "OBRingSearch_SortRings", _wrap_OBRingSearch_SortRings, METH_O, "Sort ring sizes from smallest to largest"},
	 { "OBRingSearch_RemoveRedundant", _wrap_OBRingSearch_RemoveRedundant, METH_VARARGS, "Starting with a full ring set - reduce to SSSR set"},
	 { "OBRingSearch_AddRingFromClosure", _wrap_OBRingSearch_AddRingFromClosure, METH_VARARGS, "Add a new ring from a \"closure\" bond: See OBBond::IsClosure()"},
	 { "OBRingSearch_SaveUniqueRing", _wrap_OBRingSearch_SaveUniqueRing, METH_VARARGS, NULL},
	 { "OBRingSearch_WriteRings", _wrap_OBRingSearch_WriteRings, METH_O, "For debugging only, write the rings to std::cout"},
	 { "OBRingSearch_BeginRings", _wrap_OBRingSearch_BeginRings, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator\n"
		":return: an iterator pointing to the beginning of the list of rings\n"
		""},
	 { "OBRingSearch_EndRings", _wrap_OBRingSearch_EndRings, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator\n"
		":return: an iterator pointing to the end of the list of rings\n"
		""},
	 { "OBRingSearch_swigregister", OBRingSearch_swigregister, METH_O, NULL},
	 { "OBRingSearch_swiginit", OBRingSearch_swiginit, METH_VARARGS, NULL},
	 { "new_OBRTree", _wrap_new_OBRTree, METH_VARARGS, "Construct a search tree from a possible parent entry and atom entry"},
	 { "delete_OBRTree", _wrap_delete_OBRTree, METH_O, NULL},
	 { "OBRTree_GetAtomIdx", _wrap_OBRTree_GetAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the OBAtom::GetIdx() index of the atom in this node\n"
		""},
	 { "OBRTree_PathToRoot", _wrap_OBRTree_PathToRoot, METH_VARARGS, "Recursively find the root of this tree, building up a vector of OBAtom nodes."},
	 { "OBRTree_swigregister", OBRTree_swigregister, METH_O, NULL},
	 { "OBRTree_swiginit", OBRTree_swiginit, METH_VARARGS, NULL},
	 { "delete_OBSmartsPattern", _wrap_delete_OBSmartsPattern, METH_O, NULL},
	 { "new_OBSmartsPattern", _wrap_new_OBSmartsPattern, METH_VARARGS, NULL},
	 { "OBSmartsPattern_Init", _wrap_OBSmartsPattern_Init, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Parse the ``pattern`` SMARTS string. :rtype: boolean\n"
		":return: Whether the pattern is a valid SMARTS expression\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Parse the ``pattern`` SMARTS string. :rtype: boolean\n"
		":return: Whether the pattern is a valid SMARTS expression\n"
		""},
	 { "OBSmartsPattern_GetSMARTS", _wrap_OBSmartsPattern_GetSMARTS, METH_O, "\n"
		":rtype: string\n"
		":return: the SMARTS string which is currently used\n"
		""},
	 { "OBSmartsPattern_Empty", _wrap_OBSmartsPattern_Empty, METH_O, "\n"
		" :rtype: boolean\n"
		" :return: the SMARTS string which is currently used :rtype: boolean\n"
		":return: If the SMARTS pattern is an empty expression (e.g., invalid)\n"
		""},
	 { "OBSmartsPattern_IsValid", _wrap_OBSmartsPattern_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: If the SMARTS pattern is a valid expression\n"
		""},
	 { "OBSmartsPattern_NumAtoms", _wrap_OBSmartsPattern_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of atoms in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_NumBonds", _wrap_OBSmartsPattern_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the number of bonds in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_GetBond", _wrap_OBSmartsPattern_GetBond, METH_VARARGS, "\n"
		" Access the bond ``idx`` in the internal pattern :type src: int\n"
		":param src: The index of the beginning atom :type dst: int\n"
		":param dst: The index of the end atom :type ord: int\n"
		":param ord: The bond order of this bond :type idx: int\n"
		":param idx: The index of the bond in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_GetAtomicNum", _wrap_OBSmartsPattern_GetAtomicNum, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the atomic number of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_GetCharge", _wrap_OBSmartsPattern_GetCharge, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the formal charge of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_GetVectorBinding", _wrap_OBSmartsPattern_GetVectorBinding, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the vector binding of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_Match", _wrap_OBSmartsPattern_Match, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type single: boolean, optional\n"
		":param single: Whether only a single match is required (faster). Default is false. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). This version is (more) thread safe. :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type mlist: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":param mlist: The resulting match list :type mtype: int, optional\n"
		":param mtype: The match type to use. Default is All. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). This version is (more) thread safe. :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type mlist: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":param mlist: The resulting match list :param mtype: The match type to use. Default is All. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		""},
	 { "OBSmartsPattern_HasMatch", _wrap_OBSmartsPattern_HasMatch, METH_VARARGS, "\n"
		" Thread safe check for any SMARTS match :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :rtype: boolean\n"
		":return: Whether there exists any match\n"
		""},
	 { "OBSmartsPattern_RestrictedMatch", _wrap_OBSmartsPattern_RestrictedMatch, METH_VARARGS, NULL},
	 { "OBSmartsPattern_NumMatches", _wrap_OBSmartsPattern_NumMatches, METH_O, "\n"
		":rtype: int\n"
		":return: the number of non-unique SMARTS matches To get the number of unique SMARTS matches, query GetUMapList()->size()\n"
		""},
	 { "OBSmartsPattern_GetMapList", _wrap_OBSmartsPattern_GetMapList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":return: the entire list of non-unique matches for this pattern See also: GetUMapList()\n"
		""},
	 { "OBSmartsPattern_BeginMList", _wrap_OBSmartsPattern_BeginMList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator\n"
		":return: An iterator over the (non-unique) match list, starting at the beginning\n"
		""},
	 { "OBSmartsPattern_EndMList", _wrap_OBSmartsPattern_EndMList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator\n"
		":return: An iterator over the non-unique match list, set to the end\n"
		""},
	 { "OBSmartsPattern_GetUMapList", _wrap_OBSmartsPattern_GetUMapList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":return: the entire list of unique matches for this pattern\n"
		"          A unique match is defined as one which does not cover the\n"
		"          identical atoms that a previous match has covered.\n"
		"\n"
		"      For instance, the pattern [OD1]~C~[OD1] describes a\n"
		"      carboxylate group. This pattern will match both atom number\n"
		"      permutations of the carboxylate, and if GetMapList() is called, both\n"
		"      matches will be returned. If GetUMapList() is called only unique\n"
		"      matches of the pattern will be returned.\n"
		""},
	 { "OBSmartsPattern_WriteMapList", _wrap_OBSmartsPattern_WriteMapList, METH_VARARGS, "Debugging -- write a list of matches to the output stream"},
	 { "OBSmartsPattern_swigregister", OBSmartsPattern_swigregister, METH_O, NULL},
	 { "OBSmartsPattern_swiginit", OBSmartsPattern_swiginit, METH_VARARGS, NULL},
	 { "new_OBSmartsMatcher", _wrap_new_OBSmartsMatcher, METH_NOARGS, NULL},
	 { "delete_OBSmartsMatcher", _wrap_delete_OBSmartsMatcher, METH_O, NULL},
	 { "OBSmartsMatcher_match", _wrap_OBSmartsMatcher_match, METH_VARARGS, NULL},
	 { "OBSmartsMatcher_swigregister", OBSmartsMatcher_swigregister, METH_O, NULL},
	 { "OBSmartsMatcher_swiginit", OBSmartsMatcher_swiginit, METH_VARARGS, NULL},
	 { "new_OBSSMatch", _wrap_new_OBSSMatch, METH_VARARGS, NULL},
	 { "delete_OBSSMatch", _wrap_delete_OBSSMatch, METH_O, NULL},
	 { "OBSSMatch_Match", _wrap_OBSSMatch_Match, METH_VARARGS, NULL},
	 { "OBSSMatch_swigregister", OBSSMatch_swigregister, METH_O, NULL},
	 { "OBSSMatch_swiginit", OBSSMatch_swiginit, METH_VARARGS, NULL},
	 { "SmartsLexReplace", _wrap_SmartsLexReplace, METH_VARARGS, NULL},
	 { "new_AliasData", _wrap_new_AliasData, METH_NOARGS, NULL},
	 { "AliasData_SetAlias", _wrap_AliasData_SetAlias, METH_VARARGS, "Add an alias"},
	 { "AliasData_GetAlias", _wrap_AliasData_GetAlias, METH_VARARGS, "return value of alias or its version intended to be connected at its right hand end."},
	 { "AliasData_GetColor", _wrap_AliasData_GetColor, METH_O, "Return the color which has been assigned to this alias"},
	 { "AliasData_SetColor", _wrap_AliasData_SetColor, METH_VARARGS, "Assign a color to this alias"},
	 { "AliasData_IsExpanded", _wrap_AliasData_IsExpanded, METH_O, NULL},
	 { "AliasData_RevertToAliasForm", _wrap_AliasData_RevertToAliasForm, METH_O, "\n"
		"Converts all the expanded aliases in a molecule to the alias form.\n"
		"Note that this deletes atoms and bonds. Use only as a preparation for display.\n"
		""},
	 { "AliasData_Expand", _wrap_AliasData_Expand, METH_VARARGS, "Interprets the alias text and adds atoms as appropriate to mol."},
	 { "AliasData_AddAliases", _wrap_AliasData_AddAliases, METH_O, "\n"
		"Matches univalent fragments in the molecule and adds AliasData to provide an alternative representation.\n"
		"The molecule remains a respectable OBMol. Data in superatoms.txt decides what aliases are added.\n"
		""},
	 { "delete_AliasData", _wrap_delete_AliasData, METH_O, NULL},
	 { "AliasData_swigregister", AliasData_swigregister, METH_O, NULL},
	 { "AliasData_swiginit", AliasData_swiginit, METH_VARARGS, NULL},
	 { "OBFingerprint_Default", _wrap_OBFingerprint_Default, METH_NOARGS, NULL},
	 { "OBFingerprint_FindType", _wrap_OBFingerprint_FindType, METH_O, NULL},
	 { "delete_OBFingerprint", _wrap_delete_OBFingerprint, METH_O, NULL},
	 { "OBFingerprint_SetBit", _wrap_OBFingerprint_SetBit, METH_VARARGS, "Sets the nth bit"},
	 { "OBFingerprint_GetBit", _wrap_OBFingerprint_GetBit, METH_VARARGS, "return true if the nth bit is set;"},
	 { "OBFingerprint_Fold", _wrap_OBFingerprint_Fold, METH_VARARGS, "Repeatedly ORs the top half with the bottom half until no smaller than nbits"},
	 { "OBFingerprint_GetFingerprint", _wrap_OBFingerprint_GetFingerprint, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: fingerprint in vector, which may be resized, folded to nbits (if nbits!=0)\n"
		""},
	 { "OBFingerprint_Flags", _wrap_OBFingerprint_Flags, METH_O, NULL},
	 { "OBFingerprint_SetFlags", _wrap_OBFingerprint_SetFlags, METH_VARARGS, NULL},
	 { "OBFingerprint_DescribeBits", _wrap_OBFingerprint_DescribeBits, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: a description of each bit that is set (or unset, if bSet=false)\n"
		"Since: version 2.2\n"
		""},
	 { "OBFingerprint_Tanimoto", _wrap_OBFingerprint_Tanimoto, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the Tanimoto coefficient between two vectors (vector<unsigned int>& SeekPositions)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Inline version of Tanimoto() taking a pointer for the second vector\n"
		""},
	 { "OBFingerprint_Getbitsperint", _wrap_OBFingerprint_Getbitsperint, METH_NOARGS, NULL},
	 { "OBFingerprint_FindFingerprint", _wrap_OBFingerprint_FindFingerprint, METH_O, "\n"
		":rtype: :py:class:`OBFingerprint`\n"
		":return: a pointer to a fingerprint (the default if ID is empty), or NULL if not available\n"
		"   For backward compatibility;  a synonym of OBFingerprint::FindType\n"
		""},
	 { "OBFingerprint_swigregister", OBFingerprint_swigregister, METH_O, NULL},
	 { "FptIndexHeader_headerlength_set", _wrap_FptIndexHeader_headerlength_set, METH_VARARGS, "offset to data: sizeof(FptIndexHeader)"},
	 { "FptIndexHeader_headerlength_get", _wrap_FptIndexHeader_headerlength_get, METH_O, "offset to data: sizeof(FptIndexHeader)"},
	 { "FptIndexHeader_nEntries_set", _wrap_FptIndexHeader_nEntries_set, METH_VARARGS, "number of fingerprints"},
	 { "FptIndexHeader_nEntries_get", _wrap_FptIndexHeader_nEntries_get, METH_O, "number of fingerprints"},
	 { "FptIndexHeader_words_set", _wrap_FptIndexHeader_words_set, METH_VARARGS, "number 32bit words per fingerprint"},
	 { "FptIndexHeader_words_get", _wrap_FptIndexHeader_words_get, METH_O, "number 32bit words per fingerprint"},
	 { "FptIndexHeader_fpid_set", _wrap_FptIndexHeader_fpid_set, METH_VARARGS, "ID of the fingerprint type"},
	 { "FptIndexHeader_fpid_get", _wrap_FptIndexHeader_fpid_get, METH_O, "ID of the fingerprint type"},
	 { "FptIndexHeader_seek64_set", _wrap_FptIndexHeader_seek64_set, METH_VARARGS, NULL},
	 { "FptIndexHeader_seek64_get", _wrap_FptIndexHeader_seek64_get, METH_O, NULL},
	 { "FptIndexHeader_datafilename_set", _wrap_FptIndexHeader_datafilename_set, METH_VARARGS, "the data that this is an index to"},
	 { "FptIndexHeader_datafilename_get", _wrap_FptIndexHeader_datafilename_get, METH_O, "the data that this is an index to"},
	 { "new_FptIndexHeader", _wrap_new_FptIndexHeader, METH_NOARGS, NULL},
	 { "delete_FptIndexHeader", _wrap_delete_FptIndexHeader, METH_O, NULL},
	 { "FptIndexHeader_swigregister", FptIndexHeader_swigregister, METH_O, NULL},
	 { "FptIndexHeader_swiginit", FptIndexHeader_swiginit, METH_VARARGS, NULL},
	 { "FastSearch_ReadIndexFile", _wrap_FastSearch_ReadIndexFile, METH_VARARGS, "Loads an index from a file and returns the name of the datafile"},
	 { "FastSearch_ReadIndex", _wrap_FastSearch_ReadIndex, METH_VARARGS, NULL},
	 { "delete_FastSearch", _wrap_delete_FastSearch, METH_O, NULL},
	 { "FastSearch_Find", _wrap_FastSearch_Find, METH_VARARGS, "Does substructure search and returns vector of the file positions of matches"},
	 { "FastSearch_FindMatch", _wrap_FastSearch_FindMatch, METH_VARARGS, "\n"
		"Similar to Find() but all bits of matching fingerprints have to be the same\n"
		"Since: version 2.1\n"
		""},
	 { "FastSearch_FindSimilar", _wrap_FastSearch_FindSimilar, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: A multimap containing objects whose Tanimoto coefficients with the target\n"
		"    is greater than the value specified.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: A multimap containing the nCandidates objects with largest Tanimoto\n"
		"     coefficients with the target.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: boolean\n"
		":return: A multimap containing the nCandidates objects with largest Tanimoto\n"
		"     coefficients with the target.\n"
		""},
	 { "FastSearch_GetFingerprint", _wrap_FastSearch_GetFingerprint, METH_O, "\n"
		":rtype: :py:class:`OBFingerprint`\n"
		":return: a pointer to the fingerprint type used to constuct the index\n"
		""},
	 { "FastSearch_GetIndexHeader", _wrap_FastSearch_GetIndexHeader, METH_O, "\n"
		":rtype: :py:class:`FptIndexHeader`\n"
		":return: a pointer to the index header containing size info etc.\n"
		""},
	 { "new_FastSearch", _wrap_new_FastSearch, METH_NOARGS, NULL},
	 { "FastSearch_swigregister", FastSearch_swigregister, METH_O, NULL},
	 { "FastSearch_swiginit", FastSearch_swiginit, METH_VARARGS, NULL},
	 { "new_FastSearchIndexer", _wrap_new_FastSearchIndexer, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor with a new index\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructor using existing index\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Constructor using existing index\n"
		""},
	 { "delete_FastSearchIndexer", _wrap_delete_FastSearchIndexer, METH_O, NULL},
	 { "FastSearchIndexer_Add", _wrap_FastSearchIndexer_Add, METH_VARARGS, "Called for each object"},
	 { "FastSearchIndexer_swigregister", FastSearchIndexer_swigregister, METH_O, NULL},
	 { "FastSearchIndexer_swiginit", FastSearchIndexer_swiginit, METH_VARARGS, NULL},
	 { "OBDescriptor_Default", _wrap_OBDescriptor_Default, METH_NOARGS, NULL},
	 { "new_OBDescriptor", _wrap_new_OBDescriptor, METH_VARARGS, NULL},
	 { "OBDescriptor_FindType", _wrap_OBDescriptor_FindType, METH_O, NULL},
	 { "OBDescriptor_Predict", _wrap_OBDescriptor_Predict, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the value of a numeric descriptor\n"
		""},
	 { "OBDescriptor_PredictAndSave", _wrap_OBDescriptor_PredictAndSave, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the value of the descriptor and adds it to the object's OBPairData\n"
		""},
	 { "OBDescriptor_GetStringValue", _wrap_OBDescriptor_GetStringValue, METH_VARARGS, "Provides a string value for non-numeric descriptors and returns NaN, or a string representation and returns a numeric value"},
	 { "OBDescriptor_Compare", _wrap_OBDescriptor_Compare, METH_VARARGS, "Parses the filter stream for a relational expression and returns its result when applied to the chemical object"},
	 { "OBDescriptor_Display", _wrap_OBDescriptor_Display, METH_VARARGS, "\n"
		"Write information on a plugin class to the string txt.\n"
		"If the parameter is a descriptor ID, displays the verbose description for that descriptor only\n"
		" e.g. babel -L descriptors HBA1\n"
		""},
	 { "OBDescriptor_Order", _wrap_OBDescriptor_Order, METH_VARARGS, "\n"
		"Comparison of the values of the descriptor. Used in sorting.\n"
		"Descriptors may use more complicated ordering than this default (e.g.InChIFilter)\n"
		""},
	 { "OBDescriptor_FilterCompare", _wrap_OBDescriptor_FilterCompare, METH_VARARGS, "Interprets the --filter option string and returns the combined result of all the comparisons it contains"},
	 { "OBDescriptor_AddProperties", _wrap_OBDescriptor_AddProperties, METH_VARARGS, "Reads list of descriptor IDs and calls PredictAndSave() for each."},
	 { "OBDescriptor_DeleteProperties", _wrap_OBDescriptor_DeleteProperties, METH_VARARGS, "Deletes all the OBPairDatas whose attribute names are in the list (if they exist)."},
	 { "OBDescriptor_GetValues", _wrap_OBDescriptor_GetValues, METH_VARARGS, "\n"
		"Reads list of descriptor IDs and OBPairData names and returns a list of values,\n"
		"each precede by a space or the first character in the list if it is whitespace or punctuation.\n"
		""},
	 { "OBDescriptor_GetIdentifier", _wrap_OBDescriptor_GetIdentifier, METH_O, "Read an identifier and its parameter from the filter string."},
	 { "delete_OBDescriptor", _wrap_delete_OBDescriptor, METH_O, NULL},
	 { "OBDescriptor_swigregister", OBDescriptor_swigregister, METH_O, NULL},
	 { "OBDescriptor_swiginit", OBDescriptor_swiginit, METH_VARARGS, NULL},
	 { "new_LineSearchType", _wrap_new_LineSearchType, METH_NOARGS, NULL},
	 { "delete_LineSearchType", _wrap_delete_LineSearchType, METH_O, NULL},
	 { "LineSearchType_swigregister", LineSearchType_swigregister, METH_O, NULL},
	 { "LineSearchType_swiginit", LineSearchType_swiginit, METH_VARARGS, NULL},
	 { "OBFFParameter_a_set", _wrap_OBFFParameter_a_set, METH_VARARGS, "Used to store integer atom types"},
	 { "OBFFParameter_a_get", _wrap_OBFFParameter_a_get, METH_O, "Used to store integer atom types"},
	 { "OBFFParameter_b_set", _wrap_OBFFParameter_b_set, METH_VARARGS, NULL},
	 { "OBFFParameter_b_get", _wrap_OBFFParameter_b_get, METH_O, NULL},
	 { "OBFFParameter_c_set", _wrap_OBFFParameter_c_set, METH_VARARGS, NULL},
	 { "OBFFParameter_c_get", _wrap_OBFFParameter_c_get, METH_O, NULL},
	 { "OBFFParameter_d_set", _wrap_OBFFParameter_d_set, METH_VARARGS, NULL},
	 { "OBFFParameter_d_get", _wrap_OBFFParameter_d_get, METH_O, NULL},
	 { "OBFFParameter__a_set", _wrap_OBFFParameter__a_set, METH_VARARGS, "used to store string atom types"},
	 { "OBFFParameter__a_get", _wrap_OBFFParameter__a_get, METH_O, "used to store string atom types"},
	 { "OBFFParameter__b_set", _wrap_OBFFParameter__b_set, METH_VARARGS, NULL},
	 { "OBFFParameter__b_get", _wrap_OBFFParameter__b_get, METH_O, NULL},
	 { "OBFFParameter__c_set", _wrap_OBFFParameter__c_set, METH_VARARGS, NULL},
	 { "OBFFParameter__c_get", _wrap_OBFFParameter__c_get, METH_O, NULL},
	 { "OBFFParameter__d_set", _wrap_OBFFParameter__d_set, METH_VARARGS, NULL},
	 { "OBFFParameter__d_get", _wrap_OBFFParameter__d_get, METH_O, NULL},
	 { "OBFFParameter__ipar_set", _wrap_OBFFParameter__ipar_set, METH_VARARGS, "Used to store integer type parameters (bondtypes, multiplicity, ...)"},
	 { "OBFFParameter__ipar_get", _wrap_OBFFParameter__ipar_get, METH_O, "Used to store integer type parameters (bondtypes, multiplicity, ...)"},
	 { "OBFFParameter__dpar_set", _wrap_OBFFParameter__dpar_set, METH_VARARGS, "Used to store double type parameters (force constants, bond lengths, angles, ...)"},
	 { "OBFFParameter__dpar_get", _wrap_OBFFParameter__dpar_get, METH_O, "Used to store double type parameters (force constants, bond lengths, angles, ...)"},
	 { "OBFFParameter_clear", _wrap_OBFFParameter_clear, METH_O, "Reset the atom types and set all parameters to zero"},
	 { "new_OBFFParameter", _wrap_new_OBFFParameter, METH_NOARGS, NULL},
	 { "delete_OBFFParameter", _wrap_delete_OBFFParameter, METH_O, NULL},
	 { "OBFFParameter_swigregister", OBFFParameter_swigregister, METH_O, NULL},
	 { "OBFFParameter_swiginit", OBFFParameter_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation2_energy_set", _wrap_OBFFCalculation2_energy_set, METH_VARARGS, "Used to store the energy for this OBFFCalculation"},
	 { "OBFFCalculation2_energy_get", _wrap_OBFFCalculation2_energy_get, METH_O, "Used to store the energy for this OBFFCalculation"},
	 { "OBFFCalculation2_a_set", _wrap_OBFFCalculation2_a_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_a_get", _wrap_OBFFCalculation2_a_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_b_set", _wrap_OBFFCalculation2_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_b_get", _wrap_OBFFCalculation2_b_get, METH_O, NULL},
	 { "OBFFCalculation2_idx_a_set", _wrap_OBFFCalculation2_idx_a_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_idx_a_get", _wrap_OBFFCalculation2_idx_a_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_idx_b_set", _wrap_OBFFCalculation2_idx_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_idx_b_get", _wrap_OBFFCalculation2_idx_b_get, METH_O, NULL},
	 { "OBFFCalculation2_pos_a_set", _wrap_OBFFCalculation2_pos_a_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation2_pos_a_get", _wrap_OBFFCalculation2_pos_a_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation2_pos_b_set", _wrap_OBFFCalculation2_pos_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_pos_b_get", _wrap_OBFFCalculation2_pos_b_get, METH_O, NULL},
	 { "OBFFCalculation2_force_a_set", _wrap_OBFFCalculation2_force_a_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation2_force_a_get", _wrap_OBFFCalculation2_force_a_get, METH_O, "Pointer to atom forces"},
	 { "OBFFCalculation2_force_b_set", _wrap_OBFFCalculation2_force_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_force_b_get", _wrap_OBFFCalculation2_force_b_get, METH_O, NULL},
	 { "delete_OBFFCalculation2", _wrap_delete_OBFFCalculation2, METH_O, "Destructor"},
	 { "OBFFCalculation2_SetupPointers", _wrap_OBFFCalculation2_SetupPointers, METH_O, "\n"
		":rtype: void\n"
		":return: Setup pointers to atom positions and forces (To be called  while setting up calculations). Sets optimized to true.\n"
		""},
	 { "new_OBFFCalculation2", _wrap_new_OBFFCalculation2, METH_NOARGS, NULL},
	 { "OBFFCalculation2_swigregister", OBFFCalculation2_swigregister, METH_O, NULL},
	 { "OBFFCalculation2_swiginit", OBFFCalculation2_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation3_c_set", _wrap_OBFFCalculation3_c_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_c_get", _wrap_OBFFCalculation3_c_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_idx_c_set", _wrap_OBFFCalculation3_idx_c_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_idx_c_get", _wrap_OBFFCalculation3_idx_c_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_pos_c_set", _wrap_OBFFCalculation3_pos_c_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation3_pos_c_get", _wrap_OBFFCalculation3_pos_c_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation3_force_c_set", _wrap_OBFFCalculation3_force_c_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation3_force_c_get", _wrap_OBFFCalculation3_force_c_get, METH_O, "Pointer to atom forces"},
	 { "delete_OBFFCalculation3", _wrap_delete_OBFFCalculation3, METH_O, "Destructor"},
	 { "new_OBFFCalculation3", _wrap_new_OBFFCalculation3, METH_NOARGS, NULL},
	 { "OBFFCalculation3_swigregister", OBFFCalculation3_swigregister, METH_O, NULL},
	 { "OBFFCalculation3_swiginit", OBFFCalculation3_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation4_d_set", _wrap_OBFFCalculation4_d_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_d_get", _wrap_OBFFCalculation4_d_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_idx_d_set", _wrap_OBFFCalculation4_idx_d_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_idx_d_get", _wrap_OBFFCalculation4_idx_d_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_pos_d_set", _wrap_OBFFCalculation4_pos_d_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation4_pos_d_get", _wrap_OBFFCalculation4_pos_d_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation4_force_d_set", _wrap_OBFFCalculation4_force_d_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation4_force_d_get", _wrap_OBFFCalculation4_force_d_get, METH_O, "Pointer to atom forces"},
	 { "delete_OBFFCalculation4", _wrap_delete_OBFFCalculation4, METH_O, "Destructor"},
	 { "new_OBFFCalculation4", _wrap_new_OBFFCalculation4, METH_NOARGS, NULL},
	 { "OBFFCalculation4_swigregister", OBFFCalculation4_swigregister, METH_O, NULL},
	 { "OBFFCalculation4_swiginit", OBFFCalculation4_swiginit, METH_VARARGS, NULL},
	 { "OBFFConstraint_factor_set", _wrap_OBFFConstraint_factor_set, METH_VARARGS, "Used to store the contraint energy for this OBFFConstraint"},
	 { "OBFFConstraint_factor_get", _wrap_OBFFConstraint_factor_get, METH_O, "Used to store the contraint energy for this OBFFConstraint"},
	 { "OBFFConstraint_constraint_value_set", _wrap_OBFFConstraint_constraint_value_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_constraint_value_get", _wrap_OBFFConstraint_constraint_value_get, METH_O, NULL},
	 { "OBFFConstraint_rab0_set", _wrap_OBFFConstraint_rab0_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_rab0_get", _wrap_OBFFConstraint_rab0_get, METH_O, NULL},
	 { "OBFFConstraint_rbc0_set", _wrap_OBFFConstraint_rbc0_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_rbc0_get", _wrap_OBFFConstraint_rbc0_get, METH_O, NULL},
	 { "OBFFConstraint_type_set", _wrap_OBFFConstraint_type_set, METH_VARARGS, "Used to store the contraint type for this OBFFConstraint"},
	 { "OBFFConstraint_type_get", _wrap_OBFFConstraint_type_get, METH_O, "Used to store the contraint type for this OBFFConstraint"},
	 { "OBFFConstraint_ia_set", _wrap_OBFFConstraint_ia_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ia_get", _wrap_OBFFConstraint_ia_get, METH_O, NULL},
	 { "OBFFConstraint_ib_set", _wrap_OBFFConstraint_ib_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ib_get", _wrap_OBFFConstraint_ib_get, METH_O, NULL},
	 { "OBFFConstraint_ic_set", _wrap_OBFFConstraint_ic_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ic_get", _wrap_OBFFConstraint_ic_get, METH_O, NULL},
	 { "OBFFConstraint_id_set", _wrap_OBFFConstraint_id_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_id_get", _wrap_OBFFConstraint_id_get, METH_O, NULL},
	 { "OBFFConstraint_a_set", _wrap_OBFFConstraint_a_set, METH_VARARGS, "Used to store the atoms for this OBFFCostraint"},
	 { "OBFFConstraint_a_get", _wrap_OBFFConstraint_a_get, METH_O, "Used to store the atoms for this OBFFCostraint"},
	 { "OBFFConstraint_b_set", _wrap_OBFFConstraint_b_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_b_get", _wrap_OBFFConstraint_b_get, METH_O, NULL},
	 { "OBFFConstraint_c_set", _wrap_OBFFConstraint_c_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_c_get", _wrap_OBFFConstraint_c_get, METH_O, NULL},
	 { "OBFFConstraint_d_set", _wrap_OBFFConstraint_d_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_d_get", _wrap_OBFFConstraint_d_get, METH_O, NULL},
	 { "OBFFConstraint_grada_set", _wrap_OBFFConstraint_grada_set, METH_VARARGS, "Used to store the gradients for this OBFFCalculation"},
	 { "OBFFConstraint_grada_get", _wrap_OBFFConstraint_grada_get, METH_O, "Used to store the gradients for this OBFFCalculation"},
	 { "OBFFConstraint_gradb_set", _wrap_OBFFConstraint_gradb_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradb_get", _wrap_OBFFConstraint_gradb_get, METH_O, NULL},
	 { "OBFFConstraint_gradc_set", _wrap_OBFFConstraint_gradc_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradc_get", _wrap_OBFFConstraint_gradc_get, METH_O, NULL},
	 { "OBFFConstraint_gradd_set", _wrap_OBFFConstraint_gradd_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradd_get", _wrap_OBFFConstraint_gradd_get, METH_O, NULL},
	 { "new_OBFFConstraint", _wrap_new_OBFFConstraint, METH_NOARGS, "Constructor"},
	 { "delete_OBFFConstraint", _wrap_delete_OBFFConstraint, METH_O, "Destructor"},
	 { "OBFFConstraint_GetGradient", _wrap_OBFFConstraint_GetGradient, METH_VARARGS, NULL},
	 { "OBFFConstraint_swigregister", OBFFConstraint_swigregister, METH_O, NULL},
	 { "OBFFConstraint_swiginit", OBFFConstraint_swiginit, METH_VARARGS, NULL},
	 { "new_OBFFConstraints", _wrap_new_OBFFConstraints, METH_NOARGS, "Constructor"},
	 { "delete_OBFFConstraints", _wrap_delete_OBFFConstraints, METH_O, "Destructor"},
	 { "OBFFConstraints_Clear", _wrap_OBFFConstraints_Clear, METH_O, "Clear all constraints"},
	 { "OBFFConstraints_GetConstraintEnergy", _wrap_OBFFConstraints_GetConstraintEnergy, METH_O, "Get the constraint energy"},
	 { "OBFFConstraints_GetGradient", _wrap_OBFFConstraints_GetGradient, METH_VARARGS, "Get the constraint gradient for atom with index a"},
	 { "OBFFConstraints_Setup", _wrap_OBFFConstraints_Setup, METH_VARARGS, "\n"
		"Translate indices to OBAtom* objects, this function is called from OBForceField::Setup,\n"
		"this function doesn't have to be called from anywhere else.\n"
		""},
	 { "OBFFConstraints_SetFactor", _wrap_OBFFConstraints_SetFactor, METH_VARARGS, "Set Constraint factor"},
	 { "OBFFConstraints_AddIgnore", _wrap_OBFFConstraints_AddIgnore, METH_VARARGS, "Ignore the atom while setting up calculations"},
	 { "OBFFConstraints_AddAtomConstraint", _wrap_OBFFConstraints_AddAtomConstraint, METH_VARARGS, "Fix the position of an atom"},
	 { "OBFFConstraints_AddAtomXConstraint", _wrap_OBFFConstraints_AddAtomXConstraint, METH_VARARGS, "Fix the x coordinate of the atom position"},
	 { "OBFFConstraints_AddAtomYConstraint", _wrap_OBFFConstraints_AddAtomYConstraint, METH_VARARGS, "Fix the y coordinate of the atom position"},
	 { "OBFFConstraints_AddAtomZConstraint", _wrap_OBFFConstraints_AddAtomZConstraint, METH_VARARGS, "Fix the z coordinate of the atom position"},
	 { "OBFFConstraints_AddDistanceConstraint", _wrap_OBFFConstraints_AddDistanceConstraint, METH_VARARGS, "Constrain the bond length a-b"},
	 { "OBFFConstraints_AddAngleConstraint", _wrap_OBFFConstraints_AddAngleConstraint, METH_VARARGS, "Constrain the angle a-b-c"},
	 { "OBFFConstraints_AddTorsionConstraint", _wrap_OBFFConstraints_AddTorsionConstraint, METH_VARARGS, "Constrain the torsion angle a-b-c-d"},
	 { "OBFFConstraints_DeleteConstraint", _wrap_OBFFConstraints_DeleteConstraint, METH_VARARGS, "Delete a constraint Title: index constraint index"},
	 { "OBFFConstraints_GetFactor", _wrap_OBFFConstraints_GetFactor, METH_O, "Get Constraint factor"},
	 { "OBFFConstraints_Size", _wrap_OBFFConstraints_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the number of set constraints\n"
		""},
	 { "OBFFConstraints_GetConstraintType", _wrap_OBFFConstraints_GetConstraintType, METH_VARARGS, "\n"
		"The following constraint types are known: OBFF_CONST_IGNORE (ignore\n"
		"the atom while setting up calculations, forcefield implementations\n"
		"need to check this value in their setup function), OBFF_CONST_ATOM\n"
		"(fix atom position), OBFF_CONST_ATOM_X (fix x coordinate),\n"
		"OBFF_CONST_ATOM_Y (fix y coordinate), OBFF_CONST_ATOM_Z (fix z\n"
		"coordinate), OBFF_CONST_BOND (constrain bond length), OBFF_CONST_ANGLE\n"
		"(constrain angle), OBFF_CONST_TORSION (constrain torsion angle)\n"
		":rtype: int\n"
		":return: the constraint type\n"
		""},
	 { "OBFFConstraints_GetConstraintValue", _wrap_OBFFConstraints_GetConstraintValue, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: The constraint value, this can be a bond length, angle or\n"
		"     torsion angle depending on the constraint type.\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomA", _wrap_OBFFConstraints_GetConstraintAtomA, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom a (or fixed atom) Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomB", _wrap_OBFFConstraints_GetConstraintAtomB, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom b Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomC", _wrap_OBFFConstraints_GetConstraintAtomC, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom c Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomD", _wrap_OBFFConstraints_GetConstraintAtomD, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom d Title: index constraint index\n"
		""},
	 { "OBFFConstraints_IsIgnored", _wrap_OBFFConstraints_IsIgnored, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this atom is ignored Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsFixed", _wrap_OBFFConstraints_IsFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsXFixed", _wrap_OBFFConstraints_IsXFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the x coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsYFixed", _wrap_OBFFConstraints_IsYFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the y coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsZFixed", _wrap_OBFFConstraints_IsZFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the z coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_GetIgnoredBitVec", _wrap_OBFFConstraints_GetIgnoredBitVec, METH_O, "\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: the ignored atom indexes as bitvec. (used in OBForceField::Setup() to determine if a call to OBForceField::SetupCalculations() is needed)\n"
		""},
	 { "OBFFConstraints_GetFixedBitVec", _wrap_OBFFConstraints_GetFixedBitVec, METH_O, "\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: the fixed atom indexes as bitvec. (used in OBForceField::SystematicRotorSearch() and similar)\n"
		""},
	 { "OBFFConstraints_swigregister", OBFFConstraints_swigregister, METH_O, NULL},
	 { "OBFFConstraints_swiginit", OBFFConstraints_swiginit, METH_VARARGS, NULL},
	 { "OBForceField_Default", _wrap_OBForceField_Default, METH_NOARGS, NULL},
	 { "OBForceField_FindType", _wrap_OBForceField_FindType, METH_O, NULL},
	 { "OBForceField_MakeNewInstance", _wrap_OBForceField_MakeNewInstance, METH_O, "\n"
		"Clone the current instance. May be desirable in multithreaded environments,\n"
		"Should be deleted after use\n"
		""},
	 { "delete_OBForceField", _wrap_delete_OBForceField, METH_O, "Destructor (defined in forcefield.cpp because _lbfgsState's pimpl type LBFGSState is incomplete in this header)."},
	 { "OBForceField_FindForceField", _wrap_OBForceField_FindForceField, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":type ID: string\n"
		":param ID: forcefield id (Ghemical, MMFF94, UFF, ...).\n"
		":rtype: :py:class:`OBForceField`\n"
		":return: A pointer to a forcefield (the default if ID is empty), or NULL if not available.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":type ID: string\n"
		":param ID: forcefield id (Ghemical, MMFF94, UFF, ...).\n"
		":rtype: :py:class:`OBForceField`\n"
		":return: A pointer to a forcefield (the default if ID is empty), or NULL if not available.\n"
		""},
	 { "OBForceField_SetParameterFile", _wrap_OBForceField_SetParameterFile, METH_VARARGS, NULL},
	 { "OBForceField_GetUnit", _wrap_OBForceField_GetUnit, METH_O, "\n"
		":rtype: string\n"
		":return: The unit (kcal/mol, kJ/mol, ...) in which the energy is expressed as std::string.\n"
		""},
	 { "OBForceField_HasAnalyticalGradients", _wrap_OBForceField_HasAnalyticalGradients, METH_O, NULL},
	 { "OBForceField_Setup", _wrap_OBForceField_Setup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Setup the forcefield for mol (assigns atom types, charges, etc.). Keep current constraints.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object that contains the atoms and bonds.\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Setup the forcefield for mol (assigns atom types, charges, etc.). Use new constraints.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object that contains the atoms and bonds.\n"
		":type constraints: :py:class:`OBFFConstraints`\n"
		":param constraints: The OBFFConstraints object that contains the constraints.\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_ParseParamFile", _wrap_OBForceField_ParseParamFile, METH_O, "\n"
		"Load the parameters (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetTypes", _wrap_OBForceField_SetTypes, METH_O, "\n"
		"Set the atom types (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetFormalCharges", _wrap_OBForceField_SetFormalCharges, METH_O, "\n"
		"Set the formal charges (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetPartialCharges", _wrap_OBForceField_SetPartialCharges, METH_O, "\n"
		"Set the partial charges (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetupCalculations", _wrap_OBForceField_SetupCalculations, METH_O, "\n"
		"Setup the calculations (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetupPointers", _wrap_OBForceField_SetupPointers, METH_O, "\n"
		"Setup the pointers to the atom positions in the OBFFCalculation objects. This method\n"
		"will iterate over all the calculations and call SetupPointers for each one. (This\n"
		"function should be implemented by the individual force field implementations).\n"
		""},
	 { "OBForceField_IsSetupNeeded", _wrap_OBForceField_IsSetupNeeded, METH_VARARGS, "\n"
		"Compare the internal forcefield OBMol object to mol. If the two have the\n"
		"same number of atoms and bonds, and all atomic numbers are the same,\n"
		"this function returns false, and no call to Setup is needed.\n"
		":rtype: boolean\n"
		":return: True if Setup needs to be called.\n"
		""},
	 { "OBForceField_GetAtomTypes", _wrap_OBForceField_GetAtomTypes, METH_VARARGS, "\n"
		"Get the force atom types. The atom types will be added to\n"
		"the atoms of mol as OBPairData. The attribute will be \"FFAtomType\".\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    ...\n"
		"    pFF->Setup(&mol);\n"
		"    pFF->GetAtomTypes(&mol);\n"
		"    FOR_ATOMS_OF_MOL (atom, mol) {\n"
		"      OBPairData *type = (OBPairData*) atom->GetData(\"FFAtomType\");\n"
		"      if (type)\n"
		"        cout << \"atom \" << atom->GetIdx() << \" : \" << type->GetValue() << endl;\n"
		"    }\n"
		"    ...\n"
		""},
	 { "OBForceField_GetPartialCharges", _wrap_OBForceField_GetPartialCharges, METH_VARARGS, "\n"
		"Get the force field formal charges. The formal charges will be added to\n"
		"the atoms of mol as OBPairData. The attribute will be \"FFPartialCharge\".\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    ...\n"
		"    pFF->Setup(&mol);\n"
		"    pFF->GetPartialCharges(&mol);\n"
		"    FOR_ATOMS_OF_MOL (atom, mol) {\n"
		"      OBPairData *chg = (OBPairData*) atom->GetData(\"FFPartialCharge\");\n"
		"      if (chg)\n"
		"        cout << \"atom \" << atom->GetIdx() << \" : \" << chg->GetValue() << endl;\n"
		"    }\n"
		"    ...\n"
		""},
	 { "OBForceField_GetCoordinates", _wrap_OBForceField_GetCoordinates, METH_VARARGS, "\n"
		"Get coordinates for current conformer and attach OBConformerData with energies, forces, ... to mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates to (from OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_UpdateCoordinates", _wrap_OBForceField_UpdateCoordinates, METH_VARARGS, "Deprecated: Use GetCooordinates instead."},
	 { "OBForceField_GetConformers", _wrap_OBForceField_GetConformers, METH_VARARGS, "\n"
		"Get coordinates for all conformers and attach OBConformerData with energies, forces, ... to mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates to (from OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_UpdateConformers", _wrap_OBForceField_UpdateConformers, METH_VARARGS, "Deprecated: Use GetConformers instead."},
	 { "OBForceField_SetCoordinates", _wrap_OBForceField_SetCoordinates, METH_VARARGS, "\n"
		"Set coordinates for current conformer.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: the OBMol object to copy the coordinates from (to OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: true if successful.\n"
		""},
	 { "OBForceField_SetConformers", _wrap_OBForceField_SetConformers, METH_VARARGS, "\n"
		"Set coordinates for all conformers.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates from (to OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_GetGrid", _wrap_OBForceField_GetGrid, METH_VARARGS, "\n"
		"Create a grid with spacing ``step`` and ``padding``. Place a probe atom of type probe at every grid point,\n"
		"calculate the energy and store it in the grid. These grids can then be used to create isosurfaces to\n"
		"identify locations where the probe atom has favourable interactions with the molecule.\n"
		":type step: float\n"
		":param step: The grid step size in A..\n"
		":type padding: float\n"
		":param padding: The padding for the grid in A.\n"
		":type type: string\n"
		":param type: The force field atom type for the probe.\n"
		":type pchg: float\n"
		":param pchg: The partial charge for the probe atom.\n"
		":rtype: :py:class:`OBGridData`\n"
		":return: Pointer to the grid constaining the results.\n"
		""},
	 { "OBForceField_AddIntraGroup", _wrap_OBForceField_AddIntraGroup, METH_VARARGS, "\n"
		"Enable intra-molecular interactions for group (bonds, angles, strbnd, torsions, oop).\n"
		"This function should be called before Setup().\n"
		":type group: :py:class:`OBBitVec`\n"
		":param group: OBBitVec with bits set for the indexes of the atoms which make up the group.\n"
		""},
	 { "OBForceField_AddInterGroup", _wrap_OBForceField_AddInterGroup, METH_VARARGS, "\n"
		"Enable inter-molecular interactions for group (non-bonded: vdw & ele).\n"
		"This function should be called before Setup().\n"
		":type group: :py:class:`OBBitVec`\n"
		":param group: OBBitVec with bits set for the indexes of the atoms which make up the group.\n"
		""},
	 { "OBForceField_AddInterGroups", _wrap_OBForceField_AddInterGroups, METH_VARARGS, "\n"
		"Enable inter-molecular interactions between group1 and group2 (non-bonded: vdw & ele).\n"
		"Note that this function doesn't enable bonded interactions in either group. Non-bonded\n"
		"interactions in the groups itself are also not enabled.\n"
		"This function should be called before Setup().\n"
		":type group1: :py:class:`OBBitVec`\n"
		":param group1: OBBitVec with bits set for the indexes of the atoms which make up the first group.\n"
		":type group2: :py:class:`OBBitVec`\n"
		":param group2: OBBitVec with bits set for the indexes of the atoms which make up the second group.\n"
		""},
	 { "OBForceField_ClearGroups", _wrap_OBForceField_ClearGroups, METH_O, "Clear all previously specified groups."},
	 { "OBForceField_HasGroups", _wrap_OBForceField_HasGroups, METH_O, "\n"
		":rtype: boolean\n"
		":return: true if there are groups.\n"
		""},
	 { "OBForceField_EnableCutOff", _wrap_OBForceField_EnableCutOff, METH_VARARGS, "\n"
		"Enable or disable Cut-offs. Cut-offs are disabled by default.\n"
		":type enable: boolean\n"
		":param enable: Enable when true, disable when false.\n"
		""},
	 { "OBForceField_IsCutOffEnabled", _wrap_OBForceField_IsCutOffEnabled, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if Cut-off distances are used.\n"
		""},
	 { "OBForceField_SetVDWCutOff", _wrap_OBForceField_SetVDWCutOff, METH_VARARGS, "\n"
		"Set the VDW cut-off distance to r. Note that this does not enable cut-off distances.\n"
		":type r: float\n"
		":param r: The VDW cut-off distance to be used in A.\n"
		""},
	 { "OBForceField_GetVDWCutOff", _wrap_OBForceField_GetVDWCutOff, METH_O, "\n"
		"Get the VDW cut-off distance.\n"
		":rtype: float\n"
		":return: The VDW cut-off distance in A.\n"
		""},
	 { "OBForceField_SetElectrostaticCutOff", _wrap_OBForceField_SetElectrostaticCutOff, METH_VARARGS, "\n"
		"Set the Electrostatic cut-off distance to r. Note that this does not\n"
		"enable cut-off distances.\n"
		":type r: float\n"
		":param r: The electrostatic cut-off distance to be used in A.\n"
		""},
	 { "OBForceField_GetElectrostaticCutOff", _wrap_OBForceField_GetElectrostaticCutOff, METH_O, "\n"
		"Get the Electrostatic cut-off distance.\n"
		":rtype: float\n"
		":return: The electrostatic cut-off distance in A.\n"
		""},
	 { "OBForceField_SetDielectricConstant", _wrap_OBForceField_SetDielectricConstant, METH_VARARGS, "\n"
		" Set the dielectric constant for electrostatic SetupCalculations\n"
		":type epsilon: float\n"
		":param epsilon: The relative permittivity to use (default = 1.0)\n"
		""},
	 { "OBForceField_GetDielectricConstant", _wrap_OBForceField_GetDielectricConstant, METH_O, NULL},
	 { "OBForceField_SetUpdateFrequency", _wrap_OBForceField_SetUpdateFrequency, METH_VARARGS, "\n"
		"Set the frequency by which non-bonded pairs are updated. Values from 10 to 20\n"
		"are recommended. Too low will decrease performance, too high will cause\n"
		"non-bonded interactions within cut-off not to be calculated.\n"
		":type f: int\n"
		":param f: The pair list update frequency.\n"
		""},
	 { "OBForceField_GetUpdateFrequency", _wrap_OBForceField_GetUpdateFrequency, METH_O, "\n"
		"Get the frequency by which non-bonded pairs are updated.\n"
		":rtype: int\n"
		":return: The pair list update frequency.\n"
		""},
	 { "OBForceField_UpdatePairsSimple", _wrap_OBForceField_UpdatePairsSimple, METH_O, "\n"
		"Set the bits in _vdwpairs and _elepairs to 1 for interactions that\n"
		"are within cut-off distance. This function is called in minimizing\n"
		"algorithms such as SteepestDescent and ConjugateGradients.\n"
		""},
	 { "OBForceField_GetNumPairs", _wrap_OBForceField_GetNumPairs, METH_O, "\n"
		"Get the number of non-bonded pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of atom pairs (ignores cutoff)\n"
		""},
	 { "OBForceField_GetNumElectrostaticPairs", _wrap_OBForceField_GetNumElectrostaticPairs, METH_O, "\n"
		"Get the number of enabled electrostatic pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of pairs currently enabled (within cut-off distance)\n"
		""},
	 { "OBForceField_GetNumVDWPairs", _wrap_OBForceField_GetNumVDWPairs, METH_O, "\n"
		"Get the number of enabled VDW pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of pairs currently enabled (within cut-off distance)\n"
		""},
	 { "OBForceField_EnableAllPairs", _wrap_OBForceField_EnableAllPairs, METH_O, "\n"
		"Set bits in range 0..._numpairs-1 to 1. Using this means there will\n"
		"be no cut-off. (not-working: see code for more information.\n"
		""},
	 { "OBForceField_GetGradient", _wrap_OBForceField_GetGradient, METH_VARARGS, "Get the pointer to the gradients"},
	 { "OBForceField_GetGradientPtr", _wrap_OBForceField_GetGradientPtr, METH_O, "Get the pointer to the gradients"},
	 { "OBForceField_Energy", _wrap_OBForceField_Energy, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Total energy.\n"
		" Title: Output to log:\n"
		"  OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"  OBFF_LOGLVL_LOW:    none\n"
		"\n"
		"  OBFF_LOGLVL_MEDIUM: energy for individual energy terms\n"
		"\n"
		"  OBFF_LOGLVL_HIGH:   energy for individual energy interactions\n"
		""},
	 { "OBForceField_E_Bond", _wrap_OBForceField_E_Bond, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Bond stretching energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Angle", _wrap_OBForceField_E_Angle, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Angle bending energy.\n"
		"Title: Output to log:\n"
		" see Energy()\n"
		""},
	 { "OBForceField_E_StrBnd", _wrap_OBForceField_E_StrBnd, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Stretch bending energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Torsion", _wrap_OBForceField_E_Torsion, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Torsional energy.\n"
		"  Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_OOP", _wrap_OBForceField_E_OOP, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Out-Of-Plane bending energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_VDW", _wrap_OBForceField_E_VDW, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Van der Waals energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Electrostatic", _wrap_OBForceField_E_Electrostatic, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Electrostatic energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_PrintTypes", _wrap_OBForceField_PrintTypes, METH_O, "Print the atom types to the log."},
	 { "OBForceField_PrintFormalCharges", _wrap_OBForceField_PrintFormalCharges, METH_O, "\n"
		"Print the formal charges to the log (atom.GetPartialCharge(),\n"
		"MMFF94 FC's are not always int).\n"
		""},
	 { "OBForceField_PrintPartialCharges", _wrap_OBForceField_PrintPartialCharges, METH_O, "Print the partial charges to the log."},
	 { "OBForceField_PrintVelocities", _wrap_OBForceField_PrintVelocities, METH_O, "Print the velocities to the log."},
	 { "OBForceField_SetLogFile", _wrap_OBForceField_SetLogFile, METH_VARARGS, "\n"
		"Set the stream for logging (can also be &cout for logging to screen).\n"
		":type pos: std::ostream\n"
		":param pos: Stream (when pos is 0, std::cout wil be used).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_SetLogLevel", _wrap_OBForceField_SetLogLevel, METH_VARARGS, "\n"
		"Set the log level (OBFF_LOGLVL_NONE, OBFF_LOGLVL_LOW, OBFF_LOGLVL_MEDIUM, OBFF_LOGLVL_HIGH).\n"
		"Inline if statements for logging are available:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    #define IF_OBFF_LOGLVL_LOW    if(_loglvl >= OBFF_LOGLVL_LOW)\n"
		"    #define IF_OBFF_LOGLVL_MEDIUM if(_loglvl >= OBFF_LOGLVL_MEDIUM)\n"
		"    #define IF_OBFF_LOGLVL_HIGH   if(_loglvl >= OBFF_LOGLVL_HIGH)\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    SetLogLevel(OBFF_LOGLVL_MEDIUM);\n"
		"    IF_OBFF_LOGLVL_HIGH {\n"
		"      OBFFLog(\"this text will NOT be logged...\\n\");\n"
		"    }\n"
		"\n"
		"    IF_OBFF_LOGLVL_LOW {\n"
		"      OBFFLog\"this text will be logged...\\n\");\n"
		"    }\n"
		"\n"
		"    IF_OBFF_LOGLVL_MEDIUM {\n"
		"      OBFFLog(\"this text will also be logged...\\n\");\n"
		"    }\n"
		""},
	 { "OBForceField_GetLogLevel", _wrap_OBForceField_GetLogLevel, METH_O, "\n"
		":rtype: int\n"
		":return: The log level.\n"
		""},
	 { "OBForceField_OBFFLog", _wrap_OBForceField_OBFFLog, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Print msg to the logfile.\n"
		":type msg: string\n"
		":param msg: The message to print.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Print msg to the logfile.\n"
		":type msg: string\n"
		":param msg: The message to print.\n"
		""},
	 { "OBForceField_DistanceGeometry", _wrap_OBForceField_DistanceGeometry, METH_O, "Generate coordinates for the molecule (distance geometry) Deprecated: Use OBDistanceGeometry class instead"},
	 { "OBForceField_SystematicRotorSearch", _wrap_OBForceField_SystematicRotorSearch, METH_VARARGS, "\n"
		"Generate conformers for the molecule (systematicaly rotating torsions).\n"
		"\n"
		"The initial starting structure here is important, this structure should be\n"
		"minimized for the best results. SystematicRotorSearch works by rotating around\n"
		"the rotatable bond in a molecule (see OBRotamerList class). This rotating generates\n"
		"multiple conformers. The energy for all these conformers is then evaluated and the\n"
		"lowest energy conformer is selected.\n"
		"\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for this function will interfere with the output for\n"
		"this function.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		"OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_SystematicRotorSearchInitialize", _wrap_OBForceField_SystematicRotorSearchInitialize, METH_VARARGS, "\n"
		"Generate conformers for the molecule by systematicaly rotating torsions. To be used in combination with\n"
		"SystematicRotorSearchNexConformer().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->SystematicRotorSearchInitialize(300);\n"
		"    while (pFF->SystematicRotorSearchNextConformer(300)) {\n"
		"      // do some updating in your program (show last generated conformer, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, SystematicRotorSearch() is recommended.\n"
		"\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		":rtype: int\n"
		":return: The number of conformers.\n"
		""},
	 { "OBForceField_SystematicRotorSearchNextConformer", _wrap_OBForceField_SystematicRotorSearchNextConformer, METH_VARARGS, "\n"
		"Evaluate the next conformer.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":rtype: boolean\n"
		":return: True if there are more conformers.\n"
		""},
	 { "OBForceField_RandomRotorSearch", _wrap_OBForceField_RandomRotorSearch, METH_VARARGS, "\n"
		"Generate conformers for the molecule (randomly rotating torsions).\n"
		"\n"
		"The initial starting structure here is important, this structure should be\n"
		"minimized for the best results. RandomRotorSearch works by randomly rotating around\n"
		"the rotatable bonds in a molecule (see OBRotamerList class). This rotating generates\n"
		"multiple conformers. The energy for all these conformers is then evaluated and the\n"
		"lowest energy conformer is selected.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization for each conformer.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for this function will interfere with the output for\n"
		"this function.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		"OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_RandomRotorSearchInitialize", _wrap_OBForceField_RandomRotorSearchInitialize, METH_VARARGS, "\n"
		"Generate conformers for the molecule by randomly rotating torsions. To be used in combination with\n"
		"RandomRotorSearchNexConformer().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->RandomRotorSearchInitialize(300);\n"
		"    while (pFF->RandomRotorSearchNextConformer(300)) {\n"
		"      // do some updating in your program (show last generated conformer, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, RandomRotorSearch() is recommended.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		""},
	 { "OBForceField_RandomRotorSearchNextConformer", _wrap_OBForceField_RandomRotorSearchNextConformer, METH_VARARGS, "\n"
		"Evaluate the next conformer.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":rtype: boolean\n"
		":return: True if there are more conformers.\n"
		""},
	 { "OBForceField_WeightedRotorSearch", _wrap_OBForceField_WeightedRotorSearch, METH_VARARGS, "\n"
		" Generate conformers for the molecule (randomly rotating torsions).\n"
		"\n"
		" The initial starting structure here is important, this structure should be\n"
		" minimized for the best results. WeightedRotorSearch works by randomly rotating around\n"
		" the rotatable bonds in a molecule (see OBRotamerList class). Unlike RandomRotorSearch()\n"
		" the random choice of torsions is reweighted based on the energy of the generated conformer.\n"
		" Over time, the generated conformers for each step should become increasingly better.\n"
		" The lowest energy conformer is selected.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search.\n"
		":type geomSteps: int\n"
		":param geomSteps: The number of steps to take during geometry optimization for each conformer.\n"
		" :type sampleRingBonds: boolean, optional\n"
		" :param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"	Title: Output to log:\n"
		" This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"	too much information about the energy calculations needed for this function will interfere with the output for\n"
		"	this function.\n"
		"\n"
		"\n"
		" OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		" OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		" OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		" OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_FastRotorSearch", _wrap_OBForceField_FastRotorSearch, METH_VARARGS, "\n"
		"A fast rotor search to find low energy conformations\n"
		"\n"
		"Iterate over each of the rotors, and set the\n"
		"torsion angle to that which minimizes the energy (while keeping the rest of the molecule\n"
		"fixed). In general (for molecules with more than\n"
		"one rotatable bond), this procedure will not find\n"
		"the global minimum, but it will at least get rid of any bad\n"
		"clashes, and it do so quickly.\n"
		"\n"
		"Torsions closer to the center\n"
		"of the molecule will be optimized first as these most likely\n"
		"to generate large clashes.\n"
		"\n"
		"One possible use of this procedure is to prepare a reasonable 3D structure\n"
		"of a molecule for viewing. Another is to prepare the starting structure\n"
		"for a more systematic rotor search (in which case you should geometry\n"
		"optimize the final structure).\n"
		"\n"
		":type permute: boolean, optional\n"
		":param permute: Whether or not to permute the order of the 4 most central rotors.\n"
		"                   Default is true. This does a more thorough search, but takes 4! = 24 times\n"
		"                   as long.\n"
		"Since: version 2.4\n"
		""},
	 { "OBForceField_DiverseConfGen", _wrap_OBForceField_DiverseConfGen, METH_VARARGS, "Since: version 2.4"},
	 { "OBForceField_SetLineSearchType", _wrap_OBForceField_SetLineSearchType, METH_VARARGS, "\n"
		"Set the LineSearchType. The default type is LineSearchType::Newton2Num.\n"
		":type type: int\n"
		":param type: The LineSearchType to be used in SteepestDescent and ConjugateGradients.\n"
		""},
	 { "OBForceField_GetLineSearchType", _wrap_OBForceField_GetLineSearchType, METH_O, "\n"
		"Get the LineSearchType.\n"
		":rtype: int\n"
		":return: The current LineSearchType.\n"
		""},
	 { "OBForceField_LineSearch", _wrap_OBForceField_LineSearch, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Perform a linesearch starting at atom in direction direction.\n"
		"Deprecated: Current code should use LineSearch(double *, double*) instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Perform a linesearch for the entire molecule in direction ``direction``.\n"
		" This function is called when using LineSearchType::Simple.\n"
		"\n"
		" :type currentCoords: float\n"
		" :param currentCoords: Start coordinates.\n"
		" :type direction: float\n"
		" :param direction: The search direction.\n"
		" :rtype: float\n"
		" :return: alpha, The scale of the step we moved along the direction vector.\n"
		"\n"
		" Title: Output to log:\n"
		" OBFF_LOGLVL_NONE:   none\n"
		"\n"
		" OBFF_LOGLVL_LOW:    none\n"
		"\n"
		" OBFF_LOGLVL_MEDIUM: none\n"
		"\n"
		" OBFF_LOGLVL_HIGH:   none\n"
		""},
	 { "OBForceField_Newton2NumLineSearch", _wrap_OBForceField_Newton2NumLineSearch, METH_VARARGS, "\n"
		"Perform a linesearch for the entire molecule.\n"
		"This function is called when using LineSearchType::Newton2Num.\n"
		"\n"
		":type direction: float\n"
		":param direction: The search direction.\n"
		":rtype: float\n"
		":return: alpha, The scale of the step we moved along the direction vector.\n"
		"\n"
		"Title: Output to log:\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    none\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: none\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   none\n"
		""},
	 { "OBForceField_LineSearchTakeStep", _wrap_OBForceField_LineSearchTakeStep, METH_VARARGS, "\n"
		"Set the coordinates of the atoms to origCoord + step.\n"
		":type origCoords: float\n"
		":param origCoords: Start coordinates.\n"
		":type direction: float\n"
		":param direction: The search direction.\n"
		":type step: float\n"
		":param step: The step to take.\n"
		""},
	 { "OBForceField_SteepestDescent", _wrap_OBForceField_SteepestDescent, METH_VARARGS, "\n"
		"Perform steepest descent optimalization for steps steps or until convergence criteria is reached.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps and first step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_SteepestDescentInitialize", _wrap_OBForceField_SteepestDescentInitialize, METH_VARARGS, "\n"
		"Initialize steepest descent optimalization, to be used in combination with SteepestDescentTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->SteepestDescentInitialize(100, 1e-5f);\n"
		"    while (pFF->SteepestDescentTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, SteepestDescent() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_SteepestDescentTakeNSteps", _wrap_OBForceField_SteepestDescentTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in a steepestdescent optimalization that was previously initialized with SteepestDescentInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by SteepestDescentInitialize() has been reached.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    step number, energy and energy for the previous step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradients", _wrap_OBForceField_ConjugateGradients, METH_VARARGS, "\n"
		"Perform conjugate gradient optimalization for steps steps or until convergence criteria is reached.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    information about the progress of the minimization\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradientsInitialize", _wrap_OBForceField_ConjugateGradientsInitialize, METH_VARARGS, "\n"
		"Initialize conjugate gradient optimalization and take the first step, to be\n"
		"used in combination with ConjugateGradientsTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->ConjugateGradientsInitialize(100, 1e-5f);\n"
		"    while (pFF->ConjugateGradientsTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, ConjugateGradients() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps and first step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradientsTakeNSteps", _wrap_OBForceField_ConjugateGradientsTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in a conjugate gradient optimalization that was previously\n"
		"initialized with ConjugateGradientsInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by ConjugateGradientsInitialize() has been reached.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    step number, energy and energy for the previous step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_LBFGS", _wrap_OBForceField_LBFGS, METH_VARARGS, "\n"
		"Perform L-BFGS optimization for steps steps or until convergence criteria is reached.\n"
		"\n"
		"L-BFGS is a quasi-Newton method that approximates the inverse Hessian\n"
		"from a limited history of gradient changes. For typical molecular\n"
		"systems it converges faster than steepest descent or conjugate\n"
		"gradients while using comparable memory.\n"
		"\n"
		"Note: the line search algorithm chosen via SetLineSearchType()\n"
		"(Simple, Newton2Num) does NOT apply here. L-BFGS uses an internal\n"
		"backtracking line search with the Armijo condition; mixing in an\n"
		"arbitrary line search would break the curvature approximation.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (default is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		""},
	 { "OBForceField_LBFGSInitialize", _wrap_OBForceField_LBFGSInitialize, METH_VARARGS, "\n"
		"Initialize L-BFGS optimization, to be used in combination with LBFGSTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to an OBForceField class\n"
		"    pFF->LBFGSInitialize(100, 1e-5f);\n"
		"    while (pFF->LBFGSTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, LBFGS() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (default is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		""},
	 { "OBForceField_LBFGSTakeNSteps", _wrap_OBForceField_LBFGSTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in an L-BFGS optimization previously initialized with LBFGSInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by LBFGSInitialize() has been reached.\n"
		""},
	 { "OBForceField_GenerateVelocities", _wrap_OBForceField_GenerateVelocities, METH_O, "Generate starting velocities with a Maxwellian distribution."},
	 { "OBForceField_CorrectVelocities", _wrap_OBForceField_CorrectVelocities, METH_O, "\n"
		"Correct the velocities so that the following is true:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"          3N\n"
		"         ----\n"
		"    0.5  \\    m_i * v_i^2 = 0.5 * Ndf * R * T = E_kin\n"
		"         /\n"
		"         ----\n"
		"         i=1\n"
		"\n"
		"    E_kin : kinetic energy\n"
		"    m_i : mass of atom i\n"
		"    v_i : velocity of atom i\n"
		"    Ndf : number of degrees of freedom (3 * number of atoms)\n"
		"    R : gas constant\n"
		"    T : temperature\n"
		""},
	 { "OBForceField_MolecularDynamicsTakeNSteps", _wrap_OBForceField_MolecularDynamicsTakeNSteps, METH_VARARGS, "\n"
		"Take n steps at temperature T. If no velocities are set, they will be generated.\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    while (pFF->MolecularDynamicsTakeNSteps(5, 300)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":type T: float\n"
		":param T: Absolute temperature in Kelvin.\n"
		":type timestep: float, optional\n"
		":param timestep: The time step in picoseconds. (10e-12 s)\n"
		":type method: int, optional\n"
		":param method: OBFF_ANALYTICAL_GRADIENTS (default) or OBFF_NUMERICAL_GRADIENTS\n"
		""},
	 { "OBForceField_GetConstraints", _wrap_OBForceField_GetConstraints, METH_O, "\n"
		"Get the current constraints.\n"
		":rtype: :py:class:`OBFFConstraints`\n"
		":return: The current constrains stored in the force field.\n"
		""},
	 { "OBForceField_SetConstraints", _wrap_OBForceField_SetConstraints, METH_VARARGS, "\n"
		"Set the constraints.\n"
		":type constraints: :py:class:`OBFFConstraints`\n"
		":param constraints: The new constraints to be used.\n"
		""},
	 { "OBForceField_SetFixAtom", _wrap_OBForceField_SetFixAtom, METH_VARARGS, "\n"
		"Fix the atom position until UnsetFixAtom() is called. This function\n"
		"can be used in programs that allow the user to interact with a molecule\n"
		"that is being minimized without having to check if the atom is already\n"
		"fixed in the constraints set by Setup() or SetConstraints(). Using this\n"
		"makes sure the selected atom follows the mouse cursur.\n"
		":type index: int\n"
		":param index: The index for the atom to fix.\n"
		""},
	 { "OBForceField_UnsetFixAtom", _wrap_OBForceField_UnsetFixAtom, METH_O, "\n"
		"Undo last SetFixAtom. This function will not remove the fix atom\n"
		"constraint for this atom if set by Setup() or SetConstraints().\n"
		""},
	 { "OBForceField_SetIgnoreAtom", _wrap_OBForceField_SetIgnoreAtom, METH_VARARGS, "\n"
		"Ignore the atom until UnsetIgnoreAtom() is called. This function\n"
		"can be used in programs that allow the user to interact with a molecule\n"
		"that is being minimized without having to check if the atom is already\n"
		"ignored in the constraints set by Setup() or SetConstraints(). Using this\n"
		"makes sure, in drawing mode, you can close rings without your newly\n"
		"created puching the other atoms away.\n"
		":type index: int\n"
		":param index: The index for the atom to ignore.\n"
		""},
	 { "OBForceField_UnsetIgnoreAtom", _wrap_OBForceField_UnsetIgnoreAtom, METH_O, "\n"
		"Undo last SetIgnoreAtom. This function will not remove the ignore atom\n"
		"constraint for this atom if set by Setup() or SetConstraints().\n"
		""},
	 { "OBForceField_IgnoreCalculation", _wrap_OBForceField_IgnoreCalculation, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"internal function\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"internal function\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"internal function\n"
		""},
	 { "OBForceField_DetectExplosion", _wrap_OBForceField_DetectExplosion, METH_O, "(debugging)"},
	 { "OBForceField_ValidateLineSearch", _wrap_OBForceField_ValidateLineSearch, METH_VARARGS, "(debugging)"},
	 { "OBForceField_ValidateSteepestDescent", _wrap_OBForceField_ValidateSteepestDescent, METH_VARARGS, "(debugging)"},
	 { "OBForceField_ValidateConjugateGradients", _wrap_OBForceField_ValidateConjugateGradients, METH_VARARGS, "(debugging)"},
	 { "OBForceField_Validate", _wrap_OBForceField_Validate, METH_O, "Validate the force field implementation (debugging)"},
	 { "OBForceField_ValidateGradients", _wrap_OBForceField_ValidateGradients, METH_O, "\n"
		"Validate the analytical gradients by comparing them to numerical ones. This function has to\n"
		"be implemented force field specific. (debugging)\n"
		""},
	 { "OBForceField_ValidateGradientError", _wrap_OBForceField_ValidateGradientError, METH_VARARGS, "\n"
		"Calculate the error of the analytical gradient (debugging)\n"
		":rtype: :py:class:`vector3`\n"
		":return: error = fabs(numgrad - anagrad) / anagrad * 100%\n"
		""},
	 { "OBForceField_VectorBondDerivative", _wrap_OBForceField_VectorBondDerivative, METH_VARARGS, "\n"
		" Calculate the derivative of a vector length. The vector is given by a - b,\n"
		"the length of this vector rab = sqrt(ab.x^2 + ab.y^2 + ab.z^2).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":rtype: float\n"
		":return: The distance between a and b (bondlength for bond stretching, separation for vdw, electrostatic)\n"
		""},
	 { "OBForceField_VectorDistanceDerivative", _wrap_OBForceField_VectorDistanceDerivative, METH_VARARGS, "\n"
		"To be used for VDW or Electrostatic interactions. This\n"
		"is faster than VectorBondDerivative, but does no error checking.\n"
		""},
	 { "OBForceField_VectorLengthDerivative", _wrap_OBForceField_VectorLengthDerivative, METH_VARARGS, "Deprecated:"},
	 { "OBForceField_VectorAngleDerivative", _wrap_OBForceField_VectorAngleDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a angle a-b-c. The angle is given by dot(ab,cb)/rab*rcb.\n"
		" Used for harmonic (cubic) angle potentials.\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":rtype: float\n"
		":return: The angle between a-b-c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorOOPDerivative", _wrap_OBForceField_VectorOOPDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a OOP angle a-b-c-d. b is the central atom, and a-b-c is the plane.\n"
		"The OOP angle is given by 90° - arccos(dot(corss(ab,cb),db)/rabbc*rdb).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type pos_d: float\n"
		":param pos_d: atom d (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":type force_d: float\n"
		":param force_d: - return value for the force on atom d\n"
		":rtype: float\n"
		":return: The OOP angle for a-b-c-d\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorTorsionDerivative", _wrap_OBForceField_VectorTorsionDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a torsion angle a-b-c-d. The torsion angle is given by arccos(dot(corss(ab,bc),cross(bc,cd))/rabbc*rbccd).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type pos_d: float\n"
		":param pos_d: atom d (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":type force_d: float\n"
		":param force_d: - return value for the force on atom d\n"
		":rtype: float\n"
		":return: The tosion angle for a-b-c-d\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorSubtract", _wrap_OBForceField_VectorSubtract, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: pointer to result[3], will be set to i - j\n"
		""},
	 { "OBForceField_VectorAdd", _wrap_OBForceField_VectorAdd, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: pointer to result[3], will be set to i + j\n"
		""},
	 { "OBForceField_VectorDivide", _wrap_OBForceField_VectorDivide, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type n: float\n"
		":param n: divide x,y,z with n\n"
		":type result: float\n"
		":param result: pointer to result[3]\n"
		""},
	 { "OBForceField_VectorMultiply", _wrap_OBForceField_VectorMultiply, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type n: float\n"
		":param n: multiply x,y,z with n\n"
		":type result: float\n"
		":param result: pointer to result[3]\n"
		""},
	 { "OBForceField_VectorSelfMultiply", _wrap_OBForceField_VectorSelfMultiply, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3], multiply this vector by n and set this vector to the result.\n"
		":type n: float\n"
		":param n: the scalar value to be multipled\n"
		""},
	 { "OBForceField_VectorNormalize", _wrap_OBForceField_VectorNormalize, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3] to be normalized\n"
		""},
	 { "OBForceField_VectorCopy", _wrap_OBForceField_VectorCopy, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type from: float\n"
		":param from: pointer to i[3] to be copied from\n"
		":type to: float\n"
		":param to: pointer to j[3] to be copied to\n"
		""},
	 { "OBForceField_VectorLength", _wrap_OBForceField_VectorLength, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":rtype: float\n"
		":return: the vector length\n"
		""},
	 { "OBForceField_VectorDistance", _wrap_OBForceField_VectorDistance, METH_VARARGS, NULL},
	 { "OBForceField_VectorAngle", _wrap_OBForceField_VectorAngle, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":rtype: float\n"
		":return: the vector angle ijk (deg)\n"
		""},
	 { "OBForceField_VectorTorsion", _wrap_OBForceField_VectorTorsion, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":type l: float\n"
		":param l: pointer to l[3]\n"
		":rtype: float\n"
		":return: the vector torson ijkl (deg)\n"
		""},
	 { "OBForceField_VectorOOP", _wrap_OBForceField_VectorOOP, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":type l: float\n"
		":param l: pointer to l[3]\n"
		":rtype: float\n"
		":return: the vector torson ijkl (deg)\n"
		""},
	 { "OBForceField_VectorClear", _wrap_OBForceField_VectorClear, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3], will set x,y,z to 0,0,0\n"
		""},
	 { "OBForceField_VectorDot", _wrap_OBForceField_VectorDot, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":rtype: float\n"
		":return: the dot product\n"
		""},
	 { "OBForceField_VectorCross", _wrap_OBForceField_VectorCross, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: the dot product (as a return value double[3])\n"
		""},
	 { "OBForceField_PrintVector", _wrap_OBForceField_PrintVector, METH_O, NULL},
	 { "OBForceField_SetLogToStdOut", _wrap_OBForceField_SetLogToStdOut, METH_O, NULL},
	 { "OBForceField_SetLogToStdErr", _wrap_OBForceField_SetLogToStdErr, METH_O, NULL},
	 { "OBForceField_swigregister", OBForceField_swigregister, METH_O, NULL},
	 { "new_OBBuilder", _wrap_new_OBBuilder, METH_NOARGS, NULL},
	 { "OBBuilder_Build", _wrap_OBBuilder_Build, METH_VARARGS, "\n"
		"The mol object contains all connectivity information (atomic numbers, bonds, bond orders, ..)\n"
		"but no 3D coordinates. Build generates these coordinates and assigns them.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: Molecule with the connectivity (from SMILES for example). The coordinates are also\n"
		"           changed in this mol.\n"
		":type stereoWarnings: boolean, optional\n"
		":param stereoWarnings: Warn if the stereochemistry is incorrect (default is true)\n"
		""},
	 { "OBBuilder_SetKeepRings", _wrap_OBBuilder_SetKeepRings, METH_O, "\n"
		"If the molecule already contains 3D coordinates, if you set KeepRings to true it will use\n"
		"retain the 3D coordinates of the rings. By default KeepRings is false, and ring conformations\n"
		"are obtained by lookup in a library of ring conformers. However, since the ring conformer library\n"
		"is not exhaustive, if the ring system is not found in the library, the resulting 3D structure can\n"
		"be poor, and require geometry optimisation before it is reasonable. If your starting point is\n"
		"a 3D structure, you can set KeepRings to true, and the conformation will be taken from the input.\n"
		"The remaining (acyclic) bonds will still all be built by the builder.\n"
		""},
	 { "OBBuilder_UnsetKeepRings", _wrap_OBBuilder_UnsetKeepRings, METH_O, NULL},
	 { "OBBuilder_AddRingFragment", _wrap_OBBuilder_AddRingFragment, METH_VARARGS, "Used by LoadFragments to check for invalid (all zero coordinates) fragments"},
	 { "OBBuilder_LoadFragments", _wrap_OBBuilder_LoadFragments, METH_O, "Load fragment info from file, if is it has not already been done"},
	 { "OBBuilder_GetFragmentCoord", _wrap_OBBuilder_GetFragmentCoord, METH_VARARGS, NULL},
	 { "OBBuilder_GetNewBondVector", _wrap_OBBuilder_GetNewBondVector, METH_VARARGS, "\n"
		" Get the position for a new neighbour on atom.  Returns\n"
		"non-finite vector if there is no reasonable location.\n"
		" :type atom: :py:class:`OBAtom`\n"
		" :param atom: Atom for which we want a new neighbour location.\n"
		" :rtype: :py:class:`vector3`\n"
		" :return: The position for the new atom.\n"
		""},
	 { "OBBuilder_Connect", _wrap_OBBuilder_Connect, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":type newpos: :py:class:`vector3`\n"
		":param newpos: Direction for new bond between a and b\n"
		":type bondOrder: int, optional\n"
		":param bondOrder: Bond order of the new bond between a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":type bondOrder: int, optional\n"
		":param bondOrder: Bond order of the new bond bewtween a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":param bondOrder: Bond order of the new bond bewtween a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		""},
	 { "OBBuilder_Swap", _wrap_OBBuilder_Swap, METH_VARARGS, "\n"
		"Swap group b, bonded to a with group d, bonded to c. The bonds a-b and b-c cannot be\n"
		"part of a ring. Atoms a and b will not be moved. Atoms b, d and their connected atoms\n"
		"(after deleting bonds ab and cd) will be translated/rotated.\n"
		"\n"
		"Example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"      \\ /                            /\n"
		"       b                            d\n"
		"        \\     /     Swap(a,b,c,d)    \\     /\n"
		"         a---x          ---->         a---x\n"
		"        /     \\     /                /     \\     /\n"
		"       x       c---d                x       c---b\n"
		"                                                 \\\n"
		"\n"
		"\n"
		"This function can also be used to invert chiral centers if a and c are the same atom.\n"
		"\n"
		"Example\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"       1                        3\n"
		"       |      Swap(C,1,C,3)     |\n"
		"    2>-C-<3      ----->      2>-C-<1\n"
		"       |                        |\n"
		"       4                        4\n"
		""},
	 { "OBBuilder_CorrectStereoBonds", _wrap_OBBuilder_CorrectStereoBonds, METH_O, "\n"
		" Atoms a and b must be bonded and this bond cannot be part of a ring. The bond will\n"
		" be broken and the smiles fragment will be inserted bewteen the two remaining fragments.\n"
		" The fragment that contains a will not be translated or rotated. Parameters c and d are\n"
		" the index in the smiles to which atoms a and b will be connected respectivly.\n"
		"\n"
		"        Correct double bond stereochemistry\n"
		"\n"
		":rtype: boolean\n"
		":return: Success or failure\n"
		""},
	 { "OBBuilder_CorrectStereoAtoms", _wrap_OBBuilder_CorrectStereoAtoms, METH_VARARGS, "\n"
		" Correct stereochemistry at tetrahedral atoms with at least two non-ring\n"
		"bonds. It also works for spiro atoms.\n"
		"\n"
		":rtype: boolean\n"
		":return: Success or failure\n"
		""},
	 { "OBBuilder_IsSpiroAtom", _wrap_OBBuilder_IsSpiroAtom, METH_VARARGS, "Does this atom connect two rings which are not otherwise connected?"},
	 { "OBBuilder_GetFragment", _wrap_OBBuilder_GetFragment, METH_O, "\n"
		"Get the fragment to which this atom belongs.\n"
		":type atom: :py:class:`OBAtom`\n"
		":param atom: Atom in the fragment.\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: The OBBitVec defining the fragment to which a belongs.\n"
		""},
	 { "OBBuilder_AddNbrs", _wrap_OBBuilder_AddNbrs, METH_VARARGS, NULL},
	 { "delete_OBBuilder", _wrap_delete_OBBuilder, METH_O, NULL},
	 { "OBBuilder_swigregister", OBBuilder_swigregister, METH_O, NULL},
	 { "OBBuilder_swiginit", OBBuilder_swiginit, METH_VARARGS, NULL},
	 { "OBOp_Default", _wrap_OBOp_Default, METH_NOARGS, NULL},
	 { "OBOp_FindType", _wrap_OBOp_FindType, METH_O, NULL},
	 { "OBOp_Do", _wrap_OBOp_Do, METH_VARARGS, "Required function that does the work. Normally return true, unless object is not to be output."},
	 { "OBOp_WorksWith", _wrap_OBOp_WorksWith, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this op is designed to work with the class of pOb, e.g. OBMol\n"
		""},
	 { "OBOp_ProcessVec", _wrap_OBOp_ProcessVec, METH_VARARGS, "Do something with an array of objects. Used a a callback routine in OpSort, etc."},
	 { "OBOp_OpOptions", _wrap_OBOp_OpOptions, METH_O, "\n"
		":rtype: string\n"
		":return: string describing options, for display with -H and to make checkboxes in GUI\n"
		""},
	 { "OBOp_DoOps", _wrap_OBOp_DoOps, METH_VARARGS, "\n"
		"Call Do() of all the OBOps whose ID is a key in the map.\n"
		"Called from DoTransformations(). The map has general options like -x or --multicharoption\n"
		"The key is the option name and the value, if any, is text which follows the option name.\n"
		" In some cases, there may be several parameters, space separated)\n"
		" :rtype: boolean\n"
		" :return: false indicating object should not be output, if any Do() returns false\n"
		""},
	 { "delete_OBOp", _wrap_delete_OBOp, METH_O, NULL},
	 { "OBOp_swigregister", OBOp_swigregister, METH_O, NULL},
	 { "OBChargeModel_Default", _wrap_OBChargeModel_Default, METH_NOARGS, NULL},
	 { "new_OBChargeModel", _wrap_new_OBChargeModel, METH_VARARGS, NULL},
	 { "OBChargeModel_FindType", _wrap_OBChargeModel_FindType, METH_O, NULL},
	 { "OBChargeModel_ComputeCharges", _wrap_OBChargeModel_ComputeCharges, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: whether partial charges were successfully assigned to this molecule\n"
		"Notes: The method should fill m_partialCharges and m_formalCharges as well\n"
		""},
	 { "OBChargeModel_GetFormalCharges", _wrap_OBChargeModel_GetFormalCharges, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a vector of the formal charges on each atom, indexed from 0\n"
		"    This method returns floating point formal charges since some\n"
		"    charge models consider fractional charges (e.g., 0.5 for an\n"
		"    oxygen in a carboxylate CO2- group).\n"
		"Notes: If ComputeCharges() has not been called, this will return an empty vector\n"
		""},
	 { "OBChargeModel_GetPartialCharges", _wrap_OBChargeModel_GetPartialCharges, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a vector of the partial charges on each atom, indexed from 0\n"
		"Notes: If ComputeCharges() has not been called, this will return an empty vector\n"
		""},
	 { "OBChargeModel_GetDipoleMoment", _wrap_OBChargeModel_GetDipoleMoment, METH_VARARGS, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: a vector of the dipole moment from this molecule\n"
		""},
	 { "delete_OBChargeModel", _wrap_delete_OBChargeModel, METH_O, NULL},
	 { "OBChargeModel_swigregister", OBChargeModel_swigregister, METH_O, NULL},
	 { "OBChargeModel_swiginit", OBChargeModel_swiginit, METH_VARARGS, NULL},
	 { "new_OBChemTsfm", _wrap_new_OBChemTsfm, METH_NOARGS, NULL},
	 { "delete_OBChemTsfm", _wrap_delete_OBChemTsfm, METH_O, NULL},
	 { "OBChemTsfm_Init", _wrap_OBChemTsfm_Init, METH_VARARGS, "Initialize this transformation with the supplied SMARTS patterns"},
	 { "OBChemTsfm_Apply", _wrap_OBChemTsfm_Apply, METH_VARARGS, "Apply this transformation to all matches in the supplied OBMol"},
	 { "OBChemTsfm_IsAcid", _wrap_OBChemTsfm_IsAcid, METH_O, "\n"
		"Is this transformation an acid dissociation?\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"        Ka\n"
		"    HA ----> A(-)         (the H(+) will be deleted)\n"
		"\n"
		"IsAcid() will check the charge in the end SMARTS pattern.\n"
		":rtype: boolean\n"
		":return: true if the charge is less than 0 (-1).\n"
		""},
	 { "OBChemTsfm_IsBase", _wrap_OBChemTsfm_IsBase, METH_O, "\n"
		"Is this a transformation to the conjugated acid from a base?\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"        Ka\n"
		"    HA ----> A(-)         (the H(+) will be deleted)\n"
		"\n"
		"IsBase() will check the charge in the end SMARTS pattern.\n"
		":rtype: boolean\n"
		":return: true if the charge is higher than 0 (+1).\n"
		""},
	 { "OBChemTsfm_swigregister", OBChemTsfm_swigregister, METH_O, NULL},
	 { "OBChemTsfm_swiginit", OBChemTsfm_swiginit, METH_VARARGS, NULL},
	 { "new_OBPhModel", _wrap_new_OBPhModel, METH_NOARGS, NULL},
	 { "delete_OBPhModel", _wrap_delete_OBPhModel, METH_O, NULL},
	 { "OBPhModel_AssignSeedPartialCharge", _wrap_OBPhModel_AssignSeedPartialCharge, METH_VARARGS, NULL},
	 { "OBPhModel_CorrectForPH", _wrap_OBPhModel_CorrectForPH, METH_VARARGS, NULL},
	 { "OBPhModel_swigregister", OBPhModel_swigregister, METH_O, NULL},
	 { "OBPhModel_swiginit", OBPhModel_swiginit, METH_VARARGS, NULL},
	 { "new_OBGraphSym", _wrap_new_OBGraphSym, METH_VARARGS, "Constructor"},
	 { "delete_OBGraphSym", _wrap_delete_OBGraphSym, METH_O, "Destructor"},
	 { "OBGraphSym_GetSymmetry", _wrap_OBGraphSym_GetSymmetry, METH_VARARGS, "\n"
		"Calculate the symmetry classes for the molecule. The result will be\n"
		"stored in ``symmetry_classes``.\n"
		"\n"
		"The results in ``symmetry_classes`` will be ordered by symmetry\n"
		"classes. Use the OBAtom* pointer in the std::pair to match the atoms\n"
		"with the right symmetry classes.\n"
		"\n"
		":rtype: int\n"
		":return: The number of symmetry classes.\n"
		""},
	 { "OBGraphSym_ClearSymmetry", _wrap_OBGraphSym_ClearSymmetry, METH_O, "\n"
		"Clear the symmetry classes data stored in the molecule specified when\n"
		"construting the OBGraphSym object.\n"
		""},
	 { "OBGraphSym_swigregister", OBGraphSym_swigregister, METH_O, NULL},
	 { "OBGraphSym_swiginit", OBGraphSym_swiginit, METH_VARARGS, NULL},
	 { "delete_OBIsomorphismMapper", _wrap_delete_OBIsomorphismMapper, METH_O, NULL},
	 { "OBIsomorphismMapper_GetInstance", _wrap_OBIsomorphismMapper_GetInstance, METH_VARARGS, "\n"
		"Get a pointer to an instance of the specified ``algorithm``. This pointer\n"
		"has to be delted when the instance is no longer needed.\n"
		":type query: :py:class:`OBQuery`\n"
		":param query: The search query to be mapped.\n"
		":type algorithm: string, optional\n"
		":param algorithm: The algorithm for the mapper.\n"
		":rtype: :py:class:`OBIsomorphismMapper`\n"
		":return: OBIsomorphismMapper instance or 0 if there is no subclass implementing\n"
		"    the specified ``algorithm``.\n"
		""},
	 { "OBIsomorphismMapper_MapFirst", _wrap_OBIsomorphismMapper_MapFirst, METH_VARARGS, "\n"
		"Find a single mapping in ``queried``.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type map: :py:class:`Mapping`\n"
		":param map: Reference to the object to store the result in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapUnique", _wrap_OBIsomorphismMapper_MapUnique, METH_VARARGS, "\n"
		"Find all unique mappings in ``queried``. A mapping is unique when there is no previous\n"
		"mapping covering the same queried atoms. For two mappings, some overlap is allowed but\n"
		"at least one atom should be different.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type maps: :py:class:`Mappings`\n"
		":param maps: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapAll", _wrap_OBIsomorphismMapper_MapAll, METH_VARARGS, "\n"
		"Find all mappings in ``queried``. This function is used by FindAutomorphisms()\n"
		"with a query that is a copy of the queried molecule (taking the mask into\n"
		"account).\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type maps: :py:class:`Mappings`\n"
		":param maps: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		":type maxMemory: std::size_t, optional\n"
		":param maxMemory: Memory limit for the ``maps`` object in bytes. Default is 300MB.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapGeneric", _wrap_OBIsomorphismMapper_MapGeneric, METH_VARARGS, "\n"
		"Find all mappings in ``queried``. The functor will be called when a mapping is found.\n"
		":type functor: OpenBabel::OBIsomorphismMapper::Functor\n"
		":param functor: The functor to handle found mappings.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_SetTimeout", _wrap_OBIsomorphismMapper_SetTimeout, METH_VARARGS, "Set the timeout in seconds."},
	 { "OBIsomorphismMapper_swigregister", OBIsomorphismMapper_swigregister, METH_O, NULL},
	 { "MapsTo", _wrap_MapsTo, METH_VARARGS, NULL},
	 { "FindAutomorphisms", _wrap_FindAutomorphisms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function is a wrapper for FindAutomorphisms with a functor to store all\n"
		"automorphisms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule for which to find the automorphisms.\n"
		":type aut: std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > >\n"
		":param aut: The result will be stored here.\n"
		":type symmetry_classes: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param symmetry_classes: The graph invariants to use. See OBGraphSym or use\n"
		"    the FindAutomorphisms function that computes these for you below.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A bit vector specifying which atoms to consider. An empty mask\n"
		"    will consider all atoms. The bits are indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":type maxMemory: std::size_t, optional\n"
		":param maxMemory: Maximum memory limit for ``aut``. The number of automorphisms\n"
		"    for certain graphs can be large. The default is 300MB, consider using a functor\n"
		"    to process automorphisms when they are found.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"is the main implementation for finding automorphisms and uses an\n"
		"OBIsomorphismMapper::Functor to process found isomorphisms. Wrapper functions\n"
		"are provided to find and store all automorphisms but the number of automorphisms\n"
		"can be large for certain graphs making it not feasible to store all automorphisms\n"
		"in memory (RAM).\n"
		"\n"
		"See also: 'MapGeneric'\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 6:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"is the main implementation for finding automorphisms and uses an\n"
		"OBIsomorphismMapper::Functor to process found isomorphisms. Wrapper functions\n"
		"are provided to find and store all automorphisms but the number of automorphisms\n"
		"can be large for certain graphs making it not feasible to store all automorphisms\n"
		"in memory (RAM).\n"
		"\n"
		"See also: 'MapGeneric'\n"
		"Since: version 2.3\n"
		""},
	 { "new_OBQueryAtom", _wrap_new_OBQueryAtom, METH_VARARGS, "\n"
		"Constructor.\n"
		":type atomicNum: int, optional\n"
		":param atomicNum: The atomic number for this query atom.\n"
		":type isInRing: boolean, optional\n"
		":param isInRing: Specify whether the query atom is in a ring. Default is false.\n"
		":type isAromatic: boolean, optional\n"
		":param isAromatic: Specify whether the query atom is aromatic. Default is false.\n"
		""},
	 { "delete_OBQueryAtom", _wrap_delete_OBQueryAtom, METH_O, NULL},
	 { "OBQueryAtom_GetIndex", _wrap_OBQueryAtom_GetIndex, METH_O, "\n"
		"Get the index for this query atom. Atoms are indexed starting from 0.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_GetBonds", _wrap_OBQueryAtom_GetBonds, METH_O, "\n"
		"Get the query bonds for this atom.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_GetNbrs", _wrap_OBQueryAtom_GetNbrs, METH_O, "\n"
		"Get the neighbor query atoms.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_Matches", _wrap_OBQueryAtom_Matches, METH_VARARGS, "\n"
		"This is the match method to verify if an OBQueryAtom and OBAtom class match.\n"
		"The default implementation only checks if the atomic numbers match. Reimplement\n"
		"this method in a subclass for more advances matching.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		":type atom: :py:class:`OBAtom`\n"
		":param atom: The OBAtom object to compare this OBQueryAtom with.\n"
		""},
	 { "OBQueryAtom_swigregister", OBQueryAtom_swigregister, METH_O, NULL},
	 { "OBQueryAtom_swiginit", OBQueryAtom_swiginit, METH_VARARGS, NULL},
	 { "new_OBQueryBond", _wrap_new_OBQueryBond, METH_VARARGS, "Constructor."},
	 { "delete_OBQueryBond", _wrap_delete_OBQueryBond, METH_O, NULL},
	 { "OBQueryBond_GetIndex", _wrap_OBQueryBond_GetIndex, METH_O, "Get the index for this query bonds. Query bonds are indexed starting from 0."},
	 { "OBQueryBond_GetBeginAtom", _wrap_OBQueryBond_GetBeginAtom, METH_O, "Get the begin atom."},
	 { "OBQueryBond_GetEndAtom", _wrap_OBQueryBond_GetEndAtom, METH_O, "Get the end atom."},
	 { "OBQueryBond_Matches", _wrap_OBQueryBond_Matches, METH_VARARGS, "\n"
		"This is the match method to verify if an OBQueryBond and OBBond class match.\n"
		"The default implementation checks if both bonds are aromatic and compares the\n"
		"bond orders otherwise. Reimplement this method in a subclass for more\n"
		"advances matching.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		":type bond: :py:class:`OBBond`\n"
		":param bond: The OBBond object to compare this OBQueryBond with.\n"
		""},
	 { "OBQueryBond_swigregister", OBQueryBond_swigregister, METH_O, NULL},
	 { "OBQueryBond_swiginit", OBQueryBond_swiginit, METH_VARARGS, NULL},
	 { "delete_OBQuery", _wrap_delete_OBQuery, METH_O, NULL},
	 { "OBQuery_NumAtoms", _wrap_OBQuery_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in the query.\n"
		""},
	 { "OBQuery_NumBonds", _wrap_OBQuery_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: The number of bonds in the query.\n"
		""},
	 { "OBQuery_GetAtoms", _wrap_OBQuery_GetAtoms, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > >\n"
		":return: std::vector with pointers to the query atoms.\n"
		""},
	 { "OBQuery_GetBonds", _wrap_OBQuery_GetBonds, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > >\n"
		":return: std::vector with pointers to the query bonds.\n"
		""},
	 { "OBQuery_GetBond", _wrap_OBQuery_GetBond, METH_VARARGS, "\n"
		":rtype: :py:class:`OBQueryBond`\n"
		":return: The query bond between ``begin`` and ``end``. If there is no\n"
		"    bond between ``begin`` and ``end``, this function returns 0.\n"
		""},
	 { "OBQuery_AddAtom", _wrap_OBQuery_AddAtom, METH_VARARGS, "Add a query atom to the query. This function steals the pointer."},
	 { "OBQuery_AddBond", _wrap_OBQuery_AddBond, METH_VARARGS, "Add a query atom to the query. This function steals the pointer."},
	 { "new_OBQuery", _wrap_new_OBQuery, METH_NOARGS, NULL},
	 { "OBQuery_swigregister", OBQuery_swigregister, METH_O, NULL},
	 { "OBQuery_swiginit", OBQuery_swiginit, METH_VARARGS, NULL},
	 { "CompileMoleculeQuery", _wrap_CompileMoleculeQuery, METH_VARARGS, "\n"
		"Create an OBQuery object from an OBMol object.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The query molecule.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The mask specifying the atoms to use. Indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":rtype: :py:class:`OBQuery`\n"
		":return: A pointer to an OBQuery object for the smiles string. This pointer should be deleted.\n"
		"Since: version 2.3\n"
		""},
	 { "CompileSmilesQuery", _wrap_CompileSmilesQuery, METH_VARARGS, "\n"
		"Create an OBQuery object from a smiles string.\n"
		":type smiles: string\n"
		":param smiles: The query smiles string.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The mask specifying the atoms to use. Indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":rtype: :py:class:`OBQuery`\n"
		":return: A pointer to an OBQuery object for the smiles string. This pointer should be deleted.\n"
		"Since: version 2.3\n"
		""},
	 { "CanonicalLabels", _wrap_CanonicalLabels, METH_VARARGS, "\n"
		"Calculate the canonical labels for the molecule. Stereochemistry is\n"
		"included in the algorithm and the canonical labels. The result will be\n"
		"stored in ``canonical_labels``.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type symmetry_classes: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param symmetry_classes: The symmetry_classes for the molecule. These can\n"
		"    be obtained using the OBGraphSym class.\n"
		":type canonical_labels: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param canonical_labels: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The fragment to label. When the bit for an atom is set, it is\n"
		"    included in the fragment. If no bits are set, all atoms will be included.\n"
		"    Atoms are indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":type maxSeconds: int, optional\n"
		":param maxSeconds: Timeout in seconds.\n"
		":type onlyOne: boolean, optional\n"
		":param onlyOne: If true, the first found labels are returned. These are\n"
		"    canonical labels without considering stereochemistry and other attributes\n"
		"    not included in the symmetry classes.\n"
		"\n"
		":rtype: void\n"
		":return: The canonical labels for the molecule in ``canonical_labels``.\n"
		"\n"
		"See also: 'canonical_code_algorithm'\n"
		"Since: 2.3\n"
		""},
	 { "OBStereo_MakeRefs", _wrap_OBStereo_MakeRefs, METH_VARARGS, "\n"
		"Create a Refs list filled with ``ref1``, ``ref2``, ``ref3`` & ``ref4``.\n"
		"``ref4`` is not added to the returned Refs if it is equal to NoRef.\n"
		"\n"
		":rtype: :py:class:`Refs`\n"
		":return: A Refs list containing the specified Ref values.\n"
		""},
	 { "OBStereo_ContainsSameRefs", _wrap_OBStereo_ContainsSameRefs, METH_VARARGS, "\n"
		"Check if ``refs1`` and ``refs2`` contain the same Ref values regardless\n"
		"of their order.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(2, 3, 1)) // true\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(3, 2, 1)) // true\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(3, 4, 1)) // false\n"
		"\n"
		":rtype: boolean\n"
		":return: True if ``refs1`` and ``refs2`` contain the same Ref values.\n"
		""},
	 { "OBStereo_ContainsRef", _wrap_OBStereo_ContainsRef, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if ``refs`` contains ``ref``.\n"
		""},
	 { "OBStereo_NumInversions", _wrap_OBStereo_NumInversions, METH_O, "\n"
		"Compute the inversion vector for ``refs`` and return the sum of it's\n"
		"elements. The ith element in the inversion vector is the number of\n"
		"element to the right of element i with a lower value.\n"
		"\n"
		"The number of inversions is the same as the number of interchanges\n"
		"of consecutive elements.\n"
		"\n"
		"When working with 3 refs from a tetrahedral configuration:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    permutation   inversion vector    sum\n"
		"    -------------------------------------\n"
		"    123           0 0 0               0 (even) -> clockwise\n"
		"    132           0 1 0               1 (odd)  -> anti-clockwise\n"
		"    213           1 0 0               1 (odd)  -> anti-clockwise\n"
		"    231           1 1 0               2 (even) -> clockwise\n"
		"    312           2 0 0               2 (even) -> clockwise\n"
		"    321           2 1 0               3 (odd)  -> anti-clockwise\n"
		""},
	 { "OBStereo_Permutate", _wrap_OBStereo_Permutate, METH_VARARGS, "\n"
		"Permutate element ``i`` with ``j`` in ``refs``.\n"
		"\n"
		":type refs: :py:class:`Refs`\n"
		":param refs: The sequence with N elements to permutate.\n"
		":type i: int\n"
		":param i: Element i (0...N-1) will be mutated to j and vice versa.\n"
		":type j: int\n"
		":param j: Element j (0...N-1) will be mutated to i and vice versa.\n"
		"\n"
		"Notes: This method does nothing if i equals j.\n"
		""},
	 { "OBStereo_Permutated", _wrap_OBStereo_Permutated, METH_VARARGS, "\n"
		"Get ``refs`` with element ``i`` and ``j`` permutated.\n"
		"\n"
		":type refs: :py:class:`Refs`\n"
		":param refs: The sequence with N elements to permutate.\n"
		":type i: int\n"
		":param i: Element ``i`` (0...N-1) will be mutated to ``j`` and vice versa.\n"
		":type j: int\n"
		":param j: Element ``j`` (0...N-1) will be mutated to ``i`` and vice versa.\n"
		"\n"
		":rtype: :py:class:`Refs`\n"
		":return: ``refs`` with elements ``i`` and ``j`` permutated.\n"
		"\n"
		"Notes: This method does nothing if ``i`` equals ``j``.\n"
		""},
	 { "new_OBStereo", _wrap_new_OBStereo, METH_NOARGS, NULL},
	 { "delete_OBStereo", _wrap_delete_OBStereo, METH_O, NULL},
	 { "OBStereo_swigregister", OBStereo_swigregister, METH_O, NULL},
	 { "OBStereo_swiginit", OBStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBStereoUnit", _wrap_new_OBStereoUnit, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Default constructor creating an invalid OBStereoUnit. The unit can be\n"
		"made valid by setting the type, id and para data members.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Constructor specifying all data to create a valid OBStereoUnit.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Constructor specifying all data to create a valid OBStereoUnit.\n"
		""},
	 { "OBStereoUnit_type_set", _wrap_OBStereoUnit_type_set, METH_VARARGS, "the type for this stereogenic unit"},
	 { "OBStereoUnit_type_get", _wrap_OBStereoUnit_type_get, METH_O, "the type for this stereogenic unit"},
	 { "OBStereoUnit_id_set", _wrap_OBStereoUnit_id_set, METH_VARARGS, "the atom/bond (depends on type) unique id"},
	 { "OBStereoUnit_id_get", _wrap_OBStereoUnit_id_get, METH_O, "the atom/bond (depends on type) unique id"},
	 { "OBStereoUnit_para_set", _wrap_OBStereoUnit_para_set, METH_VARARGS, "para- (=ressemble) or true-stereocenter"},
	 { "OBStereoUnit_para_get", _wrap_OBStereoUnit_para_get, METH_O, "para- (=ressemble) or true-stereocenter"},
	 { "delete_OBStereoUnit", _wrap_delete_OBStereoUnit, METH_O, NULL},
	 { "OBStereoUnit_swigregister", OBStereoUnit_swigregister, METH_O, NULL},
	 { "OBStereoUnit_swiginit", OBStereoUnit_swiginit, METH_VARARGS, NULL},
	 { "delete_OBStereoBase", _wrap_delete_OBStereoBase, METH_O, "Destructor."},
	 { "OBStereoBase_GetMolecule", _wrap_OBStereoBase_GetMolecule, METH_O, "\n"
		"Get the molecule. This can be used by subclasses when more\n"
		"information is needed (e.g. OBCisTransStereo::GetCisRef, ...).\n"
		""},
	 { "OBStereoBase_GetType", _wrap_OBStereoBase_GetType, METH_O, "Reimplemented by subclasses to return the type defined in OBStereo::Type."},
	 { "OBStereoBase_SetSpecified", _wrap_OBStereoBase_SetSpecified, METH_VARARGS, "\n"
		"Set whether the stereochemistry is specified. Comparing a specified\n"
		"OBStereoBase derived class (or it's Config struct) with an unspecified\n"
		"one, always returns true.\n"
		""},
	 { "OBStereoBase_IsSpecified", _wrap_OBStereoBase_IsSpecified, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if the stereochemistry is specified.\n"
		""},
	 { "OBStereoBase_swigregister", OBStereoBase_swigregister, METH_O, NULL},
	 { "new_OBStereoFacade", _wrap_new_OBStereoFacade, METH_VARARGS, "\n"
		"Constructor with ``mol`` and ``perceive`` parameter.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type perceive: boolean, optional\n"
		":param perceive: If true, PerceiveStereo will be called if the\n"
		"    OBMol::HasChiralityPerceived() flag is not set. (default is true)\n"
		""},
	 { "OBStereoFacade_NumTetrahedralStereo", _wrap_OBStereoFacade_NumTetrahedralStereo, METH_O, "Get the number of tetrahedral stereocenters."},
	 { "OBStereoFacade_GetAllTetrahedralStereo", _wrap_OBStereoFacade_GetAllTetrahedralStereo, METH_O, "Get all the OBTetrahedralStereo objects."},
	 { "OBStereoFacade_HasTetrahedralStereo", _wrap_OBStereoFacade_HasTetrahedralStereo, METH_VARARGS, "\n"
		"Check if atom with ``id`` is a tetrahedral center.\n"
		":rtype: boolean\n"
		":return: True if the atom with ``id`` has tetrahedral stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetTetrahedralStereo", _wrap_OBStereoFacade_GetTetrahedralStereo, METH_VARARGS, "\n"
		"Get the OBTetrahedralStereo object with ``atomId`` as center. This\n"
		"function returns 0 if there is no OBTetrahedralStereo object found\n"
		"with the specified center.\n"
		""},
	 { "OBStereoFacade_NumCisTransStereo", _wrap_OBStereoFacade_NumCisTransStereo, METH_O, "Get the number of cis/trans stereocenters."},
	 { "OBStereoFacade_GetAllCisTransStereo", _wrap_OBStereoFacade_GetAllCisTransStereo, METH_O, "Get all the OBCisTransStereo objects."},
	 { "OBStereoFacade_HasCisTransStereo", _wrap_OBStereoFacade_HasCisTransStereo, METH_VARARGS, "\n"
		"Check if bond with ``id`` is a stereogenic cis/trans double bond.\n"
		":rtype: boolean\n"
		":return: True if the bond with ``id`` has cis/trans stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetCisTransStereo", _wrap_OBStereoFacade_GetCisTransStereo, METH_VARARGS, "\n"
		"Get the OBTetrahedralStereo object with ``bondId`` as double bond.\n"
		"This function returns 0 if there is no OBCisTransStereo object found\n"
		"with the specified bond.\n"
		""},
	 { "OBStereoFacade_NumSquarePlanarStereo", _wrap_OBStereoFacade_NumSquarePlanarStereo, METH_O, "Get the number of square-planar stereocenters."},
	 { "OBStereoFacade_GetAllSquarePlanarStereo", _wrap_OBStereoFacade_GetAllSquarePlanarStereo, METH_O, "Get all the OBSquarePlanarStereo objects."},
	 { "OBStereoFacade_HasSquarePlanarStereo", _wrap_OBStereoFacade_HasSquarePlanarStereo, METH_VARARGS, "\n"
		"Check if atom with ``id`` is a stereogenic square-planar atom.\n"
		":rtype: boolean\n"
		":return: True if the atom with ``id`` has square-planar stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetSquarePlanarStereo", _wrap_OBStereoFacade_GetSquarePlanarStereo, METH_VARARGS, "\n"
		"Get the OBSquarePlanarStereo object with ``atomId`` as center. This\n"
		"function returns 0 if there is no OBSquarePlanarStereo object found\n"
		"with the specified center.\n"
		""},
	 { "delete_OBStereoFacade", _wrap_delete_OBStereoFacade, METH_O, NULL},
	 { "OBStereoFacade_swigregister", OBStereoFacade_swigregister, METH_O, NULL},
	 { "OBStereoFacade_swiginit", OBStereoFacade_swiginit, METH_VARARGS, NULL},
	 { "PerceiveStereo", _wrap_PerceiveStereo, METH_VARARGS, "\n"
		"Convert 0D/2D/3D coordinates to OBStereo objects. The right function will\n"
		"be selected based on the molecule's dimensionality\n"
		"(i.e. OBMol::GetDimension()).\n"
		"\n"
		"See also: StereoFrom3D StereoFrom2D StereoFrom0D\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom2D", _wrap_StereoFrom2D, METH_VARARGS, "\n"
		"Convert the 2D depiction of molecule ``mol`` to OBStereo objects.\n"
		"This function makes use of the lower level functions\n"
		"TetrahedralFrom2D(), CisTransFrom2D(), SquarePlanarFrom2D(), ...\n"
		"\n"
		"First, symmetry analysis taking stereochemistry into account is\n"
		"performed iteratively (see OBGraphSym). Next the 2D coordinates,\n"
		"OBBond::Wedge, OBBond::Hash, OBBond::WedgeOrHash and OBBond::CisOrTrans\n"
		"are used to construct OBStereoBase derived objects to store the\n"
		"stereochemistry. These objects will be added to ``mol``.\n"
		"\n"
		"Unless perception is forced, this function does nothing if stereochemistry\n"
		"has already been perceived (i.e. OBMol::HasChiralityPerceived()). Before\n"
		"doing the actual perception, any data of the OBGenericDataType::StereoData\n"
		"type will be deleted.\n"
		"\n"
		"          Reference:\n"
		"     [1] T. Cieplak, J.L. Wisniewski, A New Effective Algorithm for the\n"
		"     Unambiguous Identification of the Stereochemical Characteristics of\n"
		"     Compounds During Their Registration in Databases. Molecules 2000, 6,\n"
		"     915-926, http://www.mdpi.org/molecules/papers/61100915/61100915.htm\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule containing 2D coordinates.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >, optional\n"
		":param updown: A map of OBStereo::BondDirection for cis/trans bonds\n"
		":type force: boolean, optional\n"
		":param force: Force to run the perception even if the results are cached.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom0D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom3D", _wrap_StereoFrom3D, METH_VARARGS, "\n"
		"Convert the 3D coordinates of molecule ``mol`` to OBStereo objects. This\n"
		"function makes use of the lower level functions TetrahedralFrom3D(),\n"
		"CisTransFrom3D(), SquarePlanarFrom3D(), ...\n"
		"\n"
		"Unless perception is forced, this function does nothing if stereochemistry\n"
		"has already been perceived (i.e. OBMol::HasChiralityPerceived()). Before\n"
		"doing the actual perception, any data of the OBGenericDataType::StereoData\n"
		"type will be deleted.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule containing 3D coordinates.\n"
		":type force: boolean, optional\n"
		":param force: Force to run the perception even if the results are cached.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom0D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom0D", _wrap_StereoFrom0D, METH_O, "\n"
		"Add missing OBStereo objects. Unlike StereoFrom3D() and StereoFrom2D(), this\n"
		"method only adds objects for previously unidentified objects since we\n"
		"don't want to loose any information. The Config::specified flag for the\n"
		"newly added structs is always set to false.\n"
		"\n"
		"For example, a smiles is read which has two tetrahedral centers. Only one has\n"
		"stereochemisrty specified using a '@' character. StereoFrom0D() will detect the\n"
		"second tetrahedral atom and add an OBTetrahedralStereo object to the molecule.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom2D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "TetrahedralFrom3D", _wrap_TetrahedralFrom3D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom3D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 3D coordinates to OBTetrahedralStereo object\n"
		"uses the sign of the volume described by the 4 center atom neighbors. Given\n"
		"4 points :math:`, \\f$b`, :math:`and \\f$d`, the signed volume _v:math:`is defined as:     \\f[ S_v = \\left| \\begin{array}{ccc}     x_b - x_a & y_b - y_a & z_b - z_a \\\\     x_c - x_a & y_c - y_a & z_c - z_a \\\\     x_d - x_a & y_d - y_a & z_d - z_a     \\end{array} \\right| The sign of _v` changes when any of the points cross the plane defined\n"
		"by the other 3 points. To make this less abstract one could say that\n"
		"a change of sign is equal to inverting the tetrahedral stereochemistry.\n"
		"\n"
		"In case there are only 3 neighbor atoms for the tetrahedral center, the\n"
		"center atom itself is used as 4th point. This only changes the magnitude\n"
		"and not the sign of \\f$S_v:math:`because the center atom is still on the sameside of the plane.This function is also used for symmetry analysis to handle cases wherethere are two atoms in the same symmetry class that don't have the samestereochemistry. In this situation, the addToMol parameter is set tofalse and the returned objects will need to be deleted explicitly.mol The molecule.stereoUnits The stereogenic units.addToMol If true, the OBTetrahedralStereo objects will be addedto the molecule using OBBase::SetData().StereoFrom3D FindStereogenicUnitsversion 2.3`\n"
		""},
	 { "TetrahedralFrom2D", _wrap_TetrahedralFrom2D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom2D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 2D coordinates and bond properties\n"
		"(i.e. OBBond::Wedge, OBBond::Hash, OBBond::WedgeOrHash and OBBond::CisOrTrans)\n"
		"uses the sign of a triangle. Given 3 points :math:`, \\f$b` and :math:`, thesign of the trianle \\f$S_t` is defined as:\n"
		"\n"
		"     .. math::\n"
		"\n"
		"         S_t = (x_a - x_c) (y_b - y_c) - (y_a - y_c) (x_b - x_c)\n"
		"\n"
		"This is equation 6 from on the referenced web page. The 3 points used\n"
		"to calculate the triangle sign always remain in the same plane (i.e. z = 0).\n"
		"The actual meaning of _t:math:`(i.e. assignment of OBStereo::Winding) dependson the 4th atom. When the atom is in front of the plane, the sign should bechanged to have the same absolute meaning for an atom behind the plane and thesame triangle. It is important to note that none of the z coordinates is everchanged, the molecule always stays 2D (unlike methods which set a pseudo-zcoordinate).@todo document bond property interpretation!This function is also used for symmetry analysis to handle cases wherethere are two atoms in the same symmetry class that don't have the samestereochemistry. In this situation, the @p addToMol parameter is set tofalse and the returned objects will need to be deleted explicitly.     @verbatim     Reference:     [1] T. Cieplak, J.L. Wisniewski, A New Effective Algorithm for the     Unambiguous Identification of the Stereochemical Characteristics of     Compounds During Their Registration in Databases. Molecules 2000, 6,     915-926, http://www.mdpi.org/molecules/papers/61100915/61100915.htm     mol The molecule.stereoUnits The stereogenic units.addToMol If true, the OBTetrahedralStereo objects will be addedto the molecule using OBBase::SetData().StereoFrom2D FindStereogenicUnitsversion 2.3`\n"
		""},
	 { "TetrahedralFrom0D", _wrap_TetrahedralFrom0D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom0D() with the ``addToMol`` parameter is set\n"
		"to true. There is no algorithm used here, all specified flags will be\n"
		"set to false.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBTetrahedralStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom0D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "CisTransFrom3D", _wrap_CisTransFrom3D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom3D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 3D coordinates to OBCisTransStereo objects\n"
		"considers the signed distance between the attached atoms and the plane\n"
		"through the double bond at right angles to the plane of the attached\n"
		"atoms. Bonds on the same side (cis) will share the same sign for the\n"
		"signed distance.\n"
		"\n"
		"Missing atom coordinates (OBStereo::ImplicitRef) and their bond\n"
		"vectors will be computed if needed.\n"
		"\n"
		"            0      3     Get signed distance of 0 and 2 to the plane\n"
		"          \\    /      that goes through the double bond and is at\n"
		"           C==C       right angles to the stereo bonds.\n"
		"   \\\n"
		"         1      2     If the two signed distances have the same sign\n"
		"                      then they are cis; if not, then trans.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom3D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "CisTransFrom2D", _wrap_CisTransFrom2D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom2D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		"The algorithm for converting the 2D coordinates uses the same triangle\n"
		"sign as TetrahedralFrom2D(). Depending on sign of 2 triangles, the right\n"
		"OBStereo::Shape is selected.\n"
		"         0      3\n"
		"       \\    /        2 triangles: 0-1-b & 2-3-a\n"
		"        a==b    -->  same sign: U\n"
		"   \\        opposite sign: Z\n"
		"      1      2\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >, optional\n"
		":param updown: A map of OBStereo::BondDirection for cis/trans bonds\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom2D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "TetStereoToWedgeHash", _wrap_TetStereoToWedgeHash, METH_VARARGS, "\n"
		"Convert a molecule's OBTetrahedralStereo objects to a series of hash or\n"
		"wedge bonds. Note that the molecule itself is not modified; the result\n"
		"is returned in the maps ``updown`` and ``from``, which indicate\n"
		"the origin and direction of each hash or wedge bond.\n"
		"\n"
		"When converting, the following guidelines are followed when trying to\n"
		"find the best candidate bond to set up/down for each OBTetrahedralStereo\n"
		"object:\n"
		"-# Should not already be set\n"
		"-# Should not be connected to a 2nd tet center\n"
		"   (this is acceptable in theory as the wedge is only at one end, but\n"
		"    in practice it may cause confusion and thus we avoid it)\n"
		"-# Preferably is not in a cycle\n"
		"-# Preferably is a terminal H\n"
		"\n"
		"If no bond can be found that matches rules 1 and 2 (and in theory this is possible)\n"
		"then an error message is logged and the function returns false. (If you find an\n"
		"example where this occurs, please file a bug.)\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >\n"
		":param updown: A map of OBStereo::BondDirection for each hash/wedge bond\n"
		":type from: std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > >\n"
		":param from: A map of OBStereo::Ref indicating the origin of each hash/wedge bond\n"
		":rtype: boolean\n"
		":return: True or False depending on whether the conversion was successful\n"
		"Since: version 2.3\n"
		""},
	 { "GetUnspecifiedCisTrans", _wrap_GetUnspecifiedCisTrans, METH_O, "\n"
		"Return a set of double bonds corresponding to the OBCisTransStereo objects\n"
		"for which the stereochemistry is undefined.\n"
		"\n"
		"Note that this functions just iterates over the existing OBCisTransStereo\n"
		"objects - it does not try to identify new ones.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":rtype: std::set< OpenBabel::OBBond * >\n"
		":return: A set of bonds with unspecified cis/trans stereochemistry\n"
		"Since: version 2.3\n"
		""},
	 { "StereoRefToImplicit", _wrap_StereoRefToImplicit, METH_VARARGS, "\n"
		"Convert any reference to ``atomId`` in a stereo object to an OBStereo::ImplicitRef.\n"
		"This function is called from OBMol::DeleteHydrogens()\n"
		"(via OBMol::DeleteHydrogen()) to remove any explicit references to a\n"
		"hydrogen atom that has been deleted. However, the code is not specific\n"
		"to hydrogen atoms and could be used for other atoms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":type atomId: int\n"
		":param atomId: The Id of the atom to be converted to an OBStereo::ImplicitRef\n"
		"Since: version 2.3\n"
		""},
	 { "ImplicitRefToStereo", _wrap_ImplicitRefToStereo, METH_VARARGS, "\n"
		"Convert any reference to an OBStereo::ImplicitRef attached to ``centerId``\n"
		"in a stereo object to an explicit reference to ``newId``.\n"
		"This function is called from OBMol::AddHydrogens() and\n"
		"OBMol::AddHydrogen() to convert any implicit references to a\n"
		"hydrogen atom that has just been added. However, the code is not specific\n"
		"to hydrogen atoms and could be used for other atoms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":type centerId: int\n"
		":param centerId: The Id of the atom to which the new explicit atom is attached\n"
		":type newId: int\n"
		":param newId: The Id of the atom which was previously an OBStereo::ImplicitRef\n"
		"Since: version 2.4\n"
		""},
	 { "CisTransFrom0D", _wrap_CisTransFrom0D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom0D() with the ``addToMol`` parameter is set\n"
		"to true. There is no algorithm used here, all specified flags will be\n"
		"set to false.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom0D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "FindStereogenicUnits", _wrap_FindStereogenicUnits, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"\n"
		"Find the stereogenic units in a molecule using a set of rules. 1\n"
		"\n"
		"The potential stereocenters are identified first. A potential tetrahedral\n"
		"stereogenic atom is any atom meeting the following criteria:\n"
		"\n"
		"- sp3 hybridization\n"
		"- at least 3 \"heavy\" neighbors\n"
		"\n"
		"Nitrogen is treated as a special case since the barrier of inversion is\n"
		"low in many cases making the atom non-stereogenic. Only bridge-head\n"
		"nitrogen atoms (i.e. nitrogen has 3 neighbors in rings) will be\n"
		"considered stereogenic.\n"
		"\n"
		"Potential stereogenic double bonds are identified using another set of\n"
		"simple criteria:\n"
		"\n"
		"- must be a double bond\n"
		"- must not be in a ring\n"
		"- both begin and end atom should have at least one single bond\n"
		"\n"
		"True stereocenters (i.e. stereocenters with topologically different\n"
		"ligands) are identified first. For tetrahedral stereocenters, true\n"
		"stereocenters will have 4 different neighbor atom symmetry classes\n"
		"and this can be expressed using T1234 to classify these stereocenters.\n"
		"For stereogenic bonds, a similar classification C12 can be used but\n"
		"both begin and end atom have their own classification and the bond\n"
		"is only a true stereocenter if both atoms are C12.\n"
		"\n"
		"Para stereocenters are all stereocenters where there are at least two\n"
		"equivalent neighbor atom symmetry classes. These are T1123, T1112, T1111\n"
		"and T1122 for tetrahedral stereocenters and C11 for double bonds. To\n"
		"determine which of the remaining potential stereocenters really are\n"
		"stereocenters, a set of rules is used. 1\n"
		"\n"
		"Rule 1 is applied recusively:\n"
		"\n"
		"All rings are merged \"mergedRings\". A merged ring is simply a fragment consisting\n"
		"of all atoms of a ring system (bridged, spiro, adjacent, ...). If two rings in the\n"
		"SSSR set share an atom, they are merged.\n"
		"\n"
		"Each merged must at least have two para-stereocenters (or 1 true + 1 para) in order\n"
		"for the para-stereocenter to be valid. This is repeated until no new stereocenters\n"
		"are identified.\n"
		"\n"
		"rule 1a for double bonds:\n"
		"- bond atom in ring has two identical symmetry classes for it's neighbor atoms (-> para)\n"
		"- other bond atom:\n"
		"  - has two different symmetry classes for it's neighbours -> new stereocenter\n"
		"  - has two identical symmetry classes, but the ligand contains at least 1 true or para stereocenter -> new stereocenter\n"
		"\n"
		"rule 1b for tetracoord atoms:\n"
		"- at least two neighbour symmetry classes are the same (-> para)\n"
		"- other pair:\n"
		"  - has two different symmetry classes for it's neighbours -> new stereocenter\n"
		"  - has two identical symmetry classes, but the ligand contains at least 1 true or para stereocenter -> new stereocenter\n"
		"\n"
		"Rules 2 and 3 are applied sequential (i.e. only once).\n"
		"\n"
		"Rule 2a for tetracoordinate carbon:\n"
		"- 1 or 2 pair identical ligands\n"
		"- each ligand contains at least 1 true-stereocenter or 2 para-stereocenters (from rule 1)\n"
		"\n"
		"Rule 2b for tetracoordinate carbon:\n"
		"- 3 or 4 identical ligands with at least\n"
		"  - 2 true-stereocenters\n"
		"  - 2 separate assemblies of para-stereocenters (from rule 1)\n"
		"\n"
		"Rule 3 for double bonds:\n"
		"- 1 or 2 pair identical ligands (on begin and end atom)\n"
		"- each pair contains at least 1 true-stereocenter or 2 para-stereocenters (from rule 1)\n"
		"\n"
		"\n"
		"     Reference:\n"
		"     [1] M. Razinger, K. Balasubramanian, M. Perdih, M. E. Munk, Stereoisomer\n"
		"     Generation in Computer-Enhanced Structure Elucidation, J. Chem. Inf.\n"
		"     Comput. Sci. 1993, 33, 812-825\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Find the stereogenic units in a molecule making use of the automorphisms.\n"
		"\n"
		"The potential stereocenters are identified first. A potential tetrahedral\n"
		"stereogenic atom is any atom meeting the following criteria:\n"
		"\n"
		"- sp3 hybridization\n"
		"- at least 3 \"heavy\" neighbors\n"
		"\n"
		"Nitrogen is treated as a special case since the barrier of inversion is\n"
		"low in many cases making the atom non-stereogenic. Only bridge-head\n"
		"nitrogen atoms (i.e. nitrogen has 3 neighbors in rings) will be\n"
		"considered stereogenic.\n"
		"\n"
		"Potential stereogenic double bonds are identified using another set of\n"
		"simple criteria:\n"
		"\n"
		"- must be a double bond\n"
		"- must not be in a ring\n"
		"- both begin and end atom should have at least one single bond\n"
		"\n"
		"Once the potential stereocenters are found, the automorphisms are the key\n"
		"to identifying real stereogenic units. Automorphisms can be seen as\n"
		"permutations that permutate a graph back to the same graph. Such a\n"
		"permutation can only exchange atoms with the same symmetry class and it\n"
		"follows that the use of automorphisms takes symmetry into account. The\n"
		"definitions below use a concept where the automorphisms cause inversions\n"
		"of configuration to potential stereocenters. Such an inversion occurs\n"
		"whenever an automorphism exchanges two equivalent (i.e. with the same\n"
		"symmetry class) neighbor atoms attached to the potential stereogenic unit.\n"
		"\n"
		"Title: Definition for tetrahedral stereocenters:\n"
		"A potential stereocenter really is a stereocenter if there exists no automorphic\n"
		"permutation causing an inversion of the configuration of only the potential\n"
		"stereogenic unit under consideration.\n"
		"If there exists at least one automorphic permutation causing an inversion of\n"
		"the configuration, then the potential stereogenic center can be a stereogenic\n"
		"center if the number of topologically equivalent neighbors (ligands) of the\n"
		"potential stereogenic center is less than or equal to the number of different\n"
		"configurations of these ligands. 1\n"
		"\n"
		"The actual number of configurations needed for the ligands depends on the\n"
		"classification (i.e. T1234, T1123, ...) of the stereo center. These classes\n"
		"reflect the symmetry classes of the neighbor atoms of the center.\n"
		"\n"
		"- T1123: 1 true stereocenter OR 2 para stereocenters\n"
		"- T1122: 1 true stereocenter OR 2 para stereocenters (for both)\n"
		"- T1112: 2 true stereocenters OR 2 para stereocenter assemblies\n"
		"- T1111: 2 true stereocenters OR 2 para stereocenter assemblies\n"
		"\n"
		"Title: Definition for double bond stereocenters:\n"
		"A potential stereogenic double bond really is a stereogenic bond if there\n"
		"exists no automorphic permutation causing an inversion of the configuration\n"
		"of only the potential stereogenic unit under consideration. The bond can still\n"
		"be a stereogenic bond if there exists such an automorphism when the number of\n"
		"configurations of the pair of topologically equivalent geminal ligands, which\n"
		"are exchanged by the automorphism, is greater than or equal to two (i.e. the\n"
		"number of topologically equivalent geminal ligands. 1\n"
		"\n"
		"For stereogenic bonds, there is only one case but both begin and end atom\n"
		"have to be checked.\n"
		"\n"
		"- C11: 1 true stereocenter OR 1 para stereocenter\n"
		"\n"
		"These criteria are analogous to the rules from the Razinger paper on\n"
		"stereoisomer generation. Since the existence of stereocenters can depend\n"
		"on the existence of other stereocenters (in the ligands), the stereocenters\n"
		"are found by iterating until no new stereocenters are found.\n"
		"\n"
		"     Reference:\n"
		"     [1] M. Perdih, M. Razinger, Stereochemistry and Sequence Rules:\n"
		"     A Proposal for Modification of Cahn-Ingold-Prelog System,\n"
		"     Tetrahedron: Asymmetry, 1994, Vol. 5, No. 5, 835-861\n"
		""},
	 { "new_OBRotorRule", _wrap_new_OBRotorRule, METH_VARARGS, NULL},
	 { "delete_OBRotorRule", _wrap_delete_OBRotorRule, METH_O, NULL},
	 { "OBRotorRule_IsValid", _wrap_OBRotorRule_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this rotor rule is valid (i.e., is the SMARTS pattern valid)\n"
		""},
	 { "OBRotorRule_GetReferenceAtoms", _wrap_OBRotorRule_GetReferenceAtoms, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: a copy of the reference atom indexes inside the SMARTS pattern  These should be freed after use.\n"
		""},
	 { "OBRotorRule_SetDelta", _wrap_OBRotorRule_SetDelta, METH_VARARGS, "Set the resolution (delta) of a torsional step in degrees"},
	 { "OBRotorRule_GetDelta", _wrap_OBRotorRule_GetDelta, METH_O, "\n"
		":rtype: float\n"
		":return: the resolution (delta) of a torsional step in degrees\n"
		""},
	 { "OBRotorRule_GetTorsionVals", _wrap_OBRotorRule_GetTorsionVals, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a reference to the dihedral angles to evaluate (in radians)\n"
		""},
	 { "OBRotorRule_GetSmartsString", _wrap_OBRotorRule_GetSmartsString, METH_O, "\n"
		":rtype: string\n"
		":return: the text of the SMARTS pattern for this rule\n"
		""},
	 { "OBRotorRule_GetSmartsPattern", _wrap_OBRotorRule_GetSmartsPattern, METH_O, "\n"
		":rtype: :py:class:`OBSmartsPattern`\n"
		":return: the exact OBSmartsPattern object for this rule\n"
		""},
	 { "OBRotorRule_swigregister", OBRotorRule_swigregister, METH_O, NULL},
	 { "OBRotorRule_swiginit", OBRotorRule_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorRules", _wrap_new_OBRotorRules, METH_NOARGS, NULL},
	 { "delete_OBRotorRules", _wrap_delete_OBRotorRules, METH_O, NULL},
	 { "OBRotorRules_SetFilename", _wrap_OBRotorRules_SetFilename, METH_VARARGS, "Set the filename to be used for the database. Default = torlib.txt"},
	 { "OBRotorRules_GetRotorIncrements", _wrap_OBRotorRules_GetRotorIncrements, METH_VARARGS, "\n"
		" Determine the torsional angles to evaluate based on the database :type mol: :py:class:`OBMol`\n"
		":param mol: molecule to evaluate :type bond: :py:class:`OBBond`\n"
		":param bond: rotatable bond to evaluate :type refs: int\n"
		":param refs: set to be the atom indexes (in mol) of the dihedral angle :type vals: std::vector< double,std::allocator< double > >\n"
		":param vals: set to be the list of angles to evaluate (in radians) :type delta: float\n"
		":param delta: potential dihedral angle steps (in degrees)\n"
		""},
	 { "OBRotorRules_Quiet", _wrap_OBRotorRules_Quiet, METH_O, "Turn off debugging output from GetRotorIncrements()"},
	 { "OBRotorRules_swigregister", OBRotorRules_swigregister, METH_O, NULL},
	 { "OBRotorRules_swiginit", OBRotorRules_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotor", _wrap_new_OBRotor, METH_NOARGS, "Constructor."},
	 { "delete_OBRotor", _wrap_delete_OBRotor, METH_O, "Destructor."},
	 { "OBRotor_SetBond", _wrap_OBRotor_SetBond, METH_VARARGS, "Set the OBBond associated with this OBRotor."},
	 { "OBRotor_SetRings", _wrap_OBRotor_SetRings, METH_O, "\n"
		"Set the rings associated with this bond (if it's a ring bond)\n"
		"Since: Version 2.4\n"
		""},
	 { "OBRotor_SetIdx", _wrap_OBRotor_SetIdx, METH_VARARGS, "Set the index for this rotor. Used by OBRotorList"},
	 { "OBRotor_SetDihedralAtoms", _wrap_OBRotor_SetDihedralAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Set the dihedral atoms.\n"
		":type ref: std::vector< int,std::allocator< int > >\n"
		":param ref: The dihedral atom indexes. These indexes start from 1.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Set the dihedral atoms.\n"
		":type ref: int\n"
		":param ref: The dihedral atom indexes. These indexes start from 1.\n"
		""},
	 { "OBRotor_SetRotAtoms", _wrap_OBRotor_SetRotAtoms, METH_VARARGS, "\n"
		"Set the atom indexes that will be displaced when this rotor\n"
		"changes torsion angle. These indexes start from 0 and are multiplied\n"
		"by 3 for easy coordinate access.\n"
		""},
	 { "OBRotor_SetTorsionValues", _wrap_OBRotor_SetTorsionValues, METH_VARARGS, "Set the possible torsion values or angles."},
	 { "OBRotor_SetFixedBonds", _wrap_OBRotor_SetFixedBonds, METH_VARARGS, "Set the bonds that will be fixed."},
	 { "OBRotor_SetToAngle", _wrap_OBRotor_SetToAngle, METH_VARARGS, "\n"
		"Rotate the atoms in the specified ``coordinates`` to the specified angle.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to rotate.\n"
		":type setang: float\n"
		":param setang: The new torsion angle in radians.\n"
		""},
	 { "OBRotor_SetRotor", _wrap_OBRotor_SetRotor, METH_VARARGS, "\n"
		"Rotate the atoms in the specified ``coordinates``. This function does not\n"
		"require any precomputation and will compute all needed information when\n"
		"needed.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates for the molecules as pointer to double.\n"
		":type next: int\n"
		":param next: The index of the new rotor angle. This is an index for the\n"
		"    GetTorsionValues() list.\n"
		":type prev: int, optional\n"
		":param prev: If specified, the torsion current torsion angle can be\n"
		"    looked up and does not have to be calculated again.\n"
		""},
	 { "OBRotor_Precompute", _wrap_OBRotor_Precompute, METH_VARARGS, "\n"
		"Precompute the reference angle and inverse bond length of this rotor for\n"
		"a single conformer. This function should be used in combination with\n"
		"Set(double *coordinates, int idx).\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to use in the computation.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OBMol mol;\n"
		"    ...\n"
		"\n"
		"    unsigned int numCoords = mol.NumAtoms() * 3;\n"
		"    double *coords = mol.GetCoordinates();\n"
		"    OBRotor rotor;\n"
		"    rotor.SetBond(mol.GetBond(3));\n"
		"\n"
		"    // set the possible torsion values\n"
		"    std::vector<double> angles;\n"
		"    angles.push_back(0.0);\n"
		"    angles.push_back(3.1415);\n"
		"    rotor.SetTorsionValues(angles);\n"
		"\n"
		"    // precompute inverse bond length (i.e. the bond length of bond with index 3\n"
		"    // using the specified coordinates) and reference angle (i.e. torsion angle\n"
		"    //in coords)\n"
		"    rotor.Precompute(coords);\n"
		"\n"
		"    // copy coordinates to coords_1\n"
		"    double *coords_1 = new double[numCoords];\n"
		"    for (unsigned int i = 0; i < numCoords; ++i)\n"
		"      coords_1[i] = coords[i];\n"
		"    // rotate the atoms in coords_1 to angle with index 0 (i.e. 0.0 degrees)\n"
		"    // note: on input, the coordinates should be the same as the coordinates used\n"
		"    //       to precompute the inverse bond length and reference angle (in other\n"
		"    //       words, the inverse magnitude and reference angle in the specfied\n"
		"    //       coordinates should be the same as the one used for Precompute)\n"
		"    rotor.Set(coords_1, 0)\n"
		"\n"
		"    // copy coordinates to coords_2\n"
		"    double *coords_2 = new double[numCoords];\n"
		"    for (unsigned int i = 0; i < numCoords; ++i)\n"
		"      coords_2[i] = coords[i];\n"
		"    // rotate the atoms in coords_2 to angle with index 1 (i.e. 180.0 degrees)\n"
		"    rotor.Set(coords_2, 1)\n"
		"\n"
		"    delete coords_1;\n"
		"    delete coords_2;\n"
		""},
	 { "OBRotor_Precalc", _wrap_OBRotor_Precalc, METH_VARARGS, "\n"
		"Precompute the inverse bond lengths, rotation matrices for all\n"
		"specified conformers and all possible torsion values. This method is\n"
		"used in combination with Set(double *coordinates, int conformer, int idx).\n"
		":type conformers: std::vector< double *,std::allocator< double * > >\n"
		":param conformers: The pointers to the conformer coordinates\n"
		""},
	 { "OBRotor_Set", _wrap_OBRotor_Set, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Rotate the specified ``coordinates`` by using the specified rotation matrix.\n"
		"\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Rotate the ``coordinates`` to set the torsion angle of this rotor to the angle\n"
		"specified by the index ``idx``. Make sure to call Precompute before calling\n"
		"this function.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to rotate.\n"
		":type idx: int\n"
		":param idx: The index of the torsion angle in the GetTorsionValues() list.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Rotate the ``coordinates`` to set the torsion to the torsion value with the\n"
		"specified ``index``. The coordinates should be the same as the conformer used\n"
		"for calling Precalc (i.e. conformers[conformer] == coordinates). Make sure\n"
		"to call Precalc before calling this method.\n"
		":type coordinates: float\n"
		":param coordinates: The conformer coordinates.\n"
		":type conformer: int\n"
		":param conformer: The conformer index in the conformer list given to Precalc().\n"
		":type idx: int\n"
		":param idx: The torsion value index in the GetTorsionValues() list.\n"
		""},
	 { "OBRotor_GetBond", _wrap_OBRotor_GetBond, METH_O, "Get the OBBond object associated with this OBRotor."},
	 { "OBRotor_Size", _wrap_OBRotor_Size, METH_O, "\n"
		"Get the number of possible torsion angles for this OBRotor. This\n"
		"is the length of the GetTorsionValues() list.\n"
		""},
	 { "OBRotor_GetIdx", _wrap_OBRotor_GetIdx, METH_O, "Get the index for this rotor (index in an OBRotorList)."},
	 { "OBRotor_GetDihedralAtoms", _wrap_OBRotor_GetDihedralAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Get the dihedral atom indexes. These indexes start from 1.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Get the dihedral atom indexes. These indexes start from 1.\n"
		""},
	 { "OBRotor_GetTorsionValues", _wrap_OBRotor_GetTorsionValues, METH_O, "Get the possible torsion angles for this OBRotor."},
	 { "OBRotor_GetFixedBonds", _wrap_OBRotor_GetFixedBonds, METH_O, "\n"
		"Get an OBBitVec objects with bits set for all bonds that are fixed.\n"
		"Bonds are indexed from 0.\n"
		""},
	 { "OBRotor_CalcTorsion", _wrap_OBRotor_CalcTorsion, METH_VARARGS, "\n"
		"Calculate the torsion for this OBRotor using the specified coordinates.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates (e.g. OBMol::GetCoordinates()).\n"
		":rtype: float\n"
		":return: The torsion angle in radians.\n"
		""},
	 { "OBRotor_CalcBondLength", _wrap_OBRotor_CalcBondLength, METH_VARARGS, "\n"
		"Calculate the bond length for this OBRotor using the specified coordinates.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates (e.g. OBMol::GetCoordinates()).\n"
		""},
	 { "OBRotor_BeginTorIncrement", _wrap_OBRotor_BeginTorIncrement, METH_O, ""},
	 { "OBRotor_EndTorIncrement", _wrap_OBRotor_EndTorIncrement, METH_O, NULL},
	 { "OBRotor_RemoveSymTorsionValues", _wrap_OBRotor_RemoveSymTorsionValues, METH_VARARGS, ""},
	 { "OBRotor_SetDelta", _wrap_OBRotor_SetDelta, METH_VARARGS, "Deprecated: Has no effect."},
	 { "OBRotor_GetDelta", _wrap_OBRotor_GetDelta, METH_O, "Deprecated: Has no effect."},
	 { "OBRotor_GetFixedAtoms", _wrap_OBRotor_GetFixedAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_SetFixedAtoms", _wrap_OBRotor_SetFixedAtoms, METH_VARARGS, "Deprecated: See SetFixedBonds"},
	 { "OBRotor_GetEvalAtoms", _wrap_OBRotor_GetEvalAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_SetEvalAtoms", _wrap_OBRotor_SetEvalAtoms, METH_VARARGS, "Deprecated:"},
	 { "OBRotor_GetRotAtoms", _wrap_OBRotor_GetRotAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_GetResolution", _wrap_OBRotor_GetResolution, METH_O, "Deprecated: Bad name, see GetTorsionValues()"},
	 { "OBRotor_SetNumCoords", _wrap_OBRotor_SetNumCoords, METH_VARARGS, "Deprecated:"},
	 { "OBRotor_swigregister", OBRotor_swigregister, METH_O, NULL},
	 { "OBRotor_swiginit", OBRotor_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorList", _wrap_new_OBRotorList, METH_NOARGS, "Constructor."},
	 { "delete_OBRotorList", _wrap_delete_OBRotorList, METH_O, "Destructor."},
	 { "OBRotorList_Clear", _wrap_OBRotorList_Clear, METH_O, "Clear the internal list of rotors and reset."},
	 { "OBRotorList_Size", _wrap_OBRotorList_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotors in this list\n"
		""},
	 { "OBRotorList_IsFixedBond", _wrap_OBRotorList_IsFixedBond, METH_VARARGS, "\n"
		"When no atoms/bonds are fixed or when bonds are fixed, this function will\n"
		"return true if the bond is fixed. When using the deprecated fixed atoms,\n"
		"this function will return true if the bond and at least one neighboring\n"
		"bond has fixed atoms.\n"
		""},
	 { "OBRotorList_HasFixedBonds", _wrap_OBRotorList_HasFixedBonds, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this rotor list has any fixed bonds.\n"
		""},
	 { "OBRotorList_RemoveSymVals", _wrap_OBRotorList_RemoveSymVals, METH_VARARGS, "Rotates each bond to zero and 180 degrees and tests if the 2 conformers are duplicates.  if so - the symmetric torsion values are removed from consideration during a search"},
	 { "OBRotorList_HasRingRotors", _wrap_OBRotorList_HasRingRotors, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this rotor list has any ring bonds.\n"
		"Since: version 2.4\n"
		""},
	 { "OBRotorList_Setup", _wrap_OBRotorList_Setup, METH_VARARGS, "\n"
		"Setup this rotor list for the supplied molecule. This method calls\n"
		"FindRotors(), SetEvalAtoms(), and AssignTorVals().\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type sampleRings: boolean, optional\n"
		":param sampleRings: Whether to sample ring conformers - default = false\n"
		":rtype: boolean\n"
		":return: True if rotatable bonds were found.\n"
		""},
	 { "OBRotorList_SetFixedBonds", _wrap_OBRotorList_SetFixedBonds, METH_VARARGS, "Set the bonds that will be fixed."},
	 { "OBRotorList_Init", _wrap_OBRotorList_Init, METH_VARARGS, "Initialize the private OBRotorRules database from a specific file."},
	 { "OBRotorList_SetQuiet", _wrap_OBRotorList_SetQuiet, METH_O, "Turn off debugging output."},
	 { "OBRotorList_SetRotAtoms", _wrap_OBRotorList_SetRotAtoms, METH_VARARGS, "\n"
		" Set the atoms to rotate from the dihedral atoms for each rotor Uses OBRotor->GetDihedralAtoms() to call OBRotor->SetRotAtoms() and standarizes the dihedral angles via OBRotor->SetDihedralAtoms() :rtype: boolean\n"
		":return: True\n"
		""},
	 { "OBRotorList_FindRotors", _wrap_OBRotorList_FindRotors, METH_VARARGS, "\n"
		"Find all potentially rotatable bonds in the molecule. This method uses\n"
		"OBBond::IsRotor() for initial evaluation which depends on ring perception\n"
		"(i.e. ring bonds are considered rotatable). Fixed bonds, specified using the\n"
		"deprecated fixed atoms or the new fixed bonds methods are not added to\n"
		"the list. All rotatable bonds will be sorted by their graph theoretical\n"
		"distance (GTD) score (see OBMol::GetGTDVector()). This results in the\n"
		"the rotors going from the inside to the outside of the mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: whether to sample ring bonds from analysis (default = false)\n"
		":rtype: boolean\n"
		":return: True.\n"
		""},
	 { "OBRotorList_SetEvalAtoms", _wrap_OBRotorList_SetEvalAtoms, METH_VARARGS, "\n"
		" Determines which atoms should be used to calculate the internal energy if the dihedral angle of the rotor is modified :rtype: boolean\n"
		":return: True\n"
		""},
	 { "OBRotorList_AssignTorVals", _wrap_OBRotorList_AssignTorVals, METH_VARARGS, "\n"
		"Using the OBRotorRules database, set the torsion values (and delta)\n"
		"to be evaluated and tested. This method also sets the rotatable atoms\n"
		"for the rotors to the smallest possible set. For each bond there are\n"
		"two candidate sets, one on either side. The smallest of these is used\n"
		"and the torsion indexes for the rotor are inverted if needed\n"
		"(i.e. a-b-c-d -> d-c-b-a).\n"
		""},
	 { "OBRotorList_BeginRotor", _wrap_OBRotorList_BeginRotor, METH_VARARGS, "\n"
		"Initialize the ``i`` iterator and get a pointer to the first OBRotor.\n"
		":type i: OpenBabel::OBRotorIterator\n"
		":param i: OBRotorIterator object.\n"
		""},
	 { "OBRotorList_NextRotor", _wrap_OBRotorList_NextRotor, METH_VARARGS, "\n"
		"Get a pointer to the next iterator.\n"
		":type i: OpenBabel::OBRotorIterator\n"
		":param i: OBRotorIterator object.\n"
		""},
	 { "OBRotorList_BeginRotors", _wrap_OBRotorList_BeginRotors, METH_O, "Get the rotor list begin iterator."},
	 { "OBRotorList_EndRotors", _wrap_OBRotorList_EndRotors, METH_O, "Get the rotor list end iterator."},
	 { "OBRotorList_IdentifyEvalAtoms", _wrap_OBRotorList_IdentifyEvalAtoms, METH_VARARGS, "Deprecated: Not declared. Use Setup() for top-level functionality"},
	 { "OBRotorList_SetFixAtoms", _wrap_OBRotorList_SetFixAtoms, METH_VARARGS, "\n"
		"Set the list of fixed (invariant) atoms to the supplied OBBitVec\n"
		"Deprecated: See SetFixedBonds()\n"
		""},
	 { "OBRotorList_HasFixedAtoms", _wrap_OBRotorList_HasFixedAtoms, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this rotor list has any fixed (invariant) atoms\n"
		"Deprecated: See HasFixedBonds()\n"
		""},
	 { "OBRotorList_IgnoreSymmetryRemoval", _wrap_OBRotorList_IgnoreSymmetryRemoval, METH_O, "Has no effect Deprecated: Currently has no effect"},
	 { "OBRotorList_SetRotAtomsByFix", _wrap_OBRotorList_SetRotAtomsByFix, METH_VARARGS, "Set the atoms to rotate from the dihedral atoms for each rotor Insures the fixed atoms are respected, but otherwise functions like SetRotAtoms()"},
	 { "OBRotorList_swigregister", OBRotorList_swigregister, METH_O, NULL},
	 { "OBRotorList_swiginit", OBRotorList_swiginit, METH_VARARGS, NULL},
	 { "new_rotor_digit", _wrap_new_rotor_digit, METH_VARARGS, NULL},
	 { "rotor_digit_set_size", _wrap_rotor_digit_set_size, METH_VARARGS, NULL},
	 { "rotor_digit_set_state", _wrap_rotor_digit_set_state, METH_VARARGS, NULL},
	 { "rotor_digit_get_state", _wrap_rotor_digit_get_state, METH_O, NULL},
	 { "rotor_digit_size", _wrap_rotor_digit_size, METH_O, NULL},
	 { "rotor_digit_next", _wrap_rotor_digit_next, METH_O, NULL},
	 { "delete_rotor_digit", _wrap_delete_rotor_digit, METH_O, NULL},
	 { "rotor_digit_swigregister", rotor_digit_swigregister, METH_O, NULL},
	 { "rotor_digit_swiginit", rotor_digit_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorKeys", _wrap_new_OBRotorKeys, METH_NOARGS, "Constructor"},
	 { "OBRotorKeys_Clear", _wrap_OBRotorKeys_Clear, METH_O, "Clear all rotors"},
	 { "OBRotorKeys_NumKeys", _wrap_OBRotorKeys_NumKeys, METH_O, "Number of rotor keys (= number of possible conformers)"},
	 { "OBRotorKeys_AddRotor", _wrap_OBRotorKeys_AddRotor, METH_VARARGS, "\n"
		" Add a rotor :type size: int\n"
		":param size: the rotor resolution\n"
		""},
	 { "OBRotorKeys_Next", _wrap_OBRotorKeys_Next, METH_O, "\n"
		" Select the next rotor key :rtype: boolean\n"
		":return: true if there are more rotor keys\n"
		""},
	 { "OBRotorKeys_GetKey", _wrap_OBRotorKeys_GetKey, METH_O, "\n"
		" Get the currently selected rotor key :rtype: std::vector< int,std::allocator< int > >\n"
		":return: current rotor key\n"
		""},
	 { "delete_OBRotorKeys", _wrap_delete_OBRotorKeys, METH_O, NULL},
	 { "OBRotorKeys_swigregister", OBRotorKeys_swigregister, METH_O, NULL},
	 { "OBRotorKeys_swiginit", OBRotorKeys_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotamerList", _wrap_new_OBRotamerList, METH_NOARGS, NULL},
	 { "delete_OBRotamerList", _wrap_delete_OBRotamerList, METH_O, NULL},
	 { "OBRotamerList_Setup", _wrap_OBRotamerList_Setup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Set up a rotamer list based on an already created OBRotorList\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Set up a rotamer list based on the supplied reference atoms and the number of rotors :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to evaluate :type ref: unsigned char\n"
		":param ref: An array of the 4 dihedral atoms for each rotor :type nrotors: int\n"
		":param nrotors: The number of rotors (i.e., the size of ref / 4)\n"
		""},
	 { "OBRotamerList_NumRotors", _wrap_OBRotamerList_NumRotors, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotatable bonds considered\n"
		""},
	 { "OBRotamerList_NumRotamers", _wrap_OBRotamerList_NumRotamers, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotamer (conformation) coordinate sets\n"
		""},
	 { "OBRotamerList_AddRotamer", _wrap_OBRotamerList_AddRotamer, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Add a rotamer to the list based on the supplied coordinate set as a double*\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		""},
	 { "OBRotamerList_AddRotamers", _wrap_OBRotamerList_AddRotamers, METH_VARARGS, "Add ``nconf`` rotamers based on ``as`` an array of configurations much like AddRotamer()"},
	 { "OBRotamerList_GetReferenceArray", _wrap_OBRotamerList_GetReferenceArray, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: A reference array (as used by AddRotamer() as a configuration of the individual rotor bonds\n"
		""},
	 { "OBRotamerList_BeginRotamer", _wrap_OBRotamerList_BeginRotamer, METH_O, NULL},
	 { "OBRotamerList_EndRotamer", _wrap_OBRotamerList_EndRotamer, METH_O, NULL},
	 { "OBRotamerList_CreateConformerList", _wrap_OBRotamerList_CreateConformerList, METH_VARARGS, "\n"
		" Create a conformer list using the internal base set of coordinates :rtype: std::vector< double *,std::allocator< double * > >\n"
		":return: The set of coordinates by rotating the bonds in each rotamer\n"
		""},
	 { "OBRotamerList_ExpandConformerList", _wrap_OBRotamerList_ExpandConformerList, METH_VARARGS, "\n"
		" Create a conformer list using the internal base set of coordinates :rtype: void\n"
		":return: The set of coordinates as a reference in ``confs``\n"
		""},
	 { "OBRotamerList_SetCurrentCoordinates", _wrap_OBRotamerList_SetCurrentCoordinates, METH_VARARGS, NULL},
	 { "OBRotamerList_SetBaseCoordinateSets", _wrap_OBRotamerList_SetBaseCoordinateSets, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Copies the mol's conformers (the coordinates, NOT the pointers) into the object as base coordinates\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Copies the coordinates in bc, NOT the pointers, into this object :type bc: std::vector< double *,std::allocator< double * > >\n"
		":param bc: The conformer set for the molecule\n"
		"        :type N: int\n"
		"        :param N:  The number of atoms in the molecule\n"
		""},
	 { "OBRotamerList_NumBaseCoordinateSets", _wrap_OBRotamerList_NumBaseCoordinateSets, METH_O, "\n"
		":rtype: int\n"
		":return: The number of \"base\" coordinate sets (i.e., the number of conformers in the base OBMol)\n"
		""},
	 { "OBRotamerList_GetBaseCoordinateSet", _wrap_OBRotamerList_GetBaseCoordinateSet, METH_VARARGS, "Get a pointer to a specific base pointer (i.e., specific conformer)"},
	 { "OBRotamerList_NumAtoms", _wrap_OBRotamerList_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in the base OBMol\n"
		""},
	 { "OBRotamerList_swigregister", OBRotamerList_swigregister, METH_O, NULL},
	 { "OBRotamerList_swiginit", OBRotamerList_swiginit, METH_VARARGS, NULL},
	 { "new_OBSpectrophore", _wrap_new_OBSpectrophore, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Default constructor to create an OBSpectrophore object.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor to create an OBSpectrophore object by taking\n"
		"        a copy from another OBSpectrophore object.\n"
		"\n"
		"        :type sphore: :py:class:`OBSpectrophore`\n"
		"        :param sphore: A reference to the source OBSpectrophore object\n"
		""},
	 { "delete_OBSpectrophore", _wrap_delete_OBSpectrophore, METH_O, "Default destructor object"},
	 { "OBSpectrophore_SetResolution", _wrap_OBSpectrophore_SetResolution, METH_VARARGS, "\n"
		"Method to set the resolution at which Spectrophores will be\n"
		"        calculated.\n"
		"\n"
		"        Sets the resolution at which Spectrophores are calculated. The\n"
		"        resolution is an arbitrary positive number larger than 0.0, and is used\n"
		"        to increase the box size for the calculation of Spectrophores in\n"
		"        all directions. For example, a resolution of 3 means that the box size\n"
		"        will be increased by a value of 3 &Aring; in all directions. Smaller\n"
		"        values for the resolution have the effect that small differences in the\n"
		"        atomic properties will result in more enhanced differences in the\n"
		"        resulting Spectrophores, while larger resolution values will result\n"
		"        in the smoothing of local property differences. All values <= 0.0\n"
		"        are automatically reset to the default value of 3.0.\n"
		"\n"
		"        :type r: float, optional\n"
		"        :param r: The desired resolution expressed as a double number\n"
		""},
	 { "OBSpectrophore_SetAccuracy", _wrap_OBSpectrophore_SetAccuracy, METH_VARARGS, "\n"
		"Method to set the accuracy at which Spectrophores will be\n"
		"        calculated.\n"
		"\n"
		"        Sets the accuracy by which Spectrophores are calculated. The\n"
		"        accuracy is linked to the angular step increment that is used to\n"
		"        calculate Spectrophores. The accuracy parameter is an enumeration\n"
		"        type defined by OBSpectrophore::AccuracyOption. should be a number ranging\n"
		"        between 0 and 9 (inclusive), where a value of 9 means a very high\n"
		"        accuracy, and 0 means no accuracy at all. Values between 2-6 are a good\n"
		"        compromise between accuracy and speed. The following table provides the\n"
		"        link between the parameter value and the angular step size:\n"
		"\n"
		"        |**Accuracy**|**OBSpectrophore::AccuracyOption parameter**|**Minimal angular steps size**|\n"
		"        |Extremely high|AngStepSize1|1º|\n"
		"        |Extremely high|AngStepSize2|2º|\n"
		"        |Very High|AngStepSize5|5º|\n"
		"        |Very high|AngStepSize10|10º|\n"
		"        |High|AngStepSize15|15º|\n"
		"        |Default|AngStepSize20|20º|\n"
		"        |Low|AngStepSize30|30º|\n"
		"        |Very low|AngStepSize36|36º|\n"
		"        |Extremely low|AngStepSize45|45º|\n"
		"        |Extremely low|AngStepSize60|60º|\n"
		"\n"
		"\n"
		"        :type a: int, optional\n"
		"        :param a: The desired accuracy expressed as an\n"
		"            OBSpectrophore::AccuracyOption enumeration type\n"
		""},
	 { "OBSpectrophore_SetStereo", _wrap_OBSpectrophore_SetStereo, METH_VARARGS, "\n"
		"Method to set the required stereoselectivity of the resulting\n"
		"        Spectrophores.\n"
		"\n"
		"        Sets the stereoselectivity of the generated Spectrophores. This is\n"
		"        achieved by selecting the appropriate probes for the calculation. In\n"
		"        the default case that no stereoselectivity is required, only\n"
		"        non-stereospecific probes are used in the calculation. Stereospecific\n"
		"        differences between enantiomers are only captured in combination with a\n"
		"        very high accuracy, hence leading to very long computation times. The\n"
		"        following table provides the link between the parameter value and the\n"
		"        desired stereochemical treatment:\n"
		"\n"
		"        |**Stereo treatment**|**OBSpectrophore::StereoOption parameter**|\n"
		"        |Non-stereospecific probes (default)|NoStereoSpecificProbes|\n"
		"        |Only the mirror stereospecific probes|UniqueStereoSpecificProbes|\n"
		"        |Only the unique stereospecific probes|MirrorStereoSpecificProbes|\n"
		"        |All stereospecific probes (unique and mirror)|AllStereoSpecificProbes|\n"
		"\n"
		"\n"
		"        :type s: int, optional\n"
		"        :param s: The desired stereospecificity treatment expressed as an\n"
		"            OBSpectrophore::StereoOption enumeration type\n"
		""},
	 { "OBSpectrophore_SetNormalization", _wrap_OBSpectrophore_SetNormalization, METH_VARARGS, "\n"
		"Method to set the desired normalization treatment of the calculated\n"
		"        Spectrophores.\n"
		"\n"
		"        The following table provides the link between the parameter value\n"
		"        and the desired normalization treatment:\n"
		"\n"
		"        |**Normalization treatment**|**OBSpectrophore::NormalizationOption parameter**|\n"
		"        |No normalisation (default)|NoNormalization|\n"
		"        |Normalization towards zero mean|NormalizationTowardsZeroMean|\n"
		"        |Normalization towards unit standard deviation|NormalizationTowardsUnitStd|\n"
		"        |Normalization towards zero mean and unit standard deviation|NormalizationTowardsZeroMeanAndUnitStd|\n"
		"\n"
		"\n"
		"\n"
		"        :type n: int, optional\n"
		"        :param n: The desired normalization treatment expressed as an\n"
		"            OBSpectrophore::NormalizationOption enumeration type\n"
		""},
	 { "OBSpectrophore_GetAccuracy", _wrap_OBSpectrophore_GetAccuracy, METH_O, "\n"
		"Returns the accuracy at which Spectrophores will be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The accuracy expressed as an OBSpectrophore::AccuracyOption\n"
		"                     enumeration type. For a link between the returned value and\n"
		"                     the angular step size, see the\n"
		"                     OBSpectrophore::SetAccuracy(const OBSpectrophore::AccuracyOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetResolution", _wrap_OBSpectrophore_GetResolution, METH_O, "\n"
		"Returns the resolution at which Spectrophores will be calculated.\n"
		"\n"
		"        :rtype: float\n"
		"        :return: The resolution in &Aring; units\n"
		""},
	 { "OBSpectrophore_GetStereo", _wrap_OBSpectrophore_GetStereo, METH_O, "\n"
		"Returns the stereoselectivity setting at which Spectrophores\n"
		"        will be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The level of stereospecificity expressed as an\n"
		"                     OBSpectrophore::StereoOption enumeration type. For a link\n"
		"                     between the returned value and the stereoselectivity\n"
		"                     treatment, see the\n"
		"                     OBSpectrophore::SetStereo(const OBSpectrophore::StereoOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetNormalization", _wrap_OBSpectrophore_GetNormalization, METH_O, "\n"
		"Returns the normalization settings at which Spectrophores will\n"
		"        be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The normalization treatmet expressed as an\n"
		"                     OBSpectrophore::NormalizationOption enumeration type.\n"
		"                     For a link between the returned value and the normalization\n"
		"                     treatment, see the\n"
		"                     OBSpectrophore::SetNormalization(const OBSpectrophore::NormalizationOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetSpectrophore", _wrap_OBSpectrophore_GetSpectrophore, METH_VARARGS, "\n"
		" Calling this method starts the calculation of the Spectrophore.\n"
		"         After successful calculation, the Spectrophore is returned as a\n"
		"         standard vector of 48 doubles. The 48 doubles are organised into 4 sets\n"
		"         of 12 doubles each:-\n"
		"         - numbers 01-11: Spectrophore values calculated from the atomic partial charges;\n"
		"         - numbers 13-24: Spectrophore values calculated from the atomic lipophilicity properties;\n"
		"         - numbers 25-36: Spectrophore values calculated from the atomic shape deviations;\n"
		"         - numbers 37-48: Spectrophore values calculated from the atomic electrophilicity properties;\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OpenBabel::OBConversion obconversion;\n"
		"    obconversion.SetInFormat(\"sdf\");\n"
		"    OpenBabel::OBMol mol;\n"
		"    OpenBabel::OBSpectrophore s;\n"
		"    s.SetAccuracy(OpenBabel::OBSpectrophore::AngStepSize20);          // Default\n"
		"    s.SetResolution(3.0);                                             // Default\n"
		"    s.SetStereo(OpenBabel::OBSpectrophore::NoStereoSpecificProbes);   // Default\n"
		"    s.SetNormalization(OpenBabel::OBSpectrophore::NoNormalization);   // Default\n"
		"    std::ifstream ifs;\n"
		"    ifs.open(argv[1]);\n"
		"    while (obconversion.Read(&mol, &ifs))\n"
		"    {\n"
		"       std::vector<double> result = s.GetSpectrophore(&mol);\n"
		"       for (unsigned int i(0); i < result.size(); ++i)\n"
		"       {\n"
		"          if (!(i%12)) std::cerr << std::endl;\n"
		"          std::cerr << result[i] << \"  \";\n"
		"       }\n"
		"       std::cerr << std::endl;\n"
		"       mol.Clear();\n"
		"    }\n"
		"\n"
		"         :type mol: :py:class:`OBMol`\n"
		"         :param mol: Pointer to the OBMol object from which to calculate a\n"
		"                      Spectrophore. For proper functioning, the input molecule\n"
		"                      should have all explicit hydrogens assigned and the molecule\n"
		"                      should have a three-dimensional conformation assigned. It is\n"
		"                      the responsability of the programmer to make sure that the\n"
		"                      molecule is in the desired protonation state.\n"
		"\n"
		"         :rtype: std::vector< double,std::allocator< double > >\n"
		"         :return: The calculated Spectrophore as a standard vector of 48\n"
		"                      doubles. An empty vector of doubles is returned in case an\n"
		"                      error occurred during the calculation. For example, this could\n"
		"                      happen in the case that the molecule contains less than three\n"
		"                      atoms. It is therefore up to the user to check the size of the\n"
		"                      returned vector to capture errors.\n"
		""},
	 { "OBSpectrophore_swigregister", OBSpectrophore_swigregister, METH_O, NULL},
	 { "OBSpectrophore_swiginit", OBSpectrophore_swiginit, METH_VARARGS, NULL},
	 { "OBConformerFilter_IsGood", _wrap_OBConformerFilter_IsGood, METH_VARARGS, "\n"
		"Check if a conformer is good (i.e. passes the filter). The molecule, rotor\n"
		"key and coords are passed as parameters to this function and can be used\n"
		"for checking the confoormer.\n"
		":rtype: boolean\n"
		":return: True if the conformer passes the filter.\n"
		""},
	 { "delete_OBConformerFilter", _wrap_delete_OBConformerFilter, METH_O, NULL},
	 { "OBConformerFilter_swigregister", OBConformerFilter_swigregister, METH_O, NULL},
	 { "new_OBConformerFilters", _wrap_new_OBConformerFilters, METH_O, "Constructor specifiying the filters to combine."},
	 { "delete_OBConformerFilters", _wrap_delete_OBConformerFilters, METH_O, NULL},
	 { "OBConformerFilters_swigregister", OBConformerFilters_swigregister, METH_O, NULL},
	 { "OBConformerFilters_swiginit", OBConformerFilters_swiginit, METH_VARARGS, NULL},
	 { "new_OBStericConformerFilter", _wrap_new_OBStericConformerFilter, METH_VARARGS, NULL},
	 { "delete_OBStericConformerFilter", _wrap_delete_OBStericConformerFilter, METH_O, NULL},
	 { "OBStericConformerFilter_swigregister", OBStericConformerFilter_swigregister, METH_O, NULL},
	 { "OBStericConformerFilter_swiginit", OBStericConformerFilter_swiginit, METH_VARARGS, NULL},
	 { "OBConformerScore_GetPreferred", _wrap_OBConformerScore_GetPreferred, METH_O, "Preferred order for subclass scoring function."},
	 { "OBConformerScore_GetConvergence", _wrap_OBConformerScore_GetConvergence, METH_O, "Convergence criteria for subclass scoring function."},
	 { "OBConformerScore_Score", _wrap_OBConformerScore_Score, METH_VARARGS, "Score an individual conformer specified by index."},
	 { "delete_OBConformerScore", _wrap_delete_OBConformerScore, METH_O, NULL},
	 { "OBConformerScore_swigregister", OBConformerScore_swigregister, METH_O, NULL},
	 { "new_OBRMSDConformerScore", _wrap_new_OBRMSDConformerScore, METH_NOARGS, NULL},
	 { "delete_OBRMSDConformerScore", _wrap_delete_OBRMSDConformerScore, METH_O, NULL},
	 { "OBRMSDConformerScore_swigregister", OBRMSDConformerScore_swigregister, METH_O, NULL},
	 { "OBRMSDConformerScore_swiginit", OBRMSDConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBEnergyConformerScore", _wrap_new_OBEnergyConformerScore, METH_NOARGS, NULL},
	 { "OBEnergyConformerScore_GetNbEnergyCompute", _wrap_OBEnergyConformerScore_GetNbEnergyCompute, METH_O, NULL},
	 { "OBEnergyConformerScore_GetNbEnergyRequest", _wrap_OBEnergyConformerScore_GetNbEnergyRequest, METH_O, NULL},
	 { "delete_OBEnergyConformerScore", _wrap_delete_OBEnergyConformerScore, METH_O, NULL},
	 { "OBEnergyConformerScore_swigregister", OBEnergyConformerScore_swigregister, METH_O, NULL},
	 { "OBEnergyConformerScore_swiginit", OBEnergyConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBMinimizingEnergyConformerScore", _wrap_new_OBMinimizingEnergyConformerScore, METH_NOARGS, NULL},
	 { "OBMinimizingEnergyConformerScore_GetNbEnergyCompute", _wrap_OBMinimizingEnergyConformerScore_GetNbEnergyCompute, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_GetNbEnergyRequest", _wrap_OBMinimizingEnergyConformerScore_GetNbEnergyRequest, METH_O, NULL},
	 { "delete_OBMinimizingEnergyConformerScore", _wrap_delete_OBMinimizingEnergyConformerScore, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_swigregister", OBMinimizingEnergyConformerScore_swigregister, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_swiginit", OBMinimizingEnergyConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBMinimizingRMSDConformerScore", _wrap_new_OBMinimizingRMSDConformerScore, METH_NOARGS, NULL},
	 { "delete_OBMinimizingRMSDConformerScore", _wrap_delete_OBMinimizingRMSDConformerScore, METH_O, NULL},
	 { "OBMinimizingRMSDConformerScore_swigregister", OBMinimizingRMSDConformerScore_swigregister, METH_O, NULL},
	 { "OBMinimizingRMSDConformerScore_swiginit", OBMinimizingRMSDConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBConformerSearch", _wrap_new_OBConformerSearch, METH_NOARGS, NULL},
	 { "delete_OBConformerSearch", _wrap_delete_OBConformerSearch, METH_O, NULL},
	 { "OBConformerSearch_Setup", _wrap_OBConformerSearch_Setup, METH_VARARGS, "\n"
		"Setup this instance with the specified molecule.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule with initial conformer.\n"
		":type numConformers: int, optional\n"
		":param numConformers: The number of conformers that should be generated. This\n"
		"           is also the number of conformers selected for each generation.\n"
		":type numChildren: int, optional\n"
		":param numChildren: When a new generation is generated, for each of the\n"
		"           numConformers conformers, numChildren children are created.\n"
		":type mutability: int, optional\n"
		":param mutability: The mutability determines how frequent a permutation occurs\n"
		"           when generating the next generation.\n"
		":type convergence: int, optional\n"
		":param convergence: The number of identical generations before considering\n"
		"           the process converged.\n"
		""},
	 { "OBConformerSearch_SetNumConformers", _wrap_OBConformerSearch_SetNumConformers, METH_VARARGS, "Set the number of conformers."},
	 { "OBConformerSearch_SetNumChildren", _wrap_OBConformerSearch_SetNumChildren, METH_VARARGS, "Set the number of children generated for each parent."},
	 { "OBConformerSearch_SetMutability", _wrap_OBConformerSearch_SetMutability, METH_VARARGS, "Set the mutability."},
	 { "OBConformerSearch_SetConvergence", _wrap_OBConformerSearch_SetConvergence, METH_VARARGS, "\n"
		"Set the convergence (i.e. number of generations that did not change the score\n"
		"before considering the iteration converged).\n"
		""},
	 { "OBConformerSearch_SetFixedBonds", _wrap_OBConformerSearch_SetFixedBonds, METH_VARARGS, "Set the bonds to be fixed."},
	 { "OBConformerSearch_SetFilter", _wrap_OBConformerSearch_SetFilter, METH_VARARGS, "\n"
		"Set the filter method used to check if a newly generated is acceptable. Typical\n"
		"examples are a steric filter or electrostatic energy filter. The filters make a binary\n"
		"selection and one group will be scored in the next step. A number of filters\n"
		"are provided by default but implementing a custom filter is easy.\n"
		"\n"
		"There is also a OBConformerFilters class to make it easy to use multiple filters.\n"
		""},
	 { "OBConformerSearch_SetScore", _wrap_OBConformerSearch_SetScore, METH_VARARGS, "\n"
		"All acceptable conformers are scored to select the fittest conformers from\n"
		"the population. A higher score means the conformer is desired in the final\n"
		"set or next generation. Typical scoring would be force field energy to find\n"
		"the lowest energy conformer. Another example is the highest RMSD between\n"
		"conformer i and the closest conformer to create a diverse set of conformers.\n"
		"Diversity could also be expressed in other ways and implementing a scoring\n"
		"class is easy.\n"
		""},
	 { "OBConformerSearch_PrintRotors", _wrap_OBConformerSearch_PrintRotors, METH_VARARGS, "Set whether or not you want rotors to be printed prior to the conformer search."},
	 { "OBConformerSearch_Search", _wrap_OBConformerSearch_Search, METH_O, "Perform conformer search using a genetic algorithm."},
	 { "OBConformerSearch_GetRotorKeys", _wrap_OBConformerSearch_GetRotorKeys, METH_O, NULL},
	 { "OBConformerSearch_GetConformers", _wrap_OBConformerSearch_GetConformers, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetLogStream", _wrap_OBConformerSearch_SetLogStream, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetSharing", _wrap_OBConformerSearch_SetSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNbNiches", _wrap_OBConformerSearch_GetNbNiches, METH_O, NULL},
	 { "OBConformerSearch_SetNbNiches", _wrap_OBConformerSearch_SetNbNiches, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNicheRadius", _wrap_OBConformerSearch_GetNicheRadius, METH_O, NULL},
	 { "OBConformerSearch_SetNicheRadius", _wrap_OBConformerSearch_SetNicheRadius, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetAlphaSharing", _wrap_OBConformerSearch_GetAlphaSharing, METH_O, NULL},
	 { "OBConformerSearch_SetAlphaSharing", _wrap_OBConformerSearch_SetAlphaSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetSigmaSharing", _wrap_OBConformerSearch_GetSigmaSharing, METH_O, NULL},
	 { "OBConformerSearch_SetSigmaSharing", _wrap_OBConformerSearch_SetSigmaSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetCrossoverProbability", _wrap_OBConformerSearch_GetCrossoverProbability, METH_O, NULL},
	 { "OBConformerSearch_SetCrossoverProbability", _wrap_OBConformerSearch_SetCrossoverProbability, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNicheMating", _wrap_OBConformerSearch_GetNicheMating, METH_O, NULL},
	 { "OBConformerSearch_SetNicheMating", _wrap_OBConformerSearch_SetNicheMating, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetLocalOptRate", _wrap_OBConformerSearch_SetLocalOptRate, METH_VARARGS, NULL},
	 { "OBConformerSearch_swigregister", OBConformerSearch_swigregister, METH_O, NULL},
	 { "OBConformerSearch_swiginit", OBConformerSearch_swiginit, METH_VARARGS, NULL},
	 { "new_OBAlign", _wrap_new_OBAlign, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"If this constructor is used, the Target and Reference must be\n"
		"set using SetRef/SetRefMol and SetTarget/SetTargetMol before running\n"
		"the alignment.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		"\n"
		"Align two vectors of vector3 objects.\n"
		""},
	 { "OBAlign_SetRef", _wrap_OBAlign_SetRef, METH_VARARGS, "\n"
		"Set the Reference (to which the Target will be aligned) in\n"
		"terms of a vector of vector3 objects. Note that it is faster\n"
		"to perform multiple alignments to the same Reference, rather than\n"
		"multiple alignments to the same Target.\n"
		""},
	 { "OBAlign_SetTarget", _wrap_OBAlign_SetTarget, METH_VARARGS, "\n"
		"Set the Target (which will be aligned to the Reference) in\n"
		"terms of a vector of vector3 objects.\n"
		""},
	 { "OBAlign_SetRefMol", _wrap_OBAlign_SetRefMol, METH_VARARGS, "\n"
		"Set the Reference Molecule (to which the Target Molecule must\n"
		"be aligned). Note that is faster to perform multiple alignments\n"
		"to the same Reference Molecule, rather than multple alignments\n"
		"to the same Target Molecule.\n"
		""},
	 { "OBAlign_SetTargetMol", _wrap_OBAlign_SetTargetMol, METH_VARARGS, "\n"
		"Set the Target Molecule (which will be aligned to the\n"
		"Reference Molecule).\n"
		""},
	 { "OBAlign_Align", _wrap_OBAlign_Align, METH_O, "Align the Target to the Reference using a least-squares alignment."},
	 { "OBAlign_SetMethod", _wrap_OBAlign_SetMethod, METH_VARARGS, NULL},
	 { "OBAlign_GetRMSD", _wrap_OBAlign_GetRMSD, METH_O, "\n"
		"Return the root mean squared deviation of the target from\n"
		"the reference. This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "OBAlign_GetRotMatrix", _wrap_OBAlign_GetRotMatrix, METH_O, "\n"
		"Return the rotation matrix associated with the least squares\n"
		"alignment. This function should only\n"
		"be called after running the alignment (using Align()).\n"
		"\n"
		"The following example shows how to use the rotation matrix\n"
		"to rotate all of the atoms in a molecule.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    matrix3x3 rotmatrix = align.GetRotMatrix();\n"
		"    for (unsigned int i = 1; i <= mol.NumAtoms(); ++i) {\n"
		"       vector3 tmpvec = mol.GetAtom(i)->GetVector();\n"
		"       tmpvec *= rotmatrix; //apply the rotation\n"
		"       mol.GetAtom(i)->SetVector(tmpvec);\n"
		"    }\n"
		"Note that if you wish to use the rotation matrix to find the\n"
		"aligned coordinates (that is, the same coordinates returned by\n"
		"GetAlignment()), you should first translate the set of coordinates\n"
		"to the origin by subtracting the centroid, then apply the rotation,\n"
		"and finally add the centroid of the reference coordinates.\n"
		""},
	 { "OBAlign_GetAlignment", _wrap_OBAlign_GetAlignment, METH_O, "\n"
		"Return the actual alignment of the Target to the Reference\n"
		"in terms of a vector of vector3 objects. If you want an OBMol\n"
		"with the aligned coordinates, you should use UpdateCoords() instead.\n"
		"This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "OBAlign_UpdateCoords", _wrap_OBAlign_UpdateCoords, METH_VARARGS, "\n"
		"Set the coordinates of an OBMol to those from the alignment.\n"
		"This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "delete_OBAlign", _wrap_delete_OBAlign, METH_O, NULL},
	 { "OBAlign_swigregister", OBAlign_swigregister, METH_O, NULL},
	 { "OBAlign_swiginit", OBAlign_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomIter", _wrap_new__OBMolAtomIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomIter", _wrap_delete__OBMolAtomIter, METH_O, NULL},
	 { "_OBMolAtomIter___nonzero__", _wrap__OBMolAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomIter_inc", _wrap__OBMolAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement iterator -- advance to next atom and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement iterator -- return the current atom and advance\n"
		""},
	 { "_OBMolAtomIter___ref__", _wrap__OBMolAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomIter_swigregister", _OBMolAtomIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomIter_swiginit", _OBMolAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomDFSIter", _wrap_new__OBMolAtomDFSIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomDFSIter", _wrap_delete__OBMolAtomDFSIter, METH_O, NULL},
	 { "_OBMolAtomDFSIter___nonzero__", _wrap__OBMolAtomDFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomDFSIter_inc", _wrap__OBMolAtomDFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the DFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "_OBMolAtomDFSIter___ref__", _wrap__OBMolAtomDFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomDFSIter_next", _wrap__OBMolAtomDFSIter_next, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: NULL if at the last atom in a fragment, else the next atom\n"
		""},
	 { "_OBMolAtomDFSIter_swigregister", _OBMolAtomDFSIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomDFSIter_swiginit", _OBMolAtomDFSIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomBFSIter", _wrap_new__OBMolAtomBFSIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomBFSIter", _wrap_delete__OBMolAtomBFSIter, METH_O, NULL},
	 { "_OBMolAtomBFSIter___nonzero__", _wrap__OBMolAtomBFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomBFSIter_inc", _wrap__OBMolAtomBFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the BFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "_OBMolAtomBFSIter___ref__", _wrap__OBMolAtomBFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomBFSIter_CurrentDepth", _wrap__OBMolAtomBFSIter_CurrentDepth, METH_O, "\n"
		":rtype: int\n"
		":return: the current depth of the iterator Since: version 2.2\n"
		""},
	 { "_OBMolAtomBFSIter_swigregister", _OBMolAtomBFSIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomBFSIter_swiginit", _OBMolAtomBFSIter_swiginit, METH_VARARGS, NULL},
	 { "new_OBMolBondBFSIter", _wrap_new_OBMolBondBFSIter, METH_VARARGS, NULL},
	 { "delete_OBMolBondBFSIter", _wrap_delete_OBMolBondBFSIter, METH_O, NULL},
	 { "OBMolBondBFSIter___nonzero__", _wrap_OBMolBondBFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "OBMolBondBFSIter_inc", _wrap_OBMolBondBFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the BFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "OBMolBondBFSIter___ref__", _wrap_OBMolBondBFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current atom\n"
		""},
	 { "OBMolBondBFSIter_CurrentDepth", _wrap_OBMolBondBFSIter_CurrentDepth, METH_O, "\n"
		":rtype: int\n"
		":return: the current depth of the iterator Since: version 2.2\n"
		""},
	 { "OBMolBondBFSIter_swigregister", OBMolBondBFSIter_swigregister, METH_O, NULL},
	 { "OBMolBondBFSIter_swiginit", OBMolBondBFSIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolBondIter", _wrap_new__OBMolBondIter, METH_VARARGS, NULL},
	 { "delete__OBMolBondIter", _wrap_delete__OBMolBondIter, METH_O, NULL},
	 { "_OBMolBondIter___nonzero__", _wrap__OBMolBondIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more bonds)\n"
		""},
	 { "_OBMolBondIter_inc", _wrap__OBMolBondIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next bond and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current bond and advance to the next\n"
		""},
	 { "_OBMolBondIter___ref__", _wrap__OBMolBondIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current bond\n"
		""},
	 { "_OBMolBondIter_swigregister", _OBMolBondIter_swigregister, METH_O, NULL},
	 { "_OBMolBondIter_swiginit", _OBMolBondIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBAtomAtomIter", _wrap_new__OBAtomAtomIter, METH_VARARGS, NULL},
	 { "delete__OBAtomAtomIter", _wrap_delete__OBAtomAtomIter, METH_O, NULL},
	 { "_OBAtomAtomIter___nonzero__", _wrap__OBAtomAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more neighbors)\n"
		""},
	 { "_OBAtomAtomIter_inc", _wrap__OBAtomAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next neighbor and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current neighbor and advance to the next\n"
		""},
	 { "_OBAtomAtomIter___ref__", _wrap__OBAtomAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBAtomAtomIter_swigregister", _OBAtomAtomIter_swigregister, METH_O, NULL},
	 { "_OBAtomAtomIter_swiginit", _OBAtomAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBAtomBondIter", _wrap_new__OBAtomBondIter, METH_VARARGS, NULL},
	 { "delete__OBAtomBondIter", _wrap_delete__OBAtomBondIter, METH_O, NULL},
	 { "_OBAtomBondIter___nonzero__", _wrap__OBAtomBondIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more bonds)\n"
		""},
	 { "_OBAtomBondIter_inc", _wrap__OBAtomBondIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next bond and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next bond\n"
		""},
	 { "_OBAtomBondIter___ref__", _wrap__OBAtomBondIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current bond\n"
		""},
	 { "_OBAtomBondIter_swigregister", _OBAtomBondIter_swigregister, METH_O, NULL},
	 { "_OBAtomBondIter_swiginit", _OBAtomBondIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBResidueIter", _wrap_new__OBResidueIter, METH_VARARGS, NULL},
	 { "delete__OBResidueIter", _wrap_delete__OBResidueIter, METH_O, NULL},
	 { "_OBResidueIter___nonzero__", _wrap__OBResidueIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more residues)\n"
		""},
	 { "_OBResidueIter_inc", _wrap__OBResidueIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next residue and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next residue\n"
		""},
	 { "_OBResidueIter___ref__", _wrap__OBResidueIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: a reference to the current residue\n"
		""},
	 { "_OBResidueIter_swigregister", _OBResidueIter_swigregister, METH_O, NULL},
	 { "_OBResidueIter_swiginit", _OBResidueIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBResidueAtomIter", _wrap_new__OBResidueAtomIter, METH_VARARGS, NULL},
	 { "delete__OBResidueAtomIter", _wrap_delete__OBResidueAtomIter, METH_O, NULL},
	 { "_OBResidueAtomIter___nonzero__", _wrap__OBResidueAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms in this residue)\n"
		""},
	 { "_OBResidueAtomIter_inc", _wrap__OBResidueAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom (if any) and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next atom (if any)\n"
		""},
	 { "_OBResidueAtomIter___ref__", _wrap__OBResidueAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBResidueAtomIter_swigregister", _OBResidueAtomIter_swigregister, METH_O, NULL},
	 { "_OBResidueAtomIter_swiginit", _OBResidueAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAngleIter", _wrap_new__OBMolAngleIter, METH_VARARGS, NULL},
	 { "delete__OBMolAngleIter", _wrap_delete__OBMolAngleIter, METH_O, NULL},
	 { "_OBMolAngleIter___nonzero__", _wrap__OBMolAngleIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more angles)\n"
		""},
	 { "_OBMolAngleIter_inc", _wrap__OBMolAngleIter_inc, METH_O, "Preincrement -- advance to the next angle and return"},
	 { "_OBMolAngleIter___ref__", _wrap__OBMolAngleIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of three atom indexes specifying the angle See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolAngleIter_swigregister", _OBMolAngleIter_swigregister, METH_O, NULL},
	 { "_OBMolAngleIter_swiginit", _OBMolAngleIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolTorsionIter", _wrap_new__OBMolTorsionIter, METH_VARARGS, NULL},
	 { "delete__OBMolTorsionIter", _wrap_delete__OBMolTorsionIter, METH_O, NULL},
	 { "_OBMolTorsionIter___nonzero__", _wrap__OBMolTorsionIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more torsions)\n"
		""},
	 { "_OBMolTorsionIter_inc", _wrap__OBMolTorsionIter_inc, METH_O, "Preincrement -- advance to the next torsion and return"},
	 { "_OBMolTorsionIter___ref__", _wrap__OBMolTorsionIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of four atom indexes specifying the torsion See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolTorsionIter_swigregister", _OBMolTorsionIter_swigregister, METH_O, NULL},
	 { "_OBMolTorsionIter_swiginit", _OBMolTorsionIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolPairIter", _wrap_new__OBMolPairIter, METH_VARARGS, NULL},
	 { "delete__OBMolPairIter", _wrap_delete__OBMolPairIter, METH_O, NULL},
	 { "_OBMolPairIter___nonzero__", _wrap__OBMolPairIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more 1-4 atom pairs)\n"
		""},
	 { "_OBMolPairIter_inc", _wrap__OBMolPairIter_inc, METH_O, "Preincrement -- advance to the next 1-4 atom pair and return"},
	 { "_OBMolPairIter___ref__", _wrap__OBMolPairIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of two atom indexes specifying the 1-4 atom pair See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolPairIter_swigregister", _OBMolPairIter_swigregister, METH_O, NULL},
	 { "_OBMolPairIter_swiginit", _OBMolPairIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolRingIter", _wrap_new__OBMolRingIter, METH_VARARGS, NULL},
	 { "delete__OBMolRingIter", _wrap_delete__OBMolRingIter, METH_O, NULL},
	 { "_OBMolRingIter___nonzero__", _wrap__OBMolRingIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can advance (i.e., there are more rings)\n"
		""},
	 { "_OBMolRingIter_inc", _wrap__OBMolRingIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next ring (if any) and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next ring\n"
		""},
	 { "_OBMolRingIter___ref__", _wrap__OBMolRingIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBRing`\n"
		":return: A reference to the current ring (if any)\n"
		""},
	 { "_OBMolRingIter_swigregister", _OBMolRingIter_swigregister, METH_O, NULL},
	 { "_OBMolRingIter_swiginit", _OBMolRingIter_swiginit, METH_VARARGS, NULL},
	 { "MolGetAtoms", _wrap_MolGetAtoms, METH_VARARGS, NULL},
	 { "MolGetBonds", _wrap_MolGetBonds, METH_VARARGS, NULL},
	 { "AtomGetBonds", _wrap_AtomGetBonds, METH_VARARGS, NULL},
	 { "AtomGetNbrs", _wrap_AtomGetNbrs, METH_VARARGS, NULL},
	 { "new_doubleArray", _wrap_new_doubleArray, METH_O, NULL},
	 { "delete_doubleArray", _wrap_delete_doubleArray, METH_O, NULL},
	 { "doubleArray___getitem__", _wrap_doubleArray___getitem__, METH_VARARGS, NULL},
	 { "doubleArray___setitem__", _wrap_doubleArray___setitem__, METH_VARARGS, NULL},
	 { "doubleArray_cast", _wrap_doubleArray_cast, METH_O, NULL},
	 { "doubleArray_frompointer", _wrap_doubleArray_frompointer, METH_O, NULL},
	 { "doubleArray_swigregister", doubleArray_swigregister, METH_O, NULL},
	 { "doubleArray_swiginit", doubleArray_swiginit, METH_VARARGS, NULL},
	 { "delete_OBTetraNonPlanarStereo", _wrap_delete_OBTetraNonPlanarStereo, METH_O, "Destructor."},
	 { "OBTetraNonPlanarStereo_swigregister", OBTetraNonPlanarStereo_swigregister, METH_O, NULL},
	 { "delete_OBTetraPlanarStereo", _wrap_delete_OBTetraPlanarStereo, METH_O, NULL},
	 { "OBTetraPlanarStereo_swigregister", OBTetraPlanarStereo_swigregister, METH_O, NULL},
	 { "new_OBTetrahedralStereo", _wrap_new_OBTetrahedralStereo, METH_O, "Constructor."},
	 { "delete_OBTetrahedralStereo", _wrap_delete_OBTetrahedralStereo, METH_O, "Destructor."},
	 { "OBTetrahedralStereo_IsValid", _wrap_OBTetrahedralStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all (center, from\n"
		"    and ref) atom ids are set.\n"
		""},
	 { "OBTetrahedralStereo___eq__", _wrap_OBTetrahedralStereo___eq__, METH_VARARGS, "\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBTetrahedralStereo___ne__", _wrap_OBTetrahedralStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBTetrahedralStereo_SetConfig", _wrap_OBTetrahedralStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBTetrahedralStereo_GetConfig", _wrap_OBTetrahedralStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBTetrahedralStereo_swigregister", OBTetrahedralStereo_swigregister, METH_O, NULL},
	 { "OBTetrahedralStereo_swiginit", OBTetrahedralStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBCisTransStereo", _wrap_new_OBCisTransStereo, METH_O, "\n"
		"<openbabel/stereo/cistrans.h>\n"
		"Stereochemical configuration for double-bond cis/trans stereochemistry\n"
		"\n"
		"The config struct represents the stereochemistry in a well defined way.\n"
		"For cis/trans stereo bonds, the following data members define the spacial\n"
		"arrengement of the atoms.\n"
		"\n"
		"- OBStereo::Ref ``begin``: The begin atom for the double bond.\n"
		"- OBStereo::Ref ``end``: The end atom for the double bond.\n"
		"- OBStereo::Refs ``refs``: The 4 atoms connected to the double bond.\n"
		"- OBStereo::Shape ``shape``: The shape formed by the ``refs`` by connecting them\n"
		"  in the same order as they occur in ``refs``.\n"
		"\n"
		"Image: cistrans.png\n"
		"Image: SPshapes.png\n"
		"\n"
		"Only ``begin`` and ``end`` are specific for OBCisTransStereo::Config. The other\n"
		"data members occur in all OBTetraPlanarStereo derived classes.\n"
		"\n"
		"Constructor.\n"
		""},
	 { "delete_OBCisTransStereo", _wrap_delete_OBCisTransStereo, METH_O, "Destructor."},
	 { "OBCisTransStereo_IsValid", _wrap_OBCisTransStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all these\n"
		"    conditions are met:\n"
		"    - ``begin`` != OBStereo::NoRef\n"
		"    - ``end`` != OBStereo::NoRef\n"
		"    - ``refs`` contains 4 elements\n"
		""},
	 { "OBCisTransStereo___eq__", _wrap_OBCisTransStereo___eq__, METH_VARARGS, "\n"
		"Set the configuration using a Config struct.\n"
		"\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBCisTransStereo___ne__", _wrap_OBCisTransStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBCisTransStereo_IsOnSameAtom", _wrap_OBCisTransStereo_IsOnSameAtom, METH_VARARGS, "\n"
		"Check if the two atoms for ``id1`` & ``id2`` are bonded to the same\n"
		"atom. If the atoms for one of the atoms doesn't exist (anymore),\n"
		"the valence of the begin and end atom is checked. If the exising\n"
		"atom is bonded to the begin atom and end->GetExplicitDegree() == 2, the\n"
		"ids are considered to be on different atoms. The reasoning behind\n"
		"this is that hydrogens may be deleted. However, you can also use\n"
		"OBStereo::ImplicitRef explicitly in code like:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    //\n"
		"    //  F       F      F      F     0      3\n"
		"    //   \\     /       |      |     |      |\n"
		"    //    C===C        | C  C |     | 1  2 |\n"
		"    //   /     \\       |      |     |      |\n"
		"    // (H)     (H)    (H)----(H)    H------H\n"
		"    //\n"
		"    reading smiles F/C=C\\F cis-difluorethene\n"
		"    OBCisTransStereo ct(mol);\n"
		"    ct.SetCenters(1, 2);\n"
		"    ct.SetRefs(OBStereo::MakeRefs(0, OBStereo::ImplicitRef, OBStereo::ImplicitRef, 3));\n"
		"    ...\n"
		"\n"
		":rtype: boolean\n"
		":return: True if ``id1`` and ``id2`` are bonded to the same atom\n"
		"    taking implicit hydrogens into account.\n"
		""},
	 { "OBCisTransStereo_IsTrans", _wrap_OBCisTransStereo_IsTrans, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed trans configuration.\n"
		""},
	 { "OBCisTransStereo_IsCis", _wrap_OBCisTransStereo_IsCis, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed in a cis configuration.\n"
		""},
	 { "OBCisTransStereo_GetTransRef", _wrap_OBCisTransStereo_GetTransRef, METH_VARARGS, "\n"
		"Image: gettransref.png\n"
		"Get the reference id trans from reference ``id``.\n"
		""},
	 { "OBCisTransStereo_GetCisRef", _wrap_OBCisTransStereo_GetCisRef, METH_VARARGS, "\n"
		"Image: getcisref.png\n"
		"Get the reference id cis from reference ``id``.\n"
		""},
	 { "OBCisTransStereo_SetConfig", _wrap_OBCisTransStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBCisTransStereo_GetConfig", _wrap_OBCisTransStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBCisTransStereo_swigregister", OBCisTransStereo_swigregister, METH_O, NULL},
	 { "OBCisTransStereo_swiginit", OBCisTransStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBSquarePlanarStereo", _wrap_new_OBSquarePlanarStereo, METH_O, "\n"
		"<openbabel/stereo/squareplanar.h>\n"
		"Stereochemical configuration for square planar stereocenters\n"
		"\n"
		"The config struct represents the stereochemistry in a well defined way.\n"
		"For squareplanar stereocenters, the following data members define the spacial\n"
		"arrengement of the atoms.\n"
		"\n"
		"- OBStereo::Ref ``center``: The central atom.\n"
		"- OBStereo::Refs ``refs``: The 4 atoms connected to the double bond.\n"
		"- OBStereo::Shape ``shape``: The shape formed by the ``refs`` by connecting them\n"
		"  in the same order as they occur in ``refs``.\n"
		"\n"
		"Image: squareplanar.png\n"
		"Image: SPshapes.png\n"
		"\n"
		"Only ``center`` are specific for OBSquarePlanarStereo::Config. The other\n"
		"data members occur in all OBTetraPlanarStereo derived classes.\n"
		"\n"
		"Constructor.\n"
		""},
	 { "delete_OBSquarePlanarStereo", _wrap_delete_OBSquarePlanarStereo, METH_O, "Destructor."},
	 { "OBSquarePlanarStereo_IsValid", _wrap_OBSquarePlanarStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all (center and\n"
		"    and 4 reference) atom ids are set.\n"
		""},
	 { "OBSquarePlanarStereo___eq__", _wrap_OBSquarePlanarStereo___eq__, METH_VARARGS, "\n"
		"Set the configuration using a Config struct.\n"
		"\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBSquarePlanarStereo___ne__", _wrap_OBSquarePlanarStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBSquarePlanarStereo_IsTrans", _wrap_OBSquarePlanarStereo_IsTrans, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed trans configuration.\n"
		""},
	 { "OBSquarePlanarStereo_IsCis", _wrap_OBSquarePlanarStereo_IsCis, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed in a cis configuration.\n"
		""},
	 { "OBSquarePlanarStereo_GetTransRef", _wrap_OBSquarePlanarStereo_GetTransRef, METH_VARARGS, "\n"
		"Image: gettransref.png\n"
		"Get the reference id trans from reference ``id``.\n"
		""},
	 { "OBSquarePlanarStereo_GetCisRefs", _wrap_OBSquarePlanarStereo_GetCisRefs, METH_VARARGS, "Get the reference id cis from reference ``id``."},
	 { "OBSquarePlanarStereo_SetConfig", _wrap_OBSquarePlanarStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBSquarePlanarStereo_GetConfig", _wrap_OBSquarePlanarStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBSquarePlanarStereo_swigregister", OBSquarePlanarStereo_swigregister, METH_O, NULL},
	 { "OBSquarePlanarStereo_swiginit", OBSquarePlanarStereo_swiginit, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig___eq__", _wrap_OBTetrahedralConfig___eq__, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig___ne__", _wrap_OBTetrahedralConfig___ne__, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_center_set", _wrap_OBTetrahedralConfig_center_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_center_get", _wrap_OBTetrahedralConfig_center_get, METH_O, NULL},
	 { "OBTetrahedralConfig_from_or_towards_set", _wrap_OBTetrahedralConfig_from_or_towards_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_from_or_towards_get", _wrap_OBTetrahedralConfig_from_or_towards_get, METH_O, NULL},
	 { "OBTetrahedralConfig_refs_set", _wrap_OBTetrahedralConfig_refs_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_refs_get", _wrap_OBTetrahedralConfig_refs_get, METH_O, NULL},
	 { "OBTetrahedralConfig_winding_set", _wrap_OBTetrahedralConfig_winding_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_winding_get", _wrap_OBTetrahedralConfig_winding_get, METH_O, NULL},
	 { "OBTetrahedralConfig_view_set", _wrap_OBTetrahedralConfig_view_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_view_get", _wrap_OBTetrahedralConfig_view_get, METH_O, NULL},
	 { "OBTetrahedralConfig_specified_set", _wrap_OBTetrahedralConfig_specified_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_specified_get", _wrap_OBTetrahedralConfig_specified_get, METH_O, NULL},
	 { "new_OBTetrahedralConfig", _wrap_new_OBTetrahedralConfig, METH_NOARGS, NULL},
	 { "delete_OBTetrahedralConfig", _wrap_delete_OBTetrahedralConfig, METH_O, NULL},
	 { "OBTetrahedralConfig_swigregister", OBTetrahedralConfig_swigregister, METH_O, NULL},
	 { "OBTetrahedralConfig_swiginit", OBTetrahedralConfig_swiginit, METH_VARARGS, NULL},
	 { "OBCisTransConfig___eq__", _wrap_OBCisTransConfig___eq__, METH_VARARGS, NULL},
	 { "OBCisTransConfig___ne__", _wrap_OBCisTransConfig___ne__, METH_VARARGS, NULL},
	 { "OBCisTransConfig_begin_set", _wrap_OBCisTransConfig_begin_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_begin_get", _wrap_OBCisTransConfig_begin_get, METH_O, NULL},
	 { "OBCisTransConfig_end_set", _wrap_OBCisTransConfig_end_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_end_get", _wrap_OBCisTransConfig_end_get, METH_O, NULL},
	 { "OBCisTransConfig_refs_set", _wrap_OBCisTransConfig_refs_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_refs_get", _wrap_OBCisTransConfig_refs_get, METH_O, NULL},
	 { "OBCisTransConfig_shape_set", _wrap_OBCisTransConfig_shape_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_shape_get", _wrap_OBCisTransConfig_shape_get, METH_O, NULL},
	 { "OBCisTransConfig_specified_set", _wrap_OBCisTransConfig_specified_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_specified_get", _wrap_OBCisTransConfig_specified_get, METH_O, NULL},
	 { "new_OBCisTransConfig", _wrap_new_OBCisTransConfig, METH_NOARGS, NULL},
	 { "delete_OBCisTransConfig", _wrap_delete_OBCisTransConfig, METH_O, NULL},
	 { "OBCisTransConfig_swigregister", OBCisTransConfig_swigregister, METH_O, NULL},
	 { "OBCisTransConfig_swiginit", OBCisTransConfig_swiginit, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig___eq__", _wrap_OBSquarePlanarConfig___eq__, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig___ne__", _wrap_OBSquarePlanarConfig___ne__, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_center_set", _wrap_OBSquarePlanarConfig_center_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_center_get", _wrap_OBSquarePlanarConfig_center_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_refs_set", _wrap_OBSquarePlanarConfig_refs_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_refs_get", _wrap_OBSquarePlanarConfig_refs_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_shape_set", _wrap_OBSquarePlanarConfig_shape_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_shape_get", _wrap_OBSquarePlanarConfig_shape_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_specified_set", _wrap_OBSquarePlanarConfig_specified_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_specified_get", _wrap_OBSquarePlanarConfig_specified_get, METH_O, NULL},
	 { "new_OBSquarePlanarConfig", _wrap_new_OBSquarePlanarConfig, METH_NOARGS, NULL},
	 { "delete_OBSquarePlanarConfig", _wrap_delete_OBSquarePlanarConfig, METH_O, NULL},
	 { "OBSquarePlanarConfig_swigregister", OBSquarePlanarConfig_swigregister, METH_O, NULL},
	 { "OBSquarePlanarConfig_swiginit", OBSquarePlanarConfig_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};

static PyMethodDef SwigMethods_proxydocs[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "SWIG_PyStaticMethod_New", SWIG_PyStaticMethod_New, METH_O, NULL},
	 { "delete_SwigPyIterator", _wrap_delete_SwigPyIterator, METH_O, NULL},
	 { "SwigPyIterator_value", _wrap_SwigPyIterator_value, METH_O, NULL},
	 { "SwigPyIterator_incr", _wrap_SwigPyIterator_incr, METH_VARARGS, NULL},
	 { "SwigPyIterator_decr", _wrap_SwigPyIterator_decr, METH_VARARGS, NULL},
	 { "SwigPyIterator_distance", _wrap_SwigPyIterator_distance, METH_VARARGS, NULL},
	 { "SwigPyIterator_equal", _wrap_SwigPyIterator_equal, METH_VARARGS, NULL},
	 { "SwigPyIterator_copy", _wrap_SwigPyIterator_copy, METH_O, NULL},
	 { "SwigPyIterator_next", _wrap_SwigPyIterator_next, METH_O, NULL},
	 { "SwigPyIterator___next__", _wrap_SwigPyIterator___next__, METH_O, NULL},
	 { "SwigPyIterator_previous", _wrap_SwigPyIterator_previous, METH_O, NULL},
	 { "SwigPyIterator_advance", _wrap_SwigPyIterator_advance, METH_VARARGS, NULL},
	 { "SwigPyIterator___eq__", _wrap_SwigPyIterator___eq__, METH_VARARGS, NULL},
	 { "SwigPyIterator___ne__", _wrap_SwigPyIterator___ne__, METH_VARARGS, NULL},
	 { "SwigPyIterator___iadd__", _wrap_SwigPyIterator___iadd__, METH_VARARGS, NULL},
	 { "SwigPyIterator___isub__", _wrap_SwigPyIterator___isub__, METH_VARARGS, NULL},
	 { "SwigPyIterator___add__", _wrap_SwigPyIterator___add__, METH_VARARGS, NULL},
	 { "SwigPyIterator___sub__", _wrap_SwigPyIterator___sub__, METH_VARARGS, NULL},
	 { "SwigPyIterator_swigregister", SwigPyIterator_swigregister, METH_O, NULL},
	 { "vectorInt_iterator", _wrap_vectorInt_iterator, METH_O, NULL},
	 { "vectorInt___nonzero__", _wrap_vectorInt___nonzero__, METH_O, NULL},
	 { "vectorInt___bool__", _wrap_vectorInt___bool__, METH_O, NULL},
	 { "vectorInt___len__", _wrap_vectorInt___len__, METH_O, NULL},
	 { "vectorInt___getslice__", _wrap_vectorInt___getslice__, METH_VARARGS, NULL},
	 { "vectorInt___setslice__", _wrap_vectorInt___setslice__, METH_VARARGS, NULL},
	 { "vectorInt___delslice__", _wrap_vectorInt___delslice__, METH_VARARGS, NULL},
	 { "vectorInt___delitem__", _wrap_vectorInt___delitem__, METH_VARARGS, NULL},
	 { "vectorInt___getitem__", _wrap_vectorInt___getitem__, METH_VARARGS, NULL},
	 { "vectorInt___setitem__", _wrap_vectorInt___setitem__, METH_VARARGS, NULL},
	 { "vectorInt_clear", _wrap_vectorInt_clear, METH_O, NULL},
	 { "new_vectorInt", _wrap_new_vectorInt, METH_VARARGS, NULL},
	 { "delete_vectorInt", _wrap_delete_vectorInt, METH_O, NULL},
	 { "vectorInt_swigregister", vectorInt_swigregister, METH_O, NULL},
	 { "vectorInt_swiginit", vectorInt_swiginit, METH_VARARGS, NULL},
	 { "vectorUnsignedInt_iterator", _wrap_vectorUnsignedInt_iterator, METH_O, NULL},
	 { "vectorUnsignedInt___nonzero__", _wrap_vectorUnsignedInt___nonzero__, METH_O, NULL},
	 { "vectorUnsignedInt___bool__", _wrap_vectorUnsignedInt___bool__, METH_O, NULL},
	 { "vectorUnsignedInt___len__", _wrap_vectorUnsignedInt___len__, METH_O, NULL},
	 { "vectorUnsignedInt___getslice__", _wrap_vectorUnsignedInt___getslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___setslice__", _wrap_vectorUnsignedInt___setslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___delslice__", _wrap_vectorUnsignedInt___delslice__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___delitem__", _wrap_vectorUnsignedInt___delitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___getitem__", _wrap_vectorUnsignedInt___getitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt___setitem__", _wrap_vectorUnsignedInt___setitem__, METH_VARARGS, NULL},
	 { "vectorUnsignedInt_clear", _wrap_vectorUnsignedInt_clear, METH_O, NULL},
	 { "new_vectorUnsignedInt", _wrap_new_vectorUnsignedInt, METH_VARARGS, NULL},
	 { "delete_vectorUnsignedInt", _wrap_delete_vectorUnsignedInt, METH_O, NULL},
	 { "vectorUnsignedInt_swigregister", vectorUnsignedInt_swigregister, METH_O, NULL},
	 { "vectorUnsignedInt_swiginit", vectorUnsignedInt_swiginit, METH_VARARGS, NULL},
	 { "vectorvInt_iterator", _wrap_vectorvInt_iterator, METH_O, NULL},
	 { "vectorvInt___nonzero__", _wrap_vectorvInt___nonzero__, METH_O, NULL},
	 { "vectorvInt___bool__", _wrap_vectorvInt___bool__, METH_O, NULL},
	 { "vectorvInt___len__", _wrap_vectorvInt___len__, METH_O, NULL},
	 { "vectorvInt___getslice__", _wrap_vectorvInt___getslice__, METH_VARARGS, NULL},
	 { "vectorvInt___setslice__", _wrap_vectorvInt___setslice__, METH_VARARGS, NULL},
	 { "vectorvInt___delslice__", _wrap_vectorvInt___delslice__, METH_VARARGS, NULL},
	 { "vectorvInt___delitem__", _wrap_vectorvInt___delitem__, METH_VARARGS, NULL},
	 { "vectorvInt___getitem__", _wrap_vectorvInt___getitem__, METH_VARARGS, NULL},
	 { "vectorvInt___setitem__", _wrap_vectorvInt___setitem__, METH_VARARGS, NULL},
	 { "vectorvInt_clear", _wrap_vectorvInt_clear, METH_O, NULL},
	 { "new_vectorvInt", _wrap_new_vectorvInt, METH_VARARGS, NULL},
	 { "delete_vectorvInt", _wrap_delete_vectorvInt, METH_O, NULL},
	 { "vectorvInt_swigregister", vectorvInt_swigregister, METH_O, NULL},
	 { "vectorvInt_swiginit", vectorvInt_swiginit, METH_VARARGS, NULL},
	 { "vectorDouble_iterator", _wrap_vectorDouble_iterator, METH_O, NULL},
	 { "vectorDouble___nonzero__", _wrap_vectorDouble___nonzero__, METH_O, NULL},
	 { "vectorDouble___bool__", _wrap_vectorDouble___bool__, METH_O, NULL},
	 { "vectorDouble___len__", _wrap_vectorDouble___len__, METH_O, NULL},
	 { "vectorDouble___getslice__", _wrap_vectorDouble___getslice__, METH_VARARGS, NULL},
	 { "vectorDouble___setslice__", _wrap_vectorDouble___setslice__, METH_VARARGS, NULL},
	 { "vectorDouble___delslice__", _wrap_vectorDouble___delslice__, METH_VARARGS, NULL},
	 { "vectorDouble___delitem__", _wrap_vectorDouble___delitem__, METH_VARARGS, NULL},
	 { "vectorDouble___getitem__", _wrap_vectorDouble___getitem__, METH_VARARGS, NULL},
	 { "vectorDouble___setitem__", _wrap_vectorDouble___setitem__, METH_VARARGS, NULL},
	 { "vectorDouble_clear", _wrap_vectorDouble_clear, METH_O, NULL},
	 { "new_vectorDouble", _wrap_new_vectorDouble, METH_VARARGS, NULL},
	 { "delete_vectorDouble", _wrap_delete_vectorDouble, METH_O, NULL},
	 { "vectorDouble_swigregister", vectorDouble_swigregister, METH_O, NULL},
	 { "vectorDouble_swiginit", vectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorULong_iterator", _wrap_vectorULong_iterator, METH_O, NULL},
	 { "vectorULong___nonzero__", _wrap_vectorULong___nonzero__, METH_O, NULL},
	 { "vectorULong___bool__", _wrap_vectorULong___bool__, METH_O, NULL},
	 { "vectorULong___len__", _wrap_vectorULong___len__, METH_O, NULL},
	 { "vectorULong___getslice__", _wrap_vectorULong___getslice__, METH_VARARGS, NULL},
	 { "vectorULong___setslice__", _wrap_vectorULong___setslice__, METH_VARARGS, NULL},
	 { "vectorULong___delslice__", _wrap_vectorULong___delslice__, METH_VARARGS, NULL},
	 { "vectorULong___delitem__", _wrap_vectorULong___delitem__, METH_VARARGS, NULL},
	 { "vectorULong___getitem__", _wrap_vectorULong___getitem__, METH_VARARGS, NULL},
	 { "vectorULong___setitem__", _wrap_vectorULong___setitem__, METH_VARARGS, NULL},
	 { "vectorULong_clear", _wrap_vectorULong_clear, METH_O, NULL},
	 { "new_vectorULong", _wrap_new_vectorULong, METH_VARARGS, NULL},
	 { "delete_vectorULong", _wrap_delete_vectorULong, METH_O, NULL},
	 { "vectorULong_swigregister", vectorULong_swigregister, METH_O, NULL},
	 { "vectorULong_swiginit", vectorULong_swiginit, METH_VARARGS, NULL},
	 { "vectorString_iterator", _wrap_vectorString_iterator, METH_O, NULL},
	 { "vectorString___nonzero__", _wrap_vectorString___nonzero__, METH_O, NULL},
	 { "vectorString___bool__", _wrap_vectorString___bool__, METH_O, NULL},
	 { "vectorString___len__", _wrap_vectorString___len__, METH_O, NULL},
	 { "vectorString___getslice__", _wrap_vectorString___getslice__, METH_VARARGS, NULL},
	 { "vectorString___setslice__", _wrap_vectorString___setslice__, METH_VARARGS, NULL},
	 { "vectorString___delslice__", _wrap_vectorString___delslice__, METH_VARARGS, NULL},
	 { "vectorString___delitem__", _wrap_vectorString___delitem__, METH_VARARGS, NULL},
	 { "vectorString___getitem__", _wrap_vectorString___getitem__, METH_VARARGS, NULL},
	 { "vectorString___setitem__", _wrap_vectorString___setitem__, METH_VARARGS, NULL},
	 { "vectorString_clear", _wrap_vectorString_clear, METH_O, NULL},
	 { "new_vectorString", _wrap_new_vectorString, METH_VARARGS, NULL},
	 { "delete_vectorString", _wrap_delete_vectorString, METH_O, NULL},
	 { "vectorString_swigregister", vectorString_swigregister, METH_O, NULL},
	 { "vectorString_swiginit", vectorString_swiginit, METH_VARARGS, NULL},
	 { "vectorVector3_iterator", _wrap_vectorVector3_iterator, METH_O, NULL},
	 { "vectorVector3___nonzero__", _wrap_vectorVector3___nonzero__, METH_O, NULL},
	 { "vectorVector3___bool__", _wrap_vectorVector3___bool__, METH_O, NULL},
	 { "vectorVector3___len__", _wrap_vectorVector3___len__, METH_O, NULL},
	 { "vectorVector3___getslice__", _wrap_vectorVector3___getslice__, METH_VARARGS, NULL},
	 { "vectorVector3___setslice__", _wrap_vectorVector3___setslice__, METH_VARARGS, NULL},
	 { "vectorVector3___delslice__", _wrap_vectorVector3___delslice__, METH_VARARGS, NULL},
	 { "vectorVector3___delitem__", _wrap_vectorVector3___delitem__, METH_VARARGS, NULL},
	 { "vectorVector3___getitem__", _wrap_vectorVector3___getitem__, METH_VARARGS, NULL},
	 { "vectorVector3___setitem__", _wrap_vectorVector3___setitem__, METH_VARARGS, NULL},
	 { "vectorVector3_clear", _wrap_vectorVector3_clear, METH_O, NULL},
	 { "new_vectorVector3", _wrap_new_vectorVector3, METH_VARARGS, NULL},
	 { "delete_vectorVector3", _wrap_delete_vectorVector3, METH_O, NULL},
	 { "vectorVector3_swigregister", vectorVector3_swigregister, METH_O, NULL},
	 { "vectorVector3_swiginit", vectorVector3_swiginit, METH_VARARGS, NULL},
	 { "vectorvVector3_iterator", _wrap_vectorvVector3_iterator, METH_O, NULL},
	 { "vectorvVector3___nonzero__", _wrap_vectorvVector3___nonzero__, METH_O, NULL},
	 { "vectorvVector3___bool__", _wrap_vectorvVector3___bool__, METH_O, NULL},
	 { "vectorvVector3___len__", _wrap_vectorvVector3___len__, METH_O, NULL},
	 { "vectorvVector3___getslice__", _wrap_vectorvVector3___getslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___setslice__", _wrap_vectorvVector3___setslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___delslice__", _wrap_vectorvVector3___delslice__, METH_VARARGS, NULL},
	 { "vectorvVector3___delitem__", _wrap_vectorvVector3___delitem__, METH_VARARGS, NULL},
	 { "vectorvVector3___getitem__", _wrap_vectorvVector3___getitem__, METH_VARARGS, NULL},
	 { "vectorvVector3___setitem__", _wrap_vectorvVector3___setitem__, METH_VARARGS, NULL},
	 { "vectorvVector3_clear", _wrap_vectorvVector3_clear, METH_O, NULL},
	 { "new_vectorvVector3", _wrap_new_vectorvVector3, METH_VARARGS, NULL},
	 { "delete_vectorvVector3", _wrap_delete_vectorvVector3, METH_O, NULL},
	 { "vectorvVector3_swigregister", vectorvVector3_swigregister, METH_O, NULL},
	 { "vectorvVector3_swiginit", vectorvVector3_swiginit, METH_VARARGS, NULL},
	 { "vectorOBMol_iterator", _wrap_vectorOBMol_iterator, METH_O, NULL},
	 { "vectorOBMol___nonzero__", _wrap_vectorOBMol___nonzero__, METH_O, NULL},
	 { "vectorOBMol___bool__", _wrap_vectorOBMol___bool__, METH_O, NULL},
	 { "vectorOBMol___len__", _wrap_vectorOBMol___len__, METH_O, NULL},
	 { "vectorOBMol___getslice__", _wrap_vectorOBMol___getslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___setslice__", _wrap_vectorOBMol___setslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___delslice__", _wrap_vectorOBMol___delslice__, METH_VARARGS, NULL},
	 { "vectorOBMol___delitem__", _wrap_vectorOBMol___delitem__, METH_VARARGS, NULL},
	 { "vectorOBMol___getitem__", _wrap_vectorOBMol___getitem__, METH_VARARGS, NULL},
	 { "vectorOBMol___setitem__", _wrap_vectorOBMol___setitem__, METH_VARARGS, NULL},
	 { "vectorOBMol_clear", _wrap_vectorOBMol_clear, METH_O, NULL},
	 { "new_vectorOBMol", _wrap_new_vectorOBMol, METH_VARARGS, NULL},
	 { "delete_vectorOBMol", _wrap_delete_vectorOBMol, METH_O, NULL},
	 { "vectorOBMol_swigregister", vectorOBMol_swigregister, METH_O, NULL},
	 { "vectorOBMol_swiginit", vectorOBMol_swiginit, METH_VARARGS, NULL},
	 { "vectorOBBond_iterator", _wrap_vectorOBBond_iterator, METH_O, NULL},
	 { "vectorOBBond___nonzero__", _wrap_vectorOBBond___nonzero__, METH_O, NULL},
	 { "vectorOBBond___bool__", _wrap_vectorOBBond___bool__, METH_O, NULL},
	 { "vectorOBBond___len__", _wrap_vectorOBBond___len__, METH_O, NULL},
	 { "vectorOBBond___getslice__", _wrap_vectorOBBond___getslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___setslice__", _wrap_vectorOBBond___setslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___delslice__", _wrap_vectorOBBond___delslice__, METH_VARARGS, NULL},
	 { "vectorOBBond___delitem__", _wrap_vectorOBBond___delitem__, METH_VARARGS, NULL},
	 { "vectorOBBond___getitem__", _wrap_vectorOBBond___getitem__, METH_VARARGS, NULL},
	 { "vectorOBBond___setitem__", _wrap_vectorOBBond___setitem__, METH_VARARGS, NULL},
	 { "vectorOBBond_clear", _wrap_vectorOBBond_clear, METH_O, NULL},
	 { "new_vectorOBBond", _wrap_new_vectorOBBond, METH_VARARGS, NULL},
	 { "delete_vectorOBBond", _wrap_delete_vectorOBBond, METH_O, NULL},
	 { "vectorOBBond_swigregister", vectorOBBond_swigregister, METH_O, NULL},
	 { "vectorOBBond_swiginit", vectorOBBond_swiginit, METH_VARARGS, NULL},
	 { "vectorOBResidue_iterator", _wrap_vectorOBResidue_iterator, METH_O, NULL},
	 { "vectorOBResidue___nonzero__", _wrap_vectorOBResidue___nonzero__, METH_O, NULL},
	 { "vectorOBResidue___bool__", _wrap_vectorOBResidue___bool__, METH_O, NULL},
	 { "vectorOBResidue___len__", _wrap_vectorOBResidue___len__, METH_O, NULL},
	 { "vectorOBResidue___getslice__", _wrap_vectorOBResidue___getslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___setslice__", _wrap_vectorOBResidue___setslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___delslice__", _wrap_vectorOBResidue___delslice__, METH_VARARGS, NULL},
	 { "vectorOBResidue___delitem__", _wrap_vectorOBResidue___delitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue___getitem__", _wrap_vectorOBResidue___getitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue___setitem__", _wrap_vectorOBResidue___setitem__, METH_VARARGS, NULL},
	 { "vectorOBResidue_clear", _wrap_vectorOBResidue_clear, METH_O, NULL},
	 { "new_vectorOBResidue", _wrap_new_vectorOBResidue, METH_VARARGS, NULL},
	 { "delete_vectorOBResidue", _wrap_delete_vectorOBResidue, METH_O, NULL},
	 { "vectorOBResidue_swigregister", vectorOBResidue_swigregister, METH_O, NULL},
	 { "vectorOBResidue_swiginit", vectorOBResidue_swiginit, METH_VARARGS, NULL},
	 { "vectorOBRing_iterator", _wrap_vectorOBRing_iterator, METH_O, NULL},
	 { "vectorOBRing___nonzero__", _wrap_vectorOBRing___nonzero__, METH_O, NULL},
	 { "vectorOBRing___bool__", _wrap_vectorOBRing___bool__, METH_O, NULL},
	 { "vectorOBRing___len__", _wrap_vectorOBRing___len__, METH_O, NULL},
	 { "vectorOBRing___getslice__", _wrap_vectorOBRing___getslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___setslice__", _wrap_vectorOBRing___setslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___delslice__", _wrap_vectorOBRing___delslice__, METH_VARARGS, NULL},
	 { "vectorOBRing___delitem__", _wrap_vectorOBRing___delitem__, METH_VARARGS, NULL},
	 { "vectorOBRing___getitem__", _wrap_vectorOBRing___getitem__, METH_VARARGS, NULL},
	 { "vectorOBRing___setitem__", _wrap_vectorOBRing___setitem__, METH_VARARGS, NULL},
	 { "vectorOBRing_clear", _wrap_vectorOBRing_clear, METH_O, NULL},
	 { "new_vectorOBRing", _wrap_new_vectorOBRing, METH_VARARGS, NULL},
	 { "delete_vectorOBRing", _wrap_delete_vectorOBRing, METH_O, NULL},
	 { "vectorOBRing_swigregister", vectorOBRing_swigregister, METH_O, NULL},
	 { "vectorOBRing_swiginit", vectorOBRing_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBRing_iterator", _wrap_vectorpOBRing_iterator, METH_O, NULL},
	 { "vectorpOBRing___nonzero__", _wrap_vectorpOBRing___nonzero__, METH_O, NULL},
	 { "vectorpOBRing___bool__", _wrap_vectorpOBRing___bool__, METH_O, NULL},
	 { "vectorpOBRing___len__", _wrap_vectorpOBRing___len__, METH_O, NULL},
	 { "vectorpOBRing___getslice__", _wrap_vectorpOBRing___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___setslice__", _wrap_vectorpOBRing___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___delslice__", _wrap_vectorpOBRing___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBRing___delitem__", _wrap_vectorpOBRing___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing___getitem__", _wrap_vectorpOBRing___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing___setitem__", _wrap_vectorpOBRing___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBRing_clear", _wrap_vectorpOBRing_clear, METH_O, NULL},
	 { "new_vectorpOBRing", _wrap_new_vectorpOBRing, METH_VARARGS, NULL},
	 { "delete_vectorpOBRing", _wrap_delete_vectorpOBRing, METH_O, NULL},
	 { "vectorpOBRing_swigregister", vectorpOBRing_swigregister, METH_O, NULL},
	 { "vectorpOBRing_swiginit", vectorpOBRing_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBGenericData_iterator", _wrap_vectorpOBGenericData_iterator, METH_O, NULL},
	 { "vectorpOBGenericData___nonzero__", _wrap_vectorpOBGenericData___nonzero__, METH_O, NULL},
	 { "vectorpOBGenericData___bool__", _wrap_vectorpOBGenericData___bool__, METH_O, NULL},
	 { "vectorpOBGenericData___len__", _wrap_vectorpOBGenericData___len__, METH_O, NULL},
	 { "vectorpOBGenericData___getslice__", _wrap_vectorpOBGenericData___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___setslice__", _wrap_vectorpOBGenericData___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___delslice__", _wrap_vectorpOBGenericData___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___delitem__", _wrap_vectorpOBGenericData___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___getitem__", _wrap_vectorpOBGenericData___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData___setitem__", _wrap_vectorpOBGenericData___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBGenericData_clear", _wrap_vectorpOBGenericData_clear, METH_O, NULL},
	 { "new_vectorpOBGenericData", _wrap_new_vectorpOBGenericData, METH_VARARGS, NULL},
	 { "delete_vectorpOBGenericData", _wrap_delete_vectorpOBGenericData, METH_O, NULL},
	 { "vectorpOBGenericData_swigregister", vectorpOBGenericData_swigregister, METH_O, NULL},
	 { "vectorpOBGenericData_swiginit", vectorpOBGenericData_swiginit, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord_iterator", _wrap_vectorpOBInternalCoord_iterator, METH_O, NULL},
	 { "vectorpOBInternalCoord___nonzero__", _wrap_vectorpOBInternalCoord___nonzero__, METH_O, NULL},
	 { "vectorpOBInternalCoord___bool__", _wrap_vectorpOBInternalCoord___bool__, METH_O, NULL},
	 { "vectorpOBInternalCoord___len__", _wrap_vectorpOBInternalCoord___len__, METH_O, NULL},
	 { "vectorpOBInternalCoord___getslice__", _wrap_vectorpOBInternalCoord___getslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___setslice__", _wrap_vectorpOBInternalCoord___setslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___delslice__", _wrap_vectorpOBInternalCoord___delslice__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___delitem__", _wrap_vectorpOBInternalCoord___delitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___getitem__", _wrap_vectorpOBInternalCoord___getitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord___setitem__", _wrap_vectorpOBInternalCoord___setitem__, METH_VARARGS, NULL},
	 { "vectorpOBInternalCoord_clear", _wrap_vectorpOBInternalCoord_clear, METH_O, NULL},
	 { "new_vectorpOBInternalCoord", _wrap_new_vectorpOBInternalCoord, METH_VARARGS, NULL},
	 { "delete_vectorpOBInternalCoord", _wrap_delete_vectorpOBInternalCoord, METH_O, NULL},
	 { "vectorpOBInternalCoord_swigregister", vectorpOBInternalCoord_swigregister, METH_O, NULL},
	 { "vectorpOBInternalCoord_swiginit", vectorpOBInternalCoord_swiginit, METH_VARARGS, NULL},
	 { "new_pairUIntUInt", _wrap_new_pairUIntUInt, METH_VARARGS, NULL},
	 { "pairUIntUInt_first_set", _wrap_pairUIntUInt_first_set, METH_VARARGS, NULL},
	 { "pairUIntUInt_first_get", _wrap_pairUIntUInt_first_get, METH_O, NULL},
	 { "pairUIntUInt_second_set", _wrap_pairUIntUInt_second_set, METH_VARARGS, NULL},
	 { "pairUIntUInt_second_get", _wrap_pairUIntUInt_second_get, METH_O, NULL},
	 { "delete_pairUIntUInt", _wrap_delete_pairUIntUInt, METH_O, NULL},
	 { "pairUIntUInt_swigregister", pairUIntUInt_swigregister, METH_O, NULL},
	 { "pairUIntUInt_swiginit", pairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "vpairUIntUInt_iterator", _wrap_vpairUIntUInt_iterator, METH_O, NULL},
	 { "vpairUIntUInt___nonzero__", _wrap_vpairUIntUInt___nonzero__, METH_O, NULL},
	 { "vpairUIntUInt___bool__", _wrap_vpairUIntUInt___bool__, METH_O, NULL},
	 { "vpairUIntUInt___len__", _wrap_vpairUIntUInt___len__, METH_O, NULL},
	 { "vpairUIntUInt___getslice__", _wrap_vpairUIntUInt___getslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___setslice__", _wrap_vpairUIntUInt___setslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___delslice__", _wrap_vpairUIntUInt___delslice__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___delitem__", _wrap_vpairUIntUInt___delitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___getitem__", _wrap_vpairUIntUInt___getitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt___setitem__", _wrap_vpairUIntUInt___setitem__, METH_VARARGS, NULL},
	 { "vpairUIntUInt_clear", _wrap_vpairUIntUInt_clear, METH_O, NULL},
	 { "new_vpairUIntUInt", _wrap_new_vpairUIntUInt, METH_VARARGS, NULL},
	 { "delete_vpairUIntUInt", _wrap_delete_vpairUIntUInt, METH_O, NULL},
	 { "vpairUIntUInt_swigregister", vpairUIntUInt_swigregister, METH_O, NULL},
	 { "vpairUIntUInt_swiginit", vpairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_iterator", _wrap_vvpairUIntUInt_iterator, METH_O, NULL},
	 { "vvpairUIntUInt___nonzero__", _wrap_vvpairUIntUInt___nonzero__, METH_O, NULL},
	 { "vvpairUIntUInt___bool__", _wrap_vvpairUIntUInt___bool__, METH_O, NULL},
	 { "vvpairUIntUInt___len__", _wrap_vvpairUIntUInt___len__, METH_O, NULL},
	 { "vvpairUIntUInt___getslice__", _wrap_vvpairUIntUInt___getslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___setslice__", _wrap_vvpairUIntUInt___setslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___delslice__", _wrap_vvpairUIntUInt___delslice__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___delitem__", _wrap_vvpairUIntUInt___delitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___getitem__", _wrap_vvpairUIntUInt___getitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt___setitem__", _wrap_vvpairUIntUInt___setitem__, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_pop", _wrap_vvpairUIntUInt_pop, METH_O, NULL},
	 { "vvpairUIntUInt_append", _wrap_vvpairUIntUInt_append, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_empty", _wrap_vvpairUIntUInt_empty, METH_O, NULL},
	 { "vvpairUIntUInt_size", _wrap_vvpairUIntUInt_size, METH_O, NULL},
	 { "vvpairUIntUInt_swap", _wrap_vvpairUIntUInt_swap, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_begin", _wrap_vvpairUIntUInt_begin, METH_O, NULL},
	 { "vvpairUIntUInt_end", _wrap_vvpairUIntUInt_end, METH_O, NULL},
	 { "vvpairUIntUInt_rbegin", _wrap_vvpairUIntUInt_rbegin, METH_O, NULL},
	 { "vvpairUIntUInt_rend", _wrap_vvpairUIntUInt_rend, METH_O, NULL},
	 { "vvpairUIntUInt_clear", _wrap_vvpairUIntUInt_clear, METH_O, NULL},
	 { "vvpairUIntUInt_get_allocator", _wrap_vvpairUIntUInt_get_allocator, METH_O, NULL},
	 { "vvpairUIntUInt_pop_back", _wrap_vvpairUIntUInt_pop_back, METH_O, NULL},
	 { "vvpairUIntUInt_erase", _wrap_vvpairUIntUInt_erase, METH_VARARGS, NULL},
	 { "new_vvpairUIntUInt", _wrap_new_vvpairUIntUInt, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_push_back", _wrap_vvpairUIntUInt_push_back, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_front", _wrap_vvpairUIntUInt_front, METH_O, NULL},
	 { "vvpairUIntUInt_back", _wrap_vvpairUIntUInt_back, METH_O, NULL},
	 { "vvpairUIntUInt_assign", _wrap_vvpairUIntUInt_assign, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_resize", _wrap_vvpairUIntUInt_resize, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_insert", _wrap_vvpairUIntUInt_insert, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_reserve", _wrap_vvpairUIntUInt_reserve, METH_VARARGS, NULL},
	 { "vvpairUIntUInt_capacity", _wrap_vvpairUIntUInt_capacity, METH_O, NULL},
	 { "delete_vvpairUIntUInt", _wrap_delete_vvpairUIntUInt, METH_O, NULL},
	 { "vvpairUIntUInt_swigregister", vvpairUIntUInt_swigregister, METH_O, NULL},
	 { "vvpairUIntUInt_swiginit", vvpairUIntUInt_swiginit, METH_VARARGS, NULL},
	 { "toAliasData", _wrap_toAliasData, METH_O, NULL},
	 { "toAngleData", _wrap_toAngleData, METH_O, NULL},
	 { "toCommentData", _wrap_toCommentData, METH_O, NULL},
	 { "toConformerData", _wrap_toConformerData, METH_O, NULL},
	 { "toExternalBondData", _wrap_toExternalBondData, METH_O, NULL},
	 { "toGridData", _wrap_toGridData, METH_O, NULL},
	 { "toMatrixData", _wrap_toMatrixData, METH_O, NULL},
	 { "toNasaThermoData", _wrap_toNasaThermoData, METH_O, NULL},
	 { "toPairData", _wrap_toPairData, METH_O, NULL},
	 { "toPairInteger", _wrap_toPairInteger, METH_O, NULL},
	 { "toPairFloatingPoint", _wrap_toPairFloatingPoint, METH_O, NULL},
	 { "toRateData", _wrap_toRateData, METH_O, NULL},
	 { "toRotamerList", _wrap_toRotamerList, METH_O, NULL},
	 { "toRotationData", _wrap_toRotationData, METH_O, NULL},
	 { "toSerialNums", _wrap_toSerialNums, METH_O, NULL},
	 { "toSetData", _wrap_toSetData, METH_O, NULL},
	 { "toSymmetryData", _wrap_toSymmetryData, METH_O, NULL},
	 { "toTorsionData", _wrap_toTorsionData, METH_O, NULL},
	 { "toUnitCell", _wrap_toUnitCell, METH_O, NULL},
	 { "toVectorData", _wrap_toVectorData, METH_O, NULL},
	 { "toVibrationData", _wrap_toVibrationData, METH_O, NULL},
	 { "toVirtualBond", _wrap_toVirtualBond, METH_O, NULL},
	 { "toStereoBase", _wrap_toStereoBase, METH_O, NULL},
	 { "toTetrahedralStereo", _wrap_toTetrahedralStereo, METH_O, NULL},
	 { "toCisTransStereo", _wrap_toCisTransStereo, METH_O, NULL},
	 { "toSquarePlanarStereo", _wrap_toSquarePlanarStereo, METH_O, NULL},
	 { "new_OBGlobalDataBase", _wrap_new_OBGlobalDataBase, METH_NOARGS, "Constructor"},
	 { "delete_OBGlobalDataBase", _wrap_delete_OBGlobalDataBase, METH_O, "Destructor"},
	 { "OBGlobalDataBase_Init", _wrap_OBGlobalDataBase_Init, METH_O, "Read in the data file, falling back as needed"},
	 { "OBGlobalDataBase_GetSize", _wrap_OBGlobalDataBase_GetSize, METH_O, "\n"
		":rtype: int\n"
		":return: the size of the database (for error checking)\n"
		""},
	 { "OBGlobalDataBase_SetReadDirectory", _wrap_OBGlobalDataBase_SetReadDirectory, METH_VARARGS, "Set the directory before calling Init()"},
	 { "OBGlobalDataBase_SetEnvironmentVariable", _wrap_OBGlobalDataBase_SetEnvironmentVariable, METH_VARARGS, "Set the environment variable to use before calling Init()"},
	 { "OBGlobalDataBase_ParseLine", _wrap_OBGlobalDataBase_ParseLine, METH_VARARGS, "Specified by particular table classes (parses an individual data line)"},
	 { "OBGlobalDataBase_swigregister", OBGlobalDataBase_swigregister, METH_O, NULL},
	 { "OBGlobalDataBase_swiginit", OBGlobalDataBase_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomHOF", _wrap_new_OBAtomHOF, METH_VARARGS, "\n"
		"Initialize Heat of Formation for atom\n"
		"\n"
		"    :type element: string\n"
		"    :param element: The element string\n"
		"    :type charge: int\n"
		"    :param charge:  The formal charge of the particle (if an ion)\n"
		"    :type method: string\n"
		"    :param method:  The method used for determining the value\n"
		"    :type desc: string\n"
		"    :param desc:    Description of the value\n"
		"    :type T: float\n"
		"    :param T:       Temperature\n"
		"    :type value: float\n"
		"    :param value:   The value of the property (energy)\n"
		"    :type multiplicity: int\n"
		"    :param multiplicity: The multiplicity of the atomic system\n"
		"    :type unit: string\n"
		"    :param unit:    The (energy) unit\n"
		""},
	 { "delete_OBAtomHOF", _wrap_delete_OBAtomHOF, METH_O, "Destructor"},
	 { "OBAtomHOF_Element", _wrap_OBAtomHOF_Element, METH_O, "Return the chemical element"},
	 { "OBAtomHOF_Charge", _wrap_OBAtomHOF_Charge, METH_O, "Return the formal charge"},
	 { "OBAtomHOF_Method", _wrap_OBAtomHOF_Method, METH_O, "Return the method used for the measurement/calculation"},
	 { "OBAtomHOF_Desc", _wrap_OBAtomHOF_Desc, METH_O, "Return specification of the measurement/calculation type"},
	 { "OBAtomHOF_T", _wrap_OBAtomHOF_T, METH_O, "Return the temperature"},
	 { "OBAtomHOF_Value", _wrap_OBAtomHOF_Value, METH_O, "Return the (energy) value"},
	 { "OBAtomHOF_Multiplicity", _wrap_OBAtomHOF_Multiplicity, METH_O, "Return the multiplicity"},
	 { "OBAtomHOF_Unit", _wrap_OBAtomHOF_Unit, METH_O, "Return the (energy) unit"},
	 { "OBAtomHOF_swigregister", OBAtomHOF_swigregister, METH_O, NULL},
	 { "OBAtomHOF_swiginit", OBAtomHOF_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomicHeatOfFormationTable", _wrap_new_OBAtomicHeatOfFormationTable, METH_NOARGS, "Constructor"},
	 { "delete_OBAtomicHeatOfFormationTable", _wrap_delete_OBAtomicHeatOfFormationTable, METH_O, "Destructor"},
	 { "OBAtomicHeatOfFormationTable_GetHeatOfFormation", _wrap_OBAtomicHeatOfFormationTable_GetHeatOfFormation, METH_VARARGS, "\n"
		"Extract heat of formation and entropy for an atom\n"
		"      :type elem: string\n"
		"      :param elem:         The chemical element we're looking for\n"
		"      :type charge: int\n"
		"      :param charge:       At this formal charge\n"
		"      :type method: string\n"
		"      :param method:       The method used for computing/measuring\n"
		"      :type T: float\n"
		"      :param T:            The temperature\n"
		"      :type dhof0: float\n"
		"      :param dhof0:        The output energy at 0K\n"
		"      :param dhof1:        The output energy at T\n"
		"      :type S0T: float\n"
		"      :param S0T:          The entropy at T (it is 0 at 0K)\n"
		"      :rtype: int\n"
		"      :return: 1 if the contribution to the Heat of Formation for this atom\n"
		"          is known at temperature T. If 1 the values\n"
		"          including all corrections are returned in the dhof variable.\n"
		""},
	 { "OBAtomicHeatOfFormationTable_swigregister", OBAtomicHeatOfFormationTable_swigregister, METH_O, NULL},
	 { "OBAtomicHeatOfFormationTable_swiginit", OBAtomicHeatOfFormationTable_swiginit, METH_VARARGS, NULL},
	 { "new_OBTypeTable", _wrap_new_OBTypeTable, METH_NOARGS, NULL},
	 { "delete_OBTypeTable", _wrap_delete_OBTypeTable, METH_O, NULL},
	 { "OBTypeTable_SetFromType", _wrap_OBTypeTable_SetFromType, METH_VARARGS, "Set the initial atom type to be translated"},
	 { "OBTypeTable_SetToType", _wrap_OBTypeTable_SetToType, METH_VARARGS, "Set the destination atom type for translation"},
	 { "OBTypeTable_Translate", _wrap_OBTypeTable_Translate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Translate atom types\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Translate atom types :rtype: boolean\n"
		":return: whether the translation was successful\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Translate atom types :rtype: string\n"
		":return: the translated atom type, or an empty string if not possible\n"
		""},
	 { "OBTypeTable_GetFromType", _wrap_OBTypeTable_GetFromType, METH_O, "\n"
		":rtype: string\n"
		":return: the initial atom type to be translated\n"
		""},
	 { "OBTypeTable_GetToType", _wrap_OBTypeTable_GetToType, METH_O, "\n"
		":rtype: string\n"
		":return: the destination atom type for translation\n"
		""},
	 { "OBTypeTable_swigregister", OBTypeTable_swigregister, METH_O, NULL},
	 { "OBTypeTable_swiginit", OBTypeTable_swiginit, METH_VARARGS, NULL},
	 { "new_OBResidueData", _wrap_new_OBResidueData, METH_NOARGS, NULL},
	 { "OBResidueData_SetResName", _wrap_OBResidueData_SetResName, METH_VARARGS, "\n"
		" Sets the table to access the residue information for a specified  residue name :rtype: boolean\n"
		":return: whether this residue name is in the table\n"
		""},
	 { "OBResidueData_LookupBO", _wrap_OBResidueData_LookupBO, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: int\n"
		":return: the bond order for the bond specified in the current residue Deprecated: Easier to use the two-argument form\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: the bond order for the bond specified between the two specified atom labels\n"
		""},
	 { "OBResidueData_LookupType", _wrap_OBResidueData_LookupType, METH_VARARGS, "\n"
		" Look up the atom type and hybridization for the atom label specified in the first argument for the current residue :rtype: boolean\n"
		":return: whether the atom label specified is found in the current residue\n"
		""},
	 { "OBResidueData_AssignBonds", _wrap_OBResidueData_AssignBonds, METH_VARARGS, "Assign bond orders, atom types and residues for the supplied OBMol based on the residue information assigned to atoms"},
	 { "delete_OBResidueData", _wrap_delete_OBResidueData, METH_O, NULL},
	 { "OBResidueData_swigregister", OBResidueData_swigregister, METH_O, NULL},
	 { "OBResidueData_swiginit", OBResidueData_swiginit, METH_VARARGS, NULL},
	 { "OBStopwatch_Start", _wrap_OBStopwatch_Start, METH_O, "Mark the start of \"stopwatch\" timing"},
	 { "OBStopwatch_Lap", _wrap_OBStopwatch_Lap, METH_O, "\n"
		":rtype: float\n"
		":return: The time since calling OBStopwatch::Start() in seconds.\n"
		""},
	 { "OBStopwatch_Elapsed", _wrap_OBStopwatch_Elapsed, METH_O, "\n"
		":rtype: float\n"
		":return: The time since calling OBStopwatch::Start() in seconds.\n"
		""},
	 { "new_OBStopwatch", _wrap_new_OBStopwatch, METH_NOARGS, NULL},
	 { "delete_OBStopwatch", _wrap_delete_OBStopwatch, METH_O, NULL},
	 { "OBStopwatch_swigregister", OBStopwatch_swigregister, METH_O, NULL},
	 { "OBStopwatch_swiginit", OBStopwatch_swiginit, METH_VARARGS, NULL},
	 { "new_OBSqrtTbl", _wrap_new_OBSqrtTbl, METH_VARARGS, "\n"
		" Create a square root table to handle up to the square root of ``max`` (e.g., if you want the square root of 144, supply 12 for max) :type max: float\n"
		":param max: The maximum square root stored in the lookup table :type incr: float\n"
		":param incr: The floating point resolution of the lookup table\n"
		""},
	 { "delete_OBSqrtTbl", _wrap_delete_OBSqrtTbl, METH_O, NULL},
	 { "OBSqrtTbl_Sqrt", _wrap_OBSqrtTbl_Sqrt, METH_VARARGS, "\n"
		" Fast square root calculation using a lookup table :rtype: float\n"
		":return: Square root of ``d2``\n"
		""},
	 { "OBSqrtTbl_Init", _wrap_OBSqrtTbl_Init, METH_VARARGS, "\n"
		" Initialize the square root lookup table :type max: float\n"
		":param max: The maximum square root stored in the lookup table (e.g., if you want the square root of 144, supply 12 for max) :type incr: float\n"
		":param incr: The floating point resolution of the lookup table\n"
		""},
	 { "OBSqrtTbl_swigregister", OBSqrtTbl_swigregister, METH_O, NULL},
	 { "OBSqrtTbl_swiginit", OBSqrtTbl_swiginit, METH_VARARGS, NULL},
	 { "rotate_coords", _wrap_rotate_coords, METH_VARARGS, NULL},
	 { "calc_rms", _wrap_calc_rms, METH_VARARGS, NULL},
	 { "new_vector3", _wrap_new_vector3, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy Constructor\n"
		""},
	 { "delete_vector3", _wrap_delete_vector3, METH_O, "Destructor"},
	 { "vector3_begin", _wrap_vector3_begin, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" :rtype: float\n"
		" :return: iterator to beginning\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"return const_iterator to beginning\n"
		""},
	 { "vector3_end", _wrap_vector3_end, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" :rtype: float\n"
		" :return: iterator to end\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"return const_iterator to end\n"
		""},
	 { "vector3_Set", _wrap_vector3_Set, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set x,y and z-component of a vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set x,y and z-component of a vector from c[0]..c[2]\n"
		""},
	 { "vector3_SetX", _wrap_vector3_SetX, METH_VARARGS, "Access function to set the x-coordinate of the vector"},
	 { "vector3_SetY", _wrap_vector3_SetY, METH_VARARGS, "Access function to set the y-coordinate of the vector"},
	 { "vector3_SetZ", _wrap_vector3_SetZ, METH_VARARGS, "Access function to set the z-coordinate of the vector"},
	 { "vector3_GetX", _wrap_vector3_GetX, METH_O, "Access function to get the x-coordinate of the vector"},
	 { "vector3_GetY", _wrap_vector3_GetY, METH_O, "Access function to get the y-coordinate of the vector"},
	 { "vector3_GetZ", _wrap_vector3_GetZ, METH_O, "Access function to get the z-coordinate of the vector"},
	 { "vector3_Get", _wrap_vector3_Get, METH_VARARGS, "Set c[0]..c[2] to the components of the vector Warning: No error checking is performed"},
	 { "vector3_AsArray", _wrap_vector3_AsArray, METH_O, "\n"
		":rtype: float\n"
		":return: the vector as a const double *\n"
		""},
	 { "vector3___iadd__", _wrap_vector3___iadd__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Vector addition (add ``v`` to *this) :rtype: :py:class:`vector3`\n"
		":return: *this + v\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Scalar addition (add ``f`` to *this) :rtype: :py:class:`vector3`\n"
		":return: *this + f\n"
		""},
	 { "vector3___isub__", _wrap_vector3___isub__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Vector subtraction (subtract ``v`` from *this) :rtype: :py:class:`vector3`\n"
		":return: *this - v\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Scalar subtraction (subtract ``f`` from *this) :rtype: :py:class:`vector3`\n"
		":return: *this - f\n"
		""},
	 { "vector3___itruediv__", _wrap_vector3___itruediv__, METH_VARARGS, "\n"
		" Scalar division (divide *this by ``c)`` :rtype: :py:class:`vector3`\n"
		":return: *this divided by c\n"
		""},
	 { "vector3___imul__", _wrap_vector3___imul__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Scalar multiplication (multiply *this by ``c)`` :rtype: :py:class:`vector3`\n"
		":return: *this * c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Multiplication of matrix and vector :rtype: :py:class:`vector3`\n"
		":return: the result (i.e., the updated vector) TODO: Currently unimplemented\n"
		""},
	 { "vector3_randomUnitVector", _wrap_vector3_randomUnitVector, METH_O, "Create a random unit vector"},
	 { "vector3_normalize", _wrap_vector3_normalize, METH_O, "\n"
		" Scales a vector to give it length one. :rtype: :py:class:`vector3`\n"
		":return: the result (i.e., the normalized vector)\n"
		""},
	 { "vector3_CanBeNormalized", _wrap_vector3_CanBeNormalized, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether a vector can be normalized\n"
		""},
	 { "vector3_length_2", _wrap_vector3_length_2, METH_O, "\n"
		":rtype: float\n"
		":return: The length of the vector squared\n"
		""},
	 { "vector3_length", _wrap_vector3_length, METH_O, "\n"
		":rtype: float\n"
		":return: The vector length\n"
		""},
	 { "vector3_x", _wrap_vector3_x, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the x-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the x-coordinate of the vector\n"
		""},
	 { "vector3_y", _wrap_vector3_y, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the y-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the y-coordinate of the vector\n"
		""},
	 { "vector3_z", _wrap_vector3_z, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Access function to get the z-coordinate of the vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Access function to set the z-coordinate of the vector\n"
		""},
	 { "vector3___eq__", _wrap_vector3___eq__, METH_VARARGS, "\n"
		" Comparison Methods Equivalence of vectors Deprecated: This method uses unreliable floating point == comparisons    Use vector3::IsApprox() instead. :rtype: boolean\n"
		":return: true if every component is equal\n"
		""},
	 { "vector3___ne__", _wrap_vector3___ne__, METH_VARARGS, "\n"
		" Deprecated: This method uses unreliable floating point == comparisons    Use vector3::IsApprox() instead. :rtype: boolean\n"
		":return: true if at least one component of the two vectors is !=\n"
		""},
	 { "vector3_IsApprox", _wrap_vector3_IsApprox, METH_VARARGS, "\n"
		" Safe comparison for floating-point vector3 :rtype: boolean\n"
		":return: true if the vector *this is approximately equal to the vector         ``other``, to the precision ``precision``. More specifically,         this method works exactly like the OpenBabel::IsApprox()         function, replacing the absolute value for doubles by the norm         for vectors. :type other: :py:class:`vector3`\n"
		":param other: The vector for comparison :type precision: float\n"
		":param precision: This parameter plays the same role as in        OpenBabel::IsApprox().\n"
		""},
	 { "vector3_distSq", _wrap_vector3_distSq, METH_VARARGS, "\n"
		" }@ :rtype: float\n"
		":return: square of the distance between *this and vv equivalent to length_2(*this-vv)\n"
		""},
	 { "vector3_createOrthoVector", _wrap_vector3_createOrthoVector, METH_VARARGS, "\n"
		" Creates a vector of length one, orthogonal to *this. :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "vector3_swigregister", vector3_swigregister, METH_O, NULL},
	 { "vector3_swiginit", vector3_swiginit, METH_VARARGS, NULL},
	 { "__add__", _wrap___add__, METH_VARARGS, "Vector addition"},
	 { "__truediv__", _wrap___truediv__, METH_VARARGS, "Division by a scalar"},
	 { "dot", _wrap_dot, METH_VARARGS, "Dot product of two vectors"},
	 { "cross", _wrap_cross, METH_VARARGS, "Cross product of two vectors"},
	 { "vectorAngle", _wrap_vectorAngle, METH_VARARGS, "Calculate the angle between vectors (in degrees)"},
	 { "CalcTorsionAngle", _wrap_CalcTorsionAngle, METH_VARARGS, "Calculate the torsion angle between vectors (in degrees)"},
	 { "Point2PlaneSigned", _wrap_Point2PlaneSigned, METH_VARARGS, "Calculate the signed distance of point a to the plane determined by b,c,d"},
	 { "Point2Plane", _wrap_Point2Plane, METH_VARARGS, "Calculate the distance of point a to the plane determined by b,c,d"},
	 { "Point2PlaneAngle", _wrap_Point2PlaneAngle, METH_VARARGS, "Calculate the angle between point a and the plane determined by b,c,d"},
	 { "Point2Line", _wrap_Point2Line, METH_VARARGS, "Calculate the distance of a point a to a line determined by b and c"},
	 { "new_matrix3x3", _wrap_new_matrix3x3, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructs the zero-matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructs s times the unit matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Constructs a matrix from row vectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Constructs a matrix from a 3x3-array of doubles The first index represents the row, the second index the column\n"
		""},
	 { "delete_matrix3x3", _wrap_delete_matrix3x3, METH_O, "Destructor"},
	 { "matrix3x3_GetArray", _wrap_matrix3x3_GetArray, METH_VARARGS, "\n"
		"Access function Writes the matrix into the 1-dimensional array m, row by\n"
		"       row. The array must be able to hold 9 doubles, otherwise your\n"
		"       program will segfault.\n"
		""},
	 { "matrix3x3___call__", _wrap_matrix3x3___call__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: a constant reference to an element of the matrix.\n"
		"             row and column must be between 0 and 2. No check is done.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: a non-constant reference to an element of the matrix.\n"
		"             row and column must be between 0 and 2. No check is done.\n"
		""},
	 { "matrix3x3_inverse", _wrap_matrix3x3_inverse, METH_O, "Calculates the inverse of a matrix."},
	 { "matrix3x3_transpose", _wrap_matrix3x3_transpose, METH_O, "Calculates the transpose of a matrix."},
	 { "matrix3x3_determinant", _wrap_matrix3x3_determinant, METH_O, "\n"
		":rtype: float\n"
		":return: The determinant of the matrix\n"
		""},
	 { "matrix3x3_isSymmetric", _wrap_matrix3x3_isSymmetric, METH_O, "Checks if a matrix is symmetric"},
	 { "matrix3x3_isOrthogonal", _wrap_matrix3x3_isOrthogonal, METH_O, "\n"
		"Checks if a matrix is orthogonal This method checks if a matrix is orthogonal, i.e.\n"
		"       if all column vectors are normalized and\n"
		"       are mutually orthogonal. A matrix is orthogonal if,\n"
		"       and only if the transformation it describes is orthonormal.\n"
		"       An orthonormal transformation is a\n"
		"       transformation that preserves length and angle.\n"
		"\n"
		"       The check is performed using the method isUnitMatrix() to\n"
		"       check if\n"
		"\n"
		"       .. code-block:: c++\n"
		"\n"
		"   this * transpose()\n"
		"       is a unit matrix. The criterion is therefore numerically quite\n"
		"       tight.\n"
		""},
	 { "matrix3x3_isDiagonal", _wrap_matrix3x3_isDiagonal, METH_O, "\n"
		":rtype: boolean\n"
		":return: if a matrix is diagonal\n"
		""},
	 { "matrix3x3_isUnitMatrix", _wrap_matrix3x3_isUnitMatrix, METH_O, "\n"
		":rtype: boolean\n"
		":return: if a matrix is the unit matrix\n"
		""},
	 { "matrix3x3_Get", _wrap_matrix3x3_Get, METH_VARARGS, "\n"
		" Access function Warning: row or column are not in the range 0..2, zero is returned\n"
		"Deprecated: use the constant operator() instead\n"
		""},
	 { "matrix3x3_Set", _wrap_matrix3x3_Set, METH_VARARGS, "\n"
		" Access function Warning: if row or column are not in the range 0..2, nothing will happen\n"
		"Deprecated: use the non-constant operator() instead\n"
		""},
	 { "matrix3x3_SetColumn", _wrap_matrix3x3_SetColumn, METH_VARARGS, "\n"
		"Access function Warning: If column is not in the range 0..2, the vector\n"
		"       remains unchanged and an exception is thrown.\n"
		""},
	 { "matrix3x3_SetRow", _wrap_matrix3x3_SetRow, METH_VARARGS, "\n"
		"Access function Warning: If column is not in the range 0..2, the vector\n"
		"       remains unchanged and an exception is thrown.\n"
		""},
	 { "matrix3x3_GetColumn", _wrap_matrix3x3_GetColumn, METH_VARARGS, "\n"
		"Access function Warning: If col is not in the range 0..2, an exception is\n"
		"       thrown.\n"
		""},
	 { "matrix3x3_GetRow", _wrap_matrix3x3_GetRow, METH_VARARGS, "\n"
		"Access function Warning: If row is not in the range 0..2, an exception is\n"
		"       thrown.\n"
		""},
	 { "matrix3x3___imul__", _wrap_matrix3x3___imul__, METH_VARARGS, "Multiplies all entries of the matrix by a scalar c"},
	 { "matrix3x3___itruediv__", _wrap_matrix3x3___itruediv__, METH_VARARGS, "Divides all entries of the matrix by a scalar c"},
	 { "matrix3x3_SetupRotMat", _wrap_matrix3x3_SetupRotMat, METH_VARARGS, "Calculate a rotation matrix for rotation about the x, y, and z axes by the angles specified (in degrees)"},
	 { "matrix3x3_PlaneReflection", _wrap_matrix3x3_PlaneReflection, METH_VARARGS, "Calculates a matrix that represents reflection on a plane"},
	 { "matrix3x3_RotAboutAxisByAngle", _wrap_matrix3x3_RotAboutAxisByAngle, METH_VARARGS, "Calculates a rotation matrix, rotating around the specified axis by the specified angle (in degrees)"},
	 { "matrix3x3_FillOrth", _wrap_matrix3x3_FillOrth, METH_VARARGS, "Calculate an orthogonalisation matrix for a unit cell specified by the parameters alpha, beta, gamma, a, b, c where alpha, beta, and gamma are the cell angles (in degrees) and a, b, and c are the cell vector lengths Used by OBUnitCell"},
	 { "matrix3x3_findEigenvectorsIfSymmetric", _wrap_matrix3x3_findEigenvectorsIfSymmetric, METH_VARARGS, "Find the eigenvalues and -vectors of a symmetric matrix"},
	 { "__mul__", _wrap___mul__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Multiplication with a scalar\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Multiplication with a scalar\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Multiplication of matrix and vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Matrix-matrix multiplication\n"
		""},
	 { "matrix3x3_jacobi", _wrap_matrix3x3_jacobi, METH_VARARGS, "Eigenvalue calculation"},
	 { "matrix3x3_swigregister", matrix3x3_swigregister, METH_O, NULL},
	 { "matrix3x3_swiginit", matrix3x3_swiginit, METH_VARARGS, NULL},
	 { "new_transform3d", _wrap_new_transform3d, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructs a matrix from row vectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructs a matrix from a 3x3-array of doubles The first index represents the row, the second index the column\n"
		""},
	 { "transform3d___mul__", _wrap_transform3d___mul__, METH_VARARGS, NULL},
	 { "transform3d_DescribeAsString", _wrap_transform3d_DescribeAsString, METH_O, NULL},
	 { "transform3d_DescribeAsValues", _wrap_transform3d_DescribeAsValues, METH_O, NULL},
	 { "transform3d_Normalize", _wrap_transform3d_Normalize, METH_O, NULL},
	 { "delete_transform3d", _wrap_delete_transform3d, METH_O, NULL},
	 { "transform3d_swigregister", transform3d_swigregister, METH_O, NULL},
	 { "transform3d_swiginit", transform3d_swiginit, METH_VARARGS, NULL},
	 { "new_SpaceGroup", _wrap_new_SpaceGroup, METH_NOARGS, NULL},
	 { "delete_SpaceGroup", _wrap_delete_SpaceGroup, METH_O, NULL},
	 { "SpaceGroup_SetHMName", _wrap_SpaceGroup_SetHMName, METH_VARARGS, NULL},
	 { "SpaceGroup_SetHallName", _wrap_SpaceGroup_SetHallName, METH_VARARGS, NULL},
	 { "SpaceGroup_SetId", _wrap_SpaceGroup_SetId, METH_VARARGS, NULL},
	 { "SpaceGroup_AddTransform", _wrap_SpaceGroup_AddTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_GetHMName", _wrap_SpaceGroup_GetHMName, METH_O, NULL},
	 { "SpaceGroup_GetHallName", _wrap_SpaceGroup_GetHallName, METH_O, NULL},
	 { "SpaceGroup_GetId", _wrap_SpaceGroup_GetId, METH_O, NULL},
	 { "SpaceGroup_GetOriginAlternative", _wrap_SpaceGroup_GetOriginAlternative, METH_O, NULL},
	 { "SpaceGroup_Transform", _wrap_SpaceGroup_Transform, METH_VARARGS, NULL},
	 { "SpaceGroup_BeginTransform", _wrap_SpaceGroup_BeginTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_NextTransform", _wrap_SpaceGroup_NextTransform, METH_VARARGS, NULL},
	 { "SpaceGroup_GetSpaceGroup", _wrap_SpaceGroup_GetSpaceGroup, METH_VARARGS, NULL},
	 { "SpaceGroup_Find", _wrap_SpaceGroup_Find, METH_O, NULL},
	 { "SpaceGroup___eq__", _wrap_SpaceGroup___eq__, METH_VARARGS, NULL},
	 { "SpaceGroup___ne__", _wrap_SpaceGroup___ne__, METH_VARARGS, NULL},
	 { "SpaceGroup_IsValid", _wrap_SpaceGroup_IsValid, METH_O, NULL},
	 { "SpaceGroup_HEXAGONAL_ORIGIN_get", _wrap_SpaceGroup_HEXAGONAL_ORIGIN_get, METH_O, NULL},
	 { "SpaceGroup_swigregister", SpaceGroup_swigregister, METH_O, NULL},
	 { "SpaceGroup_swiginit", SpaceGroup_swiginit, METH_VARARGS, NULL},
	 { "new_OBBitVec", _wrap_new_OBBitVec, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Construct a bit vector of the default size Construct a bit vector of STARTWORDS size,\n"
		"	      cleared to all zero bits.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Construct a bit vector of maxbits bits Construct a bit vector with a size in bits\n"
		"	      of ``size_in_bits`` rounded up to the nearest word\n"
		"		  and cleared to all zero bits.\n"
		"		  :type size_in_bits: int, in\n"
		":param size_in_bits: The number of bits for which to reserve space\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Copy constructor (result has same number of bits) Construct a bit vector which is an exact\n"
		"	      duplicate of ``bv``.\n"
		"		  :type bv: :py:class:`OBBitVec`, in\n"
		":param bv: The other bit vector to copy to this\n"
		""},
	 { "OBBitVec_SetBitOn", _wrap_OBBitVec_SetBitOn, METH_VARARGS, "Set the ``bit_offset`` 'th bit to 1"},
	 { "OBBitVec_SetBitOff", _wrap_OBBitVec_SetBitOff, METH_VARARGS, "Set the ``bit_offset`` 'th bit to 0"},
	 { "OBBitVec_SetRangeOn", _wrap_OBBitVec_SetRangeOn, METH_VARARGS, "Set the range of bits from ``lo_bit_offset`` to ``hi_bit_offset`` to 1"},
	 { "OBBitVec_SetRangeOff", _wrap_OBBitVec_SetRangeOff, METH_VARARGS, "Set the range of bits from ``lo_bit_offset`` to ``hi_bit_offset`` to 0"},
	 { "OBBitVec_Fold", _wrap_OBBitVec_Fold, METH_VARARGS, "Reduce the size of the vector by or-ing the excess bits over the start"},
	 { "OBBitVec_FirstBit", _wrap_OBBitVec_FirstBit, METH_VARARGS, "\n"
		" Find the first true bit at or after ``bit_offset`` Searches the vector for the first true value, starting at the ``bit_offset`` 'th bit\n"
		"          :type bit_offset: int, in, optional\n"
		"          :param bit_offset: the first bit to consider\n"
		"    		  :rtype: int\n"
		":return: the bit offset of the first true bit at or after ``bit_offset``, or -1 if there is none\n"
		""},
	 { "OBBitVec_NextBit", _wrap_OBBitVec_NextBit, METH_VARARGS, "Find the next true bit after ``last_bit_offset``"},
	 { "OBBitVec_EndBit", _wrap_OBBitVec_EndBit, METH_O, "Return the bit offset of the last bit (for iterating) i.e. -1"},
	 { "OBBitVec_GetSize", _wrap_OBBitVec_GetSize, METH_O, "Return the number of words ( NOT the number of bits )."},
	 { "OBBitVec_CountBits", _wrap_OBBitVec_CountBits, METH_O, "Return the number of bits which are set to 1 in the vector"},
	 { "OBBitVec_IsEmpty", _wrap_OBBitVec_IsEmpty, METH_O, "Are there no bits set to 1 in this vector?"},
	 { "OBBitVec_Resize", _wrap_OBBitVec_Resize, METH_VARARGS, "\n"
		" Reserve space for ``size_in_bits`` bits Reserve space for ``size_in_bits`` bits rounded up\n"
		"	      :type size_in_bits: int, in\n"
		":param size_in_bits: the number of bits\n"
		"    	  :rtype: boolean\n"
		":return: true if enlargement was necessary, false otherwise\n"
		""},
	 { "OBBitVec_ResizeWords", _wrap_OBBitVec_ResizeWords, METH_VARARGS, "\n"
		" Reserve space for ``size_in_words`` words Reserve space for ``size_in_words`` words\n"
		"	      :type size_in_words: int, in\n"
		":param size_in_words: the number of words\n"
		"    	  :rtype: boolean\n"
		":return: true if enlargement was necessary, false otherwise\n"
		""},
	 { "OBBitVec_BitIsSet", _wrap_OBBitVec_BitIsSet, METH_VARARGS, "\n"
		" Asks if the ``bit_offset`` 'th bit is set Is the ``bit_offset`` 'th bit set ?\n"
		"          :type bit_offset: int, in\n"
		"          :param bit_offset: a zero based offset into the bit vector\n"
		"    		  :rtype: boolean\n"
		":return: true if it is set, false otherwise\n"
		""},
	 { "OBBitVec_FromVecInt", _wrap_OBBitVec_FromVecInt, METH_VARARGS, "Sets the bits listed as bit offsets"},
	 { "OBBitVec_FromString", _wrap_OBBitVec_FromString, METH_VARARGS, "Sets the bits listed as a string of integers"},
	 { "OBBitVec_ToVecInt", _wrap_OBBitVec_ToVecInt, METH_VARARGS, "List the offsets of the bits which are set"},
	 { "OBBitVec_Clear", _wrap_OBBitVec_Clear, METH_O, "Set all bits to zero"},
	 { "OBBitVec_Negate", _wrap_OBBitVec_Negate, METH_O, "\n"
		"Inverts every bit in the vector Inverts the entire vector.\n"
		"      Note that this may give unexpected results, as the vector\n"
		"	  can be considered to end in an arbitrary number of zero bits.\n"
		""},
	 { "OBBitVec_GetWords", _wrap_OBBitVec_GetWords, METH_VARARGS, "\n"
		" Return a copy of the internal vector of words, at the end of ``vec`` Copy the internal word vector.\n"
		"	      The copy is appended to ``vec``.\n"
		"		  :type vec: OpenBabel::OBBitVec::word_vector, out\n"
		":param vec: a vector of words to which to append the data\n"
		""},
	 { "OBBitVec___iand__", _wrap_OBBitVec___iand__, METH_VARARGS, "And-equals operator"},
	 { "OBBitVec___ior__", _wrap_OBBitVec___ior__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Or-equals operator\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Or-equals operator for integer Or the bit at offset ``bit_offset`` with 1\n"
		""},
	 { "OBBitVec___ixor__", _wrap_OBBitVec___ixor__, METH_VARARGS, "Exclusive-or-equals operator"},
	 { "OBBitVec___isub__", _wrap_OBBitVec___isub__, METH_VARARGS, "Minus-equals operator"},
	 { "OBBitVec___iadd__", _wrap_OBBitVec___iadd__, METH_VARARGS, "Plus-equals operator"},
	 { "__or__", _wrap___or__, METH_VARARGS, "Or operator"},
	 { "__and__", _wrap___and__, METH_VARARGS, "And operator"},
	 { "__xor__", _wrap___xor__, METH_VARARGS, "Exclusive-or operator"},
	 { "__sub__", _wrap___sub__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Vector subtraction\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Unary minus\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Minus operator\n"
		""},
	 { "__eq__", _wrap___eq__, METH_VARARGS, "Equivalency operator"},
	 { "__lt__", _wrap___lt__, METH_VARARGS, "Smaller-than operator"},
	 { "__rshift__", _wrap___rshift__, METH_VARARGS, "Input from a stream"},
	 { "delete_OBBitVec", _wrap_delete_OBBitVec, METH_O, NULL},
	 { "OBBitVec_swigregister", OBBitVec_swigregister, METH_O, NULL},
	 { "OBBitVec_swiginit", OBBitVec_swiginit, METH_VARARGS, NULL},
	 { "Tanimoto", _wrap_Tanimoto, METH_VARARGS, "The Tanimoto coefficient, which may be regarded as the proportion of the \"on-bits\" which are shared."},
	 { "OBReleaseVersion", _wrap_OBReleaseVersion, METH_NOARGS, "\n"
		":rtype: string\n"
		":return: the version of the Open Babel library for feature-detection (e.g. \"2.3.1\")\n"
		""},
	 { "new_OBGenericData", _wrap_new_OBGenericData, METH_VARARGS, NULL},
	 { "OBGenericData_Clone", _wrap_OBGenericData_Clone, METH_VARARGS, NULL},
	 { "delete_OBGenericData", _wrap_delete_OBGenericData, METH_O, NULL},
	 { "OBGenericData_SetAttribute", _wrap_OBGenericData_SetAttribute, METH_VARARGS, "Set the attribute (key), which can be used to retrieve this data"},
	 { "OBGenericData_SetOrigin", _wrap_OBGenericData_SetOrigin, METH_VARARGS, "Set the origin of this data, which can be used to filter the data"},
	 { "OBGenericData_GetAttribute", _wrap_OBGenericData_GetAttribute, METH_O, "\n"
		":rtype: string\n"
		":return: The attribute (key), which can be used to retrieve this data\n"
		""},
	 { "OBGenericData_GetDataType", _wrap_OBGenericData_GetDataType, METH_O, "\n"
		":rtype: int\n"
		":return: the data type for this object as defined in OBGenericDataType\n"
		""},
	 { "OBGenericData_GetValue", _wrap_OBGenericData_GetValue, METH_O, "Base class returns a default value (the attribute type) but should never be called"},
	 { "OBGenericData_GetOrigin", _wrap_OBGenericData_GetOrigin, METH_O, NULL},
	 { "OBGenericData_swigregister", OBGenericData_swigregister, METH_O, NULL},
	 { "OBGenericData_swiginit", OBGenericData_swiginit, METH_VARARGS, NULL},
	 { "delete_OBBase", _wrap_delete_OBBase, METH_O, NULL},
	 { "OBBase_Clear", _wrap_OBBase_Clear, METH_O, "Clear any and all data associated with this object"},
	 { "OBBase_DoTransformations", _wrap_OBBase_DoTransformations, METH_VARARGS, "Perform a set of transformations specified by the user Typically these are program options to filter or modify an object For example, see OBMol::DoTransformations() and OBMol::ClassDescription() Base type does nothing"},
	 { "OBBase_ClassDescription", _wrap_OBBase_ClassDescription, METH_NOARGS, "\n"
		":rtype: string\n"
		":return: A list of descriptions of command-line options for DoTransformations()\n"
		""},
	 { "OBBase_GetTitle", _wrap_OBBase_GetTitle, METH_VARARGS, "Base type does nothing Made virtual around r3535 to simplify code which passes around OBBase*."},
	 { "OBBase_SetTitle", _wrap_OBBase_SetTitle, METH_VARARGS, NULL},
	 { "OBBase_HasData", _wrap_OBBase_HasData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: boolean\n"
		":return: whether the generic attribute/value pair exists, for a given OBGenericDataType\n"
		""},
	 { "OBBase_DeleteData", _wrap_OBBase_DeleteData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Delete any data matching the given OBGenericDataType\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Delete the given generic data from this object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Delete all of the given generic data from this object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Deletes the generic data with the specified attribute, returning false if not found\n"
		""},
	 { "OBBase_CloneData", _wrap_OBBase_CloneData, METH_VARARGS, "Adds a copy of a data object; does nothing if d == NULL Since: version 2.2"},
	 { "OBBase_DataSize", _wrap_OBBase_DataSize, METH_O, "\n"
		":rtype: int\n"
		":return: the number of OBGenericData items attached to this molecule.\n"
		""},
	 { "OBBase_GetAllData", _wrap_OBBase_GetAllData, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: the all matching data for a given type from OBGenericDataType    or an empty vector if nothing matches Since: version 2.2\n"
		""},
	 { "OBBase_GetData", _wrap_OBBase_GetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the first matching data for a given type from OBGenericDataType    or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: any data matching the given attribute name or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: any data matching the given attribute name or NULL if nothing matches\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: all data, suitable for iterating\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		":rtype: std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >\n"
		":return: all data with a specific origin, suitable for iterating\n"
		""},
	 { "OBBase_BeginData", _wrap_OBBase_BeginData, METH_O, "\n"
		":rtype: OpenBabel::OBDataIterator\n"
		":return: An iterator pointing to the beginning of the data\n"
		""},
	 { "OBBase_EndData", _wrap_OBBase_EndData, METH_O, "\n"
		":rtype: OpenBabel::OBDataIterator\n"
		":return: An iterator pointing to the end of the data\n"
		""},
	 { "new_OBBase", _wrap_new_OBBase, METH_NOARGS, NULL},
	 { "OBBase_swigregister", OBBase_swigregister, METH_O, NULL},
	 { "OBBase_swiginit", OBBase_swiginit, METH_VARARGS, NULL},
	 { "new_OBCommentData", _wrap_new_OBCommentData, METH_VARARGS, NULL},
	 { "OBCommentData_SetData", _wrap_OBCommentData_SetData, METH_VARARGS, NULL},
	 { "OBCommentData_GetData", _wrap_OBCommentData_GetData, METH_O, NULL},
	 { "delete_OBCommentData", _wrap_delete_OBCommentData, METH_O, NULL},
	 { "OBCommentData_swigregister", OBCommentData_swigregister, METH_O, NULL},
	 { "OBCommentData_swiginit", OBCommentData_swiginit, METH_VARARGS, NULL},
	 { "new_OBExternalBond", _wrap_new_OBExternalBond, METH_VARARGS, NULL},
	 { "delete_OBExternalBond", _wrap_delete_OBExternalBond, METH_O, NULL},
	 { "OBExternalBond_GetIdx", _wrap_OBExternalBond_GetIdx, METH_O, NULL},
	 { "OBExternalBond_GetAtom", _wrap_OBExternalBond_GetAtom, METH_O, NULL},
	 { "OBExternalBond_GetBond", _wrap_OBExternalBond_GetBond, METH_O, NULL},
	 { "OBExternalBond_SetIdx", _wrap_OBExternalBond_SetIdx, METH_VARARGS, NULL},
	 { "OBExternalBond_SetAtom", _wrap_OBExternalBond_SetAtom, METH_VARARGS, NULL},
	 { "OBExternalBond_SetBond", _wrap_OBExternalBond_SetBond, METH_VARARGS, NULL},
	 { "OBExternalBond_swigregister", OBExternalBond_swigregister, METH_O, NULL},
	 { "OBExternalBond_swiginit", OBExternalBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBExternalBondData", _wrap_new_OBExternalBondData, METH_NOARGS, NULL},
	 { "OBExternalBondData_SetData", _wrap_OBExternalBondData_SetData, METH_VARARGS, NULL},
	 { "OBExternalBondData_GetData", _wrap_OBExternalBondData_GetData, METH_O, NULL},
	 { "delete_OBExternalBondData", _wrap_delete_OBExternalBondData, METH_O, NULL},
	 { "OBExternalBondData_swigregister", OBExternalBondData_swigregister, METH_O, NULL},
	 { "OBExternalBondData_swiginit", OBExternalBondData_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairData", _wrap_new_OBPairData, METH_NOARGS, NULL},
	 { "OBPairData_SetValue", _wrap_OBPairData_SetValue, METH_VARARGS, NULL},
	 { "delete_OBPairData", _wrap_delete_OBPairData, METH_O, NULL},
	 { "OBPairData_swigregister", OBPairData_swigregister, METH_O, NULL},
	 { "OBPairData_swiginit", OBPairData_swiginit, METH_VARARGS, NULL},
	 { "new_OBSetData", _wrap_new_OBSetData, METH_NOARGS, NULL},
	 { "OBSetData_AddData", _wrap_OBSetData_AddData, METH_VARARGS, "Add an OBGenericData element to the set."},
	 { "OBSetData_SetData", _wrap_OBSetData_SetData, METH_VARARGS, "Set the array of data to a new vector"},
	 { "OBSetData_GetData", _wrap_OBSetData_GetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the OBGenericData associate with the attribute name parameter.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBGenericData`\n"
		":return: the OBGenericData associate with the attribute name parameter.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Gets the entire set.\n"
		""},
	 { "OBSetData_GetBegin", _wrap_OBSetData_GetBegin, METH_O, "Get the begin iterator."},
	 { "OBSetData_GetEnd", _wrap_OBSetData_GetEnd, METH_O, "Get the end iterator."},
	 { "OBSetData_DeleteData", _wrap_OBSetData_DeleteData, METH_VARARGS, "Delete the matching OBGenericData element."},
	 { "delete_OBSetData", _wrap_delete_OBSetData, METH_O, NULL},
	 { "OBSetData_swigregister", OBSetData_swigregister, METH_O, NULL},
	 { "OBSetData_swiginit", OBSetData_swiginit, METH_VARARGS, NULL},
	 { "new_OBVirtualBond", _wrap_new_OBVirtualBond, METH_VARARGS, NULL},
	 { "OBVirtualBond_GetBgn", _wrap_OBVirtualBond_GetBgn, METH_O, NULL},
	 { "OBVirtualBond_GetEnd", _wrap_OBVirtualBond_GetEnd, METH_O, NULL},
	 { "OBVirtualBond_GetOrder", _wrap_OBVirtualBond_GetOrder, METH_O, NULL},
	 { "OBVirtualBond_GetStereo", _wrap_OBVirtualBond_GetStereo, METH_O, NULL},
	 { "delete_OBVirtualBond", _wrap_delete_OBVirtualBond, METH_O, NULL},
	 { "OBVirtualBond_swigregister", OBVirtualBond_swigregister, METH_O, NULL},
	 { "OBVirtualBond_swiginit", OBVirtualBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBRingData", _wrap_new_OBRingData, METH_VARARGS, NULL},
	 { "delete_OBRingData", _wrap_delete_OBRingData, METH_O, NULL},
	 { "OBRingData_SetData", _wrap_OBRingData_SetData, METH_VARARGS, NULL},
	 { "OBRingData_PushBack", _wrap_OBRingData_PushBack, METH_VARARGS, NULL},
	 { "OBRingData_GetData", _wrap_OBRingData_GetData, METH_O, NULL},
	 { "OBRingData_BeginRings", _wrap_OBRingData_BeginRings, METH_O, NULL},
	 { "OBRingData_EndRings", _wrap_OBRingData_EndRings, METH_O, NULL},
	 { "OBRingData_BeginRing", _wrap_OBRingData_BeginRing, METH_VARARGS, NULL},
	 { "OBRingData_NextRing", _wrap_OBRingData_NextRing, METH_VARARGS, NULL},
	 { "OBRingData_swigregister", OBRingData_swigregister, METH_O, NULL},
	 { "OBRingData_swiginit", OBRingData_swiginit, METH_VARARGS, NULL},
	 { "new_OBUnitCell", _wrap_new_OBUnitCell, METH_VARARGS, "public constructor"},
	 { "delete_OBUnitCell", _wrap_delete_OBUnitCell, METH_O, NULL},
	 { "OBUnitCell_SetData", _wrap_OBUnitCell_SetData, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Constructs the cell matrix in lower triangular form from the values supplied\n"
		":type a: float\n"
		":param a: The length a\n"
		":type b: float\n"
		":param b: The length b\n"
		":type c: float\n"
		":param c: The length c\n"
		":type alpha: float\n"
		":param alpha: The angle alpha\n"
		":type beta: float\n"
		":param beta: The angle beta\n"
		":type gamma: float\n"
		":param gamma: The angle gamma\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Constructs the cell matrix using the supplied row vectors\n"
		":type v1: :py:class:`vector3`\n"
		":param v1: The x-vector\n"
		":type v2: :py:class:`vector3`\n"
		":param v2: The y-vector\n"
		":type v3: :py:class:`vector3`\n"
		":param v3: The z-vector\n"
		"See also: OBUnitCell::GetCellVectors\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Sets the unit cell matrix\n"
		":type m: :py:class:`matrix3x3`\n"
		":param m: The unit cell matrix (row vectors)\n"
		"See also: OBUnitCell::GetCellMatrix\n"
		""},
	 { "OBUnitCell_SetOffset", _wrap_OBUnitCell_SetOffset, METH_VARARGS, "Set the offset to the origin to ``v1``"},
	 { "OBUnitCell_SetSpaceGroup", _wrap_OBUnitCell_SetSpaceGroup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the space group for this unit cell. Does not create an OBSymmetryData entry\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the space group symbol for this unit cell. Does not create an OBSymmetryData entry or attempt to convert  between different symbol notations\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Set the space group for this unit cell. Each spacegroup-symbol has a numeric equivalent which is easier to use to convert between notations. Does not create an OBSymmetryData entry or attempt to convert  between different symbol notations\n"
		""},
	 { "OBUnitCell_SetLatticeType", _wrap_OBUnitCell_SetLatticeType, METH_VARARGS, "Set the Bravais lattice type for this unit cell"},
	 { "OBUnitCell_FillUnitCell", _wrap_OBUnitCell_FillUnitCell, METH_VARARGS, "Duplicate symmetry-unique atoms to fill out the unit cell of the molecule, based on the known space group"},
	 { "OBUnitCell_GetA", _wrap_OBUnitCell_GetA, METH_VARARGS, "\n"
		" TODO: Remove nonconst overloads in OBUnitCell on next version bump :rtype: float\n"
		":return: vector a\n"
		""},
	 { "OBUnitCell_GetB", _wrap_OBUnitCell_GetB, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: vector b\n"
		""},
	 { "OBUnitCell_GetC", _wrap_OBUnitCell_GetC, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: vector c\n"
		""},
	 { "OBUnitCell_GetAlpha", _wrap_OBUnitCell_GetAlpha, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle alpha\n"
		""},
	 { "OBUnitCell_GetBeta", _wrap_OBUnitCell_GetBeta, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle beta\n"
		""},
	 { "OBUnitCell_GetGamma", _wrap_OBUnitCell_GetGamma, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: angle gamma\n"
		""},
	 { "OBUnitCell_GetOffset", _wrap_OBUnitCell_GetOffset, METH_VARARGS, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: any offset in the origin of the periodic boundaries\n"
		""},
	 { "OBUnitCell_GetSpaceGroup", _wrap_OBUnitCell_GetSpaceGroup, METH_VARARGS, "\n"
		":rtype: :py:class:`SpaceGroup`\n"
		":return: the text representation of the space group for this unit cell\n"
		""},
	 { "OBUnitCell_GetSpaceGroupName", _wrap_OBUnitCell_GetSpaceGroupName, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the text representation of the space group for this unit cell\n"
		""},
	 { "OBUnitCell_GetLatticeType", _wrap_OBUnitCell_GetLatticeType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: int\n"
		":return: lattice type (based on the ``spacegroup)``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: lattice type (based on angles and distances)\n"
		""},
	 { "OBUnitCell_GetCellVectors", _wrap_OBUnitCell_GetCellVectors, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >\n"
		":return: v1, v2, v3 cell vectors\n"
		""},
	 { "OBUnitCell_GetCellMatrix", _wrap_OBUnitCell_GetCellMatrix, METH_VARARGS, "\n"
		" Access to the cell matrix as row vectors, useful for writing input files. Equivalent to the transpose of GetOrientationMatrix() * GetOrthoMatrix() :rtype: :py:class:`matrix3x3`\n"
		":return: The cell matrix with row vectors. See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetOrthoMatrix", _wrap_OBUnitCell_GetOrthoMatrix, METH_VARARGS, "\n"
		":rtype: :py:class:`matrix3x3`\n"
		":return: The orthogonalization matrix, used for converting from fractional to Cartesian coords. See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetOrientationMatrix", _wrap_OBUnitCell_GetOrientationMatrix, METH_VARARGS, "\n"
		" Used to convert fractional and cartesian coordinates if the cell is not oriented in standard form (a parallel to x axis, b in xy plane) :rtype: :py:class:`matrix3x3`\n"
		":return: The orientation matrix See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetFractionalMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_GetFractionalMatrix", _wrap_OBUnitCell_GetFractionalMatrix, METH_VARARGS, "\n"
		":rtype: :py:class:`matrix3x3`\n"
		":return: The fractionalization matrix, used for converting from Cartesian to fractional coords. See also: OBUnitCell::GetOrthoMatrix See also: OBUnitCell::GetCellMatrix See also: OBUnitCell::GetOrientationMatrix See also: OBUnitCell::FractionalToCartesian See also: OBUnitCell::CartesianToFractional\n"
		""},
	 { "OBUnitCell_FractionalToCartesian", _wrap_OBUnitCell_FractionalToCartesian, METH_VARARGS, "\n"
		" Convenience function to convert fractional coordinates to cartesian coordinates. Returns GetOrientationMatrix() * GetOrthoMatrix() * frac + GetOffset() :type frac: :py:class:`vector3`\n"
		":param frac: Vector of fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian coordinates\n"
		""},
	 { "OBUnitCell_CartesianToFractional", _wrap_OBUnitCell_CartesianToFractional, METH_VARARGS, "\n"
		" Convenience function to convert cartesian coordinates to fractional coordinates. Returns GetFractionalMatrix() * GetOrientationMatrix().inverse() * (cart - GetOffset()) :type cart: :py:class:`vector3`\n"
		":param cart: Vector of cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional coordinates\n"
		""},
	 { "OBUnitCell_WrapCartesianCoordinate", _wrap_OBUnitCell_WrapCartesianCoordinate, METH_VARARGS, "\n"
		" Wraps cartesian coordinate to fall within the unit cell. :type cart: :py:class:`vector3`\n"
		":param cart: Vector of cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian coordinates within cell boundaries.\n"
		""},
	 { "OBUnitCell_WrapFractionalCoordinate", _wrap_OBUnitCell_WrapFractionalCoordinate, METH_VARARGS, "\n"
		" Wraps fractional coordinate to fall within the unit cell. :type frac: :py:class:`vector3`\n"
		":param frac: Vector of fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional coordinates within cell boundaries (between 0 and 1). TODO: Make OBUnitCell::WrapFractionalCoordinate static in the next ABI break\n"
		""},
	 { "OBUnitCell_UnwrapCartesianNear", _wrap_OBUnitCell_UnwrapCartesianNear, METH_VARARGS, "\n"
		" Unwraps Cartesian coordinates near a reference location :type new_loc: :py:class:`vector3`\n"
		":param new_loc: Cartesian coordinates of target :type ref_loc: :py:class:`vector3`\n"
		":param ref_loc: Cartesian coordinates of the reference location :rtype: :py:class:`vector3`\n"
		":return: Unwrapped coordinates of new_loc near ref_loc\n"
		""},
	 { "OBUnitCell_UnwrapFractionalNear", _wrap_OBUnitCell_UnwrapFractionalNear, METH_VARARGS, "\n"
		" Unwraps fractional coordinates near a reference location. :type new_loc: :py:class:`vector3`\n"
		":param new_loc: Fractional coordinates of target :type ref_loc: :py:class:`vector3`\n"
		":param ref_loc: Fractional coordinates of the reference location :rtype: :py:class:`vector3`\n"
		":return: Unwrapped coordinates of new_loc near ref_loc TODO: Add a simple test case/example, like unwrapNear(<0.9, 0.2, 0.2>, <0.3, 0.9, 0.2>) -> <-0.1, 1.2, 0.2>\n"
		""},
	 { "OBUnitCell_MinimumImageCartesian", _wrap_OBUnitCell_MinimumImageCartesian, METH_VARARGS, "\n"
		" Applies the minimum image convention to a Cartesian displacement vector :type cart: :py:class:`vector3`\n"
		":param cart: Displacement vector between two atoms in Cartesian coordinates :rtype: :py:class:`vector3`\n"
		":return: Cartesian difference, wrapped within half the unit cell\n"
		""},
	 { "OBUnitCell_MinimumImageFractional", _wrap_OBUnitCell_MinimumImageFractional, METH_VARARGS, "\n"
		" Applies the minimum image convention to a fractional displacement vector :param cart: Displacement vector between two atoms in fractional coordinates :rtype: :py:class:`vector3`\n"
		":return: Fractional difference, wrapped within half the unit cell (-0.5 to 0.5)\n"
		""},
	 { "OBUnitCell_GetSpaceGroupNumber", _wrap_OBUnitCell_GetSpaceGroupNumber, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The numeric value of the given spacegroup\n"
		""},
	 { "OBUnitCell_GetCellVolume", _wrap_OBUnitCell_GetCellVolume, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: The cell volume (in Angstroms^3)\n"
		""},
	 { "OBUnitCell_swigregister", OBUnitCell_swigregister, METH_O, NULL},
	 { "OBUnitCell_swiginit", OBUnitCell_swiginit, METH_VARARGS, NULL},
	 { "new_OBConformerData", _wrap_new_OBConformerData, METH_VARARGS, NULL},
	 { "delete_OBConformerData", _wrap_delete_OBConformerData, METH_O, NULL},
	 { "OBConformerData_SetDimension", _wrap_OBConformerData_SetDimension, METH_VARARGS, NULL},
	 { "OBConformerData_SetEnergies", _wrap_OBConformerData_SetEnergies, METH_VARARGS, NULL},
	 { "OBConformerData_SetForces", _wrap_OBConformerData_SetForces, METH_VARARGS, NULL},
	 { "OBConformerData_SetVelocities", _wrap_OBConformerData_SetVelocities, METH_VARARGS, NULL},
	 { "OBConformerData_SetDisplacements", _wrap_OBConformerData_SetDisplacements, METH_VARARGS, NULL},
	 { "OBConformerData_SetData", _wrap_OBConformerData_SetData, METH_VARARGS, NULL},
	 { "OBConformerData_GetDimension", _wrap_OBConformerData_GetDimension, METH_O, NULL},
	 { "OBConformerData_GetEnergies", _wrap_OBConformerData_GetEnergies, METH_O, NULL},
	 { "OBConformerData_GetForces", _wrap_OBConformerData_GetForces, METH_O, NULL},
	 { "OBConformerData_GetVelocities", _wrap_OBConformerData_GetVelocities, METH_O, NULL},
	 { "OBConformerData_GetDisplacements", _wrap_OBConformerData_GetDisplacements, METH_O, NULL},
	 { "OBConformerData_GetData", _wrap_OBConformerData_GetData, METH_O, NULL},
	 { "OBConformerData_swigregister", OBConformerData_swigregister, METH_O, NULL},
	 { "OBConformerData_swiginit", OBConformerData_swiginit, METH_VARARGS, NULL},
	 { "new_OBSymmetryData", _wrap_new_OBSymmetryData, METH_VARARGS, NULL},
	 { "delete_OBSymmetryData", _wrap_delete_OBSymmetryData, METH_O, NULL},
	 { "OBSymmetryData_SetData", _wrap_OBSymmetryData_SetData, METH_VARARGS, NULL},
	 { "OBSymmetryData_SetPointGroup", _wrap_OBSymmetryData_SetPointGroup, METH_VARARGS, NULL},
	 { "OBSymmetryData_SetSpaceGroup", _wrap_OBSymmetryData_SetSpaceGroup, METH_VARARGS, NULL},
	 { "OBSymmetryData_GetPointGroup", _wrap_OBSymmetryData_GetPointGroup, METH_O, NULL},
	 { "OBSymmetryData_GetSpaceGroup", _wrap_OBSymmetryData_GetSpaceGroup, METH_O, NULL},
	 { "OBSymmetryData_swigregister", OBSymmetryData_swigregister, METH_O, NULL},
	 { "OBSymmetryData_swiginit", OBSymmetryData_swiginit, METH_VARARGS, NULL},
	 { "new_OBTorsion", _wrap_new_OBTorsion, METH_VARARGS, NULL},
	 { "delete_OBTorsion", _wrap_delete_OBTorsion, METH_O, NULL},
	 { "OBTorsion_Clear", _wrap_OBTorsion_Clear, METH_O, NULL},
	 { "OBTorsion_Empty", _wrap_OBTorsion_Empty, METH_O, NULL},
	 { "OBTorsion_AddTorsion", _wrap_OBTorsion_AddTorsion, METH_VARARGS, NULL},
	 { "OBTorsion_SetAngle", _wrap_OBTorsion_SetAngle, METH_VARARGS, NULL},
	 { "OBTorsion_SetData", _wrap_OBTorsion_SetData, METH_VARARGS, NULL},
	 { "OBTorsion_GetAngle", _wrap_OBTorsion_GetAngle, METH_VARARGS, NULL},
	 { "OBTorsion_GetBondIdx", _wrap_OBTorsion_GetBondIdx, METH_O, "\n"
		" Gets the bond index of the central bond :rtype: int\n"
		":return: int bond index\n"
		""},
	 { "OBTorsion_GetSize", _wrap_OBTorsion_GetSize, METH_O, NULL},
	 { "OBTorsion_GetBC", _wrap_OBTorsion_GetBC, METH_O, "\n"
		" Gets the two central atoms of ABCD torsion   :rtype: std::pair< OpenBabel::OBAtom *,OpenBabel::OBAtom * >\n"
		":return: pair<OBAtom*,OBAtom*>\n"
		""},
	 { "OBTorsion_GetADs", _wrap_OBTorsion_GetADs, METH_O, "\n"
		" Gets the vector of distal atoms of ABCD torsion :rtype: std::vector< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double >,std::allocator< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double > > >\n"
		":return: vector of A,D atom pointers and a double\n"
		""},
	 { "OBTorsion_IsProtonRotor", _wrap_OBTorsion_IsProtonRotor, METH_O, NULL},
	 { "OBTorsion_swigregister", OBTorsion_swigregister, METH_O, NULL},
	 { "OBTorsion_swiginit", OBTorsion_swiginit, METH_VARARGS, NULL},
	 { "OBTorsionData_Clear", _wrap_OBTorsionData_Clear, METH_O, NULL},
	 { "OBTorsionData_GetData", _wrap_OBTorsionData_GetData, METH_O, "\n"
		" Gets a vector of the OBTorsion objects :rtype: std::vector< OpenBabel::OBTorsion,std::allocator< OpenBabel::OBTorsion > >\n"
		":return: the vector of torsions\n"
		""},
	 { "OBTorsionData_GetSize", _wrap_OBTorsionData_GetSize, METH_O, "\n"
		" Gets the number of torsion structs :rtype: int\n"
		":return: integer count of the number of torsions\n"
		""},
	 { "OBTorsionData_SetData", _wrap_OBTorsionData_SetData, METH_VARARGS, NULL},
	 { "OBTorsionData_FillTorsionArray", _wrap_OBTorsionData_FillTorsionArray, METH_VARARGS, NULL},
	 { "delete_OBTorsionData", _wrap_delete_OBTorsionData, METH_O, NULL},
	 { "OBTorsionData_swigregister", OBTorsionData_swigregister, METH_O, NULL},
	 { "new_OBAngle", _wrap_new_OBAngle, METH_VARARGS, NULL},
	 { "delete_OBAngle", _wrap_delete_OBAngle, METH_O, NULL},
	 { "OBAngle___eq__", _wrap_OBAngle___eq__, METH_VARARGS, NULL},
	 { "OBAngle_Clear", _wrap_OBAngle_Clear, METH_O, NULL},
	 { "OBAngle_GetAngle", _wrap_OBAngle_GetAngle, METH_O, "\n"
		" Gets the OBAngle angle value :rtype: float\n"
		":return: angle in radians\n"
		""},
	 { "OBAngle_SetAngle", _wrap_OBAngle_SetAngle, METH_VARARGS, "\n"
		" Sets the OBAngle to ``radians`` :type angle: float\n"
		":param angle: in radians\n"
		""},
	 { "OBAngle_SetAtoms", _wrap_OBAngle_SetAtoms, METH_VARARGS, NULL},
	 { "OBAngle_swigregister", OBAngle_swigregister, METH_O, NULL},
	 { "OBAngle_swiginit", OBAngle_swiginit, METH_VARARGS, NULL},
	 { "OBAngleData_Clear", _wrap_OBAngleData_Clear, METH_O, NULL},
	 { "OBAngleData_FillAngleArray", _wrap_OBAngleData_FillAngleArray, METH_VARARGS, NULL},
	 { "OBAngleData_SetData", _wrap_OBAngleData_SetData, METH_VARARGS, NULL},
	 { "OBAngleData_GetSize", _wrap_OBAngleData_GetSize, METH_O, "\n"
		" Gets the number of angles stored :rtype: int\n"
		":return: integer count of the number of angles\n"
		""},
	 { "delete_OBAngleData", _wrap_delete_OBAngleData, METH_O, NULL},
	 { "OBAngleData_swigregister", OBAngleData_swigregister, METH_O, NULL},
	 { "new_OBSerialNums", _wrap_new_OBSerialNums, METH_VARARGS, NULL},
	 { "OBSerialNums_GetData", _wrap_OBSerialNums_GetData, METH_O, NULL},
	 { "OBSerialNums_SetData", _wrap_OBSerialNums_SetData, METH_VARARGS, NULL},
	 { "delete_OBSerialNums", _wrap_delete_OBSerialNums, METH_O, NULL},
	 { "OBSerialNums_swigregister", OBSerialNums_swigregister, METH_O, NULL},
	 { "OBSerialNums_swiginit", OBSerialNums_swiginit, METH_VARARGS, NULL},
	 { "new_OBVibrationData", _wrap_new_OBVibrationData, METH_NOARGS, NULL},
	 { "delete_OBVibrationData", _wrap_delete_OBVibrationData, METH_O, NULL},
	 { "OBVibrationData_SetData", _wrap_OBVibrationData_SetData, METH_VARARGS, NULL},
	 { "OBVibrationData_GetLx", _wrap_OBVibrationData_GetLx, METH_O, NULL},
	 { "OBVibrationData_GetFrequencies", _wrap_OBVibrationData_GetFrequencies, METH_O, NULL},
	 { "OBVibrationData_GetIntensities", _wrap_OBVibrationData_GetIntensities, METH_O, NULL},
	 { "OBVibrationData_GetRamanActivities", _wrap_OBVibrationData_GetRamanActivities, METH_O, NULL},
	 { "OBVibrationData_GetNumberOfFrequencies", _wrap_OBVibrationData_GetNumberOfFrequencies, METH_O, NULL},
	 { "OBVibrationData_swigregister", OBVibrationData_swigregister, METH_O, NULL},
	 { "OBVibrationData_swiginit", OBVibrationData_swiginit, METH_VARARGS, NULL},
	 { "new_OBDOSData", _wrap_new_OBDOSData, METH_NOARGS, NULL},
	 { "delete_OBDOSData", _wrap_delete_OBDOSData, METH_O, NULL},
	 { "OBDOSData_SetData", _wrap_OBDOSData_SetData, METH_VARARGS, NULL},
	 { "OBDOSData_GetFermiEnergy", _wrap_OBDOSData_GetFermiEnergy, METH_O, NULL},
	 { "OBDOSData_GetEnergies", _wrap_OBDOSData_GetEnergies, METH_O, NULL},
	 { "OBDOSData_GetDensities", _wrap_OBDOSData_GetDensities, METH_O, NULL},
	 { "OBDOSData_GetIntegration", _wrap_OBDOSData_GetIntegration, METH_O, NULL},
	 { "OBDOSData_swigregister", OBDOSData_swigregister, METH_O, NULL},
	 { "OBDOSData_swiginit", OBDOSData_swiginit, METH_VARARGS, NULL},
	 { "OBOrbital_SetData", _wrap_OBOrbital_SetData, METH_VARARGS, NULL},
	 { "OBOrbital_GetEnergy", _wrap_OBOrbital_GetEnergy, METH_O, NULL},
	 { "OBOrbital_GetOccupation", _wrap_OBOrbital_GetOccupation, METH_O, NULL},
	 { "OBOrbital_GetSymbol", _wrap_OBOrbital_GetSymbol, METH_O, NULL},
	 { "new_OBOrbital", _wrap_new_OBOrbital, METH_NOARGS, NULL},
	 { "delete_OBOrbital", _wrap_delete_OBOrbital, METH_O, NULL},
	 { "OBOrbital_swigregister", OBOrbital_swigregister, METH_O, NULL},
	 { "OBOrbital_swiginit", OBOrbital_swiginit, METH_VARARGS, NULL},
	 { "new_OBOrbitalData", _wrap_new_OBOrbitalData, METH_NOARGS, NULL},
	 { "delete_OBOrbitalData", _wrap_delete_OBOrbitalData, METH_O, NULL},
	 { "OBOrbitalData_SetAlphaOrbitals", _wrap_OBOrbitalData_SetAlphaOrbitals, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetBetaOrbitals", _wrap_OBOrbitalData_SetBetaOrbitals, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetHOMO", _wrap_OBOrbitalData_SetHOMO, METH_VARARGS, NULL},
	 { "OBOrbitalData_SetOpenShell", _wrap_OBOrbitalData_SetOpenShell, METH_VARARGS, NULL},
	 { "OBOrbitalData_IsOpenShell", _wrap_OBOrbitalData_IsOpenShell, METH_O, NULL},
	 { "OBOrbitalData_GetAlphaHOMO", _wrap_OBOrbitalData_GetAlphaHOMO, METH_O, NULL},
	 { "OBOrbitalData_GetBetaHOMO", _wrap_OBOrbitalData_GetBetaHOMO, METH_O, NULL},
	 { "OBOrbitalData_GetAlphaOrbitals", _wrap_OBOrbitalData_GetAlphaOrbitals, METH_O, NULL},
	 { "OBOrbitalData_GetBetaOrbitals", _wrap_OBOrbitalData_GetBetaOrbitals, METH_O, NULL},
	 { "OBOrbitalData_LoadClosedShellOrbitals", _wrap_OBOrbitalData_LoadClosedShellOrbitals, METH_VARARGS, "Convenience function for common cases of closed-shell calculations -- pass the energies and symmetries This method will fill the OBOrbital objects for you"},
	 { "OBOrbitalData_LoadAlphaOrbitals", _wrap_OBOrbitalData_LoadAlphaOrbitals, METH_VARARGS, "Convenience function to load alpha orbitals in an open-shell calculation"},
	 { "OBOrbitalData_LoadBetaOrbitals", _wrap_OBOrbitalData_LoadBetaOrbitals, METH_VARARGS, "Convenience function to load beta orbitals in an open-shell calculation"},
	 { "OBOrbitalData_swigregister", OBOrbitalData_swigregister, METH_O, NULL},
	 { "OBOrbitalData_swiginit", OBOrbitalData_swiginit, METH_VARARGS, NULL},
	 { "new_OBElectronicTransitionData", _wrap_new_OBElectronicTransitionData, METH_NOARGS, NULL},
	 { "delete_OBElectronicTransitionData", _wrap_delete_OBElectronicTransitionData, METH_O, NULL},
	 { "OBElectronicTransitionData_SetData", _wrap_OBElectronicTransitionData_SetData, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetEDipole", _wrap_OBElectronicTransitionData_SetEDipole, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetRotatoryStrengthsVelocity", _wrap_OBElectronicTransitionData_SetRotatoryStrengthsVelocity, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_SetRotatoryStrengthsLength", _wrap_OBElectronicTransitionData_SetRotatoryStrengthsLength, METH_VARARGS, NULL},
	 { "OBElectronicTransitionData_GetWavelengths", _wrap_OBElectronicTransitionData_GetWavelengths, METH_O, NULL},
	 { "OBElectronicTransitionData_GetForces", _wrap_OBElectronicTransitionData_GetForces, METH_O, NULL},
	 { "OBElectronicTransitionData_GetEDipole", _wrap_OBElectronicTransitionData_GetEDipole, METH_O, NULL},
	 { "OBElectronicTransitionData_GetRotatoryStrengthsVelocity", _wrap_OBElectronicTransitionData_GetRotatoryStrengthsVelocity, METH_O, NULL},
	 { "OBElectronicTransitionData_GetRotatoryStrengthsLength", _wrap_OBElectronicTransitionData_GetRotatoryStrengthsLength, METH_O, NULL},
	 { "OBElectronicTransitionData_swigregister", OBElectronicTransitionData_swigregister, METH_O, NULL},
	 { "OBElectronicTransitionData_swiginit", OBElectronicTransitionData_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotationData", _wrap_new_OBRotationData, METH_NOARGS, NULL},
	 { "delete_OBRotationData", _wrap_delete_OBRotationData, METH_O, NULL},
	 { "OBRotationData_SetData", _wrap_OBRotationData_SetData, METH_VARARGS, NULL},
	 { "OBRotationData_GetRotConsts", _wrap_OBRotationData_GetRotConsts, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: Rotational constants in GHz\n"
		""},
	 { "OBRotationData_GetSymmetryNumber", _wrap_OBRotationData_GetSymmetryNumber, METH_O, NULL},
	 { "OBRotationData_GetRotorType", _wrap_OBRotationData_GetRotorType, METH_O, NULL},
	 { "OBRotationData_swigregister", OBRotationData_swigregister, METH_O, NULL},
	 { "OBRotationData_swiginit", OBRotationData_swiginit, METH_VARARGS, NULL},
	 { "new_OBVectorData", _wrap_new_OBVectorData, METH_NOARGS, NULL},
	 { "delete_OBVectorData", _wrap_delete_OBVectorData, METH_O, NULL},
	 { "OBVectorData_SetData", _wrap_OBVectorData_SetData, METH_VARARGS, NULL},
	 { "OBVectorData_GetData", _wrap_OBVectorData_GetData, METH_O, NULL},
	 { "OBVectorData_swigregister", OBVectorData_swigregister, METH_O, NULL},
	 { "OBVectorData_swiginit", OBVectorData_swiginit, METH_VARARGS, NULL},
	 { "new_OBMatrixData", _wrap_new_OBMatrixData, METH_NOARGS, NULL},
	 { "delete_OBMatrixData", _wrap_delete_OBMatrixData, METH_O, NULL},
	 { "OBMatrixData_SetData", _wrap_OBMatrixData_SetData, METH_VARARGS, NULL},
	 { "OBMatrixData_GetData", _wrap_OBMatrixData_GetData, METH_O, NULL},
	 { "OBMatrixData_swigregister", OBMatrixData_swigregister, METH_O, NULL},
	 { "OBMatrixData_swiginit", OBMatrixData_swiginit, METH_VARARGS, NULL},
	 { "new_OBFreeGridPoint", _wrap_new_OBFreeGridPoint, METH_VARARGS, NULL},
	 { "delete_OBFreeGridPoint", _wrap_delete_OBFreeGridPoint, METH_O, NULL},
	 { "OBFreeGridPoint_GetX", _wrap_OBFreeGridPoint_GetX, METH_O, NULL},
	 { "OBFreeGridPoint_GetY", _wrap_OBFreeGridPoint_GetY, METH_O, NULL},
	 { "OBFreeGridPoint_GetZ", _wrap_OBFreeGridPoint_GetZ, METH_O, NULL},
	 { "OBFreeGridPoint_GetV", _wrap_OBFreeGridPoint_GetV, METH_O, NULL},
	 { "OBFreeGridPoint_SetX", _wrap_OBFreeGridPoint_SetX, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetY", _wrap_OBFreeGridPoint_SetY, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetZ", _wrap_OBFreeGridPoint_SetZ, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_SetV", _wrap_OBFreeGridPoint_SetV, METH_VARARGS, NULL},
	 { "OBFreeGridPoint_swigregister", OBFreeGridPoint_swigregister, METH_O, NULL},
	 { "OBFreeGridPoint_swiginit", OBFreeGridPoint_swiginit, METH_VARARGS, NULL},
	 { "new_OBFreeGrid", _wrap_new_OBFreeGrid, METH_NOARGS, NULL},
	 { "delete_OBFreeGrid", _wrap_delete_OBFreeGrid, METH_O, NULL},
	 { "OBFreeGrid_NumPoints", _wrap_OBFreeGrid_NumPoints, METH_O, NULL},
	 { "OBFreeGrid_AddPoint", _wrap_OBFreeGrid_AddPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_BeginPoints", _wrap_OBFreeGrid_BeginPoints, METH_O, NULL},
	 { "OBFreeGrid_EndPoints", _wrap_OBFreeGrid_EndPoints, METH_O, NULL},
	 { "OBFreeGrid_BeginPoint", _wrap_OBFreeGrid_BeginPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_NextPoint", _wrap_OBFreeGrid_NextPoint, METH_VARARGS, NULL},
	 { "OBFreeGrid_Clear", _wrap_OBFreeGrid_Clear, METH_O, NULL},
	 { "OBFreeGrid_swigregister", OBFreeGrid_swigregister, METH_O, NULL},
	 { "OBFreeGrid_swiginit", OBFreeGrid_swiginit, METH_VARARGS, NULL},
	 { "new_OBPcharge", _wrap_new_OBPcharge, METH_NOARGS, NULL},
	 { "delete_OBPcharge", _wrap_delete_OBPcharge, METH_O, NULL},
	 { "OBPcharge_NumPartialCharges", _wrap_OBPcharge_NumPartialCharges, METH_O, NULL},
	 { "OBPcharge_AddPartialCharge", _wrap_OBPcharge_AddPartialCharge, METH_VARARGS, NULL},
	 { "OBPcharge_GetPartialCharge", _wrap_OBPcharge_GetPartialCharge, METH_O, NULL},
	 { "OBPcharge_swigregister", OBPcharge_swigregister, METH_O, NULL},
	 { "OBPcharge_swiginit", OBPcharge_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairInteger", _wrap_new_OBPairInteger, METH_NOARGS, NULL},
	 { "OBPairInteger_SetValue", _wrap_OBPairInteger_SetValue, METH_VARARGS, NULL},
	 { "OBPairInteger_GetGenericValue", _wrap_OBPairInteger_GetGenericValue, METH_O, NULL},
	 { "delete_OBPairInteger", _wrap_delete_OBPairInteger, METH_O, NULL},
	 { "OBPairInteger_swigregister", OBPairInteger_swigregister, METH_O, NULL},
	 { "OBPairInteger_swiginit", OBPairInteger_swiginit, METH_VARARGS, NULL},
	 { "new_OBPairFloatingPoint", _wrap_new_OBPairFloatingPoint, METH_NOARGS, NULL},
	 { "OBPairFloatingPoint_SetValue", _wrap_OBPairFloatingPoint_SetValue, METH_VARARGS, NULL},
	 { "OBPairFloatingPoint_GetGenericValue", _wrap_OBPairFloatingPoint_GetGenericValue, METH_O, NULL},
	 { "delete_OBPairFloatingPoint", _wrap_delete_OBPairFloatingPoint, METH_O, NULL},
	 { "OBPairFloatingPoint_swigregister", OBPairFloatingPoint_swigregister, METH_O, NULL},
	 { "OBPairFloatingPoint_swiginit", OBPairFloatingPoint_swiginit, METH_VARARGS, NULL},
	 { "new_OBGridData", _wrap_new_OBGridData, METH_NOARGS, "Constructor."},
	 { "delete_OBGridData", _wrap_delete_OBGridData, METH_O, "Destructor"},
	 { "OBGridData_GetXAxis", _wrap_OBGridData_GetXAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: X axis direction.\n"
		""},
	 { "OBGridData_GetYAxis", _wrap_OBGridData_GetYAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: Y axis direction.\n"
		""},
	 { "OBGridData_GetZAxis", _wrap_OBGridData_GetZAxis, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: Z axis direction.\n"
		""},
	 { "OBGridData_GetAxes", _wrap_OBGridData_GetAxes, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: void\n"
		":return: the three axes parallel to the grid edges. The\n"
		"    length of the returned vector is the step along that\n"
		"    direction.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: void\n"
		":return: the three axes parallel to the grid edges\n"
		""},
	 { "OBGridData_GetNumberOfPoints", _wrap_OBGridData_GetNumberOfPoints, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: void\n"
		":return: number of points along the three axes parallel to the grid edges.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: int\n"
		":return: total number of points in the grid.\n"
		""},
	 { "OBGridData_GetNumberOfSteps", _wrap_OBGridData_GetNumberOfSteps, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: number of points along the three axes parallel to the grid edges.\n"
		""},
	 { "OBGridData_GetValues", _wrap_OBGridData_GetValues, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: grid values as a vector of doubles.\n"
		""},
	 { "OBGridData_GetValue", _wrap_OBGridData_GetValue, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the value at position i, j, k in the grid.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the value at a position in the grid (by interpolation)\n"
		""},
	 { "OBGridData_GetUnit", _wrap_OBGridData_GetUnit, METH_O, "\n"
		":rtype: int\n"
		":return: the unit of measure for grid spacings.\n"
		""},
	 { "OBGridData_GetMinValue", _wrap_OBGridData_GetMinValue, METH_O, "\n"
		":rtype: float\n"
		":return: the minimum value of all points in the grid.\n"
		""},
	 { "OBGridData_GetMaxValue", _wrap_OBGridData_GetMaxValue, METH_O, "\n"
		":rtype: float\n"
		":return: the maximum value of all points in the grid.\n"
		""},
	 { "OBGridData_GetOriginVector", _wrap_OBGridData_GetOriginVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: vector3 of the origin (i.e., the minimum x, y, and z coords of the grid).\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":type o: float\n"
		":param o: set to the origin (i.e., the minimum x, y, and z coords of the grid).\n"
		"Deprecated: Will be removed.\n"
		"See also: GetOriginVector()\n"
		""},
	 { "OBGridData_GetMaxVector", _wrap_OBGridData_GetMaxVector, METH_O, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: The maximum point in the grid.\n"
		""},
	 { "OBGridData_GetUnrestricted", _wrap_OBGridData_GetUnrestricted, METH_O, "\n"
		":rtype: boolean\n"
		":return: the unrestricted flag.\n"
		""},
	 { "OBGridData_GetNumSymmetries", _wrap_OBGridData_GetNumSymmetries, METH_O, "\n"
		":rtype: int\n"
		":return: the number of symmetries.\n"
		""},
	 { "OBGridData_SetNumberOfPoints", _wrap_OBGridData_SetNumberOfPoints, METH_VARARGS, "Set number of points along the three axes."},
	 { "OBGridData_SetLimits", _wrap_OBGridData_SetLimits, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the limits (i.e., the origin point and the axes)\n"
		"NOTE: You must set the number of points first,\n"
		"      with SetNumberOfPoints\n"
		"      so the grid spacing can be calculated\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Deprecated: Will be removed.\n"
		"See also: SetLimits(const vector3 &origin, const vector3 &x, const vector3 &y, const vector3 &z)\n"
		""},
	 { "OBGridData_SetValue", _wrap_OBGridData_SetValue, METH_VARARGS, "Set an individual value, grid must have been initialised"},
	 { "OBGridData_SetValues", _wrap_OBGridData_SetValues, METH_VARARGS, "Set the values, this vector must match the dimensions of the grid"},
	 { "OBGridData_SetUnit", _wrap_OBGridData_SetUnit, METH_VARARGS, "Set the unit of measure"},
	 { "OBGridData_SetUnrestricted", _wrap_OBGridData_SetUnrestricted, METH_VARARGS, "Set the unrestricted flag"},
	 { "OBGridData_SetNumSymmetries", _wrap_OBGridData_SetNumSymmetries, METH_VARARGS, "Set the number of symmetries"},
	 { "OBGridData_swigregister", OBGridData_swigregister, METH_O, NULL},
	 { "OBGridData_swiginit", OBGridData_swiginit, METH_VARARGS, NULL},
	 { "OBKekulize", _wrap_OBKekulize, METH_O, "\n"
		"Kekulize a molecule by assigning bond orders of 1 or 2 to aromatic bonds\n"
		"\n"
		"Some file formats describe bond orders as aromatic. Such bonds require kekulization\n"
		"before the molecule is returned by the reader. Normally, a user should never need\n"
		"to call this function themselves.\n"
		"\n"
		"This function takes an OBMol which has atoms and bonds marked as aromatic,\n"
		"aromatic bonds whose bond orders have been set to single, and aromaticity set as perceived.\n"
		"The function assumes that atoms joined by aromatic bonds have been marked as aromatic\n"
		"(if they are so intended).\n"
		"\n"
		"The purpose of the function is to set the bond orders of the aromatic bonds to either 1 or 2\n"
		"in such a way that the valencies of all of the aromatic atoms are satisfied. Failure to do\n"
		"this will result in one or more atoms having unsatisfied valences, indicated by a radical.\n"
		"Such a failure can only occur if an atom is incorrectly marked as aromatic, or is correctly\n"
		"marked as aromatic but has incorrect valence (e.g. 'n' instead of '[nH]' in SMILES).\n"
		"\n"
		":rtype: boolean\n"
		":return: Whether kekulization was successful\n"
		""},
	 { "new_OBChainsParser", _wrap_new_OBChainsParser, METH_NOARGS, NULL},
	 { "delete_OBChainsParser", _wrap_delete_OBChainsParser, METH_O, NULL},
	 { "OBChainsParser_PerceiveChains", _wrap_OBChainsParser_PerceiveChains, METH_VARARGS, "\n"
		"Perceive macromolecular (peptide and nucleotide) residues and chains\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to parse and update\n"
		":type nukeSingleResidue: boolean, optional\n"
		":param nukeSingleResidue: If only one residue is found, clear information\n"
		"    default = false  -- single residue files should still be recognized.\n"
		""},
	 { "OBChainsParser_swigregister", OBChainsParser_swigregister, METH_O, NULL},
	 { "OBChainsParser_swiginit", OBChainsParser_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomTyper", _wrap_new_OBAtomTyper, METH_VARARGS, NULL},
	 { "delete_OBAtomTyper", _wrap_delete_OBAtomTyper, METH_O, NULL},
	 { "OBAtomTyper_AssignHyb", _wrap_OBAtomTyper_AssignHyb, METH_VARARGS, "Assign atomic hybridization (1 = sp, 2 = sp2, 3 = sp3...)"},
	 { "OBAtomTyper_AssignTypes", _wrap_OBAtomTyper_AssignTypes, METH_VARARGS, "Assign external atomic types (i.e., EXTTYP lines in atomtyp.txt)"},
	 { "OBAtomTyper_swigregister", OBAtomTyper_swigregister, METH_O, NULL},
	 { "OBAtomTyper_swiginit", OBAtomTyper_swiginit, METH_VARARGS, NULL},
	 { "new_OBAromaticTyper", _wrap_new_OBAromaticTyper, METH_NOARGS, NULL},
	 { "delete_OBAromaticTyper", _wrap_delete_OBAromaticTyper, METH_O, NULL},
	 { "OBAromaticTyper_AssignAromaticFlags", _wrap_OBAromaticTyper_AssignAromaticFlags, METH_VARARGS, "Assign aromaticity flag to atoms and bonds"},
	 { "OBAromaticTyper_swigregister", OBAromaticTyper_swigregister, METH_O, NULL},
	 { "OBAromaticTyper_swiginit", OBAromaticTyper_swiginit, METH_VARARGS, NULL},
	 { "new_OBRingTyper", _wrap_new_OBRingTyper, METH_NOARGS, NULL},
	 { "delete_OBRingTyper", _wrap_delete_OBRingTyper, METH_O, NULL},
	 { "OBRingTyper_AssignTypes", _wrap_OBRingTyper_AssignTypes, METH_VARARGS, "Assign external atomic types (ringtyp.txt)"},
	 { "OBRingTyper_swigregister", OBRingTyper_swigregister, METH_O, NULL},
	 { "OBRingTyper_swiginit", OBRingTyper_swiginit, METH_VARARGS, NULL},
	 { "CharPtrLess___call__", _wrap_CharPtrLess___call__, METH_VARARGS, NULL},
	 { "new_CharPtrLess", _wrap_new_CharPtrLess, METH_NOARGS, NULL},
	 { "delete_CharPtrLess", _wrap_delete_CharPtrLess, METH_O, NULL},
	 { "CharPtrLess_swigregister", CharPtrLess_swigregister, METH_O, NULL},
	 { "CharPtrLess_swiginit", CharPtrLess_swiginit, METH_VARARGS, NULL},
	 { "delete_OBPlugin", _wrap_delete_OBPlugin, METH_O, "Virtual destructor necessary for classes with virtual functions"},
	 { "OBPlugin_Description", _wrap_OBPlugin_Description, METH_O, "Required description of a sub-type"},
	 { "OBPlugin_TypeID", _wrap_OBPlugin_TypeID, METH_O, "Redefined by each plugin type: \"formats\", \"fingerprints\", etc."},
	 { "OBPlugin_Display", _wrap_OBPlugin_Display, METH_VARARGS, "\n"
		"Write information on a plugin class to the string txt.\n"
		"Return false if not written.\n"
		"The default implementation outputs:\n"
		" the ID, a tab character, and the first line of the Description.\n"
		"The param string can be used in derived types to provide different outputs.\n"
		""},
	 { "OBPlugin_MakeInstance", _wrap_OBPlugin_MakeInstance, METH_VARARGS, "\n"
		"Make a new instance of the class.\n"
		"See OpTransform, OBGroupContrib, SmartsDescriptor classes for derived versions.\n"
		"Usually, the first parameter is the classname, the next three are\n"
		"parameters(ID, filename, description) for a constructor, and the rest data.\n"
		""},
	 { "OBPlugin_Init", _wrap_OBPlugin_Init, METH_O, "\n"
		"Initialize the plugin.\n"
		"The default version does nothing.\n"
		""},
	 { "OBPlugin_GetPlugin", _wrap_OBPlugin_GetPlugin, METH_VARARGS, "\n"
		"Get a pointer to a plugin from its type and ID. Return NULL if not found.\n"
		"If Type is NULL, search all types. Not cast to Type*\n"
		""},
	 { "OBPlugin_GetID", _wrap_OBPlugin_GetID, METH_O, "Return the ID of the sub-type instance."},
	 { "OBPlugin_ListAsVector", _wrap_OBPlugin_ListAsVector, METH_VARARGS, "\n"
		"Output a list of sub-type classes of the the type PluginID,\n"
		"or, if PluginID is \"plugins\" or empty, a list of the base types.\n"
		"If PluginID is not recognized or is NULL, the base types are output and the return is false.\n"
		""},
	 { "OBPlugin_List", _wrap_OBPlugin_List, METH_VARARGS, "As ListAsVector but sent to an ostream with a default of cout if not specified"},
	 { "OBPlugin_ListAsString", _wrap_OBPlugin_ListAsString, METH_VARARGS, "As ListAsVector but returns a string containing the list"},
	 { "OBPlugin_FirstLine", _wrap_OBPlugin_FirstLine, METH_O, "Utility function to return only the first line of a string"},
	 { "OBPlugin_Begin", _wrap_OBPlugin_Begin, METH_O, "\n"
		"Return an iterator at the start of the map of the plugin types PluginID or,\n"
		" if there is no such map, the end of the top level plugin map.\n"
		""},
	 { "OBPlugin_End", _wrap_OBPlugin_End, METH_O, NULL},
	 { "OBPlugin_GetMap", _wrap_OBPlugin_GetMap, METH_O, "Returns the map of the subtypes"},
	 { "OBPlugin_LoadAllPlugins", _wrap_OBPlugin_LoadAllPlugins, METH_NOARGS, "Load all plugins (formats, fingerprints, forcefields etc.)"},
	 { "OBPlugin_swigregister", OBPlugin_swigregister, METH_O, NULL},
	 { "new_stringbuf", _wrap_new_stringbuf, METH_NOARGS, NULL},
	 { "delete_stringbuf", _wrap_delete_stringbuf, METH_O, NULL},
	 { "stringbuf_swigregister", stringbuf_swigregister, METH_O, NULL},
	 { "stringbuf_swiginit", stringbuf_swiginit, METH_VARARGS, NULL},
	 { "new_OBError", _wrap_new_OBError, METH_VARARGS, "Constructor for an error message e.g. OBError(__FUNCTION__, \" message \")"},
	 { "OBError_message", _wrap_OBError_message, METH_O, "\n"
		":rtype: string\n"
		":return: A formatted message string, including optional explanations, etc.\n"
		""},
	 { "__lshift__", _wrap___lshift__, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Prints a representation of the vector as a row vector of the form \"<0.1,1,2>\"\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Output a text representation of a matrix\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Output to a stream\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Output a formatted message string\n"
		""},
	 { "OBError_GetMethod", _wrap_OBError_GetMethod, METH_O, "\n"
		":rtype: string\n"
		":return: The method which caused this error\n"
		"             (typically supplied via the __FUNCTION__ compiler macro *\n"
		""},
	 { "OBError_GetError", _wrap_OBError_GetError, METH_O, "\n"
		":rtype: string\n"
		":return: The main error message\n"
		""},
	 { "OBError_GetExplanation", _wrap_OBError_GetExplanation, METH_O, "\n"
		":rtype: string\n"
		":return: A more detailed explanation of the error (optional)\n"
		""},
	 { "OBError_GetPossibleCause", _wrap_OBError_GetPossibleCause, METH_O, "\n"
		":rtype: string\n"
		":return: A possible cause for the error (optional)\n"
		""},
	 { "OBError_GetSuggestedRemedy", _wrap_OBError_GetSuggestedRemedy, METH_O, "\n"
		":rtype: string\n"
		":return: The suggested workaround or remedy for the error (optional)\n"
		""},
	 { "OBError_GetLevel", _wrap_OBError_GetLevel, METH_O, "\n"
		":rtype: int\n"
		":return: The severity level of this error\n"
		""},
	 { "OBError___eq__", _wrap_OBError___eq__, METH_VARARGS, NULL},
	 { "delete_OBError", _wrap_delete_OBError, METH_O, NULL},
	 { "OBError_swigregister", OBError_swigregister, METH_O, NULL},
	 { "OBError_swiginit", OBError_swiginit, METH_VARARGS, NULL},
	 { "new_OBMessageHandler", _wrap_new_OBMessageHandler, METH_NOARGS, NULL},
	 { "delete_OBMessageHandler", _wrap_delete_OBMessageHandler, METH_O, NULL},
	 { "OBMessageHandler_ThrowError", _wrap_OBMessageHandler_ThrowError, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Throw an error with an already-formatted OBError object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Throw an error in the specified method with an appropriate level\n"
		""},
	 { "OBMessageHandler_GetMessagesOfLevel", _wrap_OBMessageHandler_GetMessagesOfLevel, METH_VARARGS, "\n"
		":rtype: std::vector< std::string,std::allocator< std::string > >\n"
		":return: all messages matching a specified level\n"
		""},
	 { "OBMessageHandler_StartLogging", _wrap_OBMessageHandler_StartLogging, METH_O, "Start logging messages (default)"},
	 { "OBMessageHandler_StopLogging", _wrap_OBMessageHandler_StopLogging, METH_O, "Stop logging messages completely"},
	 { "OBMessageHandler_SetMaxLogEntries", _wrap_OBMessageHandler_SetMaxLogEntries, METH_VARARGS, "Set the maximum number of entries (or 0 for no limit)"},
	 { "OBMessageHandler_GetMaxLogEntries", _wrap_OBMessageHandler_GetMaxLogEntries, METH_O, "\n"
		":rtype: int\n"
		":return: the current maximum number of entries (default = 0 for no limit)\n"
		""},
	 { "OBMessageHandler_ClearLog", _wrap_OBMessageHandler_ClearLog, METH_O, "Clear the current message log entirely"},
	 { "OBMessageHandler_SetOutputLevel", _wrap_OBMessageHandler_SetOutputLevel, METH_VARARGS, "Set the level of messages to output (i.e., messages with at least this priority will be output)"},
	 { "OBMessageHandler_GetOutputLevel", _wrap_OBMessageHandler_GetOutputLevel, METH_O, "\n"
		":rtype: int\n"
		":return: the current output level\n"
		""},
	 { "OBMessageHandler_SetOutputStream", _wrap_OBMessageHandler_SetOutputStream, METH_VARARGS, NULL},
	 { "OBMessageHandler_GetOutputStream", _wrap_OBMessageHandler_GetOutputStream, METH_O, NULL},
	 { "OBMessageHandler_StartErrorWrap", _wrap_OBMessageHandler_StartErrorWrap, METH_O, "Start \"wrapping\" messages to cerr into ThrowError calls"},
	 { "OBMessageHandler_StopErrorWrap", _wrap_OBMessageHandler_StopErrorWrap, METH_O, "Turn off \"wrapping\" messages, restoring normal cerr use (default)"},
	 { "OBMessageHandler_GetErrorMessageCount", _wrap_OBMessageHandler_GetErrorMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obError level\n"
		""},
	 { "OBMessageHandler_GetWarningMessageCount", _wrap_OBMessageHandler_GetWarningMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obWarning level\n"
		""},
	 { "OBMessageHandler_GetInfoMessageCount", _wrap_OBMessageHandler_GetInfoMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obInfo level\n"
		""},
	 { "OBMessageHandler_GetAuditMessageCount", _wrap_OBMessageHandler_GetAuditMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obAuditMsg level\n"
		""},
	 { "OBMessageHandler_GetDebugMessageCount", _wrap_OBMessageHandler_GetDebugMessageCount, METH_O, "\n"
		":rtype: int\n"
		":return: Count of messages received at the obDebug level\n"
		""},
	 { "OBMessageHandler_GetMessageSummary", _wrap_OBMessageHandler_GetMessageSummary, METH_O, "\n"
		":rtype: string\n"
		":return: Summary of messages received at all levels\n"
		""},
	 { "OBMessageHandler_swigregister", OBMessageHandler_swigregister, METH_O, NULL},
	 { "OBMessageHandler_swiginit", OBMessageHandler_swiginit, METH_VARARGS, NULL},
	 { "delete_obLogBuf", _wrap_delete_obLogBuf, METH_O, "Close the output buffer, flush, and call OBMessageHandler::ThrowError()"},
	 { "new_obLogBuf", _wrap_new_obLogBuf, METH_NOARGS, NULL},
	 { "obLogBuf_swigregister", obLogBuf_swigregister, METH_O, NULL},
	 { "obLogBuf_swiginit", obLogBuf_swiginit, METH_VARARGS, NULL},
	 { "OBFormat_Default", _wrap_OBFormat_Default, METH_NOARGS, NULL},
	 { "OBFormat_FindType", _wrap_OBFormat_FindType, METH_O, NULL},
	 { "OBFormat_ReadMolecule", _wrap_OBFormat_ReadMolecule, METH_VARARGS, "\n"
		"The \"API\" interface Read function.\n"
		"Reads a single object.\n"
		"Does not make a new object on the heap;\n"
		"can be used with a pointer to an chem object on the heap or the stack.\n"
		""},
	 { "OBFormat_ReadChemObject", _wrap_OBFormat_ReadChemObject, METH_VARARGS, "\n"
		"The \"Convert\" interface Read function.\n"
		"Possibly reads multiple new objects on the heap and subjects them\n"
		"to its DoTransformations() function, which may delete them again.\n"
		"Sends result to pConv->AddChemObject()\n"
		""},
	 { "OBFormat_WriteMolecule", _wrap_OBFormat_WriteMolecule, METH_VARARGS, "\n"
		"The \"API\" interface Write function.\n"
		"Writes a single object\n"
		"Does not delete the object;\n"
		"can be used with a pointer to an chem object on the heap or the stack.\n"
		":rtype: boolean\n"
		":return: false on error.\n"
		""},
	 { "OBFormat_WriteChemObject", _wrap_OBFormat_WriteChemObject, METH_VARARGS, "\n"
		"The \"Convert\" interface Write function.\n"
		"Writes a single object\n"
		"Deletes the object after writing\n"
		":rtype: boolean\n"
		":return: false on error\n"
		"\n"
		"Defined out-of-line in format.cpp: it needs OBConversion::GetChemObject(),\n"
		"which is only forward-declared here to avoid a header cycle with\n"
		"obconversion.h. Formats that don't override this must still consume\n"
		"and delete the pending object themselves, or it leaks (see\n"
		"OBConversion::Convert()'s final flush and AddChemObject()).\n"
		""},
	 { "OBFormat_TargetClassDescription", _wrap_OBFormat_TargetClassDescription, METH_O, "\n"
		"A decription of the chemical object converted by this format.\n"
		"If not provided, the object type used by the default format is used (usually OBMol).\n"
		""},
	 { "OBFormat_GetType", _wrap_OBFormat_GetType, METH_O, "\n"
		":rtype: std::type_info\n"
		":return: the type of chemical object used by the format.\n"
		"    Defaults to that used by the default format. Useful for checking\n"
		"    that a format can handle a particular object.\n"
		""},
	 { "OBFormat_SpecificationURL", _wrap_OBFormat_SpecificationURL, METH_O, "Web address where the format is defined."},
	 { "OBFormat_GetMIMEType", _wrap_OBFormat_GetMIMEType, METH_O, "Chemical MIME type associated with this file type (if any)"},
	 { "OBFormat_Flags", _wrap_OBFormat_Flags, METH_O, "\n"
		"Decribes the capabilities of the format (Read only etc.)\n"
		"Currently, can be a bitwise OR of any of the following\n"
		"NOTREADABLE READONEONLY NOTWRITABLE WRITEONEONLY DEFAULTFORMAT\n"
		"READBINARY WRITEBINARY READXML\n"
		""},
	 { "OBFormat_SkipObjects", _wrap_OBFormat_SkipObjects, METH_VARARGS, "\n"
		"Skip past first n objects in input stream (or current one with n=0)\n"
		":rtype: int\n"
		":return: 1 on success, -1 on error and 0 if not implemented\n"
		""},
	 { "OBFormat_MakeNewInstance", _wrap_OBFormat_MakeNewInstance, METH_O, "\n"
		":rtype: :py:class:`OBFormat`\n"
		":return: a pointer to a new instance of the format, or NULL if fails.\n"
		"    Normally a single global instance is used but this may cause problems\n"
		"    if there are member variables and the format is used in more than one place\n"
		"    in the program.\n"
		""},
	 { "OBFormat_RegisterFormat", _wrap_OBFormat_RegisterFormat, METH_VARARGS, NULL},
	 { "OBFormat_Display", _wrap_OBFormat_Display, METH_VARARGS, "\n"
		"Provides a description in txt of the format specified by itr.\n"
		"If param starts with \"in\", \"read\", \"out\" or \"write\" only the\n"
		"appropriate formats are output. The others return false.\n"
		"If param contains \"verbose\", the whole description is output.\n"
		""},
	 { "OBFormat_FormatFromMIME", _wrap_OBFormat_FormatFromMIME, METH_O, NULL},
	 { "delete_OBFormat", _wrap_delete_OBFormat, METH_O, NULL},
	 { "OBFormat_swigregister", OBFormat_swigregister, METH_O, NULL},
	 { "new_OBConversion", _wrap_new_OBConversion, METH_VARARGS, "Copy constructor.  Stream *ownership* is not copied. Source remains responsible for the memory."},
	 { "delete_OBConversion", _wrap_delete_OBConversion, METH_O, NULL},
	 { "OBConversion_RegisterFormat", _wrap_OBConversion_RegisterFormat, METH_VARARGS, "Called once by each format class"},
	 { "OBConversion_FindFormat", _wrap_OBConversion_FindFormat, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Searches registered formats\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Searches registered formats\n"
		"Since: version 2.3\n"
		""},
	 { "OBConversion_FormatFromExt", _wrap_OBConversion_FormatFromExt, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Searches registered formats for an ID the same as the file extension\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Searches registered formats for an ID the same as the file extension\n"
		"Since: version 2.3\n"
		""},
	 { "OBConversion_FormatFromMIME", _wrap_OBConversion_FormatFromMIME, METH_O, "Searches registered formats for a MIME the same as the chemical MIME type passed"},
	 { "OBConversion_Description", _wrap_OBConversion_Description, METH_NOARGS, "Deprecated!.Repeatedly called to recover available Formats"},
	 { "OBConversion_GetInStream", _wrap_OBConversion_GetInStream, METH_O, "These return a filtered stream for reading/writing (possible filters include compression, decompression, and newline transformation)"},
	 { "OBConversion_GetOutStream", _wrap_OBConversion_GetOutStream, METH_O, NULL},
	 { "OBConversion_SetInStream", _wrap_OBConversion_SetInStream, METH_VARARGS, "\n"
		"Set input stream.  If takeOwnership is true, will deallocate when done.\n"
		"If isGzipped is true, will treat as a gzipped stream regardless of option settings,\n"
		""},
	 { "OBConversion_SetOutStream", _wrap_OBConversion_SetOutStream, METH_VARARGS, NULL},
	 { "OBConversion_SetInAndOutFormats", _wrap_OBConversion_SetInAndOutFormats, METH_VARARGS, "Sets the formats from their ids, e g CML"},
	 { "OBConversion_SetInFormat", _wrap_OBConversion_SetInFormat, METH_VARARGS, "Sets the input format from an id e.g. CML"},
	 { "OBConversion_SetOutFormat", _wrap_OBConversion_SetOutFormat, METH_VARARGS, "Sets the output format from an id e.g. CML"},
	 { "OBConversion_GetInFormat", _wrap_OBConversion_GetInFormat, METH_O, NULL},
	 { "OBConversion_GetOutFormat", _wrap_OBConversion_GetOutFormat, METH_O, NULL},
	 { "OBConversion_GetInGzipped", _wrap_OBConversion_GetInGzipped, METH_O, NULL},
	 { "OBConversion_GetOutGzipped", _wrap_OBConversion_GetOutGzipped, METH_O, NULL},
	 { "OBConversion_GetInFilename", _wrap_OBConversion_GetInFilename, METH_O, NULL},
	 { "OBConversion_GetOutFilename", _wrap_OBConversion_GetOutFilename, METH_O, NULL},
	 { "OBConversion_GetInPos", _wrap_OBConversion_GetInPos, METH_O, "Get the position in the input stream of the object being read"},
	 { "OBConversion_GetInLen", _wrap_OBConversion_GetInLen, METH_O, "Get the length in the input stream of the object being read"},
	 { "OBConversion_GetTitle", _wrap_OBConversion_GetTitle, METH_O, "\n"
		":rtype: string\n"
		":return: a default title which is the filename\n"
		""},
	 { "OBConversion_GetAuxConv", _wrap_OBConversion_GetAuxConv, METH_O, "Extension method: deleted in ~OBConversion()"},
	 { "OBConversion_SetAuxConv", _wrap_OBConversion_SetAuxConv, METH_VARARGS, NULL},
	 { "OBConversion_IsOption", _wrap_OBConversion_IsOption, METH_VARARGS, "\n"
		"Determine whether an option is set. :rtype: char\n"
		":return: NULL if option not and a pointer to the associated text if it is\n"
		""},
	 { "OBConversion_GetOptions", _wrap_OBConversion_GetOptions, METH_VARARGS, "Access the map with option name as key and any associated text as value"},
	 { "OBConversion_AddOption", _wrap_OBConversion_AddOption, METH_VARARGS, "Set an option of specified type, with optional text"},
	 { "OBConversion_RemoveOption", _wrap_OBConversion_RemoveOption, METH_VARARGS, NULL},
	 { "OBConversion_SetOptions", _wrap_OBConversion_SetOptions, METH_VARARGS, "Set several single character options of specified type from string like ab\"btext\"c\"ctext\""},
	 { "OBConversion_RegisterOptionParam", _wrap_OBConversion_RegisterOptionParam, METH_VARARGS, "For example -h takes 0 parameters; -f takes 1. Call in a format constructor."},
	 { "OBConversion_GetOptionParams", _wrap_OBConversion_GetOptionParams, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the number of parameters registered for the option, or 0 if not found\n"
		""},
	 { "OBConversion_CopyOptions", _wrap_OBConversion_CopyOptions, METH_VARARGS, "Copies the options (by default of all types) from one OBConversion Object to another."},
	 { "OBConversion_GetSupportedInputFormat", _wrap_OBConversion_GetSupportedInputFormat, METH_O, NULL},
	 { "OBConversion_GetSupportedOutputFormat", _wrap_OBConversion_GetSupportedOutputFormat, METH_O, NULL},
	 { "OBConversion_Convert", _wrap_OBConversion_Convert, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Conversion for single input and output stream\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Conversion with existing streams\n"
		""},
	 { "OBConversion_FullConvert", _wrap_OBConversion_FullConvert, METH_VARARGS, "\n"
		"Conversion with multiple input/output files:\n"
		"makes input and output streams, and carries out normal, batch, aggregation, and splitting conversion.\n"
		""},
	 { "OBConversion_AddChemObject", _wrap_OBConversion_AddChemObject, METH_VARARGS, "Adds to internal array during input"},
	 { "OBConversion_GetChemObject", _wrap_OBConversion_GetChemObject, METH_O, "Retrieve from internal array during output"},
	 { "OBConversion_IsLast", _wrap_OBConversion_IsLast, METH_O, "True if no more objects to be output"},
	 { "OBConversion_IsFirstInput", _wrap_OBConversion_IsFirstInput, METH_O, "True if the first input object is being processed"},
	 { "OBConversion_SetFirstInput", _wrap_OBConversion_SetFirstInput, METH_VARARGS, "Setwhether or not is the first input"},
	 { "OBConversion_GetOutputIndex", _wrap_OBConversion_GetOutputIndex, METH_O, "Retrieves number of ChemObjects that have been actually output"},
	 { "OBConversion_SetOutputIndex", _wrap_OBConversion_SetOutputIndex, METH_VARARGS, "Sets output index (maybe to control whether seen as first object)"},
	 { "OBConversion_SetMoreFilesToCome", _wrap_OBConversion_SetMoreFilesToCome, METH_O, "Used with multiple input files. Off by default."},
	 { "OBConversion_SetOneObjectOnly", _wrap_OBConversion_SetOneObjectOnly, METH_VARARGS, "Used with multiple input files. Off by default."},
	 { "OBConversion_SetLast", _wrap_OBConversion_SetLast, METH_VARARGS, "Synonym for SetOneObjectOnly()"},
	 { "OBConversion_IsLastFile", _wrap_OBConversion_IsLastFile, METH_O, "True if no more files to be read"},
	 { "OBConversion_GetCount", _wrap_OBConversion_GetCount, METH_O, "\n"
		"Number of objects read and processed\n"
		"Incremented after options are processed, so 0 for first object.  Returns -1 if Convert interface not used.\n"
		""},
	 { "OBConversion_GetDefaultFormat", _wrap_OBConversion_GetDefaultFormat, METH_NOARGS, "The default format is set in a single OBFormat class (generally it is OBMol)"},
	 { "OBConversion_Write", _wrap_OBConversion_Write, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase.\n"
		"Part of \"API\" interface.\n"
		"The output stream can be specified and the change is retained in the OBConversion instance\n"
		""},
	 { "OBConversion_WriteString", _wrap_OBConversion_WriteString, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase as a string\n"
		"Part of \"API\" interface.\n"
		"The output stream is temporarily changed to the string and then restored\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		"The optional \"trimWhitespace\" parameter allows trailing whitespace to be removed\n"
		"(e.g., in a SMILES string or InChI, etc.)\n"
		""},
	 { "OBConversion_WriteFile", _wrap_OBConversion_WriteFile, METH_VARARGS, "\n"
		"Outputs an object of a class derived from OBBase as a file (with the supplied path)\n"
		"Part of \"API\" interface.\n"
		"The output stream is changed to the supplied file and the change is retained in the\n"
		"OBConversion instance.\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		""},
	 { "OBConversion_CloseOutFile", _wrap_OBConversion_CloseOutFile, METH_O, "\n"
		"Manually closes and deletes the output stream\n"
		"The file is closed anyway when in the OBConversion destructor or when WriteFile\n"
		"is called again.\n"
		"Since: version 2.1\n"
		""},
	 { "OBConversion_Read", _wrap_OBConversion_Read, METH_VARARGS, "\n"
		"Reads an object of a class derived from OBBase into pOb.\n"
		"Part of \"API\" interface.\n"
		"The input stream can be specified and the change is retained in the OBConversion instance\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		""},
	 { "OBConversion_ReadString", _wrap_OBConversion_ReadString, METH_VARARGS, "\n"
		"Part of \"API\" interface.\n"
		"The input stream can be specified and the change is retained in the OBConversion instance\n"
		":rtype: boolean\n"
		":return: NULL on error\n"
		"Reads an object of a class derived from OBBase into pOb from the supplied string\n"
		"Part of \"API\" interface.\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		"    This method is primarily intended for scripting languages without \"stream\" classes\n"
		"    Any existing input stream will be replaced by stringstream.\n"
		""},
	 { "OBConversion_ReadFile", _wrap_OBConversion_ReadFile, METH_VARARGS, "\n"
		"Reads an object of a class derived from OBBase into pOb from the file specified\n"
		"Part of \"API\" interface.\n"
		"The output stream is changed to the supplied file and the change is retained in the\n"
		"OBConversion instance. For multi-molecule files, the remaining molecules\n"
		"can be read by repeatedly calling the Read() method.\n"
		":rtype: boolean\n"
		":return: false and pOb=NULL on error\n"
		"    This method is primarily intended for scripting languages without \"stream\" classes\n"
		""},
	 { "OBConversion_OpenInAndOutFiles", _wrap_OBConversion_OpenInAndOutFiles, METH_VARARGS, "\n"
		"Part of the \"Convert\" interface.\n"
		"Open the files and update the streams in the OBConversion object.\n"
		"This method is primarily intended for scripting languages without \"stream\" classes\n"
		"and will usually followed by a call to Convert().\n"
		"Will set format from file extension if format has not already been set.\n"
		"Files will be opened even if format cannot be determined, but not if file path is empty.\n"
		":rtype: boolean\n"
		":return: false if unsuccessful.\n"
		""},
	 { "OBConversion_ReportNumberConverted", _wrap_OBConversion_ReportNumberConverted, METH_VARARGS, "\n"
		"Sends a message like \"2 molecules converted\" to clog\n"
		"The type of object is taken from the TargetClassDescription\n"
		"of the specified class (or the output format if not specified)and\n"
		"is appropriately singular or plural.\n"
		""},
	 { "OBConversion_NumInputObjects", _wrap_OBConversion_NumInputObjects, METH_O, "\n"
		":rtype: int\n"
		":return: the number of objects in the inputstream,\n"
		"    or -1 if error or if SkipObjects for the input format is not implemented\n"
		"    Adjusts for the value of -f and -l options (first and last objects).\n"
		""},
	 { "OBConversion_swigregister", OBConversion_swigregister, METH_O, NULL},
	 { "OBConversion_swiginit", OBConversion_swiginit, METH_VARARGS, NULL},
	 { "OBBondGetSmallestRingSize", _wrap_OBBondGetSmallestRingSize, METH_VARARGS, "\n"
		"Return the size of the smallest ring in which a bond appears\n"
		"\n"
		"This function returns the size of the smallest ring in which a bond appears. The\n"
		"search is bounded by the specified bound. A value of 0 is returned if the bond\n"
		"is not in a ring or if no ring is found of size less than or equal to the bound.\n"
		"\n"
		"Note that alternative algorithms may be more appropriate if you wish to calculate\n"
		"this value for all atoms in a molecule.\n"
		"\n"
		":rtype: int\n"
		":return: The size of the smallest ring, or 0\n"
		""},
	 { "GetTypicalValence", _wrap_GetTypicalValence, METH_VARARGS, "\n"
		"Return the typical valence of an atom of a particular element\n"
		"\n"
		"This function returns the typical valence of an atom given its element, current\n"
		"valence (that is, the current sum of the bond orders of its bonds) and formal\n"
		"charge.\n"
		"\n"
		"This is typically used on atoms that are missing hydrogens, to decide how many\n"
		"implicit hydrogens should be assigned (should one have to guess). For example,\n"
		"the value 3 is returned for a positively charged carbon with no attached atoms.\n"
		"\n"
		":rtype: int\n"
		":return: A value for the typical valence\n"
		""},
	 { "OBAtomAssignTypicalImplicitHydrogens", _wrap_OBAtomAssignTypicalImplicitHydrogens, METH_O, "\n"
		"Assign implicit hydrogens to an OBAtom based on typical valences\n"
		"\n"
		"This function uses the return value of GetTypicalValence to determine how many\n"
		"implicit hydrogens to assign to an OBAtom. Note that most file formats describe\n"
		"exactly how many hydrogens are present, and do not require this function.\n"
		""},
	 { "GetSymbol", _wrap_GetSymbol, METH_O, "\n"
		":rtype: string\n"
		":return: the element symbol matching the atomic number passed\n"
		""},
	 { "GetName", _wrap_GetName, METH_O, "\n"
		":rtype: string\n"
		":return: the name of this element\n"
		""},
	 { "GetMass", _wrap_GetMass, METH_O, "\n"
		":rtype: float\n"
		":return: the average atomic mass for this element. For exact isotope masses, use GetExactMass()\n"
		""},
	 { "GetExactMass", _wrap_GetExactMass, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the exact mass of the specified isotope         or by default (i.e. \"isotope 0\") the mass of the most abundant isotope\n"
		""},
	 { "GetAtomicNum", _wrap_GetAtomicNum, METH_O, NULL},
	 { "GetAllredRochowElectroNeg", _wrap_GetAllredRochowElectroNeg, METH_O, "\n"
		":rtype: float\n"
		":return: the Allred-Rochow electronegativity for this element\n"
		""},
	 { "GetCovalentRad", _wrap_GetCovalentRad, METH_O, "\n"
		":rtype: float\n"
		":return: the covalent radius (in Angstrom) for this atomic number\n"
		""},
	 { "GetVdwRad", _wrap_GetVdwRad, METH_O, "\n"
		":rtype: float\n"
		":return: the van der Waals radius (in Angstrom) for this atomic number\n"
		""},
	 { "GetElectronAffinity", _wrap_GetElectronAffinity, METH_O, "\n"
		":rtype: float\n"
		":return: the electron affinity (in eV) for this element\n"
		""},
	 { "GetIonization", _wrap_GetIonization, METH_O, "\n"
		":rtype: float\n"
		":return: the ionization potential (in eV) for this element\n"
		""},
	 { "GetMaxBonds", _wrap_GetMaxBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the maximum expected number of bonds to this element\n"
		""},
	 { "GetElectroNeg", _wrap_GetElectroNeg, METH_O, "\n"
		":rtype: float\n"
		":return: the Pauling electronegativity for this element\n"
		""},
	 { "GetRGB", _wrap_GetRGB, METH_VARARGS, "Sets the red, green, and blue color values for this element"},
	 { "new_OBResidue", _wrap_new_OBResidue, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor\n"
		""},
	 { "delete_OBResidue", _wrap_delete_OBResidue, METH_O, "Destructor"},
	 { "OBResidue_AddAtom", _wrap_OBResidue_AddAtom, METH_VARARGS, "Add ``atom`` to this residue. Updates the atom via OBAtom::SetResidue()"},
	 { "OBResidue_InsertAtom", _wrap_OBResidue_InsertAtom, METH_VARARGS, "Add ``atom`` to this residue. Updates the atom via OBAtom::SetResidue()"},
	 { "OBResidue_RemoveAtom", _wrap_OBResidue_RemoveAtom, METH_VARARGS, "Remove ``atom`` from this residue and update the atom."},
	 { "OBResidue_SetName", _wrap_OBResidue_SetName, METH_VARARGS, "Set the name of this residue (e.g., \"ALA\"). Use 3-char PDB standard names. http://www.rcsb.org/pdb/file_formats/pdb/pdbguide2.2/part_79.html MODRES records for modified residues: http://www.rcsb.org/pdb/file_formats/pdb/pdbguide2.2/part_36.html"},
	 { "OBResidue_SetSegName", _wrap_OBResidue_SetSegName, METH_VARARGS, "Set the segment name of this residue (max four characters)"},
	 { "OBResidue_SetNum", _wrap_OBResidue_SetNum, METH_VARARGS, "Set the residue number (in the sequence)"},
	 { "OBResidue_SetChain", _wrap_OBResidue_SetChain, METH_VARARGS, "Set the chain ID for this residue"},
	 { "OBResidue_SetChainNum", _wrap_OBResidue_SetChainNum, METH_VARARGS, "Set the chain number for this residue"},
	 { "OBResidue_SetIdx", _wrap_OBResidue_SetIdx, METH_VARARGS, "Set the internal index of this residue in the parent OBMol. Intended mostly for internal use"},
	 { "OBResidue_SetInsertionCode", _wrap_OBResidue_SetInsertionCode, METH_VARARGS, "Set  PDB insertion code information for this residue. This allows consecutive residues to have the same number. Some communities that work in a well-conserved structural world use this, e.g. for immunoglobulins."},
	 { "OBResidue_SetAtomID", _wrap_OBResidue_SetAtomID, METH_VARARGS, "Set the character code ID for an ATOM record for the supplied atom This does nothing if the supplied atom is not found in the residue"},
	 { "OBResidue_SetHetAtom", _wrap_OBResidue_SetHetAtom, METH_VARARGS, NULL},
	 { "OBResidue_SetSerialNum", _wrap_OBResidue_SetSerialNum, METH_VARARGS, "Set the atomic serial number for a given atom (see OBSerialNums)"},
	 { "OBResidue_GetName", _wrap_OBResidue_GetName, METH_O, "\n"
		":rtype: string\n"
		":return: The residue name\n"
		""},
	 { "OBResidue_GetSegName", _wrap_OBResidue_GetSegName, METH_O, "\n"
		":rtype: string\n"
		":return: The residue segment name\n"
		""},
	 { "OBResidue_GetNum", _wrap_OBResidue_GetNum, METH_O, "\n"
		":rtype: int\n"
		":return: The residue number (in the sequence)\n"
		""},
	 { "OBResidue_GetNumString", _wrap_OBResidue_GetNumString, METH_O, NULL},
	 { "OBResidue_GetNumAtoms", _wrap_OBResidue_GetNumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in this residue\n"
		""},
	 { "OBResidue_GetNumHvyAtoms", _wrap_OBResidue_GetNumHvyAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of heavy atoms in this residue\n"
		""},
	 { "OBResidue_GetChain", _wrap_OBResidue_GetChain, METH_O, "\n"
		":rtype: char\n"
		":return: The ID of the chain which includes this residue\n"
		""},
	 { "OBResidue_GetChainNum", _wrap_OBResidue_GetChainNum, METH_O, "\n"
		":rtype: int\n"
		":return: The number of the chain which includes this residue\n"
		""},
	 { "OBResidue_GetIdx", _wrap_OBResidue_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The internal index of this residue in the parent OBMol\n"
		""},
	 { "OBResidue_GetResKey", _wrap_OBResidue_GetResKey, METH_O, "\n"
		":rtype: int\n"
		":return: The residue key (i.e., an entry in the OBResidueIndex namespace)\n"
		""},
	 { "OBResidue_GetBonds", _wrap_OBResidue_GetBonds, METH_VARARGS, "\n"
		":rtype: std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >\n"
		":return: all bonds in this residue. ``exterior`` includes bonds to atoms  outside this residue (default is true)\n"
		""},
	 { "OBResidue_GetAtomID", _wrap_OBResidue_GetAtomID, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the atom ID (character code) for the supplied atom or \"\"  if the atom is not found in this residue\n"
		""},
	 { "OBResidue_GetSerialNum", _wrap_OBResidue_GetSerialNum, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the serial number of the supplied atom (uses OBSerialNums)\n"
		""},
	 { "OBResidue_GetInsertionCode", _wrap_OBResidue_GetInsertionCode, METH_O, "\n"
		":rtype: char\n"
		":return: The Insertion Code (i.e., an extra position motivated by a multiple sequence alignment against a template with defined numbers)\n"
		""},
	 { "OBResidue_GetAminoAcidProperty", _wrap_OBResidue_GetAminoAcidProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this residue has the supplied amino acid property  defined from the OBAminoAcidProperty namespace\n"
		""},
	 { "OBResidue_GetAtomProperty", _wrap_OBResidue_GetAtomProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether atom ``a`` has the supplied residue atom property  defined from the OBResidueAtomProperty namespace\n"
		""},
	 { "OBResidue_GetResidueProperty", _wrap_OBResidue_GetResidueProperty, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this residue has the supplied property  defined from the OBResidueProperty namespace\n"
		""},
	 { "OBResidue_IsHetAtom", _wrap_OBResidue_IsHetAtom, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: If the given atom is a HETATM record\n"
		""},
	 { "OBResidue_IsResidueType", _wrap_OBResidue_IsResidueType, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: If this residue matches the supplied ``restype`` Set by SetResidueKeys()\n"
		""},
	 { "OBResidue_BeginAtoms", _wrap_OBResidue_BeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomIterator\n"
		":return: An iterator to the beginning of the atom list in this residue\n"
		""},
	 { "OBResidue_CBeginAtoms", _wrap_OBResidue_CBeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant iterator to the beginning of the atom list in this residue\n"
		""},
	 { "OBResidue_EndAtoms", _wrap_OBResidue_EndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomIterator\n"
		":return: An iterator to the end of the atom list in this residue\n"
		""},
	 { "OBResidue_CEndAtoms", _wrap_OBResidue_CEndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant iterator to the end of the atom list in this residue\n"
		""},
	 { "OBResidue_BeginAtom", _wrap_OBResidue_BeginAtom, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the atom list in this residue :rtype: :py:class:`OBAtom`\n"
		":return: The first atom (or NULL if none exist)\n"
		""},
	 { "OBResidue_NextAtom", _wrap_OBResidue_NextAtom, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBAtom`\n"
		":return: The next atom (or NULL if none exist)\n"
		""},
	 { "OBResidue_swigregister", OBResidue_swigregister, METH_O, NULL},
	 { "OBResidue_swiginit", OBResidue_swiginit, METH_VARARGS, NULL},
	 { "OBInternalCoord__a_set", _wrap_OBInternalCoord__a_set, METH_VARARGS, "First connection for this atom (i.e., distance)"},
	 { "OBInternalCoord__a_get", _wrap_OBInternalCoord__a_get, METH_O, "First connection for this atom (i.e., distance)"},
	 { "OBInternalCoord__b_set", _wrap_OBInternalCoord__b_set, METH_VARARGS, "Second reference atom (i.e., angle)"},
	 { "OBInternalCoord__b_get", _wrap_OBInternalCoord__b_get, METH_O, "Second reference atom (i.e., angle)"},
	 { "OBInternalCoord__c_set", _wrap_OBInternalCoord__c_set, METH_VARARGS, "Third reference atom (i.e., dihedral / torsion angle)"},
	 { "OBInternalCoord__c_get", _wrap_OBInternalCoord__c_get, METH_O, "Third reference atom (i.e., dihedral / torsion angle)"},
	 { "OBInternalCoord__dst_set", _wrap_OBInternalCoord__dst_set, METH_VARARGS, "Distance between this atom and _a"},
	 { "OBInternalCoord__dst_get", _wrap_OBInternalCoord__dst_get, METH_O, "Distance between this atom and _a"},
	 { "OBInternalCoord__ang_set", _wrap_OBInternalCoord__ang_set, METH_VARARGS, "Angle between this, _a, and _b (i.e., _a is the vertex)"},
	 { "OBInternalCoord__ang_get", _wrap_OBInternalCoord__ang_get, METH_O, "Angle between this, _a, and _b (i.e., _a is the vertex)"},
	 { "OBInternalCoord__tor_set", _wrap_OBInternalCoord__tor_set, METH_VARARGS, "Torsional/dihedral angle between this, _a, _b, and _c"},
	 { "OBInternalCoord__tor_get", _wrap_OBInternalCoord__tor_get, METH_O, "Torsional/dihedral angle between this, _a, _b, and _c"},
	 { "new_OBInternalCoord", _wrap_new_OBInternalCoord, METH_VARARGS, "Constructor"},
	 { "delete_OBInternalCoord", _wrap_delete_OBInternalCoord, METH_O, NULL},
	 { "OBInternalCoord_swigregister", OBInternalCoord_swigregister, METH_O, NULL},
	 { "OBInternalCoord_swiginit", OBInternalCoord_swiginit, METH_VARARGS, NULL},
	 { "new_OBAtomAtomIterAdaptor", _wrap_new_OBAtomAtomIterAdaptor, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___ref__", _wrap_OBAtomAtomIterAdaptor___ref__, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_inc", _wrap_OBAtomAtomIterAdaptor_inc, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___eq__", _wrap_OBAtomAtomIterAdaptor___eq__, METH_VARARGS, NULL},
	 { "OBAtomAtomIterAdaptor___ne__", _wrap_OBAtomAtomIterAdaptor___ne__, METH_VARARGS, NULL},
	 { "delete_OBAtomAtomIterAdaptor", _wrap_delete_OBAtomAtomIterAdaptor, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_swigregister", OBAtomAtomIterAdaptor_swigregister, METH_O, NULL},
	 { "OBAtomAtomIterAdaptor_swiginit", OBAtomAtomIterAdaptor_swiginit, METH_VARARGS, NULL},
	 { "OBAtom_Visit_set", _wrap_OBAtom_Visit_set, METH_VARARGS, "Used internally by graph traversal algorithms"},
	 { "OBAtom_Visit_get", _wrap_OBAtom_Visit_get, METH_O, "Used internally by graph traversal algorithms"},
	 { "new_OBAtom", _wrap_new_OBAtom, METH_NOARGS, "Constructor"},
	 { "delete_OBAtom", _wrap_delete_OBAtom, METH_O, "Destructor"},
	 { "OBAtom___eq__", _wrap_OBAtom___eq__, METH_VARARGS, "Equivalence"},
	 { "OBAtom_Duplicate", _wrap_OBAtom_Duplicate, METH_VARARGS, "Duplicate another atom. Copies all information with the exception of index Since: version 2.2"},
	 { "OBAtom_SetIdx", _wrap_OBAtom_SetIdx, METH_VARARGS, "Set atom index (i.e., in an OBMol)"},
	 { "OBAtom_SetId", _wrap_OBAtom_SetId, METH_VARARGS, NULL},
	 { "OBAtom_SetHyb", _wrap_OBAtom_SetHyb, METH_VARARGS, "Set atom hybridization (i.e., 1 = sp, 2 = sp2, 3 = sp3 ...)"},
	 { "OBAtom_SetAtomicNum", _wrap_OBAtom_SetAtomicNum, METH_VARARGS, "Set atomic number"},
	 { "OBAtom_SetIsotope", _wrap_OBAtom_SetIsotope, METH_VARARGS, "Set isotope number (actual atomic weight is tabulated automatically, 0 = most abundant)"},
	 { "OBAtom_SetImplicitHCount", _wrap_OBAtom_SetImplicitHCount, METH_VARARGS, "Set the implicit hydrogen count to ``val``"},
	 { "OBAtom_SetFormalCharge", _wrap_OBAtom_SetFormalCharge, METH_VARARGS, "Set the formal charge of the atom to ``fcharge``"},
	 { "OBAtom_SetSpinMultiplicity", _wrap_OBAtom_SetSpinMultiplicity, METH_VARARGS, "Set the atomic spin to ``spin``. See _spinmultiplicity"},
	 { "OBAtom_SetType", _wrap_OBAtom_SetType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the atomic type symbol (see OBTypeTable and OBAtomTyper for more)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the atomic type symbol (see OBTypeTable and OBAtomTyper for more)\n"
		""},
	 { "OBAtom_SetPartialCharge", _wrap_OBAtom_SetPartialCharge, METH_VARARGS, "Set the partial charge to ``pcharge``"},
	 { "OBAtom_SetCoordPtr", _wrap_OBAtom_SetCoordPtr, METH_VARARGS, "Set the position of this atom from a pointer-driven array of coordinates"},
	 { "OBAtom_SetVector", _wrap_OBAtom_SetVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the coordinate vector for this atom to ``v`` as a vector3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the coordinate vector for this atom based on ``x`` ``y`` & ``z``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Set the position of this atom based on the internal pointer array (i.e. from SetCoordPtr() )\n"
		""},
	 { "OBAtom_SetResidue", _wrap_OBAtom_SetResidue, METH_VARARGS, "Attach an OBResidue ``res`` as containing this atom"},
	 { "OBAtom_SetParent", _wrap_OBAtom_SetParent, METH_VARARGS, "Attach an OBMol ``ptr`` as the parent container for this atom"},
	 { "OBAtom_SetAromatic", _wrap_OBAtom_SetAromatic, METH_VARARGS, "Mark atom as being aromatic"},
	 { "OBAtom_SetInRing", _wrap_OBAtom_SetInRing, METH_VARARGS, "Mark an atom as belonging to at least one ring"},
	 { "OBAtom_ClearCoordPtr", _wrap_OBAtom_ClearCoordPtr, METH_O, "Clear the internal coordinate pointer"},
	 { "OBAtom_GetFormalCharge", _wrap_OBAtom_GetFormalCharge, METH_O, "\n"
		":rtype: int\n"
		":return: the formal charge for this atom\n"
		""},
	 { "OBAtom_GetAtomicNum", _wrap_OBAtom_GetAtomicNum, METH_O, "\n"
		":rtype: int\n"
		":return: the atomic number for this atom\n"
		""},
	 { "OBAtom_GetIsotope", _wrap_OBAtom_GetIsotope, METH_O, "\n"
		":rtype: int\n"
		":return: the isotope for this atom, if specified, or 0 for unspecified\n"
		""},
	 { "OBAtom_GetSpinMultiplicity", _wrap_OBAtom_GetSpinMultiplicity, METH_O, "\n"
		":rtype: int\n"
		":return: the atomic spin, e.g., 0 (default) for normal atoms - note that this value is a convention,   2 for radical  1 or 3 for carbene\n"
		""},
	 { "OBAtom_GetAtomicMass", _wrap_OBAtom_GetAtomicMass, METH_O, "\n"
		":rtype: float\n"
		":return: the atomic mass of this atom given by standard IUPAC  average molar mass\n"
		""},
	 { "OBAtom_GetExactMass", _wrap_OBAtom_GetExactMass, METH_O, "\n"
		":rtype: float\n"
		":return: the atomic mass of this atom given by the isotope (default of 0 gives the most abundant isotope)\n"
		""},
	 { "OBAtom_GetIdx", _wrap_OBAtom_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the internal atom index (e.g., inside an OBMol)\n"
		""},
	 { "OBAtom_GetIndex", _wrap_OBAtom_GetIndex, METH_O, NULL},
	 { "OBAtom_GetId", _wrap_OBAtom_GetId, METH_O, NULL},
	 { "OBAtom_GetCoordinateIdx", _wrap_OBAtom_GetCoordinateIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the index into a pointer-driven array as used by   GetCoordPtr() or SetCoordPtr()\n"
		""},
	 { "OBAtom_GetExplicitDegree", _wrap_OBAtom_GetExplicitDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of explicit bonds to this atom\n"
		""},
	 { "OBAtom_GetTotalDegree", _wrap_OBAtom_GetTotalDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The total number of bonds to this atom including bonds to implicit hydrogens\n"
		""},
	 { "OBAtom_GetExplicitValence", _wrap_OBAtom_GetExplicitValence, METH_O, "\n"
		":rtype: int\n"
		":return: The sum of the bond orders of the explicit bonds to this atom\n"
		""},
	 { "OBAtom_GetTotalValence", _wrap_OBAtom_GetTotalValence, METH_O, "\n"
		":rtype: int\n"
		":return: The sum of the bond orders of all bonds to this atom including bonds to implicit hydrogens\n"
		""},
	 { "OBAtom_GetHyb", _wrap_OBAtom_GetHyb, METH_O, "\n"
		":rtype: int\n"
		":return: The hybridization of this atom: 1 for sp, 2 for sp2, 3 for sp3, 4 for sq. planar, 5 for trig. bipy, 6 for octahedral\n"
		""},
	 { "OBAtom_GetImplicitHCount", _wrap_OBAtom_GetImplicitHCount, METH_O, "\n"
		":rtype: unsigned char\n"
		":return: The number of implicit hydrogens attached to this atom\n"
		""},
	 { "OBAtom_GetHvyDegree", _wrap_OBAtom_GetHvyDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of non-hydrogens connected to this atom\n"
		""},
	 { "OBAtom_GetHeteroDegree", _wrap_OBAtom_GetHeteroDegree, METH_O, "\n"
		":rtype: int\n"
		":return: The number of heteroatoms connected to an atom\n"
		""},
	 { "OBAtom_GetType", _wrap_OBAtom_GetType, METH_O, "\n"
		":rtype: string\n"
		":return: the atomic type (e.g., for molecular mechanics)\n"
		""},
	 { "OBAtom_GetX", _wrap_OBAtom_GetX, METH_O, "\n"
		":rtype: float\n"
		":return: the x coordinate\n"
		""},
	 { "OBAtom_GetY", _wrap_OBAtom_GetY, METH_O, "\n"
		":rtype: float\n"
		":return: the y coordinate\n"
		""},
	 { "OBAtom_GetZ", _wrap_OBAtom_GetZ, METH_O, "\n"
		":rtype: float\n"
		":return: the z coordinate\n"
		""},
	 { "OBAtom_x", _wrap_OBAtom_x, METH_O, "\n"
		":rtype: float\n"
		":return: the x coordinate\n"
		""},
	 { "OBAtom_y", _wrap_OBAtom_y, METH_O, "\n"
		":rtype: float\n"
		":return: the y coordinate\n"
		""},
	 { "OBAtom_z", _wrap_OBAtom_z, METH_O, "\n"
		":rtype: float\n"
		":return: the z coordinate\n"
		""},
	 { "OBAtom_GetCoordinate", _wrap_OBAtom_GetCoordinate, METH_O, "\n"
		":rtype: float\n"
		":return: the coordinates as a double* or NULL if none. See SetCoordPtr() for more. If no coordinate pointer is used (e.g., only vector3), NULL will be returned.\n"
		""},
	 { "OBAtom_GetVector", _wrap_OBAtom_GetVector, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the coordinates as a vector3 object\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the coordinates as a vector3 object\n"
		""},
	 { "OBAtom_GetPartialCharge", _wrap_OBAtom_GetPartialCharge, METH_O, "\n"
		":rtype: float\n"
		":return: the partial charge of this atom, calculating a Gasteiger charge if needed\n"
		""},
	 { "OBAtom_GetResidue", _wrap_OBAtom_GetResidue, METH_O, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: the residue which contains this atom, or NULL if none exists\n"
		""},
	 { "OBAtom_GetParent", _wrap_OBAtom_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: the molecule which contains this atom, or NULL if none exists\n"
		""},
	 { "OBAtom_GetNewBondVector", _wrap_OBAtom_GetNewBondVector, METH_VARARGS, "\n"
		" Create a vector for a new bond from this atom, with length given by the supplied parameter :rtype: boolean\n"
		":return: success or failure\n"
		""},
	 { "OBAtom_GetBond", _wrap_OBAtom_GetBond, METH_VARARGS, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the OBBond object between this atom and that supplied, or NULL if the two atoms are not bonded\n"
		""},
	 { "OBAtom_BeginBonds", _wrap_OBAtom_BeginBonds, METH_O, "\n"
		":rtype: OpenBabel::OBBondIterator\n"
		":return: An iterator to the beginning of the bonds to this atom\n"
		""},
	 { "OBAtom_EndBonds", _wrap_OBAtom_EndBonds, METH_O, "\n"
		":rtype: OpenBabel::OBBondIterator\n"
		":return: An iterator to the end of the bonds to this atom\n"
		""},
	 { "OBAtom_GetBonds", _wrap_OBAtom_GetBonds, METH_O, "\n"
		":rtype: OpenBabel::OBAtomBondRange\n"
		":return: A range over the bonds to this atom. This range can be used in a range-based for loop.\n"
		""},
	 { "OBAtom_BeginBond", _wrap_OBAtom_BeginBond, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the bonds :rtype: :py:class:`OBBond`\n"
		":return: The first bond to this atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_NextBond", _wrap_OBAtom_NextBond, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBBond`\n"
		":return: The next bond to this atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_GetNbrs", _wrap_OBAtom_GetNbrs, METH_O, "\n"
		":rtype: OpenBabel::OBAtomAtomRange\n"
		":return: A range over the neighbors of this atom. This range can be used in a range-based for loop.\n"
		""},
	 { "OBAtom_BeginNbrAtom", _wrap_OBAtom_BeginNbrAtom, METH_VARARGS, "\n"
		" Set the iterator ``i`` to the beginning of the bonds :rtype: :py:class:`OBAtom`\n"
		":return: The first neighboring atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_NextNbrAtom", _wrap_OBAtom_NextNbrAtom, METH_VARARGS, "\n"
		" Increment the iterator ``i`` :rtype: :py:class:`OBAtom`\n"
		":return: The next neighboring atom (or NULL if none exist)\n"
		""},
	 { "OBAtom_GetDistance", _wrap_OBAtom_GetDistance, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the distance to the atom defined by OBMol::GetAtom()\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the distance to the supplied OBAtom\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: float\n"
		":return: the distance to the coordinates of the supplied vector3 Since: version 2.4\n"
		""},
	 { "OBAtom_GetAngle", _wrap_OBAtom_GetAngle, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the angle defined by this atom -> b (vertex) -> c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the angle defined by this atom -> b (vertex) -> c\n"
		""},
	 { "OBAtom_NewResidue", _wrap_OBAtom_NewResidue, METH_O, "If no residue has been set for this atom, create a new one"},
	 { "OBAtom_AddResidue", _wrap_OBAtom_AddResidue, METH_VARARGS, "Add (set) the residue for this atom"},
	 { "OBAtom_DeleteResidue", _wrap_OBAtom_DeleteResidue, METH_O, "Delete any residue associated with this atom"},
	 { "OBAtom_AddBond", _wrap_OBAtom_AddBond, METH_VARARGS, "Add a bond to the internal list. Does not update the bond."},
	 { "OBAtom_InsertBond", _wrap_OBAtom_InsertBond, METH_VARARGS, "Insert ``bond`` into the internal list at the position from ``i`` Does not modify the bond"},
	 { "OBAtom_DeleteBond", _wrap_OBAtom_DeleteBond, METH_VARARGS, "Find ``bond`` and remove it from the internal list. Does not update the bond."},
	 { "OBAtom_ClearBond", _wrap_OBAtom_ClearBond, METH_O, "Clear all bonding information in this atom (does not delete them)"},
	 { "OBAtom_HtoMethyl", _wrap_OBAtom_HtoMethyl, METH_O, "\n"
		" If this is a hydrogen atom, transform into a methyl group :rtype: boolean\n"
		":return: success or failure\n"
		""},
	 { "OBAtom_SetHybAndGeom", _wrap_OBAtom_SetHybAndGeom, METH_VARARGS, "\n"
		" Change the hybridization of this atom and modify the geometry accordingly :rtype: boolean\n"
		":return: success or failure Deprecated: This will be removed in future versions of Open Babel\n"
		""},
	 { "OBAtom_CountFreeOxygens", _wrap_OBAtom_CountFreeOxygens, METH_O, "\n"
		":rtype: int\n"
		":return: The number of oxygen atoms connected that only have one heavy valence\n"
		""},
	 { "OBAtom_CountFreeSulfurs", _wrap_OBAtom_CountFreeSulfurs, METH_O, "\n"
		":rtype: int\n"
		":return: The number of sulfur atoms connected that only have one heavy valence Since: version 2.4\n"
		""},
	 { "OBAtom_ExplicitHydrogenCount", _wrap_OBAtom_ExplicitHydrogenCount, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The number of hydrogens explicitly bound to this atom, optionally excluding D,T and isotope explicitly set to 1\n"
		""},
	 { "OBAtom_MemberOfRingCount", _wrap_OBAtom_MemberOfRingCount, METH_O, "\n"
		":rtype: int\n"
		":return: The number of rings that contain this atom\n"
		""},
	 { "OBAtom_MemberOfRingSize", _wrap_OBAtom_MemberOfRingSize, METH_O, "\n"
		":rtype: int\n"
		":return: The size of the smallest ring that contains this atom (0 if not in a ring)\n"
		""},
	 { "OBAtom_CountRingBonds", _wrap_OBAtom_CountRingBonds, METH_O, "\n"
		":rtype: int\n"
		":return: The number of explicit ring connections to this atom\n"
		""},
	 { "OBAtom_SmallestBondAngle", _wrap_OBAtom_SmallestBondAngle, METH_O, "\n"
		":rtype: float\n"
		":return: The smallest angle of bonds to this atom\n"
		""},
	 { "OBAtom_AverageBondAngle", _wrap_OBAtom_AverageBondAngle, METH_O, "\n"
		":rtype: float\n"
		":return: The average angle of bonds to this atom\n"
		""},
	 { "OBAtom_LewisAcidBaseCounts", _wrap_OBAtom_LewisAcidBaseCounts, METH_O, "\n"
		"Lewis acid/base vacancies for this atom\n"
		":rtype: std::pair< int,int >\n"
		":return: A pair of integers, where first is acid count and second is base count\n"
		"Since: version 2.3\n"
		""},
	 { "OBAtom_HasResidue", _wrap_OBAtom_HasResidue, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is there any residue information?\n"
		""},
	 { "OBAtom_IsHetAtom", _wrap_OBAtom_IsHetAtom, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this a HETATM in a residue (returns false if not in a residue) Since: version 2.4\n"
		""},
	 { "OBAtom_IsElement", _wrap_OBAtom_IsElement, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the specified element, as specified by atom number (see OBElement namespace)?\n"
		""},
	 { "OBAtom_IsAromatic", _wrap_OBAtom_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom aromatic?\n"
		""},
	 { "OBAtom_IsInRing", _wrap_OBAtom_IsInRing, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom in a ring?\n"
		""},
	 { "OBAtom_IsInRingSize", _wrap_OBAtom_IsInRingSize, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the atom in a ring of a given size?\n"
		""},
	 { "OBAtom_IsHeteroatom", _wrap_OBAtom_IsHeteroatom, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an element in the 15th or 16th main groups  (i.e., N, O, P, S ...) ?\n"
		""},
	 { "OBAtom_IsConnected", _wrap_OBAtom_IsConnected, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom directly connected to the supplied OBAtom?\n"
		""},
	 { "OBAtom_IsOneThree", _wrap_OBAtom_IsOneThree, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom related to the supplied OBAtom in  a 1,3 bonding pattern?\n"
		""},
	 { "OBAtom_IsOneFour", _wrap_OBAtom_IsOneFour, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is this atom related to the supplied OBAtom in  a 1,4 bonding pattern?\n"
		""},
	 { "OBAtom_IsCarboxylOxygen", _wrap_OBAtom_IsCarboxylOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a carboxyl (-CO2 or CO2H) group?\n"
		""},
	 { "OBAtom_IsPhosphateOxygen", _wrap_OBAtom_IsPhosphateOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a phosphate (R-PO3) group?\n"
		""},
	 { "OBAtom_IsSulfateOxygen", _wrap_OBAtom_IsSulfateOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a sulfate (-SO3) group?\n"
		""},
	 { "OBAtom_IsNitroOxygen", _wrap_OBAtom_IsNitroOxygen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an oxygen in a nitro (-NO2) group?\n"
		""},
	 { "OBAtom_IsAmideNitrogen", _wrap_OBAtom_IsAmideNitrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a nitrogen in an amide (-C(=O)NR2) group?\n"
		""},
	 { "OBAtom_IsPolarHydrogen", _wrap_OBAtom_IsPolarHydrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen connected to a polar atom  (i.e., N, O, P, S)\n"
		""},
	 { "OBAtom_IsNonPolarHydrogen", _wrap_OBAtom_IsNonPolarHydrogen, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen connected to a non-polar atom  (i.e., C)\n"
		""},
	 { "OBAtom_IsAromaticNOxide", _wrap_OBAtom_IsAromaticNOxide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an aromatic nitrogen with at least one  double bond to an oxygen atom\n"
		""},
	 { "OBAtom_IsChiral", _wrap_OBAtom_IsChiral, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom chiral?\n"
		""},
	 { "OBAtom_IsPeriodic", _wrap_OBAtom_IsPeriodic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the atom part of a periodic unit cell?\n"
		""},
	 { "OBAtom_IsAxial", _wrap_OBAtom_IsAxial, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom an axial atom in a ring\n"
		""},
	 { "OBAtom_IsHbondAcceptor", _wrap_OBAtom_IsHbondAcceptor, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond acceptor  (considering also atom surrounding)\n"
		""},
	 { "OBAtom_IsHbondAcceptorSimple", _wrap_OBAtom_IsHbondAcceptorSimple, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond acceptor (old function)?\n"
		""},
	 { "OBAtom_IsHbondDonor", _wrap_OBAtom_IsHbondDonor, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a hydrogen-bond donor?\n"
		""},
	 { "OBAtom_IsHbondDonorH", _wrap_OBAtom_IsHbondDonorH, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this a hydrogen atom attached to a hydrogen-bond donor?\n"
		""},
	 { "OBAtom_IsMetal", _wrap_OBAtom_IsMetal, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is this atom a metal? Since: version 2.4\n"
		""},
	 { "OBAtom_HasAlphaBetaUnsat", _wrap_OBAtom_HasAlphaBetaUnsat, METH_VARARGS, "\n"
		" :rtype: boolean\n"
		" :return: Whether a neighboring atom (alpha) has an unsaturated bond   to a third atom (beta). :type includePandS: boolean, optional\n"
		":param includePandS: Whether to include phosphorus and sulfur neighbors in this determination (or to exclude them)\n"
		""},
	 { "OBAtom_HasBondOfOrder", _wrap_OBAtom_HasBondOfOrder, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this atom is connected to any bond with order == ``bo``\n"
		""},
	 { "OBAtom_CountBondsOfOrder", _wrap_OBAtom_CountBondsOfOrder, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The count of bonds connected to this atom with order == ``bo``\n"
		""},
	 { "OBAtom_HighestBondOrder", _wrap_OBAtom_HighestBondOrder, METH_O, "\n"
		":rtype: int\n"
		":return: The maximum bond order for this atom\n"
		""},
	 { "OBAtom_HasNonSingleBond", _wrap_OBAtom_HasNonSingleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this atom is connected to any bond with order >1\n"
		""},
	 { "OBAtom_HasSingleBond", _wrap_OBAtom_HasSingleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have a single bond\n"
		""},
	 { "OBAtom_HasDoubleBond", _wrap_OBAtom_HasDoubleBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have a double bond\n"
		""},
	 { "OBAtom_HasAromaticBond", _wrap_OBAtom_HasAromaticBond, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this atom have an aromatic bond\n"
		""},
	 { "OBAtom_MatchesSMARTS", _wrap_OBAtom_MatchesSMARTS, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether this atom matches the first atom in a given SMARTS pattern\n"
		""},
	 { "OBAtom_swigregister", OBAtom_swigregister, METH_O, NULL},
	 { "OBAtom_swiginit", OBAtom_swiginit, METH_VARARGS, NULL},
	 { "OBBond_Visit_set", _wrap_OBBond_Visit_set, METH_VARARGS, "Whether this bond has been visited by a graph algorithm Deprecated: Use OBBitVec objects instead to be fully thread-safe. *"},
	 { "OBBond_Visit_get", _wrap_OBBond_Visit_get, METH_O, "Whether this bond has been visited by a graph algorithm Deprecated: Use OBBitVec objects instead to be fully thread-safe. *"},
	 { "new_OBBond", _wrap_new_OBBond, METH_NOARGS, "Constructor"},
	 { "delete_OBBond", _wrap_delete_OBBond, METH_O, "Destructor"},
	 { "OBBond_SetIdx", _wrap_OBBond_SetIdx, METH_VARARGS, "\n"
		"Set the internal bond index Warning: This will not update the index in the parent OBMol.\n"
		"         Intended mainly for internal use. Use with care. *\n"
		""},
	 { "OBBond_SetId", _wrap_OBBond_SetId, METH_VARARGS, NULL},
	 { "OBBond_SetBondOrder", _wrap_OBBond_SetBondOrder, METH_VARARGS, "Set the bond order to ``order`` (i.e., 1 = single, 2 = double, 5 = aromatic)"},
	 { "OBBond_SetBegin", _wrap_OBBond_SetBegin, METH_VARARGS, "Set the beginning atom of this bond to ``begin``. Does not update ``begin``."},
	 { "OBBond_SetEnd", _wrap_OBBond_SetEnd, METH_VARARGS, "Set the ending atom of this bond to ``end``. Does not update ``end``."},
	 { "OBBond_SetParent", _wrap_OBBond_SetParent, METH_VARARGS, "Set the parent molecule to ``ptr``. Does not update parent."},
	 { "OBBond_SetLength", _wrap_OBBond_SetLength, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Change the bond length to ``length``, while keeping ``fixed`` stationary\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Change the bond length to ``length``, moving both atoms halfway Since: version 2.2\n"
		""},
	 { "OBBond_Set", _wrap_OBBond_Set, METH_VARARGS, "Set the main bond information (i.e., when creating a bond)"},
	 { "OBBond_SetAromatic", _wrap_OBBond_SetAromatic, METH_VARARGS, "Mark that this bond is aromatic. Does not update atoms or validate."},
	 { "OBBond_SetWedge", _wrap_OBBond_SetWedge, METH_VARARGS, "\n"
		"Mark that this bond has 2D \"wedge\" notation (i.e., goes in a positive\n"
		"Z direction from the beginning to end atoms)\n"
		""},
	 { "OBBond_SetHash", _wrap_OBBond_SetHash, METH_VARARGS, "\n"
		"Mark that this bond has 2D \"hash\" notation (i.e., goes in a negative\n"
		"Z direction from the beginning to end atoms)\n"
		""},
	 { "OBBond_SetWedgeOrHash", _wrap_OBBond_SetWedgeOrHash, METH_VARARGS, "Set the WedgeOrHash flag on a bond (??)"},
	 { "OBBond_SetInRing", _wrap_OBBond_SetInRing, METH_VARARGS, "Mark that this bond is in a ring. Primarily for internal use."},
	 { "OBBond_SetClosure", _wrap_OBBond_SetClosure, METH_VARARGS, "\n"
		"Mark that this bond indicates a ring closure when walking the molecule Warning: This is for internal use only. All closure bonds are marked\n"
		"         automatically by lazy evaluation when requesting\n"
		"         OBBond::IsClosure() *\n"
		""},
	 { "OBBond_GetIdx", _wrap_OBBond_GetIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The unique bond index in a molecule.\n"
		""},
	 { "OBBond_GetId", _wrap_OBBond_GetId, METH_O, NULL},
	 { "OBBond_GetBondOrder", _wrap_OBBond_GetBondOrder, METH_O, "\n"
		":rtype: int\n"
		":return: The bond order for the bond\n"
		""},
	 { "OBBond_GetFlags", _wrap_OBBond_GetFlags, METH_O, "\n"
		":rtype: int\n"
		":return: The set of property flags defined for this bond.\n"
		""},
	 { "OBBond_GetBeginAtomIdx", _wrap_OBBond_GetBeginAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The atom index for the end atom in this bond (from OBAtom::GetIdx()\n"
		""},
	 { "OBBond_GetEndAtomIdx", _wrap_OBBond_GetEndAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: The atom index for the end atom in this bond (from OBAtom::GetIdx()\n"
		""},
	 { "OBBond_GetBeginAtom", _wrap_OBBond_GetBeginAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The \"beginning\" atom for this bond\n"
		""},
	 { "OBBond_GetEndAtom", _wrap_OBBond_GetEndAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The \"end\" atom for this bond\n"
		""},
	 { "OBBond_GetNbrAtom", _wrap_OBBond_GetNbrAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: The neighboring atom to ``ptr`` (i.e., the end if ``ptr`` is the start) Warning: If ``ptr`` is not part of the bond, the beginning atom\n"
		"         will always be returned *\n"
		""},
	 { "OBBond_GetParent", _wrap_OBBond_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: The enclosing OBMol for this bond, or NULL if none is defined.\n"
		""},
	 { "OBBond_GetEquibLength", _wrap_OBBond_GetEquibLength, METH_O, "\n"
		":rtype: float\n"
		":return: The expected \"equilibrium\" length based on the covalent radii and bond order Length is given in Angstroms *\n"
		""},
	 { "OBBond_GetLength", _wrap_OBBond_GetLength, METH_O, "\n"
		":rtype: float\n"
		":return: The current length of this bond in Angstroms\n"
		""},
	 { "OBBond_GetNbrAtomIdx", _wrap_OBBond_GetNbrAtomIdx, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The index to the neighboring atom of ``ptr`` (i.e., the end if ``ptr`` is the start) Warning: If ``ptr`` is not part of the bond, the beginning atom\n"
		"         index will always be returned *\n"
		""},
	 { "OBBond_FindSmallestRing", _wrap_OBBond_FindSmallestRing, METH_O, "Find the smallest ring containing this bond (returns a NULL pointer if none exists)"},
	 { "OBBond_IsAromatic", _wrap_OBBond_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond aromatic?  (Note that the two atoms of the bond may be aromatic,   but not the bond)\n"
		""},
	 { "OBBond_IsInRing", _wrap_OBBond_IsInRing, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond part of a ring?\n"
		""},
	 { "OBBond_IsRotor", _wrap_OBBond_IsRotor, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Is the bond a rotatable bond?  Currently, this function classifies any bond with at least one heavy\n"
		"              atom, no sp-hybrid atoms (e.g., a triple bond somewhere) not in a ring\n"
		"              as a potential rotor if includeRingsBonds is false.  If true, rotors in\n"
		"              rings with more than 3 atoms may be included. No other bond typing is attempted.\n"
		"              For more detailed rotor detection, check the OBRotorList and\n"
		"              OBRotorRules classes *\n"
		""},
	 { "OBBond_IsPeriodic", _wrap_OBBond_IsPeriodic, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond an amide link (i.e., between a carbonyl C and a N)?\n"
		"              No distinction is made between primary, secondary, and tertiary amides. *\n"
		""},
	 { "OBBond_IsAmide", _wrap_OBBond_IsAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond within a periodic unit cell?\n"
		""},
	 { "OBBond_IsPrimaryAmide", _wrap_OBBond_IsPrimaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a primary amide (i.e., between carbonyl C and a NH2)?\n"
		"              In versions prior to 2.3, this function incorrectly identified secondary amides. *\n"
		""},
	 { "OBBond_IsSecondaryAmide", _wrap_OBBond_IsSecondaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a secondary amide (i.e., between a carbonyl C and a NH1)?\n"
		"              In versions prior to 2.3, this function incorrectly identified tertiary amides. *\n"
		""},
	 { "OBBond_IsTertiaryAmide", _wrap_OBBond_IsTertiaryAmide, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a teriary amide (i.e., between a carbonyl C and a NH0)?  Since: version 2.3.\n"
		""},
	 { "OBBond_IsEster", _wrap_OBBond_IsEster, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond an ester link (i.e., between a carbonyl C and an O)?\n"
		""},
	 { "OBBond_IsCarbonyl", _wrap_OBBond_IsCarbonyl, METH_O, "\n"
		":rtype: boolean\n"
		":return: Is the bond a carbonyl C=O?\n"
		""},
	 { "OBBond_IsClosure", _wrap_OBBond_IsClosure, METH_O, "\n"
		":rtype: boolean\n"
		":return: Does this bond \"close\" a ring when walking the molecular graph?\n"
		""},
	 { "OBBond_IsWedge", _wrap_OBBond_IsWedge, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this bond is a \"wedge\" in 2D representations\n"
		"             (i.e., goes in a positive Z direction from the beginning to end atoms) *\n"
		""},
	 { "OBBond_IsHash", _wrap_OBBond_IsHash, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether this bond is a \"hash\" in 2D representations\n"
		"             (i.e., goes in a negative Z direction from the beginning to end atoms) *\n"
		""},
	 { "OBBond_IsWedgeOrHash", _wrap_OBBond_IsWedgeOrHash, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this bond is either a wedge or hash.\n"
		"Notes: : This is a separate bond type\n"
		"Since: version 2.3\n"
		""},
	 { "OBBond_IsCisOrTrans", _wrap_OBBond_IsCisOrTrans, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this bond is either a cis or trans.\n"
		"Since: version 2.3\n"
		""},
	 { "OBBond_IsDoubleBondGeometry", _wrap_OBBond_IsDoubleBondGeometry, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether the geometry around this bond \"looks\" unsaturated\n"
		""},
	 { "OBBond_swigregister", OBBond_swigregister, METH_O, NULL},
	 { "OBBond_swiginit", OBBond_swiginit, METH_VARARGS, NULL},
	 { "new_OBReaction", _wrap_new_OBReaction, METH_NOARGS, NULL},
	 { "OBReaction_NumReactants", _wrap_OBReaction_NumReactants, METH_O, NULL},
	 { "OBReaction_NumProducts", _wrap_OBReaction_NumProducts, METH_O, NULL},
	 { "OBReaction_NumAgents", _wrap_OBReaction_NumAgents, METH_O, NULL},
	 { "OBReaction_AddReactant", _wrap_OBReaction_AddReactant, METH_VARARGS, NULL},
	 { "OBReaction_AddProduct", _wrap_OBReaction_AddProduct, METH_VARARGS, NULL},
	 { "OBReaction_SetTransitionState", _wrap_OBReaction_SetTransitionState, METH_VARARGS, NULL},
	 { "OBReaction_AddAgent", _wrap_OBReaction_AddAgent, METH_VARARGS, NULL},
	 { "OBReaction_GetReactant", _wrap_OBReaction_GetReactant, METH_VARARGS, NULL},
	 { "OBReaction_GetProduct", _wrap_OBReaction_GetProduct, METH_VARARGS, NULL},
	 { "OBReaction_GetAgent", _wrap_OBReaction_GetAgent, METH_VARARGS, NULL},
	 { "OBReaction_GetTransitionState", _wrap_OBReaction_GetTransitionState, METH_O, NULL},
	 { "OBReaction_GetTitle", _wrap_OBReaction_GetTitle, METH_O, NULL},
	 { "OBReaction_GetComment", _wrap_OBReaction_GetComment, METH_O, NULL},
	 { "OBReaction_SetTitle", _wrap_OBReaction_SetTitle, METH_VARARGS, NULL},
	 { "OBReaction_SetComment", _wrap_OBReaction_SetComment, METH_VARARGS, NULL},
	 { "OBReaction_IsReversible", _wrap_OBReaction_IsReversible, METH_O, NULL},
	 { "OBReaction_SetReversible", _wrap_OBReaction_SetReversible, METH_VARARGS, NULL},
	 { "OBReaction_ClassDescription", _wrap_OBReaction_ClassDescription, METH_NOARGS, NULL},
	 { "delete_OBReaction", _wrap_delete_OBReaction, METH_O, NULL},
	 { "OBReaction_swigregister", OBReaction_swigregister, METH_O, NULL},
	 { "OBReaction_swiginit", OBReaction_swiginit, METH_VARARGS, NULL},
	 { "new_OBReactionFacade", _wrap_new_OBReactionFacade, METH_O, "\n"
		"The constructor requires an OBMol. Member functions retrieve or manipulate reaction\n"
		"data stored in this molecule.\n"
		""},
	 { "delete_OBReactionFacade", _wrap_delete_OBReactionFacade, METH_O, NULL},
	 { "OBReactionFacade_AssignComponentIds", _wrap_OBReactionFacade_AssignComponentIds, METH_VARARGS, "\n"
		"These are convenience functions that set/get reaction-related data on individual atoms.\n"
		"The actual data is stored in the underlying OBMol as OBPairInteger data with the\n"
		"keys \"rxnrole\" or \"rxncomp\" (reaction role, reaction component).\n"
		"The other methods of this class are implemented using these methods. Care should be\n"
		"taken using the setter functions, as (unlike the other\n"
		"methods) their use can result in an inconsistent state. IsValid() may be used to check\n"
		"whether this is the case.\n"
		"\n"
		"\n"
		"Assigns a component Id to every atom in the molecule based on connected components.\n"
		"If ``wipe`` is ``false``, then any atoms that already have a component Id are skipped.\n"
		""},
	 { "OBReactionFacade_GetRole", _wrap_OBReactionFacade_GetRole, METH_VARARGS, "Return the reaction role of an atom."},
	 { "OBReactionFacade_GetComponentId", _wrap_OBReactionFacade_GetComponentId, METH_VARARGS, "Return the component Id of an atom."},
	 { "OBReactionFacade_SetRole", _wrap_OBReactionFacade_SetRole, METH_VARARGS, "Set the reaction role of an atom."},
	 { "OBReactionFacade_SetComponentId", _wrap_OBReactionFacade_SetComponentId, METH_VARARGS, "Set the component Id of an atom."},
	 { "OBReactionFacade_IsValid", _wrap_OBReactionFacade_IsValid, METH_O, "\n"
		"Check whether this OBMol is in a valid state to represent a reaction.\n"
		"\n"
		"Note, first of all, that this does not check whether a reaction is\n"
		"chemically valid. It simply checks whether the OBMol has the following\n"
		"properties:\n"
		"- (a) it is marked as a reaction\n"
		"- (b) every atom is marked as having a reaction role\n"
		"- (c) every atom is marked as belonging to a particular reaction component\n"
		"- (d) every atom in a connected component is marked as belong to the same\n"
		"      reaction component and having the same reaction role\n"
		"\n"
		"If (a) is not true, while it does not affect any of the methods of the\n"
		"OBReactionFacade, the reaction may not be written appropriately by an\n"
		"output format.\n"
		"\n"
		"If (b), (c) or (d) is not true, then the behavior of the toolkit is undefined\n"
		"when handling this molecule.\n"
		"\n"
		"While not checked by this function, it is also a good idea to avoid reusing\n"
		"the same component Id for different reaction roles, e.g. having a reactant\n"
		"with component Id 1 as well as a product with component Id 1. If you reassign\n"
		"roles, then the duplication of component Ids may cause two components to\n"
		"be merged into one.\n"
		""},
	 { "OBReactionFacade_AddComponent", _wrap_OBReactionFacade_AddComponent, METH_VARARGS, "\n"
		"These are a set of methods that manipulate reactions at the component level.\n"
		"If you use these high-level methods, and subsequently modify the underlying\n"
		"OBMol or its reaction data, you may need to clear cached information on the\n"
		"reaction components via ClearInternalState().\n"
		"\n"
		"\n"
		"Add a new component to the reaction\n"
		""},
	 { "OBReactionFacade_ClearInternalState", _wrap_OBReactionFacade_ClearInternalState, METH_O, "\n"
		"Clear the list of found components.\n"
		"\n"
		"On first\n"
		"use any of the high-level methods,\n"
		"the number of components for each reaction role is\n"
		"determined and each is assigned a number from 0 to the number of components.\n"
		"This information is cached for future accesses. If you modify the underlying\n"
		"OBMol in a way that invalidates this information (e.g. using the low-level\n"
		"methods), you should call ClearInternalState() or create a new OBReactionFacade.\n"
		""},
	 { "OBReactionFacade_GetComponent", _wrap_OBReactionFacade_GetComponent, METH_VARARGS, "\n"
		"Copy a component from the reaction into the provided OBMol.\n"
		"\n"
		"Note that all\n"
		"reaction data is copied, and so this may be useful for building up a new\n"
		"reaction composed of just some of the components of the original reaction.\n"
		"If copying components from multiple reactions into a single one, to avoid\n"
		"duplicate component Ids it may be easier to copy into an empty OBMol, and\n"
		"then use AddComponent() to add this OBMol into the new reaction.\n"
		"\n"
		":rtype: boolean\n"
		":return: A boolean indicating whether the reaction has a component with that number.\n"
		""},
	 { "OBReactionFacade_NumComponents", _wrap_OBReactionFacade_NumComponents, METH_VARARGS, "Return the number of reaction components with a particular reaction role."},
	 { "OBReactionFacade_ReassignComponent", _wrap_OBReactionFacade_ReassignComponent, METH_VARARGS, "\n"
		"Reassign the reaction role of a given component. If assigned to #NO_REACTIONROLE,\n"
		"this may be used to hide the reaction component when writing to a reaction\n"
		"output format.\n"
		"\n"
		"It should be noted that reassigning a component changes the numbers assigned.\n"
		"For example,\n"
		"in a reaction with 2 reactants and 2 products (each numbered from 0 to 1), if you\n"
		"reassign reactant number\n"
		"0 to be a product, it will now be product number 2 while former reactant number 1\n"
		"will now be reactant number 0.\n"
		"\n"
		":rtype: boolean\n"
		":return: A boolean indicating whether the reaction has a component with that number.\n"
		""},
	 { "OBReactionFacade_swigregister", OBReactionFacade_swigregister, METH_O, NULL},
	 { "OBReactionFacade_swiginit", OBReactionFacade_swiginit, METH_VARARGS, NULL},
	 { "new_OBMol", _wrap_new_OBMol, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor, copies atoms,bonds and OBGenericData\n"
		""},
	 { "delete_OBMol", _wrap_delete_OBMol, METH_O, "Destructor"},
	 { "OBMol___iadd__", _wrap_OBMol___iadd__, METH_VARARGS, "Copies atoms and bonds but not OBGenericData"},
	 { "OBMol_ReserveAtoms", _wrap_OBMol_ReserveAtoms, METH_VARARGS, "Reserve a minimum number of atoms for internal storage This improves performance since the internal atom vector does not grow. Reservation is purely a hint, so an unreasonably large value (e.g. a count read from a malformed file) is silently clamped to a sane upper bound to avoid OOM on untrusted input."},
	 { "OBMol_DestroyAtom", _wrap_OBMol_DestroyAtom, METH_VARARGS, "Free an OBAtom pointer if defined. Does no bookkeeping See also: DeleteAtom which ensures internal connections"},
	 { "OBMol_DestroyBond", _wrap_OBMol_DestroyBond, METH_VARARGS, "Free an OBBond pointer if defined. Does no bookkeeping See also: DeleteBond which ensures internal connections"},
	 { "OBMol_DestroyResidue", _wrap_OBMol_DestroyResidue, METH_VARARGS, "Free an OBResidue pointer if defined. Does no bookkeeping See also: DeleteResidue which ensures internal connections"},
	 { "OBMol_AddAtom", _wrap_OBMol_AddAtom, METH_VARARGS, "\n"
		" Add the specified atom to this molecule :type atom: :py:class:`OBAtom`\n"
		":param atom:        the atom to add :type forceNewId: boolean, optional\n"
		":param forceNewId:  whether to make a new atom Id even if the atom already has one (default is false) :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_InsertAtom", _wrap_OBMol_InsertAtom, METH_VARARGS, "Add a new atom to this molecule (like AddAtom) Calls BeginModify() before insertion and EndModify() after insertion"},
	 { "OBMol_AddBond", _wrap_OBMol_AddBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Add a new bond to the molecule with the specified parameters :type beginIdx: int\n"
		":param beginIdx:  the atom index of the \"start\" atom :type endIdx: int\n"
		":param endIdx:    the atom index of the \"end\" atom :type order: int\n"
		":param order:     the bond order (see OBBond::GetBondOrder()) :type flags: int, optional\n"
		":param flags:     any bond flags such as stereochemistry (default = none) :type insertpos: int, optional\n"
		":param insertpos: the position index to insert the bond (default = none) :rtype: boolean\n"
		":return: Whether the new bond creation was successful\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Add the specified residue to this molecule and update connections :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_AddResidue", _wrap_OBMol_AddResidue, METH_VARARGS, "\n"
		" Add the specified residue to this molecule and update connections :rtype: boolean\n"
		":return: Whether the method was successful\n"
		""},
	 { "OBMol_NewAtom", _wrap_OBMol_NewAtom, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Create a new OBAtom in this molecule and ensure connections (e.g. OBAtom::GetParent(). A new unique id will be assigned to this atom.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Create a new OBAtom in this molecule and ensure connections. (e.g. OBAtom::GetParent(). The ``id`` will be assigned to this atom.\n"
		""},
	 { "OBMol_NewBond", _wrap_OBMol_NewBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Create a new OBBond in this molecule and ensure connections (e.g. OBBond::GetParent(). A new unique id will be assigned to this bond.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Create a new OBBond in this molecule and ensure connections (e.g. OBBond::GetParent(). The ``id`` will be assigned to this bond.\n"
		""},
	 { "OBMol_NewResidue", _wrap_OBMol_NewResidue, METH_O, "Create a new OBResidue in this molecule and ensure connections."},
	 { "OBMol_DeleteAtom", _wrap_OBMol_DeleteAtom, METH_VARARGS, "\n"
		" Deletes an atom from this molecule and all appropriate bonds. Updates the molecule and atom and bond indexes accordingly. Warning: Does not update any residues which may contain this atom :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_DeleteBond", _wrap_OBMol_DeleteBond, METH_VARARGS, "\n"
		" Deletes an bond from this molecule and updates accordingly :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_DeleteResidue", _wrap_OBMol_DeleteResidue, METH_VARARGS, "\n"
		" Deletes a residue from this molecule and updates accordingly. :rtype: boolean\n"
		":return: Whether deletion was successful\n"
		""},
	 { "OBMol_BeginModify", _wrap_OBMol_BeginModify, METH_O, "Call when making many modifications -- clears conformer/rotomer data. The method \"turns off\" perception routines, improving performance. Changes in molecular structure will be re-considered after modifications."},
	 { "OBMol_EndModify", _wrap_OBMol_EndModify, METH_VARARGS, "Call when done with modificaions -- re-perceive data as needed. This method \"turns on\" perception routines and re-evaluates molecular structure."},
	 { "OBMol_GetMod", _wrap_OBMol_GetMod, METH_O, "\n"
		":rtype: int\n"
		":return: The number of nested BeginModify() calls. Used internally.\n"
		""},
	 { "OBMol_IncrementMod", _wrap_OBMol_IncrementMod, METH_O, "Increase the number of nested BeginModify calls. Dangerous! Instead, properly use BeginModify as needed."},
	 { "OBMol_DecrementMod", _wrap_OBMol_DecrementMod, METH_O, "Decrease the number of nested BeginModify calls. Dangerous! Instead, properly use EndModify as needed."},
	 { "OBMol_GetFlags", _wrap_OBMol_GetFlags, METH_O, "\n"
		":rtype: int\n"
		":return: the entire set of flags. (Internal use, mainly.)\n"
		""},
	 { "OBMol_GetTitle", _wrap_OBMol_GetTitle, METH_VARARGS, "\n"
		" :rtype: char\n"
		" :return: the title of this molecule (often the filename) :type replaceNewlines: boolean, optional\n"
		":param replaceNewlines: whether to replace any newline characters with spaces\n"
		""},
	 { "OBMol_NumAtoms", _wrap_OBMol_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of atoms (i.e. OBAtom children)\n"
		""},
	 { "OBMol_NumBonds", _wrap_OBMol_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the number of bonds (i.e. OBBond children)\n"
		""},
	 { "OBMol_NumHvyAtoms", _wrap_OBMol_NumHvyAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of non-hydrogen atoms\n"
		""},
	 { "OBMol_NumResidues", _wrap_OBMol_NumResidues, METH_O, "\n"
		":rtype: int\n"
		":return: the number of residues (i.e. OBResidue substituents)\n"
		""},
	 { "OBMol_NumRotors", _wrap_OBMol_NumRotors, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the number of rotatable bonds. If sampleRingBonds is true, will include rotors within rings (see OBBond::IsRotor() for details)\n"
		""},
	 { "OBMol_GetAtom", _wrap_OBMol_GetAtom, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the atom at index ``idx`` or NULL if it does not exist. Warning: Atom indexing will change. Use iterator methods instead.\n"
		""},
	 { "OBMol_GetAtomById", _wrap_OBMol_GetAtomById, METH_VARARGS, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the atom with ``id`` or NULL if it does not exist.\n"
		""},
	 { "OBMol_GetFirstAtom", _wrap_OBMol_GetFirstAtom, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: the first atom in this molecule, or NULL if none exist. Deprecated: Will be removed in favor of more standard iterator methods\n"
		""},
	 { "OBMol_GetBondById", _wrap_OBMol_GetBondById, METH_VARARGS, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond with ``id`` or NULL if it does not exist.\n"
		""},
	 { "OBMol_GetBond", _wrap_OBMol_GetBond, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond at index ``idx`` or NULL if it does not exist. Warning: Bond indexing may change. Use iterator methods instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond connecting the atom indexed by ``a`` and ``b`` or NULL if none exists. Warning: Atom indexing will change. Use atom objects and iterators instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: :py:class:`OBBond`\n"
		":return: the bond between the atoms ``bgn`` and ``end`` or NULL if none exists\n"
		""},
	 { "OBMol_GetResidue", _wrap_OBMol_GetResidue, METH_VARARGS, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: the residue indexed by ``idx``, or NULL if none exists Warning: Residue indexing may change. Use iterator methods instead.\n"
		""},
	 { "OBMol_GetInternalCoord", _wrap_OBMol_GetInternalCoord, METH_O, NULL},
	 { "OBMol_GetTorsion", _wrap_OBMol_GetTorsion, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the dihedral angle (in degrees) between the four atoms supplied a1-a2-a3-a4)\n"
		"    WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"    in degrees\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: float\n"
		":return: the dihedral angle (in degrees) between the four atoms ``a``, ``b``, ``c``, and ``d)``\n"
		"    WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"    in degrees\n"
		""},
	 { "OBMol_GetAngle", _wrap_OBMol_GetAngle, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the angle (in degrees) between the three atoms ``a``, ``b`` and ``c``  (where  a-> b (vertex) -> c )\n"
		""},
	 { "OBMol_AreInSameRing", _wrap_OBMol_AreInSameRing, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the size of the smallest ring if a and b are in the same ring, 0 otherwise Since: version 2.4\n"
		""},
	 { "OBMol_GetFormula", _wrap_OBMol_GetFormula, METH_O, "\n"
		":rtype: string\n"
		":return: the stochoimetric formula (e.g., C4H6O)\n"
		""},
	 { "OBMol_GetSpacedFormula", _wrap_OBMol_GetSpacedFormula, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: the stochoimetric formula in spaced format e.g. C 4 H 6 O 1\n"
		""},
	 { "OBMol_GetMolWt", _wrap_OBMol_GetMolWt, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the standard molar mass given by IUPAC atomic masses (amu)\n"
		""},
	 { "OBMol_GetExactMass", _wrap_OBMol_GetExactMass, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the mass given by isotopes (or most abundant isotope, if not specified)\n"
		""},
	 { "OBMol_GetTotalCharge", _wrap_OBMol_GetTotalCharge, METH_O, "\n"
		":rtype: int\n"
		":return: the total charge on this molecule (i.e., 0 = neutral, +1, -1...)\n"
		""},
	 { "OBMol_GetTotalSpinMultiplicity", _wrap_OBMol_GetTotalSpinMultiplicity, METH_O, "\n"
		":rtype: int\n"
		":return: the total spin on this molecule (i.e., 1 = singlet, 2 = doublet...)\n"
		""},
	 { "OBMol_GetDimension", _wrap_OBMol_GetDimension, METH_O, "\n"
		":rtype: int\n"
		":return: the dimensionality of coordinates (i.e., 0 = unknown or no coord, 2=2D, 3=3D)\n"
		""},
	 { "OBMol_GetCoordinates", _wrap_OBMol_GetCoordinates, METH_O, "\n"
		":rtype: float\n"
		":return: the set of all atomic coordinates. See OBAtom::GetCoordPtr for more\n"
		""},
	 { "OBMol_GetSSSR", _wrap_OBMol_GetSSSR, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >\n"
		":return: the Smallest Set of Smallest Rings has been run (see OBRing class)\n"
		""},
	 { "OBMol_GetLSSR", _wrap_OBMol_GetLSSR, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >\n"
		":return: the Largest Set of Smallest Rings has been run (see OBRing class)\n"
		""},
	 { "OBMol_AutomaticFormalCharge", _wrap_OBMol_AutomaticFormalCharge, METH_O, "Get the current flag for whether formal charges are set with pH correction"},
	 { "OBMol_AutomaticPartialCharge", _wrap_OBMol_AutomaticPartialCharge, METH_O, "Get the current flag for whether partial charges are auto-determined"},
	 { "OBMol_SetTitle", _wrap_OBMol_SetTitle, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the title of this molecule to ``title``\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the title of this molecule to ``title``\n"
		""},
	 { "OBMol_SetFormula", _wrap_OBMol_SetFormula, METH_VARARGS, "Set the stochiometric formula for this molecule"},
	 { "OBMol_SetEnergy", _wrap_OBMol_SetEnergy, METH_VARARGS, "Set the heat of formation for this molecule (in kcal/mol)"},
	 { "OBMol_SetDimension", _wrap_OBMol_SetDimension, METH_VARARGS, "Set the dimension of this molecule (i.e., 0, 1 , 2, 3)"},
	 { "OBMol_SetTotalCharge", _wrap_OBMol_SetTotalCharge, METH_VARARGS, "Set the total charge of this molecule to ``charge``"},
	 { "OBMol_SetTotalSpinMultiplicity", _wrap_OBMol_SetTotalSpinMultiplicity, METH_VARARGS, "Set the total spin multiplicity of this molecule to ``spinMultiplicity`` Overrides the calculation from spin multiplicity of OBAtoms"},
	 { "OBMol_SetInternalCoord", _wrap_OBMol_SetInternalCoord, METH_VARARGS, "Set the internal coordinates to ``int_coord`` (Does not call InternalToCartesian to update the 3D cartesian coordinates). The size of the ``int_coord`` has to be the same as the number of atoms in molecule (+ NULL at the beginning)."},
	 { "OBMol_SetAutomaticFormalCharge", _wrap_OBMol_SetAutomaticFormalCharge, METH_VARARGS, "Set the flag for determining automatic formal charges with pH (default=true)"},
	 { "OBMol_SetAutomaticPartialCharge", _wrap_OBMol_SetAutomaticPartialCharge, METH_VARARGS, "Set the flag for determining partial charges automatically (default=true)"},
	 { "OBMol_SetAromaticPerceived", _wrap_OBMol_SetAromaticPerceived, METH_VARARGS, "Mark that aromaticity has been perceived for this molecule (see OBAromaticTyper)"},
	 { "OBMol_SetSSSRPerceived", _wrap_OBMol_SetSSSRPerceived, METH_VARARGS, "Mark that Smallest Set of Smallest Rings has been run (see OBRing class)"},
	 { "OBMol_SetLSSRPerceived", _wrap_OBMol_SetLSSRPerceived, METH_VARARGS, "Mark that Largest Set of Smallest Rings has been run (see OBRing class)"},
	 { "OBMol_SetRingAtomsAndBondsPerceived", _wrap_OBMol_SetRingAtomsAndBondsPerceived, METH_VARARGS, "Mark that rings have been perceived (see OBRing class for details)"},
	 { "OBMol_SetAtomTypesPerceived", _wrap_OBMol_SetAtomTypesPerceived, METH_VARARGS, "Mark that atom types have been perceived (see OBAtomTyper for details)"},
	 { "OBMol_SetRingTypesPerceived", _wrap_OBMol_SetRingTypesPerceived, METH_VARARGS, "Mark that ring types have been perceived (see OBRingTyper for details)"},
	 { "OBMol_SetChainsPerceived", _wrap_OBMol_SetChainsPerceived, METH_VARARGS, "Mark that chains and residues have been perceived (see OBChainsParser)"},
	 { "OBMol_SetChiralityPerceived", _wrap_OBMol_SetChiralityPerceived, METH_VARARGS, "Mark that chirality has been perceived"},
	 { "OBMol_SetPartialChargesPerceived", _wrap_OBMol_SetPartialChargesPerceived, METH_VARARGS, "Mark that partial charges have been assigned"},
	 { "OBMol_SetHybridizationPerceived", _wrap_OBMol_SetHybridizationPerceived, METH_VARARGS, "Mark that hybridization of all atoms has been assigned"},
	 { "OBMol_SetClosureBondsPerceived", _wrap_OBMol_SetClosureBondsPerceived, METH_VARARGS, "Mark that ring closure bonds have been assigned by graph traversal"},
	 { "OBMol_SetHydrogensAdded", _wrap_OBMol_SetHydrogensAdded, METH_VARARGS, "Mark that explicit hydrogen atoms have been added"},
	 { "OBMol_SetCorrectedForPH", _wrap_OBMol_SetCorrectedForPH, METH_VARARGS, NULL},
	 { "OBMol_SetSpinMultiplicityAssigned", _wrap_OBMol_SetSpinMultiplicityAssigned, METH_VARARGS, NULL},
	 { "OBMol_SetIsPatternStructure", _wrap_OBMol_SetIsPatternStructure, METH_VARARGS, "The OBMol is a pattern, not a complete molecule. Left unchanged by Clear()."},
	 { "OBMol_SetIsReaction", _wrap_OBMol_SetIsReaction, METH_VARARGS, NULL},
	 { "OBMol_SetPeriodicMol", _wrap_OBMol_SetPeriodicMol, METH_VARARGS, "Mark that distance calculations, etc., should apply periodic boundary conditions through the minimimum image convention. Does not automatically recalculate bonding."},
	 { "OBMol_HasFlag", _wrap_OBMol_HasFlag, METH_VARARGS, NULL},
	 { "OBMol_SetFlag", _wrap_OBMol_SetFlag, METH_VARARGS, NULL},
	 { "OBMol_UnsetFlag", _wrap_OBMol_UnsetFlag, METH_VARARGS, NULL},
	 { "OBMol_SetFlags", _wrap_OBMol_SetFlags, METH_VARARGS, NULL},
	 { "OBMol_ClassDescription", _wrap_OBMol_ClassDescription, METH_NOARGS, NULL},
	 { "OBMol_RenumberAtoms", _wrap_OBMol_RenumberAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Renumber the atoms of this molecule according to the order in the supplied vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Renumber the atoms of this molecule using the initial indexes in the supplied vector\n"
		""},
	 { "OBMol_SetCoordinates", _wrap_OBMol_SetCoordinates, METH_VARARGS, "Set the coordinates for all atoms in this conformer. See also: OBMol::GetCoordinates()"},
	 { "OBMol_ToInertialFrame", _wrap_OBMol_ToInertialFrame, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translate one conformer and rotate by a rotation matrix (which is returned) to the inertial frame-of-reference\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Translate all conformers to the inertial frame-of-reference\n"
		""},
	 { "OBMol_Translate", _wrap_OBMol_Translate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translates all conformers in the molecule by the supplied vector\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Translates one conformer in the molecule by the supplied vector\n"
		""},
	 { "OBMol_Rotate", _wrap_OBMol_Rotate, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Rotate all conformers using the supplied matrix ``u`` (a 3x3 array of double)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Rotate all conformers using the supplied matrix ``m`` (a linear 3x3 row-major array of double)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Rotate a specific conformer ``nconf`` using the supplied rotation matrix ``m``\n"
		""},
	 { "OBMol_DeleteHydrogens", _wrap_OBMol_DeleteHydrogens, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Suppress hydrogens by converting explicit hydrogen atoms to implicit :rtype: boolean\n"
		":return: Success\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Suppress explicit hydrogen atoms on the supplied atom :rtype: boolean\n"
		":return: Success\n"
		""},
	 { "OBMol_DeletePolarHydrogens", _wrap_OBMol_DeletePolarHydrogens, METH_O, "Suppress explicit hydrogen atoms connected to a polar atom See also: OBAtom::IsPolarHydrogen Since: version 2.4"},
	 { "OBMol_DeleteNonPolarHydrogens", _wrap_OBMol_DeleteNonPolarHydrogens, METH_O, "Suppress explicit hydrogen atoms connected to a non-polar atom See also: OBAtom::IsNonPolarHydrogen"},
	 { "OBMol_DeleteHydrogen", _wrap_OBMol_DeleteHydrogen, METH_VARARGS, "Suppress the supplied atom if it is a hydrogen (Helper function for DeleteHydrogens)"},
	 { "OBMol_AddHydrogens", _wrap_OBMol_AddHydrogens, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Convert implicit hydrogens to explicit atoms in the molecular graph :type polaronly: boolean, optional\n"
		":param polaronly:    Whether to add hydrogens only to polar atoms (i.e., not to C atoms) :type correctForPH: boolean, optional\n"
		":param correctForPH: Whether to call CorrectForPH() first :type pH: float, optional\n"
		":param pH: The pH to use for CorrectForPH() modification :rtype: boolean\n"
		":return: Whether any hydrogens were added\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" For a particular atom, convert implicit hydrogens to explicit atoms in the molecular graph\n"
		""},
	 { "OBMol_AddPolarHydrogens", _wrap_OBMol_AddPolarHydrogens, METH_O, "For polar atoms only, convert implicit hydrogens to explicit atoms in the molecular graph"},
	 { "OBMol_AddNonPolarHydrogens", _wrap_OBMol_AddNonPolarHydrogens, METH_O, "For non-polar atoms only, convert implicit hydrogens to explicit atoms in the molecular graph Since: version 2.4"},
	 { "OBMol_AddNewHydrogens", _wrap_OBMol_AddNewHydrogens, METH_VARARGS, "For polar and/or non-polar atoms, convert implicit hydrogens to explicit atoms in the molecular graph Since: verison 2.4"},
	 { "OBMol_StripSalts", _wrap_OBMol_StripSalts, METH_VARARGS, "If ``threshold`` is not specified or is zero, remove all but the largest contiguous fragment. If ``threshold`` is non-zero, remove any fragments with fewer than ``threshold`` atoms."},
	 { "OBMol_Separate", _wrap_OBMol_Separate, METH_VARARGS, "Copies each disconnected fragment as a separate OBMol"},
	 { "OBMol_GetNextFragment", _wrap_OBMol_GetNextFragment, METH_VARARGS, "Iterative component of Separate to copy one fragment at a time"},
	 { "OBMol_CopySubstructure", _wrap_OBMol_CopySubstructure, METH_VARARGS, NULL},
	 { "OBMol_ConvertDativeBonds", _wrap_OBMol_ConvertDativeBonds, METH_O, "Converts the charged form of coordinate bonds, e.g.[N+]([O-])=O to N(=O)=O"},
	 { "OBMol_MakeDativeBonds", _wrap_OBMol_MakeDativeBonds, METH_O, "\n"
		" Converts 5-valent N and P only. Return true if conversion occurred. :rtype: boolean\n"
		":return: has charged form of dative bonds(e.g.[N+]([O-])=O from N(=O)=O). Since: version 2.4\n"
		""},
	 { "OBMol_ConvertZeroBonds", _wrap_OBMol_ConvertZeroBonds, METH_O, "\n"
		"Convert zero-order bonds to single or double bonds and adjust adjacent atom\n"
		"charges in an attempt to achieve the correct valence state.\n"
		":rtype: boolean\n"
		":return: Whether any modifications were made\n"
		"Since: version 2.4\n"
		""},
	 { "OBMol_CorrectForPH", _wrap_OBMol_CorrectForPH, METH_VARARGS, "Correct for pH by applying the OBPhModel transformations"},
	 { "OBMol_AssignSpinMultiplicity", _wrap_OBMol_AssignSpinMultiplicity, METH_VARARGS, NULL},
	 { "OBMol_AssignTotalChargeToAtoms", _wrap_OBMol_AssignTotalChargeToAtoms, METH_VARARGS, "Put the specified molecular charge on appropriate atoms. Assumes all the hydrogen is explicitly included in the molecule. Since: version 2.4"},
	 { "OBMol_Center", _wrap_OBMol_Center, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Translate to the center of all coordinates (for this conformer)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: :py:class:`vector3`\n"
		":return: the center of the supplied conformer ``nconf`` See also: Center() to actually center all conformers at the origin\n"
		""},
	 { "OBMol_FindSSSR", _wrap_OBMol_FindSSSR, METH_O, "Find Smallest Set of Smallest Rings (see OBRing class for more details)"},
	 { "OBMol_FindLSSR", _wrap_OBMol_FindLSSR, METH_O, "Find Largest Set of Smallest Rings"},
	 { "OBMol_FindRingAtomsAndBonds", _wrap_OBMol_FindRingAtomsAndBonds, METH_O, "Find all ring atoms and bonds. Does not need to call FindSSSR()."},
	 { "OBMol_FindChildren", _wrap_OBMol_FindChildren, METH_VARARGS, NULL},
	 { "OBMol_FindLargestFragment", _wrap_OBMol_FindLargestFragment, METH_VARARGS, "\n"
		" Find the largest fragment in OBMol (which may include multiple non-connected fragments) :type frag: :py:class:`OBBitVec`\n"
		":param frag:   Return (by reference) a bit vector indicating the atoms in the largest fragment\n"
		""},
	 { "OBMol_ContigFragList", _wrap_OBMol_ContigFragList, METH_VARARGS, "Sort a list of contig fragments by size from largest to smallest Each vector<int> contains the atom numbers of a contig fragment"},
	 { "OBMol_Align", _wrap_OBMol_Align, METH_VARARGS, "Aligns atom a on p1 and atom b along p1->p2 vector"},
	 { "OBMol_ConnectTheDots", _wrap_OBMol_ConnectTheDots, METH_O, "Adds single bonds based on atom proximity"},
	 { "OBMol_PerceiveBondOrders", _wrap_OBMol_PerceiveBondOrders, METH_O, "Attempts to perceive multiple bonds based on geometries"},
	 { "OBMol_FindAngles", _wrap_OBMol_FindAngles, METH_O, "Fills out an OBAngleData with angles from the molecule"},
	 { "OBMol_FindTorsions", _wrap_OBMol_FindTorsions, METH_O, "Fills out an OBTorsionData with angles from the molecule"},
	 { "OBMol_GetGTDVector", _wrap_OBMol_GetGTDVector, METH_VARARGS, NULL},
	 { "OBMol_GetGIVector", _wrap_OBMol_GetGIVector, METH_VARARGS, NULL},
	 { "OBMol_GetGIDVector", _wrap_OBMol_GetGIDVector, METH_VARARGS, NULL},
	 { "OBMol_Has2D", _wrap_OBMol_Has2D, METH_VARARGS, "Are there non-zero coordinates in two dimensions (i.e. X and Y)- and, if Not3D is true, no Z coordinates?"},
	 { "OBMol_Has3D", _wrap_OBMol_Has3D, METH_O, "Are there non-zero coordinates in all three dimensions (i.e. X, Y, Z)?"},
	 { "OBMol_HasNonZeroCoords", _wrap_OBMol_HasNonZeroCoords, METH_O, "Are there any non-zero coordinates?"},
	 { "OBMol_HasAromaticPerceived", _wrap_OBMol_HasAromaticPerceived, METH_O, "Has aromatic perception been performed?"},
	 { "OBMol_HasSSSRPerceived", _wrap_OBMol_HasSSSRPerceived, METH_O, "Has the smallest set of smallest rings (FindSSSR) been performed?"},
	 { "OBMol_HasLSSRPerceived", _wrap_OBMol_HasLSSRPerceived, METH_O, "Has the largest set of smallest rings (FindLSSR) been performed?"},
	 { "OBMol_HasRingAtomsAndBondsPerceived", _wrap_OBMol_HasRingAtomsAndBondsPerceived, METH_O, "Have ring atoms and bonds been assigned?"},
	 { "OBMol_HasAtomTypesPerceived", _wrap_OBMol_HasAtomTypesPerceived, METH_O, "Have atom types been assigned by OBAtomTyper?"},
	 { "OBMol_HasRingTypesPerceived", _wrap_OBMol_HasRingTypesPerceived, METH_O, "Have ring types been assigned by OBRingTyper?"},
	 { "OBMol_HasChiralityPerceived", _wrap_OBMol_HasChiralityPerceived, METH_O, "Has atom chirality been assigned?"},
	 { "OBMol_HasPartialChargesPerceived", _wrap_OBMol_HasPartialChargesPerceived, METH_O, "Have atomic Gasteiger partial charges been assigned by OBGastChrg?"},
	 { "OBMol_HasHybridizationPerceived", _wrap_OBMol_HasHybridizationPerceived, METH_O, "Has atomic hybridization been assigned by OBAtomTyper?"},
	 { "OBMol_HasClosureBondsPerceived", _wrap_OBMol_HasClosureBondsPerceived, METH_O, "Have ring \"closure\" bonds been assigned? (e.g., OBBond::IsClosure())"},
	 { "OBMol_HasChainsPerceived", _wrap_OBMol_HasChainsPerceived, METH_O, "Have biomolecule chains and residues been assigned by OBChainsParser?"},
	 { "OBMol_HasHydrogensAdded", _wrap_OBMol_HasHydrogensAdded, METH_O, "Have hydrogens been added to the molecule?"},
	 { "OBMol_IsCorrectedForPH", _wrap_OBMol_IsCorrectedForPH, METH_O, "Has the molecule been corrected for pH by CorrectForPH?"},
	 { "OBMol_HasSpinMultiplicityAssigned", _wrap_OBMol_HasSpinMultiplicityAssigned, METH_O, "Has total spin multiplicity been assigned?"},
	 { "OBMol_IsReaction", _wrap_OBMol_IsReaction, METH_O, "Does this OBMol represent a reaction?"},
	 { "OBMol_IsPeriodic", _wrap_OBMol_IsPeriodic, METH_O, "Is this molecule periodic? Should periodic boundary conditions be applied?"},
	 { "OBMol_Empty", _wrap_OBMol_Empty, METH_O, "Are there any atoms in this molecule?"},
	 { "OBMol_NumConformers", _wrap_OBMol_NumConformers, METH_O, "\n"
		":rtype: int\n"
		":return: the number of conformers in this molecule\n"
		""},
	 { "OBMol_SetConformers", _wrap_OBMol_SetConformers, METH_VARARGS, "Set the entire set of conformers for this molecule to ``v``"},
	 { "OBMol_AddConformer", _wrap_OBMol_AddConformer, METH_VARARGS, "Add a new set of coordinates ``f`` as a new conformer"},
	 { "OBMol_SetConformer", _wrap_OBMol_SetConformer, METH_VARARGS, "Set the molecule's current conformer to ``i`` Does nothing if ``i`` is larger than NumConformers()"},
	 { "OBMol_CopyConformer", _wrap_OBMol_CopyConformer, METH_VARARGS, "Copy the conformer ``nconf`` into the array ``c`` Warning: Does no checking to see if ``c`` is large enough"},
	 { "OBMol_DeleteConformer", _wrap_OBMol_DeleteConformer, METH_VARARGS, "Delete the conformer ``nconf``"},
	 { "OBMol_GetConformer", _wrap_OBMol_GetConformer, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the coordinates to conformer ``i``\n"
		""},
	 { "OBMol_SetEnergies", _wrap_OBMol_SetEnergies, METH_VARARGS, "Set the entire set of conformer energies"},
	 { "OBMol_GetEnergies", _wrap_OBMol_GetEnergies, METH_O, "Set the entire set of conformer energies"},
	 { "OBMol_GetEnergy", _wrap_OBMol_GetEnergy, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the heat of formation for this molecule (in kcal/mol)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Get the energy for conformer ci Title: ci conformer index\n"
		""},
	 { "OBMol_BeginConformer", _wrap_OBMol_BeginConformer, METH_VARARGS, "\n"
		" Set the iterator to the beginning of the conformer list :rtype: float\n"
		":return: the array of coordinates for the first conformer\n"
		""},
	 { "OBMol_NextConformer", _wrap_OBMol_NextConformer, METH_VARARGS, "\n"
		" Advance the iterator to the next confomer, if possible :rtype: float\n"
		":return: The array of coordinates for the next conformer, or NULL if none exist\n"
		""},
	 { "OBMol_GetConformers", _wrap_OBMol_GetConformers, METH_O, "\n"
		":rtype: std::vector< double *,std::allocator< double * > >\n"
		":return: the entire set of conformers for this molecule as a vector of floating point arrays\n"
		""},
	 { "OBMol_CBeginAtoms", _wrap_OBMol_CBeginAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant atom iterator pointing to the beginning of the atom list\n"
		""},
	 { "OBMol_CEndAtoms", _wrap_OBMol_CEndAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBAtomConstIterator\n"
		":return: A constant atom iterator pointing to the end of the atom list\n"
		""},
	 { "OBMol_GetAtoms", _wrap_OBMol_GetAtoms, METH_O, "\n"
		":rtype: OpenBabel::OBMolAtomRange\n"
		":return: A range over the atoms. This range can be used in a range-based for loop.\n"
		""},
	 { "OBMol_GetBonds", _wrap_OBMol_GetBonds, METH_O, "\n"
		":rtype: OpenBabel::OBMolBondRange\n"
		":return: A range over the bonds. This range can be used in a range-based for loop.\n"
		""},
	 { "OBMol_BeginInternalCoord", _wrap_OBMol_BeginInternalCoord, METH_VARARGS, "\n"
		" Set the iterator to the beginning of the internal coordinate list :rtype: :py:class:`OBInternalCoord`\n"
		":return: the first internal coordinate record, or NULL if none exist See also: SetInternalCoord\n"
		""},
	 { "OBMol_NextInternalCoord", _wrap_OBMol_NextInternalCoord, METH_VARARGS, "\n"
		" Advance the iterator to the next internal coordinate record :rtype: :py:class:`OBInternalCoord`\n"
		":return: the next first internal coordinate record, or NULL if none exist See also: SetInternalCoord\n"
		""},
	 { "OBMol_SetTorsion", _wrap_OBMol_SetTorsion, METH_VARARGS, "\n"
		"Set the torsion defined by these atoms, rotating bonded neighbors\n"
		"Title: ang The torsion angle in radians\n"
		"WARNING: SetTorsion takes an angle in radians while GetTorsion returns it\n"
		"in degrees\n"
		""},
	 { "OBMol_swigregister", OBMol_swigregister, METH_O, NULL},
	 { "OBMol_swiginit", OBMol_swiginit, METH_VARARGS, NULL},
	 { "ThrowError", _wrap_ThrowError, METH_VARARGS, NULL},
	 { "CartesianToInternal", _wrap_CartesianToInternal, METH_VARARGS, "Convert Cartesian XYZ to a set of OBInternalCoord coordinates"},
	 { "InternalToCartesian", _wrap_InternalToCartesian, METH_VARARGS, "Convert set of OBInternalCoord coordinates into Cartesian XYZ"},
	 { "NewExtension", _wrap_NewExtension, METH_VARARGS, NULL},
	 { "get_rmat", _wrap_get_rmat, METH_VARARGS, NULL},
	 { "ob_make_rmat", _wrap_ob_make_rmat, METH_VARARGS, NULL},
	 { "qtrfit", _wrap_qtrfit, METH_VARARGS, NULL},
	 { "superimpose", _wrap_superimpose, METH_VARARGS, NULL},
	 { "OBRing_ring_id_set", _wrap_OBRing_ring_id_set, METH_VARARGS, "a unique id to ensure a stable sort in SSSR determination"},
	 { "OBRing_ring_id_get", _wrap_OBRing_ring_id_get, METH_O, "a unique id to ensure a stable sort in SSSR determination"},
	 { "OBRing__path_set", _wrap_OBRing__path_set, METH_VARARGS, "the path of this ring (i.e., the atom indexes)"},
	 { "OBRing__path_get", _wrap_OBRing__path_get, METH_O, "the path of this ring (i.e., the atom indexes)"},
	 { "OBRing__pathset_set", _wrap_OBRing__pathset_set, METH_VARARGS, "the path of this ring as a redundant bit vector"},
	 { "OBRing__pathset_get", _wrap_OBRing__pathset_get, METH_O, "the path of this ring as a redundant bit vector"},
	 { "new_OBRing", _wrap_new_OBRing, METH_VARARGS, "Initialize a ring from a set of atom indexes ``path`` and with ``size``"},
	 { "OBRing_Size", _wrap_OBRing_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the size of this ring (i.e., how many atoms in the cycle)\n"
		""},
	 { "OBRing_PathSize", _wrap_OBRing_PathSize, METH_O, "\n"
		":rtype: int\n"
		":return: the size of this ring (i.e., how many atoms in the cycle) Deprecated: Use Size() instead\n"
		""},
	 { "OBRing_IsAromatic", _wrap_OBRing_IsAromatic, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this ring is aromatic If all atoms in this ring are aromatic, the ring will be considered aromatic TODO: This method uses implicit bonding -- bond info is not stored in OBRing\n"
		""},
	 { "OBRing_SetType", _wrap_OBRing_SetType, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Set the ring type (see OBRingTyper for more)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Set the ring type (see OBRingTyper for more)\n"
		""},
	 { "OBRing_GetType", _wrap_OBRing_GetType, METH_O, "\n"
		":rtype: string\n"
		":return: the ring type\n"
		""},
	 { "OBRing_GetRootAtom", _wrap_OBRing_GetRootAtom, METH_O, "\n"
		":rtype: int\n"
		":return: the index for the root atom. O for furan, S for thiazole, N for pyrrole. For 6 membered aromatic rings, the first non carbon atom is used for root. For 5 members rings the O, S or N (BOSum=3, valence=3) will be used for root\n"
		""},
	 { "OBRing_IsMember", _wrap_OBRing_IsMember, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: Whether atom ``a`` is a member of this ring\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: Whether both atoms in bond ``b`` are in this ring TODO: This method uses implicit bonding -- bond info is not stored in OBRing\n"
		""},
	 { "OBRing_IsInRing", _wrap_OBRing_IsInRing, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: Whether ``i`` as an atom index is in this ring\n"
		""},
	 { "OBRing_SetParent", _wrap_OBRing_SetParent, METH_VARARGS, "Set the parent of this ring to ``m``"},
	 { "OBRing_GetParent", _wrap_OBRing_GetParent, METH_O, "\n"
		":rtype: :py:class:`OBMol`\n"
		":return: the parent of this ring, or NULL if none has been defined\n"
		""},
	 { "OBRing_findCenterAndNormal", _wrap_OBRing_findCenterAndNormal, METH_VARARGS, "\n"
		" Set the supplied vectors to the ``center`` of this ring, along with the ``normal`` (in both directions). :type center: :py:class:`vector3`\n"
		":param center: The center of the ring :type norm1: :py:class:`vector3`\n"
		":param norm1: The normal of the best-fit plane for this ring :type norm2: :py:class:`vector3`\n"
		":param norm2: -1 * norm1 (i.e., the opposite direction of norm1) :rtype: boolean\n"
		":return: True (success)\n"
		""},
	 { "delete_OBRing", _wrap_delete_OBRing, METH_O, NULL},
	 { "OBRing_swigregister", OBRing_swigregister, METH_O, NULL},
	 { "OBRing_swiginit", OBRing_swiginit, METH_VARARGS, NULL},
	 { "CompareRingSize", _wrap_CompareRingSize, METH_VARARGS, "\n"
		" Comparison function for rings, used by OBRingSearch::SortRings() :rtype: boolean\n"
		":return: true if a.size() > b.size()\n"
		""},
	 { "new_OBRingSearch", _wrap_new_OBRingSearch, METH_NOARGS, NULL},
	 { "delete_OBRingSearch", _wrap_delete_OBRingSearch, METH_O, NULL},
	 { "OBRingSearch_SortRings", _wrap_OBRingSearch_SortRings, METH_O, "Sort ring sizes from smallest to largest"},
	 { "OBRingSearch_RemoveRedundant", _wrap_OBRingSearch_RemoveRedundant, METH_VARARGS, "Starting with a full ring set - reduce to SSSR set"},
	 { "OBRingSearch_AddRingFromClosure", _wrap_OBRingSearch_AddRingFromClosure, METH_VARARGS, "Add a new ring from a \"closure\" bond: See OBBond::IsClosure()"},
	 { "OBRingSearch_SaveUniqueRing", _wrap_OBRingSearch_SaveUniqueRing, METH_VARARGS, NULL},
	 { "OBRingSearch_WriteRings", _wrap_OBRingSearch_WriteRings, METH_O, "For debugging only, write the rings to std::cout"},
	 { "OBRingSearch_BeginRings", _wrap_OBRingSearch_BeginRings, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator\n"
		":return: an iterator pointing to the beginning of the list of rings\n"
		""},
	 { "OBRingSearch_EndRings", _wrap_OBRingSearch_EndRings, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator\n"
		":return: an iterator pointing to the end of the list of rings\n"
		""},
	 { "OBRingSearch_swigregister", OBRingSearch_swigregister, METH_O, NULL},
	 { "OBRingSearch_swiginit", OBRingSearch_swiginit, METH_VARARGS, NULL},
	 { "new_OBRTree", _wrap_new_OBRTree, METH_VARARGS, "Construct a search tree from a possible parent entry and atom entry"},
	 { "delete_OBRTree", _wrap_delete_OBRTree, METH_O, NULL},
	 { "OBRTree_GetAtomIdx", _wrap_OBRTree_GetAtomIdx, METH_O, "\n"
		":rtype: int\n"
		":return: the OBAtom::GetIdx() index of the atom in this node\n"
		""},
	 { "OBRTree_PathToRoot", _wrap_OBRTree_PathToRoot, METH_VARARGS, "Recursively find the root of this tree, building up a vector of OBAtom nodes."},
	 { "OBRTree_swigregister", OBRTree_swigregister, METH_O, NULL},
	 { "OBRTree_swiginit", OBRTree_swiginit, METH_VARARGS, NULL},
	 { "delete_OBSmartsPattern", _wrap_delete_OBSmartsPattern, METH_O, NULL},
	 { "new_OBSmartsPattern", _wrap_new_OBSmartsPattern, METH_VARARGS, NULL},
	 { "OBSmartsPattern_Init", _wrap_OBSmartsPattern_Init, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Parse the ``pattern`` SMARTS string. :rtype: boolean\n"
		":return: Whether the pattern is a valid SMARTS expression\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Parse the ``pattern`` SMARTS string. :rtype: boolean\n"
		":return: Whether the pattern is a valid SMARTS expression\n"
		""},
	 { "OBSmartsPattern_GetSMARTS", _wrap_OBSmartsPattern_GetSMARTS, METH_O, "\n"
		":rtype: string\n"
		":return: the SMARTS string which is currently used\n"
		""},
	 { "OBSmartsPattern_Empty", _wrap_OBSmartsPattern_Empty, METH_O, "\n"
		" :rtype: boolean\n"
		" :return: the SMARTS string which is currently used :rtype: boolean\n"
		":return: If the SMARTS pattern is an empty expression (e.g., invalid)\n"
		""},
	 { "OBSmartsPattern_IsValid", _wrap_OBSmartsPattern_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: If the SMARTS pattern is a valid expression\n"
		""},
	 { "OBSmartsPattern_NumAtoms", _wrap_OBSmartsPattern_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: the number of atoms in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_NumBonds", _wrap_OBSmartsPattern_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: the number of bonds in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_GetBond", _wrap_OBSmartsPattern_GetBond, METH_VARARGS, "\n"
		" Access the bond ``idx`` in the internal pattern :type src: int\n"
		":param src: The index of the beginning atom :type dst: int\n"
		":param dst: The index of the end atom :type ord: int\n"
		":param ord: The bond order of this bond :type idx: int\n"
		":param idx: The index of the bond in the SMARTS pattern\n"
		""},
	 { "OBSmartsPattern_GetAtomicNum", _wrap_OBSmartsPattern_GetAtomicNum, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the atomic number of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_GetCharge", _wrap_OBSmartsPattern_GetCharge, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the formal charge of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_GetVectorBinding", _wrap_OBSmartsPattern_GetVectorBinding, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: the vector binding of the atom ``idx`` in the internal pattern\n"
		""},
	 { "OBSmartsPattern_Match", _wrap_OBSmartsPattern_Match, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type single: boolean, optional\n"
		":param single: Whether only a single match is required (faster). Default is false. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). This version is (more) thread safe. :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type mlist: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":param mlist: The resulting match list :type mtype: int, optional\n"
		":param mtype: The match type to use. Default is All. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		" Perform SMARTS matching for the pattern specified using Init(). This version is (more) thread safe. :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :type mlist: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":param mlist: The resulting match list :param mtype: The match type to use. Default is All. :rtype: boolean\n"
		":return: Whether matches occurred\n"
		""},
	 { "OBSmartsPattern_HasMatch", _wrap_OBSmartsPattern_HasMatch, METH_VARARGS, "\n"
		" Thread safe check for any SMARTS match :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to use for matching :rtype: boolean\n"
		":return: Whether there exists any match\n"
		""},
	 { "OBSmartsPattern_RestrictedMatch", _wrap_OBSmartsPattern_RestrictedMatch, METH_VARARGS, NULL},
	 { "OBSmartsPattern_NumMatches", _wrap_OBSmartsPattern_NumMatches, METH_O, "\n"
		":rtype: int\n"
		":return: the number of non-unique SMARTS matches To get the number of unique SMARTS matches, query GetUMapList()->size()\n"
		""},
	 { "OBSmartsPattern_GetMapList", _wrap_OBSmartsPattern_GetMapList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":return: the entire list of non-unique matches for this pattern See also: GetUMapList()\n"
		""},
	 { "OBSmartsPattern_BeginMList", _wrap_OBSmartsPattern_BeginMList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator\n"
		":return: An iterator over the (non-unique) match list, starting at the beginning\n"
		""},
	 { "OBSmartsPattern_EndMList", _wrap_OBSmartsPattern_EndMList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >::iterator\n"
		":return: An iterator over the non-unique match list, set to the end\n"
		""},
	 { "OBSmartsPattern_GetUMapList", _wrap_OBSmartsPattern_GetUMapList, METH_O, "\n"
		":rtype: std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > >\n"
		":return: the entire list of unique matches for this pattern\n"
		"          A unique match is defined as one which does not cover the\n"
		"          identical atoms that a previous match has covered.\n"
		"\n"
		"      For instance, the pattern [OD1]~C~[OD1] describes a\n"
		"      carboxylate group. This pattern will match both atom number\n"
		"      permutations of the carboxylate, and if GetMapList() is called, both\n"
		"      matches will be returned. If GetUMapList() is called only unique\n"
		"      matches of the pattern will be returned.\n"
		""},
	 { "OBSmartsPattern_WriteMapList", _wrap_OBSmartsPattern_WriteMapList, METH_VARARGS, "Debugging -- write a list of matches to the output stream"},
	 { "OBSmartsPattern_swigregister", OBSmartsPattern_swigregister, METH_O, NULL},
	 { "OBSmartsPattern_swiginit", OBSmartsPattern_swiginit, METH_VARARGS, NULL},
	 { "new_OBSmartsMatcher", _wrap_new_OBSmartsMatcher, METH_NOARGS, NULL},
	 { "delete_OBSmartsMatcher", _wrap_delete_OBSmartsMatcher, METH_O, NULL},
	 { "OBSmartsMatcher_match", _wrap_OBSmartsMatcher_match, METH_VARARGS, NULL},
	 { "OBSmartsMatcher_swigregister", OBSmartsMatcher_swigregister, METH_O, NULL},
	 { "OBSmartsMatcher_swiginit", OBSmartsMatcher_swiginit, METH_VARARGS, NULL},
	 { "new_OBSSMatch", _wrap_new_OBSSMatch, METH_VARARGS, NULL},
	 { "delete_OBSSMatch", _wrap_delete_OBSSMatch, METH_O, NULL},
	 { "OBSSMatch_Match", _wrap_OBSSMatch_Match, METH_VARARGS, NULL},
	 { "OBSSMatch_swigregister", OBSSMatch_swigregister, METH_O, NULL},
	 { "OBSSMatch_swiginit", OBSSMatch_swiginit, METH_VARARGS, NULL},
	 { "SmartsLexReplace", _wrap_SmartsLexReplace, METH_VARARGS, NULL},
	 { "new_AliasData", _wrap_new_AliasData, METH_NOARGS, NULL},
	 { "AliasData_SetAlias", _wrap_AliasData_SetAlias, METH_VARARGS, "Add an alias"},
	 { "AliasData_GetAlias", _wrap_AliasData_GetAlias, METH_VARARGS, "return value of alias or its version intended to be connected at its right hand end."},
	 { "AliasData_GetColor", _wrap_AliasData_GetColor, METH_O, "Return the color which has been assigned to this alias"},
	 { "AliasData_SetColor", _wrap_AliasData_SetColor, METH_VARARGS, "Assign a color to this alias"},
	 { "AliasData_IsExpanded", _wrap_AliasData_IsExpanded, METH_O, NULL},
	 { "AliasData_RevertToAliasForm", _wrap_AliasData_RevertToAliasForm, METH_O, "\n"
		"Converts all the expanded aliases in a molecule to the alias form.\n"
		"Note that this deletes atoms and bonds. Use only as a preparation for display.\n"
		""},
	 { "AliasData_Expand", _wrap_AliasData_Expand, METH_VARARGS, "Interprets the alias text and adds atoms as appropriate to mol."},
	 { "AliasData_AddAliases", _wrap_AliasData_AddAliases, METH_O, "\n"
		"Matches univalent fragments in the molecule and adds AliasData to provide an alternative representation.\n"
		"The molecule remains a respectable OBMol. Data in superatoms.txt decides what aliases are added.\n"
		""},
	 { "delete_AliasData", _wrap_delete_AliasData, METH_O, NULL},
	 { "AliasData_swigregister", AliasData_swigregister, METH_O, NULL},
	 { "AliasData_swiginit", AliasData_swiginit, METH_VARARGS, NULL},
	 { "OBFingerprint_Default", _wrap_OBFingerprint_Default, METH_NOARGS, NULL},
	 { "OBFingerprint_FindType", _wrap_OBFingerprint_FindType, METH_O, NULL},
	 { "delete_OBFingerprint", _wrap_delete_OBFingerprint, METH_O, NULL},
	 { "OBFingerprint_SetBit", _wrap_OBFingerprint_SetBit, METH_VARARGS, "Sets the nth bit"},
	 { "OBFingerprint_GetBit", _wrap_OBFingerprint_GetBit, METH_VARARGS, "return true if the nth bit is set;"},
	 { "OBFingerprint_Fold", _wrap_OBFingerprint_Fold, METH_VARARGS, "Repeatedly ORs the top half with the bottom half until no smaller than nbits"},
	 { "OBFingerprint_GetFingerprint", _wrap_OBFingerprint_GetFingerprint, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: fingerprint in vector, which may be resized, folded to nbits (if nbits!=0)\n"
		""},
	 { "OBFingerprint_Flags", _wrap_OBFingerprint_Flags, METH_O, NULL},
	 { "OBFingerprint_SetFlags", _wrap_OBFingerprint_SetFlags, METH_VARARGS, NULL},
	 { "OBFingerprint_DescribeBits", _wrap_OBFingerprint_DescribeBits, METH_VARARGS, "\n"
		":rtype: string\n"
		":return: a description of each bit that is set (or unset, if bSet=false)\n"
		"Since: version 2.2\n"
		""},
	 { "OBFingerprint_Tanimoto", _wrap_OBFingerprint_Tanimoto, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: float\n"
		":return: the Tanimoto coefficient between two vectors (vector<unsigned int>& SeekPositions)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Inline version of Tanimoto() taking a pointer for the second vector\n"
		""},
	 { "OBFingerprint_Getbitsperint", _wrap_OBFingerprint_Getbitsperint, METH_NOARGS, NULL},
	 { "OBFingerprint_FindFingerprint", _wrap_OBFingerprint_FindFingerprint, METH_O, "\n"
		":rtype: :py:class:`OBFingerprint`\n"
		":return: a pointer to a fingerprint (the default if ID is empty), or NULL if not available\n"
		"   For backward compatibility;  a synonym of OBFingerprint::FindType\n"
		""},
	 { "OBFingerprint_swigregister", OBFingerprint_swigregister, METH_O, NULL},
	 { "FptIndexHeader_headerlength_set", _wrap_FptIndexHeader_headerlength_set, METH_VARARGS, "offset to data: sizeof(FptIndexHeader)"},
	 { "FptIndexHeader_headerlength_get", _wrap_FptIndexHeader_headerlength_get, METH_O, "offset to data: sizeof(FptIndexHeader)"},
	 { "FptIndexHeader_nEntries_set", _wrap_FptIndexHeader_nEntries_set, METH_VARARGS, "number of fingerprints"},
	 { "FptIndexHeader_nEntries_get", _wrap_FptIndexHeader_nEntries_get, METH_O, "number of fingerprints"},
	 { "FptIndexHeader_words_set", _wrap_FptIndexHeader_words_set, METH_VARARGS, "number 32bit words per fingerprint"},
	 { "FptIndexHeader_words_get", _wrap_FptIndexHeader_words_get, METH_O, "number 32bit words per fingerprint"},
	 { "FptIndexHeader_fpid_set", _wrap_FptIndexHeader_fpid_set, METH_VARARGS, "ID of the fingerprint type"},
	 { "FptIndexHeader_fpid_get", _wrap_FptIndexHeader_fpid_get, METH_O, "ID of the fingerprint type"},
	 { "FptIndexHeader_seek64_set", _wrap_FptIndexHeader_seek64_set, METH_VARARGS, NULL},
	 { "FptIndexHeader_seek64_get", _wrap_FptIndexHeader_seek64_get, METH_O, NULL},
	 { "FptIndexHeader_datafilename_set", _wrap_FptIndexHeader_datafilename_set, METH_VARARGS, "the data that this is an index to"},
	 { "FptIndexHeader_datafilename_get", _wrap_FptIndexHeader_datafilename_get, METH_O, "the data that this is an index to"},
	 { "new_FptIndexHeader", _wrap_new_FptIndexHeader, METH_NOARGS, NULL},
	 { "delete_FptIndexHeader", _wrap_delete_FptIndexHeader, METH_O, NULL},
	 { "FptIndexHeader_swigregister", FptIndexHeader_swigregister, METH_O, NULL},
	 { "FptIndexHeader_swiginit", FptIndexHeader_swiginit, METH_VARARGS, NULL},
	 { "FastSearch_ReadIndexFile", _wrap_FastSearch_ReadIndexFile, METH_VARARGS, "Loads an index from a file and returns the name of the datafile"},
	 { "FastSearch_ReadIndex", _wrap_FastSearch_ReadIndex, METH_VARARGS, NULL},
	 { "delete_FastSearch", _wrap_delete_FastSearch, METH_O, NULL},
	 { "FastSearch_Find", _wrap_FastSearch_Find, METH_VARARGS, "Does substructure search and returns vector of the file positions of matches"},
	 { "FastSearch_FindMatch", _wrap_FastSearch_FindMatch, METH_VARARGS, "\n"
		"Similar to Find() but all bits of matching fingerprints have to be the same\n"
		"Since: version 2.1\n"
		""},
	 { "FastSearch_FindSimilar", _wrap_FastSearch_FindSimilar, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":rtype: boolean\n"
		":return: A multimap containing objects whose Tanimoto coefficients with the target\n"
		"    is greater than the value specified.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":rtype: boolean\n"
		":return: A multimap containing the nCandidates objects with largest Tanimoto\n"
		"     coefficients with the target.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		":rtype: boolean\n"
		":return: A multimap containing the nCandidates objects with largest Tanimoto\n"
		"     coefficients with the target.\n"
		""},
	 { "FastSearch_GetFingerprint", _wrap_FastSearch_GetFingerprint, METH_O, "\n"
		":rtype: :py:class:`OBFingerprint`\n"
		":return: a pointer to the fingerprint type used to constuct the index\n"
		""},
	 { "FastSearch_GetIndexHeader", _wrap_FastSearch_GetIndexHeader, METH_O, "\n"
		":rtype: :py:class:`FptIndexHeader`\n"
		":return: a pointer to the index header containing size info etc.\n"
		""},
	 { "new_FastSearch", _wrap_new_FastSearch, METH_NOARGS, NULL},
	 { "FastSearch_swigregister", FastSearch_swigregister, METH_O, NULL},
	 { "FastSearch_swiginit", FastSearch_swiginit, METH_VARARGS, NULL},
	 { "new_FastSearchIndexer", _wrap_new_FastSearchIndexer, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Constructor with a new index\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Constructor using existing index\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Constructor using existing index\n"
		""},
	 { "delete_FastSearchIndexer", _wrap_delete_FastSearchIndexer, METH_O, NULL},
	 { "FastSearchIndexer_Add", _wrap_FastSearchIndexer_Add, METH_VARARGS, "Called for each object"},
	 { "FastSearchIndexer_swigregister", FastSearchIndexer_swigregister, METH_O, NULL},
	 { "FastSearchIndexer_swiginit", FastSearchIndexer_swiginit, METH_VARARGS, NULL},
	 { "OBDescriptor_Default", _wrap_OBDescriptor_Default, METH_NOARGS, NULL},
	 { "new_OBDescriptor", _wrap_new_OBDescriptor, METH_VARARGS, NULL},
	 { "OBDescriptor_FindType", _wrap_OBDescriptor_FindType, METH_O, NULL},
	 { "OBDescriptor_Predict", _wrap_OBDescriptor_Predict, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the value of a numeric descriptor\n"
		""},
	 { "OBDescriptor_PredictAndSave", _wrap_OBDescriptor_PredictAndSave, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: the value of the descriptor and adds it to the object's OBPairData\n"
		""},
	 { "OBDescriptor_GetStringValue", _wrap_OBDescriptor_GetStringValue, METH_VARARGS, "Provides a string value for non-numeric descriptors and returns NaN, or a string representation and returns a numeric value"},
	 { "OBDescriptor_Compare", _wrap_OBDescriptor_Compare, METH_VARARGS, "Parses the filter stream for a relational expression and returns its result when applied to the chemical object"},
	 { "OBDescriptor_Display", _wrap_OBDescriptor_Display, METH_VARARGS, "\n"
		"Write information on a plugin class to the string txt.\n"
		"If the parameter is a descriptor ID, displays the verbose description for that descriptor only\n"
		" e.g. babel -L descriptors HBA1\n"
		""},
	 { "OBDescriptor_Order", _wrap_OBDescriptor_Order, METH_VARARGS, "\n"
		"Comparison of the values of the descriptor. Used in sorting.\n"
		"Descriptors may use more complicated ordering than this default (e.g.InChIFilter)\n"
		""},
	 { "OBDescriptor_FilterCompare", _wrap_OBDescriptor_FilterCompare, METH_VARARGS, "Interprets the --filter option string and returns the combined result of all the comparisons it contains"},
	 { "OBDescriptor_AddProperties", _wrap_OBDescriptor_AddProperties, METH_VARARGS, "Reads list of descriptor IDs and calls PredictAndSave() for each."},
	 { "OBDescriptor_DeleteProperties", _wrap_OBDescriptor_DeleteProperties, METH_VARARGS, "Deletes all the OBPairDatas whose attribute names are in the list (if they exist)."},
	 { "OBDescriptor_GetValues", _wrap_OBDescriptor_GetValues, METH_VARARGS, "\n"
		"Reads list of descriptor IDs and OBPairData names and returns a list of values,\n"
		"each precede by a space or the first character in the list if it is whitespace or punctuation.\n"
		""},
	 { "OBDescriptor_GetIdentifier", _wrap_OBDescriptor_GetIdentifier, METH_O, "Read an identifier and its parameter from the filter string."},
	 { "delete_OBDescriptor", _wrap_delete_OBDescriptor, METH_O, NULL},
	 { "OBDescriptor_swigregister", OBDescriptor_swigregister, METH_O, NULL},
	 { "OBDescriptor_swiginit", OBDescriptor_swiginit, METH_VARARGS, NULL},
	 { "new_LineSearchType", _wrap_new_LineSearchType, METH_NOARGS, NULL},
	 { "delete_LineSearchType", _wrap_delete_LineSearchType, METH_O, NULL},
	 { "LineSearchType_swigregister", LineSearchType_swigregister, METH_O, NULL},
	 { "LineSearchType_swiginit", LineSearchType_swiginit, METH_VARARGS, NULL},
	 { "OBFFParameter_a_set", _wrap_OBFFParameter_a_set, METH_VARARGS, "Used to store integer atom types"},
	 { "OBFFParameter_a_get", _wrap_OBFFParameter_a_get, METH_O, "Used to store integer atom types"},
	 { "OBFFParameter_b_set", _wrap_OBFFParameter_b_set, METH_VARARGS, NULL},
	 { "OBFFParameter_b_get", _wrap_OBFFParameter_b_get, METH_O, NULL},
	 { "OBFFParameter_c_set", _wrap_OBFFParameter_c_set, METH_VARARGS, NULL},
	 { "OBFFParameter_c_get", _wrap_OBFFParameter_c_get, METH_O, NULL},
	 { "OBFFParameter_d_set", _wrap_OBFFParameter_d_set, METH_VARARGS, NULL},
	 { "OBFFParameter_d_get", _wrap_OBFFParameter_d_get, METH_O, NULL},
	 { "OBFFParameter__a_set", _wrap_OBFFParameter__a_set, METH_VARARGS, "used to store string atom types"},
	 { "OBFFParameter__a_get", _wrap_OBFFParameter__a_get, METH_O, "used to store string atom types"},
	 { "OBFFParameter__b_set", _wrap_OBFFParameter__b_set, METH_VARARGS, NULL},
	 { "OBFFParameter__b_get", _wrap_OBFFParameter__b_get, METH_O, NULL},
	 { "OBFFParameter__c_set", _wrap_OBFFParameter__c_set, METH_VARARGS, NULL},
	 { "OBFFParameter__c_get", _wrap_OBFFParameter__c_get, METH_O, NULL},
	 { "OBFFParameter__d_set", _wrap_OBFFParameter__d_set, METH_VARARGS, NULL},
	 { "OBFFParameter__d_get", _wrap_OBFFParameter__d_get, METH_O, NULL},
	 { "OBFFParameter__ipar_set", _wrap_OBFFParameter__ipar_set, METH_VARARGS, "Used to store integer type parameters (bondtypes, multiplicity, ...)"},
	 { "OBFFParameter__ipar_get", _wrap_OBFFParameter__ipar_get, METH_O, "Used to store integer type parameters (bondtypes, multiplicity, ...)"},
	 { "OBFFParameter__dpar_set", _wrap_OBFFParameter__dpar_set, METH_VARARGS, "Used to store double type parameters (force constants, bond lengths, angles, ...)"},
	 { "OBFFParameter__dpar_get", _wrap_OBFFParameter__dpar_get, METH_O, "Used to store double type parameters (force constants, bond lengths, angles, ...)"},
	 { "OBFFParameter_clear", _wrap_OBFFParameter_clear, METH_O, "Reset the atom types and set all parameters to zero"},
	 { "new_OBFFParameter", _wrap_new_OBFFParameter, METH_NOARGS, NULL},
	 { "delete_OBFFParameter", _wrap_delete_OBFFParameter, METH_O, NULL},
	 { "OBFFParameter_swigregister", OBFFParameter_swigregister, METH_O, NULL},
	 { "OBFFParameter_swiginit", OBFFParameter_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation2_energy_set", _wrap_OBFFCalculation2_energy_set, METH_VARARGS, "Used to store the energy for this OBFFCalculation"},
	 { "OBFFCalculation2_energy_get", _wrap_OBFFCalculation2_energy_get, METH_O, "Used to store the energy for this OBFFCalculation"},
	 { "OBFFCalculation2_a_set", _wrap_OBFFCalculation2_a_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_a_get", _wrap_OBFFCalculation2_a_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_b_set", _wrap_OBFFCalculation2_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_b_get", _wrap_OBFFCalculation2_b_get, METH_O, NULL},
	 { "OBFFCalculation2_idx_a_set", _wrap_OBFFCalculation2_idx_a_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_idx_a_get", _wrap_OBFFCalculation2_idx_a_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation2_idx_b_set", _wrap_OBFFCalculation2_idx_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_idx_b_get", _wrap_OBFFCalculation2_idx_b_get, METH_O, NULL},
	 { "OBFFCalculation2_pos_a_set", _wrap_OBFFCalculation2_pos_a_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation2_pos_a_get", _wrap_OBFFCalculation2_pos_a_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation2_pos_b_set", _wrap_OBFFCalculation2_pos_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_pos_b_get", _wrap_OBFFCalculation2_pos_b_get, METH_O, NULL},
	 { "OBFFCalculation2_force_a_set", _wrap_OBFFCalculation2_force_a_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation2_force_a_get", _wrap_OBFFCalculation2_force_a_get, METH_O, "Pointer to atom forces"},
	 { "OBFFCalculation2_force_b_set", _wrap_OBFFCalculation2_force_b_set, METH_VARARGS, NULL},
	 { "OBFFCalculation2_force_b_get", _wrap_OBFFCalculation2_force_b_get, METH_O, NULL},
	 { "delete_OBFFCalculation2", _wrap_delete_OBFFCalculation2, METH_O, "Destructor"},
	 { "OBFFCalculation2_SetupPointers", _wrap_OBFFCalculation2_SetupPointers, METH_O, "\n"
		":rtype: void\n"
		":return: Setup pointers to atom positions and forces (To be called  while setting up calculations). Sets optimized to true.\n"
		""},
	 { "new_OBFFCalculation2", _wrap_new_OBFFCalculation2, METH_NOARGS, NULL},
	 { "OBFFCalculation2_swigregister", OBFFCalculation2_swigregister, METH_O, NULL},
	 { "OBFFCalculation2_swiginit", OBFFCalculation2_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation3_c_set", _wrap_OBFFCalculation3_c_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_c_get", _wrap_OBFFCalculation3_c_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_idx_c_set", _wrap_OBFFCalculation3_idx_c_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_idx_c_get", _wrap_OBFFCalculation3_idx_c_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation3_pos_c_set", _wrap_OBFFCalculation3_pos_c_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation3_pos_c_get", _wrap_OBFFCalculation3_pos_c_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation3_force_c_set", _wrap_OBFFCalculation3_force_c_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation3_force_c_get", _wrap_OBFFCalculation3_force_c_get, METH_O, "Pointer to atom forces"},
	 { "delete_OBFFCalculation3", _wrap_delete_OBFFCalculation3, METH_O, "Destructor"},
	 { "new_OBFFCalculation3", _wrap_new_OBFFCalculation3, METH_NOARGS, NULL},
	 { "OBFFCalculation3_swigregister", OBFFCalculation3_swigregister, METH_O, NULL},
	 { "OBFFCalculation3_swiginit", OBFFCalculation3_swiginit, METH_VARARGS, NULL},
	 { "OBFFCalculation4_d_set", _wrap_OBFFCalculation4_d_set, METH_VARARGS, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_d_get", _wrap_OBFFCalculation4_d_get, METH_O, "Used to store the atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_idx_d_set", _wrap_OBFFCalculation4_idx_d_set, METH_VARARGS, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_idx_d_get", _wrap_OBFFCalculation4_idx_d_get, METH_O, "Used to store the index of atoms for this OBFFCalculation"},
	 { "OBFFCalculation4_pos_d_set", _wrap_OBFFCalculation4_pos_d_set, METH_VARARGS, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation4_pos_d_get", _wrap_OBFFCalculation4_pos_d_get, METH_O, "Pointer to atom coordinates as double[3]"},
	 { "OBFFCalculation4_force_d_set", _wrap_OBFFCalculation4_force_d_set, METH_VARARGS, "Pointer to atom forces"},
	 { "OBFFCalculation4_force_d_get", _wrap_OBFFCalculation4_force_d_get, METH_O, "Pointer to atom forces"},
	 { "delete_OBFFCalculation4", _wrap_delete_OBFFCalculation4, METH_O, "Destructor"},
	 { "new_OBFFCalculation4", _wrap_new_OBFFCalculation4, METH_NOARGS, NULL},
	 { "OBFFCalculation4_swigregister", OBFFCalculation4_swigregister, METH_O, NULL},
	 { "OBFFCalculation4_swiginit", OBFFCalculation4_swiginit, METH_VARARGS, NULL},
	 { "OBFFConstraint_factor_set", _wrap_OBFFConstraint_factor_set, METH_VARARGS, "Used to store the contraint energy for this OBFFConstraint"},
	 { "OBFFConstraint_factor_get", _wrap_OBFFConstraint_factor_get, METH_O, "Used to store the contraint energy for this OBFFConstraint"},
	 { "OBFFConstraint_constraint_value_set", _wrap_OBFFConstraint_constraint_value_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_constraint_value_get", _wrap_OBFFConstraint_constraint_value_get, METH_O, NULL},
	 { "OBFFConstraint_rab0_set", _wrap_OBFFConstraint_rab0_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_rab0_get", _wrap_OBFFConstraint_rab0_get, METH_O, NULL},
	 { "OBFFConstraint_rbc0_set", _wrap_OBFFConstraint_rbc0_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_rbc0_get", _wrap_OBFFConstraint_rbc0_get, METH_O, NULL},
	 { "OBFFConstraint_type_set", _wrap_OBFFConstraint_type_set, METH_VARARGS, "Used to store the contraint type for this OBFFConstraint"},
	 { "OBFFConstraint_type_get", _wrap_OBFFConstraint_type_get, METH_O, "Used to store the contraint type for this OBFFConstraint"},
	 { "OBFFConstraint_ia_set", _wrap_OBFFConstraint_ia_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ia_get", _wrap_OBFFConstraint_ia_get, METH_O, NULL},
	 { "OBFFConstraint_ib_set", _wrap_OBFFConstraint_ib_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ib_get", _wrap_OBFFConstraint_ib_get, METH_O, NULL},
	 { "OBFFConstraint_ic_set", _wrap_OBFFConstraint_ic_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_ic_get", _wrap_OBFFConstraint_ic_get, METH_O, NULL},
	 { "OBFFConstraint_id_set", _wrap_OBFFConstraint_id_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_id_get", _wrap_OBFFConstraint_id_get, METH_O, NULL},
	 { "OBFFConstraint_a_set", _wrap_OBFFConstraint_a_set, METH_VARARGS, "Used to store the atoms for this OBFFCostraint"},
	 { "OBFFConstraint_a_get", _wrap_OBFFConstraint_a_get, METH_O, "Used to store the atoms for this OBFFCostraint"},
	 { "OBFFConstraint_b_set", _wrap_OBFFConstraint_b_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_b_get", _wrap_OBFFConstraint_b_get, METH_O, NULL},
	 { "OBFFConstraint_c_set", _wrap_OBFFConstraint_c_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_c_get", _wrap_OBFFConstraint_c_get, METH_O, NULL},
	 { "OBFFConstraint_d_set", _wrap_OBFFConstraint_d_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_d_get", _wrap_OBFFConstraint_d_get, METH_O, NULL},
	 { "OBFFConstraint_grada_set", _wrap_OBFFConstraint_grada_set, METH_VARARGS, "Used to store the gradients for this OBFFCalculation"},
	 { "OBFFConstraint_grada_get", _wrap_OBFFConstraint_grada_get, METH_O, "Used to store the gradients for this OBFFCalculation"},
	 { "OBFFConstraint_gradb_set", _wrap_OBFFConstraint_gradb_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradb_get", _wrap_OBFFConstraint_gradb_get, METH_O, NULL},
	 { "OBFFConstraint_gradc_set", _wrap_OBFFConstraint_gradc_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradc_get", _wrap_OBFFConstraint_gradc_get, METH_O, NULL},
	 { "OBFFConstraint_gradd_set", _wrap_OBFFConstraint_gradd_set, METH_VARARGS, NULL},
	 { "OBFFConstraint_gradd_get", _wrap_OBFFConstraint_gradd_get, METH_O, NULL},
	 { "new_OBFFConstraint", _wrap_new_OBFFConstraint, METH_NOARGS, "Constructor"},
	 { "delete_OBFFConstraint", _wrap_delete_OBFFConstraint, METH_O, "Destructor"},
	 { "OBFFConstraint_GetGradient", _wrap_OBFFConstraint_GetGradient, METH_VARARGS, NULL},
	 { "OBFFConstraint_swigregister", OBFFConstraint_swigregister, METH_O, NULL},
	 { "OBFFConstraint_swiginit", OBFFConstraint_swiginit, METH_VARARGS, NULL},
	 { "new_OBFFConstraints", _wrap_new_OBFFConstraints, METH_NOARGS, "Constructor"},
	 { "delete_OBFFConstraints", _wrap_delete_OBFFConstraints, METH_O, "Destructor"},
	 { "OBFFConstraints_Clear", _wrap_OBFFConstraints_Clear, METH_O, "Clear all constraints"},
	 { "OBFFConstraints_GetConstraintEnergy", _wrap_OBFFConstraints_GetConstraintEnergy, METH_O, "Get the constraint energy"},
	 { "OBFFConstraints_GetGradient", _wrap_OBFFConstraints_GetGradient, METH_VARARGS, "Get the constraint gradient for atom with index a"},
	 { "OBFFConstraints_Setup", _wrap_OBFFConstraints_Setup, METH_VARARGS, "\n"
		"Translate indices to OBAtom* objects, this function is called from OBForceField::Setup,\n"
		"this function doesn't have to be called from anywhere else.\n"
		""},
	 { "OBFFConstraints_SetFactor", _wrap_OBFFConstraints_SetFactor, METH_VARARGS, "Set Constraint factor"},
	 { "OBFFConstraints_AddIgnore", _wrap_OBFFConstraints_AddIgnore, METH_VARARGS, "Ignore the atom while setting up calculations"},
	 { "OBFFConstraints_AddAtomConstraint", _wrap_OBFFConstraints_AddAtomConstraint, METH_VARARGS, "Fix the position of an atom"},
	 { "OBFFConstraints_AddAtomXConstraint", _wrap_OBFFConstraints_AddAtomXConstraint, METH_VARARGS, "Fix the x coordinate of the atom position"},
	 { "OBFFConstraints_AddAtomYConstraint", _wrap_OBFFConstraints_AddAtomYConstraint, METH_VARARGS, "Fix the y coordinate of the atom position"},
	 { "OBFFConstraints_AddAtomZConstraint", _wrap_OBFFConstraints_AddAtomZConstraint, METH_VARARGS, "Fix the z coordinate of the atom position"},
	 { "OBFFConstraints_AddDistanceConstraint", _wrap_OBFFConstraints_AddDistanceConstraint, METH_VARARGS, "Constrain the bond length a-b"},
	 { "OBFFConstraints_AddAngleConstraint", _wrap_OBFFConstraints_AddAngleConstraint, METH_VARARGS, "Constrain the angle a-b-c"},
	 { "OBFFConstraints_AddTorsionConstraint", _wrap_OBFFConstraints_AddTorsionConstraint, METH_VARARGS, "Constrain the torsion angle a-b-c-d"},
	 { "OBFFConstraints_DeleteConstraint", _wrap_OBFFConstraints_DeleteConstraint, METH_VARARGS, "Delete a constraint Title: index constraint index"},
	 { "OBFFConstraints_GetFactor", _wrap_OBFFConstraints_GetFactor, METH_O, "Get Constraint factor"},
	 { "OBFFConstraints_Size", _wrap_OBFFConstraints_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the number of set constraints\n"
		""},
	 { "OBFFConstraints_GetConstraintType", _wrap_OBFFConstraints_GetConstraintType, METH_VARARGS, "\n"
		"The following constraint types are known: OBFF_CONST_IGNORE (ignore\n"
		"the atom while setting up calculations, forcefield implementations\n"
		"need to check this value in their setup function), OBFF_CONST_ATOM\n"
		"(fix atom position), OBFF_CONST_ATOM_X (fix x coordinate),\n"
		"OBFF_CONST_ATOM_Y (fix y coordinate), OBFF_CONST_ATOM_Z (fix z\n"
		"coordinate), OBFF_CONST_BOND (constrain bond length), OBFF_CONST_ANGLE\n"
		"(constrain angle), OBFF_CONST_TORSION (constrain torsion angle)\n"
		":rtype: int\n"
		":return: the constraint type\n"
		""},
	 { "OBFFConstraints_GetConstraintValue", _wrap_OBFFConstraints_GetConstraintValue, METH_VARARGS, "\n"
		":rtype: float\n"
		":return: The constraint value, this can be a bond length, angle or\n"
		"     torsion angle depending on the constraint type.\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomA", _wrap_OBFFConstraints_GetConstraintAtomA, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom a (or fixed atom) Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomB", _wrap_OBFFConstraints_GetConstraintAtomB, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom b Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomC", _wrap_OBFFConstraints_GetConstraintAtomC, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom c Title: index constraint index\n"
		""},
	 { "OBFFConstraints_GetConstraintAtomD", _wrap_OBFFConstraints_GetConstraintAtomD, METH_VARARGS, "\n"
		":rtype: int\n"
		":return: The constraint atom d Title: index constraint index\n"
		""},
	 { "OBFFConstraints_IsIgnored", _wrap_OBFFConstraints_IsIgnored, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this atom is ignored Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsFixed", _wrap_OBFFConstraints_IsFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsXFixed", _wrap_OBFFConstraints_IsXFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the x coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsYFixed", _wrap_OBFFConstraints_IsYFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the y coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_IsZFixed", _wrap_OBFFConstraints_IsZFixed, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if the z coordinate for this atom is fixed Title: a atom index\n"
		""},
	 { "OBFFConstraints_GetIgnoredBitVec", _wrap_OBFFConstraints_GetIgnoredBitVec, METH_O, "\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: the ignored atom indexes as bitvec. (used in OBForceField::Setup() to determine if a call to OBForceField::SetupCalculations() is needed)\n"
		""},
	 { "OBFFConstraints_GetFixedBitVec", _wrap_OBFFConstraints_GetFixedBitVec, METH_O, "\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: the fixed atom indexes as bitvec. (used in OBForceField::SystematicRotorSearch() and similar)\n"
		""},
	 { "OBFFConstraints_swigregister", OBFFConstraints_swigregister, METH_O, NULL},
	 { "OBFFConstraints_swiginit", OBFFConstraints_swiginit, METH_VARARGS, NULL},
	 { "OBForceField_Default", _wrap_OBForceField_Default, METH_NOARGS, NULL},
	 { "OBForceField_FindType", _wrap_OBForceField_FindType, METH_O, NULL},
	 { "OBForceField_MakeNewInstance", _wrap_OBForceField_MakeNewInstance, METH_O, "\n"
		"Clone the current instance. May be desirable in multithreaded environments,\n"
		"Should be deleted after use\n"
		""},
	 { "delete_OBForceField", _wrap_delete_OBForceField, METH_O, "Destructor (defined in forcefield.cpp because _lbfgsState's pimpl type LBFGSState is incomplete in this header)."},
	 { "OBForceField_FindForceField", _wrap_OBForceField_FindForceField, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		":type ID: string\n"
		":param ID: forcefield id (Ghemical, MMFF94, UFF, ...).\n"
		":rtype: :py:class:`OBForceField`\n"
		":return: A pointer to a forcefield (the default if ID is empty), or NULL if not available.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		":type ID: string\n"
		":param ID: forcefield id (Ghemical, MMFF94, UFF, ...).\n"
		":rtype: :py:class:`OBForceField`\n"
		":return: A pointer to a forcefield (the default if ID is empty), or NULL if not available.\n"
		""},
	 { "OBForceField_SetParameterFile", _wrap_OBForceField_SetParameterFile, METH_VARARGS, NULL},
	 { "OBForceField_GetUnit", _wrap_OBForceField_GetUnit, METH_O, "\n"
		":rtype: string\n"
		":return: The unit (kcal/mol, kJ/mol, ...) in which the energy is expressed as std::string.\n"
		""},
	 { "OBForceField_HasAnalyticalGradients", _wrap_OBForceField_HasAnalyticalGradients, METH_O, NULL},
	 { "OBForceField_Setup", _wrap_OBForceField_Setup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Setup the forcefield for mol (assigns atom types, charges, etc.). Keep current constraints.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object that contains the atoms and bonds.\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Setup the forcefield for mol (assigns atom types, charges, etc.). Use new constraints.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object that contains the atoms and bonds.\n"
		":type constraints: :py:class:`OBFFConstraints`\n"
		":param constraints: The OBFFConstraints object that contains the constraints.\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_ParseParamFile", _wrap_OBForceField_ParseParamFile, METH_O, "\n"
		"Load the parameters (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetTypes", _wrap_OBForceField_SetTypes, METH_O, "\n"
		"Set the atom types (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetFormalCharges", _wrap_OBForceField_SetFormalCharges, METH_O, "\n"
		"Set the formal charges (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetPartialCharges", _wrap_OBForceField_SetPartialCharges, METH_O, "\n"
		"Set the partial charges (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetupCalculations", _wrap_OBForceField_SetupCalculations, METH_O, "\n"
		"Setup the calculations (this function is overloaded by the individual forcefields,\n"
		"and is called autoamically from OBForceField::Setup()).\n"
		""},
	 { "OBForceField_SetupPointers", _wrap_OBForceField_SetupPointers, METH_O, "\n"
		"Setup the pointers to the atom positions in the OBFFCalculation objects. This method\n"
		"will iterate over all the calculations and call SetupPointers for each one. (This\n"
		"function should be implemented by the individual force field implementations).\n"
		""},
	 { "OBForceField_IsSetupNeeded", _wrap_OBForceField_IsSetupNeeded, METH_VARARGS, "\n"
		"Compare the internal forcefield OBMol object to mol. If the two have the\n"
		"same number of atoms and bonds, and all atomic numbers are the same,\n"
		"this function returns false, and no call to Setup is needed.\n"
		":rtype: boolean\n"
		":return: True if Setup needs to be called.\n"
		""},
	 { "OBForceField_GetAtomTypes", _wrap_OBForceField_GetAtomTypes, METH_VARARGS, "\n"
		"Get the force atom types. The atom types will be added to\n"
		"the atoms of mol as OBPairData. The attribute will be \"FFAtomType\".\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    ...\n"
		"    pFF->Setup(&mol);\n"
		"    pFF->GetAtomTypes(&mol);\n"
		"    FOR_ATOMS_OF_MOL (atom, mol) {\n"
		"      OBPairData *type = (OBPairData*) atom->GetData(\"FFAtomType\");\n"
		"      if (type)\n"
		"        cout << \"atom \" << atom->GetIdx() << \" : \" << type->GetValue() << endl;\n"
		"    }\n"
		"    ...\n"
		""},
	 { "OBForceField_GetPartialCharges", _wrap_OBForceField_GetPartialCharges, METH_VARARGS, "\n"
		"Get the force field formal charges. The formal charges will be added to\n"
		"the atoms of mol as OBPairData. The attribute will be \"FFPartialCharge\".\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    ...\n"
		"    pFF->Setup(&mol);\n"
		"    pFF->GetPartialCharges(&mol);\n"
		"    FOR_ATOMS_OF_MOL (atom, mol) {\n"
		"      OBPairData *chg = (OBPairData*) atom->GetData(\"FFPartialCharge\");\n"
		"      if (chg)\n"
		"        cout << \"atom \" << atom->GetIdx() << \" : \" << chg->GetValue() << endl;\n"
		"    }\n"
		"    ...\n"
		""},
	 { "OBForceField_GetCoordinates", _wrap_OBForceField_GetCoordinates, METH_VARARGS, "\n"
		"Get coordinates for current conformer and attach OBConformerData with energies, forces, ... to mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates to (from OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_UpdateCoordinates", _wrap_OBForceField_UpdateCoordinates, METH_VARARGS, "Deprecated: Use GetCooordinates instead."},
	 { "OBForceField_GetConformers", _wrap_OBForceField_GetConformers, METH_VARARGS, "\n"
		"Get coordinates for all conformers and attach OBConformerData with energies, forces, ... to mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates to (from OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_UpdateConformers", _wrap_OBForceField_UpdateConformers, METH_VARARGS, "Deprecated: Use GetConformers instead."},
	 { "OBForceField_SetCoordinates", _wrap_OBForceField_SetCoordinates, METH_VARARGS, "\n"
		"Set coordinates for current conformer.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: the OBMol object to copy the coordinates from (to OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: true if successful.\n"
		""},
	 { "OBForceField_SetConformers", _wrap_OBForceField_SetConformers, METH_VARARGS, "\n"
		"Set coordinates for all conformers.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The OBMol object to copy the coordinates from (to OBForceField::_mol).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_GetGrid", _wrap_OBForceField_GetGrid, METH_VARARGS, "\n"
		"Create a grid with spacing ``step`` and ``padding``. Place a probe atom of type probe at every grid point,\n"
		"calculate the energy and store it in the grid. These grids can then be used to create isosurfaces to\n"
		"identify locations where the probe atom has favourable interactions with the molecule.\n"
		":type step: float\n"
		":param step: The grid step size in A..\n"
		":type padding: float\n"
		":param padding: The padding for the grid in A.\n"
		":type type: string\n"
		":param type: The force field atom type for the probe.\n"
		":type pchg: float\n"
		":param pchg: The partial charge for the probe atom.\n"
		":rtype: :py:class:`OBGridData`\n"
		":return: Pointer to the grid constaining the results.\n"
		""},
	 { "OBForceField_AddIntraGroup", _wrap_OBForceField_AddIntraGroup, METH_VARARGS, "\n"
		"Enable intra-molecular interactions for group (bonds, angles, strbnd, torsions, oop).\n"
		"This function should be called before Setup().\n"
		":type group: :py:class:`OBBitVec`\n"
		":param group: OBBitVec with bits set for the indexes of the atoms which make up the group.\n"
		""},
	 { "OBForceField_AddInterGroup", _wrap_OBForceField_AddInterGroup, METH_VARARGS, "\n"
		"Enable inter-molecular interactions for group (non-bonded: vdw & ele).\n"
		"This function should be called before Setup().\n"
		":type group: :py:class:`OBBitVec`\n"
		":param group: OBBitVec with bits set for the indexes of the atoms which make up the group.\n"
		""},
	 { "OBForceField_AddInterGroups", _wrap_OBForceField_AddInterGroups, METH_VARARGS, "\n"
		"Enable inter-molecular interactions between group1 and group2 (non-bonded: vdw & ele).\n"
		"Note that this function doesn't enable bonded interactions in either group. Non-bonded\n"
		"interactions in the groups itself are also not enabled.\n"
		"This function should be called before Setup().\n"
		":type group1: :py:class:`OBBitVec`\n"
		":param group1: OBBitVec with bits set for the indexes of the atoms which make up the first group.\n"
		":type group2: :py:class:`OBBitVec`\n"
		":param group2: OBBitVec with bits set for the indexes of the atoms which make up the second group.\n"
		""},
	 { "OBForceField_ClearGroups", _wrap_OBForceField_ClearGroups, METH_O, "Clear all previously specified groups."},
	 { "OBForceField_HasGroups", _wrap_OBForceField_HasGroups, METH_O, "\n"
		":rtype: boolean\n"
		":return: true if there are groups.\n"
		""},
	 { "OBForceField_EnableCutOff", _wrap_OBForceField_EnableCutOff, METH_VARARGS, "\n"
		"Enable or disable Cut-offs. Cut-offs are disabled by default.\n"
		":type enable: boolean\n"
		":param enable: Enable when true, disable when false.\n"
		""},
	 { "OBForceField_IsCutOffEnabled", _wrap_OBForceField_IsCutOffEnabled, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if Cut-off distances are used.\n"
		""},
	 { "OBForceField_SetVDWCutOff", _wrap_OBForceField_SetVDWCutOff, METH_VARARGS, "\n"
		"Set the VDW cut-off distance to r. Note that this does not enable cut-off distances.\n"
		":type r: float\n"
		":param r: The VDW cut-off distance to be used in A.\n"
		""},
	 { "OBForceField_GetVDWCutOff", _wrap_OBForceField_GetVDWCutOff, METH_O, "\n"
		"Get the VDW cut-off distance.\n"
		":rtype: float\n"
		":return: The VDW cut-off distance in A.\n"
		""},
	 { "OBForceField_SetElectrostaticCutOff", _wrap_OBForceField_SetElectrostaticCutOff, METH_VARARGS, "\n"
		"Set the Electrostatic cut-off distance to r. Note that this does not\n"
		"enable cut-off distances.\n"
		":type r: float\n"
		":param r: The electrostatic cut-off distance to be used in A.\n"
		""},
	 { "OBForceField_GetElectrostaticCutOff", _wrap_OBForceField_GetElectrostaticCutOff, METH_O, "\n"
		"Get the Electrostatic cut-off distance.\n"
		":rtype: float\n"
		":return: The electrostatic cut-off distance in A.\n"
		""},
	 { "OBForceField_SetDielectricConstant", _wrap_OBForceField_SetDielectricConstant, METH_VARARGS, "\n"
		" Set the dielectric constant for electrostatic SetupCalculations\n"
		":type epsilon: float\n"
		":param epsilon: The relative permittivity to use (default = 1.0)\n"
		""},
	 { "OBForceField_GetDielectricConstant", _wrap_OBForceField_GetDielectricConstant, METH_O, NULL},
	 { "OBForceField_SetUpdateFrequency", _wrap_OBForceField_SetUpdateFrequency, METH_VARARGS, "\n"
		"Set the frequency by which non-bonded pairs are updated. Values from 10 to 20\n"
		"are recommended. Too low will decrease performance, too high will cause\n"
		"non-bonded interactions within cut-off not to be calculated.\n"
		":type f: int\n"
		":param f: The pair list update frequency.\n"
		""},
	 { "OBForceField_GetUpdateFrequency", _wrap_OBForceField_GetUpdateFrequency, METH_O, "\n"
		"Get the frequency by which non-bonded pairs are updated.\n"
		":rtype: int\n"
		":return: The pair list update frequency.\n"
		""},
	 { "OBForceField_UpdatePairsSimple", _wrap_OBForceField_UpdatePairsSimple, METH_O, "\n"
		"Set the bits in _vdwpairs and _elepairs to 1 for interactions that\n"
		"are within cut-off distance. This function is called in minimizing\n"
		"algorithms such as SteepestDescent and ConjugateGradients.\n"
		""},
	 { "OBForceField_GetNumPairs", _wrap_OBForceField_GetNumPairs, METH_O, "\n"
		"Get the number of non-bonded pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of atom pairs (ignores cutoff)\n"
		""},
	 { "OBForceField_GetNumElectrostaticPairs", _wrap_OBForceField_GetNumElectrostaticPairs, METH_O, "\n"
		"Get the number of enabled electrostatic pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of pairs currently enabled (within cut-off distance)\n"
		""},
	 { "OBForceField_GetNumVDWPairs", _wrap_OBForceField_GetNumVDWPairs, METH_O, "\n"
		"Get the number of enabled VDW pairs in _mol.\n"
		":rtype: int\n"
		":return: The number of pairs currently enabled (within cut-off distance)\n"
		""},
	 { "OBForceField_EnableAllPairs", _wrap_OBForceField_EnableAllPairs, METH_O, "\n"
		"Set bits in range 0..._numpairs-1 to 1. Using this means there will\n"
		"be no cut-off. (not-working: see code for more information.\n"
		""},
	 { "OBForceField_GetGradient", _wrap_OBForceField_GetGradient, METH_VARARGS, "Get the pointer to the gradients"},
	 { "OBForceField_GetGradientPtr", _wrap_OBForceField_GetGradientPtr, METH_O, "Get the pointer to the gradients"},
	 { "OBForceField_Energy", _wrap_OBForceField_Energy, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Total energy.\n"
		" Title: Output to log:\n"
		"  OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"  OBFF_LOGLVL_LOW:    none\n"
		"\n"
		"  OBFF_LOGLVL_MEDIUM: energy for individual energy terms\n"
		"\n"
		"  OBFF_LOGLVL_HIGH:   energy for individual energy interactions\n"
		""},
	 { "OBForceField_E_Bond", _wrap_OBForceField_E_Bond, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Bond stretching energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Angle", _wrap_OBForceField_E_Angle, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Angle bending energy.\n"
		"Title: Output to log:\n"
		" see Energy()\n"
		""},
	 { "OBForceField_E_StrBnd", _wrap_OBForceField_E_StrBnd, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Stretch bending energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Torsion", _wrap_OBForceField_E_Torsion, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Torsional energy.\n"
		"  Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_OOP", _wrap_OBForceField_E_OOP, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Out-Of-Plane bending energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_VDW", _wrap_OBForceField_E_VDW, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Van der Waals energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_E_Electrostatic", _wrap_OBForceField_E_Electrostatic, METH_VARARGS, "\n"
		":param gradients: Set to true when the gradients need to be calculated\n"
		"    (needs to be done before calling GetGradient()).\n"
		":rtype: float\n"
		":return: Electrostatic energy.\n"
		" Title: Output to log:\n"
		"  see Energy()\n"
		""},
	 { "OBForceField_PrintTypes", _wrap_OBForceField_PrintTypes, METH_O, "Print the atom types to the log."},
	 { "OBForceField_PrintFormalCharges", _wrap_OBForceField_PrintFormalCharges, METH_O, "\n"
		"Print the formal charges to the log (atom.GetPartialCharge(),\n"
		"MMFF94 FC's are not always int).\n"
		""},
	 { "OBForceField_PrintPartialCharges", _wrap_OBForceField_PrintPartialCharges, METH_O, "Print the partial charges to the log."},
	 { "OBForceField_PrintVelocities", _wrap_OBForceField_PrintVelocities, METH_O, "Print the velocities to the log."},
	 { "OBForceField_SetLogFile", _wrap_OBForceField_SetLogFile, METH_VARARGS, "\n"
		"Set the stream for logging (can also be &cout for logging to screen).\n"
		":type pos: std::ostream\n"
		":param pos: Stream (when pos is 0, std::cout wil be used).\n"
		":rtype: boolean\n"
		":return: True if successful.\n"
		""},
	 { "OBForceField_SetLogLevel", _wrap_OBForceField_SetLogLevel, METH_VARARGS, "\n"
		"Set the log level (OBFF_LOGLVL_NONE, OBFF_LOGLVL_LOW, OBFF_LOGLVL_MEDIUM, OBFF_LOGLVL_HIGH).\n"
		"Inline if statements for logging are available:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    #define IF_OBFF_LOGLVL_LOW    if(_loglvl >= OBFF_LOGLVL_LOW)\n"
		"    #define IF_OBFF_LOGLVL_MEDIUM if(_loglvl >= OBFF_LOGLVL_MEDIUM)\n"
		"    #define IF_OBFF_LOGLVL_HIGH   if(_loglvl >= OBFF_LOGLVL_HIGH)\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    SetLogLevel(OBFF_LOGLVL_MEDIUM);\n"
		"    IF_OBFF_LOGLVL_HIGH {\n"
		"      OBFFLog(\"this text will NOT be logged...\\n\");\n"
		"    }\n"
		"\n"
		"    IF_OBFF_LOGLVL_LOW {\n"
		"      OBFFLog\"this text will be logged...\\n\");\n"
		"    }\n"
		"\n"
		"    IF_OBFF_LOGLVL_MEDIUM {\n"
		"      OBFFLog(\"this text will also be logged...\\n\");\n"
		"    }\n"
		""},
	 { "OBForceField_GetLogLevel", _wrap_OBForceField_GetLogLevel, METH_O, "\n"
		":rtype: int\n"
		":return: The log level.\n"
		""},
	 { "OBForceField_OBFFLog", _wrap_OBForceField_OBFFLog, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Print msg to the logfile.\n"
		":type msg: string\n"
		":param msg: The message to print.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Print msg to the logfile.\n"
		":type msg: string\n"
		":param msg: The message to print.\n"
		""},
	 { "OBForceField_DistanceGeometry", _wrap_OBForceField_DistanceGeometry, METH_O, "Generate coordinates for the molecule (distance geometry) Deprecated: Use OBDistanceGeometry class instead"},
	 { "OBForceField_SystematicRotorSearch", _wrap_OBForceField_SystematicRotorSearch, METH_VARARGS, "\n"
		"Generate conformers for the molecule (systematicaly rotating torsions).\n"
		"\n"
		"The initial starting structure here is important, this structure should be\n"
		"minimized for the best results. SystematicRotorSearch works by rotating around\n"
		"the rotatable bond in a molecule (see OBRotamerList class). This rotating generates\n"
		"multiple conformers. The energy for all these conformers is then evaluated and the\n"
		"lowest energy conformer is selected.\n"
		"\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for this function will interfere with the output for\n"
		"this function.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		"OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_SystematicRotorSearchInitialize", _wrap_OBForceField_SystematicRotorSearchInitialize, METH_VARARGS, "\n"
		"Generate conformers for the molecule by systematicaly rotating torsions. To be used in combination with\n"
		"SystematicRotorSearchNexConformer().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->SystematicRotorSearchInitialize(300);\n"
		"    while (pFF->SystematicRotorSearchNextConformer(300)) {\n"
		"      // do some updating in your program (show last generated conformer, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, SystematicRotorSearch() is recommended.\n"
		"\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		":rtype: int\n"
		":return: The number of conformers.\n"
		""},
	 { "OBForceField_SystematicRotorSearchNextConformer", _wrap_OBForceField_SystematicRotorSearchNextConformer, METH_VARARGS, "\n"
		"Evaluate the next conformer.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":rtype: boolean\n"
		":return: True if there are more conformers.\n"
		""},
	 { "OBForceField_RandomRotorSearch", _wrap_OBForceField_RandomRotorSearch, METH_VARARGS, "\n"
		"Generate conformers for the molecule (randomly rotating torsions).\n"
		"\n"
		"The initial starting structure here is important, this structure should be\n"
		"minimized for the best results. RandomRotorSearch works by randomly rotating around\n"
		"the rotatable bonds in a molecule (see OBRotamerList class). This rotating generates\n"
		"multiple conformers. The energy for all these conformers is then evaluated and the\n"
		"lowest energy conformer is selected.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization for each conformer.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for this function will interfere with the output for\n"
		"this function.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		"OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_RandomRotorSearchInitialize", _wrap_OBForceField_RandomRotorSearchInitialize, METH_VARARGS, "\n"
		"Generate conformers for the molecule by randomly rotating torsions. To be used in combination with\n"
		"RandomRotorSearchNexConformer().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->RandomRotorSearchInitialize(300);\n"
		"    while (pFF->RandomRotorSearchNextConformer(300)) {\n"
		"      // do some updating in your program (show last generated conformer, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, RandomRotorSearch() is recommended.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: Whether to sample ring torsions.\n"
		""},
	 { "OBForceField_RandomRotorSearchNextConformer", _wrap_OBForceField_RandomRotorSearchNextConformer, METH_VARARGS, "\n"
		"Evaluate the next conformer.\n"
		":type geomSteps: int, optional\n"
		":param geomSteps: The number of steps to take during geometry optimization.\n"
		":rtype: boolean\n"
		":return: True if there are more conformers.\n"
		""},
	 { "OBForceField_WeightedRotorSearch", _wrap_OBForceField_WeightedRotorSearch, METH_VARARGS, "\n"
		" Generate conformers for the molecule (randomly rotating torsions).\n"
		"\n"
		" The initial starting structure here is important, this structure should be\n"
		" minimized for the best results. WeightedRotorSearch works by randomly rotating around\n"
		" the rotatable bonds in a molecule (see OBRotamerList class). Unlike RandomRotorSearch()\n"
		" the random choice of torsions is reweighted based on the energy of the generated conformer.\n"
		" Over time, the generated conformers for each step should become increasingly better.\n"
		" The lowest energy conformer is selected.\n"
		"\n"
		":type conformers: int\n"
		":param conformers: The number of random conformers to consider during the search.\n"
		":type geomSteps: int\n"
		":param geomSteps: The number of steps to take during geometry optimization for each conformer.\n"
		" :type sampleRingBonds: boolean, optional\n"
		" :param sampleRingBonds: Whether to sample ring torsions.\n"
		"\n"
		"	Title: Output to log:\n"
		" This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"	too much information about the energy calculations needed for this function will interfere with the output for\n"
		"	this function.\n"
		"\n"
		"\n"
		" OBFF_LOGLVL_NONE:   None.\n"
		"\n"
		" OBFF_LOGLVL_LOW:    Number of rotatable bonds, energies for the conformers, which one is the lowest, ...\n"
		"\n"
		" OBFF_LOGLVL_MEDIUM: See note above.\n"
		"\n"
		" OBFF_LOGLVL_HIGH:   See note above.\n"
		""},
	 { "OBForceField_FastRotorSearch", _wrap_OBForceField_FastRotorSearch, METH_VARARGS, "\n"
		"A fast rotor search to find low energy conformations\n"
		"\n"
		"Iterate over each of the rotors, and set the\n"
		"torsion angle to that which minimizes the energy (while keeping the rest of the molecule\n"
		"fixed). In general (for molecules with more than\n"
		"one rotatable bond), this procedure will not find\n"
		"the global minimum, but it will at least get rid of any bad\n"
		"clashes, and it do so quickly.\n"
		"\n"
		"Torsions closer to the center\n"
		"of the molecule will be optimized first as these most likely\n"
		"to generate large clashes.\n"
		"\n"
		"One possible use of this procedure is to prepare a reasonable 3D structure\n"
		"of a molecule for viewing. Another is to prepare the starting structure\n"
		"for a more systematic rotor search (in which case you should geometry\n"
		"optimize the final structure).\n"
		"\n"
		":type permute: boolean, optional\n"
		":param permute: Whether or not to permute the order of the 4 most central rotors.\n"
		"                   Default is true. This does a more thorough search, but takes 4! = 24 times\n"
		"                   as long.\n"
		"Since: version 2.4\n"
		""},
	 { "OBForceField_DiverseConfGen", _wrap_OBForceField_DiverseConfGen, METH_VARARGS, "Since: version 2.4"},
	 { "OBForceField_SetLineSearchType", _wrap_OBForceField_SetLineSearchType, METH_VARARGS, "\n"
		"Set the LineSearchType. The default type is LineSearchType::Newton2Num.\n"
		":type type: int\n"
		":param type: The LineSearchType to be used in SteepestDescent and ConjugateGradients.\n"
		""},
	 { "OBForceField_GetLineSearchType", _wrap_OBForceField_GetLineSearchType, METH_O, "\n"
		"Get the LineSearchType.\n"
		":rtype: int\n"
		":return: The current LineSearchType.\n"
		""},
	 { "OBForceField_LineSearch", _wrap_OBForceField_LineSearch, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Perform a linesearch starting at atom in direction direction.\n"
		"Deprecated: Current code should use LineSearch(double *, double*) instead.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Perform a linesearch for the entire molecule in direction ``direction``.\n"
		" This function is called when using LineSearchType::Simple.\n"
		"\n"
		" :type currentCoords: float\n"
		" :param currentCoords: Start coordinates.\n"
		" :type direction: float\n"
		" :param direction: The search direction.\n"
		" :rtype: float\n"
		" :return: alpha, The scale of the step we moved along the direction vector.\n"
		"\n"
		" Title: Output to log:\n"
		" OBFF_LOGLVL_NONE:   none\n"
		"\n"
		" OBFF_LOGLVL_LOW:    none\n"
		"\n"
		" OBFF_LOGLVL_MEDIUM: none\n"
		"\n"
		" OBFF_LOGLVL_HIGH:   none\n"
		""},
	 { "OBForceField_Newton2NumLineSearch", _wrap_OBForceField_Newton2NumLineSearch, METH_VARARGS, "\n"
		"Perform a linesearch for the entire molecule.\n"
		"This function is called when using LineSearchType::Newton2Num.\n"
		"\n"
		":type direction: float\n"
		":param direction: The search direction.\n"
		":rtype: float\n"
		":return: alpha, The scale of the step we moved along the direction vector.\n"
		"\n"
		"Title: Output to log:\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    none\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: none\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   none\n"
		""},
	 { "OBForceField_LineSearchTakeStep", _wrap_OBForceField_LineSearchTakeStep, METH_VARARGS, "\n"
		"Set the coordinates of the atoms to origCoord + step.\n"
		":type origCoords: float\n"
		":param origCoords: Start coordinates.\n"
		":type direction: float\n"
		":param direction: The search direction.\n"
		":type step: float\n"
		":param step: The step to take.\n"
		""},
	 { "OBForceField_SteepestDescent", _wrap_OBForceField_SteepestDescent, METH_VARARGS, "\n"
		"Perform steepest descent optimalization for steps steps or until convergence criteria is reached.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps and first step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_SteepestDescentInitialize", _wrap_OBForceField_SteepestDescentInitialize, METH_VARARGS, "\n"
		"Initialize steepest descent optimalization, to be used in combination with SteepestDescentTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->SteepestDescentInitialize(100, 1e-5f);\n"
		"    while (pFF->SteepestDescentTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, SteepestDescent() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_SteepestDescentTakeNSteps", _wrap_OBForceField_SteepestDescentTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in a steepestdescent optimalization that was previously initialized with SteepestDescentInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by SteepestDescentInitialize() has been reached.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    step number, energy and energy for the previous step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradients", _wrap_OBForceField_ConjugateGradients, METH_VARARGS, "\n"
		"Perform conjugate gradient optimalization for steps steps or until convergence criteria is reached.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    information about the progress of the minimization\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradientsInitialize", _wrap_OBForceField_ConjugateGradientsInitialize, METH_VARARGS, "\n"
		"Initialize conjugate gradient optimalization and take the first step, to be\n"
		"used in combination with ConjugateGradientsTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    pFF->ConjugateGradientsInitialize(100, 1e-5f);\n"
		"    while (pFF->ConjugateGradientsTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, ConjugateGradients() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (defualt is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    header including number of steps and first step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_ConjugateGradientsTakeNSteps", _wrap_OBForceField_ConjugateGradientsTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in a conjugate gradient optimalization that was previously\n"
		"initialized with ConjugateGradientsInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by ConjugateGradientsInitialize() has been reached.\n"
		"\n"
		"Title: Output to log:\n"
		"This function should only be called with the log level set to OBFF_LOGLVL_NONE or OBFF_LOGLVL_LOW. Otherwise\n"
		"too much information about the energy calculations needed for the minimization will interfere with the list\n"
		"of energies for succesive steps.\n"
		"\n"
		"\n"
		"OBFF_LOGLVL_NONE:   none\n"
		"\n"
		"OBFF_LOGLVL_LOW:    step number, energy and energy for the previous step\n"
		"\n"
		"OBFF_LOGLVL_MEDIUM: see note above\n"
		"\n"
		"OBFF_LOGLVL_HIGH:   see note above\n"
		""},
	 { "OBForceField_LBFGS", _wrap_OBForceField_LBFGS, METH_VARARGS, "\n"
		"Perform L-BFGS optimization for steps steps or until convergence criteria is reached.\n"
		"\n"
		"L-BFGS is a quasi-Newton method that approximates the inverse Hessian\n"
		"from a limited history of gradient changes. For typical molecular\n"
		"systems it converges faster than steepest descent or conjugate\n"
		"gradients while using comparable memory.\n"
		"\n"
		"Note: the line search algorithm chosen via SetLineSearchType()\n"
		"(Simple, Newton2Num) does NOT apply here. L-BFGS uses an internal\n"
		"backtracking line search with the Armijo condition; mixing in an\n"
		"arbitrary line search would break the curvature approximation.\n"
		"\n"
		":type steps: int\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (default is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		""},
	 { "OBForceField_LBFGSInitialize", _wrap_OBForceField_LBFGSInitialize, METH_VARARGS, "\n"
		"Initialize L-BFGS optimization, to be used in combination with LBFGSTakeNSteps().\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to an OBForceField class\n"
		"    pFF->LBFGSInitialize(100, 1e-5f);\n"
		"    while (pFF->LBFGSTakeNSteps(5)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		"If you don't need any updating in your program, LBFGS() is recommended.\n"
		"\n"
		":type steps: int, optional\n"
		":param steps: The number of steps.\n"
		":type econv: float, optional\n"
		":param econv: Energy convergence criteria. (default is 1e-6)\n"
		":type method: int, optional\n"
		":param method: Deprecated. (see HasAnalyticalGradients())\n"
		""},
	 { "OBForceField_LBFGSTakeNSteps", _wrap_OBForceField_LBFGSTakeNSteps, METH_VARARGS, "\n"
		"Take n steps in an L-BFGS optimization previously initialized with LBFGSInitialize().\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":rtype: boolean\n"
		":return: False if convergence or the number of steps given by LBFGSInitialize() has been reached.\n"
		""},
	 { "OBForceField_GenerateVelocities", _wrap_OBForceField_GenerateVelocities, METH_O, "Generate starting velocities with a Maxwellian distribution."},
	 { "OBForceField_CorrectVelocities", _wrap_OBForceField_CorrectVelocities, METH_O, "\n"
		"Correct the velocities so that the following is true:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"          3N\n"
		"         ----\n"
		"    0.5  \\    m_i * v_i^2 = 0.5 * Ndf * R * T = E_kin\n"
		"         /\n"
		"         ----\n"
		"         i=1\n"
		"\n"
		"    E_kin : kinetic energy\n"
		"    m_i : mass of atom i\n"
		"    v_i : velocity of atom i\n"
		"    Ndf : number of degrees of freedom (3 * number of atoms)\n"
		"    R : gas constant\n"
		"    T : temperature\n"
		""},
	 { "OBForceField_MolecularDynamicsTakeNSteps", _wrap_OBForceField_MolecularDynamicsTakeNSteps, METH_VARARGS, "\n"
		"Take n steps at temperature T. If no velocities are set, they will be generated.\n"
		"\n"
		"example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    // pFF is a pointer to a OBForceField class\n"
		"    while (pFF->MolecularDynamicsTakeNSteps(5, 300)) {\n"
		"      // do some updating in your program (redraw structure, ...)\n"
		"    }\n"
		"\n"
		":type n: int\n"
		":param n: The number of steps to take.\n"
		":type T: float\n"
		":param T: Absolute temperature in Kelvin.\n"
		":type timestep: float, optional\n"
		":param timestep: The time step in picoseconds. (10e-12 s)\n"
		":type method: int, optional\n"
		":param method: OBFF_ANALYTICAL_GRADIENTS (default) or OBFF_NUMERICAL_GRADIENTS\n"
		""},
	 { "OBForceField_GetConstraints", _wrap_OBForceField_GetConstraints, METH_O, "\n"
		"Get the current constraints.\n"
		":rtype: :py:class:`OBFFConstraints`\n"
		":return: The current constrains stored in the force field.\n"
		""},
	 { "OBForceField_SetConstraints", _wrap_OBForceField_SetConstraints, METH_VARARGS, "\n"
		"Set the constraints.\n"
		":type constraints: :py:class:`OBFFConstraints`\n"
		":param constraints: The new constraints to be used.\n"
		""},
	 { "OBForceField_SetFixAtom", _wrap_OBForceField_SetFixAtom, METH_VARARGS, "\n"
		"Fix the atom position until UnsetFixAtom() is called. This function\n"
		"can be used in programs that allow the user to interact with a molecule\n"
		"that is being minimized without having to check if the atom is already\n"
		"fixed in the constraints set by Setup() or SetConstraints(). Using this\n"
		"makes sure the selected atom follows the mouse cursur.\n"
		":type index: int\n"
		":param index: The index for the atom to fix.\n"
		""},
	 { "OBForceField_UnsetFixAtom", _wrap_OBForceField_UnsetFixAtom, METH_O, "\n"
		"Undo last SetFixAtom. This function will not remove the fix atom\n"
		"constraint for this atom if set by Setup() or SetConstraints().\n"
		""},
	 { "OBForceField_SetIgnoreAtom", _wrap_OBForceField_SetIgnoreAtom, METH_VARARGS, "\n"
		"Ignore the atom until UnsetIgnoreAtom() is called. This function\n"
		"can be used in programs that allow the user to interact with a molecule\n"
		"that is being minimized without having to check if the atom is already\n"
		"ignored in the constraints set by Setup() or SetConstraints(). Using this\n"
		"makes sure, in drawing mode, you can close rings without your newly\n"
		"created puching the other atoms away.\n"
		":type index: int\n"
		":param index: The index for the atom to ignore.\n"
		""},
	 { "OBForceField_UnsetIgnoreAtom", _wrap_OBForceField_UnsetIgnoreAtom, METH_O, "\n"
		"Undo last SetIgnoreAtom. This function will not remove the ignore atom\n"
		"constraint for this atom if set by Setup() or SetConstraints().\n"
		""},
	 { "OBForceField_IgnoreCalculation", _wrap_OBForceField_IgnoreCalculation, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"internal function\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"internal function\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"internal function\n"
		""},
	 { "OBForceField_DetectExplosion", _wrap_OBForceField_DetectExplosion, METH_O, "(debugging)"},
	 { "OBForceField_ValidateLineSearch", _wrap_OBForceField_ValidateLineSearch, METH_VARARGS, "(debugging)"},
	 { "OBForceField_ValidateSteepestDescent", _wrap_OBForceField_ValidateSteepestDescent, METH_VARARGS, "(debugging)"},
	 { "OBForceField_ValidateConjugateGradients", _wrap_OBForceField_ValidateConjugateGradients, METH_VARARGS, "(debugging)"},
	 { "OBForceField_Validate", _wrap_OBForceField_Validate, METH_O, "Validate the force field implementation (debugging)"},
	 { "OBForceField_ValidateGradients", _wrap_OBForceField_ValidateGradients, METH_O, "\n"
		"Validate the analytical gradients by comparing them to numerical ones. This function has to\n"
		"be implemented force field specific. (debugging)\n"
		""},
	 { "OBForceField_ValidateGradientError", _wrap_OBForceField_ValidateGradientError, METH_VARARGS, "\n"
		"Calculate the error of the analytical gradient (debugging)\n"
		":rtype: :py:class:`vector3`\n"
		":return: error = fabs(numgrad - anagrad) / anagrad * 100%\n"
		""},
	 { "OBForceField_VectorBondDerivative", _wrap_OBForceField_VectorBondDerivative, METH_VARARGS, "\n"
		" Calculate the derivative of a vector length. The vector is given by a - b,\n"
		"the length of this vector rab = sqrt(ab.x^2 + ab.y^2 + ab.z^2).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":rtype: float\n"
		":return: The distance between a and b (bondlength for bond stretching, separation for vdw, electrostatic)\n"
		""},
	 { "OBForceField_VectorDistanceDerivative", _wrap_OBForceField_VectorDistanceDerivative, METH_VARARGS, "\n"
		"To be used for VDW or Electrostatic interactions. This\n"
		"is faster than VectorBondDerivative, but does no error checking.\n"
		""},
	 { "OBForceField_VectorLengthDerivative", _wrap_OBForceField_VectorLengthDerivative, METH_VARARGS, "Deprecated:"},
	 { "OBForceField_VectorAngleDerivative", _wrap_OBForceField_VectorAngleDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a angle a-b-c. The angle is given by dot(ab,cb)/rab*rcb.\n"
		" Used for harmonic (cubic) angle potentials.\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":rtype: float\n"
		":return: The angle between a-b-c\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorOOPDerivative", _wrap_OBForceField_VectorOOPDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a OOP angle a-b-c-d. b is the central atom, and a-b-c is the plane.\n"
		"The OOP angle is given by 90° - arccos(dot(corss(ab,cb),db)/rabbc*rdb).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type pos_d: float\n"
		":param pos_d: atom d (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":type force_d: float\n"
		":param force_d: - return value for the force on atom d\n"
		":rtype: float\n"
		":return: The OOP angle for a-b-c-d\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorTorsionDerivative", _wrap_OBForceField_VectorTorsionDerivative, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Calculate the derivative of a torsion angle a-b-c-d. The torsion angle is given by arccos(dot(corss(ab,bc),cross(bc,cd))/rabbc*rbccd).\n"
		":type pos_a: float\n"
		":param pos_a: atom a (coordinates)\n"
		":type pos_b: float\n"
		":param pos_b: atom b (coordinates)\n"
		":type pos_c: float\n"
		":param pos_c: atom c (coordinates)\n"
		":type pos_d: float\n"
		":param pos_d: atom d (coordinates)\n"
		":type force_a: float\n"
		":param force_a: - return value for the force on atom a\n"
		":type force_b: float\n"
		":param force_b: - return value for the force on atom b\n"
		":type force_c: float\n"
		":param force_c: - return value for the force on atom c\n"
		":type force_d: float\n"
		":param force_d: - return value for the force on atom d\n"
		":rtype: float\n"
		":return: The tosion angle for a-b-c-d\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Deprecated: \n"
		""},
	 { "OBForceField_VectorSubtract", _wrap_OBForceField_VectorSubtract, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: pointer to result[3], will be set to i - j\n"
		""},
	 { "OBForceField_VectorAdd", _wrap_OBForceField_VectorAdd, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: pointer to result[3], will be set to i + j\n"
		""},
	 { "OBForceField_VectorDivide", _wrap_OBForceField_VectorDivide, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type n: float\n"
		":param n: divide x,y,z with n\n"
		":type result: float\n"
		":param result: pointer to result[3]\n"
		""},
	 { "OBForceField_VectorMultiply", _wrap_OBForceField_VectorMultiply, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type n: float\n"
		":param n: multiply x,y,z with n\n"
		":type result: float\n"
		":param result: pointer to result[3]\n"
		""},
	 { "OBForceField_VectorSelfMultiply", _wrap_OBForceField_VectorSelfMultiply, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3], multiply this vector by n and set this vector to the result.\n"
		":type n: float\n"
		":param n: the scalar value to be multipled\n"
		""},
	 { "OBForceField_VectorNormalize", _wrap_OBForceField_VectorNormalize, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3] to be normalized\n"
		""},
	 { "OBForceField_VectorCopy", _wrap_OBForceField_VectorCopy, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type from: float\n"
		":param from: pointer to i[3] to be copied from\n"
		":type to: float\n"
		":param to: pointer to j[3] to be copied to\n"
		""},
	 { "OBForceField_VectorLength", _wrap_OBForceField_VectorLength, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":rtype: float\n"
		":return: the vector length\n"
		""},
	 { "OBForceField_VectorDistance", _wrap_OBForceField_VectorDistance, METH_VARARGS, NULL},
	 { "OBForceField_VectorAngle", _wrap_OBForceField_VectorAngle, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":rtype: float\n"
		":return: the vector angle ijk (deg)\n"
		""},
	 { "OBForceField_VectorTorsion", _wrap_OBForceField_VectorTorsion, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":type l: float\n"
		":param l: pointer to l[3]\n"
		":rtype: float\n"
		":return: the vector torson ijkl (deg)\n"
		""},
	 { "OBForceField_VectorOOP", _wrap_OBForceField_VectorOOP, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type k: float\n"
		":param k: pointer to k[3]\n"
		":type l: float\n"
		":param l: pointer to l[3]\n"
		":rtype: float\n"
		":return: the vector torson ijkl (deg)\n"
		""},
	 { "OBForceField_VectorClear", _wrap_OBForceField_VectorClear, METH_O, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3], will set x,y,z to 0,0,0\n"
		""},
	 { "OBForceField_VectorDot", _wrap_OBForceField_VectorDot, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":rtype: float\n"
		":return: the dot product\n"
		""},
	 { "OBForceField_VectorCross", _wrap_OBForceField_VectorCross, METH_VARARGS, "\n"
		" inline fuction to speed up minimization speed\n"
		":type i: float\n"
		":param i: pointer to i[3]\n"
		":type j: float\n"
		":param j: pointer to j[3]\n"
		":type result: float\n"
		":param result: the dot product (as a return value double[3])\n"
		""},
	 { "OBForceField_PrintVector", _wrap_OBForceField_PrintVector, METH_O, NULL},
	 { "OBForceField_SetLogToStdOut", _wrap_OBForceField_SetLogToStdOut, METH_O, NULL},
	 { "OBForceField_SetLogToStdErr", _wrap_OBForceField_SetLogToStdErr, METH_O, NULL},
	 { "OBForceField_swigregister", OBForceField_swigregister, METH_O, NULL},
	 { "new_OBBuilder", _wrap_new_OBBuilder, METH_NOARGS, NULL},
	 { "OBBuilder_Build", _wrap_OBBuilder_Build, METH_VARARGS, "\n"
		"The mol object contains all connectivity information (atomic numbers, bonds, bond orders, ..)\n"
		"but no 3D coordinates. Build generates these coordinates and assigns them.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: Molecule with the connectivity (from SMILES for example). The coordinates are also\n"
		"           changed in this mol.\n"
		":type stereoWarnings: boolean, optional\n"
		":param stereoWarnings: Warn if the stereochemistry is incorrect (default is true)\n"
		""},
	 { "OBBuilder_SetKeepRings", _wrap_OBBuilder_SetKeepRings, METH_O, "\n"
		"If the molecule already contains 3D coordinates, if you set KeepRings to true it will use\n"
		"retain the 3D coordinates of the rings. By default KeepRings is false, and ring conformations\n"
		"are obtained by lookup in a library of ring conformers. However, since the ring conformer library\n"
		"is not exhaustive, if the ring system is not found in the library, the resulting 3D structure can\n"
		"be poor, and require geometry optimisation before it is reasonable. If your starting point is\n"
		"a 3D structure, you can set KeepRings to true, and the conformation will be taken from the input.\n"
		"The remaining (acyclic) bonds will still all be built by the builder.\n"
		""},
	 { "OBBuilder_UnsetKeepRings", _wrap_OBBuilder_UnsetKeepRings, METH_O, NULL},
	 { "OBBuilder_AddRingFragment", _wrap_OBBuilder_AddRingFragment, METH_VARARGS, "Used by LoadFragments to check for invalid (all zero coordinates) fragments"},
	 { "OBBuilder_LoadFragments", _wrap_OBBuilder_LoadFragments, METH_O, "Load fragment info from file, if is it has not already been done"},
	 { "OBBuilder_GetFragmentCoord", _wrap_OBBuilder_GetFragmentCoord, METH_VARARGS, NULL},
	 { "OBBuilder_GetNewBondVector", _wrap_OBBuilder_GetNewBondVector, METH_VARARGS, "\n"
		" Get the position for a new neighbour on atom.  Returns\n"
		"non-finite vector if there is no reasonable location.\n"
		" :type atom: :py:class:`OBAtom`\n"
		" :param atom: Atom for which we want a new neighbour location.\n"
		" :rtype: :py:class:`vector3`\n"
		" :return: The position for the new atom.\n"
		""},
	 { "OBBuilder_Connect", _wrap_OBBuilder_Connect, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":type newpos: :py:class:`vector3`\n"
		":param newpos: Direction for new bond between a and b\n"
		":type bondOrder: int, optional\n"
		":param bondOrder: Bond order of the new bond between a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":type bondOrder: int, optional\n"
		":param bondOrder: Bond order of the new bond bewtween a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Atoms a and b are part of two fragments that are not connected in mol.\n"
		"Connect will translate and rotate the fragment that contains b so that\n"
		"a and b are separated by a bond. This bond is also added.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to be modified\n"
		":type a: int\n"
		":param a: Index for atom in fragment that should not be rotated.\n"
		":type b: int\n"
		":param b: Index for atom in fragment that should be rotated.\n"
		":param bondOrder: Bond order of the new bond bewtween a and b.\n"
		":rtype: boolean\n"
		":return: true if successful or fails when failed (most likely cause\n"
		"    for failing: a and b are in the same fragment, they are connected)\n"
		""},
	 { "OBBuilder_Swap", _wrap_OBBuilder_Swap, METH_VARARGS, "\n"
		"Swap group b, bonded to a with group d, bonded to c. The bonds a-b and b-c cannot be\n"
		"part of a ring. Atoms a and b will not be moved. Atoms b, d and their connected atoms\n"
		"(after deleting bonds ab and cd) will be translated/rotated.\n"
		"\n"
		"Example:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"      \\ /                            /\n"
		"       b                            d\n"
		"        \\     /     Swap(a,b,c,d)    \\     /\n"
		"         a---x          ---->         a---x\n"
		"        /     \\     /                /     \\     /\n"
		"       x       c---d                x       c---b\n"
		"                                                 \\\n"
		"\n"
		"\n"
		"This function can also be used to invert chiral centers if a and c are the same atom.\n"
		"\n"
		"Example\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"       1                        3\n"
		"       |      Swap(C,1,C,3)     |\n"
		"    2>-C-<3      ----->      2>-C-<1\n"
		"       |                        |\n"
		"       4                        4\n"
		""},
	 { "OBBuilder_CorrectStereoBonds", _wrap_OBBuilder_CorrectStereoBonds, METH_O, "\n"
		" Atoms a and b must be bonded and this bond cannot be part of a ring. The bond will\n"
		" be broken and the smiles fragment will be inserted bewteen the two remaining fragments.\n"
		" The fragment that contains a will not be translated or rotated. Parameters c and d are\n"
		" the index in the smiles to which atoms a and b will be connected respectivly.\n"
		"\n"
		"        Correct double bond stereochemistry\n"
		"\n"
		":rtype: boolean\n"
		":return: Success or failure\n"
		""},
	 { "OBBuilder_CorrectStereoAtoms", _wrap_OBBuilder_CorrectStereoAtoms, METH_VARARGS, "\n"
		" Correct stereochemistry at tetrahedral atoms with at least two non-ring\n"
		"bonds. It also works for spiro atoms.\n"
		"\n"
		":rtype: boolean\n"
		":return: Success or failure\n"
		""},
	 { "OBBuilder_IsSpiroAtom", _wrap_OBBuilder_IsSpiroAtom, METH_VARARGS, "Does this atom connect two rings which are not otherwise connected?"},
	 { "OBBuilder_GetFragment", _wrap_OBBuilder_GetFragment, METH_O, "\n"
		"Get the fragment to which this atom belongs.\n"
		":type atom: :py:class:`OBAtom`\n"
		":param atom: Atom in the fragment.\n"
		":rtype: :py:class:`OBBitVec`\n"
		":return: The OBBitVec defining the fragment to which a belongs.\n"
		""},
	 { "OBBuilder_AddNbrs", _wrap_OBBuilder_AddNbrs, METH_VARARGS, NULL},
	 { "delete_OBBuilder", _wrap_delete_OBBuilder, METH_O, NULL},
	 { "OBBuilder_swigregister", OBBuilder_swigregister, METH_O, NULL},
	 { "OBBuilder_swiginit", OBBuilder_swiginit, METH_VARARGS, NULL},
	 { "OBOp_Default", _wrap_OBOp_Default, METH_NOARGS, NULL},
	 { "OBOp_FindType", _wrap_OBOp_FindType, METH_O, NULL},
	 { "OBOp_Do", _wrap_OBOp_Do, METH_VARARGS, "Required function that does the work. Normally return true, unless object is not to be output."},
	 { "OBOp_WorksWith", _wrap_OBOp_WorksWith, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: true if this op is designed to work with the class of pOb, e.g. OBMol\n"
		""},
	 { "OBOp_ProcessVec", _wrap_OBOp_ProcessVec, METH_VARARGS, "Do something with an array of objects. Used a a callback routine in OpSort, etc."},
	 { "OBOp_OpOptions", _wrap_OBOp_OpOptions, METH_O, "\n"
		":rtype: string\n"
		":return: string describing options, for display with -H and to make checkboxes in GUI\n"
		""},
	 { "OBOp_DoOps", _wrap_OBOp_DoOps, METH_VARARGS, "\n"
		"Call Do() of all the OBOps whose ID is a key in the map.\n"
		"Called from DoTransformations(). The map has general options like -x or --multicharoption\n"
		"The key is the option name and the value, if any, is text which follows the option name.\n"
		" In some cases, there may be several parameters, space separated)\n"
		" :rtype: boolean\n"
		" :return: false indicating object should not be output, if any Do() returns false\n"
		""},
	 { "delete_OBOp", _wrap_delete_OBOp, METH_O, NULL},
	 { "OBOp_swigregister", OBOp_swigregister, METH_O, NULL},
	 { "OBChargeModel_Default", _wrap_OBChargeModel_Default, METH_NOARGS, NULL},
	 { "new_OBChargeModel", _wrap_new_OBChargeModel, METH_VARARGS, NULL},
	 { "OBChargeModel_FindType", _wrap_OBChargeModel_FindType, METH_O, NULL},
	 { "OBChargeModel_ComputeCharges", _wrap_OBChargeModel_ComputeCharges, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: whether partial charges were successfully assigned to this molecule\n"
		"Notes: The method should fill m_partialCharges and m_formalCharges as well\n"
		""},
	 { "OBChargeModel_GetFormalCharges", _wrap_OBChargeModel_GetFormalCharges, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a vector of the formal charges on each atom, indexed from 0\n"
		"    This method returns floating point formal charges since some\n"
		"    charge models consider fractional charges (e.g., 0.5 for an\n"
		"    oxygen in a carboxylate CO2- group).\n"
		"Notes: If ComputeCharges() has not been called, this will return an empty vector\n"
		""},
	 { "OBChargeModel_GetPartialCharges", _wrap_OBChargeModel_GetPartialCharges, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a vector of the partial charges on each atom, indexed from 0\n"
		"Notes: If ComputeCharges() has not been called, this will return an empty vector\n"
		""},
	 { "OBChargeModel_GetDipoleMoment", _wrap_OBChargeModel_GetDipoleMoment, METH_VARARGS, "\n"
		":rtype: :py:class:`vector3`\n"
		":return: a vector of the dipole moment from this molecule\n"
		""},
	 { "delete_OBChargeModel", _wrap_delete_OBChargeModel, METH_O, NULL},
	 { "OBChargeModel_swigregister", OBChargeModel_swigregister, METH_O, NULL},
	 { "OBChargeModel_swiginit", OBChargeModel_swiginit, METH_VARARGS, NULL},
	 { "new_OBChemTsfm", _wrap_new_OBChemTsfm, METH_NOARGS, NULL},
	 { "delete_OBChemTsfm", _wrap_delete_OBChemTsfm, METH_O, NULL},
	 { "OBChemTsfm_Init", _wrap_OBChemTsfm_Init, METH_VARARGS, "Initialize this transformation with the supplied SMARTS patterns"},
	 { "OBChemTsfm_Apply", _wrap_OBChemTsfm_Apply, METH_VARARGS, "Apply this transformation to all matches in the supplied OBMol"},
	 { "OBChemTsfm_IsAcid", _wrap_OBChemTsfm_IsAcid, METH_O, "\n"
		"Is this transformation an acid dissociation?\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"        Ka\n"
		"    HA ----> A(-)         (the H(+) will be deleted)\n"
		"\n"
		"IsAcid() will check the charge in the end SMARTS pattern.\n"
		":rtype: boolean\n"
		":return: true if the charge is less than 0 (-1).\n"
		""},
	 { "OBChemTsfm_IsBase", _wrap_OBChemTsfm_IsBase, METH_O, "\n"
		"Is this a transformation to the conjugated acid from a base?\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"        Ka\n"
		"    HA ----> A(-)         (the H(+) will be deleted)\n"
		"\n"
		"IsBase() will check the charge in the end SMARTS pattern.\n"
		":rtype: boolean\n"
		":return: true if the charge is higher than 0 (+1).\n"
		""},
	 { "OBChemTsfm_swigregister", OBChemTsfm_swigregister, METH_O, NULL},
	 { "OBChemTsfm_swiginit", OBChemTsfm_swiginit, METH_VARARGS, NULL},
	 { "new_OBPhModel", _wrap_new_OBPhModel, METH_NOARGS, NULL},
	 { "delete_OBPhModel", _wrap_delete_OBPhModel, METH_O, NULL},
	 { "OBPhModel_AssignSeedPartialCharge", _wrap_OBPhModel_AssignSeedPartialCharge, METH_VARARGS, NULL},
	 { "OBPhModel_CorrectForPH", _wrap_OBPhModel_CorrectForPH, METH_VARARGS, NULL},
	 { "OBPhModel_swigregister", OBPhModel_swigregister, METH_O, NULL},
	 { "OBPhModel_swiginit", OBPhModel_swiginit, METH_VARARGS, NULL},
	 { "new_OBGraphSym", _wrap_new_OBGraphSym, METH_VARARGS, "Constructor"},
	 { "delete_OBGraphSym", _wrap_delete_OBGraphSym, METH_O, "Destructor"},
	 { "OBGraphSym_GetSymmetry", _wrap_OBGraphSym_GetSymmetry, METH_VARARGS, "\n"
		"Calculate the symmetry classes for the molecule. The result will be\n"
		"stored in ``symmetry_classes``.\n"
		"\n"
		"The results in ``symmetry_classes`` will be ordered by symmetry\n"
		"classes. Use the OBAtom* pointer in the std::pair to match the atoms\n"
		"with the right symmetry classes.\n"
		"\n"
		":rtype: int\n"
		":return: The number of symmetry classes.\n"
		""},
	 { "OBGraphSym_ClearSymmetry", _wrap_OBGraphSym_ClearSymmetry, METH_O, "\n"
		"Clear the symmetry classes data stored in the molecule specified when\n"
		"construting the OBGraphSym object.\n"
		""},
	 { "OBGraphSym_swigregister", OBGraphSym_swigregister, METH_O, NULL},
	 { "OBGraphSym_swiginit", OBGraphSym_swiginit, METH_VARARGS, NULL},
	 { "delete_OBIsomorphismMapper", _wrap_delete_OBIsomorphismMapper, METH_O, NULL},
	 { "OBIsomorphismMapper_GetInstance", _wrap_OBIsomorphismMapper_GetInstance, METH_VARARGS, "\n"
		"Get a pointer to an instance of the specified ``algorithm``. This pointer\n"
		"has to be delted when the instance is no longer needed.\n"
		":type query: :py:class:`OBQuery`\n"
		":param query: The search query to be mapped.\n"
		":type algorithm: string, optional\n"
		":param algorithm: The algorithm for the mapper.\n"
		":rtype: :py:class:`OBIsomorphismMapper`\n"
		":return: OBIsomorphismMapper instance or 0 if there is no subclass implementing\n"
		"    the specified ``algorithm``.\n"
		""},
	 { "OBIsomorphismMapper_MapFirst", _wrap_OBIsomorphismMapper_MapFirst, METH_VARARGS, "\n"
		"Find a single mapping in ``queried``.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type map: :py:class:`Mapping`\n"
		":param map: Reference to the object to store the result in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapUnique", _wrap_OBIsomorphismMapper_MapUnique, METH_VARARGS, "\n"
		"Find all unique mappings in ``queried``. A mapping is unique when there is no previous\n"
		"mapping covering the same queried atoms. For two mappings, some overlap is allowed but\n"
		"at least one atom should be different.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type maps: :py:class:`Mappings`\n"
		":param maps: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapAll", _wrap_OBIsomorphismMapper_MapAll, METH_VARARGS, "\n"
		"Find all mappings in ``queried``. This function is used by FindAutomorphisms()\n"
		"with a query that is a copy of the queried molecule (taking the mask into\n"
		"account).\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type maps: :py:class:`Mappings`\n"
		":param maps: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		":type maxMemory: std::size_t, optional\n"
		":param maxMemory: Memory limit for the ``maps`` object in bytes. Default is 300MB.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_MapGeneric", _wrap_OBIsomorphismMapper_MapGeneric, METH_VARARGS, "\n"
		"Find all mappings in ``queried``. The functor will be called when a mapping is found.\n"
		":type functor: OpenBabel::OBIsomorphismMapper::Functor\n"
		":param functor: The functor to handle found mappings.\n"
		":type queried: :py:class:`OBMol`\n"
		":param queried: The molecule to search.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A mask to restrict the search to a part of the queried molecule.\n"
		"    The default empty mask will result in all atoms being considered. The mask\n"
		"    indexes start from 1 (i.e. OBAtom::GetIdx()).\n"
		""},
	 { "OBIsomorphismMapper_SetTimeout", _wrap_OBIsomorphismMapper_SetTimeout, METH_VARARGS, "Set the timeout in seconds."},
	 { "OBIsomorphismMapper_swigregister", OBIsomorphismMapper_swigregister, METH_O, NULL},
	 { "MapsTo", _wrap_MapsTo, METH_VARARGS, NULL},
	 { "FindAutomorphisms", _wrap_FindAutomorphisms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function is a wrapper for FindAutomorphisms with a functor to store all\n"
		"automorphisms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule for which to find the automorphisms.\n"
		":type aut: std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > >\n"
		":param aut: The result will be stored here.\n"
		":type symmetry_classes: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param symmetry_classes: The graph invariants to use. See OBGraphSym or use\n"
		"    the FindAutomorphisms function that computes these for you below.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: A bit vector specifying which atoms to consider. An empty mask\n"
		"    will consider all atoms. The bits are indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":type maxMemory: std::size_t, optional\n"
		":param maxMemory: Maximum memory limit for ``aut``. The number of automorphisms\n"
		"    for certain graphs can be large. The default is 300MB, consider using a functor\n"
		"    to process automorphisms when they are found.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"function will first find the graph invariants (i.e. symmetry_classes) using\n"
		"the mask. This function is a wrapper for FindAutomorphisms with a functor to\n"
		"store all automorphisms.\n"
		"\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"is the main implementation for finding automorphisms and uses an\n"
		"OBIsomorphismMapper::Functor to process found isomorphisms. Wrapper functions\n"
		"are provided to find and store all automorphisms but the number of automorphisms\n"
		"can be large for certain graphs making it not feasible to store all automorphisms\n"
		"in memory (RAM).\n"
		"\n"
		"See also: 'MapGeneric'\n"
		"Since: version 2.3\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 6:*\n"
		"\n"
		"Find the automorphisms of a molecule by using an OBIsomorphismMapper. This\n"
		"is the main implementation for finding automorphisms and uses an\n"
		"OBIsomorphismMapper::Functor to process found isomorphisms. Wrapper functions\n"
		"are provided to find and store all automorphisms but the number of automorphisms\n"
		"can be large for certain graphs making it not feasible to store all automorphisms\n"
		"in memory (RAM).\n"
		"\n"
		"See also: 'MapGeneric'\n"
		"Since: version 2.3\n"
		""},
	 { "new_OBQueryAtom", _wrap_new_OBQueryAtom, METH_VARARGS, "\n"
		"Constructor.\n"
		":type atomicNum: int, optional\n"
		":param atomicNum: The atomic number for this query atom.\n"
		":type isInRing: boolean, optional\n"
		":param isInRing: Specify whether the query atom is in a ring. Default is false.\n"
		":type isAromatic: boolean, optional\n"
		":param isAromatic: Specify whether the query atom is aromatic. Default is false.\n"
		""},
	 { "delete_OBQueryAtom", _wrap_delete_OBQueryAtom, METH_O, NULL},
	 { "OBQueryAtom_GetIndex", _wrap_OBQueryAtom_GetIndex, METH_O, "\n"
		"Get the index for this query atom. Atoms are indexed starting from 0.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_GetBonds", _wrap_OBQueryAtom_GetBonds, METH_O, "\n"
		"Get the query bonds for this atom.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_GetNbrs", _wrap_OBQueryAtom_GetNbrs, METH_O, "\n"
		"Get the neighbor query atoms.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		""},
	 { "OBQueryAtom_Matches", _wrap_OBQueryAtom_Matches, METH_VARARGS, "\n"
		"This is the match method to verify if an OBQueryAtom and OBAtom class match.\n"
		"The default implementation only checks if the atomic numbers match. Reimplement\n"
		"this method in a subclass for more advances matching.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		":type atom: :py:class:`OBAtom`\n"
		":param atom: The OBAtom object to compare this OBQueryAtom with.\n"
		""},
	 { "OBQueryAtom_swigregister", OBQueryAtom_swigregister, METH_O, NULL},
	 { "OBQueryAtom_swiginit", OBQueryAtom_swiginit, METH_VARARGS, NULL},
	 { "new_OBQueryBond", _wrap_new_OBQueryBond, METH_VARARGS, "Constructor."},
	 { "delete_OBQueryBond", _wrap_delete_OBQueryBond, METH_O, NULL},
	 { "OBQueryBond_GetIndex", _wrap_OBQueryBond_GetIndex, METH_O, "Get the index for this query bonds. Query bonds are indexed starting from 0."},
	 { "OBQueryBond_GetBeginAtom", _wrap_OBQueryBond_GetBeginAtom, METH_O, "Get the begin atom."},
	 { "OBQueryBond_GetEndAtom", _wrap_OBQueryBond_GetEndAtom, METH_O, "Get the end atom."},
	 { "OBQueryBond_Matches", _wrap_OBQueryBond_Matches, METH_VARARGS, "\n"
		"This is the match method to verify if an OBQueryBond and OBBond class match.\n"
		"The default implementation checks if both bonds are aromatic and compares the\n"
		"bond orders otherwise. Reimplement this method in a subclass for more\n"
		"advances matching.\n"
		"This method is used by OBIsomorphismMapper implementations.\n"
		":type bond: :py:class:`OBBond`\n"
		":param bond: The OBBond object to compare this OBQueryBond with.\n"
		""},
	 { "OBQueryBond_swigregister", OBQueryBond_swigregister, METH_O, NULL},
	 { "OBQueryBond_swiginit", OBQueryBond_swiginit, METH_VARARGS, NULL},
	 { "delete_OBQuery", _wrap_delete_OBQuery, METH_O, NULL},
	 { "OBQuery_NumAtoms", _wrap_OBQuery_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in the query.\n"
		""},
	 { "OBQuery_NumBonds", _wrap_OBQuery_NumBonds, METH_O, "\n"
		":rtype: int\n"
		":return: The number of bonds in the query.\n"
		""},
	 { "OBQuery_GetAtoms", _wrap_OBQuery_GetAtoms, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > >\n"
		":return: std::vector with pointers to the query atoms.\n"
		""},
	 { "OBQuery_GetBonds", _wrap_OBQuery_GetBonds, METH_O, "\n"
		":rtype: std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > >\n"
		":return: std::vector with pointers to the query bonds.\n"
		""},
	 { "OBQuery_GetBond", _wrap_OBQuery_GetBond, METH_VARARGS, "\n"
		":rtype: :py:class:`OBQueryBond`\n"
		":return: The query bond between ``begin`` and ``end``. If there is no\n"
		"    bond between ``begin`` and ``end``, this function returns 0.\n"
		""},
	 { "OBQuery_AddAtom", _wrap_OBQuery_AddAtom, METH_VARARGS, "Add a query atom to the query. This function steals the pointer."},
	 { "OBQuery_AddBond", _wrap_OBQuery_AddBond, METH_VARARGS, "Add a query atom to the query. This function steals the pointer."},
	 { "new_OBQuery", _wrap_new_OBQuery, METH_NOARGS, NULL},
	 { "OBQuery_swigregister", OBQuery_swigregister, METH_O, NULL},
	 { "OBQuery_swiginit", OBQuery_swiginit, METH_VARARGS, NULL},
	 { "CompileMoleculeQuery", _wrap_CompileMoleculeQuery, METH_VARARGS, "\n"
		"Create an OBQuery object from an OBMol object.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The query molecule.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The mask specifying the atoms to use. Indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":rtype: :py:class:`OBQuery`\n"
		":return: A pointer to an OBQuery object for the smiles string. This pointer should be deleted.\n"
		"Since: version 2.3\n"
		""},
	 { "CompileSmilesQuery", _wrap_CompileSmilesQuery, METH_VARARGS, "\n"
		"Create an OBQuery object from a smiles string.\n"
		":type smiles: string\n"
		":param smiles: The query smiles string.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The mask specifying the atoms to use. Indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":rtype: :py:class:`OBQuery`\n"
		":return: A pointer to an OBQuery object for the smiles string. This pointer should be deleted.\n"
		"Since: version 2.3\n"
		""},
	 { "CanonicalLabels", _wrap_CanonicalLabels, METH_VARARGS, "\n"
		"Calculate the canonical labels for the molecule. Stereochemistry is\n"
		"included in the algorithm and the canonical labels. The result will be\n"
		"stored in ``canonical_labels``.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type symmetry_classes: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param symmetry_classes: The symmetry_classes for the molecule. These can\n"
		"    be obtained using the OBGraphSym class.\n"
		":type canonical_labels: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":param canonical_labels: Reference to the object to store the results in.\n"
		":type mask: :py:class:`OBBitVec`, optional\n"
		":param mask: The fragment to label. When the bit for an atom is set, it is\n"
		"    included in the fragment. If no bits are set, all atoms will be included.\n"
		"    Atoms are indexed from 1 (i.e. OBAtom::GetIdx()).\n"
		":type maxSeconds: int, optional\n"
		":param maxSeconds: Timeout in seconds.\n"
		":type onlyOne: boolean, optional\n"
		":param onlyOne: If true, the first found labels are returned. These are\n"
		"    canonical labels without considering stereochemistry and other attributes\n"
		"    not included in the symmetry classes.\n"
		"\n"
		":rtype: void\n"
		":return: The canonical labels for the molecule in ``canonical_labels``.\n"
		"\n"
		"See also: 'canonical_code_algorithm'\n"
		"Since: 2.3\n"
		""},
	 { "OBStereo_MakeRefs", _wrap_OBStereo_MakeRefs, METH_VARARGS, "\n"
		"Create a Refs list filled with ``ref1``, ``ref2``, ``ref3`` & ``ref4``.\n"
		"``ref4`` is not added to the returned Refs if it is equal to NoRef.\n"
		"\n"
		":rtype: :py:class:`Refs`\n"
		":return: A Refs list containing the specified Ref values.\n"
		""},
	 { "OBStereo_ContainsSameRefs", _wrap_OBStereo_ContainsSameRefs, METH_VARARGS, "\n"
		"Check if ``refs1`` and ``refs2`` contain the same Ref values regardless\n"
		"of their order.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(2, 3, 1)) // true\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(3, 2, 1)) // true\n"
		"    OBStereo::ContainsSameRefs(OBStereo::MakeRefs(1, 2, 3), OBStereo::MakeRefs(3, 4, 1)) // false\n"
		"\n"
		":rtype: boolean\n"
		":return: True if ``refs1`` and ``refs2`` contain the same Ref values.\n"
		""},
	 { "OBStereo_ContainsRef", _wrap_OBStereo_ContainsRef, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if ``refs`` contains ``ref``.\n"
		""},
	 { "OBStereo_NumInversions", _wrap_OBStereo_NumInversions, METH_O, "\n"
		"Compute the inversion vector for ``refs`` and return the sum of it's\n"
		"elements. The ith element in the inversion vector is the number of\n"
		"element to the right of element i with a lower value.\n"
		"\n"
		"The number of inversions is the same as the number of interchanges\n"
		"of consecutive elements.\n"
		"\n"
		"When working with 3 refs from a tetrahedral configuration:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    permutation   inversion vector    sum\n"
		"    -------------------------------------\n"
		"    123           0 0 0               0 (even) -> clockwise\n"
		"    132           0 1 0               1 (odd)  -> anti-clockwise\n"
		"    213           1 0 0               1 (odd)  -> anti-clockwise\n"
		"    231           1 1 0               2 (even) -> clockwise\n"
		"    312           2 0 0               2 (even) -> clockwise\n"
		"    321           2 1 0               3 (odd)  -> anti-clockwise\n"
		""},
	 { "OBStereo_Permutate", _wrap_OBStereo_Permutate, METH_VARARGS, "\n"
		"Permutate element ``i`` with ``j`` in ``refs``.\n"
		"\n"
		":type refs: :py:class:`Refs`\n"
		":param refs: The sequence with N elements to permutate.\n"
		":type i: int\n"
		":param i: Element i (0...N-1) will be mutated to j and vice versa.\n"
		":type j: int\n"
		":param j: Element j (0...N-1) will be mutated to i and vice versa.\n"
		"\n"
		"Notes: This method does nothing if i equals j.\n"
		""},
	 { "OBStereo_Permutated", _wrap_OBStereo_Permutated, METH_VARARGS, "\n"
		"Get ``refs`` with element ``i`` and ``j`` permutated.\n"
		"\n"
		":type refs: :py:class:`Refs`\n"
		":param refs: The sequence with N elements to permutate.\n"
		":type i: int\n"
		":param i: Element ``i`` (0...N-1) will be mutated to ``j`` and vice versa.\n"
		":type j: int\n"
		":param j: Element ``j`` (0...N-1) will be mutated to ``i`` and vice versa.\n"
		"\n"
		":rtype: :py:class:`Refs`\n"
		":return: ``refs`` with elements ``i`` and ``j`` permutated.\n"
		"\n"
		"Notes: This method does nothing if ``i`` equals ``j``.\n"
		""},
	 { "new_OBStereo", _wrap_new_OBStereo, METH_NOARGS, NULL},
	 { "delete_OBStereo", _wrap_delete_OBStereo, METH_O, NULL},
	 { "OBStereo_swigregister", OBStereo_swigregister, METH_O, NULL},
	 { "OBStereo_swiginit", OBStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBStereoUnit", _wrap_new_OBStereoUnit, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Default constructor creating an invalid OBStereoUnit. The unit can be\n"
		"made valid by setting the type, id and para data members.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Constructor specifying all data to create a valid OBStereoUnit.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Constructor specifying all data to create a valid OBStereoUnit.\n"
		""},
	 { "OBStereoUnit_type_set", _wrap_OBStereoUnit_type_set, METH_VARARGS, "the type for this stereogenic unit"},
	 { "OBStereoUnit_type_get", _wrap_OBStereoUnit_type_get, METH_O, "the type for this stereogenic unit"},
	 { "OBStereoUnit_id_set", _wrap_OBStereoUnit_id_set, METH_VARARGS, "the atom/bond (depends on type) unique id"},
	 { "OBStereoUnit_id_get", _wrap_OBStereoUnit_id_get, METH_O, "the atom/bond (depends on type) unique id"},
	 { "OBStereoUnit_para_set", _wrap_OBStereoUnit_para_set, METH_VARARGS, "para- (=ressemble) or true-stereocenter"},
	 { "OBStereoUnit_para_get", _wrap_OBStereoUnit_para_get, METH_O, "para- (=ressemble) or true-stereocenter"},
	 { "delete_OBStereoUnit", _wrap_delete_OBStereoUnit, METH_O, NULL},
	 { "OBStereoUnit_swigregister", OBStereoUnit_swigregister, METH_O, NULL},
	 { "OBStereoUnit_swiginit", OBStereoUnit_swiginit, METH_VARARGS, NULL},
	 { "delete_OBStereoBase", _wrap_delete_OBStereoBase, METH_O, "Destructor."},
	 { "OBStereoBase_GetMolecule", _wrap_OBStereoBase_GetMolecule, METH_O, "\n"
		"Get the molecule. This can be used by subclasses when more\n"
		"information is needed (e.g. OBCisTransStereo::GetCisRef, ...).\n"
		""},
	 { "OBStereoBase_GetType", _wrap_OBStereoBase_GetType, METH_O, "Reimplemented by subclasses to return the type defined in OBStereo::Type."},
	 { "OBStereoBase_SetSpecified", _wrap_OBStereoBase_SetSpecified, METH_VARARGS, "\n"
		"Set whether the stereochemistry is specified. Comparing a specified\n"
		"OBStereoBase derived class (or it's Config struct) with an unspecified\n"
		"one, always returns true.\n"
		""},
	 { "OBStereoBase_IsSpecified", _wrap_OBStereoBase_IsSpecified, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if the stereochemistry is specified.\n"
		""},
	 { "OBStereoBase_swigregister", OBStereoBase_swigregister, METH_O, NULL},
	 { "new_OBStereoFacade", _wrap_new_OBStereoFacade, METH_VARARGS, "\n"
		"Constructor with ``mol`` and ``perceive`` parameter.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type perceive: boolean, optional\n"
		":param perceive: If true, PerceiveStereo will be called if the\n"
		"    OBMol::HasChiralityPerceived() flag is not set. (default is true)\n"
		""},
	 { "OBStereoFacade_NumTetrahedralStereo", _wrap_OBStereoFacade_NumTetrahedralStereo, METH_O, "Get the number of tetrahedral stereocenters."},
	 { "OBStereoFacade_GetAllTetrahedralStereo", _wrap_OBStereoFacade_GetAllTetrahedralStereo, METH_O, "Get all the OBTetrahedralStereo objects."},
	 { "OBStereoFacade_HasTetrahedralStereo", _wrap_OBStereoFacade_HasTetrahedralStereo, METH_VARARGS, "\n"
		"Check if atom with ``id`` is a tetrahedral center.\n"
		":rtype: boolean\n"
		":return: True if the atom with ``id`` has tetrahedral stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetTetrahedralStereo", _wrap_OBStereoFacade_GetTetrahedralStereo, METH_VARARGS, "\n"
		"Get the OBTetrahedralStereo object with ``atomId`` as center. This\n"
		"function returns 0 if there is no OBTetrahedralStereo object found\n"
		"with the specified center.\n"
		""},
	 { "OBStereoFacade_NumCisTransStereo", _wrap_OBStereoFacade_NumCisTransStereo, METH_O, "Get the number of cis/trans stereocenters."},
	 { "OBStereoFacade_GetAllCisTransStereo", _wrap_OBStereoFacade_GetAllCisTransStereo, METH_O, "Get all the OBCisTransStereo objects."},
	 { "OBStereoFacade_HasCisTransStereo", _wrap_OBStereoFacade_HasCisTransStereo, METH_VARARGS, "\n"
		"Check if bond with ``id`` is a stereogenic cis/trans double bond.\n"
		":rtype: boolean\n"
		":return: True if the bond with ``id`` has cis/trans stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetCisTransStereo", _wrap_OBStereoFacade_GetCisTransStereo, METH_VARARGS, "\n"
		"Get the OBTetrahedralStereo object with ``bondId`` as double bond.\n"
		"This function returns 0 if there is no OBCisTransStereo object found\n"
		"with the specified bond.\n"
		""},
	 { "OBStereoFacade_NumSquarePlanarStereo", _wrap_OBStereoFacade_NumSquarePlanarStereo, METH_O, "Get the number of square-planar stereocenters."},
	 { "OBStereoFacade_GetAllSquarePlanarStereo", _wrap_OBStereoFacade_GetAllSquarePlanarStereo, METH_O, "Get all the OBSquarePlanarStereo objects."},
	 { "OBStereoFacade_HasSquarePlanarStereo", _wrap_OBStereoFacade_HasSquarePlanarStereo, METH_VARARGS, "\n"
		"Check if atom with ``id`` is a stereogenic square-planar atom.\n"
		":rtype: boolean\n"
		":return: True if the atom with ``id`` has square-planar stereochemistry.\n"
		""},
	 { "OBStereoFacade_GetSquarePlanarStereo", _wrap_OBStereoFacade_GetSquarePlanarStereo, METH_VARARGS, "\n"
		"Get the OBSquarePlanarStereo object with ``atomId`` as center. This\n"
		"function returns 0 if there is no OBSquarePlanarStereo object found\n"
		"with the specified center.\n"
		""},
	 { "delete_OBStereoFacade", _wrap_delete_OBStereoFacade, METH_O, NULL},
	 { "OBStereoFacade_swigregister", OBStereoFacade_swigregister, METH_O, NULL},
	 { "OBStereoFacade_swiginit", OBStereoFacade_swiginit, METH_VARARGS, NULL},
	 { "PerceiveStereo", _wrap_PerceiveStereo, METH_VARARGS, "\n"
		"Convert 0D/2D/3D coordinates to OBStereo objects. The right function will\n"
		"be selected based on the molecule's dimensionality\n"
		"(i.e. OBMol::GetDimension()).\n"
		"\n"
		"See also: StereoFrom3D StereoFrom2D StereoFrom0D\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom2D", _wrap_StereoFrom2D, METH_VARARGS, "\n"
		"Convert the 2D depiction of molecule ``mol`` to OBStereo objects.\n"
		"This function makes use of the lower level functions\n"
		"TetrahedralFrom2D(), CisTransFrom2D(), SquarePlanarFrom2D(), ...\n"
		"\n"
		"First, symmetry analysis taking stereochemistry into account is\n"
		"performed iteratively (see OBGraphSym). Next the 2D coordinates,\n"
		"OBBond::Wedge, OBBond::Hash, OBBond::WedgeOrHash and OBBond::CisOrTrans\n"
		"are used to construct OBStereoBase derived objects to store the\n"
		"stereochemistry. These objects will be added to ``mol``.\n"
		"\n"
		"Unless perception is forced, this function does nothing if stereochemistry\n"
		"has already been perceived (i.e. OBMol::HasChiralityPerceived()). Before\n"
		"doing the actual perception, any data of the OBGenericDataType::StereoData\n"
		"type will be deleted.\n"
		"\n"
		"          Reference:\n"
		"     [1] T. Cieplak, J.L. Wisniewski, A New Effective Algorithm for the\n"
		"     Unambiguous Identification of the Stereochemical Characteristics of\n"
		"     Compounds During Their Registration in Databases. Molecules 2000, 6,\n"
		"     915-926, http://www.mdpi.org/molecules/papers/61100915/61100915.htm\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule containing 2D coordinates.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >, optional\n"
		":param updown: A map of OBStereo::BondDirection for cis/trans bonds\n"
		":type force: boolean, optional\n"
		":param force: Force to run the perception even if the results are cached.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom0D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom3D", _wrap_StereoFrom3D, METH_VARARGS, "\n"
		"Convert the 3D coordinates of molecule ``mol`` to OBStereo objects. This\n"
		"function makes use of the lower level functions TetrahedralFrom3D(),\n"
		"CisTransFrom3D(), SquarePlanarFrom3D(), ...\n"
		"\n"
		"Unless perception is forced, this function does nothing if stereochemistry\n"
		"has already been perceived (i.e. OBMol::HasChiralityPerceived()). Before\n"
		"doing the actual perception, any data of the OBGenericDataType::StereoData\n"
		"type will be deleted.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule containing 3D coordinates.\n"
		":type force: boolean, optional\n"
		":param force: Force to run the perception even if the results are cached.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom0D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "StereoFrom0D", _wrap_StereoFrom0D, METH_O, "\n"
		"Add missing OBStereo objects. Unlike StereoFrom3D() and StereoFrom2D(), this\n"
		"method only adds objects for previously unidentified objects since we\n"
		"don't want to loose any information. The Config::specified flag for the\n"
		"newly added structs is always set to false.\n"
		"\n"
		"For example, a smiles is read which has two tetrahedral centers. Only one has\n"
		"stereochemisrty specified using a '@' character. StereoFrom0D() will detect the\n"
		"second tetrahedral atom and add an OBTetrahedralStereo object to the molecule.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		"\n"
		"See also: StereoFrom3D StereoFrom2D PerceiveStereo\n"
		"Since: version 2.3\n"
		""},
	 { "TetrahedralFrom3D", _wrap_TetrahedralFrom3D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom3D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 3D coordinates to OBTetrahedralStereo object\n"
		"uses the sign of the volume described by the 4 center atom neighbors. Given\n"
		"4 points :math:`, \\f$b`, :math:`and \\f$d`, the signed volume _v:math:`is defined as:     \\f[ S_v = \\left| \\begin{array}{ccc}     x_b - x_a & y_b - y_a & z_b - z_a \\\\     x_c - x_a & y_c - y_a & z_c - z_a \\\\     x_d - x_a & y_d - y_a & z_d - z_a     \\end{array} \\right| The sign of _v` changes when any of the points cross the plane defined\n"
		"by the other 3 points. To make this less abstract one could say that\n"
		"a change of sign is equal to inverting the tetrahedral stereochemistry.\n"
		"\n"
		"In case there are only 3 neighbor atoms for the tetrahedral center, the\n"
		"center atom itself is used as 4th point. This only changes the magnitude\n"
		"and not the sign of \\f$S_v:math:`because the center atom is still on the sameside of the plane.This function is also used for symmetry analysis to handle cases wherethere are two atoms in the same symmetry class that don't have the samestereochemistry. In this situation, the addToMol parameter is set tofalse and the returned objects will need to be deleted explicitly.mol The molecule.stereoUnits The stereogenic units.addToMol If true, the OBTetrahedralStereo objects will be addedto the molecule using OBBase::SetData().StereoFrom3D FindStereogenicUnitsversion 2.3`\n"
		""},
	 { "TetrahedralFrom2D", _wrap_TetrahedralFrom2D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom2D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 2D coordinates and bond properties\n"
		"(i.e. OBBond::Wedge, OBBond::Hash, OBBond::WedgeOrHash and OBBond::CisOrTrans)\n"
		"uses the sign of a triangle. Given 3 points :math:`, \\f$b` and :math:`, thesign of the trianle \\f$S_t` is defined as:\n"
		"\n"
		"     .. math::\n"
		"\n"
		"         S_t = (x_a - x_c) (y_b - y_c) - (y_a - y_c) (x_b - x_c)\n"
		"\n"
		"This is equation 6 from on the referenced web page. The 3 points used\n"
		"to calculate the triangle sign always remain in the same plane (i.e. z = 0).\n"
		"The actual meaning of _t:math:`(i.e. assignment of OBStereo::Winding) dependson the 4th atom. When the atom is in front of the plane, the sign should bechanged to have the same absolute meaning for an atom behind the plane and thesame triangle. It is important to note that none of the z coordinates is everchanged, the molecule always stays 2D (unlike methods which set a pseudo-zcoordinate).@todo document bond property interpretation!This function is also used for symmetry analysis to handle cases wherethere are two atoms in the same symmetry class that don't have the samestereochemistry. In this situation, the @p addToMol parameter is set tofalse and the returned objects will need to be deleted explicitly.     @verbatim     Reference:     [1] T. Cieplak, J.L. Wisniewski, A New Effective Algorithm for the     Unambiguous Identification of the Stereochemical Characteristics of     Compounds During Their Registration in Databases. Molecules 2000, 6,     915-926, http://www.mdpi.org/molecules/papers/61100915/61100915.htm     mol The molecule.stereoUnits The stereogenic units.addToMol If true, the OBTetrahedralStereo objects will be addedto the molecule using OBBase::SetData().StereoFrom2D FindStereogenicUnitsversion 2.3`\n"
		""},
	 { "TetrahedralFrom0D", _wrap_TetrahedralFrom0D, METH_VARARGS, "\n"
		"Get a vector with all OBTetrahedralStereo objects for the molecule. This\n"
		"function is used by StereoFrom0D() with the ``addToMol`` parameter is set\n"
		"to true. There is no algorithm used here, all specified flags will be\n"
		"set to false.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBTetrahedralStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom0D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "CisTransFrom3D", _wrap_CisTransFrom3D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom3D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"The algorithm to convert the 3D coordinates to OBCisTransStereo objects\n"
		"considers the signed distance between the attached atoms and the plane\n"
		"through the double bond at right angles to the plane of the attached\n"
		"atoms. Bonds on the same side (cis) will share the same sign for the\n"
		"signed distance.\n"
		"\n"
		"Missing atom coordinates (OBStereo::ImplicitRef) and their bond\n"
		"vectors will be computed if needed.\n"
		"\n"
		"            0      3     Get signed distance of 0 and 2 to the plane\n"
		"          \\    /      that goes through the double bond and is at\n"
		"           C==C       right angles to the stereo bonds.\n"
		"   \\\n"
		"         1      2     If the two signed distances have the same sign\n"
		"                      then they are cis; if not, then trans.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom3D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "CisTransFrom2D", _wrap_CisTransFrom2D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom2D() with the ``addToMol`` parameter is set\n"
		"to true.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		"The algorithm for converting the 2D coordinates uses the same triangle\n"
		"sign as TetrahedralFrom2D(). Depending on sign of 2 triangles, the right\n"
		"OBStereo::Shape is selected.\n"
		"         0      3\n"
		"       \\    /        2 triangles: 0-1-b & 2-3-a\n"
		"        a==b    -->  same sign: U\n"
		"   \\        opposite sign: Z\n"
		"      1      2\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >, optional\n"
		":param updown: A map of OBStereo::BondDirection for cis/trans bonds\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom2D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "TetStereoToWedgeHash", _wrap_TetStereoToWedgeHash, METH_VARARGS, "\n"
		"Convert a molecule's OBTetrahedralStereo objects to a series of hash or\n"
		"wedge bonds. Note that the molecule itself is not modified; the result\n"
		"is returned in the maps ``updown`` and ``from``, which indicate\n"
		"the origin and direction of each hash or wedge bond.\n"
		"\n"
		"When converting, the following guidelines are followed when trying to\n"
		"find the best candidate bond to set up/down for each OBTetrahedralStereo\n"
		"object:\n"
		"-# Should not already be set\n"
		"-# Should not be connected to a 2nd tet center\n"
		"   (this is acceptable in theory as the wedge is only at one end, but\n"
		"    in practice it may cause confusion and thus we avoid it)\n"
		"-# Preferably is not in a cycle\n"
		"-# Preferably is a terminal H\n"
		"\n"
		"If no bond can be found that matches rules 1 and 2 (and in theory this is possible)\n"
		"then an error message is logged and the function returns false. (If you find an\n"
		"example where this occurs, please file a bug.)\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type updown: std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > >\n"
		":param updown: A map of OBStereo::BondDirection for each hash/wedge bond\n"
		":type from: std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > >\n"
		":param from: A map of OBStereo::Ref indicating the origin of each hash/wedge bond\n"
		":rtype: boolean\n"
		":return: True or False depending on whether the conversion was successful\n"
		"Since: version 2.3\n"
		""},
	 { "GetUnspecifiedCisTrans", _wrap_GetUnspecifiedCisTrans, METH_O, "\n"
		"Return a set of double bonds corresponding to the OBCisTransStereo objects\n"
		"for which the stereochemistry is undefined.\n"
		"\n"
		"Note that this functions just iterates over the existing OBCisTransStereo\n"
		"objects - it does not try to identify new ones.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":rtype: std::set< OpenBabel::OBBond * >\n"
		":return: A set of bonds with unspecified cis/trans stereochemistry\n"
		"Since: version 2.3\n"
		""},
	 { "StereoRefToImplicit", _wrap_StereoRefToImplicit, METH_VARARGS, "\n"
		"Convert any reference to ``atomId`` in a stereo object to an OBStereo::ImplicitRef.\n"
		"This function is called from OBMol::DeleteHydrogens()\n"
		"(via OBMol::DeleteHydrogen()) to remove any explicit references to a\n"
		"hydrogen atom that has been deleted. However, the code is not specific\n"
		"to hydrogen atoms and could be used for other atoms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":type atomId: int\n"
		":param atomId: The Id of the atom to be converted to an OBStereo::ImplicitRef\n"
		"Since: version 2.3\n"
		""},
	 { "ImplicitRefToStereo", _wrap_ImplicitRefToStereo, METH_VARARGS, "\n"
		"Convert any reference to an OBStereo::ImplicitRef attached to ``centerId``\n"
		"in a stereo object to an explicit reference to ``newId``.\n"
		"This function is called from OBMol::AddHydrogens() and\n"
		"OBMol::AddHydrogen() to convert any implicit references to a\n"
		"hydrogen atom that has just been added. However, the code is not specific\n"
		"to hydrogen atoms and could be used for other atoms.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule\n"
		":type centerId: int\n"
		":param centerId: The Id of the atom to which the new explicit atom is attached\n"
		":type newId: int\n"
		":param newId: The Id of the atom which was previously an OBStereo::ImplicitRef\n"
		"Since: version 2.4\n"
		""},
	 { "CisTransFrom0D", _wrap_CisTransFrom0D, METH_VARARGS, "\n"
		"Get a vector with all OBCisTransStereo objects for the molecule. This\n"
		"function is used by StereoFrom0D() with the ``addToMol`` parameter is set\n"
		"to true. There is no algorithm used here, all specified flags will be\n"
		"set to false.\n"
		"\n"
		"This function is also used for symmetry analysis to handle cases where\n"
		"there are two atoms in the same symmetry class that don't have the same\n"
		"stereochemistry. In this situation, the ``addToMol`` parameter is set to\n"
		"false and the returned objects will need to be deleted explicitly.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type stereoUnits: OpenBabel::OBStereoUnitSet\n"
		":param stereoUnits: The stereogenic units.\n"
		":type addToMol: boolean, optional\n"
		":param addToMol: If true, the OBCisTransStereo objects will be added\n"
		"    to the molecule using OBBase::SetData().\n"
		"\n"
		"See also: StereoFrom0D FindStereogenicUnits\n"
		"Since: version 2.3\n"
		""},
	 { "FindStereogenicUnits", _wrap_FindStereogenicUnits, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"\n"
		"Find the stereogenic units in a molecule using a set of rules. 1\n"
		"\n"
		"The potential stereocenters are identified first. A potential tetrahedral\n"
		"stereogenic atom is any atom meeting the following criteria:\n"
		"\n"
		"- sp3 hybridization\n"
		"- at least 3 \"heavy\" neighbors\n"
		"\n"
		"Nitrogen is treated as a special case since the barrier of inversion is\n"
		"low in many cases making the atom non-stereogenic. Only bridge-head\n"
		"nitrogen atoms (i.e. nitrogen has 3 neighbors in rings) will be\n"
		"considered stereogenic.\n"
		"\n"
		"Potential stereogenic double bonds are identified using another set of\n"
		"simple criteria:\n"
		"\n"
		"- must be a double bond\n"
		"- must not be in a ring\n"
		"- both begin and end atom should have at least one single bond\n"
		"\n"
		"True stereocenters (i.e. stereocenters with topologically different\n"
		"ligands) are identified first. For tetrahedral stereocenters, true\n"
		"stereocenters will have 4 different neighbor atom symmetry classes\n"
		"and this can be expressed using T1234 to classify these stereocenters.\n"
		"For stereogenic bonds, a similar classification C12 can be used but\n"
		"both begin and end atom have their own classification and the bond\n"
		"is only a true stereocenter if both atoms are C12.\n"
		"\n"
		"Para stereocenters are all stereocenters where there are at least two\n"
		"equivalent neighbor atom symmetry classes. These are T1123, T1112, T1111\n"
		"and T1122 for tetrahedral stereocenters and C11 for double bonds. To\n"
		"determine which of the remaining potential stereocenters really are\n"
		"stereocenters, a set of rules is used. 1\n"
		"\n"
		"Rule 1 is applied recusively:\n"
		"\n"
		"All rings are merged \"mergedRings\". A merged ring is simply a fragment consisting\n"
		"of all atoms of a ring system (bridged, spiro, adjacent, ...). If two rings in the\n"
		"SSSR set share an atom, they are merged.\n"
		"\n"
		"Each merged must at least have two para-stereocenters (or 1 true + 1 para) in order\n"
		"for the para-stereocenter to be valid. This is repeated until no new stereocenters\n"
		"are identified.\n"
		"\n"
		"rule 1a for double bonds:\n"
		"- bond atom in ring has two identical symmetry classes for it's neighbor atoms (-> para)\n"
		"- other bond atom:\n"
		"  - has two different symmetry classes for it's neighbours -> new stereocenter\n"
		"  - has two identical symmetry classes, but the ligand contains at least 1 true or para stereocenter -> new stereocenter\n"
		"\n"
		"rule 1b for tetracoord atoms:\n"
		"- at least two neighbour symmetry classes are the same (-> para)\n"
		"- other pair:\n"
		"  - has two different symmetry classes for it's neighbours -> new stereocenter\n"
		"  - has two identical symmetry classes, but the ligand contains at least 1 true or para stereocenter -> new stereocenter\n"
		"\n"
		"Rules 2 and 3 are applied sequential (i.e. only once).\n"
		"\n"
		"Rule 2a for tetracoordinate carbon:\n"
		"- 1 or 2 pair identical ligands\n"
		"- each ligand contains at least 1 true-stereocenter or 2 para-stereocenters (from rule 1)\n"
		"\n"
		"Rule 2b for tetracoordinate carbon:\n"
		"- 3 or 4 identical ligands with at least\n"
		"  - 2 true-stereocenters\n"
		"  - 2 separate assemblies of para-stereocenters (from rule 1)\n"
		"\n"
		"Rule 3 for double bonds:\n"
		"- 1 or 2 pair identical ligands (on begin and end atom)\n"
		"- each pair contains at least 1 true-stereocenter or 2 para-stereocenters (from rule 1)\n"
		"\n"
		"\n"
		"     Reference:\n"
		"     [1] M. Razinger, K. Balasubramanian, M. Perdih, M. E. Munk, Stereoisomer\n"
		"     Generation in Computer-Enhanced Structure Elucidation, J. Chem. Inf.\n"
		"     Comput. Sci. 1993, 33, 812-825\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Find the stereogenic units in a molecule making use of the automorphisms.\n"
		"\n"
		"The potential stereocenters are identified first. A potential tetrahedral\n"
		"stereogenic atom is any atom meeting the following criteria:\n"
		"\n"
		"- sp3 hybridization\n"
		"- at least 3 \"heavy\" neighbors\n"
		"\n"
		"Nitrogen is treated as a special case since the barrier of inversion is\n"
		"low in many cases making the atom non-stereogenic. Only bridge-head\n"
		"nitrogen atoms (i.e. nitrogen has 3 neighbors in rings) will be\n"
		"considered stereogenic.\n"
		"\n"
		"Potential stereogenic double bonds are identified using another set of\n"
		"simple criteria:\n"
		"\n"
		"- must be a double bond\n"
		"- must not be in a ring\n"
		"- both begin and end atom should have at least one single bond\n"
		"\n"
		"Once the potential stereocenters are found, the automorphisms are the key\n"
		"to identifying real stereogenic units. Automorphisms can be seen as\n"
		"permutations that permutate a graph back to the same graph. Such a\n"
		"permutation can only exchange atoms with the same symmetry class and it\n"
		"follows that the use of automorphisms takes symmetry into account. The\n"
		"definitions below use a concept where the automorphisms cause inversions\n"
		"of configuration to potential stereocenters. Such an inversion occurs\n"
		"whenever an automorphism exchanges two equivalent (i.e. with the same\n"
		"symmetry class) neighbor atoms attached to the potential stereogenic unit.\n"
		"\n"
		"Title: Definition for tetrahedral stereocenters:\n"
		"A potential stereocenter really is a stereocenter if there exists no automorphic\n"
		"permutation causing an inversion of the configuration of only the potential\n"
		"stereogenic unit under consideration.\n"
		"If there exists at least one automorphic permutation causing an inversion of\n"
		"the configuration, then the potential stereogenic center can be a stereogenic\n"
		"center if the number of topologically equivalent neighbors (ligands) of the\n"
		"potential stereogenic center is less than or equal to the number of different\n"
		"configurations of these ligands. 1\n"
		"\n"
		"The actual number of configurations needed for the ligands depends on the\n"
		"classification (i.e. T1234, T1123, ...) of the stereo center. These classes\n"
		"reflect the symmetry classes of the neighbor atoms of the center.\n"
		"\n"
		"- T1123: 1 true stereocenter OR 2 para stereocenters\n"
		"- T1122: 1 true stereocenter OR 2 para stereocenters (for both)\n"
		"- T1112: 2 true stereocenters OR 2 para stereocenter assemblies\n"
		"- T1111: 2 true stereocenters OR 2 para stereocenter assemblies\n"
		"\n"
		"Title: Definition for double bond stereocenters:\n"
		"A potential stereogenic double bond really is a stereogenic bond if there\n"
		"exists no automorphic permutation causing an inversion of the configuration\n"
		"of only the potential stereogenic unit under consideration. The bond can still\n"
		"be a stereogenic bond if there exists such an automorphism when the number of\n"
		"configurations of the pair of topologically equivalent geminal ligands, which\n"
		"are exchanged by the automorphism, is greater than or equal to two (i.e. the\n"
		"number of topologically equivalent geminal ligands. 1\n"
		"\n"
		"For stereogenic bonds, there is only one case but both begin and end atom\n"
		"have to be checked.\n"
		"\n"
		"- C11: 1 true stereocenter OR 1 para stereocenter\n"
		"\n"
		"These criteria are analogous to the rules from the Razinger paper on\n"
		"stereoisomer generation. Since the existence of stereocenters can depend\n"
		"on the existence of other stereocenters (in the ligands), the stereocenters\n"
		"are found by iterating until no new stereocenters are found.\n"
		"\n"
		"     Reference:\n"
		"     [1] M. Perdih, M. Razinger, Stereochemistry and Sequence Rules:\n"
		"     A Proposal for Modification of Cahn-Ingold-Prelog System,\n"
		"     Tetrahedron: Asymmetry, 1994, Vol. 5, No. 5, 835-861\n"
		""},
	 { "new_OBRotorRule", _wrap_new_OBRotorRule, METH_VARARGS, NULL},
	 { "delete_OBRotorRule", _wrap_delete_OBRotorRule, METH_O, NULL},
	 { "OBRotorRule_IsValid", _wrap_OBRotorRule_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this rotor rule is valid (i.e., is the SMARTS pattern valid)\n"
		""},
	 { "OBRotorRule_GetReferenceAtoms", _wrap_OBRotorRule_GetReferenceAtoms, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: a copy of the reference atom indexes inside the SMARTS pattern  These should be freed after use.\n"
		""},
	 { "OBRotorRule_SetDelta", _wrap_OBRotorRule_SetDelta, METH_VARARGS, "Set the resolution (delta) of a torsional step in degrees"},
	 { "OBRotorRule_GetDelta", _wrap_OBRotorRule_GetDelta, METH_O, "\n"
		":rtype: float\n"
		":return: the resolution (delta) of a torsional step in degrees\n"
		""},
	 { "OBRotorRule_GetTorsionVals", _wrap_OBRotorRule_GetTorsionVals, METH_O, "\n"
		":rtype: std::vector< double,std::allocator< double > >\n"
		":return: a reference to the dihedral angles to evaluate (in radians)\n"
		""},
	 { "OBRotorRule_GetSmartsString", _wrap_OBRotorRule_GetSmartsString, METH_O, "\n"
		":rtype: string\n"
		":return: the text of the SMARTS pattern for this rule\n"
		""},
	 { "OBRotorRule_GetSmartsPattern", _wrap_OBRotorRule_GetSmartsPattern, METH_O, "\n"
		":rtype: :py:class:`OBSmartsPattern`\n"
		":return: the exact OBSmartsPattern object for this rule\n"
		""},
	 { "OBRotorRule_swigregister", OBRotorRule_swigregister, METH_O, NULL},
	 { "OBRotorRule_swiginit", OBRotorRule_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorRules", _wrap_new_OBRotorRules, METH_NOARGS, NULL},
	 { "delete_OBRotorRules", _wrap_delete_OBRotorRules, METH_O, NULL},
	 { "OBRotorRules_SetFilename", _wrap_OBRotorRules_SetFilename, METH_VARARGS, "Set the filename to be used for the database. Default = torlib.txt"},
	 { "OBRotorRules_GetRotorIncrements", _wrap_OBRotorRules_GetRotorIncrements, METH_VARARGS, "\n"
		" Determine the torsional angles to evaluate based on the database :type mol: :py:class:`OBMol`\n"
		":param mol: molecule to evaluate :type bond: :py:class:`OBBond`\n"
		":param bond: rotatable bond to evaluate :type refs: int\n"
		":param refs: set to be the atom indexes (in mol) of the dihedral angle :type vals: std::vector< double,std::allocator< double > >\n"
		":param vals: set to be the list of angles to evaluate (in radians) :type delta: float\n"
		":param delta: potential dihedral angle steps (in degrees)\n"
		""},
	 { "OBRotorRules_Quiet", _wrap_OBRotorRules_Quiet, METH_O, "Turn off debugging output from GetRotorIncrements()"},
	 { "OBRotorRules_swigregister", OBRotorRules_swigregister, METH_O, NULL},
	 { "OBRotorRules_swiginit", OBRotorRules_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotor", _wrap_new_OBRotor, METH_NOARGS, "Constructor."},
	 { "delete_OBRotor", _wrap_delete_OBRotor, METH_O, "Destructor."},
	 { "OBRotor_SetBond", _wrap_OBRotor_SetBond, METH_VARARGS, "Set the OBBond associated with this OBRotor."},
	 { "OBRotor_SetRings", _wrap_OBRotor_SetRings, METH_O, "\n"
		"Set the rings associated with this bond (if it's a ring bond)\n"
		"Since: Version 2.4\n"
		""},
	 { "OBRotor_SetIdx", _wrap_OBRotor_SetIdx, METH_VARARGS, "Set the index for this rotor. Used by OBRotorList"},
	 { "OBRotor_SetDihedralAtoms", _wrap_OBRotor_SetDihedralAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Set the dihedral atoms.\n"
		":type ref: std::vector< int,std::allocator< int > >\n"
		":param ref: The dihedral atom indexes. These indexes start from 1.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Set the dihedral atoms.\n"
		":type ref: int\n"
		":param ref: The dihedral atom indexes. These indexes start from 1.\n"
		""},
	 { "OBRotor_SetRotAtoms", _wrap_OBRotor_SetRotAtoms, METH_VARARGS, "\n"
		"Set the atom indexes that will be displaced when this rotor\n"
		"changes torsion angle. These indexes start from 0 and are multiplied\n"
		"by 3 for easy coordinate access.\n"
		""},
	 { "OBRotor_SetTorsionValues", _wrap_OBRotor_SetTorsionValues, METH_VARARGS, "Set the possible torsion values or angles."},
	 { "OBRotor_SetFixedBonds", _wrap_OBRotor_SetFixedBonds, METH_VARARGS, "Set the bonds that will be fixed."},
	 { "OBRotor_SetToAngle", _wrap_OBRotor_SetToAngle, METH_VARARGS, "\n"
		"Rotate the atoms in the specified ``coordinates`` to the specified angle.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to rotate.\n"
		":type setang: float\n"
		":param setang: The new torsion angle in radians.\n"
		""},
	 { "OBRotor_SetRotor", _wrap_OBRotor_SetRotor, METH_VARARGS, "\n"
		"Rotate the atoms in the specified ``coordinates``. This function does not\n"
		"require any precomputation and will compute all needed information when\n"
		"needed.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates for the molecules as pointer to double.\n"
		":type next: int\n"
		":param next: The index of the new rotor angle. This is an index for the\n"
		"    GetTorsionValues() list.\n"
		":type prev: int, optional\n"
		":param prev: If specified, the torsion current torsion angle can be\n"
		"    looked up and does not have to be calculated again.\n"
		""},
	 { "OBRotor_Precompute", _wrap_OBRotor_Precompute, METH_VARARGS, "\n"
		"Precompute the reference angle and inverse bond length of this rotor for\n"
		"a single conformer. This function should be used in combination with\n"
		"Set(double *coordinates, int idx).\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to use in the computation.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OBMol mol;\n"
		"    ...\n"
		"\n"
		"    unsigned int numCoords = mol.NumAtoms() * 3;\n"
		"    double *coords = mol.GetCoordinates();\n"
		"    OBRotor rotor;\n"
		"    rotor.SetBond(mol.GetBond(3));\n"
		"\n"
		"    // set the possible torsion values\n"
		"    std::vector<double> angles;\n"
		"    angles.push_back(0.0);\n"
		"    angles.push_back(3.1415);\n"
		"    rotor.SetTorsionValues(angles);\n"
		"\n"
		"    // precompute inverse bond length (i.e. the bond length of bond with index 3\n"
		"    // using the specified coordinates) and reference angle (i.e. torsion angle\n"
		"    //in coords)\n"
		"    rotor.Precompute(coords);\n"
		"\n"
		"    // copy coordinates to coords_1\n"
		"    double *coords_1 = new double[numCoords];\n"
		"    for (unsigned int i = 0; i < numCoords; ++i)\n"
		"      coords_1[i] = coords[i];\n"
		"    // rotate the atoms in coords_1 to angle with index 0 (i.e. 0.0 degrees)\n"
		"    // note: on input, the coordinates should be the same as the coordinates used\n"
		"    //       to precompute the inverse bond length and reference angle (in other\n"
		"    //       words, the inverse magnitude and reference angle in the specfied\n"
		"    //       coordinates should be the same as the one used for Precompute)\n"
		"    rotor.Set(coords_1, 0)\n"
		"\n"
		"    // copy coordinates to coords_2\n"
		"    double *coords_2 = new double[numCoords];\n"
		"    for (unsigned int i = 0; i < numCoords; ++i)\n"
		"      coords_2[i] = coords[i];\n"
		"    // rotate the atoms in coords_2 to angle with index 1 (i.e. 180.0 degrees)\n"
		"    rotor.Set(coords_2, 1)\n"
		"\n"
		"    delete coords_1;\n"
		"    delete coords_2;\n"
		""},
	 { "OBRotor_Precalc", _wrap_OBRotor_Precalc, METH_VARARGS, "\n"
		"Precompute the inverse bond lengths, rotation matrices for all\n"
		"specified conformers and all possible torsion values. This method is\n"
		"used in combination with Set(double *coordinates, int conformer, int idx).\n"
		":type conformers: std::vector< double *,std::allocator< double * > >\n"
		":param conformers: The pointers to the conformer coordinates\n"
		""},
	 { "OBRotor_Set", _wrap_OBRotor_Set, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Rotate the specified ``coordinates`` by using the specified rotation matrix.\n"
		"\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Rotate the ``coordinates`` to set the torsion angle of this rotor to the angle\n"
		"specified by the index ``idx``. Make sure to call Precompute before calling\n"
		"this function.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates to rotate.\n"
		":type idx: int\n"
		":param idx: The index of the torsion angle in the GetTorsionValues() list.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Rotate the ``coordinates`` to set the torsion to the torsion value with the\n"
		"specified ``index``. The coordinates should be the same as the conformer used\n"
		"for calling Precalc (i.e. conformers[conformer] == coordinates). Make sure\n"
		"to call Precalc before calling this method.\n"
		":type coordinates: float\n"
		":param coordinates: The conformer coordinates.\n"
		":type conformer: int\n"
		":param conformer: The conformer index in the conformer list given to Precalc().\n"
		":type idx: int\n"
		":param idx: The torsion value index in the GetTorsionValues() list.\n"
		""},
	 { "OBRotor_GetBond", _wrap_OBRotor_GetBond, METH_O, "Get the OBBond object associated with this OBRotor."},
	 { "OBRotor_Size", _wrap_OBRotor_Size, METH_O, "\n"
		"Get the number of possible torsion angles for this OBRotor. This\n"
		"is the length of the GetTorsionValues() list.\n"
		""},
	 { "OBRotor_GetIdx", _wrap_OBRotor_GetIdx, METH_O, "Get the index for this rotor (index in an OBRotorList)."},
	 { "OBRotor_GetDihedralAtoms", _wrap_OBRotor_GetDihedralAtoms, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"Get the dihedral atom indexes. These indexes start from 1.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Get the dihedral atom indexes. These indexes start from 1.\n"
		""},
	 { "OBRotor_GetTorsionValues", _wrap_OBRotor_GetTorsionValues, METH_O, "Get the possible torsion angles for this OBRotor."},
	 { "OBRotor_GetFixedBonds", _wrap_OBRotor_GetFixedBonds, METH_O, "\n"
		"Get an OBBitVec objects with bits set for all bonds that are fixed.\n"
		"Bonds are indexed from 0.\n"
		""},
	 { "OBRotor_CalcTorsion", _wrap_OBRotor_CalcTorsion, METH_VARARGS, "\n"
		"Calculate the torsion for this OBRotor using the specified coordinates.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates (e.g. OBMol::GetCoordinates()).\n"
		":rtype: float\n"
		":return: The torsion angle in radians.\n"
		""},
	 { "OBRotor_CalcBondLength", _wrap_OBRotor_CalcBondLength, METH_VARARGS, "\n"
		"Calculate the bond length for this OBRotor using the specified coordinates.\n"
		":type coordinates: float\n"
		":param coordinates: The coordinates (e.g. OBMol::GetCoordinates()).\n"
		""},
	 { "OBRotor_BeginTorIncrement", _wrap_OBRotor_BeginTorIncrement, METH_O, ""},
	 { "OBRotor_EndTorIncrement", _wrap_OBRotor_EndTorIncrement, METH_O, NULL},
	 { "OBRotor_RemoveSymTorsionValues", _wrap_OBRotor_RemoveSymTorsionValues, METH_VARARGS, ""},
	 { "OBRotor_SetDelta", _wrap_OBRotor_SetDelta, METH_VARARGS, "Deprecated: Has no effect."},
	 { "OBRotor_GetDelta", _wrap_OBRotor_GetDelta, METH_O, "Deprecated: Has no effect."},
	 { "OBRotor_GetFixedAtoms", _wrap_OBRotor_GetFixedAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_SetFixedAtoms", _wrap_OBRotor_SetFixedAtoms, METH_VARARGS, "Deprecated: See SetFixedBonds"},
	 { "OBRotor_GetEvalAtoms", _wrap_OBRotor_GetEvalAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_SetEvalAtoms", _wrap_OBRotor_SetEvalAtoms, METH_VARARGS, "Deprecated:"},
	 { "OBRotor_GetRotAtoms", _wrap_OBRotor_GetRotAtoms, METH_O, "Deprecated:"},
	 { "OBRotor_GetResolution", _wrap_OBRotor_GetResolution, METH_O, "Deprecated: Bad name, see GetTorsionValues()"},
	 { "OBRotor_SetNumCoords", _wrap_OBRotor_SetNumCoords, METH_VARARGS, "Deprecated:"},
	 { "OBRotor_swigregister", OBRotor_swigregister, METH_O, NULL},
	 { "OBRotor_swiginit", OBRotor_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorList", _wrap_new_OBRotorList, METH_NOARGS, "Constructor."},
	 { "delete_OBRotorList", _wrap_delete_OBRotorList, METH_O, "Destructor."},
	 { "OBRotorList_Clear", _wrap_OBRotorList_Clear, METH_O, "Clear the internal list of rotors and reset."},
	 { "OBRotorList_Size", _wrap_OBRotorList_Size, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotors in this list\n"
		""},
	 { "OBRotorList_IsFixedBond", _wrap_OBRotorList_IsFixedBond, METH_VARARGS, "\n"
		"When no atoms/bonds are fixed or when bonds are fixed, this function will\n"
		"return true if the bond is fixed. When using the deprecated fixed atoms,\n"
		"this function will return true if the bond and at least one neighboring\n"
		"bond has fixed atoms.\n"
		""},
	 { "OBRotorList_HasFixedBonds", _wrap_OBRotorList_HasFixedBonds, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this rotor list has any fixed bonds.\n"
		""},
	 { "OBRotorList_RemoveSymVals", _wrap_OBRotorList_RemoveSymVals, METH_VARARGS, "Rotates each bond to zero and 180 degrees and tests if the 2 conformers are duplicates.  if so - the symmetric torsion values are removed from consideration during a search"},
	 { "OBRotorList_HasRingRotors", _wrap_OBRotorList_HasRingRotors, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this rotor list has any ring bonds.\n"
		"Since: version 2.4\n"
		""},
	 { "OBRotorList_Setup", _wrap_OBRotorList_Setup, METH_VARARGS, "\n"
		"Setup this rotor list for the supplied molecule. This method calls\n"
		"FindRotors(), SetEvalAtoms(), and AssignTorVals().\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type sampleRings: boolean, optional\n"
		":param sampleRings: Whether to sample ring conformers - default = false\n"
		":rtype: boolean\n"
		":return: True if rotatable bonds were found.\n"
		""},
	 { "OBRotorList_SetFixedBonds", _wrap_OBRotorList_SetFixedBonds, METH_VARARGS, "Set the bonds that will be fixed."},
	 { "OBRotorList_Init", _wrap_OBRotorList_Init, METH_VARARGS, "Initialize the private OBRotorRules database from a specific file."},
	 { "OBRotorList_SetQuiet", _wrap_OBRotorList_SetQuiet, METH_O, "Turn off debugging output."},
	 { "OBRotorList_SetRotAtoms", _wrap_OBRotorList_SetRotAtoms, METH_VARARGS, "\n"
		" Set the atoms to rotate from the dihedral atoms for each rotor Uses OBRotor->GetDihedralAtoms() to call OBRotor->SetRotAtoms() and standarizes the dihedral angles via OBRotor->SetDihedralAtoms() :rtype: boolean\n"
		":return: True\n"
		""},
	 { "OBRotorList_FindRotors", _wrap_OBRotorList_FindRotors, METH_VARARGS, "\n"
		"Find all potentially rotatable bonds in the molecule. This method uses\n"
		"OBBond::IsRotor() for initial evaluation which depends on ring perception\n"
		"(i.e. ring bonds are considered rotatable). Fixed bonds, specified using the\n"
		"deprecated fixed atoms or the new fixed bonds methods are not added to\n"
		"the list. All rotatable bonds will be sorted by their graph theoretical\n"
		"distance (GTD) score (see OBMol::GetGTDVector()). This results in the\n"
		"the rotors going from the inside to the outside of the mol.\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule.\n"
		":type sampleRingBonds: boolean, optional\n"
		":param sampleRingBonds: whether to sample ring bonds from analysis (default = false)\n"
		":rtype: boolean\n"
		":return: True.\n"
		""},
	 { "OBRotorList_SetEvalAtoms", _wrap_OBRotorList_SetEvalAtoms, METH_VARARGS, "\n"
		" Determines which atoms should be used to calculate the internal energy if the dihedral angle of the rotor is modified :rtype: boolean\n"
		":return: True\n"
		""},
	 { "OBRotorList_AssignTorVals", _wrap_OBRotorList_AssignTorVals, METH_VARARGS, "\n"
		"Using the OBRotorRules database, set the torsion values (and delta)\n"
		"to be evaluated and tested. This method also sets the rotatable atoms\n"
		"for the rotors to the smallest possible set. For each bond there are\n"
		"two candidate sets, one on either side. The smallest of these is used\n"
		"and the torsion indexes for the rotor are inverted if needed\n"
		"(i.e. a-b-c-d -> d-c-b-a).\n"
		""},
	 { "OBRotorList_BeginRotor", _wrap_OBRotorList_BeginRotor, METH_VARARGS, "\n"
		"Initialize the ``i`` iterator and get a pointer to the first OBRotor.\n"
		":type i: OpenBabel::OBRotorIterator\n"
		":param i: OBRotorIterator object.\n"
		""},
	 { "OBRotorList_NextRotor", _wrap_OBRotorList_NextRotor, METH_VARARGS, "\n"
		"Get a pointer to the next iterator.\n"
		":type i: OpenBabel::OBRotorIterator\n"
		":param i: OBRotorIterator object.\n"
		""},
	 { "OBRotorList_BeginRotors", _wrap_OBRotorList_BeginRotors, METH_O, "Get the rotor list begin iterator."},
	 { "OBRotorList_EndRotors", _wrap_OBRotorList_EndRotors, METH_O, "Get the rotor list end iterator."},
	 { "OBRotorList_IdentifyEvalAtoms", _wrap_OBRotorList_IdentifyEvalAtoms, METH_VARARGS, "Deprecated: Not declared. Use Setup() for top-level functionality"},
	 { "OBRotorList_SetFixAtoms", _wrap_OBRotorList_SetFixAtoms, METH_VARARGS, "\n"
		"Set the list of fixed (invariant) atoms to the supplied OBBitVec\n"
		"Deprecated: See SetFixedBonds()\n"
		""},
	 { "OBRotorList_HasFixedAtoms", _wrap_OBRotorList_HasFixedAtoms, METH_O, "\n"
		":rtype: boolean\n"
		":return: whether this rotor list has any fixed (invariant) atoms\n"
		"Deprecated: See HasFixedBonds()\n"
		""},
	 { "OBRotorList_IgnoreSymmetryRemoval", _wrap_OBRotorList_IgnoreSymmetryRemoval, METH_O, "Has no effect Deprecated: Currently has no effect"},
	 { "OBRotorList_SetRotAtomsByFix", _wrap_OBRotorList_SetRotAtomsByFix, METH_VARARGS, "Set the atoms to rotate from the dihedral atoms for each rotor Insures the fixed atoms are respected, but otherwise functions like SetRotAtoms()"},
	 { "OBRotorList_swigregister", OBRotorList_swigregister, METH_O, NULL},
	 { "OBRotorList_swiginit", OBRotorList_swiginit, METH_VARARGS, NULL},
	 { "new_rotor_digit", _wrap_new_rotor_digit, METH_VARARGS, NULL},
	 { "rotor_digit_set_size", _wrap_rotor_digit_set_size, METH_VARARGS, NULL},
	 { "rotor_digit_set_state", _wrap_rotor_digit_set_state, METH_VARARGS, NULL},
	 { "rotor_digit_get_state", _wrap_rotor_digit_get_state, METH_O, NULL},
	 { "rotor_digit_size", _wrap_rotor_digit_size, METH_O, NULL},
	 { "rotor_digit_next", _wrap_rotor_digit_next, METH_O, NULL},
	 { "delete_rotor_digit", _wrap_delete_rotor_digit, METH_O, NULL},
	 { "rotor_digit_swigregister", rotor_digit_swigregister, METH_O, NULL},
	 { "rotor_digit_swiginit", rotor_digit_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotorKeys", _wrap_new_OBRotorKeys, METH_NOARGS, "Constructor"},
	 { "OBRotorKeys_Clear", _wrap_OBRotorKeys_Clear, METH_O, "Clear all rotors"},
	 { "OBRotorKeys_NumKeys", _wrap_OBRotorKeys_NumKeys, METH_O, "Number of rotor keys (= number of possible conformers)"},
	 { "OBRotorKeys_AddRotor", _wrap_OBRotorKeys_AddRotor, METH_VARARGS, "\n"
		" Add a rotor :type size: int\n"
		":param size: the rotor resolution\n"
		""},
	 { "OBRotorKeys_Next", _wrap_OBRotorKeys_Next, METH_O, "\n"
		" Select the next rotor key :rtype: boolean\n"
		":return: true if there are more rotor keys\n"
		""},
	 { "OBRotorKeys_GetKey", _wrap_OBRotorKeys_GetKey, METH_O, "\n"
		" Get the currently selected rotor key :rtype: std::vector< int,std::allocator< int > >\n"
		":return: current rotor key\n"
		""},
	 { "delete_OBRotorKeys", _wrap_delete_OBRotorKeys, METH_O, NULL},
	 { "OBRotorKeys_swigregister", OBRotorKeys_swigregister, METH_O, NULL},
	 { "OBRotorKeys_swiginit", OBRotorKeys_swiginit, METH_VARARGS, NULL},
	 { "new_OBRotamerList", _wrap_new_OBRotamerList, METH_NOARGS, NULL},
	 { "delete_OBRotamerList", _wrap_delete_OBRotamerList, METH_O, NULL},
	 { "OBRotamerList_Setup", _wrap_OBRotamerList_Setup, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Set up a rotamer list based on an already created OBRotorList\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Set up a rotamer list based on the supplied reference atoms and the number of rotors :type mol: :py:class:`OBMol`\n"
		":param mol: The molecule to evaluate :type ref: unsigned char\n"
		":param ref: An array of the 4 dihedral atoms for each rotor :type nrotors: int\n"
		":param nrotors: The number of rotors (i.e., the size of ref / 4)\n"
		""},
	 { "OBRotamerList_NumRotors", _wrap_OBRotamerList_NumRotors, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotatable bonds considered\n"
		""},
	 { "OBRotamerList_NumRotamers", _wrap_OBRotamerList_NumRotamers, METH_O, "\n"
		":rtype: int\n"
		":return: the number of rotamer (conformation) coordinate sets\n"
		""},
	 { "OBRotamerList_AddRotamer", _wrap_OBRotamerList_AddRotamer, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Add a rotamer to the list based on the supplied coordinate set as a double*\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"Add a rotamer to the list based on ``key`` as a configuration of the individual rotor bonds\n"
		""},
	 { "OBRotamerList_AddRotamers", _wrap_OBRotamerList_AddRotamers, METH_VARARGS, "Add ``nconf`` rotamers based on ``as`` an array of configurations much like AddRotamer()"},
	 { "OBRotamerList_GetReferenceArray", _wrap_OBRotamerList_GetReferenceArray, METH_VARARGS, "\n"
		":rtype: void\n"
		":return: A reference array (as used by AddRotamer() as a configuration of the individual rotor bonds\n"
		""},
	 { "OBRotamerList_BeginRotamer", _wrap_OBRotamerList_BeginRotamer, METH_O, NULL},
	 { "OBRotamerList_EndRotamer", _wrap_OBRotamerList_EndRotamer, METH_O, NULL},
	 { "OBRotamerList_CreateConformerList", _wrap_OBRotamerList_CreateConformerList, METH_VARARGS, "\n"
		" Create a conformer list using the internal base set of coordinates :rtype: std::vector< double *,std::allocator< double * > >\n"
		":return: The set of coordinates by rotating the bonds in each rotamer\n"
		""},
	 { "OBRotamerList_ExpandConformerList", _wrap_OBRotamerList_ExpandConformerList, METH_VARARGS, "\n"
		" Create a conformer list using the internal base set of coordinates :rtype: void\n"
		":return: The set of coordinates as a reference in ``confs``\n"
		""},
	 { "OBRotamerList_SetCurrentCoordinates", _wrap_OBRotamerList_SetCurrentCoordinates, METH_VARARGS, NULL},
	 { "OBRotamerList_SetBaseCoordinateSets", _wrap_OBRotamerList_SetBaseCoordinateSets, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		" Copies the mol's conformers (the coordinates, NOT the pointers) into the object as base coordinates\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		" Copies the coordinates in bc, NOT the pointers, into this object :type bc: std::vector< double *,std::allocator< double * > >\n"
		":param bc: The conformer set for the molecule\n"
		"        :type N: int\n"
		"        :param N:  The number of atoms in the molecule\n"
		""},
	 { "OBRotamerList_NumBaseCoordinateSets", _wrap_OBRotamerList_NumBaseCoordinateSets, METH_O, "\n"
		":rtype: int\n"
		":return: The number of \"base\" coordinate sets (i.e., the number of conformers in the base OBMol)\n"
		""},
	 { "OBRotamerList_GetBaseCoordinateSet", _wrap_OBRotamerList_GetBaseCoordinateSet, METH_VARARGS, "Get a pointer to a specific base pointer (i.e., specific conformer)"},
	 { "OBRotamerList_NumAtoms", _wrap_OBRotamerList_NumAtoms, METH_O, "\n"
		":rtype: int\n"
		":return: The number of atoms in the base OBMol\n"
		""},
	 { "OBRotamerList_swigregister", OBRotamerList_swigregister, METH_O, NULL},
	 { "OBRotamerList_swiginit", OBRotamerList_swiginit, METH_VARARGS, NULL},
	 { "new_OBSpectrophore", _wrap_new_OBSpectrophore, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Default constructor to create an OBSpectrophore object.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Copy constructor to create an OBSpectrophore object by taking\n"
		"        a copy from another OBSpectrophore object.\n"
		"\n"
		"        :type sphore: :py:class:`OBSpectrophore`\n"
		"        :param sphore: A reference to the source OBSpectrophore object\n"
		""},
	 { "delete_OBSpectrophore", _wrap_delete_OBSpectrophore, METH_O, "Default destructor object"},
	 { "OBSpectrophore_SetResolution", _wrap_OBSpectrophore_SetResolution, METH_VARARGS, "\n"
		"Method to set the resolution at which Spectrophores will be\n"
		"        calculated.\n"
		"\n"
		"        Sets the resolution at which Spectrophores are calculated. The\n"
		"        resolution is an arbitrary positive number larger than 0.0, and is used\n"
		"        to increase the box size for the calculation of Spectrophores in\n"
		"        all directions. For example, a resolution of 3 means that the box size\n"
		"        will be increased by a value of 3 &Aring; in all directions. Smaller\n"
		"        values for the resolution have the effect that small differences in the\n"
		"        atomic properties will result in more enhanced differences in the\n"
		"        resulting Spectrophores, while larger resolution values will result\n"
		"        in the smoothing of local property differences. All values <= 0.0\n"
		"        are automatically reset to the default value of 3.0.\n"
		"\n"
		"        :type r: float, optional\n"
		"        :param r: The desired resolution expressed as a double number\n"
		""},
	 { "OBSpectrophore_SetAccuracy", _wrap_OBSpectrophore_SetAccuracy, METH_VARARGS, "\n"
		"Method to set the accuracy at which Spectrophores will be\n"
		"        calculated.\n"
		"\n"
		"        Sets the accuracy by which Spectrophores are calculated. The\n"
		"        accuracy is linked to the angular step increment that is used to\n"
		"        calculate Spectrophores. The accuracy parameter is an enumeration\n"
		"        type defined by OBSpectrophore::AccuracyOption. should be a number ranging\n"
		"        between 0 and 9 (inclusive), where a value of 9 means a very high\n"
		"        accuracy, and 0 means no accuracy at all. Values between 2-6 are a good\n"
		"        compromise between accuracy and speed. The following table provides the\n"
		"        link between the parameter value and the angular step size:\n"
		"\n"
		"        |**Accuracy**|**OBSpectrophore::AccuracyOption parameter**|**Minimal angular steps size**|\n"
		"        |Extremely high|AngStepSize1|1º|\n"
		"        |Extremely high|AngStepSize2|2º|\n"
		"        |Very High|AngStepSize5|5º|\n"
		"        |Very high|AngStepSize10|10º|\n"
		"        |High|AngStepSize15|15º|\n"
		"        |Default|AngStepSize20|20º|\n"
		"        |Low|AngStepSize30|30º|\n"
		"        |Very low|AngStepSize36|36º|\n"
		"        |Extremely low|AngStepSize45|45º|\n"
		"        |Extremely low|AngStepSize60|60º|\n"
		"\n"
		"\n"
		"        :type a: int, optional\n"
		"        :param a: The desired accuracy expressed as an\n"
		"            OBSpectrophore::AccuracyOption enumeration type\n"
		""},
	 { "OBSpectrophore_SetStereo", _wrap_OBSpectrophore_SetStereo, METH_VARARGS, "\n"
		"Method to set the required stereoselectivity of the resulting\n"
		"        Spectrophores.\n"
		"\n"
		"        Sets the stereoselectivity of the generated Spectrophores. This is\n"
		"        achieved by selecting the appropriate probes for the calculation. In\n"
		"        the default case that no stereoselectivity is required, only\n"
		"        non-stereospecific probes are used in the calculation. Stereospecific\n"
		"        differences between enantiomers are only captured in combination with a\n"
		"        very high accuracy, hence leading to very long computation times. The\n"
		"        following table provides the link between the parameter value and the\n"
		"        desired stereochemical treatment:\n"
		"\n"
		"        |**Stereo treatment**|**OBSpectrophore::StereoOption parameter**|\n"
		"        |Non-stereospecific probes (default)|NoStereoSpecificProbes|\n"
		"        |Only the mirror stereospecific probes|UniqueStereoSpecificProbes|\n"
		"        |Only the unique stereospecific probes|MirrorStereoSpecificProbes|\n"
		"        |All stereospecific probes (unique and mirror)|AllStereoSpecificProbes|\n"
		"\n"
		"\n"
		"        :type s: int, optional\n"
		"        :param s: The desired stereospecificity treatment expressed as an\n"
		"            OBSpectrophore::StereoOption enumeration type\n"
		""},
	 { "OBSpectrophore_SetNormalization", _wrap_OBSpectrophore_SetNormalization, METH_VARARGS, "\n"
		"Method to set the desired normalization treatment of the calculated\n"
		"        Spectrophores.\n"
		"\n"
		"        The following table provides the link between the parameter value\n"
		"        and the desired normalization treatment:\n"
		"\n"
		"        |**Normalization treatment**|**OBSpectrophore::NormalizationOption parameter**|\n"
		"        |No normalisation (default)|NoNormalization|\n"
		"        |Normalization towards zero mean|NormalizationTowardsZeroMean|\n"
		"        |Normalization towards unit standard deviation|NormalizationTowardsUnitStd|\n"
		"        |Normalization towards zero mean and unit standard deviation|NormalizationTowardsZeroMeanAndUnitStd|\n"
		"\n"
		"\n"
		"\n"
		"        :type n: int, optional\n"
		"        :param n: The desired normalization treatment expressed as an\n"
		"            OBSpectrophore::NormalizationOption enumeration type\n"
		""},
	 { "OBSpectrophore_GetAccuracy", _wrap_OBSpectrophore_GetAccuracy, METH_O, "\n"
		"Returns the accuracy at which Spectrophores will be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The accuracy expressed as an OBSpectrophore::AccuracyOption\n"
		"                     enumeration type. For a link between the returned value and\n"
		"                     the angular step size, see the\n"
		"                     OBSpectrophore::SetAccuracy(const OBSpectrophore::AccuracyOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetResolution", _wrap_OBSpectrophore_GetResolution, METH_O, "\n"
		"Returns the resolution at which Spectrophores will be calculated.\n"
		"\n"
		"        :rtype: float\n"
		"        :return: The resolution in &Aring; units\n"
		""},
	 { "OBSpectrophore_GetStereo", _wrap_OBSpectrophore_GetStereo, METH_O, "\n"
		"Returns the stereoselectivity setting at which Spectrophores\n"
		"        will be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The level of stereospecificity expressed as an\n"
		"                     OBSpectrophore::StereoOption enumeration type. For a link\n"
		"                     between the returned value and the stereoselectivity\n"
		"                     treatment, see the\n"
		"                     OBSpectrophore::SetStereo(const OBSpectrophore::StereoOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetNormalization", _wrap_OBSpectrophore_GetNormalization, METH_O, "\n"
		"Returns the normalization settings at which Spectrophores will\n"
		"        be calculated.\n"
		"\n"
		"        :rtype: int\n"
		"        :return: The normalization treatmet expressed as an\n"
		"                     OBSpectrophore::NormalizationOption enumeration type.\n"
		"                     For a link between the returned value and the normalization\n"
		"                     treatment, see the\n"
		"                     OBSpectrophore::SetNormalization(const OBSpectrophore::NormalizationOption)\n"
		"                     method.\n"
		""},
	 { "OBSpectrophore_GetSpectrophore", _wrap_OBSpectrophore_GetSpectrophore, METH_VARARGS, "\n"
		" Calling this method starts the calculation of the Spectrophore.\n"
		"         After successful calculation, the Spectrophore is returned as a\n"
		"         standard vector of 48 doubles. The 48 doubles are organised into 4 sets\n"
		"         of 12 doubles each:-\n"
		"         - numbers 01-11: Spectrophore values calculated from the atomic partial charges;\n"
		"         - numbers 13-24: Spectrophore values calculated from the atomic lipophilicity properties;\n"
		"         - numbers 25-36: Spectrophore values calculated from the atomic shape deviations;\n"
		"         - numbers 37-48: Spectrophore values calculated from the atomic electrophilicity properties;\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    OpenBabel::OBConversion obconversion;\n"
		"    obconversion.SetInFormat(\"sdf\");\n"
		"    OpenBabel::OBMol mol;\n"
		"    OpenBabel::OBSpectrophore s;\n"
		"    s.SetAccuracy(OpenBabel::OBSpectrophore::AngStepSize20);          // Default\n"
		"    s.SetResolution(3.0);                                             // Default\n"
		"    s.SetStereo(OpenBabel::OBSpectrophore::NoStereoSpecificProbes);   // Default\n"
		"    s.SetNormalization(OpenBabel::OBSpectrophore::NoNormalization);   // Default\n"
		"    std::ifstream ifs;\n"
		"    ifs.open(argv[1]);\n"
		"    while (obconversion.Read(&mol, &ifs))\n"
		"    {\n"
		"       std::vector<double> result = s.GetSpectrophore(&mol);\n"
		"       for (unsigned int i(0); i < result.size(); ++i)\n"
		"       {\n"
		"          if (!(i%12)) std::cerr << std::endl;\n"
		"          std::cerr << result[i] << \"  \";\n"
		"       }\n"
		"       std::cerr << std::endl;\n"
		"       mol.Clear();\n"
		"    }\n"
		"\n"
		"         :type mol: :py:class:`OBMol`\n"
		"         :param mol: Pointer to the OBMol object from which to calculate a\n"
		"                      Spectrophore. For proper functioning, the input molecule\n"
		"                      should have all explicit hydrogens assigned and the molecule\n"
		"                      should have a three-dimensional conformation assigned. It is\n"
		"                      the responsability of the programmer to make sure that the\n"
		"                      molecule is in the desired protonation state.\n"
		"\n"
		"         :rtype: std::vector< double,std::allocator< double > >\n"
		"         :return: The calculated Spectrophore as a standard vector of 48\n"
		"                      doubles. An empty vector of doubles is returned in case an\n"
		"                      error occurred during the calculation. For example, this could\n"
		"                      happen in the case that the molecule contains less than three\n"
		"                      atoms. It is therefore up to the user to check the size of the\n"
		"                      returned vector to capture errors.\n"
		""},
	 { "OBSpectrophore_swigregister", OBSpectrophore_swigregister, METH_O, NULL},
	 { "OBSpectrophore_swiginit", OBSpectrophore_swiginit, METH_VARARGS, NULL},
	 { "OBConformerFilter_IsGood", _wrap_OBConformerFilter_IsGood, METH_VARARGS, "\n"
		"Check if a conformer is good (i.e. passes the filter). The molecule, rotor\n"
		"key and coords are passed as parameters to this function and can be used\n"
		"for checking the confoormer.\n"
		":rtype: boolean\n"
		":return: True if the conformer passes the filter.\n"
		""},
	 { "delete_OBConformerFilter", _wrap_delete_OBConformerFilter, METH_O, NULL},
	 { "OBConformerFilter_swigregister", OBConformerFilter_swigregister, METH_O, NULL},
	 { "new_OBConformerFilters", _wrap_new_OBConformerFilters, METH_O, "Constructor specifiying the filters to combine."},
	 { "delete_OBConformerFilters", _wrap_delete_OBConformerFilters, METH_O, NULL},
	 { "OBConformerFilters_swigregister", OBConformerFilters_swigregister, METH_O, NULL},
	 { "OBConformerFilters_swiginit", OBConformerFilters_swiginit, METH_VARARGS, NULL},
	 { "new_OBStericConformerFilter", _wrap_new_OBStericConformerFilter, METH_VARARGS, NULL},
	 { "delete_OBStericConformerFilter", _wrap_delete_OBStericConformerFilter, METH_O, NULL},
	 { "OBStericConformerFilter_swigregister", OBStericConformerFilter_swigregister, METH_O, NULL},
	 { "OBStericConformerFilter_swiginit", OBStericConformerFilter_swiginit, METH_VARARGS, NULL},
	 { "OBConformerScore_GetPreferred", _wrap_OBConformerScore_GetPreferred, METH_O, "Preferred order for subclass scoring function."},
	 { "OBConformerScore_GetConvergence", _wrap_OBConformerScore_GetConvergence, METH_O, "Convergence criteria for subclass scoring function."},
	 { "OBConformerScore_Score", _wrap_OBConformerScore_Score, METH_VARARGS, "Score an individual conformer specified by index."},
	 { "delete_OBConformerScore", _wrap_delete_OBConformerScore, METH_O, NULL},
	 { "OBConformerScore_swigregister", OBConformerScore_swigregister, METH_O, NULL},
	 { "new_OBRMSDConformerScore", _wrap_new_OBRMSDConformerScore, METH_NOARGS, NULL},
	 { "delete_OBRMSDConformerScore", _wrap_delete_OBRMSDConformerScore, METH_O, NULL},
	 { "OBRMSDConformerScore_swigregister", OBRMSDConformerScore_swigregister, METH_O, NULL},
	 { "OBRMSDConformerScore_swiginit", OBRMSDConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBEnergyConformerScore", _wrap_new_OBEnergyConformerScore, METH_NOARGS, NULL},
	 { "OBEnergyConformerScore_GetNbEnergyCompute", _wrap_OBEnergyConformerScore_GetNbEnergyCompute, METH_O, NULL},
	 { "OBEnergyConformerScore_GetNbEnergyRequest", _wrap_OBEnergyConformerScore_GetNbEnergyRequest, METH_O, NULL},
	 { "delete_OBEnergyConformerScore", _wrap_delete_OBEnergyConformerScore, METH_O, NULL},
	 { "OBEnergyConformerScore_swigregister", OBEnergyConformerScore_swigregister, METH_O, NULL},
	 { "OBEnergyConformerScore_swiginit", OBEnergyConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBMinimizingEnergyConformerScore", _wrap_new_OBMinimizingEnergyConformerScore, METH_NOARGS, NULL},
	 { "OBMinimizingEnergyConformerScore_GetNbEnergyCompute", _wrap_OBMinimizingEnergyConformerScore_GetNbEnergyCompute, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_GetNbEnergyRequest", _wrap_OBMinimizingEnergyConformerScore_GetNbEnergyRequest, METH_O, NULL},
	 { "delete_OBMinimizingEnergyConformerScore", _wrap_delete_OBMinimizingEnergyConformerScore, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_swigregister", OBMinimizingEnergyConformerScore_swigregister, METH_O, NULL},
	 { "OBMinimizingEnergyConformerScore_swiginit", OBMinimizingEnergyConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBMinimizingRMSDConformerScore", _wrap_new_OBMinimizingRMSDConformerScore, METH_NOARGS, NULL},
	 { "delete_OBMinimizingRMSDConformerScore", _wrap_delete_OBMinimizingRMSDConformerScore, METH_O, NULL},
	 { "OBMinimizingRMSDConformerScore_swigregister", OBMinimizingRMSDConformerScore_swigregister, METH_O, NULL},
	 { "OBMinimizingRMSDConformerScore_swiginit", OBMinimizingRMSDConformerScore_swiginit, METH_VARARGS, NULL},
	 { "new_OBConformerSearch", _wrap_new_OBConformerSearch, METH_NOARGS, NULL},
	 { "delete_OBConformerSearch", _wrap_delete_OBConformerSearch, METH_O, NULL},
	 { "OBConformerSearch_Setup", _wrap_OBConformerSearch_Setup, METH_VARARGS, "\n"
		"Setup this instance with the specified molecule.\n"
		"\n"
		":type mol: :py:class:`OBMol`\n"
		":param mol: The molecule with initial conformer.\n"
		":type numConformers: int, optional\n"
		":param numConformers: The number of conformers that should be generated. This\n"
		"           is also the number of conformers selected for each generation.\n"
		":type numChildren: int, optional\n"
		":param numChildren: When a new generation is generated, for each of the\n"
		"           numConformers conformers, numChildren children are created.\n"
		":type mutability: int, optional\n"
		":param mutability: The mutability determines how frequent a permutation occurs\n"
		"           when generating the next generation.\n"
		":type convergence: int, optional\n"
		":param convergence: The number of identical generations before considering\n"
		"           the process converged.\n"
		""},
	 { "OBConformerSearch_SetNumConformers", _wrap_OBConformerSearch_SetNumConformers, METH_VARARGS, "Set the number of conformers."},
	 { "OBConformerSearch_SetNumChildren", _wrap_OBConformerSearch_SetNumChildren, METH_VARARGS, "Set the number of children generated for each parent."},
	 { "OBConformerSearch_SetMutability", _wrap_OBConformerSearch_SetMutability, METH_VARARGS, "Set the mutability."},
	 { "OBConformerSearch_SetConvergence", _wrap_OBConformerSearch_SetConvergence, METH_VARARGS, "\n"
		"Set the convergence (i.e. number of generations that did not change the score\n"
		"before considering the iteration converged).\n"
		""},
	 { "OBConformerSearch_SetFixedBonds", _wrap_OBConformerSearch_SetFixedBonds, METH_VARARGS, "Set the bonds to be fixed."},
	 { "OBConformerSearch_SetFilter", _wrap_OBConformerSearch_SetFilter, METH_VARARGS, "\n"
		"Set the filter method used to check if a newly generated is acceptable. Typical\n"
		"examples are a steric filter or electrostatic energy filter. The filters make a binary\n"
		"selection and one group will be scored in the next step. A number of filters\n"
		"are provided by default but implementing a custom filter is easy.\n"
		"\n"
		"There is also a OBConformerFilters class to make it easy to use multiple filters.\n"
		""},
	 { "OBConformerSearch_SetScore", _wrap_OBConformerSearch_SetScore, METH_VARARGS, "\n"
		"All acceptable conformers are scored to select the fittest conformers from\n"
		"the population. A higher score means the conformer is desired in the final\n"
		"set or next generation. Typical scoring would be force field energy to find\n"
		"the lowest energy conformer. Another example is the highest RMSD between\n"
		"conformer i and the closest conformer to create a diverse set of conformers.\n"
		"Diversity could also be expressed in other ways and implementing a scoring\n"
		"class is easy.\n"
		""},
	 { "OBConformerSearch_PrintRotors", _wrap_OBConformerSearch_PrintRotors, METH_VARARGS, "Set whether or not you want rotors to be printed prior to the conformer search."},
	 { "OBConformerSearch_Search", _wrap_OBConformerSearch_Search, METH_O, "Perform conformer search using a genetic algorithm."},
	 { "OBConformerSearch_GetRotorKeys", _wrap_OBConformerSearch_GetRotorKeys, METH_O, NULL},
	 { "OBConformerSearch_GetConformers", _wrap_OBConformerSearch_GetConformers, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetLogStream", _wrap_OBConformerSearch_SetLogStream, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetSharing", _wrap_OBConformerSearch_SetSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNbNiches", _wrap_OBConformerSearch_GetNbNiches, METH_O, NULL},
	 { "OBConformerSearch_SetNbNiches", _wrap_OBConformerSearch_SetNbNiches, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNicheRadius", _wrap_OBConformerSearch_GetNicheRadius, METH_O, NULL},
	 { "OBConformerSearch_SetNicheRadius", _wrap_OBConformerSearch_SetNicheRadius, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetAlphaSharing", _wrap_OBConformerSearch_GetAlphaSharing, METH_O, NULL},
	 { "OBConformerSearch_SetAlphaSharing", _wrap_OBConformerSearch_SetAlphaSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetSigmaSharing", _wrap_OBConformerSearch_GetSigmaSharing, METH_O, NULL},
	 { "OBConformerSearch_SetSigmaSharing", _wrap_OBConformerSearch_SetSigmaSharing, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetCrossoverProbability", _wrap_OBConformerSearch_GetCrossoverProbability, METH_O, NULL},
	 { "OBConformerSearch_SetCrossoverProbability", _wrap_OBConformerSearch_SetCrossoverProbability, METH_VARARGS, NULL},
	 { "OBConformerSearch_GetNicheMating", _wrap_OBConformerSearch_GetNicheMating, METH_O, NULL},
	 { "OBConformerSearch_SetNicheMating", _wrap_OBConformerSearch_SetNicheMating, METH_VARARGS, NULL},
	 { "OBConformerSearch_SetLocalOptRate", _wrap_OBConformerSearch_SetLocalOptRate, METH_VARARGS, NULL},
	 { "OBConformerSearch_swigregister", OBConformerSearch_swigregister, METH_O, NULL},
	 { "OBConformerSearch_swiginit", OBConformerSearch_swiginit, METH_VARARGS, NULL},
	 { "new_OBAlign", _wrap_new_OBAlign, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"\n"
		"If this constructor is used, the Target and Reference must be\n"
		"set using SetRef/SetRefMol and SetTarget/SetTargetMol before running\n"
		"the alignment.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 3:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 4:*\n"
		"\n"
		"Align two molecules.\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 5:*\n"
		"\n"
		"Align two vectors of vector3 objects.\n"
		""},
	 { "OBAlign_SetRef", _wrap_OBAlign_SetRef, METH_VARARGS, "\n"
		"Set the Reference (to which the Target will be aligned) in\n"
		"terms of a vector of vector3 objects. Note that it is faster\n"
		"to perform multiple alignments to the same Reference, rather than\n"
		"multiple alignments to the same Target.\n"
		""},
	 { "OBAlign_SetTarget", _wrap_OBAlign_SetTarget, METH_VARARGS, "\n"
		"Set the Target (which will be aligned to the Reference) in\n"
		"terms of a vector of vector3 objects.\n"
		""},
	 { "OBAlign_SetRefMol", _wrap_OBAlign_SetRefMol, METH_VARARGS, "\n"
		"Set the Reference Molecule (to which the Target Molecule must\n"
		"be aligned). Note that is faster to perform multiple alignments\n"
		"to the same Reference Molecule, rather than multple alignments\n"
		"to the same Target Molecule.\n"
		""},
	 { "OBAlign_SetTargetMol", _wrap_OBAlign_SetTargetMol, METH_VARARGS, "\n"
		"Set the Target Molecule (which will be aligned to the\n"
		"Reference Molecule).\n"
		""},
	 { "OBAlign_Align", _wrap_OBAlign_Align, METH_O, "Align the Target to the Reference using a least-squares alignment."},
	 { "OBAlign_SetMethod", _wrap_OBAlign_SetMethod, METH_VARARGS, NULL},
	 { "OBAlign_GetRMSD", _wrap_OBAlign_GetRMSD, METH_O, "\n"
		"Return the root mean squared deviation of the target from\n"
		"the reference. This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "OBAlign_GetRotMatrix", _wrap_OBAlign_GetRotMatrix, METH_O, "\n"
		"Return the rotation matrix associated with the least squares\n"
		"alignment. This function should only\n"
		"be called after running the alignment (using Align()).\n"
		"\n"
		"The following example shows how to use the rotation matrix\n"
		"to rotate all of the atoms in a molecule.\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    matrix3x3 rotmatrix = align.GetRotMatrix();\n"
		"    for (unsigned int i = 1; i <= mol.NumAtoms(); ++i) {\n"
		"       vector3 tmpvec = mol.GetAtom(i)->GetVector();\n"
		"       tmpvec *= rotmatrix; //apply the rotation\n"
		"       mol.GetAtom(i)->SetVector(tmpvec);\n"
		"    }\n"
		"Note that if you wish to use the rotation matrix to find the\n"
		"aligned coordinates (that is, the same coordinates returned by\n"
		"GetAlignment()), you should first translate the set of coordinates\n"
		"to the origin by subtracting the centroid, then apply the rotation,\n"
		"and finally add the centroid of the reference coordinates.\n"
		""},
	 { "OBAlign_GetAlignment", _wrap_OBAlign_GetAlignment, METH_O, "\n"
		"Return the actual alignment of the Target to the Reference\n"
		"in terms of a vector of vector3 objects. If you want an OBMol\n"
		"with the aligned coordinates, you should use UpdateCoords() instead.\n"
		"This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "OBAlign_UpdateCoords", _wrap_OBAlign_UpdateCoords, METH_VARARGS, "\n"
		"Set the coordinates of an OBMol to those from the alignment.\n"
		"This function should only\n"
		"be called after running the alignment (using Align()).\n"
		""},
	 { "delete_OBAlign", _wrap_delete_OBAlign, METH_O, NULL},
	 { "OBAlign_swigregister", OBAlign_swigregister, METH_O, NULL},
	 { "OBAlign_swiginit", OBAlign_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomIter", _wrap_new__OBMolAtomIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomIter", _wrap_delete__OBMolAtomIter, METH_O, NULL},
	 { "_OBMolAtomIter___nonzero__", _wrap__OBMolAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomIter_inc", _wrap__OBMolAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement iterator -- advance to next atom and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement iterator -- return the current atom and advance\n"
		""},
	 { "_OBMolAtomIter___ref__", _wrap__OBMolAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomIter_swigregister", _OBMolAtomIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomIter_swiginit", _OBMolAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomDFSIter", _wrap_new__OBMolAtomDFSIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomDFSIter", _wrap_delete__OBMolAtomDFSIter, METH_O, NULL},
	 { "_OBMolAtomDFSIter___nonzero__", _wrap__OBMolAtomDFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomDFSIter_inc", _wrap__OBMolAtomDFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the DFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "_OBMolAtomDFSIter___ref__", _wrap__OBMolAtomDFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomDFSIter_next", _wrap__OBMolAtomDFSIter_next, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: NULL if at the last atom in a fragment, else the next atom\n"
		""},
	 { "_OBMolAtomDFSIter_swigregister", _OBMolAtomDFSIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomDFSIter_swiginit", _OBMolAtomDFSIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAtomBFSIter", _wrap_new__OBMolAtomBFSIter, METH_VARARGS, NULL},
	 { "delete__OBMolAtomBFSIter", _wrap_delete__OBMolAtomBFSIter, METH_O, NULL},
	 { "_OBMolAtomBFSIter___nonzero__", _wrap__OBMolAtomBFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "_OBMolAtomBFSIter_inc", _wrap__OBMolAtomBFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the BFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "_OBMolAtomBFSIter___ref__", _wrap__OBMolAtomBFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBMolAtomBFSIter_CurrentDepth", _wrap__OBMolAtomBFSIter_CurrentDepth, METH_O, "\n"
		":rtype: int\n"
		":return: the current depth of the iterator Since: version 2.2\n"
		""},
	 { "_OBMolAtomBFSIter_swigregister", _OBMolAtomBFSIter_swigregister, METH_O, NULL},
	 { "_OBMolAtomBFSIter_swiginit", _OBMolAtomBFSIter_swiginit, METH_VARARGS, NULL},
	 { "new_OBMolBondBFSIter", _wrap_new_OBMolBondBFSIter, METH_VARARGS, NULL},
	 { "delete_OBMolBondBFSIter", _wrap_delete_OBMolBondBFSIter, METH_O, NULL},
	 { "OBMolBondBFSIter___nonzero__", _wrap_OBMolBondBFSIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms)\n"
		""},
	 { "OBMolBondBFSIter_inc", _wrap_OBMolBondBFSIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom in the BFS list and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current atom and advance to the next atom\n"
		""},
	 { "OBMolBondBFSIter___ref__", _wrap_OBMolBondBFSIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current atom\n"
		""},
	 { "OBMolBondBFSIter_CurrentDepth", _wrap_OBMolBondBFSIter_CurrentDepth, METH_O, "\n"
		":rtype: int\n"
		":return: the current depth of the iterator Since: version 2.2\n"
		""},
	 { "OBMolBondBFSIter_swigregister", OBMolBondBFSIter_swigregister, METH_O, NULL},
	 { "OBMolBondBFSIter_swiginit", OBMolBondBFSIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolBondIter", _wrap_new__OBMolBondIter, METH_VARARGS, NULL},
	 { "delete__OBMolBondIter", _wrap_delete__OBMolBondIter, METH_O, NULL},
	 { "_OBMolBondIter___nonzero__", _wrap__OBMolBondIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more bonds)\n"
		""},
	 { "_OBMolBondIter_inc", _wrap__OBMolBondIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next bond and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current bond and advance to the next\n"
		""},
	 { "_OBMolBondIter___ref__", _wrap__OBMolBondIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current bond\n"
		""},
	 { "_OBMolBondIter_swigregister", _OBMolBondIter_swigregister, METH_O, NULL},
	 { "_OBMolBondIter_swiginit", _OBMolBondIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBAtomAtomIter", _wrap_new__OBAtomAtomIter, METH_VARARGS, NULL},
	 { "delete__OBAtomAtomIter", _wrap_delete__OBAtomAtomIter, METH_O, NULL},
	 { "_OBAtomAtomIter___nonzero__", _wrap__OBAtomAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more neighbors)\n"
		""},
	 { "_OBAtomAtomIter_inc", _wrap__OBAtomAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next neighbor and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current neighbor and advance to the next\n"
		""},
	 { "_OBAtomAtomIter___ref__", _wrap__OBAtomAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBAtomAtomIter_swigregister", _OBAtomAtomIter_swigregister, METH_O, NULL},
	 { "_OBAtomAtomIter_swiginit", _OBAtomAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBAtomBondIter", _wrap_new__OBAtomBondIter, METH_VARARGS, NULL},
	 { "delete__OBAtomBondIter", _wrap_delete__OBAtomBondIter, METH_O, NULL},
	 { "_OBAtomBondIter___nonzero__", _wrap__OBAtomBondIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more bonds)\n"
		""},
	 { "_OBAtomBondIter_inc", _wrap__OBAtomBondIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next bond and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next bond\n"
		""},
	 { "_OBAtomBondIter___ref__", _wrap__OBAtomBondIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBBond`\n"
		":return: a reference to the current bond\n"
		""},
	 { "_OBAtomBondIter_swigregister", _OBAtomBondIter_swigregister, METH_O, NULL},
	 { "_OBAtomBondIter_swiginit", _OBAtomBondIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBResidueIter", _wrap_new__OBResidueIter, METH_VARARGS, NULL},
	 { "delete__OBResidueIter", _wrap_delete__OBResidueIter, METH_O, NULL},
	 { "_OBResidueIter___nonzero__", _wrap__OBResidueIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more residues)\n"
		""},
	 { "_OBResidueIter_inc", _wrap__OBResidueIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next residue and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next residue\n"
		""},
	 { "_OBResidueIter___ref__", _wrap__OBResidueIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBResidue`\n"
		":return: a reference to the current residue\n"
		""},
	 { "_OBResidueIter_swigregister", _OBResidueIter_swigregister, METH_O, NULL},
	 { "_OBResidueIter_swiginit", _OBResidueIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBResidueAtomIter", _wrap_new__OBResidueAtomIter, METH_VARARGS, NULL},
	 { "delete__OBResidueAtomIter", _wrap_delete__OBResidueAtomIter, METH_O, NULL},
	 { "_OBResidueAtomIter___nonzero__", _wrap__OBResidueAtomIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more atoms in this residue)\n"
		""},
	 { "_OBResidueAtomIter_inc", _wrap__OBResidueAtomIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next atom (if any) and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next atom (if any)\n"
		""},
	 { "_OBResidueAtomIter___ref__", _wrap__OBResidueAtomIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBAtom`\n"
		":return: a reference to the current atom\n"
		""},
	 { "_OBResidueAtomIter_swigregister", _OBResidueAtomIter_swigregister, METH_O, NULL},
	 { "_OBResidueAtomIter_swiginit", _OBResidueAtomIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolAngleIter", _wrap_new__OBMolAngleIter, METH_VARARGS, NULL},
	 { "delete__OBMolAngleIter", _wrap_delete__OBMolAngleIter, METH_O, NULL},
	 { "_OBMolAngleIter___nonzero__", _wrap__OBMolAngleIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more angles)\n"
		""},
	 { "_OBMolAngleIter_inc", _wrap__OBMolAngleIter_inc, METH_O, "Preincrement -- advance to the next angle and return"},
	 { "_OBMolAngleIter___ref__", _wrap__OBMolAngleIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of three atom indexes specifying the angle See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolAngleIter_swigregister", _OBMolAngleIter_swigregister, METH_O, NULL},
	 { "_OBMolAngleIter_swiginit", _OBMolAngleIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolTorsionIter", _wrap_new__OBMolTorsionIter, METH_VARARGS, NULL},
	 { "delete__OBMolTorsionIter", _wrap_delete__OBMolTorsionIter, METH_O, NULL},
	 { "_OBMolTorsionIter___nonzero__", _wrap__OBMolTorsionIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more torsions)\n"
		""},
	 { "_OBMolTorsionIter_inc", _wrap__OBMolTorsionIter_inc, METH_O, "Preincrement -- advance to the next torsion and return"},
	 { "_OBMolTorsionIter___ref__", _wrap__OBMolTorsionIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of four atom indexes specifying the torsion See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolTorsionIter_swigregister", _OBMolTorsionIter_swigregister, METH_O, NULL},
	 { "_OBMolTorsionIter_swiginit", _OBMolTorsionIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolPairIter", _wrap_new__OBMolPairIter, METH_VARARGS, NULL},
	 { "delete__OBMolPairIter", _wrap_delete__OBMolPairIter, METH_O, NULL},
	 { "_OBMolPairIter___nonzero__", _wrap__OBMolPairIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can still advance (i.e., visit more 1-4 atom pairs)\n"
		""},
	 { "_OBMolPairIter_inc", _wrap__OBMolPairIter_inc, METH_O, "Preincrement -- advance to the next 1-4 atom pair and return"},
	 { "_OBMolPairIter___ref__", _wrap__OBMolPairIter___ref__, METH_O, "\n"
		":rtype: std::vector< unsigned int,std::allocator< unsigned int > >\n"
		":return: A vector of two atom indexes specifying the 1-4 atom pair See also: OBAtom::GetIdx()\n"
		""},
	 { "_OBMolPairIter_swigregister", _OBMolPairIter_swigregister, METH_O, NULL},
	 { "_OBMolPairIter_swiginit", _OBMolPairIter_swiginit, METH_VARARGS, NULL},
	 { "new__OBMolRingIter", _wrap_new__OBMolRingIter, METH_VARARGS, NULL},
	 { "delete__OBMolRingIter", _wrap_delete__OBMolRingIter, METH_O, NULL},
	 { "_OBMolRingIter___nonzero__", _wrap__OBMolRingIter___nonzero__, METH_O, "\n"
		":rtype: boolean\n"
		":return: Whether the iterator can advance (i.e., there are more rings)\n"
		""},
	 { "_OBMolRingIter_inc", _wrap__OBMolRingIter_inc, METH_VARARGS, "\n"
		"*Overload 1:*\n"
		"Preincrement -- advance to the next ring (if any) and return\n"
		"\n"
		"|\n"
		"\n"
		"*Overload 2:*\n"
		"Postincrement -- return the current state and advance to the next ring\n"
		""},
	 { "_OBMolRingIter___ref__", _wrap__OBMolRingIter___ref__, METH_O, "\n"
		":rtype: :py:class:`OBRing`\n"
		":return: A reference to the current ring (if any)\n"
		""},
	 { "_OBMolRingIter_swigregister", _OBMolRingIter_swigregister, METH_O, NULL},
	 { "_OBMolRingIter_swiginit", _OBMolRingIter_swiginit, METH_VARARGS, NULL},
	 { "MolGetAtoms", _wrap_MolGetAtoms, METH_VARARGS, NULL},
	 { "MolGetBonds", _wrap_MolGetBonds, METH_VARARGS, NULL},
	 { "AtomGetBonds", _wrap_AtomGetBonds, METH_VARARGS, NULL},
	 { "AtomGetNbrs", _wrap_AtomGetNbrs, METH_VARARGS, NULL},
	 { "new_doubleArray", _wrap_new_doubleArray, METH_O, NULL},
	 { "delete_doubleArray", _wrap_delete_doubleArray, METH_O, NULL},
	 { "doubleArray___getitem__", _wrap_doubleArray___getitem__, METH_VARARGS, NULL},
	 { "doubleArray___setitem__", _wrap_doubleArray___setitem__, METH_VARARGS, NULL},
	 { "doubleArray_cast", _wrap_doubleArray_cast, METH_O, NULL},
	 { "doubleArray_frompointer", _wrap_doubleArray_frompointer, METH_O, NULL},
	 { "doubleArray_swigregister", doubleArray_swigregister, METH_O, NULL},
	 { "doubleArray_swiginit", doubleArray_swiginit, METH_VARARGS, NULL},
	 { "delete_OBTetraNonPlanarStereo", _wrap_delete_OBTetraNonPlanarStereo, METH_O, "Destructor."},
	 { "OBTetraNonPlanarStereo_swigregister", OBTetraNonPlanarStereo_swigregister, METH_O, NULL},
	 { "delete_OBTetraPlanarStereo", _wrap_delete_OBTetraPlanarStereo, METH_O, NULL},
	 { "OBTetraPlanarStereo_swigregister", OBTetraPlanarStereo_swigregister, METH_O, NULL},
	 { "new_OBTetrahedralStereo", _wrap_new_OBTetrahedralStereo, METH_O, "Constructor."},
	 { "delete_OBTetrahedralStereo", _wrap_delete_OBTetrahedralStereo, METH_O, "Destructor."},
	 { "OBTetrahedralStereo_IsValid", _wrap_OBTetrahedralStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all (center, from\n"
		"    and ref) atom ids are set.\n"
		""},
	 { "OBTetrahedralStereo___eq__", _wrap_OBTetrahedralStereo___eq__, METH_VARARGS, "\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBTetrahedralStereo___ne__", _wrap_OBTetrahedralStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBTetrahedralStereo_SetConfig", _wrap_OBTetrahedralStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBTetrahedralStereo_GetConfig", _wrap_OBTetrahedralStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBTetrahedralStereo_swigregister", OBTetrahedralStereo_swigregister, METH_O, NULL},
	 { "OBTetrahedralStereo_swiginit", OBTetrahedralStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBCisTransStereo", _wrap_new_OBCisTransStereo, METH_O, "\n"
		"<openbabel/stereo/cistrans.h>\n"
		"Stereochemical configuration for double-bond cis/trans stereochemistry\n"
		"\n"
		"The config struct represents the stereochemistry in a well defined way.\n"
		"For cis/trans stereo bonds, the following data members define the spacial\n"
		"arrengement of the atoms.\n"
		"\n"
		"- OBStereo::Ref ``begin``: The begin atom for the double bond.\n"
		"- OBStereo::Ref ``end``: The end atom for the double bond.\n"
		"- OBStereo::Refs ``refs``: The 4 atoms connected to the double bond.\n"
		"- OBStereo::Shape ``shape``: The shape formed by the ``refs`` by connecting them\n"
		"  in the same order as they occur in ``refs``.\n"
		"\n"
		"Image: cistrans.png\n"
		"Image: SPshapes.png\n"
		"\n"
		"Only ``begin`` and ``end`` are specific for OBCisTransStereo::Config. The other\n"
		"data members occur in all OBTetraPlanarStereo derived classes.\n"
		"\n"
		"Constructor.\n"
		""},
	 { "delete_OBCisTransStereo", _wrap_delete_OBCisTransStereo, METH_O, "Destructor."},
	 { "OBCisTransStereo_IsValid", _wrap_OBCisTransStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all these\n"
		"    conditions are met:\n"
		"    - ``begin`` != OBStereo::NoRef\n"
		"    - ``end`` != OBStereo::NoRef\n"
		"    - ``refs`` contains 4 elements\n"
		""},
	 { "OBCisTransStereo___eq__", _wrap_OBCisTransStereo___eq__, METH_VARARGS, "\n"
		"Set the configuration using a Config struct.\n"
		"\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBCisTransStereo___ne__", _wrap_OBCisTransStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBCisTransStereo_IsOnSameAtom", _wrap_OBCisTransStereo_IsOnSameAtom, METH_VARARGS, "\n"
		"Check if the two atoms for ``id1`` & ``id2`` are bonded to the same\n"
		"atom. If the atoms for one of the atoms doesn't exist (anymore),\n"
		"the valence of the begin and end atom is checked. If the exising\n"
		"atom is bonded to the begin atom and end->GetExplicitDegree() == 2, the\n"
		"ids are considered to be on different atoms. The reasoning behind\n"
		"this is that hydrogens may be deleted. However, you can also use\n"
		"OBStereo::ImplicitRef explicitly in code like:\n"
		"\n"
		".. code-block:: c++\n"
		"\n"
		"    //\n"
		"    //  F       F      F      F     0      3\n"
		"    //   \\     /       |      |     |      |\n"
		"    //    C===C        | C  C |     | 1  2 |\n"
		"    //   /     \\       |      |     |      |\n"
		"    // (H)     (H)    (H)----(H)    H------H\n"
		"    //\n"
		"    reading smiles F/C=C\\F cis-difluorethene\n"
		"    OBCisTransStereo ct(mol);\n"
		"    ct.SetCenters(1, 2);\n"
		"    ct.SetRefs(OBStereo::MakeRefs(0, OBStereo::ImplicitRef, OBStereo::ImplicitRef, 3));\n"
		"    ...\n"
		"\n"
		":rtype: boolean\n"
		":return: True if ``id1`` and ``id2`` are bonded to the same atom\n"
		"    taking implicit hydrogens into account.\n"
		""},
	 { "OBCisTransStereo_IsTrans", _wrap_OBCisTransStereo_IsTrans, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed trans configuration.\n"
		""},
	 { "OBCisTransStereo_IsCis", _wrap_OBCisTransStereo_IsCis, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed in a cis configuration.\n"
		""},
	 { "OBCisTransStereo_GetTransRef", _wrap_OBCisTransStereo_GetTransRef, METH_VARARGS, "\n"
		"Image: gettransref.png\n"
		"Get the reference id trans from reference ``id``.\n"
		""},
	 { "OBCisTransStereo_GetCisRef", _wrap_OBCisTransStereo_GetCisRef, METH_VARARGS, "\n"
		"Image: getcisref.png\n"
		"Get the reference id cis from reference ``id``.\n"
		""},
	 { "OBCisTransStereo_SetConfig", _wrap_OBCisTransStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBCisTransStereo_GetConfig", _wrap_OBCisTransStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBCisTransStereo_swigregister", OBCisTransStereo_swigregister, METH_O, NULL},
	 { "OBCisTransStereo_swiginit", OBCisTransStereo_swiginit, METH_VARARGS, NULL},
	 { "new_OBSquarePlanarStereo", _wrap_new_OBSquarePlanarStereo, METH_O, "\n"
		"<openbabel/stereo/squareplanar.h>\n"
		"Stereochemical configuration for square planar stereocenters\n"
		"\n"
		"The config struct represents the stereochemistry in a well defined way.\n"
		"For squareplanar stereocenters, the following data members define the spacial\n"
		"arrengement of the atoms.\n"
		"\n"
		"- OBStereo::Ref ``center``: The central atom.\n"
		"- OBStereo::Refs ``refs``: The 4 atoms connected to the double bond.\n"
		"- OBStereo::Shape ``shape``: The shape formed by the ``refs`` by connecting them\n"
		"  in the same order as they occur in ``refs``.\n"
		"\n"
		"Image: squareplanar.png\n"
		"Image: SPshapes.png\n"
		"\n"
		"Only ``center`` are specific for OBSquarePlanarStereo::Config. The other\n"
		"data members occur in all OBTetraPlanarStereo derived classes.\n"
		"\n"
		"Constructor.\n"
		""},
	 { "delete_OBSquarePlanarStereo", _wrap_delete_OBSquarePlanarStereo, METH_O, "Destructor."},
	 { "OBSquarePlanarStereo_IsValid", _wrap_OBSquarePlanarStereo_IsValid, METH_O, "\n"
		":rtype: boolean\n"
		":return: True if this object is valid. This object is valid if all (center and\n"
		"    and 4 reference) atom ids are set.\n"
		""},
	 { "OBSquarePlanarStereo___eq__", _wrap_OBSquarePlanarStereo___eq__, METH_VARARGS, "\n"
		"Set the configuration using a Config struct.\n"
		"\n"
		"Compare the stereochemistry stored in the Config struct with the\n"
		"stereochemistry specified in the Config struct from ``other``.\n"
		"\n"
		""},
	 { "OBSquarePlanarStereo___ne__", _wrap_OBSquarePlanarStereo___ne__, METH_VARARGS, "\n"
		"Not equal to operator. This is the inverse of the Equal to operator==.\n"
		"\n"
		":rtype: boolean\n"
		":return: True if the two Config structs represent a different stereochemistry.\n"
		""},
	 { "OBSquarePlanarStereo_IsTrans", _wrap_OBSquarePlanarStereo_IsTrans, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed trans configuration.\n"
		""},
	 { "OBSquarePlanarStereo_IsCis", _wrap_OBSquarePlanarStereo_IsCis, METH_VARARGS, "\n"
		":rtype: boolean\n"
		":return: True if the two reference ids are placed in a cis configuration.\n"
		""},
	 { "OBSquarePlanarStereo_GetTransRef", _wrap_OBSquarePlanarStereo_GetTransRef, METH_VARARGS, "\n"
		"Image: gettransref.png\n"
		"Get the reference id trans from reference ``id``.\n"
		""},
	 { "OBSquarePlanarStereo_GetCisRefs", _wrap_OBSquarePlanarStereo_GetCisRefs, METH_VARARGS, "Get the reference id cis from reference ``id``."},
	 { "OBSquarePlanarStereo_SetConfig", _wrap_OBSquarePlanarStereo_SetConfig, METH_VARARGS, NULL},
	 { "OBSquarePlanarStereo_GetConfig", _wrap_OBSquarePlanarStereo_GetConfig, METH_VARARGS, NULL},
	 { "OBSquarePlanarStereo_swigregister", OBSquarePlanarStereo_swigregister, METH_O, NULL},
	 { "OBSquarePlanarStereo_swiginit", OBSquarePlanarStereo_swiginit, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig___eq__", _wrap_OBTetrahedralConfig___eq__, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig___ne__", _wrap_OBTetrahedralConfig___ne__, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_center_set", _wrap_OBTetrahedralConfig_center_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_center_get", _wrap_OBTetrahedralConfig_center_get, METH_O, NULL},
	 { "OBTetrahedralConfig_from_or_towards_set", _wrap_OBTetrahedralConfig_from_or_towards_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_from_or_towards_get", _wrap_OBTetrahedralConfig_from_or_towards_get, METH_O, NULL},
	 { "OBTetrahedralConfig_refs_set", _wrap_OBTetrahedralConfig_refs_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_refs_get", _wrap_OBTetrahedralConfig_refs_get, METH_O, NULL},
	 { "OBTetrahedralConfig_winding_set", _wrap_OBTetrahedralConfig_winding_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_winding_get", _wrap_OBTetrahedralConfig_winding_get, METH_O, NULL},
	 { "OBTetrahedralConfig_view_set", _wrap_OBTetrahedralConfig_view_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_view_get", _wrap_OBTetrahedralConfig_view_get, METH_O, NULL},
	 { "OBTetrahedralConfig_specified_set", _wrap_OBTetrahedralConfig_specified_set, METH_VARARGS, NULL},
	 { "OBTetrahedralConfig_specified_get", _wrap_OBTetrahedralConfig_specified_get, METH_O, NULL},
	 { "new_OBTetrahedralConfig", _wrap_new_OBTetrahedralConfig, METH_NOARGS, NULL},
	 { "delete_OBTetrahedralConfig", _wrap_delete_OBTetrahedralConfig, METH_O, NULL},
	 { "OBTetrahedralConfig_swigregister", OBTetrahedralConfig_swigregister, METH_O, NULL},
	 { "OBTetrahedralConfig_swiginit", OBTetrahedralConfig_swiginit, METH_VARARGS, NULL},
	 { "OBCisTransConfig___eq__", _wrap_OBCisTransConfig___eq__, METH_VARARGS, NULL},
	 { "OBCisTransConfig___ne__", _wrap_OBCisTransConfig___ne__, METH_VARARGS, NULL},
	 { "OBCisTransConfig_begin_set", _wrap_OBCisTransConfig_begin_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_begin_get", _wrap_OBCisTransConfig_begin_get, METH_O, NULL},
	 { "OBCisTransConfig_end_set", _wrap_OBCisTransConfig_end_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_end_get", _wrap_OBCisTransConfig_end_get, METH_O, NULL},
	 { "OBCisTransConfig_refs_set", _wrap_OBCisTransConfig_refs_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_refs_get", _wrap_OBCisTransConfig_refs_get, METH_O, NULL},
	 { "OBCisTransConfig_shape_set", _wrap_OBCisTransConfig_shape_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_shape_get", _wrap_OBCisTransConfig_shape_get, METH_O, NULL},
	 { "OBCisTransConfig_specified_set", _wrap_OBCisTransConfig_specified_set, METH_VARARGS, NULL},
	 { "OBCisTransConfig_specified_get", _wrap_OBCisTransConfig_specified_get, METH_O, NULL},
	 { "new_OBCisTransConfig", _wrap_new_OBCisTransConfig, METH_NOARGS, NULL},
	 { "delete_OBCisTransConfig", _wrap_delete_OBCisTransConfig, METH_O, NULL},
	 { "OBCisTransConfig_swigregister", OBCisTransConfig_swigregister, METH_O, NULL},
	 { "OBCisTransConfig_swiginit", OBCisTransConfig_swiginit, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig___eq__", _wrap_OBSquarePlanarConfig___eq__, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig___ne__", _wrap_OBSquarePlanarConfig___ne__, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_center_set", _wrap_OBSquarePlanarConfig_center_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_center_get", _wrap_OBSquarePlanarConfig_center_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_refs_set", _wrap_OBSquarePlanarConfig_refs_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_refs_get", _wrap_OBSquarePlanarConfig_refs_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_shape_set", _wrap_OBSquarePlanarConfig_shape_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_shape_get", _wrap_OBSquarePlanarConfig_shape_get, METH_O, NULL},
	 { "OBSquarePlanarConfig_specified_set", _wrap_OBSquarePlanarConfig_specified_set, METH_VARARGS, NULL},
	 { "OBSquarePlanarConfig_specified_get", _wrap_OBSquarePlanarConfig_specified_get, METH_O, NULL},
	 { "new_OBSquarePlanarConfig", _wrap_new_OBSquarePlanarConfig, METH_NOARGS, NULL},
	 { "delete_OBSquarePlanarConfig", _wrap_delete_OBSquarePlanarConfig, METH_O, NULL},
	 { "OBSquarePlanarConfig_swigregister", OBSquarePlanarConfig_swigregister, METH_O, NULL},
	 { "OBSquarePlanarConfig_swiginit", OBSquarePlanarConfig_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_a_3__doubleArrayTo_p_a_3__double(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((double (*)[3])  ((doubleArray (*)[3]) x));
}
static void *_p_doubleArrayTo_p_double(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((double *)  ((doubleArray *) x));
}
static void *_p_OpenBabel__OBAtomTo_p_OpenBabel__OBBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBBase *)  ((OpenBabel::OBAtom *) x));
}
static void *_p_OpenBabel__OBBondTo_p_OpenBabel__OBBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBBase *)  ((OpenBabel::OBBond *) x));
}
static void *_p_OpenBabel__OBMolTo_p_OpenBabel__OBBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBBase *)  ((OpenBabel::OBMol *) x));
}
static void *_p_OpenBabel__OBReactionTo_p_OpenBabel__OBBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBBase *)  ((OpenBabel::OBReaction *) x));
}
static void *_p_OpenBabel__OBResidueTo_p_OpenBabel__OBBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBBase *)  ((OpenBabel::OBResidue *) x));
}
static void *_p_OpenBabel__OBConformerFiltersTo_p_OpenBabel__OBConformerFilter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerFilter *)  ((OpenBabel::OBConformerFilters *) x));
}
static void *_p_OpenBabel__OBStericConformerFilterTo_p_OpenBabel__OBConformerFilter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerFilter *)  ((OpenBabel::OBStericConformerFilter *) x));
}
static void *_p_OpenBabel__OBEnergyConformerScoreTo_p_OpenBabel__OBConformerScore(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerScore *)  ((OpenBabel::OBEnergyConformerScore *) x));
}
static void *_p_OpenBabel__OBMinimizingEnergyConformerScoreTo_p_OpenBabel__OBConformerScore(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerScore *)  ((OpenBabel::OBMinimizingEnergyConformerScore *) x));
}
static void *_p_OpenBabel__OBMinimizingRMSDConformerScoreTo_p_OpenBabel__OBConformerScore(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerScore *)  ((OpenBabel::OBMinimizingRMSDConformerScore *) x));
}
static void *_p_OpenBabel__OBRMSDConformerScoreTo_p_OpenBabel__OBConformerScore(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBConformerScore *)  ((OpenBabel::OBRMSDConformerScore *) x));
}
static void *_p_OpenBabel__OBFFCalculation3To_p_OpenBabel__OBFFCalculation2(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBFFCalculation2 *)  ((OpenBabel::OBFFCalculation3 *) x));
}
static void *_p_OpenBabel__OBFFCalculation4To_p_OpenBabel__OBFFCalculation2(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBFFCalculation2 *) (OpenBabel::OBFFCalculation3 *) ((OpenBabel::OBFFCalculation4 *) x));
}
static void *_p_OpenBabel__OBFFCalculation4To_p_OpenBabel__OBFFCalculation3(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBFFCalculation3 *)  ((OpenBabel::OBFFCalculation4 *) x));
}
static void *_p_OpenBabel__AliasDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::AliasData *) x));
}
static void *_p_OpenBabel__OBAngleDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBAngleData *) x));
}
static void *_p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *) (OpenBabel::OBStereoBase *)(OpenBabel::OBTetraPlanarStereo *) ((OpenBabel::OBCisTransStereo *) x));
}
static void *_p_OpenBabel__OBCommentDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBCommentData *) x));
}
static void *_p_OpenBabel__OBConformerDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBConformerData *) x));
}
static void *_p_OpenBabel__OBDOSDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBDOSData *) x));
}
static void *_p_OpenBabel__OBElectronicTransitionDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBElectronicTransitionData *) x));
}
static void *_p_OpenBabel__OBExternalBondDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBExternalBondData *) x));
}
static void *_p_OpenBabel__OBFreeGridTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBFreeGrid *) x));
}
static void *_p_OpenBabel__OBGridDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBGridData *) x));
}
static void *_p_OpenBabel__OBMatrixDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBMatrixData *) x));
}
static void *_p_OpenBabel__OBOrbitalDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBOrbitalData *) x));
}
static void *_p_OpenBabel__OBPairDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBPairData *) x));
}
static void *_p_OpenBabel__OBPairTemplateT_float_tTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBPairTemplate< float > *) x));
}
static void *_p_OpenBabel__OBPairTemplateT_int_tTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBPairTemplate< int > *) x));
}
static void *_p_OpenBabel__OBPchargeTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBPcharge *) x));
}
static void *_p_OpenBabel__OBRingDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBRingData *) x));
}
static void *_p_OpenBabel__OBRotamerListTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBRotamerList *) x));
}
static void *_p_OpenBabel__OBRotationDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBRotationData *) x));
}
static void *_p_OpenBabel__OBSerialNumsTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBSerialNums *) x));
}
static void *_p_OpenBabel__OBSetDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBSetData *) x));
}
static void *_p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *) (OpenBabel::OBStereoBase *)(OpenBabel::OBTetraPlanarStereo *) ((OpenBabel::OBSquarePlanarStereo *) x));
}
static void *_p_OpenBabel__OBStereoBaseTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBStereoBase *) x));
}
static void *_p_OpenBabel__OBSymmetryDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBSymmetryData *) x));
}
static void *_p_OpenBabel__OBTetraNonPlanarStereoTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *) (OpenBabel::OBStereoBase *) ((OpenBabel::OBTetraNonPlanarStereo *) x));
}
static void *_p_OpenBabel__OBTetraPlanarStereoTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *) (OpenBabel::OBStereoBase *) ((OpenBabel::OBTetraPlanarStereo *) x));
}
static void *_p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *) (OpenBabel::OBStereoBase *)(OpenBabel::OBTetraNonPlanarStereo *) ((OpenBabel::OBTetrahedralStereo *) x));
}
static void *_p_OpenBabel__OBTorsionDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBTorsionData *) x));
}
static void *_p_OpenBabel__OBUnitCellTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBUnitCell *) x));
}
static void *_p_OpenBabel__OBVectorDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBVectorData *) x));
}
static void *_p_OpenBabel__OBVibrationDataTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBVibrationData *) x));
}
static void *_p_OpenBabel__OBVirtualBondTo_p_OpenBabel__OBGenericData(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGenericData *)  ((OpenBabel::OBVirtualBond *) x));
}
static void *_p_OpenBabel__OBAtomTyperTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBAtomTyper *) x));
}
static void *_p_OpenBabel__OBAtomicHeatOfFormationTableTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBAtomicHeatOfFormationTable *) x));
}
static void *_p_OpenBabel__OBPhModelTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBPhModel *) x));
}
static void *_p_OpenBabel__OBResidueDataTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBResidueData *) x));
}
static void *_p_OpenBabel__OBRingTyperTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBRingTyper *) x));
}
static void *_p_OpenBabel__OBRotorRulesTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBRotorRules *) x));
}
static void *_p_OpenBabel__OBTypeTableTo_p_OpenBabel__OBGlobalDataBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBGlobalDataBase *)  ((OpenBabel::OBTypeTable *) x));
}
static void *_p_OpenBabel__OBChargeModelTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBChargeModel *) x));
}
static void *_p_OpenBabel__OBDescriptorTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBDescriptor *) x));
}
static void *_p_OpenBabel__OBFingerprintTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBFingerprint *) x));
}
static void *_p_OpenBabel__OBForceFieldTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBForceField *) x));
}
static void *_p_OpenBabel__OBFormatTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBFormat *) x));
}
static void *_p_OpenBabel__OBOpTo_p_OpenBabel__OBPlugin(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBPlugin *)  ((OpenBabel::OBOp *) x));
}
static void *_p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBStereoBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBStereoBase *) (OpenBabel::OBTetraPlanarStereo *) ((OpenBabel::OBCisTransStereo *) x));
}
static void *_p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBStereoBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBStereoBase *) (OpenBabel::OBTetraPlanarStereo *) ((OpenBabel::OBSquarePlanarStereo *) x));
}
static void *_p_OpenBabel__OBTetraNonPlanarStereoTo_p_OpenBabel__OBStereoBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBStereoBase *)  ((OpenBabel::OBTetraNonPlanarStereo *) x));
}
static void *_p_OpenBabel__OBTetraPlanarStereoTo_p_OpenBabel__OBStereoBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBStereoBase *)  ((OpenBabel::OBTetraPlanarStereo *) x));
}
static void *_p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBStereoBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBStereoBase *) (OpenBabel::OBTetraNonPlanarStereo *) ((OpenBabel::OBTetrahedralStereo *) x));
}
static void *_p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBTetraNonPlanarStereo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBTetraNonPlanarStereo *)  ((OpenBabel::OBTetrahedralStereo *) x));
}
static void *_p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBTetraPlanarStereo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBTetraPlanarStereo *)  ((OpenBabel::OBCisTransStereo *) x));
}
static void *_p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBTetraPlanarStereo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::OBTetraPlanarStereo *)  ((OpenBabel::OBSquarePlanarStereo *) x));
}
static void *_p_OpenBabel__transform3dTo_p_OpenBabel__matrix3x3(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::matrix3x3 *)  ((OpenBabel::transform3d *) x));
}
static void *_p_OpenBabel__transform3dTo_p_OpenBabel__vector3(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((OpenBabel::vector3 *)  ((OpenBabel::transform3d *) x));
}
static void *_p_p_doubleArrayTo_p_p_double(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((double **)  ((doubleArray **) x));
}
static void *_p_OpenBabel__obLogBufTo_p_std__stringbuf(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((std::stringbuf *)  ((OpenBabel::obLogBuf *) x));
}
static swig_type_info _swigt__p_ConstRefIter = {"_p_ConstRefIter", "ConstRefIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Iter = {"_p_Iter", "Iter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Mapping = {"_p_Mapping", "Mapping *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Mappings = {"_p_Mappings", "Mappings *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpMap = {"_p_OpMap", "OpMap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__AliasData = {"_p_OpenBabel__AliasData", "OpenBabel::AliasData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__CharPtrLess = {"_p_OpenBabel__CharPtrLess", "OpenBabel::CharPtrLess *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__FastSearch = {"_p_OpenBabel__FastSearch", "OpenBabel::FastSearch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__FastSearchIndexer = {"_p_OpenBabel__FastSearchIndexer", "OpenBabel::FastSearchIndexer *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__FptIndex = {"_p_OpenBabel__FptIndex", "OpenBabel::FptIndex *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__FptIndexHeader = {"_p_OpenBabel__FptIndexHeader", "OpenBabel::FptIndexHeader *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__LineSearchType = {"_p_OpenBabel__LineSearchType", "OpenBabel::LineSearchType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAlign = {"_p_OpenBabel__OBAlign", "OpenBabel::OBAlign *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAngle = {"_p_OpenBabel__OBAngle", "OpenBabel::OBAngle *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAngleData = {"_p_OpenBabel__OBAngleData", "OpenBabel::OBAngleData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAromaticTyper = {"_p_OpenBabel__OBAromaticTyper", "OpenBabel::OBAromaticTyper *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtom = {"_p_OpenBabel__OBAtom", "OpenBabel::OBNodeBase *|OpenBabel::OBAtom *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomAtomIter = {"_p_OpenBabel__OBAtomAtomIter", "OpenBabel::OBAtomAtomIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomAtomIterAdaptor = {"_p_OpenBabel__OBAtomAtomIterAdaptor", "OpenBabel::OBAtomAtomIterAdaptor *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomBondIter = {"_p_OpenBabel__OBAtomBondIter", "OpenBabel::OBAtomBondIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomHOF = {"_p_OpenBabel__OBAtomHOF", "OpenBabel::OBAtomHOF *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomTyper = {"_p_OpenBabel__OBAtomTyper", "OpenBabel::OBAtomTyper *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBAtomicHeatOfFormationTable = {"_p_OpenBabel__OBAtomicHeatOfFormationTable", "OpenBabel::OBAtomicHeatOfFormationTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBBase = {"_p_OpenBabel__OBBase", "OpenBabel::OBBase *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBBitVec = {"_p_OpenBabel__OBBitVec", "OpenBabel::OBBitVec *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBBond = {"_p_OpenBabel__OBBond", "std::vector< OpenBabel::OBBond >::value_type *|OpenBabel::OBBond *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBBuilder = {"_p_OpenBabel__OBBuilder", "OpenBabel::OBBuilder *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBChainsParser = {"_p_OpenBabel__OBChainsParser", "OpenBabel::OBChainsParser *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBChargeModel = {"_p_OpenBabel__OBChargeModel", "OpenBabel::OBChargeModel *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBChemTsfm = {"_p_OpenBabel__OBChemTsfm", "OpenBabel::OBChemTsfm *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBCisTransConfig = {"_p_OpenBabel__OBCisTransConfig", "OpenBabel::OBCisTransConfig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBCisTransStereo = {"_p_OpenBabel__OBCisTransStereo", "OpenBabel::OBCisTransStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBCommentData = {"_p_OpenBabel__OBCommentData", "OpenBabel::OBCommentData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConformerData = {"_p_OpenBabel__OBConformerData", "OpenBabel::OBConformerData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConformerFilter = {"_p_OpenBabel__OBConformerFilter", "OpenBabel::OBConformerFilter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConformerFilters = {"_p_OpenBabel__OBConformerFilters", "OpenBabel::OBConformerFilters *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConformerScore = {"_p_OpenBabel__OBConformerScore", "OpenBabel::OBConformerScore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConformerSearch = {"_p_OpenBabel__OBConformerSearch", "OpenBabel::OBConformerSearch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBConversion = {"_p_OpenBabel__OBConversion", "OpenBabel::OBConversion *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBDOSData = {"_p_OpenBabel__OBDOSData", "OpenBabel::OBDOSData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBDescriptor = {"_p_OpenBabel__OBDescriptor", "OpenBabel::OBDescriptor *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBElectronicTransitionData = {"_p_OpenBabel__OBElectronicTransitionData", "OpenBabel::OBElectronicTransitionData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBEnergyConformerScore = {"_p_OpenBabel__OBEnergyConformerScore", "OpenBabel::OBEnergyConformerScore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBError = {"_p_OpenBabel__OBError", "OpenBabel::OBError *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBExternalBond = {"_p_OpenBabel__OBExternalBond", "OpenBabel::OBExternalBond *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBExternalBondData = {"_p_OpenBabel__OBExternalBondData", "OpenBabel::OBExternalBondData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFCalculation2 = {"_p_OpenBabel__OBFFCalculation2", "OpenBabel::OBFFCalculation2 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFCalculation3 = {"_p_OpenBabel__OBFFCalculation3", "OpenBabel::OBFFCalculation3 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFCalculation4 = {"_p_OpenBabel__OBFFCalculation4", "OpenBabel::OBFFCalculation4 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFConstraint = {"_p_OpenBabel__OBFFConstraint", "OpenBabel::OBFFConstraint *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFConstraints = {"_p_OpenBabel__OBFFConstraints", "OpenBabel::OBFFConstraints *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFFParameter = {"_p_OpenBabel__OBFFParameter", "OpenBabel::OBFFParameter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFingerprint = {"_p_OpenBabel__OBFingerprint", "OpenBabel::OBFingerprint *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBForceField = {"_p_OpenBabel__OBForceField", "OpenBabel::OBForceField *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFormat = {"_p_OpenBabel__OBFormat", "OpenBabel::OBFormat *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFreeGrid = {"_p_OpenBabel__OBFreeGrid", "OpenBabel::OBFreeGrid *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBFreeGridPoint = {"_p_OpenBabel__OBFreeGridPoint", "OpenBabel::OBFreeGridPoint *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBGenericData = {"_p_OpenBabel__OBGenericData", "std::vector< OpenBabel::OBGenericData * >::value_type|OpenBabel::OBGenericData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBGlobalDataBase = {"_p_OpenBabel__OBGlobalDataBase", "OpenBabel::OBGlobalDataBase *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBGraphSym = {"_p_OpenBabel__OBGraphSym", "OpenBabel::OBGraphSym *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBGridData = {"_p_OpenBabel__OBGridData", "OpenBabel::OBGridData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBInternalCoord = {"_p_OpenBabel__OBInternalCoord", "std::vector< OpenBabel::OBInternalCoord * >::value_type|OpenBabel::OBInternalCoord *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBIsomorphismMapper = {"_p_OpenBabel__OBIsomorphismMapper", "OpenBabel::OBIsomorphismMapper *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBIsomorphismMapper__Functor = {"_p_OpenBabel__OBIsomorphismMapper__Functor", "OpenBabel::OBIsomorphismMapper::Functor *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMatrixData = {"_p_OpenBabel__OBMatrixData", "OpenBabel::OBMatrixData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMessageHandler = {"_p_OpenBabel__OBMessageHandler", "OpenBabel::OBMessageHandler *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMinimizingEnergyConformerScore = {"_p_OpenBabel__OBMinimizingEnergyConformerScore", "OpenBabel::OBMinimizingEnergyConformerScore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMinimizingRMSDConformerScore = {"_p_OpenBabel__OBMinimizingRMSDConformerScore", "OpenBabel::OBMinimizingRMSDConformerScore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMol = {"_p_OpenBabel__OBMol", "std::vector< OpenBabel::OBMol >::value_type *|OpenBabel::OBMol *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolAngleIter = {"_p_OpenBabel__OBMolAngleIter", "OpenBabel::OBMolAngleIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolAtomBFSIter = {"_p_OpenBabel__OBMolAtomBFSIter", "OpenBabel::OBMolAtomBFSIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolAtomDFSIter = {"_p_OpenBabel__OBMolAtomDFSIter", "OpenBabel::OBMolAtomDFSIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolAtomIter = {"_p_OpenBabel__OBMolAtomIter", "OpenBabel::OBMolAtomIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolBondBFSIter = {"_p_OpenBabel__OBMolBondBFSIter", "OpenBabel::OBMolBondBFSIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolBondIter = {"_p_OpenBabel__OBMolBondIter", "OpenBabel::OBMolBondIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolPairIter = {"_p_OpenBabel__OBMolPairIter", "OpenBabel::OBMolPairIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolRingIter = {"_p_OpenBabel__OBMolRingIter", "OpenBabel::OBMolRingIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBMolTorsionIter = {"_p_OpenBabel__OBMolTorsionIter", "OpenBabel::OBMolTorsionIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBNasaThermoData = {"_p_OpenBabel__OBNasaThermoData", "OpenBabel::OBNasaThermoData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBOp = {"_p_OpenBabel__OBOp", "OpenBabel::OBOp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBOrbital = {"_p_OpenBabel__OBOrbital", "OpenBabel::OBOrbital *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBOrbitalData = {"_p_OpenBabel__OBOrbitalData", "OpenBabel::OBOrbitalData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPairData = {"_p_OpenBabel__OBPairData", "OpenBabel::OBPairData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPairTemplateT_bool_t = {"_p_OpenBabel__OBPairTemplateT_bool_t", "OpenBabel::OBPairBool *|OpenBabel::OBPairTemplate< bool > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPairTemplateT_double_t = {"_p_OpenBabel__OBPairTemplateT_double_t", "OpenBabel::OBPairFloatingPoint *|OpenBabel::OBPairTemplate< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPairTemplateT_float_t = {"_p_OpenBabel__OBPairTemplateT_float_t", "OpenBabel::OBPairTemplate< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPairTemplateT_int_t = {"_p_OpenBabel__OBPairTemplateT_int_t", "OpenBabel::OBPairInteger *|OpenBabel::OBPairTemplate< int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPcharge = {"_p_OpenBabel__OBPcharge", "OpenBabel::OBPcharge *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPhModel = {"_p_OpenBabel__OBPhModel", "OpenBabel::OBPhModel *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBPlugin = {"_p_OpenBabel__OBPlugin", "OpenBabel::OBPlugin *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBQuery = {"_p_OpenBabel__OBQuery", "OpenBabel::OBQuery *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBQueryAtom = {"_p_OpenBabel__OBQueryAtom", "OpenBabel::OBQueryAtom *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBQueryBond = {"_p_OpenBabel__OBQueryBond", "OpenBabel::OBQueryBond *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRMSDConformerScore = {"_p_OpenBabel__OBRMSDConformerScore", "OpenBabel::OBRMSDConformerScore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRTree = {"_p_OpenBabel__OBRTree", "OpenBabel::OBRTree *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t = {"_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t", "OpenBabel::OBAtomAtomRange *|OpenBabel::OBRange< OpenBabel::OBAtom *,OpenBabel::OBAtomAtomIterAdaptor > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t = {"_p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t", "OpenBabel::OBMolAtomRange *|OpenBabel::OBRange< OpenBabel::OBAtom *,std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t = {"_p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t", "OpenBabel::OBAtomBondRange *|OpenBabel::OBMolBondRange *|OpenBabel::OBRange< OpenBabel::OBBond *,std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRateData = {"_p_OpenBabel__OBRateData", "OpenBabel::OBRateData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBReaction = {"_p_OpenBabel__OBReaction", "OpenBabel::OBReaction *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBReactionFacade = {"_p_OpenBabel__OBReactionFacade", "OpenBabel::OBReactionFacade *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBResidue = {"_p_OpenBabel__OBResidue", "std::vector< OpenBabel::OBResidue >::value_type *|OpenBabel::OBResidue *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBResidueAtomIter = {"_p_OpenBabel__OBResidueAtomIter", "OpenBabel::OBResidueAtomIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBResidueData = {"_p_OpenBabel__OBResidueData", "OpenBabel::OBResidueData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBResidueIter = {"_p_OpenBabel__OBResidueIter", "OpenBabel::OBResidueIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRing = {"_p_OpenBabel__OBRing", "std::vector< OpenBabel::OBRing * >::value_type|std::vector< OpenBabel::OBRing >::value_type *|OpenBabel::OBRing *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRingData = {"_p_OpenBabel__OBRingData", "OpenBabel::OBRingData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRingSearch = {"_p_OpenBabel__OBRingSearch", "OpenBabel::OBRingSearch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRingTyper = {"_p_OpenBabel__OBRingTyper", "OpenBabel::OBRingTyper *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotamerList = {"_p_OpenBabel__OBRotamerList", "OpenBabel::OBRotamerList *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotationData = {"_p_OpenBabel__OBRotationData", "OpenBabel::OBRotationData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotor = {"_p_OpenBabel__OBRotor", "OpenBabel::OBRotor *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotorKeys = {"_p_OpenBabel__OBRotorKeys", "OpenBabel::OBRotorKeys *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotorList = {"_p_OpenBabel__OBRotorList", "OpenBabel::OBRotorList *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotorRule = {"_p_OpenBabel__OBRotorRule", "OpenBabel::OBRotorRule *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBRotorRules = {"_p_OpenBabel__OBRotorRules", "OpenBabel::OBRotorRules *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSSMatch = {"_p_OpenBabel__OBSSMatch", "OpenBabel::OBSSMatch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSerialNums = {"_p_OpenBabel__OBSerialNums", "OpenBabel::OBSerialNums *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSetData = {"_p_OpenBabel__OBSetData", "OpenBabel::OBSetData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSmartsMatcher = {"_p_OpenBabel__OBSmartsMatcher", "OpenBabel::OBSmartsMatcher *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSmartsPattern = {"_p_OpenBabel__OBSmartsPattern", "OpenBabel::OBSmartsPattern *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSpectrophore = {"_p_OpenBabel__OBSpectrophore", "OpenBabel::OBSpectrophore *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSqrtTbl = {"_p_OpenBabel__OBSqrtTbl", "OpenBabel::OBSqrtTbl *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSquarePlanarConfig = {"_p_OpenBabel__OBSquarePlanarConfig", "OpenBabel::OBSquarePlanarConfig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSquarePlanarStereo = {"_p_OpenBabel__OBSquarePlanarStereo", "OpenBabel::OBSquarePlanarStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStereo = {"_p_OpenBabel__OBStereo", "OpenBabel::OBStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStereoBase = {"_p_OpenBabel__OBStereoBase", "OpenBabel::OBStereoBase *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStereoFacade = {"_p_OpenBabel__OBStereoFacade", "OpenBabel::OBStereoFacade *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStereoUnit = {"_p_OpenBabel__OBStereoUnit", "OpenBabel::OBStereoUnit *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStericConformerFilter = {"_p_OpenBabel__OBStericConformerFilter", "OpenBabel::OBStericConformerFilter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBStopwatch = {"_p_OpenBabel__OBStopwatch", "OpenBabel::OBStopwatch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBSymmetryData = {"_p_OpenBabel__OBSymmetryData", "OpenBabel::OBSymmetryData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTetraNonPlanarStereo = {"_p_OpenBabel__OBTetraNonPlanarStereo", "OpenBabel::OBTetraNonPlanarStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTetraPlanarStereo = {"_p_OpenBabel__OBTetraPlanarStereo", "OpenBabel::OBTetraPlanarStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTetrahedralConfig = {"_p_OpenBabel__OBTetrahedralConfig", "OpenBabel::OBTetrahedralConfig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTetrahedralStereo = {"_p_OpenBabel__OBTetrahedralStereo", "OpenBabel::OBTetrahedralStereo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTorsion = {"_p_OpenBabel__OBTorsion", "OpenBabel::OBTorsion *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTorsionData = {"_p_OpenBabel__OBTorsionData", "OpenBabel::OBTorsionData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBTypeTable = {"_p_OpenBabel__OBTypeTable", "OpenBabel::OBTypeTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBUnitCell = {"_p_OpenBabel__OBUnitCell", "OpenBabel::OBUnitCell *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBVectorData = {"_p_OpenBabel__OBVectorData", "OpenBabel::OBVectorData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBVibrationData = {"_p_OpenBabel__OBVibrationData", "OpenBabel::OBVibrationData *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__OBVirtualBond = {"_p_OpenBabel__OBVirtualBond", "OpenBabel::OBVirtualBond *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__Pattern = {"_p_OpenBabel__Pattern", "OpenBabel::Pattern *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__SpaceGroup = {"_p_OpenBabel__SpaceGroup", "OpenBabel::SpaceGroup *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__Template = {"_p_OpenBabel__Template", "OpenBabel::Template *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__matrix3x3 = {"_p_OpenBabel__matrix3x3", "OpenBabel::matrix3x3 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__obLogBuf = {"_p_OpenBabel__obLogBuf", "OpenBabel::obLogBuf *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t = {"_p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t", "OpenBabel::quad< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__rotor_digit = {"_p_OpenBabel__rotor_digit", "OpenBabel::rotor_digit *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__transform3d = {"_p_OpenBabel__transform3d", "OpenBabel::transform3d *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t = {"_p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t", "OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,OpenBabel::OBAtom * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_OpenBabel__vector3 = {"_p_OpenBabel__vector3", "std::vector< OpenBabel::vector3 >::value_type *|OpenBabel::vector3 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_PluginIterator = {"_p_PluginIterator", "PluginIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_PluginMapType = {"_p_PluginMapType", "PluginMapType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Ref = {"_p_Ref", "Ref *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_RefIter = {"_p_RefIter", "RefIter *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Refs = {"_p_Refs", "Refs *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Unit = {"_p_Unit", "Unit *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_a_3__double = {"_p_a_3__double", "double (*)[3]", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_a_3__doubleArray = {"_p_a_3__doubleArray", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_allocator_type = {"_p_allocator_type", "allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_bool = {"_p_bool", "bool *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_difference_type = {"_p_difference_type", "difference_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_double = {"_p_double", "OpenBabel::vector3::const_iterator|OpenBabel::vector3::iterator|double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_doubleArray = {"_p_doubleArray", "doubleArray *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_first_type = {"_p_first_type", "first_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_iterator_category = {"_p_iterator_category", "iterator_category *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBChargeModel = {"_p_p_OpenBabel__OBChargeModel", "OpenBabel::OBChargeModel **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBDescriptor = {"_p_p_OpenBabel__OBDescriptor", "OpenBabel::OBDescriptor **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBFingerprint = {"_p_p_OpenBabel__OBFingerprint", "OpenBabel::OBFingerprint **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBForceField = {"_p_p_OpenBabel__OBForceField", "OpenBabel::OBForceField **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBFormat = {"_p_p_OpenBabel__OBFormat", "OpenBabel::OBFormat **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_OpenBabel__OBOp = {"_p_p_OpenBabel__OBOp", "OpenBabel::OBOp **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_PyObject = {"_p_p_PyObject", "PyObject **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_double = {"_p_p_double", "double **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_doubleArray = {"_p_p_doubleArray", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_p_int = {"_p_p_int", "int **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_second_type = {"_p_second_type", "second_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_size_type = {"_p_size_type", "size_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t = {"_p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t", "std::vector< std::vector< std::pair< unsigned int,unsigned int > > >::allocator_type *|std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__dequeT_int_t = {"_p_std__dequeT_int_t", "std::deque< int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__invalid_argument = {"_p_std__invalid_argument", "std::invalid_argument *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__istream = {"_p_std__istream", "std::istream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator = {"_p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator", "OpenBabel::transform3dIterator *|std::list< OpenBabel::transform3d *,std::allocator< OpenBabel::transform3d * > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t = {"_p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t", "std::list< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t = {"_p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t", "std::map< OpenBabel::OBBond *,enum OpenBabel::OBStereo::BondDirection,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,enum OpenBabel::OBStereo::BondDirection > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t = {"_p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t", "std::map< OpenBabel::OBBond *,OpenBabel::OBStereo::Ref,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,OpenBabel::OBStereo::Ref > > > *|std::map< OpenBabel::OBBond *,unsigned long,std::less< OpenBabel::OBBond * >,std::allocator< std::pair< OpenBabel::OBBond *const,unsigned long > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t = {"_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t", "OpenBabel::OBPlugin::PluginMapType *|std::map< char const *,OpenBabel::OBPlugin *,OpenBabel::CharPtrLess,std::allocator< std::pair< char const *const,OpenBabel::OBPlugin * > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator = {"_p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator", "OpenBabel::Formatpos *|OpenBabel::OBPlugin::PluginIterator *|std::map< char const *,OpenBabel::OBPlugin *,OpenBabel::CharPtrLess,std::allocator< std::pair< char const *const,OpenBabel::OBPlugin * > > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t = {"_p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t", "std::map< int,OpenBabel::OBAtom *,std::less< int >,std::allocator< std::pair< int const,OpenBabel::OBAtom * > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t = {"_p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t", "OpenBabel::OBOp::OpMap *|std::map< std::string,std::string,std::less< std::string >,std::allocator< std::pair< std::string const,std::string > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t = {"_p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t", "OpenBabel::mapRotorEnergy *|std::map< std::vector< int,std::allocator< int > >,double,std::less< std::vector< int > >,std::allocator< std::pair< std::vector< int,std::allocator< int > > const,double > > > *|std::map< std::vector< int,std::allocator< int > >,double,std::less< std::vector< int,std::allocator< int > > >,std::allocator< std::pair< std::vector< int,std::allocator< int > > const,double > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__multimapT_double_unsigned_long_t = {"_p_std__multimapT_double_unsigned_long_t", "std::multimap< double,unsigned long > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__ostream = {"_p_std__ostream", "std::ostream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t = {"_p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t", "std::pair< OpenBabel::OBAtom *,OpenBabel::OBAtom * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_int_int_t = {"_p_std__pairT_int_int_t", "std::pair< int,int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_std__string_std__string_t = {"_p_std__pairT_std__string_std__string_t", "std::pair< std::string,std::string > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_unsigned_int_unsigned_int_t = {"_p_std__pairT_unsigned_int_unsigned_int_t", "std::pair< unsigned int,unsigned int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setT_OpenBabel__OBBond_p_t = {"_p_std__setT_OpenBabel__OBBond_p_t", "std::set< OpenBabel::OBBond * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__shared_ptrT_OpenBabel__OBMol_t = {"_p_std__shared_ptrT_OpenBabel__OBMol_t", "std::shared_ptr< OpenBabel::OBMol > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__streampos = {"_p_std__streampos", "std::streampos *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__string = {"_p_std__string", "std::string *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__stringbuf = {"_p_std__stringbuf", "std::stringbuf *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__type_info = {"_p_std__type_info", "std::type_info *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t = {"_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t", "std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator = {"_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator", "OpenBabel::OBAtomConstIterator *|std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator = {"_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator", "OpenBabel::OBAtomIterator *|std::vector< OpenBabel::OBAtom *,std::allocator< OpenBabel::OBAtom * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t = {"_p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t", "std::vector< OpenBabel::OBBase *,std::allocator< OpenBabel::OBBase * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t = {"_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t", "std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator = {"_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator", "OpenBabel::OBAtomAtomIterAdaptor::Iter *|std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::const_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator = {"_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator", "OpenBabel::OBBondIterator *|std::vector< OpenBabel::OBBond *,std::allocator< OpenBabel::OBBond * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBBond_t = {"_p_std__vectorT_OpenBabel__OBBond_t", "std::vector< OpenBabel::OBBond,std::allocator< OpenBabel::OBBond > > *|std::vector< OpenBabel::OBBond > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t = {"_p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t", "std::vector< OpenBabel::OBCisTransStereo *,std::allocator< OpenBabel::OBCisTransStereo * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t = {"_p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t", "std::vector< OpenBabel::OBConformerFilter *,std::allocator< OpenBabel::OBConformerFilter * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t = {"_p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t", "std::vector< OpenBabel::OBExternalBond,std::allocator< OpenBabel::OBExternalBond > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator = {"_p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator", "OpenBabel::OBFreeGridPointIterator *|std::vector< OpenBabel::OBFreeGridPoint *,std::allocator< OpenBabel::OBFreeGridPoint * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t = {"_p_std__vectorT_OpenBabel__OBGenericData_p_t", "std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > > *|std::vector< OpenBabel::OBGenericData * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator = {"_p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator", "OpenBabel::OBDataIterator *|std::vector< OpenBabel::OBGenericData *,std::allocator< OpenBabel::OBGenericData * > >::iterator *|std::vector< OpenBabel::OBGenericData * >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t = {"_p_std__vectorT_OpenBabel__OBInternalCoord_p_t", "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > > *|std::vector< OpenBabel::OBInternalCoord * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator = {"_p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator", "std::vector< OpenBabel::OBInternalCoord *,std::allocator< OpenBabel::OBInternalCoord * > >::iterator *|std::vector< OpenBabel::OBInternalCoord * >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBMol_t = {"_p_std__vectorT_OpenBabel__OBMol_t", "std::vector< OpenBabel::OBMol,std::allocator< OpenBabel::OBMol > > *|std::vector< OpenBabel::OBMol > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t = {"_p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t", "std::vector< OpenBabel::OBOrbital,std::allocator< OpenBabel::OBOrbital > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t = {"_p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t", "std::vector< OpenBabel::OBQueryAtom *,std::allocator< OpenBabel::OBQueryAtom * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t = {"_p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t", "std::vector< OpenBabel::OBQueryBond *,std::allocator< OpenBabel::OBQueryBond * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator = {"_p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator", "OpenBabel::OBResidueIterator *|std::vector< OpenBabel::OBResidue *,std::allocator< OpenBabel::OBResidue * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBResidue_t = {"_p_std__vectorT_OpenBabel__OBResidue_t", "std::vector< OpenBabel::OBResidue,std::allocator< OpenBabel::OBResidue > > *|std::vector< OpenBabel::OBResidue > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBRing_p_t = {"_p_std__vectorT_OpenBabel__OBRing_p_t", "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > > *|std::vector< OpenBabel::OBRing * > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBRing_p_t__iterator = {"_p_std__vectorT_OpenBabel__OBRing_p_t__iterator", "std::vector< OpenBabel::OBRing *,std::allocator< OpenBabel::OBRing * > >::iterator *|std::vector< OpenBabel::OBRing * >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBRing_t = {"_p_std__vectorT_OpenBabel__OBRing_t", "std::vector< OpenBabel::OBRing,std::allocator< OpenBabel::OBRing > > *|std::vector< OpenBabel::OBRing > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator = {"_p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator", "OpenBabel::OBRotorIterator *|std::vector< OpenBabel::OBRotor *,std::allocator< OpenBabel::OBRotor * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t = {"_p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t", "std::vector< OpenBabel::OBSquarePlanarStereo *,std::allocator< OpenBabel::OBSquarePlanarStereo * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t = {"_p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t", "OpenBabel::OBStereoUnitSet *|std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t = {"_p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t", "std::vector< OpenBabel::OBTetrahedralStereo *,std::allocator< OpenBabel::OBTetrahedralStereo * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t = {"_p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t", "std::vector< OpenBabel::OBTorsion,std::allocator< OpenBabel::OBTorsion > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t = {"_p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t", "std::vector< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double >,std::allocator< OpenBabel::triple< OpenBabel::OBAtom *,OpenBabel::OBAtom *,double > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_OpenBabel__vector3_t = {"_p_std__vectorT_OpenBabel__vector3_t", "std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > *|std::vector< OpenBabel::vector3 > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t = {"_p_std__vectorT_double_p_std__allocatorT_double_p_t_t", "std::vector< double *,std::allocator< double * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator = {"_p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator", "std::vector< double *,std::allocator< double * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_double_t = {"_p_std__vectorT_double_t", "std::vector< double,std::allocator< double > > *|std::vector< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_int_t = {"_p_std__vectorT_int_t", "OpenBabel::RotorKey *|std::vector< int,std::allocator< int > > *|std::vector< int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t = {"_p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t", "std::vector< std::pair< int,int >,std::allocator< std::pair< int,int > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t = {"_p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t", "std::vector< std::pair< std::string,std::string >,std::allocator< std::pair< std::string,std::string > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t = {"_p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t", "OpenBabel::Automorphism *|OpenBabel::OBIsomorphismMapper::Mapping *|std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > *|std::vector< std::pair< unsigned int,unsigned int > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__string_t = {"_p_std__vectorT_std__string_t", "std::vector< std::string,std::allocator< std::string > > *|std::vector< std::string > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t = {"_p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t", "OpenBabel::OBStereoUnitSetOfSets *|std::vector< std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > >,std::allocator< std::vector< OpenBabel::OBStereoUnit,std::allocator< OpenBabel::OBStereoUnit > > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t = {"_p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t", "std::vector< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > >,std::allocator< std::vector< OpenBabel::vector3,std::allocator< OpenBabel::vector3 > > > > *|std::vector< std::vector< OpenBabel::vector3 > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__vectorT_int_t_t = {"_p_std__vectorT_std__vectorT_int_t_t", "OpenBabel::RotorKeys *|std::vector< std::vector< int,std::allocator< int > >,std::allocator< std::vector< int,std::allocator< int > > > > *|std::vector< std::vector< int > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t = {"_p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t", "OpenBabel::Automorphisms *|OpenBabel::OBIsomorphismMapper::Mappings *|std::vector< OpenBabel::OBIsomorphismMapper::Mapping,std::allocator< OpenBabel::OBIsomorphismMapper::Mapping > > *|std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > *|std::vector< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > >,std::allocator< std::vector< std::pair< unsigned int,unsigned int >,std::allocator< std::pair< unsigned int,unsigned int > > > > > *|std::vector< std::vector< std::pair< unsigned int,unsigned int > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t = {"_p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t", "std::vector< std::vector< unsigned int,std::allocator< unsigned int > >,std::allocator< std::vector< unsigned int,std::allocator< unsigned int > > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t = {"_p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t", "OpenBabel::OBBitVec::word_vector *|std::vector< uint32_t,std::allocator< uint32_t > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator = {"_p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator", "std::vector< unsigned char *,std::allocator< unsigned char * > >::iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_unsigned_int_t = {"_p_std__vectorT_unsigned_int_t", "std::vector< unsigned int,std::allocator< unsigned int > > *|std::vector< unsigned int > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_unsigned_long_t = {"_p_std__vectorT_unsigned_long_t", "OpenBabel::OBStereo::Refs *|std::vector< unsigned long,std::allocator< unsigned long > > *|std::vector< unsigned long > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t = {"_p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t", "std::vector< unsigned short,std::allocator< unsigned short > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_swig__SwigPyIterator = {"_p_swig__SwigPyIterator", "swig::SwigPyIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_type = {"_p_value_type", "value_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_word_vector = {"_p_word_vector", "word_vector *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_ConstRefIter,
  &_swigt__p_Iter,
  &_swigt__p_Mapping,
  &_swigt__p_Mappings,
  &_swigt__p_OpMap,
  &_swigt__p_OpenBabel__AliasData,
  &_swigt__p_OpenBabel__CharPtrLess,
  &_swigt__p_OpenBabel__FastSearch,
  &_swigt__p_OpenBabel__FastSearchIndexer,
  &_swigt__p_OpenBabel__FptIndex,
  &_swigt__p_OpenBabel__FptIndexHeader,
  &_swigt__p_OpenBabel__LineSearchType,
  &_swigt__p_OpenBabel__OBAlign,
  &_swigt__p_OpenBabel__OBAngle,
  &_swigt__p_OpenBabel__OBAngleData,
  &_swigt__p_OpenBabel__OBAromaticTyper,
  &_swigt__p_OpenBabel__OBAtom,
  &_swigt__p_OpenBabel__OBAtomAtomIter,
  &_swigt__p_OpenBabel__OBAtomAtomIterAdaptor,
  &_swigt__p_OpenBabel__OBAtomBondIter,
  &_swigt__p_OpenBabel__OBAtomHOF,
  &_swigt__p_OpenBabel__OBAtomTyper,
  &_swigt__p_OpenBabel__OBAtomicHeatOfFormationTable,
  &_swigt__p_OpenBabel__OBBase,
  &_swigt__p_OpenBabel__OBBitVec,
  &_swigt__p_OpenBabel__OBBond,
  &_swigt__p_OpenBabel__OBBuilder,
  &_swigt__p_OpenBabel__OBChainsParser,
  &_swigt__p_OpenBabel__OBChargeModel,
  &_swigt__p_OpenBabel__OBChemTsfm,
  &_swigt__p_OpenBabel__OBCisTransConfig,
  &_swigt__p_OpenBabel__OBCisTransStereo,
  &_swigt__p_OpenBabel__OBCommentData,
  &_swigt__p_OpenBabel__OBConformerData,
  &_swigt__p_OpenBabel__OBConformerFilter,
  &_swigt__p_OpenBabel__OBConformerFilters,
  &_swigt__p_OpenBabel__OBConformerScore,
  &_swigt__p_OpenBabel__OBConformerSearch,
  &_swigt__p_OpenBabel__OBConversion,
  &_swigt__p_OpenBabel__OBDOSData,
  &_swigt__p_OpenBabel__OBDescriptor,
  &_swigt__p_OpenBabel__OBElectronicTransitionData,
  &_swigt__p_OpenBabel__OBEnergyConformerScore,
  &_swigt__p_OpenBabel__OBError,
  &_swigt__p_OpenBabel__OBExternalBond,
  &_swigt__p_OpenBabel__OBExternalBondData,
  &_swigt__p_OpenBabel__OBFFCalculation2,
  &_swigt__p_OpenBabel__OBFFCalculation3,
  &_swigt__p_OpenBabel__OBFFCalculation4,
  &_swigt__p_OpenBabel__OBFFConstraint,
  &_swigt__p_OpenBabel__OBFFConstraints,
  &_swigt__p_OpenBabel__OBFFParameter,
  &_swigt__p_OpenBabel__OBFingerprint,
  &_swigt__p_OpenBabel__OBForceField,
  &_swigt__p_OpenBabel__OBFormat,
  &_swigt__p_OpenBabel__OBFreeGrid,
  &_swigt__p_OpenBabel__OBFreeGridPoint,
  &_swigt__p_OpenBabel__OBGenericData,
  &_swigt__p_OpenBabel__OBGlobalDataBase,
  &_swigt__p_OpenBabel__OBGraphSym,
  &_swigt__p_OpenBabel__OBGridData,
  &_swigt__p_OpenBabel__OBInternalCoord,
  &_swigt__p_OpenBabel__OBIsomorphismMapper,
  &_swigt__p_OpenBabel__OBIsomorphismMapper__Functor,
  &_swigt__p_OpenBabel__OBMatrixData,
  &_swigt__p_OpenBabel__OBMessageHandler,
  &_swigt__p_OpenBabel__OBMinimizingEnergyConformerScore,
  &_swigt__p_OpenBabel__OBMinimizingRMSDConformerScore,
  &_swigt__p_OpenBabel__OBMol,
  &_swigt__p_OpenBabel__OBMolAngleIter,
  &_swigt__p_OpenBabel__OBMolAtomBFSIter,
  &_swigt__p_OpenBabel__OBMolAtomDFSIter,
  &_swigt__p_OpenBabel__OBMolAtomIter,
  &_swigt__p_OpenBabel__OBMolBondBFSIter,
  &_swigt__p_OpenBabel__OBMolBondIter,
  &_swigt__p_OpenBabel__OBMolPairIter,
  &_swigt__p_OpenBabel__OBMolRingIter,
  &_swigt__p_OpenBabel__OBMolTorsionIter,
  &_swigt__p_OpenBabel__OBNasaThermoData,
  &_swigt__p_OpenBabel__OBOp,
  &_swigt__p_OpenBabel__OBOrbital,
  &_swigt__p_OpenBabel__OBOrbitalData,
  &_swigt__p_OpenBabel__OBPairData,
  &_swigt__p_OpenBabel__OBPairTemplateT_bool_t,
  &_swigt__p_OpenBabel__OBPairTemplateT_double_t,
  &_swigt__p_OpenBabel__OBPairTemplateT_float_t,
  &_swigt__p_OpenBabel__OBPairTemplateT_int_t,
  &_swigt__p_OpenBabel__OBPcharge,
  &_swigt__p_OpenBabel__OBPhModel,
  &_swigt__p_OpenBabel__OBPlugin,
  &_swigt__p_OpenBabel__OBQuery,
  &_swigt__p_OpenBabel__OBQueryAtom,
  &_swigt__p_OpenBabel__OBQueryBond,
  &_swigt__p_OpenBabel__OBRMSDConformerScore,
  &_swigt__p_OpenBabel__OBRTree,
  &_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t,
  &_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t,
  &_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t,
  &_swigt__p_OpenBabel__OBRateData,
  &_swigt__p_OpenBabel__OBReaction,
  &_swigt__p_OpenBabel__OBReactionFacade,
  &_swigt__p_OpenBabel__OBResidue,
  &_swigt__p_OpenBabel__OBResidueAtomIter,
  &_swigt__p_OpenBabel__OBResidueData,
  &_swigt__p_OpenBabel__OBResidueIter,
  &_swigt__p_OpenBabel__OBRing,
  &_swigt__p_OpenBabel__OBRingData,
  &_swigt__p_OpenBabel__OBRingSearch,
  &_swigt__p_OpenBabel__OBRingTyper,
  &_swigt__p_OpenBabel__OBRotamerList,
  &_swigt__p_OpenBabel__OBRotationData,
  &_swigt__p_OpenBabel__OBRotor,
  &_swigt__p_OpenBabel__OBRotorKeys,
  &_swigt__p_OpenBabel__OBRotorList,
  &_swigt__p_OpenBabel__OBRotorRule,
  &_swigt__p_OpenBabel__OBRotorRules,
  &_swigt__p_OpenBabel__OBSSMatch,
  &_swigt__p_OpenBabel__OBSerialNums,
  &_swigt__p_OpenBabel__OBSetData,
  &_swigt__p_OpenBabel__OBSmartsMatcher,
  &_swigt__p_OpenBabel__OBSmartsPattern,
  &_swigt__p_OpenBabel__OBSpectrophore,
  &_swigt__p_OpenBabel__OBSqrtTbl,
  &_swigt__p_OpenBabel__OBSquarePlanarConfig,
  &_swigt__p_OpenBabel__OBSquarePlanarStereo,
  &_swigt__p_OpenBabel__OBStereo,
  &_swigt__p_OpenBabel__OBStereoBase,
  &_swigt__p_OpenBabel__OBStereoFacade,
  &_swigt__p_OpenBabel__OBStereoUnit,
  &_swigt__p_OpenBabel__OBStericConformerFilter,
  &_swigt__p_OpenBabel__OBStopwatch,
  &_swigt__p_OpenBabel__OBSymmetryData,
  &_swigt__p_OpenBabel__OBTetraNonPlanarStereo,
  &_swigt__p_OpenBabel__OBTetraPlanarStereo,
  &_swigt__p_OpenBabel__OBTetrahedralConfig,
  &_swigt__p_OpenBabel__OBTetrahedralStereo,
  &_swigt__p_OpenBabel__OBTorsion,
  &_swigt__p_OpenBabel__OBTorsionData,
  &_swigt__p_OpenBabel__OBTypeTable,
  &_swigt__p_OpenBabel__OBUnitCell,
  &_swigt__p_OpenBabel__OBVectorData,
  &_swigt__p_OpenBabel__OBVibrationData,
  &_swigt__p_OpenBabel__OBVirtualBond,
  &_swigt__p_OpenBabel__Pattern,
  &_swigt__p_OpenBabel__SpaceGroup,
  &_swigt__p_OpenBabel__Template,
  &_swigt__p_OpenBabel__matrix3x3,
  &_swigt__p_OpenBabel__obLogBuf,
  &_swigt__p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  &_swigt__p_OpenBabel__rotor_digit,
  &_swigt__p_OpenBabel__transform3d,
  &_swigt__p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  &_swigt__p_OpenBabel__vector3,
  &_swigt__p_PluginIterator,
  &_swigt__p_PluginMapType,
  &_swigt__p_Ref,
  &_swigt__p_RefIter,
  &_swigt__p_Refs,
  &_swigt__p_Unit,
  &_swigt__p_a_3__double,
  &_swigt__p_a_3__doubleArray,
  &_swigt__p_allocator_type,
  &_swigt__p_bool,
  &_swigt__p_char,
  &_swigt__p_difference_type,
  &_swigt__p_double,
  &_swigt__p_doubleArray,
  &_swigt__p_first_type,
  &_swigt__p_int,
  &_swigt__p_iterator_category,
  &_swigt__p_p_OpenBabel__OBChargeModel,
  &_swigt__p_p_OpenBabel__OBDescriptor,
  &_swigt__p_p_OpenBabel__OBFingerprint,
  &_swigt__p_p_OpenBabel__OBForceField,
  &_swigt__p_p_OpenBabel__OBFormat,
  &_swigt__p_p_OpenBabel__OBOp,
  &_swigt__p_p_PyObject,
  &_swigt__p_p_double,
  &_swigt__p_p_doubleArray,
  &_swigt__p_p_int,
  &_swigt__p_second_type,
  &_swigt__p_size_type,
  &_swigt__p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,
  &_swigt__p_std__dequeT_int_t,
  &_swigt__p_std__invalid_argument,
  &_swigt__p_std__istream,
  &_swigt__p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator,
  &_swigt__p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t,
  &_swigt__p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t,
  &_swigt__p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t,
  &_swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t,
  &_swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator,
  &_swigt__p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t,
  &_swigt__p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t,
  &_swigt__p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t,
  &_swigt__p_std__multimapT_double_unsigned_long_t,
  &_swigt__p_std__ostream,
  &_swigt__p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  &_swigt__p_std__pairT_int_int_t,
  &_swigt__p_std__pairT_std__string_std__string_t,
  &_swigt__p_std__pairT_unsigned_int_unsigned_int_t,
  &_swigt__p_std__setT_OpenBabel__OBBond_p_t,
  &_swigt__p_std__shared_ptrT_OpenBabel__OBMol_t,
  &_swigt__p_std__streampos,
  &_swigt__p_std__string,
  &_swigt__p_std__stringbuf,
  &_swigt__p_std__type_info,
  &_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBBond_t,
  &_swigt__p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t,
  &_swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t,
  &_swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBMol_t,
  &_swigt__p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBResidue_t,
  &_swigt__p_std__vectorT_OpenBabel__OBRing_p_t,
  &_swigt__p_std__vectorT_OpenBabel__OBRing_p_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBRing_t,
  &_swigt__p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator,
  &_swigt__p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t,
  &_swigt__p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t,
  &_swigt__p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t,
  &_swigt__p_std__vectorT_OpenBabel__vector3_t,
  &_swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t,
  &_swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator,
  &_swigt__p_std__vectorT_double_t,
  &_swigt__p_std__vectorT_int_t,
  &_swigt__p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t,
  &_swigt__p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t,
  &_swigt__p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t,
  &_swigt__p_std__vectorT_std__string_t,
  &_swigt__p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t,
  &_swigt__p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t,
  &_swigt__p_std__vectorT_std__vectorT_int_t_t,
  &_swigt__p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,
  &_swigt__p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t,
  &_swigt__p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t,
  &_swigt__p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator,
  &_swigt__p_std__vectorT_unsigned_int_t,
  &_swigt__p_std__vectorT_unsigned_long_t,
  &_swigt__p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t,
  &_swigt__p_swig__SwigPyIterator,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
  &_swigt__p_value_type,
  &_swigt__p_void,
  &_swigt__p_word_vector,
};

static swig_cast_info _swigc__p_ConstRefIter[] = {  {&_swigt__p_ConstRefIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Iter[] = {  {&_swigt__p_Iter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Mapping[] = {  {&_swigt__p_Mapping, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Mappings[] = {  {&_swigt__p_Mappings, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpMap[] = {  {&_swigt__p_OpMap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__AliasData[] = {  {&_swigt__p_OpenBabel__AliasData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__CharPtrLess[] = {  {&_swigt__p_OpenBabel__CharPtrLess, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__FastSearch[] = {  {&_swigt__p_OpenBabel__FastSearch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__FastSearchIndexer[] = {  {&_swigt__p_OpenBabel__FastSearchIndexer, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__FptIndex[] = {  {&_swigt__p_OpenBabel__FptIndex, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__FptIndexHeader[] = {  {&_swigt__p_OpenBabel__FptIndexHeader, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__LineSearchType[] = {  {&_swigt__p_OpenBabel__LineSearchType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAlign[] = {  {&_swigt__p_OpenBabel__OBAlign, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAngle[] = {  {&_swigt__p_OpenBabel__OBAngle, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAngleData[] = {  {&_swigt__p_OpenBabel__OBAngleData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAromaticTyper[] = {  {&_swigt__p_OpenBabel__OBAromaticTyper, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtom[] = {  {&_swigt__p_OpenBabel__OBAtom, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomAtomIter[] = {  {&_swigt__p_OpenBabel__OBAtomAtomIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomAtomIterAdaptor[] = {  {&_swigt__p_OpenBabel__OBAtomAtomIterAdaptor, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomBondIter[] = {  {&_swigt__p_OpenBabel__OBAtomBondIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomHOF[] = {  {&_swigt__p_OpenBabel__OBAtomHOF, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomTyper[] = {  {&_swigt__p_OpenBabel__OBAtomTyper, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBAtomicHeatOfFormationTable[] = {  {&_swigt__p_OpenBabel__OBAtomicHeatOfFormationTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBBase[] = {  {&_swigt__p_OpenBabel__OBBase, 0, 0, 0},  {&_swigt__p_OpenBabel__OBAtom, _p_OpenBabel__OBAtomTo_p_OpenBabel__OBBase, 0, 0},  {&_swigt__p_OpenBabel__OBBond, _p_OpenBabel__OBBondTo_p_OpenBabel__OBBase, 0, 0},  {&_swigt__p_OpenBabel__OBMol, _p_OpenBabel__OBMolTo_p_OpenBabel__OBBase, 0, 0},  {&_swigt__p_OpenBabel__OBReaction, _p_OpenBabel__OBReactionTo_p_OpenBabel__OBBase, 0, 0},  {&_swigt__p_OpenBabel__OBResidue, _p_OpenBabel__OBResidueTo_p_OpenBabel__OBBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBBitVec[] = {  {&_swigt__p_OpenBabel__OBBitVec, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBBond[] = {  {&_swigt__p_OpenBabel__OBBond, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBBuilder[] = {  {&_swigt__p_OpenBabel__OBBuilder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBChainsParser[] = {  {&_swigt__p_OpenBabel__OBChainsParser, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBChargeModel[] = {  {&_swigt__p_OpenBabel__OBChargeModel, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBChemTsfm[] = {  {&_swigt__p_OpenBabel__OBChemTsfm, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBCisTransConfig[] = {  {&_swigt__p_OpenBabel__OBCisTransConfig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBCisTransStereo[] = {  {&_swigt__p_OpenBabel__OBCisTransStereo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBCommentData[] = {  {&_swigt__p_OpenBabel__OBCommentData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConformerData[] = {  {&_swigt__p_OpenBabel__OBConformerData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConformerFilter[] = {  {&_swigt__p_OpenBabel__OBConformerFilter, 0, 0, 0},  {&_swigt__p_OpenBabel__OBConformerFilters, _p_OpenBabel__OBConformerFiltersTo_p_OpenBabel__OBConformerFilter, 0, 0},  {&_swigt__p_OpenBabel__OBStericConformerFilter, _p_OpenBabel__OBStericConformerFilterTo_p_OpenBabel__OBConformerFilter, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConformerFilters[] = {  {&_swigt__p_OpenBabel__OBConformerFilters, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConformerScore[] = {  {&_swigt__p_OpenBabel__OBConformerScore, 0, 0, 0},  {&_swigt__p_OpenBabel__OBEnergyConformerScore, _p_OpenBabel__OBEnergyConformerScoreTo_p_OpenBabel__OBConformerScore, 0, 0},  {&_swigt__p_OpenBabel__OBMinimizingEnergyConformerScore, _p_OpenBabel__OBMinimizingEnergyConformerScoreTo_p_OpenBabel__OBConformerScore, 0, 0},  {&_swigt__p_OpenBabel__OBMinimizingRMSDConformerScore, _p_OpenBabel__OBMinimizingRMSDConformerScoreTo_p_OpenBabel__OBConformerScore, 0, 0},  {&_swigt__p_OpenBabel__OBRMSDConformerScore, _p_OpenBabel__OBRMSDConformerScoreTo_p_OpenBabel__OBConformerScore, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConformerSearch[] = {  {&_swigt__p_OpenBabel__OBConformerSearch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBConversion[] = {  {&_swigt__p_OpenBabel__OBConversion, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBDOSData[] = {  {&_swigt__p_OpenBabel__OBDOSData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBDescriptor[] = {  {&_swigt__p_OpenBabel__OBDescriptor, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBElectronicTransitionData[] = {  {&_swigt__p_OpenBabel__OBElectronicTransitionData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBEnergyConformerScore[] = {  {&_swigt__p_OpenBabel__OBEnergyConformerScore, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBError[] = {  {&_swigt__p_OpenBabel__OBError, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBExternalBond[] = {  {&_swigt__p_OpenBabel__OBExternalBond, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBExternalBondData[] = {  {&_swigt__p_OpenBabel__OBExternalBondData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFCalculation2[] = {  {&_swigt__p_OpenBabel__OBFFCalculation2, 0, 0, 0},  {&_swigt__p_OpenBabel__OBFFCalculation3, _p_OpenBabel__OBFFCalculation3To_p_OpenBabel__OBFFCalculation2, 0, 0},  {&_swigt__p_OpenBabel__OBFFCalculation4, _p_OpenBabel__OBFFCalculation4To_p_OpenBabel__OBFFCalculation2, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFCalculation3[] = {  {&_swigt__p_OpenBabel__OBFFCalculation3, 0, 0, 0},  {&_swigt__p_OpenBabel__OBFFCalculation4, _p_OpenBabel__OBFFCalculation4To_p_OpenBabel__OBFFCalculation3, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFCalculation4[] = {  {&_swigt__p_OpenBabel__OBFFCalculation4, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFConstraint[] = {  {&_swigt__p_OpenBabel__OBFFConstraint, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFConstraints[] = {  {&_swigt__p_OpenBabel__OBFFConstraints, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFFParameter[] = {  {&_swigt__p_OpenBabel__OBFFParameter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFingerprint[] = {  {&_swigt__p_OpenBabel__OBFingerprint, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBForceField[] = {  {&_swigt__p_OpenBabel__OBForceField, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFormat[] = {  {&_swigt__p_OpenBabel__OBFormat, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFreeGrid[] = {  {&_swigt__p_OpenBabel__OBFreeGrid, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBFreeGridPoint[] = {  {&_swigt__p_OpenBabel__OBFreeGridPoint, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBGenericData[] = {  {&_swigt__p_OpenBabel__OBGenericData, 0, 0, 0},  {&_swigt__p_OpenBabel__AliasData, _p_OpenBabel__AliasDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBAngleData, _p_OpenBabel__OBAngleDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBCisTransStereo, _p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBCommentData, _p_OpenBabel__OBCommentDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBConformerData, _p_OpenBabel__OBConformerDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBDOSData, _p_OpenBabel__OBDOSDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBElectronicTransitionData, _p_OpenBabel__OBElectronicTransitionDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBExternalBondData, _p_OpenBabel__OBExternalBondDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBFreeGrid, _p_OpenBabel__OBFreeGridTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBGridData, _p_OpenBabel__OBGridDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBMatrixData, _p_OpenBabel__OBMatrixDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBOrbitalData, _p_OpenBabel__OBOrbitalDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBPairData, _p_OpenBabel__OBPairDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBPairTemplateT_float_t, _p_OpenBabel__OBPairTemplateT_float_tTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBPairTemplateT_int_t, _p_OpenBabel__OBPairTemplateT_int_tTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBPcharge, _p_OpenBabel__OBPchargeTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBRingData, _p_OpenBabel__OBRingDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBRotamerList, _p_OpenBabel__OBRotamerListTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBRotationData, _p_OpenBabel__OBRotationDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBSerialNums, _p_OpenBabel__OBSerialNumsTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBSetData, _p_OpenBabel__OBSetDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBSquarePlanarStereo, _p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBStereoBase, _p_OpenBabel__OBStereoBaseTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBSymmetryData, _p_OpenBabel__OBSymmetryDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBTetraNonPlanarStereo, _p_OpenBabel__OBTetraNonPlanarStereoTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBTetraPlanarStereo, _p_OpenBabel__OBTetraPlanarStereoTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBTetrahedralStereo, _p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBTorsionData, _p_OpenBabel__OBTorsionDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBUnitCell, _p_OpenBabel__OBUnitCellTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBVectorData, _p_OpenBabel__OBVectorDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBVibrationData, _p_OpenBabel__OBVibrationDataTo_p_OpenBabel__OBGenericData, 0, 0},  {&_swigt__p_OpenBabel__OBVirtualBond, _p_OpenBabel__OBVirtualBondTo_p_OpenBabel__OBGenericData, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBGlobalDataBase[] = {  {&_swigt__p_OpenBabel__OBGlobalDataBase, 0, 0, 0},  {&_swigt__p_OpenBabel__OBAtomTyper, _p_OpenBabel__OBAtomTyperTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBAtomicHeatOfFormationTable, _p_OpenBabel__OBAtomicHeatOfFormationTableTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBPhModel, _p_OpenBabel__OBPhModelTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBResidueData, _p_OpenBabel__OBResidueDataTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBRingTyper, _p_OpenBabel__OBRingTyperTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBRotorRules, _p_OpenBabel__OBRotorRulesTo_p_OpenBabel__OBGlobalDataBase, 0, 0},  {&_swigt__p_OpenBabel__OBTypeTable, _p_OpenBabel__OBTypeTableTo_p_OpenBabel__OBGlobalDataBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBGraphSym[] = {  {&_swigt__p_OpenBabel__OBGraphSym, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBGridData[] = {  {&_swigt__p_OpenBabel__OBGridData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBInternalCoord[] = {  {&_swigt__p_OpenBabel__OBInternalCoord, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBIsomorphismMapper[] = {  {&_swigt__p_OpenBabel__OBIsomorphismMapper, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBIsomorphismMapper__Functor[] = {  {&_swigt__p_OpenBabel__OBIsomorphismMapper__Functor, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMatrixData[] = {  {&_swigt__p_OpenBabel__OBMatrixData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMessageHandler[] = {  {&_swigt__p_OpenBabel__OBMessageHandler, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMinimizingEnergyConformerScore[] = {  {&_swigt__p_OpenBabel__OBMinimizingEnergyConformerScore, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMinimizingRMSDConformerScore[] = {  {&_swigt__p_OpenBabel__OBMinimizingRMSDConformerScore, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMol[] = {  {&_swigt__p_OpenBabel__OBMol, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolAngleIter[] = {  {&_swigt__p_OpenBabel__OBMolAngleIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolAtomBFSIter[] = {  {&_swigt__p_OpenBabel__OBMolAtomBFSIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolAtomDFSIter[] = {  {&_swigt__p_OpenBabel__OBMolAtomDFSIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolAtomIter[] = {  {&_swigt__p_OpenBabel__OBMolAtomIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolBondBFSIter[] = {  {&_swigt__p_OpenBabel__OBMolBondBFSIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolBondIter[] = {  {&_swigt__p_OpenBabel__OBMolBondIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolPairIter[] = {  {&_swigt__p_OpenBabel__OBMolPairIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolRingIter[] = {  {&_swigt__p_OpenBabel__OBMolRingIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBMolTorsionIter[] = {  {&_swigt__p_OpenBabel__OBMolTorsionIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBNasaThermoData[] = {  {&_swigt__p_OpenBabel__OBNasaThermoData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBOp[] = {  {&_swigt__p_OpenBabel__OBOp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBOrbital[] = {  {&_swigt__p_OpenBabel__OBOrbital, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBOrbitalData[] = {  {&_swigt__p_OpenBabel__OBOrbitalData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPairData[] = {  {&_swigt__p_OpenBabel__OBPairData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPairTemplateT_bool_t[] = {  {&_swigt__p_OpenBabel__OBPairTemplateT_bool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPairTemplateT_double_t[] = {  {&_swigt__p_OpenBabel__OBPairTemplateT_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPairTemplateT_float_t[] = {  {&_swigt__p_OpenBabel__OBPairTemplateT_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPairTemplateT_int_t[] = {  {&_swigt__p_OpenBabel__OBPairTemplateT_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPcharge[] = {  {&_swigt__p_OpenBabel__OBPcharge, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPhModel[] = {  {&_swigt__p_OpenBabel__OBPhModel, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBPlugin[] = {  {&_swigt__p_OpenBabel__OBPlugin, 0, 0, 0},  {&_swigt__p_OpenBabel__OBChargeModel, _p_OpenBabel__OBChargeModelTo_p_OpenBabel__OBPlugin, 0, 0},  {&_swigt__p_OpenBabel__OBDescriptor, _p_OpenBabel__OBDescriptorTo_p_OpenBabel__OBPlugin, 0, 0},  {&_swigt__p_OpenBabel__OBFingerprint, _p_OpenBabel__OBFingerprintTo_p_OpenBabel__OBPlugin, 0, 0},  {&_swigt__p_OpenBabel__OBForceField, _p_OpenBabel__OBForceFieldTo_p_OpenBabel__OBPlugin, 0, 0},  {&_swigt__p_OpenBabel__OBFormat, _p_OpenBabel__OBFormatTo_p_OpenBabel__OBPlugin, 0, 0},  {&_swigt__p_OpenBabel__OBOp, _p_OpenBabel__OBOpTo_p_OpenBabel__OBPlugin, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBQuery[] = {  {&_swigt__p_OpenBabel__OBQuery, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBQueryAtom[] = {  {&_swigt__p_OpenBabel__OBQueryAtom, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBQueryBond[] = {  {&_swigt__p_OpenBabel__OBQueryBond, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRMSDConformerScore[] = {  {&_swigt__p_OpenBabel__OBRMSDConformerScore, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRTree[] = {  {&_swigt__p_OpenBabel__OBRTree, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t[] = {  {&_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t[] = {  {&_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t[] = {  {&_swigt__p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRateData[] = {  {&_swigt__p_OpenBabel__OBRateData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBReaction[] = {  {&_swigt__p_OpenBabel__OBReaction, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBReactionFacade[] = {  {&_swigt__p_OpenBabel__OBReactionFacade, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBResidue[] = {  {&_swigt__p_OpenBabel__OBResidue, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBResidueAtomIter[] = {  {&_swigt__p_OpenBabel__OBResidueAtomIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBResidueData[] = {  {&_swigt__p_OpenBabel__OBResidueData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBResidueIter[] = {  {&_swigt__p_OpenBabel__OBResidueIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRing[] = {  {&_swigt__p_OpenBabel__OBRing, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRingData[] = {  {&_swigt__p_OpenBabel__OBRingData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRingSearch[] = {  {&_swigt__p_OpenBabel__OBRingSearch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRingTyper[] = {  {&_swigt__p_OpenBabel__OBRingTyper, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotamerList[] = {  {&_swigt__p_OpenBabel__OBRotamerList, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotationData[] = {  {&_swigt__p_OpenBabel__OBRotationData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotor[] = {  {&_swigt__p_OpenBabel__OBRotor, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotorKeys[] = {  {&_swigt__p_OpenBabel__OBRotorKeys, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotorList[] = {  {&_swigt__p_OpenBabel__OBRotorList, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotorRule[] = {  {&_swigt__p_OpenBabel__OBRotorRule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBRotorRules[] = {  {&_swigt__p_OpenBabel__OBRotorRules, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSSMatch[] = {  {&_swigt__p_OpenBabel__OBSSMatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSerialNums[] = {  {&_swigt__p_OpenBabel__OBSerialNums, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSetData[] = {  {&_swigt__p_OpenBabel__OBSetData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSmartsMatcher[] = {  {&_swigt__p_OpenBabel__OBSmartsMatcher, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSmartsPattern[] = {  {&_swigt__p_OpenBabel__OBSmartsPattern, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSpectrophore[] = {  {&_swigt__p_OpenBabel__OBSpectrophore, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSqrtTbl[] = {  {&_swigt__p_OpenBabel__OBSqrtTbl, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSquarePlanarConfig[] = {  {&_swigt__p_OpenBabel__OBSquarePlanarConfig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSquarePlanarStereo[] = {  {&_swigt__p_OpenBabel__OBSquarePlanarStereo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStereo[] = {  {&_swigt__p_OpenBabel__OBStereo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStereoBase[] = {  {&_swigt__p_OpenBabel__OBStereoBase, 0, 0, 0},  {&_swigt__p_OpenBabel__OBCisTransStereo, _p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBStereoBase, 0, 0},  {&_swigt__p_OpenBabel__OBSquarePlanarStereo, _p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBStereoBase, 0, 0},  {&_swigt__p_OpenBabel__OBTetraNonPlanarStereo, _p_OpenBabel__OBTetraNonPlanarStereoTo_p_OpenBabel__OBStereoBase, 0, 0},  {&_swigt__p_OpenBabel__OBTetraPlanarStereo, _p_OpenBabel__OBTetraPlanarStereoTo_p_OpenBabel__OBStereoBase, 0, 0},  {&_swigt__p_OpenBabel__OBTetrahedralStereo, _p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBStereoBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStereoFacade[] = {  {&_swigt__p_OpenBabel__OBStereoFacade, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStereoUnit[] = {  {&_swigt__p_OpenBabel__OBStereoUnit, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStericConformerFilter[] = {  {&_swigt__p_OpenBabel__OBStericConformerFilter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBStopwatch[] = {  {&_swigt__p_OpenBabel__OBStopwatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBSymmetryData[] = {  {&_swigt__p_OpenBabel__OBSymmetryData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTetraNonPlanarStereo[] = {  {&_swigt__p_OpenBabel__OBTetraNonPlanarStereo, 0, 0, 0},  {&_swigt__p_OpenBabel__OBTetrahedralStereo, _p_OpenBabel__OBTetrahedralStereoTo_p_OpenBabel__OBTetraNonPlanarStereo, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTetraPlanarStereo[] = {  {&_swigt__p_OpenBabel__OBTetraPlanarStereo, 0, 0, 0},  {&_swigt__p_OpenBabel__OBCisTransStereo, _p_OpenBabel__OBCisTransStereoTo_p_OpenBabel__OBTetraPlanarStereo, 0, 0},  {&_swigt__p_OpenBabel__OBSquarePlanarStereo, _p_OpenBabel__OBSquarePlanarStereoTo_p_OpenBabel__OBTetraPlanarStereo, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTetrahedralConfig[] = {  {&_swigt__p_OpenBabel__OBTetrahedralConfig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTetrahedralStereo[] = {  {&_swigt__p_OpenBabel__OBTetrahedralStereo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTorsion[] = {  {&_swigt__p_OpenBabel__OBTorsion, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTorsionData[] = {  {&_swigt__p_OpenBabel__OBTorsionData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBTypeTable[] = {  {&_swigt__p_OpenBabel__OBTypeTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBUnitCell[] = {  {&_swigt__p_OpenBabel__OBUnitCell, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBVectorData[] = {  {&_swigt__p_OpenBabel__OBVectorData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBVibrationData[] = {  {&_swigt__p_OpenBabel__OBVibrationData, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__OBVirtualBond[] = {  {&_swigt__p_OpenBabel__OBVirtualBond, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__Pattern[] = {  {&_swigt__p_OpenBabel__Pattern, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__SpaceGroup[] = {  {&_swigt__p_OpenBabel__SpaceGroup, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__Template[] = {  {&_swigt__p_OpenBabel__Template, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__matrix3x3[] = {  {&_swigt__p_OpenBabel__matrix3x3, 0, 0, 0},  {&_swigt__p_OpenBabel__transform3d, _p_OpenBabel__transform3dTo_p_OpenBabel__matrix3x3, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__obLogBuf[] = {  {&_swigt__p_OpenBabel__obLogBuf, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t[] = {  {&_swigt__p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__rotor_digit[] = {  {&_swigt__p_OpenBabel__rotor_digit, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__transform3d[] = {  {&_swigt__p_OpenBabel__transform3d, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t[] = {  {&_swigt__p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_OpenBabel__vector3[] = {  {&_swigt__p_OpenBabel__vector3, 0, 0, 0},  {&_swigt__p_OpenBabel__transform3d, _p_OpenBabel__transform3dTo_p_OpenBabel__vector3, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_PluginIterator[] = {  {&_swigt__p_PluginIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_PluginMapType[] = {  {&_swigt__p_PluginMapType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Ref[] = {  {&_swigt__p_Ref, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_RefIter[] = {  {&_swigt__p_RefIter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Refs[] = {  {&_swigt__p_Refs, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Unit[] = {  {&_swigt__p_Unit, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_a_3__doubleArray[] = {{&_swigt__p_a_3__doubleArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_a_3__double[] = {  {&_swigt__p_a_3__double, 0, 0, 0},  {&_swigt__p_a_3__doubleArray, _p_a_3__doubleArrayTo_p_a_3__double, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_allocator_type[] = {  {&_swigt__p_allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_bool[] = {  {&_swigt__p_bool, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_difference_type[] = {  {&_swigt__p_difference_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},  {&_swigt__p_doubleArray, _p_doubleArrayTo_p_double, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_doubleArray[] = {  {&_swigt__p_doubleArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_first_type[] = {  {&_swigt__p_first_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_iterator_category[] = {  {&_swigt__p_iterator_category, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBChargeModel[] = {  {&_swigt__p_p_OpenBabel__OBChargeModel, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBDescriptor[] = {  {&_swigt__p_p_OpenBabel__OBDescriptor, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBFingerprint[] = {  {&_swigt__p_p_OpenBabel__OBFingerprint, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBForceField[] = {  {&_swigt__p_p_OpenBabel__OBForceField, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBFormat[] = {  {&_swigt__p_p_OpenBabel__OBFormat, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_OpenBabel__OBOp[] = {  {&_swigt__p_p_OpenBabel__OBOp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_PyObject[] = {  {&_swigt__p_p_PyObject, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_doubleArray[] = {{&_swigt__p_p_doubleArray, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_double[] = {  {&_swigt__p_p_double, 0, 0, 0},  {&_swigt__p_p_doubleArray, _p_p_doubleArrayTo_p_p_double, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_int[] = {  {&_swigt__p_p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_second_type[] = {  {&_swigt__p_second_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_size_type[] = {  {&_swigt__p_size_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t[] = {  {&_swigt__p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__dequeT_int_t[] = {  {&_swigt__p_std__dequeT_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__invalid_argument[] = {  {&_swigt__p_std__invalid_argument, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__istream[] = {  {&_swigt__p_std__istream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator[] = {  {&_swigt__p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t[] = {  {&_swigt__p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t[] = {  {&_swigt__p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t[] = {  {&_swigt__p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t[] = {  {&_swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator[] = {  {&_swigt__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t[] = {  {&_swigt__p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t[] = {  {&_swigt__p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t[] = {  {&_swigt__p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__multimapT_double_unsigned_long_t[] = {  {&_swigt__p_std__multimapT_double_unsigned_long_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__ostream[] = {  {&_swigt__p_std__ostream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t[] = {  {&_swigt__p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_int_int_t[] = {  {&_swigt__p_std__pairT_int_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_std__string_std__string_t[] = {  {&_swigt__p_std__pairT_std__string_std__string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_unsigned_int_unsigned_int_t[] = {  {&_swigt__p_std__pairT_unsigned_int_unsigned_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setT_OpenBabel__OBBond_p_t[] = {  {&_swigt__p_std__setT_OpenBabel__OBBond_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__shared_ptrT_OpenBabel__OBMol_t[] = {  {&_swigt__p_std__shared_ptrT_OpenBabel__OBMol_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__streampos[] = {  {&_swigt__p_std__streampos, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__string[] = {  {&_swigt__p_std__string, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__stringbuf[] = {  {&_swigt__p_std__stringbuf, 0, 0, 0},  {&_swigt__p_OpenBabel__obLogBuf, _p_OpenBabel__obLogBufTo_p_std__stringbuf, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__type_info[] = {  {&_swigt__p_std__type_info, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBBond_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBBond_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBGenericData_p_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBInternalCoord_p_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBMol_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBMol_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBResidue_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBResidue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBRing_p_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBRing_p_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBRing_p_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBRing_p_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBRing_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBRing_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_OpenBabel__vector3_t[] = {  {&_swigt__p_std__vectorT_OpenBabel__vector3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_double_p_std__allocatorT_double_p_t_t[] = {  {&_swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_double_t[] = {  {&_swigt__p_std__vectorT_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_int_t[] = {  {&_swigt__p_std__vectorT_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t[] = {  {&_swigt__p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t[] = {  {&_swigt__p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t[] = {  {&_swigt__p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__string_t[] = {  {&_swigt__p_std__vectorT_std__string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t[] = {  {&_swigt__p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t[] = {  {&_swigt__p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__vectorT_int_t_t[] = {  {&_swigt__p_std__vectorT_std__vectorT_int_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t[] = {  {&_swigt__p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t[] = {  {&_swigt__p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t[] = {  {&_swigt__p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator[] = {  {&_swigt__p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_unsigned_int_t[] = {  {&_swigt__p_std__vectorT_unsigned_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_unsigned_long_t[] = {  {&_swigt__p_std__vectorT_unsigned_long_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t[] = {  {&_swigt__p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_swig__SwigPyIterator[] = {  {&_swigt__p_swig__SwigPyIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_type[] = {  {&_swigt__p_value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_word_vector[] = {  {&_swigt__p_word_vector, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_ConstRefIter,
  _swigc__p_Iter,
  _swigc__p_Mapping,
  _swigc__p_Mappings,
  _swigc__p_OpMap,
  _swigc__p_OpenBabel__AliasData,
  _swigc__p_OpenBabel__CharPtrLess,
  _swigc__p_OpenBabel__FastSearch,
  _swigc__p_OpenBabel__FastSearchIndexer,
  _swigc__p_OpenBabel__FptIndex,
  _swigc__p_OpenBabel__FptIndexHeader,
  _swigc__p_OpenBabel__LineSearchType,
  _swigc__p_OpenBabel__OBAlign,
  _swigc__p_OpenBabel__OBAngle,
  _swigc__p_OpenBabel__OBAngleData,
  _swigc__p_OpenBabel__OBAromaticTyper,
  _swigc__p_OpenBabel__OBAtom,
  _swigc__p_OpenBabel__OBAtomAtomIter,
  _swigc__p_OpenBabel__OBAtomAtomIterAdaptor,
  _swigc__p_OpenBabel__OBAtomBondIter,
  _swigc__p_OpenBabel__OBAtomHOF,
  _swigc__p_OpenBabel__OBAtomTyper,
  _swigc__p_OpenBabel__OBAtomicHeatOfFormationTable,
  _swigc__p_OpenBabel__OBBase,
  _swigc__p_OpenBabel__OBBitVec,
  _swigc__p_OpenBabel__OBBond,
  _swigc__p_OpenBabel__OBBuilder,
  _swigc__p_OpenBabel__OBChainsParser,
  _swigc__p_OpenBabel__OBChargeModel,
  _swigc__p_OpenBabel__OBChemTsfm,
  _swigc__p_OpenBabel__OBCisTransConfig,
  _swigc__p_OpenBabel__OBCisTransStereo,
  _swigc__p_OpenBabel__OBCommentData,
  _swigc__p_OpenBabel__OBConformerData,
  _swigc__p_OpenBabel__OBConformerFilter,
  _swigc__p_OpenBabel__OBConformerFilters,
  _swigc__p_OpenBabel__OBConformerScore,
  _swigc__p_OpenBabel__OBConformerSearch,
  _swigc__p_OpenBabel__OBConversion,
  _swigc__p_OpenBabel__OBDOSData,
  _swigc__p_OpenBabel__OBDescriptor,
  _swigc__p_OpenBabel__OBElectronicTransitionData,
  _swigc__p_OpenBabel__OBEnergyConformerScore,
  _swigc__p_OpenBabel__OBError,
  _swigc__p_OpenBabel__OBExternalBond,
  _swigc__p_OpenBabel__OBExternalBondData,
  _swigc__p_OpenBabel__OBFFCalculation2,
  _swigc__p_OpenBabel__OBFFCalculation3,
  _swigc__p_OpenBabel__OBFFCalculation4,
  _swigc__p_OpenBabel__OBFFConstraint,
  _swigc__p_OpenBabel__OBFFConstraints,
  _swigc__p_OpenBabel__OBFFParameter,
  _swigc__p_OpenBabel__OBFingerprint,
  _swigc__p_OpenBabel__OBForceField,
  _swigc__p_OpenBabel__OBFormat,
  _swigc__p_OpenBabel__OBFreeGrid,
  _swigc__p_OpenBabel__OBFreeGridPoint,
  _swigc__p_OpenBabel__OBGenericData,
  _swigc__p_OpenBabel__OBGlobalDataBase,
  _swigc__p_OpenBabel__OBGraphSym,
  _swigc__p_OpenBabel__OBGridData,
  _swigc__p_OpenBabel__OBInternalCoord,
  _swigc__p_OpenBabel__OBIsomorphismMapper,
  _swigc__p_OpenBabel__OBIsomorphismMapper__Functor,
  _swigc__p_OpenBabel__OBMatrixData,
  _swigc__p_OpenBabel__OBMessageHandler,
  _swigc__p_OpenBabel__OBMinimizingEnergyConformerScore,
  _swigc__p_OpenBabel__OBMinimizingRMSDConformerScore,
  _swigc__p_OpenBabel__OBMol,
  _swigc__p_OpenBabel__OBMolAngleIter,
  _swigc__p_OpenBabel__OBMolAtomBFSIter,
  _swigc__p_OpenBabel__OBMolAtomDFSIter,
  _swigc__p_OpenBabel__OBMolAtomIter,
  _swigc__p_OpenBabel__OBMolBondBFSIter,
  _swigc__p_OpenBabel__OBMolBondIter,
  _swigc__p_OpenBabel__OBMolPairIter,
  _swigc__p_OpenBabel__OBMolRingIter,
  _swigc__p_OpenBabel__OBMolTorsionIter,
  _swigc__p_OpenBabel__OBNasaThermoData,
  _swigc__p_OpenBabel__OBOp,
  _swigc__p_OpenBabel__OBOrbital,
  _swigc__p_OpenBabel__OBOrbitalData,
  _swigc__p_OpenBabel__OBPairData,
  _swigc__p_OpenBabel__OBPairTemplateT_bool_t,
  _swigc__p_OpenBabel__OBPairTemplateT_double_t,
  _swigc__p_OpenBabel__OBPairTemplateT_float_t,
  _swigc__p_OpenBabel__OBPairTemplateT_int_t,
  _swigc__p_OpenBabel__OBPcharge,
  _swigc__p_OpenBabel__OBPhModel,
  _swigc__p_OpenBabel__OBPlugin,
  _swigc__p_OpenBabel__OBQuery,
  _swigc__p_OpenBabel__OBQueryAtom,
  _swigc__p_OpenBabel__OBQueryBond,
  _swigc__p_OpenBabel__OBRMSDConformerScore,
  _swigc__p_OpenBabel__OBRTree,
  _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_OpenBabel__OBAtomAtomIterAdaptor_t,
  _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBAtom_p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator_t,
  _swigc__p_OpenBabel__OBRangeT_OpenBabel__OBBond_p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator_t,
  _swigc__p_OpenBabel__OBRateData,
  _swigc__p_OpenBabel__OBReaction,
  _swigc__p_OpenBabel__OBReactionFacade,
  _swigc__p_OpenBabel__OBResidue,
  _swigc__p_OpenBabel__OBResidueAtomIter,
  _swigc__p_OpenBabel__OBResidueData,
  _swigc__p_OpenBabel__OBResidueIter,
  _swigc__p_OpenBabel__OBRing,
  _swigc__p_OpenBabel__OBRingData,
  _swigc__p_OpenBabel__OBRingSearch,
  _swigc__p_OpenBabel__OBRingTyper,
  _swigc__p_OpenBabel__OBRotamerList,
  _swigc__p_OpenBabel__OBRotationData,
  _swigc__p_OpenBabel__OBRotor,
  _swigc__p_OpenBabel__OBRotorKeys,
  _swigc__p_OpenBabel__OBRotorList,
  _swigc__p_OpenBabel__OBRotorRule,
  _swigc__p_OpenBabel__OBRotorRules,
  _swigc__p_OpenBabel__OBSSMatch,
  _swigc__p_OpenBabel__OBSerialNums,
  _swigc__p_OpenBabel__OBSetData,
  _swigc__p_OpenBabel__OBSmartsMatcher,
  _swigc__p_OpenBabel__OBSmartsPattern,
  _swigc__p_OpenBabel__OBSpectrophore,
  _swigc__p_OpenBabel__OBSqrtTbl,
  _swigc__p_OpenBabel__OBSquarePlanarConfig,
  _swigc__p_OpenBabel__OBSquarePlanarStereo,
  _swigc__p_OpenBabel__OBStereo,
  _swigc__p_OpenBabel__OBStereoBase,
  _swigc__p_OpenBabel__OBStereoFacade,
  _swigc__p_OpenBabel__OBStereoUnit,
  _swigc__p_OpenBabel__OBStericConformerFilter,
  _swigc__p_OpenBabel__OBStopwatch,
  _swigc__p_OpenBabel__OBSymmetryData,
  _swigc__p_OpenBabel__OBTetraNonPlanarStereo,
  _swigc__p_OpenBabel__OBTetraPlanarStereo,
  _swigc__p_OpenBabel__OBTetrahedralConfig,
  _swigc__p_OpenBabel__OBTetrahedralStereo,
  _swigc__p_OpenBabel__OBTorsion,
  _swigc__p_OpenBabel__OBTorsionData,
  _swigc__p_OpenBabel__OBTypeTable,
  _swigc__p_OpenBabel__OBUnitCell,
  _swigc__p_OpenBabel__OBVectorData,
  _swigc__p_OpenBabel__OBVibrationData,
  _swigc__p_OpenBabel__OBVirtualBond,
  _swigc__p_OpenBabel__Pattern,
  _swigc__p_OpenBabel__SpaceGroup,
  _swigc__p_OpenBabel__Template,
  _swigc__p_OpenBabel__matrix3x3,
  _swigc__p_OpenBabel__obLogBuf,
  _swigc__p_OpenBabel__quadT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  _swigc__p_OpenBabel__rotor_digit,
  _swigc__p_OpenBabel__transform3d,
  _swigc__p_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  _swigc__p_OpenBabel__vector3,
  _swigc__p_PluginIterator,
  _swigc__p_PluginMapType,
  _swigc__p_Ref,
  _swigc__p_RefIter,
  _swigc__p_Refs,
  _swigc__p_Unit,
  _swigc__p_a_3__double,
  _swigc__p_a_3__doubleArray,
  _swigc__p_allocator_type,
  _swigc__p_bool,
  _swigc__p_char,
  _swigc__p_difference_type,
  _swigc__p_double,
  _swigc__p_doubleArray,
  _swigc__p_first_type,
  _swigc__p_int,
  _swigc__p_iterator_category,
  _swigc__p_p_OpenBabel__OBChargeModel,
  _swigc__p_p_OpenBabel__OBDescriptor,
  _swigc__p_p_OpenBabel__OBFingerprint,
  _swigc__p_p_OpenBabel__OBForceField,
  _swigc__p_p_OpenBabel__OBFormat,
  _swigc__p_p_OpenBabel__OBOp,
  _swigc__p_p_PyObject,
  _swigc__p_p_double,
  _swigc__p_p_doubleArray,
  _swigc__p_p_int,
  _swigc__p_second_type,
  _swigc__p_size_type,
  _swigc__p_std__allocatorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,
  _swigc__p_std__dequeT_int_t,
  _swigc__p_std__invalid_argument,
  _swigc__p_std__istream,
  _swigc__p_std__listT_OpenBabel__transform3d_p_std__allocatorT_OpenBabel__transform3d_p_t_t__const_iterator,
  _swigc__p_std__listT_OpenBabel__vector3_std__allocatorT_OpenBabel__vector3_t_t,
  _swigc__p_std__mapT_OpenBabel__OBBond_p_OpenBabel__OBStereo__BondDirection_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_OpenBabel__OBStereo__BondDirection_t_t_t,
  _swigc__p_std__mapT_OpenBabel__OBBond_p_unsigned_long_std__lessT_OpenBabel__OBBond_p_t_std__allocatorT_std__pairT_OpenBabel__OBBond_pconst_unsigned_long_t_t_t,
  _swigc__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t,
  _swigc__p_std__mapT_char_const_p_OpenBabel__OBPlugin_p_OpenBabel__CharPtrLess_std__allocatorT_std__pairT_char_const_pconst_OpenBabel__OBPlugin_p_t_t_t__const_iterator,
  _swigc__p_std__mapT_int_OpenBabel__OBAtom_p_std__lessT_int_t_std__allocatorT_std__pairT_int_const_OpenBabel__OBAtom_p_t_t_t,
  _swigc__p_std__mapT_std__string_std__string_std__lessT_std__string_t_std__allocatorT_std__pairT_std__string_const_std__string_t_t_t,
  _swigc__p_std__mapT_std__vectorT_int_t_double_std__lessT_std__vectorT_int_t_t_std__allocatorT_std__pairT_std__vectorT_int_t_const_double_t_t_t,
  _swigc__p_std__multimapT_double_unsigned_long_t,
  _swigc__p_std__ostream,
  _swigc__p_std__pairT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_t,
  _swigc__p_std__pairT_int_int_t,
  _swigc__p_std__pairT_std__string_std__string_t,
  _swigc__p_std__pairT_unsigned_int_unsigned_int_t,
  _swigc__p_std__setT_OpenBabel__OBBond_p_t,
  _swigc__p_std__shared_ptrT_OpenBabel__OBMol_t,
  _swigc__p_std__streampos,
  _swigc__p_std__string,
  _swigc__p_std__stringbuf,
  _swigc__p_std__type_info,
  _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__const_iterator,
  _swigc__p_std__vectorT_OpenBabel__OBAtom_p_std__allocatorT_OpenBabel__OBAtom_p_t_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBBase_p_std__allocatorT_OpenBabel__OBBase_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__const_iterator,
  _swigc__p_std__vectorT_OpenBabel__OBBond_p_std__allocatorT_OpenBabel__OBBond_p_t_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBBond_t,
  _swigc__p_std__vectorT_OpenBabel__OBCisTransStereo_p_std__allocatorT_OpenBabel__OBCisTransStereo_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBConformerFilter_p_std__allocatorT_OpenBabel__OBConformerFilter_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBExternalBond_std__allocatorT_OpenBabel__OBExternalBond_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBFreeGridPoint_p_std__allocatorT_OpenBabel__OBFreeGridPoint_p_t_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBGenericData_p_t,
  _swigc__p_std__vectorT_OpenBabel__OBGenericData_p_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBInternalCoord_p_t,
  _swigc__p_std__vectorT_OpenBabel__OBInternalCoord_p_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBMol_t,
  _swigc__p_std__vectorT_OpenBabel__OBOrbital_std__allocatorT_OpenBabel__OBOrbital_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBQueryAtom_p_std__allocatorT_OpenBabel__OBQueryAtom_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBQueryBond_p_std__allocatorT_OpenBabel__OBQueryBond_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBResidue_p_std__allocatorT_OpenBabel__OBResidue_p_t_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBResidue_t,
  _swigc__p_std__vectorT_OpenBabel__OBRing_p_t,
  _swigc__p_std__vectorT_OpenBabel__OBRing_p_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBRing_t,
  _swigc__p_std__vectorT_OpenBabel__OBRotor_p_std__allocatorT_OpenBabel__OBRotor_p_t_t__iterator,
  _swigc__p_std__vectorT_OpenBabel__OBSquarePlanarStereo_p_std__allocatorT_OpenBabel__OBSquarePlanarStereo_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBTetrahedralStereo_p_std__allocatorT_OpenBabel__OBTetrahedralStereo_p_t_t,
  _swigc__p_std__vectorT_OpenBabel__OBTorsion_std__allocatorT_OpenBabel__OBTorsion_t_t,
  _swigc__p_std__vectorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_std__allocatorT_OpenBabel__tripleT_OpenBabel__OBAtom_p_OpenBabel__OBAtom_p_double_t_t_t,
  _swigc__p_std__vectorT_OpenBabel__vector3_t,
  _swigc__p_std__vectorT_double_p_std__allocatorT_double_p_t_t,
  _swigc__p_std__vectorT_double_p_std__allocatorT_double_p_t_t__iterator,
  _swigc__p_std__vectorT_double_t,
  _swigc__p_std__vectorT_int_t,
  _swigc__p_std__vectorT_std__pairT_int_int_t_std__allocatorT_std__pairT_int_int_t_t_t,
  _swigc__p_std__vectorT_std__pairT_std__string_std__string_t_std__allocatorT_std__pairT_std__string_std__string_t_t_t,
  _swigc__p_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t,
  _swigc__p_std__vectorT_std__string_t,
  _swigc__p_std__vectorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_std__allocatorT_std__vectorT_OpenBabel__OBStereoUnit_std__allocatorT_OpenBabel__OBStereoUnit_t_t_t_t,
  _swigc__p_std__vectorT_std__vectorT_OpenBabel__vector3_t_t,
  _swigc__p_std__vectorT_std__vectorT_int_t_t,
  _swigc__p_std__vectorT_std__vectorT_std__pairT_unsigned_int_unsigned_int_t_t_t,
  _swigc__p_std__vectorT_std__vectorT_unsigned_int_t_std__allocatorT_std__vectorT_unsigned_int_t_t_t,
  _swigc__p_std__vectorT_uint32_t_std__allocatorT_uint32_t_t_t,
  _swigc__p_std__vectorT_unsigned_char_p_std__allocatorT_unsigned_char_p_t_t__iterator,
  _swigc__p_std__vectorT_unsigned_int_t,
  _swigc__p_std__vectorT_unsigned_long_t,
  _swigc__p_std__vectorT_unsigned_short_std__allocatorT_unsigned_short_t_t,
  _swigc__p_swig__SwigPyIterator,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
  _swigc__p_value_type,
  _swigc__p_void,
  _swigc__p_word_vector,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        SWIG_Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------
   * Patch %callback methods' docstrings to hold the callback ptrs
   * -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods, const swig_const_info *const_table, swig_type_info **types, swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        const swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif




#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * Method creation and docstring support functions
   * ----------------------------------------------------------------------------- */
  
  /* -----------------------------------------------------------------------------
   * Function to find the method definition with the correct docstring for the
   * proxy module as opposed to the low-level API
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name) {
    /* Find the function in the modified method table */
    size_t offset = 0;
    int found = 0;
    while (SwigMethods_proxydocs[offset].ml_meth != NULL) {
      if (strcmp(SwigMethods_proxydocs[offset].ml_name, name) == 0) {
        found = 1;
        break;
      }
      offset++;
    }
    /* Use the copy with the modified docstring if available */
    return found ? &SwigMethods_proxydocs[offset] : NULL;
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyInstanceMethod_New() used in Python 3
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
#if PY_VERSION_HEX >= 0x03000000
    return PyInstanceMethod_New(func);
#else
    return PyMethod_New(func, NULL, NULL);
#endif
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyStaticMethod_New()
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
    return PyStaticMethod_New(func);
  }
  
#ifdef __cplusplus
}
#endif




/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

SWIGINTERN int SWIG_mod_exec(PyObject *module);

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
#if PY_VERSION_HEX >= 0x03000000
  static PyModuleDef_Slot SwigSlots[] = {
    {
      Py_mod_exec, (void *)SWIG_mod_exec 
    },
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
    {
      Py_mod_gil, Py_MOD_GIL_NOT_USED 
    },
#endif
#endif
    {
      0, NULL 
    }
  };
  
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    0,
    SwigMethods,
    SwigSlots,
    NULL,
    NULL,
    NULL
  };
  
  return PyModuleDef_Init(&SWIG_module);
#else
  PyObject *m = Py_InitModule(SWIG_name, SwigMethods);
  if (m && SWIG_mod_exec(m) != 0) {
    Py_DECREF(m);
  }
#endif
}

SWIGINTERN int SWIG_mod_exec(PyObject *m) {
  PyObject *d, *md, *globals;
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType_Type();
  assert(metatype);
  
  SwigPyStaticVar_Type();
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_runtime_data_module();
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_Type();
  SwigVarLink_Type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_Type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
  PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
#endif
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
    return -1;
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  SWIG_Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  
  // thread safe initialization
  swig::container_owner_attribute();
  
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
    return -1;
  }
  PyDict_SetItemString(md, "cvar", globals);
  SWIG_Python_SetConstant(d, "M_PI",SWIG_From_double(static_cast< double >(3.14159265358979323846)));
  SWIG_Python_SetConstant(d, "RAD_TO_DEG",SWIG_From_double(static_cast< double >((180.0/3.14159265358979323846))));
  SWIG_Python_SetConstant(d, "DEG_TO_RAD",SWIG_From_double(static_cast< double >((3.14159265358979323846/180.0))));
  SWIG_Python_SetConstant(d, "SETWORD",SWIG_From_int(static_cast< int >(32)));
  SWIG_Python_SetConstant(d, "WORDROLL",SWIG_From_int(static_cast< int >(5)));
  SWIG_Python_SetConstant(d, "WORDMASK",SWIG_From_int(static_cast< int >(31)));
  SWIG_Python_SetConstant(d, "STARTWORDS",SWIG_From_int(static_cast< int >(10)));
  SWIG_Python_SetConstant(d, "UndefinedData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::UndefinedData)));
  SWIG_Python_SetConstant(d, "PairData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::PairData)));
  SWIG_Python_SetConstant(d, "EnergyData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::EnergyData)));
  SWIG_Python_SetConstant(d, "CommentData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CommentData)));
  SWIG_Python_SetConstant(d, "ConformerData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::ConformerData)));
  SWIG_Python_SetConstant(d, "ExternalBondData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::ExternalBondData)));
  SWIG_Python_SetConstant(d, "RotamerList",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::RotamerList)));
  SWIG_Python_SetConstant(d, "VirtualBondData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::VirtualBondData)));
  SWIG_Python_SetConstant(d, "RingData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::RingData)));
  SWIG_Python_SetConstant(d, "TorsionData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::TorsionData)));
  SWIG_Python_SetConstant(d, "AngleData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::AngleData)));
  SWIG_Python_SetConstant(d, "SerialNums",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::SerialNums)));
  SWIG_Python_SetConstant(d, "UnitCell",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::UnitCell)));
  SWIG_Python_SetConstant(d, "SpinData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::SpinData)));
  SWIG_Python_SetConstant(d, "ChargeData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::ChargeData)));
  SWIG_Python_SetConstant(d, "SymmetryData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::SymmetryData)));
  SWIG_Python_SetConstant(d, "OccupationData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::OccupationData)));
  SWIG_Python_SetConstant(d, "DensityData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::DensityData)));
  SWIG_Python_SetConstant(d, "ElectronicData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::ElectronicData)));
  SWIG_Python_SetConstant(d, "VibrationData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::VibrationData)));
  SWIG_Python_SetConstant(d, "RotationData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::RotationData)));
  SWIG_Python_SetConstant(d, "NuclearData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::NuclearData)));
  SWIG_Python_SetConstant(d, "SetData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::SetData)));
  SWIG_Python_SetConstant(d, "GridData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::GridData)));
  SWIG_Python_SetConstant(d, "VectorData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::VectorData)));
  SWIG_Python_SetConstant(d, "MatrixData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::MatrixData)));
  SWIG_Python_SetConstant(d, "StereoData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::StereoData)));
  SWIG_Python_SetConstant(d, "DOSData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::DOSData)));
  SWIG_Python_SetConstant(d, "ElectronicTransitionData",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::ElectronicTransitionData)));
  SWIG_Python_SetConstant(d, "CustomData0",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData0)));
  SWIG_Python_SetConstant(d, "CustomData1",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData1)));
  SWIG_Python_SetConstant(d, "CustomData2",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData2)));
  SWIG_Python_SetConstant(d, "CustomData3",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData3)));
  SWIG_Python_SetConstant(d, "CustomData4",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData4)));
  SWIG_Python_SetConstant(d, "CustomData5",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData5)));
  SWIG_Python_SetConstant(d, "CustomData6",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData6)));
  SWIG_Python_SetConstant(d, "CustomData7",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData7)));
  SWIG_Python_SetConstant(d, "CustomData8",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData8)));
  SWIG_Python_SetConstant(d, "CustomData9",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData9)));
  SWIG_Python_SetConstant(d, "CustomData10",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData10)));
  SWIG_Python_SetConstant(d, "CustomData11",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData11)));
  SWIG_Python_SetConstant(d, "CustomData12",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData12)));
  SWIG_Python_SetConstant(d, "CustomData13",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData13)));
  SWIG_Python_SetConstant(d, "CustomData14",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData14)));
  SWIG_Python_SetConstant(d, "CustomData15",SWIG_From_int(static_cast< int >(OpenBabel::OBGenericDataType::CustomData15)));
  SWIG_Python_SetConstant(d, "any",SWIG_From_int(static_cast< int >(OpenBabel::any)));
  SWIG_Python_SetConstant(d, "fileformatInput",SWIG_From_int(static_cast< int >(OpenBabel::fileformatInput)));
  SWIG_Python_SetConstant(d, "userInput",SWIG_From_int(static_cast< int >(OpenBabel::userInput)));
  SWIG_Python_SetConstant(d, "perceived",SWIG_From_int(static_cast< int >(OpenBabel::perceived)));
  SWIG_Python_SetConstant(d, "external",SWIG_From_int(static_cast< int >(OpenBabel::external)));
  SWIG_Python_SetConstant(d, "local",SWIG_From_int(static_cast< int >(OpenBabel::local)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Undefined",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Undefined)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Triclinic",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Triclinic)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Monoclinic",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Monoclinic)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Orthorhombic",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Orthorhombic)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Tetragonal",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Tetragonal)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Rhombohedral",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Rhombohedral)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Hexagonal",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Hexagonal)));
  SWIG_Python_SetConstant(d, "OBUnitCell_Cubic",SWIG_From_int(static_cast< int >(OpenBabel::OBUnitCell::Cubic)));
  SWIG_Python_SetConstant(d, "OBRotationData_UNKNOWN",SWIG_From_int(static_cast< int >(OpenBabel::OBRotationData::UNKNOWN)));
  SWIG_Python_SetConstant(d, "OBRotationData_ASYMMETRIC",SWIG_From_int(static_cast< int >(OpenBabel::OBRotationData::ASYMMETRIC)));
  SWIG_Python_SetConstant(d, "OBRotationData_SYMMETRIC",SWIG_From_int(static_cast< int >(OpenBabel::OBRotationData::SYMMETRIC)));
  SWIG_Python_SetConstant(d, "OBRotationData_LINEAR",SWIG_From_int(static_cast< int >(OpenBabel::OBRotationData::LINEAR)));
  SWIG_Python_SetConstant(d, "OBGridData_BOHR",SWIG_From_int(static_cast< int >(OpenBabel::OBGridData::BOHR)));
  SWIG_Python_SetConstant(d, "OBGridData_ANGSTROM",SWIG_From_int(static_cast< int >(OpenBabel::OBGridData::ANGSTROM)));
  SWIG_Python_SetConstant(d, "OBGridData_OTHER",SWIG_From_int(static_cast< int >(OpenBabel::OBGridData::OTHER)));
  SWIG_Python_SetConstant(d, "MaxMonoAtom",SWIG_From_int(static_cast< int >(20)));
  SWIG_Python_SetConstant(d, "MaxMonoBond",SWIG_From_int(static_cast< int >(20)));
  SWIG_Python_SetConstant(d, "obError",SWIG_From_int(static_cast< int >(OpenBabel::obError)));
  SWIG_Python_SetConstant(d, "obWarning",SWIG_From_int(static_cast< int >(OpenBabel::obWarning)));
  SWIG_Python_SetConstant(d, "obInfo",SWIG_From_int(static_cast< int >(OpenBabel::obInfo)));
  SWIG_Python_SetConstant(d, "obAuditMsg",SWIG_From_int(static_cast< int >(OpenBabel::obAuditMsg)));
  SWIG_Python_SetConstant(d, "obDebug",SWIG_From_int(static_cast< int >(OpenBabel::obDebug)));
  SWIG_Python_SetConstant(d, "always",SWIG_From_int(static_cast< int >(OpenBabel::always)));
  SWIG_Python_SetConstant(d, "onceOnly",SWIG_From_int(static_cast< int >(OpenBabel::onceOnly)));
  SWIG_Python_SetConstant(d, "NOTREADABLE",SWIG_From_int(static_cast< int >(0x01)));
  SWIG_Python_SetConstant(d, "READONEONLY",SWIG_From_int(static_cast< int >(0x02)));
  SWIG_Python_SetConstant(d, "READBINARY",SWIG_From_int(static_cast< int >(0x04)));
  SWIG_Python_SetConstant(d, "ZEROATOMSOK",SWIG_From_int(static_cast< int >(0x08)));
  SWIG_Python_SetConstant(d, "NOTWRITABLE",SWIG_From_int(static_cast< int >(0x10)));
  SWIG_Python_SetConstant(d, "WRITEONEONLY",SWIG_From_int(static_cast< int >(0x20)));
  SWIG_Python_SetConstant(d, "WRITEBINARY",SWIG_From_int(static_cast< int >(0x40)));
  SWIG_Python_SetConstant(d, "READXML",SWIG_From_int(static_cast< int >(0x80)));
  SWIG_Python_SetConstant(d, "DEPICTION2D",SWIG_From_int(static_cast< int >(0x100)));
  SWIG_Python_SetConstant(d, "DEFAULTFORMAT",SWIG_From_int(static_cast< int >(0x4000)));
  SWIG_Python_SetConstant(d, "OBConversion_INOPTIONS",SWIG_From_int(static_cast< int >(OpenBabel::OBConversion::INOPTIONS)));
  SWIG_Python_SetConstant(d, "OBConversion_OUTOPTIONS",SWIG_From_int(static_cast< int >(OpenBabel::OBConversion::OUTOPTIONS)));
  SWIG_Python_SetConstant(d, "OBConversion_GENOPTIONS",SWIG_From_int(static_cast< int >(OpenBabel::OBConversion::GENOPTIONS)));
  SWIG_Python_SetConstant(d, "OBConversion_ALL",SWIG_From_int(static_cast< int >(OpenBabel::OBConversion::ALL)));
  SWIG_Python_SetConstant(d, "MAXSETNO",SWIG_From_int(static_cast< int >(40)));
  SWIG_Python_SetConstant(d, "MAXELEM",SWIG_From_int(static_cast< int >(29)));
  SWIG_Python_SetConstant(d, "MAXRES",SWIG_From_int(static_cast< int >(54)));
  SWIG_Python_SetConstant(d, "AA_ALA",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "AA_GLY",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "AA_LEU",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "AA_SER",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "AA_VAL",SWIG_From_int(static_cast< int >((1 << 5))));
  SWIG_Python_SetConstant(d, "AA_THR",SWIG_From_int(static_cast< int >((1 << 6))));
  SWIG_Python_SetConstant(d, "AA_LYS",SWIG_From_int(static_cast< int >((1 << 7))));
  SWIG_Python_SetConstant(d, "AA_ASP",SWIG_From_int(static_cast< int >((1 << 8))));
  SWIG_Python_SetConstant(d, "AA_ILE",SWIG_From_int(static_cast< int >((1 << 9))));
  SWIG_Python_SetConstant(d, "AA_ASN",SWIG_From_int(static_cast< int >((1 << 10))));
  SWIG_Python_SetConstant(d, "AA_GLU",SWIG_From_int(static_cast< int >((1 << 11))));
  SWIG_Python_SetConstant(d, "AA_PRO",SWIG_From_int(static_cast< int >((1 << 12))));
  SWIG_Python_SetConstant(d, "AA_ARG",SWIG_From_int(static_cast< int >((1 << 13))));
  SWIG_Python_SetConstant(d, "AA_PHE",SWIG_From_int(static_cast< int >((1 << 14))));
  SWIG_Python_SetConstant(d, "AA_GLN",SWIG_From_int(static_cast< int >((1 << 15))));
  SWIG_Python_SetConstant(d, "AA_TYR",SWIG_From_int(static_cast< int >((1 << 16))));
  SWIG_Python_SetConstant(d, "AA_HIS",SWIG_From_int(static_cast< int >((1 << 17))));
  SWIG_Python_SetConstant(d, "AA_CYS",SWIG_From_int(static_cast< int >((1 << 18))));
  SWIG_Python_SetConstant(d, "AA_MET",SWIG_From_int(static_cast< int >((1 << 19))));
  SWIG_Python_SetConstant(d, "AA_TRP",SWIG_From_int(static_cast< int >((1 << 20))));
  SWIG_Python_SetConstant(d, "ACIDIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::ACIDIC)));
  SWIG_Python_SetConstant(d, "ACYCLIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::ACYCLIC)));
  SWIG_Python_SetConstant(d, "ALIPHATIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::ALIPHATIC)));
  SWIG_Python_SetConstant(d, "AROMATIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::AROMATIC)));
  SWIG_Python_SetConstant(d, "BASIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::BASIC)));
  SWIG_Python_SetConstant(d, "BURIED",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::BURIED)));
  SWIG_Python_SetConstant(d, "CHARGED",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::CHARGED)));
  SWIG_Python_SetConstant(d, "CYCLIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::CYCLIC)));
  SWIG_Python_SetConstant(d, "HYDROPHOBIC",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::HYDROPHOBIC)));
  SWIG_Python_SetConstant(d, "LARGE",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::LARGE)));
  SWIG_Python_SetConstant(d, "MEDIUM",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::MEDIUM)));
  SWIG_Python_SetConstant(d, "NEGATIVE",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::NEGATIVE)));
  SWIG_Python_SetConstant(d, "NEUTRAL",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::NEUTRAL)));
  SWIG_Python_SetConstant(d, "POLAR",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::POLAR)));
  SWIG_Python_SetConstant(d, "POSITIVE",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::POSITIVE)));
  SWIG_Python_SetConstant(d, "SMALL",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::SMALL)));
  SWIG_Python_SetConstant(d, "SURFACE",SWIG_From_int(static_cast< int >(OpenBabel::OBAminoAcidProperty::SURFACE)));
  SWIG_Python_SetConstant(d, "ALPHA_CARBON",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::ALPHA_CARBON)));
  SWIG_Python_SetConstant(d, "AMINO_BACKBONE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::AMINO_BACKBONE)));
  SWIG_Python_SetConstant(d, "BACKBONE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::BACKBONE)));
  SWIG_Python_SetConstant(d, "CYSTEINE_SULPHUR",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::CYSTEINE_SULPHUR)));
  SWIG_Python_SetConstant(d, "LIGAND",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::LIGAND)));
  SWIG_Python_SetConstant(d, "NUCLEIC_BACKBONE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::NUCLEIC_BACKBONE)));
  SWIG_Python_SetConstant(d, "SHAPELY_BACKBONE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::SHAPELY_BACKBONE)));
  SWIG_Python_SetConstant(d, "SHAPELY_SPECIAL",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::SHAPELY_SPECIAL)));
  SWIG_Python_SetConstant(d, "SIDECHAIN",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::SIDECHAIN)));
  SWIG_Python_SetConstant(d, "SUGAR_PHOSPHATE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueAtomProperty::SUGAR_PHOSPHATE)));
  SWIG_Python_SetConstant(d, "ALA",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ALA)));
  SWIG_Python_SetConstant(d, "GLY",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::GLY)));
  SWIG_Python_SetConstant(d, "LEU",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::LEU)));
  SWIG_Python_SetConstant(d, "SER",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::SER)));
  SWIG_Python_SetConstant(d, "VAL",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::VAL)));
  SWIG_Python_SetConstant(d, "THR",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::THR)));
  SWIG_Python_SetConstant(d, "LYS",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::LYS)));
  SWIG_Python_SetConstant(d, "ASP",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ASP)));
  SWIG_Python_SetConstant(d, "ILE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ILE)));
  SWIG_Python_SetConstant(d, "ASN",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ASN)));
  SWIG_Python_SetConstant(d, "GLU",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::GLU)));
  SWIG_Python_SetConstant(d, "PRO",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::PRO)));
  SWIG_Python_SetConstant(d, "ARG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ARG)));
  SWIG_Python_SetConstant(d, "PHE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::PHE)));
  SWIG_Python_SetConstant(d, "GLN",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::GLN)));
  SWIG_Python_SetConstant(d, "TYR",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::TYR)));
  SWIG_Python_SetConstant(d, "HIS",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::HIS)));
  SWIG_Python_SetConstant(d, "CYS",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::CYS)));
  SWIG_Python_SetConstant(d, "MET",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::MET)));
  SWIG_Python_SetConstant(d, "TRP",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::TRP)));
  SWIG_Python_SetConstant(d, "ASX",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ASX)));
  SWIG_Python_SetConstant(d, "GLX",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::GLX)));
  SWIG_Python_SetConstant(d, "PCA",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::PCA)));
  SWIG_Python_SetConstant(d, "HYP",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::HYP)));
  SWIG_Python_SetConstant(d, "resA",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::A)));
  SWIG_Python_SetConstant(d, "resC",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::C)));
  SWIG_Python_SetConstant(d, "resG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::G)));
  SWIG_Python_SetConstant(d, "resT",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::T)));
  SWIG_Python_SetConstant(d, "resU",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::U)));
  SWIG_Python_SetConstant(d, "UPLUS",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::UPLUS)));
  SWIG_Python_SetConstant(d, "resI",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::I)));
  SWIG_Python_SetConstant(d, "_1MA",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_1MA)));
  SWIG_Python_SetConstant(d, "_5MC",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_5MC)));
  SWIG_Python_SetConstant(d, "OMC",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::OMC)));
  SWIG_Python_SetConstant(d, "_1MG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_1MG)));
  SWIG_Python_SetConstant(d, "_2MG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_2MG)));
  SWIG_Python_SetConstant(d, "M2G",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::M2G)));
  SWIG_Python_SetConstant(d, "_7MG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_7MG)));
  SWIG_Python_SetConstant(d, "OMG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::OMG)));
  SWIG_Python_SetConstant(d, "YG",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::YG)));
  SWIG_Python_SetConstant(d, "H2U",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::H2U)));
  SWIG_Python_SetConstant(d, "_5MU",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::_5MU)));
  SWIG_Python_SetConstant(d, "PSU",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::PSU)));
  SWIG_Python_SetConstant(d, "UNK",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::UNK)));
  SWIG_Python_SetConstant(d, "ACE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::ACE)));
  SWIG_Python_SetConstant(d, "FOR",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::FOR)));
  SWIG_Python_SetConstant(d, "HOH",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::HOH)));
  SWIG_Python_SetConstant(d, "DOD",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::DOD)));
  SWIG_Python_SetConstant(d, "SO4",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::SO4)));
  SWIG_Python_SetConstant(d, "PO4",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::PO4)));
  SWIG_Python_SetConstant(d, "NAD",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::NAD)));
  SWIG_Python_SetConstant(d, "COA",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::COA)));
  SWIG_Python_SetConstant(d, "NAP",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::NAP)));
  SWIG_Python_SetConstant(d, "NDP",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueIndex::NDP)));
  SWIG_Python_SetConstant(d, "AMINO",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::AMINO)));
  SWIG_Python_SetConstant(d, "AMINO_NUCLEO",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::AMINO_NUCLEO)));
  SWIG_Python_SetConstant(d, "COENZYME",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::COENZYME)));
  SWIG_Python_SetConstant(d, "ION",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::ION)));
  SWIG_Python_SetConstant(d, "NUCLEO",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::NUCLEO)));
  SWIG_Python_SetConstant(d, "PROTEIN",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::PROTEIN)));
  SWIG_Python_SetConstant(d, "PURINE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::PURINE)));
  SWIG_Python_SetConstant(d, "PYRIMIDINE",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::PYRIMIDINE)));
  SWIG_Python_SetConstant(d, "SOLVENT",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::SOLVENT)));
  SWIG_Python_SetConstant(d, "WATER",SWIG_From_int(static_cast< int >(OpenBabel::OBResidueProperty::WATER)));
  SWIG_Python_SetConstant(d, "OB_4RING_ATOM",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "OB_3RING_ATOM",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "OB_AROMATIC_ATOM",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "OB_RING_ATOM",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "OB_DONOR_ATOM",SWIG_From_int(static_cast< int >((1 << 7))));
  SWIG_Python_SetConstant(d, "OB_ACCEPTOR_ATOM",SWIG_From_int(static_cast< int >((1 << 8))));
  SWIG_Python_SetConstant(d, "OBATOM_TYPE_LEN",SWIG_From_int(static_cast< int >(6)));
  SWIG_Python_SetConstant(d, "OB_AROMATIC_BOND",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "OB_WEDGE_BOND",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "OB_HASH_BOND",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "OB_RING_BOND",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "OB_CLOSURE_BOND",SWIG_From_int(static_cast< int >((1 << 10))));
  SWIG_Python_SetConstant(d, "OB_WEDGE_OR_HASH_BOND",SWIG_From_int(static_cast< int >((1 << 11))));
  SWIG_Python_SetConstant(d, "OBBond_Aromatic",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::Aromatic)));
  SWIG_Python_SetConstant(d, "OBBond_Ring",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::Ring)));
  SWIG_Python_SetConstant(d, "OBBond_Closure",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::Closure)));
  SWIG_Python_SetConstant(d, "OBBond_Wedge",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::Wedge)));
  SWIG_Python_SetConstant(d, "OBBond_Hash",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::Hash)));
  SWIG_Python_SetConstant(d, "OBBond_WedgeOrHash",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::WedgeOrHash)));
  SWIG_Python_SetConstant(d, "OBBond_CisOrTrans",SWIG_From_int(static_cast< int >(OpenBabel::OBBond::CisOrTrans)));
  SWIG_Python_SetConstant(d, "NO_REACTIONROLE",SWIG_From_int(static_cast< int >(OpenBabel::NO_REACTIONROLE)));
  SWIG_Python_SetConstant(d, "REACTANT",SWIG_From_int(static_cast< int >(OpenBabel::REACTANT)));
  SWIG_Python_SetConstant(d, "AGENT",SWIG_From_int(static_cast< int >(OpenBabel::AGENT)));
  SWIG_Python_SetConstant(d, "PRODUCT",SWIG_From_int(static_cast< int >(OpenBabel::PRODUCT)));
  SWIG_Python_SetConstant(d, "OB_SSSR_MOL",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "OB_RINGFLAGS_MOL",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "OB_AROMATIC_MOL",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "OB_ATOMTYPES_MOL",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "OB_CHIRALITY_MOL",SWIG_From_int(static_cast< int >((1 << 5))));
  SWIG_Python_SetConstant(d, "OB_PCHARGE_MOL",SWIG_From_int(static_cast< int >((1 << 6))));
  SWIG_Python_SetConstant(d, "OB_HYBRID_MOL",SWIG_From_int(static_cast< int >((1 << 8))));
  SWIG_Python_SetConstant(d, "OB_CLOSURE_MOL",SWIG_From_int(static_cast< int >((1 << 11))));
  SWIG_Python_SetConstant(d, "OB_H_ADDED_MOL",SWIG_From_int(static_cast< int >((1 << 12))));
  SWIG_Python_SetConstant(d, "OB_PH_CORRECTED_MOL",SWIG_From_int(static_cast< int >((1 << 13))));
  SWIG_Python_SetConstant(d, "OB_CHAINS_MOL",SWIG_From_int(static_cast< int >((1 << 15))));
  SWIG_Python_SetConstant(d, "OB_TCHARGE_MOL",SWIG_From_int(static_cast< int >((1 << 16))));
  SWIG_Python_SetConstant(d, "OB_TSPIN_MOL",SWIG_From_int(static_cast< int >((1 << 17))));
  SWIG_Python_SetConstant(d, "OB_RINGTYPES_MOL",SWIG_From_int(static_cast< int >((1 << 18))));
  SWIG_Python_SetConstant(d, "OB_PATTERN_STRUCTURE",SWIG_From_int(static_cast< int >((1 << 19))));
  SWIG_Python_SetConstant(d, "OB_LSSR_MOL",SWIG_From_int(static_cast< int >((1 << 20))));
  SWIG_Python_SetConstant(d, "OB_ATOMSPIN_MOL",SWIG_From_int(static_cast< int >((1 << 21))));
  SWIG_Python_SetConstant(d, "OB_REACTION_MOL",SWIG_From_int(static_cast< int >((1 << 22))));
  SWIG_Python_SetConstant(d, "OB_PERIODIC_MOL",SWIG_From_int(static_cast< int >((1 << 23))));
  SWIG_Python_SetConstant(d, "OB_CURRENT_CONFORMER",SWIG_From_int(static_cast< int >(-1)));
  SWIG_Python_SetConstant(d, "AllHydrogen",SWIG_From_int(static_cast< int >(OpenBabel::AllHydrogen)));
  SWIG_Python_SetConstant(d, "PolarHydrogen",SWIG_From_int(static_cast< int >(OpenBabel::PolarHydrogen)));
  SWIG_Python_SetConstant(d, "NonPolarHydrogen",SWIG_From_int(static_cast< int >(OpenBabel::NonPolarHydrogen)));
  SWIG_Python_SetConstant(d, "BUFF_SIZE",SWIG_From_int(static_cast< int >(32768)));
  SWIG_Python_SetConstant(d, "OBSmartsPattern_All",SWIG_From_int(static_cast< int >(OpenBabel::OBSmartsPattern::All)));
  SWIG_Python_SetConstant(d, "OBSmartsPattern_Single",SWIG_From_int(static_cast< int >(OpenBabel::OBSmartsPattern::Single)));
  SWIG_Python_SetConstant(d, "OBSmartsPattern_AllUnique",SWIG_From_int(static_cast< int >(OpenBabel::OBSmartsPattern::AllUnique)));
  SWIG_Python_SetConstant(d, "OBFingerprint_FPT_UNIQUEBITS",SWIG_From_int(static_cast< int >(OpenBabel::OBFingerprint::FPT_UNIQUEBITS)));
  SWIG_Python_SetConstant(d, "OBFingerprint_FPT_NOINFO",SWIG_From_int(static_cast< int >(OpenBabel::OBFingerprint::FPT_NOINFO)));
  SWIG_Python_SetConstant(d, "OBFF_LOGLVL_NONE",SWIG_From_int(static_cast< int >(0)));
  SWIG_Python_SetConstant(d, "OBFF_LOGLVL_LOW",SWIG_From_int(static_cast< int >(1)));
  SWIG_Python_SetConstant(d, "OBFF_LOGLVL_MEDIUM",SWIG_From_int(static_cast< int >(2)));
  SWIG_Python_SetConstant(d, "OBFF_LOGLVL_HIGH",SWIG_From_int(static_cast< int >(3)));
  SWIG_Python_SetConstant(d, "OBFF_ENERGY",SWIG_From_int(static_cast< int >((1 << 0))));
  SWIG_Python_SetConstant(d, "OBFF_EBOND",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "OBFF_EANGLE",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "OBFF_ESTRBND",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "OBFF_ETORSION",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "OBFF_EOOP",SWIG_From_int(static_cast< int >((1 << 5))));
  SWIG_Python_SetConstant(d, "OBFF_EVDW",SWIG_From_int(static_cast< int >((1 << 6))));
  SWIG_Python_SetConstant(d, "OBFF_EELECTROSTATIC",SWIG_From_int(static_cast< int >((1 << 7))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_IGNORE",SWIG_From_int(static_cast< int >((1 << 0))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_ATOM",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_ATOM_X",SWIG_From_int(static_cast< int >((1 << 2))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_ATOM_Y",SWIG_From_int(static_cast< int >((1 << 3))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_ATOM_Z",SWIG_From_int(static_cast< int >((1 << 4))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_DISTANCE",SWIG_From_int(static_cast< int >((1 << 5))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_ANGLE",SWIG_From_int(static_cast< int >((1 << 6))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_TORSION",SWIG_From_int(static_cast< int >((1 << 7))));
  SWIG_Python_SetConstant(d, "OBFF_CONST_CHIRAL",SWIG_From_int(static_cast< int >((1 << 8))));
  SWIG_Python_SetConstant(d, "OBFF_NUMERICAL_GRADIENT",SWIG_From_int(static_cast< int >((1 << 0))));
  SWIG_Python_SetConstant(d, "OBFF_ANALYTICAL_GRADIENT",SWIG_From_int(static_cast< int >((1 << 1))));
  SWIG_Python_SetConstant(d, "LineSearchType_Simple",SWIG_From_int(static_cast< int >(OpenBabel::LineSearchType::Simple)));
  SWIG_Python_SetConstant(d, "LineSearchType_Newton2Num",SWIG_From_int(static_cast< int >(OpenBabel::LineSearchType::Newton2Num)));
  SWIG_Python_SetConstant(d, "OBStereo_CisTrans",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::CisTrans)));
  SWIG_Python_SetConstant(d, "OBStereo_ExtendedCisTrans",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ExtendedCisTrans)));
  SWIG_Python_SetConstant(d, "OBStereo_SquarePlanar",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::SquarePlanar)));
  SWIG_Python_SetConstant(d, "OBStereo_Tetrahedral",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::Tetrahedral)));
  SWIG_Python_SetConstant(d, "OBStereo_ExtendedTetrahedral",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ExtendedTetrahedral)));
  SWIG_Python_SetConstant(d, "OBStereo_TrigonalBipyramidal",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::TrigonalBipyramidal)));
  SWIG_Python_SetConstant(d, "OBStereo_Octahedral",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::Octahedral)));
  SWIG_Python_SetConstant(d, "OBStereo_NotStereo",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::NotStereo)));
  SWIG_Python_SetConstant(d, "OBStereo_UpBond",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::UpBond)));
  SWIG_Python_SetConstant(d, "OBStereo_DownBond",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::DownBond)));
  SWIG_Python_SetConstant(d, "OBStereo_UnknownDir",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::UnknownDir)));
  SWIG_Python_SetConstant(d, "OBStereo_ShapeU",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ShapeU)));
  SWIG_Python_SetConstant(d, "OBStereo_ShapeZ",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ShapeZ)));
  SWIG_Python_SetConstant(d, "OBStereo_Shape4",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::Shape4)));
  SWIG_Python_SetConstant(d, "OBStereo_ViewFrom",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ViewFrom)));
  SWIG_Python_SetConstant(d, "OBStereo_ViewTowards",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ViewTowards)));
  SWIG_Python_SetConstant(d, "OBStereo_Clockwise",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::Clockwise)));
  SWIG_Python_SetConstant(d, "OBStereo_AntiClockwise",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::AntiClockwise)));
  SWIG_Python_SetConstant(d, "OBStereo_UnknownWinding",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::UnknownWinding)));
  SWIG_Python_SetConstant(d, "OBStereo_NoRef",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::NoRef)));
  SWIG_Python_SetConstant(d, "OBStereo_ImplicitRef",SWIG_From_int(static_cast< int >(OpenBabel::OBStereo::ImplicitRef)));
  SWIG_Python_SetConstant(d, "N_PROPERTIES",SWIG_From_int(static_cast< int >(4)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize1",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize1)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize2",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize2)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize5",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize5)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize10",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize10)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize15",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize15)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize20",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize20)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize30",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize30)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize36",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize36)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize45",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize45)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AngStepSize60",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AngStepSize60)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_NoNormalization",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::NoNormalization)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_NormalizationTowardsZeroMean",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::NormalizationTowardsZeroMean)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_NormalizationTowardsUnitStd",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::NormalizationTowardsUnitStd)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_NormalizationTowardsZeroMeanAndUnitStd",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::NormalizationTowardsZeroMeanAndUnitStd)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_NoStereoSpecificProbes",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::NoStereoSpecificProbes)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_UniqueStereoSpecificProbes",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::UniqueStereoSpecificProbes)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_MirrorStereoSpecificProbes",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::MirrorStereoSpecificProbes)));
  SWIG_Python_SetConstant(d, "OBSpectrophore_AllStereoSpecificProbes",SWIG_From_int(static_cast< int >(OpenBabel::OBSpectrophore::AllStereoSpecificProbes)));
  SWIG_Python_SetConstant(d, "OBConformerScore_HighScore",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::HighScore)));
  SWIG_Python_SetConstant(d, "OBConformerScore_LowScore",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::LowScore)));
  SWIG_Python_SetConstant(d, "OBConformerScore_Highest",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::Highest)));
  SWIG_Python_SetConstant(d, "OBConformerScore_Lowest",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::Lowest)));
  SWIG_Python_SetConstant(d, "OBConformerScore_Sum",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::Sum)));
  SWIG_Python_SetConstant(d, "OBConformerScore_Average",SWIG_From_int(static_cast< int >(OpenBabel::OBConformerScore::Average)));
  SWIG_Python_SetConstant(d, "OBAlign_Kabsch",SWIG_From_int(static_cast< int >(OpenBabel::OBAlign::Kabsch)));
  SWIG_Python_SetConstant(d, "OBAlign_QCP",SWIG_From_int(static_cast< int >(OpenBabel::OBAlign::QCP)));
  SWIG_addvarlink(globals, "ttab", Swig_var_ttab_get, Swig_var_ttab_set);
  SWIG_addvarlink(globals, "resdat", Swig_var_resdat_get, Swig_var_resdat_set);
  SWIG_addvarlink(globals, "VZero", Swig_var_VZero_get, Swig_var_VZero_set);
  SWIG_addvarlink(globals, "VX", Swig_var_VX_get, Swig_var_VX_set);
  SWIG_addvarlink(globals, "VY", Swig_var_VY_get, Swig_var_VY_set);
  SWIG_addvarlink(globals, "VZ", Swig_var_VZ_get, Swig_var_VZ_set);
  SWIG_addvarlink(globals, "chainsparser", Swig_var_chainsparser_get, Swig_var_chainsparser_set);
  SWIG_addvarlink(globals, "atomtyper", Swig_var_atomtyper_get, Swig_var_atomtyper_set);
  SWIG_addvarlink(globals, "aromtyper", Swig_var_aromtyper_get, Swig_var_aromtyper_set);
  SWIG_addvarlink(globals, "obErrorLog", Swig_var_obErrorLog_get, Swig_var_obErrorLog_set);
  SWIG_addvarlink(globals, "Dummy", Swig_var_Dummy_get, Swig_var_Dummy_set);
  SWIG_addvarlink(globals, "Hydrogen", Swig_var_Hydrogen_get, Swig_var_Hydrogen_set);
  SWIG_addvarlink(globals, "H", Swig_var_H_get, Swig_var_H_set);
  SWIG_addvarlink(globals, "Helium", Swig_var_Helium_get, Swig_var_Helium_set);
  SWIG_addvarlink(globals, "He", Swig_var_He_get, Swig_var_He_set);
  SWIG_addvarlink(globals, "Lithium", Swig_var_Lithium_get, Swig_var_Lithium_set);
  SWIG_addvarlink(globals, "Li", Swig_var_Li_get, Swig_var_Li_set);
  SWIG_addvarlink(globals, "Beryllium", Swig_var_Beryllium_get, Swig_var_Beryllium_set);
  SWIG_addvarlink(globals, "Be", Swig_var_Be_get, Swig_var_Be_set);
  SWIG_addvarlink(globals, "Boron", Swig_var_Boron_get, Swig_var_Boron_set);
  SWIG_addvarlink(globals, "B", Swig_var_B_get, Swig_var_B_set);
  SWIG_addvarlink(globals, "Carbon", Swig_var_Carbon_get, Swig_var_Carbon_set);
  SWIG_addvarlink(globals, "C", Swig_var_C_get, Swig_var_C_set);
  SWIG_addvarlink(globals, "Nitrogen", Swig_var_Nitrogen_get, Swig_var_Nitrogen_set);
  SWIG_addvarlink(globals, "N", Swig_var_N_get, Swig_var_N_set);
  SWIG_addvarlink(globals, "Oxygen", Swig_var_Oxygen_get, Swig_var_Oxygen_set);
  SWIG_addvarlink(globals, "O", Swig_var_O_get, Swig_var_O_set);
  SWIG_addvarlink(globals, "Fluorine", Swig_var_Fluorine_get, Swig_var_Fluorine_set);
  SWIG_addvarlink(globals, "F", Swig_var_F_get, Swig_var_F_set);
  SWIG_addvarlink(globals, "Neon", Swig_var_Neon_get, Swig_var_Neon_set);
  SWIG_addvarlink(globals, "Ne", Swig_var_Ne_get, Swig_var_Ne_set);
  SWIG_addvarlink(globals, "Sodium", Swig_var_Sodium_get, Swig_var_Sodium_set);
  SWIG_addvarlink(globals, "Na", Swig_var_Na_get, Swig_var_Na_set);
  SWIG_addvarlink(globals, "Magnesium", Swig_var_Magnesium_get, Swig_var_Magnesium_set);
  SWIG_addvarlink(globals, "Mg", Swig_var_Mg_get, Swig_var_Mg_set);
  SWIG_addvarlink(globals, "Aluminium", Swig_var_Aluminium_get, Swig_var_Aluminium_set);
  SWIG_addvarlink(globals, "Al", Swig_var_Al_get, Swig_var_Al_set);
  SWIG_addvarlink(globals, "Silicon", Swig_var_Silicon_get, Swig_var_Silicon_set);
  SWIG_addvarlink(globals, "Si", Swig_var_Si_get, Swig_var_Si_set);
  SWIG_addvarlink(globals, "Phosphorus", Swig_var_Phosphorus_get, Swig_var_Phosphorus_set);
  SWIG_addvarlink(globals, "P", Swig_var_P_get, Swig_var_P_set);
  SWIG_addvarlink(globals, "Sulfur", Swig_var_Sulfur_get, Swig_var_Sulfur_set);
  SWIG_addvarlink(globals, "S", Swig_var_S_get, Swig_var_S_set);
  SWIG_addvarlink(globals, "Chlorine", Swig_var_Chlorine_get, Swig_var_Chlorine_set);
  SWIG_addvarlink(globals, "Cl", Swig_var_Cl_get, Swig_var_Cl_set);
  SWIG_addvarlink(globals, "Argon", Swig_var_Argon_get, Swig_var_Argon_set);
  SWIG_addvarlink(globals, "Ar", Swig_var_Ar_get, Swig_var_Ar_set);
  SWIG_addvarlink(globals, "Potassium", Swig_var_Potassium_get, Swig_var_Potassium_set);
  SWIG_addvarlink(globals, "K", Swig_var_K_get, Swig_var_K_set);
  SWIG_addvarlink(globals, "Calcium", Swig_var_Calcium_get, Swig_var_Calcium_set);
  SWIG_addvarlink(globals, "Ca", Swig_var_Ca_get, Swig_var_Ca_set);
  SWIG_addvarlink(globals, "Scandium", Swig_var_Scandium_get, Swig_var_Scandium_set);
  SWIG_addvarlink(globals, "Sc", Swig_var_Sc_get, Swig_var_Sc_set);
  SWIG_addvarlink(globals, "Titanium", Swig_var_Titanium_get, Swig_var_Titanium_set);
  SWIG_addvarlink(globals, "Ti", Swig_var_Ti_get, Swig_var_Ti_set);
  SWIG_addvarlink(globals, "Vanadium", Swig_var_Vanadium_get, Swig_var_Vanadium_set);
  SWIG_addvarlink(globals, "V", Swig_var_V_get, Swig_var_V_set);
  SWIG_addvarlink(globals, "Chromium", Swig_var_Chromium_get, Swig_var_Chromium_set);
  SWIG_addvarlink(globals, "Cr", Swig_var_Cr_get, Swig_var_Cr_set);
  SWIG_addvarlink(globals, "Manganese", Swig_var_Manganese_get, Swig_var_Manganese_set);
  SWIG_addvarlink(globals, "Mn", Swig_var_Mn_get, Swig_var_Mn_set);
  SWIG_addvarlink(globals, "Iron", Swig_var_Iron_get, Swig_var_Iron_set);
  SWIG_addvarlink(globals, "Fe", Swig_var_Fe_get, Swig_var_Fe_set);
  SWIG_addvarlink(globals, "Cobalt", Swig_var_Cobalt_get, Swig_var_Cobalt_set);
  SWIG_addvarlink(globals, "Co", Swig_var_Co_get, Swig_var_Co_set);
  SWIG_addvarlink(globals, "Nickel", Swig_var_Nickel_get, Swig_var_Nickel_set);
  SWIG_addvarlink(globals, "Ni", Swig_var_Ni_get, Swig_var_Ni_set);
  SWIG_addvarlink(globals, "Copper", Swig_var_Copper_get, Swig_var_Copper_set);
  SWIG_addvarlink(globals, "Cu", Swig_var_Cu_get, Swig_var_Cu_set);
  SWIG_addvarlink(globals, "Zinc", Swig_var_Zinc_get, Swig_var_Zinc_set);
  SWIG_addvarlink(globals, "Zn", Swig_var_Zn_get, Swig_var_Zn_set);
  SWIG_addvarlink(globals, "Gallium", Swig_var_Gallium_get, Swig_var_Gallium_set);
  SWIG_addvarlink(globals, "Ga", Swig_var_Ga_get, Swig_var_Ga_set);
  SWIG_addvarlink(globals, "Germanium", Swig_var_Germanium_get, Swig_var_Germanium_set);
  SWIG_addvarlink(globals, "Ge", Swig_var_Ge_get, Swig_var_Ge_set);
  SWIG_addvarlink(globals, "Arsenic", Swig_var_Arsenic_get, Swig_var_Arsenic_set);
  SWIG_addvarlink(globals, "As", Swig_var_As_get, Swig_var_As_set);
  SWIG_addvarlink(globals, "Selenium", Swig_var_Selenium_get, Swig_var_Selenium_set);
  SWIG_addvarlink(globals, "Se", Swig_var_Se_get, Swig_var_Se_set);
  SWIG_addvarlink(globals, "Bromine", Swig_var_Bromine_get, Swig_var_Bromine_set);
  SWIG_addvarlink(globals, "Br", Swig_var_Br_get, Swig_var_Br_set);
  SWIG_addvarlink(globals, "Krypton", Swig_var_Krypton_get, Swig_var_Krypton_set);
  SWIG_addvarlink(globals, "Kr", Swig_var_Kr_get, Swig_var_Kr_set);
  SWIG_addvarlink(globals, "Rubidium", Swig_var_Rubidium_get, Swig_var_Rubidium_set);
  SWIG_addvarlink(globals, "Rb", Swig_var_Rb_get, Swig_var_Rb_set);
  SWIG_addvarlink(globals, "Strontium", Swig_var_Strontium_get, Swig_var_Strontium_set);
  SWIG_addvarlink(globals, "Sr", Swig_var_Sr_get, Swig_var_Sr_set);
  SWIG_addvarlink(globals, "Yttrium", Swig_var_Yttrium_get, Swig_var_Yttrium_set);
  SWIG_addvarlink(globals, "Y", Swig_var_Y_get, Swig_var_Y_set);
  SWIG_addvarlink(globals, "Zirconium", Swig_var_Zirconium_get, Swig_var_Zirconium_set);
  SWIG_addvarlink(globals, "Zr", Swig_var_Zr_get, Swig_var_Zr_set);
  SWIG_addvarlink(globals, "Niobium", Swig_var_Niobium_get, Swig_var_Niobium_set);
  SWIG_addvarlink(globals, "Nb", Swig_var_Nb_get, Swig_var_Nb_set);
  SWIG_addvarlink(globals, "Molybdenum", Swig_var_Molybdenum_get, Swig_var_Molybdenum_set);
  SWIG_addvarlink(globals, "Mo", Swig_var_Mo_get, Swig_var_Mo_set);
  SWIG_addvarlink(globals, "Technetium", Swig_var_Technetium_get, Swig_var_Technetium_set);
  SWIG_addvarlink(globals, "Tc", Swig_var_Tc_get, Swig_var_Tc_set);
  SWIG_addvarlink(globals, "Ruthenium", Swig_var_Ruthenium_get, Swig_var_Ruthenium_set);
  SWIG_addvarlink(globals, "Ru", Swig_var_Ru_get, Swig_var_Ru_set);
  SWIG_addvarlink(globals, "Rhodium", Swig_var_Rhodium_get, Swig_var_Rhodium_set);
  SWIG_addvarlink(globals, "Rh", Swig_var_Rh_get, Swig_var_Rh_set);
  SWIG_addvarlink(globals, "Palladium", Swig_var_Palladium_get, Swig_var_Palladium_set);
  SWIG_addvarlink(globals, "Pd", Swig_var_Pd_get, Swig_var_Pd_set);
  SWIG_addvarlink(globals, "Silver", Swig_var_Silver_get, Swig_var_Silver_set);
  SWIG_addvarlink(globals, "Ag", Swig_var_Ag_get, Swig_var_Ag_set);
  SWIG_addvarlink(globals, "Cadmium", Swig_var_Cadmium_get, Swig_var_Cadmium_set);
  SWIG_addvarlink(globals, "Cd", Swig_var_Cd_get, Swig_var_Cd_set);
  SWIG_addvarlink(globals, "Indium", Swig_var_Indium_get, Swig_var_Indium_set);
  SWIG_addvarlink(globals, "In", Swig_var_In_get, Swig_var_In_set);
  SWIG_addvarlink(globals, "Tin", Swig_var_Tin_get, Swig_var_Tin_set);
  SWIG_addvarlink(globals, "Sn", Swig_var_Sn_get, Swig_var_Sn_set);
  SWIG_addvarlink(globals, "Antimony", Swig_var_Antimony_get, Swig_var_Antimony_set);
  SWIG_addvarlink(globals, "Sb", Swig_var_Sb_get, Swig_var_Sb_set);
  SWIG_addvarlink(globals, "Tellurium", Swig_var_Tellurium_get, Swig_var_Tellurium_set);
  SWIG_addvarlink(globals, "Te", Swig_var_Te_get, Swig_var_Te_set);
  SWIG_addvarlink(globals, "Iodine", Swig_var_Iodine_get, Swig_var_Iodine_set);
  SWIG_addvarlink(globals, "I", Swig_var_I_get, Swig_var_I_set);
  SWIG_addvarlink(globals, "Xenon", Swig_var_Xenon_get, Swig_var_Xenon_set);
  SWIG_addvarlink(globals, "Xe", Swig_var_Xe_get, Swig_var_Xe_set);
  SWIG_addvarlink(globals, "Caesium", Swig_var_Caesium_get, Swig_var_Caesium_set);
  SWIG_addvarlink(globals, "Cs", Swig_var_Cs_get, Swig_var_Cs_set);
  SWIG_addvarlink(globals, "Barium", Swig_var_Barium_get, Swig_var_Barium_set);
  SWIG_addvarlink(globals, "Ba", Swig_var_Ba_get, Swig_var_Ba_set);
  SWIG_addvarlink(globals, "Lanthanum", Swig_var_Lanthanum_get, Swig_var_Lanthanum_set);
  SWIG_addvarlink(globals, "La", Swig_var_La_get, Swig_var_La_set);
  SWIG_addvarlink(globals, "Cerium", Swig_var_Cerium_get, Swig_var_Cerium_set);
  SWIG_addvarlink(globals, "Ce", Swig_var_Ce_get, Swig_var_Ce_set);
  SWIG_addvarlink(globals, "Praseodymium", Swig_var_Praseodymium_get, Swig_var_Praseodymium_set);
  SWIG_addvarlink(globals, "Pr", Swig_var_Pr_get, Swig_var_Pr_set);
  SWIG_addvarlink(globals, "Neodymium", Swig_var_Neodymium_get, Swig_var_Neodymium_set);
  SWIG_addvarlink(globals, "Nd", Swig_var_Nd_get, Swig_var_Nd_set);
  SWIG_addvarlink(globals, "Promethium", Swig_var_Promethium_get, Swig_var_Promethium_set);
  SWIG_addvarlink(globals, "Pm", Swig_var_Pm_get, Swig_var_Pm_set);
  SWIG_addvarlink(globals, "Samarium", Swig_var_Samarium_get, Swig_var_Samarium_set);
  SWIG_addvarlink(globals, "Sm", Swig_var_Sm_get, Swig_var_Sm_set);
  SWIG_addvarlink(globals, "Europium", Swig_var_Europium_get, Swig_var_Europium_set);
  SWIG_addvarlink(globals, "Eu", Swig_var_Eu_get, Swig_var_Eu_set);
  SWIG_addvarlink(globals, "Gadolinium", Swig_var_Gadolinium_get, Swig_var_Gadolinium_set);
  SWIG_addvarlink(globals, "Gd", Swig_var_Gd_get, Swig_var_Gd_set);
  SWIG_addvarlink(globals, "Terbium", Swig_var_Terbium_get, Swig_var_Terbium_set);
  SWIG_addvarlink(globals, "Tb", Swig_var_Tb_get, Swig_var_Tb_set);
  SWIG_addvarlink(globals, "Dysprosium", Swig_var_Dysprosium_get, Swig_var_Dysprosium_set);
  SWIG_addvarlink(globals, "Dy", Swig_var_Dy_get, Swig_var_Dy_set);
  SWIG_addvarlink(globals, "Holmium", Swig_var_Holmium_get, Swig_var_Holmium_set);
  SWIG_addvarlink(globals, "Ho", Swig_var_Ho_get, Swig_var_Ho_set);
  SWIG_addvarlink(globals, "Erbium", Swig_var_Erbium_get, Swig_var_Erbium_set);
  SWIG_addvarlink(globals, "Er", Swig_var_Er_get, Swig_var_Er_set);
  SWIG_addvarlink(globals, "Thulium", Swig_var_Thulium_get, Swig_var_Thulium_set);
  SWIG_addvarlink(globals, "Tm", Swig_var_Tm_get, Swig_var_Tm_set);
  SWIG_addvarlink(globals, "Ytterbium", Swig_var_Ytterbium_get, Swig_var_Ytterbium_set);
  SWIG_addvarlink(globals, "Yb", Swig_var_Yb_get, Swig_var_Yb_set);
  SWIG_addvarlink(globals, "Lutetium", Swig_var_Lutetium_get, Swig_var_Lutetium_set);
  SWIG_addvarlink(globals, "Lu", Swig_var_Lu_get, Swig_var_Lu_set);
  SWIG_addvarlink(globals, "Hafnium", Swig_var_Hafnium_get, Swig_var_Hafnium_set);
  SWIG_addvarlink(globals, "Hf", Swig_var_Hf_get, Swig_var_Hf_set);
  SWIG_addvarlink(globals, "Tantalum", Swig_var_Tantalum_get, Swig_var_Tantalum_set);
  SWIG_addvarlink(globals, "Ta", Swig_var_Ta_get, Swig_var_Ta_set);
  SWIG_addvarlink(globals, "Tungsten", Swig_var_Tungsten_get, Swig_var_Tungsten_set);
  SWIG_addvarlink(globals, "W", Swig_var_W_get, Swig_var_W_set);
  SWIG_addvarlink(globals, "Rhenium", Swig_var_Rhenium_get, Swig_var_Rhenium_set);
  SWIG_addvarlink(globals, "Re", Swig_var_Re_get, Swig_var_Re_set);
  SWIG_addvarlink(globals, "Osmium", Swig_var_Osmium_get, Swig_var_Osmium_set);
  SWIG_addvarlink(globals, "Os", Swig_var_Os_get, Swig_var_Os_set);
  SWIG_addvarlink(globals, "Iridium", Swig_var_Iridium_get, Swig_var_Iridium_set);
  SWIG_addvarlink(globals, "Ir", Swig_var_Ir_get, Swig_var_Ir_set);
  SWIG_addvarlink(globals, "Platinum", Swig_var_Platinum_get, Swig_var_Platinum_set);
  SWIG_addvarlink(globals, "Pt", Swig_var_Pt_get, Swig_var_Pt_set);
  SWIG_addvarlink(globals, "Gold", Swig_var_Gold_get, Swig_var_Gold_set);
  SWIG_addvarlink(globals, "Au", Swig_var_Au_get, Swig_var_Au_set);
  SWIG_addvarlink(globals, "Mercury", Swig_var_Mercury_get, Swig_var_Mercury_set);
  SWIG_addvarlink(globals, "Hg", Swig_var_Hg_get, Swig_var_Hg_set);
  SWIG_addvarlink(globals, "Thallium", Swig_var_Thallium_get, Swig_var_Thallium_set);
  SWIG_addvarlink(globals, "Tl", Swig_var_Tl_get, Swig_var_Tl_set);
  SWIG_addvarlink(globals, "Lead", Swig_var_Lead_get, Swig_var_Lead_set);
  SWIG_addvarlink(globals, "Pb", Swig_var_Pb_get, Swig_var_Pb_set);
  SWIG_addvarlink(globals, "Bismuth", Swig_var_Bismuth_get, Swig_var_Bismuth_set);
  SWIG_addvarlink(globals, "Bi", Swig_var_Bi_get, Swig_var_Bi_set);
  SWIG_addvarlink(globals, "Polonium", Swig_var_Polonium_get, Swig_var_Polonium_set);
  SWIG_addvarlink(globals, "Po", Swig_var_Po_get, Swig_var_Po_set);
  SWIG_addvarlink(globals, "Astatine", Swig_var_Astatine_get, Swig_var_Astatine_set);
  SWIG_addvarlink(globals, "At", Swig_var_At_get, Swig_var_At_set);
  SWIG_addvarlink(globals, "Radon", Swig_var_Radon_get, Swig_var_Radon_set);
  SWIG_addvarlink(globals, "Rn", Swig_var_Rn_get, Swig_var_Rn_set);
  SWIG_addvarlink(globals, "Francium", Swig_var_Francium_get, Swig_var_Francium_set);
  SWIG_addvarlink(globals, "Fr", Swig_var_Fr_get, Swig_var_Fr_set);
  SWIG_addvarlink(globals, "Radium", Swig_var_Radium_get, Swig_var_Radium_set);
  SWIG_addvarlink(globals, "Ra", Swig_var_Ra_get, Swig_var_Ra_set);
  SWIG_addvarlink(globals, "Actinium", Swig_var_Actinium_get, Swig_var_Actinium_set);
  SWIG_addvarlink(globals, "Ac", Swig_var_Ac_get, Swig_var_Ac_set);
  SWIG_addvarlink(globals, "Thorium", Swig_var_Thorium_get, Swig_var_Thorium_set);
  SWIG_addvarlink(globals, "Th", Swig_var_Th_get, Swig_var_Th_set);
  SWIG_addvarlink(globals, "Protactinium", Swig_var_Protactinium_get, Swig_var_Protactinium_set);
  SWIG_addvarlink(globals, "Pa", Swig_var_Pa_get, Swig_var_Pa_set);
  SWIG_addvarlink(globals, "Uranium", Swig_var_Uranium_get, Swig_var_Uranium_set);
  SWIG_addvarlink(globals, "U", Swig_var_U_get, Swig_var_U_set);
  SWIG_addvarlink(globals, "Neptunium", Swig_var_Neptunium_get, Swig_var_Neptunium_set);
  SWIG_addvarlink(globals, "Np", Swig_var_Np_get, Swig_var_Np_set);
  SWIG_addvarlink(globals, "Plutonium", Swig_var_Plutonium_get, Swig_var_Plutonium_set);
  SWIG_addvarlink(globals, "Pu", Swig_var_Pu_get, Swig_var_Pu_set);
  SWIG_addvarlink(globals, "Americium", Swig_var_Americium_get, Swig_var_Americium_set);
  SWIG_addvarlink(globals, "Am", Swig_var_Am_get, Swig_var_Am_set);
  SWIG_addvarlink(globals, "Curium", Swig_var_Curium_get, Swig_var_Curium_set);
  SWIG_addvarlink(globals, "Cm", Swig_var_Cm_get, Swig_var_Cm_set);
  SWIG_addvarlink(globals, "Berkelium", Swig_var_Berkelium_get, Swig_var_Berkelium_set);
  SWIG_addvarlink(globals, "Bk", Swig_var_Bk_get, Swig_var_Bk_set);
  SWIG_addvarlink(globals, "Californium", Swig_var_Californium_get, Swig_var_Californium_set);
  SWIG_addvarlink(globals, "Cf", Swig_var_Cf_get, Swig_var_Cf_set);
  SWIG_addvarlink(globals, "Einsteinium", Swig_var_Einsteinium_get, Swig_var_Einsteinium_set);
  SWIG_addvarlink(globals, "Es", Swig_var_Es_get, Swig_var_Es_set);
  SWIG_addvarlink(globals, "Fermium", Swig_var_Fermium_get, Swig_var_Fermium_set);
  SWIG_addvarlink(globals, "Fm", Swig_var_Fm_get, Swig_var_Fm_set);
  SWIG_addvarlink(globals, "Mendelevium", Swig_var_Mendelevium_get, Swig_var_Mendelevium_set);
  SWIG_addvarlink(globals, "Md", Swig_var_Md_get, Swig_var_Md_set);
  SWIG_addvarlink(globals, "Nobelium", Swig_var_Nobelium_get, Swig_var_Nobelium_set);
  SWIG_addvarlink(globals, "No", Swig_var_No_get, Swig_var_No_set);
  SWIG_addvarlink(globals, "Lawrencium", Swig_var_Lawrencium_get, Swig_var_Lawrencium_set);
  SWIG_addvarlink(globals, "Lr", Swig_var_Lr_get, Swig_var_Lr_set);
  SWIG_addvarlink(globals, "Rutherfordium", Swig_var_Rutherfordium_get, Swig_var_Rutherfordium_set);
  SWIG_addvarlink(globals, "Rf", Swig_var_Rf_get, Swig_var_Rf_set);
  SWIG_addvarlink(globals, "Dubnium", Swig_var_Dubnium_get, Swig_var_Dubnium_set);
  SWIG_addvarlink(globals, "Db", Swig_var_Db_get, Swig_var_Db_set);
  SWIG_addvarlink(globals, "Seaborgium", Swig_var_Seaborgium_get, Swig_var_Seaborgium_set);
  SWIG_addvarlink(globals, "Sg", Swig_var_Sg_get, Swig_var_Sg_set);
  SWIG_addvarlink(globals, "Bohrium", Swig_var_Bohrium_get, Swig_var_Bohrium_set);
  SWIG_addvarlink(globals, "Bh", Swig_var_Bh_get, Swig_var_Bh_set);
  SWIG_addvarlink(globals, "Hassium", Swig_var_Hassium_get, Swig_var_Hassium_set);
  SWIG_addvarlink(globals, "Hs", Swig_var_Hs_get, Swig_var_Hs_set);
  SWIG_addvarlink(globals, "Meitnerium", Swig_var_Meitnerium_get, Swig_var_Meitnerium_set);
  SWIG_addvarlink(globals, "Mt", Swig_var_Mt_get, Swig_var_Mt_set);
  SWIG_addvarlink(globals, "Darmstadtium", Swig_var_Darmstadtium_get, Swig_var_Darmstadtium_set);
  SWIG_addvarlink(globals, "Ds", Swig_var_Ds_get, Swig_var_Ds_set);
  SWIG_addvarlink(globals, "Roentgenium", Swig_var_Roentgenium_get, Swig_var_Roentgenium_set);
  SWIG_addvarlink(globals, "Rg", Swig_var_Rg_get, Swig_var_Rg_set);
  SWIG_addvarlink(globals, "Copernicium", Swig_var_Copernicium_get, Swig_var_Copernicium_set);
  SWIG_addvarlink(globals, "Cn", Swig_var_Cn_get, Swig_var_Cn_set);
  SWIG_addvarlink(globals, "Nihonium", Swig_var_Nihonium_get, Swig_var_Nihonium_set);
  SWIG_addvarlink(globals, "Nh", Swig_var_Nh_get, Swig_var_Nh_set);
  SWIG_addvarlink(globals, "Flerovium", Swig_var_Flerovium_get, Swig_var_Flerovium_set);
  SWIG_addvarlink(globals, "Fl", Swig_var_Fl_get, Swig_var_Fl_set);
  SWIG_addvarlink(globals, "Moscovium", Swig_var_Moscovium_get, Swig_var_Moscovium_set);
  SWIG_addvarlink(globals, "Mc", Swig_var_Mc_get, Swig_var_Mc_set);
  SWIG_addvarlink(globals, "Livermorium", Swig_var_Livermorium_get, Swig_var_Livermorium_set);
  SWIG_addvarlink(globals, "Lv", Swig_var_Lv_get, Swig_var_Lv_set);
  SWIG_addvarlink(globals, "Tennessine", Swig_var_Tennessine_get, Swig_var_Tennessine_set);
  SWIG_addvarlink(globals, "Ts", Swig_var_Ts_get, Swig_var_Ts_set);
  SWIG_addvarlink(globals, "Oganesson", Swig_var_Oganesson_get, Swig_var_Oganesson_set);
  SWIG_addvarlink(globals, "Og", Swig_var_Og_get, Swig_var_Og_set);
  SWIG_addvarlink(globals, "NoId", Swig_var_NoId_get, Swig_var_NoId_set);
  SWIG_addvarlink(globals, "AliasDataType", Swig_var_AliasDataType_get, Swig_var_AliasDataType_set);
  SWIG_addvarlink(globals, "KCAL_TO_KJ", Swig_var_KCAL_TO_KJ_get, Swig_var_KCAL_TO_KJ_set);
  SWIG_addvarlink(globals, "GAS_CONSTANT", Swig_var_GAS_CONSTANT_get, Swig_var_GAS_CONSTANT_set);
  SWIG_addvarlink(globals, "OBGraphSym_NoSymmetryClass", Swig_var_OBGraphSym_NoSymmetryClass_get, Swig_var_OBGraphSym_NoSymmetryClass_set);
  return 0;
}

